(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Equational_Reasoning 
  imports Trs
begin

definition estep :: "('f,'v)trs \<Rightarrow> ('f,'v)trs"
  where "estep E \<equiv> rstep (E \<union> E^-1)"

lemma estep_sym_closure_conv: "estep E = (rstep E)\<^sup>\<leftrightarrow>"
  unfolding estep_def rstep_union reverseTrs ..

(* TODO: perhaps rename to something else, since usually the semantic
   property is called theory, but since Birkhoff is true, we equivalence here *)
definition "theory" :: "('f,'v)trs \<Rightarrow> ('f,'v)trs"
  where "theory E \<equiv> (estep E)^*"

lemma join_conversion: "join (rstep R) \<subseteq> (estep R)^*"
proof
  fix a b
  assume "(a,b) \<in> join (rstep R)"
  hence steps: "(a,b) \<in> (rstep R)^* O ((rstep R)^*)^-1" by auto
  show "(a,b) \<in> (estep R)^*"
    by (rule set_mp[OF _ steps], unfold rtrancl_converse[symmetric] estep_sym_closure_conv, regexp)
qed

lemma theory_is_conversion: "theory E = conversion (rstep E)"
  unfolding conversion_def theory_def estep_sym_closure_conv ..

definition equivalent :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> bool"
  where "equivalent E E' \<equiv> theory E = theory E'"

definition subsumes :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> bool"
  where "subsumes E E' \<equiv> theory E \<subseteq> theory E'"

lemma equivalent_subsumes: "equivalent E E' = (subsumes E E' \<and> subsumes E' E)"
  unfolding subsumes_def equivalent_def by auto

lemma sym_estep: "sym (estep E)" 
  unfolding estep_sym_closure_conv sym_def by auto

lemma subst_closed_estep: "subst_closed (estep E)" unfolding estep_def rstep_union by  (rule subst_closed_union, insert subst_closed_rstep, auto)

lemma ctxt_closed_estep: "ctxt_closed (estep E)" unfolding estep_def rstep_union by  (rule ctxt_closed_union, insert ctxt_closed_rstep, auto)


lemma subsumes_via_rule_conversion: "subsumes E E' = (\<forall> (s,t) \<in> E. (s,t) \<in> (estep E')^*)"
proof
  assume subs: "subsumes E E'"
  show "\<forall> (s,t) \<in> E. (s,t) \<in> (estep E')^*"
  proof(clarify)
    fix s t
    assume "(s,t) \<in> E"
    hence "(s,t) \<in> theory E" unfolding theory_def estep_def by auto
    with subs have "(s,t) \<in> theory E'" unfolding subsumes_def by auto
    thus "(s,t) \<in> (estep E')^*" unfolding theory_def .
  qed
next
  assume "\<forall> (s,t) \<in> E. (s,t) \<in> (estep E')^*"
  hence steps: "\<And> s t. (s,t) \<in> E \<Longrightarrow> (s,t) \<in> (estep E')^*" by auto
  show "subsumes E E'"
    unfolding subsumes_def theory_def
  proof
    fix s t 
    assume "(s,t) \<in> (estep E)^*"
    thus "(s,t) \<in> (estep E')^*"
    proof(induct)
      case base
      show ?case unfolding estep_def by auto
    next
      case (step t u)
      from step(2)[unfolded estep_def] obtain l r C \<sigma> where t: "t = C\<langle>l\<cdot>\<sigma>\<rangle>"
        and u: "u = C\<langle>r \<cdot> \<sigma>\<rangle>" and lr: "(l,r) \<in> E \<union> E^-1" by auto
      from lr have steps: "(l,r) \<in> (estep E')^*"
      proof
        assume "(l,r) \<in> E"
        from steps[OF this] show ?thesis .
      next
        assume "(l,r) \<in> E^-1" 
        hence "(r,l) \<in> E" by simp
        from steps[OF this] have "(r,l) \<in> (estep E')^*" .
        with sym_rtrancl[OF sym_estep, of E']
        show ?thesis unfolding sym_def by auto
      qed
      with subst_closed_rtrancl[OF subst_closed_estep[of E'], unfolded subst_closed_def] have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> (estep E')^*" by auto
      from ctxt_closed_imp_ctxt[OF rtrancl_one_step_ctxt_closed[unfolded ctxt_closed_one_step_ctxt_closed_conv[symmetric], OF ctxt_closed_estep] this]
      have "(t,u) \<in> (estep E')^*" unfolding t u .
      with step(3) show ?case by simp
    qed
  qed
qed

definition completed_rewrite_system :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> bool"
  where "completed_rewrite_system E R \<equiv> equivalent E R \<and> CR (rstep R) \<and> SN (rstep R)"

lemma completion_via_WCR_SN_simulation: assumes one: "subsumes E R" 
  and two: "subsumes R E"
  and three: "WCR (rstep R)"
  and four: "SN (rstep R)"
  shows "completed_rewrite_system E R"
  unfolding completed_rewrite_system_def equivalent_subsumes 
  using Newman[OF four three] four one two by auto

end
  
          