(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Flat_Context_Closure_Impl
imports
  Flat_Context_Closure
  QDP_Framework_Impl
begin

definition fresh_vars_list :: "nat \<Rightarrow> string list \<Rightarrow> string list" where
  "fresh_vars_list n tabus \<equiv> fresh_strings_list ''x'' 0 tabus n"

lemma fresh_vars_list_sound[simp]:
  "fresh_vars_list n tabus = fresh_vars n (set tabus)"
  unfolding fresh_vars_list_def fresh_vars_def by auto

lemma flat_ctxt_i_fresh_vars_list_conv[simp]:
  "flat_ctxt_i (map Var (fresh_vars_list n tabus)) =
    flat_ctxt_i (map Var (fresh_vars n (set tabus)))"
  by (intro ext) (simp add: flat_ctxt_i_def)

lemma fresh_vars_list_fresh: "set (fresh_vars_list n vs) \<inter> set vs = {}"
  using fresh_name_gen_for_strings_list[of "''x''" 0]
  unfolding fresh_vars_list_def fresh_name_gen_list_def by auto

lemma fresh_vars_list_length: "length (fresh_vars_list n vs) = n"
  using fresh_name_gen_for_strings_list[of "''x''" 0]
  unfolding fresh_name_gen_list_def fresh_vars_list_def by simp

lemma fresh_vars_list_distinct: "distinct (fresh_vars_list n vs)"
  using Var_inj[of "fresh_strings_list ''x'' 0 vs n"]
  fresh_name_gen_for_strings_list[of "''x''" 0]
  distinct_map[of "Var" "fresh_strings_list ''x'' 0 vs n",symmetric]
  unfolding fresh_name_gen_list_def fresh_vars_list_def by auto

definition
  flat_ctxts_list :: "string list \<Rightarrow> ('f \<times> nat) list \<Rightarrow> ('f, string) ctxt list"
where
  "flat_ctxts_list tabus fs \<equiv> concat (map (\<lambda>(f, n). (
    let xs = map Var (fresh_vars_list n tabus) in
    map (flat_ctxt_i xs f) [(0::nat)..<n]
  )) fs)"

lemma flat_ctxts_list_sound[simp]:
  assumes "distinct fs"
  shows "set (flat_ctxts_list tabus fs) = flat_ctxts (set tabus) (set fs)"
  using assms by (induct fs) (auto simp: flat_ctxts_list_def flat_ctxts_def)

definition
  flat_ctxt_closure_list :: "('f \<times> nat) list \<Rightarrow> ('f, string) rules \<Rightarrow> ('f, string) rules"
where
  "flat_ctxt_closure_list fs trs \<equiv> (
    let fcs       = flat_ctxts_list (vars_trs_list trs) fs in
    let (ras, rs) = partition (\<lambda>(l, r). root l \<noteq> root r) trs in
    concat (map (\<lambda>(l, r). map (\<lambda>C. (C\<langle>l\<rangle>, C\<langle>r\<rangle>)) fcs) ras) @ rs)"

lemma flat_ctxt_closure_list_sound_1[simp]: 
  assumes "distinct fs"
  shows "set (
    concat (map (\<lambda>(l, r).
      map (\<lambda>C. (C\<langle>l\<rangle>, C\<langle>r\<rangle>)) (flat_ctxts_list (vars_trs_list trs) fs)
    ) (fst (partition (\<lambda>(l, r). root l \<noteq> root r) trs)))
  ) = (\<Union>(l, r)\<in>(set trs)a.
    {(C\<langle>l\<rangle>, C\<langle>r\<rangle>) | C. C \<in> flat_ctxts (vars_trs (set trs)) (set fs)}
  )"
proof -
  let ?C = "\<lambda>(l,r)C. (C\<langle>l\<rangle>, C\<langle>r\<rangle>)"
  let ?ra = "fst(partition (\<lambda>(l, r). root l \<noteq> root r) trs)"
  let ?fcs = "flat_ctxts_list (vars_trs_list trs) fs"
  have "set (concat (map (\<lambda>(l, r). map (?C(l, r)) ?fcs) ?ra)) =
    (\<Union>x\<in>set(map (\<lambda>(l, r). map (?C(l, r)) ?fcs) ?ra). set x)" by simp
  also have "\<dots> = (\<Union>x\<in>(\<lambda>(l, r). map (?C(l, r)) ?fcs) ` set ?ra. set x)" by simp
  also have "\<dots> = (\<Union>(l, r)\<in>set ?ra. set((\<lambda>(l, r). map (?C(l, r)) ?fcs) (l, r)))" by best
  also have "\<dots> = (\<Union>(l, r)\<in>set ?ra. set(map (\<lambda>C. (C\<langle>l\<rangle>, C\<langle>r\<rangle>)) ?fcs))" by simp
  also have "\<dots> = (\<Union>(l, r)\<in>set ?ra. {(C\<langle>l\<rangle>, C\<langle>r\<rangle>) | C. C \<in> set (?fcs)})" by auto
  also have "\<dots> = (\<Union>(l, r)\<in>(set trs)a. {(C\<langle>l\<rangle>, C\<langle>r\<rangle>) | C. C \<in> set (?fcs)})"
    using root_altering_list[of "trs"] by simp
  finally show ?thesis
  using flat_ctxts_list_sound[of "fs" "vars_trs_list trs",OF `distinct fs`] by auto
qed

lemma flat_ctxt_closure_list_sound[simp]:
  assumes "distinct fs"
  shows "set (flat_ctxt_closure_list fs trs) = FC (set fs) (set trs)"
proof -
  let ?fcs = "flat_ctxts_list (vars_trs_list trs) fs"
  let ?ras = "fst(partition (\<lambda>(l, r). root l \<noteq> root r) trs)"
  let ?rs  = "snd(partition (\<lambda>(l, r). root l \<noteq> root r) trs)"
  have a: "set (
    let fcs       = ?fcs in
    let (ras, rs) = partition (\<lambda>(l, r). root l \<noteq> root r) trs in
    concat (map (\<lambda>(l, r). map (\<lambda>C. (C\<langle>l\<rangle>, C\<langle>r\<rangle>)) fcs) ras) @ rs) =
    set (concat (map (\<lambda>(l, r). map (\<lambda>C. (C\<langle>l\<rangle>, C\<langle>r\<rangle>)) ?fcs) ?ras)) \<union> set ?rs"
    by simp
  show ?thesis
    unfolding flat_ctxt_closure_list_def flat_ctxt_closure_def
    unfolding a
    unfolding flat_ctxt_closure_list_sound_1[OF `distinct fs`]
    unfolding root_altering_list_aux by simp
qed

(* Closure under Flat Contexts *)
definition
  check_rule_preserving ::
    "('f::show, 'v::show) ctxt list
     \<Rightarrow> ('f, 'v) rule list
     \<Rightarrow> ('f, 'v) rule
     \<Rightarrow> shows check"
where
  "check_rule_preserving fcs rs' rule \<equiv> (
     check (rule \<in> set rs' \<or> (\<forall>C\<in>set fcs. (C\<langle>fst rule\<rangle>, C\<langle>snd rule\<rangle>) \<in> set rs'))
       (''the rule '' +#+ shows_rule rule +@+
         '' is neither contained in the resulting set of rules nor closed under all flat contexts''
         +#+ shows_nl)
   )"

definition
  check_rule_reflecting ::
    "('f::show, 'v::show) ctxt list
     \<Rightarrow> ('f, 'v) rule list
     \<Rightarrow> ('f, 'v) rule
     \<Rightarrow> shows check"
where
  "check_rule_reflecting fcs rs rule \<equiv> (
     check (\<exists>(l, r)\<in>set rs. \<exists>C\<in>set (\<box> # fcs). fst rule = C\<langle>l\<rangle> \<and> snd rule = C\<langle>r\<rangle>)
       (''the rule '' +#+ shows_rule rule +@+
         '' is neither contained in the original set of rules nor obtained by applying a flat context''
         +#+ shows_nl)
   )"

lemma check_rule_preserving_sound:
  assumes "isOK (check_rule_preserving fcs rs' (l, r))"
  shows "(l, r) \<in> set rs' \<or> (\<forall>C\<in>set fcs. (C\<langle>l\<rangle>, C\<langle>r\<rangle>) \<in> set rs')"
proof -
  from assms have "(l, r) \<in> set rs' \<or> (\<forall>C\<in>set fcs. (C\<langle>l\<rangle>, C\<langle>r\<rangle>) \<in> set rs')"
    unfolding check_rule_preserving_def by simp
  thus ?thesis by simp
qed

lemma check_rule_reflecting_sound:
  assumes "isOK (check_rule_reflecting fcs rs (l, r))"
  shows "\<exists>(u, v) \<in> set rs.\<exists>C\<in>set(\<box>#fcs). l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>"
proof -
  from assms have "\<exists>(u, v)\<in>set rs. \<exists>C\<in>set (\<box> # fcs). l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>"
    unfolding check_rule_reflecting_def by simp
  thus ?thesis .
qed

lemma check_rules_preserving_sound:
  assumes "isOK (check_allm (check_rule_preserving fcs lR) R)"
  shows "\<forall>(l, r)\<in>set R. (l, r) \<in> set lR \<or> (\<forall>C\<in>set fcs. (C\<langle>l\<rangle>, C\<langle>r\<rangle>) \<in> set lR)"
proof -
  from assms have "\<forall>rule\<in>set R. isOK (check_rule_preserving fcs lR rule)" by simp
  thus ?thesis using check_rule_preserving_sound by best
qed

lemma check_rules_reflecting_sound:
  assumes "isOK (check_allm (check_rule_reflecting fcs R) lR)"
  shows "\<forall>(l, r)\<in>set lR. \<exists>(u, v)\<in>set R. \<exists>C\<in>set (\<box>#fcs). l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>"
proof -
  from assms have "\<forall>rule\<in>set lR. isOK(check_rule_reflecting fcs R rule)" by simp
  thus ?thesis using check_rule_reflecting_sound[of fcs R] by fast
qed

definition
  check_flat_ctxt :: "('v::show) list \<Rightarrow> ('f::show, 'v) ctxt \<Rightarrow> shows check"
where
  "check_flat_ctxt vs C \<equiv>
     (case C of
       More f ss1 \<box> ss2 \<Rightarrow> do {
         let ss = ss1 @ ss2;
         check (distinct ss) (shows C +@+ '' contains duplicate variables'' +#+ shows_nl);
         check (\<forall>s\<in>set ss. is_Var s) (shows C
           +@+ '' is not flat, i.e., has depth greater than one'' +#+ shows_nl);
         check (\<forall>t\<in>set (ss1 @ ss2). \<not> (the_Var t) \<in> set vs) (shows C +@+
           '' has to contain only fresh variables'' +#+ shows_nl)
       }
     | _ \<Rightarrow> error (shows C +@+ '' is not a flat context'' +#+ shows_nl))"

fun
  is_flat_ctxt_list :: "'v list \<Rightarrow> ('f \<times> nat) list \<Rightarrow> ('f, 'v) ctxt \<Rightarrow> bool"
where
  "is_flat_ctxt_list vs fas (More f ss1 \<box> ss2) = (let ss = ss1 @ ss2 in (
     (f, Suc (length ss)) \<in> set fas
       \<and> (\<forall>s\<in>set ss. is_Var s) \<and> distinct ss \<and> list_inter (map the_Var ss) vs = []
   ))"
| "is_flat_ctxt_list vs fas _ = False"

lemma is_flat_ctxt_list_cases[consumes 1,case_names True]:
  assumes "is_flat_ctxt_list vs fas C" and "\<And>f ss1 ss2. C = More f ss1 \<box> ss2 \<Longrightarrow> P" shows "P"
  using assms by (cases rule: is_flat_ctxt_list.cases[of "(vs,fas,C)"]) (auto simp: Let_def)

lemma is_flat_ctxt_list_sound:
  "is_flat_ctxt_list vs fas C = is_flat_ctxt (set vs) (set fas) C"
proof
  assume fc: "is_flat_ctxt_list vs fas C" thus "is_flat_ctxt (set vs) (set fas) C"
  proof (cases rule: is_flat_ctxt_list_cases)
    case (True f ss1 ss2)
    from fc have "(f,Suc(length(ss1@ss2))) \<in> set fas"
      and "\<forall>s\<in>set (ss1 @ ss2). is_Var s" and "distinct (ss1 @ ss2)"
      and "set (map the_Var (ss1 @ ss2)) \<inter> set vs = {}"
      unfolding True list_inter[symmetric] by (auto simp: Let_def)
    moreover have "set (map the_Var (ss1 @ ss2)) = vars_ctxt C" unfolding True
      using terms_to_vars[OF `\<forall>s\<in>set (ss1 @ ss2). is_Var s`] by (simp add: Let_def)
    ultimately show ?thesis unfolding True by simp
  qed
next
  assume fc: "is_flat_ctxt (set vs) (set fas) C" thus "is_flat_ctxt_list vs fas C"
  proof (cases rule: is_flat_ctxt_cases)
    case (True f ss1 ss2)
    from fc have "(f, Suc (length (ss1 @ ss2))) \<in> set fas"
      and "\<forall>s\<in>set (ss1 @ ss2). is_Var s" and "distinct (ss1 @ ss2)"
      and "list_inter (map the_Var (ss1 @ ss2)) vs = []"
      unfolding True
      using terms_to_vars[of "ss1 @ ss2"] list_inter[of "map the_Var (ss1@ss2)" "vs"] 
      by (auto simp del: list_inter simp: Let_def)
    thus ?thesis unfolding True by simp
  qed
qed

definition
  check_flat_ctxt_complete :: "('f::show, 'v::show) ctxt list \<Rightarrow> ('f \<times> nat) \<Rightarrow> shows check"
where
  "check_flat_ctxt_complete fcs fa \<equiv>
     check (\<forall>i\<in>set [0..<snd fa]. \<exists>C\<in>set fcs. hole_at (snd fa) i (fst fa) C)
       (''the list of flat contexts is incomplete'' +#+ shows_nl)"

definition
  check_is_flat_ctxt ::
    "('v::show) list \<Rightarrow> ('f::show \<times> nat) list \<Rightarrow> ('f, 'v) ctxt \<Rightarrow> shows check"
where
  "check_is_flat_ctxt vs fas C \<equiv> (
     check (is_flat_ctxt_list vs fas C) (shows C +@+ '' is not a flat context'' +#+ shows_nl))"

lemma check_is_flat_ctxt_sound:
  assumes "isOK (check_allm (check_is_flat_ctxt vs fas) fcs)"
  shows "\<forall>C\<in>set fcs. is_flat_ctxt (set vs) (set fas) C"
proof -
  from assms have "\<forall>C\<in>set fcs. is_flat_ctxt_list vs fas C"
    by (auto simp add: check_is_flat_ctxt_def)
  thus ?thesis unfolding is_flat_ctxt_list_sound .
qed

lemma check_flat_ctxt_complete_sound:
  assumes "isOK (check_allm (check_flat_ctxt_complete fcs) F)"
  shows "\<forall>(f, n) \<in> set F. \<forall>i<n. \<exists>ss1 ss2.
    More f ss1 \<box> ss2 \<in> set fcs \<and> length ss1 = i \<and> length ss2 = n - i - 1"
proof (intro ballI2 allI impI)
  fix f n i assume "(f, n) \<in> set F" and "i < n"
  with assms have "\<exists>C\<in>set fcs. hole_at n i f C"
    using check_flat_ctxt_complete_def[of fcs "(f, n)"] by (auto simp: list_all_iff)
  then obtain C where "C \<in> set fcs" and C: "hole_at n i f C" unfolding list_ex_iff by auto
  from C show "\<exists>ss1 ss2. More f ss1 \<box> ss2 \<in> set fcs \<and> length ss1 = i \<and> length ss2 = n - i - 1"
  proof (cases rule: hole_at.cases[of "(n, i, f, C)"])
    case (1 n' i' f' g ss1 ss2) with `C \<in> set fcs` and C show ?thesis by auto
  next
    case ("2_1" n' i' f') with C show ?thesis by simp
  next
    case ("2_2" n' i' f' g ss1 h ts1 D ts2 ss2) with C show ?thesis by simp
  qed
qed

definition fcc_tt ::
  "('tp, 'f::show, 'v::show) tp_ops \<Rightarrow>
   ('f, 'v) ctxt list \<Rightarrow> ('f, 'v) rules \<Rightarrow> 'tp \<Rightarrow> 'tp result"
where
  "fcc_tt I fcs cR tp \<equiv>
    let
      R   = tp_ops.R I tp;
      Q   = tp_ops.Q I tp;
      vs  = vars_trs_list R;
      fas = funas_trs_list R
    in
    check_return (do {
      check (Q = []) (''Q is not empty'' +#+ shows_nl);
      check (tp_ops.Rw I tp = []) (''weak rules are not empty'' +#+ shows_nl);
      check (fcs \<noteq> [])
        (''at least one flat context is required for flat context closure'' +#+ shows_nl);
      check_allm (check_flat_ctxt vs) fcs;
      check_allm (check_is_flat_ctxt vs fas) fcs;
      check_allm (check_flat_ctxt_complete fcs) fas;
      check_allm (check_rule_preserving fcs cR) R;
      check_allm (check_rule_reflecting fcs R) cR
    }) (tp_ops.mk I [] cR [])"

context tp_spec begin

lemma check_fcc_tt[simp]:
  assumes "fcc_tt I fcs R' tp = return tp'"
  shows "Flat_Context_Closure.fcc_tt (set fcs) (qreltrs tp) (qreltrs tp')"
proof -
  note cond = assms[unfolded fcc_tt_def Let_def]
  let ?R  = "set (R tp)"
  let ?R' = "set R'"

  from assms have "tp' = mk [] R' []" by (simp add: fcc_tt_def Let_def)
  hence tp': "qreltrs tp' = ({}, ?R', {})"
    by (simp add: set_empty[symmetric] tp_spec_sound mk_simps)
  
  from cond have tp: "qreltrs tp = ({}, ?R, {})" by (simp add: tp_spec_sound)

  have "Flat_Context_Closure.fcc_tt_cond (set fcs) ?R ?R'"
  proof -
    let ?F  = "funas_trs ?R"
    let ?fs = "funas_trs_list (R tp)"

    have "fcs \<noteq> []" using cond by simp

    moreover have "\<forall>C\<in>set fcs. is_flat_ctxt (vars_trs ?R) ?F C"
    proof -
      from cond
        have "isOK (check_allm (check_is_flat_ctxt
          (vars_trs_list (R tp)) ?fs) fcs)" by simp
      from check_is_flat_ctxt_sound[OF this] show ?thesis by simp
    qed

    moreover have "\<forall>(g, n)\<in>?F. \<forall>i<n. \<exists>ss1 ss2.
      More g ss1 \<box> ss2 \<in> set fcs \<and> length ss1 = i \<and> length ss2 = n - i - 1"
    proof -
      from cond have "isOK (check_allm (check_flat_ctxt_complete fcs) ?fs)" by simp
      from check_flat_ctxt_complete_sound[OF this] show ?thesis by simp
    qed

    moreover have "\<forall>(l, r)\<in>?R. (l, r) \<in> ?R' \<or> (\<forall>C\<in>set fcs. (C\<langle>l\<rangle>, C\<langle>r\<rangle>) \<in> ?R')"
    proof -
      from cond have "isOK (check_allm (check_rule_preserving fcs R') (R tp))"
        by simp
      from check_rules_preserving_sound[OF this] show ?thesis by simp
    qed

    ultimately show ?thesis unfolding fcc_tt_cond_def Let_def by blast
  qed
  thus ?thesis unfolding tp tp' by simp
qed

lemma fcc_tt:
  "sound_tt_impl (fcc_tt I fcs R')"
  unfolding sound_tt_impl_def
proof (intro allI impI)
  fix tp tp'
  let ?R = "set (R tp)"

  assume ok: "fcc_tt I fcs R' tp = return tp'"
    and SN: "SN_qrel (qreltrs tp')"

  from ok have "tp' = mk [] R' []" by (simp add: fcc_tt_def Let_def)
  hence tp': "qreltrs tp' = ({}, set R', {})"
    by (simp add: tp_spec_sound set_empty[symmetric] mk_simps)
  
  from ok have tp: "qreltrs tp = ({}, ?R, {})"
    by (simp add: tp_spec_sound fcc_tt_def Let_def)

  from fcc_tt_sound[OF check_fcc_tt[OF ok, unfolded tp tp'] finite_set
    SN[unfolded tp']]
    show "SN_qrel (qreltrs tp)" unfolding tp by simp
qed

end

definition
  block_trs_list :: "'f \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) rules"
where
  "block_trs_list f trs = map (block_rule f) trs"

lemmas block_defs = block_trs_list_def block_rule_def

definition
  check_superset_of_blocked ::
    "'f::show \<Rightarrow> ('f, 'v::show) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> shows check"
where
  "check_superset_of_blocked f P' P \<equiv> (
     check_all (\<lambda>rule. (block_rule f rule) \<in> set P') P
       <+? (\<lambda>rule. ''the rule '' +#+ shows_rule (block_rule f rule)
         +@+ '' is missing'' +#+ shows_nl)
       <+? (\<lambda>e. shows_trs P  +@+ ''is not a subset of'' +#+
         shows_trs P' +@+ e +@+ shows_nl))"

lemma check_superset_of_blocked_sound:
  assumes "isOK (check_superset_of_blocked f P' P)"
  shows "set (block_trs_list f P) \<subseteq> set P'"
proof (rule subsetI2)
  from assms[unfolded check_superset_of_blocked_def]
  have a: "\<forall>r\<in>set P. block_rule f r \<in> set P'" by simp
  fix s t assume "(s, t) \<in> set (block_trs_list f P)"
  hence "(s, t) \<in> block_rule f ` set P" unfolding block_defs by simp
  then obtain r where st: "block_rule f r = (s, t)" and "r \<in> set P" by auto
  with a have "block_rule f r \<in> set P'" by blast
  thus "(s, t) \<in> set P'" unfolding st .
qed

definition
  check_no_defined_root_defs :: "(('f::show) \<times> nat) list \<Rightarrow> ('f, 'v::show) term \<Rightarrow> shows check"
where
  "check_no_defined_root_defs F t = check (\<not> (the (root t)) \<in> set F) (
     ''the root of '' +#+ shows t +@+ shows_string '' is defined'')"

lemma check_no_defined_root_defs_sound:
  "isOK (check_no_defined_root_defs (defs_list R) t) =
    (\<not> defined (set R) (the (root t)))"
  unfolding check_no_defined_root_defs_def using defs_list_sound by force

definition
  unblock_trs_list :: "'f \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) rules"
where
  "unblock_trs_list f trs \<equiv> map (unblock_rule f) trs"

lemma block_trs_list_sound[simp]: "set (block_trs_list f R) = block_rule f ` set R"
  unfolding block_trs_list_def by (induct R) simp_all

lemma unblock_block_trs_list_ident[simp]:
  "unblock_trs_list f (block_trs_list f trs) = trs"
  unfolding unblock_trs_list_def block_trs_list_def
  by (induct trs) (simp_all add: unblock_rule_def block_rule_def)

lemma block_trs_list_unblock_trs_list[simp]:
  assumes "set (block_trs_list f P) \<subseteq> set P'"
  shows "set P \<subseteq> set (unblock_trs_list f P')"
proof (rule subsetI2)
  fix s t assume "(s, t) \<in> set P"
  hence "block_rule f (s, t) \<in> block_rule f ` set P" by simp
  with assms have "block_rule f (s, t) \<in> set P'" unfolding block_defs by auto
  hence "unblock_rule f (block_rule f (s, t)) \<in> unblock_rule f ` set P'" by simp
  thus "(s, t) \<in> set (unblock_trs_list f P')"
    unfolding block_rule_def unblock_rule_def unblock_trs_list_def by simp
qed

lemma check_rules_preserving_sound_spec:
  fixes f::"'f::show" and fcs::"('f, 'v::show) ctxt list"
  defines "fcs' \<equiv> More f [] \<box> [] # fcs"
  assumes "isOK (check_allm (check_rule_preserving fcs' R') R)"
  shows "\<forall>(l, r)\<in>set R. (l, r) \<in> set R' \<or> (Fun f [l], Fun f [r]) \<in> set R'"
proof -
  from check_rules_preserving_sound[OF assms(2)]
  show ?thesis unfolding fcs'_def by force
qed

definition
  fcc_proc ::
    "('dpp, 'f::show, 'v::show) dpp_ops \<Rightarrow>
    'f \<Rightarrow> ('f, 'v) ctxt list \<Rightarrow>
    ('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow>
    'dpp proc"
where
  "fcc_proc I f fcs P' Pw' R' dpp = check_return (do {
    let P  = dpp_ops.P I dpp;
    let Pw = dpp_ops.Pw I dpp;
    let Pb = list_union P Pw;
    let Q  = dpp_ops.Q I dpp;
    let Rs = dpp_ops.R I dpp;
    let R  = dpp_ops.Rw I dpp;

    let fa   = (f, 1);
    let Cf   = More f [] \<box> [];
    let fcs' = Cf # fcs;
    let vs   = vars_trs_list R;
    let fs   = list_union (funas_trs_list R) (funas_args_trs_list Pb);
    let fas  = fa # fs;
    let ds   = defs_list R;

    check (Q = []) (''Q is not empty'' +#+ shows_nl);
    check (Rs = []) (''the strict rules are not empty'' +#+ shows_nl);
    check (\<not> fa \<in> set ds) (shows f +@+ ''is not fresh'' +#+ shows_nl);
    check_left_linear_trs R;
    check_wf_trs R;
    check_allm (\<lambda>r.
      check_no_var (fst r) \<guillemotright>
      check_no_var (snd r) \<guillemotright>
      check_no_defined_root_defs ds (snd r)) Pb;
    check_allm (check_flat_ctxt vs) fcs';
    check_allm (check_is_flat_ctxt vs fas) fcs';
    check_allm (check_flat_ctxt_complete fcs') fas;
    check_allm (check_rule_preserving fcs' R') R;
    check_allm (check_rule_reflecting fcs' R) R';
    check_superset_of_blocked f P' P;
    check_superset_of_blocked f Pw' Pw
  }) (dpp_ops.mk I P' Pw' [] [] R')"

context dpp_spec
begin

lemma check_fcc_proc[simp]:
  assumes "fcc_proc I f fcs P' Pw' R' d = return d'"
  shows "Flat_Context_Closure.fcc_proc f fcs (dpp_ops.dpp I d) (dpp_ops.dpp I d')"
proof -
  note cond = assms[unfolded fcc_proc_def Let_def]
  let ?P   = "set (dpp_ops.P I d)"
  let ?P'  = "set P'"
  let ?Pw  = "set (dpp_ops.Pw I d)"
  let ?Pw' = "set Pw'"
  let ?R   = "set (dpp_ops.Rw I d)"
  let ?R'  = "set R'"

  from assms have "d' = dpp_ops.mk I P' Pw' [] [] R'" by (simp add: fcc_proc_def Let_def)
  hence d': "dpp_ops.dpp I d' = (?P', ?Pw', {}, {}, ?R')"
    by (simp add: dpp_spec_sound set_empty[symmetric] mk_simps)
  
  from cond have d: "dpp_ops.dpp I d = (?P, ?Pw, {}, {}, ?R)" by (simp add: dpp_spec_sound)

  have "Flat_Context_Closure.fcc_cond f fcs ?P ?Pw ?R ?P' ?Pw' ?R'"
  proof -
    let ?Cf = "More f [] \<box> []"
    let ?F' = "funas_trs ?R \<union> funas_args_trs (?P \<union> ?Pw)"
    let ?fs = "list_union
      (funas_trs_list (dpp_ops.Rw I d))
      (funas_args_trs_list (list_union (dpp_ops.P I d) (dpp_ops.Pw I d)))"
    let ?F  = "{(f, 1)} \<union> ?F'"

    have "wf_trs ?R"
      and "left_linear_trs ?R"
      and "\<not> defined ?R (f, 1)"
      using cond by simp_all

    moreover have "\<forall>C\<in>set (?Cf # fcs). is_flat_ctxt (vars_trs ?R) ?F C"
    proof -
      from cond
        have "isOK (check_allm (check_is_flat_ctxt
          (vars_trs_list (dpp_ops.Rw I d)) ((f, 1) # ?fs)) (?Cf # fcs))" by simp
      from check_is_flat_ctxt_sound[OF this] show ?thesis by simp
    qed

    moreover have "\<forall>(g, n)\<in>?F'. \<forall>i<n. \<exists>ss1 ss2.
      More g ss1 \<box> ss2 \<in> set (?Cf # fcs) \<and> length ss1 = i \<and> length ss2 = n - i - 1"
    proof -
      from cond have "isOK (check_allm (check_flat_ctxt_complete (?Cf # fcs))
        ((f, 1) # ?fs))" by simp
      from check_flat_ctxt_complete_sound[OF this] show ?thesis by simp
    qed

    moreover have "\<forall>(l, r)\<in>?R. (l, r) \<in> ?R' \<or> (\<forall>C\<in>set (?Cf # fcs). (C\<langle>l\<rangle>, C\<langle>r\<rangle>) \<in> ?R')"
    proof -
      from cond have "isOK (check_allm (check_rule_preserving (?Cf # fcs) R') (dpp_ops.Rw I d))"
        by simp
      from check_rules_preserving_sound[OF this] show ?thesis by simp
    qed

    moreover have "\<forall>(l, r)\<in>?R'. \<exists>(u, v)\<in>?R. \<exists>C\<in>set (\<box> # ?Cf # fcs). l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>"
    proof -
      from cond have "isOK (check_allm (check_rule_reflecting (?Cf # fcs) (dpp_ops.Rw I d)) R')"
        by simp
      from check_rules_reflecting_sound[OF this] show ?thesis by simp
    qed

    moreover have "block_rule f ` ?P \<subseteq> ?P'"
    proof -
      from cond have "isOK (check_superset_of_blocked f P' (dpp_ops.P I d))" by simp
      from check_superset_of_blocked_sound[OF this] show ?thesis by simp
    qed

    moreover have "block_rule f ` ?Pw \<subseteq> ?Pw'"
    proof -
      from cond have "isOK (check_superset_of_blocked f Pw' (dpp_ops.Pw I d))" by simp
      from check_superset_of_blocked_sound[OF this] show ?thesis by simp
    qed

    moreover have "\<forall>(s, t)\<in>set (dpp_ops.P I d) \<union> set (dpp_ops.Pw I d). is_Fun s \<and> is_Fun t"
    proof -
      from cond have "isOK (check_allm (\<lambda>r. check_no_var (fst r) \<guillemotright> check_no_var (snd r))
        (list_union (dpp_ops.P I d) (dpp_ops.Pw I d)))" by simp
      thus ?thesis by force
    qed

    moreover have "\<forall>(s, t)\<in>set (dpp_ops.P I d) \<union> set (dpp_ops.Pw I d).
        \<not> defined (set (dpp_ops.Rw I d)) (the (root t))"
    proof -
      from cond have "isOK (check_allm (\<lambda>r. check_no_defined_root_defs
        (defs_list (dpp_ops.Rw I d)) (snd r)) (list_union (dpp_ops.P I d) (dpp_ops.Pw I d)))"
        by simp
      with check_no_defined_root_defs_sound show ?thesis by force
    qed

    ultimately show ?thesis unfolding fcc_cond_def Let_def by blast
  qed
  thus ?thesis unfolding d d' by simp
qed

lemma fcc_proc:
  "sound_proc_impl (fcc_proc I f fcs P' Pw' R')"
proof
  fix d d'
  let ?P  = "set (dpp_ops.P I d)"
  let ?Pw = "set (dpp_ops.Pw I d)"
  let ?R  = "set (dpp_ops.Rw I d)"

  assume ok: "fcc_proc I f fcs P' Pw' R' d = return d'"
    and fin: "finite_dpp (dpp_ops.dpp I d')"

  from ok have "d' = dpp_ops.mk I P' Pw' [] [] R'" by (simp add: fcc_proc_def Let_def)
  hence d': "dpp_ops.dpp I d' = (set P', set Pw', {}, {}, set R')"
    by (simp add: dpp_spec_sound set_empty[symmetric] mk_simps)
  
  from ok have d: "dpp_ops.dpp I d = (?P, ?Pw, {}, {}, ?R)"
    by (simp add: dpp_spec_sound fcc_proc_def Let_def)

  from fcc_proc_sound[OF check_fcc_proc[OF ok, unfolded d d'] finite_set
    fin[unfolded d']]
  show "finite_dpp (dpp_ops.dpp I d)" unfolding d by simp
qed

end

end


