(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Innermost_Usable_Rules
imports
  Icap 
  QDP_Framework
begin

fun is_ur_closed_term' :: "('f,string)trs \<Rightarrow> ('f,string)trs \<Rightarrow> ('f,string)terms \<Rightarrow> ('f,string)cap_fun \<Rightarrow> ('f,string)terms \<Rightarrow> ('f,string)term \<Rightarrow> bool"
  where "is_ur_closed_term' R U Q ecap S (Var x) = True"
     |  "is_ur_closed_term' R U Q ecap S (Fun f ts) = 
              ((\<forall> t \<in> set ts. is_ur_closed_term' R U Q ecap S t) \<and>                
              (\<forall> (l,r) \<in> R. (l,r) \<in> U \<or> \<not> (\<exists> \<mu>.
              mgu_class (Fun f (map (ecap R Q S) ts)) l = Some \<mu> \<and> (S \<union> mv_yvar ` (set (args l))) \<cdot>set \<mu> \<subseteq> NF_terms Q)))"

lemma is_ur_closed_subt: "t \<unrhd> s \<Longrightarrow> is_ur_closed_term' R U Q ecap S t \<Longrightarrow> is_ur_closed_term' R U Q ecap S s"
proof (induct rule: supteq.induct)
  case (refl t)
  thus ?case .
next
  case (subt u ss t f)
  from subt(4) subt(1) have "is_ur_closed_term' R U Q ecap S u" by auto
  from subt(3)[OF this] show ?case .
qed

abbreviation is_ur_closed_term_mv' :: "('f,string)trs \<Rightarrow> ('f,string)trs \<Rightarrow> ('f,string)terms \<Rightarrow> ('f,string)cap_fun \<Rightarrow> ('f,string)terms \<Rightarrow> ('f,string)term \<Rightarrow> bool"
  where "is_ur_closed_term_mv' R U Q ecap S t \<equiv> is_ur_closed_term' R U Q ecap (mv_xvar ` S) (mv_xvar t)"

lemma is_ur_closed_mv_subt: "t \<unrhd> s \<Longrightarrow> is_ur_closed_term_mv' R U Q ecap S t \<Longrightarrow> is_ur_closed_term_mv' R U Q ecap S s"
proof (induct rule: supteq.induct)
  case (refl t)
  thus ?case .
next
  case (subt u ss t f)
  from subt(4) subt(1) have "is_ur_closed_term_mv' R U Q ecap S u" by auto
  from subt(3)[OF this] show ?case .
qed


lemma is_ur_closed_term_mv_icap': "is_ur_closed_term_mv' R U Q icap' S t = is_ur_closed_term_mv' R U Q icap S t" (is "?l t = ?r t")
proof (induct t)
  case (Var x) show ?case by simp
next
  case (Fun f ts)  
  hence rec: "(\<forall> t \<in> set ts. ?l t) = (\<forall> t \<in> set ts. ?r t)" by auto
  let ?m = "mv_xvar"
  let ?S = "?m ` S"
  have id: "map (icap' R Q ?S) (map ?m ts) = map (icap R Q ?S) (map ?m ts)"
    unfolding map_eq_conv set_map using icap'[of R Q S] by auto
  show "?l (Fun f ts) = ?r (Fun f ts)"    
    unfolding map_vars.simps is_ur_closed_term'.simps
    unfolding set_map ball_simps
    unfolding rec 
    unfolding id ..
qed


locale R_Q_U_ecap = 
  fixes R U :: "('f,string)trs"
  and Q :: "('f,string)terms"
  and ecap :: "('f,string)cap_fun" 
  assumes ecap: "is_ecap ecap"
    and NF_Q_R: "NF_terms Q \<subseteq> NF_trs R"
begin

abbreviation NFQ :: "('f,string)terms" where "NFQ \<equiv> NF_terms Q"
abbreviation e_cap where "e_cap \<equiv> ecap R Q"
abbreviation is_ur_closed_term where "is_ur_closed_term \<equiv> is_ur_closed_term' R U Q ecap"
abbreviation is_ur_closed_term_mv where "is_ur_closed_term_mv \<equiv> is_ur_closed_term_mv' R U Q ecap"

lemma is_ur_closed_term_main: assumes steps: "\<And> i. i < n \<Longrightarrow> (ts i, ts (Suc i)) \<in> qrstep Q {lr i}"
  and varcond: "\<And> i. i < n \<Longrightarrow> vars_term (snd (lr i)) \<subseteq> vars_term (fst (lr i)) \<and> is_Fun (fst (lr i))"
  and \<sigma>: "\<sigma> ` vars_term t \<subseteq> NFQ"
  and lr: "\<And> i. i < n \<Longrightarrow> lr i \<in> R"
  and t0: "ts 0 = t \<cdot> \<sigma>"
  and S: "S \<cdot>set \<sigma> \<subseteq> NFQ"
  and U: "\<And> l r. (l,r) \<in> U \<Longrightarrow> is_ur_closed_term_mv (set (args l)) r"
  and closed: "is_ur_closed_term_mv S t"
  and i: "i < n"
  shows "lr i \<in> U"
proof -
  let ?vc = "\<lambda> (lr :: nat \<Rightarrow> ('f,string)rule) i. vars_term (snd (lr i)) \<subseteq> vars_term (fst (lr i)) \<and> is_Fun (fst (lr i))"
  obtain vc where vc: "vc = ?vc" by auto
  let ?m = "[\<lambda> (nt :: nat \<times> ('f,string)term). fst nt, \<lambda> nt. size (snd nt)]"
  let ?P = "\<lambda> (n,t). (\<forall> \<sigma> u S ts lr. (\<forall> i < n. vc lr i) \<longrightarrow> ts 0 = t \<cdot> \<sigma> \<longrightarrow> ts n = u \<longrightarrow> S \<cdot>set \<sigma> \<subseteq> NFQ \<longrightarrow> (\<sigma> ` vars_term t \<subseteq> NFQ) \<longrightarrow> is_ur_closed_term_mv S t \<longrightarrow> (\<forall> i < n. (ts i, ts (Suc i)) \<in> qrstep Q {lr i}) \<longrightarrow> (\<forall> i < n. lr i \<in> R) \<longrightarrow> (\<forall> i < n. lr i \<in> U))"
  {
    fix n t
    have "?P (n,t)"
    proof (induct rule: wf_induct[OF wf_measures[of ?m], of ?P])
      case (1 nt)
      obtain n t where nt: "nt = (n,t)" by force
      note 1 = 1[unfolded nt, rule_format]
      show ?case
        unfolding nt split
      proof (intro allI, intro impI)
        fix \<sigma> u S ts lr 
        assume vc_lr: "\<forall> i < n. vc lr i"
          and t\<sigma>: "ts 0 = t \<cdot> \<sigma>" 
          and tsn: "ts n = u" 
          and S: "S \<cdot>set \<sigma> \<subseteq> NFQ" 
          and \<sigma>: "\<sigma> ` vars_term t \<subseteq> NFQ" 
          and ur: "is_ur_closed_term_mv S t"
          and steps: "\<forall> i < n. (ts i, ts (Suc i)) \<in> qrstep Q {lr i}"
          and R: "\<forall> i < n. lr i \<in> R"
        let ?S = "mv_xvar ` S"
        let ?t = "mv_xvar t"
        show "\<forall> i < n. lr i \<in> U"
        proof (cases n)
          case 0 thus ?thesis by auto
        next
          case (Suc m)
          {
            fix i
            assume i: "i < m"
            with vc_lr Suc have vc: "\<And> i. i < m \<Longrightarrow> vc lr i" by auto
            have m: "((m,t),(n,t)) \<in> measures ?m" unfolding Suc by auto
            have "lr i \<in> U"
              by (rule 1[OF m, unfolded split, rule_format], rule vc, simp, rule t\<sigma>, rule refl, insert S \<sigma> ur i R steps Suc, auto)
          } note IH = this
          have main: "lr m \<in> U"
          proof (cases "\<forall> i < m. (ts i, ts (Suc i)) \<in> nrqrstep Q {lr i}")
            case False
            then obtain i where i: "i < m" and "(ts i, ts (Suc i)) \<notin> nrqrstep Q {lr i}" by auto
            with steps[THEN spec[of _ i]] have "(ts i, ts (Suc i)) \<in> rqrstep Q {lr i}" unfolding Suc qrstep_iff_rqrstep_or_nrqrstep by auto
            then obtain l r \<tau> where l\<tau>: "l \<cdot> \<tau> = ts i" and r\<tau>: "r \<cdot> \<tau> = ts (Suc i)"  and lr: "(l,r) = lr i" and NF: "\<forall> u \<lhd> l \<cdot> \<tau>. u \<in> NFQ" by auto
            from IH[OF i] U lr
            have ur: "is_ur_closed_term_mv (set (args l)) r" by auto
            from i Suc lr R R have lr': "(l,r) \<in> R" by auto
            let ?ts = "\<lambda> j. ts (Suc i + j)"
            let ?lr = "\<lambda> j. lr (Suc i + j)"
            from i Suc have i': "i < n" by simp
            from vc_lr[rule_format, OF i', unfolded vc]
            have vcond: "vars_term r \<subseteq> vars_term l" "is_Fun l" unfolding lr[symmetric] by auto
            then obtain f ls where l: "l = Fun f ls" by (cases l, auto)
            {
              fix x
              assume "x \<in> vars_term r"
              with vcond have "x \<in> vars_term l" by auto
              hence "Var x \<lhd> l" unfolding l by auto
              hence "Var x \<cdot> \<tau> \<lhd> l \<cdot> \<tau>" by (rule supt_stable)
              with NF have "\<tau> x \<in> NFQ" by auto
            }
            hence \<tau>: "\<tau> ` vars_term r \<subseteq> NFQ" by auto
            from NF[unfolded NF_terms_args_conv[symmetric]] have NF: "set (args l) \<cdot>set \<tau> \<subseteq> NFQ" by (cases l, auto)
            let ?m' = "m - Suc i"
            have IH: "((n - Suc i, r), (n,t)) \<in> measures ?m" unfolding Suc by auto
            have "lr (Suc i + ?m') \<in> U"
              by (rule 1[OF IH, unfolded split Suc, rule_format, OF _ _ refl NF \<tau> ur, of ?lr ?ts ?m'], insert R[unfolded Suc] steps[unfolded Suc] vc_lr i Suc, auto simp: r\<tau> vc)
            with i show ?thesis by auto
          next
            case True note nrr = this
            {
              assume "is_Var t"
              then obtain x where t: "t = Var x" by (cases t, auto)
              with \<sigma> have "\<sigma> x \<in> NFQ" by auto
              with NF_Q_R have "\<sigma> x \<in> NF_trs R" by auto
              hence "\<sigma> x \<in> NF (qrstep Q R)" using qrstep_subset_rstep[of Q R] 
                unfolding NF_def by blast
              hence NF: "t \<cdot> \<sigma> \<in> NF (qrstep Q R)" unfolding t by simp
              from steps[THEN spec[of _ 0]] have "(t \<cdot> \<sigma>, ts (Suc 0)) \<in> qrstep Q R" 
                using R[THEN spec[of _ 0]] R unfolding Suc qrstep_rule_conv[where R = R] t\<sigma> by auto
              with NF have False by auto
            }
            then obtain f ss where t: "t = Fun f ss" by (cases t, auto)
            with t\<sigma> have ts0: "ts 0 = Fun f (map (\<lambda>s. s \<cdot> \<sigma>) ss)" by auto
            let ?n = "length ss" 
            let ?LR = "{(l,r) | l r. vars_term r \<subseteq> vars_term l \<and> is_Fun l}"
            let ?Q = "\<lambda> ts' is' i. ts i = Fun f ts' \<and> length ts' = ?n \<and> length is' = ?n \<and> (\<forall> j \<in> set is'. j \<le> i) \<and> (\<forall> j < ?n. (ss ! j \<cdot> \<sigma>, ts' ! j) \<in> (qrstep Q (R \<inter> ?LR))^^(is' ! j))"
            {
              fix i
              assume i: "i \<le> m"
              from i have "\<exists> ts' is'. ?Q ts' is' i"
              proof (induct i)
                case 0
                show ?case unfolding ts0
                  by (rule exI, rule exI[of _ "replicate ?n 0"], intro conjI, rule refl, auto)
              next
                case (Suc i)
                hence "i \<le> m" by auto
                from Suc(1)[OF this] obtain ts' is' where Q: "?Q ts' is' i" by blast
                from Suc have i: "i < m" by auto
                with R[THEN spec[of _ i]] have R: "lr i \<in> R" unfolding `n = Suc m` by auto
                from True[rule_format, OF i] have step: "(ts i, ts (Suc i)) \<in> nrqrstep Q ({lr i})" by force
                from vc_lr i `n = Suc m` have "vc lr i" by auto
                with vc have vcond: "lr i \<in> ?LR" by auto
                have "(ts i, ts (Suc i)) \<in> nrqrstep Q (R \<inter> ?LR)"
                  by (rule set_mp[OF nrqrstep_mono step], insert R vcond, auto)
                with Q have step: "(Fun f ts', ts (Suc i)) \<in> nrqrstep Q (R \<inter> ?LR)" by auto
                then obtain l r D \<tau> where step: "(l,r) \<in> (R \<inter> ?LR)" "Fun f ts' = D\<langle>l\<cdot>\<tau>\<rangle>" "ts (Suc i) = D\<langle>r \<cdot> \<tau>\<rangle>" "D \<noteq> \<box>" "\<forall> u \<lhd> l\<cdot>\<tau>. u \<in> NFQ" 
                  unfolding nrqrstep_def by blast
                from step obtain bef C aft where "D = More f bef C aft" by (cases D, auto)
                note step = step[unfolded this]
                from step have ts': "ts' = bef @ C\<langle>l\<cdot>\<tau>\<rangle> # aft" by simp
                let ?j = "length bef"
                let ?ts' = "bef @ C\<langle>r\<cdot>\<tau>\<rangle> # aft"
                let ?is' = "is'[?j := Suc(is' ! ?j)]"
                show ?case
                proof (rule exI[of _ ?ts'], rule exI[of _ ?is'], intro conjI)
                  show "ts (Suc i) = Fun f ?ts'" unfolding step by simp
                next
                  show "length ?ts' = ?n" using Q step by auto
                next
                  show "length ?is' = ?n" using Q step by auto
                next
                  have j: "?j < length is'" using Q step by auto
                  show "\<forall> j \<in> set ?is'. j \<le> Suc i" 
                  proof
                    fix j
                    assume j': "j \<in> set ?is'"
                    have "j \<in> insert (Suc (is' ! ?j)) (set is')" 
                      by (rule set_mp[OF set_update_subset_insert j'])
                    hence "j = Suc (is' ! ?j) \<or> j \<in> set is'" by auto
                    thus "j \<le> Suc i"
                    proof
                      assume j': "j = Suc (is' ! ?j)"
                      from j have "is' ! ?j \<in> set is'" by auto
                      with Q have "is' ! ?j \<le> i" by auto
                      thus ?thesis using j' by auto
                    next
                      assume j: "j \<in> set is'"
                      with Q show ?thesis by auto
                    qed
                  qed
                next
                  show "\<forall> j < ?n. (ss ! j \<cdot> \<sigma>, ?ts' ! j) \<in> qrstep Q (R \<inter> ?LR) ^^ (?is' ! j)"
                  proof (intro allI impI)
                    fix j
                    assume j: "j < ?n"
                    with Q have steps: "(ss ! j \<cdot> \<sigma>, ts' ! j) \<in> qrstep Q (R \<inter> ?LR) ^^ (is' ! j)" by simp
                    show "(ss ! j \<cdot> \<sigma>, ?ts' ! j) \<in> qrstep Q (R \<inter> ?LR) ^^ (?is' ! j)"
                    proof (cases "?j = j")
                      case False
                      with steps show ?thesis unfolding ts' nth_append nth_list_update_neq[OF False]
                        by auto
                    next
                      case True
                      hence j: "j = length bef" "length bef < length is'" using j step Q by auto
                      from step have onestep: "(ts' ! j, ?ts' ! j) \<in> qrstep Q (R \<inter> ?LR)" unfolding j ts' by auto
                      from j have is': "?is' ! j = Suc (is' ! j)" by auto
                      show ?thesis unfolding is' using steps onestep by auto
                    qed
                  qed
                qed
              qed
            } note details = this
            hence "\<forall> i. \<exists> ts' is'. i \<le> m \<longrightarrow> ?Q ts' is' i" by auto
            from choice[OF this] obtain ts' where "\<forall> i. \<exists> is'. i \<le> m \<longrightarrow> ?Q (ts' i) is' i" by auto
            from choice[OF this] obtain is' where Q: "\<And> i. i \<le> m \<Longrightarrow> ?Q (ts' i) (is' i) i" by auto
            show ?thesis 
            proof (cases "(ts m, ts (Suc m)) \<in> rqrstep Q {lr m}")
              case True
              then obtain l r \<tau> where lr: "(l,r) \<in> {lr m}" and l\<tau>: "ts m = l \<cdot> \<tau>" and r\<tau>: "ts (Suc m) = r \<cdot> \<tau>" and NF: "\<forall> s \<lhd> l \<cdot> \<tau>. s \<in> NFQ" by auto
              from lr R[THEN spec[of _ m]] have lr': "(l,r) \<in> R" unfolding Suc by auto
              let ?ss = "map mv_xvar ss"
              let ?\<sigma> = "mv_subst \<sigma>"
              from ur[unfolded t, unfolded is_ur_closed_term'.simps] lr' R
              have "(l,r) \<in> U \<or> \<not> (\<exists> \<mu>. mgu_class (Fun f (map (e_cap ?S) ?ss)) l = Some \<mu> \<and> (?S \<union> mv_yvar ` set (args l)) \<cdot>set \<mu> \<subseteq> NFQ)" (is "_ \<or> \<not> ?check") by auto
              thus ?thesis
              proof
                assume "(l,r) \<in> U" with lr show ?thesis by auto
              next                
                assume ncheck: "\<not> ?check"
                from l\<tau> Q[of m] have l\<tau>: "l \<cdot> \<tau> = Fun f (ts' m)" by auto
                from S have S': "mv_xvar ` S \<cdot>set ?\<sigma> \<subseteq> NFQ" by (auto simp: mv_xvar[of _ \<sigma>])
                have inst: "l \<cdot> \<tau> \<in> fresh_instances_subst (Fun f (map (e_cap ?S) ?ss)) ?\<sigma>" unfolding l\<tau>
                  unfolding fresh_instances_subst.simps
                proof (rule, unfold length_map, intro exI conjI, rule refl)
                  show "length (ts' m) = ?n" using Q[of m] by simp
                next
                  show "\<forall> i < ?n. ts' m ! i \<in> fresh_instances_subst (map (e_cap ?S) ?ss ! i) ?\<sigma>"
                  proof (intro allI impI)
                    fix i
                    assume i: "i < ?n"
                    show "ts' m ! i \<in> fresh_instances_subst (map (e_cap ?S) ?ss ! i) ?\<sigma>"
                      unfolding map_map o_def
                      unfolding nth_map[OF i]
                    proof (rule ecap_steps[OF ecap])
                      from Q[of m] i have steps: "(ss ! i \<cdot> \<sigma>, ts' m ! i) \<in> qrstep Q (R \<inter> ?LR) ^^ is' m ! i" (is "?p \<in> _") by auto
                      have "?p \<in> qrstep Q R ^^ is' m ! i"
                        by (rule set_mp[OF rel_pow_mono steps], auto)
                      thus "(mv_xvar (ss ! i) \<cdot> ?\<sigma>, ts' m ! i) \<in> (qrstep Q R)^*" 
                        unfolding mv_xvar[of _ \<sigma>]
                        by (rule rel_pow_imp_rtrancl)
                    qed (insert S', auto)
                  qed
                qed
                {
                  fix x
                  assume x: "Inr x \<in> vars_term (Fun f (map (e_cap ?S) ?ss))"
                  then obtain s where s: "s \<in> set ss" and x: "Inr x \<in> vars_term (e_cap ?S (mv_xvar s))" by auto
                  from x vars_term_ecap[OF ecap, of R Q ?S "mv_xvar s"]
                  have "x \<in> range x_var" unfolding vars_term_map_vars by auto
                }
                from mgu_class_complete[OF inst this]
                obtain \<mu> \<delta> where mgu: "mgu_class (Fun f (map (e_cap ?S) ?ss)) l = Some \<mu>"
                  and \<sigma>: "\<And> x. x \<in> range x_var \<Longrightarrow> ?\<sigma> x = \<mu> x \<cdot> \<delta>"
                  and \<tau>: "\<And> t. t \<cdot> \<tau> = mv_yvar t \<cdot> \<mu> \<cdot> \<delta>" by blast
                  {
                    fix t
                    assume "t \<in> ?S \<cdot>set \<mu>"
                    then obtain s where s: "s \<in> S" and t: "t = mv_xvar s \<cdot> \<mu>" by auto
                    from s S have NF: "s \<cdot> \<sigma> \<in> NFQ" by auto
                    also have "s \<cdot> \<sigma> = mv_xvar s \<cdot> (\<mu> \<circ>s \<delta>)"
                      unfolding mv_xvar[of _ \<sigma>]
                      by (rule term_subst_eq, unfold subst_compose_def, rule \<sigma>,
                        unfold vars_term_map_vars, auto)
                    also have "... = mv_xvar s \<cdot> \<mu> \<cdot> \<delta>" by simp                    
                    finally have "t \<in> NFQ" unfolding t
                      by (rule NF_instance)
                  } 
                  hence NFS: "?S \<cdot>set \<mu> \<subseteq> NFQ" by auto
                  have "\<forall> u \<in> set (args l). u \<cdot> \<tau> \<in> NFQ" using NF[unfolded NF_terms_args_conv[symmetric]] by (cases l, auto)
                  from this[unfolded \<tau>]
                  have NFl: "mv_yvar ` set (args l) \<cdot>set \<mu> \<subseteq> NFQ" using NF_instance[of _ \<delta> "Id_on Q"] by auto
                have "?check" 
                  by (intro exI conjI, rule mgu, insert NFS NFl, auto)
                with ncheck show ?thesis ..
              qed
            next
              assume "(ts m,ts (Suc m)) \<notin> rqrstep Q {lr m}"
              with steps[THEN spec[of _ m]] have "(ts m, ts (Suc m)) \<in> nrqrstep Q {lr m}" unfolding Suc qrstep_iff_rqrstep_or_nrqrstep by auto
              then obtain l r D \<tau> where step: "(l,r) \<in> {lr m}" "ts m = D\<langle>l\<cdot>\<tau>\<rangle>" "ts (Suc m) = D\<langle>r \<cdot> \<tau>\<rangle>" "D \<noteq> \<box>" "\<forall> u \<lhd> l\<cdot>\<tau>. u \<in> NFQ" 
                unfolding nrqrstep_def by auto
              from Q[of m]  have tsm: "ts m = Fun f (ts' m)" by auto
              from tsm step obtain bef C aft where D: "D = More f bef C aft" by (cases D, auto)
              note step = step[unfolded D]
              let ?i = "length bef"
              from step tsm Q[of m] have i: "?i < ?n" by auto
              from Q[of m] i obtain j where steps: "(ss ! ?i \<cdot> \<sigma>, ts' m ! ?i) \<in> qrstep Q (R \<inter> ?LR) ^^ j" and j: "j \<le> m" by auto
              from steps[unfolded rel_pow_fun_conv] obtain ss' where first: "ss' 0 = ss ! ?i \<cdot> \<sigma>" and last: "ss' j = ts' m ! ?i"
                and steps: "\<forall> i < j. (ss' i, ss' (Suc i)) \<in> qrstep Q (R \<inter> ?LR)" by auto
              {
                fix i
                assume i: "i < j"
                from steps[THEN spec, THEN mp[OF _ this], unfolded qrstep_rule_conv[where R = "R \<inter> ?LR"]] 
                have "\<exists> lr. lr \<in> R \<and> lr \<in> ?LR \<and> (ss' i, ss' (Suc i)) \<in> qrstep Q {lr}" by auto
              }
              hence "\<forall> i. \<exists> lr. i < j \<longrightarrow> lr \<in> R \<and> lr \<in> ?LR \<and> (ss' i, ss' (Suc i)) \<in> qrstep Q {lr}" by auto
              from choice[OF this] obtain lr' where lr'': "\<And> i. i < j \<Longrightarrow> lr' i \<in> R" "\<And> i. i < j \<Longrightarrow> lr' i \<in> ?LR" and steps: "\<And> i. i < j \<Longrightarrow> (ss' i, ss' (Suc i)) \<in> qrstep Q {lr' i}" by blast
              let ?ss = "\<lambda> i. if i \<le> j then ss' i else C\<langle>r\<cdot>\<tau>\<rangle>"
              let ?lr = "\<lambda> i. if i < j then lr' i else lr m"
              have lr': "\<And> i. ?lr i \<in> R" using lr'' R[THEN spec[of _ m]] unfolding Suc by auto
              have vc_lr: "\<And> i. vc ?lr i" using lr'' vc_lr Suc unfolding vc 
                by force
              from tsm step have tsm: "ts' m = bef @ C\<langle>l\<cdot>\<tau>\<rangle> # aft" by auto
              from step have "(C\<langle>l\<cdot>\<tau>\<rangle>, C\<langle>r\<cdot>\<tau>\<rangle>) \<in> qrstep Q {lr m}" by auto
              hence "(ts' m ! ?i, C\<langle>r\<cdot>\<tau>\<rangle>) \<in> qrstep Q {lr m}" unfolding tsm by auto
              with last have step: "(?ss j, ?ss (Suc j)) \<in> qrstep Q {lr m}" by auto
              from j Suc have jn: "Suc j \<le> n" by auto
              from i t have ssi: "ss ! ?i \<in> set ss" by auto
              hence size: "size (ss ! ?i) < size t" unfolding t by (auto simp: size_simps)
              from ur[unfolded t] ssi have ur: "is_ur_closed_term_mv S (ss ! ?i)" by auto
              from \<sigma>[unfolded t] ssi have \<sigma>: "\<sigma> ` vars_term (ss ! ?i) \<subseteq> NFQ" by auto
              have meas: "((Suc j,ss ! ?i), (n,t)) \<in> measures ?m" using jn size by auto
              have "?lr j \<in> U"
              proof (rule 1[OF meas, unfolded split, rule_format, OF vc_lr _ refl S \<sigma> ur _ lr', of ?ss],
                unfold first)
                fix i
                assume i: "i < Suc j"
                show "(?ss i, ?ss (Suc i)) \<in> qrstep Q {?lr i}"
                proof (cases "i < j")
                  case False
                  with i have i: "i = j" by auto
                  show ?thesis unfolding i using step by auto
                next
                  case True
                  from steps[OF this] True show ?thesis by auto
                qed
              qed auto
              thus ?thesis by simp
            qed
          qed
          show ?thesis unfolding Suc 
          proof (intro allI impI)
            fix i
            assume "i < Suc m"
            hence "i < m \<or> i = m" by auto
            with IH main show "lr i \<in> U" by auto
          qed
        qed
      qed
    qed
  }
  hence main: "?P (n,t)" .
  show ?thesis
    by (rule main[unfolded split vc, rule_format, of _ ts, OF _ t0 refl S \<sigma> closed steps lr i], insert varcond, auto)
qed


lemma is_ur_closed_term: assumes steps: "(t \<cdot> \<sigma>, u) \<in> (qrstep Q R')^*"
  and \<sigma>: "\<sigma> ` vars_term t \<subseteq> NFQ"
  and S: "S \<cdot>set \<sigma> \<subseteq> NFQ"
  and U: "\<And> l r. (l,r) \<in> U \<Longrightarrow> is_ur_closed_term_mv (set (args l)) r"
  and R': "R' \<subseteq> R"
  and varcond_or_SN: "wf_trs R' \<or> SN_on (qrstep Q R') {t \<cdot> \<sigma>}"
  and closed: "is_ur_closed_term_mv S t"
  shows "(t \<cdot> \<sigma>, u) \<in> (qrstep Q (R' \<inter> U))^*"
proof -
  from steps[unfolded qrsteps_rules_conv] obtain n ts lr where first: "ts 0 = t \<cdot> \<sigma>" and last: "ts n = u" and steps: "\<And> i. i < n \<Longrightarrow> (ts i, ts (Suc i)) \<in> qrstep Q {lr i}" and lr: "\<And> i. i < n \<Longrightarrow> lr i \<in> R'" by blast+
  with R' have lr': "\<And> i. i < n \<Longrightarrow> lr i \<in> R" by auto
  {
    fix i
    assume i: "i < n"
    have "lr i \<in> U"
    proof (rule is_ur_closed_term_main[where ts = ts and n = n, OF steps _ \<sigma> lr' first S U closed i])
      fix i'
      assume i': "i' < n"
      from varcond_or_SN show "vars_term (snd (lr i')) \<subseteq> vars_term (fst (lr i')) \<and> is_Fun (fst (lr i'))"
      proof
        assume SN: "SN_on (qrstep Q R') {t \<cdot> \<sigma>}"
        have i'steps: "(t \<cdot> \<sigma>, ts i') \<in> (qrstep Q R')^*"
          unfolding qrsteps_rules_conv
        proof (intro exI conjI, rule first, rule refl)
          show "\<forall> i < i'. (ts i, ts (Suc i)) \<in> qrstep Q {lr i} \<and> lr i \<in> R'"
          proof (intro allI impI)
            fix i
            assume "i < i'" with i' have "i < n" by simp
            from steps[OF this] lr[OF this]
            show "(ts i, ts (Suc i)) \<in> qrstep Q {lr i} \<and> lr i \<in> R'" by simp
          qed
        qed        
        have SN: "SN_on (qrstep Q R') {ts i'}"
          by (rule steps_preserve_SN_on[OF i'steps SN])
        from steps[OF i'] obtain l r C \<tau> where l\<tau>: "ts i' = C\<langle>l \<cdot> \<tau>\<rangle>" and r\<tau>: "C\<langle>r \<cdot> \<tau>\<rangle> = ts (Suc i')"  and lr': "(l,r) = lr i'" and NF: "\<forall> u \<lhd> l \<cdot> \<tau>. u \<in> NFQ" by auto
        from lr' lr[OF i'] have mem: "(l,r) \<in> R'" by auto
        from only_useful_rules[OF NF, of r] SN_on_imp_wwf_rule[OF SN l\<tau> mem NF]
        have vars: "vars_term r \<subseteq> vars_term l" "is_Fun l" 
          unfolding wwf_rule_def by auto
        with lr'[symmetric] show ?thesis by auto
      next
        assume "wf_trs R'"
        from this[unfolded wf_trs_def, rule_format, of "fst (lr i')" "snd (lr i')"] lr[OF i']
        show ?thesis by force
      qed
    qed
  }
  with steps lr have steps: "\<And> i. i < n \<Longrightarrow> (ts i, ts (Suc i)) \<in> qrstep Q (R' \<inter> U)" unfolding qrstep_rule_conv[where R = "R' \<inter> U"] by blast
  show ?thesis unfolding rtrancl_fun_conv using first last steps by auto
qed

lemma usable_rules_proc: assumes fin: "finite_dpp (P,Pw,Q,Rs \<inter> U,Rw \<inter> U)"
  and P: "\<And> l r. (l,r) \<in> P \<union> Pw \<Longrightarrow> is_ur_closed_term_mv {l} r"
  and vars: "\<And> l r. (l,r) \<in> P \<union> Pw \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and U: "\<And> l r. (l,r) \<in> U \<Longrightarrow> is_ur_closed_term_mv (set (args l)) r"
  and R: "Rs \<union> Rw \<subseteq> R"
  shows "finite_dpp (P, Pw, Q, Rs, Rw)"
proof (rule ccontr)
  assume "\<not> ?thesis"
  then obtain s t \<sigma> where chain: "min_ichain (P,Pw,Q,Rs,Rw) s t \<sigma>" 
    unfolding finite_dpp_def by auto
  from chain have min: "minimal_cond Q (Rs \<union> Rw) s t \<sigma>" unfolding min_ichain.simps by auto     
  have "min_ichain (P,Pw,Q,Rs \<inter> U,Rw \<inter> U) s t \<sigma>"
    unfolding min_ichain.simps
  proof
    show "minimal_cond Q (Rs \<inter> U \<union> Rw \<inter> U) s t \<sigma>" 
      by (rule minimal_cond_mono[OF ci_subsetI min], auto)
  next
    from chain have chain: "ichain (P,Pw,Q,Rs,Rw) s t \<sigma>" unfolding min_ichain.simps by simp
    note chain = chain[unfolded ichain.simps]
    from chain have inP: "\<And> i. (s i, t i) \<in> P \<union> Pw" by auto
    from chain have NF: "\<And> i. s i \<cdot> \<sigma> i \<in> NFQ" by auto
    {
      fix i x
      assume "x \<in> vars_term (t i)"
      with vars[OF inP] have "x \<in> vars_term (s i)" by auto
      hence "Var x \<unlhd> s i" by auto
      hence "Var x \<cdot> \<sigma> i \<unlhd> s i \<cdot> \<sigma> i" by (rule supteq_stable)
      from NF_subterm[OF NF this]
      have "\<sigma> i x \<in> NFQ" by simp
    } 
    hence NFt: "\<And> i. \<sigma> i ` vars_term (t i) \<subseteq> NFQ" by auto
    from NF have NFs: "\<And> i. {s i} \<cdot>set \<sigma> i \<subseteq> NFQ" by auto
    note ur = P[OF inP]    
    let ?B = "qrstep Q (Rs \<union> Rw)"
    let ?B' = "qrstep Q (Rs \<inter> U \<union> Rw \<inter> U)"
    let ?R = "qrstep Q Rs"
    let ?R' = "qrstep Q (Rs \<inter> U)"
    {
      fix i n ts lr j
      assume first: "ts 0 = t i \<cdot> \<sigma> i" and steps: "\<And> i. i < n \<Longrightarrow> (ts i, ts (Suc i)) \<in> qrstep Q {lr i}"
      and lr: "\<And> i. i < n \<Longrightarrow> lr i \<in> Rs \<union> Rw"
      and j: "j < n"
      have "lr j \<in> U"
      proof (rule is_ur_closed_term_main[of n ts lr, OF steps _ NFt set_mp[OF R lr] first NFs U ur j])
        fix i'
        assume i': "i' < n"
        have i'steps: "(t i \<cdot> \<sigma> i, ts i') \<in> (qrstep Q (Rs \<union> Rw))^*" 
          unfolding qrsteps_rules_conv
        proof (intro exI conjI, rule first, rule refl)
          show "\<forall> i < i'. (ts i, ts (Suc i)) \<in> qrstep Q {lr i} \<and> lr i \<in> Rs \<union> Rw"
          proof (intro allI impI)
            fix i
            assume "i < i'" with i' have "i < n" by simp
            from steps[OF this] lr[OF this]
            show "(ts i, ts (Suc i)) \<in> qrstep Q {lr i} \<and> lr i \<in> Rs \<union> Rw" by simp
          qed
        qed
        have SN: "SN_on (qrstep Q (Rs \<union> Rw)) {ts i'}"
          by (rule steps_preserve_SN_on[OF i'steps], insert min, unfold minimal_cond_def, auto)
        from steps[OF i'] obtain l r C \<tau> where l\<tau>: "ts i' = C\<langle>l \<cdot> \<tau>\<rangle>" and r\<tau>: "C\<langle>r \<cdot> \<tau>\<rangle> = ts (Suc i')"  and lr': "(l,r) = lr i'" and NF: "\<forall> u \<lhd> l \<cdot> \<tau>. u \<in> NFQ" by auto
        from lr' lr[OF i'] have mem: "(l,r) \<in> Rs \<union> Rw" by auto
        from only_useful_rules[OF NF, of r] SN_on_imp_wwf_rule[OF SN l\<tau> mem NF]
        have vars: "vars_term r \<subseteq> vars_term l" "is_Fun l" 
          unfolding wwf_rule_def by auto
        with lr'[symmetric]
        show "vars_term (snd (lr i')) \<subseteq> vars_term (fst (lr i')) \<and> is_Fun (fst (lr i'))" by auto
      qed
    } note ur = this
    show "ichain (P,Pw,Q,Rs \<inter> U,Rw \<inter> U) s t \<sigma>"
      unfolding ichain.simps
    proof (intro conjI allI, rule inP)
      fix i
      from chain have steps: "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> ?B^*" by simp
      from steps[unfolded qrsteps_rules_conv]
      obtain n ts lr
        where first: "ts 0 = t i \<cdot> \<sigma> i" and last: "ts n = s (Suc i) \<cdot> \<sigma> (Suc i)"
        and steps: "\<And> i. i < n \<Longrightarrow> (ts i, ts (Suc i)) \<in> qrstep Q {lr i}"
        and lr: "\<And> i. i < n \<Longrightarrow> lr i \<in> Rs \<union> Rw" by blast
      from ur[of ts i n lr, OF first steps lr] lr have lr: "\<And> i. i < n \<Longrightarrow> lr i \<in> Rs \<inter> U \<union> Rw \<inter> U"
        by auto
      show "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> ?B'^*"
        unfolding qrsteps_rules_conv
        by (rule exI[of _ n], rule exI[of _ ts], rule exI[of _ lr], intro conjI, rule first, rule last, insert steps lr, auto)
    next
      fix i
      show "s i \<cdot> \<sigma> i \<in> NFQ" by (rule NF)
    next
      from chain have "(INFM i. (s i, t i) \<in> P) \<or>
        (INFM i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in>
        ?B^* O ?R O ?B^*)" (is "?l \<or> ?r") by simp
      thus "(INFM i. (s i, t i) \<in> P) \<or>
        (INFM i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in>
        ?B'^* O ?R' O ?B'^*)" (is "?l' \<or> ?r'")
      proof
        assume ?l thus ?thesis by auto
      next
        assume ?r
        have ?r' unfolding INFM_nat_le
        proof
          fix i
          from `?r`[unfolded INFM_nat_le, THEN spec[of _ i]] obtain j where j: "j \<ge> i"
            and steps: "(t j \<cdot> \<sigma> j, s (Suc j) \<cdot> \<sigma> (Suc j)) \<in> ?B^* O ?R O ?B^*" (is "(?t j,?s j) \<in> _") by blast
          from qrsteps_rules_conv'[OF steps]
          obtain n ts lr k
            where first: "ts 0 = t j \<cdot> \<sigma> j" and last: "ts n = s (Suc j) \<cdot> \<sigma> (Suc j)"
            and steps: "\<And> i . i < n \<Longrightarrow> (ts i, ts (Suc i)) \<in> qrstep Q {lr i}"
            and lr: "\<And> i. i < n \<Longrightarrow> lr i \<in> Rs \<union> Rw"
            and k: "k < n" and k': "lr k \<in> Rs" by blast
          from ur[of ts j n lr, OF first steps lr] have lr': "\<And> i. i < n \<Longrightarrow> lr i \<in> U" by auto
          with lr have lr: "\<And> i. i < n \<Longrightarrow> lr i \<in> Rs \<inter> U \<union> Rw \<inter> U" by auto
          from k' lr'[OF k] have k': "lr k \<in> Rs \<inter> U" by auto
          show "\<exists> j \<ge> i. (?t j, ?s j) \<in> ?B'^* O ?R' O ?B'^*"
            by (intro exI conjI, rule j, rule qrsteps_rules_conv''[of ts _ n _ _ lr, OF first last steps lr k k'])
        qed
        thus ?thesis ..
      qed
    qed
  qed
  with fin show False unfolding finite_dpp_def by blast
qed
end

  

end