(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Proof_Checker
imports Processors 
begin

type_synonym ('f, 'l, 'v) ruleLL   = "(('f, 'l) lab, 'v) rule"
type_synonym ('f, 'l, 'v) trsLL   = "(('f, 'l) lab, 'v) rules"
type_synonym ('f, 'l, 'v) termsLL = "(('f, 'l) lab, 'v) term list"
type_synonym ('f, 'l, 'v) rseqL   = "((('f, 'l) lab, 'v) rule \<times> (('f, 'l) lab, 'v) rseq) list"
type_synonym ('f, 'l, 'v) dppLL =
  "('f, 'l, 'v) trsLL \<times> ('f, 'l, 'v) trsLL \<times>
  ('f, 'l, 'v) termsLL \<times>
  ('f, 'l, 'v) trsLL \<times> ('f, 'l, 'v) trsLL"

datatype ('f, 'l, 'v) dp_proof =
  PisEmpty
| SubtermCriterionProc "('f, 'l) lab projL" "('f, 'l, 'v) rseqL"
    "('f, 'l, 'v) trsLL" "('f, 'l, 'v) dp_proof"
| RedPairProc "('f,'l)lab root_redtriple_impl + ('f, 'l) lab redtriple_impl" "('f, 'l, 'v) trsLL"  "('f, 'l, 'v) dp_proof"
| RedPairUrProc "('f, 'l) lab root_redtriple_impl + ('f, 'l) lab redtriple_impl"
    "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) dp_proof" 
| UsableRulesProc "('f, 'l, 'v) trsLL" "('f, 'l, 'v) dp_proof" 
| DepGraphProc "(('f, 'l, 'v) dp_proof option \<times> ('f, 'l, 'v) trsLL) list"
| MonoRedPairProc "('f, 'l) lab redtriple_impl"
    "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) dp_proof" 
| MonoRedPairUrProc "('f, 'l) lab redtriple_impl"
    "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) dp_proof"
| SizeChangeSubterm "((('f, 'l) lab, 'v) rule \<times> ((nat \<times> nat) list \<times> (nat \<times> nat) list)) list"
| SizeChangeRedPair "('f, 'l) lab redtriple_impl" "('f, 'l, 'v) trsLL"
    "((('f, 'l) lab, 'v) rule \<times> ((nat \<times> nat) list \<times> (nat \<times> nat) list)) list"
| UncurryProc "nat option" "(('f, 'l) lab, 'v) uncurry_info"
    "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) dp_proof"
| FccProc "('f, 'l) lab" "(('f, 'l) lab, 'v) ctxt list"
    "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) dp_proof"
| SplitProc
    "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trsLL" 
    "('f, 'l, 'v) dp_proof" "('f, 'l, 'v) dp_proof"
| SemlabProc
    "(('f, 'l) lab, 'v) sl_variant" "('f, 'l, 'v) trsLL"
    "(('f, 'l) lab, 'v) term list" "('f, 'l, 'v) trsLL"
    "('f, 'l, 'v) dp_proof"
| SwitchInnermostProc "(('f,'l)lab,'v)crit_pair_joins option" "('f, 'l, 'v)dp_proof"
| RewritingProc   "('f, 'l, 'v) trsLL" "('f,'l,'v)ruleLL" "('f,'l,'v)ruleLL" "('f,'l,'v)ruleLL" pos "('f, 'l, 'v) dp_proof"
| InstantiationProc   "('f,'l,'v)ruleLL" "('f,'l,'v)trsLL" "('f, 'l, 'v) dp_proof"
| ForwardInstantiationProc   "('f,'l,'v)ruleLL" "('f,'l,'v)trsLL" "('f,'l,'v)trsLL option" "('f, 'l, 'v) dp_proof"
| NarrowingProc   "('f,'l,'v)ruleLL" pos "('f,'l,'v)trsLL" "('f, 'l, 'v) dp_proof"
| AssumeFinite "('f, 'l, 'v) dppLL"

definition missing :: "string \<Rightarrow> shows \<Rightarrow> shows" where
  "missing s x \<equiv> ''the '' +#+ s +#+ '' '' +#+ x +@+ shows_string '' is missing''"

definition toomuch :: "string \<Rightarrow> shows \<Rightarrow> shows" where
  "toomuch s x \<equiv> ''superfluous '' +#+ s +#+ '' '' +#+ x +@+ x"

fun
  check_dpp_subsumes ::
    "('d, ('f::show, 'l::show) lab, 'v::show) dpp_ops
    \<Rightarrow> ('f, 'l, 'v) dppLL \<Rightarrow> 'd \<Rightarrow> shows check"
where
  "check_dpp_subsumes I (p', pw', q', r', rw') d = (do {
    let p  = dpp_ops.P I d;
    let pw = dpp_ops.Pw I d;
    let q  = dpp_ops.Q I d;
    let r  = dpp_ops.R I d;
    let rw = dpp_ops.Rw I d;
    let pb' = p' @ pw';
    let rb  = r  @ rw;
    let rb' = r' @ rw';
    check_subseteq p p'   <+? (\<lambda>r. toomuch ''pair'' (shows_rule r));
    check_subseteq pw pb' <+? (\<lambda>r. toomuch ''weak pair'' (shows_rule r));
    check_subseteq q q'   <+? (\<lambda>r. toomuch ''term'' (shows_term r));
    check_subseteq q' q   <+? (\<lambda>r. missing ''term'' (shows_term r));
    check_subseteq r r'   <+? (\<lambda>r. toomuch ''strict rule'' (shows_rule r));
    check_subseteq rb rb' <+? (\<lambda>r. toomuch ''strict/weak rule'' (shows_rule r));
    check_subseteq rb' rb <+? (\<lambda>r. missing ''strict/weak rule'' (shows_rule r));
    return ()
  }) <+? (\<lambda>s. ''finiteness of the problem'' +#+ shows_nl +@+ shows_dpp I d +@+ shows_nl
           +@+ ''may not be concluded from assuming finiteness of the problem'' +#+ shows_nl
           +@+ shows_dpp I (dpp_ops.mk I p' pw' q' r' rw') +@+ shows_nl
           +@+ s +@+ shows_nl)"

(*The first argument is a flag that specifies whether partial proofs (i.e, containing
assumptions) are accepted or not.*)
function
  check_dp_proof ::
    "bool \<Rightarrow> shows \<Rightarrow> ('dpp, ('f::{show,linorder},label_type) lab, string) dpp_ops \<Rightarrow> 'dpp \<Rightarrow>
    ('f, label_type, string) dp_proof \<Rightarrow>
    shows check"
where 
  "check_dp_proof assms i I dpp PisEmpty = debug (i []) ''P is empty''
     (if dpp_ops.P I dpp = [] \<and> (dpp_ops.Pw I dpp = [] \<or> dpp_ops.R I dpp = []) 
          then return () 
          else error (i +@+ '': P is not empty in the following DP-problem '' +#+
         shows_nl +@+ shows_dpp I dpp))"
| "check_dp_proof assms i I dpp (SubtermCriterionProc p rseq rP prf) =
    debug (i []) ''SubtermCriterionProc'' (do {
      let P  = dpp_ops.pairs I dpp;
      let Pr = list_diff P rP;
      dpp' \<leftarrow> subterm_criterion_proc I (create_proj p) (create_rseq_map rseq) Pr dpp
        <+? (\<lambda>s. i +@+ '': error when applying the subterm criterion to the DP problem''
          +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+
          ''and trying to remove the pairs '' +#+ shows_nl +@+ shows_rules Pr
          +@+ shows_nl +@+ s);
      check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
        <+? (\<lambda>s. i +@+ '': error below the subterm criterion'' +#+ shows_nl
          +@+ indent s)
    })"
| "check_dp_proof assms i I dpp (RedPairProc redp rP prf) = debug (i []) ''RedPairProc'' (do {
         let P = dpp_ops.pairs I dpp;
         let Pr = list_diff P rP;
         let Proc = (case redp of Inr rp \<Rightarrow> redtriple_proc I (get_redtriple rp)
                        | Inl rrp \<Rightarrow> root_redtriple_proc I (get_root_redtriple rrp));
         dpp' \<leftarrow> Proc Pr dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the reduction pair processor to remove from the DP problem''
             +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+
             '' the pairs '' +#+ shows_nl +@+ shows_rules Pr +@+ shows_nl +@+ s);
         check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the reduction pair processor'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_dp_proof assms i I dpp (UsableRulesProc U prf) = debug (i []) ''UsableRulesProc'' (do {
         dpp' \<leftarrow> usable_rules_proc I U dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the usable rules processor to restrict the DP problem''
             +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+
             '' to the usable rules '' +#+ shows_nl +@+ shows_rules U +@+ shows_nl +@+ s);
         check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the usable rules processor'' +#+ shows_nl
             +@+ indent s)
   })" 
|  "check_dp_proof assms i I dpp (MonoRedPairProc redp rP rR prf) = debug (i []) ''MonoRedPairProc'' (do {
         let P = dpp_ops.pairs I dpp;
         let Pr = list_diff P rP;
         let R = dpp_ops.rules I dpp;
         let Rr = list_diff R rR;
         dpp' \<leftarrow> mono_redpair_proc I (get_redtriple redp) Pr Rr dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the mono reduction pair processor to remove from the DP problem''
             +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+
             '' the pairs '' +#+ shows_nl +@+ shows_rules Pr +@+ shows_nl +@+ 
             '' and the rules '' +#+ shows_nl +@+ shows_rules Rr +@+ shows_nl +@+ s);
         check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the mono reduction pair processor'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_dp_proof assms i I dpp (DepGraphProc edpts) =
     debug (i []) ''DepGraphProc'' (do {
        pdpps \<leftarrow> dep_graph_proc I dpp edpts
         <+? (\<lambda>s. i +@+ '': error while trying to perform SCC-decomposition  on '' +#+
           shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+ s);
        check_allm_index (\<lambda> (prof,dpp') j. check_dp_proof assms (i +@+ ''.'' +#+ shows (Suc j)) I dpp' prof) pdpps
          <+? (\<lambda>s. i +@+ '': error below the dependency graph processor'' +#+ shows_nl
            +@+ indent s)
  })"
| "check_dp_proof assms i I dpp (RedPairUrProc redp rP ur prf) = debug (i []) ''RedPairUrProc'' (do {
         let P = dpp_ops.pairs I dpp;
         let Pr = list_diff P rP;
         let Proc = (case redp of Inr rp \<Rightarrow> ur_af_redtriple_proc I (get_redtriple rp)
                        | Inl rrp \<Rightarrow> ur_af_root_redtriple_proc I (get_root_redtriple rrp));
         dpp' \<leftarrow> Proc Pr ur dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the reduction pair processor with usable rules to remove from the DP problem''
             +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+
             '' the pairs '' +#+ shows_nl +@+ shows_rules Pr +@+ shows_nl +@+ s);
         check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the reduction pair processor'' +#+ shows_nl
             +@+ indent s)
   })" 
|  "check_dp_proof assms i I dpp (MonoRedPairUrProc redp rP rR ur prf) = debug (i []) ''MonoRedPairUrProc'' (do {
         let P = dpp_ops.pairs I dpp;
         let Pr = list_diff P rP;
         let R = dpp_ops.rules I dpp;
         let Rr = list_diff R rR;
         dpp' \<leftarrow> mono_ur_redpair_proc I (get_redtriple redp) Pr Rr ur dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the mono reduction pair processor with usable rules to remove from the DP problem''
             +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+
             '' the pairs '' +#+ shows_nl +@+ shows_rules Pr +@+ shows_nl +@+ 
             '' and the rules '' +#+ shows_nl +@+ shows_rules Rr +@+ shows_nl +@+ s);
         check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the mono reduction pair processor with usable rules'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_dp_proof assms i I dpp (UncurryProc mode u_info P' R' prf) = 
    debug (i []) ''UncurryProc'' (do {
      dpp' \<leftarrow> uncurry_proc_both I mode u_info P' R' dpp
        <+? (\<lambda>s. i +@+ '': error when applying the uncurrying processor on the DP problem''
          +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+ s);
      check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
        <+? (\<lambda>s. i +@+ '': error below the uncurrying processor'' +#+ shows_nl
          +@+ indent s)
    })"
| "check_dp_proof assms i I dpp (SizeChangeSubterm graphs) =
     debug (i []) ''SizeChangeSubterm'' (sct_subterm_proc I graphs dpp)"
| "check_dp_proof assms i I dpp (SizeChangeRedPair redp ur graphs) =
     debug (i []) ''SizeChangeRedPair'' (sct_ur_af_proc I (get_redtriple redp) graphs ur dpp)"
| "check_dp_proof assms i I dpp (FccProc f fcs P' R' prf) = debug (i []) ''FccProc'' (do {
    dpp' \<leftarrow> fcc_proc I f fcs P' [] R' dpp
      <+? (\<lambda>s. i +@+
        '': error when applying the flat context closure processor on the DP problem'' +#+
        shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+ s);
    check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
      <+? (\<lambda>s. i +@+ '': error below the flat context closure processor'' +#+ shows_nl
        +@+ indent s)
  })"
| "check_dp_proof assms i I dpp (SplitProc Prem Rrem prf1 prf2) = debug (i []) ''SplitProc'' (do {
    let (dpp1,dpp2) = split_proc I dpp Prem Rrem;
    check_dp_proof assms (i +@+ shows_string ''.1'') I dpp1 prf1
      <+? (\<lambda>s. i +@+ '': error below the split processor'' +#+ shows_nl
        +@+ indent s);
    check_dp_proof assms (i +@+ shows_string ''.2'') I dpp2 prf2
      <+? (\<lambda>s. i +@+ '': error below the split processor'' +#+ shows_nl
        +@+ indent s)
  })"
| "check_dp_proof assms i I dpp (SemlabProc sli lP lQ lR prf) = 
    debug (i []) ''SemlabProc'' (do {
      dpp' \<leftarrow> semlab_fin_proc I sli lP lQ lR dpp
        <+? (\<lambda>s. i +@+ '': error when applying the semlab processor on the DP problem''
          +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+ s);
      check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
        <+? (\<lambda>s. i +@+ '': error below the semlab processor'' +#+ shows_nl
          +@+ indent s)
    })"
| "check_dp_proof assms i I dpp (SwitchInnermostProc joins prf) = 
    debug (i []) ''SwitchInnermostProc'' (do {
      dpp' \<leftarrow> switch_innermost_proc I joins dpp
        <+? (\<lambda>s. i +@+ '': error when applying the processor to switch to innermost on the DP problem''
          +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+ s);
      check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
        <+? (\<lambda>s. i +@+ '': error below the switch to innermost processor'' +#+ shows_nl
          +@+ indent s)
    })"
| "check_dp_proof assms i I dpp (AssumeFinite dpp') =
    debug (i []) ''Finiteness Assumption'' (
      if assms
        then check_dpp_subsumes I dpp' dpp
          <+? (\<lambda>s. i +@+ '': error in finiteness assumption'' +#+ shows_nl +@+ s +@+ shows_nl)
        else error (i +@+ '': the proof contains a local finiteness assumption'' +#+ shows_nl)
    )"
| "check_dp_proof assms i I dpp (RewritingProc U st st' lr p prf) = debug (i []) ''RewritingProc'' (do {
         dpp' \<leftarrow> rewriting_proc I U st st' lr p dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the rewriting processor to rewrite the pair ''
             +#+ shows_nl +@+ shows_rule st +@+ shows_nl +@+
             '' to the pair '' +#+ shows_nl +@+ shows_rule st' +@+ shows_nl +@+ s);
         check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the rewriting processor'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_dp_proof assms i I dpp (NarrowingProc st p sts prf) = debug (i []) ''NarrowingProc'' (do {
         dpp' \<leftarrow> narrowing_proc I st p sts dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the processor processor to narrow the pair ''
             +#+ shows_nl +@+ shows_rule st +@+ shows_nl +@+
             '' to the pairs '' +#+ shows_nl +@+ shows_trs sts +@+ shows_nl +@+ s);
         check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the narrowing processor'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_dp_proof assms i I dpp (InstantiationProc st sts prf) = debug (i []) ''InstantiationProc'' (do {
         dpp' \<leftarrow> instantiation_proc I st sts dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the instantiation processor to instantiate the pair ''
             +#+ shows_nl +@+ shows_rule st +@+ shows_nl +@+
             '' to the pairs '' +#+ shows_nl +@+ shows_trs sts +@+ shows_nl +@+ s);
         check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the instantiation processor'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_dp_proof assms i I dpp (ForwardInstantiationProc st sts u_opt prf) = debug (i []) ''ForwardInstantiationProc'' (do {
         dpp' \<leftarrow> forward_instantiation_proc I st sts u_opt dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the forward_instantiation processor to instantiate the pair ''
             +#+ shows_nl +@+ shows_rule st +@+ shows_nl +@+
             '' to the pairs '' +#+ shows_nl +@+ shows_trs sts +@+ shows_nl +@+ s);
         check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the forward instantiation processor'' +#+ shows_nl
             +@+ indent s)
   })" 
  by pat_completeness auto

termination
proof (rule, unfold update_error_return)
  fix a i I dpp 
    and edpts::"(('f::{show,linorder},'l::{show,linorder}, string) dp_proof option
      \<times> ('f, 'l, string ) trsLL) list" 
    and pdpps pdpp i' prof dpp' j
  assume proc: "dep_graph_proc I dpp edpts = Inr pdpps"
    and mem: "pdpp \<in> set pdpps" and pdpp: "(prof, dpp') = pdpp"
  from proc mem[unfolded pdpp[symmetric]]
    have "Some prof \<in> set (map fst edpts)" unfolding dep_graph_proc_def by force
  then obtain P where mem: "(Some prof, P) \<in> set edpts" by auto
  let ?m = "prod_size (option_size size)
              (list_size
                (prod_size (Term.term.term_size (lab_size (\<lambda>x. 0) (\<lambda>x. 0)) (\<lambda>x. 0))
                  (Term.term.term_size (lab_size (\<lambda>x. 0) (\<lambda>x. 0)) (\<lambda>x. 0))))"
  have "size prof < ?m (Some prof, P)" by auto
  also have "... \<le> list_size ?m edpts" using mem by (induct edpts, auto)
  finally have "size prof < size (DepGraphProc edpts)" by simp
  thus "((a, i +@+ ''.'' +#+ shows (Suc j), I, dpp', prof), a, i, I, dpp, DepGraphProc edpts)
    \<in> measures [\<lambda>(a, i, I, dpp, prof). size prof]" by auto
qed auto

lemma isOK_check_dp_proof_AssumeFinite_False[simp]:
  "isOK (check_dp_proof False i I dpp (AssumeFinite dpp')) = False" by simp

lemma check_dpp_subsumesE[elim]:
  assumes 1: "dpp_spec I"
    and 2: "isOK (check_dpp_subsumes I (p, pw, q, r, rw) d)"
    and 3: "set (dpp_ops.P I d) \<subseteq> set p
    \<Longrightarrow> set (dpp_ops.Pw I d) \<subseteq> set p \<union> set pw
    \<Longrightarrow> set (dpp_ops.Q I d) = set q
    \<Longrightarrow> set (dpp_ops.R I d) \<subseteq> set r
    \<Longrightarrow> set (dpp_ops.R I d) \<union> set (dpp_ops.Rw I d) = set r \<union> set rw
    \<Longrightarrow> P"
  shows "P"
  by (rule 3, insert 1 2) (simp add: Let_def | blast)+

fun finite_dppL :: "('f, 'l, 'v) dppLL \<Rightarrow> bool" where
  "finite_dppL (p, pw, q, r, rw) = finite_dpp (set p, set pw, set q, set r, set rw)"

lemma check_dpp_subsumes_sound:
  assumes I: "dpp_spec I"
    and ok: "isOK (check_dpp_subsumes I dpp' dpp)"
    and fin: "finite_dppL dpp'"
  shows "finite_dpp (dpp_ops.dpp I dpp)"
proof -
  obtain p pw q r rw where dpp': "dpp' = (p, pw, q, r, rw)" by (cases dpp') force+
  let ?P   = "set (dpp_ops.P I dpp)"
  let ?Pw  = "set (dpp_ops.Pw I dpp)"
  let ?Q   = "set (dpp_ops.Q I dpp)"
  let ?R   = "set (dpp_ops.R I dpp)"
  let ?Rw  = "set (dpp_ops.Rw I dpp)"
  let ?P'  = "set p"
  let ?Pw' = "set pw"
  let ?Q'  = "set q"
  let ?R'  = "set r"
  let ?Rw' = "set rw"
  from I and ok
    have "?P \<subseteq> ?P'" and "?Pw \<subseteq> ?P' \<union> ?Pw'" and q: "?Q = ?Q'" and "?R \<subseteq> ?R'"
    and "?R \<union> ?Rw = ?R' \<union> ?Rw'"
    unfolding dpp' by blast+
  note 1 = this
  show ?thesis unfolding dpp_spec.dpp_sound[OF I]
    by (rule  finite_dpp_mono[OF fin[unfolded dpp', simplified], unfolded q[symmetric]])
       (insert 1, blast+)
qed

fun dp_assms :: "bool \<Rightarrow> ('f, 'l, 'v) dp_proof \<Rightarrow> ('f, 'l, 'v) dppLL list" where
  "dp_assms a PisEmpty = []"
| "dp_assms a (SubtermCriterionProc _ _ _ p) = dp_assms a p"
| "dp_assms a (RedPairProc _ _ p) = dp_assms a p"
| "dp_assms a (RedPairUrProc _ _ _ p) = dp_assms a p"
| "dp_assms a (UsableRulesProc _ p) = dp_assms a p"
| "dp_assms a (DepGraphProc ps) =
    concat (concat (map (map (dp_assms a) \<circ> option_to_list \<circ> fst) ps))"
| "dp_assms a (MonoRedPairProc _ _ _ p) = dp_assms a p"
| "dp_assms a (MonoRedPairUrProc _ _ _ _ p) = dp_assms a p"
| "dp_assms a (SizeChangeSubterm _) = []"
| "dp_assms a (SizeChangeRedPair _ _ _) = []"
| "dp_assms a (UncurryProc _ _ _ _ p) = dp_assms a p"
| "dp_assms a (FccProc _ _ _ _ p) = dp_assms a p"
| "dp_assms a (SplitProc _ _ p q) = dp_assms a p @ dp_assms a q"
| "dp_assms a (SemlabProc _ _ _ _ p) = dp_assms a p"
| "dp_assms a (SwitchInnermostProc _ p) = dp_assms a p"
| "dp_assms a (RewritingProc _ _ _ _ _ p) = dp_assms a p"
| "dp_assms a (NarrowingProc _ _ _ p) = dp_assms a p"
| "dp_assms a (InstantiationProc _ _ p) = dp_assms a p"
| "dp_assms a (ForwardInstantiationProc _ _ _ p) = dp_assms a p"
| "dp_assms a (AssumeFinite d) = (if a then [d] else [])"


lemma dp_assms_False[simp]: "dp_assms False prf = []"
  by (induct False "prf" rule: dp_assms.induct) simp_all

lemma check_dp_proof_with_dp_assms:
  assumes I: "dpp_spec I"
    and fin: "\<forall>p\<in>set (dp_assms a prf). finite_dppL p"
    and ok: "isOK (check_dp_proof a i I dpp prf)"
  shows "finite_dpp (dpp_ops.dpp I dpp)"
using ok I fin
proof (induct a i I dpp "prf" rule: check_dp_proof.induct)
  case (1 a i I dpp)
  interpret dpp_spec I by fact
  from 1(1) show ?case by (auto simp:
    QDP_Framework.finite_dpp_def dpp_spec_sound QDP_Framework.ichain.simps)
next
  case (2 a i I dpp p rseq rP prof)
  interpret dpp_spec I by fact
  let ?Pr = "list_diff (dpp_ops.pairs I dpp) rP"
  let ?i = "i +@+ shows_string ''.1''"
  from 2(2)
    obtain dpp' where proc: "subterm_criterion_proc I (create_proj p) (create_rseq_map rseq) ?Pr dpp = return dpp'"
    and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from 2 have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case
    by (rule sound_proc_impl[OF subterm_criterion_proc,
      OF proc 2(1)[OF HOL.refl HOL.refl _ rec 2(3) fin]])
      (unfold update_error_return, rule proc)
next
  case (3 a i I dpp rpc rP prof)
  interpret dpp_spec I by fact
  let ?Pr = "list_diff (dpp_ops.pairs I dpp) rP"
  let ?i = "i +@+ shows_string ''.1''"
  from 3 have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case 
  proof (cases rpc)
    case (Inr rp)
    with 3(2)
    obtain dpp' where proc: "redtriple_proc I (get_redtriple rp) ?Pr dpp = return dpp'"
      and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
      by (auto simp: Let_def)
    from 3(1)[OF HOL.refl HOL.refl HOL.refl _ rec 3(3) fin, unfolded Inr update_error_return sum.simps, OF proc]
    have fin: "finite_dpp (dpp_ops.dpp I dpp')" .
    show ?thesis      
      by (rule sound_proc_impl[OF redtriple_proc[OF get_redtriple] _ fin], 
        rule proc)
  next
    case (Inl rrp)
    with 3(2)
    obtain dpp' where proc: "root_redtriple_proc I (get_root_redtriple rrp) ?Pr dpp = return dpp'"
      and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
      by (auto simp: Let_def)
    from 3(1)[OF HOL.refl HOL.refl HOL.refl _ rec 3(3) fin, unfolded Inl update_error_return sum.simps, OF proc] 
    have fin: "finite_dpp (dpp_ops.dpp I dpp')" .
    show ?thesis
      by (rule sound_proc_impl[OF root_redtriple_proc[OF get_root_redtriple] _ fin], rule proc)
  qed
next
  case (4 a i I dpp ur prof)
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from 4(2) obtain dpp'
    where proc: "usable_rules_proc I ur dpp = return dpp'"
    and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from 4 have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case
    by (rule sound_proc_impl[OF usable_rules_proc, OF _ proc 4(1)[OF _ rec _ fin]], insert proc 4(3), auto)  
next
  case (5 a i I dpp rp rP rR prof)
  interpret dpp_spec I by fact
  let ?Pr = "list_diff (dpp_ops.pairs I dpp) rP"
  let ?Rr = "list_diff (dpp_ops.rules I dpp) rR"
  let ?i = "i +@+ shows_string ''.1''"
  from 5(2)
    obtain dpp' where proc: "mono_redpair_proc I (get_redtriple rp) ?Pr ?Rr dpp = return dpp'"
    and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from 5 have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case      
    by (rule sound_proc_impl[OF mono_redpair_proc[OF get_redtriple],
      OF proc 5(1)[OF HOL.refl HOL.refl HOL.refl HOL.refl _ rec 5(3) fin]],
      unfold update_error_return, rule proc)
next
  case (6 a i I dpp edpts)
  interpret dpp_spec I by fact
  from 6(2) obtain pdpps where proc: "dep_graph_proc I dpp edpts = return pdpps" 
    and rec: "\<And>j. j < length pdpps \<Longrightarrow>
      isOK ((\<lambda>(prof, dpp') j.
        check_dp_proof a (i +@+ ''.'' +#+ shows (Suc j)) I dpp' prof) (pdpps ! j) j)"
    by auto
  show ?case
  proof (rule dep_graph_proc[OF 6(3) proc])
    fix p dpp'
    assume mem: "(p, dpp') \<in> set pdpps"
    from this[unfolded set_conv_nth]
      obtain j where j: "j < length pdpps" and pair: "pdpps ! j = (p, dpp')" by auto
    from rec[OF j] have rec: "isOK (check_dp_proof a (i +@+ ''.'' +#+ shows (Suc j)) I dpp' p)"
      unfolding pair by simp
    from proc[unfolded dep_graph_proc_def, simplified]
      have pdpps: "pdpps =
      map (\<lambda>aP. (the (fst aP), dpp_ops.intersect_pairs I dpp (snd aP)))
          [aP\<leftarrow>edpts . \<exists>y. fst aP = Some y]" by simp
    from mem and 6(4) and proc
      have fin: "\<forall>a\<in>set (dp_assms a p). finite_dppL a" by (simp add: dep_graph_proc_def) force
    show "finite_dpp (dpp_ops.dpp I dpp')" 
      by (rule 6(1)[unfolded update_error_return, OF proc mem HOL.refl rec 6(3) fin])
  qed
next
  case (7 a i I dpp rpc rP ur prof)
  interpret dpp_spec I by fact
  let ?Pr = "list_diff (dpp_ops.pairs I dpp) rP"
  let ?i = "i +@+ shows_string ''.1''"
  from 7 have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case 
  proof (cases rpc)
    case (Inr rp)
    with 7(2) obtain dpp'
      where proc: "ur_af_redtriple_proc I (get_redtriple rp) ?Pr ur dpp = return dpp'"
      and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
      by (auto simp: Let_def)
    show ?thesis      
      by (rule sound_proc_impl[OF ur_af_redtriple_proc[OF get_redtriple],
        OF proc 7(1)[OF HOL.refl HOL.refl HOL.refl _ rec 7(3) fin]],
        unfold Inr update_error_return sum.simps, rule proc)
  next
    case (Inl rrp)
    with 7(2) obtain dpp'
      where proc: "ur_af_root_redtriple_proc I (get_root_redtriple rrp) ?Pr ur dpp = return dpp'"
      and rec: "isOK (check_dp_proof a ?i I dpp' prof)" 
      by (auto simp: Let_def)
    show ?thesis      
      by (rule sound_proc_impl[OF ur_af_root_redtriple_proc[OF get_root_redtriple],
        OF proc 7(1)[OF HOL.refl HOL.refl HOL.refl _ rec 7(3) fin]],
        unfold Inl update_error_return sum.simps, rule proc)
  qed
next
  case (8 a i I dpp rp rP rR ur prof)
  interpret dpp_spec I by fact
  let ?Pr = "list_diff (dpp_ops.pairs I dpp) rP"
  let ?Rr = "list_diff (dpp_ops.rules I dpp) rR"
  let ?i = "i +@+ shows_string ''.1''"
  from 8(2) obtain dpp'
    where proc: "mono_ur_redpair_proc I (get_redtriple rp) ?Pr ?Rr ur dpp = return dpp'"
    and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from 8 have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case
    by (rule sound_proc_impl[OF mono_ur_redpair_proc[OF get_redtriple],
      OF proc 8(1)[OF HOL.refl HOL.refl HOL.refl HOL.refl _ rec 8(3) fin]],
      unfold update_error_return, rule proc)
next
  case (9 a i I dpp mode u_info P' R' prof)
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from 9(2) obtain
    dpp' where proc: "uncurry_proc_both I mode u_info P' R'  dpp = return dpp'"
    and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from 9 have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case      
    by (rule sound_proc_impl[OF uncurry_proc_both[OF 9(3)], OF proc 9(1)],
      unfold update_error_return, rule proc, rule rec, rule 9(3), rule fin)
next
  case (10 a i I dpp graphs)
  interpret dpp_spec I by fact
  from 10(1) have proc: "isOK (sct_subterm_proc I graphs dpp)" by auto
  show ?case
    by (rule sct_subterm_proc_sound[OF proc])
next
  case (11 a i I dpp rp ur graphs)
  interpret dpp_spec I by fact
  from 11(1) have proc: "isOK (sct_ur_af_proc I (get_redtriple rp) graphs ur dpp)" by auto
  show ?case
    by (rule sct_ur_af_proc[OF get_redtriple, OF proc])
next
  case (12 a i I dpp f fcs P' R' prof)
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from 12(2) obtain dpp'
    where proc: "fcc_proc I f fcs P' [] R' dpp = return dpp'"
    and rec: "isOK (check_dp_proof a ?i I dpp' prof)" by auto
  from 12 have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  from 12(1)[OF _ rec 12(3) fin, simplified, OF proc] have "finite_dpp (dpp_ops.dpp I dpp')"
    by (simp add: dpp_spec_sound)
  from sound_proc_impl[OF fcc_proc, OF proc this] show ?case .
next
  case (13 a i I dpp P' R' prof1 prof2)
  interpret dpp_spec I by fact
  let ?i = "\<lambda>s. i +@+ shows_string s"
  let ?i1 = "?i ''.1''"
  let ?i2 = "?i ''.2''"
  obtain dpp1 dpp2 where s: "split_proc I dpp P' R' = (dpp1, dpp2)" by force
  note ss = s[symmetric]
  from 13(3) s have ok1: "isOK (check_dp_proof a ?i1 I dpp1 prof1)" 
    and ok2: "isOK (check_dp_proof a ?i2 I dpp2 prof2)" by auto
  from 13 have fin1: "\<forall>p\<in>set (dp_assms a prof1). finite_dppL p"
    and fin2: "\<forall>p\<in>set (dp_assms a prof2). finite_dppL p" by simp_all
  from 13(1)[OF ss HOL.refl ok1 13(4) fin1]
    have fin1': "finite_dpp (dpp_ops.dpp I dpp1)" by auto
  from 13(2)[OF ss HOL.refl _ ok2 13(4) fin2] ok1
    have fin2': "finite_dpp (dpp_ops.dpp I dpp2)" by auto
  show ?case
    by (rule split_proc[OF s fin1' fin2'])
next
  case (14 a i I dpp sli lP lQ lR prof)
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from 14(2) obtain dpp'
    where proc: "semlab_fin_proc I sli lP lQ lR dpp = return dpp'"
    and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from 14 have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case
    by (rule sound_proc_impl[OF semlab_fin_proc[OF 14(3)], OF proc 14(1)],
      unfold update_error_return, rule proc, rule rec, rule 14(3), rule fin)
next
  case (15 a i I dpp joins prof)
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from 15(2) obtain dpp'
    where proc: "switch_innermost_proc I joins dpp = return dpp'"
    and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from 15 have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case    
    by (rule sound_proc_impl[OF switch_innermost_proc[OF 15(3)] _ 15(1)], rule proc, unfold update_error_return, rule proc, rule rec, rule 15(3), rule fin)
next
  case (16 a i I dpp dpp')
  interpret dpp_spec I by fact
  have "a = True \<or> a = False" by simp
  thus ?case
  proof
    assume False: "a = False" thus ?thesis using 16(1) by simp
  next
    assume True: "a = True"
    show ?thesis
    proof (cases dpp')
      case fields
      show ?thesis using 16[unfolded True,simplified] and check_dpp_subsumes_sound by best
    qed
  qed
next
  case (17 a i I dpp U st st' lr p prof)
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from 17(2) obtain dpp'
    where proc: "rewriting_proc I U st st' lr p dpp = return dpp'"
    and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from 17 have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case    
    by (rule sound_proc_impl[OF rewriting_proc[OF 17(3)] _ 17(1)], rule proc, unfold update_error_return, rule proc, rule rec, rule 17(3), rule fin)  
next
  case (18 a i I dpp st p sts prof)
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from 18(2) obtain dpp'
    where proc: "narrowing_proc I st p sts dpp = return dpp'"
    and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from 18 have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case    
    by (rule sound_proc_impl[OF narrowing_proc[OF 18(3)] _ 18(1)], rule proc, unfold update_error_return, rule proc, rule rec, rule 18(3), rule fin)  
next
  case (19 a i I dpp st sts prof)
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from 19(2) obtain dpp'
    where proc: "instantiation_proc I st sts dpp = return dpp'"
    and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from 19 have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case    
    by (rule sound_proc_impl[OF instantiation_proc[OF 19(3)] _ 19(1)], rule proc, unfold update_error_return, rule proc, rule rec, rule 19(3), rule fin)  
next
  case (20 a i I dpp st sts u_opt prof)
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from 20(2) obtain dpp'
    where proc: "forward_instantiation_proc I st sts u_opt dpp = return dpp'"
    and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from 20 have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case    
    by (rule sound_proc_impl[OF forward_instantiation_proc[OF 20(3)] _ 20(1)], rule proc, unfold update_error_return, rule proc, rule rec, rule 20(3), rule fin)  
qed

(*the old soundness lemma is an instance of the new one*)
lemma check_dp_proof:
  assumes "dpp_spec I"
    and "isOK (check_dp_proof False i I dpp prf)"
  shows "finite_dpp (dpp_ops.dpp I dpp)"
  using check_dp_proof_with_dp_assms[OF assms(1) _ assms(2)] by simp

type_synonym ('f, 'l, 'v) qreltrsLL =
  "('f, 'l, 'v) termsLL \<times> ('f, 'l, 'v) trsLL \<times> ('f, 'l, 'v) trsLL"

datatype ('f, 'l, 'v) trs_proof =
  DPTrans "(('f, 'l) lab, 'v) rules" "('f, 'l, 'v) dp_proof"
| RuleRemoval "('f, 'l) lab redtriple_impl" "('f, 'l, 'v) trsLL" "('f, 'l, 'v)trs_proof"
| StringReversal "('f, 'l, 'v) trs_proof"
| Bounds "(('f, 'l) lab, 'v) bounds_info"
| Uncurry "(('f, 'l) lab, 'v) uncurry_info" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trs_proof"
| Semlab
    "(('f, 'l) lab, 'v) sl_variant" "(('f, 'l) lab, 'v) term list"
    "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trs_proof"
| RisEmpty
| Fcc "(('f, 'l) lab, 'v) ctxt list" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trs_proof"
| Split "('f, 'l, 'v) trsLL" "('f, 'l, 'v)trs_proof" "('f, 'l, 'v)trs_proof"
| SwitchInnermost "(('f,'l)lab, 'v)crit_pair_joins option" "('f, 'l, 'v)trs_proof" 
| AssumeSN "('f, 'l, 'v) qreltrsLL"

fun
  check_tp_subsumes ::
    "('tp, ('f::show, 'l::show) lab, 'v::show) tp_ops \<Rightarrow> ('f, 'l, 'v) qreltrsLL \<Rightarrow>
    'tp \<Rightarrow> shows check"
where
  "check_tp_subsumes I (q', r', rw') tp = (do {
    let q   = tp_ops.Q I tp;
    let r   = tp_ops.R I tp;
    let rw  = tp_ops.Rw I tp;
    let rb' = r' @ rw';
    check_subseteq q q'   <+? (\<lambda>r. toomuch ''term'' (shows_term r));
    check_subseteq q' q   <+? (\<lambda>r. missing ''term'' (shows_term r));
    check_subseteq r r'   <+? (\<lambda>r. toomuch ''rule'' (shows_rule r));
    check_subseteq rw rb' <+? (\<lambda>r. toomuch ''relative rule'' (shows_rule r));
    return ()
  }) <+? (\<lambda>s. ''termination of the problem'' +#+ shows_nl +@+ shows_tp I tp +@+ shows_nl
           +@+ ''may not be concluded from assuming termination of the problem'' +#+ shows_nl
           +@+ shows_tp I (tp_ops.mk I q' r' rw') +@+ shows_nl
           +@+ s +@+ shows_nl)"

fun
  check_trs_proof ::
    "bool \<Rightarrow> shows \<Rightarrow>
     ('tp, ('f::{show,linorder}, label_type) lab, string) tp_ops \<Rightarrow>
     ('dpp, ('f, label_type) lab, string) dpp_ops \<Rightarrow> 'tp \<Rightarrow>
     ('f, label_type, string) trs_proof \<Rightarrow>
     shows check" 
where
  "check_trs_proof a i I J tp RisEmpty = debug (i []) ''R is empty''
     (if tp_ops.R I tp = []
          then return () 
          else error (i +@+ '': R is not empty in the following termination-problem '' +#+
         shows_nl +@+ shows_tp I tp))"
|  "check_trs_proof a i I J tp (RuleRemoval redp rR prf) = debug (i []) ''Rule Removal'' (do {
         let R = tp_ops.rules I tp;
         let Rr = list_diff R rR;
         tp' \<leftarrow> rule_removal_tt I (get_redtriple redp) Rr tp 
           <+? (\<lambda>s. i +@+ '': error when applying the rule removal technique on ''
             +#+ shows_nl +@+ shows_tp I tp +@+ shows_nl +@+
             '' to remove the rules '' +#+ shows_nl +@+ shows_rules Rr +@+ shows_nl +@+ s);
         check_trs_proof a (i +@+ shows_string ''.1'') I J tp' prf
           <+? (\<lambda>s. i +@+ '': error below the rule removal technique'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_trs_proof a i I J tp (DPTrans P prf) = debug (i []) ''DP trans'' (do {
     dpp \<leftarrow> dependency_pairs_tt I J tp Sharp P
       <+? (\<lambda>s. i +@+ '': error when switching from the TRS'' +#+ shows_nl +@+ shows_tp I tp
         +@+ shows_nl +@+ ''to the initial DP problem with pairs '' +#+ shows_nl +@+ shows_rules P +@+ shows_nl +@+ s);
     check_dp_proof a (i +@+ shows_string ''.1'') J dpp prf
       <+? (\<lambda>s. i +@+ '': error below switch to dependency pairs'' +#+ shows_nl +@+ indent s)
   })" 
|  "check_trs_proof a i I J tp (StringReversal prf) = debug (i []) ''String Reversal'' (do {
         tp' \<leftarrow> string_reversal_tt I tp 
           <+? (\<lambda>s. i +@+ '': error when applying string reversal on ''
             +#+ shows_nl +@+ shows_tp I tp +@+ shows_nl +@+ s);
         check_trs_proof a (i +@+ shows_string ''.1'') I J tp' prf
           <+? (\<lambda>s. i +@+ '': error below the string reversal technique'' +#+ shows_nl
             +@+ indent s)
   })" 
|  "check_trs_proof a i I J tp (Semlab sli lQ lAll prf) = debug (i []) ''Semlab'' (do {
         tp' \<leftarrow> semlab_fin_tt I sli lQ lAll tp 
           <+? (\<lambda>s. i +@+ '': error when applying semantic labelling on ''
             +#+ shows_nl +@+ shows_tp I tp +@+ shows_nl +@+ s);
         check_trs_proof a (i +@+ shows_string ''.1'') I J tp' prf
           <+? (\<lambda>s. i +@+ '': error below the sem.lab technique'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_trs_proof a i I J tp (Bounds info) = debug (i []) ''Bounds'' (do {
     bounds_tt I info tp 
       <+? (\<lambda>s. i +@+ '': error when applying bounds on the termination problem '' +#+ shows_nl +@+ shows_tp I tp +@+ shows_nl +@+ s)})" 
| "check_trs_proof a i I J tp (Uncurry u_info rR prf) = debug (i []) ''Uncurry'' (do {
     tp' \<leftarrow> uncurry_tt I u_info rR tp
       <+? (\<lambda>s. i +@+ '': error when applying currying transformation'' +#+ shows_nl +@+ s);
     check_trs_proof a (i +@+ shows_string ''.1'') I J tp' prf
       <+? (\<lambda>s. i +@+ '': error below the uncurrying technique'' +#+ shows_nl +@+ indent s)
   })"
| "check_trs_proof a i I J tp (Fcc fcs R' prf) = debug (i []) ''Fcc'' (do {
    tp' \<leftarrow> fcc_tt I fcs R' tp
      <+? (\<lambda>s. i +@+ '': error when applying flat context closure'' +#+ shows_nl +@+ s);
    check_trs_proof a (i +@+ shows_string ''.1'') I J tp' prf
      <+? (\<lambda>s. i +@+ '': error below flat context closure'' +#+ shows_nl +@+ indent s)
  })"
| "check_trs_proof a i I J tp (Split Rrem prf1 prf2) = debug (i []) ''Split'' (do {
    let (tp1,tp2) = split_tt I tp Rrem;
    check_trs_proof a (i +@+ shows_string ''.1'') I J tp1 prf1
      <+? (\<lambda>s. i +@+ '': error below the splitting'' +#+ shows_nl
        +@+ indent s);
    check_trs_proof a (i +@+ shows_string ''.2'') I J tp2 prf2
      <+? (\<lambda>s. i +@+ '': error below the splitting'' +#+ shows_nl
        +@+ indent s)
  })"
|  "check_trs_proof a i I J tp (SwitchInnermost joins prf) = debug (i []) ''Switch Innermost'' (do {
         tp' \<leftarrow> switch_innermost_tt I joins tp 
           <+? (\<lambda>s. i +@+ '': error when switching to innermost on ''
             +#+ shows_nl +@+ shows_tp I tp +@+ shows_nl +@+ s);
         check_trs_proof a (i +@+ shows_string ''.1'') I J tp' prf
           <+? (\<lambda>s. i +@+ '': error below the switch to innermost'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_trs_proof a i I J tp (AssumeSN tp') =
    debug (i []) ''Termination Assumption'' (
      if a
        then check_tp_subsumes I tp' tp
          <+? (\<lambda>s. i +@+ '': error in termination assumption'' +#+ shows_nl +@+ s +@+ shows_nl)
        else error (i +@+ '': the proof contains a local termination assumption'' +#+ shows_nl)
    )"


lemma check_tp_subsumesE[elim]:
  assumes 1: "tp_spec I"
    and 2: "isOK (check_tp_subsumes I (q, r, rw) t)"
    and 3: "set (tp_ops.Q I t) = set q
    \<Longrightarrow> set (tp_ops.R I t) \<subseteq> set r
    \<Longrightarrow> set (tp_ops.Rw I t) \<subseteq> set r \<union> set rw
    \<Longrightarrow> P"
  shows "P"
  by (rule 3, insert 1 2) (simp add: Let_def | blast)+

fun SN_qrelL :: "('f, 'l, 'v) qreltrsLL \<Rightarrow> bool" where
  "SN_qrelL (q, r, rw) = SN_qrel (set q, set r, set rw)"

lemma check_tp_subsumes_sound:
  assumes I: "tp_spec I"
    and ok: "isOK (check_tp_subsumes I tp' tp)"
    and fin: "SN_qrelL tp'"
  shows "SN_qrel (tp_ops.qreltrs I tp)"
proof -
  obtain q r rw where tp': "tp' = (q, r, rw)" by (cases tp') force+
  let ?Q   = "set (tp_ops.Q I tp)"
  let ?R   = "set (tp_ops.R I tp)"
  let ?Rw  = "set (tp_ops.Rw I tp)"
  let ?Q'  = "set q"
  let ?R'  = "set r"
  let ?Rw' = "set rw"
  from I and ok
    have q: "?Q = ?Q'" and "?R \<subseteq> ?R'" and "?Rw \<subseteq> ?R' \<union> ?Rw'"
    unfolding tp' by blast+
  note 1 = this
  show ?thesis unfolding tp_spec.qreltrs_sound[OF I]
    by (rule  SN_qrel_mono[OF _ _ _ fin[unfolded tp', simplified], unfolded q[symmetric]])
       (insert 1, blast+)
qed

type_synonym ('f, 'l, 'v) prob = "('f, 'l, 'v) qreltrsLL + ('f, 'l, 'v) dppLL"

fun assms :: "bool \<Rightarrow> ('f, 'l, 'v) trs_proof \<Rightarrow> ('f, 'l, 'v) prob list" where
  "assms a (DPTrans _ p) = map Inr (dp_assms a p)"
| "assms a (RuleRemoval _ _ p) = assms a p"
| "assms a (StringReversal p) = assms a p"
| "assms a (Bounds _) = []"
| "assms a (Uncurry _ _ p) = assms a p"
| "assms a (Semlab _ _ _ p) = assms a p"
| "assms a RisEmpty = []"
| "assms a (Fcc _ _ p) = assms a p"
| "assms a (Split _ p q) = assms a p @ assms a q"
| "assms a (SwitchInnermost _ p) = assms a p"
| "assms a (AssumeSN tp) = (if a then [Inl tp] else [])"

lemma assms_False[simp]: "assms False prf = []"
  by (induct False "prf" rule: assms.induct) simp_all

fun holds :: "('f, 'l, 'v) prob \<Rightarrow> bool" where
  "holds (Inl p) = SN_qrelL p"
| "holds (Inr p) = finite_dppL p"

lemma check_trs_proof_with_assms_sound:
  assumes I: "tp_spec I" and J: "dpp_spec J"
    and fin: "\<forall>p\<in>set (assms a prf). holds p"
  and ok: "isOK (check_trs_proof a i I J tp prf)"
  shows "SN_qrel (tp_ops.qreltrs I tp)"
proof -
  interpret tp_spec I by fact
  from ok fin show ?thesis
  proof (induct "prf" arbitrary: tp i)
    case (RisEmpty i)
    thus ?case by (auto simp: SN_qrel_def rel_SN_def rel_step.simps)
  next
    case (RuleRemoval rp rR prof tp i)
    then obtain tp' i' Rr where tt: "rule_removal_tt I (get_redtriple rp) Rr tp = return tp'"
    and rec: "isOK (check_trs_proof a i' I J tp' prof)"
      by (auto simp: Let_def)
    from RuleRemoval have fin: "\<forall>p\<in>set (assms a prof). holds p" by simp
    show ?case
      by (rule sound_tt_impl[OF rule_removal_tt[OF get_redtriple],
        OF tt RuleRemoval(1)[OF rec fin]])
  next
    case (DPTrans P prof tp i)
    then obtain dpp i' shp where tt: "dependency_pairs_tt I J tp shp P = return dpp" 
      and rec: "isOK (check_dp_proof a i' J dpp prof)" by auto
    from DPTrans have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
    show ?case
      by (rule dependency_pairs_tt[OF I J tt
        check_dp_proof_with_dp_assms[OF J fin rec]])
  next
    case (StringReversal prof tp i)
    then obtain tp' i' where tt: "string_reversal_tt I tp = return tp'"
      and rec: "isOK (check_trs_proof a i' I J tp' prof)"
      by (auto simp: Let_def)
    from StringReversal have fin: "\<forall>p\<in>set (assms a prof). holds p" by simp
    show ?case
      by (rule sound_tt_impl[OF string_reversal_tt, OF tt StringReversal(1)[OF rec fin]])
  next
    case (Semlab sli lQ lAll prof tp i)
    then obtain tp' i' where tt: "semlab_fin_tt I sli lQ lAll tp = return tp'"
    and rec: "isOK(check_trs_proof a i' I J tp' prof)"
      by (auto simp: Let_def)
    from Semlab have fin: "\<forall>p\<in>set (assms a prof). holds p" by simp
    show ?case
      by (rule sound_tt_impl[OF semlab_fin_tt[OF I], OF tt Semlab(1)[OF rec fin]])
  next
    case (Bounds info)
    hence tt: "isOK (bounds_tt I info tp)"
      by (auto simp: Let_def)
    show ?case
      by (rule bounds_tt[OF I tt])
  next
    case (Uncurry mode info prof tp i)
    then obtain tp' i' where tt: "uncurry_tt I mode info tp = return tp'"
      and rec: "isOK (check_trs_proof a i' I J tp' prof)"
      by (auto simp: Let_def)
    from Uncurry have fin: "\<forall>p\<in>set (assms a prof). holds p" by simp
    show ?case
      by (rule sound_tt_impl[OF uncurry_tt, OF _ tt Uncurry(1)[OF rec fin]], unfold_locales)
  next
    case (Fcc fcs R' prof tp i)
    then obtain tp' i' where tt: "fcc_tt I fcs R' tp = return tp'"
      and rec: "isOK (check_trs_proof a i' I J tp' prof)" by auto
    from Fcc have fin: "\<forall>p\<in>set (assms a prof). holds p" by simp
    from Fcc(1)[OF rec fin] have "SN_qrel (tp_ops.qreltrs I tp')" .
    from sound_tt_impl[OF fcc_tt, OF tt this] show ?case .
  next
    case (Split R' prof1 prof2 tp i)
    let ?i = "\<lambda>s. i +@+ shows_string s"
    let ?i1 = "?i ''.1''"
    let ?i2 = "?i ''.2''"
    obtain tp1 tp2 where s: "split_tt I tp R' = (tp1,tp2)" by force
    from Split(3) s have ok1: "isOK (check_trs_proof a ?i1 I J tp1 prof1)" 
      and ok2: "isOK (check_trs_proof a ?i2 I J tp2 prof2)" by auto
    from Split have fin1: "\<forall>p\<in>set (assms a prof1). holds p"
      and fin2: "\<forall>p\<in>set (assms a prof2). holds p" by simp_all
    show ?case
      by (rule split_tt[OF s Split(1)[OF ok1 fin1] Split(2)[OF ok2 fin2]])
  next
    case (SwitchInnermost joins prof tp i)
    then obtain tp' i' where tt: "switch_innermost_tt I joins tp = return tp'"
      and rec: "isOK (check_trs_proof a i' I J tp' prof)"
      by (auto simp: Let_def)
    from SwitchInnermost have fin: "\<forall>p\<in>set (assms a prof). holds p" by simp
    show ?case
      by (rule sound_tt_impl[OF switch_innermost_tt[OF I], OF infinite_lab tt SwitchInnermost(1)[OF rec fin]]) 
  next
    case (AssumeSN tp' tp i)
    have "a = True \<or> a = False" by simp
    thus ?case
    proof
      assume False: "a = False" thus ?thesis using AssumeSN(1) by simp
    next
      assume True: "a = True"
      show ?thesis
      proof (cases tp')
        case fields
        show ?thesis using AssumeSN[unfolded True, simplified] and check_tp_subsumes_sound
          and `tp_spec I`
          by best
      qed
    qed
  qed
qed

lemma check_trs_proof_sound:
  assumes "tp_spec I" and "dpp_spec J"
    and "isOK (check_trs_proof False i I J tp prf)"
  shows "SN_qrel (tp_ops.qreltrs I tp)"
  using check_trs_proof_with_assms_sound[OF assms(1,2) _ assms(3)] by simp

datatype ('f,'l,'v)cr_proof = SN_WCR "(('f,'l)lab,'v)crit_pair_joins option" "('f, 'l, 'v) trs_proof" 

fun
  check_cr_proof ::
    "bool \<Rightarrow> shows \<Rightarrow>
     ('tp, ('f::{show,linorder}, label_type) lab, string) tp_ops \<Rightarrow>
     ('dpp, ('f, label_type) lab, string) dpp_ops \<Rightarrow> (('f,label_type)lab,string) rules \<Rightarrow>
     ('f, label_type, string) cr_proof \<Rightarrow>
     shows check" 
where
  "check_cr_proof a i I J R (SN_WCR joins prf) = debug (i []) ''SN_WCR'' (do {
         let tp = tp_ops.mk I [] R [];
         check_trs_proof a (i +@+ shows_string ''.1'') I J tp prf
           <+? (\<lambda>s. i +@+ '': error below strong normalization + wcr'' +#+ shows_nl
             +@+ indent s);
         check_critical_pairs R (critical_pairs_impl R R) joins
           <+? (\<lambda>s. i +@+ shows ''error when proving local confluence of '' +@+ shows_tp I tp)
   })" 

fun cr_assms :: "bool \<Rightarrow> ('f, 'l, 'v)cr_proof \<Rightarrow> ('f, 'l, 'v) prob list" where
  "cr_assms a (SN_WCR _ p) = assms a p"

lemma check_cr_proof_with_assms_sound: 
  assumes I: "tp_spec I" and J: "dpp_spec J"
    and fin: "\<forall>p\<in>set (cr_assms a prf). holds p"
  and ok: "isOK (check_cr_proof a i I J R prf)"
  shows "CR (rstep (set R))"
proof -
  interpret tp_spec I by fact
  from ok fin show ?thesis
  proof (induct "prf" arbitrary: i)
    case (SN_WCR joins prof i)
    from SN_WCR(1)
    have cp: "isOK(check_critical_pairs R (critical_pairs_impl R R) joins)"
      and ok: "isOK(check_trs_proof a (i +@+ shows_string ''.1'') I J (mk [] R []) prof)" by (auto simp: Let_def)
    from SN_WCR(2) have a: "\<forall> a \<in> set (assms a prof). holds a" by auto
    from check_critical_pairs[OF cp] have WCR: "WCR (rstep (set R))" .
    from check_trs_proof_with_assms_sound[OF I J a ok] have SN: "SN (rstep (set R))" by simp
    from Newman[OF SN WCR] show ?case .
  qed
qed

lemma cr_assms_False[simp]: "cr_assms False prf = []"
  by (induct False "prf" rule: cr_assms.induct) simp_all

lemma check_cr_proof_sound: 
  assumes I: "tp_spec I" and J: "dpp_spec J"
  and ok: "isOK (check_cr_proof False i I J R prf)"
  shows "CR (rstep (set R))"
  by (rule check_cr_proof_with_assms_sound[OF I J _ ok], simp)


type_synonym ('f,'v)subsumption_proof = "(('f,'v)equation \<times> ('f,'v)eseq) list"

datatype ('f,'l,'v)completion_proof = SN_WCR_Eq "(('f,'l)lab,'v)crit_pair_joins option" "('f, 'l, 'v) trs_proof" "(('f,'l)lab,'v)subsumption_proof" "(('f,'l)lab,'v)subsumption_proof option"

fun
  check_completion_proof ::
    "bool \<Rightarrow> shows \<Rightarrow>
     ('tp, ('f::{show,linorder}, label_type) lab, string) tp_ops \<Rightarrow>
     ('dpp, ('f, label_type) lab, string) dpp_ops \<Rightarrow> (('f,label_type)lab,string) equation list \<Rightarrow> (('f,label_type)lab,string) rules \<Rightarrow>
     ('f, label_type, string) completion_proof \<Rightarrow>
     shows check" 
where
  "check_completion_proof a i I J E R (SN_WCR_Eq joins prf conv1 conv2) = debug (i []) ''SN_WCR_Eq'' (do {
         let tp = tp_ops.mk I [] R [];
         check_trs_proof a (i +@+ shows_string ''.1'') I J tp prf
           <+? (\<lambda>s. i +@+ '': error below strong normalization + wcr'' +#+ shows_nl
             +@+ indent s);
         check_subsumption_guided R E conv1
           <+? (\<lambda> s. i +@+ shows '': error when showing that rewrite relation can be simulated by equations'' +@+ indent s);
         check_subsumption E R conv2
           <+? (\<lambda> s. i +@+ shows '': error when showing that equations can be simulated by rewrite system'' +@+ indent s);
         check_critical_pairs R (critical_pairs_impl R R) joins
           <+? (\<lambda>s. i +@+ shows '': error when proving local confluence of '' +@+ shows_tp I tp +@+ shows_nl +@+ indent s)
   })" 

fun completion_assms :: "bool \<Rightarrow> ('f, 'l, 'v)completion_proof \<Rightarrow> ('f, 'l, 'v) prob list" where
  "completion_assms a (SN_WCR_Eq _ p _ _) = assms a p"

lemma check_completion_proof_with_assms_sound: 
  assumes I: "tp_spec I" and J: "dpp_spec J"
    and fin: "\<forall>p\<in>set (completion_assms a prf). holds p"
  and ok: "isOK (check_completion_proof a i I J E R prf)"
  shows "completed_rewrite_system (set E) (set R)"
proof -
  interpret tp_spec I by fact
  from ok fin show ?thesis
  proof (induct "prf" arbitrary: i)
    case (SN_WCR_Eq joins prof c1 c2 i)
    from SN_WCR_Eq(1)
    have cp: "isOK(check_critical_pairs R (critical_pairs_impl R R) joins)"
      and ok: "isOK(check_trs_proof a (i +@+ shows_string ''.1'') I J (mk [] R []) prof)" 
      and ok1: "isOK(check_subsumption_guided R E c1)"
      and ok2: "isOK(check_subsumption E R c2)" by (auto simp: Let_def)
    from SN_WCR_Eq(2) have a: "\<forall> a \<in> set (assms a prof). holds a" by auto
    from check_critical_pairs[OF cp] have WCR: "WCR (rstep (set R))" .
    from check_trs_proof_with_assms_sound[OF I J a ok] have SN: "SN (rstep (set R))" by simp     
    show ?case 
      by (rule completion_via_WCR_SN_simulation[OF check_subsumption[OF ok2] check_subsumption_guided[OF ok1]  WCR SN])
  qed
qed

lemma completion_assms_False[simp]: "completion_assms False prf = []"
  by (induct False "prf" rule: completion_assms.induct) simp_all

lemma check_completion_proof_sound: 
  assumes I: "tp_spec I" and J: "dpp_spec J"
  and ok: "isOK (check_completion_proof False i I J E R prf)"
  shows "completed_rewrite_system (set E) (set R)"
  by (rule check_completion_proof_with_assms_sound[OF I J _ ok], simp)


datatype ('f, 'v) strategy =
  No_Strategy
| Innermost
| Innermost_Q "('f, 'v) term list"

fun strategy_to_Q :: "('f, 'v) strategy \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) term list" where
  "strategy_to_Q No_Strategy _ = []"
| "strategy_to_Q Innermost R = remdups (map fst R)"
| "strategy_to_Q (Innermost_Q Q) _ = Q"

datatype ('f, 'v) input_problem =
  TRS "('f, 'v)strategy" "('f, 'v) rules" "('f, 'v) rules"
| DP "('f, 'v) rules" "('f, 'v) rules" "('f, 'v)strategy" "('f, 'v) rules" "('f, 'v) rules"
| EQS_TRS "('f, 'v) equation list" "('f, 'v) rules" 

datatype ('f, 'l, 'v) cert_problem = 
  TRS_Proof "(('f, 'l) lab, 'v) strategy"
            "(('f, 'l) lab, 'v) rules"
            "(('f, 'l) lab, 'v) rules option"
            "('f, 'l, 'v) trs_proof"
| DP_Proof  "(('f, 'l) lab, 'v) rules"
  "(('f, 'l) lab, 'v) rules"
  "(('f, 'l) lab, 'v) strategy"
  "(('f, 'l) lab, 'v) rules"
  "(('f, 'l) lab, 'v) rules"
  "('f, 'l, 'v) dp_proof"
| DP_Disproof  "(('f, 'l) lab, 'v) rules"
  "(('f, 'l) lab, 'v) strategy"
  "(('f, 'l) lab, 'v) rules"
  "(('f, 'l) lab, 'v) dp_disproof"
| TRS_Disproof "(('f, 'l) lab, 'v) strategy"
               "(('f, 'l) lab, 'v) rules"
               "(('f, 'l) lab, 'v) trs_disproof"
| Rel_TRS_Disproof "(('f, 'l) lab, 'v) strategy"
                   "(('f, 'l) lab, 'v) rules"
                   "(('f, 'l) lab, 'v) rules"
                   "(('f, 'l) lab, 'v) reltrs_disproof"
| TRS_Confluence "(('f,'l)lab,'v) rules" "('f,'l,'v) cr_proof"
| Completion_Proof "(('f,'l)lab,'v) equation list" "(('f,'l)lab,'v)rules" "('f,'l,'v) completion_proof"

fun input :: "('f, 'l, 'v) cert_problem \<Rightarrow> (('f, 'l) lab, 'v) input_problem" where
  "input (TRS_Proof Q R (Some S) _) = TRS Q R S"
| "input (TRS_Proof Q R None _) = TRS Q R []"
| "input (TRS_Disproof Q R _) = TRS Q R []"
| "input (Rel_TRS_Disproof Q R S _) = TRS Q R S"
| "input (DP_Proof P Pw Q R Rw _) = DP P Pw Q R Rw"
| "input (DP_Disproof P Q R _) = DP P [] Q [] R"
| "input (TRS_Confluence R _) = TRS No_Strategy R []"
| "input (Completion_Proof E R _) = EQS_TRS E R"

fun desired_property :: "('f, 'l, 'v) cert_problem \<Rightarrow> bool" where
  "desired_property (TRS_Proof Q R (Some S) _ ) = SN_qrel (set (strategy_to_Q Q R), set R, set S)"
| "desired_property (TRS_Proof Q R None _ ) = SN (qrstep (set (strategy_to_Q Q R)) (set R))"
| "desired_property (TRS_Disproof Q R _) = (\<not> SN (qrstep (set (strategy_to_Q Q R)) (set R)))"
| "desired_property (Rel_TRS_Disproof Q R S _) = (\<not> SN_qrel (set (strategy_to_Q Q R), set R, set S))"
| "desired_property (DP_Proof P Pw Q R Rw _) = finite_dpp (set P, set Pw, set (strategy_to_Q Q (R @ Rw)), set R, set Rw)"
| "desired_property (DP_Disproof P Q R _) = infinite_dpp (set P, set (strategy_to_Q Q R), set R)"
| "desired_property (TRS_Confluence R _) = CR (rstep (set R))"
| "desired_property (Completion_Proof E R _) = completed_rewrite_system (set E) (set R)"

fun xml_tag :: "('f, 'l, 'v) cert_problem \<Rightarrow> string" where
  "xml_tag (TRS_Proof _ _ (Some _) _ ) = ''<relativeTerminationProof>''"
| "xml_tag (TRS_Proof _ _ None _ ) = ''<trsTerminationProof>''"
| "xml_tag (TRS_Disproof _ _ _) = ''<trsNonterminationProof>''"
| "xml_tag (Rel_TRS_Disproof _ _ _ _) = ''<relativeNonterminationProof>''"
| "xml_tag (DP_Proof _ _ _ _ _ _) = ''<dpProof>''"
| "xml_tag (DP_Disproof _ _ _ _) = ''<dpNonterminationProof>''"
| "xml_tag (TRS_Confluence _ _) = ''<crProof>''"
| "xml_tag (Completion_Proof _ _ _) = ''<completionProof>''"

fun cert_assms :: "bool \<Rightarrow> ('f, 'l, 'v) cert_problem \<Rightarrow> ('f, 'l, 'v) prob list" where
  "cert_assms a (TRS_Proof _ _ _ p) = assms a p"
| "cert_assms a (TRS_Confluence _ p) = cr_assms a p"
| "cert_assms a (Completion_Proof _ _ p) = completion_assms a p"
| "cert_assms a (DP_Proof _ _ _ _ _ p) = map Inr (dp_assms a p)" 
| "cert_assms _ _ = []"

lemma cert_assms_False[simp]: "cert_assms False p = []" by (cases p) simp_all

datatype cert_result = Certified string | Unsupported string | Error string

fun
  certify_cert_problem ::
    "bool \<Rightarrow> ('tp, ('f::{show,linorder}, label_type) lab, string) tp_ops \<Rightarrow>
     ('dpp, ('f, label_type) lab, string) dpp_ops \<Rightarrow> 
     ('f, label_type, string) cert_problem \<Rightarrow> cert_result"
where
  "certify_cert_problem a I J (TRS_Proof Q R None prf) =
    (case check_trs_proof a (shows_string ''1'') I J (tp_ops.mk I (strategy_to_Q Q R) R []) prf of
      Inl err \<Rightarrow> Error ((''termination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (TRS_Proof Q R None prf)))"
| "certify_cert_problem a I J (TRS_Proof Q R (Some S) prf) = 
    (case check_trs_proof a (shows_string ''1'') I J (tp_ops.mk I (strategy_to_Q Q R) R S) prf of
      Inl err \<Rightarrow> Error ((''relative termination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (TRS_Proof Q R (Some S) prf)))"
| "certify_cert_problem a I J (TRS_Disproof Q R prf) =
    (case check_trs_disproof (shows_string ''1'') I J (tp_ops.mk I (strategy_to_Q Q R) R []) prf of
      Inl err \<Rightarrow> Error ((''nontermination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (TRS_Disproof Q R prf)))"
| "certify_cert_problem a I J (Rel_TRS_Disproof Q R S prf) =
    (case check_reltrs_disproof (shows_string ''1'') I J (tp_ops.mk I (strategy_to_Q Q R) R S) prf of
      Inl err \<Rightarrow>
        Error ((''relative nontermination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Rel_TRS_Disproof Q R S prf)))"
| "certify_cert_problem a I J (DP_Proof P Pw Q R Rw prf) =
    (case check_dp_proof a (shows_string ''1'') J (dpp_ops.mk J P Pw (strategy_to_Q Q (R @ Rw)) R Rw) prf of
      Inl err \<Rightarrow> Error ((''finiteness proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (DP_Proof P Pw Q R Rw prf)))"
| "certify_cert_problem a I J (DP_Disproof P Q R prf) =
    (case check_dp_disproof (shows_string ''1'') J (dpp_ops.mk J P [] (strategy_to_Q Q R) [] R) prf of
      Inl err \<Rightarrow> Error ((''infiniteness proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (DP_Disproof P Q R prf)))"
| "certify_cert_problem a I J (TRS_Confluence R prf) =
    (case check_cr_proof a (shows_string ''1'') I J R prf of
      Inl err \<Rightarrow> Error ((''confluence proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (TRS_Confluence R prf)))"
| "certify_cert_problem a I J (Completion_Proof E R prf) =
    (case check_completion_proof a (shows_string ''1'') I J E R prf of
      Inl err \<Rightarrow> Error ((''completion proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Completion_Proof E R prf)))"

lemma certify_cert_problem_with_assms_sound:
  assumes I: "tp_spec I" and J: "dpp_spec J"
    and answer: "certify_cert_problem a I J cp = Certified info"
  shows "((\<forall>p\<in>set (cert_assms a cp). holds p) \<longrightarrow> desired_property cp) \<and> xml_tag cp = info"
proof -
  let ?i = "shows_string ''1''"
  interpret tp_spec I by (rule I)  
  show ?thesis
  proof (cases cp)
    case (TRS_Proof Q R sOpt prof)
    let ?tp = "tp_ops.mk I (strategy_to_Q Q R) R"
    show ?thesis
    proof (cases sOpt)
      case None
      from answer have ok: "isOK (check_trs_proof a ?i I J (?tp []) prof) \<and> xml_tag cp = info" 
        unfolding TRS_Proof None
        by (cases "check_trs_proof a ?i I J (?tp []) prof") auto
      from check_trs_proof_with_assms_sound[OF I J _ ok[THEN conjunct1]]
       and ok show ?thesis unfolding TRS_Proof None by simp
    next
      case (Some S)
      from answer have ok: "isOK (check_trs_proof a ?i I J (?tp S) prof) \<and> xml_tag cp = info"
        unfolding TRS_Proof Some
        by (cases "check_trs_proof a ?i I J (?tp S) prof", auto)
      from check_trs_proof_with_assms_sound[OF I J _ ok[THEN conjunct1]] 
        and ok show ?thesis unfolding TRS_Proof Some by simp
    qed
  next
    case (TRS_Disproof Q R "prf")
    let ?tp = "tp_ops.mk I (strategy_to_Q Q R) R []"
    from answer have ok: "isOK (check_trs_disproof ?i I J ?tp prf) \<and> xml_tag cp = info"
      unfolding TRS_Disproof
      by (cases "check_trs_disproof ?i I J ?tp prf") auto
    from check_trs_disproof_sound[OF I J ok[THEN conjunct1]]
      ok show ?thesis unfolding mk_simps TRS_Disproof by simp
  next
    case (Rel_TRS_Disproof Q R S "prf")
    let ?tp = "tp_ops.mk I (strategy_to_Q Q R) R S"
    from answer have ok: "isOK (check_reltrs_disproof ?i I J ?tp prf) \<and> xml_tag cp = info"
      unfolding Rel_TRS_Disproof
      by (cases "check_reltrs_disproof ?i I J ?tp prf") auto
    from check_reltrs_disproof_sound[OF I J ok[THEN conjunct1]]
      ok show ?thesis unfolding mk_simps Rel_TRS_Disproof by simp
  next
    case (DP_Proof P Pw Q R Rw "prf")
    let ?dp = "dpp_ops.mk J P Pw (strategy_to_Q Q (R @ Rw)) R Rw"
    from answer have ok: "isOK (check_dp_proof a ?i J ?dp prf) \<and> xml_tag cp = info"
      unfolding DP_Proof
      by (cases "check_dp_proof a ?i J ?dp prf", auto)
    interpret J: dpp_spec J by (rule J)
    show ?thesis
    proof (intro conjI impI)
      assume ass: "\<forall> p \<in> set (cert_assms a cp). holds p"
      show "desired_property cp"
        unfolding DP_Proof desired_property.simps
      proof (rule check_dp_proof_with_dp_assms[OF J _ ok[THEN conjunct1], unfolded 
            J.dpp_spec_sound J.mk_simps cert_assms.simps])
        from ass
        show "\<forall> p \<in> set (dp_assms a prf). finite_dppL p"
          unfolding DP_Proof cert_assms.simps
          by (induct "dp_assms a prf", auto)
      qed
    qed (insert ok, auto)
  next
    case (DP_Disproof P Q R "prf")
    let ?dp = "dpp_ops.mk J P [] (strategy_to_Q Q R) [] R"
    from answer have ok: "isOK (check_dp_disproof ?i J ?dp prf) \<and> xml_tag cp = info"
      unfolding DP_Disproof
      by (cases "check_dp_disproof ?i J ?dp prf", auto)
    interpret J: dpp_spec J by (rule J)
    show ?thesis
    proof (intro conjI impI)
      show "desired_property cp"
        unfolding DP_Disproof desired_property.simps
        using check_dp_disproof[OF J ok[THEN conjunct1], unfolded 
            J.dpp_spec_sound J.mk_simps]
        by auto
    qed (insert ok, auto)
  next
    case (TRS_Confluence R "prf")
    from answer have ok: "isOK (check_cr_proof a ?i I J R prf) \<and> xml_tag cp = info"
      unfolding TRS_Confluence
      by (cases "check_cr_proof a ?i I J R prf", auto)
    note res = check_cr_proof_with_assms_sound[OF I J _ ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding TRS_Confluence by auto
  next
    case (Completion_Proof E R "prf")
    from answer have ok: "isOK (check_completion_proof a ?i I J E R prf) \<and> xml_tag cp = info"
      unfolding Completion_Proof
      by (cases "check_completion_proof a ?i I J E R prf", auto)
    note res = check_completion_proof_with_assms_sound[OF I J _ ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding Completion_Proof by auto
  qed
qed

lemma certify_cert_problem_sound:
  assumes I: "tp_spec I" and J: "dpp_spec J"
    and answer: "certify_cert_problem False I J cp = Certified info"
  shows "desired_property cp \<and> xml_tag cp = info"
  using certify_cert_problem_with_assms_sound[OF I J answer] by simp

end
