(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Usable_Rules
imports
  Edg
  Reduction_Pair
  Substitution
  Trs_Impl 
begin


definition qrewrite :: "('f,'v)terms \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term list"
  where "qrewrite Q R t \<equiv> filter (\<lambda> s. (t,s) \<in> qrstep Q (set R)) (rewrite R t)" 

lemma qrewrite_sound: assumes "t \<in> set (qrewrite Q R s)"
  shows "(s,t) \<in> qrstep Q (set R)" using assms unfolding qrewrite_def by auto

lemma qrewrite_complete:
  assumes SN: "SN_on (qrstep Q (set R)) {s}"
  and qstep: "(s,t) \<in> qrstep Q (set R)"
  shows "t \<in> set (qrewrite Q R s)" 
  using qstep
proof 
  fix C \<sigma> l r
  assume Q: "\<forall> u \<lhd> l \<cdot> \<sigma>. u \<in> NF_terms Q" and lr: "(l,r) \<in> set R" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>"
    and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
  from only_useful_rules[OF Q] SN_on_imp_wwf_rule[OF SN s lr Q, unfolded wwf_rule_def] have vars: "vars_term r \<subseteq> vars_term l" by auto
  have step: "(s,t) \<in> rstep_r_p_s (set R) (l,r) (hole_pos C) \<sigma>"
    unfolding rstep_r_p_s_def Let_def s t
    by (simp add: lr)
  have step: "t \<in> set (rewrite R s)"
    by (rule rewrite_complete[OF step vars])
  thus ?thesis using qstep unfolding qrewrite_def by auto
qed

fun i_trans_check :: "('f,'v)terms \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f \<times> nat)set \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
where "i_trans_check Q R ur us (Fun f ts) = ((f,length ts) \<in> us \<and> \<not> (\<exists> (l,r) \<in> set R - ur. match_cap_below l (set R) (Fun f ts)))"

declare i_trans_check.simps[simp del]

function i_trans :: "(('f,'v)terms \<times> ('f,'v)rules \<times> ('f,'v)trs \<times> ('f \<times> nat)set \<times> ('f \<times> 'f \<times> nat)) \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term"
where "i_trans _ (Var x) = (Var x)" 
   |  "i_trans (Q, R, ur, us, cn) (Fun f ts) = (if SN_on (qrstep Q (set R)) {Fun f ts} then (
         if i_trans_check Q R ur us (Fun f ts) then Fun f (map (i_trans (Q, R, ur, us, cn)) ts) 
                   else (comb cn) (Fun f (map (i_trans (Q, R, ur, us, cn)) ts) # map (i_trans (Q, R, ur, us, cn)) (qrewrite Q R (Fun f ts)))
  )  else Fun f ts)"
by (pat_completeness, auto)


termination i_trans
proof -
  {
    fix Q :: "('f,'v)terms"
    and R :: "('f,'v)rules" 
    and ur :: "('f,'v)trs" 
    and us :: "('f \<times> nat)set" 
    and cn :: "'f \<times> 'f \<times> nat"
    have "\<forall> t. i_trans_dom ((Q,R,ur,us,cn),t)"
    proof (clarify)
      fix t
      let ?Ruc = "(Q,R,ur,us,cn)"
      have "SN (restrict_SN_supt (qrstep Q (set R)))" 
        by (rule ctxt_closed_imp_SN_restrict_SN_supt[OF ctxt_closed_qrstep])
      hence "wf ((restrict_SN_supt (qrstep Q (set R)))\<inverse>)" by (simp add: SN_iff_wf)
      thus "i_trans_dom (?Ruc, t)"
      proof (induct)
	fix t
	assume ind: "\<And>s. (s, t) \<in> (restrict_SN_supt (qrstep Q (set R)))\<inverse> \<Longrightarrow> i_trans_dom (?Ruc, s)"
	show "i_trans_dom (?Ruc, t)"
	proof 
	  fix s
	  assume step: "i_trans_rel s (?Ruc, t)"
	  thus "i_trans_dom s"
	  proof
	    fix Q' :: "('f,'v)terms" and R' :: "('f,'v)rules" and ur' :: "('f,'v)trs" and us' :: "('f \<times> nat)set" and cn' :: "'f \<times> 'f \<times> nat" 
	      and  f :: 'f and ts :: "('f,'v)term list" and t' :: "('f,'v)term"
	    let ?Ruc' = "(Q',R',ur',us',cn')"
	    assume s: "s = (?Ruc', t')"
	    assume t: "(?Ruc,t) = (?Ruc', Fun f ts)"
	    assume t': "t' \<in> set ts"
	    assume SN: "SN_on (qrstep Q' (set R')) {Fun f ts}"
	    from t' t SN have "t \<rhd> t'" and "SN_on (qrstep Q (set R)) {t}" by auto
	    hence "(t',t) \<in> (restrict_SN_supt (qrstep Q (set R)))\<inverse>" unfolding restrict_SN_supt_def supt_def restrict_SN_def by auto
	    with ind s t show "i_trans_dom s" by auto
	  next
	    fix Q' :: "('f,'v)terms" and R' :: "('f,'v)rules" and ur' :: "('f,'v)trs" and us' :: "('f \<times> nat)set" and cn' :: "'f \<times> 'f \<times> nat" 
	      and  f :: 'f and ts :: "('f,'v)term list" and t' :: "('f,'v)term"
	    let ?Ruc' = "(Q',R',ur',us',cn')"
	    assume s: "s = (?Ruc', t')"
	    assume t: "(?Ruc,t) = (?Ruc', Fun f ts)"
	    assume t': "t' \<in> set ts"
	    assume SN: "SN_on (qrstep Q' (set R')) {Fun f ts}"
	    from t' t SN have "t \<rhd> t'" and "SN_on (qrstep Q (set R)) {t}" by auto
	    hence "(t',t) \<in> (restrict_SN_supt (qrstep Q (set R)))\<inverse>" unfolding restrict_SN_supt_def supt_def restrict_SN_def by auto
	    with ind s t show "i_trans_dom s" by auto
	  next
	    fix Q' :: "('f,'v)terms" and R' :: "('f,'v)rules" and ur' :: "('f,'v)trs" and us' :: "('f \<times> nat)set" and cn' :: "'f \<times> 'f \<times> nat" 
	      and  f :: 'f and ts :: "('f,'v)term list" and t' :: "('f,'v)term"
	    let ?Ruc' = "(Q',R',ur',us',cn')"
	    assume s: "s = (?Ruc', t')"
	    assume t: "(?Ruc,t) = (?Ruc', Fun f ts)"
	    assume t': "t' \<in> set (qrewrite Q' R' (Fun f ts))"
	    assume SN: "SN_on (qrstep Q' (set R')) {Fun f ts}"
	    from t t' qrewrite_sound have step: "(t, t') \<in> qrstep Q (set R)" by blast
	    from SN t have "SN_on (qrstep Q (set R)) {t}" by auto
	    with step have "(t',t) \<in> (restrict_SN_supt (qrstep Q (set R)))\<inverse>" unfolding restrict_SN_supt_def
	      restrict_SN_def by auto
	    with ind s t show "i_trans_dom s" by auto
	  qed
	qed
      qed
    qed
  }
  thus ?thesis by force
qed


abbreviation
  i_trans_subst ::
    "(('f,'v)terms \<times> ('f, 'v) rules \<times> ('f, 'v) trs \<times> ('f \<times> nat) set \<times> ('f \<times> 'f \<times> nat)) \<Rightarrow> ('f, 'v) subst \<Rightarrow>
    ('f, 'v) subst"
where
  "i_trans_subst params \<sigma> \<equiv> (\<lambda>x. i_trans params (\<sigma> x))"

declare size_simps[simp]
fun ur_closed_term_af :: "('f,'v)term \<Rightarrow> ('f,'v)terms \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f \<times> nat)set \<Rightarrow> 'f af \<Rightarrow> bool"
where "ur_closed_term_af (Var x) _ _ _ _ _ = True"
   |  "ur_closed_term_af (Fun f ts) Q R ur us \<pi> = ((f,length ts) \<in> us \<and> (\<forall> i < length ts. \<pi> (f, (length ts)) i \<longrightarrow> ur_closed_term_af (ts ! i) Q R ur us \<pi>) \<and> 
  (\<forall> (l,r) \<in> R. match_cap_below l R (Fun f ts) \<longrightarrow> (l,r) \<in> ur) 
)"
declare size_simps[simp del]

abbreviation ur_closed_af :: "('f,'v)trs \<Rightarrow> ('f \<times> nat)set \<Rightarrow> ('f,'v)terms \<Rightarrow> ('f,'v)trs \<Rightarrow> 'f af \<Rightarrow> bool"
where "ur_closed_af ur us Q R \<pi> \<equiv> 
   (\<forall> (l,r) \<in> ur. ur_closed_term_af r Q R ur us \<pi>)"

notation Ground_Context.equiv_class ("[_]")

context ce_af_redpair
begin

lemma i_transI:
  assumes ur: "ur_closed_term_af t Q (set R) ur us \<pi>"
    and sn: "SN_on (qrstep Q (set R)) {t \<cdot> \<sigma>}"
  shows "(t \<cdot> (i_trans_subst (Q, R, ur, us, com) \<sigma>), i_trans (Q, R, ur, us, com) (t \<cdot> \<sigma>)) \<in> NS^*"
using assms
proof (induct t, simp)
  case (Fun f ts)  
  hence sClosed: "\<And> i. i < length ts \<and> \<pi> (f, (length ts)) i \<Longrightarrow> ur_closed_term_af (ts ! i) Q (set R) ur us \<pi>" by auto
  let ?Ruc = "(Q, R, ur, us, com)"
  from `SN_on (qrstep Q (set R)) {Fun f ts \<cdot> \<sigma>}` have sns: "\<And> s. s \<in> set ts \<Longrightarrow> SN_on (qrstep Q (set R)) {s \<cdot> \<sigma>}"
    by (simp add: SN_imp_SN_arg_gen[OF ctxt_closed_qrstep])
  have "i_trans ?Ruc (Fun f ts \<cdot> \<sigma>)
       = i_trans ?Ruc (Fun f (map (\<lambda> t. t \<cdot> \<sigma>) ts))" (is "?l = _") by auto
  also have "\<dots> = Fun f (map (i_trans ?Ruc) (map (\<lambda> t. t \<cdot> \<sigma>) ts))"
  proof -
    from `ur_closed_term_af (Fun f ts) Q (set R) ur us \<pi>`
    have inst: "\<forall> (l,r) \<in> set R. match_cap_below l (set R) (Fun f ts) \<longrightarrow> (l,r) \<in> ur" and us: "(f, length ts) \<in> us" by auto
    have "i_trans_check Q R ur us (Fun f ts \<cdot> \<sigma>)" (is ?check)
    proof (rule ccontr)
      assume "\<not> ?check"
      from this us obtain l r \<delta> where nUsable: "(l,r) \<in> set R - ur" and elem: "l \<cdot> \<delta> \<in> [GCFun f (map (\<lambda> t. cap (set R) (t \<cdot> \<sigma>)) ts)]" 
	(is "_ \<in> [?hs]")
	by (auto simp:i_trans_check.simps Ground_Context.match_def)
      have "\<And> t. [cap (set R) (t \<cdot> \<sigma>)] \<subseteq> [cap (set R) t]" by (rule cap_instance_subset)
      hence "[?hs] \<subseteq> [GCFun f (map (cap (set R)) ts)]" (is "_ \<subseteq> [?h]") by auto
      with elem have "l \<cdot> \<delta> \<in> [?h]" by auto
      hence "Ground_Context.match ?h l" unfolding Ground_Context.match_def by auto
      with inst nUsable show False by auto
    qed
    with Fun show ?thesis by auto
  qed
  finally have eq1: "?l = Fun f (map (\<lambda> t. i_trans ?Ruc (t \<cdot> \<sigma>)) ts)" (is "_ = ?r") by simp
  let ?tis = "\<lambda> t. t \<cdot> (i_trans_subst ?Ruc \<sigma>)"
  let ?its = "\<lambda> t. i_trans ?Ruc (t \<cdot> \<sigma>)"
  have "(Fun f (map ?tis ts), Fun f (map ?its ts)) \<in> NS^*" 
  proof -
    have ind: "\<forall> i < length ts. \<pi> (f, (length ts)) i \<longrightarrow> (?tis (ts ! i), ?its (ts ! i)) \<in> NS^*"
    proof (intro allI impI)
      fix i
      assume ilen: "i < length ts" and ipi: "\<pi> (f, (length ts)) i"
      with sns have "ts ! i \<in> set ts" and "SN_on (qrstep Q (set R)) {ts ! i \<cdot> \<sigma>}" by auto
      with Fun sClosed ilen ipi show "(?tis (ts ! i), ?its (ts ! i)) \<in> NS^*" by auto
    qed
    from wm_NS have cc: "one_step_ctxt_closed (NS^*)" using rtrancl_one_step_ctxt_closed  by auto
    show ?thesis 
    proof (rule af_steps_imp_orient, rule trans_rtrancl, rule refl_rtrancl, rule cc,simp)
      show "\<forall>i<length (map ?tis ts). \<pi> (f, (length ts)) i \<longrightarrow> (map ?tis ts ! i, map ?its ts ! i) \<in> NS^*"
	using ind by auto
    next
      show "\<forall>bef s t aft. length (map ?tis ts) = Suc (length bef + length aft) \<longrightarrow> \<pi> (f, (length ts)) (length bef) \<or> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> NS\<^sup>*"
	using af_compat unfolding af_compatible_def by force
    qed
  qed
  also have "Fun f (map ?tis ts) = Fun f ts \<cdot> i_trans_subst ?Ruc \<sigma>" by auto
  finally show ?case using eq1 by auto
qed

end

lemma i_transII:
  assumes sn: "SN_on (qrstep Q (set R)) {t \<cdot> \<sigma>}" 
  shows "(i_trans (Q, R, ur, us, cn) (t \<cdot> \<sigma>), t \<cdot> i_trans_subst (Q, R, ur, us, cn) \<sigma>) \<in> (rstep (ce_trs cn))^*
  \<and> (funas_term t \<subseteq> us \<or> (i_trans (Q, R, ur, us, cn) (t \<cdot> \<sigma>), t \<cdot> i_trans_subst (Q, R, ur, us, cn) \<sigma>) \<in> (rstep (ce_trs cn))^+) " 
  (is "(i_trans ?Ruc _,_) \<in> ?rel \<and> (_ \<or> _ \<in> ?Rel)  ")
using sn
proof (induct t)
  case (Var x) thus ?case by auto
next
  case (Fun f ts)
  from `SN_on (qrstep Q (set R)) {Fun f ts \<cdot> \<sigma>}` have sns: "\<And> s. s \<in> set ts \<Longrightarrow> SN_on (qrstep Q (set R)) {s \<cdot> \<sigma>}"
    by (simp add: SN_imp_SN_arg_gen[OF ctxt_closed_qrstep])
  let ?ts = "map (\<lambda> t. i_trans ?Ruc (t \<cdot> \<sigma>)) ts"
  let ?ss = "map (\<lambda> t. t \<cdot> i_trans_subst ?Ruc \<sigma>) ts"
  have len: "length ?ts = length ?ss" and other: "\<forall> i. i< length ?ts \<longrightarrow> (?ts ! i, ?ss ! i) \<in> ?rel" using sns Fun  by auto
  from len other have "(Fun f ?ts, Fun f ?ss) \<in> ?rel" using rstep_rtrancl_all_ctxt_closed[where r = "ce_trs cn"] unfolding all_ctxt_closed_def by blast 
  hence nearlyDone: "(Fun f ?ts, Fun f ts \<cdot> i_trans_subst ?Ruc \<sigma>) \<in> ?rel" (is "(_,?ti) \<in> _") by auto
  show ?case
  proof (cases "i_trans_check Q R ur us (Fun f ts \<cdot> \<sigma>)")
    case False
    with Fun have id: "i_trans ?Ruc (Fun f ts \<cdot> \<sigma>) = 
       (comb cn) (Fun f ?ts # map (i_trans ?Ruc) (qrewrite Q R (Fun f ts \<cdot> \<sigma>)))" (is "?it = (comb cn) ?list") 
      by (simp add: o_def) 
    have "Fun f ?ts \<in> set ?list" by simp
    hence steps: "((comb cn) ?list, Fun f ?ts) \<in> (rstep (ce_trs cn))^+" (is "_ \<in> ?r^+") by (rule ce_trs_sound) 
    from nearlyDone steps have steps: "(i_trans ?Ruc (Fun f ts \<cdot> \<sigma>), ?ti) \<in> ?r^+" unfolding id by auto
    show ?thesis using steps by auto
  next
    case True
    with Fun nearlyDone have part1: "(i_trans ?Ruc (Fun f ts \<cdot> \<sigma>), Fun f ts \<cdot> i_trans_subst ?Ruc \<sigma>) \<in> ?rel"
      by (simp add: o_def) 
    let ?fus = "funas_term (Fun f ts) \<subseteq> us"
    show ?thesis 
    proof (intro conjI, rule part1, unfold disj_commute[of ?fus], rule disjCI)
      assume not: "\<not> ?fus"
      with Fun True have id: "i_trans ?Ruc (Fun f ts \<cdot> \<sigma>) = Fun f ?ts" by simp 
      from True have fus: "(f,length ts) \<in> us" by (simp add: i_trans_check.simps)
      with not have "\<exists> t. t \<in> set ts \<and> \<not> funas_term t \<subseteq> us" by auto
      then obtain t where t: "t \<in> set ts" and tus: "\<not> funas_term t \<subseteq> us" by blast
      from t obtain i where i: "i < length ?ts" and ti: "t = ts ! i" by (simp only: set_conv_nth, auto)
      from Fun(1)[OF t _] sns t ti tus have "(i_trans ?Ruc (ts ! i \<cdot> \<sigma>), ts ! i \<cdot> i_trans_subst ?Ruc \<sigma>) \<in> ?Rel" by auto   
      with i have ind: "(?ts ! i, ?ss ! i) \<in> ?Rel" by auto
      have "(Fun f ?ts, Fun f ?ss) \<in> ?Rel" 
        by (rule rtrancl_trancl_into_trancl[OF len other i ind one_step_ctxt_closed_rstep])
      with id show "(i_trans ?Ruc (Fun f ts \<cdot> \<sigma>), Fun f ts \<cdot> i_trans_subst ?Ruc \<sigma>) \<in> ?Rel" by auto
    qed
  qed
qed

context ce_redpair
begin

definition c :: "'a" where "c \<equiv> SOME c. True"
definition d :: "'a" where "d \<equiv> SOME d. True"
definition n :: nat where "n \<equiv> SOME n. \<forall> m. m \<ge> n \<longrightarrow> (\<forall> c d.  ce_trs (c,d,m) \<subseteq> NS)"

lemma NS_ce_compat_n: "m \<ge> n \<Longrightarrow> ce_trs (c,d,m) \<subseteq> NS"
proof -
  assume m: "n \<le> m"
  from mp[OF spec[OF someI_ex[OF NS_ce_compat]] m[unfolded n_def]] show "ce_trs (c,d,m) \<subseteq> NS" by blast
qed

lemma ce_orient: "m \<ge> n \<Longrightarrow> (rstep (ce_trs (c,d,m)))^* \<subseteq> NS^*"
  by (rule rtrancl_mono, rule trs_inclusion, rule wm_NS, rule subst_NS, rule NS_ce_compat_n, simp)

abbreviation mono where "mono m \<equiv> one_step_ctxt_closed S \<and> ce_trs (c,d,m) \<subseteq> S"
abbreviation monoNS where "monoNS \<equiv> (NS \<union> S)^*"
abbreviation monoS where "monoS \<equiv> monoNS O S O monoNS"

abbreviation Snwf where "Snwf us \<equiv> S \<union> {lr | lr. \<not> funas_term (fst lr) \<subseteq> us}"

lemma mono_ce: assumes m: "mono m"   
  shows "rstep (ce_trs (c,d,m)) \<subseteq> S"
  by (rule trs_inclusion, insert m, auto simp: subst_S)

definition mode_cond where "mode_cond mmode m \<equiv> mmode \<longrightarrow> mono m"
definition mode_left where "mode_left mmode \<equiv> if mmode then NS \<union> S else NS"
definition mode_NS where "mode_NS mmode \<equiv> if mmode then monoNS else NS^*"

lemma mode_cond_left: "mode_cond mmode m \<Longrightarrow> one_step_ctxt_closed (mode_left mmode) \<and> subst_closed (mode_left mmode)"
  by (cases mmode,
    auto simp: mode_cond_def mode_left_def subst_S subst_NS wm_NS intro: subst_closed_union mono_NSS)

lemma mode_left_NS: "mode_left mode \<subseteq> mode_NS mode"
  unfolding mode_left_def mode_NS_def
  by (cases mode, auto)

lemma mode_NS_mode[simp]: "NS^* O mode_NS mode = mode_NS mode"
proof (cases mode)
  case False thus ?thesis unfolding mode_left_def mode_NS_def by auto
next
  case True 
  hence id: "mode = True" by simp
  show ?thesis unfolding mode_left_def mode_NS_def id
    by (simp, regexp)
qed

lemma mode_mode_NS[simp]: "mode_NS mode O NS^* = mode_NS mode"
proof (cases mode)
  case False thus ?thesis unfolding mode_left_def mode_NS_def by auto
next
  case True 
  hence id: "mode = True" by simp
  show ?thesis unfolding mode_left_def mode_NS_def id
    by (simp, regexp)
qed
end

lemma (in ce_af_redpair) i_transIII: assumes 
  mode: "mode_cond mode m" and
  rule: "lr \<in> set R" and
  m: "m \<ge> n" and
  sn: "SN_on (qrstep Q (set R)) {Fun f ts}" and
  step: "(Fun f ts,s) \<in> rqrstep Q {lr}" and
  ur: "i_trans_check Q R ur us (Fun f ts)" and
  ur_closed: "ur_closed_af ur us Q (set R) \<pi>" and
  orient: "ur \<subseteq> mode_left mode"
  shows "((i_trans (Q,R,ur,us,(c,d,m)) (Fun f ts), i_trans (Q,R,ur,us,(c,d,m)) s) \<in> mode_NS mode) \<and> 
    (mode \<and> lr \<in> Snwf us \<longrightarrow>  (i_trans (Q,R,ur,us,(c,d,m)) (Fun f ts), i_trans (Q,R,ur,us,(c,d,m)) s) \<in> monoS) \<and> lr \<in> ur"
proof -
  let ?cn = "(c,d,m)"
  let ?Ruc = "(Q,R,ur,us,?cn)"
  let ?I = "i_trans ?Ruc"
  obtain l r where lr: "lr = (l,r)" by (cases lr, auto)
  note rule = rule[unfolded lr]
  from step[unfolded lr rqrstep_def] obtain \<sigma> where left: "l \<cdot> \<sigma> = Fun f ts" and right: "r \<cdot> \<sigma> = s" by auto
  let ?Is = "i_trans_subst ?Ruc \<sigma>"
  {
    assume "is_Var l"
    with rule obtain x where "(Var x, r) \<in> set R" by (cases l, auto)
    hence "\<not> (SN_on (qrstep Q (set R)) {Fun f ts})" 
      by (rule left_Var_imp_not_SN_qrstep)
    hence False using sn by auto
  }
  then obtain g ss where l: "l = Fun g ss" by (cases l, auto)
  from ur have nmatch: "\<forall> (l,r) \<in> set R - ur. \<not> match_cap_below l (set R) (Fun f ts)" by (simp only: i_trans_check.simps, blast)
  from l left have fg: "f = g" and tsss: "ts = map (\<lambda> s. s \<cdot> \<sigma>) ss" (is "_ = ?ts") and nvar: "\<not> (is_Var l)" by auto
  have "\<forall> i. ss ! i \<cdot> \<sigma> \<in> [cap (set R) (ss ! i \<cdot> \<sigma>)]"
  proof 
    fix i
    show "ss ! i \<cdot> \<sigma> \<in> [cap (set R) (ss ! i \<cdot> \<sigma>)]"
      by (rule cap_sound[where s = "ss ! i \<cdot> \<sigma>" and t = "ss ! i \<cdot> \<sigma>" and \<sigma> = Var], auto)
  qed
  hence "match_cap_below (Fun f ss) (set R) (Fun f (map (\<lambda> s. s \<cdot> \<sigma>) ss))"
    by (simp add: Ground_Context.match_def, blast)
  with l fg tsss rule nmatch have urule: "(l,r) \<in> ur" by blast
  with left[symmetric] right[symmetric] subst_closureI[where \<sigma> = \<sigma>] ctxt_closureI[where C = Hole] 
  have urStep: "(Fun f ts, s) \<in> rstep ur" by force
  from urule ur_closed have fr: "ur_closed_term_af r Q (set R) ur us \<pi>" by auto
  from left sn have snl: "SN_on(qrstep Q (set R)) {l \<cdot> \<sigma>}" by auto
  from i_transII[OF snl] left have oneB: "funas_term (fst lr) \<subseteq> us \<or> (?I (Fun f ts), l \<cdot> ?Is) \<in> (rstep (ce_trs ?cn))^+" unfolding lr by auto
  from i_transII[OF snl] ce_orient[OF m] have "(?I (l \<cdot> \<sigma>), l \<cdot> ?Is) \<in> NS^*" by blast
  with left have one: "(?I (Fun f ts), l \<cdot> ?Is) \<in> NS^*" by auto
  from step rule lr have "(Fun f ts, s) \<in> rqrstep Q (set R)" unfolding rqrstep_def by auto
  hence "(Fun f ts, s) \<in> qrstep Q (set R)" unfolding qrstep_iff_rqrstep_or_nrqrstep by auto
  with right sn step_preserves_SN_on[of _ _ "qrstep Q (set R)"] have sns: "SN_on (qrstep Q (set R)) {r \<cdot> \<sigma>}" by auto
  have "(r \<cdot> ?Is , ?I (r \<cdot> \<sigma>)) \<in> NS^*"  by (rule i_transI[OF fr sns])
  with right have stepRight: "(r \<cdot> ?Is, ?I s) \<in> NS^*" by auto
  from urule have urstep: "(l \<cdot> ?Is, r \<cdot> ?Is) \<in> rstep ur" by auto
  with trs_inclusion[OF _ _ orient] mode_cond_left[OF mode] have lstep: "(l \<cdot> ?Is, r \<cdot> ?Is) \<in> mode_left mode" by auto
  with mode_left_NS have rstep: "(l \<cdot> ?Is, r \<cdot> ?Is) \<in> mode_NS mode" by auto
  with stepRight have two: "(l \<cdot> ?Is, ?I s) \<in> mode_NS mode O NS^*" by blast
  from one two have "(?I (Fun f ts), ?I s) \<in> NS^* O mode_NS mode O NS^*" by blast
  hence part1: "(?I (Fun f ts), ?I s) \<in> mode_NS mode" by simp
  show ?thesis
  proof (intro conjI impI, rule part1, unfold lr, erule conjE)
    show "(l,r) \<in> ur" by (rule urule)
  next
    assume mono: mode and S: "(l,r) \<in> Snwf us"
    from S have "(l,r) \<in> S \<or> \<not> funas_term l \<subseteq> us" by auto
    hence "(?I (Fun f ts), r \<cdot> ?Is) \<in> monoS"
    proof
      assume "(l,r) \<in> S"
      with subst_S have "(l \<cdot> ?Is, r \<cdot> ?Is) \<in> S" unfolding subst_closed_def by auto
      moreover have "(?I (Fun f ts), l \<cdot> ?Is) \<in> monoNS" using one rtrancl_mono by blast
      ultimately show ?thesis by auto
    next
      assume "\<not> funas_term l \<subseteq> us"
      with oneB have steps: "(?I (Fun f ts), l \<cdot> ?Is) \<in> (rstep (ce_trs ?cn))^+" unfolding lr by auto
      from mode[unfolded mode_cond_def] `mode` 
        trancl_mono[OF steps mono_ce]
      have "(?I (Fun f ts), l \<cdot> ?Is) \<in> S^+" by auto
      moreover have "(l \<cdot> ?Is, r \<cdot> ?Is) \<in> CS (NS \<union> S)" using S urule orient `mode` unfolding mode_left_def by auto
      hence "(l \<cdot> ?Is, r \<cdot> ?Is) \<in> NS \<union> S" using subst_NS subst_S unfolding cs_union_distr subst_closed_def by force
      ultimately have "(?I (Fun f ts), r \<cdot> ?Is) \<in> S^+ O (NS \<union> S)" by auto
      moreover have "S^+ O (NS \<union> S) \<subseteq> monoS" by regexp
      ultimately show ?thesis by blast
    qed
    with stepRight have "(?I (Fun f ts), ?I s) \<in> monoS O NS^*" by blast
    moreover have "monoS O NS^* \<subseteq> monoS" by regexp
    ultimately show "(?I (Fun f ts), ?I s) \<in> monoS" by blast
  qed
qed


lemma i_transIV: assumes 
  sn: "SN_on (qrstep Q (set R)) {Fun f ts}" and
  step: "(Fun f ts,s) \<in> qrstep Q (set R)" and
  ur: "\<not> i_trans_check Q R ur us (Fun f ts)" 
  shows "(i_trans (Q,R,ur,us,cn) (Fun f ts), i_trans (Q,R,ur,us,cn) s) \<in> (rstep (ce_trs cn))^+" (is "_ \<in> ?rel")
proof -
  let ?I = "i_trans (Q,R,ur,us,cn)"
  from ur sn have id: "?I (Fun f ts) = (comb cn) (Fun f (map ?I ts) # map ?I (qrewrite Q R (Fun f ts)))" (is "_ = (comb cn) ?list") 
    by auto
  from qrewrite_complete[OF sn step] have "s \<in> set (qrewrite Q R (Fun f ts))" .
  hence "(?I s) \<in> set ?list" by auto
  hence "(comb cn ?list, ?I s) \<in> ?rel" by (rule ce_trs_sound)
  with id show ?thesis by simp
qed

context ce_redpair
begin

lemma mode_NS: "NS^* \<subseteq> mode_NS mode"
  unfolding mode_NS_def
  by (cases mode, insert rtrancl_mono[of NS], auto)

lemma mode_NS_ctxt: "mode_cond mode m \<Longrightarrow> one_step_ctxt_closed (mode_NS mode)"
proof (cases mode)
  case False thus ?thesis unfolding mode_NS_def using 
    rtrancl_one_step_ctxt_closed[OF wm_NS] by simp
next
  case True
  assume "mode_cond mode m"
  hence "one_step_ctxt_closed S" unfolding mode_cond_def using True by auto
  from rtrancl_one_step_ctxt_closed[OF mono_NSS[OF this]] True show ?thesis unfolding mode_NS_def by simp
qed

lemma mode_S_ctxt: "mode \<Longrightarrow> mode_cond mode m \<Longrightarrow> one_step_ctxt_closed monoS"
proof -
  assume m: mode and c: "mode_cond mode m"
  from mode_NS_ctxt[OF c] have NS: "one_step_ctxt_closed monoNS" unfolding mode_NS_def
    using m by auto
  from c m have S: "one_step_ctxt_closed S" unfolding mode_cond_def by auto
  from comp_one_step_ctxt_closed[OF comp_one_step_ctxt_closed[OF NS S] NS]
  show ?thesis .
qed

lemma mode_ce: "mode \<Longrightarrow> mode_cond mode m \<Longrightarrow> (rstep (ce_trs (c,d,m)))^+ \<subseteq> monoS"
proof -
  assume mode and mode: "mode_cond mode m"
  with mode have "mono m" unfolding mode_cond_def by auto
  from trancl_mono[OF _ mono_ce[OF this]] have "(rstep (ce_trs (c,d,m)))^+ \<subseteq> S^+" by auto
  moreover have "S^+ \<subseteq> monoS" by regexp
  ultimately show ?thesis by auto
qed
end

lemma (in ce_af_redpair) i_transV: 
  assumes 
  mode: "mode_cond mode m" and
  m: "m \<ge> n" and
  rule: "lr \<in> set R" and
  sn: "SN_on (qrstep Q (set R)) {t}" and
  step: "(t,s) \<in> qrstep Q {lr}" and
  ur_closed: "ur_closed_af ur us Q (set R) \<pi>" and
  orient: "ur \<subseteq> mode_left mode" 
  shows "(i_trans (Q,R,ur,us,(c,d,m)) t, i_trans (Q,R,ur,us,(c,d,m)) s) \<in> mode_NS mode \<and> 
  (mode \<and> (lr \<notin> ur \<or> lr \<in> Snwf us) \<longrightarrow> (i_trans (Q,R,ur,us,(c,d,m)) t, i_trans (Q,R,ur,us,(c,d,m)) s) \<in> monoS)"
proof -
  let ?L = "mode_left mode"
  let ?R = "mode_NS mode"
  let ?cn = "(c,d,m)"
  let ?I = "i_trans (Q,R,ur,us,?cn)"
  let ?cond = "mode \<and> (lr \<notin> ur \<or> lr \<in> Snwf us)"
  let ?goal = "\<lambda> t s. 
    (?I t, ?I s) \<in> ?R \<and> 
  (?cond \<longrightarrow> (i_trans (Q,R,ur,us,(c,d,m)) t, i_trans (Q,R,ur,us,(c,d,m)) s) \<in> monoS)"
  from step sn show "?goal t s"
  proof induct 
    case (IH C \<sigma> l r)
    with rule have lr: "lr = (l,r)" and IH2: "(l,r) \<in> set R" by auto
    from IH(3)
    show ?case
    proof (induct C)
      case Hole
      let ?ls = "\<box>\<langle>l \<cdot> \<sigma>\<rangle>" 
      let ?rs = "\<box>\<langle>r \<cdot> \<sigma>\<rangle>" 
      from IH(1) IH(2)
      have step: "(?ls, ?rs) \<in> qrstep Q (set R)" using rule
        by (intro qrstepI[where C = Hole], auto)
      {
        assume "is_Var ?ls"
        then obtain x where "?ls = Var x" by (cases ?ls, auto)
        then obtain x where l: "l = Var x" by (cases l, auto)
        from left_Var_imp_not_SN_qrstep[OF IH2[unfolded l]] IH(3)
        have False ..
      }
      hence "\<not> (is_Var ?ls)" ..
      from this obtain f ls where ls: "?ls = Fun f ls" by (cases ?ls, auto)
      from IH(3) Hole ls have snls: "SN_on (qrstep Q (set R)) {Fun f ls}" by auto
      from step ls have stepls: "(Fun f ls, ?rs) \<in> qrstep Q (set R)" by auto
      from IH(1) have steppls: "(Fun f ls, ?rs) \<in> rqrstep Q {(l,r)}" unfolding ls[symmetric] rqrstep_def by auto
      show "?goal ?ls ?rs"
      proof (cases "i_trans_check Q R ur us (Fun f ls)")
	case True
        from i_transIII[OF mode IH2 m snls steppls True ur_closed orient]
        show ?thesis using IH2 unfolding ls lr by blast
      next
	case False
	from snls stepls this 
	have steps: "(?I ?ls, ?I ?rs) \<in> (rstep (ce_trs ?cn))^+" unfolding ls by (rule i_transIV)
        from steps ce_orient[OF m] have "(?I ?ls, ?I ?rs) \<in> NS^*" by auto
        hence piece0:"(?I ?ls, ?I ?rs) \<in> mode_NS mode" using mode_NS by blast
        show ?thesis 
          by (intro conjI[OF piece0] impI, elim conjE, insert steps mode_ce[OF _ mode], auto)
      qed
    next
      case (More f bef D aft)
      obtain t s where t: "t = l \<cdot> \<sigma>" and s: "s = r \<cdot> \<sigma>" by auto
      from More
      have snbta: "SN_on (qrstep Q (set R)) {Fun f (bef @ D\<langle>t\<rangle> # aft)}" (is "SN_on _ {Fun f ?bta}") 
        unfolding t by auto
      let ?Dt = "D\<langle>t\<rangle>"
      let ?Ds = "D\<langle>s\<rangle>"
      let ?bsa = "bef @ ?Ds # aft"
      let ?C = "More f bef D aft"
      have step: "(Fun f ?bta, Fun f ?bsa) \<in> qrstep Q (set R)" 
        unfolding s t
        by (intro qrstepI[OF IH(1) IH2, where C = ?C], auto)
      show ?case 
      proof (cases "i_trans_check Q R ur us (Fun f ?bta)")
	case False
	from snbta step False 
	have steps: "(?I (Fun f ?bta), ?I (Fun f ?bsa)) \<in> (rstep (ce_trs ?cn))^+" by (rule i_transIV)
        from steps ce_orient[OF m] have "(?I (Fun f ?bta), ?I (Fun f ?bsa)) \<in> NS^*" by auto
        hence piece0: "(?I (Fun f ?bta), ?I (Fun f ?bsa)) \<in> mode_NS mode" using mode_NS by blast
        show ?thesis unfolding s[symmetric] t[symmetric]
          by (rule conjI, insert piece0, simp, intro impI, elim conjE, insert steps mode_ce[OF _ mode], auto)
      next
	case True
        hence fus: "(f, Suc (length bef + length aft)) \<in> us" by (simp add: i_trans_check.simps)
	have "?Dt \<in> set ?bta" by auto 
	with snbta have "SN_on (qrstep Q (set R)) {?Dt}" by (rule SN_imp_SN_arg_gen[OF ctxt_closed_qrstep])
	with More have "?goal ?Dt ?Ds" unfolding s t by auto
        hence always: "(?I ?Dt, ?I ?Ds) \<in> ?R" and choice: "?cond \<longrightarrow> (?I ?Dt, ?I ?Ds) \<in> monoS" by auto
        from always have piece0: "(Fun f (map ?I bef @ ?I ?Dt # map ?I aft), 
	        Fun f (map ?I bef @ ?I ?Ds # map ?I aft)) \<in> ?R" (is "(?Il, ?Ir) \<in> _")
          using mode_NS_ctxt[OF mode] unfolding one_step_ctxt_closed_def by auto
        from choice have choice: "?cond \<longrightarrow> (?Il, ?Ir) \<in> monoS" using mode_S_ctxt[OF _ mode]
          unfolding one_step_ctxt_closed_def by blast
	from True snbta have left: "?I (Fun f ?bta) = ?Il" by auto
	from step snbta have snbsa: "SN_on (qrstep Q (set R)) {Fun f ?bsa}" by (rule step_preserves_SN_on)
	from IH2 ctxt_closureI[where C = D] have "(?Dt,?Ds) \<in> rstep (set R)"
          unfolding rstep_CC_CS s t by auto
	hence subset: "[cap (set R) ?Ds] \<subseteq> [cap (set R) ?Dt]" by (rule cap_rewrite)		
	have "i_trans_check Q R ur us (Fun f ?bsa)"
	proof (simp add: i_trans_check.simps fus, rule ballI) 
	  fix lr'
	  assume lr': "lr' \<in> set R - ur"
	  show "\<not> (\<lambda>(l, r). Ground_Context.match (GCFun f (map (cap (set R)) bef @ cap (set R) ?Ds # map (cap (set R)) aft)) l) lr'"
	  proof (cases lr')
	    case (Pair l r)
	    from lr' Pair have "(l,r) \<in> set R" by auto
	    from Pair lr' have nMatch: 
	      "\<not> Ground_Context.match (GCFun f (map (cap (set R)) bef @ cap (set R) ?Dt # map (cap (set R)) aft)) l" (is "\<not> Ground_Context.match ?cbta l") 
	      using True by (auto simp: i_trans_check.simps)
	    {
	      assume "Ground_Context.match (GCFun f (map (cap (set R)) bef @ cap (set R) ?Ds # map (cap (set R)) aft)) l" 
	      from this subset have "Ground_Context.match ?cbta l" by (rule match_below)
	      with nMatch have False ..
	    }
	    with lr' Pair show ?thesis by auto 
	  qed
	qed
	with snbsa  have "?I (Fun f ?bsa) = ?Ir" by auto	
	with left piece0 choice show ?thesis unfolding s t by (simp)
      qed
    qed
  qed
qed

no_notation Ground_Context.equiv_class ("[_]")
	
context ce_af_redpair
begin

lemma i_transVI: assumes 
  mode: "mode_cond mode m" and
  m: "m \<ge> n" and
  sn: "SN_on (qrstep Q (set R)) {t}" and
  step: "(t,s) \<in> (qrstep Q (set R))^*" and
  ur_closed: "ur_closed_af ur us Q (set R) \<pi>" and
  orient: "ur \<subseteq> mode_left mode"
  shows "(i_trans (Q,R,ur,us,(c,d,m)) t, i_trans (Q,R,ur,us,(c,d,m)) s) \<in> mode_NS mode"
 (is "?goal t s")
using step 
proof (induct rule: rtrancl_induct)
  case base
  show ?case by (cases mode, unfold mode_NS_def, auto)
next
  case (step s u)
  let ?cn = "(c,d,m)"
  let ?I = "i_trans (Q,R,ur,us,?cn)"
  from step(1) sn have SN: "SN_on (qrstep Q (set R)) {s}" 
    by (rule steps_preserve_SN_on[of _ _ "qrstep Q (set R)"])
  from step(2)[unfolded qrstep_rule_conv[where R = "set R"]] obtain lr where lr: "lr \<in> set R" and sustep: "(s,u) \<in> qrstep Q {lr}" by auto
  from i_transV[OF mode m lr SN sustep ur_closed orient]
  have isteps2: "(?I s, ?I u) \<in> mode_NS mode" by auto
  with step(3) have "(?I t, ?I u) \<in> mode_NS mode O mode_NS mode" by auto
  thus ?case unfolding mode_NS_def by (cases mode, auto)
qed


lemma i_transVI_S: assumes 
  mode: "mode_cond mode m" and
  modeT: "mode" and
  m: "m \<ge> n" and
  sn: "SN_on (qrstep Q (set R)) {t}" and
  step: "(t,s) \<in> (qrstep Q (set R))^* O qrstep Q (set R') O (qrstep Q (set R))^*" and
  ur_closed: "ur_closed_af ur us Q (set R) \<pi>" and
  orient: "ur \<subseteq> mode_left mode" and
  R': "set R' \<inter> ur \<subseteq> Snwf us" and
  RR': "set R' \<subseteq> set R"
  shows "(i_trans (Q,R,ur,us,(c,d,m)) t, i_trans (Q,R,ur,us,(c,d,m)) s) \<in> monoS"
proof -
  let ?R = "qrstep Q (set R)"
  let ?R' = "qrstep Q (set R')"
  from step obtain u v where tu: "(t,u) \<in> ?R^*" and uv: "(u,v) \<in> ?R'" and vs: "(v,s) \<in> ?R^*" by auto
  note itVI = i_transVI[OF mode m _ _ ur_closed orient]
  let ?cn = "(c,d,m)"
  let ?I = "i_trans (Q,R,ur,us,?cn)"
  from itVI[OF sn tu] have one: "(?I t,?I u) \<in> mode_NS mode" .
  from steps_preserve_SN_on[OF tu sn] have sn: "SN_on ?R {u}" .
  from uv[unfolded qrstep_rule_conv[where R = "set R'"]] obtain lr where lr: "lr \<in> set R'"
    and uv: "(u,v) \<in> qrstep Q {lr}" by auto
  with RR' R' have lr': "lr \<in> set R" and uv': "(u,v) \<in> ?R" and lr: "lr \<notin> ur \<or> lr \<in> Snwf us" 
    unfolding qrstep_rule_conv[where R = "set R"] by auto
  from i_transV[OF mode m lr' sn uv ur_closed orient] modeT lr
  have two: "(?I u,?I v) \<in> monoS" by auto
  from step_preserves_SN_on[OF uv' sn] have sn: "SN_on ?R {v}" .
  from itVI[OF sn vs] have three: "(?I v, ?I s) \<in> mode_NS mode" .
  from one two three have steps: "(?I t, ?I s) \<in> mode_NS mode O monoS O mode_NS mode" (is "_ \<in> ?Rel") by blast
  from modeT have m: "mode_NS mode = monoNS" unfolding mode_NS_def by auto
  have "?Rel \<subseteq> monoS" unfolding m by regexp
  with steps show "(?I t, ?I s) \<in> monoS" by blast
qed
end

declare i_trans.simps[simp del]    

abbreviation ur_P_closed_af :: "('f,'v)terms \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f \<times> nat)set \<Rightarrow> ('f,'v)trs \<Rightarrow> 'f af \<Rightarrow> bool" 
  where "ur_P_closed_af Q R ur us P \<pi> \<equiv> \<forall> st \<in> P. ur_closed_term_af (snd st) Q R ur us \<pi>"

  
lemma (in ce_af_root_redtriple) i_trans_sound_dp: 
  fixes Q R ur us m and P :: "('f,'v)trs"
  defines Ruc: "Ruc \<equiv> (Q,R,ur,us,(c,d,m))"
  assumes
      m: "m \<ge> n"
  and P: "(s,t) \<in> P"
  and t: "is_Fun t"
         "\<not> defined (set R) (the (root t))"
  and u: "is_Fun u"
  and steps: "(t \<cdot> \<sigma>, u \<cdot> \<tau>) \<in> (qrstep Q (set R))^*"
  and SN: "SN_on (qrstep Q (set R)) {t \<cdot> \<sigma>}"
  and ur_closed: "ur_closed_af ur us Q (set R) \<pi>"
  and ur_P_closed: "\<And> s t i. (s,t) \<in> P \<Longrightarrow> i < num_args t \<Longrightarrow> \<pi>' (the (root t)) i \<Longrightarrow> ur_closed_term_af (args t ! i) Q (set R) ur us \<pi>"
  and orient: "ur \<subseteq> NS"
  shows  "(t \<cdot> i_trans_subst Ruc \<sigma>, u \<cdot> i_trans_subst Ruc \<tau>) \<in> NST^*"
proof -
  let ?cn = "(c,d,m)"
  let ?s = "s \<cdot> \<sigma>"
  let ?t = "t \<cdot> \<sigma>"
  let ?u = "u \<cdot> \<tau>"
  let ?I = "i_trans Ruc"
  let ?Is = "i_trans_subst Ruc"
  let ?ssig = "s \<cdot> ?Is \<sigma>"
  let ?tsig = "t \<cdot> ?Is \<sigma>"
  let ?utau = "u \<cdot> ?Is \<tau>"
  from t obtain f ts where t: "t = Fun f ts" "\<not> defined (set R) (f,length ts)"
    by (cases t, auto)
  let ?ts = "map (\<lambda> t. t \<cdot> \<sigma>) ts"
  from SN left_Var_imp_not_SN_qrstep[of _ _ "set R" Q "t \<cdot> \<sigma>"]
  have "\<forall> (l,r) \<in> set R. is_Fun l" by force
  from qrsteps_imp_nrqrsteps[OF this ndef_useful_rules steps[unfolded t subst_apply_term.simps]] t(2) have steps: "(Fun f ?ts, u \<cdot> \<tau>) \<in> (nrqrstep Q (set R))^*" by auto
  let ?urc = "\<lambda> t. ur_closed_term_af t Q (set R) ur us \<pi>"
  from ur_P_closed[OF P] have urc: "\<And> i. i < length ts \<Longrightarrow> \<pi>' (f, (length ts)) i \<Longrightarrow> ?urc (ts ! i)" unfolding t by auto
  from nrqrsteps_preserve_root_fun[OF steps] u obtain us where u: "u = Fun f us" by (cases u, auto)
  let ?us = "map (\<lambda> u. u \<cdot> \<tau>) us"
  from steps have steps: "(Fun f ?ts, Fun f ?us) \<in> (nrqrstep Q (set R))^*" unfolding u by auto
  from nrqrsteps_num_args[OF steps] have len: "length ts = length us" by auto
  let ?ts' = "map (\<lambda> t. t \<cdot> ?Is \<sigma>) ts"
  let ?us' = "map (\<lambda> u. u \<cdot> ?Is \<tau>) us"
  let ?pi' = "\<pi>' (f, (length ts))"
  let ?rel = "\<lambda> i. if ?pi' i then NS^* else UNIV"
  have "(Fun f ?ts', Fun f ?us') \<in> NST^*" 
  proof (rule args_steps_imp_steps_gen[of ?rel])
    fix bef aft :: "('f,'v)term list" and s t
    assume rel: "(s,t) \<in> ?rel (length bef)" and len': "length ?us' = Suc (length bef + length aft)"
    thus "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> NST^*"
    proof (cases "?pi' (length bef)")
      case False 
      with af_compat'[unfolded af_compatible_def, rule_format, of f bef aft s t] len' len show ?thesis by auto
    next
      case True
      with rel have "(s,t) \<in> NS^*" by auto
      from compat_NSs_root[OF this] show ?thesis .
    qed
  next
    fix i
    assume i: "i < length ?us'"
    show "(?ts' ! i, ?us' ! i) \<in> (?rel i)^*"
    proof (cases "?pi' i")
      case False
      thus ?thesis by auto
    next
      case True
      with i len have ts: "?ts' ! i = ts ! i \<cdot> ?Is \<sigma>" and us: "?us' ! i = us ! i \<cdot> ?Is \<tau>" and i: "i < length ts" 
        by auto
      hence mem: "?ts ! i \<in> set ?ts" by auto
      note SN = SN_imp_SN_arg_gen[OF ctxt_closed_qrstep SN[unfolded t subst_apply_term.simps] mem] 
      from SN i have SN: "SN_on (qrstep Q (set R)) {ts ! i \<cdot> \<sigma>}" by auto
      from nrqrsteps_imp_arg_qrsteps[OF steps, of i]
      have steps: "(ts ! i \<cdot> \<sigma>, us ! i \<cdot> \<tau>) \<in> (qrstep Q (set R))^*" 
        using i len by auto
      have mode: "mode_cond False m" unfolding mode_cond_def by auto
      from orient have orient: "ur \<subseteq> mode_left False" unfolding mode_left_def by auto
      from i_transVI[OF mode m SN steps ur_closed orient]
      have one: "(?I (ts ! i \<cdot> \<sigma>), ?I (us ! i \<cdot> \<tau>)) \<in> NS^*" unfolding mode_NS_def unfolding Ruc by simp
      from steps SN have sns: "SN_on (qrstep Q (set R)) {us ! i \<cdot> \<tau>}" by (rule steps_preserve_SN_on[of _ _ "qrstep Q (set R)"])
      from urc[OF i True] have urc: "?urc (ts ! i)" .
      from i_transI[OF urc SN] have two: "(ts ! i \<cdot> ?Is \<sigma>, ?I (ts ! i \<cdot> \<sigma>)) \<in> NS^*" 
        unfolding Ruc .
      from i_transII[OF sns, THEN conjunct1] ce_orient[OF m]
      have three: "(?I (us ! i \<cdot> \<tau>), us ! i \<cdot> ?Is \<tau>) \<in> NS^*" unfolding Ruc by auto        
      from one two three have "(ts ! i \<cdot> ?Is \<sigma>, us ! i \<cdot> ?Is \<tau>) \<in> NS^* O NS^* O NS^*" by blast
      hence steps: "(ts ! i \<cdot> ?Is \<sigma>, us ! i \<cdot> ?Is \<tau>) \<in> NS^*" by simp
      thus ?thesis unfolding ts us using True by auto
    qed
  qed (insert len, auto)
  thus ?thesis unfolding t u by auto
qed

lemma (in ce_af_root_redtriple) root_redtriple_ur_sound:
  assumes rules: "ur \<subseteq> NS"
  and nonStrict: "P \<union> Pw \<subseteq> NST \<union> S"
  and strict: "Ps \<subseteq> S"
  and Pcond: "\<And> s t. (s,t) \<in> P \<union> Pw \<Longrightarrow> is_Fun s \<and> is_Fun t \<and> \<not> defined (set (R @ Rw)) (the (root t))"
  and ur_closed: "ur_closed_af ur us Q (set (R @ Rw)) \<pi>"
  and ur_P_closed: "\<And> s t i. (s,t) \<in> P \<union> Pw \<Longrightarrow> i < num_args t \<Longrightarrow> \<pi>' (the (root t)) i \<Longrightarrow> ur_closed_term_af (args t ! i) Q (set (R @ Rw)) ur us \<pi>"
  and minChain: "min_ichain (P,Pw,Q,set R,set Rw) s t \<sigma>"
  shows  "\<exists> j. min_ichain (P - Ps, Pw - Ps, Q, set R, set Rw) (shift s j) (shift t j) (shift \<sigma> j)"
proof (rule min_ichain_split_P[OF minChain], rule)
  assume "min_ichain (Ps \<inter> (P \<union> Pw), P \<union> Pw - Ps, Q, {}, set R \<union> set Rw) s t \<sigma>"
  hence chain: "min_ichain (S \<inter> (P \<union> Pw), P \<union> Pw - S, Q, {}, set (R @ Rw)) s t \<sigma>"
    by (rule min_ichain_mono, insert strict, auto)
  have id: "S \<inter> (P \<union> Pw) \<union> (P \<union> Pw - S) = P \<union> Pw" by auto
  obtain \<tau> where \<tau>: "\<tau> = (\<lambda> i. i_trans_subst (Q, R @ Rw, ur, us, c, d, n) (\<sigma> i))" by auto
  obtain u where u: "u = (\<lambda> i. t i \<cdot> \<tau> i)" by auto
  have n: "n \<le> n"  by simp
  {
    fix i    
    from chain[unfolded min_ichain.simps ichain.simps minimal_cond_def id]
    have P: "(s i, t i) \<in> P \<union> Pw" and R: "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (qrstep Q (set (R @ Rw)))^*" and SN: "SN_on (qrstep Q (set (R @ Rw))) {t i \<cdot> \<sigma> i}" 
      and P': "(s (Suc i), t (Suc i)) \<in> P \<union> Pw" by auto
    from Pcond[OF P] Pcond[OF P'] have nvart: "is_Fun (t i)" and nvars: "is_Fun (s (Suc i))"
      and ndef: "\<not> defined (set (R @ Rw)) (the (root (t i)))" by auto
    from i_trans_sound_dp[OF n P nvart ndef nvars R SN ur_closed ur_P_closed rules] 
    have "(u i, s (Suc i) \<cdot> \<tau> (Suc i)) \<in> NST^*" unfolding u \<tau> by auto
  } note steps = this
  {
    fix i
    assume "(s (Suc i), t (Suc i)) \<in> S" 
    hence "(s (Suc i) \<cdot> \<tau> (Suc i), u (Suc i)) \<in> S" using subst_S[unfolded subst_closed_def] unfolding u by auto
    from compat_NSTs[OF steps[of i] this]
    have "(u i, u (Suc i)) \<in> S" .
  } note S = this
  {
    fix i
    assume "(s (Suc i), t (Suc i)) \<notin> S"
    with chain nonStrict have "(s (Suc i), t (Suc i)) \<in> NST" by (auto simp: ichain.simps)
    hence "(s (Suc i) \<cdot> \<tau> (Suc i), u (Suc i)) \<in> NST" using subst_NST[unfolded subst_closed_def] unfolding u by auto
    with steps[of i]
    have "(u i, u (Suc i)) \<in> NST^*" by auto
  } note NST = this
  from S NST have steps: "\<forall> i. (u i, u (Suc i)) \<in> NST^* \<union> S" by auto
  from SN have SN: "SN_on S {u 0}" unfolding SN_def by auto
  interpret compat S "NST" by (unfold_locales, rule compat_NST)
  from non_strict_ending[OF steps trCompat, OF SN] obtain j
    where notS: "\<And> i. i \<ge> j \<Longrightarrow> (u i, u (Suc i)) \<notin> S" by auto
  from chain have "INFM i. (s i, t i) \<in> S \<inter> (P \<union> Pw)" by (auto simp: ichain.simps)
  from this[unfolded INFM_nat_le, THEN spec[of _ "Suc j"]] obtain i where i: "i \<ge> Suc j" and inS: "(s i, t i) \<in> S" by auto
  from i obtain k where i: "i = Suc k" and k: "k \<ge> j" by (cases i, auto)
  from S[OF inS[unfolded i]] notS[OF k] show False by simp
qed

context ce_af_redtriple
begin

lemma i_trans_sound_dp: 
  fixes Q R ur us m
  defines Ruc: "Ruc \<equiv> (Q,R,ur,us,(c,d,m))"
  assumes
  mode: "mode_cond mode m"
  and m: "m \<ge> n"
  and P: "(s,t) \<in> P"
  and steps: "(t \<cdot> \<sigma>, u \<cdot> \<tau>) \<in> (qrstep Q (set R))^*"
  and SN: "SN_on (qrstep Q (set R)) {t \<cdot> \<sigma>}"
  and ur_closed: "ur_closed_af ur us Q (set R) \<pi>"
  and ur_P_closed: "ur_P_closed_af Q (set R) ur us P \<pi>"
  and orient: "ur \<subseteq> mode_left mode"
  shows  "(t \<cdot> i_trans_subst Ruc \<sigma>, u \<cdot> i_trans_subst Ruc \<tau>) \<in> mode_NS mode"
proof -
  let ?cn = "(c,d,m)"
  let ?s = "s \<cdot> \<sigma>"
  let ?t = "t \<cdot> \<sigma>"
  let ?u = "u \<cdot> \<tau>"
  let ?I = "i_trans Ruc"
  let ?Is = "i_trans_subst Ruc"
  let ?ssig = "s \<cdot> ?Is \<sigma>"
  let ?tsig = "t \<cdot> ?Is \<sigma>"
  let ?utau = "u \<cdot> ?Is \<tau>"
  from i_transVI[OF mode m SN steps ur_closed orient]
  have one: "(?I ?t, ?I ?u) \<in> mode_NS mode" unfolding Ruc .  
  from P ur_P_closed have "ur_closed_term_af t Q (set R) ur us \<pi>" by auto
  from i_transI[OF this SN] have two: "(?tsig, ?I ?t) \<in> NS^*" unfolding Ruc by simp
  from steps SN have sns: "SN_on (qrstep Q (set R)) {?u}" by (rule steps_preserve_SN_on[of _ _ "qrstep Q (set R)"])
  from i_transII[OF sns, THEN conjunct1] ce_orient[OF m]
  have three: "(?I ?u, ?utau ) \<in> NS^*" unfolding Ruc by auto    
  from one two three have "(?tsig,?utau) \<in> NS^* O mode_NS mode O NS^*" by blast
  thus ?thesis by simp
qed

lemma redtriple_ur_sound:
  assumes rules: "ur \<subseteq> NS"
  and nonStrict: "P \<union> Pw \<subseteq> NST \<union> S"
  and strict: "Ps \<subseteq> S"
  and ur_closed: "ur_closed_af ur us Q (set (R @ Rw)) \<pi>"
  and ur_P_closed: "ur_P_closed_af Q (set (R @ Rw)) ur us (P \<union> Pw) \<pi>"
  and minChain: "min_ichain (P,Pw,Q,set R,set Rw) s t \<sigma>"
  shows  "\<exists> j. min_ichain (P - Ps, Pw - Ps, Q, set R, set Rw) (shift s j) (shift t j) (shift \<sigma> j)"
proof (rule min_ichain_split_P[OF minChain], rule)
  assume "min_ichain (Ps \<inter> (P \<union> Pw), P \<union> Pw - Ps, Q, {}, set R \<union> set Rw) s t \<sigma>"
  hence chain: "min_ichain (S \<inter> (P \<union> Pw), P \<union> Pw - S, Q, {}, set (R @ Rw)) s t \<sigma>"
    by (rule min_ichain_mono, insert strict, auto)
  have id: "S \<inter> (P \<union> Pw) \<union> (P \<union> Pw - S) = P \<union> Pw" by auto
  obtain \<tau> where \<tau>: "\<tau> = (\<lambda> i. i_trans_subst (Q, R @ Rw, ur, us, c, d, n) (\<sigma> i))" by auto
  have n: "n \<le> n"  by simp
  obtain mode where mode: "\<not> mode" by blast
  hence cond: "mode_cond mode n" unfolding mode_cond_def by auto
  from mode rules have orient: "ur \<subseteq> mode_left mode" unfolding mode_left_def by auto
  {
    fix i    
    from chain[unfolded min_ichain.simps ichain.simps minimal_cond_def id]
    have P: "(s i, t i) \<in> P \<union> Pw" and R: "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (qrstep Q (set (R @ Rw)))^*" and SN: "SN_on (qrstep Q (set (R @ Rw))) {t i \<cdot> \<sigma> i}" by auto
    from i_trans_sound_dp[OF cond n P R SN ur_closed ur_P_closed orient]
    have "(t i \<cdot> \<tau> i, s (Suc i) \<cdot> \<tau> (Suc i)) \<in> NS^*" unfolding \<tau> mode_NS_def using mode by auto
    hence "(t i \<cdot> \<tau> i, s (Suc i) \<cdot> \<tau> (Suc i)) \<in> (rstep NS)^*" 
      using rtrancl_mono[OF subset_rstep[of NS]] by auto
  } note steps = this
  from redtriple_sound[OF subset_refl nonStrict, of "{}"]
  have nchain: "\<not> ichain (S \<inter> (P \<union> Pw), P \<union> Pw - S, {}, {}, NS) s t \<tau>" .
  have chain: "ichain (S \<inter> (P \<union> Pw), P \<union> Pw - S, {}, {}, NS) s t \<tau>"
    using chain steps
    unfolding min_ichain.simps ichain.simps id INFM_nat
    by auto
  with nchain show False ..
qed
end

context mono_ce_af_redtriple
begin
lemma mono_redpair_ur_sound:
  assumes mono: "one_step_ctxt_closed S"
  and rules: "ur \<subseteq> NS \<union> S"
  and nonStrict: "P \<union> Pw \<subseteq> NS \<union> S"
  and strictP: "Ps - {lr | lr. \<not> funas_term (fst lr) \<subseteq> us} \<subseteq> S"
  and strictR: "(Rs \<inter> ur) - {lr | lr. \<not> funas_term (fst lr) \<subseteq> us} \<subseteq> S"
  and ur_closed: "ur_closed_af ur us Q (set (R @ Rw)) \<pi>"
  and ur_P_closed: "ur_P_closed_af Q (set (R @ Rw)) ur us (P \<union> Pw) \<pi>"
  and minChain: "min_ichain (P,Pw,Q,set R,set Rw) s t \<sigma>"
  shows  "\<exists> j. min_ichain (P - Ps, Pw - Ps, Q, set R - Rs, set Rw - Rs) (shift s j) (shift t j) (shift \<sigma> j)"
proof (rule min_ichain_split[OF minChain], rule)
  assume chain: "min_ichain (Ps \<inter> (P \<union> Pw), P \<union> Pw - Ps, Q, Rs \<inter> (set R \<union> set Rw), set R \<union> set Rw - Rs) s t \<sigma>"
  hence chain: "min_ichain (Ps \<inter> (P \<union> Pw), P \<union> Pw - Ps, Q, Rs \<inter> (set (R @ Rw)), set (R @ Rw) - Rs) s t \<sigma>" by auto    
  have id: "Ps \<inter> (P \<union> Pw) \<union> (P \<union> Pw - Ps) = P \<union> Pw" by auto
  have idR: "Rs \<inter> set (R @ Rw) \<union> (set (R @ Rw) - Rs) = set (R @ Rw)" by auto
  have idS: "S \<union> NS = NS \<union> S" by auto
  from S_ce_compat obtain n' where S_compat: "\<And> m. m \<ge> n' \<Longrightarrow> ce_trs (c,d,m) \<subseteq> S" by blast
  obtain k where k: "k = max n n'" by auto
  hence n: "n \<le> k"  by simp
  obtain mode where mode: "mode" by blast
  from S_compat k mode have cond: "mode_cond mode k" unfolding mode_cond_def using mono by auto
  obtain \<tau> where \<tau>: "\<tau> = (\<lambda> i. i_trans_subst (Q, R @ Rw, ur, us, c, d, k) (\<sigma> i))" by auto
  from mode rules have orient: "ur \<subseteq> mode_left mode" unfolding mode_left_def by auto
  note chain = chain[unfolded min_ichain.simps ichain.simps minimal_cond_def id idR]
  {
    fix i    
    from chain
    have P: "(s i, t i) \<in> P \<union> Pw" and R: "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (qrstep Q (set (R @ Rw)))^*" and SN: "SN_on (qrstep Q (set (R @ Rw))) {t i \<cdot> \<sigma> i}" by auto
    from i_trans_sound_dp[OF cond n P R SN ur_closed ur_P_closed orient]
    have "(t i \<cdot> \<tau> i, s (Suc i) \<cdot> \<tau> (Suc i)) \<in> (NS \<union> S)^*" unfolding \<tau> mode_NS_def using mode by auto
    hence "(t i \<cdot> \<tau> i, s (Suc i) \<cdot> \<tau> (Suc i)) \<in> (rstep (NS \<union> S))^*" 
      using rtrancl_mono[OF subset_rstep[of "NS \<union> S"]] by auto 
  } note steps = this
  have chain: "ichain (S,NS,{},S,NS) s t \<tau>"
    unfolding ichain.simps qrstep_rstep_conv idS
  proof (intro conjI allI)
    fix i
    from chain nonStrict show "(s i, t i) \<in> NS \<union> S" by auto 
  next
    fix i
    show "(s i \<cdot> \<tau> i) \<in> NF_terms {}" by simp
  next
    fix i
    show "(t i \<cdot> \<tau> i, s (Suc i) \<cdot> \<tau> (Suc i)) \<in> (rstep (NS \<union> S))^*"
      using steps[of i] by simp
  next
    let ?Q = "qrstep Q (set (R @ Rw))"
    let ?QQ = "?Q^* O qrstep Q (Rs \<inter> set (R @ Rw)) O ?Q^*"
    let ?R = "rstep (NS \<union> S)"
    let ?RR = "?R^* O rstep S O ?R^*"
    show "(INFM i. (s i, t i) \<in> S) \<or> 
      (INFM i. (t i \<cdot> \<tau> i, s (Suc i) \<cdot> \<tau> (Suc i)) \<in> ?RR)" unfolding INFM_disj_distrib[symmetric] 
      unfolding INFM_nat
    proof (intro allI)
      fix m
      from chain 
      have "(INFM i. (s i, t i) \<in> Ps \<inter> (P \<union> Pw)) \<or>
        (INFM i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> ?QQ)" by blast
      from this[unfolded INFM_disj_distrib[symmetric], unfolded INFM_nat]
      obtain l where l: "Suc m < l" and alt:"(s l, t l) \<in> Ps \<inter> (P \<union> Pw) \<or> 
        (t l \<cdot> \<sigma> l, s (Suc l) \<cdot> \<sigma> (Suc l)) \<in> ?QQ" by blast
      hence l': "m < l" by auto
      let ?I = "i_trans (Q, R @ Rw, ur, us, c, d, k)"
      obtain I where I: "?I = I" by auto        
      from alt show "\<exists> l > m. (s l, t l) \<in> S \<or> (t l \<cdot> \<tau> l, s (Suc l) \<cdot> \<tau> (Suc l)) \<in> ?RR"
      proof
        assume Ssteps: "(t l \<cdot> \<sigma> l, s (Suc l) \<cdot> \<sigma> (Suc l)) \<in> ?QQ"
        let ?ts = "t l \<cdot> \<sigma> l"
        let ?tt = "t l \<cdot> \<tau> l"
        let ?ss = "s (Suc l) \<cdot> \<sigma> (Suc l)"      
        let ?st = "s (Suc l) \<cdot> \<tau> (Suc l)"              
        from chain have SN: "SN_on (qrstep Q (set (R @ Rw))) {?ts}" by simp
        from chain have P: "(s l, t l) \<in> P \<union> Pw" by auto
        from P ur_P_closed have urt: "ur_closed_term_af (t l) Q (set (R @ Rw)) ur us \<pi>" by force
        from chain have steps: "(?ts, ?ss) \<in> ?Q^*" by auto
        from steps SN have SNs: "SN_on (qrstep Q (set (R @ Rw))) {?ss}" by (rule steps_preserve_SN_on[of _ _ "qrstep Q (set (R @ Rw))"])
        from i_transI[OF urt SN] have ns0: "(?tt, ?I ?ts) \<in> NS^*" unfolding \<tau> .
        from i_transII[OF SNs, of ur us "(c,d,k)", THEN conjunct1] ce_orient[OF n]
        have ns2: "(?I ?ss, ?st) \<in> NS^*" unfolding \<tau> by auto
        let ?Rs = "filter (\<lambda> lr. lr \<in> Rs) (R @ Rw)"
        have Rs: "Rs \<inter> (set (R @ Rw)) = set ?Rs" by auto         
        have "(?I ?ts, ?I ?ss) \<in> monoNS O S O monoNS"
          by (rule i_transVI_S[OF cond mode n SN Ssteps[unfolded Rs] ur_closed orient], insert strictR, auto)
        from ns0 this ns2 have dec: "(?tt, ?st) \<in> NS^* O monoNS O S O monoNS O NS^*" unfolding I by auto
        have "NS^* O monoNS O S O monoNS O NS^* \<subseteq> monoNS O S O monoNS" by regexp
        with dec have "(?tt, ?st) \<in> monoNS O S O monoNS" by blast
        with rtrancl_mono[OF subset_rstep[of "NS \<union> S"]] subset_rstep[of S] have "(?tt, ?st) \<in> ?RR" by auto
        with l' show ?thesis by auto
      next
        assume P: "(s l, t l) \<in> Ps \<inter> (P \<union> Pw)"
        with strictP have "(s l, t l) \<in> S \<or> \<not> funas_term (s l) \<subseteq> us" by force
        thus ?thesis
        proof
          assume "(s l, t l) \<in> S" thus ?thesis using l' by auto
        next
          assume nwf: "\<not> funas_term (s l) \<subseteq> us"
          from l obtain ll where ll: "l = Suc ll" by (cases l, auto)
          let ?ts = "t ll \<cdot> \<sigma> ll"
          let ?tt = "t ll \<cdot> \<tau> ll"
          let ?ss = "s (Suc ll) \<cdot> \<sigma> (Suc ll)"      
          let ?st = "s (Suc ll) \<cdot> \<tau> (Suc ll)"              
          from chain have SN: "SN_on (qrstep Q (set (R @ Rw))) {?ts}" by simp
          from chain have P: "(s ll, t ll) \<in> P \<union> Pw" by auto
          from P ur_P_closed have urt: "ur_closed_term_af (t ll) Q (set (R @ Rw)) ur us \<pi>" by force
          from chain have steps: "(?ts, ?ss) \<in> ?Q^*" by auto
          from steps SN have SNs: "SN_on (qrstep Q (set (R @ Rw))) {?ss}" by (rule steps_preserve_SN_on[of _ _ "qrstep Q (set (R @ Rw))"])
          from i_transI[OF urt SN] have ns0: "(?tt, ?I ?ts) \<in> NS^*" unfolding \<tau> .
          from i_transVI[OF cond n SN steps ur_closed orient] have ns1: "(?I ?ts, ?I ?ss) \<in> monoNS"
            using mode unfolding mode_NS_def by auto
          from i_transII[OF SNs, of ur us "(c,d,k)", THEN conjunct2] nwf[unfolded ll]
          have "(?I ?ss, ?st) \<in> (rstep (ce_trs (c,d,k)))^+" unfolding \<tau> by auto
          from trancl_mono[OF this mono_ce[OF cond[unfolded mode_cond_def, THEN mp[OF _ mode]]]]
          have strict: "(?I ?ss, ?st) \<in> S^+" .
          from ns0 ns1 strict have strict: "(?tt,?st) \<in> NS^* O monoNS O S^+" by auto
          have "NS^*O monoNS O S^+ \<subseteq> monoNS O S O monoNS" by regexp
          from set_mp[OF this strict] have "(?tt,?st) \<in> monoNS O S O monoNS" .
          with rtrancl_mono[OF subset_rstep[of "NS \<union> S"]] subset_rstep[of S] have strict: "(?tt, ?st) \<in> ?RR" by auto
          from l ll have "m < ll" by auto
          with strict show ?thesis by blast
        qed
      qed
    qed
  qed
  have idS: "S \<inter> (NS \<union> S) = S" by auto
  from ichain_mono[OF chain subset_refl _ _ subset_refl, of "NS \<union> S - S" "{}" "NS \<union> S - S"]
    mono_redpair_sound_ichain[OF mono subset_refl subset_refl, of "{}" s t \<tau>]
  show False unfolding idS by blast
qed
end


end