(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Usable_Rules_Impl
imports
  Usable_Rules
  Edg_Impl
  QDP_Framework_Impl
  "Q-Restricted_Rewriting_Impl"
  Term_Order_Impl
begin

fun matchCapRMBelow :: "('f \<times> nat \<Rightarrow> ('f,'v)rules) \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
  where "matchCapRMBelow rm l (Fun f ts) = Ground_Context.match (GCFun f (map (capRM2 rm) ts)) l"

lemma matchCapRMBelow:
  assumes noLeftVars: "\<And> lr. lr \<in> set R \<Longrightarrow> is_Fun (fst lr)"
    and rm: "\<And>f n. set (rm (f,n)) = {(l,r) | l r. (l,r) \<in> set R \<and> root l = Some (f, n)}"
  shows "matchCapRMBelow rm l (Fun f ts) = match_cap_below l (set R) (Fun f ts)"
proof -
  from capRM2_sound[OF noLeftVars rm] have id: "capRM2 rm = cap (set R)" 
    by (intro ext, auto)
  show ?thesis unfolding matchCapRMBelow.simps match_cap_below.simps id ..
qed

fun
  check_ur_closed_term_rm_af :: "(('f \<times> nat) \<Rightarrow> ('f, 'v) rules) \<Rightarrow> ('f::show, 'v::show) rules \<Rightarrow> 'f af \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
where
  "check_ur_closed_term_rm_af _ _ _ (Var x) = OK" |
  "check_ur_closed_term_rm_af rm ur \<pi> (Fun f ts) = do {
     let n = length ts;
     let pi = \<pi> (f, n);
     check_allm_index (\<lambda>t i. if pi i then check_ur_closed_term_rm_af rm ur \<pi> t else OK) ts;
     check_allm (\<lambda>lr.
       check (lr \<in> set ur \<or> \<not> matchCapRMBelow rm (fst lr) (Fun f ts)) 
        (''due to the subterm '' +#+ shows(Fun f ts) +@+ '' of some usable rhs, rule '' +#+ shows_rule lr +@+ shows_string '' should be usable.'')
     ) (rm (f,n))
   }"


lemma check_ur_closed_term_rm_af_sound:
  assumes noLeftVars: "\<And> lr. lr \<in> set R \<Longrightarrow> is_Fun (fst lr)"
  and rm: "\<And> f n. set (rm (f,n)) = {(l,r) | l r. (l,r) \<in> set R \<and> root l = Some (f, n)}"
  and check: "isOK(check_ur_closed_term_rm_af rm ur \<pi> t)" 
    and tus: "funas_term t \<subseteq> us"
  shows "ur_closed_term_af t Q (set R) (set ur) us \<pi>"
using check tus proof (induct t)
  case (Fun f ts)
  from Fun(3) have fus: "(f, length ts) \<in> us" by auto
  from Fun(3) have "\<forall> i < length ts. funas_term (ts ! i) \<subseteq> us" by force
  with Fun(1) Fun(2)
  have rls: "\<forall> (l,r) \<in> set (rm (f,length ts)). (l,r) \<in> set ur \<or> \<not> matchCapRMBelow rm l (Fun f ts)"
    and ind: " (\<forall>i<length ts. \<pi> (f, (length ts)) i \<longrightarrow> ur_closed_term_af (ts ! i) Q (set R) (set ur) us \<pi>)" by (auto simp: Let_def) 
  have "\<forall> (l,r) \<in> set R. (l,r) \<in> set ur \<or> \<not> matchCapRMBelow rm l (Fun f ts)" (is "\<forall> lr \<in> set R. ?P lr") 
  proof
    fix lr
    assume lr: "lr \<in> set R"
    show "?P lr"
    proof (cases lr)
      case (Pair l r)
      show ?thesis
      proof (cases "(l,r) \<in> set (rm (f,length ts))")
	case True
	with Pair rls show ?thesis by auto
      next
	case False
	from noLeftVars[OF lr] Pair obtain g ss where l: "l = Fun g ss" by (cases l, auto)
	with rm False lr Pair have "\<not> (f = g \<and> length ss = length ts)" by auto
	with l rm have "\<not> matchCapRMBelow rm l  (Fun f ts)" by (auto simp: matchCapRMBelow.simps Ground_Context.match_def)
	with Pair rls show ?thesis by auto
      qed
    qed
  qed
  with ind fus show ?case using matchCapRMBelow[OF noLeftVars rm] by auto
qed simp 


definition
  check_ur_P_closed_rm_af ::
    "('f \<times> nat \<Rightarrow> ('f,'v)rules) \<Rightarrow> ('f :: show, 'v :: show) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> 'f af \<Rightarrow> shows check"
where
  "check_ur_P_closed_rm_af rm ur P \<pi> \<equiv> do {
       check_allm (\<lambda>lr. check_ur_closed_term_rm_af rm ur \<pi> (snd lr)) ur
         <+? (\<lambda>s. ''error when checking closure properties of rhs of usable rules'' +#+ shows_nl +@+ s);
       check_allm (\<lambda>st. check_ur_closed_term_rm_af rm ur \<pi> (snd st)) P
         <+? (\<lambda>s. ''error when checking closure properties of rhs of DPs'' +#+ shows_nl +@+ s)
     }"

lemma check_ur_P_closed_rm_af_sound:
  assumes noLeftVars: "\<And> lr. lr \<in> set R \<Longrightarrow> is_Fun (fst lr)"
  and rm: "\<And> f n. set (rm (f,n)) = {(l,r) | l r. (l,r) \<in> set R \<and> root l = Some (f, n)}"
  and check: "isOK(check_ur_P_closed_rm_af rm ur P \<pi>)"
    and us: "(\<Union> ( (funas_term o snd) ` (set ur \<union> set P))) \<subseteq> us"
  shows "ur_closed_af (set ur) us Q (set R) \<pi> \<and> ur_P_closed_af Q (set R) (set ur) us (set P) \<pi>"
proof -
  note closure = check_ur_closed_term_rm_af_sound[OF noLeftVars rm]
  {
    fix l r
    assume lr: "(l,r) \<in> set ur"
    with us have rus: "funas_term r \<subseteq> us" by force
    from closure[OF _ _ rus] check lr
    have "ur_closed_term_af r Q (set R) (set ur) us \<pi>" unfolding check_ur_P_closed_rm_af_def by auto
  } note part1 = this
  {
    fix l r
    assume lr: "(l,r) \<in> set P"
    with us have rus: "funas_term r \<subseteq> us" by force
    from closure[OF _ _ rus] check lr
    have "ur_closed_term_af r Q (set R) (set ur) us \<pi>" unfolding check_ur_P_closed_rm_af_def by auto
  } note part2 = this
  from part1 part2 show ?thesis by auto
qed

definition
  ur_af_root_redtriple_proc ::
    "('dpp, 'f, 'v) dpp_ops \<Rightarrow> ('f::show, 'v::show)root_redtriple \<Rightarrow> 
     ('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> 'dpp proc"
where
  "ur_af_root_redtriple_proc I rp Premove ur dpp = check_return (do {
     let (ps, pns) = dpp_ops.split_pairs I dpp Premove;
     let P = dpp_ops.pairs I dpp;
     let rm = dpp_ops.rules_map I dpp;     
     let r = dpp_ops.rules I dpp;
     let pi = root_redtriple.af rp;
     let pi' = root_redtriple.aft rp;
      check_allm (\<lambda>(l, r). do {
        check_no_var l;
        check_no_var r;
        check_no_defined_root (dpp_spec.is_defined I dpp) r
      }) P;
     root_redtriple.valid rp;
     check_ur_P_closed_rm_af rm ur (concat (map (\<lambda>st.
         case snd st of Fun f ts \<Rightarrow> (let n = length ts; is = pi' (f, n) in map (\<lambda> i. (fst st, ts ! i)) (filter is [0..<n]))
     ) P)) pi;
     check_allm (root_redtriple.ns rp) ur
       <+? (\<lambda>s. ''problem when orienting usable rules'' +#+ shows_nl +@+ s);
     check_allm (root_redtriple.nst rp) pns
       <+? (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s);
     check_allm (root_redtriple.s rp) ps
       <+? (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s)
   } <+? (\<lambda>s. ''could not apply the reduction pair processor with the following'' +#+ shows_nl +@+
     (root_redtriple.desc rp) +@+ shows_nl +@+ s))
   (dpp_spec.delete_pairs I dpp Premove)"


definition
  ur_af_redtriple_proc ::
    "('dpp, 'f, 'v) dpp_ops \<Rightarrow> ('f::show, 'v::show) redtriple \<Rightarrow> 
     ('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> 'dpp proc"
where
  "ur_af_redtriple_proc I rp Premove ur dpp = check_return (do {
     let (ps, pns) = dpp_ops.split_pairs I dpp Premove;
     let P = dpp_ops.pairs I dpp;
     let rm = dpp_ops.rules_map I dpp;     
     redtriple.valid rp;
     check_ur_P_closed_rm_af rm ur P (redtriple.af rp);
     check_allm (redtriple.ns rp) ur
       <+? (\<lambda>s. ''problem when orienting usable rules'' +#+ shows_nl +@+ s);
     check_allm (redtriple.nst rp) pns
       <+? (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s);
     check_allm (redtriple.s rp) ps
       <+? (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s)
   } <+? (\<lambda>s. ''could not apply the reduction pair processor with the following'' +#+ shows_nl +@+
     (redtriple.desc rp) +@+ shows_nl +@+ s))
   (dpp_spec.delete_pairs I dpp Premove)"

definition
  mono_ur_redpair_proc ::
    "('dpp, 'f, 'v) dpp_ops \<Rightarrow> ('f::show, 'v::show) redtriple \<Rightarrow> 
     ('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> 'dpp proc"
where
  "mono_ur_redpair_proc I rp Premove Rremove ur dpp = check_return (do {
     let P = dpp_ops.pairs I dpp;
     let us = concat (map (funas_term_list o snd) (P @ ur));
     let filt = (\<lambda> lr. (\<forall> f \<in> set (funas_term_list (fst lr)). f \<in> set us));
     let (pms, pns) = dpp_ops.split_pairs I dpp Premove;
     let (ps, pnwf) = partition filt pms;
     let (urms, urns) = partition (\<lambda> u. u \<in> set Rremove) ur;
     let (urs, urnwf) = partition filt urms;
     let rm = dpp_ops.rules_map I dpp;     
     redtriple.valid rp;
     redtriple.mono rp;
     check_ur_P_closed_rm_af rm ur P full_af;
     check_allm (redtriple.ns rp) (urns @ urnwf)
       <+? (\<lambda>s. ''problem when orienting usable rules'' +#+ shows_nl +@+ s);
     check_allm (redtriple.s rp) urs
       <+? (\<lambda>s. ''problem when orienting usable rules'' +#+ shows_nl +@+ s);
     check_allm (redtriple.ns rp) (pns @ pnwf)
       <+? (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s);
     check_allm (redtriple.s rp) ps
       <+? (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s)
   } <+? (\<lambda>s. ''could not apply the reduction pair processor with the following'' +#+ shows_nl +@+
     (redtriple.desc rp) +@+ shows_nl +@+ s))
   (dpp_spec.delete_pairs_rules I dpp Premove Rremove)"

context dpp_spec
begin

lemma ur_af_root_redtriple_proc:
  assumes grp: "generic_root_redtriple_impl grp"
  shows "sound_proc_impl (ur_af_root_redtriple_proc I (grp rp) ps ur)"
proof 
  fix d d'
  assume fin: "finite_dpp (dpp d')"
  and ok: "ur_af_root_redtriple_proc I (grp rp) ps ur d = return d'"
  let ?rp = "grp rp"
  let ?S = "set ps"
  let ?P = "set (P d)"
  let ?Pw = "set (Pw d)"
  let ?S' = "(?P \<union> ?Pw) \<inter> ?S"
  let ?Pb = "set (pairs d)"
  let ?PP = "?P \<union> ?Pw"
  let ?Q = "set (Q d)"
  let ?R = "set (R d)"
  let ?Rw = "set (Rw d)"
  let ?Rb = "set (rules d)"
  let ?RR = "set (R d @ Rw d)"
  have RR_conv: "?RR = ?Rb" unfolding dpp_spec_sound by auto
  have PP_conv: "?PP = ?Pb" unfolding dpp_spec_sound by auto
  interpret generic_root_redtriple_impl grp by (rule grp)
  show "finite_dpp (dpp d)"
  proof (rule finite_under_var_cond_impl)
    assume var: "\<And>lr. lr \<in> ?Rb \<Longrightarrow> is_Fun (fst lr)"
    show ?thesis
    proof (cases "split_pairs d ps")
      case (Pair Ps Pr)    
      from split_pairs_sound[OF this] have Ps: "set Ps = (?P \<union> ?Pw) \<inter> ?S" 
        and Pr: "set Pr = (?P \<union> ?Pw) - ?S" by (auto simp: dpp_spec_sound)
      let ?pi = "root_redtriple.af ?rp"
      let ?pi' = "root_redtriple.aft ?rp"
      note ok = ok[unfolded ur_af_root_redtriple_proc_def Let_def Pair, simplified]
      let ?Pargs = "concat (map (\<lambda>st.
         case snd st of Fun f ts \<Rightarrow> map (\<lambda> i. (fst st, ts ! i)) (filter (?pi' (f, length ts)) [0..<length ts])) (pairs d))"
      from ok have valid: "isOK(root_redtriple.valid ?rp)" 
        and NS: "isOK (check_allm (root_redtriple.ns ?rp) ur)"
        and NST: "isOK (check_allm (root_redtriple.nst ?rp) Pr)"
        and S: "isOK(check_allm (root_redtriple.s ?rp) Ps)" 
        and ur: "isOK( check_ur_P_closed_rm_af (rules_map d) ur ?Pargs ?pi)"
        and d': "d' = delete_pairs d ps" by auto
      from ok have condP: "\<forall>(l, r)\<in> ?P \<union> ?Pw.
        is_Fun r \<and> is_Fun l \<and> 
        \<not> defined (?R \<union> ?Rw) (the (root r))"
        by (simp add: split_def Let_def
          check_no_defined_root_sound[OF is_defined_sound, unfolded rules_map_defined_conv]
          rules_sound pairs_sound)      
      from generate_root_redtriple[OF valid S NS NST, unfolded Ps] 
      obtain S NS NST where "ce_af_root_redtriple_order S NS NST ?pi ?pi'" and S: "(?P \<union> ?Pw) \<inter> ?S \<subseteq> S" and NS: "?P \<union> ?Pw - ?S \<subseteq> NST" and R: "set ur \<subseteq> NS" by  (auto simp: dpp_spec_sound Pr)
      hence NS: "?P \<union> ?Pw \<subseteq> NST \<union> S" by auto
      let ?us = "(\<Union> ( (funas_term o snd) ` (set ur \<union> set ?Pargs)))"
      have ur_cl: "ur_closed_af (set ur) ?us ?Q ?RR ?pi \<and> 
        ur_P_closed_af ?Q ?RR (set ur) ?us (set ?Pargs) ?pi" (is "?ur1 \<and> ?ur2")
        unfolding RR_conv 
        by (rule check_ur_P_closed_rm_af_sound[OF var _ ur subset_refl],
          unfold dpp_spec_sound, auto)
      hence ?ur1 and ?ur2 by auto    
      {
        fix s t i
        assume mem: "(s,t) \<in> ?PP" and i: "i < num_args t" and pi: "?pi' (the (root t)) i"
        from i obtain f ts where t: "t = Fun f ts" by (cases t, auto)
        with mem i pi have "(s,ts ! i) \<in> set ?Pargs" by force
        with `?ur2`
        have "ur_closed_term_af (args t ! i) ?Q ?RR (set ur) ?us ?pi" unfolding t by (force simp: dpp_spec_sound)
      } note ur2 = this
      interpret ce_af_root_redtriple_order S NS NST ?pi ?pi' by fact
      note proc =  root_redtriple_ur_sound[OF R NS S]
      show ?thesis
        unfolding finite_dpp_def
      proof(rule, elim exE) 
        fix s t \<sigma>
        assume "min_ichain (dpp d) s t \<sigma>"
        note proc = proc[OF _ _ _ this[unfolded dpp_spec_sound]]
        note proc = proc[OF _ `?ur1` ur2]
        {
          fix s t
          assume "(s,t) \<in> ?PP"
          hence "is_Fun s \<and> is_Fun t \<and> \<not> defined ?RR (the (root t))"
            using condP by auto
        }
        note proc = proc[OF this]
        from proc obtain i where chain: "min_ichain (?P - ?S',?Pw - ?S',?Q,?R,?Rw) (shift s i) (shift t i) (shift \<sigma> i)" ..
        have "min_ichain (?P - ?S,?Pw - ?S,?Q,?R,?Rw) (shift s i) (shift t i) (shift \<sigma> i)" 
          by (rule min_ichain_mono[OF chain], auto)
        with fin[unfolded d' delete_P_Pw_sound finite_dpp_def]
        show False by blast
      qed
    qed
  qed
qed

lemma ur_af_redtriple_proc:
  assumes grp: "generic_redtriple_impl grp"
  shows "sound_proc_impl (ur_af_redtriple_proc I (grp rp) ps ur)"
proof 
  fix d d'
  assume fin: "finite_dpp (dpp d')"
  and ok: "ur_af_redtriple_proc I (grp rp) ps ur d = return d'"
  let ?rp = "grp rp"
  let ?S = "set ps"
  let ?P = "set (P d)"
  let ?Pw = "set (Pw d)"
  let ?S' = "(?P \<union> ?Pw) \<inter> ?S"
  let ?Pb = "set (pairs d)"
  let ?PP = "?P \<union> ?Pw"
  let ?Q = "set (Q d)"
  let ?R = "set (R d)"
  let ?Rw = "set (Rw d)"
  let ?Rb = "set (rules d)"
  let ?RR = "set (R d @ Rw d)"
  have RR_conv: "?RR = ?Rb" unfolding dpp_spec_sound by auto
  have PP_conv: "?PP = ?Pb" unfolding dpp_spec_sound by auto
  interpret generic_redtriple_impl grp by (rule grp)
  show "finite_dpp (dpp d)"
  proof (rule finite_under_var_cond_impl)
    assume var: "\<And>lr. lr \<in> ?Rb \<Longrightarrow> is_Fun (fst lr)"
    show ?thesis
    proof (cases "split_pairs d ps")
      case (Pair Ps Pr)    
      from split_pairs_sound[OF this] have Ps: "set Ps = (?P \<union> ?Pw) \<inter> ?S" 
        and Pr: "set Pr = (?P \<union> ?Pw) - ?S" by (auto simp: dpp_spec_sound)
      let ?pi = "redtriple.af ?rp"
      note ok = ok[unfolded ur_af_redtriple_proc_def Let_def Pair, simplified]
      from ok have valid: "isOK(redtriple.valid ?rp)" 
        and compat: "isOK(check_ur_P_closed_rm_af (rules_map d) ur (pairs d) ?pi)"
        and NS: "isOK (check_allm (redtriple.ns ?rp) ur)"
        and NST: "isOK (check_allm (redtriple.nst ?rp) Pr)"
        and S: "isOK(check_allm (redtriple.s ?rp) Ps)" 
        and d': "d' = delete_pairs d ps" by auto
      from generate_redtriple[OF valid S NS NST, unfolded Ps] 
      obtain S NS NST where "ce_af_redtriple_order S NS NST (redtriple.af ?rp)" and S: "(?P \<union> ?Pw) \<inter> ?S \<subseteq> S" and NS: "?P \<union> ?Pw - ?S \<subseteq> NST" and R: "set ur \<subseteq> NS" by  (auto simp: dpp_spec_sound Pr)
      hence NS: "?P \<union> ?Pw \<subseteq> NST \<union> S" by auto
      let ?us = "(\<Union> ( (funas_term o snd) ` (set ur \<union> ?Pb)))"
      have ur_cl: "ur_closed_af (set ur) ?us ?Q ?RR ?pi \<and> 
        ur_P_closed_af ?Q ?RR (set ur) ?us ?PP ?pi" (is "?ur1 \<and> ?ur2")
        unfolding RR_conv PP_conv
        by (rule check_ur_P_closed_rm_af_sound[OF var _ compat subset_refl],   
          unfold dpp_spec_sound, auto)
      hence ?ur1 and ?ur2 by auto
      interpret ce_af_redtriple_order S NS NST ?pi by fact
      note proc =  redtriple_ur_sound[OF R NS S `?ur1` `?ur2`] 
      show ?thesis
        unfolding finite_dpp_def
      proof(rule, elim exE) 
        fix s t \<sigma>
        assume "min_ichain (dpp d) s t \<sigma>"
        from proc[OF this[unfolded dpp_spec_sound]]
        obtain i where chain: "min_ichain (?P - ?S',?Pw - ?S',?Q,?R,?Rw) (shift s i) (shift t i) (shift \<sigma> i)" ..
        have "min_ichain (?P - ?S,?Pw - ?S,?Q,?R,?Rw) (shift s i) (shift t i) (shift \<sigma> i)" 
          by (rule min_ichain_mono[OF chain], auto)
        with fin[unfolded d' delete_P_Pw_sound finite_dpp_def]
        show False by blast
      qed
    qed
  qed
qed

lemma mono_ur_redpair_proc:
  assumes grp: "generic_redtriple_impl grp"
  shows "sound_proc_impl (mono_ur_redpair_proc I (grp rp) ps rs ur)"
proof 
  fix d d'
  assume fin: "finite_dpp (dpp d')"
    and ok: "mono_ur_redpair_proc I (grp rp) ps rs ur d = return d'"
  let ?rp = "grp rp"
  let ?S = "set ps"
  let ?P = "set (dpp_ops.P I d)"
  let ?Pw = "set (dpp_ops.Pw I d)"
  let ?S' = "(?P \<union> ?Pw) \<inter> ?S"
  let ?Pb = "set (dpp_ops.pairs I d)"
  let ?PP = "?P \<union> ?Pw"
  let ?Q = "set (dpp_ops.Q I d)"
  let ?R = "set (dpp_ops.R I d)"
  let ?Rw = "set (dpp_ops.Rw I d)"
  let ?Sr = "set rs"
  let ?Rb = "set (dpp_ops.rules I d)"
  let ?RR = "set (dpp_ops.R I d @ dpp_ops.Rw I d)"
  have RR_conv: "?RR = ?Rb" unfolding dpp_spec_sound by auto
  have PP_conv: "?PP = ?Pb" unfolding dpp_spec_sound by auto
  interpret generic_redtriple_impl grp by (rule grp)
  show "finite_dpp (dpp_ops.dpp I d)"
  proof (rule finite_under_var_cond_impl)
    assume var: "\<And>lr. lr \<in> ?Rb \<Longrightarrow> is_Fun (fst lr)"
    obtain us where us: "us = concat (map (funas_term_list o snd) (dpp_ops.pairs I d @ ur))" by auto
    let ?filt = "\<lambda> lr. (\<forall> f \<in> set (funas_term_list (fst lr)). f \<in> set us)"
    let ?WF = "{lr. ?filt lr}"
    obtain Pms Pns where p1: "dpp_ops.split_pairs I d ps = (Pms,Pns)" by force
    obtain Ps Pnwf where p2: "partition ?filt Pms = (Ps,Pnwf)" by force
    obtain Rms Rns where r1: "partition (\<lambda> r. r \<in> set rs) ur = (Rms,Rns)" by force
    obtain Rs Rnwf where r2: "partition ?filt Rms = (Rs,Rnwf)" by force    
    from r1 r2 have ur: "set ur = set Rs \<union> set Rnwf \<union> set Rns" and Rs: "set Rs = (set ur \<inter> set rs) \<inter> ?WF" by auto
    from split_pairs_sound[OF p1] have Pms: "set Pms = (?P \<union> ?Pw) \<inter> ?S" and Pns: "set Pns = (?P \<union> ?Pw) - ?S" by (auto simp: dpp_spec_sound)
    from p2 have Ps: "set Ps = ((?P \<union> ?Pw) \<inter> ?S) \<inter> ?WF" and Pnwf: "set Pnwf = ((?P \<union> ?Pw) \<inter> ?S) - ?WF" unfolding Pms[symmetric] by auto
    note P = Ps Pnwf Pns
    note ok = ok[unfolded mono_ur_redpair_proc_def Let_def p1 r1 split p2[unfolded us] r2[unfolded us], simplified]
    let ?\<pi> = "full_af"
    from ok have valid: "isOK(redtriple.valid ?rp)" 
      and mono: "isOK(redtriple.mono ?rp)"
      and compat: "isOK(check_ur_P_closed_rm_af (dpp_ops.rules_map I d) ur (dpp_ops.pairs I d) ?\<pi>)"
      and NS: "isOK (check_allm (redtriple.ns ?rp) (Rns @ Rnwf @ Pns @ Pnwf))"
      and S: "isOK(check_allm (redtriple.s ?rp) (Ps @ Rs))" 
      and d': "d' = dpp_spec.delete_pairs_rules I d ps rs" by auto
    let ?us = "(\<Union> ( (funas_term o snd) ` (set ur \<union> ?Pb)))"
    have ur_cl: "ur_closed_af (set ur) ?us ?Q ?RR ?\<pi> \<and> 
      ur_P_closed_af ?Q ?RR (set ur) ?us ?PP ?\<pi>" (is "?ur1 \<and> ?ur2")
      unfolding RR_conv PP_conv
      by (rule check_ur_P_closed_rm_af_sound[OF var _ compat subset_refl],   
        unfold dpp_spec_sound, auto)
    hence ?ur1 and ?ur2 by auto
    from generate_redtriple[OF valid S NS, of Nil] mono
    obtain S NS NST where "mono_ce_af_redtriple_order S NS NST (redtriple.af ?rp)" and S: "set Ps \<union> set Rs \<subseteq> S" and NS: "set Rns \<union> set Rnwf \<union> set Pns \<union> set Pnwf \<subseteq> NS" and mono: "one_step_ctxt_closed S"
      by auto
    then interpret old: mono_ce_af_redtriple_order S NS NST "redtriple.af ?rp" by simp
    interpret mono_ce_af_redtriple S NS NST full_af 
      by (unfold_locales, rule full_af)
    from S NS have ur: "set ur \<subseteq> NS \<union> S" unfolding ur by auto
    have P: "?P \<union> ?Pw = set Ps \<union> set Pns \<union> set Pnwf" unfolding P by blast
    from S NS have p: "?P \<union> ?Pw \<subseteq> NS \<union> S" unfolding P by auto
    let ?NWF = "{lr | lr. \<not> funas_term (fst lr) \<subseteq> ?us}"
    have swap: "set ur \<union> ?Pb = ?Pb \<union> set ur" by auto
    have "?S' - ?NWF \<subseteq> set Ps" unfolding Ps us pairs_sound[symmetric] swap by auto
    also have "... \<subseteq> S" using S by auto
    finally have Ps: "?S' - ?NWF \<subseteq> S" .
    have "(?Sr \<inter> set ur) - ?NWF = (set ur \<inter> ?Sr) - ?NWF" by auto
    also have "... \<subseteq> set Rs" unfolding Rs us swap by auto
    also have "... \<subseteq> S" using S by auto
    finally have Rs: "(?Sr \<inter> set ur) - ?NWF \<subseteq> S" .
    note proc = mono_redpair_ur_sound[OF mono ur p Ps Rs `?ur1` `?ur2`] 
    show ?thesis
      unfolding finite_dpp_def
    proof(rule, elim exE) 
      fix s t \<sigma>
      assume "min_ichain (dpp_ops.dpp I d) s t \<sigma>"
      from proc[OF this[unfolded dpp_spec_sound]]
      obtain i where chain: "min_ichain (?P - ?S',?Pw - ?S',?Q,?R - ?Sr,?Rw - ?Sr) (shift s i) (shift t i) (shift \<sigma> i)" ..
      have "min_ichain (?P - ?S,?Pw - ?S,?Q,?R - ?Sr,?Rw - ?Sr) (shift s i) (shift t i) (shift \<sigma> i)" 
        by (rule min_ichain_mono[OF chain], auto)
      with fin[unfolded d' delete_simps finite_dpp_def]
      show False by blast
    qed
  qed
qed
end


end