(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Util2
  imports Main
begin

definition replace :: "'a \<Rightarrow> 'a set \<Rightarrow> 'a set \<Rightarrow> 'a set"
  where "replace a bs M \<equiv> if a \<in> M then M - {a} \<union> bs else M"

definition replace_impl :: "'a \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list"
  where "replace_impl a bs M \<equiv> if a \<in> set M then bs @ filter (\<lambda> b. b \<noteq> a) M else M"

lemma replace_impl[simp]: "set (replace_impl a bs M) = replace a (set bs) (set M)"
  unfolding replace_impl_def replace_def by auto


(* TODO: this function is currently also available in RBT_Map_Set_Extension in AFP/Transive_Closure *)
(* however, the import of the Collection Framework via RBT_Map_Set_Extension has some undesired 
   behaviour elsewhere
   \<Rightarrow> for next AFP update, move sorted_list_subset out of Transitive_Closure into Util ! *)
fun sorted_list_subset :: "'a :: linorder list \<Rightarrow> 'a list \<Rightarrow> 'a option"
  where "sorted_list_subset (a # as) (b # bs) = 
            (if a = b then sorted_list_subset as (b # bs)
        else if a > b then sorted_list_subset (a # as) bs
        else Some a)"
    |  "sorted_list_subset [] _ = None"
    |  "sorted_list_subset (a # _) [] = Some a"
   
lemma sorted_list_subset: assumes "sorted as" and "sorted bs"
  shows "(sorted_list_subset as bs = None) = (set as \<subseteq> set bs)"
  using assms 
proof (induct rule: sorted_list_subset.induct)
  case (2 bs)
  thus ?case by auto
next
  case (3 a as)
  thus ?case by auto
next
  case (1 a as b bs)
  from 1(3) have sas: "sorted as" and a: "\<And> a'. a' \<in> set as \<Longrightarrow> a \<le> a'" unfolding linorder_class.sorted.simps[of "a # as"] by auto
  from 1(4) have sbs: "sorted bs" and b: "\<And> b'. b' \<in> set bs \<Longrightarrow> b \<le> b'" unfolding linorder_class.sorted.simps[of "b # bs"] by auto
  show ?case
  proof (cases "a = b")
    case True
    from 1(1)[OF this sas 1(4)] True show ?thesis by auto
  next
    case False note oFalse = this
    show ?thesis 
    proof (cases "a > b")
      case True
      with a b have "b \<notin> set as" by force
      with 1(2)[OF False True 1(3) sbs] False True show ?thesis by auto
    next
      case False
      with oFalse have "a < b" by auto
      with a b have "a \<notin> set bs" by force
      with oFalse False show ?thesis by auto
    qed
  qed
qed

end