object Natural {

  def apply(numeral: BigInt): Natural = new Natural(numeral max 0)
  def apply(numeral: Int): Natural = Natural(BigInt(numeral))
  def apply(numeral: String): Natural = Natural(BigInt(numeral))

}

class Natural private(private val value: BigInt) {

  override def hashCode(): Int = this.value.hashCode()

  override def equals(that: Any): Boolean = that match {
    case that: Natural => this equals that
    case _ => false
  }

  override def toString(): String = this.value.toString

  def equals(that: Natural): Boolean = this.value == that.value

  def as_BigInt: BigInt = this.value
  def as_Int: Int = if (this.value >= scala.Int.MinValue && this.value <= scala.Int.MaxValue)
      this.value.intValue
    else error("Int value out of range: " + this.value.toString)

  def +(that: Natural): Natural = new Natural(this.value + that.value)
  def -(that: Natural): Natural = Natural(this.value - that.value)
  def *(that: Natural): Natural = new Natural(this.value * that.value)

  def /%(that: Natural): (Natural, Natural) = if (that.value == 0) (new Natural(0), this)
    else {
      val (k, l) = this.value /% that.value
      (new Natural(k), new Natural(l))
    }

  def <=(that: Natural): Boolean = this.value <= that.value

  def <(that: Natural): Boolean = this.value < that.value

}


object Ceta {

def id[A]: A => A = (x: A) => x

trait equal[A] {
  val `Ceta.equal`: (A, A) => Boolean
}
def equal[A](a: A, b: A)(implicit A: equal[A]): Boolean = A.`Ceta.equal`(a, b)

def eq[A : equal](a: A, b: A): Boolean = equal[A](a, b)

abstract sealed class nat
final case object Zero_nat extends nat
final case class Suc(a: nat) extends nat

abstract sealed class color
final case object R extends color
final case object B extends color

abstract sealed class rbta[+A, +B]
final case object Emptya extends rbta[Nothing, Nothing]
final case class
  Branch[A, B](a: color, b: rbta[A, B], c: A, d: B, e: rbta[A, B])
  extends rbta[A, B]

trait ord[A] {
  val `Ceta.less_eq`: (A, A) => Boolean
  val `Ceta.less`: (A, A) => Boolean
}
def less_eq[A](a: A, b: A)(implicit A: ord[A]): Boolean = A.`Ceta.less_eq`(a, b)
def less[A](a: A, b: A)(implicit A: ord[A]): Boolean = A.`Ceta.less`(a, b)

trait preorder[A] extends ord[A] {
}

trait order[A] extends preorder[A] {
}

trait linorder[A] extends order[A] {
}

abstract sealed class rbt[+A, +B]
final case class Rbt[A, B](a: rbta[A, B]) extends rbt[A, B]

abstract sealed class rat
final case class Frct(a: (BigInt, BigInt)) extends rat

def nat_aux(i: BigInt, n: nat): nat =
  (if (i <= BigInt(0)) n else nat_aux(i - BigInt(1), Suc(n)))

def nat(i: BigInt): nat = nat_aux(i, Zero_nat)

def hd[A](x0: List[A]): A = x0 match {
  case x :: xs => x
}

def tl[A](x0: List[A]): List[A] = x0 match {
  case Nil => Nil
  case x :: xs => xs
}

abstract sealed class order_tag
final case object Lex extends order_tag
final case object Mul extends order_tag

def nulla[A](x0: List[A]): Boolean = x0 match {
  case Nil => true
  case x :: xs => false
}

def the[A](x0: Option[A]): A = x0 match {
  case Some(x) => x
}

def foldr[A, B](f: A => B => B, x1: List[A], a: B): B = (f, x1, a) match {
  case (f, Nil, a) => a
  case (f, x :: xs, a) => (f(x))(foldr[A, B](f, xs, a))
}

def concat[A](xss: List[List[A]]): List[A] =
  foldr[List[A], List[A]]((a: List[A]) => (b: List[A]) => a ++ b, xss, Nil)

abstract sealed class sum[+A, +B]
final case class Inl[A, B](a: A) extends sum[A, B]
final case class Inr[B, A](a: B) extends sum[A, B]

abstract sealed class nibble
final case object Nibble0 extends nibble
final case object Nibble1 extends nibble
final case object Nibble2 extends nibble
final case object Nibble3 extends nibble
final case object Nibble4 extends nibble
final case object Nibble5 extends nibble
final case object Nibble6 extends nibble
final case object Nibble7 extends nibble
final case object Nibble8 extends nibble
final case object Nibble9 extends nibble
final case object NibbleA extends nibble
final case object NibbleB extends nibble
final case object NibbleC extends nibble
final case object NibbleD extends nibble
final case object NibbleE extends nibble
final case object NibbleF extends nibble

def newline: Char = '\12'

abstract sealed class xml
final case class
  Xml(a: List[Char], b: List[(List[Char], List[Char])], c: List[xml],
       d: Option[List[Char]])
  extends xml

def is_none[A](x0: Option[A]): Boolean = x0 match {
  case Some(x) => false
  case None => true
}

def equal_lista[A : equal](x0: List[A], x1: List[A]): Boolean = (x0, x1) match {
  case (a :: list, Nil) => false
  case (Nil, a :: list) => false
  case (aa :: lista, a :: list) => eq[A](aa, a) && equal_lista[A](lista, list)
  case (Nil, Nil) => true
}

implicit def equal_char: equal[Char] = new equal[Char] {
  val `Ceta.equal` = (a: Char, b: Char) => a == b
}

def comp[A, B, C](f: A => B, g: C => A): C => B = (x: C) => f(g(x))

trait show[A] {
  val `Ceta.shows_prec`: (nat, A, List[Char]) => List[Char]
  val `Ceta.shows_list`: (List[A], List[Char]) => List[Char]
}
def shows_prec[A](a: nat, b: A, c: List[Char])(implicit A: show[A]): List[Char]
  = A.`Ceta.shows_prec`(a, b, c)
def shows_list[A](a: List[A], b: List[Char])(implicit A: show[A]): List[Char] =
  A.`Ceta.shows_list`(a, b)

def shows_string: (List[Char]) => (List[Char]) => List[Char] =
  (a: List[Char]) => (b: List[Char]) => a ++ b

def shows_list_char(cs: List[Char]): (List[Char]) => List[Char] =
  shows_string.apply(cs)

def shows_prec_char(d: nat, c: Char): (List[Char]) => List[Char] =
  (a: List[Char]) => c :: a

implicit def show_char: show[Char] = new show[Char] {
  val `Ceta.shows_prec` = (a: nat, b: Char, c: List[Char]) =>
    (shows_prec_char(a, b)).apply(c)
  val `Ceta.shows_list` = (a: List[Char], b: List[Char]) =>
    (shows_list_char(a)).apply(b)
}

def replicate[A](xa0: nat, x: A): List[A] = (xa0, x) match {
  case (Zero_nat, x) => Nil
  case (Suc(n), x) => x :: replicate[A](n, x)
}

def shows_map[A](s: A => (List[Char]) => List[Char], x1: List[A]):
      (List[Char]) => List[Char] =
  (s, x1) match {
  case (s, Nil) => id[List[Char]]
  case (s, x :: xs) =>
    comp[List[Char], List[Char], List[Char]](s(x), shows_map[A](s, xs))
}

def fst[A, B](x0: (A, B)): A = x0 match {
  case (a, b) => a
}

def snd[A, B](x0: (A, B)): B = x0 match {
  case (a, b) => b
}

def dquote: List[Char] = List('\"')

def shows_prec_list[A : show](d: nat, l: List[A]): (List[Char]) => List[Char] =
  (a: List[Char]) => shows_list[A](l, a)

def shows_attr(av: (List[Char], List[Char])): (List[Char]) => List[Char] =
  comp[List[Char], List[Char],
        List[Char]](shows_prec_list[Char](Zero_nat,
   fst[List[Char], List[Char]](av)),
                     comp[List[Char], List[Char],
                           List[Char]](shows_string.apply(List('=')),
shows_string.apply(dquote ++ (snd[List[Char], List[Char]](av) ++ dquote))))

def shows_attrs(as: List[(List[Char], List[Char])]):
      (List[Char]) => List[Char] =
  shows_map[(List[Char],
              List[Char])]((a: (List[Char], List[Char])) =>
                             comp[List[Char], List[Char],
                                   List[Char]](shows_string.apply(List(' ')),
        shows_attr(a)),
                            as)

def shows_XML_indent(ind: List[Char], i: nat, x2: xml):
      (List[Char]) => List[Char] =
  (ind, i, x2) match {
  case (ind, i, Xml(n, a, c, t)) =>
    comp[List[Char], List[Char],
          List[Char]](shows_string.apply(ind),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('<')),
  comp[List[Char], List[Char],
        List[Char]](shows_prec_list[Char](Zero_nat, n),
                     comp[List[Char], List[Char],
                           List[Char]](shows_attrs(a),
(if (nulla[xml](c) && is_none[List[Char]](t)) shows_string.apply(List('/', '>'))
  else comp[List[Char], List[Char],
             List[Char]](shows_string.apply(List('>')),
                          comp[List[Char], List[Char],
                                List[Char]]((if (! (nulla[xml](c)))
      comp[List[Char], List[Char],
            List[Char]](shows_map[xml]((aa: xml) =>
 shows_XML_indent(ind ++ replicate[Char](i, ' '), i, aa),
c),
                         shows_string.apply(ind))
      else id[List[Char]]),
     comp[List[Char], List[Char],
           List[Char]]((if (is_none[List[Char]](t)) id[List[Char]]
                         else shows_string.apply(the[List[Char]](t))),
                        comp[List[Char], List[Char],
                              List[Char]](shows_string.apply(List('<', '/')),
   comp[List[Char], List[Char],
         List[Char]](shows_prec_list[Char](Zero_nat, n),
                      shows_string.apply(List('>'))))))))))))
}

def shows_prec_xml(d: nat, xml: xml): (List[Char]) => List[Char] =
  shows_XML_indent(List(newline), nat(BigInt(2)), xml)

def text(tag: List[Char], x1: xml): sum[List[Char], List[Char]] = (tag, x1)
  match {
  case (tag, Xml(n, atts, cs, t)) =>
    (if (equal_lista[Char](n, tag) &&
           (nulla[(List[Char], List[Char])](atts) &&
             (nulla[xml](cs) &&
               (! (is_none[List[Char]](t)) &&
                 ! (nulla[Char](the[List[Char]](t)))))))
      Inr[List[Char], List[Char]](the[List[Char]](t))
      else Inl[List[Char],
                List[Char]](concat[Char](List(List('c', 'o', 'u', 'l', 'd', ' ',
            'n', 'o', 't', ' ', 'e', 'x', 't', 'r', 'a', 'c', 't', ' ', 't',
            'e', 'x', 't', ' ', 'f', 'o', 'r', ' '),
       tag, List(' ', 'f', 'r', 'o', 'm', ' '), List(newline),
       (shows_prec_xml(Zero_nat, Xml(n, atts, cs, t))).apply(Nil)))))
}

def take[A](n: nat, x1: List[A]): List[A] = (n, x1) match {
  case (n, Nil) => Nil
  case (n, x :: xs) =>
    n match {
      case Zero_nat => Nil
      case Suc(m) => x :: take[A](m, xs)
    }
}

def one_nat: nat = Suc(Zero_nat)

def bindb[A, B, C](m: sum[A, B], f: B => sum[A, C]): sum[A, C] =
  m match {
    case Inl(a) => Inl[A, C](a)
    case Inr(a) => f(a)
  }

def digit2int(x: Char): sum[List[Char], BigInt] =
  (if (x == '0') Inr[BigInt, List[Char]](BigInt(0))
    else (if (x == '1') Inr[BigInt, List[Char]](BigInt(1))
           else (if (x == '2') Inr[BigInt, List[Char]](BigInt(2))
                  else (if (x == '3') Inr[BigInt, List[Char]](BigInt(3))
                         else (if (x == '4') Inr[BigInt, List[Char]](BigInt(4))
                                else (if (x == '5')
                                       Inr[BigInt, List[Char]](BigInt(5))
                                       else (if (x == '6')
      Inr[BigInt, List[Char]](BigInt(6))
      else (if (x == '7') Inr[BigInt, List[Char]](BigInt(7))
             else (if (x == '8') Inr[BigInt, List[Char]](BigInt(8))
                    else (if (x == '9') Inr[BigInt, List[Char]](BigInt(9))
                           else Inl[List[Char],
                                     BigInt](List(x) ++
       List(' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'd', 'i', 'g',
             'i', 't'))))))))))))

def string2int_aux(n: BigInt, x1: List[Char]): sum[List[Char], BigInt] = (n, x1)
  match {
  case (n, Nil) => Inr[BigInt, List[Char]](n)
  case (n, d :: s) =>
    bindb[List[Char], BigInt,
           BigInt](digit2int(d),
                    (m: BigInt) => string2int_aux(BigInt(10) * n + m, s))
}

def string2int(s: List[Char]): sum[List[Char], BigInt] =
  (if (nulla[Char](s))
    Inl[List[Char],
         BigInt](List('c', 'a', 'n', 'n', 'o', 't', ' ', 'c', 'o', 'n', 'v',
                       'e', 'r', 't', ' ', 'e', 'm', 'p', 't', 'y', ' ', 's',
                       't', 'r', 'i', 'n', 'g', ' ', 'i', 'n', 't', 'o', ' ',
                       'n', 'u', 'm', 'b', 'e', 'r'))
    else (if (equal_lista[Char](take[Char](one_nat, s), List('-')))
           bindb[List[Char], BigInt,
                  BigInt](string2int_aux(BigInt(0), tl[Char](s)),
                           (i: BigInt) =>
                             Inr[BigInt, List[Char]](BigInt(0) - i))
           else string2int_aux(BigInt(0), s)))

def int(tag: List[Char], x: xml): sum[List[Char], BigInt] =
  bindb[List[Char], List[Char],
         BigInt](text(tag, x), (a: List[Char]) => string2int(a))

def mapa[A, B](f: A => B, x1: List[A]): List[B] = (f, x1) match {
  case (f, Nil) => Nil
  case (f, x :: xs) => f(x) :: mapa[A, B](f, xs)
}

def sequence[A, B](x0: List[sum[A, B]]): sum[A, List[B]] = x0 match {
  case Nil => Inr[List[B], A](Nil)
  case m :: ms =>
    bindb[A, B,
           List[B]](m, (x: B) =>
                         bindb[A, List[B],
                                List[B]](sequence[A, B](ms),
  (xs: List[B]) => Inr[List[B], A](x :: xs)))
}

def mapm[A, B, C](f: A => sum[B, C], ms: List[A]): sum[B, List[C]] =
  sequence[B, C](mapa[A, sum[B, C]](f, ms))

def map[A]:
      (xml => sum[List[Char], A]) => (List[xml]) => sum[List[Char], List[A]] =
  (a: xml => sum[List[Char], A]) =>
    (b: List[xml]) => mapm[xml, List[Char], A](a, b)

def nata(tag: List[Char], x: xml): sum[List[Char], nat] =
  bindb[List[Char], List[Char],
         nat](text(tag, x),
               (txt: List[Char]) =>
                 bindb[List[Char], BigInt,
                        nat](string2int(txt),
                              (i: BigInt) => Inr[nat, List[Char]](nat(i))))

def abs_int(i: BigInt): BigInt = (if (i < BigInt(0)) (- i) else i)

def sgn_int(i: BigInt): BigInt =
  (if (i == BigInt(0)) BigInt(0)
    else (if (BigInt(0) < i) BigInt(1) else (- BigInt(1))))

def apsnd[A, B, C](f: A => B, x1: (C, A)): (C, B) = (f, x1) match {
  case (f, (x, y)) => (x, f(y))
}

def divmod_int(k: BigInt, l: BigInt): (BigInt, BigInt) =
  (if (k == BigInt(0)) (BigInt(0), BigInt(0))
    else (if (l == BigInt(0)) (BigInt(0), k)
           else apsnd[BigInt, BigInt,
                       BigInt]((a: BigInt) => sgn_int(l) * a,
                                (if (sgn_int(k) == sgn_int(l))
                                  ((k: BigInt) => (l: BigInt) => if (l == 0)
                                    (BigInt(0), k) else
                                    (k.abs /% l.abs)).apply(k).apply(l)
                                  else {
 val (r, s): (BigInt, BigInt) =
   ((k: BigInt) => (l: BigInt) => if (l == 0) (BigInt(0), k) else
     (k.abs /% l.abs)).apply(k).apply(l);
 (if (s == BigInt(0)) ((- r), BigInt(0))
   else ((- r) - BigInt(1), abs_int(l) - s))
                                       }))))

def mod_int(a: BigInt, b: BigInt): BigInt =
  snd[BigInt, BigInt](divmod_int(a, b))

def gcd_int(k: BigInt, l: BigInt): BigInt =
  abs_int((if (l == BigInt(0)) k
            else gcd_int(l, mod_int(abs_int(k), abs_int(l)))))

def div_int(a: BigInt, b: BigInt): BigInt =
  fst[BigInt, BigInt](divmod_int(a, b))

def normalize(p: (BigInt, BigInt)): (BigInt, BigInt) =
  (if (BigInt(0) < snd[BigInt, BigInt](p))
    {
      val a: BigInt = gcd_int(fst[BigInt, BigInt](p), snd[BigInt, BigInt](p));
      (div_int(fst[BigInt, BigInt](p), a), div_int(snd[BigInt, BigInt](p), a))
    }
    else (if (snd[BigInt, BigInt](p) == BigInt(0)) (BigInt(0), BigInt(1))
           else {
                  val a: BigInt =
                    (- (gcd_int(fst[BigInt, BigInt](p),
                                 snd[BigInt, BigInt](p))));
                  (div_int(fst[BigInt, BigInt](p), a),
                    div_int(snd[BigInt, BigInt](p), a))
                }))

def fract(a: BigInt, b: BigInt): rat = Frct(normalize((a, b)))

def xml_change[A, B](p: xml => sum[List[Char], A], f: A => B, x: xml):
      sum[List[Char], B] =
  bindb[List[Char], A, B](p(x), (a: A) => Inr[B, List[Char]](f(a)))

def maps[A, B](f: A => List[B], x1: List[A]): List[B] = (f, x1) match {
  case (f, Nil) => Nil
  case (f, x :: xs) => f(x) ++ maps[A, B](f, xs)
}

def lookupa[A : equal, B](uu: A, x1: List[(A, B)]): Option[B] = (uu, x1) match {
  case (uu, Nil) => None
  case (aa, (a, b) :: xs) =>
    (if (eq[A](aa, a)) Some[B](b) else lookupa[A, B](aa, xs))
}

def get_name(x0: xml): List[Char] = x0 match {
  case Xml(name, uu, uv, uw) => name
}

implicit def equal_list[A : equal]: equal[List[A]] = new equal[List[A]] {
  val `Ceta.equal` = (a: List[A], b: List[A]) => equal_lista[A](a, b)
}

def xml_options[A](ps: List[(List[Char], xml => sum[List[Char], A])], x: xml):
      sum[List[Char], A] =
  lookupa[List[Char], xml => sum[List[Char], A]](get_name(x), ps) match {
    case None =>
      Inl[List[Char],
           A](concat[Char](List(List('e', 'x', 'p', 'e', 'c', 't', 'e', 'd',
                                      ' ', 'o', 'n', 'e', ' ', 'o', 'f', ':',
                                      ' '),
                                 maps[(List[Char], xml => sum[List[Char], A]),
                                       Char]((p:
        (List[Char], xml => sum[List[Char], A]))
       => fst[List[Char], xml => sum[List[Char], A]](p) ++ List(' '),
      ps),
                                 List(newline),
                                 List('b', 'u', 't', ' ', 'f', 'o', 'u', 'n',
                                       'd'),
                                 List(newline),
                                 (shows_prec_xml(Zero_nat, x)).apply(Nil))))
    case Some(p) => p(x)
  }

def nth[A](x0: List[A], x1: nat): A = (x0, x1) match {
  case (x :: xs, Suc(n)) => nth[A](xs, n)
  case (x :: xs, Zero_nat) => x
}

def equal_nata(x0: nat, x1: nat): Boolean = (x0, x1) match {
  case (Suc(nat), Zero_nat) => false
  case (Zero_nat, Suc(nat)) => false
  case (Suc(nata), Suc(nat)) => equal_nata(nata, nat)
  case (Zero_nat, Zero_nat) => true
}

def plus_nat(x0: nat, n: nat): nat = (x0, n) match {
  case (Suc(m), n) => plus_nat(m, Suc(n))
  case (Zero_nat, n) => n
}

def size_list[A](x0: List[A]): nat = x0 match {
  case Nil => Zero_nat
  case a :: list => plus_nat(size_list[A](list), Suc(Zero_nat))
}

def no_transform[A](tag: List[Char], xml: xml): sum[List[Char], A] =
  Inl[List[Char],
       A](concat[Char](List(List('c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't',
                                  ' ', 't', 'r', 'a', 'n', 's', 'f', 'o', 'r',
                                  'm', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l',
                                  'l', 'o', 'w', 'i', 'n', 'g', '\12', ' ', ' ',
                                  ' ', ' ', ' ', 'x', 'm', 'l', ' ', 'e', 'l',
                                  'e', 'm', 'e', 'n', 't', ' ', '(', 'e', 'x',
                                  'p', 'e', 'c', 't', 'e', 'd', ' '),
                             tag, List(')'), List(newline),
                             (shows_prec_xml(Zero_nat, xml)).apply(Nil))))

def xml2elements[A, B,
                  C](tag: List[Char], p1: xml => sum[List[Char], A],
                      p2: xml => sum[List[Char], B], f: A => B => C, x4: xml):
      sum[List[Char], C] =
  (tag, p1, p2, f, x4) match {
  case (tag, p1, p2, f, Xml(name, atts, cs, text)) =>
    (if (equal_lista[Char](name, tag) &&
           (nulla[(List[Char], List[Char])](atts) &&
             (equal_nata(size_list[xml](cs), nat(BigInt(2))) &&
               is_none[List[Char]](text))))
      bindb[List[Char], A,
             C](p1(nth[xml](cs, Zero_nat)),
                 (x1: A) =>
                   bindb[List[Char], B,
                          C](p2(nth[xml](cs, one_nat)),
                              (x2: B) => Inr[C, List[Char]]((f(x1))(x2))))
      else no_transform[C](tag, Xml(name, atts, cs, text)))
}

def rat: xml => sum[List[Char], rat] =
  (a: xml) =>
    xml_options[rat](List((List('i', 'n', 't', 'e', 'g', 'e', 'r'),
                            (aa: xml) =>
                              xml_change[BigInt,
  rat]((ab: xml) => int(List('i', 'n', 't', 'e', 'g', 'e', 'r'), ab),
        (x: BigInt) => fract(x, BigInt(1)), aa)),
                           (List('r', 'a', 't', 'i', 'o', 'n', 'a', 'l'),
                             (aa: xml) =>
                               xml2elements[BigInt, BigInt,
     rat](List('r', 'a', 't', 'i', 'o', 'n', 'a', 'l'),
           (ab: xml) =>
             int(List('n', 'u', 'm', 'e', 'r', 'a', 't', 'o', 'r'), ab),
           (ab: xml) =>
             int(List('d', 'e', 'n', 'o', 'm', 'i', 'n', 'a', 't', 'o', 'r'),
                  ab),
           (ab: BigInt) => (b: BigInt) => fract(ab, b), aa))),
                      a)

def fold[A, B](f: A => B => B, x1: List[A]): B => B = (f, x1) match {
  case (f, Nil) => id[B]
  case (f, x :: xs) => comp[B, B, B](fold[A, B](f, xs), f(x))
}

def rev[A](xs: List[A]): List[A] =
  (fold[A, List[A]]((a: A) => (b: List[A]) => a :: b, xs)).apply(Nil)

def inserta[A : equal](y: A, a: A => Boolean, x: A): Boolean =
  eq[A](y, x) || a(x)

trait bot[A] extends order[A] {
  val `Ceta.bot`: A
}
def bot[A](implicit A: bot[A]): A = A.`Ceta.bot`

def bot_boola: Boolean = false

def less_bool(x0: Boolean, b: Boolean): Boolean = (x0, b) match {
  case (true, b) => false
  case (false, b) => b
}

def less_eq_bool(x0: Boolean, b: Boolean): Boolean = (x0, b) match {
  case (true, b) => b
  case (false, b) => true
}

implicit def ord_bool: ord[Boolean] = new ord[Boolean] {
  val `Ceta.less_eq` = (a: Boolean, b: Boolean) => less_eq_bool(a, b)
  val `Ceta.less` = (a: Boolean, b: Boolean) => less_bool(a, b)
}

implicit def preorder_bool: preorder[Boolean] = new preorder[Boolean] {
  val `Ceta.less_eq` = (a: Boolean, b: Boolean) => less_eq_bool(a, b)
  val `Ceta.less` = (a: Boolean, b: Boolean) => less_bool(a, b)
}

implicit def order_bool: order[Boolean] = new order[Boolean] {
  val `Ceta.less_eq` = (a: Boolean, b: Boolean) => less_eq_bool(a, b)
  val `Ceta.less` = (a: Boolean, b: Boolean) => less_bool(a, b)
}

implicit def bot_bool: bot[Boolean] = new bot[Boolean] {
  val `Ceta.bot` = bot_boola
  val `Ceta.less_eq` = (a: Boolean, b: Boolean) => less_eq_bool(a, b)
  val `Ceta.less` = (a: Boolean, b: Boolean) => less_bool(a, b)
}

def bot_fun[A, B : bot]: A => B = (_: A) => bot[B]

def set[A : equal](x0: List[A]): A => Boolean = x0 match {
  case Nil => bot_fun[A, Boolean]
  case x :: xs => (a: A) => inserta[A](x, set[A](xs), a)
}

def less_eq_nat(x0: nat, n: nat): Boolean = (x0, n) match {
  case (Suc(m), n) => less_nat(m, n)
  case (Zero_nat, n) => true
}

def less_nat(m: nat, x1: nat): Boolean = (m, x1) match {
  case (m, Suc(n)) => less_eq_nat(m, n)
  case (n, Zero_nat) => false
}

def upt(i: nat, j: nat): List[nat] =
  (if (less_nat(i, j)) i :: upt(Suc(i), j) else Nil)

def zip[A, B](xs: List[A], ys: List[B]): List[(A, B)] = (xs, ys) match {
  case (x :: xs, y :: ys) => (x, y) :: zip[A, B](xs, ys)
  case (xs, Nil) => Nil
  case (Nil, ys) => Nil
}

abstract sealed class term[+A, +B]
final case class Var[B, A](a: B) extends term[A, B]
final case class Fun[A, B](a: A, b: List[term[A, B]]) extends term[A, B]

abstract sealed class ctxt[+A, +B]
final case object Hole extends ctxt[Nothing, Nothing]
final case class
  More[A, B](a: A, b: List[term[A, B]], c: ctxt[A, B], d: List[term[A, B]])
  extends ctxt[A, B]

def string2bool(s: List[Char]): sum[List[Char], Boolean] =
  (if (equal_lista[Char](s, List('t', 'r', 'u', 'e')))
    Inr[Boolean, List[Char]](true)
    else (if (equal_lista[Char](s, List('f', 'a', 'l', 's', 'e')))
           Inr[Boolean, List[Char]](false)
           else Inl[List[Char],
                     Boolean](List('c', 'a', 'n', 'n', 'o', 't', ' ', 'c', 'o',
                                    'n', 'v', 'e', 'r', 't', ' ') ++
                                (s ++ List(' ', 'i', 'n', 't', 'o', ' ', 'B',
    'o', 'o', 'l', 'e', 'a', 'n')))))

def bool(tag: List[Char], node: xml): sum[List[Char], Boolean] =
  bindb[List[Char], List[Char],
         Boolean](text(tag, node), (a: List[Char]) => string2bool(a))

def drop[A](n: nat, x1: List[A]): List[A] = (n, x1) match {
  case (n, Nil) => Nil
  case (n, x :: xs) =>
    n match {
      case Zero_nat => x :: xs
      case Suc(m) => drop[A](m, xs)
    }
}

def last[A](x0: List[A]): A = x0 match {
  case x :: xs => (if (nulla[A](xs)) x else last[A](xs))
}

def empty[A : linorder, B]: rbt[A, B] = Rbt[A, B](Emptya)

trait times[A] {
  val `Ceta.times`: (A, A) => A
}
def times[A](a: A, b: A)(implicit A: times[A]): A = A.`Ceta.times`(a, b)

trait dvd[A] extends times[A] {
}

def args[A, B](x0: term[A, B]): List[term[A, B]] = x0 match {
  case Var(x) => Nil
  case Fun(f, ts) => ts
}

def root[A, B](x0: term[A, B]): Option[(A, nat)] = x0 match {
  case Var(x) => None
  case Fun(f, ts) => Some[(A, nat)]((f, size_list[term[A, B]](ts)))
}

def membera[A : equal](x0: List[A], y: A): Boolean = (x0, y) match {
  case (Nil, y) => false
  case (x :: xs, y) => eq[A](x, y) || membera[A](xs, y)
}

def list_ex[A](p: A => Boolean, x1: List[A]): Boolean = (p, x1) match {
  case (p, Nil) => false
  case (p, x :: xs) => p(x) || list_ex[A](p, xs)
}

def equal_terma[A : equal, B : equal](x0: term[A, B], x1: term[A, B]): Boolean =
  (x0, x1) match {
  case (Fun(f, list), Var(v)) => false
  case (Var(v), Fun(f, list)) => false
  case (Fun(fa, lista), Fun(f, list)) =>
    eq[A](fa, f) && equal_lista[term[A, B]](lista, list)
  case (Var(va), Var(v)) => eq[B](va, v)
}

implicit def equal_term[A : equal, B : equal]: equal[term[A, B]] = new
  equal[term[A, B]] {
  val `Ceta.equal` = (a: term[A, B], b: term[A, B]) => equal_terma[A, B](a, b)
}

def supt_impl[A : equal, B : equal](x0: term[A, B], t: term[A, B]): Boolean =
  (x0, t) match {
  case (Var(x), t) => false
  case (Fun(f, ss), t) =>
    membera[term[A, B]](ss, t) ||
      list_ex[term[A, B]]((s: term[A, B]) => supt_impl[A, B](s, t), ss)
}

def supt[A : equal, B : equal]: ((term[A, B], term[A, B])) => Boolean =
  (a: (term[A, B], term[A, B])) =>
    {
      val (aa, b): (term[A, B], term[A, B]) = a;
      supt_impl[A, B](aa, b)
    }

def find[A](uu: A => Boolean, x1: List[A]): Option[A] = (uu, x1) match {
  case (uu, Nil) => None
  case (p, x :: xs) => (if (p(x)) Some[A](x) else find[A](p, xs))
}

def bind[A, B](x0: Option[A], f: A => Option[B]): Option[B] = (x0, f) match {
  case (None, f) => None
  case (Some(x), f) => f(x)
}

def mapM[A, B](f: A => Option[B], x1: List[A]): Option[List[B]] = (f, x1) match
  {
  case (f, Nil) => Some[List[B]](Nil)
  case (f, x :: xs) =>
    bind[B, List[B]](f(x),
                      (y: B) =>
                        bind[List[B],
                              List[B]](mapM[A, B](f, xs),
(ys: List[B]) => Some[List[B]](y :: ys)))
}

trait one[A] {
  val `Ceta.one`: A
}
def one[A](implicit A: one[A]): A = A.`Ceta.one`

implicit def times_int: times[BigInt] = new times[BigInt] {
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
}

implicit def dvd_int: dvd[BigInt] = new dvd[BigInt] {
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
}

def one_inta: BigInt = BigInt(1)

implicit def one_int: one[BigInt] = new one[BigInt] {
  val `Ceta.one` = one_inta
}

implicit def ord_int: ord[BigInt] = new ord[BigInt] {
  val `Ceta.less_eq` = (a: BigInt, b: BigInt) => a <= b
  val `Ceta.less` = (a: BigInt, b: BigInt) => a < b
}

def foldl[A, B](f: A => B => A, a: A, x2: List[B]): A = (f, a, x2) match {
  case (f, a, Nil) => a
  case (f, a, x :: xs) => foldl[A, B](f, (f(a))(x), xs)
}

def vec[A](n: nat, x: List[A]): Boolean = equal_nata(size_list[A](x), n)

def list_all[A](p: A => Boolean, x1: List[A]): Boolean = (p, x1) match {
  case (p, Nil) => true
  case (p, x :: xs) => p(x) && list_all[A](p, xs)
}

def mat[A](nr: nat, nc: nat, m: List[List[A]]): Boolean =
  equal_nata(size_list[List[A]](m), nc) &&
    list_all[List[A]]((a: List[A]) => vec[A](nr, a), m)

implicit def ord_nat: ord[nat] = new ord[nat] {
  val `Ceta.less_eq` = (a: nat, b: nat) => less_eq_nat(a, b)
  val `Ceta.less` = (a: nat, b: nat) => less_nat(a, b)
}

def impl_of[A : linorder, B](x0: rbt[A, B]): rbta[A, B] = x0 match {
  case Rbt(x) => x
}

def balance[A, B](x0: rbta[A, B], s: A, t: B, x3: rbta[A, B]): rbta[A, B] =
  (x0, s, t, x3) match {
  case (Branch(R, a, w, x, b), s, t, Branch(R, c, y, z, d)) =>
    Branch[A, B](R, Branch[A, B](B, a, w, x, b), s, t,
                  Branch[A, B](B, c, y, z, d))
  case (Branch(R, Branch(R, a, w, x, b), s, t, c), y, z, Emptya) =>
    Branch[A, B](R, Branch[A, B](B, a, w, x, b), s, t,
                  Branch[A, B](B, c, y, z, Emptya))
  case (Branch(R, Branch(R, a, w, x, b), s, t, c), y, z,
         Branch(B, va, vb, vc, vd))
    => Branch[A, B](R, Branch[A, B](B, a, w, x, b), s, t,
                     Branch[A, B](B, c, y, z, Branch[A, B](B, va, vb, vc, vd)))
  case (Branch(R, Emptya, w, x, Branch(R, b, s, t, c)), y, z, Emptya) =>
    Branch[A, B](R, Branch[A, B](B, Emptya, w, x, b), s, t,
                  Branch[A, B](B, c, y, z, Emptya))
  case (Branch(R, Branch(B, va, vb, vc, vd), w, x, Branch(R, b, s, t, c)), y, z,
         Emptya)
    => Branch[A, B](R, Branch[A, B](B, Branch[A, B](B, va, vb, vc, vd), w, x,
                                     b),
                     s, t, Branch[A, B](B, c, y, z, Emptya))
  case (Branch(R, Emptya, w, x, Branch(R, b, s, t, c)), y, z,
         Branch(B, va, vb, vc, vd))
    => Branch[A, B](R, Branch[A, B](B, Emptya, w, x, b), s, t,
                     Branch[A, B](B, c, y, z, Branch[A, B](B, va, vb, vc, vd)))
  case (Branch(R, Branch(B, ve, vf, vg, vh), w, x, Branch(R, b, s, t, c)), y, z,
         Branch(B, va, vb, vc, vd))
    => Branch[A, B](R, Branch[A, B](B, Branch[A, B](B, ve, vf, vg, vh), w, x,
                                     b),
                     s, t,
                     Branch[A, B](B, c, y, z, Branch[A, B](B, va, vb, vc, vd)))
  case (Emptya, w, x, Branch(R, b, s, t, Branch(R, c, y, z, d))) =>
    Branch[A, B](R, Branch[A, B](B, Emptya, w, x, b), s, t,
                  Branch[A, B](B, c, y, z, d))
  case (Branch(B, va, vb, vc, vd), w, x,
         Branch(R, b, s, t, Branch(R, c, y, z, d)))
    => Branch[A, B](R, Branch[A, B](B, Branch[A, B](B, va, vb, vc, vd), w, x,
                                     b),
                     s, t, Branch[A, B](B, c, y, z, d))
  case (Emptya, w, x, Branch(R, Branch(R, b, s, t, c), y, z, Emptya)) =>
    Branch[A, B](R, Branch[A, B](B, Emptya, w, x, b), s, t,
                  Branch[A, B](B, c, y, z, Emptya))
  case (Emptya, w, x,
         Branch(R, Branch(R, b, s, t, c), y, z, Branch(B, va, vb, vc, vd)))
    => Branch[A, B](R, Branch[A, B](B, Emptya, w, x, b), s, t,
                     Branch[A, B](B, c, y, z, Branch[A, B](B, va, vb, vc, vd)))
  case (Branch(B, va, vb, vc, vd), w, x,
         Branch(R, Branch(R, b, s, t, c), y, z, Emptya))
    => Branch[A, B](R, Branch[A, B](B, Branch[A, B](B, va, vb, vc, vd), w, x,
                                     b),
                     s, t, Branch[A, B](B, c, y, z, Emptya))
  case (Branch(B, va, vb, vc, vd), w, x,
         Branch(R, Branch(R, b, s, t, c), y, z, Branch(B, ve, vf, vg, vh)))
    => Branch[A, B](R, Branch[A, B](B, Branch[A, B](B, va, vb, vc, vd), w, x,
                                     b),
                     s, t,
                     Branch[A, B](B, c, y, z, Branch[A, B](B, ve, vf, vg, vh)))
  case (Emptya, s, t, Emptya) => Branch[A, B](B, Emptya, s, t, Emptya)
  case (Emptya, s, t, Branch(B, va, vb, vc, vd)) =>
    Branch[A, B](B, Emptya, s, t, Branch[A, B](B, va, vb, vc, vd))
  case (Emptya, s, t, Branch(v, Emptya, vb, vc, Emptya)) =>
    Branch[A, B](B, Emptya, s, t, Branch[A, B](v, Emptya, vb, vc, Emptya))
  case (Emptya, s, t, Branch(v, Branch(B, ve, vf, vg, vh), vb, vc, Emptya)) =>
    Branch[A, B](B, Emptya, s, t,
                  Branch[A, B](v, Branch[A, B](B, ve, vf, vg, vh), vb, vc,
                                Emptya))
  case (Emptya, s, t, Branch(v, Emptya, vb, vc, Branch(B, vf, vg, vh, vi))) =>
    Branch[A, B](B, Emptya, s, t,
                  Branch[A, B](v, Emptya, vb, vc,
                                Branch[A, B](B, vf, vg, vh, vi)))
  case (Emptya, s, t,
         Branch(v, Branch(B, ve, vj, vk, vl), vb, vc,
                 Branch(B, vf, vg, vh, vi)))
    => Branch[A, B](B, Emptya, s, t,
                     Branch[A, B](v, Branch[A, B](B, ve, vj, vk, vl), vb, vc,
                                   Branch[A, B](B, vf, vg, vh, vi)))
  case (Branch(B, va, vb, vc, vd), s, t, Emptya) =>
    Branch[A, B](B, Branch[A, B](B, va, vb, vc, vd), s, t, Emptya)
  case (Branch(B, va, vb, vc, vd), s, t, Branch(B, ve, vf, vg, vh)) =>
    Branch[A, B](B, Branch[A, B](B, va, vb, vc, vd), s, t,
                  Branch[A, B](B, ve, vf, vg, vh))
  case (Branch(B, va, vb, vc, vd), s, t, Branch(v, Emptya, vf, vg, Emptya)) =>
    Branch[A, B](B, Branch[A, B](B, va, vb, vc, vd), s, t,
                  Branch[A, B](v, Emptya, vf, vg, Emptya))
  case (Branch(B, va, vb, vc, vd), s, t,
         Branch(v, Branch(B, vi, vj, vk, vl), vf, vg, Emptya))
    => Branch[A, B](B, Branch[A, B](B, va, vb, vc, vd), s, t,
                     Branch[A, B](v, Branch[A, B](B, vi, vj, vk, vl), vf, vg,
                                   Emptya))
  case (Branch(B, va, vb, vc, vd), s, t,
         Branch(v, Emptya, vf, vg, Branch(B, vj, vk, vl, vm)))
    => Branch[A, B](B, Branch[A, B](B, va, vb, vc, vd), s, t,
                     Branch[A, B](v, Emptya, vf, vg,
                                   Branch[A, B](B, vj, vk, vl, vm)))
  case (Branch(B, va, vb, vc, vd), s, t,
         Branch(v, Branch(B, vi, vn, vo, vp), vf, vg,
                 Branch(B, vj, vk, vl, vm)))
    => Branch[A, B](B, Branch[A, B](B, va, vb, vc, vd), s, t,
                     Branch[A, B](v, Branch[A, B](B, vi, vn, vo, vp), vf, vg,
                                   Branch[A, B](B, vj, vk, vl, vm)))
  case (Branch(v, Emptya, vb, vc, Emptya), s, t, Emptya) =>
    Branch[A, B](B, Branch[A, B](v, Emptya, vb, vc, Emptya), s, t, Emptya)
  case (Branch(v, Emptya, vb, vc, Branch(B, ve, vf, vg, vh)), s, t, Emptya) =>
    Branch[A, B](B, Branch[A, B](v, Emptya, vb, vc,
                                  Branch[A, B](B, ve, vf, vg, vh)),
                  s, t, Emptya)
  case (Branch(v, Branch(B, vf, vg, vh, vi), vb, vc, Emptya), s, t, Emptya) =>
    Branch[A, B](B, Branch[A, B](v, Branch[A, B](B, vf, vg, vh, vi), vb, vc,
                                  Emptya),
                  s, t, Emptya)
  case (Branch(v, Branch(B, vf, vg, vh, vi), vb, vc, Branch(B, ve, vj, vk, vl)),
         s, t, Emptya)
    => Branch[A, B](B, Branch[A, B](v, Branch[A, B](B, vf, vg, vh, vi), vb, vc,
                                     Branch[A, B](B, ve, vj, vk, vl)),
                     s, t, Emptya)
  case (Branch(v, Emptya, vf, vg, Emptya), s, t, Branch(B, va, vb, vc, vd)) =>
    Branch[A, B](B, Branch[A, B](v, Emptya, vf, vg, Emptya), s, t,
                  Branch[A, B](B, va, vb, vc, vd))
  case (Branch(v, Emptya, vf, vg, Branch(B, vi, vj, vk, vl)), s, t,
         Branch(B, va, vb, vc, vd))
    => Branch[A, B](B, Branch[A, B](v, Emptya, vf, vg,
                                     Branch[A, B](B, vi, vj, vk, vl)),
                     s, t, Branch[A, B](B, va, vb, vc, vd))
  case (Branch(v, Branch(B, vj, vk, vl, vm), vf, vg, Emptya), s, t,
         Branch(B, va, vb, vc, vd))
    => Branch[A, B](B, Branch[A, B](v, Branch[A, B](B, vj, vk, vl, vm), vf, vg,
                                     Emptya),
                     s, t, Branch[A, B](B, va, vb, vc, vd))
  case (Branch(v, Branch(B, vj, vk, vl, vm), vf, vg, Branch(B, vi, vn, vo, vp)),
         s, t, Branch(B, va, vb, vc, vd))
    => Branch[A, B](B, Branch[A, B](v, Branch[A, B](B, vj, vk, vl, vm), vf, vg,
                                     Branch[A, B](B, vi, vn, vo, vp)),
                     s, t, Branch[A, B](B, va, vb, vc, vd))
}

def ins[A : linorder, B](f: A => B => B => B, k: A, v: B, x3: rbta[A, B]):
      rbta[A, B] =
  (f, k, v, x3) match {
  case (f, k, v, Emptya) => Branch[A, B](R, Emptya, k, v, Emptya)
  case (f, k, v, Branch(B, l, x, y, r)) =>
    (if (less[A](k, x)) balance[A, B](ins[A, B](f, k, v, l), x, y, r)
      else (if (less[A](x, k)) balance[A, B](l, x, y, ins[A, B](f, k, v, r))
             else Branch[A, B](B, l, x, ((f(k))(y))(v), r)))
  case (f, k, v, Branch(R, l, x, y, r)) =>
    (if (less[A](k, x)) Branch[A, B](R, ins[A, B](f, k, v, l), x, y, r)
      else (if (less[A](x, k)) Branch[A, B](R, l, x, y, ins[A, B](f, k, v, r))
             else Branch[A, B](R, l, x, ((f(k))(y))(v), r)))
}

def paint[A, B](c: color, x1: rbta[A, B]): rbta[A, B] = (c, x1) match {
  case (c, Emptya) => Emptya
  case (c, Branch(uu, l, k, v, r)) => Branch[A, B](c, l, k, v, r)
}

def insert_with_key[A : linorder,
                     B](f: A => B => B => B, k: A, v: B, t: rbta[A, B]):
      rbta[A, B] =
  paint[A, B](B, ins[A, B](f, k, v, t))

def insertc[A : linorder, B]: A => B => (rbta[A, B]) => rbta[A, B] =
  (a: A) =>
    (b: B) =>
      (c: rbta[A, B]) =>
        insert_with_key[A, B]((_: A) => (_: B) => (nv: B) => nv, a, b, c)

def insert[A : linorder, B](k: A, v: B, t: rbt[A, B]): rbt[A, B] =
  Rbt[A, B](insertc[A, B].apply(k).apply(v).apply(impl_of[A, B](t)))

def lookupc[A : linorder, B](x0: rbta[A, B], k: A): Option[B] = (x0, k) match {
  case (Emptya, k) => None
  case (Branch(uu, l, x, y, r), k) =>
    (if (less[A](k, x)) lookupc[A, B](l, k)
      else (if (less[A](x, k)) lookupc[A, B](r, k) else Some[B](y)))
}

def lookup[A : linorder, B](t: rbt[A, B]): A => Option[B] =
  (a: A) => lookupc[A, B](impl_of[A, B](t), a)

def quotient_of(x0: rat): (BigInt, BigInt) = x0 match {
  case Frct(x) => x
}

def times_rata(p: rat, q: rat): rat =
  Frct({
         val (a, c): (BigInt, BigInt) = quotient_of(p)
         val (b, d): (BigInt, BigInt) = quotient_of(q);
         normalize((a * b, c * d))
       })

implicit def times_rat: times[rat] = new times[rat] {
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
}

implicit def dvd_rat: dvd[rat] = new dvd[rat] {
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
}

def one_rata: rat = Frct((one_inta, one_inta))

implicit def one_rat: one[rat] = new one[rat] {
  val `Ceta.one` = one_rata
}

def less_rat(p: rat, q: rat): Boolean =
  {
    val (a, c): (BigInt, BigInt) = quotient_of(p)
    val (b, d): (BigInt, BigInt) = quotient_of(q);
    a * d < c * b
  }

def less_eq_rat(p: rat, q: rat): Boolean =
  {
    val (a, c): (BigInt, BigInt) = quotient_of(p)
    val (b, d): (BigInt, BigInt) = quotient_of(q);
    a * d <= c * b
  }

implicit def ord_rat: ord[rat] = new ord[rat] {
  val `Ceta.less_eq` = (a: rat, b: rat) => less_eq_rat(a, b)
  val `Ceta.less` = (a: rat, b: rat) => less_rat(a, b)
}

def proja[A, B](x0: term[A, B], i: nat): term[A, B] = (x0, i) match {
  case (Fun(f, ts), i) =>
    (if (less_nat(i, size_list[term[A, B]](ts))) nth[term[A, B]](ts, i)
      else Fun[A, B](f, ts))
}

def member[A](x: A, a: A => Boolean): Boolean = a(x)

def shows_between(l: (List[Char]) => List[Char], p: (List[Char]) => List[Char],
                   r: (List[Char]) => List[Char]):
      (List[Char]) => List[Char] =
  comp[List[Char], List[Char],
        List[Char]](l, comp[List[Char], List[Char], List[Char]](p, r))

def quote(s: (List[Char]) => List[Char]): (List[Char]) => List[Char] =
  shows_between(shows_prec_char(Zero_nat, '\''), s,
                 shows_prec_char(Zero_nat, '\''))

def debug[A](i: List[Char], t: List[Char], x: A): A = x

def fun_upd[A : equal, B](f: A => B, a: A, b: B): A => B =
  (x: A) => (if (eq[A](x, a)) b else f(x))

trait plus[A] {
  val `Ceta.plus`: (A, A) => A
}
def plus[A](a: A, b: A)(implicit A: plus[A]): A = A.`Ceta.plus`(a, b)

trait zero[A] {
  val `Ceta.zero`: A
}
def zero[A](implicit A: zero[A]): A = A.`Ceta.zero`

implicit def plus_int: plus[BigInt] = new plus[BigInt] {
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
}

def zero_inta: BigInt = BigInt(0)

implicit def zero_int: zero[BigInt] = new zero[BigInt] {
  val `Ceta.zero` = zero_inta
}

def filter[A](p: A => Boolean, x1: List[A]): List[A] = (p, x1) match {
  case (p, Nil) => Nil
  case (p, x :: xs) => (if (p(x)) x :: filter[A](p, xs) else filter[A](p, xs))
}

def insertb[A : equal](x: A, xs: List[A]): List[A] =
  (if (membera[A](xs, x)) xs else x :: xs)

abstract sealed class pos
final case object Empty extends pos
final case class PCons(a: nat, b: pos) extends pos

trait power[A] extends one[A] with times[A] {
}

def powera[A](one: A, times: A => A => A, a: A, x3: nat): A =
  (one, times, a, x3) match {
  case (one, times, a, Suc(n)) => (times(a))(powera[A](one, times, a, n))
  case (one, times, a, Zero_nat) => one
}

def power[A : power]: A => nat => A =
  (a: A) =>
    (b: nat) => powera[A](one[A], (aa: A) => (ba: A) => times[A](aa, ba), a, b)

def prc_nat[A](pr: A => nat => nat):
      A => nat => A => nat => (Boolean, Boolean) =
  (f: A) =>
    (n: nat) =>
      (g: A) =>
        (m: nat) =>
          {
            val pf: nat = (pr(f))(n)
            val pg: nat = (pr(g))(m);
            (less_nat(pg, pf), less_eq_nat(pg, pf))
          }

def prl_nat[A](pr: A => nat => nat): A => Boolean =
  (f: A) => equal_nata((pr(f))(Zero_nat), Zero_nat)

def plus_rata(p: rat, q: rat): rat =
  Frct({
         val (a, c): (BigInt, BigInt) = quotient_of(p)
         val (b, d): (BigInt, BigInt) = quotient_of(q);
         normalize((a * d + b * c, c * d))
       })

implicit def plus_rat: plus[rat] = new plus[rat] {
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
}

def zero_rata: rat = Frct((zero_inta, one_inta))

implicit def zero_rat: zero[rat] = new zero[rat] {
  val `Ceta.zero` = zero_rata
}

def lterms[A, B](pi: ((A, nat)) => List[(nat, nat)]):
      (term[A, B]) => List[(term[A, B], nat)] =
  (a: term[A, B]) =>
    {
      val (Fun(f, ts)): term[A, B] = a;
      mapa[(nat, nat),
            (term[A, B],
              nat)]((aa: (nat, nat)) =>
                      {
                        val (i, ab): (nat, nat) = aa;
                        (proja[A, B](Fun[A, B](f, ts), i), ab)
                      },
                     pi((f, size_list[term[A, B]](ts))))
    }

def is_Var[A, B](x0: term[A, B]): Boolean = x0 match {
  case Var(x) => true
  case Fun(f, ts) => false
}

def supteq[A : equal, B : equal]: ((term[A, B], term[A, B])) => Boolean =
  (a: (term[A, B], term[A, B])) =>
    {
      val (x, y): (term[A, B], term[A, B]) = a;
      equal_terma[A, B](x, y) || supt_impl[A, B](x, y)
    }

def fun_of[A : equal, B](vec: List[(A, B)], x: A): B =
  the[B](lookupa[A, B](x, vec))

implicit def equal_int: equal[BigInt] = new equal[BigInt] {
  val `Ceta.equal` = (a: BigInt, b: BigInt) => a == b
}

implicit def power_int: power[BigInt] = new power[BigInt] {
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
  val `Ceta.one` = one_inta
}

abstract sealed class lab[+A, +B]
final case class Lab[A, B](a: lab[A, B], b: B) extends lab[A, B]
final case class FunLab[A, B](a: lab[A, B], b: List[lab[A, B]]) extends
  lab[A, B]
final case class UnLab[A, B](a: A) extends lab[A, B]
final case class Sharp[A, B](a: lab[A, B]) extends lab[A, B]

def remdups[A : equal](x0: List[A]): List[A] = x0 match {
  case Nil => Nil
  case x :: xs =>
    (if (membera[A](xs, x)) remdups[A](xs) else x :: remdups[A](xs))
}

def remove1[A : equal](x: A, xa1: List[A]): List[A] = (x, xa1) match {
  case (x, Nil) => Nil
  case (x, y :: xs) => (if (eq[A](x, y)) xs else y :: remove1[A](x, xs))
}

def it_add[A, B,
            C](update: A => B => C => C,
                iterate: (A => B => C => C) => C => C => C, m1: C, m2: C):
      C =
  ((iterate(update))(m2))(m1)

def vec0I[A](ze: A, n: nat): List[A] = replicate[A](n, ze)

def mat0I[A](ze: A, nr: nat, nc: nat): List[List[A]] =
  replicate[List[A]](nc, vec0I[A](ze, nr))

def minus_nat(m: nat, n: nat): nat = (m, n) match {
  case (Suc(m), Suc(n)) => minus_nat(m, n)
  case (Zero_nat, n) => Zero_nat
  case (m, Zero_nat) => m
}

def vec1I[A](ze: A, on: A, n: nat, i: nat): List[A] =
  replicate[A](i, ze) ++
    (on :: replicate[A](minus_nat(minus_nat(n, one_nat), i), ze))

def mat1I[A](ze: A, on: A, n: nat): List[List[A]] =
  mapa[nat, List[A]]((a: nat) => vec1I[A](ze, on, n, a), upt(Zero_nat, n))

def lm_store_acc[A : equal,
                  B](x0: List[(A, B)], kv: (A, B), accu: List[(A, B)]):
      List[(A, B)] =
  (x0, kv, accu) match {
  case (Nil, kv, accu) => kv :: accu
  case ((a, uu) :: rec, (k, v), accu) =>
    eq[A](k, a) match {
      case true => accu
      case false => lm_store_acc[A, B](rec, (k, v), accu)
    }
}

def lm_store[A : equal, B](m: List[(A, B)], kv: (A, B)): List[(A, B)] =
  lm_store_acc[A, B](m, kv, m)

def l2m_store[A : equal, B : equal,
               C](x0: List[(A, List[(B, C)])], x1: ((A, B), C)):
      List[(A, List[(B, C)])] =
  (x0, x1) match {
  case (Nil, ((k1, k2), v)) => List((k1, List((k2, v))))
  case ((a, kvs) :: rec, ((k1, k2), v)) =>
    eq[A](k1, a) match {
      case true => (a, lm_store[B, C](kvs, (k2, v))) :: rec
      case false => (a, kvs) :: l2m_store[A, B, C](rec, ((k1, k2), v))
    }
}

abstract sealed class memory_ext[+A, +B, +C, +D]
final case class
  Memory_ext[A, B, C,
              D](a: Unit => A, b: A => B => Option[C], c: A => ((B, C)) => A,
                  d: D)
  extends memory_ext[A, B, C, D]

def l2m_lookup[A : equal, B : equal,
                C](x0: List[(A, List[(B, C)])], uu: (A, B)):
      Option[C] =
  (x0, uu) match {
  case (Nil, uu) => None
  case ((a, kvs) :: rec, (k1, k2)) =>
    eq[A](k1, a) match {
      case true => lookupa[B, C](k2, kvs)
      case false => l2m_lookup[A, B, C](rec, (k1, k2))
    }
}

def l2m[A : equal, B : equal, C]:
      memory_ext[List[(A, List[(B, C)])], (A, B), C, Unit] =
  Memory_ext[List[(A, List[(B, C)])], (A, B), C,
              Unit]((_: Unit) => Nil,
                     (a: List[(A, List[(B, C)])]) =>
                       (b: (A, B)) => l2m_lookup[A, B, C](a, b),
                     (a: List[(A, List[(B, C)])]) =>
                       (b: ((A, B), C)) => l2m_store[A, B, C](a, b),
                     ())

def less_eq_prod_aux[A : equal : ord, B : ord](x0: (A, B), x1: (A, B)):
      Boolean =
  (x0, x1) match {
  case ((a1, a2), (b1, b2)) =>
    less[A](a1, b1) || eq[A](a1, b1) && less_eq[B](a2, b2)
}

def equal_proda[A : equal, B : equal](x0: (A, B), x1: (A, B)): Boolean =
  (x0, x1) match {
  case ((aa, ba), (a, b)) => eq[A](aa, a) && eq[B](ba, b)
}

def less_prod[A : equal : ord, B : equal : ord](a: (A, B), b: (A, B)): Boolean =
  ! (equal_proda[A, B](a, b)) && less_eq_prod_aux[A, B](a, b)

def less_eq_prod[A : equal : ord, B : ord](a: (A, B), b: (A, B)): Boolean =
  less_eq_prod_aux[A, B](a, b)

implicit def ord_prod[A : equal : ord, B : equal : ord]: ord[(A, B)] = new
  ord[(A, B)] {
  val `Ceta.less_eq` = (a: (A, B), b: (A, B)) => less_eq_prod[A, B](a, b)
  val `Ceta.less` = (a: (A, B), b: (A, B)) => less_prod[A, B](a, b)
}

def rm_lookup[A : linorder, B](k: A, m: rbt[A, B]): Option[B] =
  (lookup[A, B](m)).apply(k)

def rm_update[A : linorder, B]: A => B => (rbt[A, B]) => rbt[A, B] =
  (a: A) => (b: B) => (c: rbt[A, B]) => insert[A, B](a, b, c)

def rm_update_dj[A : linorder, B]: A => B => (rbt[A, B]) => rbt[A, B] =
  rm_update[A, B]

def aux[A : equal,
         B : linorder](key: A => Option[B], ma: rbt[B, List[A]], v: A,
                        m: rbt[B, List[A]]):
      rbt[B, List[A]] =
  key(v) match {
    case None => m
    case Some(k) =>
      rm_lookup[B, List[A]](k, ma) match {
        case None => m
        case Some(ws) =>
          (if (membera[A](ws, v))
            rm_lookup[B, List[A]](k, m) match {
              case None =>
                rm_update_dj[B, List[A]].apply(k).apply(List(v)).apply(m)
              case Some(vs) =>
                rm_update[B, List[A]].apply(k).apply(insertb[A](v, vs)).apply(m)
            }
            else m)
      }
  }

implicit def equal_nat: equal[nat] = new equal[nat] {
  val `Ceta.equal` = (a: nat, b: nat) => equal_nata(a, b)
}

implicit def preorder_nat: preorder[nat] = new preorder[nat] {
  val `Ceta.less_eq` = (a: nat, b: nat) => less_eq_nat(a, b)
  val `Ceta.less` = (a: nat, b: nat) => less_nat(a, b)
}

implicit def order_nat: order[nat] = new order[nat] {
  val `Ceta.less_eq` = (a: nat, b: nat) => less_eq_nat(a, b)
  val `Ceta.less` = (a: nat, b: nat) => less_nat(a, b)
}

def equal_rata(a: rat, b: rat): Boolean =
  equal_proda[BigInt, BigInt](quotient_of(a), quotient_of(b))

implicit def equal_rat: equal[rat] = new equal[rat] {
  val `Ceta.equal` = (a: rat, b: rat) => equal_rata(a, b)
}

implicit def power_rat: power[rat] = new power[rat] {
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
  val `Ceta.one` = one_rata
}

def key[A, B, C](x0: (A, (term[B, C], term[B, C]))): Option[(B, nat)] = x0 match
  {
  case (uu, (Fun(f, ts), uv)) => Some[(B, nat)]((f, size_list[term[B, C]](ts)))
  case (uw, (Var(ux), uy)) => None
}

trait non_strict_order[A] {
  val `Ceta.ge`: (A, A) => Boolean
}
def ge[A](a: A, b: A)(implicit A: non_strict_order[A]): Boolean =
  A.`Ceta.ge`(a, b)

def shows_sep[A](s: A => (List[Char]) => List[Char],
                  sep: (List[Char]) => List[Char], x2: List[A]):
      (List[Char]) => List[Char] =
  (s, sep, x2) match {
  case (s, sep, Nil) => shows_string.apply(Nil)
  case (s, sep, List(x)) => s(x)
  case (s, sep, x :: v :: va) =>
    comp[List[Char], List[Char],
          List[Char]](s(x),
                       comp[List[Char], List[Char],
                             List[Char]](sep, shows_sep[A](s, sep, v :: va)))
}

def shows_list_gen[A](elt: A => (List[Char]) => List[Char], e: List[Char],
                       l: List[Char], s: List[Char], r: List[Char],
                       xs: List[A]):
      (List[Char]) => List[Char] =
  (if (nulla[A](xs)) shows_string.apply(e)
    else shows_between(shows_string.apply(l),
                        shows_sep[A](elt, shows_string.apply(s), xs),
                        shows_string.apply(r)))

def shows_list_aux[A](s: A => (List[Char]) => List[Char], xs: List[A]):
      (List[Char]) => List[Char] =
  shows_list_gen[A](s, List('[', ']'), List('['), List(','), List(']'), xs)

def divmod_nat(m: nat, n: nat): (nat, nat) =
  (if (equal_nata(n, Zero_nat) || less_nat(m, n)) (Zero_nat, m)
    else {
           val (q, a): (nat, nat) = divmod_nat(minus_nat(m, n), n);
           (Suc(q), a)
         })

def div_nat(m: nat, n: nat): nat = fst[nat, nat](divmod_nat(m, n))

def mod_nat(m: nat, n: nat): nat = snd[nat, nat](divmod_nat(m, n))

def digit2string(n: nat): List[Char] =
  (if (equal_nata(n, Zero_nat)) List('0')
    else (if (equal_nata(n, one_nat)) List('1')
           else (if (equal_nata(n, nat(BigInt(2)))) List('2')
                  else (if (equal_nata(n, nat(BigInt(3)))) List('3')
                         else (if (equal_nata(n, nat(BigInt(4)))) List('4')
                                else (if (equal_nata(n, nat(BigInt(5))))
                                       List('5')
                                       else (if (equal_nata(n, nat(BigInt(6))))
      List('6')
      else (if (equal_nata(n, nat(BigInt(7)))) List('7')
             else (if (equal_nata(n, nat(BigInt(8)))) List('8')
                    else (if (equal_nata(n, nat(BigInt(9)))) List('9')
                           else List('n', 'o', 't', ' ', 'a', ' ', 'd', 'i',
                                      'g', 'i', 't')))))))))))

def shows_nat(n: nat): (List[Char]) => List[Char] =
  (if (less_nat(n, nat(BigInt(10)))) shows_string.apply(digit2string(n))
    else comp[List[Char], List[Char],
               List[Char]](shows_nat(div_nat(n, nat(BigInt(10)))),
                            shows_string.apply(digit2string(mod_nat(n,
                             nat(BigInt(10)))))))

def shows_prec_nat(d: nat, n: nat): (List[Char]) => List[Char] = shows_nat(n)

def shows_prec_int(d: nat, i: BigInt): (List[Char]) => List[Char] =
  (if (i < BigInt(0))
    comp[List[Char], List[Char],
          List[Char]](shows_prec_char(Zero_nat, '-'),
                       shows_prec_nat(Zero_nat, nat(BigInt(0) - i)))
    else shows_prec_nat(Zero_nat, nat(i)))

def shows_list_int(is: List[BigInt]): (List[Char]) => List[Char] =
  shows_list_aux[BigInt]((a: BigInt) => shows_prec_int(Zero_nat, a), is)

implicit def show_int: show[BigInt] = new show[BigInt] {
  val `Ceta.shows_prec` = (a: nat, b: BigInt, c: List[Char]) =>
    (shows_prec_int(a, b)).apply(c)
  val `Ceta.shows_list` = (a: List[BigInt], b: List[Char]) =>
    (shows_list_int(a)).apply(b)
}

def shows_list_nat(ns: List[nat]): (List[Char]) => List[Char] =
  shows_list_aux[nat]((a: nat) => shows_prec_nat(Zero_nat, a), ns)

implicit def show_nat: show[nat] = new show[nat] {
  val `Ceta.shows_prec` = (a: nat, b: nat, c: List[Char]) =>
    (shows_prec_nat(a, b)).apply(c)
  val `Ceta.shows_list` = (a: List[nat], b: List[Char]) =>
    (shows_list_nat(a)).apply(b)
}

def show_rata(x: rat): List[Char] =
  {
    val (den, num): (BigInt, BigInt) = quotient_of(x);
    (if (num == BigInt(1)) (shows_prec_int(Zero_nat, den)).apply(Nil)
      else (comp[List[Char], List[Char],
                  List[Char]](comp[List[Char], List[Char],
                                    List[Char]](shows_prec_int(Zero_nat, den),
         shows_prec_char(Zero_nat, '/')),
                               shows_prec_int(Zero_nat, num))).apply(Nil))
  }

def shows_prec_rat(d: nat, x: rat): (List[Char]) => List[Char] =
  shows_string.apply(show_rata(x))

def shows_list_rat(rs: List[rat]): (List[Char]) => List[Char] =
  shows_list_aux[rat]((a: rat) => shows_prec_rat(Zero_nat, a), rs)

implicit def show_rat: show[rat] = new show[rat] {
  val `Ceta.shows_prec` = (a: nat, b: rat, c: List[Char]) =>
    (shows_prec_rat(a, b)).apply(c)
  val `Ceta.shows_list` = (a: List[rat], b: List[Char]) =>
    (shows_list_rat(a)).apply(b)
}

def subt_at[A, B](s: term[A, B], x1: pos): term[A, B] = (s, x1) match {
  case (s, Empty) => s
  case (Fun(f, ss), PCons(i, p)) => subt_at[A, B](nth[term[A, B]](ss, i), p)
}

def the_Var[A, B](x0: term[A, B]): B = x0 match {
  case Var(x) => x
}

def map_term[A, B, C](fun: A => B, x1: term[A, C]): term[B, C] = (fun, x1) match
  {
  case (fun, Var(x)) => Var[C, B](x)
  case (fun, Fun(f, ts)) =>
    Fun[B, C](fun(f),
               mapa[term[A, C],
                     term[B, C]]((a: term[A, C]) => map_term[A, B, C](fun, a),
                                  ts))
}

def map_rule[A, B, C](fg: A => B, lr: (term[A, C], term[A, C])):
      (term[B, C], term[B, C]) =
  (map_term[A, B, C](fg, fst[term[A, C], term[A, C]](lr)),
    map_term[A, B, C](fg, snd[term[A, C], term[A, C]](lr)))

def extract[A](p: A => Boolean, x1: List[A]): Option[(List[A], (A, List[A]))] =
  (p, x1) match {
  case (p, Nil) => None
  case (p, x :: xs) =>
    (if (p(x)) Some[(List[A], (A, List[A]))]((Nil, (x, xs)))
      else extract[A](p, xs) match {
             case None => None
             case Some((ys, (y, zs))) =>
               Some[(List[A], (A, List[A]))]((x :: ys, (y, zs)))
           })
}

def distinct[A : equal](x0: List[A]): Boolean = x0 match {
  case Nil => true
  case x :: xs => ! (membera[A](xs, x)) && distinct[A](xs)
}

def merge[A, B : linorder](key: A => B, x1: List[A], bs: List[A]): List[A] =
  (key, x1, bs) match {
  case (key, a :: as, b :: bs) =>
    (if (less[B](key(b), key(a))) b :: merge[A, B](key, a :: as, bs)
      else a :: merge[A, B](key, as, b :: bs))
  case (key, Nil, bs) => bs
  case (key, v :: va, Nil) => v :: va
}

def merge_pairs[A, B : linorder](key: A => B, x1: List[List[A]]):
      List[List[A]] =
  (key, x1) match {
  case (key, a :: b :: xs) =>
    merge[A, B](key, a, b) :: merge_pairs[A, B](key, xs)
  case (key, Nil) => Nil
  case (key, List(v)) => List(v)
}

def merge_all[A, B : linorder](key: A => B, x1: List[List[A]]): List[A] =
  (key, x1) match {
  case (key, Nil) => Nil
  case (key, List(x)) => x
  case (key, v :: vb :: vc) =>
    merge_all[A, B](key, merge_pairs[A, B](key, v :: vb :: vc))
}

def asc[A, B : linorder](key: A => B, a: A, f: (List[A]) => List[A],
                          x3: List[A]):
      List[List[A]] =
  (key, a, f, x3) match {
  case (key, a, f, b :: bs) =>
    (if (! (less[B](key(b), key(a))))
      asc[A, B](key, b,
                 comp[List[A], List[A], List[A]](f, (ba: List[A]) => a :: ba),
                 bs)
      else f(List(a)) :: sequences[A, B](key, b :: bs))
  case (key, a, f, Nil) => f(List(a)) :: sequences[A, B](key, Nil)
}

def desc[A, B : linorder](key: A => B, a: A, as: List[A], x3: List[A]):
      List[List[A]] =
  (key, a, as, x3) match {
  case (key, a, as, b :: bs) =>
    (if (less[B](key(b), key(a))) desc[A, B](key, b, a :: as, bs)
      else (a :: as) :: sequences[A, B](key, b :: bs))
  case (key, a, as, Nil) => (a :: as) :: sequences[A, B](key, Nil)
}

def sequences[A, B : linorder](key: A => B, x1: List[A]): List[List[A]] =
  (key, x1) match {
  case (key, a :: b :: xs) =>
    (if (less[B](key(b), key(a))) desc[A, B](key, b, List(a), xs)
      else asc[A, B](key, b, (ba: List[A]) => a :: ba, xs))
  case (key, Nil) => List(Nil)
  case (key, List(v)) => List(List(v))
}

def sort_key[A, B : linorder](key: A => B): (List[A]) => List[A] =
  comp[List[List[A]], List[A],
        List[A]]((a: List[List[A]]) => merge_all[A, B](key, a),
                  (a: List[A]) => sequences[A, B](key, a))

def vec_comp_all[A](r: A => A => Boolean, v: List[A], w: List[A]): Boolean =
  list_all[(A, A)]((a: (A, A)) => {
                                    val (aa, b): (A, A) = a;
                                    (r(aa))(b)
                                  },
                    zip[A, A](v, w))

def mat_comp_all[A](r: A => A => Boolean, m1: List[List[A]], m2: List[List[A]]):
      Boolean =
  list_all[(List[A],
             List[A])]((a: (List[A], List[A])) =>
                         {
                           val (aa, b): (List[A], List[A]) = a;
                           vec_comp_all[A](r, aa, b)
                         },
                        zip[List[A], List[A]](m1, m2))

def mat_ge[A : non_strict_order]:
      (List[List[A]]) => (List[List[A]]) => Boolean =
  (a: List[List[A]]) =>
    (b: List[List[A]]) =>
      mat_comp_all[A]((aa: A) => (ba: A) => ge[A](aa, ba), a, b)

def times_nat(x0: nat, n: nat): nat = (x0, n) match {
  case (Zero_nat, n) => Zero_nat
  case (Suc(m), n) => plus_nat(n, times_nat(m, n))
}

def maxa[A](less_eq: A => A => Boolean, a: A, b: A): A =
  (if ((less_eq(a))(b)) b else a)

def max[A : ord]: A => A => A =
  (a: A) => (b: A) => maxa[A]((aa: A) => (ba: A) => less_eq[A](aa, ba), a, b)

def mina[A](less_eq: A => A => Boolean, a: A, b: A): A =
  (if ((less_eq(a))(b)) a else b)

def min[A : ord]: A => A => A =
  (a: A) => (b: A) => mina[A]((aa: A) => (ba: A) => less_eq[A](aa, ba), a, b)

trait top[A] extends order[A] {
  val `Ceta.top`: A
}
def top[A](implicit A: top[A]): A = A.`Ceta.top`

def minus_rat(p: rat, q: rat): rat =
  Frct({
         val (a, c): (BigInt, BigInt) = quotient_of(p)
         val (b, d): (BigInt, BigInt) = quotient_of(q);
         normalize((a * d - b * c, c * d))
       })

def shows_list_list[A : show](ls: List[List[A]]): (List[Char]) => List[Char] =
  shows_list_aux[List[A]]((a: List[A]) => shows_prec_list[A](Zero_nat, a), ls)

implicit def show_list[A : show]: show[List[A]] = new show[List[A]] {
  val `Ceta.shows_prec` = (a: nat, b: List[A], c: List[Char]) =>
    (shows_prec_list[A](a, b)).apply(c)
  val `Ceta.shows_list` = (a: List[List[A]], b: List[Char]) =>
    (shows_list_list[A](a)).apply(b)
}

def shows_paren(p: (List[Char]) => List[Char]): (List[Char]) => List[Char] =
  comp[List[Char], List[Char],
        List[Char]](shows_prec_char(Zero_nat, '('),
                     comp[List[Char], List[Char],
                           List[Char]](p, shows_prec_char(Zero_nat, ')')))

def shows_prec_prod[A : show, B : show](d: nat, p: (A, B)):
      (List[Char]) => List[Char] =
  shows_paren(comp[List[Char], List[Char],
                    List[Char]]((a: List[Char]) =>
                                  shows_prec[A](Zero_nat, fst[A, B](p), a),
                                 comp[List[Char], List[Char],
                                       List[Char]](shows_prec_char(Zero_nat,
                            ','),
            (a: List[Char]) => shows_prec[B](Zero_nat, snd[A, B](p), a))))

def shows_list_prod[A : show, B : show](ps: List[(A, B)]):
      (List[Char]) => List[Char] =
  shows_list_aux[(A, B)]((a: (A, B)) => shows_prec_prod[A, B](Zero_nat, a), ps)

implicit def show_prod[A : show, B : show]: show[(A, B)] = new show[(A, B)] {
  val `Ceta.shows_prec` = (a: nat, b: (A, B), c: List[Char]) =>
    (shows_prec_prod[A, B](a, b)).apply(c)
  val `Ceta.shows_list` = (a: List[(A, B)], b: List[Char]) =>
    (shows_list_prod[A, B](a)).apply(b)
}

def shows_nl: (List[Char]) => List[Char] = shows_prec_char(Zero_nat, newline)

def map_vars[A, B, C](m: A => B, x1: term[C, A]): term[C, B] = (m, x1) match {
  case (m, Var(x)) => Var[B, C](m(x))
  case (m, Fun(f, ts)) =>
    Fun[C, B](f, mapa[term[C, A],
                       term[C, B]]((a: term[C, A]) => map_vars[A, B, C](m, a),
                                    ts))
}

def num_args[A, B](x0: term[A, B]): nat = x0 match {
  case Var(x) => Zero_nat
  case Fun(f, ts) => size_list[term[A, B]](ts)
}

def unapp[A : equal, B](a: A, x1: term[A, B]): (term[A, B], List[term[A, B]]) =
  (a, x1) match {
  case (a, Var(x)) => (Var[B, A](x), Nil)
  case (a, Fun(f, ss)) =>
    (if (eq[A](f, a) && equal_nata(size_list[term[A, B]](ss), nat(BigInt(2))))
      {
        val (r, ts): (term[A, B], List[term[A, B]]) =
          unapp[A, B](a, nth[term[A, B]](ss, Zero_nat));
        (r, ts ++ List(nth[term[A, B]](ss, one_nat)))
      }
      else (Fun[A, B](f, ss), Nil))
}

def max_list(x0: List[nat]): nat = x0 match {
  case Nil => Zero_nat
  case x :: xs => max[nat].apply(x).apply(max_list(xs))
}

def min_list[A : linorder](x0: List[A]): A = x0 match {
  case List(x) => x
  case x :: v :: va => min[A].apply(x).apply(min_list[A](v :: va))
}

def xml3elements[A, B, C,
                  D](tag: List[Char], p1: xml => sum[List[Char], A],
                      p2: xml => sum[List[Char], B],
                      p3: xml => sum[List[Char], C], f: A => B => C => D,
                      x5: xml):
      sum[List[Char], D] =
  (tag, p1, p2, p3, f, x5) match {
  case (tag, p1, p2, p3, f, Xml(name, atts, cs, text)) =>
    (if (equal_lista[Char](name, tag) &&
           (nulla[(List[Char], List[Char])](atts) &&
             (equal_nata(size_list[xml](cs), nat(BigInt(3))) &&
               is_none[List[Char]](text))))
      bindb[List[Char], A,
             D](p1(nth[xml](cs, Zero_nat)),
                 (x1: A) =>
                   bindb[List[Char], B,
                          D](p2(nth[xml](cs, one_nat)),
                              (x2: B) =>
                                bindb[List[Char], C,
                                       D](p3(nth[xml](cs, nat(BigInt(2)))),
   (x3: C) => Inr[D, List[Char]](((f(x1))(x2))(x3)))))
      else no_transform[D](tag, Xml(name, atts, cs, text)))
}

def position: xml => sum[List[Char], nat] =
  (a: xml) =>
    xml_change[nat, nat]((aa: xml) =>
                           nata(List('p', 'o', 's', 'i', 't', 'i', 'o', 'n'),
                                 aa),
                          (n: nat) => minus_nat(n, one_nat), a)

def xml_many_elements[A, B](tag: List[Char], p: xml => sum[List[Char], A],
                             f: (List[A]) => B, x3: xml):
      sum[List[Char], B] =
  (tag, p, f, x3) match {
  case (tag, p, f, Xml(name, atts, cs, text)) =>
    (if (equal_lista[Char](name, tag) &&
           (nulla[(List[Char], List[Char])](atts) && is_none[List[Char]](text)))
      bindb[List[Char], List[A],
             B](map[A].apply(p).apply(cs),
                 comp[B, sum[List[Char], B],
                       List[A]]((a: B) => Inr[B, List[Char]](a), f))
      else no_transform[B](tag, Xml(name, atts, cs, text)))
}

abstract sealed class af_entry
final case class Collapse(a: nat) extends af_entry
final case class AFList(a: List[nat]) extends af_entry

def afs[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], List[((A, nat), af_entry)]] =
  (a: xml) =>
    xml_many_elements[((A, nat), af_entry),
                       List[((A, nat),
                              af_entry)]](List('a', 'r', 'g', 'u', 'm', 'e',
        'n', 't', 'F', 'i', 'l', 't', 'e', 'r'),
   (aa: xml) =>
     xml3elements[A, nat, af_entry,
                   ((A, nat),
                     af_entry)](List('a', 'r', 'g', 'u', 'm', 'e', 'n', 't',
                                      'F', 'i', 'l', 't', 'e', 'r', 'E', 'n',
                                      't', 'r', 'y'),
                                 xml2name,
                                 (ab: xml) =>
                                   nata(List('a', 'r', 'i', 't', 'y'), ab),
                                 (ab: xml) =>
                                   xml_options[af_entry](List((List('c', 'o',
                             'l', 'l', 'a', 'p', 's', 'i', 'n', 'g'),
                        (ac: xml) =>
                          xml_change[nat, af_entry]((ad: xml) =>
              nata(List('c', 'o', 'l', 'l', 'a', 'p', 's', 'i', 'n', 'g'), ad),
             (i: nat) => Collapse(minus_nat(i, one_nat)), ac)),
                       (List('n', 'o', 'n', 'C', 'o', 'l', 'l', 'a', 'p', 's',
                              'i', 'n', 'g'),
                         (ac: xml) =>
                           xml_many_elements[nat,
      af_entry](List('n', 'o', 'n', 'C', 'o', 'l', 'l', 'a', 'p', 's', 'i', 'n',
                      'g'),
                 position, (ad: List[nat]) => AFList(ad), ac))),
                  ab),
                                 (f: A) =>
                                   (ab: nat) => (b: af_entry) => ((f, ab), b),
                                 aa),
   id[List[((A, nat), af_entry)]], a)

def list2position(x0: List[nat]): pos = x0 match {
  case Nil => Empty
  case n :: ns => PCons(n, list2position(ns))
}

def pos: xml => sum[List[Char], pos] =
  (a: xml) =>
    xml_many_elements[nat, pos](List('p', 'o', 's', 'i', 't', 'i', 'o', 'n',
                                      'I', 'n', 'T', 'e', 'r', 'm'),
                                 position, (aa: List[nat]) => list2position(aa),
                                 a)

def vara[A](x: xml): sum[List[Char], term[A, List[Char]]] =
  bindb[List[Char], List[Char],
         term[A, List[Char]]](text(List('v', 'a', 'r'), x),
                               comp[term[A, List[Char]],
                                     sum[List[Char], term[A, List[Char]]],
                                     List[Char]]((a: term[A, List[Char]]) =>
           Inr[term[A, List[Char]], List[Char]](a),
          (a: List[Char]) => Var[List[Char], A](a)))

def xml1element[A, B](tag: List[Char], p1: xml => sum[List[Char], A], f: A => B,
                       x3: xml):
      sum[List[Char], B] =
  (tag, p1, f, x3) match {
  case (tag, p1, f, Xml(name, atts, cs, text)) =>
    (if (equal_lista[Char](name, tag) &&
           (nulla[(List[Char], List[Char])](atts) &&
             (equal_nata(size_list[xml](cs), one_nat) &&
               is_none[List[Char]](text))))
      bindb[List[Char], A,
             B](p1(nth[xml](cs, Zero_nat)),
                 (x1: A) => Inr[B, List[Char]](f(x1)))
      else no_transform[B](tag, Xml(name, atts, cs, text)))
}

def xml1many_elements_gen[A, B,
                           C](tag: List[Char], p1: xml => sum[List[Char], A],
                               p2: A => xml => sum[List[Char], B],
                               f: A => (List[B]) => C, x4: xml):
      sum[List[Char], C] =
  (tag, p1, p2, f, x4) match {
  case (tag, p1, p2, f, Xml(name, atts, cs, text)) =>
    (if (equal_lista[Char](name, tag) &&
           (nulla[(List[Char], List[Char])](atts) &&
             (is_none[List[Char]](text) && ! (nulla[xml](cs)))))
      bindb[List[Char], A,
             C](p1(hd[xml](cs)),
                 (x: A) =>
                   bindb[List[Char], List[B],
                          C](map[B].apply(p2(x)).apply(tl[xml](cs)),
                              (xs: List[B]) => Inr[C, List[Char]]((f(x))(xs))))
      else no_transform[C](tag, Xml(name, atts, cs, text)))
}

def xml1many_elements[A, B,
                       C](tag: List[Char], p1: xml => sum[List[Char], A],
                           p2: xml => sum[List[Char], B]):
      (A => (List[B]) => C) => xml => sum[List[Char], C] =
  (a: A => (List[B]) => C) =>
    (b: xml) => xml1many_elements_gen[A, B, C](tag, p1, (_: A) => p2, a, b)

def term[A](xml2name: xml => sum[List[Char], A], x: xml):
      sum[List[Char], term[A, List[Char]]] =
  xml_options[term[A, List[Char]]](List((List('v', 'a', 'r'),
  (a: xml) => vara[A](a)),
 (List('f', 'u', 'n', 'a', 'p', 'p'),
   (xml1many_elements[A, term[A, List[Char]],
                       term[A, List[Char]]](List('f', 'u', 'n', 'a', 'p', 'p'),
     xml2name,
     (a: xml) =>
       xml1element[term[A, List[Char]],
                    term[A, List[Char]]](List('a', 'r', 'g'),
  (aa: xml) => term[A](xml2name, aa), id[term[A, List[Char]]],
  a))).apply((a: A) =>
               (b: List[term[A, List[Char]]]) => Fun[A, List[Char]](a, b)))),
                                    x)

def rule[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], (term[A, List[Char]], term[A, List[Char]])] =
  (a: xml) =>
    xml2elements[term[A, List[Char]], term[A, List[Char]],
                  (term[A, List[Char]],
                    term[A, List[Char]])](List('r', 'u', 'l', 'e'),
   (aa: xml) =>
     xml1element[term[A, List[Char]],
                  term[A, List[Char]]](List('l', 'h', 's'),
(ab: xml) => term[A](xml2name, ab), id[term[A, List[Char]]], aa),
   (aa: xml) =>
     xml1element[term[A, List[Char]],
                  term[A, List[Char]]](List('r', 'h', 's'),
(ab: xml) => term[A](xml2name, ab), id[term[A, List[Char]]], aa),
   (aa: term[A, List[Char]]) => (b: term[A, List[Char]]) => (aa, b), a)

def map_filter[A, B](f: A => Option[B], x1: List[A]): List[B] = (f, x1) match {
  case (f, Nil) => Nil
  case (f, x :: xs) =>
    f(x) match {
      case None => map_filter[A, B](f, xs)
      case Some(y) => y :: map_filter[A, B](f, xs)
    }
}

def scg_position: xml => sum[List[Char], nat] =
  (a: xml) => nata(List('p', 'o', 's', 'i', 't', 'i', 'o', 'n'), a)

def scg[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             ((term[A, List[Char]], term[A, List[Char]]),
               (List[(nat, nat)], List[(nat, nat)]))] =
  (xml1many_elements[(term[A, List[Char]], term[A, List[Char]]),
                      (Boolean, (nat, nat)),
                      ((term[A, List[Char]], term[A, List[Char]]),
                        (List[(nat, nat)],
                          List[(nat, nat)]))](List('s', 'i', 'z', 'e', 'C', 'h',
            'a', 'n', 'g', 'e', 'G', 'r', 'a', 'p', 'h'),
       rule[A](xml2name),
       (a: xml) =>
         xml3elements[nat, Boolean, nat,
                       (Boolean,
                         (nat, nat))](List('e', 'd', 'g', 'e'), scg_position,
                                       (aa: xml) =>
 bool(List('s', 't', 'r', 'i', 'c', 't'), aa),
                                       scg_position,
                                       (p: nat) =>
 (s: Boolean) => (q: nat) => (s, (p, q)),
                                       a))).apply((lr:
             (term[A, List[Char]], term[A, List[Char]]))
            => (edges: List[(Boolean, (nat, nat))]) =>
                 (lr, (map_filter[(Boolean, (nat, nat)),
                                   (nat, nat)]((x: (Boolean, (nat, nat))) =>
         (if (fst[Boolean, (nat, nat)](x))
           Some[(nat, nat)](snd[Boolean, (nat, nat)](x)) else None),
        edges),
                        map_filter[(Boolean, (nat, nat)),
                                    (nat, nat)]((x: (Boolean, (nat, nat))) =>
          (if (! (fst[Boolean, (nat, nat)](x)))
            Some[(nat, nat)](snd[Boolean, (nat, nat)](x)) else None),
         edges))))

abstract sealed class gctxt[+A, +B]
final case object GCHole extends gctxt[Nothing, Nothing]
final case class GCFun[A, B](a: A, b: List[gctxt[A, B]]) extends gctxt[A, B]

def guard(b: Boolean): Option[Unit] = (if (b) Some[Unit](()) else None)

def merge_lists[A : equal, B](x0: List[gctxt[A, B]], x1: List[gctxt[A, B]]):
      Option[List[gctxt[A, B]]] =
  (x0, x1) match {
  case (c :: cs, Nil) => None
  case (Nil, d :: ds) => None
  case (GCFun(f, ss) :: cs, GCFun(g, ts) :: ds) =>
    bind[Unit,
          List[gctxt[A, B]]](guard(eq[A](f, g)),
                              (_: Unit) =>
                                bind[List[gctxt[A, B]],
                                      List[gctxt[A,
          B]]](merge_lists[A, B](ss, ts),
                (us: List[gctxt[A, B]]) =>
                  bind[List[gctxt[A, B]],
                        List[gctxt[A, B]]](merge_lists[A, B](cs, ds),
    (es: List[gctxt[A, B]]) =>
      Some[List[gctxt[A, B]]](GCFun[A, B](f, us) :: es))))
  case (c :: cs, GCHole :: ds) =>
    bind[List[gctxt[A, B]],
          List[gctxt[A, B]]](merge_lists[A, B](cs, ds),
                              (es: List[gctxt[A, B]]) =>
                                Some[List[gctxt[A, B]]](c :: es))
  case (GCHole :: cs, d :: ds) =>
    bind[List[gctxt[A, B]],
          List[gctxt[A, B]]](merge_lists[A, B](cs, ds),
                              (es: List[gctxt[A, B]]) =>
                                Some[List[gctxt[A, B]]](d :: es))
  case (Nil, Nil) => Some[List[gctxt[A, B]]](Nil)
}

def mergea[A : equal, B](c: gctxt[A, B], d: gctxt[A, B]): Option[gctxt[A, B]] =
  bind[List[gctxt[A, B]],
        gctxt[A, B]](merge_lists[A, B](List(c), List(d)),
                      (es: List[gctxt[A, B]]) =>
                        Some[gctxt[A, B]](nth[gctxt[A, B]](es, Zero_nat)))

def merge_var[A : equal,
               B : equal](x: A, c: gctxt[B, A], xa2: List[(gctxt[B, A], A)]):
      Option[((gctxt[B, A], A), List[(gctxt[B, A], A)])] =
  (x, c, xa2) match {
  case (x, c, (d, y) :: ps) =>
    (if (eq[A](x, y))
      bind[gctxt[B, A],
            ((gctxt[B, A], A),
              List[(gctxt[B, A],
                     A)])](mergea[B, A](c, d),
                            (e: gctxt[B, A]) => merge_var[A, B](x, e, ps))
      else bind[((gctxt[B, A], A), List[(gctxt[B, A], A)]),
                 ((gctxt[B, A], A),
                   List[(gctxt[B, A],
                          A)])](merge_var[A, B](x, c, ps),
                                 (a: ((gctxt[B, A], A), List[(gctxt[B, A], A)]))
                                   => {
val (b, psa): ((gctxt[B, A], A), List[(gctxt[B, A], A)]) = a;
Some[((gctxt[B, A], A), List[(gctxt[B, A], A)])]((b, (d, y) :: psa))
                                      }))
  case (x, c, Nil) =>
    Some[((gctxt[B, A], A), List[(gctxt[B, A], A)])](((c, x), Nil))
}

def merge_alla[A : equal, B : equal](x0: List[(gctxt[A, B], B)]):
      Option[List[(gctxt[A, B], B)]] =
  x0 match {
  case Nil => Some[List[(gctxt[A, B], B)]](Nil)
  case (c, x) :: ps =>
    bind[((gctxt[A, B], B), List[(gctxt[A, B], B)]),
          List[(gctxt[A, B],
                 B)]](merge_var[B, A](x, c, ps),
                       (a: ((gctxt[A, B], B), List[(gctxt[A, B], B)])) =>
                         {
                           val (cx, psa):
                                 ((gctxt[A, B], B), List[(gctxt[A, B], B)])
                             = a;
                           bind[List[(gctxt[A, B], B)],
                                 List[(gctxt[A, B],
B)]](merge_alla[A, B](psa),
      (psb: List[(gctxt[A, B], B)]) => Some[List[(gctxt[A, B], B)]](cx :: psb))
                         })
}

def zip_option[A, B](x0: List[A], x1: List[B]): Option[List[(A, B)]] = (x0, x1)
  match {
  case (Nil, Nil) => Some[List[(A, B)]](Nil)
  case (x :: xs, y :: ys) =>
    bind[List[(A, B)],
          List[(A, B)]](zip_option[A, B](xs, ys),
                         (zs: List[(A, B)]) => Some[List[(A, B)]]((x, y) :: zs))
  case (x :: xs, Nil) => None
  case (Nil, y :: ys) => None
}

def match_lista[A : equal, B](x0: List[(gctxt[A, B], term[A, B])]):
      Option[List[(gctxt[A, B], B)]] =
  x0 match {
  case (GCFun(f, ss), Var(x)) :: ps =>
    bind[List[(gctxt[A, B], B)],
          List[(gctxt[A, B],
                 B)]](match_lista[A, B](ps),
                       (psa: List[(gctxt[A, B], B)]) =>
                         Some[List[(gctxt[A, B],
                                     B)]]((GCFun[A, B](f, ss), x) :: psa))
  case (GCFun(f, ss), Fun(g, ts)) :: ps =>
    bind[Unit,
          List[(gctxt[A, B],
                 B)]](guard(eq[A](f, g)),
                       (_: Unit) =>
                         bind[List[(gctxt[A, B], term[A, B])],
                               List[(gctxt[A, B],
                                      B)]](zip_option[gctxt[A, B],
               term[A, B]](ss, ts),
    (psa: List[(gctxt[A, B], term[A, B])]) => match_lista[A, B](psa ++ ps)))
  case (GCHole, t) :: ps => match_lista[A, B](ps)
  case Nil => Some[List[(gctxt[A, B], B)]](Nil)
}

def matchd[A : equal, B : equal](x0: (gctxt[A, B], term[A, B])):
      Option[List[(gctxt[A, B], B)]] =
  x0 match {
  case (c, t) =>
    bind[List[(gctxt[A, B], B)],
          List[(gctxt[A, B],
                 B)]](match_lista[A, B](List((c, t))),
                       (a: List[(gctxt[A, B], B)]) => merge_alla[A, B](a))
}

def matchc[A : equal, B : equal](c: gctxt[A, B], t: term[A, B]): Boolean =
  ! (is_none[List[(gctxt[A, B], B)]](matchd[A, B]((c, t))))

def capRM2[A : equal,
            B : equal](uu: ((A, nat)) => List[(term[A, B], term[A, B])],
                        x1: term[A, B]):
      gctxt[A, B] =
  (uu, x1) match {
  case (uu, Var(uv)) => GCHole
  case (rm, Fun(f, ts)) =>
    {
      val h: gctxt[A, B] =
        GCFun[A, B](f, mapa[term[A, B],
                             gctxt[A, B]]((a: term[A, B]) =>
    capRM2[A, B](rm, a),
   ts))
      val n: nat = size_list[term[A, B]](ts);
      (if (list_ex[(term[A, B],
                     term[A, B])]((r: (term[A, B], term[A, B])) =>
                                    matchc[A,
    B](h, fst[term[A, B], term[A, B]](r)),
                                   rm((f, n))))
        GCHole else h)
    }
}

def capRM[A : equal,
           B : equal](nlv: Boolean,
                       rm: ((A, nat)) => List[(term[A, B], term[A, B])]):
      (term[A, B]) => gctxt[A, B] =
  (if (nlv) (a: term[A, B]) => capRM2[A, B](rm, a)
    else (_: term[A, B]) => GCHole)

def vars_term_lista[A, B](x0: term[A, B]): List[B] = x0 match {
  case Var(x) => List(x)
  case Fun(uu, ts) =>
    maps[term[A, B], B]((a: term[A, B]) => vars_term_lista[A, B](a), ts)
}

def vars_term_list[A, B : equal]: (term[A, B]) => List[B] =
  comp[List[B], List[B],
        term[A, B]]((a: List[B]) => remdups[B](a),
                     (a: term[A, B]) => vars_term_lista[A, B](a))

def varsL[A, B : equal](list: List[(term[A, B], term[A, B])], x: B): Boolean =
  foldr[(term[A, B], term[A, B]),
         Boolean]((eq: (term[A, B], term[A, B])) =>
                    (b: Boolean) =>
                      membera[B](vars_term_list[A,
         B].apply(fst[term[A, B], term[A, B]](eq)),
                                  x) ||
                        (membera[B](vars_term_list[A,
            B].apply(snd[term[A, B], term[A, B]](eq)),
                                     x) ||
                          b),
                   list, false)

def decompose[A : equal, B](s: term[A, B], t: term[A, B]):
      Option[List[(term[A, B], term[A, B])]] =
  (s, t) match {
    case (Var(_), b) => None
    case (Fun(f, ss), Var(_)) => None
    case (Fun(f, ss), Fun(g, ts)) =>
      (if (eq[A](f, g)) zip_option[term[A, B], term[A, B]](ss, ts) else None)
  }

def simple_subst[A : equal, B](x: A, t: term[B, A], xa2: term[B, A]):
      term[B, A] =
  (x, t, xa2) match {
  case (x, t, Var(y)) => (if (eq[A](x, y)) t else Var[A, B](y))
  case (x, t, Fun(f, ss)) =>
    Fun[B, A](f, mapa[term[B, A],
                       term[B, A]]((a: term[B, A]) =>
                                     simple_subst[A, B](x, t, a),
                                    ss))
}

def simple_subst_list[A : equal,
                       B](x: A, u: term[B, A],
                           list: List[(term[B, A], term[B, A])]):
      List[(term[B, A], term[B, A])] =
  mapa[(term[B, A], term[B, A]),
        (term[B, A],
          term[B, A])]((p: (term[B, A], term[B, A])) =>
                         (simple_subst[A,
B](x, u, fst[term[B, A], term[B, A]](p)),
                           simple_subst[A,
 B](x, u, snd[term[B, A], term[B, A]](p))),
                        list)

def unify[A : equal,
           B : equal](x0: List[(term[A, B], term[A, B])],
                       seen: List[(term[A, B], term[A, B])]):
      Option[List[(term[A, B], term[A, B])]] =
  (x0, seen) match {
  case (Nil, seen) => Some[List[(term[A, B], term[A, B])]](seen)
  case ((Fun(f, ss), Fun(g, ts)) :: unifs, seen) =>
    decompose[A, B](Fun[A, B](f, ss), Fun[A, B](g, ts)) match {
      case None => None
      case Some(eqs) => unify[A, B](eqs ++ unifs, seen)
    }
  case ((Fun(f, ss), Var(x)) :: unifs, seen) =>
    unify[A, B]((Var[B, A](x), Fun[A, B](f, ss)) :: unifs, seen)
  case ((Var(x), t) :: unifs, seen) =>
    (if (equal_terma[A, B](t, Var[B, A](x))) unify[A, B](unifs, seen)
      else (if (membera[B](vars_term_list[A, B].apply(t), x)) None
             else (if (varsL[A, B](unifs ++ seen, x))
                    unify[A, B](simple_subst_list[B, A](x, t, unifs ++ seen) ++
                                  List((Var[B, A](x), t)),
                                 Nil)
                    else unify[A, B](unifs, (Var[B, A](x), t) :: seen))))
}

def subst_of[A, B : equal](x0: List[(term[A, B], term[A, B])]):
      B => term[A, B] =
  x0 match {
  case Nil => (a: B) => Var[B, A](a)
  case (Var(x), t) :: ts =>
    (y: B) => (if (eq[B](x, y)) t else (subst_of[A, B](ts)).apply(y))
}

def mgu[A : equal, B : equal](s: term[A, B], t: term[A, B]):
      Option[B => term[A, B]] =
  unify[A, B](List((s, t)), Nil) match {
    case None => None
    case Some(res) => Some[B => term[A, B]](subst_of[A, B](res))
  }

def class_to_term_intern[A, B](iv: nat => A, i: nat, x2: term[B, sum[Unit, A]]):
      (nat, term[B, A]) =
  (iv, i, x2) match {
  case (iv, i, Fun(f, ts)) =>
    {
      val (k, ss): (nat, List[term[B, A]]) =
        foldr[term[B, sum[Unit, A]],
               (nat, List[term[B, A]])]((t: term[B, sum[Unit, A]]) =>
  (a: (nat, List[term[B, A]])) =>
    {
      val (j, ss): (nat, List[term[B, A]]) = a
      val (k, s): (nat, term[B, A]) = class_to_term_intern[A, B](iv, j, t);
      (k, s :: ss)
    },
 ts, (i, Nil));
      (k, Fun[B, A](f, ss))
    }
  case (iv, i, Var(Inl(uu))) => (plus_nat(i, one_nat), Var[A, B](iv(i)))
  case (iv, i, Var(Inr(x))) => (i, Var[A, B](x))
}

def class_to_term[A](c: Char, t: term[A, sum[Unit, List[Char]]]):
      term[A, List[Char]] =
  snd[nat, term[A, List[Char]]](class_to_term_intern[List[Char],
              A]((i: nat) => c :: (shows_prec_nat(Zero_nat, i)).apply(Nil),
                  Zero_nat, t))

def mgu_class[A : equal](cs: term[A, sum[Unit, List[Char]]],
                          t: term[A, List[Char]]):
      Option[(List[Char]) => term[A, List[Char]]] =
  mgu[A, List[Char]](class_to_term[A]('z', cs),
                      map_vars[List[Char], List[Char],
                                A]((a: List[Char]) => 'y' :: a, t))

def partition[A](p: A => Boolean, x1: List[A]): (List[A], List[A]) = (p, x1)
  match {
  case (p, Nil) => (Nil, Nil)
  case (p, x :: xs) =>
    {
      val (yes, no): (List[A], List[A]) = partition[A](p, xs);
      (if (p(x)) (x :: yes, no) else (yes, x :: no))
    }
}

def removeAll[A : equal](x: A, xa1: List[A]): List[A] = (x, xa1) match {
  case (x, Nil) => Nil
  case (x, y :: xs) =>
    (if (eq[A](x, y)) removeAll[A](x, xs) else y :: removeAll[A](x, xs))
}

def sub_vec[A]: nat => (List[A]) => List[A] =
  (a: nat) => (b: List[A]) => take[A](a, b)

def sub_mat[A](nr: nat, nc: nat, m: List[List[A]]): List[List[A]] =
  mapa[List[A], List[A]](sub_vec[A].apply(nr), take[List[A]](nc, m))

def vec_pre_gtI[A](gt: A => A => Boolean, v: List[A], w: List[A]): Boolean =
  list_ex[(A, A)]((a: (A, A)) => {
                                   val (aa, b): (A, A) = a;
                                   (gt(aa))(b)
                                 },
                   zip[A, A](v, w))

def mat_pre_gtI[A](gt: A => A => Boolean, m1: List[List[A]], m2: List[List[A]]):
      Boolean =
  list_ex[(List[A],
            List[A])]((a: (List[A], List[A])) =>
                        {
                          val (aa, b): (List[A], List[A]) = a;
                          vec_pre_gtI[A](gt, aa, b)
                        },
                       zip[List[A], List[A]](m1, m2))

def mat_gtI[A : non_strict_order](gt: A => A => Boolean, sd: nat,
                                   m1: List[List[A]], m2: List[List[A]]):
      Boolean =
  mat_ge[A].apply(m1).apply(m2) &&
    mat_pre_gtI[A](gt, sub_mat[A](sd, sd, m1), sub_mat[A](sd, sd, m2))

def vec_map[A]: (A => A) => (List[A]) => List[A] =
  (a: A => A) => (b: List[A]) => mapa[A, A](a, b)

def mat_map[A](f: A => A): (List[List[A]]) => List[List[A]] =
  (a: List[List[A]]) => mapa[List[A], List[A]](vec_map[A].apply(f), a)

def emptya[A, B, C, D](x0: memory_ext[A, B, C, D]): Unit => A = x0 match {
  case Memory_ext(empty, lookup, store, more) => empty
}

def store[A, B, C, D](x0: memory_ext[A, B, C, D]): A => ((B, C)) => A = x0 match
  {
  case Memory_ext(empty, lookup, store, more) => store
}

implicit def
  preorder_prod[A : equal : order, B : equal : order]: preorder[(A, B)] = new
  preorder[(A, B)] {
  val `Ceta.less_eq` = (a: (A, B), b: (A, B)) => less_eq_prod[A, B](a, b)
  val `Ceta.less` = (a: (A, B), b: (A, B)) => less_prod[A, B](a, b)
}

implicit def order_prod[A : equal : order, B : equal : order]: order[(A, B)] =
  new order[(A, B)] {
  val `Ceta.less_eq` = (a: (A, B), b: (A, B)) => less_eq_prod[A, B](a, b)
  val `Ceta.less` = (a: (A, B), b: (A, B)) => less_prod[A, B](a, b)
}

trait semigroup_mult[A] extends times[A] {
}

trait semigroup_add[A] extends plus[A] {
}

trait ab_semigroup_add[A] extends semigroup_add[A] {
}

trait semiring[A] extends ab_semigroup_add[A] with semigroup_mult[A] {
}

trait monoid_add[A] extends semigroup_add[A] with zero[A] {
}

trait ordered_ab_semigroup[A]
  extends ab_semigroup_add[A] with non_strict_order[A] {
}

trait max_ordered_monoid_add[A]
  extends monoid_add[A] with ordered_ab_semigroup[A] {
  val `Ceta.max0`: A => A
}
def max0[A](a: A)(implicit A: max_ordered_monoid_add[A]): A = A.`Ceta.max0`(a)

def s_ins[A, B, C](update: A => Unit => B => C, x: A, s: B): C =
  ((update(x))(()))(s)

def sharp_term[A, B](shp: A => A, x1: term[A, B]): term[A, B] = (shp, x1) match
  {
  case (shp, Var(x)) => Var[B, A](x)
  case (shp, Fun(f, ss)) => Fun[A, B](shp(f), ss)
}

def aarity[A](sm: A => nat => List[A], f: A, n: nat): nat =
  minus_nat(size_list[A]((sm(f))(n)), one_nat)

def list_diff[A : equal](x0: List[A], ys: List[A]): List[A] = (x0, ys) match {
  case (Nil, ys) => Nil
  case (x :: xs, ys) =>
    {
      val zs: List[A] = list_diff[A](xs, ys);
      (if (membera[A](ys, x)) zs else x :: zs)
    }
}

def expecting[A : show, B](msg: List[Char], ts: List[A]):
      sum[List[Char], (B, List[A])] =
  Inl[List[Char],
       (B, List[A])](List('e', 'x', 'p', 'e', 'c', 't', 'i', 'n', 'g', ':',
                           ' ') ++
                       (msg ++
                         (List(',', ' ', 'b', 'u', 't', ' ', 'f', 'o', 'u', 'n',
                                'd', ':', ' ') ++
                           (quote(shows_prec_list[A](Zero_nat,
              take[A](nat(BigInt(30)), ts)))).apply(Nil))))

def eoi[A : show](x0: List[A]): sum[List[Char], (Unit, List[A])] = x0 match {
  case Nil => Inr[(Unit, List[A]), List[Char]](((), Nil))
  case v :: va =>
    expecting[A, Unit](List('e', 'n', 'd', ' ', 'o', 'f', ' ', 'i', 'n', 'p',
                             'u', 't'),
                        v :: va)
}

def xml_choice[A](e: List[Char], xa1: List[xml => sum[List[Char], A]], x: xml):
      sum[List[Char], A] =
  (e, xa1, x) match {
  case (e, Nil, x) =>
    Inl[List[Char],
         A](concat[Char](List(List('e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ',
                                    'p', 'a', 'r', 's', 'i', 'n', 'g', ' ', 'c',
                                    'h', 'o', 'i', 'c', 'e', ' ', 'f', 'o', 'r',
                                    ' '),
                               e, List(newline),
                               (shows_prec_xml(Zero_nat, x)).apply(Nil))))
  case (e, p :: ps, x) =>
    p(x) match {
      case Inl(_) => xml_choice[A](e, ps, x)
      case Inr(a) => Inr[A, List[Char]](a)
    }
}

abstract sealed class alter[+A, +B, +C, +D, +E, +F, +G, +H, +I, +J]
final case class Alter1[A, B, C, D, E, F, G, H, I, J](a: A) extends
  alter[A, B, C, D, E, F, G, H, I, J]
final case class Alter2[B, A, C, D, E, F, G, H, I, J](a: B) extends
  alter[A, B, C, D, E, F, G, H, I, J]
final case class Alter3[C, A, B, D, E, F, G, H, I, J](a: C) extends
  alter[A, B, C, D, E, F, G, H, I, J]
final case class Alter4[D, A, B, C, E, F, G, H, I, J](a: D) extends
  alter[A, B, C, D, E, F, G, H, I, J]
final case class Alter5[E, A, B, C, D, F, G, H, I, J](a: E) extends
  alter[A, B, C, D, E, F, G, H, I, J]
final case class Alter6[F, A, B, C, D, E, G, H, I, J](a: F) extends
  alter[A, B, C, D, E, F, G, H, I, J]
final case class Alter7[G, A, B, C, D, E, F, H, I, J](a: G) extends
  alter[A, B, C, D, E, F, G, H, I, J]
final case class Alter8[H, A, B, C, D, E, F, G, I, J](a: H) extends
  alter[A, B, C, D, E, F, G, H, I, J]
final case class Alter9[I, A, B, C, D, E, F, G, H, J](a: I) extends
  alter[A, B, C, D, E, F, G, H, I, J]
final case class Alter10[J, A, B, C, D, E, F, G, H, I](a: J) extends
  alter[A, B, C, D, E, F, G, H, I, J]

def xml0elements[A](tag: List[Char], x: A, xa2: xml): sum[List[Char], A] =
  (tag, x, xa2) match {
  case (tag, x, Xml(name, atts, cs, text)) =>
    (if (equal_lista[Char](name, tag) &&
           (nulla[(List[Char], List[Char])](atts) &&
             (nulla[xml](cs) && is_none[List[Char]](text))))
      Inr[A, List[Char]](x)
      else no_transform[A](tag, Xml(name, atts, cs, text)))
}

def xml4elements[A, B, C, D,
                  E](tag: List[Char], p1: xml => sum[List[Char], A],
                      p2: xml => sum[List[Char], B],
                      p3: xml => sum[List[Char], C],
                      p4: xml => sum[List[Char], D], f: A => B => C => D => E,
                      x6: xml):
      sum[List[Char], E] =
  (tag, p1, p2, p3, p4, f, x6) match {
  case (tag, p1, p2, p3, p4, f, Xml(name, atts, cs, text)) =>
    (if (equal_lista[Char](name, tag) &&
           (nulla[(List[Char], List[Char])](atts) &&
             (equal_nata(size_list[xml](cs), nat(BigInt(4))) &&
               is_none[List[Char]](text))))
      bindb[List[Char], A,
             E](p1(nth[xml](cs, Zero_nat)),
                 (x1: A) =>
                   bindb[List[Char], B,
                          E](p2(nth[xml](cs, one_nat)),
                              (x2: B) =>
                                bindb[List[Char], C,
                                       E](p3(nth[xml](cs, nat(BigInt(2)))),
   (x3: C) =>
     bindb[List[Char], D,
            E](p4(nth[xml](cs, nat(BigInt(3)))),
                (x4: D) => Inr[E, List[Char]]((((f(x1))(x2))(x3))(x4))))))
      else no_transform[E](tag, Xml(name, atts, cs, text)))
}

def ctxt[A](xml2name: xml => sum[List[Char], A], x: xml):
      sum[List[Char], ctxt[A, List[Char]]] =
  xml_options[ctxt[A, List[Char]]](List((List('b', 'o', 'x'),
  (a: xml) => xml0elements[ctxt[A, List[Char]]](List('b', 'o', 'x'), Hole, a)),
 (List('f', 'u', 'n', 'C', 'o', 'n', 't', 'e', 'x', 't'),
   (a: xml) =>
     xml4elements[A, List[term[A, List[Char]]], ctxt[A, List[Char]],
                   List[term[A, List[Char]]],
                   ctxt[A, List[Char]]](List('f', 'u', 'n', 'C', 'o', 'n', 't',
      'e', 'x', 't'),
 xml2name,
 (aa: xml) =>
   xml_many_elements[term[A, List[Char]],
                      List[term[A, List[Char]]]](List('b', 'e', 'f', 'o', 'r',
               'e'),
          (ab: xml) => term[A](xml2name, ab), id[List[term[A, List[Char]]]],
          aa),
 (aa: xml) => ctxt[A](xml2name, aa),
 (aa: xml) =>
   xml_many_elements[term[A, List[Char]],
                      List[term[A, List[Char]]]](List('a', 'f', 't', 'e', 'r'),
          (ab: xml) => term[A](xml2name, ab), id[List[term[A, List[Char]]]],
          aa),
 (aa: A) =>
   (b: List[term[A, List[Char]]]) =>
     (c: ctxt[A, List[Char]]) =>
       (d: List[term[A, List[Char]]]) => More[A, List[Char]](aa, b, c, d),
 a))),
                                    x)

def subst[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], List[(List[Char], term[A, List[Char]])]] =
  (a: xml) =>
    xml_many_elements[(List[Char], term[A, List[Char]]),
                       List[(List[Char],
                              term[A, List[Char]])]](List('s', 'u', 'b', 's',
                   't', 'i', 't', 'u', 't', 'i', 'o', 'n'),
              (aa: xml) =>
                xml2elements[List[Char], term[A, List[Char]],
                              (List[Char],
                                term[A, List[Char]])](List('s', 'u', 'b', 's',
                    't', 'E', 'n', 't', 'r', 'y'),
               (ab: xml) => text(List('v', 'a', 'r'), ab),
               (ab: xml) => term[A](xml2name, ab),
               (ab: List[Char]) => (b: term[A, List[Char]]) => (ab, b), aa),
              id[List[(List[Char], term[A, List[Char]])]], a)

def xml3to4elements[A, B, C, D,
                     E](tag: List[Char], p1: xml => sum[List[Char], A],
                         p2: xml => sum[List[Char], B],
                         p3: xml => sum[List[Char], C],
                         p4: xml => sum[List[Char], D],
                         f: A => B => Option[C] => D => E, x6: xml):
      sum[List[Char], E] =
  (tag, p1, p2, p3, p4, f, x6) match {
  case (tag, p1, p2, p3, p4, f, Xml(name, atts, cs, text)) =>
    {
      val l: nat = size_list[xml](cs);
      (if (equal_lista[Char](name, tag) &&
             (nulla[(List[Char], List[Char])](atts) &&
               (less_eq_nat(nat(BigInt(3)), l) &&
                 (less_eq_nat(l, nat(BigInt(4))) &&
                   is_none[List[Char]](text)))))
        bindb[List[Char], A,
               E](p1(nth[xml](cs, Zero_nat)),
                   (x1: A) =>
                     bindb[List[Char], B,
                            E](p2(nth[xml](cs, one_nat)),
                                (x2: B) =>
                                  (if (equal_nata(l, nat(BigInt(4))))
                                    bindb[List[Char], C,
   E](p3(nth[xml](cs, nat(BigInt(2)))),
       (x3: C) =>
         bindb[List[Char], D,
                E](p4(nth[xml](cs, nat(BigInt(3)))),
                    (x4: D) =>
                      Inr[E, List[Char]]((((f(x1))(x2))(Some[C](x3)))(x4))))
                                    else bindb[List[Char], D,
        E](p4(nth[xml](cs, nat(BigInt(2)))),
            (x4: D) => Inr[E, List[Char]]((((f(x1))(x2))(None))(x4))))))
        else no_transform[E](tag, Xml(name, atts, cs, text)))
    }
}

def relstep[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             (pos, ((term[A, List[Char]], term[A, List[Char]]),
                     (Boolean, term[A, List[Char]])))] =
  (a: xml) =>
    xml3to4elements[pos, (term[A, List[Char]], term[A, List[Char]]), Unit,
                     term[A, List[Char]],
                     (pos, ((term[A, List[Char]], term[A, List[Char]]),
                             (Boolean,
                               term[A, List[Char]])))](List('r', 'e', 'w', 'r',
                     'i', 't', 'e', 'S', 't', 'e', 'p'),
                pos, rule[A](xml2name),
                (aa: xml) =>
                  xml0elements[Unit](List('r', 'e', 'l', 'a', 't', 'i', 'v',
   'e'),
                                      (), aa),
                (aa: xml) => term[A](xml2name, aa),
                (p: pos) =>
                  (r: (term[A, List[Char]], term[A, List[Char]])) =>
                    (rel: Option[Unit]) =>
                      (t: term[A, List[Char]]) =>
                        (p, (r, (is_none[Unit](rel), t))),
                a)

def relsteps[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             (term[A, List[Char]],
               List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                            (Boolean, term[A, List[Char]])))])] =
  (xml1many_elements[term[A, List[Char]],
                      (pos, ((term[A, List[Char]], term[A, List[Char]]),
                              (Boolean, term[A, List[Char]]))),
                      (term[A, List[Char]],
                        List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                     (Boolean,
                                       term[A,
     List[Char]])))])](List('r', 'e', 'w', 'r', 'i', 't', 'e', 'S', 'e', 'q',
                             'u', 'e', 'n', 'c', 'e'),
                        (a: xml) =>
                          xml1element[term[A, List[Char]],
                                       term[A,
     List[Char]]](List('s', 't', 'a', 'r', 't', 'T', 'e', 'r', 'm'),
                   (aa: xml) => term[A](xml2name, aa), id[term[A, List[Char]]],
                   a),
                        relstep[A](xml2name))).apply((a: term[A, List[Char]]) =>
               (b: List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                (Boolean, term[A, List[Char]])))])
                 => (a, b))

def loop[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             (term[A, List[Char]],
               (List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                             (Boolean, term[A, List[Char]])))],
                 (List[(List[Char], term[A, List[Char]])],
                   ctxt[A, List[Char]])))] =
  (a: xml) =>
    xml3elements[(term[A, List[Char]],
                   List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                (Boolean, term[A, List[Char]])))]),
                  List[(List[Char], term[A, List[Char]])], ctxt[A, List[Char]],
                  (term[A, List[Char]],
                    (List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                  (Boolean, term[A, List[Char]])))],
                      (List[(List[Char], term[A, List[Char]])],
                        ctxt[A, List[Char]])))](List('l', 'o', 'o', 'p'),
         relsteps[A](xml2name), subst[A](xml2name),
         (aa: xml) => ctxt[A](xml2name, aa),
         (aa: (term[A, List[Char]],
                List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                             (Boolean, term[A, List[Char]])))]))
           => {
                val (s, rseq):
                      (term[A, List[Char]],
                        List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                     (Boolean, term[A, List[Char]])))])
                  = aa;
                (sigma: List[(List[Char], term[A, List[Char]])]) =>
                  (c: ctxt[A, List[Char]]) => (s, (rseq, (sigma, c)))
              },
         a)

abstract sealed class projL[+A]
final case class Projection[A](a: List[((A, nat), nat)]) extends projL[A]

def proj[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], projL[A]] =
  (a: xml) =>
    xml_change[List[((A, nat), af_entry)],
                projL[A]](afs[A](xml2name),
                           (afl: List[((A, nat), af_entry)]) =>
                             Projection[A](mapa[((A, nat), af_entry),
         ((A, nat),
           nat)]((aa: ((A, nat), af_entry)) =>
                   aa match {
                     case (fa, Collapse(ab)) => (fa, ab)
                     case (fa, AFList(_)) => (fa, snd[A, nat](fa))
                   },
                  afl)),
                           a)

def rstep[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             (pos, ((term[A, List[Char]], term[A, List[Char]]),
                     term[A, List[Char]]))] =
  (a: xml) =>
    xml3elements[pos, (term[A, List[Char]], term[A, List[Char]]),
                  term[A, List[Char]],
                  (pos, ((term[A, List[Char]], term[A, List[Char]]),
                          term[A, List[Char]]))](List('r', 'e', 'w', 'r', 'i',
               't', 'e', 'S', 't', 'e', 'p'),
          pos, rule[A](xml2name), (aa: xml) => term[A](xml2name, aa),
          (p: pos) =>
            (r: (term[A, List[Char]], term[A, List[Char]])) =>
              (t: term[A, List[Char]]) => (p, (r, t)),
          a)

def start_term[A : equal](xml2name: xml => sum[List[Char], A],
                           t: term[A, List[Char]]):
      xml => sum[List[Char], term[A, List[Char]]] =
  (a: xml) =>
    xml1element[term[A, List[Char]],
                 term[A, List[Char]]](List('s', 't', 'a', 'r', 't', 'T', 'e',
    'r', 'm'),
                                       (xml: xml) =>
 bindb[List[Char], term[A, List[Char]],
        term[A, List[Char]]](term[A](xml2name, xml),
                              (s: term[A, List[Char]]) =>
                                (if (equal_terma[A, List[Char]](s, t))
                                  Inr[term[A, List[Char]], List[Char]](t)
                                  else Inl[List[Char],
    term[A, List[Char]]](List('<', 's', 't', 'a', 'r', 't', 'T', 'e', 'r', 'm',
                               '>', ' ', 'd', 'o', 'e', 's', ' ', 'n', 'o', 't',
                               ' ', 'm', 'a', 't', 'c', 'h', ' ', 'l', 'h',
                               's')))),
                                       id[term[A, List[Char]]], a)

def proj_term[A, B](p: ((A, nat)) => nat, x1: term[A, B]): term[A, B] = (p, x1)
  match {
  case (p, Var(x)) => Var[B, A](x)
  case (p, Fun(f, ts)) =>
    {
      val n: nat = size_list[term[A, B]](ts)
      val i: nat = p((f, n));
      (if (less_nat(i, n)) nth[term[A, B]](ts, i) else Fun[A, B](f, ts))
    }
}

def rseq[A : equal](xml2name: xml => sum[List[Char], A], pi: ((A, nat)) => nat,
                     r: (term[A, List[Char]], term[A, List[Char]])):
      xml =>
        sum[List[Char],
             ((term[A, List[Char]], term[A, List[Char]]),
               List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                            term[A, List[Char]]))])] =
  (xml1many_elements[term[A, List[Char]],
                      (pos, ((term[A, List[Char]], term[A, List[Char]]),
                              term[A, List[Char]])),
                      ((term[A, List[Char]], term[A, List[Char]]),
                        List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                     term[A,
   List[Char]]))])](List('r', 'e', 'w', 'r', 'i', 't', 'e', 'S', 'e', 'q', 'u',
                          'e', 'n', 'c', 'e'),
                     start_term[A](xml2name,
                                    proj_term[A,
       List[Char]](pi, fst[term[A, List[Char]], term[A, List[Char]]](r))),
                     rstep[A](xml2name))).apply((_: term[A, List[Char]]) =>
          (a: List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                           term[A, List[Char]]))])
            => (r, a))

def match_vars[A : equal, B : equal,
                C : equal](w: term[A, B], x1: List[(C, term[A, B])]):
      Option[C => term[A, B]] =
  (w, x1) match {
  case (w, Nil) => Some[C => term[A, B]]((_: C) => w)
  case (w, (x, t) :: ps) =>
    lookupa[C, term[A, B]](x, ps) match {
      case None =>
        match_vars[A, B, C](w, ps) match {
          case None => None
          case Some(sigma) =>
            Some[C => term[A, B]](fun_upd[C, term[A, B]](sigma, x, t))
        }
      case Some(s) =>
        (if (equal_terma[A, B](t, s)) match_vars[A, B, C](w, ps) else None)
    }
}

def match_term_list[A : equal, B, C](x0: List[(term[A, B], term[A, C])]):
      Option[List[(C, term[A, B])]] =
  x0 match {
  case Nil => Some[List[(C, term[A, B])]](Nil)
  case (Var(uu), Fun(uv, uw)) :: ps => None
  case (Fun(f, ts), Fun(g, ls)) :: ps =>
    bind[Unit,
          List[(C, term[A, B])]](guard(eq[A](f, g)),
                                  (_: Unit) =>
                                    bind[List[(term[A, B], term[A, C])],
  List[(C, term[A, B])]](zip_option[term[A, B], term[A, C]](ts, ls),
                          (eqs: List[(term[A, B], term[A, C])]) =>
                            match_term_list[A, B, C](eqs ++ ps)))
  case (t, Var(x)) :: ps =>
    bind[List[(C, term[A, B])],
          List[(C, term[A, B])]](match_term_list[A, B, C](ps),
                                  (eqs: List[(C, term[A, B])]) =>
                                    Some[List[(C, term[A, B])]]((x, t) :: eqs))
}

def match_list[A : equal, B : equal,
                C : equal](w: term[A, B], tls: List[(term[A, B], term[A, C])]):
      Option[C => term[A, B]] =
  match_term_list[A, B, C](tls) match {
    case None => None
    case Some(a) => match_vars[A, B, C](w, a)
  }

def matchb[A : equal, B : equal, C : equal](t: term[A, B], l: term[A, C]):
      Option[C => term[A, B]] =
  match_list[A, B, C](t, List((t, l)))

def matches[A : equal, B : equal, C : equal](t: term[A, B], l: term[A, C]):
      Boolean =
  matchb[A, B, C](t, l) match {
    case None => false
    case Some(_) => true
  }

def supteq_list[A, B](x0: term[A, B]): List[term[A, B]] = x0 match {
  case Var(x) => List(Var[B, A](x))
  case Fun(f, ts) =>
    Fun[A, B](f, ts) ::
      maps[term[A, B], term[A, B]]((a: term[A, B]) => supteq_list[A, B](a), ts)
}

def rm_alpha[A : linorder, B]: (rbt[A, B]) => A => Option[B] =
  (a: rbt[A, B]) => lookup[A, B](a)

def rm_set_lookup[A : linorder, B](rm: rbt[A, List[B]]): A => List[B] =
  (a: A) =>
    rm_alpha[A, List[B]].apply(rm).apply(a) match {
      case None => Nil
      case Some(rules) => rules
    }

def is_NF[A : equal : linorder,
           B : equal](var_cond: Boolean, r: List[term[A, B]],
                       m: rbt[A, List[term[A, B]]]):
      (term[A, B]) => Boolean =
  (if (var_cond) (_: term[A, B]) => false
    else (if (nulla[term[A, B]](r)) (_: term[A, B]) => true
           else (t: term[A, B]) =>
                  list_all[term[A, B]]((u: term[A, B]) =>
 list_all[term[A, B]]((l: term[A, B]) => ! (matches[A, B, B](u, l)),
                       (rm_set_lookup[A, term[A,
       B]](m)).apply(fst[A, nat](the[(A, nat)](root[A, B](u))))),
supteq_list[A, B](t))))

implicit def semigroup_mult_int: semigroup_mult[BigInt] = new
  semigroup_mult[BigInt] {
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
}

implicit def semigroup_add_int: semigroup_add[BigInt] = new
  semigroup_add[BigInt] {
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
}

implicit def ab_semigroup_add_int: ab_semigroup_add[BigInt] = new
  ab_semigroup_add[BigInt] {
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
}

implicit def semiring_int: semiring[BigInt] = new semiring[BigInt] {
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
}

def label[A, B](f: lab[A, B], n: nat, x2: sum[B, List[lab[A, B]]]): lab[A, B] =
  (f, n, x2) match {
  case (f, n, Inl(l)) => Lab[A, B](f, l)
  case (f, n, Inr(l)) => FunLab[A, B](f, l)
}

def mat_max0[A : max_ordered_monoid_add]: (List[List[A]]) => List[List[A]] =
  mat_map[A]((a: A) => max0[A](a))

def lookupb[A, B, C, D](x0: memory_ext[A, B, C, D]): A => B => Option[C] = x0
  match {
  case Memory_ext(empty, lookup, store, more) => lookup
}

def rito_map_to_sorted_list[A, B,
                             C](iterate:
                                  (A => B => (List[(A, B)]) => List[(A, B)]) =>
                                    C => (List[(A, B)]) => List[(A, B)],
                                 m: C):
      List[(A, B)] =
  ((iterate((u: A) => (v: B) => (a: List[(A, B)]) => (u, v) :: a))(m))(Nil)

def itoi_reverse_iterateo[A, B, C,
                           D](it: (A => Boolean) =>
                                    (B => C => A => A) => D => A => A):
      (B => C => A => A) => D => A => A =
  it((_: A) => true)

def rm_reverse_iterateoi[A, B,
                          C](c: A => Boolean, f: B => C => A => A,
                              x2: rbta[B, C], sigma: A):
      A =
  (c, f, x2, sigma) match {
  case (c, f, Emptya, sigma) => sigma
  case (c, f, Branch(col, l, k, v, r), sigma) =>
    (if (c(sigma))
      {
        val sigmaa: A = rm_reverse_iterateoi[A, B, C](c, f, r, sigma);
        (if (c(sigmaa))
          rm_reverse_iterateoi[A, B, C](c, f, l, ((f(k))(v))(sigmaa))
          else sigmaa)
      }
      else sigma)
}

def rm_reverse_iterateoia[A, B : linorder,
                           C](c: A => Boolean, f: B => C => A => A,
                               r: rbt[B, C]):
      A => A =
  (a: A) => rm_reverse_iterateoi[A, B, C](c, f, impl_of[B, C](r), a)

def rm_reverse_iterateo[A : linorder, B, C]:
      (A => B => C => C) => (rbt[A, B]) => C => C =
  itoi_reverse_iterateo[C, A, B,
                         rbt[A, B]]((a: C => Boolean) =>
                                      (b: A => B => C => C) =>
(c: rbt[A, B]) => rm_reverse_iterateoia[C, A, B](a, b, c))

def rm_to_list[A : linorder, B]: (rbt[A, B]) => List[(A, B)] =
  (a: rbt[A, B]) =>
    rito_map_to_sorted_list[A, B,
                             rbt[A, B]](rm_reverse_iterateo[A, B, List[(A, B)]],
 a)

def values[A : linorder, B](m: rbt[A, List[B]]): List[B] =
  maps[(A, List[B]),
        B]((a: (A, List[B])) => snd[A, List[B]](a),
            rm_to_list[A, List[B]].apply(m))

implicit def linorder_nat: linorder[nat] = new linorder[nat] {
  val `Ceta.less_eq` = (a: nat, b: nat) => less_eq_nat(a, b)
  val `Ceta.less` = (a: nat, b: nat) => less_nat(a, b)
}

abstract sealed class tpoly[+A, +B]
final case class PVar[A, B](a: A) extends tpoly[A, B]
final case class PNum[B, A](a: B) extends tpoly[A, B]
final case class PSum[A, B](a: List[tpoly[A, B]]) extends tpoly[A, B]
final case class PMult[A, B](a: List[tpoly[A, B]]) extends tpoly[A, B]

implicit def semigroup_mult_rat: semigroup_mult[rat] = new semigroup_mult[rat] {
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
}

implicit def semigroup_add_rat: semigroup_add[rat] = new semigroup_add[rat] {
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
}

implicit def ab_semigroup_add_rat: ab_semigroup_add[rat] = new
  ab_semigroup_add[rat] {
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
}

implicit def semiring_rat: semiring[rat] = new semiring[rat] {
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
}

trait mult_zero[A] extends times[A] with zero[A] {
}

def s_memb[A, B, C](lookup: A => B => Option[C], x: A, s: B): Boolean =
  ! (is_none[C]((lookup(x))(s)))

def ctxt_apply[A, B](x0: ctxt[A, B], s: term[A, B]): term[A, B] = (x0, s) match
  {
  case (Hole, s) => s
  case (More(f, ss1, c, ss2), s) =>
    Fun[A, B](f, ss1 ++ (ctxt_apply[A, B](c, s) :: ss2))
}

def equal_ctxt[A : equal, B : equal](x0: ctxt[A, B], x1: ctxt[A, B]): Boolean =
  (x0, x1) match {
  case (More(f, list1, ctxt, list2), Hole) => false
  case (Hole, More(f, list1, ctxt, list2)) => false
  case (More(fa, list1a, ctxta, list2a), More(f, list1, ctxt, list2)) =>
    eq[A](fa, f) &&
      (equal_lista[term[A, B]](list1a, list1) &&
        (equal_ctxt[A, B](ctxta, ctxt) &&
          equal_lista[term[A, B]](list2a, list2)))
  case (Hole, Hole) => true
}

def hvf_top[A : equal, B](a: A, n: nat, x2: term[A, B]): Boolean = (a, n, x2)
  match {
  case (a, n, Fun(f, ts)) =>
    (if (eq[A](f, a) && equal_nata(size_list[term[A, B]](ts), n))
      ! (is_Var[A, B](hd[term[A, B]](ts))) else true)
  case (a, n, Var(uu)) => false
}

def list_inter[A : equal](x0: List[A], bs: List[A]): List[A] = (x0, bs) match {
  case (Nil, bs) => Nil
  case (a :: as, bs) =>
    (if (membera[A](bs, a)) a :: list_inter[A](as, bs)
      else list_inter[A](as, bs))
}

def list_union[A : equal](x0: List[A], ys: List[A]): List[A] = (x0, ys) match {
  case (Nil, ys) => ys
  case (x :: xs, ys) =>
    {
      val zs: List[A] = list_union[A](xs, ys);
      (if (membera[A](zs, x)) zs else x :: zs)
    }
}

def binda[A, B,
           C](m: (List[A]) => sum[List[Char], (B, List[A])],
               f: B => (List[A]) => sum[List[Char], (C, List[A])], ts: List[A]):
      sum[List[Char], (C, List[A])] =
  bindb[List[Char], (B, List[A]),
         (C, List[A])](m(ts),
                        (a: (B, List[A])) =>
                          {
                            val (aa, b): (B, List[A]) = a;
                            (f(aa))(b)
                          })

def many(p: Char => Boolean, x1: List[Char]):
      sum[List[Char], (List[Char], List[Char])] =
  (p, x1) match {
  case (p, Nil) => Inr[(List[Char], List[Char]), List[Char]]((Nil, Nil))
  case (p, t :: ts) =>
    (if (p(t))
      bindb[List[Char], (List[Char], List[Char]),
             (List[Char],
               List[Char])](many(p, ts),
                             (a: (List[Char], List[Char])) =>
                               {
                                 val (rs, tsa): (List[Char], List[Char]) = a;
                                 Inr[(List[Char], List[Char]),
                                      List[Char]]((t :: rs, tsa))
                               })
      else Inr[(List[Char], List[Char]), List[Char]]((Nil, t :: ts)))
}

def wspace: List[Char] = List(newline, '\11', '\15', ' ')

def trim(x0: List[Char]): List[Char] = x0 match {
  case Nil => Nil
  case c :: s => (if (membera[Char](wspace, c)) trim(s) else c :: s)
}

abstract sealed class domain
final case class Naturala(a: nat) extends domain
final case object Arctic extends domain
final case object Arctic_rat extends domain
final case class Int_mat(a: nat, b: nat) extends domain
final case class Arc_mat(a: nat) extends domain
final case class Arc_rat_mat(a: nat) extends domain
final case class Rational(a: rat, b: nat) extends domain
final case class Rat_mat(a: nat, b: nat) extends domain

def estep[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             (pos, ((term[A, List[Char]], term[A, List[Char]]),
                     (Boolean, term[A, List[Char]])))] =
  (a: xml) =>
    xml4elements[pos, (term[A, List[Char]], term[A, List[Char]]), Boolean,
                  term[A, List[Char]],
                  (pos, ((term[A, List[Char]], term[A, List[Char]]),
                          (Boolean,
                            term[A, List[Char]])))](List('e', 'q', 'u', 'a',
                  't', 'i', 'o', 'n', 'S', 't', 'e', 'p'),
             pos, rule[A](xml2name),
             (aa: xml) =>
               xml_options[Boolean](List((List('l', 'e', 'f', 't', 'R', 'i',
        'g', 'h', 't'),
   (ab: xml) =>
     xml0elements[Boolean](List('l', 'e', 'f', 't', 'R', 'i', 'g', 'h', 't'),
                            true, ab)),
  (List('r', 'i', 'g', 'h', 't', 'L', 'e', 'f', 't'),
    (ab: xml) =>
      xml0elements[Boolean](List('r', 'i', 'g', 'h', 't', 'L', 'e', 'f', 't'),
                             false, ab))),
                                     aa),
             (aa: xml) => term[A](xml2name, aa),
             (p: pos) =>
               (r: (term[A, List[Char]], term[A, List[Char]])) =>
                 (b: Boolean) => (t: term[A, List[Char]]) => (p, (r, (b, t))),
             a)

def labela[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], sum[List[nat], List[A]]] =
  (a: xml) =>
    xml_options[sum[List[nat],
                     List[A]]](List((List('n', 'u', 'm', 'b', 'e', 'r', 'L',
   'a', 'b', 'e', 'l'),
                                      (aa: xml) =>
xml_many_elements[nat, sum[List[nat],
                            List[A]]](List('n', 'u', 'm', 'b', 'e', 'r', 'L',
    'a', 'b', 'e', 'l'),
                                       (ab: xml) =>
 nata(List('n', 'u', 'm', 'b', 'e', 'r'), ab),
                                       (ab: List[nat]) =>
 Inl[List[nat], List[A]](ab),
                                       aa)),
                                     (List('s', 'y', 'm', 'b', 'o', 'l', 'L',
    'a', 'b', 'e', 'l'),
                                       (aa: xml) =>
 xml_many_elements[A, sum[List[nat],
                           List[A]]](List('s', 'y', 'm', 'b', 'o', 'l', 'L',
   'a', 'b', 'e', 'l'),
                                      xml2name,
                                      (ab: List[A]) =>
Inr[List[A], List[nat]](ab),
                                      aa))),
                                a)

def rules[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], List[(term[A, List[Char]], term[A, List[Char]])]] =
  (a: xml) =>
    xml_many_elements[(term[A, List[Char]], term[A, List[Char]]),
                       List[(term[A, List[Char]],
                              term[A, List[Char]])]](List('r', 'u', 'l', 'e',
                   's'),
              rule[A](xml2name),
              id[List[(term[A, List[Char]], term[A, List[Char]])]], a)

def state: xml => sum[List[Char], List[Char]] =
  (a: xml) => text(List('s', 't', 'a', 't', 'e'), a)

def xml_lab[A : show, B : show](x0: lab[A, List[B]]): xml = x0 match {
  case UnLab(x) =>
    Xml(List('n', 'a', 'm', 'e'), Nil, Nil,
         Some[List[Char]](shows_prec[A](Zero_nat, x, Nil)))
  case Sharp(x) =>
    Xml(List('s', 'h', 'a', 'r', 'p'), Nil, List(xml_lab[A, B](x)), None)
  case FunLab(x, l) =>
    Xml(List('l', 'a', 'b', 'e', 'l', 'e', 'd', 'S', 'y', 'm', 'b', 'o', 'l'),
         Nil,
         List(xml_lab[A, B](x),
               Xml(List('s', 'y', 'm', 'b', 'o', 'l', 'L', 'a', 'b', 'e', 'l'),
                    Nil,
                    mapa[lab[A, List[B]],
                          xml]((a: lab[A, List[B]]) => xml_lab[A, B](a), l),
                    None)),
         None)
  case Lab(x, l) =>
    Xml(List('l', 'a', 'b', 'e', 'l', 'e', 'd', 'S', 'y', 'm', 'b', 'o', 'l'),
         Nil,
         List(xml_lab[A, B](x),
               Xml(List('n', 'u', 'm', 'b', 'e', 'r', 'L', 'a', 'b', 'e', 'l'),
                    Nil,
                    mapa[B, xml]((n: B) =>
                                   Xml(List('n', 'u', 'm', 'b', 'e', 'r'), Nil,
Nil, Some[List[Char]](shows_prec[B](Zero_nat, n, Nil))),
                                  l),
                    None)),
         None)
}

def xml_term[A : show, B : show, C : show](x0: term[lab[A, List[B]], C]): xml =
  x0 match {
  case Var(x) =>
    Xml(List('v', 'a', 'r'), Nil, Nil,
         Some[List[Char]](shows_prec[C](Zero_nat, x, Nil)))
  case Fun(f, ts) =>
    Xml(List('f', 'u', 'n', 'a', 'p', 'p'), Nil,
         xml_lab[A, B](f) ::
           mapa[term[lab[A, List[B]], C],
                 xml]((t: term[lab[A, List[B]], C]) =>
                        Xml(List('a', 'r', 'g'), Nil,
                             List(xml_term[A, B, C](t)), None),
                       ts),
         None)
}

abstract sealed class strategy[+A, +B]
final case object No_Strategy extends strategy[Nothing, Nothing]
final case object Innermost extends strategy[Nothing, Nothing]
final case class Innermost_Q[A, B](a: List[term[A, B]]) extends strategy[A, B]

def xml_Q[A : show, B : show, C : show](x0: strategy[lab[A, List[B]], C]):
      List[xml] =
  x0 match {
  case No_Strategy => Nil
  case Innermost =>
    List(Xml(List('s', 't', 'r', 'a', 't', 'e', 'g', 'y'), Nil,
              List(Xml(List('i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't'), Nil,
                        Nil, None)),
              None))
  case Innermost_Q(q) =>
    List(Xml(List('s', 't', 'r', 'a', 't', 'e', 'g', 'y'), Nil,
              List(Xml(List('i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', 'L',
                             'h', 's', 's'),
                        Nil,
                        mapa[term[lab[A, List[B]], C],
                              xml]((a: term[lab[A, List[B]], C]) =>
                                     xml_term[A, B, C](a),
                                    q),
                        None)),
              None))
}

def isOK[A, B](m: sum[A, B]): Boolean =
  m match {
    case Inl(_) => false
    case Inr(_) => true
  }

implicit def mult_zero_int: mult_zero[BigInt] = new mult_zero[BigInt] {
  val `Ceta.zero` = zero_inta
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
}

def list_update[A](x0: List[A], i: nat, y: A): List[A] = (x0, i, y) match {
  case (x :: xs, Suc(i), y) => x :: list_update[A](xs, i, y)
  case (x :: xs, Zero_nat, y) => y :: xs
  case (Nil, i, y) => Nil
}

def base[A](x0: (A, nat)): A = x0 match {
  case (f, h) => f
}

def lift[A](h: nat, f: A): (A, nat) = (f, h)

def roof[A, B : equal](x0: (term[A, B], term[A, B])): (term[A, B]) => Boolean =
  x0 match {
  case (l, r) =>
    {
      val xs: List[B] = vars_term_list[A, B].apply(r);
      (t: term[A, B]) =>
        {
          val xt: List[B] = vars_term_list[A, B].apply(t);
          list_all[B]((a: B) => membera[B](xt, a), xs)
        }
    }
}

def mat_monoI[A](geq1: A => Boolean, sd: nat, m: List[List[A]]): Boolean =
  list_all[List[A]]((a: List[A]) => list_ex[A](geq1, a), sub_mat[A](sd, sd, m))

def transpose[A](nr: nat, x1: List[List[A]]): List[List[A]] = (nr, x1) match {
  case (nr, Nil) => replicate[List[A]](nr, Nil)
  case (nr, v :: m) =>
    mapa[(A, List[A]),
          List[A]]((a: (A, List[A])) =>
                     {
                       val (aa, b): (A, List[A]) = a;
                       aa :: b
                     },
                    zip[A, List[A]](v, transpose[A](nr, m)))
}

def scalar_prodI[A](ze: A, pl: A => A => A, ti: A => A => A, v: List[A],
                     w: List[A]):
      A =
  foldr[(A, A),
         A]((a: (A, A)) => {
                             val (x, y): (A, A) = a;
                             pl((ti(x))(y))
                           },
             zip[A, A](v, w), ze)

def matT_vec_multI[A](ze: A, pl: A => A => A, ti: A => A => A, m: List[List[A]],
                       v: List[A]):
      List[A] =
  mapa[List[A], A]((w: List[A]) => scalar_prodI[A](ze, pl, ti, w, v), m)

def mat_multI[A](ze: A, pl: A => A => A, ti: A => A => A, nr: nat,
                  m1: List[List[A]], m2: List[List[A]]):
      List[List[A]] =
  mapa[List[A],
        List[A]]((a: List[A]) =>
                   matT_vec_multI[A](ze, pl, ti, transpose[A](nr, m1), a),
                  m2)

def vec_plusI[A](pl: A => A => A, v: List[A], w: List[A]): List[A] =
  mapa[(A, A),
        A]((xy: (A, A)) => (pl(fst[A, A](xy)))(snd[A, A](xy)), zip[A, A](v, w))

def mat_plusI[A](pl: A => A => A, m1: List[List[A]], m2: List[List[A]]):
      List[List[A]] =
  mapa[(List[A], List[A]),
        List[A]]((uv: (List[A], List[A])) =>
                   vec_plusI[A](pl, fst[List[A], List[A]](uv),
                                 snd[List[A], List[A]](uv)),
                  zip[List[A], List[A]](m1, m2))

def equal_color(x0: color, x1: color): Boolean = (x0, x1) match {
  case (B, R) => false
  case (R, B) => false
  case (B, B) => true
  case (R, R) => true
}

def equal_rbtb[A : equal, B : equal](x0: rbta[A, B], x1: rbta[A, B]): Boolean =
  (x0, x1) match {
  case (Branch(color, rbt1, a, b, rbt2), Emptya) => false
  case (Emptya, Branch(color, rbt1, a, b, rbt2)) => false
  case (Branch(colora, rbt1a, aa, ba, rbt2a), Branch(color, rbt1, a, b, rbt2))
    => equal_color(colora, color) &&
         (equal_rbtb[A, B](rbt1a, rbt1) &&
           (eq[A](aa, a) && (eq[B](ba, b) && equal_rbtb[A, B](rbt2a, rbt2))))
  case (Emptya, Emptya) => true
}

def equal_rbta[A : equal : linorder, B : equal](ra: rbt[A, B], r: rbt[A, B]):
      Boolean =
  equal_rbtb[A, B](impl_of[A, B](ra), impl_of[A, B](r))

implicit def equal_rbt[A : equal : linorder, B : equal]: equal[rbt[A, B]] = new
  equal[rbt[A, B]] {
  val `Ceta.equal` = (a: rbt[A, B], b: rbt[A, B]) => equal_rbta[A, B](a, b)
}

implicit def mult_zero_rat: mult_zero[rat] = new mult_zero[rat] {
  val `Ceta.zero` = zero_rata
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
}

trait comm_monoid_add[A] extends ab_semigroup_add[A] with monoid_add[A] {
}

trait semiring_0[A]
  extends comm_monoid_add[A] with mult_zero[A] with semiring[A] {
}

trait monoid_mult[A] extends semigroup_mult[A] with power[A] {
}

trait zero_neq_one[A] extends one[A] with zero[A] {
}

trait semiring_1[A]
  extends monoid_mult[A] with semiring_0[A] with zero_neq_one[A] {
}

def fun_ofa[A, B : equal](defa: A, xs: List[(B, A)], a: B): A =
  lookupa[B, A](a, xs) match {
    case None => defa
    case Some(b) => b
  }

def s_empty[A](empt: A): A = empt

def s_union[A](add: A): A = add

def linear_term_impl[A : equal, B](xs: List[A], x1: term[B, A]):
      Option[List[A]] =
  (xs, x1) match {
  case (xs, Var(x)) => (if (membera[A](xs, x)) None else Some[List[A]](x :: xs))
  case (xs, Fun(uu, Nil)) => Some[List[A]](xs)
  case (xs, Fun(f, t :: ts)) =>
    linear_term_impl[A, B](xs, t) match {
      case None => None
      case Some(ys) => linear_term_impl[A, B](ys, Fun[B, A](f, ts))
    }
}

def linear_term[A, B : equal](t: term[A, B]): Boolean =
  linear_term_impl[B, A](Nil, t) match {
    case None => false
    case Some(_) => true
  }

def map_term_wa[A, B, C](fg: A => nat => B, x1: term[A, C]): term[B, C] =
  (fg, x1) match {
  case (fg, Var(x)) => Var[C, B](x)
  case (fg, Fun(f, ts)) =>
    Fun[B, C]((fg(f))(size_list[term[A, C]](ts)),
               mapa[term[A, C],
                     term[B, C]]((a: term[A, C]) => map_term_wa[A, B, C](fg, a),
                                  ts))
}

abstract sealed class ta_rule[+A, +B]
final case class TA_rule[B, A](a: B, b: List[A], c: A) extends ta_rule[A, B]

def rq[A, B](x0: ta_rule[A, B]): A = x0 match {
  case TA_rule(f, qs, q) => q
}

def poss_list[A, B](x0: term[A, B]): List[pos] = x0 match {
  case Var(x) => List(Empty)
  case Fun(f, ss) =>
    Empty ::
      maps[(nat, List[pos]),
            pos]((a: (nat, List[pos])) =>
                   {
                     val (i, aa): (nat, List[pos]) = a;
                     mapa[pos, pos]((ab: pos) => PCons(i, ab), aa)
                   },
                  zip[nat, List[pos]](upt(Zero_nat, size_list[term[A, B]](ss)),
                                       mapa[term[A, B],
     List[pos]]((a: term[A, B]) => poss_list[A, B](a), ss)))
}

def ctxt_of_pos_term[A, B](x0: pos, t: term[A, B]): ctxt[A, B] = (x0, t) match {
  case (Empty, t) => Hole
  case (PCons(i, ps), Fun(f, ts)) =>
    More[A, B](f, take[term[A, B]](i, ts),
                ctxt_of_pos_term[A, B](ps, nth[term[A, B]](ts, i)),
                drop[term[A, B]](Suc(i), ts))
}

def subst_apply_term[A, B, C](x0: term[A, B], sigma: B => term[A, C]):
      term[A, C] =
  (x0, sigma) match {
  case (Var(x), sigma) => sigma(x)
  case (Fun(f, ss), sigma) =>
    Fun[A, C](f, mapa[term[A, B],
                       term[A, C]]((t: term[A, B]) =>
                                     subst_apply_term[A, B, C](t, sigma),
                                    ss))
}

def rewrite[A : equal,
             B : equal](r: List[(term[A, B], term[A, B])], s: term[A, B]):
      List[term[A, B]] =
  maps[(term[A, B], term[A, B]),
        term[A, B]]((a: (term[A, B], term[A, B])) =>
                      {
                        val (l, ra): (term[A, B], term[A, B]) = a;
                        maps[pos, term[A,
B]]((p: pos) =>
      matchb[A, B, B](subt_at[A, B](s, p), l) match {
        case None => Nil
        case Some(sigma) =>
          List(ctxt_apply[A, B](ctxt_of_pos_term[A, B](p, s),
                                 subst_apply_term[A, B, B](ra, sigma)))
      },
     poss_list[A, B](s))
                      },
                     r)

def hvf_term[A : equal, B](a: A, t: term[A, B]): Boolean =
  unapp[A, B](a, t) match {
    case (Var(_), ts) => nulla[term[A, B]](ts)
    case (Fun(_, us), ts) =>
      list_all[term[A, B]]((b: term[A, B]) => hvf_term[A, B](a, b), us ++ ts)
  }

def oneof_aux(allowed: List[List[Char]], x1: List[List[Char]], ts: List[Char]):
      sum[List[Char], (List[Char], List[Char])] =
  (allowed, x1, ts) match {
  case (allowed, Nil, ts) =>
    expecting[Char,
               List[Char]](List('o', 'n', 'e', ' ', 'o', 'f', ' ') ++
                             (shows_list_list[Char](allowed)).apply(Nil),
                            ts)
  case (allowed, x :: xs, ts) =>
    (if (equal_lista[Char](mapa[(Char, Char),
                                 Char]((a: (Char, Char)) => snd[Char, Char](a),
zip[Char, Char](x, ts)),
                            x))
      Inr[(List[Char], List[Char]),
           List[Char]]((x, trim(drop[Char](size_list[Char](x), ts))))
      else oneof_aux(allowed, xs, ts))
}

def oneof(xs: List[List[Char]]):
      (List[Char]) => sum[List[Char], (List[Char], List[Char])] =
  (a: List[Char]) => oneof_aux(xs, xs, a)

abstract sealed class xmldoc
final case class Xmldoc(a: List[Char], b: xml) extends xmldoc

def xml5elements[A, B, C, D, E,
                  F](tag: List[Char], p1: xml => sum[List[Char], A],
                      p2: xml => sum[List[Char], B],
                      p3: xml => sum[List[Char], C],
                      p4: xml => sum[List[Char], D],
                      p5: xml => sum[List[Char], E],
                      f: A => B => C => D => E => F, x7: xml):
      sum[List[Char], F] =
  (tag, p1, p2, p3, p4, p5, f, x7) match {
  case (tag, p1, p2, p3, p4, p5, f, Xml(name, atts, cs, text)) =>
    (if (equal_lista[Char](name, tag) &&
           (nulla[(List[Char], List[Char])](atts) &&
             (equal_nata(size_list[xml](cs), nat(BigInt(5))) &&
               is_none[List[Char]](text))))
      bindb[List[Char], A,
             F](p1(nth[xml](cs, Zero_nat)),
                 (x1: A) =>
                   bindb[List[Char], B,
                          F](p2(nth[xml](cs, one_nat)),
                              (x2: B) =>
                                bindb[List[Char], C,
                                       F](p3(nth[xml](cs, nat(BigInt(2)))),
   (x3: C) =>
     bindb[List[Char], D,
            F](p4(nth[xml](cs, nat(BigInt(3)))),
                (x4: D) =>
                  bindb[List[Char], E,
                         F](p5(nth[xml](cs, nat(BigInt(4)))),
                             (x5: E) =>
                               Inr[F, List[Char]](((((f(x1))(x2))(x3))(x4))(x5)))))))
      else no_transform[F](tag, Xml(name, atts, cs, text)))
}

def rsteps[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             (term[A, List[Char]],
               List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                            term[A, List[Char]]))])] =
  (xml1many_elements[term[A, List[Char]],
                      (pos, ((term[A, List[Char]], term[A, List[Char]]),
                              term[A, List[Char]])),
                      (term[A, List[Char]],
                        List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                     term[A,
   List[Char]]))])](List('r', 'e', 'w', 'r', 'i', 't', 'e', 'S', 'e', 'q', 'u',
                          'e', 'n', 'c', 'e'),
                     (a: xml) =>
                       xml1element[term[A, List[Char]],
                                    term[A,
  List[Char]]](List('s', 't', 'a', 'r', 't', 'T', 'e', 'r', 'm'),
                (aa: xml) => term[A](xml2name, aa), id[term[A, List[Char]]], a),
                     rstep[A](xml2name))).apply((a: term[A, List[Char]]) =>
          (b: List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                           term[A, List[Char]]))])
            => (a, b))

def catcha[A, B, C](m: sum[A, B], f: A => sum[C, B]): sum[C, B] =
  m match {
    case Inl(a) => f(a)
    case Inr(a) => Inr[B, C](a)
  }

def forallm[A, B](f: A => sum[B, Unit], x1: List[A]): sum[(A, B), Unit] =
  (f, x1) match {
  case (f, Nil) => Inr[Unit, (A, B)](())
  case (f, x :: xs) =>
    bindb[(A, B), Unit,
           Unit](catcha[B, Unit,
                         (A, B)](f(x), (xa: B) => Inl[(A, B), Unit]((x, xa))),
                  (_: Unit) => forallm[A, B](f, xs))
}

def check[A](b: Boolean, e: A): sum[A, Unit] =
  (if (b) Inr[Unit, A](()) else Inl[A, Unit](e))

def check_edges[A](ss: A => (List[Char]) => List[Char], g: ((A, A)) => Boolean,
                    c: List[A], d: List[A]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(A, (List[Char]) => List[Char]), Unit,
          (List[Char]) =>
            List[Char]](forallm[A, (List[Char]) =>
                                     List[Char]]((n: A) =>
           catcha[(A, (List[Char]) => List[Char]), Unit,
                   (List[Char]) =>
                     List[Char]](forallm[A,
  (List[Char]) =>
    List[Char]]((m: A) =>
                  check[(List[Char]) =>
                          List[Char]](! (g((m, n))),
                                       comp[List[Char], List[Char],
     List[Char]](shows_string.apply(List('e', 'd', 'g', 'e', ' ', 'f', 'r', 'o',
  'm', ' ')),
                  comp[List[Char], List[Char],
                        List[Char]](ss(m),
                                     comp[List[Char], List[Char],
   List[Char]](shows_string.apply(List(' ', 't', 'o', ' ')), ss(n))))),
                 d),
                                  (x: (A, (List[Char]) => List[Char])) =>
                                    Inl[(List[Char]) => List[Char],
 Unit](snd[A, (List[Char]) => List[Char]](x))),
          c),
                         (x: (A, (List[Char]) => List[Char])) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](snd[A, (List[Char]) => List[Char]](x)))

implicit def monoid_add_int: monoid_add[BigInt] = new monoid_add[BigInt] {
  val `Ceta.zero` = zero_inta
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
}

implicit def comm_monoid_add_int: comm_monoid_add[BigInt] = new
  comm_monoid_add[BigInt] {
  val `Ceta.zero` = zero_inta
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
}

implicit def semiring_0_int: semiring_0[BigInt] = new semiring_0[BigInt] {
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
  val `Ceta.zero` = zero_inta
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
}

implicit def monoid_mult_int: monoid_mult[BigInt] = new monoid_mult[BigInt] {
  val `Ceta.one` = one_inta
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
}

implicit def zero_neq_one_int: zero_neq_one[BigInt] = new zero_neq_one[BigInt] {
  val `Ceta.zero` = zero_inta
  val `Ceta.one` = one_inta
}

implicit def semiring_1_int: semiring_1[BigInt] = new semiring_1[BigInt] {
  val `Ceta.zero` = zero_inta
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
  val `Ceta.one` = one_inta
}

def rel_pow_impl[A, B](succ: (List[A]) => List[A], un: (List[A]) => B => B,
                        memb: A => B => Boolean, newa: List[A], have: B,
                        x5: nat):
      B =
  (succ, un, memb, newa, have, x5) match {
  case (succ, un, memb, newa, have, Zero_nat) => (un(newa))(have)
  case (succ, un, memb, newa, have, Suc(m)) =>
    (if (nulla[A](newa)) have
      else {
             val maybe: List[A] = succ(newa)
             val havea: B = (un(newa))(have)
             val newb: List[A] =
               filter[A]((n: A) => ! ((memb(n))(havea)), maybe);
             rel_pow_impl[A, B](succ, un, memb, newb, havea, m)
           })
}

def rtrancl_impl[A, B](gen_succ: (List[(A, A)]) => (List[A]) => List[A],
                        un: (List[A]) => B => B, memb: A => B => Boolean,
                        emp: B, rel: List[(A, A)]):
      (List[A]) => B =
  {
    val succ: (List[A]) => List[A] = gen_succ(rel)
    val n: nat = size_list[(A, A)](rel);
    (as: List[A]) => rel_pow_impl[A, B](succ, un, memb, as, emp, n)
  }

def rtrancl_list_impl[A : equal]: (List[(A, A)]) => (List[A]) => List[A] =
  (a: List[(A, A)]) =>
    rtrancl_impl[A, List[A]]((r: List[(A, A)]) =>
                               (as: List[A]) =>
                                 remdups[A](map_filter[(A, A),
                A]((x: (A, A)) =>
                     (if ({
                            val (aa, _): (A, A) = x;
                            membera[A](as, aa)
                          })
                       Some[A](snd[A, A](x)) else None),
                    r)),
                              (xs: List[A]) =>
                                (ys: List[A]) =>
                                  filter[A]((x: A) => ! (membera[A](ys, x)),
     xs) ++
                                    ys,
                              (x: A) => (xs: List[A]) => membera[A](xs, x), Nil,
                              a)

def w_impl[A : equal, B](d: List[(A, term[B, A])]): (term[B, A]) => List[A] =
  {
    val filt: (List[(A, A)]) => List[(A, A)] =
      (a: List[(A, A)]) =>
        filter[(A, A)]((aa: (A, A)) =>
                         {
                           val (x, y): (A, A) = aa;
                           ! (eq[A](x, y))
                         },
                        a)
    val xvs: List[(A, A)] =
      maps[(A, term[B, A]),
            (A, A)]((a: (A, term[B, A])) =>
                      {
                        val (x, t): (A, term[B, A]) = a;
                        mapa[A, (A, A)]((aa: A) => (x, aa),
 vars_term_list[B, A].apply(t))
                      },
                     d)
    val rel: List[(A, A)] = filt(xvs)
    val rtran: (List[A]) => List[A] = rtrancl_list_impl[A].apply(rel);
    (t: term[B, A]) => rtran(vars_term_list[B, A].apply(t))
  }

def matcha[A, B]: ((term[A, B], term[A, B])) => (term[A, B]) => Boolean =
  (_: (term[A, B], term[A, B])) => (_: term[A, B]) => true

implicit def
  linorder_prod[A : equal : linorder, B : equal : linorder]: linorder[(A, B)] =
  new linorder[(A, B)] {
  val `Ceta.less_eq` = (a: (A, B), b: (A, B)) => less_eq_prod[A, B](a, b)
  val `Ceta.less` = (a: (A, B), b: (A, B)) => less_prod[A, B](a, b)
}

def top_boola: Boolean = true

implicit def top_bool: top[Boolean] = new top[Boolean] {
  val `Ceta.top` = top_boola
  val `Ceta.less_eq` = (a: Boolean, b: Boolean) => less_eq_bool(a, b)
  val `Ceta.less` = (a: Boolean, b: Boolean) => less_bool(a, b)
}

def top_fun[A, B : top]: A => B = (_: A) => top[B]

def equal_posa(x0: pos, x1: pos): Boolean = (x0, x1) match {
  case (PCons(nat, pos), Empty) => false
  case (Empty, PCons(nat, pos)) => false
  case (PCons(nata, posa), PCons(nat, pos)) =>
    equal_nata(nata, nat) && equal_posa(posa, pos)
  case (Empty, Empty) => true
}

implicit def equal_pos: equal[pos] = new equal[pos] {
  val `Ceta.equal` = (a: pos, b: pos) => equal_posa(a, b)
}

abstract sealed class arithFun
final case class Arg(a: nat) extends arithFun
final case class Const(a: nat) extends arithFun
final case class Sum(a: List[arithFun]) extends arithFun
final case class Max(a: List[arithFun]) extends arithFun
final case class Min(a: List[arithFun]) extends arithFun
final case class Prod(a: List[arithFun]) extends arithFun

abstract sealed class sl_inter[+A]
final case class SL_Inter[A](a: nat, b: List[((A, nat), arithFun)]) extends
  sl_inter[A]

abstract sealed class sl_variant[+A, +B]
final case class Rootlab[A, B](a: Option[(A, nat)]) extends sl_variant[A, B]
final case class Finitelab[A, B](a: sl_inter[A]) extends sl_variant[A, B]
final case class QuasiFinitelab[A, B](a: sl_inter[A], b: B) extends
  sl_variant[A, B]

abstract sealed class arctic
final case object MinInfty extends arctic
final case class Num_arc(a: BigInt) extends arctic

abstract sealed class arctic_rat
final case object MinInfty_rat extends arctic_rat
final case class Num_arc_rat(a: rat) extends arctic_rat

abstract sealed class redtriple_impl[+A]
final case class Int_carrier[A](a: List[((A, nat), (BigInt, List[BigInt]))])
  extends redtriple_impl[A]
final case class
  Int_nl_carrier[A](a: List[((A, nat), List[(List[(nat, nat)], BigInt)])])
  extends redtriple_impl[A]
final case class Rat_carrier[A](a: List[((A, nat), (rat, List[rat]))]) extends
  redtriple_impl[A]
final case class
  Rat_nl_carrier[A](a: rat, b: List[((A, nat), List[(List[(nat, nat)], rat)])])
  extends redtriple_impl[A]
final case class Arctic_carrier[A](a: List[((A, nat), (arctic, List[arctic]))])
  extends redtriple_impl[A]
final case class
  Arctic_rat_carrier[A](a: List[((A, nat), (arctic_rat, List[arctic_rat]))])
  extends redtriple_impl[A]
final case class
  Int_mat_carrier[A](a: nat, b: nat,
                      c: List[((A, nat),
                                (List[List[BigInt]],
                                  List[List[List[BigInt]]]))])
  extends redtriple_impl[A]
final case class
  Rat_mat_carrier[A](a: nat, b: nat,
                      c: List[((A, nat),
                                (List[List[rat]], List[List[List[rat]]]))])
  extends redtriple_impl[A]
final case class
  Arc_mat_carrier[A](a: nat,
                      b: List[((A, nat),
                                (List[List[arctic]],
                                  List[List[List[arctic]]]))])
  extends redtriple_impl[A]
final case class
  Arc_rat_mat_carrier[A](a: nat,
                          b: List[((A, nat),
                                    (List[List[arctic_rat]],
                                      List[List[List[arctic_rat]]]))])
  extends redtriple_impl[A]
final case class
  Rpo[A](a: List[((A, nat), (nat, order_tag))], b: List[((A, nat), af_entry)])
  extends redtriple_impl[A]

abstract sealed class list_order_type
final case object MS_Ext extends list_order_type
final case object Max_Ext extends list_order_type
final case object Min_Ext extends list_order_type

abstract sealed class root_redtriple_impl[+A]
final case class
  Scnp[A](a: list_order_type, b: List[((A, nat), List[(nat, nat)])],
           c: redtriple_impl[A])
  extends root_redtriple_impl[A]

abstract sealed class dp_proof[+A, +B, +C]
final case object PisEmpty extends dp_proof[Nothing, Nothing, Nothing]
final case class
  SubtermCriterionProc[A, B,
                        C](a: projL[lab[A, B]],
                            b: List[((term[lab[A, B], C], term[lab[A, B], C]),
                                      List[(pos,
     ((term[lab[A, B], C], term[lab[A, B], C]), term[lab[A, B], C]))])],
                            c: List[(term[lab[A, B], C], term[lab[A, B], C])],
                            d: dp_proof[A, B, C])
  extends dp_proof[A, B, C]
final case class
  RedPairProc[A, B,
               C](a: sum[root_redtriple_impl[lab[A, B]],
                          redtriple_impl[lab[A, B]]],
                   b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                   c: dp_proof[A, B, C])
  extends dp_proof[A, B, C]
final case class
  RedPairUrProc[A, B,
                 C](a: sum[root_redtriple_impl[lab[A, B]],
                            redtriple_impl[lab[A, B]]],
                     b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                     c: List[(term[lab[A, B], C], term[lab[A, B], C])],
                     d: dp_proof[A, B, C])
  extends dp_proof[A, B, C]
final case class
  UsableRulesProc[A, B,
                   C](a: List[(term[lab[A, B], C], term[lab[A, B], C])],
                       b: dp_proof[A, B, C])
  extends dp_proof[A, B, C]
final case class
  DepGraphProc[A, B,
                C](a: List[(Option[dp_proof[A, B, C]],
                             List[(term[lab[A, B], C], term[lab[A, B], C])])])
  extends dp_proof[A, B, C]
final case class
  MonoRedPairProc[A, B,
                   C](a: redtriple_impl[lab[A, B]],
                       b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                       c: List[(term[lab[A, B], C], term[lab[A, B], C])],
                       d: dp_proof[A, B, C])
  extends dp_proof[A, B, C]
final case class
  MonoRedPairUrProc[A, B,
                     C](a: redtriple_impl[lab[A, B]],
                         b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                         c: List[(term[lab[A, B], C], term[lab[A, B], C])],
                         d: List[(term[lab[A, B], C], term[lab[A, B], C])],
                         e: dp_proof[A, B, C])
  extends dp_proof[A, B, C]
final case class
  SizeChangeSubterm[A, B,
                     C](a: List[((term[lab[A, B], C], term[lab[A, B], C]),
                                  (List[(nat, nat)], List[(nat, nat)]))])
  extends dp_proof[A, B, C]
final case class
  SizeChangeRedPair[A, B,
                     C](a: redtriple_impl[lab[A, B]],
                         b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                         c: List[((term[lab[A, B], C], term[lab[A, B], C]),
                                   (List[(nat, nat)], List[(nat, nat)]))])
  extends dp_proof[A, B, C]
final case class
  UncurryProc[A, B,
               C](a: Option[nat],
                   b: (lab[A, B],
                        (List[((lab[A, B], nat), List[lab[A, B]])],
                          (List[(term[lab[A, B], C], term[lab[A, B], C])],
                            List[(term[lab[A, B], C], term[lab[A, B], C])]))),
                   c: List[(term[lab[A, B], C], term[lab[A, B], C])],
                   d: List[(term[lab[A, B], C], term[lab[A, B], C])],
                   e: dp_proof[A, B, C])
  extends dp_proof[A, B, C]
final case class
  FccProc[A, B,
           C](a: lab[A, B], b: List[ctxt[lab[A, B], C]],
               c: List[(term[lab[A, B], C], term[lab[A, B], C])],
               d: List[(term[lab[A, B], C], term[lab[A, B], C])],
               e: dp_proof[A, B, C])
  extends dp_proof[A, B, C]
final case class
  SplitProc[A, B,
             C](a: List[(term[lab[A, B], C], term[lab[A, B], C])],
                 b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                 c: dp_proof[A, B, C], d: dp_proof[A, B, C])
  extends dp_proof[A, B, C]
final case class
  SemlabProc[A, B,
              C](a: sl_variant[lab[A, B], C],
                  b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                  c: List[term[lab[A, B], C]],
                  d: List[(term[lab[A, B], C], term[lab[A, B], C])],
                  e: dp_proof[A, B, C])
  extends dp_proof[A, B, C]
final case class
  SwitchInnermostProc[A, B,
                       C](a: Option[List[(term[lab[A, B], C],
   (List[(pos, ((term[lab[A, B], C], term[lab[A, B], C]), term[lab[A, B], C]))],
     (term[lab[A, B], C],
       List[(pos, ((term[lab[A, B], C], term[lab[A, B], C]),
                    term[lab[A, B], C]))])))]],
                           b: dp_proof[A, B, C])
  extends dp_proof[A, B, C]
final case class
  RewritingProc[A, B,
                 C](a: List[(term[lab[A, B], C], term[lab[A, B], C])],
                     b: (term[lab[A, B], C], term[lab[A, B], C]),
                     c: (term[lab[A, B], C], term[lab[A, B], C]),
                     d: (term[lab[A, B], C], term[lab[A, B], C]), e: pos,
                     f: dp_proof[A, B, C])
  extends dp_proof[A, B, C]
final case class
  InstantiationProc[A, B,
                     C](a: (term[lab[A, B], C], term[lab[A, B], C]),
                         b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                         c: dp_proof[A, B, C])
  extends dp_proof[A, B, C]
final case class
  ForwardInstantiationProc[A, B,
                            C](a: (term[lab[A, B], C], term[lab[A, B], C]),
                                b: List[(term[lab[A, B], C],
  term[lab[A, B], C])],
                                c: Option[List[(term[lab[A, B], C],
         term[lab[A, B], C])]],
                                d: dp_proof[A, B, C])
  extends dp_proof[A, B, C]
final case class
  NarrowingProc[A, B,
                 C](a: (term[lab[A, B], C], term[lab[A, B], C]), b: pos,
                     c: List[(term[lab[A, B], C], term[lab[A, B], C])],
                     d: dp_proof[A, B, C])
  extends dp_proof[A, B, C]
final case class
  AssumeFinite[A, B,
                C](a: (List[(term[lab[A, B], C], term[lab[A, B], C])],
                        (List[(term[lab[A, B], C], term[lab[A, B], C])],
                          (List[term[lab[A, B], C]],
                            (List[(term[lab[A, B], C], term[lab[A, B], C])],
                              List[(term[lab[A, B], C],
                                     term[lab[A, B], C])])))))
  extends dp_proof[A, B, C]

abstract sealed class boundstype
final case object Roof extends boundstype
final case object Match extends boundstype

abstract sealed class tree_automaton[+A, +B]
final case class
  Tree_Automaton[A, B](a: List[A], b: List[ta_rule[A, B]], c: List[(A, A)])
  extends tree_automaton[A, B]

abstract sealed class bounds_info[+A, +B]
final case class
  Bounds_Info[B, A](a: boundstype, b: nat, c: List[B],
                     d: tree_automaton[B, (A, nat)])
  extends bounds_info[A, B]

abstract sealed class trs_proof[+A, +B, +C]
final case class
  DPTrans[A, B,
           C](a: List[(term[lab[A, B], C], term[lab[A, B], C])],
               b: dp_proof[A, B, C])
  extends trs_proof[A, B, C]
final case class
  RuleRemoval[A, B,
               C](a: redtriple_impl[lab[A, B]],
                   b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                   c: trs_proof[A, B, C])
  extends trs_proof[A, B, C]
final case class StringReversal[A, B, C](a: trs_proof[A, B, C]) extends
  trs_proof[A, B, C]
final case class Bounds[A, B, C](a: bounds_info[lab[A, B], C]) extends
  trs_proof[A, B, C]
final case class
  Uncurry[A, B,
           C](a: (lab[A, B],
                   (List[((lab[A, B], nat), List[lab[A, B]])],
                     (List[(term[lab[A, B], C], term[lab[A, B], C])],
                       List[(term[lab[A, B], C], term[lab[A, B], C])]))),
               b: List[(term[lab[A, B], C], term[lab[A, B], C])],
               c: trs_proof[A, B, C])
  extends trs_proof[A, B, C]
final case class
  Semlab[A, B,
          C](a: sl_variant[lab[A, B], C], b: List[term[lab[A, B], C]],
              c: List[(term[lab[A, B], C], term[lab[A, B], C])],
              d: trs_proof[A, B, C])
  extends trs_proof[A, B, C]
final case object RisEmpty extends trs_proof[Nothing, Nothing, Nothing]
final case class
  Fcc[A, B,
       C](a: List[ctxt[lab[A, B], C]],
           b: List[(term[lab[A, B], C], term[lab[A, B], C])],
           c: trs_proof[A, B, C])
  extends trs_proof[A, B, C]
final case class
  Split[A, B,
         C](a: List[(term[lab[A, B], C], term[lab[A, B], C])],
             b: trs_proof[A, B, C], c: trs_proof[A, B, C])
  extends trs_proof[A, B, C]
final case class
  SwitchInnermost[A, B,
                   C](a: Option[List[(term[lab[A, B], C],
                                       (List[(pos,
       ((term[lab[A, B], C], term[lab[A, B], C]), term[lab[A, B], C]))],
 (term[lab[A, B], C],
   List[(pos, ((term[lab[A, B], C], term[lab[A, B], C]),
                term[lab[A, B], C]))])))]],
                       b: trs_proof[A, B, C])
  extends trs_proof[A, B, C]
final case class
  AssumeSN[A, B,
            C](a: (List[term[lab[A, B], C]],
                    (List[(term[lab[A, B], C], term[lab[A, B], C])],
                      List[(term[lab[A, B], C], term[lab[A, B], C])])))
  extends trs_proof[A, B, C]

def itoi_iterateo[A, B, C,
                   D](it: (A => Boolean) => (B => C => A => A) => D => A => A):
      (B => C => A => A) => D => A => A =
  it((_: A) => true)

def rm_iterateoi[A, B,
                  C](c: A => Boolean, f: B => C => A => A, x2: rbta[B, C],
                      sigma: A):
      A =
  (c, f, x2, sigma) match {
  case (c, f, Emptya, sigma) => sigma
  case (c, f, Branch(col, l, k, v, r), sigma) =>
    (if (c(sigma))
      {
        val sigmaa: A = rm_iterateoi[A, B, C](c, f, l, sigma);
        (if (c(sigmaa)) rm_iterateoi[A, B, C](c, f, r, ((f(k))(v))(sigmaa))
          else sigmaa)
      }
      else sigma)
}

def rm_iterateoia[A, B : linorder,
                   C](c: A => Boolean, f: B => C => A => A, r: rbt[B, C]):
      A => A =
  (a: A) => rm_iterateoi[A, B, C](c, f, impl_of[B, C](r), a)

def rm_iterateo[A : linorder, B, C]:
      (A => B => C => C) => (rbt[A, B]) => C => C =
  itoi_iterateo[C, A, B,
                 rbt[A, B]]((a: C => Boolean) =>
                              (b: A => B => C => C) =>
                                (c: rbt[A, B]) =>
                                  rm_iterateoia[C, A, B](a, b, c))

def rm_iterate[A : linorder, B, C]:
      (A => B => C => C) => (rbt[A, B]) => C => C =
  rm_iterateo[A, B, C]

def rm_add[A : linorder, B]: (rbt[A, B]) => (rbt[A, B]) => rbt[A, B] =
  (a: rbt[A, B]) =>
    (b: rbt[A, B]) =>
      it_add[A, B,
              rbt[A, B]](rm_update[A, B], rm_iterate[A, B, rbt[A, B]], a, b)

def rs_ins[A : linorder]: A => (rbt[A, Unit]) => rbt[A, Unit] =
  (a: A) =>
    (b: rbt[A, Unit]) =>
      s_ins[A, rbt[A, Unit], rbt[A, Unit]](rm_update[A, Unit], a, b)

def equal_order_tag(x0: order_tag, x1: order_tag): Boolean = (x0, x1) match {
  case (Mul, Lex) => false
  case (Lex, Mul) => false
  case (Mul, Mul) => true
  case (Lex, Lex) => true
}

def diag_l[A, B](x0: (A, B)): ((A, A), B) = x0 match {
  case (a, b) => ((a, a), b)
}

def lex_ext_efficient_m[A, B](f: A => A => B => ((Boolean, Boolean), B),
                               x1: List[A], x2: List[A], m: B):
      ((Boolean, Boolean), B) =
  (f, x1, x2, m) match {
  case (f, Nil, Nil, m) => ((false, true), m)
  case (f, uu :: uv, Nil, m) => ((true, true), m)
  case (f, Nil, uw :: ux, m) => ((false, false), m)
  case (f, a :: as, b :: bs, m) =>
    {
      val (aa, ba): ((Boolean, Boolean), B) = ((f(a))(b))(m);
      ({
         val (s, ns): (Boolean, Boolean) = aa;
         (ma: B) =>
           (if (s) ((true, true), ma)
             else (if (ns) lex_ext_efficient_m[A, B](f, as, bs, ma)
                    else ((false, false), ma)))
       })(ba)
    }
}

def filter_not_stri_efficient_m_aux[A, B](f: A => B => ((Boolean, Boolean), B),
   x1: List[A], bs: List[A], m: B):
      (List[A], B) =
  (f, x1, bs, m) match {
  case (f, Nil, bs, m) => (bs, m)
  case (f, a :: as, bs, m) =>
    {
      val (b, c): ((Boolean, Boolean), B) = (f(a))(m);
      ({
         val (s, _): (Boolean, Boolean) = b;
         (ma: B) =>
           (if (s) filter_not_stri_efficient_m_aux[A, B](f, as, bs, ma)
             else filter_not_stri_efficient_m_aux[A, B](f, as, a :: bs, ma))
       })(c)
    }
}

def filter_not_stri_efficient_m[A, B](f: A => B => ((Boolean, Boolean), B),
                                       as: List[A], m: B):
      (List[A], B) =
  {
    val (ys, a): (List[A], B) =
      filter_not_stri_efficient_m_aux[A, B](f, as, Nil, m);
    (rev[A](ys), a)
  }

def ex_dom_efficient_m[A : equal,
                        B](f: A => A => B => ((Boolean, Boolean), B), y: A,
                            x2: List[A], uw: List[A], ux: List[A], m: B):
      ((Boolean, Boolean), B) =
  (f, y, x2, uw, ux, m) match {
  case (f, y, Nil, uw, ux, m) => ((false, false), m)
  case (f, y, b :: bs, xs, ys, m) =>
    {
      val (a, c): ((Boolean, Boolean), B) = ((f(b))(y))(m);
      ({
         val (s, ns): (Boolean, Boolean) = a;
         (ma: B) =>
           (if (s)
             {
               val (ysa, m2): (List[A], B) =
                 filter_not_stri_efficient_m[A, B](f(b), ys, ma)
               val (aa, ba): ((Boolean, Boolean), B) =
                 mul_ext_efficient_m[A, B](f, remove1[A](b, xs), ysa, m2);
               ({
                  val (_, ns1): (Boolean, Boolean) = aa;
                  (m3: B) =>
                    (if (ns1) ((true, true), m3)
                      else ex_dom_efficient_m[A, B](f, y, bs, xs, ys, m3))
                })(ba)
             }
             else (if (ns)
                    {
                      val (aa, ba): ((Boolean, Boolean), B) =
                        mul_ext_efficient_m[A, B](f, remove1[A](b, xs), ys, ma);
                      ({
                         val (s1, ns1): (Boolean, Boolean) = aa;
                         (mb: B) =>
                           (if (s1) ((true, true), mb)
                             else {
                                    val (ab, bb): ((Boolean, Boolean), B) =
                                      ex_dom_efficient_m[A,
                  B](f, y, bs, xs, ys, mb);
                                    ({
                                       val (s2, ns2): (Boolean, Boolean) = ab;
                                       (m3: B) =>
 (if (s2) ((true, true), m3) else ((false, ns1 || ns2), m3))
                                     })(bb)
                                  })
                       })(ba)
                    }
                    else ex_dom_efficient_m[A, B](f, y, bs, xs, ys, ma)))
       })(c)
    }
}

def mul_ext_efficient_m[A : equal,
                         B](f: A => A => B => ((Boolean, Boolean), B),
                             x1: List[A], x2: List[A], m: B):
      ((Boolean, Boolean), B) =
  (f, x1, x2, m) match {
  case (f, Nil, Nil, m) => ((false, true), m)
  case (f, Nil, v :: va, m) => ((false, false), m)
  case (f, v :: va, Nil, m) => ((true, true), m)
  case (f, v :: va, y :: ys, m) =>
    ex_dom_efficient_m[A, B](f, y, v :: va, v :: va, ys, m)
}

def all_stri_efficient_m[A, B](f: A => B => ((Boolean, Boolean), B),
                                x1: List[A], m: B):
      (Boolean, B) =
  (f, x1, m) match {
  case (f, Nil, m) => (true, m)
  case (f, a :: as, m) =>
    {
      val (aa, b): ((Boolean, Boolean), B) = (f(a))(m);
      ({
         val (s, _): (Boolean, Boolean) = aa;
         (ma: B) =>
           (if (s) all_stri_efficient_m[A, B](f, as, ma) else (false, ma))
       })(b)
    }
}

def any_nstri_efficient_m[A, B](f: A => B => ((Boolean, Boolean), B),
                                 x1: List[A], m: B):
      (Boolean, B) =
  (f, x1, m) match {
  case (f, Nil, m) => (false, m)
  case (f, a :: as, m) =>
    {
      val (aa, b): ((Boolean, Boolean), B) = (f(a))(m);
      ({
         val (_, ns): (Boolean, Boolean) = aa;
         (ma: B) =>
           (if (ns) (true, ma) else any_nstri_efficient_m[A, B](f, as, ma))
       })(b)
    }
}

def rpo_efficient_m[A, B : equal,
                     C : equal](model:
                                  memory_ext[A, (term[B, C], term[B, C]),
      (Boolean, Boolean), Unit],
                                 pr: (B =>
nat => B => nat => (Boolean, Boolean),
                                       B => Boolean),
                                 tag: B => nat => order_tag, s: term[B, C],
                                 t: term[B, C], m: A):
      ((Boolean, Boolean), A) =
  (lookupb[A, (term[B, C], term[B, C]), (Boolean, Boolean),
            Unit](model)).apply(m).apply((s, t))
    match {
    case None =>
      {
        val (res, res_m): ((Boolean, Boolean), A) =
          (s, t) match {
            case (Var(x), Var(y)) => ((false, eq[C](x, y)), m)
            case (Var(x), Fun(g, ts)) =>
              ((false,
                 nulla[term[B, C]](ts) &&
                   (snd[B => nat => B => nat => (Boolean, Boolean),
                         B => Boolean](pr)).apply(g)),
                m)
            case (Fun(f, ss), Var(y)) =>
              diag_l[Boolean,
                      A](any_nstri_efficient_m[term[B, C],
        A]((sa: term[B, C]) =>
             (a: A) =>
               rpo_efficient_m[A, B, C](model, pr, tag, sa, Var[C, B](y), a),
            ss, m))
            case (Fun(f, ss), Fun(g, ts)) =>
              any_nstri_efficient_m[term[B, C],
                                     A]((sa: term[B, C]) =>
  (a: A) => rpo_efficient_m[A, B, C](model, pr, tag, sa, Fun[B, C](g, ts), a),
 ss, m)
                match {
                case (true, ma) => ((true, true), ma)
                case (false, ma) =>
                  {
                    val maa: nat = size_list[term[B, C]](ss)
                    val n: nat = size_list[term[B, C]](ts);
                    (fst[B => nat => B => nat => (Boolean, Boolean),
                          B => Boolean](pr)).apply(f).apply(maa).apply(g).apply(n)
                      match {
                      case (prs, true) =>
                        all_stri_efficient_m[term[B, C],
      A]((a: term[B, C]) =>
           (b: A) =>
             rpo_efficient_m[A, B, C](model, pr, tag, Fun[B, C](f, ss), a, b),
          ts, ma)
                          match {
                          case (true, mb) =>
                            (if (prs) ((true, true), mb)
                              else {
                                     val cf: order_tag = (tag(f))(maa);
                                     (if (equal_order_tag(cf, (tag(g))(n)))
                                       (if (equal_order_tag(cf, Mul))
 mul_ext_efficient_m[term[B, C],
                      A]((a: term[B, C]) =>
                           (b: term[B, C]) =>
                             (c: A) =>
                               rpo_efficient_m[A, B,
        C](model, pr, tag, a, b, c),
                          ss, ts, mb)
 else lex_ext_efficient_m[term[B, C],
                           A]((a: term[B, C]) =>
                                (b: term[B, C]) =>
                                  (c: A) =>
                                    rpo_efficient_m[A, B,
             C](model, pr, tag, a, b, c),
                               ss, ts, mb))
                                       else {
      val b: Boolean = equal_nata(n, Zero_nat);
      ((! (equal_nata(maa, Zero_nat)) && b, b), mb)
    })
                                   })
                          case (false, mb) => ((false, false), mb)
                        }
                      case (prs, false) => ((false, false), ma)
                    }
                  }
              }
          };
        (res, (store[A, (term[B, C], term[B, C]), (Boolean, Boolean),
                      Unit](model)).apply(res_m).apply(((s, t), res)))
      }
    case Some(v) => (v, m)
  }

def efficient_rpo_2[A : equal,
                     B : equal](pr: (A => nat => A => nat => (Boolean, Boolean),
                                      A => Boolean),
                                 tag: A => nat => order_tag, p: term[A, B],
                                 q: term[A, B]):
      (Boolean, Boolean) =
  fst[(Boolean, Boolean),
       List[(term[A, B],
              List[(term[A, B],
                     (Boolean,
                       Boolean))])]](rpo_efficient_m[List[(term[A, B],
                    List[(term[A, B], (Boolean, Boolean))])],
              A, B](l2m[term[A, B], term[A, B], (Boolean, Boolean)], pr, tag, p,
                     q, (emptya[List[(term[A, B],
                                       List[(term[A, B], (Boolean, Boolean))])],
                                 (term[A, B], term[A, B]), (Boolean, Boolean),
                                 Unit](l2m[term[A, B], term[A, B],
    (Boolean, Boolean)])).apply(())))

def rpo_unbounded_impl[A : equal, B : equal]:
      ((A => nat => A => nat => (Boolean, Boolean), A => Boolean)) =>
        (A => nat => order_tag) =>
          (term[A, B]) => (term[A, B]) => (Boolean, Boolean) =
  (a: (A => nat => A => nat => (Boolean, Boolean), A => Boolean)) =>
    (b: A => nat => order_tag) =>
      (c: term[A, B]) => (d: term[A, B]) => efficient_rpo_2[A, B](a, b, c, d)

def rpo_unbounded[A : equal, B : equal]:
      ((A => nat => A => nat => (Boolean, Boolean), A => Boolean)) =>
        (A => nat => order_tag) =>
          (term[A, B]) => (term[A, B]) => (Boolean, Boolean) =
  rpo_unbounded_impl[A, B]

implicit def monoid_add_rat: monoid_add[rat] = new monoid_add[rat] {
  val `Ceta.zero` = zero_rata
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
}

implicit def comm_monoid_add_rat: comm_monoid_add[rat] = new
  comm_monoid_add[rat] {
  val `Ceta.zero` = zero_rata
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
}

implicit def semiring_0_rat: semiring_0[rat] = new semiring_0[rat] {
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
  val `Ceta.zero` = zero_rata
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
}

implicit def monoid_mult_rat: monoid_mult[rat] = new monoid_mult[rat] {
  val `Ceta.one` = one_rata
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
}

implicit def zero_neq_one_rat: zero_neq_one[rat] = new zero_neq_one[rat] {
  val `Ceta.zero` = zero_rata
  val `Ceta.one` = one_rata
}

implicit def semiring_1_rat: semiring_1[rat] = new semiring_1[rat] {
  val `Ceta.zero` = zero_rata
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
  val `Ceta.one` = one_rata
}

def equal_suma[A : equal, B : equal](x0: sum[A, B], x1: sum[A, B]): Boolean =
  (x0, x1) match {
  case (Inr(b), Inl(a)) => false
  case (Inl(a), Inr(b)) => false
  case (Inr(ba), Inr(b)) => eq[B](ba, b)
  case (Inl(aa), Inl(a)) => eq[A](aa, a)
}

implicit def equal_sum[A : equal, B : equal]: equal[sum[A, B]] = new
  equal[sum[A, B]] {
  val `Ceta.equal` = (a: sum[A, B], b: sum[A, B]) => equal_suma[A, B](a, b)
}

def in_poss[A, B](x0: pos, uu: term[A, B]): Boolean = (x0, uu) match {
  case (Empty, uu) => true
  case (PCons(i, p), Fun(f, ts)) =>
    less_nat(i, size_list[term[A, B]](ts)) &&
      in_poss[A, B](p, nth[term[A, B]](ts, i))
  case (PCons(i, p), Var(uv)) => false
}

abstract sealed class redtriple_ext[+A, +B, +C]
final case class
  Redtriple_ext[A, B,
                 C](a: sum[(List[Char]) => List[Char], Unit],
                     b: ((term[A, B], term[A, B])) =>
                          sum[(List[Char]) => List[Char], Unit],
                     c: ((term[A, B], term[A, B])) =>
                          sum[(List[Char]) => List[Char], Unit],
                     d: ((term[A, B], term[A, B])) =>
                          sum[(List[Char]) => List[Char], Unit],
                     e: ((A, nat)) => nat => Boolean,
                     f: sum[(List[Char]) => List[Char], Unit],
                     g: (List[Char]) => List[Char], h: C)
  extends redtriple_ext[A, B, C]

def s[A, B, C](x0: redtriple_ext[A, B, C]):
      ((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit] =
  x0 match {
  case Redtriple_ext(valid, s, ns, nst, af, mono, desc, more) => s
}

def instance_rule[A : equal, B : equal,
                   C : equal](lr: (term[A, B], term[A, B]),
                               st: (term[A, C], term[A, C])):
      Boolean =
  ! (is_none[C => term[A, B]](match_list[A, B,
  C](fst[term[A, B], term[A, B]](lr),
      List((fst[term[A, B], term[A, B]](lr), fst[term[A, C], term[A, C]](st)),
            (snd[term[A, B], term[A, B]](lr),
              snd[term[A, C], term[A, C]](st))))))

def equal_laba[A : equal, B : equal](x0: lab[A, B], x1: lab[A, B]): Boolean =
  (x0, x1) match {
  case (Sharp(lab), UnLab(f)) => false
  case (UnLab(f), Sharp(lab)) => false
  case (Sharp(laba), FunLab(lab, list)) => false
  case (FunLab(laba, list), Sharp(lab)) => false
  case (UnLab(f), FunLab(lab, list)) => false
  case (FunLab(lab, list), UnLab(f)) => false
  case (Sharp(laba), Lab(lab, l)) => false
  case (Lab(laba, l), Sharp(lab)) => false
  case (UnLab(f), Lab(lab, l)) => false
  case (Lab(lab, l), UnLab(f)) => false
  case (FunLab(laba, list), Lab(lab, l)) => false
  case (Lab(laba, l), FunLab(lab, list)) => false
  case (Sharp(laba), Sharp(lab)) => equal_laba[A, B](laba, lab)
  case (UnLab(fa), UnLab(f)) => eq[A](fa, f)
  case (FunLab(laba, lista), FunLab(lab, list)) =>
    equal_laba[A, B](laba, lab) && equal_lista[lab[A, B]](lista, list)
  case (Lab(laba, la), Lab(lab, l)) =>
    equal_laba[A, B](laba, lab) && eq[B](la, l)
}

implicit def equal_lab[A : equal, B : equal]: equal[lab[A, B]] = new
  equal[lab[A, B]] {
  val `Ceta.equal` = (a: lab[A, B], b: lab[A, B]) => equal_laba[A, B](a, b)
}

def gen_label[A, B](f: lab[A, B], x1: nat): lab[A, B] = (f, x1) match {
  case (f, Zero_nat) => f
  case (f, Suc(n)) => FunLab[A, B](gen_label[A, B](f, n), Nil)
}

def label_depth[A, B](x0: lab[A, B]): nat = x0 match {
  case UnLab(uu) => Zero_nat
  case Lab(f, uv) => Suc(label_depth[A, B](f))
  case FunLab(f, uw) => Suc(label_depth[A, B](f))
  case Sharp(f) => Suc(label_depth[A, B](f))
}

def fmap[A : equal,
          B : equal](a: lab[A, B], nn: nat,
                      sml: List[((lab[A, B], nat), List[lab[A, B]])]):
      (lab[A, B]) => nat => lab[A, B] =
  {
    val m: nat =
      Suc(max_list(mapa[lab[A, B],
                         nat]((aa: lab[A, B]) => label_depth[A, B](aa),
                               a :: maps[((lab[A, B], nat), List[lab[A, B]]),
  lab[A, B]]((aa: ((lab[A, B], nat), List[lab[A, B]])) =>
               snd[(lab[A, B], nat), List[lab[A, B]]](aa),
              sml))));
    (f: lab[A, B]) =>
      (n: nat) =>
        (if (equal_proda[lab[A, B], nat]((f, n), (a, nn))) a
          else gen_label[A, B](f, m))
  }

def concat_lists[A](x0: List[List[A]]): List[List[A]] = x0 match {
  case Nil => List(Nil)
  case as :: xs =>
    {
      val a: List[List[A]] = concat_lists[A](xs);
      maps[List[A],
            List[A]]((vec: List[A]) =>
                       mapa[A, List[A]]((aa: A) => aa :: vec, as),
                      a)
    }
}

def enum_vectors[A, B](c: List[A], x1: List[B]): List[List[(B, A)]] = (c, x1)
  match {
  case (c, Nil) => List(Nil)
  case (c, x :: xs) =>
    {
      val a: List[List[(B, A)]] = enum_vectors[A, B](c, xs);
      maps[List[(B, A)],
            List[(B, A)]]((vec: List[(B, A)]) =>
                            mapa[A, List[(B, A)]]((ca: A) => (x, ca) :: vec, c),
                           a)
    }
}

def manyof(cs: List[Char]):
      (List[Char]) => sum[List[Char], (List[Char], List[Char])] =
  (a: List[Char]) => many((aa: Char) => membera[Char](cs, aa), a)

def spaces: (List[Char]) => sum[List[Char], (Unit, List[Char])] =
  (a: List[Char]) =>
    binda[Char, List[Char],
           Unit](manyof(wspace),
                  (_: List[Char]) =>
                    (ts: List[Char]) =>
                      Inr[(Unit, List[Char]), List[Char]](((), ts)),
                  a)

def xml2nd_choice[A, B,
                   C](tag: List[Char], p1: xml => sum[List[Char], A],
                       cn: List[Char], p2: xml => sum[List[Char], B],
                       f: A => Option[B] => C, x5: xml):
      sum[List[Char], C] =
  (tag, p1, cn, p2, f, x5) match {
  case (tag, p1, cn, p2, f, Xml(name, atts, cs, text)) =>
    {
      val l: nat = size_list[xml](cs);
      (if (equal_lista[Char](name, tag) &&
             (nulla[(List[Char], List[Char])](atts) &&
               (less_eq_nat(nat(BigInt(2)), l) && is_none[List[Char]](text))))
        bindb[List[Char], A,
               C](p1(nth[xml](cs, Zero_nat)),
                   (x1: A) =>
                     bindb[List[Char], Boolean,
                            C](bool(cn, nth[xml](cs, one_nat)),
                                (b: Boolean) =>
                                  (if (b)
                                    bindb[List[Char], B,
   C](p2(nth[xml](cs, minus_nat(l, one_nat))),
       (x2: B) => Inr[C, List[Char]]((f(x1))(Some[B](x2))))
                                    else Inr[C, List[Char]]((f(x1))(None)))))
        else no_transform[C](tag, Xml(name, atts, cs, text)))
    }
}

def alter1a[A, B, C, D, E, F, G, H, I,
             J](x0: alter[A, B, C, D, E, F, G, H, I, J]):
      A =
  x0 match {
  case Alter1(x) => x
}

def alter2a[A, B, C, D, E, F, G, H, I,
             J](x0: alter[A, B, C, D, E, F, G, H, I, J]):
      B =
  x0 match {
  case Alter2(x) => x
}

def alter3a[A, B, C, D, E, F, G, H, I,
             J](x0: alter[A, B, C, D, E, F, G, H, I, J]):
      C =
  x0 match {
  case Alter3(x) => x
}

def alter4a[A, B, C, D, E, F, G, H, I,
             J](x0: alter[A, B, C, D, E, F, G, H, I, J]):
      D =
  x0 match {
  case Alter4(x) => x
}

def alter5a[A, B, C, D, E, F, G, H, I,
             J](x0: alter[A, B, C, D, E, F, G, H, I, J]):
      E =
  x0 match {
  case Alter5(x) => x
}

def alter6a[A, B, C, D, E, F, G, H, I,
             J](x0: alter[A, B, C, D, E, F, G, H, I, J]):
      F =
  x0 match {
  case Alter6(x) => x
}

def alter7a[A, B, C, D, E, F, G, H, I,
             J](x0: alter[A, B, C, D, E, F, G, H, I, J]):
      G =
  x0 match {
  case Alter7(x) => x
}

def alter8a[A, B, C, D, E, F, G, H, I,
             J](x0: alter[A, B, C, D, E, F, G, H, I, J]):
      H =
  x0 match {
  case Alter8(x) => x
}

def alter9a[A, B, C, D, E, F, G, H, I,
             J](x0: alter[A, B, C, D, E, F, G, H, I, J]):
      I =
  x0 match {
  case Alter9(x) => x
}

abstract sealed class dpp_ops_ext[+A, +B, +C, +D]
final case class
  Dpp_ops_ext[A, B, C,
               D](a: A => (((term[B, C], term[B, C])) => Boolean,
                            (((term[B, C], term[B, C])) => Boolean,
                              ((term[B, C]) => Boolean,
                                (((term[B, C], term[B, C])) => Boolean,
                                  ((term[B, C], term[B, C])) => Boolean)))),
                   b: A => List[(term[B, C], term[B, C])],
                   c: A => List[(term[B, C], term[B, C])],
                   d: A => List[(term[B, C], term[B, C])],
                   e: A => List[term[B, C]],
                   f: A => List[(term[B, C], term[B, C])],
                   g: A => List[(term[B, C], term[B, C])],
                   h: A => List[(term[B, C], term[B, C])], i: A => Boolean,
                   j: A => Boolean, k: A => Boolean,
                   l: A => (term[B, C]) => Boolean, m: A => Boolean,
                   n: A => ((B, nat)) => List[(term[B, C], term[B, C])],
                   o: A => ((B, nat)) => List[(term[B, C], term[B, C])],
                   p: A => (List[(term[B, C], term[B, C])]) => A,
                   q: A => ((term[B, C], term[B, C])) =>
                             (List[(term[B, C], term[B, C])]) => A,
                   r: A => (List[(term[B, C], term[B, C])]) => A,
                   s: A => (List[(term[B, C], term[B, C])]) =>
                             (List[(term[B, C], term[B, C])]) => A,
                   t: A => (List[(term[B, C], term[B, C])]) =>
                             (List[(term[B, C], term[B, C])]) => A,
                   u: A => (List[(term[B, C], term[B, C])]) =>
                             (List[(term[B, C], term[B, C])],
                               List[(term[B, C], term[B, C])]),
                   v: A => (List[(term[B, C], term[B, C])]) =>
                             (List[(term[B, C], term[B, C])],
                               List[(term[B, C], term[B, C])]),
                   w: (List[(term[B, C], term[B, C])]) =>
                        (List[(term[B, C], term[B, C])]) =>
                          (List[term[B, C]]) =>
                            (List[(term[B, C], term[B, C])]) =>
                              (List[(term[B, C], term[B, C])]) => A,
                   x: D)
  extends dpp_ops_ext[A, B, C, D]

def rules_mapb[A, B, C, D](x0: dpp_ops_ext[A, B, C, D]):
      A => ((B, nat)) => List[(term[B, C], term[B, C])] =
  x0 match {
  case Dpp_ops_ext(dpp, p, pw, pairs, q, r, rw, rules, q_empty,
                    rules_no_left_var, rules_non_collapsing, is_QNF,
                    nFQ_subset_NF_rules, rules_map, reverse_rules_map,
                    intersect_pairs, replace_pair, intersect_rules, delete_P_Pw,
                    delete_R_Rw, split_pairs, split_rules, mk, more)
    => rules_map
}

def rules_no_left_vara[A, B, C, D](x0: dpp_ops_ext[A, B, C, D]): A => Boolean =
  x0 match {
  case Dpp_ops_ext(dpp, p, pw, pairs, q, r, rw, rules, q_empty,
                    rules_no_left_var, rules_non_collapsing, is_QNF,
                    nFQ_subset_NF_rules, rules_map, reverse_rules_map,
                    intersect_pairs, replace_pair, intersect_rules, delete_P_Pw,
                    delete_R_Rw, split_pairs, split_rules, mk, more)
    => rules_no_left_var
}

def capRM_dpp[A, B : equal, C : equal](i: dpp_ops_ext[A, B, C, Unit], dpp: A):
      (term[B, C]) => gctxt[B, C] =
  capRM[B, C]((rules_no_left_vara[A, B, C, Unit](i)).apply(dpp),
               (rules_mapb[A, B, C, Unit](i)).apply(dpp))

def rm_empty[A : linorder, B]: rbt[A, B] = empty[A, B]

def elem_list_to_rm[A, B : linorder](key: A => B, x1: List[A]):
      rbt[B, List[A]] =
  (key, x1) match {
  case (key, Nil) => rm_empty[B, List[A]]
  case (key, d :: ds) =>
    {
      val rm: rbt[B, List[A]] = elem_list_to_rm[A, B](key, ds)
      val k: B = key(d);
      rm_alpha[B, List[A]].apply(rm).apply(k) match {
        case None => rm_update_dj[B, List[A]].apply(k).apply(List(d)).apply(rm)
        case Some(data) =>
          rm_update[B, List[A]].apply(k).apply(d :: data).apply(rm)
      }
    }
}

def term_map[A : linorder, B](ts: List[term[A, B]]): rbt[A, List[term[A, B]]] =
  elem_list_to_rm[term[A, B],
                   A](comp[Option[(A, nat)], A,
                            term[A, B]](comp[(A, nat), A,
      Option[(A, nat)]]((a: (A, nat)) => fst[A, nat](a),
                         (a: Option[(A, nat)]) => the[(A, nat)](a)),
 (a: term[A, B]) => root[A, B](a)),
                       ts)

def conflict_pos[A : equal,
                  B : equal](mu: A => term[B, A],
                              x1: (term[B, A], (term[B, A], nat))):
      List[(term[B, A], (term[B, A], nat))] =
  (mu, x1) match {
  case (mu, (Fun(f, ts), (Fun(g, ss), n))) =>
    (if (eq[B](f, g) &&
           equal_nata(size_list[term[B, A]](ts), size_list[term[B, A]](ss)))
      maps[(term[B, A], term[B, A]),
            (term[B, A],
              (term[B, A],
                nat))]((a: (term[B, A], term[B, A])) =>
                         {
                           val (s, t): (term[B, A], term[B, A]) = a;
                           conflict_pos[A, B](mu, (s, (t, n)))
                         },
                        zip[term[B, A], term[B, A]](ts, ss))
      else List((Fun[B, A](f, ts), (Fun[B, A](g, ss), n))))
  case (mu, (Var(x), (Fun(g, ss), n))) =>
    List((Var[A, B](x), (Fun[B, A](g, ss), n)))
  case (mu, (Fun(f, ts), (Var(y), Zero_nat))) =>
    List((Var[A, B](y), (Fun[B, A](f, ts), Zero_nat)))
  case (mu, (Var(x), (Var(y), Zero_nat))) =>
    (if (eq[A](x, y)) Nil else List((Var[A, B](x), (Var[A, B](y), Zero_nat))))
  case (mu, (s, (Var(x), Suc(n)))) => conflict_pos[A, B](mu, (s, (mu(x), n)))
}

def height[A](x0: (A, nat)): nat = x0 match {
  case (f, h) => h
}

def equal_optiona[A : equal](x0: Option[A], x1: Option[A]): Boolean = (x0, x1)
  match {
  case (Some(a), None) => false
  case (None, Some(a)) => false
  case (Some(aa), Some(a)) => eq[A](aa, a)
  case (None, None) => true
}

implicit def equal_option[A : equal]: equal[Option[A]] = new equal[Option[A]] {
  val `Ceta.equal` = (a: Option[A], b: Option[A]) => equal_optiona[A](a, b)
}

def one_poly[A, B : semiring_1]: List[(List[(A, nat)], B)] = List((Nil, one[B]))

def eq_monom[A : equal](x0: List[(A, nat)], n: List[(A, nat)]): Boolean =
  (x0, n) match {
  case (Nil, n) => nulla[(A, nat)](n)
  case ((x, p) :: m, n) =>
    extract[(A, nat)]((yq: (A, nat)) => eq[A](fst[A, nat](yq), x), n) match {
      case None => false
      case Some((n1, ((_, q), n2))) =>
        equal_nata(p, q) && eq_monom[A](m, n1 ++ n2)
    }
}

def poly_add[A : equal,
              B : equal : semiring_0](x0: List[(List[(A, nat)], B)],
                                       q: List[(List[(A, nat)], B)]):
      List[(List[(A, nat)], B)] =
  (x0, q) match {
  case (Nil, q) => q
  case ((m, c) :: p, q) =>
    extract[(List[(A, nat)],
              B)]((mc: (List[(A, nat)], B)) =>
                    eq_monom[A](fst[List[(A, nat)], B](mc), m),
                   q)
      match {
      case None => (m, c) :: poly_add[A, B](p, q)
      case Some((q1, ((_, d), q2))) =>
        (if (eq[B](plus[B](c, d), zero[B])) poly_add[A, B](p, q1 ++ q2)
          else (m, plus[B](c, d)) :: poly_add[A, B](p, q1 ++ q2))
    }
}

def monom_mult[A : equal](x0: List[(A, nat)], n: List[(A, nat)]):
      List[(A, nat)] =
  (x0, n) match {
  case (Nil, n) => n
  case ((x, p) :: m, n) =>
    extract[(A, nat)]((yq: (A, nat)) => eq[A](fst[A, nat](yq), x), n) match {
      case None => (x, p) :: monom_mult[A](m, n)
      case Some((n1, ((_, q), n2))) =>
        (x, plus_nat(p, q)) :: monom_mult[A](m, n1 ++ n2)
    }
}

def monom_mult_poly[A : equal,
                     B : equal : semiring_0](uu: (List[(A, nat)], B),
      x1: List[(List[(A, nat)], B)]):
      List[(List[(A, nat)], B)] =
  (uu, x1) match {
  case (uu, Nil) => Nil
  case ((ma, c), (m, d) :: p) =>
    (if (eq[B](times[B](c, d), zero[B])) monom_mult_poly[A, B]((ma, c), p)
      else (monom_mult[A](ma, m), times[B](c, d)) ::
             monom_mult_poly[A, B]((ma, c), p))
}

def poly_mult[A : equal,
               B : equal : semiring_0](x0: List[(List[(A, nat)], B)],
q: List[(List[(A, nat)], B)]):
      List[(List[(A, nat)], B)] =
  (x0, q) match {
  case (Nil, q) => Nil
  case (mc :: p, q) =>
    poly_add[A, B](monom_mult_poly[A, B](mc, q), poly_mult[A, B](p, q))
}

def zero_poly[A, B]: List[(List[(A, nat)], B)] = Nil

trait ab_semigroup_mult[A] extends semigroup_mult[A] {
}

trait comm_semiring[A] extends ab_semigroup_mult[A] with semiring[A] {
}

trait comm_semiring_0[A] extends comm_semiring[A] with semiring_0[A] {
}

trait comm_monoid_mult[A] extends ab_semigroup_mult[A] with monoid_mult[A] {
}

trait comm_semiring_1[A]
  extends comm_monoid_mult[A] with comm_semiring_0[A] with dvd[A] with
    semiring_1[A]
  {
}

def poly_of[A : equal, B : equal : comm_semiring_1](x0: tpoly[A, B]):
      List[(List[(A, nat)], B)] =
  x0 match {
  case PNum(i) => (if (eq[B](i, zero[B])) Nil else List((Nil, i)))
  case PVar(x) => List((List((x, one_nat)), one[B]))
  case PSum(Nil) => zero_poly[A, B]
  case PSum(p :: ps) =>
    poly_add[A, B](poly_of[A, B](p), poly_of[A, B](PSum[A, B](ps)))
  case PMult(Nil) => one_poly[A, B]
  case PMult(p :: ps) =>
    poly_mult[A, B](poly_of[A, B](p), poly_of[A, B](PMult[A, B](ps)))
}

def rs_memb[A : linorder]: A => (rbt[A, Unit]) => Boolean =
  (a: A) =>
    (b: rbt[A, Unit]) =>
      s_memb[A, rbt[A, Unit],
              Unit]((aa: A) => (ba: rbt[A, Unit]) => rm_lookup[A, Unit](aa, ba),
                     a, b)

def shows_rpo_repr[A : show](prs: List[((A, nat), (nat, order_tag))]):
      (List[Char]) => List[Char] =
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List('R', 'P', 'O', ' ', 'w', 'i', 't',
     'h', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g',
     ' ', 'p', 'r', 'e', 'c', 'e', 'd', 'e', 'n', 'c', 'e')),
                     comp[List[Char], List[Char],
                           List[Char]](shows_nl,
comp[List[Char], List[Char],
      List[Char]](shows_map[((A, nat),
                              (nat, order_tag))]((a:
            ((A, nat), (nat, order_tag)))
           => {
                val (aa, b): ((A, nat), (nat, order_tag)) = a;
                ({
                   val (f, n): (A, nat) = aa;
                   (ab: (nat, order_tag)) =>
                     {
                       val (pr, _): (nat, order_tag) = ab;
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('p', 'r', 'e',
                          'c', 'e', 'd', 'e', 'n', 'c', 'e', '(')),
  comp[List[Char], List[Char],
        List[Char]]((ac: List[Char]) => shows_prec[A](Zero_nat, f, ac),
                     comp[List[Char], List[Char],
                           List[Char]](shows_string.apply(List('[')),
comp[List[Char], List[Char],
      List[Char]](shows_prec_nat(Zero_nat, n),
                   comp[List[Char], List[Char],
                         List[Char]](shows_string.apply(List(']', ')', ' ', '=',
                      ' ')),
                                      comp[List[Char], List[Char],
    List[Char]](shows_prec_nat(Zero_nat, pr), shows_nl))))))
                     }
                 })(b)
              },
          prs),
                   comp[List[Char], List[Char],
                         List[Char]](shows_nl,
                                      comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('p', 'r', 'e', 'c', 'e', 'd', 'e', 'n',
 'c', 'e', '(', '_', ')', ' ', '=', ' ', '0')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_nl,
                                    comp[List[Char], List[Char],
  List[Char]](shows_string.apply(List('a', 'n', 'd', ' ', 't', 'h', 'e', ' ',
                                       'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n',
                                       'g', ' ', 's', 't', 'a', 't', 'u', 's')),
               comp[List[Char], List[Char],
                     List[Char]](shows_nl,
                                  comp[List[Char], List[Char],
List[Char]](shows_map[((A, nat),
                        (nat, order_tag))]((a: ((A, nat), (nat, order_tag))) =>
     {
       val (aa, b): ((A, nat), (nat, order_tag)) = a;
       ({
          val (f, n): (A, nat) = aa;
          (ab: (nat, order_tag)) =>
            {
              val (_, s): (nat, order_tag) = ab;
              comp[List[Char], List[Char],
                    List[Char]](shows_string.apply(List('s', 't', 'a', 't', 'u',
                 's', '(')),
                                 comp[List[Char], List[Char],
                                       List[Char]]((ac: List[Char]) =>
             shows_prec[A](Zero_nat, f, ac),
            comp[List[Char], List[Char],
                  List[Char]](shows_string.apply(List('[')),
                               comp[List[Char], List[Char],
                                     List[Char]](shows_prec_nat(Zero_nat, n),
          comp[List[Char], List[Char],
                List[Char]](shows_string.apply(List(']', ')', ' ', '=', ' ')),
                             comp[List[Char], List[Char],
                                   List[Char]](shows_string.apply(s match {
                            case Lex => List('l', 'e', 'x')
                            case Mul => List('m', 'u', 'l')
                          }),
        shows_nl))))))
            }
        })(b)
     },
    prs),
             comp[List[Char], List[Char],
                   List[Char]](shows_nl,
                                comp[List[Char], List[Char],
                                      List[Char]](shows_string.apply(List('s',
                                   't', 'a', 't', 'u', 's', '(', '_', ')', ' ',
                                   '=', ' ', 'l', 'e', 'x')),
           shows_nl)))))))))))

def scnp_af_to_af[A](pia: ((A, nat)) => List[(nat, nat)],
                      pi: ((A, nat)) => nat => Boolean):
      ((A, nat)) => nat => Boolean =
  (a: (A, nat)) =>
    {
      val (f, n): (A, nat) = a
      val is: List[nat] =
        mapa[(nat, nat),
              nat]((aa: (nat, nat)) => fst[nat, nat](aa), pia((f, n)));
      (if (list_ex[nat]((aa: nat) => less_eq_nat(n, aa), is))
        {
          val isa: nat => Boolean = pi((f, n));
          (i: nat) => isa(i) || membera[nat](is, i)
        }
        else (aa: nat) => membera[nat](is, aa))
    }

def s_isEmpty[A](isEmpty: A): A = isEmpty

def subst_compose[A, B, C, D](sigma: A => term[B, C], tau: C => term[B, D]):
      A => term[B, D] =
  (x: A) => subst_apply_term[B, C, D](sigma(x), tau)

def mk_subst[A : equal, B](xts: List[(A, term[B, A])]): A => term[B, A] =
  (x: A) =>
    lookupa[A, term[B, A]](x, xts) match {
      case None => Var[A, B](x)
      case Some(t) => t
    }

def shows_term[A, B](fun: A => (List[Char]) => List[Char],
                      vara: B => (List[Char]) => List[Char], x2: term[A, B]):
      (List[Char]) => List[Char] =
  (fun, vara, x2) match {
  case (fun, vara, Var(x)) => vara(x)
  case (fun, vara, Fun(f, ts)) =>
    comp[List[Char], List[Char],
          List[Char]](fun(f),
                       shows_list_gen[(List[Char]) =>
List[Char]](id[(List[Char]) => List[Char]], Nil, List('('), List(',', ' '),
             List(')'),
             mapa[term[A, B],
                   (List[Char]) =>
                     List[Char]]((a: term[A, B]) =>
                                   shows_term[A, B](fun, vara, a),
                                  ts)))
}

def shows_prec_term[A : show, B : show](d: nat):
      (term[A, B]) => (List[Char]) => List[Char] =
  (a: term[A, B]) =>
    shows_term[A, B]((aa: A) =>
                       (b: List[Char]) => shows_prec[A](Zero_nat, aa, b),
                      (aa: B) =>
                        (b: List[Char]) => shows_prec[B](Zero_nat, aa, b),
                      a)

def shows_list_term[A : show, B : show](is: List[term[A, B]]):
      (List[Char]) => List[Char] =
  shows_list_aux[term[A, B]](shows_prec_term[A, B](Zero_nat), is)

implicit def show_term[A : show, B : show]: show[term[A, B]] = new
  show[term[A, B]] {
  val `Ceta.shows_prec` = (a: nat, b: term[A, B], c: List[Char]) =>
    (shows_prec_term[A, B](a)).apply(b).apply(c)
  val `Ceta.shows_list` = (a: List[term[A, B]], b: List[Char]) =>
    (shows_list_term[A, B](a)).apply(b)
}

def full_af[A](fn: (A, nat), i: nat): Boolean = true

def af[A, B, C](x0: redtriple_ext[A, B, C]): ((A, nat)) => nat => Boolean = x0
  match {
  case Redtriple_ext(valid, s, ns, nst, af, mono, desc, more) => af
}

def ns[A, B, C](x0: redtriple_ext[A, B, C]):
      ((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit] =
  x0 match {
  case Redtriple_ext(valid, s, ns, nst, af, mono, desc, more) => ns
}

abstract sealed class root_redtriple_ext[+A, +B, +C]
final case class
  Root_redtriple_ext[A, B,
                      C](a: sum[(List[Char]) => List[Char], Unit],
                          b: ((term[A, B], term[A, B])) =>
                               sum[(List[Char]) => List[Char], Unit],
                          c: ((term[A, B], term[A, B])) =>
                               sum[(List[Char]) => List[Char], Unit],
                          d: ((term[A, B], term[A, B])) =>
                               sum[(List[Char]) => List[Char], Unit],
                          e: ((A, nat)) => nat => Boolean,
                          f: ((A, nat)) => nat => Boolean,
                          g: (List[Char]) => List[Char], h: C)
  extends root_redtriple_ext[A, B, C]

def sa[A, B, C](x0: root_redtriple_ext[A, B, C]):
      ((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit] =
  x0 match {
  case Root_redtriple_ext(valid, s, ns, nst, af, aft, desc, more) => s
}

implicit def equal_prod[A : equal, B : equal]: equal[(A, B)] = new equal[(A, B)]
  {
  val `Ceta.equal` = (a: (A, B), b: (A, B)) => equal_proda[A, B](a, b)
}

def defs_list[A : equal, B](x0: List[(term[A, B], term[A, B])]):
      List[(A, nat)] =
  x0 match {
  case Nil => Nil
  case (Fun(f, ss), r) :: ts =>
    {
      val fs: List[(A, nat)] = defs_list[A, B](ts)
      val n: nat = size_list[term[A, B]](ss);
      (if (membera[(A, nat)](fs, (f, n))) fs else (f, n) :: fs)
    }
  case (Var(x), r) :: ts => defs_list[A, B](ts)
}

def shows_pos(x0: pos): (List[Char]) => List[Char] = x0 match {
  case PCons(i, Empty) => shows_prec_nat(Zero_nat, i)
  case PCons(i, PCons(v, va)) =>
    comp[List[Char], List[Char],
          List[Char]](shows_prec_nat(Zero_nat, i), shows_pos(PCons(v, va)))
  case Empty => shows_string.apply(List('e'))
}

def shows_rule[A, B](fun: A => (List[Char]) => List[Char],
                      vara: B => (List[Char]) => List[Char], arr: List[Char],
                      x3: (term[A, B], term[A, B])):
      (List[Char]) => List[Char] =
  (fun, vara, arr, x3) match {
  case (fun, vara, arr, (l, r)) =>
    comp[List[Char], List[Char],
          List[Char]](shows_term[A, B](fun, vara, l),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(arr),
  shows_term[A, B](fun, vara, r)))
}

def shows_rules[A, B](fun: A => (List[Char]) => List[Char],
                       vara: B => (List[Char]) => List[Char], arr: List[Char],
                       trs: List[(term[A, B], term[A, B])]):
      (List[Char]) => List[Char] =
  comp[List[Char], List[Char],
        List[Char]](shows_list_gen[(term[A, B],
                                     term[A,
   B])]((a: (term[A, B], term[A, B])) => shows_rule[A, B](fun, vara, arr, a),
         Nil, Nil, List(newline), Nil, trs),
                     shows_nl)

def shows_trs[A, B](fun: A => (List[Char]) => List[Char],
                     vara: B => (List[Char]) => List[Char], name: List[Char],
                     arr: List[Char], r: List[(term[A, B], term[A, B])]):
      (List[Char]) => List[Char] =
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(name),
                     comp[List[Char], List[Char],
                           List[Char]](shows_nl,
shows_rules[A, B](fun, vara, arr, r)))

def apply_args[A, B](a: A, t: term[A, B], x2: List[term[A, B]]): term[A, B] =
  (a, t, x2) match {
  case (a, t, Nil) => t
  case (a, t, s :: ss) => apply_args[A, B](a, Fun[A, B](a, List(t, s)), ss)
}

def get_symbol[A](sm: A => nat => List[A], f: A, n: nat, i: nat): A =
  nth[A]((sm(f))(n), i)

def replace_impl[A : equal](a: A, bs: List[A], m: List[A]): List[A] =
  (if (membera[A](m, a)) bs ++ filter[A]((b: A) => ! (eq[A](b, a)), m) else m)

def letters: List[Char] =
  List('a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
        'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', 'A', 'B',
        'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
        'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '_', '0', '1', '2',
        '3', '4', '5', '6', '7', '8', '9', '&', ';', ':')

def alter10a[A, B, C, D, E, F, G, H, I,
              J](x0: alter[A, B, C, D, E, F, G, H, I, J]):
      J =
  x0 match {
  case Alter10(x) => x
}

def sum_lpoly[A](uu: A => A => A, x1: List[A], ys: List[A]): List[A] =
  (uu, x1, ys) match {
  case (uu, Nil, ys) => ys
  case (uv, v :: va, Nil) => v :: va
  case (pl, x :: xs, y :: ys) => (pl(x))(y) :: sum_lpoly[A](pl, xs, ys)
}

def lpoly_of[A : equal](uu: A, uv: A, uw: A => A => A, ux: A => A => A,
                         x4: tpoly[nat, A]):
      sum[List[Char], (A, List[A])] =
  (uu, uv, uw, ux, x4) match {
  case (uu, uv, uw, ux, PNum(i)) => Inr[(A, List[A]), List[Char]]((i, Nil))
  case (ze, on, pl, ti, PVar(x)) =>
    Inr[(A, List[A]), List[Char]]((ze, replicate[A](x, ze) ++ List(on)))
  case (ze, on, pl, ti, PSum(Nil)) => Inr[(A, List[A]), List[Char]]((ze, Nil))
  case (ze, on, pl, ti, PSum(p :: ps)) =>
    bindb[List[Char], (A, List[A]),
           (A, List[A])](lpoly_of[A](ze, on, pl, ti, p),
                          (a: (A, List[A])) =>
                            {
                              val (cp, ncp): (A, List[A]) = a;
                              bindb[List[Char], (A, List[A]),
                                     (A, List[A])](lpoly_of[A](ze, on, pl, ti,
                        PSum[nat, A](ps)),
            (aa: (A, List[A])) =>
              {
                val (cq, ncq): (A, List[A]) = aa;
                Inr[(A, List[A]),
                     List[Char]](((pl(cp))(cq), sum_lpoly[A](pl, ncp, ncq)))
              })
                            })
  case (ze, on, pl, ti, PMult(Nil)) => Inr[(A, List[A]), List[Char]]((on, Nil))
  case (ze, on, pl, ti, PMult(p :: ps)) =>
    bindb[List[Char], (A, List[A]),
           (A, List[A])](lpoly_of[A](ze, on, pl, ti, p),
                          (a: (A, List[A])) =>
                            {
                              val (cp, ncp): (A, List[A]) = a;
                              bindb[List[Char], (A, List[A]),
                                     (A, List[A])](lpoly_of[A](ze, on, pl, ti,
                        PMult[nat, A](ps)),
            (aa: (A, List[A])) =>
              {
                val (cq, ncq): (A, List[A]) = aa;
                (if (list_all[A]((ab: A) => eq[A](ze, ab), ncp))
                  Inr[(A, List[A]),
                       List[Char]](((ti(cp))(cq), mapa[A, A](ti(cp), ncq)))
                  else (if (list_all[A]((ab: A) => eq[A](ze, ab), ncq))
                         Inr[(A, List[A]),
                              List[Char]](((ti(cp))(cq),
    mapa[A, A]((x: A) => (ti(x))(cq), ncp)))
                         else Inl[List[Char],
                                   (A, List[A])](List('c', 'a', 'n', 'n', 'o',
               't', ' ', 't', 'r', 'a', 'n', 's', 'f', 'o', 'r', 'm', ' ', 'n',
               'o', 'n', '-', 'l', 'i', 'n', 'e', 'a', 'r', ' ', 'p', 'o', 'l',
               'y', 'n', 'o', 'm', 'i', 'a', 'l', ' ', 't', 'o', ' ', 'l', 'i',
               'n', 'e', 'a', 'r', ' ', 'p', 'o', 'l', 'y', 'n', 'o', 'm', 'i',
               'a', 'l'))))
              })
                            })
}

def strategy[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], strategy[A, List[Char]]] =
  (a: xml) =>
    xml1element[strategy[A, List[Char]],
                 strategy[A, List[Char]]](List('s', 't', 'r', 'a', 't', 'e',
        'g', 'y'),
   (aa: xml) =>
     xml_options[strategy[A, List[Char]]](List((List('i', 'n', 'n', 'e', 'r',
              'm', 'o', 's', 't'),
         (ab: xml) =>
           xml0elements[strategy[A, List[Char]]](List('i', 'n', 'n', 'e', 'r',
               'm', 'o', 's', 't'),
          Innermost, ab)),
        (List('i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', 'L', 'h', 's', 's'),
          (ab: xml) =>
            xml_many_elements[term[A, List[Char]],
                               strategy[A,
 List[Char]]](List('i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', 'L', 'h', 's',
                    's'),
               (ac: xml) => term[A](xml2name, ac),
               (ac: List[term[A, List[Char]]]) =>
                 Innermost_Q[A, List[Char]](ac),
               ab))),
   aa),
   id[strategy[A, List[Char]]], a)

def icap_impl[A : equal](nf: Boolean, isQnf: (term[A, List[Char]]) => Boolean,
                          ls: List[term[A, List[Char]]],
                          s: List[term[A, List[Char]]],
                          sx: (List[Char]) => Boolean, x5: term[A, List[Char]]):
      term[A, sum[Unit, List[Char]]] =
  (nf, isQnf, ls, s, sx, x5) match {
  case (nf, isQnf, ls, s, sx, Var(x)) =>
    (if (nf && sx(x)) Var[sum[Unit, List[Char]], A](Inr[List[Char], Unit](x))
      else Var[sum[Unit, List[Char]], A](Inl[Unit, List[Char]](())))
  case (nf, isQnf, ls, s, sx, Fun(f, ts)) =>
    {
      val t: term[A, sum[Unit, List[Char]]] =
        Fun[A, sum[Unit,
                    List[Char]]](f, mapa[term[A, List[Char]],
  term[A, sum[Unit,
               List[Char]]]]((a: term[A, List[Char]]) =>
                               icap_impl[A](nf, isQnf, ls, s, sx, a),
                              ts));
      (if (list_ex[term[A, List[Char]]]((l: term[A, List[Char]]) =>
  mgu_class[A](t, l) match {
    case None => false
    case Some(mu) =>
      list_all[term[A, List[Char]]]((u: term[A, List[Char]]) =>
                                      isQnf(subst_apply_term[A, List[Char],
                      List[Char]](map_vars[List[Char], List[Char],
    A]((a: List[Char]) => 'y' :: a, u),
                                   mu)),
                                     args[A, List[Char]](l)) &&
        list_all[term[A, List[Char]]]((u: term[A, List[Char]]) =>
isQnf(subst_apply_term[A, List[Char], List[Char]](u, mu)),
                                       s)
  },
 ls))
        Var[sum[Unit, List[Char]], A](Inl[Unit, List[Char]](())) else t)
    }
}

def nF_subset[A : equal, B : equal](ls: List[term[A, B]], q: List[term[A, B]]):
      Boolean =
  list_all[term[A, B]]((l: term[A, B]) =>
                         list_ex[term[A, B]]((t: term[A, B]) =>
       list_ex[term[A, B]]((u: term[A, B]) => matches[A, B, B](u, t),
                            supteq_list[A, B](l)),
      q),
                        ls)

def less_list[A : equal : order](xs: List[A], x1: List[A]): Boolean = (xs, x1)
  match {
  case (x :: xs, y :: ys) =>
    less[A](x, y) || eq[A](x, y) && less_list[A](xs, ys)
  case (Nil, x :: xs) => true
  case (xs, Nil) => false
}

def less_eq_list[A : equal : order](x0: List[A], xs: List[A]): Boolean =
  (x0, xs) match {
  case (x :: xs, y :: ys) =>
    less[A](x, y) || eq[A](x, y) && less_eq_list[A](xs, ys)
  case (Nil, xs) => true
  case (x :: xs, Nil) => false
}

implicit def ord_list[A : equal : order]: ord[List[A]] = new ord[List[A]] {
  val `Ceta.less_eq` = (a: List[A], b: List[A]) => less_eq_list[A](a, b)
  val `Ceta.less` = (a: List[A], b: List[A]) => less_list[A](a, b)
}

def mat_arc_posI[A](ap: A => Boolean, m: List[List[A]]): Boolean =
  ap(nth[A](nth[List[A]](m, Zero_nat), Zero_nat))

def poly_var[A : equal, B](p: List[(List[(A, nat)], B)], v: A): Boolean =
  membera[A](maps[(List[(A, nat)], B),
                   A](comp[List[(A, nat)], List[A],
                            (List[(A, nat)],
                              B)]((a: List[(A, nat)]) =>
                                    mapa[(A, nat),
  A]((aa: (A, nat)) => fst[A, nat](aa), a),
                                   (a: (List[(A, nat)], B)) =>
                                     fst[List[(A, nat)], B](a)),
                       p),
              v)

abstract sealed class cert_result
final case class Certified(a: List[Char]) extends cert_result
final case class Unsupported(a: List[Char]) extends cert_result
final case class Error(a: List[Char]) extends cert_result

def rs_empty[A : linorder]: rbt[A, Unit] =
  s_empty[rbt[A, Unit]](rm_empty[A, Unit])

def rs_union[A : linorder]: (rbt[A, Unit]) => (rbt[A, Unit]) => rbt[A, Unit] =
  s_union[(rbt[A, Unit]) => (rbt[A, Unit]) => rbt[A, Unit]](rm_add[A, Unit])

def rules_with[A, B : equal : linorder,
                C](p: A => Boolean,
                    m: rbt[(B, nat), List[(A, (term[B, C], term[B, C]))]]):
      List[(term[B, C], term[B, C])] =
  map_filter[(A, (term[B, C], term[B, C])),
              (term[B, C],
                term[B, C])]((x: (A, (term[B, C], term[B, C]))) =>
                               (if ((comp[A, Boolean,
   (A, (term[B, C],
         term[B, C]))](p, (a: (A, (term[B, C], term[B, C]))) =>
                            fst[A, (term[B, C], term[B, C])](a))).apply(x))
                                 Some[(term[B, C],
term[B, C])](snd[A, (term[B, C], term[B, C])](x))
                                 else None),
                              values[(B, nat),
                                      (A, (term[B, C], term[B, C]))](m))

def s_iteratei[A, B,
                C](iteratei:
                     (A => Boolean) => (B => Unit => A => A) => C => A => A,
                    c: A => Boolean, f: B => A => A, s: C, sigma_0: A):
      A =
  (((iteratei(c))((k: B) => (_: Unit) => f(k)))(s))(sigma_0)

abstract sealed class tp_ops_ext[+A, +B, +C, +D]
final case class
  Tp_ops_ext[A, B, C,
              D](a: A => ((term[B, C]) => Boolean,
                           (((term[B, C], term[B, C])) => Boolean,
                             ((term[B, C], term[B, C])) => Boolean)),
                  b: A => List[term[B, C]],
                  c: A => List[(term[B, C], term[B, C])],
                  d: A => List[(term[B, C], term[B, C])],
                  e: A => List[(term[B, C], term[B, C])], f: A => Boolean,
                  g: A => (term[B, C]) => Boolean, h: A => Boolean,
                  i: A => ((B, nat)) => List[(term[B, C], term[B, C])],
                  j: A => (List[(term[B, C], term[B, C])]) =>
                            (List[(term[B, C], term[B, C])]) => A,
                  k: A => (List[(term[B, C], term[B, C])]) =>
                            (List[(term[B, C], term[B, C])],
                              List[(term[B, C], term[B, C])]),
                  l: (List[term[B, C]]) =>
                       (List[(term[B, C], term[B, C])]) =>
                         (List[(term[B, C], term[B, C])]) => A,
                  m: D)
  extends tp_ops_ext[A, B, C, D]

def qa[A, B, C, D](x0: tp_ops_ext[A, B, C, D]): A => List[term[B, C]] = x0 match
  {
  case Tp_ops_ext(qreltrs, q, r, rw, rules, q_empty, is_QNF,
                   nFQ_subset_NF_rules, rules_map, delete_R_Rw, split_rules, mk,
                   more)
    => q
}

def mka[A, B, C, D](x0: tp_ops_ext[A, B, C, D]):
      (List[term[B, C]]) =>
        (List[(term[B, C], term[B, C])]) =>
          (List[(term[B, C], term[B, C])]) => A =
  x0 match {
  case Tp_ops_ext(qreltrs, q, r, rw, rules, q_empty, is_QNF,
                   nFQ_subset_NF_rules, rules_map, delete_R_Rw, split_rules, mk,
                   more)
    => mk
}

def delete_R_Rwa[A, B, C, D](x0: tp_ops_ext[A, B, C, D]):
      A => (List[(term[B, C], term[B, C])]) =>
             (List[(term[B, C], term[B, C])]) => A =
  x0 match {
  case Tp_ops_ext(qreltrs, q, r, rw, rules, q_empty, is_QNF,
                   nFQ_subset_NF_rules, rules_map, delete_R_Rw, split_rules, mk,
                   more)
    => delete_R_Rw
}

def split_rulesa[A, B, C, D](x0: tp_ops_ext[A, B, C, D]):
      A => (List[(term[B, C], term[B, C])]) =>
             (List[(term[B, C], term[B, C])], List[(term[B, C], term[B, C])]) =
  x0 match {
  case Tp_ops_ext(qreltrs, q, r, rw, rules, q_empty, is_QNF,
                   nFQ_subset_NF_rules, rules_map, delete_R_Rw, split_rules, mk,
                   more)
    => split_rules
}

def split_tt[A, B : show,
              C : show](i: tp_ops_ext[A, B, C, Unit], tp: A,
                         r_remove: List[(term[B, C], term[B, C])]):
      (A, A) =
  {
    val (r, rw):
          (List[(term[B, C], term[B, C])], List[(term[B, C], term[B, C])])
      = (split_rulesa[A, B, C, Unit](i)).apply(tp).apply(r_remove)
    val tp1: A =
      (mka[A, B, C,
            Unit](i)).apply((qa[A, B, C, Unit](i)).apply(tp)).apply(r).apply(rw)
    val a: A =
      (delete_R_Rwa[A, B, C,
                     Unit](i)).apply(tp).apply(r_remove).apply(r_remove);
    (tp1, a)
  }

def afa[A, B, C](x0: root_redtriple_ext[A, B, C]):
      ((A, nat)) => nat => Boolean =
  x0 match {
  case Root_redtriple_ext(valid, s, ns, nst, af, aft, desc, more) => af
}

def aft[A, B, C](x0: root_redtriple_ext[A, B, C]):
      ((A, nat)) => nat => Boolean =
  x0 match {
  case Root_redtriple_ext(valid, s, ns, nst, af, aft, desc, more) => aft
}

def nsa[A, B, C](x0: root_redtriple_ext[A, B, C]):
      ((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit] =
  x0 match {
  case Root_redtriple_ext(valid, s, ns, nst, af, aft, desc, more) => ns
}

def nst[A, B, C](x0: redtriple_ext[A, B, C]):
      ((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit] =
  x0 match {
  case Redtriple_ext(valid, s, ns, nst, af, mono, desc, more) => nst
}

def aarity_term[A : equal, B](a: A, sm: A => nat => List[A], t: term[A, B]):
      Option[nat] =
  unapp[A, B](a, t) match {
    case (Var(_), ts) => None
    case (Fun(f, ss), ts) =>
      Some[nat](minus_nat(aarity[A](sm, f, size_list[term[A, B]](ss)),
                           size_list[term[A, B]](ts)))
  }

def uncurry_top[A : equal,
                 B](a: A, n: nat, sm: A => nat => List[A], x3: term[A, B]):
      term[A, B] =
  (a, n, sm, x3) match {
  case (a, n, sm, Fun(f, ts)) =>
    {
      val mt: (List[term[A, B]]) => List[term[A, B]] =
        (aa: List[term[A, B]]) =>
          mapa[term[A, B],
                term[A, B]]((ab: term[A, B]) =>
                              map_term_wa[A, A,
   B]((fa: A) => (na: nat) => get_symbol[A](sm, fa, na, Zero_nat), ab),
                             aa)
      val (h, m): (A, nat) = the[(A, nat)](root[A, B](hd[term[A, B]](ts)));
      (if (eq[A](f, a) &&
             (equal_nata(size_list[term[A, B]](ts), n) &&
               (! (is_Var[A, B](hd[term[A, B]](ts))) &&
                 ! (equal_nata(aarity[A](sm, h, m), Zero_nat)))))
        {
          val (Fun(g, ss)): term[A, B] = hd[term[A, B]](ts);
          Fun[A, B](get_symbol[A](sm, g, size_list[term[A, B]](ss), one_nat),
                     mt(ss ++ tl[term[A, B]](ts)))
        }
        else Fun[A, B](get_symbol[A](sm, f, size_list[term[A, B]](ts),
                                      Zero_nat),
                        mt(ts)))
    }
  case (a, n, sm, Var(x)) => Var[B, A](x)
}

def option_to_list[A](x0: Option[A]): List[A] = x0 match {
  case Some(a) => List(a)
  case None => Nil
}

def xml1to2elements[A, B,
                     C](tag: List[Char], p1: xml => sum[List[Char], A],
                         p2: xml => sum[List[Char], B], f: A => Option[B] => C,
                         x4: xml):
      sum[List[Char], C] =
  (tag, p1, p2, f, x4) match {
  case (tag, p1, p2, f, Xml(name, atts, cs, text)) =>
    {
      val l: nat = size_list[xml](cs);
      (if (equal_lista[Char](name, tag) &&
             (nulla[(List[Char], List[Char])](atts) &&
               (less_eq_nat(one_nat, l) &&
                 (less_eq_nat(l, nat(BigInt(2))) &&
                   is_none[List[Char]](text)))))
        bindb[List[Char], A,
               C](p1(nth[xml](cs, Zero_nat)),
                   (x1: A) =>
                     (if (equal_nata(l, nat(BigInt(2))))
                       bindb[List[Char], B,
                              C](p2(nth[xml](cs, one_nat)),
                                  (x2: B) =>
                                    Inr[C, List[Char]]((f(x1))(Some[B](x2))))
                       else Inr[C, List[Char]]((f(x1))(None))))
        else no_transform[C](tag, Xml(name, atts, cs, text)))
    }
}

def arith_fun(xml: xml): sum[List[Char], arithFun] =
  xml1element[arithFun,
               arithFun](List('a', 'r', 'i', 't', 'h', 'F', 'u', 'n', 'c', 't',
                               'i', 'o', 'n'),
                          (a: xml) =>
                            xml_options[arithFun](List((List('n', 'a', 't', 'u',
                      'r', 'a', 'l'),
                 (aa: xml) =>
                   xml_change[nat, arithFun]((ab: xml) =>
       nata(List('n', 'a', 't', 'u', 'r', 'a', 'l'), ab),
      (ab: nat) => Const(ab), aa)),
                (List('v', 'a', 'r', 'i', 'a', 'b', 'l', 'e'),
                  (aa: xml) =>
                    xml_change[nat, arithFun]((ab: xml) =>
        nata(List('v', 'a', 'r', 'i', 'a', 'b', 'l', 'e'), ab),
       (n: nat) => Arg(minus_nat(n, one_nat)), aa)),
                (List('s', 'u', 'm'),
                  (aa: xml) =>
                    xml_many_elements[arithFun,
                                       arithFun](List('s', 'u', 'm'),
          (ab: xml) => arith_fun(ab), (ab: List[arithFun]) => Sum(ab), aa)),
                (List('p', 'r', 'o', 'd', 'u', 'c', 't'),
                  (aa: xml) =>
                    xml_many_elements[arithFun,
                                       arithFun](List('p', 'r', 'o', 'd', 'u',
               'c', 't'),
          (ab: xml) => arith_fun(ab), (ab: List[arithFun]) => Prod(ab), aa)),
                (List('m', 'i', 'n'),
                  (aa: xml) =>
                    xml_many_elements[arithFun,
                                       arithFun](List('m', 'i', 'n'),
          (ab: xml) => arith_fun(ab), (ab: List[arithFun]) => Min(ab), aa)),
                (List('m', 'a', 'x'),
                  (aa: xml) =>
                    xml_many_elements[arithFun,
                                       arithFun](List('m', 'a', 'x'),
          (ab: xml) => arith_fun(ab), (ab: List[arithFun]) => Max(ab), aa))),
           a),
                          id[arithFun], xml)

def int_coeff: xml => sum[List[Char], BigInt] =
  (a: xml) => int(List('i', 'n', 't', 'e', 'g', 'e', 'r'), a)

def vec_coeff[A](xml2coeff: xml => sum[List[Char], A]):
      xml => sum[List[Char], List[A]] =
  (a: xml) =>
    xml_many_elements[A, List[A]](List('v', 'e', 'c', 't', 'o', 'r'),
                                   (aa: xml) =>
                                     xml1element[A,
          A](List('c', 'o', 'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n', 't'),
              xml2coeff, id[A], aa),
                                   id[List[A]], a)

def mat_coeff[A](ze: A, xml2coeff: xml => sum[List[Char], A]):
      xml => sum[List[Char], List[List[A]]] =
  (a: xml) =>
    xml_options[List[List[A]]](List((List('m', 'a', 't', 'r', 'i', 'x'),
                                      (aa: xml) =>
xml_many_elements[List[A],
                   List[List[A]]](List('m', 'a', 't', 'r', 'i', 'x'),
                                   vec_coeff[A](xml2coeff), id[List[List[A]]],
                                   aa)),
                                     (List('v', 'e', 'c', 't', 'o', 'r'),
                                       (aa: xml) =>
 xml_change[List[A],
             List[List[A]]](vec_coeff[A](xml2coeff),
                             (v: List[A]) =>
                               v :: replicate[List[A]](minus_nat(size_list[A](v),
                          one_nat),
                replicate[A](size_list[A](v), ze)),
                             aa))),
                                a)

def status_precedence[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], List[((A, nat), (nat, order_tag))]] =
  (a: xml) =>
    xml_many_elements[((A, nat), (nat, order_tag)),
                       List[((A, nat),
                              (nat, order_tag))]](List('s', 't', 'a', 't', 'u',
                's', 'P', 'r', 'e', 'c', 'e', 'd', 'e', 'n', 'c', 'e'),
           (aa: xml) =>
             xml4elements[A, nat, nat, order_tag,
                           ((A, nat),
                             (nat, order_tag))](List('s', 't', 'a', 't', 'u',
              's', 'P', 'r', 'e', 'c', 'e', 'd', 'e', 'n', 'c', 'e', 'E', 'n',
              't', 'r', 'y'),
         xml2name, (ab: xml) => nata(List('a', 'r', 'i', 't', 'y'), ab),
         (ab: xml) =>
           nata(List('p', 'r', 'e', 'c', 'e', 'd', 'e', 'n', 'c', 'e'), ab),
         (ab: xml) =>
           xml_options[order_tag](List((List('l', 'e', 'x'),
 (ac: xml) => xml0elements[order_tag](List('l', 'e', 'x'), Lex, ac)),
(List('m', 'u', 'l'),
  (ac: xml) => xml0elements[order_tag](List('m', 'u', 'l'), Mul, ac))),
                                   ab),
         (f: A) => (ab: nat) => (p: nat) => (s: order_tag) => ((f, ab), (p, s)),
         aa),
           id[List[((A, nat), (nat, order_tag))]], a)

def path_order[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], redtriple_impl[A]] =
  (a: xml) =>
    xml1to2elements[List[((A, nat), (nat, order_tag))],
                     List[((A, nat), af_entry)],
                     redtriple_impl[A]](List('p', 'a', 't', 'h', 'O', 'r', 'd',
      'e', 'r'),
 status_precedence[A](xml2name), afs[A](xml2name),
 (prec_tau: List[((A, nat), (nat, order_tag))]) =>
   (afo: Option[List[((A, nat), af_entry)]]) =>
     {
       val aa: List[((A, nat), af_entry)] =
         afo match {
           case None => Nil
           case Some(af) => af
         };
       Rpo[A](prec_tau, aa)
     },
 a)

abstract sealed class poly_c_ext[+A, +B]
final case class
  Poly_c_ext[A, B](a: A => Boolean, b: A, c: A, d: A, e: A => A => A,
                    f: A => A => A, g: A => A, h: A => A => Boolean,
                    i: A => A => Boolean, j: Boolean, k: List[Char],
                    l: A => Boolean, m: A => Boolean, n: B)
  extends poly_c_ext[A, B]

def gt[A, B](x0: poly_c_ext[A, B]): A => A => Boolean = x0 match {
  case Poly_c_ext(wf, zero, one, default, plus, times, max0, ge, gt, mono,
                   description, arc_pos, check_mono, more)
    => gt
}

def fit_length[A](uu: A, x1: nat, uv: List[A]): List[A] = (uu, x1, uv) match {
  case (uu, Zero_nat, uv) => Nil
  case (a, Suc(v), Nil) => replicate[A](Suc(v), a)
  case (a, Suc(n), b :: as) => b :: fit_length[A](a, n, as)
}

def polynomial[A](xml2coeff: xml => sum[List[Char], A], x: xml):
      sum[List[Char], tpoly[nat, A]] =
  xml1element[tpoly[nat, A],
               tpoly[nat, A]](List('p', 'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a',
                                    'l'),
                               (a: xml) =>
                                 xml_options[tpoly[nat,
            A]](List((List('c', 'o', 'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n',
                            't'),
                       (aa: xml) =>
                         xml1element[A, tpoly[nat,
       A]](List('c', 'o', 'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n', 't'),
            xml2coeff, (ab: A) => PNum[A, nat](ab), aa)),
                      (List('v', 'a', 'r', 'i', 'a', 'b', 'l', 'e'),
                        (aa: xml) =>
                          xml_change[nat, tpoly[nat,
         A]]((ab: xml) =>
               nata(List('v', 'a', 'r', 'i', 'a', 'b', 'l', 'e'), ab),
              (n: nat) => PVar[nat, A](minus_nat(n, one_nat)), aa)),
                      (List('s', 'u', 'm'),
                        (aa: xml) =>
                          xml_many_elements[tpoly[nat, A],
     tpoly[nat, A]](List('s', 'u', 'm'),
                     (ab: xml) => polynomial[A](xml2coeff, ab),
                     (ab: List[tpoly[nat, A]]) => PSum[nat, A](ab), aa)),
                      (List('p', 'r', 'o', 'd', 'u', 'c', 't'),
                        (aa: xml) =>
                          xml_many_elements[tpoly[nat, A],
     tpoly[nat, A]](List('p', 'r', 'o', 'd', 'u', 'c', 't'),
                     (ab: xml) => polynomial[A](xml2coeff, ab),
                     (ab: List[tpoly[nat, A]]) => PMult[nat, A](ab), aa))),
                 a),
                               id[tpoly[nat, A]], x)

def onea[A, B](x0: poly_c_ext[A, B]): A = x0 match {
  case Poly_c_ext(wf, zero, one, default, plus, times, max0, ge, gt, mono,
                   description, arc_pos, check_mono, more)
    => one
}

implicit def ab_semigroup_mult_int: ab_semigroup_mult[BigInt] = new
  ab_semigroup_mult[BigInt] {
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
}

implicit def comm_semiring_int: comm_semiring[BigInt] = new
  comm_semiring[BigInt] {
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
}

implicit def comm_semiring_0_int: comm_semiring_0[BigInt] = new
  comm_semiring_0[BigInt] {
  val `Ceta.zero` = zero_inta
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
}

implicit def comm_monoid_mult_int: comm_monoid_mult[BigInt] = new
  comm_monoid_mult[BigInt] {
  val `Ceta.one` = one_inta
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
}

implicit def comm_semiring_1_int: comm_semiring_1[BigInt] = new
  comm_semiring_1[BigInt] {
  val `Ceta.zero` = zero_inta
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
  val `Ceta.one` = one_inta
}

def monoa[A, B](x0: poly_c_ext[A, B]): Boolean = x0 match {
  case Poly_c_ext(wf, zero, one, default, plus, times, max0, ge, gt, mono,
                   description, arc_pos, check_mono, more)
    => mono
}

def plusa[A, B](x0: poly_c_ext[A, B]): A => A => A = x0 match {
  case Poly_c_ext(wf, zero, one, default, plus, times, max0, ge, gt, mono,
                   description, arc_pos, check_mono, more)
    => plus
}

def zeroa[A, B](x0: poly_c_ext[A, B]): A = x0 match {
  case Poly_c_ext(wf, zero, one, default, plus, times, max0, ge, gt, mono,
                   description, arc_pos, check_mono, more)
    => zero
}

implicit def ab_semigroup_mult_rat: ab_semigroup_mult[rat] = new
  ab_semigroup_mult[rat] {
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
}

implicit def comm_semiring_rat: comm_semiring[rat] = new comm_semiring[rat] {
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
}

implicit def comm_semiring_0_rat: comm_semiring_0[rat] = new
  comm_semiring_0[rat] {
  val `Ceta.zero` = zero_rata
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
}

implicit def comm_monoid_mult_rat: comm_monoid_mult[rat] = new
  comm_monoid_mult[rat] {
  val `Ceta.one` = one_rata
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
}

implicit def comm_semiring_1_rat: comm_semiring_1[rat] = new
  comm_semiring_1[rat] {
  val `Ceta.zero` = zero_rata
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
  val `Ceta.one` = one_rata
}

def arctic_coeff: xml => sum[List[Char], arctic] =
  (a: xml) =>
    xml_options[arctic](List((List('i', 'n', 't', 'e', 'g', 'e', 'r'),
                               (aa: xml) =>
                                 xml_change[BigInt,
     arctic]((ab: xml) => int(List('i', 'n', 't', 'e', 'g', 'e', 'r'), ab),
              (ab: BigInt) => Num_arc(ab), aa)),
                              (List('m', 'i', 'n', 'u', 's', 'I', 'n', 'f', 'i',
                                     'n', 'i', 't', 'y'),
                                (aa: xml) =>
                                  xml0elements[arctic](List('m', 'i', 'n', 'u',
                     's', 'I', 'n', 'f', 'i', 'n', 'i', 't', 'y'),
                MinInfty, aa))),
                         a)

def timesa[A, B](x0: poly_c_ext[A, B]): A => A => A = x0 match {
  case Poly_c_ext(wf, zero, one, default, plus, times, max0, ge, gt, mono,
                   description, arc_pos, check_mono, more)
    => times
}

def arc_pos[A, B](x0: poly_c_ext[A, B]): A => Boolean = x0 match {
  case Poly_c_ext(wf, zero, one, default, plus, times, max0, ge, gt, mono,
                   description, arc_pos, check_mono, more)
    => arc_pos
}

def default[A, B](x0: poly_c_ext[A, B]): A = x0 match {
  case Poly_c_ext(wf, zero, one, default, plus, times, max0, ge, gt, mono,
                   description, arc_pos, check_mono, more)
    => default
}

def arctic_rat_coeff: xml => sum[List[Char], arctic_rat] =
  (a: xml) =>
    xml_options[arctic_rat](List((List('i', 'n', 't', 'e', 'g', 'e', 'r'),
                                   (aa: xml) =>
                                     xml_change[rat,
         arctic_rat](rat, (ab: rat) => Num_arc_rat(ab), aa)),
                                  (List('r', 'a', 't', 'i', 'o', 'n', 'a', 'l'),
                                    (aa: xml) =>
                                      xml_change[rat,
          arctic_rat](rat, (ab: rat) => Num_arc_rat(ab), aa)),
                                  (List('m', 'i', 'n', 'u', 's', 'I', 'n', 'f',
 'i', 'n', 'i', 't', 'y'),
                                    (aa: xml) =>
                                      xml0elements[arctic_rat](List('m', 'i',
                             'n', 'u', 's', 'I', 'n', 'f', 'i', 'n', 'i', 't',
                             'y'),
                        MinInfty_rat, aa))),
                             a)

def one_arctic: arctic = Num_arc(BigInt(0))

def mat_mono[A](sd: nat, m: A => Boolean): (List[List[A]]) => Boolean =
  (a: List[List[A]]) => mat_monoI[A](m, sd, a)

def check_mono[A, B](x0: poly_c_ext[A, B]): A => Boolean = x0 match {
  case Poly_c_ext(wf, zero, one, default, plus, times, max0, ge, gt, mono,
                   description, arc_pos, check_mono, more)
    => check_mono
}

def equal_arctica(x0: arctic, x1: arctic): Boolean = (x0, x1) match {
  case (Num_arc(int), MinInfty) => false
  case (MinInfty, Num_arc(int)) => false
  case (Num_arc(inta), Num_arc(int)) => inta == int
  case (MinInfty, MinInfty) => true
}

implicit def equal_arctic: equal[arctic] = new equal[arctic] {
  val `Ceta.equal` = (a: arctic, b: arctic) => equal_arctica(a, b)
}

def plus_arctic(x0: arctic, y: arctic): arctic = (x0, y) match {
  case (MinInfty, y) => y
  case (Num_arc(v), MinInfty) => Num_arc(v)
  case (Num_arc(x), Num_arc(y)) => Num_arc(max[BigInt].apply(x).apply(y))
}

def zero_arctic: arctic = MinInfty

def description[A, B](x0: poly_c_ext[A, B]): List[Char] = x0 match {
  case Poly_c_ext(wf, zero, one, default, plus, times, max0, ge, gt, mono,
                   description, arc_pos, check_mono, more)
    => description
}

def times_arctic(x0: arctic, y: arctic): arctic = (x0, y) match {
  case (MinInfty, y) => MinInfty
  case (Num_arc(v), MinInfty) => MinInfty
  case (Num_arc(x), Num_arc(y)) => Num_arc(x + y)
}

def rat_domain: xml => sum[List[Char], rat] =
  (a: xml) =>
    xml1element[rat, rat](List('r', 'a', 't', 'i', 'o', 'n', 'a', 'l', 's'),
                           (aa: xml) =>
                             xml1element[rat,
  rat](List('d', 'e', 'l', 't', 'a'), rat, id[rat], aa),
                           id[rat], a)

def basic_domain: xml => sum[List[Char], nat => domain] =
  (a: xml) =>
    xml_options[nat =>
                  domain](List((List('n', 'a', 't', 'u', 'r', 'a', 'l', 's'),
                                 (aa: xml) =>
                                   xml0elements[nat =>
          domain](List('n', 'a', 't', 'u', 'r', 'a', 'l', 's'),
                   (ab: nat) => Naturala(ab), aa)),
                                (List('a', 'r', 'c', 't', 'i', 'c'),
                                  (aa: xml) =>
                                    xml1element[nat => domain,
         nat =>
           domain](List('a', 'r', 'c', 't', 'i', 'c'),
                    (ab: xml) =>
                      xml1element[nat => domain,
                                   nat =>
                                     domain](List('d', 'o', 'm', 'a', 'i', 'n'),
      (ac: xml) =>
        xml_options[nat =>
                      domain](List((List('n', 'a', 't', 'u', 'r', 'a', 'l',
  's'),
                                     (ad: xml) =>
                                       xml0elements[nat =>
              domain](List('n', 'a', 't', 'u', 'r', 'a', 'l', 's'),
                       (_: nat) => Arctic, ad)),
                                    (List('i', 'n', 't', 'e', 'g', 'e', 'r',
   's'),
                                      (ad: xml) =>
xml0elements[nat =>
               domain](List('i', 'n', 't', 'e', 'g', 'e', 'r', 's'),
                        (_: nat) => Arctic, ad)),
                                    (List('r', 'a', 't', 'i', 'o', 'n', 'a',
   'l', 's'),
                                      (ad: xml) =>
xml_change[rat, nat =>
                  domain](rat_domain, (_: rat) => (_: nat) => Arctic_rat, ad))),
                               ac),
      id[nat => domain], ab),
                    id[nat => domain], aa)),
                                (List('r', 'a', 't', 'i', 'o', 'n', 'a', 'l',
                                       's'),
                                  (aa: xml) =>
                                    xml_change[rat,
        nat =>
          domain](rat_domain, (ab: rat) => (b: nat) => Rational(ab, b), aa))),
                           a)

def interpretation_type: xml => sum[List[Char], domain] =
  (a: xml) =>
    xml1element[domain,
                 domain](List('t', 'y', 'p', 'e'),
                          (aa: xml) =>
                            xml_options[domain](List((List('p', 'o', 'l', 'y',
                    'n', 'o', 'm', 'i', 'a', 'l'),
               (ab: xml) =>
                 xml2elements[nat => domain, nat,
                               domain](List('p', 'o', 'l', 'y', 'n', 'o', 'm',
     'i', 'a', 'l'),
(ac: xml) =>
  xml1element[nat => domain,
               nat =>
                 domain](List('d', 'o', 'm', 'a', 'i', 'n'),
                          (ad: xml) =>
                            xml_options[nat =>
  domain](List((List('n', 'a', 't', 'u', 'r', 'a', 'l', 's'), basic_domain),
                (List('a', 'r', 'c', 't', 'i', 'c'), basic_domain),
                (List('r', 'a', 't', 'i', 'o', 'n', 'a', 'l', 's'),
                  basic_domain),
                (List('m', 'a', 't', 'r', 'i', 'c', 'e', 's'),
                  (ae: xml) =>
                    xml3elements[nat, nat, nat => domain,
                                  nat =>
                                    domain](List('m', 'a', 't', 'r', 'i', 'c',
          'e', 's'),
     (af: xml) => nata(List('d', 'i', 'm', 'e', 'n', 's', 'i', 'o', 'n'), af),
     (af: xml) =>
       nata(List('s', 't', 'r', 'i', 'c', 't', 'D', 'i', 'm', 'e', 'n', 's',
                  'i', 'o', 'n'),
             af),
     (af: xml) =>
       xml1element[nat => domain,
                    nat =>
                      domain](List('d', 'o', 'm', 'a', 'i', 'n'), basic_domain,
                               id[nat => domain], af),
     (di: nat) =>
       (sd: nat) =>
         (domain: nat => domain) =>
           (d: nat) =>
             domain(d) match {
               case Naturala(_) => Int_mat(di, sd)
               case Arctic => Arc_mat(di)
               case Arctic_rat => Arc_rat_mat(di)
               case Rational(_, _) => Rat_mat(di, sd)
             },
     ae))),
           ad),
                          id[nat => domain], ac),
(ac: xml) => nata(List('d', 'e', 'g', 'r', 'e', 'e'), ac),
(typea: nat => domain) => typea, ab)),
              (List('m', 'a', 't', 'r', 'i', 'x', 'I', 'n', 't', 'e', 'r', 'p',
                     'r', 'e', 't', 'a', 't', 'i', 'o', 'n'),
                (ab: xml) =>
                  xml3elements[nat => domain, nat, nat,
                                domain](List('m', 'a', 't', 'r', 'i', 'x', 'I',
      'n', 't', 'e', 'r', 'p', 'r', 'e', 't', 'a', 't', 'i', 'o', 'n'),
 (ac: xml) =>
   xml1element[nat => domain,
                nat =>
                  domain](List('d', 'o', 'm', 'a', 'i', 'n'), basic_domain,
                           id[nat => domain], ac),
 (ac: xml) => nata(List('d', 'i', 'm', 'e', 'n', 's', 'i', 'o', 'n'), ac),
 (ac: xml) =>
   nata(List('s', 't', 'r', 'i', 'c', 't', 'D', 'i', 'm', 'e', 'n', 's', 'i',
              'o', 'n'),
         ac),
 (domain: nat => domain) =>
   (di: nat) =>
     (sd: nat) =>
       domain(Zero_nat) match {
         case Naturala(_) => Int_mat(di, sd)
         case Arctic => Arc_mat(di)
         case Arctic_rat => Arc_rat_mat(di)
         case Rational(_, _) => Rat_mat(di, sd)
       },
 ab))),
         aa),
                          id[domain], a)

def one_arctic_rat: arctic_rat = Num_arc_rat(fract(BigInt(0), BigInt(1)))

def equal_arctic_rata(x0: arctic_rat, x1: arctic_rat): Boolean = (x0, x1) match
  {
  case (Num_arc_rat(rat), MinInfty_rat) => false
  case (MinInfty_rat, Num_arc_rat(rat)) => false
  case (Num_arc_rat(rata), Num_arc_rat(rat)) => equal_rata(rata, rat)
  case (MinInfty_rat, MinInfty_rat) => true
}

implicit def equal_arctic_rat: equal[arctic_rat] = new equal[arctic_rat] {
  val `Ceta.equal` = (a: arctic_rat, b: arctic_rat) => equal_arctic_rata(a, b)
}

def plus_arctic_rat(x0: arctic_rat, y: arctic_rat): arctic_rat = (x0, y) match {
  case (MinInfty_rat, y) => y
  case (Num_arc_rat(v), MinInfty_rat) => Num_arc_rat(v)
  case (Num_arc_rat(x), Num_arc_rat(y)) =>
    Num_arc_rat(max[rat].apply(x).apply(y))
}

def zero_arctic_rat: arctic_rat = MinInfty_rat

def times_arctic_rat(x0: arctic_rat, y: arctic_rat): arctic_rat = (x0, y) match
  {
  case (MinInfty_rat, y) => MinInfty_rat
  case (Num_arc_rat(v), MinInfty_rat) => MinInfty_rat
  case (Num_arc_rat(x), Num_arc_rat(y)) => Num_arc_rat(plus_rata(x, y))
}

def int_mono(x: BigInt): Boolean = BigInt(1) <= x

def max0_int(x: BigInt): BigInt = max[BigInt].apply(BigInt(0)).apply(x)

def int_name: List[Char] =
  List('n', 'a', 't', 'u', 'r', 'a', 'l', 's', ' ', 'w', 'i', 't', 'h', ' ',
        'n', 'e', 'g', 'a', 't', 'i', 'v', 'e', ' ', 'c', 'o', 'n', 's', 't',
        'a', 'n', 't', 's')

def int_components: poly_c_ext[BigInt, Unit] =
  Poly_c_ext[BigInt,
              Unit]((_: BigInt) => true, BigInt(0), BigInt(1), BigInt(1),
                     (a: BigInt) => (b: BigInt) => a + b,
                     (a: BigInt) => (b: BigInt) => a * b,
                     (a: BigInt) => max0_int(a),
                     (x: BigInt) => (y: BigInt) => y <= x,
                     (x: BigInt) => (y: BigInt) => y < x, true, int_name,
                     (_: BigInt) => true, (a: BigInt) => int_mono(a), ())

def max0_rat(x: rat): rat = max[rat].apply(fract(BigInt(0), BigInt(1))).apply(x)

def rat_mono(x: rat): Boolean = less_eq_rat(fract(BigInt(1), BigInt(1)), x)

def rat_name: List[Char] =
  List('r', 'a', 't', 'i', 'o', 'n', 'a', 'l', 's', ' ', '(', 'w', 'i', 't',
        'h', ' ', 'n', 'e', 'g', 'a', 't', 'i', 'v', 'e', ' ', 'c', 'o', 'n',
        's', 't', 'a', 'n', 't', 's', ')')

def rat_components: poly_c_ext[rat, Unit] =
  Poly_c_ext[rat, Unit]((_: rat) => true, fract(BigInt(0), BigInt(1)),
                         fract(BigInt(1), BigInt(1)),
                         fract(BigInt(1), BigInt(1)),
                         (a: rat) => (b: rat) => plus_rata(a, b),
                         (a: rat) => (b: rat) => times_rata(a, b),
                         (a: rat) => max0_rat(a),
                         (x: rat) => (y: rat) => less_eq_rat(y, x),
                         (x: rat) => (y: rat) => less_rat(y, x), true, rat_name,
                         (_: rat) => true, (a: rat) => rat_mono(a), ())

def ge_int(x: BigInt, y: BigInt): Boolean = y <= x

implicit def non_strict_order_int: non_strict_order[BigInt] = new
  non_strict_order[BigInt] {
  val `Ceta.ge` = (a: BigInt, b: BigInt) => ge_int(a, b)
}

def ge_rat(x: rat, y: rat): Boolean = less_eq_rat(y, x)

implicit def non_strict_order_rat: non_strict_order[rat] = new
  non_strict_order[rat] {
  val `Ceta.ge` = (a: rat, b: rat) => ge_rat(a, b)
}

def ge_arctic(x: arctic, xa1: arctic): Boolean = (x, xa1) match {
  case (x, MinInfty) => true
  case (MinInfty, Num_arc(uu)) => false
  case (Num_arc(x), Num_arc(y)) => y <= x
}

def gt_arctic(uu: arctic, x1: arctic): Boolean = (uu, x1) match {
  case (uu, MinInfty) => true
  case (MinInfty, Num_arc(uv)) => false
  case (Num_arc(x), Num_arc(y)) => y < x
}

def pos_arctic(x0: arctic): Boolean = x0 match {
  case MinInfty => false
  case Num_arc(n) => BigInt(0) <= n
}

def arctic_name: List[Char] =
  List('a', 'r', 'c', 't', 'i', 'c', ' ', 's', 'e', 'm', 'i', 'r', 'i', 'n',
        'g')

def arctic_components: poly_c_ext[arctic, Unit] =
  Poly_c_ext[arctic,
              Unit]((_: arctic) => true, zero_arctic, one_arctic, one_arctic,
                     (a: arctic) => (b: arctic) => plus_arctic(a, b),
                     (a: arctic) => (b: arctic) => times_arctic(a, b),
                     id[arctic], (a: arctic) => (b: arctic) => ge_arctic(a, b),
                     (a: arctic) => (b: arctic) => gt_arctic(a, b), false,
                     arctic_name, (a: arctic) => pos_arctic(a),
                     (_: arctic) => false, ())

implicit def non_strict_order_arctic: non_strict_order[arctic] = new
  non_strict_order[arctic] {
  val `Ceta.ge` = (a: arctic, b: arctic) => ge_arctic(a, b)
}

def ge_arctic_rat(x: arctic_rat, xa1: arctic_rat): Boolean = (x, xa1) match {
  case (x, MinInfty_rat) => true
  case (MinInfty_rat, Num_arc_rat(uu)) => false
  case (Num_arc_rat(x), Num_arc_rat(y)) => less_eq_rat(y, x)
}

def pos_arctic_rat(x0: arctic_rat): Boolean = x0 match {
  case MinInfty_rat => false
  case Num_arc_rat(n) => less_eq_rat(fract(BigInt(0), BigInt(1)), n)
}

def arctic_rat_name: List[Char] =
  List('a', 'r', 'c', 't', 'i', 'c', ' ', 'r', 'a', 't', 'i', 'o', 'n', 'a',
        'l', 's')

def weak_gt_arctic_rat(uu: arctic_rat, x1: arctic_rat): Boolean = (uu, x1) match
  {
  case (uu, MinInfty_rat) => true
  case (MinInfty_rat, Num_arc_rat(uv)) => false
  case (Num_arc_rat(x), Num_arc_rat(y)) => less_rat(y, x)
}

def arctic_rat_components: poly_c_ext[arctic_rat, Unit] =
  Poly_c_ext[arctic_rat,
              Unit]((_: arctic_rat) => true, zero_arctic_rat, one_arctic_rat,
                     one_arctic_rat,
                     (a: arctic_rat) =>
                       (b: arctic_rat) => plus_arctic_rat(a, b),
                     (a: arctic_rat) =>
                       (b: arctic_rat) => times_arctic_rat(a, b),
                     id[arctic_rat],
                     (a: arctic_rat) => (b: arctic_rat) => ge_arctic_rat(a, b),
                     (a: arctic_rat) =>
                       (b: arctic_rat) => weak_gt_arctic_rat(a, b),
                     false, arctic_rat_name,
                     (a: arctic_rat) => pos_arctic_rat(a),
                     (_: arctic_rat) => false, ())

implicit def ordered_ab_semigroup_int: ordered_ab_semigroup[BigInt] = new
  ordered_ab_semigroup[BigInt] {
  val `Ceta.ge` = (a: BigInt, b: BigInt) => ge_int(a, b)
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
}

implicit def max_ordered_monoid_add_int: max_ordered_monoid_add[BigInt] = new
  max_ordered_monoid_add[BigInt] {
  val `Ceta.max0` = (a: BigInt) => max0_int(a)
  val `Ceta.ge` = (a: BigInt, b: BigInt) => ge_int(a, b)
  val `Ceta.zero` = zero_inta
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
}

implicit def ordered_ab_semigroup_rat: ordered_ab_semigroup[rat] = new
  ordered_ab_semigroup[rat] {
  val `Ceta.ge` = (a: rat, b: rat) => ge_rat(a, b)
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
}

implicit def max_ordered_monoid_add_rat: max_ordered_monoid_add[rat] = new
  max_ordered_monoid_add[rat] {
  val `Ceta.max0` = (a: rat) => max0_rat(a)
  val `Ceta.ge` = (a: rat, b: rat) => ge_rat(a, b)
  val `Ceta.zero` = zero_rata
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
}

implicit def non_strict_order_arctic_rat: non_strict_order[arctic_rat] = new
  non_strict_order[arctic_rat] {
  val `Ceta.ge` = (a: arctic_rat, b: arctic_rat) => ge_arctic_rat(a, b)
}

def interpretation[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], redtriple_impl[A]] =
  (a: xml) =>
    xml1many_elements_gen[domain,
                           alter[((A, nat), (BigInt, List[BigInt])),
                                  ((A, nat), (rat, List[rat])),
                                  ((A, nat), (arctic, List[arctic])),
                                  ((A, nat),
                                    (List[List[BigInt]],
                                      List[List[List[BigInt]]])),
                                  ((A, nat),
                                    (List[List[rat]], List[List[List[rat]]])),
                                  ((A, nat),
                                    (List[List[arctic]],
                                      List[List[List[arctic]]])),
                                  ((A, nat), (arctic_rat, List[arctic_rat])),
                                  ((A, nat),
                                    (List[List[arctic_rat]],
                                      List[List[List[arctic_rat]]])),
                                  ((A, nat), List[(List[(nat, nat)], BigInt)]),
                                  ((A, nat), List[(List[(nat, nat)], rat)])],
                           redtriple_impl[A]](List('i', 'n', 't', 'e', 'r', 'p',
            'r', 'e', 't', 'a', 't', 'i', 'o', 'n'),
       interpretation_type,
       (aa: domain) =>
         aa match {
           case Naturala(deg) =>
             (if (less_eq_nat(deg, one_nat))
               (ab: xml) =>
                 xml3elements[A, nat, (BigInt, List[BigInt]),
                               alter[((A, nat), (BigInt, List[BigInt])),
                                      ((A, nat), (rat, List[rat])),
                                      ((A, nat), (arctic, List[arctic])),
                                      ((A, nat),
(List[List[BigInt]], List[List[List[BigInt]]])),
                                      ((A, nat),
(List[List[rat]], List[List[List[rat]]])),
                                      ((A, nat),
(List[List[arctic]], List[List[List[arctic]]])),
                                      ((A, nat),
(arctic_rat, List[arctic_rat])),
                                      ((A, nat),
(List[List[arctic_rat]], List[List[List[arctic_rat]]])),
                                      ((A, nat),
List[(List[(nat, nat)], BigInt)]),
                                      ((A, nat),
List[(List[(nat, nat)],
       rat)])]](List('i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'), xml2name,
                 (ac: xml) => nata(List('a', 'r', 'i', 't', 'y'), ac),
                 (x: xml) =>
                   bindb[List[Char], tpoly[nat, BigInt],
                          (BigInt,
                            List[BigInt])](polynomial[BigInt](int_coeff, x),
    (ac: tpoly[nat, BigInt]) =>
      lpoly_of[BigInt](zeroa[BigInt, Unit](int_components),
                        onea[BigInt, Unit](int_components),
                        plusa[BigInt, Unit](int_components),
                        timesa[BigInt, Unit](int_components), ac)),
                 (f: A) =>
                   (ac: nat) =>
                     (poly: (BigInt, List[BigInt])) =>
                       Alter1[((A, nat), (BigInt, List[BigInt])),
                               ((A, nat), (rat, List[rat])),
                               ((A, nat), (arctic, List[arctic])),
                               ((A, nat),
                                 (List[List[BigInt]],
                                   List[List[List[BigInt]]])),
                               ((A, nat),
                                 (List[List[rat]], List[List[List[rat]]])),
                               ((A, nat),
                                 (List[List[arctic]],
                                   List[List[List[arctic]]])),
                               ((A, nat), (arctic_rat, List[arctic_rat])),
                               ((A, nat),
                                 (List[List[arctic_rat]],
                                   List[List[List[arctic_rat]]])),
                               ((A, nat), List[(List[(nat, nat)], BigInt)]),
                               ((A, nat),
                                 List[(List[(nat, nat)],
rat)])](((f, ac),
          (fst[BigInt, List[BigInt]](poly),
            fit_length[BigInt](zeroa[BigInt, Unit](int_components), ac,
                                snd[BigInt, List[BigInt]](poly))))),
                 ab)
               else (ab: xml) =>
                      xml3elements[A, nat, List[(List[(nat, nat)], BigInt)],
                                    alter[((A, nat), (BigInt, List[BigInt])),
   ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
   ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
   ((A, nat), (List[List[rat]], List[List[List[rat]]])),
   ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
   ((A, nat), (arctic_rat, List[arctic_rat])),
   ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
   ((A, nat), List[(List[(nat, nat)], BigInt)]),
   ((A, nat),
     List[(List[(nat, nat)],
            rat)])]](List('i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'),
                      xml2name,
                      (ac: xml) => nata(List('a', 'r', 'i', 't', 'y'), ac),
                      (x: xml) =>
                        bindb[List[Char], tpoly[nat, BigInt],
                               List[(List[(nat, nat)],
                                      BigInt)]](polynomial[BigInt](int_coeff,
                            x),
         (p: tpoly[nat, BigInt]) =>
           Inr[List[(List[(nat, nat)], BigInt)],
                List[Char]](poly_of[nat, BigInt](p))),
                      (f: A) =>
                        (ac: nat) =>
                          (poly: List[(List[(nat, nat)], BigInt)]) =>
                            Alter9[((A, nat), List[(List[(nat, nat)], BigInt)]),
                                    ((A, nat), (BigInt, List[BigInt])),
                                    ((A, nat), (rat, List[rat])),
                                    ((A, nat), (arctic, List[arctic])),
                                    ((A, nat),
                                      (List[List[BigInt]],
List[List[List[BigInt]]])),
                                    ((A, nat),
                                      (List[List[rat]], List[List[List[rat]]])),
                                    ((A, nat),
                                      (List[List[arctic]],
List[List[List[arctic]]])),
                                    ((A, nat), (arctic_rat, List[arctic_rat])),
                                    ((A, nat),
                                      (List[List[arctic_rat]],
List[List[List[arctic_rat]]])),
                                    ((A, nat),
                                      List[(List[(nat, nat)],
     rat)])](((f, ac), poly)),
                      ab))
           case Arctic =>
             (ab: xml) =>
               xml3elements[A, nat, (arctic, List[arctic]),
                             alter[((A, nat), (BigInt, List[BigInt])),
                                    ((A, nat), (rat, List[rat])),
                                    ((A, nat), (arctic, List[arctic])),
                                    ((A, nat),
                                      (List[List[BigInt]],
List[List[List[BigInt]]])),
                                    ((A, nat),
                                      (List[List[rat]], List[List[List[rat]]])),
                                    ((A, nat),
                                      (List[List[arctic]],
List[List[List[arctic]]])),
                                    ((A, nat), (arctic_rat, List[arctic_rat])),
                                    ((A, nat),
                                      (List[List[arctic_rat]],
List[List[List[arctic_rat]]])),
                                    ((A, nat),
                                      List[(List[(nat, nat)], BigInt)]),
                                    ((A, nat),
                                      List[(List[(nat, nat)],
     rat)])]](List('i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'), xml2name,
               (ac: xml) => nata(List('a', 'r', 'i', 't', 'y'), ac),
               (x: xml) =>
                 bindb[List[Char], tpoly[nat, arctic],
                        (arctic,
                          List[arctic])](polynomial[arctic](arctic_coeff, x),
  (ac: tpoly[nat, arctic]) =>
    lpoly_of[arctic](zeroa[arctic, Unit](arctic_components),
                      onea[arctic, Unit](arctic_components),
                      plusa[arctic, Unit](arctic_components),
                      timesa[arctic, Unit](arctic_components), ac)),
               (f: A) =>
                 (ac: nat) =>
                   (poly: (arctic, List[arctic])) =>
                     Alter3[((A, nat), (arctic, List[arctic])),
                             ((A, nat), (BigInt, List[BigInt])),
                             ((A, nat), (rat, List[rat])),
                             ((A, nat),
                               (List[List[BigInt]], List[List[List[BigInt]]])),
                             ((A, nat),
                               (List[List[rat]], List[List[List[rat]]])),
                             ((A, nat),
                               (List[List[arctic]], List[List[List[arctic]]])),
                             ((A, nat), (arctic_rat, List[arctic_rat])),
                             ((A, nat),
                               (List[List[arctic_rat]],
                                 List[List[List[arctic_rat]]])),
                             ((A, nat), List[(List[(nat, nat)], BigInt)]),
                             ((A, nat),
                               List[(List[(nat, nat)],
                                      rat)])](((f, ac),
        (fst[arctic, List[arctic]](poly),
          fit_length[arctic](zeroa[arctic, Unit](arctic_components), ac,
                              snd[arctic, List[arctic]](poly))))),
               ab)
           case Arctic_rat =>
             (ab: xml) =>
               xml3elements[A, nat, (arctic_rat, List[arctic_rat]),
                             alter[((A, nat), (BigInt, List[BigInt])),
                                    ((A, nat), (rat, List[rat])),
                                    ((A, nat), (arctic, List[arctic])),
                                    ((A, nat),
                                      (List[List[BigInt]],
List[List[List[BigInt]]])),
                                    ((A, nat),
                                      (List[List[rat]], List[List[List[rat]]])),
                                    ((A, nat),
                                      (List[List[arctic]],
List[List[List[arctic]]])),
                                    ((A, nat), (arctic_rat, List[arctic_rat])),
                                    ((A, nat),
                                      (List[List[arctic_rat]],
List[List[List[arctic_rat]]])),
                                    ((A, nat),
                                      List[(List[(nat, nat)], BigInt)]),
                                    ((A, nat),
                                      List[(List[(nat, nat)],
     rat)])]](List('i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'), xml2name,
               (ac: xml) => nata(List('a', 'r', 'i', 't', 'y'), ac),
               (x: xml) =>
                 bindb[List[Char], tpoly[nat, arctic_rat],
                        (arctic_rat,
                          List[arctic_rat])](polynomial[arctic_rat](arctic_rat_coeff,
                             x),
      (ac: tpoly[nat, arctic_rat]) =>
        lpoly_of[arctic_rat](zeroa[arctic_rat, Unit](arctic_rat_components),
                              onea[arctic_rat, Unit](arctic_rat_components),
                              plusa[arctic_rat, Unit](arctic_rat_components),
                              timesa[arctic_rat, Unit](arctic_rat_components),
                              ac)),
               (f: A) =>
                 (ac: nat) =>
                   (poly: (arctic_rat, List[arctic_rat])) =>
                     Alter7[((A, nat), (arctic_rat, List[arctic_rat])),
                             ((A, nat), (BigInt, List[BigInt])),
                             ((A, nat), (rat, List[rat])),
                             ((A, nat), (arctic, List[arctic])),
                             ((A, nat),
                               (List[List[BigInt]], List[List[List[BigInt]]])),
                             ((A, nat),
                               (List[List[rat]], List[List[List[rat]]])),
                             ((A, nat),
                               (List[List[arctic]], List[List[List[arctic]]])),
                             ((A, nat),
                               (List[List[arctic_rat]],
                                 List[List[List[arctic_rat]]])),
                             ((A, nat), List[(List[(nat, nat)], BigInt)]),
                             ((A, nat),
                               List[(List[(nat, nat)],
                                      rat)])](((f, ac),
        (fst[arctic_rat, List[arctic_rat]](poly),
          fit_length[arctic_rat](zeroa[arctic_rat, Unit](arctic_rat_components),
                                  ac, snd[arctic_rat,
   List[arctic_rat]](poly))))),
               ab)
           case Int_mat(n, sd) =>
             (ab: xml) =>
               xml3elements[A, nat,
                             (List[List[BigInt]], List[List[List[BigInt]]]),
                             alter[((A, nat), (BigInt, List[BigInt])),
                                    ((A, nat), (rat, List[rat])),
                                    ((A, nat), (arctic, List[arctic])),
                                    ((A, nat),
                                      (List[List[BigInt]],
List[List[List[BigInt]]])),
                                    ((A, nat),
                                      (List[List[rat]], List[List[List[rat]]])),
                                    ((A, nat),
                                      (List[List[arctic]],
List[List[List[arctic]]])),
                                    ((A, nat), (arctic_rat, List[arctic_rat])),
                                    ((A, nat),
                                      (List[List[arctic_rat]],
List[List[List[arctic_rat]]])),
                                    ((A, nat),
                                      List[(List[(nat, nat)], BigInt)]),
                                    ((A, nat),
                                      List[(List[(nat, nat)],
     rat)])]](List('i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'), xml2name,
               (ac: xml) => nata(List('a', 'r', 'i', 't', 'y'), ac),
               (x: xml) =>
                 bindb[List[Char], tpoly[nat, List[List[BigInt]]],
                        (List[List[BigInt]],
                          List[List[List[BigInt]]])](polynomial[List[List[BigInt]]](mat_coeff[BigInt](BigInt(0),
                       int_coeff),
     x),
              (ac: tpoly[nat, List[List[BigInt]]]) =>
                lpoly_of[List[List[BigInt]]](zeroa[List[List[BigInt]],
            Unit](Poly_c_ext[List[List[BigInt]],
                              Unit]((ad: List[List[BigInt]]) =>
                                      mat[BigInt](n, n, ad),
                                     mat0I[BigInt](BigInt(0), n, n),
                                     mat1I[BigInt](BigInt(0), BigInt(1), n),
                                     mat1I[BigInt](BigInt(0),
            default[BigInt, Unit](int_components), n),
                                     (ad: List[List[BigInt]]) =>
                                       (b: List[List[BigInt]]) =>
 mat_plusI[BigInt]((ae: BigInt) => (ba: BigInt) => ae + ba, ad, b),
                                     (ad: List[List[BigInt]]) =>
                                       (b: List[List[BigInt]]) =>
 mat_multI[BigInt](BigInt(0), (ae: BigInt) => (ba: BigInt) => ae + ba,
                    (ae: BigInt) => (ba: BigInt) => ae * ba, n, ad, b),
                                     mat_max0[BigInt], mat_ge[BigInt],
                                     (ad: List[List[BigInt]]) =>
                                       (b: List[List[BigInt]]) =>
 mat_gtI[BigInt](gt[BigInt, Unit](int_components), sd, ad, b),
                                     monoa[BigInt, Unit](int_components),
                                     concat[Char](List((shows_prec_nat(Zero_nat,
                                n)).apply(Nil),
                List('x'), (shows_prec_nat(Zero_nat, n)).apply(Nil),
                List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'w', 'i',
                      't', 'h', ' ', 's', 't', 'r', 'i', 'c', 't', ' ', 'd',
                      'i', 'm', 'e', 'n', 's', 'i', 'o', 'n', ' '),
                (shows_prec_nat(Zero_nat, sd)).apply(Nil),
                List(' ', 'o', 'v', 'e', 'r', ' '),
                description[BigInt, Unit](int_components))),
                                     (_: List[List[BigInt]]) => true,
                                     mat_mono[BigInt](sd,
               check_mono[BigInt, Unit](int_components)),
                                     ())),
      onea[List[List[BigInt]],
            Unit](Poly_c_ext[List[List[BigInt]],
                              Unit]((ad: List[List[BigInt]]) =>
                                      mat[BigInt](n, n, ad),
                                     mat0I[BigInt](BigInt(0), n, n),
                                     mat1I[BigInt](BigInt(0), BigInt(1), n),
                                     mat1I[BigInt](BigInt(0),
            default[BigInt, Unit](int_components), n),
                                     (ad: List[List[BigInt]]) =>
                                       (b: List[List[BigInt]]) =>
 mat_plusI[BigInt]((ae: BigInt) => (ba: BigInt) => ae + ba, ad, b),
                                     (ad: List[List[BigInt]]) =>
                                       (b: List[List[BigInt]]) =>
 mat_multI[BigInt](BigInt(0), (ae: BigInt) => (ba: BigInt) => ae + ba,
                    (ae: BigInt) => (ba: BigInt) => ae * ba, n, ad, b),
                                     mat_max0[BigInt], mat_ge[BigInt],
                                     (ad: List[List[BigInt]]) =>
                                       (b: List[List[BigInt]]) =>
 mat_gtI[BigInt](gt[BigInt, Unit](int_components), sd, ad, b),
                                     monoa[BigInt, Unit](int_components),
                                     concat[Char](List((shows_prec_nat(Zero_nat,
                                n)).apply(Nil),
                List('x'), (shows_prec_nat(Zero_nat, n)).apply(Nil),
                List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'w', 'i',
                      't', 'h', ' ', 's', 't', 'r', 'i', 'c', 't', ' ', 'd',
                      'i', 'm', 'e', 'n', 's', 'i', 'o', 'n', ' '),
                (shows_prec_nat(Zero_nat, sd)).apply(Nil),
                List(' ', 'o', 'v', 'e', 'r', ' '),
                description[BigInt, Unit](int_components))),
                                     (_: List[List[BigInt]]) => true,
                                     mat_mono[BigInt](sd,
               check_mono[BigInt, Unit](int_components)),
                                     ())),
      plusa[List[List[BigInt]],
             Unit](Poly_c_ext[List[List[BigInt]],
                               Unit]((ad: List[List[BigInt]]) =>
                                       mat[BigInt](n, n, ad),
                                      mat0I[BigInt](BigInt(0), n, n),
                                      mat1I[BigInt](BigInt(0), BigInt(1), n),
                                      mat1I[BigInt](BigInt(0),
             default[BigInt, Unit](int_components), n),
                                      (ad: List[List[BigInt]]) =>
(b: List[List[BigInt]]) =>
  mat_plusI[BigInt]((ae: BigInt) => (ba: BigInt) => ae + ba, ad, b),
                                      (ad: List[List[BigInt]]) =>
(b: List[List[BigInt]]) =>
  mat_multI[BigInt](BigInt(0), (ae: BigInt) => (ba: BigInt) => ae + ba,
                     (ae: BigInt) => (ba: BigInt) => ae * ba, n, ad, b),
                                      mat_max0[BigInt], mat_ge[BigInt],
                                      (ad: List[List[BigInt]]) =>
(b: List[List[BigInt]]) =>
  mat_gtI[BigInt](gt[BigInt, Unit](int_components), sd, ad, b),
                                      monoa[BigInt, Unit](int_components),
                                      concat[Char](List((shows_prec_nat(Zero_nat,
                                 n)).apply(Nil),
                 List('x'), (shows_prec_nat(Zero_nat, n)).apply(Nil),
                 List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'w',
                       'i', 't', 'h', ' ', 's', 't', 'r', 'i', 'c', 't', ' ',
                       'd', 'i', 'm', 'e', 'n', 's', 'i', 'o', 'n', ' '),
                 (shows_prec_nat(Zero_nat, sd)).apply(Nil),
                 List(' ', 'o', 'v', 'e', 'r', ' '),
                 description[BigInt, Unit](int_components))),
                                      (_: List[List[BigInt]]) => true,
                                      mat_mono[BigInt](sd,
                check_mono[BigInt, Unit](int_components)),
                                      ())),
      timesa[List[List[BigInt]],
              Unit](Poly_c_ext[List[List[BigInt]],
                                Unit]((ad: List[List[BigInt]]) =>
mat[BigInt](n, n, ad),
                                       mat0I[BigInt](BigInt(0), n, n),
                                       mat1I[BigInt](BigInt(0), BigInt(1), n),
                                       mat1I[BigInt](BigInt(0),
              default[BigInt, Unit](int_components), n),
                                       (ad: List[List[BigInt]]) =>
 (b: List[List[BigInt]]) =>
   mat_plusI[BigInt]((ae: BigInt) => (ba: BigInt) => ae + ba, ad, b),
                                       (ad: List[List[BigInt]]) =>
 (b: List[List[BigInt]]) =>
   mat_multI[BigInt](BigInt(0), (ae: BigInt) => (ba: BigInt) => ae + ba,
                      (ae: BigInt) => (ba: BigInt) => ae * ba, n, ad, b),
                                       mat_max0[BigInt], mat_ge[BigInt],
                                       (ad: List[List[BigInt]]) =>
 (b: List[List[BigInt]]) =>
   mat_gtI[BigInt](gt[BigInt, Unit](int_components), sd, ad, b),
                                       monoa[BigInt, Unit](int_components),
                                       concat[Char](List((shows_prec_nat(Zero_nat,
                                  n)).apply(Nil),
                  List('x'), (shows_prec_nat(Zero_nat, n)).apply(Nil),
                  List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'w',
                        'i', 't', 'h', ' ', 's', 't', 'r', 'i', 'c', 't', ' ',
                        'd', 'i', 'm', 'e', 'n', 's', 'i', 'o', 'n', ' '),
                  (shows_prec_nat(Zero_nat, sd)).apply(Nil),
                  List(' ', 'o', 'v', 'e', 'r', ' '),
                  description[BigInt, Unit](int_components))),
                                       (_: List[List[BigInt]]) => true,
                                       mat_mono[BigInt](sd,
                 check_mono[BigInt, Unit](int_components)),
                                       ())),
      ac)),
               (f: A) =>
                 (ac: nat) =>
                   (poly: (List[List[BigInt]], List[List[List[BigInt]]])) =>
                     Alter4[((A, nat),
                              (List[List[BigInt]], List[List[List[BigInt]]])),
                             ((A, nat), (BigInt, List[BigInt])),
                             ((A, nat), (rat, List[rat])),
                             ((A, nat), (arctic, List[arctic])),
                             ((A, nat),
                               (List[List[rat]], List[List[List[rat]]])),
                             ((A, nat),
                               (List[List[arctic]], List[List[List[arctic]]])),
                             ((A, nat), (arctic_rat, List[arctic_rat])),
                             ((A, nat),
                               (List[List[arctic_rat]],
                                 List[List[List[arctic_rat]]])),
                             ((A, nat), List[(List[(nat, nat)], BigInt)]),
                             ((A, nat),
                               List[(List[(nat, nat)],
                                      rat)])](((f, ac),
        (fst[List[List[BigInt]], List[List[List[BigInt]]]](poly),
          fit_length[List[List[BigInt]]](zeroa[List[List[BigInt]],
        Unit](Poly_c_ext[List[List[BigInt]],
                          Unit]((ad: List[List[BigInt]]) =>
                                  mat[BigInt](n, n, ad),
                                 mat0I[BigInt](BigInt(0), n, n),
                                 mat1I[BigInt](BigInt(0), BigInt(1), n),
                                 mat1I[BigInt](BigInt(0),
        default[BigInt, Unit](int_components), n),
                                 (ad: List[List[BigInt]]) =>
                                   (b: List[List[BigInt]]) =>
                                     mat_plusI[BigInt]((ae: BigInt) =>
                 (ba: BigInt) => ae + ba,
                ad, b),
                                 (ad: List[List[BigInt]]) =>
                                   (b: List[List[BigInt]]) =>
                                     mat_multI[BigInt](BigInt(0),
                (ae: BigInt) => (ba: BigInt) => ae + ba,
                (ae: BigInt) => (ba: BigInt) => ae * ba, n, ad, b),
                                 mat_max0[BigInt], mat_ge[BigInt],
                                 (ad: List[List[BigInt]]) =>
                                   (b: List[List[BigInt]]) =>
                                     mat_gtI[BigInt](gt[BigInt,
                 Unit](int_components),
              sd, ad, b),
                                 monoa[BigInt, Unit](int_components),
                                 concat[Char](List((shows_prec_nat(Zero_nat,
                            n)).apply(Nil),
            List('x'), (shows_prec_nat(Zero_nat, n)).apply(Nil),
            List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'w', 'i',
                  't', 'h', ' ', 's', 't', 'r', 'i', 'c', 't', ' ', 'd', 'i',
                  'm', 'e', 'n', 's', 'i', 'o', 'n', ' '),
            (shows_prec_nat(Zero_nat, sd)).apply(Nil),
            List(' ', 'o', 'v', 'e', 'r', ' '),
            description[BigInt, Unit](int_components))),
                                 (_: List[List[BigInt]]) => true,
                                 mat_mono[BigInt](sd,
           check_mono[BigInt, Unit](int_components)),
                                 ())),
  ac, snd[List[List[BigInt]], List[List[List[BigInt]]]](poly))))),
               ab)
           case Arc_mat(n) =>
             (ab: xml) =>
               xml3elements[A, nat,
                             (List[List[arctic]], List[List[List[arctic]]]),
                             alter[((A, nat), (BigInt, List[BigInt])),
                                    ((A, nat), (rat, List[rat])),
                                    ((A, nat), (arctic, List[arctic])),
                                    ((A, nat),
                                      (List[List[BigInt]],
List[List[List[BigInt]]])),
                                    ((A, nat),
                                      (List[List[rat]], List[List[List[rat]]])),
                                    ((A, nat),
                                      (List[List[arctic]],
List[List[List[arctic]]])),
                                    ((A, nat), (arctic_rat, List[arctic_rat])),
                                    ((A, nat),
                                      (List[List[arctic_rat]],
List[List[List[arctic_rat]]])),
                                    ((A, nat),
                                      List[(List[(nat, nat)], BigInt)]),
                                    ((A, nat),
                                      List[(List[(nat, nat)],
     rat)])]](List('i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'), xml2name,
               (ac: xml) => nata(List('a', 'r', 'i', 't', 'y'), ac),
               (x: xml) =>
                 bindb[List[Char], tpoly[nat, List[List[arctic]]],
                        (List[List[arctic]],
                          List[List[List[arctic]]])](polynomial[List[List[arctic]]](mat_coeff[arctic](MinInfty,
                       arctic_coeff),
     x),
              (ac: tpoly[nat, List[List[arctic]]]) =>
                lpoly_of[List[List[arctic]]](zeroa[List[List[arctic]],
            Unit](Poly_c_ext[List[List[arctic]],
                              Unit]((ad: List[List[arctic]]) =>
                                      mat[arctic](n, n, ad),
                                     mat0I[arctic](zero_arctic, n, n),
                                     mat1I[arctic](zero_arctic, one_arctic, n),
                                     mat1I[arctic](zero_arctic,
            default[arctic, Unit](arctic_components), n),
                                     (ad: List[List[arctic]]) =>
                                       (b: List[List[arctic]]) =>
 mat_plusI[arctic]((ae: arctic) => (ba: arctic) => plus_arctic(ae, ba), ad, b),
                                     (ad: List[List[arctic]]) =>
                                       (b: List[List[arctic]]) =>
 mat_multI[arctic](zero_arctic,
                    (ae: arctic) => (ba: arctic) => plus_arctic(ae, ba),
                    (ae: arctic) => (ba: arctic) => times_arctic(ae, ba), n, ad,
                    b),
                                     id[List[List[arctic]]], mat_ge[arctic],
                                     (ad: List[List[arctic]]) =>
                                       (b: List[List[arctic]]) =>
 mat_comp_all[arctic](gt[arctic, Unit](arctic_components), ad, b),
                                     false,
                                     concat[Char](List((shows_prec_nat(Zero_nat,
                                n)).apply(Nil),
                List('x'), (shows_prec_nat(Zero_nat, n)).apply(Nil),
                List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'o', 'v',
                      'e', 'r', ' '),
                description[arctic, Unit](arctic_components))),
                                     (ad: List[List[arctic]]) =>
                                       mat_arc_posI[arctic](arc_pos[arctic,
                             Unit](arctic_components),
                     ad),
                                     (_: List[List[arctic]]) => false, ())),
      onea[List[List[arctic]],
            Unit](Poly_c_ext[List[List[arctic]],
                              Unit]((ad: List[List[arctic]]) =>
                                      mat[arctic](n, n, ad),
                                     mat0I[arctic](zero_arctic, n, n),
                                     mat1I[arctic](zero_arctic, one_arctic, n),
                                     mat1I[arctic](zero_arctic,
            default[arctic, Unit](arctic_components), n),
                                     (ad: List[List[arctic]]) =>
                                       (b: List[List[arctic]]) =>
 mat_plusI[arctic]((ae: arctic) => (ba: arctic) => plus_arctic(ae, ba), ad, b),
                                     (ad: List[List[arctic]]) =>
                                       (b: List[List[arctic]]) =>
 mat_multI[arctic](zero_arctic,
                    (ae: arctic) => (ba: arctic) => plus_arctic(ae, ba),
                    (ae: arctic) => (ba: arctic) => times_arctic(ae, ba), n, ad,
                    b),
                                     id[List[List[arctic]]], mat_ge[arctic],
                                     (ad: List[List[arctic]]) =>
                                       (b: List[List[arctic]]) =>
 mat_comp_all[arctic](gt[arctic, Unit](arctic_components), ad, b),
                                     false,
                                     concat[Char](List((shows_prec_nat(Zero_nat,
                                n)).apply(Nil),
                List('x'), (shows_prec_nat(Zero_nat, n)).apply(Nil),
                List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'o', 'v',
                      'e', 'r', ' '),
                description[arctic, Unit](arctic_components))),
                                     (ad: List[List[arctic]]) =>
                                       mat_arc_posI[arctic](arc_pos[arctic,
                             Unit](arctic_components),
                     ad),
                                     (_: List[List[arctic]]) => false, ())),
      plusa[List[List[arctic]],
             Unit](Poly_c_ext[List[List[arctic]],
                               Unit]((ad: List[List[arctic]]) =>
                                       mat[arctic](n, n, ad),
                                      mat0I[arctic](zero_arctic, n, n),
                                      mat1I[arctic](zero_arctic, one_arctic, n),
                                      mat1I[arctic](zero_arctic,
             default[arctic, Unit](arctic_components), n),
                                      (ad: List[List[arctic]]) =>
(b: List[List[arctic]]) =>
  mat_plusI[arctic]((ae: arctic) => (ba: arctic) => plus_arctic(ae, ba), ad, b),
                                      (ad: List[List[arctic]]) =>
(b: List[List[arctic]]) =>
  mat_multI[arctic](zero_arctic,
                     (ae: arctic) => (ba: arctic) => plus_arctic(ae, ba),
                     (ae: arctic) => (ba: arctic) => times_arctic(ae, ba), n,
                     ad, b),
                                      id[List[List[arctic]]], mat_ge[arctic],
                                      (ad: List[List[arctic]]) =>
(b: List[List[arctic]]) =>
  mat_comp_all[arctic](gt[arctic, Unit](arctic_components), ad, b),
                                      false,
                                      concat[Char](List((shows_prec_nat(Zero_nat,
                                 n)).apply(Nil),
                 List('x'), (shows_prec_nat(Zero_nat, n)).apply(Nil),
                 List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'o',
                       'v', 'e', 'r', ' '),
                 description[arctic, Unit](arctic_components))),
                                      (ad: List[List[arctic]]) =>
mat_arc_posI[arctic](arc_pos[arctic, Unit](arctic_components), ad),
                                      (_: List[List[arctic]]) => false, ())),
      timesa[List[List[arctic]],
              Unit](Poly_c_ext[List[List[arctic]],
                                Unit]((ad: List[List[arctic]]) =>
mat[arctic](n, n, ad),
                                       mat0I[arctic](zero_arctic, n, n),
                                       mat1I[arctic](zero_arctic, one_arctic,
              n),
                                       mat1I[arctic](zero_arctic,
              default[arctic, Unit](arctic_components), n),
                                       (ad: List[List[arctic]]) =>
 (b: List[List[arctic]]) =>
   mat_plusI[arctic]((ae: arctic) => (ba: arctic) => plus_arctic(ae, ba), ad,
                      b),
                                       (ad: List[List[arctic]]) =>
 (b: List[List[arctic]]) =>
   mat_multI[arctic](zero_arctic,
                      (ae: arctic) => (ba: arctic) => plus_arctic(ae, ba),
                      (ae: arctic) => (ba: arctic) => times_arctic(ae, ba), n,
                      ad, b),
                                       id[List[List[arctic]]], mat_ge[arctic],
                                       (ad: List[List[arctic]]) =>
 (b: List[List[arctic]]) =>
   mat_comp_all[arctic](gt[arctic, Unit](arctic_components), ad, b),
                                       false,
                                       concat[Char](List((shows_prec_nat(Zero_nat,
                                  n)).apply(Nil),
                  List('x'), (shows_prec_nat(Zero_nat, n)).apply(Nil),
                  List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'o',
                        'v', 'e', 'r', ' '),
                  description[arctic, Unit](arctic_components))),
                                       (ad: List[List[arctic]]) =>
 mat_arc_posI[arctic](arc_pos[arctic, Unit](arctic_components), ad),
                                       (_: List[List[arctic]]) => false, ())),
      ac)),
               (f: A) =>
                 (ac: nat) =>
                   (poly: (List[List[arctic]], List[List[List[arctic]]])) =>
                     Alter6[((A, nat),
                              (List[List[arctic]], List[List[List[arctic]]])),
                             ((A, nat), (BigInt, List[BigInt])),
                             ((A, nat), (rat, List[rat])),
                             ((A, nat), (arctic, List[arctic])),
                             ((A, nat),
                               (List[List[BigInt]], List[List[List[BigInt]]])),
                             ((A, nat),
                               (List[List[rat]], List[List[List[rat]]])),
                             ((A, nat), (arctic_rat, List[arctic_rat])),
                             ((A, nat),
                               (List[List[arctic_rat]],
                                 List[List[List[arctic_rat]]])),
                             ((A, nat), List[(List[(nat, nat)], BigInt)]),
                             ((A, nat),
                               List[(List[(nat, nat)],
                                      rat)])](((f, ac),
        (fst[List[List[arctic]], List[List[List[arctic]]]](poly),
          fit_length[List[List[arctic]]](zeroa[List[List[arctic]],
        Unit](Poly_c_ext[List[List[arctic]],
                          Unit]((ad: List[List[arctic]]) =>
                                  mat[arctic](n, n, ad),
                                 mat0I[arctic](zero_arctic, n, n),
                                 mat1I[arctic](zero_arctic, one_arctic, n),
                                 mat1I[arctic](zero_arctic,
        default[arctic, Unit](arctic_components), n),
                                 (ad: List[List[arctic]]) =>
                                   (b: List[List[arctic]]) =>
                                     mat_plusI[arctic]((ae: arctic) =>
                 (ba: arctic) => plus_arctic(ae, ba),
                ad, b),
                                 (ad: List[List[arctic]]) =>
                                   (b: List[List[arctic]]) =>
                                     mat_multI[arctic](zero_arctic,
                (ae: arctic) => (ba: arctic) => plus_arctic(ae, ba),
                (ae: arctic) => (ba: arctic) => times_arctic(ae, ba), n, ad, b),
                                 id[List[List[arctic]]], mat_ge[arctic],
                                 (ad: List[List[arctic]]) =>
                                   (b: List[List[arctic]]) =>
                                     mat_comp_all[arctic](gt[arctic,
                      Unit](arctic_components),
                   ad, b),
                                 false,
                                 concat[Char](List((shows_prec_nat(Zero_nat,
                            n)).apply(Nil),
            List('x'), (shows_prec_nat(Zero_nat, n)).apply(Nil),
            List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'o', 'v',
                  'e', 'r', ' '),
            description[arctic, Unit](arctic_components))),
                                 (ad: List[List[arctic]]) =>
                                   mat_arc_posI[arctic](arc_pos[arctic,
                         Unit](arctic_components),
                 ad),
                                 (_: List[List[arctic]]) => false, ())),
  ac, snd[List[List[arctic]], List[List[List[arctic]]]](poly))))),
               ab)
           case Arc_rat_mat(n) =>
             (ab: xml) =>
               xml3elements[A, nat,
                             (List[List[arctic_rat]],
                               List[List[List[arctic_rat]]]),
                             alter[((A, nat), (BigInt, List[BigInt])),
                                    ((A, nat), (rat, List[rat])),
                                    ((A, nat), (arctic, List[arctic])),
                                    ((A, nat),
                                      (List[List[BigInt]],
List[List[List[BigInt]]])),
                                    ((A, nat),
                                      (List[List[rat]], List[List[List[rat]]])),
                                    ((A, nat),
                                      (List[List[arctic]],
List[List[List[arctic]]])),
                                    ((A, nat), (arctic_rat, List[arctic_rat])),
                                    ((A, nat),
                                      (List[List[arctic_rat]],
List[List[List[arctic_rat]]])),
                                    ((A, nat),
                                      List[(List[(nat, nat)], BigInt)]),
                                    ((A, nat),
                                      List[(List[(nat, nat)],
     rat)])]](List('i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'), xml2name,
               (ac: xml) => nata(List('a', 'r', 'i', 't', 'y'), ac),
               (x: xml) =>
                 bindb[List[Char], tpoly[nat, List[List[arctic_rat]]],
                        (List[List[arctic_rat]],
                          List[List[List[arctic_rat]]])](polynomial[List[List[arctic_rat]]](mat_coeff[arctic_rat](MinInfty_rat,
                                   arctic_rat_coeff),
             x),
                  (ac: tpoly[nat, List[List[arctic_rat]]]) =>
                    lpoly_of[List[List[arctic_rat]]](zeroa[List[List[arctic_rat]],
                    Unit](Poly_c_ext[List[List[arctic_rat]],
                                      Unit]((ad: List[List[arctic_rat]]) =>
      mat[arctic_rat](n, n, ad),
     mat0I[arctic_rat](zero_arctic_rat, n, n),
     mat1I[arctic_rat](zero_arctic_rat, one_arctic_rat, n),
     mat1I[arctic_rat](zero_arctic_rat,
                        default[arctic_rat, Unit](arctic_rat_components), n),
     (ad: List[List[arctic_rat]]) =>
       (b: List[List[arctic_rat]]) =>
         mat_plusI[arctic_rat]((ae: arctic_rat) =>
                                 (ba: arctic_rat) => plus_arctic_rat(ae, ba),
                                ad, b),
     (ad: List[List[arctic_rat]]) =>
       (b: List[List[arctic_rat]]) =>
         mat_multI[arctic_rat](zero_arctic_rat,
                                (ae: arctic_rat) =>
                                  (ba: arctic_rat) => plus_arctic_rat(ae, ba),
                                (ae: arctic_rat) =>
                                  (ba: arctic_rat) => times_arctic_rat(ae, ba),
                                n, ad, b),
     id[List[List[arctic_rat]]], mat_ge[arctic_rat],
     (ad: List[List[arctic_rat]]) =>
       (b: List[List[arctic_rat]]) =>
         mat_comp_all[arctic_rat](gt[arctic_rat, Unit](arctic_rat_components),
                                   ad, b),
     false,
     concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil), List('x'),
                        (shows_prec_nat(Zero_nat, n)).apply(Nil),
                        List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ',
                              'o', 'v', 'e', 'r', ' '),
                        description[arctic_rat, Unit](arctic_rat_components))),
     (ad: List[List[arctic_rat]]) =>
       mat_arc_posI[arctic_rat](arc_pos[arctic_rat,
 Unit](arctic_rat_components),
                                 ad),
     (_: List[List[arctic_rat]]) => false, ())),
              onea[List[List[arctic_rat]],
                    Unit](Poly_c_ext[List[List[arctic_rat]],
                                      Unit]((ad: List[List[arctic_rat]]) =>
      mat[arctic_rat](n, n, ad),
     mat0I[arctic_rat](zero_arctic_rat, n, n),
     mat1I[arctic_rat](zero_arctic_rat, one_arctic_rat, n),
     mat1I[arctic_rat](zero_arctic_rat,
                        default[arctic_rat, Unit](arctic_rat_components), n),
     (ad: List[List[arctic_rat]]) =>
       (b: List[List[arctic_rat]]) =>
         mat_plusI[arctic_rat]((ae: arctic_rat) =>
                                 (ba: arctic_rat) => plus_arctic_rat(ae, ba),
                                ad, b),
     (ad: List[List[arctic_rat]]) =>
       (b: List[List[arctic_rat]]) =>
         mat_multI[arctic_rat](zero_arctic_rat,
                                (ae: arctic_rat) =>
                                  (ba: arctic_rat) => plus_arctic_rat(ae, ba),
                                (ae: arctic_rat) =>
                                  (ba: arctic_rat) => times_arctic_rat(ae, ba),
                                n, ad, b),
     id[List[List[arctic_rat]]], mat_ge[arctic_rat],
     (ad: List[List[arctic_rat]]) =>
       (b: List[List[arctic_rat]]) =>
         mat_comp_all[arctic_rat](gt[arctic_rat, Unit](arctic_rat_components),
                                   ad, b),
     false,
     concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil), List('x'),
                        (shows_prec_nat(Zero_nat, n)).apply(Nil),
                        List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ',
                              'o', 'v', 'e', 'r', ' '),
                        description[arctic_rat, Unit](arctic_rat_components))),
     (ad: List[List[arctic_rat]]) =>
       mat_arc_posI[arctic_rat](arc_pos[arctic_rat,
 Unit](arctic_rat_components),
                                 ad),
     (_: List[List[arctic_rat]]) => false, ())),
              plusa[List[List[arctic_rat]],
                     Unit](Poly_c_ext[List[List[arctic_rat]],
                                       Unit]((ad: List[List[arctic_rat]]) =>
       mat[arctic_rat](n, n, ad),
      mat0I[arctic_rat](zero_arctic_rat, n, n),
      mat1I[arctic_rat](zero_arctic_rat, one_arctic_rat, n),
      mat1I[arctic_rat](zero_arctic_rat,
                         default[arctic_rat, Unit](arctic_rat_components), n),
      (ad: List[List[arctic_rat]]) =>
        (b: List[List[arctic_rat]]) =>
          mat_plusI[arctic_rat]((ae: arctic_rat) =>
                                  (ba: arctic_rat) => plus_arctic_rat(ae, ba),
                                 ad, b),
      (ad: List[List[arctic_rat]]) =>
        (b: List[List[arctic_rat]]) =>
          mat_multI[arctic_rat](zero_arctic_rat,
                                 (ae: arctic_rat) =>
                                   (ba: arctic_rat) => plus_arctic_rat(ae, ba),
                                 (ae: arctic_rat) =>
                                   (ba: arctic_rat) => times_arctic_rat(ae, ba),
                                 n, ad, b),
      id[List[List[arctic_rat]]], mat_ge[arctic_rat],
      (ad: List[List[arctic_rat]]) =>
        (b: List[List[arctic_rat]]) =>
          mat_comp_all[arctic_rat](gt[arctic_rat, Unit](arctic_rat_components),
                                    ad, b),
      false,
      concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil), List('x'),
                         (shows_prec_nat(Zero_nat, n)).apply(Nil),
                         List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ',
                               'o', 'v', 'e', 'r', ' '),
                         description[arctic_rat, Unit](arctic_rat_components))),
      (ad: List[List[arctic_rat]]) =>
        mat_arc_posI[arctic_rat](arc_pos[arctic_rat,
  Unit](arctic_rat_components),
                                  ad),
      (_: List[List[arctic_rat]]) => false, ())),
              timesa[List[List[arctic_rat]],
                      Unit](Poly_c_ext[List[List[arctic_rat]],
Unit]((ad: List[List[arctic_rat]]) => mat[arctic_rat](n, n, ad),
       mat0I[arctic_rat](zero_arctic_rat, n, n),
       mat1I[arctic_rat](zero_arctic_rat, one_arctic_rat, n),
       mat1I[arctic_rat](zero_arctic_rat,
                          default[arctic_rat, Unit](arctic_rat_components), n),
       (ad: List[List[arctic_rat]]) =>
         (b: List[List[arctic_rat]]) =>
           mat_plusI[arctic_rat]((ae: arctic_rat) =>
                                   (ba: arctic_rat) => plus_arctic_rat(ae, ba),
                                  ad, b),
       (ad: List[List[arctic_rat]]) =>
         (b: List[List[arctic_rat]]) =>
           mat_multI[arctic_rat](zero_arctic_rat,
                                  (ae: arctic_rat) =>
                                    (ba: arctic_rat) => plus_arctic_rat(ae, ba),
                                  (ae: arctic_rat) =>
                                    (ba: arctic_rat) =>
                                      times_arctic_rat(ae, ba),
                                  n, ad, b),
       id[List[List[arctic_rat]]], mat_ge[arctic_rat],
       (ad: List[List[arctic_rat]]) =>
         (b: List[List[arctic_rat]]) =>
           mat_comp_all[arctic_rat](gt[arctic_rat, Unit](arctic_rat_components),
                                     ad, b),
       false,
       concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil), List('x'),
                          (shows_prec_nat(Zero_nat, n)).apply(Nil),
                          List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ',
                                'o', 'v', 'e', 'r', ' '),
                          description[arctic_rat,
                                       Unit](arctic_rat_components))),
       (ad: List[List[arctic_rat]]) =>
         mat_arc_posI[arctic_rat](arc_pos[arctic_rat,
   Unit](arctic_rat_components),
                                   ad),
       (_: List[List[arctic_rat]]) => false, ())),
              ac)),
               (f: A) =>
                 (ac: nat) =>
                   (poly: (List[List[arctic_rat]],
                            List[List[List[arctic_rat]]]))
                     => Alter8[((A, nat),
                                 (List[List[arctic_rat]],
                                   List[List[List[arctic_rat]]])),
                                ((A, nat), (BigInt, List[BigInt])),
                                ((A, nat), (rat, List[rat])),
                                ((A, nat), (arctic, List[arctic])),
                                ((A, nat),
                                  (List[List[BigInt]],
                                    List[List[List[BigInt]]])),
                                ((A, nat),
                                  (List[List[rat]], List[List[List[rat]]])),
                                ((A, nat),
                                  (List[List[arctic]],
                                    List[List[List[arctic]]])),
                                ((A, nat), (arctic_rat, List[arctic_rat])),
                                ((A, nat), List[(List[(nat, nat)], BigInt)]),
                                ((A, nat),
                                  List[(List[(nat, nat)],
 rat)])](((f, ac),
           (fst[List[List[arctic_rat]], List[List[List[arctic_rat]]]](poly),
             fit_length[List[List[arctic_rat]]](zeroa[List[List[arctic_rat]],
               Unit](Poly_c_ext[List[List[arctic_rat]],
                                 Unit]((ad: List[List[arctic_rat]]) =>
 mat[arctic_rat](n, n, ad),
mat0I[arctic_rat](zero_arctic_rat, n, n),
mat1I[arctic_rat](zero_arctic_rat, one_arctic_rat, n),
mat1I[arctic_rat](zero_arctic_rat,
                   default[arctic_rat, Unit](arctic_rat_components), n),
(ad: List[List[arctic_rat]]) =>
  (b: List[List[arctic_rat]]) =>
    mat_plusI[arctic_rat]((ae: arctic_rat) =>
                            (ba: arctic_rat) => plus_arctic_rat(ae, ba),
                           ad, b),
(ad: List[List[arctic_rat]]) =>
  (b: List[List[arctic_rat]]) =>
    mat_multI[arctic_rat](zero_arctic_rat,
                           (ae: arctic_rat) =>
                             (ba: arctic_rat) => plus_arctic_rat(ae, ba),
                           (ae: arctic_rat) =>
                             (ba: arctic_rat) => times_arctic_rat(ae, ba),
                           n, ad, b),
id[List[List[arctic_rat]]], mat_ge[arctic_rat],
(ad: List[List[arctic_rat]]) =>
  (b: List[List[arctic_rat]]) =>
    mat_comp_all[arctic_rat](gt[arctic_rat, Unit](arctic_rat_components), ad,
                              b),
false,
concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil), List('x'),
                   (shows_prec_nat(Zero_nat, n)).apply(Nil),
                   List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'o',
                         'v', 'e', 'r', ' '),
                   description[arctic_rat, Unit](arctic_rat_components))),
(ad: List[List[arctic_rat]]) =>
  mat_arc_posI[arctic_rat](arc_pos[arctic_rat, Unit](arctic_rat_components),
                            ad),
(_: List[List[arctic_rat]]) => false, ())),
         ac, snd[List[List[arctic_rat]],
                  List[List[List[arctic_rat]]]](poly))))),
               ab)
           case Rational(_, deg) =>
             (if (less_eq_nat(deg, one_nat))
               (ab: xml) =>
                 xml3elements[A, nat, (rat, List[rat]),
                               alter[((A, nat), (BigInt, List[BigInt])),
                                      ((A, nat), (rat, List[rat])),
                                      ((A, nat), (arctic, List[arctic])),
                                      ((A, nat),
(List[List[BigInt]], List[List[List[BigInt]]])),
                                      ((A, nat),
(List[List[rat]], List[List[List[rat]]])),
                                      ((A, nat),
(List[List[arctic]], List[List[List[arctic]]])),
                                      ((A, nat),
(arctic_rat, List[arctic_rat])),
                                      ((A, nat),
(List[List[arctic_rat]], List[List[List[arctic_rat]]])),
                                      ((A, nat),
List[(List[(nat, nat)], BigInt)]),
                                      ((A, nat),
List[(List[(nat, nat)],
       rat)])]](List('i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'), xml2name,
                 (ac: xml) => nata(List('a', 'r', 'i', 't', 'y'), ac),
                 (x: xml) =>
                   bindb[List[Char], tpoly[nat, rat],
                          (rat, List[rat])](polynomial[rat](rat, x),
     (ac: tpoly[nat, rat]) =>
       lpoly_of[rat](zeroa[rat, Unit](rat_components),
                      onea[rat, Unit](rat_components),
                      plusa[rat, Unit](rat_components),
                      timesa[rat, Unit](rat_components), ac)),
                 (f: A) =>
                   (ac: nat) =>
                     (poly: (rat, List[rat])) =>
                       Alter2[((A, nat), (rat, List[rat])),
                               ((A, nat), (BigInt, List[BigInt])),
                               ((A, nat), (arctic, List[arctic])),
                               ((A, nat),
                                 (List[List[BigInt]],
                                   List[List[List[BigInt]]])),
                               ((A, nat),
                                 (List[List[rat]], List[List[List[rat]]])),
                               ((A, nat),
                                 (List[List[arctic]],
                                   List[List[List[arctic]]])),
                               ((A, nat), (arctic_rat, List[arctic_rat])),
                               ((A, nat),
                                 (List[List[arctic_rat]],
                                   List[List[List[arctic_rat]]])),
                               ((A, nat), List[(List[(nat, nat)], BigInt)]),
                               ((A, nat),
                                 List[(List[(nat, nat)],
rat)])](((f, ac),
          (fst[rat, List[rat]](poly),
            fit_length[rat](zeroa[rat, Unit](rat_components), ac,
                             snd[rat, List[rat]](poly))))),
                 ab)
               else (ab: xml) =>
                      xml3elements[A, nat, List[(List[(nat, nat)], rat)],
                                    alter[((A, nat), (BigInt, List[BigInt])),
   ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
   ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
   ((A, nat), (List[List[rat]], List[List[List[rat]]])),
   ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
   ((A, nat), (arctic_rat, List[arctic_rat])),
   ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
   ((A, nat), List[(List[(nat, nat)], BigInt)]),
   ((A, nat),
     List[(List[(nat, nat)],
            rat)])]](List('i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'),
                      xml2name,
                      (ac: xml) => nata(List('a', 'r', 'i', 't', 'y'), ac),
                      (x: xml) =>
                        bindb[List[Char], tpoly[nat, rat],
                               List[(List[(nat, nat)],
                                      rat)]](polynomial[rat](rat, x),
      (p: tpoly[nat, rat]) =>
        Inr[List[(List[(nat, nat)], rat)], List[Char]](poly_of[nat, rat](p))),
                      (f: A) =>
                        (ac: nat) =>
                          (poly: List[(List[(nat, nat)], rat)]) =>
                            Alter10[((A, nat), List[(List[(nat, nat)], rat)]),
                                     ((A, nat), (BigInt, List[BigInt])),
                                     ((A, nat), (rat, List[rat])),
                                     ((A, nat), (arctic, List[arctic])),
                                     ((A, nat),
                                       (List[List[BigInt]],
 List[List[List[BigInt]]])),
                                     ((A, nat),
                                       (List[List[rat]],
 List[List[List[rat]]])),
                                     ((A, nat),
                                       (List[List[arctic]],
 List[List[List[arctic]]])),
                                     ((A, nat), (arctic_rat, List[arctic_rat])),
                                     ((A, nat),
                                       (List[List[arctic_rat]],
 List[List[List[arctic_rat]]])),
                                     ((A, nat),
                                       List[(List[(nat, nat)],
      BigInt)])](((f, ac), poly)),
                      ab))
           case Rat_mat(n, sd) =>
             (ab: xml) =>
               xml3elements[A, nat, (List[List[rat]], List[List[List[rat]]]),
                             alter[((A, nat), (BigInt, List[BigInt])),
                                    ((A, nat), (rat, List[rat])),
                                    ((A, nat), (arctic, List[arctic])),
                                    ((A, nat),
                                      (List[List[BigInt]],
List[List[List[BigInt]]])),
                                    ((A, nat),
                                      (List[List[rat]], List[List[List[rat]]])),
                                    ((A, nat),
                                      (List[List[arctic]],
List[List[List[arctic]]])),
                                    ((A, nat), (arctic_rat, List[arctic_rat])),
                                    ((A, nat),
                                      (List[List[arctic_rat]],
List[List[List[arctic_rat]]])),
                                    ((A, nat),
                                      List[(List[(nat, nat)], BigInt)]),
                                    ((A, nat),
                                      List[(List[(nat, nat)],
     rat)])]](List('i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'), xml2name,
               (ac: xml) => nata(List('a', 'r', 'i', 't', 'y'), ac),
               (x: xml) =>
                 bindb[List[Char], tpoly[nat, List[List[rat]]],
                        (List[List[rat]],
                          List[List[List[rat]]])](polynomial[List[List[rat]]](mat_coeff[rat](fract(BigInt(0),
                    BigInt(1)),
              rat),
                                       x),
           (ac: tpoly[nat, List[List[rat]]]) =>
             lpoly_of[List[List[rat]]](zeroa[List[List[rat]],
      Unit](Poly_c_ext[List[List[rat]],
                        Unit]((ad: List[List[rat]]) => mat[rat](n, n, ad),
                               mat0I[rat](fract(BigInt(0), BigInt(1)), n, n),
                               mat1I[rat](fract(BigInt(0), BigInt(1)),
   fract(BigInt(1), BigInt(1)), n),
                               mat1I[rat](fract(BigInt(0), BigInt(1)),
   default[rat, Unit](rat_components), n),
                               (ad: List[List[rat]]) =>
                                 (b: List[List[rat]]) =>
                                   mat_plusI[rat]((ae: rat) =>
            (ba: rat) => plus_rata(ae, ba),
           ad, b),
                               (ad: List[List[rat]]) =>
                                 (b: List[List[rat]]) =>
                                   mat_multI[rat](fract(BigInt(0), BigInt(1)),
           (ae: rat) => (ba: rat) => plus_rata(ae, ba),
           (ae: rat) => (ba: rat) => times_rata(ae, ba), n, ad, b),
                               mat_max0[rat], mat_ge[rat],
                               (ad: List[List[rat]]) =>
                                 (b: List[List[rat]]) =>
                                   mat_gtI[rat](gt[rat, Unit](rat_components),
         sd, ad, b),
                               monoa[rat, Unit](rat_components),
                               concat[Char](List((shows_prec_nat(Zero_nat,
                          n)).apply(Nil),
          List('x'), (shows_prec_nat(Zero_nat, n)).apply(Nil),
          List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'w', 'i', 't',
                'h', ' ', 's', 't', 'r', 'i', 'c', 't', ' ', 'd', 'i', 'm', 'e',
                'n', 's', 'i', 'o', 'n', ' '),
          (shows_prec_nat(Zero_nat, sd)).apply(Nil),
          List(' ', 'o', 'v', 'e', 'r', ' '),
          description[rat, Unit](rat_components))),
                               (_: List[List[rat]]) => true,
                               mat_mono[rat](sd,
      check_mono[rat, Unit](rat_components)),
                               ())),
onea[List[List[rat]],
      Unit](Poly_c_ext[List[List[rat]],
                        Unit]((ad: List[List[rat]]) => mat[rat](n, n, ad),
                               mat0I[rat](fract(BigInt(0), BigInt(1)), n, n),
                               mat1I[rat](fract(BigInt(0), BigInt(1)),
   fract(BigInt(1), BigInt(1)), n),
                               mat1I[rat](fract(BigInt(0), BigInt(1)),
   default[rat, Unit](rat_components), n),
                               (ad: List[List[rat]]) =>
                                 (b: List[List[rat]]) =>
                                   mat_plusI[rat]((ae: rat) =>
            (ba: rat) => plus_rata(ae, ba),
           ad, b),
                               (ad: List[List[rat]]) =>
                                 (b: List[List[rat]]) =>
                                   mat_multI[rat](fract(BigInt(0), BigInt(1)),
           (ae: rat) => (ba: rat) => plus_rata(ae, ba),
           (ae: rat) => (ba: rat) => times_rata(ae, ba), n, ad, b),
                               mat_max0[rat], mat_ge[rat],
                               (ad: List[List[rat]]) =>
                                 (b: List[List[rat]]) =>
                                   mat_gtI[rat](gt[rat, Unit](rat_components),
         sd, ad, b),
                               monoa[rat, Unit](rat_components),
                               concat[Char](List((shows_prec_nat(Zero_nat,
                          n)).apply(Nil),
          List('x'), (shows_prec_nat(Zero_nat, n)).apply(Nil),
          List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'w', 'i', 't',
                'h', ' ', 's', 't', 'r', 'i', 'c', 't', ' ', 'd', 'i', 'm', 'e',
                'n', 's', 'i', 'o', 'n', ' '),
          (shows_prec_nat(Zero_nat, sd)).apply(Nil),
          List(' ', 'o', 'v', 'e', 'r', ' '),
          description[rat, Unit](rat_components))),
                               (_: List[List[rat]]) => true,
                               mat_mono[rat](sd,
      check_mono[rat, Unit](rat_components)),
                               ())),
plusa[List[List[rat]],
       Unit](Poly_c_ext[List[List[rat]],
                         Unit]((ad: List[List[rat]]) => mat[rat](n, n, ad),
                                mat0I[rat](fract(BigInt(0), BigInt(1)), n, n),
                                mat1I[rat](fract(BigInt(0), BigInt(1)),
    fract(BigInt(1), BigInt(1)), n),
                                mat1I[rat](fract(BigInt(0), BigInt(1)),
    default[rat, Unit](rat_components), n),
                                (ad: List[List[rat]]) =>
                                  (b: List[List[rat]]) =>
                                    mat_plusI[rat]((ae: rat) =>
             (ba: rat) => plus_rata(ae, ba),
            ad, b),
                                (ad: List[List[rat]]) =>
                                  (b: List[List[rat]]) =>
                                    mat_multI[rat](fract(BigInt(0), BigInt(1)),
            (ae: rat) => (ba: rat) => plus_rata(ae, ba),
            (ae: rat) => (ba: rat) => times_rata(ae, ba), n, ad, b),
                                mat_max0[rat], mat_ge[rat],
                                (ad: List[List[rat]]) =>
                                  (b: List[List[rat]]) =>
                                    mat_gtI[rat](gt[rat, Unit](rat_components),
          sd, ad, b),
                                monoa[rat, Unit](rat_components),
                                concat[Char](List((shows_prec_nat(Zero_nat,
                           n)).apply(Nil),
           List('x'), (shows_prec_nat(Zero_nat, n)).apply(Nil),
           List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'w', 'i', 't',
                 'h', ' ', 's', 't', 'r', 'i', 'c', 't', ' ', 'd', 'i', 'm',
                 'e', 'n', 's', 'i', 'o', 'n', ' '),
           (shows_prec_nat(Zero_nat, sd)).apply(Nil),
           List(' ', 'o', 'v', 'e', 'r', ' '),
           description[rat, Unit](rat_components))),
                                (_: List[List[rat]]) => true,
                                mat_mono[rat](sd,
       check_mono[rat, Unit](rat_components)),
                                ())),
timesa[List[List[rat]],
        Unit](Poly_c_ext[List[List[rat]],
                          Unit]((ad: List[List[rat]]) => mat[rat](n, n, ad),
                                 mat0I[rat](fract(BigInt(0), BigInt(1)), n, n),
                                 mat1I[rat](fract(BigInt(0), BigInt(1)),
     fract(BigInt(1), BigInt(1)), n),
                                 mat1I[rat](fract(BigInt(0), BigInt(1)),
     default[rat, Unit](rat_components), n),
                                 (ad: List[List[rat]]) =>
                                   (b: List[List[rat]]) =>
                                     mat_plusI[rat]((ae: rat) =>
              (ba: rat) => plus_rata(ae, ba),
             ad, b),
                                 (ad: List[List[rat]]) =>
                                   (b: List[List[rat]]) =>
                                     mat_multI[rat](fract(BigInt(0), BigInt(1)),
             (ae: rat) => (ba: rat) => plus_rata(ae, ba),
             (ae: rat) => (ba: rat) => times_rata(ae, ba), n, ad, b),
                                 mat_max0[rat], mat_ge[rat],
                                 (ad: List[List[rat]]) =>
                                   (b: List[List[rat]]) =>
                                     mat_gtI[rat](gt[rat, Unit](rat_components),
           sd, ad, b),
                                 monoa[rat, Unit](rat_components),
                                 concat[Char](List((shows_prec_nat(Zero_nat,
                            n)).apply(Nil),
            List('x'), (shows_prec_nat(Zero_nat, n)).apply(Nil),
            List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'w', 'i',
                  't', 'h', ' ', 's', 't', 'r', 'i', 'c', 't', ' ', 'd', 'i',
                  'm', 'e', 'n', 's', 'i', 'o', 'n', ' '),
            (shows_prec_nat(Zero_nat, sd)).apply(Nil),
            List(' ', 'o', 'v', 'e', 'r', ' '),
            description[rat, Unit](rat_components))),
                                 (_: List[List[rat]]) => true,
                                 mat_mono[rat](sd,
        check_mono[rat, Unit](rat_components)),
                                 ())),
ac)),
               (f: A) =>
                 (ac: nat) =>
                   (poly: (List[List[rat]], List[List[List[rat]]])) =>
                     Alter5[((A, nat),
                              (List[List[rat]], List[List[List[rat]]])),
                             ((A, nat), (BigInt, List[BigInt])),
                             ((A, nat), (rat, List[rat])),
                             ((A, nat), (arctic, List[arctic])),
                             ((A, nat),
                               (List[List[BigInt]], List[List[List[BigInt]]])),
                             ((A, nat),
                               (List[List[arctic]], List[List[List[arctic]]])),
                             ((A, nat), (arctic_rat, List[arctic_rat])),
                             ((A, nat),
                               (List[List[arctic_rat]],
                                 List[List[List[arctic_rat]]])),
                             ((A, nat), List[(List[(nat, nat)], BigInt)]),
                             ((A, nat),
                               List[(List[(nat, nat)],
                                      rat)])](((f, ac),
        (fst[List[List[rat]], List[List[List[rat]]]](poly),
          fit_length[List[List[rat]]](zeroa[List[List[rat]],
     Unit](Poly_c_ext[List[List[rat]],
                       Unit]((ad: List[List[rat]]) => mat[rat](n, n, ad),
                              mat0I[rat](fract(BigInt(0), BigInt(1)), n, n),
                              mat1I[rat](fract(BigInt(0), BigInt(1)),
  fract(BigInt(1), BigInt(1)), n),
                              mat1I[rat](fract(BigInt(0), BigInt(1)),
  default[rat, Unit](rat_components), n),
                              (ad: List[List[rat]]) =>
                                (b: List[List[rat]]) =>
                                  mat_plusI[rat]((ae: rat) =>
           (ba: rat) => plus_rata(ae, ba),
          ad, b),
                              (ad: List[List[rat]]) =>
                                (b: List[List[rat]]) =>
                                  mat_multI[rat](fract(BigInt(0), BigInt(1)),
          (ae: rat) => (ba: rat) => plus_rata(ae, ba),
          (ae: rat) => (ba: rat) => times_rata(ae, ba), n, ad, b),
                              mat_max0[rat], mat_ge[rat],
                              (ad: List[List[rat]]) =>
                                (b: List[List[rat]]) =>
                                  mat_gtI[rat](gt[rat, Unit](rat_components),
        sd, ad, b),
                              monoa[rat, Unit](rat_components),
                              concat[Char](List((shows_prec_nat(Zero_nat,
                         n)).apply(Nil),
         List('x'), (shows_prec_nat(Zero_nat, n)).apply(Nil),
         List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'w', 'i', 't',
               'h', ' ', 's', 't', 'r', 'i', 'c', 't', ' ', 'd', 'i', 'm', 'e',
               'n', 's', 'i', 'o', 'n', ' '),
         (shows_prec_nat(Zero_nat, sd)).apply(Nil),
         List(' ', 'o', 'v', 'e', 'r', ' '),
         description[rat, Unit](rat_components))),
                              (_: List[List[rat]]) => true,
                              mat_mono[rat](sd,
     check_mono[rat, Unit](rat_components)),
                              ())),
                                       ac, snd[List[List[rat]],
        List[List[List[rat]]]](poly))))),
               ab)
         },
       (typea: domain) =>
         (pi: List[alter[((A, nat), (BigInt, List[BigInt])),
                          ((A, nat), (rat, List[rat])),
                          ((A, nat), (arctic, List[arctic])),
                          ((A, nat),
                            (List[List[BigInt]], List[List[List[BigInt]]])),
                          ((A, nat), (List[List[rat]], List[List[List[rat]]])),
                          ((A, nat),
                            (List[List[arctic]], List[List[List[arctic]]])),
                          ((A, nat), (arctic_rat, List[arctic_rat])),
                          ((A, nat),
                            (List[List[arctic_rat]],
                              List[List[List[arctic_rat]]])),
                          ((A, nat), List[(List[(nat, nat)], BigInt)]),
                          ((A, nat), List[(List[(nat, nat)], rat)])]])
           => typea match {
                case Naturala(deg) =>
                  (if (less_eq_nat(deg, one_nat))
                    Int_carrier[A](mapa[alter[((A, nat),
        (BigInt, List[BigInt])),
       ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
       ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
       ((A, nat), (List[List[rat]], List[List[List[rat]]])),
       ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
       ((A, nat), (arctic_rat, List[arctic_rat])),
       ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
       ((A, nat), List[(List[(nat, nat)], BigInt)]),
       ((A, nat), List[(List[(nat, nat)], rat)])],
 ((A, nat),
   (BigInt,
     List[BigInt]))]((aa: alter[((A, nat), (BigInt, List[BigInt])),
                                 ((A, nat), (rat, List[rat])),
                                 ((A, nat), (arctic, List[arctic])),
                                 ((A, nat),
                                   (List[List[BigInt]],
                                     List[List[List[BigInt]]])),
                                 ((A, nat),
                                   (List[List[rat]], List[List[List[rat]]])),
                                 ((A, nat),
                                   (List[List[arctic]],
                                     List[List[List[arctic]]])),
                                 ((A, nat), (arctic_rat, List[arctic_rat])),
                                 ((A, nat),
                                   (List[List[arctic_rat]],
                                     List[List[List[arctic_rat]]])),
                                 ((A, nat), List[(List[(nat, nat)], BigInt)]),
                                 ((A, nat), List[(List[(nat, nat)], rat)])])
                       => alter1a[((A, nat), (BigInt, List[BigInt])),
                                   ((A, nat), (rat, List[rat])),
                                   ((A, nat), (arctic, List[arctic])),
                                   ((A, nat),
                                     (List[List[BigInt]],
                                       List[List[List[BigInt]]])),
                                   ((A, nat),
                                     (List[List[rat]], List[List[List[rat]]])),
                                   ((A, nat),
                                     (List[List[arctic]],
                                       List[List[List[arctic]]])),
                                   ((A, nat), (arctic_rat, List[arctic_rat])),
                                   ((A, nat),
                                     (List[List[arctic_rat]],
                                       List[List[List[arctic_rat]]])),
                                   ((A, nat), List[(List[(nat, nat)], BigInt)]),
                                   ((A, nat),
                                     List[(List[(nat, nat)], rat)])](aa),
                      pi))
                    else Int_nl_carrier[A](mapa[alter[((A, nat),
                (BigInt, List[BigInt])),
               ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
               ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
               ((A, nat), (List[List[rat]], List[List[List[rat]]])),
               ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
               ((A, nat), (arctic_rat, List[arctic_rat])),
               ((A, nat),
                 (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
               ((A, nat), List[(List[(nat, nat)], BigInt)]),
               ((A, nat), List[(List[(nat, nat)], rat)])],
         ((A, nat),
           List[(List[(nat, nat)],
                  BigInt)])]((aa: alter[((A, nat), (BigInt, List[BigInt])),
 ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
 ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
 ((A, nat), (List[List[rat]], List[List[List[rat]]])),
 ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
 ((A, nat), (arctic_rat, List[arctic_rat])),
 ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
 ((A, nat), List[(List[(nat, nat)], BigInt)]),
 ((A, nat), List[(List[(nat, nat)], rat)])])
                               => alter9a[((A, nat), (BigInt, List[BigInt])),
   ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
   ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
   ((A, nat), (List[List[rat]], List[List[List[rat]]])),
   ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
   ((A, nat), (arctic_rat, List[arctic_rat])),
   ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
   ((A, nat), List[(List[(nat, nat)], BigInt)]),
   ((A, nat), List[(List[(nat, nat)], rat)])](aa),
                              pi)))
                case Arctic =>
                  Arctic_carrier[A](mapa[alter[((A, nat),
         (BigInt, List[BigInt])),
        ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
        ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
        ((A, nat), (List[List[rat]], List[List[List[rat]]])),
        ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
        ((A, nat), (arctic_rat, List[arctic_rat])),
        ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
        ((A, nat), List[(List[(nat, nat)], BigInt)]),
        ((A, nat), List[(List[(nat, nat)], rat)])],
  ((A, nat),
    (arctic,
      List[arctic]))]((aa: alter[((A, nat), (BigInt, List[BigInt])),
                                  ((A, nat), (rat, List[rat])),
                                  ((A, nat), (arctic, List[arctic])),
                                  ((A, nat),
                                    (List[List[BigInt]],
                                      List[List[List[BigInt]]])),
                                  ((A, nat),
                                    (List[List[rat]], List[List[List[rat]]])),
                                  ((A, nat),
                                    (List[List[arctic]],
                                      List[List[List[arctic]]])),
                                  ((A, nat), (arctic_rat, List[arctic_rat])),
                                  ((A, nat),
                                    (List[List[arctic_rat]],
                                      List[List[List[arctic_rat]]])),
                                  ((A, nat), List[(List[(nat, nat)], BigInt)]),
                                  ((A, nat), List[(List[(nat, nat)], rat)])])
                        => alter3a[((A, nat), (BigInt, List[BigInt])),
                                    ((A, nat), (rat, List[rat])),
                                    ((A, nat), (arctic, List[arctic])),
                                    ((A, nat),
                                      (List[List[BigInt]],
List[List[List[BigInt]]])),
                                    ((A, nat),
                                      (List[List[rat]], List[List[List[rat]]])),
                                    ((A, nat),
                                      (List[List[arctic]],
List[List[List[arctic]]])),
                                    ((A, nat), (arctic_rat, List[arctic_rat])),
                                    ((A, nat),
                                      (List[List[arctic_rat]],
List[List[List[arctic_rat]]])),
                                    ((A, nat),
                                      List[(List[(nat, nat)], BigInt)]),
                                    ((A, nat),
                                      List[(List[(nat, nat)], rat)])](aa),
                       pi))
                case Arctic_rat =>
                  Arctic_rat_carrier[A](mapa[alter[((A, nat),
             (BigInt, List[BigInt])),
            ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
            ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
            ((A, nat), (List[List[rat]], List[List[List[rat]]])),
            ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
            ((A, nat), (arctic_rat, List[arctic_rat])),
            ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
            ((A, nat), List[(List[(nat, nat)], BigInt)]),
            ((A, nat), List[(List[(nat, nat)], rat)])],
      ((A, nat),
        (arctic_rat,
          List[arctic_rat]))]((aa: alter[((A, nat), (BigInt, List[BigInt])),
  ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
  ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
  ((A, nat), (List[List[rat]], List[List[List[rat]]])),
  ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
  ((A, nat), (arctic_rat, List[arctic_rat])),
  ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
  ((A, nat), List[(List[(nat, nat)], BigInt)]),
  ((A, nat), List[(List[(nat, nat)], rat)])])
                                => alter7a[((A, nat), (BigInt, List[BigInt])),
    ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
    ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
    ((A, nat), (List[List[rat]], List[List[List[rat]]])),
    ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
    ((A, nat), (arctic_rat, List[arctic_rat])),
    ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
    ((A, nat), List[(List[(nat, nat)], BigInt)]),
    ((A, nat), List[(List[(nat, nat)], rat)])](aa),
                               pi))
                case Int_mat(n, sd) =>
                  Int_mat_carrier[A](n, sd,
                                      mapa[alter[((A, nat),
           (BigInt, List[BigInt])),
          ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
          ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
          ((A, nat), (List[List[rat]], List[List[List[rat]]])),
          ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
          ((A, nat), (arctic_rat, List[arctic_rat])),
          ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
          ((A, nat), List[(List[(nat, nat)], BigInt)]),
          ((A, nat), List[(List[(nat, nat)], rat)])],
    ((A, nat),
      (List[List[BigInt]],
        List[List[List[BigInt]]]))]((aa: alter[((A, nat),
         (BigInt, List[BigInt])),
        ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
        ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
        ((A, nat), (List[List[rat]], List[List[List[rat]]])),
        ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
        ((A, nat), (arctic_rat, List[arctic_rat])),
        ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
        ((A, nat), List[(List[(nat, nat)], BigInt)]),
        ((A, nat), List[(List[(nat, nat)], rat)])])
                                      => alter4a[((A, nat),
           (BigInt, List[BigInt])),
          ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
          ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
          ((A, nat), (List[List[rat]], List[List[List[rat]]])),
          ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
          ((A, nat), (arctic_rat, List[arctic_rat])),
          ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
          ((A, nat), List[(List[(nat, nat)], BigInt)]),
          ((A, nat), List[(List[(nat, nat)], rat)])](aa),
                                     pi))
                case Arc_mat(n) =>
                  Arc_mat_carrier[A](n, mapa[alter[((A, nat),
             (BigInt, List[BigInt])),
            ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
            ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
            ((A, nat), (List[List[rat]], List[List[List[rat]]])),
            ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
            ((A, nat), (arctic_rat, List[arctic_rat])),
            ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
            ((A, nat), List[(List[(nat, nat)], BigInt)]),
            ((A, nat), List[(List[(nat, nat)], rat)])],
      ((A, nat),
        (List[List[arctic]],
          List[List[List[arctic]]]))]((aa:
 alter[((A, nat), (BigInt, List[BigInt])), ((A, nat), (rat, List[rat])),
        ((A, nat), (arctic, List[arctic])),
        ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
        ((A, nat), (List[List[rat]], List[List[List[rat]]])),
        ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
        ((A, nat), (arctic_rat, List[arctic_rat])),
        ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
        ((A, nat), List[(List[(nat, nat)], BigInt)]),
        ((A, nat), List[(List[(nat, nat)], rat)])])
=> alter6a[((A, nat), (BigInt, List[BigInt])), ((A, nat), (rat, List[rat])),
            ((A, nat), (arctic, List[arctic])),
            ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
            ((A, nat), (List[List[rat]], List[List[List[rat]]])),
            ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
            ((A, nat), (arctic_rat, List[arctic_rat])),
            ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
            ((A, nat), List[(List[(nat, nat)], BigInt)]),
            ((A, nat), List[(List[(nat, nat)], rat)])](aa),
                                       pi))
                case Arc_rat_mat(n) =>
                  Arc_rat_mat_carrier[A](n,
  mapa[alter[((A, nat), (BigInt, List[BigInt])), ((A, nat), (rat, List[rat])),
              ((A, nat), (arctic, List[arctic])),
              ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
              ((A, nat), (List[List[rat]], List[List[List[rat]]])),
              ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
              ((A, nat), (arctic_rat, List[arctic_rat])),
              ((A, nat),
                (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
              ((A, nat), List[(List[(nat, nat)], BigInt)]),
              ((A, nat), List[(List[(nat, nat)], rat)])],
        ((A, nat),
          (List[List[arctic_rat]],
            List[List[List[arctic_rat]]]))]((aa:
       alter[((A, nat), (BigInt, List[BigInt])), ((A, nat), (rat, List[rat])),
              ((A, nat), (arctic, List[arctic])),
              ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
              ((A, nat), (List[List[rat]], List[List[List[rat]]])),
              ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
              ((A, nat), (arctic_rat, List[arctic_rat])),
              ((A, nat),
                (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
              ((A, nat), List[(List[(nat, nat)], BigInt)]),
              ((A, nat), List[(List[(nat, nat)], rat)])])
      => alter8a[((A, nat), (BigInt, List[BigInt])),
                  ((A, nat), (rat, List[rat])),
                  ((A, nat), (arctic, List[arctic])),
                  ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
                  ((A, nat), (List[List[rat]], List[List[List[rat]]])),
                  ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
                  ((A, nat), (arctic_rat, List[arctic_rat])),
                  ((A, nat),
                    (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
                  ((A, nat), List[(List[(nat, nat)], BigInt)]),
                  ((A, nat), List[(List[(nat, nat)], rat)])](aa),
     pi))
                case Rational(d, deg) =>
                  (if (less_eq_nat(deg, one_nat))
                    Rat_carrier[A](mapa[alter[((A, nat),
        (BigInt, List[BigInt])),
       ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
       ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
       ((A, nat), (List[List[rat]], List[List[List[rat]]])),
       ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
       ((A, nat), (arctic_rat, List[arctic_rat])),
       ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
       ((A, nat), List[(List[(nat, nat)], BigInt)]),
       ((A, nat), List[(List[(nat, nat)], rat)])],
 ((A, nat),
   (rat, List[rat]))]((aa: alter[((A, nat), (BigInt, List[BigInt])),
                                  ((A, nat), (rat, List[rat])),
                                  ((A, nat), (arctic, List[arctic])),
                                  ((A, nat),
                                    (List[List[BigInt]],
                                      List[List[List[BigInt]]])),
                                  ((A, nat),
                                    (List[List[rat]], List[List[List[rat]]])),
                                  ((A, nat),
                                    (List[List[arctic]],
                                      List[List[List[arctic]]])),
                                  ((A, nat), (arctic_rat, List[arctic_rat])),
                                  ((A, nat),
                                    (List[List[arctic_rat]],
                                      List[List[List[arctic_rat]]])),
                                  ((A, nat), List[(List[(nat, nat)], BigInt)]),
                                  ((A, nat), List[(List[(nat, nat)], rat)])])
                        => alter2a[((A, nat), (BigInt, List[BigInt])),
                                    ((A, nat), (rat, List[rat])),
                                    ((A, nat), (arctic, List[arctic])),
                                    ((A, nat),
                                      (List[List[BigInt]],
List[List[List[BigInt]]])),
                                    ((A, nat),
                                      (List[List[rat]], List[List[List[rat]]])),
                                    ((A, nat),
                                      (List[List[arctic]],
List[List[List[arctic]]])),
                                    ((A, nat), (arctic_rat, List[arctic_rat])),
                                    ((A, nat),
                                      (List[List[arctic_rat]],
List[List[List[arctic_rat]]])),
                                    ((A, nat),
                                      List[(List[(nat, nat)], BigInt)]),
                                    ((A, nat),
                                      List[(List[(nat, nat)], rat)])](aa),
                       pi))
                    else Rat_nl_carrier[A](d,
    mapa[alter[((A, nat), (BigInt, List[BigInt])), ((A, nat), (rat, List[rat])),
                ((A, nat), (arctic, List[arctic])),
                ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
                ((A, nat), (List[List[rat]], List[List[List[rat]]])),
                ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
                ((A, nat), (arctic_rat, List[arctic_rat])),
                ((A, nat),
                  (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
                ((A, nat), List[(List[(nat, nat)], BigInt)]),
                ((A, nat), List[(List[(nat, nat)], rat)])],
          ((A, nat),
            List[(List[(nat, nat)],
                   rat)])]((aa: alter[((A, nat), (BigInt, List[BigInt])),
                                       ((A, nat), (rat, List[rat])),
                                       ((A, nat), (arctic, List[arctic])),
                                       ((A, nat),
 (List[List[BigInt]], List[List[List[BigInt]]])),
                                       ((A, nat),
 (List[List[rat]], List[List[List[rat]]])),
                                       ((A, nat),
 (List[List[arctic]], List[List[List[arctic]]])),
                                       ((A, nat),
 (arctic_rat, List[arctic_rat])),
                                       ((A, nat),
 (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
                                       ((A, nat),
 List[(List[(nat, nat)], BigInt)]),
                                       ((A, nat),
 List[(List[(nat, nat)], rat)])])
                             => alter10a[((A, nat), (BigInt, List[BigInt])),
  ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
  ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
  ((A, nat), (List[List[rat]], List[List[List[rat]]])),
  ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
  ((A, nat), (arctic_rat, List[arctic_rat])),
  ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
  ((A, nat), List[(List[(nat, nat)], BigInt)]),
  ((A, nat), List[(List[(nat, nat)], rat)])](aa),
                            pi)))
                case Rat_mat(n, sd) =>
                  Rat_mat_carrier[A](n, sd,
                                      mapa[alter[((A, nat),
           (BigInt, List[BigInt])),
          ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
          ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
          ((A, nat), (List[List[rat]], List[List[List[rat]]])),
          ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
          ((A, nat), (arctic_rat, List[arctic_rat])),
          ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
          ((A, nat), List[(List[(nat, nat)], BigInt)]),
          ((A, nat), List[(List[(nat, nat)], rat)])],
    ((A, nat),
      (List[List[rat]],
        List[List[List[rat]]]))]((aa: alter[((A, nat), (BigInt, List[BigInt])),
     ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
     ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
     ((A, nat), (List[List[rat]], List[List[List[rat]]])),
     ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
     ((A, nat), (arctic_rat, List[arctic_rat])),
     ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
     ((A, nat), List[(List[(nat, nat)], BigInt)]),
     ((A, nat), List[(List[(nat, nat)], rat)])])
                                   => alter5a[((A, nat),
        (BigInt, List[BigInt])),
       ((A, nat), (rat, List[rat])), ((A, nat), (arctic, List[arctic])),
       ((A, nat), (List[List[BigInt]], List[List[List[BigInt]]])),
       ((A, nat), (List[List[rat]], List[List[List[rat]]])),
       ((A, nat), (List[List[arctic]], List[List[List[arctic]]])),
       ((A, nat), (arctic_rat, List[arctic_rat])),
       ((A, nat), (List[List[arctic_rat]], List[List[List[arctic_rat]]])),
       ((A, nat), List[(List[(nat, nat)], BigInt)]),
       ((A, nat), List[(List[(nat, nat)], rat)])](aa),
                                  pi))
              },
       a)

def redtriple[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], redtriple_impl[A]] =
  (a: xml) =>
    xml1element[redtriple_impl[A],
                 redtriple_impl[A]](List('r', 'e', 'd', 'P', 'a', 'i', 'r'),
                                     (aa: xml) =>
                                       xml_options[redtriple_impl[A]](List((List('i',
  'n', 't', 'e', 'r', 'p', 'r', 'e', 't', 'a', 't', 'i', 'o', 'n'),
                                     interpretation[A](xml2name)),
                                    (List('p', 'a', 't', 'h', 'O', 'r', 'd',
   'e', 'r'),
                                      path_order[A](xml2name))),
                               aa),
                                     id[redtriple_impl[A]], a)

def joinable_critical_pairs[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             List[(term[A, List[Char]],
                    (List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                  term[A, List[Char]]))],
                      (term[A, List[Char]],
                        List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                     term[A, List[Char]]))])))]] =
  {
    val rew: xml =>
               sum[List[Char],
                    (term[A, List[Char]],
                      List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                   term[A, List[Char]]))])]
      = rsteps[A](xml2name);
    (a: xml) =>
      xml_many_elements[(term[A, List[Char]],
                          (List[(pos, ((term[A, List[Char]],
 term[A, List[Char]]),
term[A, List[Char]]))],
                            (term[A, List[Char]],
                              List[(pos, ((term[A, List[Char]],
    term[A, List[Char]]),
   term[A, List[Char]]))]))),
                         List[(term[A, List[Char]],
                                (List[(pos,
((term[A, List[Char]], term[A, List[Char]]), term[A, List[Char]]))],
                                  (term[A, List[Char]],
                                    List[(pos,
   ((term[A, List[Char]], term[A, List[Char]]),
     term[A, List[Char]]))])))]](List('j', 'o', 'i', 'n', 'a', 'b', 'l', 'e',
                                       'C', 'r', 'i', 't', 'i', 'c', 'a', 'l',
                                       'P', 'a', 'i', 'r', 's'),
                                  (aa: xml) =>
                                    xml2elements[(term[A, List[Char]],
           List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                        term[A, List[Char]]))]),
          (term[A, List[Char]],
            List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                         term[A, List[Char]]))]),
          (term[A, List[Char]],
            (List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                          term[A, List[Char]]))],
              (term[A, List[Char]],
                List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                             term[A, List[Char]]))])))](List('j', 'o', 'i', 'n',
                      'a', 'b', 'l', 'e', 'C', 'r', 'i', 't', 'i', 'c', 'a',
                      'l', 'P', 'a', 'i', 'r'),
                 rew, rew,
                 (ab: (term[A, List[Char]],
                        List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                     term[A, List[Char]]))]))
                   => {
                        val (s, sseq):
                              (term[A, List[Char]],
                                List[(pos, ((term[A, List[Char]],
      term[A, List[Char]]),
     term[A, List[Char]]))])
                          = ab;
                        (ac: (term[A, List[Char]],
                               List[(pos, ((term[A, List[Char]],
     term[A, List[Char]]),
    term[A, List[Char]]))]))
                          => {
                               val (t, tseq):
                                     (term[A, List[Char]],
                                       List[(pos,
      ((term[A, List[Char]], term[A, List[Char]]), term[A, List[Char]]))])
                                 = ac;
                               (s, (sseq, (t, tseq)))
                             }
                      },
                 aa),
                                  id[List[(term[A, List[Char]],
    (List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                  term[A, List[Char]]))],
      (term[A, List[Char]],
        List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                     term[A, List[Char]]))])))]],
                                  a)
  }

def wcr_proof[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             Option[List[(term[A, List[Char]],
                           (List[(pos, ((term[A, List[Char]],
  term[A, List[Char]]),
 term[A, List[Char]]))],
                             (term[A, List[Char]],
                               List[(pos, ((term[A, List[Char]],
     term[A, List[Char]]),
    term[A, List[Char]]))])))]]] =
  (a: xml) =>
    xml1element[Option[List[(term[A, List[Char]],
                              (List[(pos, ((term[A, List[Char]],
     term[A, List[Char]]),
    term[A, List[Char]]))],
                                (term[A, List[Char]],
                                  List[(pos,
 ((term[A, List[Char]], term[A, List[Char]]), term[A, List[Char]]))])))]],
                 Option[List[(term[A, List[Char]],
                               (List[(pos, ((term[A, List[Char]],
      term[A, List[Char]]),
     term[A, List[Char]]))],
                                 (term[A, List[Char]],
                                   List[(pos,
  ((term[A, List[Char]], term[A, List[Char]]),
    term[A, List[Char]]))])))]]](List('w', 'c', 'r', 'P', 'r', 'o', 'o', 'f'),
                                  (aa: xml) =>
                                    xml_options[Option[List[(term[A,
                           List[Char]],
                      (List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                    term[A, List[Char]]))],
                        (term[A, List[Char]],
                          List[(pos, ((term[A, List[Char]],
term[A, List[Char]]),
                                       term[A,
     List[Char]]))])))]]](List((List('j', 'o', 'i', 'n', 'a', 'b', 'l', 'e',
                                      'C', 'r', 'i', 't', 'i', 'c', 'a', 'l',
                                      'P', 'a', 'i', 'r', 's'),
                                 (ab: xml) =>
                                   xml_change[List[(term[A, List[Char]],
             (List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                           term[A, List[Char]]))],
               (term[A, List[Char]],
                 List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                              term[A, List[Char]]))])))],
       Option[List[(term[A, List[Char]],
                     (List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                   term[A, List[Char]]))],
                       (term[A, List[Char]],
                         List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                      term[A,
    List[Char]]))])))]]](joinable_critical_pairs[A](xml2name),
                          (ac: List[(term[A, List[Char]],
                                      (List[(pos,
      ((term[A, List[Char]], term[A, List[Char]]), term[A, List[Char]]))],
(term[A, List[Char]],
  List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
               term[A, List[Char]]))])))])
                            => Some[List[(term[A, List[Char]],
   (List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                 term[A, List[Char]]))],
     (term[A, List[Char]],
       List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                    term[A, List[Char]]))])))]](ac),
                          ab)),
                                (List('j', 'o', 'i', 'n', 'a', 'b', 'l', 'e',
                                       'C', 'r', 'i', 't', 'i', 'c', 'a', 'l',
                                       'P', 'a', 'i', 'r', 's', 'A', 'u', 't',
                                       'o'),
                                  (ab: xml) =>
                                    xml0elements[Option[List[(term[A,
                            List[Char]],
                       (List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                     term[A, List[Char]]))],
                         (term[A, List[Char]],
                           List[(pos, ((term[A, List[Char]],
 term[A, List[Char]]),
term[A, List[Char]]))])))]]](List('j', 'o', 'i', 'n', 'a', 'b', 'l', 'e', 'C',
                                   'r', 'i', 't', 'i', 'c', 'a', 'l', 'P', 'a',
                                   'i', 'r', 's', 'A', 'u', 't', 'o'),
                              None, ab))),
                           aa),
                                  id[Option[List[(term[A, List[Char]],
           (List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                         term[A, List[Char]]))],
             (term[A, List[Char]],
               List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                            term[A, List[Char]]))])))]]],
                                  a)

def xml2dp_input[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             (List[(term[A, List[Char]], term[A, List[Char]])],
               (strategy[A, List[Char]],
                 List[(term[A, List[Char]], term[A, List[Char]])]))] =
  (a: xml) =>
    xml_choice[(List[(term[A, List[Char]], term[A, List[Char]])],
                 (strategy[A, List[Char]],
                   List[(term[A, List[Char]],
                          term[A, List[Char]])]))](List('d', 'p', 'I', 'n', 'p',
                 'u', 't'),
            List((aa: xml) =>
                   xml4elements[List[(term[A, List[Char]],
                                       term[A, List[Char]])],
                                 List[(term[A, List[Char]],
term[A, List[Char]])],
                                 strategy[A, List[Char]], Boolean,
                                 (List[(term[A, List[Char]],
 term[A, List[Char]])],
                                   (strategy[A, List[Char]],
                                     List[(term[A, List[Char]],
    term[A, List[Char]])]))](List('d', 'p', 'I', 'n', 'p', 'u', 't'),
                              (ab: xml) =>
                                xml1element[List[(term[A, List[Char]],
           term[A, List[Char]])],
     List[(term[A, List[Char]],
            term[A, List[Char]])]](List('t', 'r', 's'), rules[A](xml2name),
                                    id[List[(term[A, List[Char]],
      term[A, List[Char]])]],
                                    ab),
                              (ab: xml) =>
                                xml1element[List[(term[A, List[Char]],
           term[A, List[Char]])],
     List[(term[A, List[Char]],
            term[A, List[Char]])]](List('d', 'p', 's'), rules[A](xml2name),
                                    id[List[(term[A, List[Char]],
      term[A, List[Char]])]],
                                    ab),
                              strategy[A](xml2name),
                              (ab: xml) =>
                                bool(List('m', 'i', 'n', 'i', 'm', 'a', 'l'),
                                      ab),
                              (r: List[(term[A, List[Char]],
 term[A, List[Char]])])
                                => (p: List[(term[A, List[Char]],
      term[A, List[Char]])])
                                     => (s: strategy[A, List[Char]]) =>
  (_: Boolean) => (p, (s, r)),
                              aa),
                  (aa: xml) =>
                    xml3elements[List[(term[A, List[Char]],
term[A, List[Char]])],
                                  List[(term[A, List[Char]],
 term[A, List[Char]])],
                                  Boolean,
                                  (List[(term[A, List[Char]],
  term[A, List[Char]])],
                                    (strategy[A, List[Char]],
                                      List[(term[A, List[Char]],
     term[A, List[Char]])]))](List('d', 'p', 'I', 'n', 'p', 'u', 't'),
                               (ab: xml) =>
                                 xml1element[List[(term[A, List[Char]],
            term[A, List[Char]])],
      List[(term[A, List[Char]],
             term[A, List[Char]])]](List('t', 'r', 's'), rules[A](xml2name),
                                     id[List[(term[A, List[Char]],
       term[A, List[Char]])]],
                                     ab),
                               (ab: xml) =>
                                 xml1element[List[(term[A, List[Char]],
            term[A, List[Char]])],
      List[(term[A, List[Char]],
             term[A, List[Char]])]](List('d', 'p', 's'), rules[A](xml2name),
                                     id[List[(term[A, List[Char]],
       term[A, List[Char]])]],
                                     ab),
                               (ab: xml) =>
                                 bool(List('m', 'i', 'n', 'i', 'm', 'a', 'l'),
                                       ab),
                               (r: List[(term[A, List[Char]],
  term[A, List[Char]])])
                                 => (p: List[(term[A, List[Char]],
       term[A, List[Char]])])
                                      => (_: Boolean) => (p, (No_Strategy, r)),
                               aa)),
            a)

def xml2trs_input[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             (strategy[A, List[Char]],
               (List[(term[A, List[Char]], term[A, List[Char]])],
                 Option[List[(term[A, List[Char]], term[A, List[Char]])]]))] =
  (x: xml) =>
    {
      val e: sum[List[Char],
                  (strategy[A, List[Char]],
                    (List[(term[A, List[Char]], term[A, List[Char]])],
                      Option[List[(term[A, List[Char]],
                                    term[A, List[Char]])]]))]
        = no_transform[(strategy[A, List[Char]],
                         (List[(term[A, List[Char]], term[A, List[Char]])],
                           Option[List[(term[A, List[Char]],
 term[A, List[Char]])]]))](List('t', 'r', 's', 'I', 'n', 'p', 'u', 't'), x);
      x match {
        case Xml(tag, Nil, Nil, option) => e
        case Xml(tag, Nil, c :: cs, None) =>
          (if (less_eq_nat(size_list[xml](cs), nat(BigInt(2))) &&
                 equal_lista[Char](tag, List('t', 'r', 's', 'I', 'n', 'p', 'u',
      't')))
            bindb[List[Char], List[(term[A, List[Char]], term[A, List[Char]])],
                   (strategy[A, List[Char]],
                     (List[(term[A, List[Char]], term[A, List[Char]])],
                       Option[List[(term[A, List[Char]],
                                     term[A,
   List[Char]])]]))](xml1element[List[(term[A, List[Char]],
term[A, List[Char]])],
                                  List[(term[A, List[Char]],
 term[A, List[Char]])]](List('t', 'r', 's'), rules[A](xml2name),
                         id[List[(term[A, List[Char]], term[A, List[Char]])]],
                         c),
                      (r: List[(term[A, List[Char]], term[A, List[Char]])]) =>
                        bindb[List[Char], (strategy[A, List[Char]], List[xml]),
                               (strategy[A, List[Char]],
                                 (List[(term[A, List[Char]],
 term[A, List[Char]])],
                                   Option[List[(term[A, List[Char]],
         term[A, List[Char]])]]))]((if (nulla[xml](cs) ||
  ! (equal_lista[Char](get_name(hd[xml](cs)),
                        List('s', 't', 'r', 'a', 't', 'e', 'g', 'y'))))
                                     Inr[(strategy[A, List[Char]], List[xml]),
  List[Char]]((No_Strategy, cs))
                                     else bindb[List[Char],
         strategy[A, List[Char]],
         (strategy[A, List[Char]],
           List[xml])]((strategy[A](xml2name)).apply(hd[xml](cs)),
                        (s: strategy[A, List[Char]]) =>
                          Inr[(strategy[A, List[Char]], List[xml]),
                               List[Char]]((s, tl[xml](cs))))),
                                    (a: (strategy[A, List[Char]], List[xml])) =>
                                      {
val (str, csa): (strategy[A, List[Char]], List[xml]) = a;
bindb[List[Char],
       (Option[List[(term[A, List[Char]], term[A, List[Char]])]], List[xml]),
       (strategy[A, List[Char]],
         (List[(term[A, List[Char]], term[A, List[Char]])],
           Option[List[(term[A, List[Char]],
                         term[A, List[Char]])]]))]((if (nulla[xml](csa) ||
                  ! (equal_lista[Char](get_name(hd[xml](csa)),
List('r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'R', 'u', 'l', 'e', 's'))))
             Inr[(Option[List[(term[A, List[Char]], term[A, List[Char]])]],
                   List[xml]),
                  List[Char]]((None, csa))
             else bindb[List[Char],
                         List[(term[A, List[Char]], term[A, List[Char]])],
                         (Option[List[(term[A, List[Char]],
term[A, List[Char]])]],
                           List[xml])](xml1element[List[(term[A, List[Char]],
                  term[A, List[Char]])],
            List[(term[A, List[Char]],
                   term[A, List[Char]])]](List('r', 'e', 'l', 'a', 't', 'i',
        'v', 'e', 'R', 'u', 'l', 'e', 's'),
   rules[A](xml2name), id[List[(term[A, List[Char]], term[A, List[Char]])]],
   hd[xml](csa)),
(s: List[(term[A, List[Char]], term[A, List[Char]])]) =>
  Inr[(Option[List[(term[A, List[Char]], term[A, List[Char]])]], List[xml]),
       List[Char]]((Some[List[(term[A, List[Char]], term[A, List[Char]])]](s),
                     tl[xml](csa))))),
            (aa: (Option[List[(term[A, List[Char]], term[A, List[Char]])]],
                   List[xml]))
              => {
                   val (s_opt, csb):
                         (Option[List[(term[A, List[Char]],
term[A, List[Char]])]],
                           List[xml])
                     = aa;
                   (if (nulla[xml](csb))
                     Inr[(strategy[A, List[Char]],
                           (List[(term[A, List[Char]], term[A, List[Char]])],
                             Option[List[(term[A, List[Char]],
   term[A, List[Char]])]])),
                          List[Char]]((str, (r, s_opt)))
                     else e)
                 })
                                      }))
            else e)
        case Xml(tag, Nil, c :: cs, Some(_)) => e
        case Xml(tag, _ :: _, list3, option) => e
      }
    }

def xml2completion_input[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             (List[(term[A, List[Char]], term[A, List[Char]])],
               List[(term[A, List[Char]], term[A, List[Char]])])] =
  (a: xml) =>
    xml2elements[List[(term[A, List[Char]], term[A, List[Char]])],
                  List[(term[A, List[Char]], term[A, List[Char]])],
                  (List[(term[A, List[Char]], term[A, List[Char]])],
                    List[(term[A, List[Char]],
                           term[A, List[Char]])])](List('c', 'o', 'm', 'p', 'l',
                 'e', 't', 'i', 'o', 'n', 'I', 'n', 'p', 'u', 't'),
            (aa: xml) =>
              xml1element[List[(term[A, List[Char]], term[A, List[Char]])],
                           List[(term[A, List[Char]],
                                  term[A,
List[Char]])]](List('e', 'q', 'u', 'a', 't', 'i', 'o', 'n', 's'),
                rules[A](xml2name),
                id[List[(term[A, List[Char]], term[A, List[Char]])]], aa),
            (aa: xml) =>
              xml1element[List[(term[A, List[Char]], term[A, List[Char]])],
                           List[(term[A, List[Char]],
                                  term[A,
List[Char]])]](List('t', 'r', 's'), rules[A](xml2name),
                id[List[(term[A, List[Char]], term[A, List[Char]])]], aa),
            (aa: List[(term[A, List[Char]], term[A, List[Char]])]) =>
              (b: List[(term[A, List[Char]], term[A, List[Char]])]) => (aa, b),
            a)

def xml2input[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             sum[(List[(term[A, List[Char]], term[A, List[Char]])],
                   (strategy[A, List[Char]],
                     List[(term[A, List[Char]], term[A, List[Char]])])),
                  sum[(strategy[A, List[Char]],
                        (List[(term[A, List[Char]], term[A, List[Char]])],
                          Option[List[(term[A, List[Char]],
term[A, List[Char]])]])),
                       (List[(term[A, List[Char]], term[A, List[Char]])],
                         List[(term[A, List[Char]], term[A, List[Char]])])]]] =
  (a: xml) =>
    xml1element[sum[(List[(term[A, List[Char]], term[A, List[Char]])],
                      (strategy[A, List[Char]],
                        List[(term[A, List[Char]], term[A, List[Char]])])),
                     sum[(strategy[A, List[Char]],
                           (List[(term[A, List[Char]], term[A, List[Char]])],
                             Option[List[(term[A, List[Char]],
   term[A, List[Char]])]])),
                          (List[(term[A, List[Char]], term[A, List[Char]])],
                            List[(term[A, List[Char]], term[A, List[Char]])])]],
                 sum[(List[(term[A, List[Char]], term[A, List[Char]])],
                       (strategy[A, List[Char]],
                         List[(term[A, List[Char]], term[A, List[Char]])])),
                      sum[(strategy[A, List[Char]],
                            (List[(term[A, List[Char]], term[A, List[Char]])],
                              Option[List[(term[A, List[Char]],
    term[A, List[Char]])]])),
                           (List[(term[A, List[Char]], term[A, List[Char]])],
                             List[(term[A, List[Char]],
                                    term[A,
  List[Char]])])]]](List('i', 'n', 'p', 'u', 't'),
                     (aa: xml) =>
                       xml_choice[sum[(List[(term[A, List[Char]],
      term[A, List[Char]])],
(strategy[A, List[Char]], List[(term[A, List[Char]], term[A, List[Char]])])),
                                       sum[(strategy[A, List[Char]],
     (List[(term[A, List[Char]], term[A, List[Char]])],
       Option[List[(term[A, List[Char]], term[A, List[Char]])]])),
    (List[(term[A, List[Char]], term[A, List[Char]])],
      List[(term[A, List[Char]],
             term[A, List[Char]])])]]](List('i', 'n', 'p', 'u', 't'),
List((ab: xml) =>
       xml_change[(strategy[A, List[Char]],
                    (List[(term[A, List[Char]], term[A, List[Char]])],
                      Option[List[(term[A, List[Char]],
                                    term[A, List[Char]])]])),
                   sum[(List[(term[A, List[Char]], term[A, List[Char]])],
                         (strategy[A, List[Char]],
                           List[(term[A, List[Char]], term[A, List[Char]])])),
                        sum[(strategy[A, List[Char]],
                              (List[(term[A, List[Char]], term[A, List[Char]])],
                                Option[List[(term[A, List[Char]],
      term[A, List[Char]])]])),
                             (List[(term[A, List[Char]], term[A, List[Char]])],
                               List[(term[A, List[Char]],
                                      term[A,
    List[Char]])])]]](xml2trs_input[A](xml2name),
                       comp[sum[(strategy[A, List[Char]],
                                  (List[(term[A, List[Char]],
  term[A, List[Char]])],
                                    Option[List[(term[A, List[Char]],
          term[A, List[Char]])]])),
                                 (List[(term[A, List[Char]],
 term[A, List[Char]])],
                                   List[(term[A, List[Char]],
  term[A, List[Char]])])],
                             sum[(List[(term[A, List[Char]],
 term[A, List[Char]])],
                                   (strategy[A, List[Char]],
                                     List[(term[A, List[Char]],
    term[A, List[Char]])])),
                                  sum[(strategy[A, List[Char]],
(List[(term[A, List[Char]], term[A, List[Char]])],
  Option[List[(term[A, List[Char]], term[A, List[Char]])]])),
                                       (List[(term[A, List[Char]],
       term[A, List[Char]])],
 List[(term[A, List[Char]], term[A, List[Char]])])]],
                             (strategy[A, List[Char]],
                               (List[(term[A, List[Char]],
                                       term[A, List[Char]])],
                                 Option[List[(term[A, List[Char]],
       term[A, List[Char]])]]))]((ac: sum[(strategy[A, List[Char]],
    (List[(term[A, List[Char]], term[A, List[Char]])],
      Option[List[(term[A, List[Char]], term[A, List[Char]])]])),
   (List[(term[A, List[Char]], term[A, List[Char]])],
     List[(term[A, List[Char]], term[A, List[Char]])])])
                                   => Inr[sum[(strategy[A, List[Char]],
        (List[(term[A, List[Char]], term[A, List[Char]])],
          Option[List[(term[A, List[Char]], term[A, List[Char]])]])),
       (List[(term[A, List[Char]], term[A, List[Char]])],
         List[(term[A, List[Char]], term[A, List[Char]])])],
   (List[(term[A, List[Char]], term[A, List[Char]])],
     (strategy[A, List[Char]],
       List[(term[A, List[Char]], term[A, List[Char]])]))](ac),
                                  (ac: (strategy[A, List[Char]],
 (List[(term[A, List[Char]], term[A, List[Char]])],
   Option[List[(term[A, List[Char]], term[A, List[Char]])]])))
                                    => Inl[(strategy[A, List[Char]],
     (List[(term[A, List[Char]], term[A, List[Char]])],
       Option[List[(term[A, List[Char]], term[A, List[Char]])]])),
    (List[(term[A, List[Char]], term[A, List[Char]])],
      List[(term[A, List[Char]], term[A, List[Char]])])](ac)),
                       ab),
      (ab: xml) =>
        xml_change[(List[(term[A, List[Char]], term[A, List[Char]])],
                     (strategy[A, List[Char]],
                       List[(term[A, List[Char]], term[A, List[Char]])])),
                    sum[(List[(term[A, List[Char]], term[A, List[Char]])],
                          (strategy[A, List[Char]],
                            List[(term[A, List[Char]], term[A, List[Char]])])),
                         sum[(strategy[A, List[Char]],
                               (List[(term[A, List[Char]],
                                       term[A, List[Char]])],
                                 Option[List[(term[A, List[Char]],
       term[A, List[Char]])]])),
                              (List[(term[A, List[Char]], term[A, List[Char]])],
                                List[(term[A, List[Char]],
                                       term[A,
     List[Char]])])]]](xml2dp_input[A](xml2name),
                        (ac: (List[(term[A, List[Char]], term[A, List[Char]])],
                               (strategy[A, List[Char]],
                                 List[(term[A, List[Char]],
term[A, List[Char]])])))
                          => Inl[(List[(term[A, List[Char]],
 term[A, List[Char]])],
                                   (strategy[A, List[Char]],
                                     List[(term[A, List[Char]],
    term[A, List[Char]])])),
                                  sum[(strategy[A, List[Char]],
(List[(term[A, List[Char]], term[A, List[Char]])],
  Option[List[(term[A, List[Char]], term[A, List[Char]])]])),
                                       (List[(term[A, List[Char]],
       term[A, List[Char]])],
 List[(term[A, List[Char]], term[A, List[Char]])])]](ac),
                        ab),
      (ab: xml) =>
        xml_change[(List[(term[A, List[Char]], term[A, List[Char]])],
                     List[(term[A, List[Char]], term[A, List[Char]])]),
                    sum[(List[(term[A, List[Char]], term[A, List[Char]])],
                          (strategy[A, List[Char]],
                            List[(term[A, List[Char]], term[A, List[Char]])])),
                         sum[(strategy[A, List[Char]],
                               (List[(term[A, List[Char]],
                                       term[A, List[Char]])],
                                 Option[List[(term[A, List[Char]],
       term[A, List[Char]])]])),
                              (List[(term[A, List[Char]], term[A, List[Char]])],
                                List[(term[A, List[Char]],
                                       term[A,
     List[Char]])])]]](xml2completion_input[A](xml2name),
                        comp[sum[(strategy[A, List[Char]],
                                   (List[(term[A, List[Char]],
   term[A, List[Char]])],
                                     Option[List[(term[A, List[Char]],
           term[A, List[Char]])]])),
                                  (List[(term[A, List[Char]],
  term[A, List[Char]])],
                                    List[(term[A, List[Char]],
   term[A, List[Char]])])],
                              sum[(List[(term[A, List[Char]],
  term[A, List[Char]])],
                                    (strategy[A, List[Char]],
                                      List[(term[A, List[Char]],
     term[A, List[Char]])])),
                                   sum[(strategy[A, List[Char]],
 (List[(term[A, List[Char]], term[A, List[Char]])],
   Option[List[(term[A, List[Char]], term[A, List[Char]])]])),
(List[(term[A, List[Char]], term[A, List[Char]])],
  List[(term[A, List[Char]], term[A, List[Char]])])]],
                              (List[(term[A, List[Char]], term[A, List[Char]])],
                                List[(term[A, List[Char]],
                                       term[A,
     List[Char]])])]((ac: sum[(strategy[A, List[Char]],
                                (List[(term[A, List[Char]],
term[A, List[Char]])],
                                  Option[List[(term[A, List[Char]],
        term[A, List[Char]])]])),
                               (List[(term[A, List[Char]],
                                       term[A, List[Char]])],
                                 List[(term[A, List[Char]],
term[A, List[Char]])])])
                       => Inr[sum[(strategy[A, List[Char]],
                                    (List[(term[A, List[Char]],
    term[A, List[Char]])],
                                      Option[List[(term[A, List[Char]],
            term[A, List[Char]])]])),
                                   (List[(term[A, List[Char]],
   term[A, List[Char]])],
                                     List[(term[A, List[Char]],
    term[A, List[Char]])])],
                               (List[(term[A, List[Char]],
                                       term[A, List[Char]])],
                                 (strategy[A, List[Char]],
                                   List[(term[A, List[Char]],
  term[A, List[Char]])]))](ac),
                      (ac: (List[(term[A, List[Char]], term[A, List[Char]])],
                             List[(term[A, List[Char]], term[A, List[Char]])]))
                        => Inr[(List[(term[A, List[Char]],
                                       term[A, List[Char]])],
                                 List[(term[A, List[Char]],
term[A, List[Char]])]),
                                (strategy[A, List[Char]],
                                  (List[(term[A, List[Char]],
  term[A, List[Char]])],
                                    Option[List[(term[A, List[Char]],
          term[A, List[Char]])]]))](ac)),
                        ab)),
aa),
                     id[sum[(List[(term[A, List[Char]], term[A, List[Char]])],
                              (strategy[A, List[Char]],
                                List[(term[A, List[Char]],
                                       term[A, List[Char]])])),
                             sum[(strategy[A, List[Char]],
                                   (List[(term[A, List[Char]],
   term[A, List[Char]])],
                                     Option[List[(term[A, List[Char]],
           term[A, List[Char]])]])),
                                  (List[(term[A, List[Char]],
  term[A, List[Char]])],
                                    List[(term[A, List[Char]],
   term[A, List[Char]])])]]],
                     a)

def xml_rules[A : show, B : show,
               C : show](tag: List[Char],
                          rls: List[(term[lab[A, List[B]], C],
                                      term[lab[A, List[B]], C])]):
      xml =
  Xml(tag, Nil,
       List(Xml(List('r', 'u', 'l', 'e', 's'), Nil,
                 mapa[(term[lab[A, List[B]], C], term[lab[A, List[B]], C]),
                       xml]((a: (term[lab[A, List[B]], C],
                                  term[lab[A, List[B]], C]))
                              => {
                                   val (l, r):
 (term[lab[A, List[B]], C], term[lab[A, List[B]], C])
                                     = a;
                                   Xml(List('r', 'u', 'l', 'e'), Nil,
List(Xml(List('l', 'h', 's'), Nil, List(xml_term[A, B, C](l)), None),
      Xml(List('r', 'h', 's'), Nil, List(xml_term[A, B, C](r)), None)),
None)
                                 },
                             rls),
                 None)),
       None)

def gen_list_to_set_aux[A, B](ins: A => B => B, accs: B, x2: List[A]): B =
  (ins, accs, x2) match {
  case (ins, accs, Nil) => accs
  case (ins, accs, x :: l) => gen_list_to_set_aux[A, B](ins, (ins(x))(accs), l)
}

def gen_list_to_set[A, B](empt: A, ins: B => A => A): (List[B]) => A =
  (a: List[B]) => gen_list_to_set_aux[B, A](ins, empt, a)

def list_to_rs[A : linorder]: (List[A]) => rbt[A, Unit] =
  gen_list_to_set[rbt[A, Unit], A](rs_empty[A], rs_ins[A])

implicit def preorder_list[A : equal : order]: preorder[List[A]] = new
  preorder[List[A]] {
  val `Ceta.less_eq` = (a: List[A], b: List[A]) => less_eq_list[A](a, b)
  val `Ceta.less` = (a: List[A], b: List[A]) => less_list[A](a, b)
}

implicit def order_list[A : equal : order]: order[List[A]] = new order[List[A]]
  {
  val `Ceta.less_eq` = (a: List[A], b: List[A]) => less_eq_list[A](a, b)
  val `Ceta.less` = (a: List[A], b: List[A]) => less_list[A](a, b)
}

implicit def linorder_list[A : equal : linorder]: linorder[List[A]] = new
  linorder[List[A]] {
  val `Ceta.less_eq` = (a: List[A], b: List[A]) => less_eq_list[A](a, b)
  val `Ceta.less` = (a: List[A], b: List[A]) => less_list[A](a, b)
}

implicit def ord_char: ord[Char] = new ord[Char] {
  val `Ceta.less_eq` = (a: Char, b: Char) => a <= b
  val `Ceta.less` = (a: Char, b: Char) => a < b
}

implicit def preorder_char: preorder[Char] = new preorder[Char] {
  val `Ceta.less_eq` = (a: Char, b: Char) => a <= b
  val `Ceta.less` = (a: Char, b: Char) => a < b
}

implicit def order_char: order[Char] = new order[Char] {
  val `Ceta.less_eq` = (a: Char, b: Char) => a <= b
  val `Ceta.less` = (a: Char, b: Char) => a < b
}

implicit def linorder_char: linorder[Char] = new linorder[Char] {
  val `Ceta.less_eq` = (a: Char, b: Char) => a <= b
  val `Ceta.less` = (a: Char, b: Char) => a < b
}

def icap_impla[A : equal : linorder](r: List[(term[A, List[Char]],
       term[A, List[Char]])],
                                      q: List[term[A, List[Char]]]):
      (List[term[A, List[Char]]]) =>
        (term[A, List[Char]]) => term[A, sum[Unit, List[Char]]] =
  {
    val ls: List[term[A, List[Char]]] =
      mapa[(term[A, List[Char]], term[A, List[Char]]),
            term[A, List[Char]]]((a: (term[A, List[Char]], term[A, List[Char]]))
                                   => fst[term[A, List[Char]],
   term[A, List[Char]]](a),
                                  r)
    val isnf: (term[A, List[Char]]) => Boolean =
      is_NF[A, List[Char]](list_ex[term[A,
 List[Char]]]((a: term[A, List[Char]]) => is_Var[A, List[Char]](a), q),
                            q, term_map[A, List[Char]](q))
    val nf: Boolean = nF_subset[A, List[Char]](ls, q)
    val ic: (List[term[A, List[Char]]]) =>
              ((List[Char]) => Boolean) =>
                (term[A, List[Char]]) => term[A, sum[Unit, List[Char]]]
      = (a: List[term[A, List[Char]]]) =>
          (b: (List[Char]) => Boolean) =>
            (c: term[A, List[Char]]) => icap_impl[A](nf, isnf, ls, a, b, c);
    (s: List[term[A, List[Char]]]) =>
      {
        val sa: List[term[A, List[Char]]] =
          mapa[term[A, List[Char]],
                term[A, List[Char]]]((a: term[A, List[Char]]) =>
                                       map_vars[List[Char], List[Char],
         A]((aa: List[Char]) => 'x' :: aa, a),
                                      s)
        val sxs: rbt[List[Char], Unit] =
          list_to_rs[List[Char]].apply(maps[term[A, List[Char]],
     List[Char]](vars_term_list[A, List[Char]], sa))
        val sx: (List[Char]) => Boolean =
          (x: List[Char]) => rs_memb[List[Char]].apply(x).apply(sxs);
        (t: term[A, List[Char]]) =>
          ((ic(sa))(sx))(map_vars[List[Char], List[Char],
                                   A]((a: List[Char]) => 'x' :: a, t))
      }
  }

def lexord_code[A : equal](lt: A => A => Boolean, uu: List[A], x2: List[A]):
      Boolean =
  (lt, uu, x2) match {
  case (lt, uu, Nil) => false
  case (lt, Nil, v :: va) => true
  case (lt, x :: xs, y :: ys) =>
    (lt(x))(y) || eq[A](x, y) && lexord_code[A](lt, xs, ys)
}

def less_lab[A : equal : ord, B : equal : ord](x0: lab[A, B], x1: lab[A, B]):
      Boolean =
  (x0, x1) match {
  case (UnLab(f), UnLab(g)) => less[A](f, g)
  case (UnLab(f), Lab(v, va)) => true
  case (UnLab(f), FunLab(v, va)) => true
  case (UnLab(f), Sharp(v)) => true
  case (Lab(f, la), Lab(g, l)) =>
    less[B](la, l) || eq[B](la, l) && less_lab[A, B](f, g)
  case (Lab(f, l), UnLab(uv)) => false
  case (Lab(f, l), FunLab(v, va)) => true
  case (Lab(f, l), Sharp(v)) => true
  case (Sharp(f), Sharp(g)) => less_lab[A, B](f, g)
  case (Sharp(f), UnLab(ux)) => false
  case (Sharp(f), Lab(uy, uz)) => false
  case (Sharp(f), FunLab(v, vb)) => true
  case (FunLab(f, fs), FunLab(g, gs)) =>
    less_lab[A, B](f, g) ||
      equal_laba[A, B](f, g) &&
        lexord_code[lab[A, B]]((a: lab[A, B]) =>
                                 (b: lab[A, B]) => less_lab[A, B](a, b),
                                fs, gs)
  case (FunLab(f, fs), Lab(v, va)) => false
  case (FunLab(f, fs), UnLab(v)) => false
  case (FunLab(f, fs), Sharp(v)) => false
}

def less_eq_lab[A : equal : ord, B : equal : ord](xs: lab[A, B], ys: lab[A, B]):
      Boolean =
  less_lab[A, B](xs, ys) || equal_laba[A, B](xs, ys)

implicit def ord_lab[A : equal : ord, B : equal : ord]: ord[lab[A, B]] = new
  ord[lab[A, B]] {
  val `Ceta.less_eq` = (a: lab[A, B], b: lab[A, B]) => less_eq_lab[A, B](a, b)
  val `Ceta.less` = (a: lab[A, B], b: lab[A, B]) => less_lab[A, B](a, b)
}

def gea[A, B](x0: poly_c_ext[A, B]): A => A => Boolean = x0 match {
  case Poly_c_ext(wf, zero, one, default, plus, times, max0, ge, gt, mono,
                   description, arc_pos, check_mono, more)
    => ge
}

def wf[A, B](x0: poly_c_ext[A, B]): A => Boolean = x0 match {
  case Poly_c_ext(wf, zero, one, default, plus, times, max0, ge, gt, mono,
                   description, arc_pos, check_mono, more)
    => wf
}

trait ordered_semiring_0[A] extends semiring_0[A] with ordered_ab_semigroup[A] {
}

trait ordered_semiring_1[A] extends semiring_1[A] with ordered_semiring_0[A] {
}

trait max_ordered_semiring_1[A]
  extends max_ordered_monoid_add[A] with ordered_semiring_1[A] {
}

trait poly_carrier[A] extends comm_semiring_1[A] with max_ordered_semiring_1[A]
  {
}

def default_I[A : poly_carrier](defa: A, n: nat): List[(List[(nat, nat)], A)] =
  (Nil, defa) ::
    mapa[nat, (List[(nat, nat)],
                A)]((i: nat) => (List((i, one_nat)), one[A]), upt(Zero_nat, n))

def poly_power[A : equal,
                B : equal : comm_semiring_1](uu: List[(List[(A, nat)], B)],
      x1: nat):
      List[(List[(A, nat)], B)] =
  (uu, x1) match {
  case (uu, Zero_nat) => one_poly[A, B]
  case (p, Suc(n)) => poly_mult[A, B](p, poly_power[A, B](p, n))
}

def monom_subst[A, B : equal,
                 C : equal : comm_semiring_1](sigma:
        A => List[(List[(B, nat)], C)],
       x1: List[(A, nat)]):
      List[(List[(B, nat)], C)] =
  (sigma, x1) match {
  case (sigma, Nil) => one_poly[B, C]
  case (sigma, (x, p) :: m) =>
    poly_mult[B, C](poly_power[B, C](sigma(x), p),
                     monom_subst[A, B, C](sigma, m))
}

def poly_subst[A, B : equal,
                C : equal : comm_semiring_1](sigma:
       A => List[(List[(B, nat)], C)],
      x1: List[(List[(A, nat)], C)]):
      List[(List[(B, nat)], C)] =
  (sigma, x1) match {
  case (sigma, Nil) => zero_poly[B, C]
  case (sigma, (m, c) :: p) =>
    poly_add[B, C](poly_mult[B, C](List((Nil, c)),
                                    monom_subst[A, B, C](sigma, m)),
                    poly_subst[A, B, C](sigma, p))
}

def eval_term[A, B : equal : poly_carrier,
               C : equal](uu: A => nat => List[(List[(nat, nat)], B)],
                           x1: term[A, C]):
      List[(List[(C, nat)], B)] =
  (uu, x1) match {
  case (uu, Var(x)) => List((List((x, one_nat)), one[B]))
  case (i, Fun(f, ts)) =>
    {
      val ps: List[List[(List[(C, nat)], B)]] =
        mapa[term[A, C],
              List[(List[(C, nat)],
                     B)]]((a: term[A, C]) => eval_term[A, B, C](i, a), ts)
      val n: nat = size_list[term[A, C]](ts);
      poly_subst[nat, C,
                  B]((ia: nat) =>
                       (if (less_nat(ia, n))
                         nth[List[(List[(C, nat)], B)]](ps, ia)
                         else zero_poly[C, B]),
                      (i(f))(n))
    }
}

def eval_monom[A, B : comm_semiring_1](alpha: A => B, x1: List[(A, nat)]): B =
  (alpha, x1) match {
  case (alpha, Nil) => one[B]
  case (alpha, (x, p) :: m) =>
    times[B](eval_monom[A, B](alpha, m), power[B].apply(alpha(x)).apply(p))
}

def eval_poly[A, B : comm_semiring_1](alpha: A => B,
                                       x1: List[(List[(A, nat)], B)]):
      B =
  (alpha, x1) match {
  case (alpha, Nil) => zero[B]
  case (alpha, mc :: p) =>
    plus[B](times[B](eval_monom[A, B](alpha, fst[List[(A, nat)], B](mc)),
                      snd[List[(A, nat)], B](mc)),
             eval_poly[A, B](alpha, p))
}

abstract sealed class cr_proof[+A, +B, +C]
final case class
  SN_WCR[A, B,
          C](a: Option[List[(term[lab[A, B], C],
                              (List[(pos, ((term[lab[A, B], C],
     term[lab[A, B], C]),
    term[lab[A, B], C]))],
                                (term[lab[A, B], C],
                                  List[(pos,
 ((term[lab[A, B], C], term[lab[A, B], C]), term[lab[A, B], C]))])))]],
              b: trs_proof[A, B, C])
  extends cr_proof[A, B, C]

def check_supt[A : equal : show,
                B : equal : show](s: term[A, B], t: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  check[(List[Char]) =>
          List[Char]](member[(term[A, B], term[A, B])]((s, t), supt[A, B]),
                       comp[List[Char], List[Char],
                             List[Char]]((shows_prec_term[A,
                   B](Zero_nat)).apply(t),
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List(' ', 'i', 's', ' ', 'n', 'o', 't',
     ' ', 'a', ' ', 'p', 'r', 'o', 'p', 'e', 'r', ' ', 's', 'u', 'b', 't', 'e',
     'r', 'm', ' ', 'o', 'f', ' ')),
                     (shows_prec_term[A, B](Zero_nat)).apply(s))))

def shows_ctxt[A, B](fun: A => (List[Char]) => List[Char],
                      vara: B => (List[Char]) => List[Char], x2: ctxt[A, B]):
      (List[Char]) => List[Char] =
  (fun, vara, x2) match {
  case (fun, vara, Hole) => shows_string.apply(List('[', ']'))
  case (fun, vara, More(f, ss1, d, ss2)) =>
    comp[List[Char], List[Char],
          List[Char]](fun(f),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('(')),
  comp[List[Char], List[Char],
        List[Char]](shows_list_gen[term[A,
 B]]((a: term[A, B]) => shows_term[A, B](fun, vara, a), Nil, Nil,
      List(',', ' '), List(',', ' '), ss1),
                     comp[List[Char], List[Char],
                           List[Char]](shows_ctxt[A, B](fun, vara, d),
shows_list_gen[term[A, B]]((a: term[A, B]) => shows_term[A, B](fun, vara, a),
                            List(')'), List(',', ' '), List(',', ' '),
                            List(')'), ss2)))))
}

def desca[A, B, C](x0: redtriple_ext[A, B, C]): (List[Char]) => List[Char] = x0
  match {
  case Redtriple_ext(valid, s, ns, nst, af, mono, desc, more) => desc
}

def mono[A, B, C](x0: redtriple_ext[A, B, C]):
      sum[(List[Char]) => List[Char], Unit] =
  x0 match {
  case Redtriple_ext(valid, s, ns, nst, af, mono, desc, more) => mono
}

def nsta[A, B, C](x0: root_redtriple_ext[A, B, C]):
      ((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit] =
  x0 match {
  case Root_redtriple_ext(valid, s, ns, nst, af, aft, desc, more) => nst
}

def eq_rule_mod_vars[A : equal,
                      B : equal](lr: (term[A, B], term[A, B]),
                                  st: (term[A, B], term[A, B])):
      Boolean =
  instance_rule[A, B, B](lr, st) && instance_rule[A, B, B](st, lr)

def generate_var(i: nat): List[Char] =
  'x' :: (shows_prec_nat(Zero_nat, i)).apply(Nil)

def uncurry_term[A : equal, B](a: A, sm: A => nat => List[A], t: term[A, B]):
      term[A, B] =
  unapp[A, B](a, t) match {
    case (Var(x), ts) =>
      apply_args[A, B](a, Var[B, A](x),
                        mapa[term[A, B],
                              term[A, B]]((b: term[A, B]) =>
    uncurry_term[A, B](a, sm, b),
   ts))
    case (Fun(f, ss), ts) =>
      {
        val n: nat = size_list[term[A, B]](ss)
        val uss: List[term[A, B]] =
          mapa[term[A, B],
                term[A, B]]((b: term[A, B]) => uncurry_term[A, B](a, sm, b), ss)
        val uts: List[term[A, B]] =
          mapa[term[A, B],
                term[A, B]]((b: term[A, B]) => uncurry_term[A, B](a, sm, b), ts)
        val aa: nat = aarity[A](sm, f, n)
        val m: nat = min[nat].apply(size_list[term[A, B]](ts)).apply(aa)
        val fm: A = get_symbol[A](sm, f, n, m);
        apply_args[A, B](a, Fun[A, B](fm, uss ++ take[term[A, B]](m, uts)),
                          drop[term[A, B]](m, uts))
      }
  }

def parse_name: (List[Char]) => sum[List[Char], (List[Char], List[Char])] =
  (a: List[Char]) =>
    binda[Char, List[Char],
           List[Char]](manyof(letters),
                        (n: List[Char]) =>
                          (aa: List[Char]) =>
                            binda[Char, Unit,
                                   List[Char]](spaces,
        (_: Unit) =>
          (if (nulla[Char](n))
            (_: List[Char]) =>
              Inl[List[Char], (List[Char], List[Char])](letters)
            else (ts: List[Char]) =>
                   Inr[(List[Char], List[Char]), List[Char]]((n, ts))),
        aa),
                        a)

def parse_text:
      (List[Char]) => sum[List[Char], (Option[List[Char]], List[Char])] =
  (a: List[Char]) =>
    binda[Char, List[Char],
           Option[List[Char]]]((aa: List[Char]) =>
                                 many((c: Char) =>
! (equal_lista[Char](List(c), List('<'))),
                                       aa),
                                (ts: List[Char]) =>
                                  {
                                    val text: List[Char] = trim(ts);
                                    (if (nulla[Char](text))
                                      (tsa: List[Char]) =>
Inr[(Option[List[Char]], List[Char]), List[Char]]((None, tsa))
                                      else (tsa: List[Char]) =>
     Inr[(Option[List[Char]], List[Char]),
          List[Char]]((Some[List[Char]](rev[Char](trim(rev[Char](text)))),
                        tsa)))
                                  },
                                a)

def parse_attribute_value:
      (List[Char]) => sum[List[Char], (List[Char], List[Char])] =
  (a: List[Char]) =>
    binda[Char, List[Char],
           List[Char]](oneof(List(dquote)),
                        (_: List[Char]) =>
                          (aa: List[Char]) =>
                            binda[Char, List[Char],
                                   List[Char]]((ab: List[Char]) =>
         many((c: Char) => ! (equal_lista[Char](List(c), dquote)), ab),
        (v: List[Char]) =>
          (ab: List[Char]) =>
            binda[Char, List[Char],
                   List[Char]](oneof(List(dquote)),
                                (_: List[Char]) =>
                                  (ts: List[Char]) =>
                                    Inr[(List[Char], List[Char]),
 List[Char]]((v, ts)),
                                ab),
        aa),
                        a)

def parse_attributes(x0: List[Char]):
      sum[List[Char], (List[(List[Char], List[Char])], List[Char])] =
  x0 match {
  case Nil =>
    Inr[(List[(List[Char], List[Char])], List[Char]), List[Char]]((Nil, Nil))
  case c :: s =>
    (if (membera[Char](List('/', '>'), c))
      Inr[(List[(List[Char], List[Char])], List[Char]),
           List[Char]]((Nil, c :: s))
      else binda[Char, List[Char],
                  List[(List[Char],
                         List[Char])]](parse_name,
(k: List[Char]) =>
  (a: List[Char]) =>
    binda[Char, List[Char],
           List[(List[Char],
                  List[Char])]](oneof(List(List('='))),
                                 (_: List[Char]) =>
                                   (aa: List[Char]) =>
                                     binda[Char, List[Char],
    List[(List[Char],
           List[Char])]](parse_attribute_value,
                          (v: List[Char]) =>
                            (ab: List[Char]) =>
                              binda[Char, List[(List[Char], List[Char])],
                                     List[(List[Char],
    List[Char])]]((ac: List[Char]) => parse_attributes(ac),
                   (atts: List[(List[Char], List[Char])]) =>
                     (ts: List[Char]) =>
                       Inr[(List[(List[Char], List[Char])], List[Char]),
                            List[Char]](((k, v) :: atts, ts)),
                   ab),
                          aa),
                                 a),
c :: s))
}

def parse_nodes(ts: List[Char]): sum[List[Char], (List[xml], List[Char])] =
  (if (nulla[Char](ts) ||
         equal_lista[Char](take[Char](nat(BigInt(2)), ts), List('<', '/')))
    Inr[(List[xml], List[Char]), List[Char]]((Nil, ts))
    else binda[Char, List[Char],
                List[xml]](oneof(List(List('<'))),
                            (_: List[Char]) =>
                              (a: List[Char]) =>
                                binda[Char, List[Char],
                                       List[xml]](parse_name,
           (n: List[Char]) =>
             (aa: List[Char]) =>
               binda[Char, List[(List[Char], List[Char])],
                      List[xml]]((ab: List[Char]) => parse_attributes(ab),
                                  (atts: List[(List[Char], List[Char])]) =>
                                    (ab: List[Char]) =>
                                      binda[Char, List[Char],
     List[xml]](oneof(List(List('/', '>'), List('>'))),
                 (e: List[Char]) =>
                   (tsa: List[Char]) =>
                     (if (equal_lista[Char](e, List('/', '>')))
                       binda[Char, List[xml],
                              List[xml]]((ac: List[Char]) => parse_nodes(ac),
  (cs: List[xml]) =>
    (tsb: List[Char]) =>
      Inr[(List[xml], List[Char]),
           List[Char]]((Xml(n, atts, Nil, None) :: cs, tsb)),
  tsa)
                       else binda[Char, Option[List[Char]],
                                   List[xml]](parse_text,
       (t: Option[List[Char]]) =>
         (ac: List[Char]) =>
           binda[Char, List[xml],
                  List[xml]]((ad: List[Char]) => parse_nodes(ad),
                              (cs: List[xml]) =>
                                (ad: List[Char]) =>
                                  binda[Char, List[Char],
 List[xml]](oneof(List(List('<', '/') ++ (n ++ List('>')))),
             (_: List[Char]) =>
               (ae: List[Char]) =>
                 binda[Char, List[xml],
                        List[xml]]((af: List[Char]) => parse_nodes(af),
                                    (ns: List[xml]) =>
                                      (tsb: List[Char]) =>
Inr[(List[xml], List[Char]), List[Char]]((Xml(n, atts, cs, t) :: ns, tsb)),
                                    ae),
             ad),
                              ac),
       tsa)),
                 ab),
                                  aa),
           a),
                            ts))

def parse_node: (List[Char]) => sum[List[Char], (xml, List[Char])] =
  (a: List[Char]) =>
    binda[Char, List[Char],
           xml](oneof(List(List('<'))),
                 (_: List[Char]) =>
                   (aa: List[Char]) =>
                     binda[Char, List[Char],
                            xml](parse_name,
                                  (n: List[Char]) =>
                                    (ab: List[Char]) =>
                                      binda[Char,
     List[(List[Char], List[Char])],
     xml]((ac: List[Char]) => parse_attributes(ac),
           (atts: List[(List[Char], List[Char])]) =>
             (ac: List[Char]) =>
               binda[Char, List[Char],
                      xml](oneof(List(List('/', '>'), List('>'))),
                            (e: List[Char]) =>
                              (if (equal_lista[Char](e, List('/', '>')))
                                (ts: List[Char]) =>
                                  Inr[(xml, List[Char]),
                                       List[Char]]((Xml(n, atts, Nil, None),
             ts))
                                else (ad: List[Char]) =>
                                       binda[Char, Option[List[Char]],
      xml](parse_text,
            (t: Option[List[Char]]) =>
              (ae: List[Char]) =>
                binda[Char, List[xml],
                       xml]((af: List[Char]) => parse_nodes(af),
                             (cs: List[xml]) =>
                               (af: List[Char]) =>
                                 binda[Char, List[Char],
xml](oneof(List(List('<', '/') ++ (n ++ List('>')))),
      (_: List[Char]) =>
        (ts: List[Char]) =>
          Inr[(xml, List[Char]), List[Char]]((Xml(n, atts, cs, t), ts)),
      af),
                             ae),
            ad)),
                            ac),
           ab),
                                  aa),
                 a)

def scan_upto(end: List[Char], x1: List[Char]):
      sum[List[Char], (List[Char], List[Char])] =
  (end, x1) match {
  case (end, Nil) =>
    Inl[List[Char],
         (List[Char],
           List[Char])](List('d', 'i', 'd', ' ', 'n', 'o', 't', ' ', 'f', 'i',
                              'n', 'd', ' ', 'e', 'n', 'd', '-', 'm', 'a', 'r',
                              'k', 'e', 'r', ' ') ++
                          (quote(shows_prec_list[Char](Zero_nat,
                end))).apply(Nil))
  case (end, t :: ts) =>
    (if (equal_lista[Char](mapa[(Char, Char),
                                 Char]((a: (Char, Char)) => snd[Char, Char](a),
zip[Char, Char](end, t :: ts)),
                            end))
      Inr[(List[Char], List[Char]),
           List[Char]]((end, drop[Char](size_list[Char](end), t :: ts)))
      else bindb[List[Char], (List[Char], List[Char]),
                  (List[Char],
                    List[Char])](scan_upto(end, ts),
                                  (a: (List[Char], List[Char])) =>
                                    {
                                      val (res, tsa): (List[Char], List[Char]) =
a;
                                      Inr[(List[Char], List[Char]),
   List[Char]]((t :: res, tsa))
                                    }))
}

def parse_header(ts: List[Char]): sum[List[Char], (List[Char], List[Char])] =
  (if (equal_lista[Char](take[Char](nat(BigInt(2)), trim(ts)), List('<', '?')))
    binda[Char, List[Char],
           List[Char]]((a: List[Char]) => scan_upto(List('?', '>'), a),
                        (h: List[Char]) =>
                          (a: List[Char]) =>
                            binda[Char, List[Char],
                                   List[Char]]((aa: List[Char]) =>
         parse_header(aa),
        (hs: List[Char]) =>
          (tsa: List[Char]) =>
            Inr[(List[Char], List[Char]), List[Char]]((h ++ hs, tsa)),
        a),
                        ts)
    else binda[Char, Unit,
                List[Char]](spaces,
                             (_: Unit) =>
                               (tsa: List[Char]) =>
                                 Inr[(List[Char], List[Char]),
                                      List[Char]]((Nil, tsa)),
                             ts))

def update_tokens[A](f: (List[A]) => List[A], ts: List[A]):
      sum[List[Char], (List[A], List[A])] =
  Inr[(List[A], List[A]), List[Char]]((ts, f(ts)))

def remove_comments_aux(uu: Boolean, x1: List[Char]): List[Char] = (uu, x1)
  match {
  case (false, c :: cs) =>
    (if (c == '<' &&
           equal_lista[Char](take[Char](nat(BigInt(3)), cs),
                              List('!', '-', '-')))
      remove_comments_aux(true, tl[Char](cs))
      else c :: remove_comments_aux(false, cs))
  case (true, c :: cs) =>
    (if (c == '-' &&
           equal_lista[Char](take[Char](nat(BigInt(2)), cs), List('-', '>')))
      remove_comments_aux(false, drop[Char](nat(BigInt(2)), cs))
      else remove_comments_aux(true, cs))
  case (uu, Nil) => Nil
}

def remove_comments_aux_acc(a: List[Char], uu: Boolean, x2: List[Char]):
      List[Char] =
  (a, uu, x2) match {
  case (a, false, c :: cs) =>
    (if (c == '<' &&
           equal_lista[Char](take[Char](nat(BigInt(3)), cs),
                              List('!', '-', '-')))
      remove_comments_aux_acc(a, true, tl[Char](cs))
      else remove_comments_aux_acc(c :: a, false, cs))
  case (a, true, c :: cs) =>
    (if (c == '-' &&
           equal_lista[Char](take[Char](nat(BigInt(2)), cs), List('-', '>')))
      remove_comments_aux_acc(a, false, drop[Char](nat(BigInt(2)), cs))
      else remove_comments_aux_acc(a, true, cs))
  case (a, uu, Nil) => a
}

def remove_comments: (List[Char]) => List[Char] =
  (if (true)
    comp[List[Char], List[Char],
          List[Char]]((a: List[Char]) => rev[Char](a),
                       (a: List[Char]) =>
                         remove_comments_aux_acc(Nil, false, a))
    else (a: List[Char]) => remove_comments_aux(false, a))

def parse_doc: (List[Char]) => sum[List[Char], (xmldoc, List[Char])] =
  debug[(List[Char]) =>
          sum[List[Char],
               (xmldoc,
                 List[Char])]](List('0'),
                                List('p', 'a', 'r', 's', 'i', 'n', 'g', ' ',
                                      't', 'x', 't', ' ', 't', 'o', ' ', 'x',
                                      'm', 'l'),
                                (a: List[Char]) =>
                                  binda[Char, List[Char],
 xmldoc]((aa: List[Char]) => update_tokens[Char](remove_comments, aa),
          (_: List[Char]) =>
            (aa: List[Char]) =>
              binda[Char, List[Char],
                     xmldoc]((ab: List[Char]) => parse_header(ab),
                              (h: List[Char]) =>
                                (ab: List[Char]) =>
                                  binda[Char, xml,
 xmldoc](parse_node,
          (xml: xml) =>
            (ac: List[Char]) =>
              binda[Char, Unit,
                     xmldoc]((ad: List[Char]) => eoi[Char](ad),
                              (_: Unit) =>
                                (ts: List[Char]) =>
                                  Inr[(xmldoc, List[Char]),
                                       List[Char]]((Xmldoc(h, xml), ts)),
                              ac),
          ab),
                              aa),
          a))

abstract sealed class filtered[+A]
final case class FPair[A](a: A, b: nat) extends filtered[A]

def conversion[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                          (Boolean, term[A, List[Char]])))]] =
  (xml1many_elements[term[A, List[Char]],
                      (pos, ((term[A, List[Char]], term[A, List[Char]]),
                              (Boolean, term[A, List[Char]]))),
                      List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                   (Boolean,
                                     term[A,
   List[Char]])))]](List('c', 'o', 'n', 'v', 'e', 'r', 's', 'i', 'o', 'n'),
                     (a: xml) =>
                       xml1element[term[A, List[Char]],
                                    term[A,
  List[Char]]](List('s', 't', 'a', 'r', 't', 'T', 'e', 'r', 'm'),
                (aa: xml) => term[A](xml2name, aa), id[term[A, List[Char]]], a),
                     estep[A](xml2name))).apply((_: term[A, List[Char]]) =>
          (x: List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                           (Boolean, term[A, List[Char]])))])
            => x)

def plain_name(x: xml): sum[List[Char], lab[List[Char], List[nat]]] =
  xml_options[lab[List[Char],
                   List[nat]]](List((List('n', 'a', 'm', 'e'),
                                      (a: xml) =>
xml_change[List[Char],
            lab[List[Char],
                 List[nat]]]((aa: xml) => text(List('n', 'a', 'm', 'e'), aa),
                              (aa: List[Char]) =>
                                UnLab[List[Char], List[nat]](aa),
                              a)),
                                     (List('s', 'h', 'a', 'r', 'p'),
                                       (a: xml) =>
 xml1element[lab[List[Char], List[nat]],
              lab[List[Char],
                   List[nat]]](List('s', 'h', 'a', 'r', 'p'),
                                (aa: xml) => plain_name(aa),
                                (aa: lab[List[Char], List[nat]]) =>
                                  Sharp[List[Char], List[nat]](aa),
                                a)),
                                     (List('l', 'a', 'b', 'e', 'l', 'e', 'd',
    'S', 'y', 'm', 'b', 'o', 'l'),
                                       (a: xml) =>
 xml2elements[lab[List[Char], List[nat]],
               sum[List[nat], List[lab[List[Char], List[nat]]]],
               lab[List[Char],
                    List[nat]]](List('l', 'a', 'b', 'e', 'l', 'e', 'd', 'S',
                                      'y', 'm', 'b', 'o', 'l'),
                                 (aa: xml) => plain_name(aa),
                                 labela[lab[List[Char],
     List[nat]]]((aa: xml) => plain_name(aa)),
                                 (f: lab[List[Char], List[nat]]) =>
                                   (aa: sum[List[nat],
     List[lab[List[Char], List[nat]]]])
                                     => aa match {
  case Inl(ab) => Lab[List[Char], List[nat]](f, ab)
  case Inr(ab) => FunLab[List[Char], List[nat]](f, ab)
},
                                 a))),
                                x)

def xml1or2many_elements[A, B, C,
                          D](tag: List[Char], p1: xml => sum[List[Char], A],
                              p2: xml => sum[List[Char], B],
                              p3: xml => sum[List[Char], C],
                              f: A => Option[B] => (List[C]) => D, x5: xml):
      sum[List[Char], D] =
  (tag, p1, p2, p3, f, x5) match {
  case (tag, p1, p2, p3, f, Xml(name, atts, cs, text)) =>
    (if (equal_lista[Char](name, tag) &&
           (nulla[(List[Char], List[Char])](atts) &&
             (is_none[List[Char]](text) &&
               less_eq_nat(nat(BigInt(2)), size_list[xml](cs)))))
      bindb[List[Char], A,
             D](p1(nth[xml](cs, Zero_nat)),
                 (x: A) =>
                   catcha[List[Char], D,
                           List[Char]](bindb[List[Char], B,
      D](p2(nth[xml](cs, one_nat)),
          (y: B) =>
            bindb[List[Char], List[C],
                   D](map[C].apply(p3).apply(tl[xml](tl[xml](cs))),
                       (xs: List[C]) =>
                         Inr[D, List[Char]](((f(x))(Some[B](y)))(xs)))),
(_: List[Char]) =>
  bindb[List[Char], List[C],
         D](map[C].apply(p3).apply(tl[xml](cs)),
             (xs: List[C]) => Inr[D, List[Char]](((f(x))(None))(xs)))))
      else no_transform[D](tag, Xml(name, atts, cs, text)))
}

def sl_variant[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], sl_variant[A, List[Char]]] =
  (a: xml) =>
    xml1element[sl_variant[A, List[Char]],
                 sl_variant[A, List[Char]]](List('m', 'o', 'd', 'e', 'l'),
     (aa: xml) =>
       xml_options[sl_variant[A, List[Char]]](List((List('f', 'i', 'n', 'i',
                  't', 'e', 'M', 'o', 'd', 'e', 'l'),
             (ab: xml) =>
               xml1or2many_elements[nat, Unit, ((A, nat), arithFun),
                                     sl_variant[A,
         List[Char]]](List('f', 'i', 'n', 'i', 't', 'e', 'M', 'o', 'd', 'e',
                            'l'),
                       (ac: xml) =>
                         xml_change[nat, nat]((ad: xml) =>
        nata(List('c', 'a', 'r', 'r', 'i', 'e', 'r', 'S', 'i', 'z', 'e'), ad),
       (n: nat) => minus_nat(n, one_nat), ac),
                       (ac: xml) =>
                         xml1element[Unit,
                                      Unit](List('t', 'u', 'p', 'l', 'e', 'O',
          'r', 'd', 'e', 'r'),
     (ad: xml) =>
       xml0elements[Unit](List('p', 'o', 'i', 'n', 't', 'W', 'i', 's', 'e'), (),
                           ad),
     id[Unit], ac),
                       (ac: xml) =>
                         xml3elements[A, nat, arithFun,
                                       ((A, nat),
 arithFun)](List('i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'), xml2name,
             (ad: xml) => nata(List('a', 'r', 'i', 't', 'y'), ad),
             (ad: xml) => arith_fun(ad),
             (f: A) => (ad: nat) => (b: arithFun) => ((f, ad), b), ac),
                       (n: nat) =>
                         (model: Option[Unit]) =>
                           (xs: List[((A, nat), arithFun)]) =>
                             (if (is_none[Unit](model))
                               Finitelab[A, List[Char]](SL_Inter[A](n, xs))
                               else QuasiFinitelab[A,
            List[Char]](SL_Inter[A](n, xs), Nil)),
                       ab)),
            (List('r', 'o', 'o', 't', 'L', 'a', 'b', 'e', 'l', 'i', 'n', 'g'),
              (ab: xml) =>
                xml_choice[sl_variant[A, List[Char]]](List('r', 'o', 'o', 't',
                    'L', 'a', 'b', 'e', 'l', 'i', 'n', 'g'),
               List((ac: xml) =>
                      xml0elements[sl_variant[A,
       List[Char]]](List('r', 'o', 'o', 't', 'L', 'a', 'b', 'e', 'l', 'i', 'n',
                          'g'),
                     Rootlab[A, List[Char]](None), ac),
                     (ac: xml) =>
                       xml1element[A, sl_variant[A,
          List[Char]]](List('r', 'o', 'o', 't', 'L', 'a', 'b', 'e', 'l', 'i',
                             'n', 'g'),
                        xml2name,
                        (f: A) =>
                          Rootlab[A, List[Char]](Some[(A, nat)]((f, one_nat))),
                        ac)),
               ab))),
       aa),
     id[sl_variant[A, List[Char]]], a)

def xml2many_elements[A, B, C,
                       D](tag: List[Char], p1: xml => sum[List[Char], A],
                           p2: xml => sum[List[Char], B],
                           p3: xml => sum[List[Char], C],
                           f: A => B => (List[C]) => D, x5: xml):
      sum[List[Char], D] =
  (tag, p1, p2, p3, f, x5) match {
  case (tag, p1, p2, p3, f, Xml(name, atts, cs, text)) =>
    (if (equal_lista[Char](name, tag) &&
           (nulla[(List[Char], List[Char])](atts) &&
             (is_none[List[Char]](text) &&
               less_eq_nat(nat(BigInt(2)), size_list[xml](cs)))))
      bindb[List[Char], A,
             D](p1(nth[xml](cs, Zero_nat)),
                 (x: A) =>
                   bindb[List[Char], B,
                          D](p2(nth[xml](cs, one_nat)),
                              (y: B) =>
                                bindb[List[Char], List[C],
                                       D](map[C].apply(p3).apply(tl[xml](tl[xml](cs))),
   (xs: List[C]) => Inr[D, List[Char]](((f(x))(y))(xs)))))
      else no_transform[D](tag, Xml(name, atts, cs, text)))
}

def transition_lhs[A](xml2name: xml => sum[List[Char], A], x: xml):
      sum[List[Char], sum[((A, nat), List[List[Char]]), List[Char]]] =
  (if ({
         val (Xml(_, _, cs, _)): xml = x;
         equal_nata(size_list[xml](cs), one_nat)
       })
    xml1element[List[Char],
                 sum[((A, nat), List[List[Char]]),
                      List[Char]]](List('l', 'h', 's'), state,
                                    (a: List[Char]) =>
                                      Inr[List[Char],
   ((A, nat), List[List[Char]])](a),
                                    x)
    else xml2many_elements[A, nat, List[Char],
                            sum[((A, nat), List[List[Char]]),
                                 List[Char]]](List('l', 'h', 's'), xml2name,
       (a: xml) => nata(List('h', 'e', 'i', 'g', 'h', 't'), a), state,
       (f: A) =>
         (h: nat) =>
           (qs: List[List[Char]]) =>
             Inl[((A, nat), List[List[Char]]), List[Char]](((f, h), qs)),
       x))

def transition[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             sum[ta_rule[List[Char], (A, nat)], (List[Char], List[Char])]] =
  (a: xml) =>
    xml2elements[sum[((A, nat), List[List[Char]]), List[Char]], List[Char],
                  sum[ta_rule[List[Char], (A, nat)],
                       (List[Char],
                         List[Char])]](List('t', 'r', 'a', 'n', 's', 'i', 't',
     'i', 'o', 'n'),
(aa: xml) => transition_lhs[A](xml2name, aa),
(aa: xml) =>
  xml1element[List[Char],
               List[Char]](List('r', 'h', 's'), state, id[List[Char]], aa),
(lhs: sum[((A, nat), List[List[Char]]), List[Char]]) =>
  (q: List[Char]) =>
    lhs match {
      case Inl((f, qs)) =>
        Inl[ta_rule[List[Char], (A, nat)],
             (List[Char], List[Char])](TA_rule[(A, nat), List[Char]](f, qs, q))
      case Inr(qa) =>
        Inr[(List[Char], List[Char]), ta_rule[List[Char], (A, nat)]]((qa, q))
    },
a)

def less_term[A : equal : ord, B : equal : ord](x0: term[A, B], x1: term[A, B]):
      Boolean =
  (x0, x1) match {
  case (Var(f), Var(g)) => less[B](f, g)
  case (Var(f), Fun(v, va)) => true
  case (Fun(f, fs), Fun(g, gs)) =>
    less[A](f, g) ||
      eq[A](f, g) &&
        lexord_code[term[A, B]]((a: term[A, B]) =>
                                  (b: term[A, B]) => less_term[A, B](a, b),
                                 fs, gs)
  case (Fun(f, fs), Var(v)) => false
}

def less_eq_term[A : equal : ord,
                  B : equal : ord](xs: term[A, B], ys: term[A, B]):
      Boolean =
  less_term[A, B](xs, ys) || equal_terma[A, B](xs, ys)

implicit def ord_term[A : equal : ord, B : equal : ord]: ord[term[A, B]] = new
  ord[term[A, B]] {
  val `Ceta.less_eq` = (a: term[A, B], b: term[A, B]) =>
    less_eq_term[A, B](a, b)
  val `Ceta.less` = (a: term[A, B], b: term[A, B]) => less_term[A, B](a, b)
}

def all_interval_nat(p: nat => Boolean, i: nat, j: nat): Boolean =
  less_eq_nat(j, i) || p(i) && all_interval_nat(p, Suc(i), j)

def gen_list_to_map_aux[A, B,
                         C](upd: A => B => C => C, accs: C, x2: List[(A, B)]):
      C =
  (upd, accs, x2) match {
  case (upd, accs, Nil) => accs
  case (upd, accs, (k, v) :: l) =>
    gen_list_to_map_aux[A, B, C](upd, ((upd(k))(v))(accs), l)
}

def gen_list_to_map[A, B, C](empt: A, upd: B => C => A => A, l: List[(B, C)]):
      A =
  gen_list_to_map_aux[B, C, A](upd, empt, rev[(B, C)](l))

def delete_value[A : equal,
                  B : linorder](key: A => Option[B], v: A, m: rbt[B, List[A]]):
      rbt[B, List[A]] =
  key(v) match {
    case None => m
    case Some(k) =>
      rm_lookup[B, List[A]](k, m) match {
        case None => m
        case Some(vs) =>
          rm_update[B, List[A]].apply(k).apply(removeAll[A](v, vs)).apply(m)
      }
  }

def insert_value[A : equal,
                  B : linorder](key: A => Option[B], v: A, m: rbt[B, List[A]]):
      rbt[B, List[A]] =
  key(v) match {
    case None => m
    case Some(k) =>
      rm_lookup[B, List[A]](k, m) match {
        case None => rm_update_dj[B, List[A]].apply(k).apply(List(v)).apply(m)
        case Some(vs) =>
          rm_update[B, List[A]].apply(k).apply(insertb[A](v, vs)).apply(m)
      }
  }

def shows_monom[A : show](x0: List[(A, nat)]): (List[Char]) => List[Char] = x0
  match {
  case List((x, p)) =>
    (if (equal_nata(p, one_nat))
      (a: List[Char]) => shows_prec[A](Zero_nat, x, a)
      else comp[List[Char], List[Char],
                 List[Char]]((a: List[Char]) => shows_prec[A](Zero_nat, x, a),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_string.apply(List('^')),
         shows_prec_nat(Zero_nat, p))))
  case (x, p) :: v :: va =>
    comp[List[Char], List[Char],
          List[Char]]((if (equal_nata(p, one_nat))
                        (a: List[Char]) => shows_prec[A](Zero_nat, x, a)
                        else comp[List[Char], List[Char],
                                   List[Char]]((a: List[Char]) =>
         shows_prec[A](Zero_nat, x, a),
        comp[List[Char], List[Char],
              List[Char]](shows_string.apply(List('^')),
                           shows_prec_nat(Zero_nat, p)))),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('*')),
  shows_monom[A](v :: va)))
  case Nil => shows_string.apply(List('1'))
}

def shows_poly[A : show,
                B : equal : poly_carrier : show](x0: List[(List[(A, nat)], B)]):
      (List[Char]) => List[Char] =
  x0 match {
  case Nil => shows_string.apply(List('0'))
  case (m, c) :: p =>
    comp[List[Char], List[Char],
          List[Char]]((if (eq[B](c, one[B])) shows_monom[A](m)
                        else (if (nulla[(A, nat)](m))
                               (a: List[Char]) => shows_prec[B](Zero_nat, c, a)
                               else comp[List[Char], List[Char],
  List[Char]]((a: List[Char]) => shows_prec[B](Zero_nat, c, a),
               comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List('*')),
                                  shows_monom[A](m))))),
                       (if (nulla[(List[(A, nat)], B)](p))
                         shows_string.apply(Nil)
                         else comp[List[Char], List[Char],
                                    List[Char]](shows_string.apply(List(' ',
                                 '+', ' ')),
         shows_poly[A, B](p))))
}

def poly_split[A : equal,
                B : zero](m: List[(A, nat)], p: List[(List[(A, nat)], B)]):
      (B, List[(List[(A, nat)], B)]) =
  extract[(List[(A, nat)],
            B)]((a: (List[(A, nat)], B)) =>
                  {
                    val (n, _): (List[(A, nat)], B) = a;
                    eq_monom[A](m, n)
                  },
                 p)
    match {
    case None => (zero[B], p)
    case Some((p1, ((_, c), p2))) => (c, p1 ++ p2)
  }

def missing(s: List[Char], x: (List[Char]) => List[Char]):
      (List[Char]) => List[Char] =
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List('t', 'h', 'e', ' ')),
                     comp[List[Char], List[Char],
                           List[Char]](shows_string.apply(s),
comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List(' ')),
                   comp[List[Char], List[Char],
                         List[Char]](x, shows_string.apply(List(' ', 'i', 's',
                         ' ', 'm', 'i', 's', 's', 'i', 'n', 'g'))))))

def toomuch(s: List[Char], x: (List[Char]) => List[Char]):
      (List[Char]) => List[Char] =
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List('s', 'u', 'p', 'e', 'r', 'f', 'l',
     'u', 'o', 'u', 's', ' ')),
                     comp[List[Char], List[Char],
                           List[Char]](shows_string.apply(s),
comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List(' ')),
                   comp[List[Char], List[Char], List[Char]](x, x))))

abstract sealed class completion_proof[+A, +B, +C]
final case class
  SN_WCR_Eq[A, B,
             C](a: Option[List[(term[lab[A, B], C],
                                 (List[(pos,
 ((term[lab[A, B], C], term[lab[A, B], C]), term[lab[A, B], C]))],
                                   (term[lab[A, B], C],
                                     List[(pos,
    ((term[lab[A, B], C], term[lab[A, B], C]), term[lab[A, B], C]))])))]],
                 b: trs_proof[A, B, C],
                 c: List[((term[lab[A, B], C], term[lab[A, B], C]),
                           List[(pos, ((term[lab[A, B], C], term[lab[A, B], C]),
(Boolean, term[lab[A, B], C])))])],
                 d: Option[List[((term[lab[A, B], C], term[lab[A, B], C]),
                                  List[(pos,
 ((term[lab[A, B], C], term[lab[A, B], C]), (Boolean, term[lab[A, B], C])))])]])
  extends completion_proof[A, B, C]

abstract sealed class dp_disproof[+A, +B]
final case class
  DP_Loop[A, B](a: term[A, B],
                 b: List[(pos, ((term[A, B], term[A, B]),
                                 (Boolean, term[A, B])))],
                 c: List[(B, term[A, B])], d: ctxt[A, B])
  extends dp_disproof[A, B]
final case class
  DP_Rule_Removal[A, B](a: List[(term[A, B], term[A, B])],
                         b: List[(term[A, B], term[A, B])],
                         c: dp_disproof[A, B])
  extends dp_disproof[A, B]

abstract sealed class trs_disproof[+A, +B]
final case class
  TRS_Loop[A, B](a: term[A, B],
                  b: List[(pos, ((term[A, B], term[A, B]), term[A, B]))],
                  c: List[(B, term[A, B])], d: ctxt[A, B])
  extends trs_disproof[A, B]
final case object TRS_Not_Well_Formed extends trs_disproof[Nothing, Nothing]
final case class
  TRS_Rule_Removal[A, B](a: List[(term[A, B], term[A, B])],
                          b: trs_disproof[A, B])
  extends trs_disproof[A, B]
final case class TRS_String_Reversal[A, B](a: trs_disproof[A, B]) extends
  trs_disproof[A, B]
final case class
  TRS_DP_Trans[A, B](a: List[(term[A, B], term[A, B])], b: dp_disproof[A, B])
  extends trs_disproof[A, B]

abstract sealed class reltrs_disproof[+A, +B]
final case class
  Rel_Loop[A, B](a: term[A, B],
                  b: List[(pos, ((term[A, B], term[A, B]),
                                  (Boolean, term[A, B])))],
                  c: List[(B, term[A, B])], d: ctxt[A, B])
  extends reltrs_disproof[A, B]
final case object Rel_Not_Well_Formed extends reltrs_disproof[Nothing, Nothing]
final case class
  Rel_Rule_Removal[A, B](a: List[(term[A, B], term[A, B])],
                          b: List[(term[A, B], term[A, B])],
                          c: reltrs_disproof[A, B])
  extends reltrs_disproof[A, B]
final case class Rel_R_Not_SN[A, B](a: trs_disproof[A, B]) extends
  reltrs_disproof[A, B]

abstract sealed class cert_problem[+A, +B, +C]
final case class
  TRS_Proof[A, B,
             C](a: strategy[lab[A, B], C],
                 b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                 c: Option[List[(term[lab[A, B], C], term[lab[A, B], C])]],
                 d: trs_proof[A, B, C])
  extends cert_problem[A, B, C]
final case class
  DP_Proof[A, B,
            C](a: List[(term[lab[A, B], C], term[lab[A, B], C])],
                b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                c: strategy[lab[A, B], C],
                d: List[(term[lab[A, B], C], term[lab[A, B], C])],
                e: List[(term[lab[A, B], C], term[lab[A, B], C])],
                f: dp_proof[A, B, C])
  extends cert_problem[A, B, C]
final case class
  DP_Disproof[A, B,
               C](a: List[(term[lab[A, B], C], term[lab[A, B], C])],
                   b: strategy[lab[A, B], C],
                   c: List[(term[lab[A, B], C], term[lab[A, B], C])],
                   d: dp_disproof[lab[A, B], C])
  extends cert_problem[A, B, C]
final case class
  TRS_Disproof[A, B,
                C](a: strategy[lab[A, B], C],
                    b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                    c: trs_disproof[lab[A, B], C])
  extends cert_problem[A, B, C]
final case class
  Rel_TRS_Disproof[A, B,
                    C](a: strategy[lab[A, B], C],
                        b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                        c: List[(term[lab[A, B], C], term[lab[A, B], C])],
                        d: reltrs_disproof[lab[A, B], C])
  extends cert_problem[A, B, C]
final case class
  TRS_Confluence[A, B,
                  C](a: List[(term[lab[A, B], C], term[lab[A, B], C])],
                      b: cr_proof[A, B, C])
  extends cert_problem[A, B, C]
final case class
  Completion_Proof[A, B,
                    C](a: List[(term[lab[A, B], C], term[lab[A, B], C])],
                        b: List[(term[lab[A, B], C], term[lab[A, B], C])],
                        c: completion_proof[A, B, C])
  extends cert_problem[A, B, C]

def xml_tag[A, B, C](x0: cert_problem[A, B, C]): List[Char] = x0 match {
  case TRS_Proof(uu, uv, Some(uw), ux) =>
    List('<', 'r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'T', 'e', 'r', 'm', 'i',
          'n', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f', '>')
  case TRS_Proof(uy, uz, None, va) =>
    List('<', 't', 'r', 's', 'T', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o',
          'n', 'P', 'r', 'o', 'o', 'f', '>')
  case TRS_Disproof(vb, vc, vd) =>
    List('<', 't', 'r', 's', 'N', 'o', 'n', 't', 'e', 'r', 'm', 'i', 'n', 'a',
          't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f', '>')
  case Rel_TRS_Disproof(ve, vf, vg, vh) =>
    List('<', 'r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'N', 'o', 'n', 't', 'e',
          'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f',
          '>')
  case DP_Proof(vi, vj, vk, vl, vm, vn) =>
    List('<', 'd', 'p', 'P', 'r', 'o', 'o', 'f', '>')
  case DP_Disproof(vo, vp, vq, vr) =>
    List('<', 'd', 'p', 'N', 'o', 'n', 't', 'e', 'r', 'm', 'i', 'n', 'a', 't',
          'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f', '>')
  case TRS_Confluence(vs, vt) =>
    List('<', 'c', 'r', 'P', 'r', 'o', 'o', 'f', '>')
  case Completion_Proof(vu, vv, vw) =>
    List('<', 'c', 'o', 'm', 'p', 'l', 'e', 't', 'i', 'o', 'n', 'P', 'r', 'o',
          'o', 'f', '>')
}

def list_to_rm[A : linorder, B]: (List[(A, B)]) => rbt[A, B] =
  (a: List[(A, B)]) =>
    gen_list_to_map[rbt[A, B], A, B](rm_empty[A, B], rm_update[A, B], a)

def rm_isEmpty[A : equal : linorder, B : equal](m: rbt[A, B]): Boolean =
  eq[rbt[A, B]](m, empty[A, B])

def equal_unita(u: Unit, v: Unit): Boolean = true

implicit def equal_unit: equal[Unit] = new equal[Unit] {
  val `Ceta.equal` = (a: Unit, b: Unit) => equal_unita(a, b)
}

def rs_isEmpty[A : equal : linorder]: (rbt[A, Unit]) => Boolean =
  s_isEmpty[(rbt[A, Unit]) =>
              Boolean]((a: rbt[A, Unit]) => rm_isEmpty[A, Unit](a))

def rito_set_to_sorted_list[A, B](iterate:
                                    (A => (List[A]) => List[A]) =>
                                      B => (List[A]) => List[A],
                                   m: B):
      List[A] =
  ((iterate((a: A) => (b: List[A]) => a :: b))(m))(Nil)

def s_reverse_iterateo[A : linorder, B,
                        C](iterate: (A => Unit => B => B) => C => B => B,
                            f: A => B => B, s: C, sigma_0: B):
      B =
  ((iterate((k: A) => (_: Unit) => f(k)))(s))(sigma_0)

def rs_reverse_iterateo[A : linorder, B]:
      (A => B => B) => (rbt[A, Unit]) => B => B =
  (a: A => B => B) =>
    (b: rbt[A, Unit]) =>
      (c: B) =>
        s_reverse_iterateo[A, B,
                            rbt[A, Unit]](rm_reverse_iterateo[A, Unit, B], a, b,
   c)

def rs_to_list[A : linorder]: (rbt[A, Unit]) => List[A] =
  (a: rbt[A, Unit]) =>
    rito_set_to_sorted_list[A, rbt[A, Unit]](rs_reverse_iterateo[A, List[A]], a)

def delete_values[A : equal,
                   B : linorder](uu: A => Option[B], x1: List[A],
                                  m: rbt[B, List[A]]):
      rbt[B, List[A]] =
  (uu, x1, m) match {
  case (uu, Nil, m) => m
  case (key, v :: vs, m) =>
    delete_value[A, B](key, v, delete_values[A, B](key, vs, m))
}

def delete_rules[A : equal, B : equal : linorder,
                  C : equal](a: A, rs: List[(term[B, C], term[B, C])]):
      (rbt[(B, nat), List[(A, (term[B, C], term[B, C]))]]) =>
        rbt[(B, nat), List[(A, (term[B, C], term[B, C]))]] =
  (b: rbt[(B, nat), List[(A, (term[B, C], term[B, C]))]]) =>
    delete_values[(A, (term[B, C], term[B, C])),
                   (B, nat)]((aa: (A, (term[B, C], term[B, C]))) =>
                               key[A, B, C](aa),
                              mapa[(term[B, C], term[B, C]),
                                    (A, (term[B, C],
  term[B, C]))]((ba: (term[B, C], term[B, C])) => (a, ba), rs),
                              b)

def insert_values[A : equal,
                   B : linorder](uu: A => Option[B], x1: List[A],
                                  m: rbt[B, List[A]]):
      rbt[B, List[A]] =
  (uu, x1, m) match {
  case (uu, Nil, m) => m
  case (key, v :: vs, m) =>
    insert_value[A, B](key, v, insert_values[A, B](key, vs, m))
}

def insert_rules[A : equal, B : equal : linorder,
                  C : equal](a: A, rs: List[(term[B, C], term[B, C])]):
      (rbt[(B, nat), List[(A, (term[B, C], term[B, C]))]]) =>
        rbt[(B, nat), List[(A, (term[B, C], term[B, C]))]] =
  (b: rbt[(B, nat), List[(A, (term[B, C], term[B, C]))]]) =>
    insert_values[(A, (term[B, C], term[B, C])),
                   (B, nat)]((aa: (A, (term[B, C], term[B, C]))) =>
                               key[A, B, C](aa),
                              mapa[(term[B, C], term[B, C]),
                                    (A, (term[B, C],
  term[B, C]))]((ba: (term[B, C], term[B, C])) => (a, ba), rs),
                              b)

def qc[A, B, C, D](x0: dpp_ops_ext[A, B, C, D]): A => List[term[B, C]] = x0
  match {
  case Dpp_ops_ext(dpp, p, pw, pairs, q, r, rw, rules, q_empty,
                    rules_no_left_var, rules_non_collapsing, is_QNF,
                    nFQ_subset_NF_rules, rules_map, reverse_rules_map,
                    intersect_pairs, replace_pair, intersect_rules, delete_P_Pw,
                    delete_R_Rw, split_pairs, split_rules, mk, more)
    => q
}

def mkc[A, B, C, D](x0: dpp_ops_ext[A, B, C, D]):
      (List[(term[B, C], term[B, C])]) =>
        (List[(term[B, C], term[B, C])]) =>
          (List[term[B, C]]) =>
            (List[(term[B, C], term[B, C])]) =>
              (List[(term[B, C], term[B, C])]) => A =
  x0 match {
  case Dpp_ops_ext(dpp, p, pw, pairs, q, r, rw, rules, q_empty,
                    rules_no_left_var, rules_non_collapsing, is_QNF,
                    nFQ_subset_NF_rules, rules_map, reverse_rules_map,
                    intersect_pairs, replace_pair, intersect_rules, delete_P_Pw,
                    delete_R_Rw, split_pairs, split_rules, mk, more)
    => mk
}

def delete_P_Pwa[A, B, C, D](x0: dpp_ops_ext[A, B, C, D]):
      A => (List[(term[B, C], term[B, C])]) =>
             (List[(term[B, C], term[B, C])]) => A =
  x0 match {
  case Dpp_ops_ext(dpp, p, pw, pairs, q, r, rw, rules, q_empty,
                    rules_no_left_var, rules_non_collapsing, is_QNF,
                    nFQ_subset_NF_rules, rules_map, reverse_rules_map,
                    intersect_pairs, replace_pair, intersect_rules, delete_P_Pw,
                    delete_R_Rw, split_pairs, split_rules, mk, more)
    => delete_P_Pw
}

def delete_R_Rwc[A, B, C, D](x0: dpp_ops_ext[A, B, C, D]):
      A => (List[(term[B, C], term[B, C])]) =>
             (List[(term[B, C], term[B, C])]) => A =
  x0 match {
  case Dpp_ops_ext(dpp, p, pw, pairs, q, r, rw, rules, q_empty,
                    rules_no_left_var, rules_non_collapsing, is_QNF,
                    nFQ_subset_NF_rules, rules_map, reverse_rules_map,
                    intersect_pairs, replace_pair, intersect_rules, delete_P_Pw,
                    delete_R_Rw, split_pairs, split_rules, mk, more)
    => delete_R_Rw
}

def split_pairsa[A, B, C, D](x0: dpp_ops_ext[A, B, C, D]):
      A => (List[(term[B, C], term[B, C])]) =>
             (List[(term[B, C], term[B, C])], List[(term[B, C], term[B, C])]) =
  x0 match {
  case Dpp_ops_ext(dpp, p, pw, pairs, q, r, rw, rules, q_empty,
                    rules_no_left_var, rules_non_collapsing, is_QNF,
                    nFQ_subset_NF_rules, rules_map, reverse_rules_map,
                    intersect_pairs, replace_pair, intersect_rules, delete_P_Pw,
                    delete_R_Rw, split_pairs, split_rules, mk, more)
    => split_pairs
}

def split_rulesc[A, B, C, D](x0: dpp_ops_ext[A, B, C, D]):
      A => (List[(term[B, C], term[B, C])]) =>
             (List[(term[B, C], term[B, C])], List[(term[B, C], term[B, C])]) =
  x0 match {
  case Dpp_ops_ext(dpp, p, pw, pairs, q, r, rw, rules, q_empty,
                    rules_no_left_var, rules_non_collapsing, is_QNF,
                    nFQ_subset_NF_rules, rules_map, reverse_rules_map,
                    intersect_pairs, replace_pair, intersect_rules, delete_P_Pw,
                    delete_R_Rw, split_pairs, split_rules, mk, more)
    => split_rules
}

def split_proc[A, B : show,
                C : show](i: dpp_ops_ext[A, B, C, Unit], d: A,
                           p_remove: List[(term[B, C], term[B, C])],
                           r_remove: List[(term[B, C], term[B, C])]):
      (A, A) =
  {
    val (p, pw):
          (List[(term[B, C], term[B, C])], List[(term[B, C], term[B, C])])
      = (split_pairsa[A, B, C, Unit](i)).apply(d).apply(p_remove)
    val (r, rw):
          (List[(term[B, C], term[B, C])], List[(term[B, C], term[B, C])])
      = (split_rulesc[A, B, C, Unit](i)).apply(d).apply(r_remove)
    val dpp1: A =
      (mkc[A, B, C,
            Unit](i)).apply(p).apply(pw).apply((qc[A, B, C,
            Unit](i)).apply(d)).apply(r).apply(rw)
    val a: A =
      (delete_R_Rwc[A, B, C,
                     Unit](i)).apply((delete_P_Pwa[A, B, C,
            Unit](i)).apply(d).apply(p_remove).apply(p_remove)).apply(r_remove).apply(r_remove);
    (dpp1, a)
  }

def map_subst_domain[A, B, C, D](m: A => B, sigma: B => term[C, D]):
      A => term[C, D] =
  (v: A) => sigma(m(v))

def subst_apply_ctxt[A, B, C](x0: ctxt[A, B], sigma: B => term[A, C]):
      ctxt[A, C] =
  (x0, sigma) match {
  case (Hole, sigma) => Hole
  case (More(f, ss1, d, ss2), sigma) =>
    More[A, C](f, mapa[term[A, B],
                        term[A, C]]((t: term[A, B]) =>
                                      subst_apply_term[A, B, C](t, sigma),
                                     ss1),
                subst_apply_ctxt[A, B, C](d, sigma),
                mapa[term[A, B],
                      term[A, C]]((t: term[A, B]) =>
                                    subst_apply_term[A, B, C](t, sigma),
                                   ss2))
}

def descaa[A, B, C](x0: root_redtriple_ext[A, B, C]):
      (List[Char]) => List[Char] =
  x0 match {
  case Root_redtriple_ext(valid, s, ns, nst, af, aft, desc, more) => desc
}

def valid[A, B, C](x0: redtriple_ext[A, B, C]):
      sum[(List[Char]) => List[Char], Unit] =
  x0 match {
  case Redtriple_ext(valid, s, ns, nst, af, mono, desc, more) => valid
}

def check_subseteq[A : equal](xs: List[A], ys: List[A]): sum[A, Unit] =
  catcha[(A, A), Unit,
          A](forallm[A, A]((x: A) =>
                             (if (membera[A](ys, x)) Inr[Unit, A](())
                               else Inl[A, Unit](x)),
                            xs),
              (x: (A, A)) => Inl[A, Unit](snd[A, A](x)))

def check_wf_trs[A : show, B : equal : show](r: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]](catcha[((term[A, B], term[A, B]),
                                 (List[Char]) => List[Char]),
                                Unit,
                                (List[Char]) =>
                                  List[Char]](forallm[(term[A, B], term[A, B]),
               (List[Char]) =>
                 List[Char]]((rule: (term[A, B], term[A, B])) =>
                               catcha[(List[Char]) => List[Char], Unit,
                                       (List[Char]) =>
 List[Char]](bindb[(List[Char]) => List[Char], Unit,
                    Unit](check[(List[Char]) =>
                                  List[Char]](! (is_Var[A,
                 B](fst[term[A, B], term[A, B]](rule))),
       shows_string.apply(List('v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ', 'l',
                                'e', 'f', 't', '-', 'h', 'a', 'n', 'd', ' ',
                                's', 'i', 'd', 'e', ' ', 'i', 'n'))),
                           (_: Unit) =>
                             catcha[B, Unit,
                                     (List[Char]) =>
                                       List[Char]](check_subseteq[B](vars_term_list[A,
     B].apply(snd[term[A, B], term[A, B]](rule)),
                              vars_term_list[A,
      B].apply(fst[term[A, B], term[A, B]](rule))),
            (x: B) =>
              Inl[(List[Char]) => List[Char],
                   Unit](comp[List[Char], List[Char],
                               List[Char]](shows_string.apply(List('f', 'r',
                            'e', 'e', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l',
                            'e', ' ')),
    comp[List[Char], List[Char],
          List[Char]]((a: List[Char]) => shows_prec[B](Zero_nat, x, a),
                       shows_string.apply(List(' ', 'i', 'n', ' ', 'r', 'i',
        'g', 'h', 't', '-', 'h', 'a', 'n', 'd', ' ', 's', 'i', 'd', 'e', ' ',
        'o', 'f'))))))),
              (x: (List[Char]) => List[Char]) =>
                Inl[(List[Char]) => List[Char],
                     Unit](comp[List[Char], List[Char],
                                 List[Char]](x,
      comp[List[Char], List[Char],
            List[Char]](shows_string.apply(List(' ', 'r', 'u', 'l', 'e', ' ')),
                         comp[List[Char], List[Char],
                               List[Char]](shows_rule[A,
               B]((a: A) => (b: List[Char]) => shows_prec[A](Zero_nat, a, b),
                   (a: B) => (b: List[Char]) => shows_prec[B](Zero_nat, a, b),
                   List(' ', '-', '>', ' '), rule),
    shows_nl))))),
                              r),
       (x: ((term[A, B], term[A, B]), (List[Char]) => List[Char])) =>
         Inl[(List[Char]) => List[Char],
              Unit](snd[(term[A, B], term[A, B]),
                         (List[Char]) => List[Char]](x))),
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('t', 'h', 'e', ' ', 'T', 'R', 'S', ' ',
 'i', 's', ' ', 'n', 'o', 't', ' ', 'w', 'e', 'l', 'l', '-', 'f', 'o', 'r', 'm',
 'e', 'd')),
                 comp[List[Char], List[Char], List[Char]](shows_nl, x))))

def map_rules_wa[A, B, C](fg: A => nat => B, r: List[(term[A, C], term[A, C])]):
      List[(term[B, C], term[B, C])] =
  mapa[(term[A, C], term[A, C]),
        (term[B, C],
          term[B, C])]((a: (term[A, C], term[A, C])) =>
                         {
                           val (l, ra): (term[A, C], term[A, C]) = a;
                           (map_term_wa[A, B, C](fg, l),
                             map_term_wa[A, B, C](fg, ra))
                         },
                        r)

def generate_f_xs[A](f: A, n: nat): term[A, List[Char]] =
  Fun[A, List[Char]](f, mapa[nat, term[A,
List[Char]]]((i: nat) => Var[List[Char], A](generate_var(i)), upt(Zero_nat, n)))

def bounds_type: xml => sum[List[Char], boundstype] =
  (a: xml) =>
    xml1element[boundstype,
                 boundstype](List('t', 'y', 'p', 'e'),
                              (aa: xml) =>
                                xml_options[boundstype](List((List('r', 'o',
                            'o', 'f'),
                       (ab: xml) =>
                         xml0elements[boundstype](List('r', 'o', 'o', 'f'),
           Roof, ab)),
                      (List('m', 'a', 't', 'c', 'h'),
                        (ab: xml) =>
                          xml0elements[boundstype](List('m', 'a', 't', 'c',
                 'h'),
            Match, ab))),
                 aa),
                              id[boundstype], a)

def bounds_bound: xml => sum[List[Char], nat] =
  (a: xml) => nata(List('b', 'o', 'u', 'n', 'd'), a)

def final_states: xml => sum[List[Char], List[List[Char]]] =
  (a: xml) =>
    xml_many_elements[List[Char],
                       List[List[Char]]](List('f', 'i', 'n', 'a', 'l', 'S', 't',
       'a', 't', 'e', 's'),
  state, id[List[List[Char]]], a)

def transitions[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             (List[ta_rule[List[Char], (A, nat)]],
               List[(List[Char], List[Char])])] =
  (a: xml) =>
    xml_many_elements[sum[ta_rule[List[Char], (A, nat)],
                           (List[Char], List[Char])],
                       (List[ta_rule[List[Char], (A, nat)]],
                         List[(List[Char],
                                List[Char])])](List('t', 'r', 'a', 'n', 's',
             'i', 't', 'i', 'o', 'n', 's'),
        transition[A](xml2name),
        (rls: List[sum[ta_rule[List[Char], (A, nat)],
                        (List[Char], List[Char])]])
          => {
               val (rules, eps):
                     (List[sum[ta_rule[List[Char], (A, nat)],
                                (List[Char], List[Char])]],
                       List[sum[ta_rule[List[Char], (A, nat)],
                                 (List[Char], List[Char])]])
                 = partition[sum[ta_rule[List[Char], (A, nat)],
                                  (List[Char],
                                    List[Char])]]((aa:
             sum[ta_rule[List[Char], (A, nat)], (List[Char], List[Char])])
            => aa match {
                 case Inl(_) => true
                 case Inr(_) => false
               },
           rls)
               val ruls: List[ta_rule[List[Char], (A, nat)]] =
                 mapa[sum[ta_rule[List[Char], (A, nat)],
                           (List[Char], List[Char])],
                       ta_rule[List[Char],
                                (A, nat)]]((aa:
      sum[ta_rule[List[Char], (A, nat)], (List[Char], List[Char])])
     => {
          val (Inl(r)):
                sum[ta_rule[List[Char], (A, nat)], (List[Char], List[Char])]
            = aa;
          r
        },
    rules)
               val aa: List[(List[Char], List[Char])] =
                 mapa[sum[ta_rule[List[Char], (A, nat)],
                           (List[Char], List[Char])],
                       (List[Char],
                         List[Char])]((aa:
 sum[ta_rule[List[Char], (A, nat)], (List[Char], List[Char])])
=> {
     val (Inr(e)): sum[ta_rule[List[Char], (A, nat)], (List[Char], List[Char])]
       = aa;
     e
   },
                                       eps);
               (ruls, aa)
             },
        a)

def bounds_tree_automaton[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], tree_automaton[List[Char], (A, nat)]] =
  (a: xml) =>
    xml2elements[List[List[Char]],
                  (List[ta_rule[List[Char], (A, nat)]],
                    List[(List[Char], List[Char])]),
                  tree_automaton[List[Char],
                                  (A, nat)]](List('t', 'r', 'e', 'e', 'A', 'u',
           't', 'o', 'm', 'a', 't', 'o', 'n'),
      final_states, transitions[A](xml2name),
      (f: List[List[Char]]) =>
        (aa: (List[ta_rule[List[Char], (A, nat)]],
               List[(List[Char], List[Char])]))
          => {
               val (ab, b):
                     (List[ta_rule[List[Char], (A, nat)]],
                       List[(List[Char], List[Char])])
                 = aa;
               Tree_Automaton[List[Char], (A, nat)](f, ab, b)
             },
      a)

def bounds_info[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], bounds_info[A, List[Char]]] =
  (a: xml) =>
    xml4elements[boundstype, nat, List[List[Char]],
                  tree_automaton[List[Char], (A, nat)],
                  bounds_info[A, List[Char]]](List('b', 'o', 'u', 'n', 'd',
            's'),
       bounds_type, bounds_bound, final_states,
       bounds_tree_automaton[A](xml2name),
       (aa: boundstype) =>
         (b: nat) =>
           (c: List[List[Char]]) =>
             (d: tree_automaton[List[Char], (A, nat)]) =>
               Bounds_Info[List[Char], A](aa, b, c, d),
       a)

def label_decomp[A, B](x0: lab[A, B]): (lab[A, B], sum[B, List[lab[A, B]]]) = x0
  match {
  case Lab(f, l) => (f, Inl[B, List[lab[A, B]]](l))
  case FunLab(f, l) => (f, Inr[List[lab[A, B]], B](l))
}

def shows_lab[A, B](fun: A => (List[Char]) => List[Char],
                     lab: B => (List[Char]) => List[Char], x2: lab[A, B]):
      (List[Char]) => List[Char] =
  (fun, lab, x2) match {
  case (fun, lab, UnLab(f)) => fun(f)
  case (fun, lab, Lab(f, l)) =>
    comp[List[Char], List[Char],
          List[Char]](shows_lab[A, B](fun, lab, f),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('[')),
  comp[List[Char], List[Char],
        List[Char]](lab(l), shows_string.apply(List(']')))))
  case (fun, lab, Sharp(f)) =>
    comp[List[Char], List[Char],
          List[Char]](shows_lab[A, B](fun, lab, f),
                       shows_string.apply(List('#')))
  case (fun, lab, FunLab(f, l)) =>
    comp[List[Char], List[Char],
          List[Char]](shows_lab[A, B](fun, lab, f),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('[')),
  comp[List[Char], List[Char],
        List[Char]](shows_map[lab[A, B]]((a: lab[A, B]) =>
   shows_lab[A, B](fun, lab, a),
  l),
                     shows_string.apply(List(']')))))
}

def shows_prec_lab[A : show, B : show](d: nat, l: lab[A, B]):
      (List[Char]) => List[Char] =
  shows_lab[A, B]((a: A) => (b: List[Char]) => shows_prec[A](Zero_nat, a, b),
                   (a: B) => (b: List[Char]) => shows_prec[B](Zero_nat, a, b),
                   l)

def shows_list_lab[A : show, B : show](fs: List[lab[A, B]]):
      (List[Char]) => List[Char] =
  shows_list_aux[lab[A, B]]((a: lab[A, B]) => shows_prec_lab[A, B](Zero_nat, a),
                             fs)

implicit def show_lab[A : show, B : show]: show[lab[A, B]] = new show[lab[A, B]]
  {
  val `Ceta.shows_prec` = (a: nat, b: lab[A, B], c: List[Char]) =>
    (shows_prec_lab[A, B](a, b)).apply(c)
  val `Ceta.shows_list` = (a: List[lab[A, B]], b: List[Char]) =>
    (shows_list_lab[A, B](a)).apply(b)
}

implicit def
  preorder_lab[A : equal : order, B : equal : order]: preorder[lab[A, B]] = new
  preorder[lab[A, B]] {
  val `Ceta.less_eq` = (a: lab[A, B], b: lab[A, B]) => less_eq_lab[A, B](a, b)
  val `Ceta.less` = (a: lab[A, B], b: lab[A, B]) => less_lab[A, B](a, b)
}

implicit def order_lab[A : equal : order, B : equal : order]: order[lab[A, B]] =
  new order[lab[A, B]] {
  val `Ceta.less_eq` = (a: lab[A, B], b: lab[A, B]) => less_eq_lab[A, B](a, b)
  val `Ceta.less` = (a: lab[A, B], b: lab[A, B]) => less_lab[A, B](a, b)
}

def max0a[A, B](x0: poly_c_ext[A, B]): A => A = x0 match {
  case Poly_c_ext(wf, zero, one, default, plus, times, max0, ge, gt, mono,
                   description, arc_pos, check_mono, more)
    => max0
}

def simplify_mp_impl[A : equal,
                      B : equal](mu: A => term[B, A], v_incr: List[A],
                                  mp1: List[(term[B, A], term[B, A])],
                                  solved: List[(term[B, A], term[B, A])]):
      Option[(List[(term[B, A], term[B, A])], nat)] =
  mp1 match {
    case Nil => Some[(List[(term[B, A], term[B, A])], nat)]((solved, Zero_nat))
    case (Var(v), Var(x)) :: mp =>
      simplify_mp_impl[A, B](mu, v_incr, mp,
                              (Var[A, B](v), Var[A, B](x)) :: solved)
    case (Var(v), Fun(_, _)) :: mp =>
      (if (membera[A](v_incr, v))
        {
          val m: (List[(term[B, A], term[B, A])]) =>
                   List[(term[B, A], term[B, A])]
            = (a: List[(term[B, A], term[B, A])]) =>
                mapa[(term[B, A], term[B, A]),
                      (term[B, A],
                        term[B, A])]((aa: (term[B, A], term[B, A])) =>
                                       {
 val (s, ab): (term[B, A], term[B, A]) = aa;
 (subst_apply_term[B, A, A](s, mu), ab)
                                       },
                                      a);
          bind[(List[(term[B, A], term[B, A])], nat),
                (List[(term[B, A], term[B, A])],
                  nat)](simplify_mp_impl[A, B](mu, v_incr, m(mp1), m(solved)),
                         (a: (List[(term[B, A], term[B, A])], nat)) =>
                           {
                             val (smp, i): (List[(term[B, A], term[B, A])], nat)
                               = a;
                             Some[(List[(term[B, A], term[B, A])],
                                    nat)]((smp, Suc(i)))
                           })
        }
        else None)
    case (Fun(fa, lista), Var(x)) :: mp =>
      simplify_mp_impl[A, B](mu, v_incr, mp,
                              (Fun[B, A](fa, lista), Var[A, B](x)) :: solved)
    case (Fun(fa, lista), Fun(f, ls)) :: mp =>
      (if (eq[B](f, fa))
        zip_option[term[B, A], term[B, A]](lista, ls) match {
          case None => None
          case Some(pairs) =>
            simplify_mp_impl[A, B](mu, v_incr, pairs ++ mp, solved)
        }
        else None)
  }

def simplify_mp[A : equal, B : equal](mu: List[(A, term[B, A])], v: List[A]):
      (List[(term[B, A], term[B, A])]) =>
        (List[(term[B, A], term[B, A])]) =>
          Option[(List[(term[B, A], term[B, A])], nat)] =
  {
    val mua: A => term[B, A] = mk_subst[A, B](mu);
    (a: List[(term[B, A], term[B, A])]) =>
      (b: List[(term[B, A], term[B, A])]) =>
        simplify_mp_impl[A, B](mua, v, a, b)
  }

def v_incr_impl[A : equal, B](d: List[(A, term[B, A])]): List[A] =
  {
    val (ixs, uxs): (List[(A, term[B, A])], List[(A, term[B, A])]) =
      partition[(A, term[B, A])]((a: (A, term[B, A])) =>
                                   {
                                     val (_, t): (A, term[B, A]) = a;
                                     ! (is_Var[B, A](t))
                                   },
                                  d)
    val rel: List[(A, A)] =
      mapa[(A, term[B, A]),
            (A, A)]((a: (A, term[B, A])) =>
                      {
                        val (x, y): (A, term[B, A]) = a;
                        (the_Var[B, A](y), x)
                      },
                     uxs)
    val a: List[A] =
      mapa[(A, term[B, A]),
            A]((a: (A, term[B, A])) => fst[A, term[B, A]](a), ixs);
    rtrancl_list_impl[A].apply(rel).apply(a)
  }

def equal_boola(p: Boolean, pa: Boolean): Boolean = (p, pa) match {
  case (p, true) => p
  case (p, false) => ! p
  case (true, p) => p
  case (false, p) => ! p
}

implicit def equal_bool: equal[Boolean] = new equal[Boolean] {
  val `Ceta.equal` = (a: Boolean, b: Boolean) => equal_boola(a, b)
}

def rm_iteratei[A, B : linorder, C]:
      (A => Boolean) => (B => C => A => A) => (rbt[B, C]) => A => A =
  (a: A => Boolean) =>
    (b: B => C => A => A) => (c: rbt[B, C]) => rm_iterateoia[A, B, C](a, b, c)

def rs_iteratei[A, B : linorder]:
      (A => Boolean) => (B => A => A) => (rbt[B, Unit]) => A => A =
  (a: A => Boolean) =>
    (b: B => A => A) =>
      (c: rbt[B, Unit]) =>
        (d: A) =>
          s_iteratei[A, B, rbt[B, Unit]](rm_iteratei[A, B, Unit], a, b, c, d)

def scnp_af_desc[A : show](af: List[((A, nat), List[(nat, nat)])]):
      (List[Char]) => List[Char] =
  shows_prec_list[Char](Zero_nat,
                         List('S', 'C', 'N', 'P', '-', 'v', 'e', 'r', 's', 'i',
                               'o', 'n', ' ', 'o', 'f', ' '))

def set_ext[A](s_ns: A => A => (Boolean, Boolean)):
      (List[A]) => (List[A]) => (Boolean, Boolean) =
  (as: List[A]) =>
    (bs: List[A]) =>
      (! (nulla[A](as)) &&
         list_all[A]((b: A) =>
                       list_ex[A]((a: A) => fst[Boolean, Boolean]((s_ns(a))(b)),
                                   as),
                      bs),
        list_all[A]((b: A) =>
                      list_ex[A]((a: A) => snd[Boolean, Boolean]((s_ns(a))(b)),
                                  as),
                     bs))

def check_no_var[A : show, B : show](t: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  check[(List[Char]) =>
          List[Char]](! (is_Var[A, B](t)),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('v', 'a', 'r',
                          'i', 'a', 'b', 'l', 'e', ' ', 'f', 'o', 'u', 'n',
                          'd')),
  shows_nl))

def check_supteq[A : equal : show,
                  B : equal : show](s: term[A, B], t: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  check[(List[Char]) =>
          List[Char]](member[(term[A, B], term[A, B])]((s, t), supteq[A, B]),
                       comp[List[Char], List[Char],
                             List[Char]]((shows_prec_term[A,
                   B](Zero_nat)).apply(t),
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List(' ', 'i', 's', ' ', 'n', 'o', 't',
     ' ', 'a', ' ', 's', 'u', 'b', 't', 'e', 'r', 'm', ' ', 'o', 'f', ' ')),
                     (shows_prec_term[A, B](Zero_nat)).apply(s))))

def valida[A, B, C](x0: root_redtriple_ext[A, B, C]):
      sum[(List[Char]) => List[Char], Unit] =
  x0 match {
  case Root_redtriple_ext(valid, s, ns, nst, af, aft, desc, more) => valid
}

def first_rewrite[A : equal,
                   B : equal](r: List[(term[A, B], term[A, B])], s: term[A, B]):
      Option[term[A, B]] =
  rewrite[A, B](r, s) match {
    case Nil => None
    case t :: _ => Some[term[A, B]](t)
  }

def compute_NF[A](f: A => Option[A], a: A): Option[A] =
  f(a) match {
    case None => Some[A](a)
    case Some(aa) => compute_NF[A](f, aa)
  }

def compute_rstep_NF[A : equal,
                      B : equal](r: List[(term[A, B], term[A, B])],
                                  s: term[A, B]):
      Option[term[A, B]] =
  compute_NF[term[A, B]]((a: term[A, B]) => first_rewrite[A, B](r, a), s)

def check_join_NF[A : equal : show,
                   B : equal : show](r: List[(term[A, B], term[A, B])],
                                      s: term[A, B], t: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  (compute_rstep_NF[A, B](r, s), compute_rstep_NF[A, B](r, t)) match {
    case (None, b) =>
      Inl[(List[Char]) => List[Char],
           Unit](shows_prec_list[Char](Zero_nat,
List('s', 't', 'r', 'a', 'n', 'g', 'e', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i',
      'n', ' ', 'n', 'o', 'r', 'm', 'a', 'l', ' ', 'f', 'o', 'r', 'm', ' ', 'c',
      'o', 'm', 'p', 'u', 't', 'a', 't', 'i', 'o', 'n')))
    case (Some(sa), None) =>
      Inl[(List[Char]) => List[Char],
           Unit](shows_prec_list[Char](Zero_nat,
List('s', 't', 'r', 'a', 'n', 'g', 'e', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i',
      'n', ' ', 'n', 'o', 'r', 'm', 'a', 'l', ' ', 'f', 'o', 'r', 'm', ' ', 'c',
      'o', 'm', 'p', 'u', 't', 'a', 't', 'i', 'o', 'n')))
    case (Some(sa), Some(ta)) =>
      check[(List[Char]) =>
              List[Char]](equal_terma[A, B](sa, ta),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('t', 'h',
                              'e', ' ', 'n', 'o', 'r', 'm', 'a', 'l', ' ', 'f',
                              'o', 'r', 'm', ' ')),
      comp[List[Char], List[Char],
            List[Char]]((shows_prec_term[A, B](Zero_nat)).apply(sa),
                         comp[List[Char], List[Char],
                               List[Char]](shows_prec_list[Char](Zero_nat,
                          List(' ', 'o', 'f', ' ')),
    comp[List[Char], List[Char],
          List[Char]]((shows_prec_term[A, B](Zero_nat)).apply(s),
                       comp[List[Char], List[Char],
                             List[Char]](shows_prec_list[Char](Zero_nat,
                        List(' ', 'd', 'i', 'f', 'f', 'e', 'r', 's', ' ', 'f',
                              'r', 'o', 'm', ' ')),
  comp[List[Char], List[Char],
        List[Char]](shows_nl,
                     comp[List[Char], List[Char],
                           List[Char]](shows_prec_list[Char](Zero_nat,
                      List('t', 'h', 'e', ' ', 'n', 'o', 'r', 'm', 'a', 'l',
                            ' ', 'f', 'o', 'r', 'm', ' ')),
comp[List[Char], List[Char],
      List[Char]]((shows_prec_term[A, B](Zero_nat)).apply(ta),
                   comp[List[Char], List[Char],
                         List[Char]](shows_prec_list[Char](Zero_nat,
                    List(' ', 'o', 'f', ' ')),
                                      (shows_prec_term[A,
                B](Zero_nat)).apply(t)))))))))))
  }

def vars_rule_lista[A, B](r: (term[A, B], term[A, B])): List[B] =
  vars_term_lista[A, B](fst[term[A, B], term[A, B]](r)) ++
    vars_term_lista[A, B](snd[term[A, B], term[A, B]](r))

def vars_trs_list[A, B : equal](trs: List[(term[A, B], term[A, B])]): List[B] =
  remdups[B](maps[(term[A, B], term[A, B]),
                   B]((a: (term[A, B], term[A, B])) => vars_rule_lista[A, B](a),
                       trs))

def check_pairwise[A, B](c: A => A => sum[B, Unit], x1: List[A]): sum[B, Unit] =
  (c, x1) match {
  case (c, Nil) => Inr[Unit, B](())
  case (c, x :: xs) =>
    bindb[B, Unit,
           Unit](catcha[(A, B), Unit,
                         B](forallm[A, B](c(x), xs),
                             (xa: (A, B)) => Inl[B, Unit](snd[A, B](xa))),
                  (_: Unit) => check_pairwise[A, B](c, xs))
}

def check_disjoint[A : equal](xs: List[A], ys: List[A]): sum[A, Unit] =
  catcha[(A, A), Unit,
          A](forallm[A, A]((x: A) =>
                             (if (! (membera[A](ys, x))) Inr[Unit, A](())
                               else Inl[A, Unit](x)),
                            xs),
              (x: (A, A)) => Inl[A, Unit](snd[A, A](x)))

def check_partition[A : equal](xss: List[List[A]]): sum[A, Unit] =
  check_pairwise[List[A],
                  A]((a: List[A]) => (b: List[A]) => check_disjoint[A](a, b),
                      xss)

def check_inj[A : equal : show,
               B : equal : show](a: lab[A, B], nn: nat,
                                  sml: List[((lab[A, B], nat),
      List[lab[A, B]])]):
      sum[(List[Char]) => List[Char], Unit] =
  {
    val symbols: List[List[(lab[A, B], nat)]] =
      mapa[((lab[A, B], nat), List[lab[A, B]]),
            List[(lab[A, B],
                   nat)]]((aa: ((lab[A, B], nat), List[lab[A, B]])) =>
                            {
                              val (ab, b): ((lab[A, B], nat), List[lab[A, B]]) =
                                aa;
                              ({
                                 val (_, n): (lab[A, B], nat) = ab;
                                 (fs: List[lab[A, B]]) =>
                                   mapa[(lab[A, B], nat),
 (lab[A, B],
   nat)]((ac: (lab[A, B], nat)) =>
           {
             val (g, i): (lab[A, B], nat) = ac;
             (g, plus_nat(n, times_nat(i, minus_nat(nn, one_nat))))
           },
          zip[lab[A, B], nat](fs, upt(Zero_nat, size_list[lab[A, B]](fs))))
                               })(b)
                            },
                           sml)
    val fsymbols: List[(lab[A, B], nat)] = concat[(lab[A, B], nat)](symbols);
    bindb[(List[Char]) => List[Char], Unit,
           Unit](catcha[(lab[A, B], nat), Unit,
                         (List[Char]) =>
                           List[Char]](check_partition[(lab[A, B],
                 nat)](symbols),
(x: (lab[A, B], nat)) =>
  Inl[(List[Char]) => List[Char],
       Unit](comp[List[Char], List[Char],
                   List[Char]](shows_prec_list[Char](Zero_nat,
              List('s', 'y', 'm', 'b', 'o', 'l', ' ')),
                                comp[List[Char], List[Char],
                                      List[Char]](shows_prec_prod[lab[A, B],
                           nat](Zero_nat, x),
           shows_prec_list[Char](Zero_nat,
                                  List(' ', 'o', 'c', 'c', 'u', 'r', 's', ' ',
't', 'w', 'i', 'c', 'e', ')')))))),
                  (_: Unit) =>
                    bindb[(List[Char]) => List[Char], Unit,
                           Unit](check[(List[Char]) =>
 List[Char]](! (membera[(lab[A, B], nat)](fsymbols, (a, nn))),
              comp[List[Char], List[Char],
                    List[Char]](shows_prec_list[Char](Zero_nat,
               List('a', 'p', 'p', 'l', 'i', 'c', 'a', 't', 'i', 'o', 'n', ' ',
                     's', 'y', 'm', 'b', 'o', 'l')),
                                 comp[List[Char], List[Char],
                                       List[Char]](shows_prec_lab[A,
                           B](Zero_nat, a),
            shows_prec_list[Char](Zero_nat,
                                   List(' ', 'm', 'u', 's', 't', ' ', 'n', 'o',
 't', ' ', 'o', 'c', 'c', 'u', 'r', ' ', 'a', 's', ' ', 'n', 'e', 'w', ' ', 's',
 'y', 'm', 'b', 'o', 'l'))))),
                                  (_: Unit) =>
                                    bindb[(List[Char]) => List[Char], Unit,
   Unit](check[(List[Char]) =>
                 List[Char]](! (membera[(lab[A, B],
  nat)](mapa[((lab[A, B], nat), List[lab[A, B]]),
              (lab[A, B],
                nat)]((aa: ((lab[A, B], nat), List[lab[A, B]])) =>
                        fst[(lab[A, B], nat), List[lab[A, B]]](aa),
                       sml),
         (a, nn))),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_prec_list[Char](Zero_nat,
                               List('a', 'p', 'p', 'l', 'i', 'c', 'a', 't', 'i',
                                     'o', 'n', ' ', 's', 'y', 'm', 'b', 'o',
                                     'l')),
         comp[List[Char], List[Char],
               List[Char]](shows_prec_lab[A, B](Zero_nat, a),
                            shows_prec_list[Char](Zero_nat,
           List(' ', 'm', 'u', 's', 't', ' ', 'n', 'o', 't', ' ', 'b', 'e', ' ',
                 'u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd'))))),
          (_: Unit) =>
            (if (less_eq_nat(nn, one_nat))
              catcha[(List[(lab[A, B], nat)], (List[Char]) => List[Char]), Unit,
                      (List[Char]) =>
                        List[Char]](forallm[List[(lab[A, B], nat)],
     (List[Char]) =>
       List[Char]]((aa: List[(lab[A, B], nat)]) =>
                     check_pairwise[(lab[A, B], nat),
                                     (List[Char]) =>
                                       List[Char]]((gn1: (lab[A, B], nat)) =>
             (gn2: (lab[A, B], nat)) =>
               check[(List[Char]) =>
                       List[Char]](! (equal_proda[lab[A, B], nat](gn1, gn2)),
                                    comp[List[Char], List[Char],
  List[Char]](shows_prec_list[Char](Zero_nat,
                                     List('s', 'y', 'm', 'b', 'o', 'l', ' ')),
               comp[List[Char], List[Char],
                     List[Char]](shows_prec_prod[lab[A, B], nat](Zero_nat, gn1),
                                  shows_prec_list[Char](Zero_nat,
                 List(' ', 'o', 'c', 'c', 'u', 'r', 's', ' ', 't', 'w', 'i',
                       'c', 'e'))))),
            aa),
                    symbols),
                                     (x:
(List[(lab[A, B], nat)], (List[Char]) => List[Char]))
                                       => Inl[(List[Char]) => List[Char],
       Unit](snd[List[(lab[A, B], nat)], (List[Char]) => List[Char]](x)))
              else Inr[Unit, (List[Char]) => List[Char]](())))))
  }

def uncurry_info[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             (A, (List[((A, nat), List[A])],
                   (List[(term[A, List[Char]], term[A, List[Char]])],
                     List[(term[A, List[Char]], term[A, List[Char]])])))] =
  (a: xml) =>
    xml4elements[A, List[((A, nat), List[A])],
                  List[(term[A, List[Char]], term[A, List[Char]])],
                  List[(term[A, List[Char]], term[A, List[Char]])],
                  (A, (List[((A, nat), List[A])],
                        (List[(term[A, List[Char]], term[A, List[Char]])],
                          List[(term[A, List[Char]],
                                 term[A, List[Char]])])))](List('u', 'n', 'c',
                         'u', 'r', 'r', 'y', 'I', 'n', 'f', 'o', 'r', 'm', 'a',
                         't', 'i', 'o', 'n'),
                    xml2name,
                    (aa: xml) =>
                      xml_many_elements[((A, nat), List[A]),
 List[((A, nat),
        List[A])]](List('u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', 'S', 'y',
                         'm', 'b', 'o', 'l', 's'),
                    (ab: xml) =>
                      xml2many_elements[A, nat, A,
 ((A, nat),
   List[A])](List('u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', 'S', 'y', 'm',
                   'b', 'o', 'l', 'E', 'n', 't', 'r', 'y'),
              xml2name, (ac: xml) => nata(List('a', 'r', 'i', 't', 'y'), ac),
              xml2name, (f: A) => (n: nat) => (ac: List[A]) => ((f, n), ac),
              ab),
                    id[List[((A, nat), List[A])]], aa),
                    (aa: xml) =>
                      xml1element[List[(term[A, List[Char]],
 term[A, List[Char]])],
                                   List[(term[A, List[Char]],
  term[A, List[Char]])]](List('u', 'n', 'c', 'u', 'r', 'r', 'y', 'R', 'u', 'l',
                               'e', 's'),
                          rules[A](xml2name),
                          id[List[(term[A, List[Char]], term[A, List[Char]])]],
                          aa),
                    (aa: xml) =>
                      xml1element[List[(term[A, List[Char]],
 term[A, List[Char]])],
                                   List[(term[A, List[Char]],
  term[A, List[Char]])]](List('e', 't', 'a', 'R', 'u', 'l', 'e', 's'),
                          rules[A](xml2name),
                          id[List[(term[A, List[Char]], term[A, List[Char]])]],
                          aa),
                    (aa: A) =>
                      (sml: List[((A, nat), List[A])]) =>
                        (u: List[(term[A, List[Char]], term[A, List[Char]])]) =>
                          (e: List[(term[A, List[Char]], term[A, List[Char]])])
                            => (aa, (sml, (u, e))),
                    a)

def flat_contexts[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], List[ctxt[A, List[Char]]]] =
  (a: xml) =>
    xml_many_elements[ctxt[A, List[Char]],
                       List[ctxt[A, List[Char]]]](List('f', 'l', 'a', 't', 'C',
                'o', 'n', 't', 'e', 'x', 't', 's'),
           (aa: xml) => ctxt[A](xml2name, aa), id[List[ctxt[A, List[Char]]]], a)

def projected_rseq[A : equal](xml2name: xml => sum[List[Char], A],
                               pi: ((A, nat)) => nat, x2: xml):
      sum[List[Char],
           ((term[A, List[Char]], term[A, List[Char]]),
             List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                          term[A, List[Char]]))])] =
  (xml2name, pi, x2) match {
  case (xml2name, pi, Xml(name, atts, cs, text)) =>
    {
      val tag: List[Char] =
        List('p', 'r', 'o', 'j', 'e', 'c', 't', 'e', 'd', 'R', 'e', 'w', 'r',
              'i', 't', 'e', 'S', 'e', 'q', 'u', 'e', 'n', 'c', 'e');
      (if (equal_lista[Char](name, tag) &&
             (nulla[(List[Char], List[Char])](atts) &&
               (is_none[List[Char]](text) &&
                 equal_nata(size_list[xml](cs), nat(BigInt(2))))))
        bindb[List[Char], (term[A, List[Char]], term[A, List[Char]]),
               ((term[A, List[Char]], term[A, List[Char]]),
                 List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                              term[A, List[Char]]))])]((rule[A](xml2name)).apply(nth[xml](cs,
           Zero_nat)),
                (r: (term[A, List[Char]], term[A, List[Char]])) =>
                  (rseq[A](xml2name, pi, r)).apply(nth[xml](cs, one_nat)))
        else no_transform[((term[A, List[Char]], term[A, List[Char]]),
                            List[(pos, ((term[A, List[Char]],
  term[A, List[Char]]),
 term[A, List[Char]]))])](tag, Xml(name, atts, cs, text)))
    }
}

def xml1many2elements_gen[A, B, C, D,
                           E](tag: List[Char], p1: xml => sum[List[Char], A],
                               p2: A => xml => sum[List[Char], B],
                               p3: xml => sum[List[Char], C],
                               p4: xml => sum[List[Char], D],
                               f: A => (List[B]) => C => D => E, x6: xml):
      sum[List[Char], E] =
  (tag, p1, p2, p3, p4, f, x6) match {
  case (tag, p1, p2, p3, p4, f, Xml(name, atts, cs, text)) =>
    {
      val ds: List[xml] = rev[xml](cs)
      val l: nat = size_list[xml](cs);
      (if (equal_lista[Char](name, tag) &&
             (nulla[(List[Char], List[Char])](atts) &&
               (is_none[List[Char]](text) && less_eq_nat(nat(BigInt(3)), l))))
        bindb[List[Char], A,
               E](p1(nth[xml](cs, Zero_nat)),
                   (x: A) =>
                     bindb[List[Char], List[B],
                            E](map[B].apply(p2(x)).apply(tl[xml](take[xml](minus_nat(l,
      nat(BigInt(2))),
                                    cs))),
                                (xs: List[B]) =>
                                  bindb[List[Char], C,
 E](p3(nth[xml](ds, one_nat)),
     (y: C) =>
       bindb[List[Char], D,
              E](p4(nth[xml](ds, Zero_nat)),
                  (z: D) => Inr[E, List[Char]]((((f(x))(xs))(y))(z))))))
        else no_transform[E](tag, Xml(name, atts, cs, text)))
    }
}

def strategy_to_Q[A : equal,
                   B : equal](x0: strategy[A, B],
                               uu: List[(term[A, B], term[A, B])]):
      List[term[A, B]] =
  (x0, uu) match {
  case (No_Strategy, uu) => Nil
  case (Innermost, r) =>
    remdups[term[A, B]](mapa[(term[A, B], term[A, B]),
                              term[A, B]]((a: (term[A, B], term[A, B])) =>
    fst[term[A, B], term[A, B]](a),
   r))
  case (Innermost_Q(q), uv) => q
}

def create_proj[A : equal](x0: projL[A], f: (A, nat)): nat = (x0, f) match {
  case (Projection(p), f) =>
    lookupa[(A, nat), nat](f, p) match {
      case None => Zero_nat
      case Some(n) => n
    }
}

def ordering_constraint_proofa[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], redtriple_impl[A]] =
  (a: xml) =>
    xml1element[redtriple_impl[A],
                 redtriple_impl[A]](List('o', 'r', 'd', 'e', 'r', 'i', 'n', 'g',
  'C', 'o', 'n', 's', 't', 'r', 'a', 'i', 'n', 't', 'P', 'r', 'o', 'o', 'f'),
                                     redtriple[A](xml2name),
                                     id[redtriple_impl[A]], a)

def level_mapping[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], List[((A, nat), List[(nat, nat)])]] =
  (a: xml) =>
    xml_many_elements[((A, nat), List[(nat, nat)]),
                       List[((A, nat),
                              List[(nat, nat)])]](List('l', 'e', 'v', 'e', 'l',
                'M', 'a', 'p', 'p', 'i', 'n', 'g'),
           (aa: xml) =>
             xml2many_elements[A, nat, (nat, nat),
                                ((A, nat),
                                  List[(nat,
 nat)])](List('l', 'e', 'v', 'e', 'l', 'M', 'a', 'p', 'p', 'i', 'n', 'g', 'E',
               'n', 't', 'r', 'y'),
          xml2name, (ab: xml) => nata(List('a', 'r', 'i', 't', 'y'), ab),
          (ab: xml) =>
            xml2elements[nat, nat,
                          (nat, nat)](List('p', 'o', 's', 'i', 't', 'i', 'o',
    'n', 'L', 'e', 'v', 'e', 'l', 'E', 'n', 't', 'r', 'y'),
                                       (ac: xml) =>
 nata(List('p', 'o', 's', 'i', 't', 'i', 'o', 'n'), ac),
                                       (ac: xml) =>
 nata(List('l', 'e', 'v', 'e', 'l'), ac),
                                       (ac: nat) => (b: nat) => (ac, b), ab),
          (f: A) =>
            (ab: nat) =>
              (ps: List[(nat, nat)]) =>
                ((f, ab),
                  mapa[(nat, nat),
                        (nat, nat)]((b: (nat, nat)) =>
                                      {
val (p, ba): (nat, nat) = b;
((if (equal_nata(p, Zero_nat)) ab else minus_nat(p, one_nat)), ba)
                                      },
                                     ps)),
          aa),
           id[List[((A, nat), List[(nat, nat)])]], a)

def root_redtriple[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], root_redtriple_impl[A]] =
  (a: xml) =>
    xml1element[root_redtriple_impl[A],
                 root_redtriple_impl[A]](List('r', 'e', 'd', 'P', 'a', 'i',
       'r'),
  (aa: xml) =>
    xml3elements[list_order_type, List[((A, nat), List[(nat, nat)])],
                  redtriple_impl[A],
                  root_redtriple_impl[A]](List('s', 'c', 'n', 'p'),
   (ab: xml) =>
     xml1element[list_order_type,
                  list_order_type](List('s', 't', 'a', 't', 'u', 's'),
                                    (ac: xml) =>
                                      xml_options[list_order_type](List((List('m',
                                       's'),
                                  (ad: xml) =>
                                    xml0elements[list_order_type](List('m',
                                's'),
                           MS_Ext, ad)),
                                 (List('m', 'i', 'n'),
                                   (ad: xml) =>
                                     xml0elements[list_order_type](List('m',
                                 'i', 'n'),
                            Min_Ext, ad)),
                                 (List('m', 'a', 'x'),
                                   (ad: xml) =>
                                     xml0elements[list_order_type](List('m',
                                 'a', 'x'),
                            Max_Ext, ad))),
                            ac),
                                    id[list_order_type], ab),
   level_mapping[A](xml2name), redtriple[A](xml2name),
   (ab: list_order_type) =>
     (b: List[((A, nat), List[(nat, nat)])]) =>
       (c: redtriple_impl[A]) => Scnp[A](ab, b, c),
   aa),
  id[root_redtriple_impl[A]], a)

def root_ordering_constraint_proof[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], root_redtriple_impl[A]] =
  (a: xml) =>
    xml1element[root_redtriple_impl[A],
                 root_redtriple_impl[A]](List('o', 'r', 'd', 'e', 'r', 'i', 'n',
       'g', 'C', 'o', 'n', 's', 't', 'r', 'a', 'i', 'n', 't', 'P', 'r', 'o',
       'o', 'f'),
  root_redtriple[A](xml2name), id[root_redtriple_impl[A]], a)

def ordering_constraint_proof[A](xml2name: xml => sum[List[Char], A]):
      xml => sum[List[Char], sum[root_redtriple_impl[A], redtriple_impl[A]]] =
  (a: xml) =>
    xml_choice[sum[root_redtriple_impl[A],
                    redtriple_impl[A]]](List('r', 'e', 'd', 'u', 'c', 't', 'i',
      'o', 'n', ' ', 'p', 'a', 'i', 'r'),
 List((aa: xml) =>
        xml_change[redtriple_impl[A],
                    sum[root_redtriple_impl[A],
                         redtriple_impl[A]]](ordering_constraint_proofa[A](xml2name),
      (ab: redtriple_impl[A]) =>
        Inr[redtriple_impl[A], root_redtriple_impl[A]](ab),
      aa),
       (aa: xml) =>
         xml_change[root_redtriple_impl[A],
                     sum[root_redtriple_impl[A],
                          redtriple_impl[A]]](root_ordering_constraint_proof[A](xml2name),
       (ab: root_redtriple_impl[A]) =>
         Inl[root_redtriple_impl[A], redtriple_impl[A]](ab),
       aa)),
 a)

def xml2dp_proof[A : equal,
                  B : equal](xml2name: xml => sum[List[Char], lab[A, B]],
                              x: xml):
      sum[List[Char], dp_proof[A, B, List[Char]]] =
  xml1element[dp_proof[A, B, List[Char]],
               dp_proof[A, B,
                         List[Char]]](List('d', 'p', 'P', 'r', 'o', 'o', 'f'),
                                       (a: xml) =>
 xml_options[dp_proof[A, B,
                       List[Char]]](List((List('p', 'I', 's', 'E', 'm', 'p',
        't', 'y'),
   (aa: xml) =>
     xml0elements[dp_proof[A, B,
                            List[Char]]](List('p', 'I', 's', 'E', 'm', 'p', 't',
       'y'),
  PisEmpty, aa)),
  (List('d', 'e', 'p', 'G', 'r', 'a', 'p', 'h', 'P', 'r', 'o', 'c'),
    (aa: xml) =>
      xml_many_elements[(Option[dp_proof[A, B, List[Char]]],
                          List[(term[lab[A, B], List[Char]],
                                 term[lab[A, B], List[Char]])]),
                         dp_proof[A, B,
                                   List[Char]]](List('d', 'e', 'p', 'G', 'r',
              'a', 'p', 'h', 'P', 'r', 'o', 'c'),
         (ab: xml) =>
           xml2nd_choice[List[(term[lab[A, B], List[Char]],
                                term[lab[A, B], List[Char]])],
                          dp_proof[A, B, List[Char]],
                          (Option[dp_proof[A, B, List[Char]]],
                            List[(term[lab[A, B], List[Char]],
                                   term[lab[A, B],
 List[Char]])])](List('c', 'o', 'm', 'p', 'o', 'n', 'e', 'n', 't'),
                  (ac: xml) =>
                    xml1element[List[(term[lab[A, B], List[Char]],
                                       term[lab[A, B], List[Char]])],
                                 List[(term[lab[A, B], List[Char]],
term[lab[A, B],
      List[Char]])]](List('d', 'p', 's'), rules[lab[A, B]](xml2name),
                      id[List[(term[lab[A, B], List[Char]],
                                term[lab[A, B], List[Char]])]],
                      ac),
                  List('r', 'e', 'a', 'l', 'S', 'c', 'c'),
                  (ac: xml) => xml2dp_proof[A, B](xml2name, ac),
                  (dps: List[(term[lab[A, B], List[Char]],
                               term[lab[A, B], List[Char]])])
                    => (prfOpt: Option[dp_proof[A, B, List[Char]]]) =>
                         (prfOpt, dps),
                  ab),
         (ab: List[(Option[dp_proof[A, B, List[Char]]],
                     List[(term[lab[A, B], List[Char]],
                            term[lab[A, B], List[Char]])])])
           => DepGraphProc[A, B, List[Char]](ab),
         aa)),
  (List('r', 'e', 'd', 'P', 'a', 'i', 'r', 'P', 'r', 'o', 'c'),
    (aa: xml) =>
      xml3elements[sum[root_redtriple_impl[lab[A, B]],
                        redtriple_impl[lab[A, B]]],
                    List[(term[lab[A, B], List[Char]],
                           term[lab[A, B], List[Char]])],
                    dp_proof[A, B, List[Char]],
                    dp_proof[A, B,
                              List[Char]]](List('r', 'e', 'd', 'P', 'a', 'i',
         'r', 'P', 'r', 'o', 'c'),
    ordering_constraint_proof[lab[A, B]](xml2name),
    (ab: xml) =>
      xml1element[List[(term[lab[A, B], List[Char]],
                         term[lab[A, B], List[Char]])],
                   List[(term[lab[A, B], List[Char]],
                          term[lab[A, B],
                                List[Char]])]](List('d', 'p', 's'),
        rules[lab[A, B]](xml2name),
        id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
        ab),
    (ab: xml) => xml2dp_proof[A, B](xml2name, ab),
    (ab: sum[root_redtriple_impl[lab[A, B]], redtriple_impl[lab[A, B]]]) =>
      (b: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]) =>
        (c: dp_proof[A, B, List[Char]]) =>
          RedPairProc[A, B, List[Char]](ab, b, c),
    aa)),
  (List('u', 's', 'a', 'b', 'l', 'e', 'R', 'u', 'l', 'e', 's', 'P', 'r', 'o',
         'c'),
    (aa: xml) =>
      xml2elements[List[(term[lab[A, B], List[Char]],
                          term[lab[A, B], List[Char]])],
                    dp_proof[A, B, List[Char]],
                    dp_proof[A, B,
                              List[Char]]](List('u', 's', 'a', 'b', 'l', 'e',
         'R', 'u', 'l', 'e', 's', 'P', 'r', 'o', 'c'),
    (ab: xml) =>
      xml1element[List[(term[lab[A, B], List[Char]],
                         term[lab[A, B], List[Char]])],
                   List[(term[lab[A, B], List[Char]],
                          term[lab[A, B],
                                List[Char]])]](List('u', 's', 'a', 'b', 'l',
             'e', 'R', 'u', 'l', 'e', 's'),
        rules[lab[A, B]](xml2name),
        id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
        ab),
    (ab: xml) => xml2dp_proof[A, B](xml2name, ab),
    (ab: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]) =>
      (b: dp_proof[A, B, List[Char]]) =>
        UsableRulesProc[A, B, List[Char]](ab, b),
    aa)),
  (List('r', 'e', 'w', 'r', 'i', 't', 'i', 'n', 'g', 'P', 'r', 'o', 'c'),
    (aa: xml) =>
      xml4elements[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]]),
                    (pos, ((term[lab[A, B], List[Char]],
                             term[lab[A, B], List[Char]]),
                            term[lab[A, B], List[Char]])),
                    List[(term[lab[A, B], List[Char]],
                           term[lab[A, B], List[Char]])],
                    dp_proof[A, B, List[Char]],
                    dp_proof[A, B,
                              List[Char]]](List('r', 'e', 'w', 'r', 'i', 't',
         'i', 'n', 'g', 'P', 'r', 'o', 'c'),
    rule[lab[A, B]](xml2name), rstep[lab[A, B]](xml2name),
    (ab: xml) =>
      xml1element[List[(term[lab[A, B], List[Char]],
                         term[lab[A, B], List[Char]])],
                   List[(term[lab[A, B], List[Char]],
                          term[lab[A, B],
                                List[Char]])]](List('u', 's', 'a', 'b', 'l',
             'e', 'R', 'u', 'l', 'e', 's'),
        rules[lab[A, B]](xml2name),
        id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
        ab),
    (ab: xml) => xml2dp_proof[A, B](xml2name, ab),
    (ab: (term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])) =>
      {
        val (s, t): (term[lab[A, B], List[Char]], term[lab[A, B], List[Char]]) =
          ab;
        (ac: (pos, ((term[lab[A, B], List[Char]], term[lab[A, B], List[Char]]),
                     term[lab[A, B], List[Char]])))
          => {
               val (p, (lr, ta)):
                     (pos, ((term[lab[A, B], List[Char]],
                              term[lab[A, B], List[Char]]),
                             term[lab[A, B], List[Char]]))
                 = ac;
               (u: List[(term[lab[A, B], List[Char]],
                          term[lab[A, B], List[Char]])])
                 => (ad: dp_proof[A, B, List[Char]]) =>
                      RewritingProc[A, B,
                                     List[Char]](u, (s, t), (s, ta), lr, p, ad)
             }
      },
    aa)),
  (List('n', 'a', 'r', 'r', 'o', 'w', 'i', 'n', 'g', 'P', 'r', 'o', 'c'),
    (aa: xml) =>
      xml4elements[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]]),
                    pos,
                    List[(term[lab[A, B], List[Char]],
                           term[lab[A, B], List[Char]])],
                    dp_proof[A, B, List[Char]],
                    dp_proof[A, B,
                              List[Char]]](List('n', 'a', 'r', 'r', 'o', 'w',
         'i', 'n', 'g', 'P', 'r', 'o', 'c'),
    rule[lab[A, B]](xml2name), pos,
    (ab: xml) =>
      xml1element[List[(term[lab[A, B], List[Char]],
                         term[lab[A, B], List[Char]])],
                   List[(term[lab[A, B], List[Char]],
                          term[lab[A, B],
                                List[Char]])]](List('n', 'a', 'r', 'r', 'o',
             'w', 'i', 'n', 'g', 's'),
        rules[lab[A, B]](xml2name),
        id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
        ab),
    (ab: xml) => xml2dp_proof[A, B](xml2name, ab),
    (ab: (term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])) =>
      (b: pos) =>
        (c: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]) =>
          (d: dp_proof[A, B, List[Char]]) =>
            NarrowingProc[A, B, List[Char]](ab, b, c, d),
    aa)),
  (List('i', 'n', 's', 't', 'a', 'n', 't', 'i', 'a', 't', 'i', 'o', 'n', 'P',
         'r', 'o', 'c'),
    (aa: xml) =>
      xml3elements[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]]),
                    List[(term[lab[A, B], List[Char]],
                           term[lab[A, B], List[Char]])],
                    dp_proof[A, B, List[Char]],
                    dp_proof[A, B,
                              List[Char]]](List('i', 'n', 's', 't', 'a', 'n',
         't', 'i', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'c'),
    rule[lab[A, B]](xml2name),
    (ab: xml) =>
      xml1element[List[(term[lab[A, B], List[Char]],
                         term[lab[A, B], List[Char]])],
                   List[(term[lab[A, B], List[Char]],
                          term[lab[A, B],
                                List[Char]])]](List('i', 'n', 's', 't', 'a',
             'n', 't', 'i', 'a', 't', 'i', 'o', 'n', 's'),
        rules[lab[A, B]](xml2name),
        id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
        ab),
    (ab: xml) => xml2dp_proof[A, B](xml2name, ab),
    (ab: (term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])) =>
      (b: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]) =>
        (c: dp_proof[A, B, List[Char]]) =>
          InstantiationProc[A, B, List[Char]](ab, b, c),
    aa)),
  (List('f', 'o', 'r', 'w', 'a', 'r', 'd', 'I', 'n', 's', 't', 'a', 'n', 't',
         'i', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'c'),
    (aa: xml) =>
      xml3to4elements[(term[lab[A, B], List[Char]],
                        term[lab[A, B], List[Char]]),
                       List[(term[lab[A, B], List[Char]],
                              term[lab[A, B], List[Char]])],
                       List[(term[lab[A, B], List[Char]],
                              term[lab[A, B], List[Char]])],
                       dp_proof[A, B, List[Char]],
                       dp_proof[A, B,
                                 List[Char]]](List('f', 'o', 'r', 'w', 'a', 'r',
            'd', 'I', 'n', 's', 't', 'a', 'n', 't', 'i', 'a', 't', 'i', 'o',
            'n', 'P', 'r', 'o', 'c'),
       rule[lab[A, B]](xml2name),
       (ab: xml) =>
         xml1element[List[(term[lab[A, B], List[Char]],
                            term[lab[A, B], List[Char]])],
                      List[(term[lab[A, B], List[Char]],
                             term[lab[A, B],
                                   List[Char]])]](List('i', 'n', 's', 't', 'a',
                'n', 't', 'i', 'a', 't', 'i', 'o', 'n', 's'),
           rules[lab[A, B]](xml2name),
           id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
           ab),
       (ab: xml) =>
         xml1element[List[(term[lab[A, B], List[Char]],
                            term[lab[A, B], List[Char]])],
                      List[(term[lab[A, B], List[Char]],
                             term[lab[A, B],
                                   List[Char]])]](List('u', 's', 'a', 'b', 'l',
                'e', 'R', 'u', 'l', 'e', 's'),
           rules[lab[A, B]](xml2name),
           id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
           ab),
       (ab: xml) => xml2dp_proof[A, B](xml2name, ab),
       (ab: (term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])) =>
         (b: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])])
           => (c: Option[List[(term[lab[A, B], List[Char]],
                                term[lab[A, B], List[Char]])]])
                => (d: dp_proof[A, B, List[Char]]) =>
                     ForwardInstantiationProc[A, B, List[Char]](ab, b, c, d),
       aa)),
  (List('s', 'e', 'm', 'l', 'a', 'b', 'P', 'r', 'o', 'c'),
    (aa: xml) =>
      xml4elements[sl_variant[lab[A, B], List[Char]],
                    List[(term[lab[A, B], List[Char]],
                           term[lab[A, B], List[Char]])],
                    List[(term[lab[A, B], List[Char]],
                           term[lab[A, B], List[Char]])],
                    dp_proof[A, B, List[Char]],
                    dp_proof[A, B,
                              List[Char]]](List('s', 'e', 'm', 'l', 'a', 'b',
         'P', 'r', 'o', 'c'),
    sl_variant[lab[A, B]](xml2name),
    (ab: xml) =>
      xml1element[List[(term[lab[A, B], List[Char]],
                         term[lab[A, B], List[Char]])],
                   List[(term[lab[A, B], List[Char]],
                          term[lab[A, B],
                                List[Char]])]](List('d', 'p', 's'),
        rules[lab[A, B]](xml2name),
        id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
        ab),
    (ab: xml) =>
      xml1element[List[(term[lab[A, B], List[Char]],
                         term[lab[A, B], List[Char]])],
                   List[(term[lab[A, B], List[Char]],
                          term[lab[A, B],
                                List[Char]])]](List('t', 'r', 's'),
        rules[lab[A, B]](xml2name),
        id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
        ab),
    (ab: xml) => xml2dp_proof[A, B](xml2name, ab),
    (sli: sl_variant[lab[A, B], List[Char]]) =>
      (lp: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]) =>
        (ab: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])])
          => (b: dp_proof[A, B, List[Char]]) =>
               SemlabProc[A, B, List[Char]](sli, lp, Nil, ab, b),
    aa)),
  (List('s', 'u', 'b', 't', 'e', 'r', 'm', 'P', 'r', 'o', 'c'),
    (aa: xml) =>
      xml1many2elements_gen[projL[lab[A, B]],
                             ((term[lab[A, B], List[Char]],
                                term[lab[A, B], List[Char]]),
                               List[(pos, ((term[lab[A, B], List[Char]],
     term[lab[A, B], List[Char]]),
    term[lab[A, B], List[Char]]))]),
                             List[(term[lab[A, B], List[Char]],
                                    term[lab[A, B], List[Char]])],
                             dp_proof[A, B, List[Char]],
                             dp_proof[A, B,
                                       List[Char]]](List('s', 'u', 'b', 't',
                  'e', 'r', 'm', 'P', 'r', 'o', 'c'),
             proj[lab[A, B]](xml2name),
             (pi: projL[lab[A, B]]) =>
               (ab: xml) =>
                 projected_rseq[lab[A, B]](xml2name,
    (ac: (lab[A, B], nat)) => create_proj[lab[A, B]](pi, ac), ab),
             (ab: xml) =>
               xml1element[List[(term[lab[A, B], List[Char]],
                                  term[lab[A, B], List[Char]])],
                            List[(term[lab[A, B], List[Char]],
                                   term[lab[A, B],
 List[Char]])]](List('d', 'p', 's'), rules[lab[A, B]](xml2name),
                 id[List[(term[lab[A, B], List[Char]],
                           term[lab[A, B], List[Char]])]],
                 ab),
             (ab: xml) => xml2dp_proof[A, B](xml2name, ab),
             (ab: projL[lab[A, B]]) =>
               (b: List[((term[lab[A, B], List[Char]],
                           term[lab[A, B], List[Char]]),
                          List[(pos, ((term[lab[A, B], List[Char]],
term[lab[A, B], List[Char]]),
                                       term[lab[A, B], List[Char]]))])])
                 => (c: List[(term[lab[A, B], List[Char]],
                               term[lab[A, B], List[Char]])])
                      => (d: dp_proof[A, B, List[Char]]) =>
                           SubtermCriterionProc[A, B, List[Char]](ab, b, c, d),
             aa)),
  (List('r', 'e', 'd', 'P', 'a', 'i', 'r', 'U', 'r', 'P', 'r', 'o', 'c'),
    (aa: xml) =>
      xml4elements[sum[root_redtriple_impl[lab[A, B]],
                        redtriple_impl[lab[A, B]]],
                    List[(term[lab[A, B], List[Char]],
                           term[lab[A, B], List[Char]])],
                    List[(term[lab[A, B], List[Char]],
                           term[lab[A, B], List[Char]])],
                    dp_proof[A, B, List[Char]],
                    dp_proof[A, B,
                              List[Char]]](List('r', 'e', 'd', 'P', 'a', 'i',
         'r', 'U', 'r', 'P', 'r', 'o', 'c'),
    ordering_constraint_proof[lab[A, B]](xml2name),
    (ab: xml) =>
      xml1element[List[(term[lab[A, B], List[Char]],
                         term[lab[A, B], List[Char]])],
                   List[(term[lab[A, B], List[Char]],
                          term[lab[A, B],
                                List[Char]])]](List('d', 'p', 's'),
        rules[lab[A, B]](xml2name),
        id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
        ab),
    (ab: xml) =>
      xml1element[List[(term[lab[A, B], List[Char]],
                         term[lab[A, B], List[Char]])],
                   List[(term[lab[A, B], List[Char]],
                          term[lab[A, B],
                                List[Char]])]](List('u', 's', 'a', 'b', 'l',
             'e', 'R', 'u', 'l', 'e', 's'),
        rules[lab[A, B]](xml2name),
        id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
        ab),
    (ab: xml) => xml2dp_proof[A, B](xml2name, ab),
    (ab: sum[root_redtriple_impl[lab[A, B]], redtriple_impl[lab[A, B]]]) =>
      (b: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]) =>
        (c: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]) =>
          (d: dp_proof[A, B, List[Char]]) =>
            RedPairUrProc[A, B, List[Char]](ab, b, c, d),
    aa)),
  (List('m', 'o', 'n', 'o', 'R', 'e', 'd', 'P', 'a', 'i', 'r', 'U', 'r', 'P',
         'r', 'o', 'c'),
    (aa: xml) =>
      xml5elements[redtriple_impl[lab[A, B]],
                    List[(term[lab[A, B], List[Char]],
                           term[lab[A, B], List[Char]])],
                    List[(term[lab[A, B], List[Char]],
                           term[lab[A, B], List[Char]])],
                    List[(term[lab[A, B], List[Char]],
                           term[lab[A, B], List[Char]])],
                    dp_proof[A, B, List[Char]],
                    dp_proof[A, B,
                              List[Char]]](List('m', 'o', 'n', 'o', 'R', 'e',
         'd', 'P', 'a', 'i', 'r', 'U', 'r', 'P', 'r', 'o', 'c'),
    ordering_constraint_proofa[lab[A, B]](xml2name),
    (ab: xml) =>
      xml1element[List[(term[lab[A, B], List[Char]],
                         term[lab[A, B], List[Char]])],
                   List[(term[lab[A, B], List[Char]],
                          term[lab[A, B],
                                List[Char]])]](List('d', 'p', 's'),
        rules[lab[A, B]](xml2name),
        id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
        ab),
    (ab: xml) =>
      xml1element[List[(term[lab[A, B], List[Char]],
                         term[lab[A, B], List[Char]])],
                   List[(term[lab[A, B], List[Char]],
                          term[lab[A, B],
                                List[Char]])]](List('t', 'r', 's'),
        rules[lab[A, B]](xml2name),
        id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
        ab),
    (ab: xml) =>
      xml1element[List[(term[lab[A, B], List[Char]],
                         term[lab[A, B], List[Char]])],
                   List[(term[lab[A, B], List[Char]],
                          term[lab[A, B],
                                List[Char]])]](List('u', 's', 'a', 'b', 'l',
             'e', 'R', 'u', 'l', 'e', 's'),
        rules[lab[A, B]](xml2name),
        id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
        ab),
    (ab: xml) => xml2dp_proof[A, B](xml2name, ab),
    (ab: redtriple_impl[lab[A, B]]) =>
      (b: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]) =>
        (c: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]) =>
          (d: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])])
            => (e: dp_proof[A, B, List[Char]]) =>
                 MonoRedPairUrProc[A, B, List[Char]](ab, b, c, d, e),
    aa)),
  (List('m', 'o', 'n', 'o', 'R', 'e', 'd', 'P', 'a', 'i', 'r', 'P', 'r', 'o',
         'c'),
    (aa: xml) =>
      xml4elements[redtriple_impl[lab[A, B]],
                    List[(term[lab[A, B], List[Char]],
                           term[lab[A, B], List[Char]])],
                    List[(term[lab[A, B], List[Char]],
                           term[lab[A, B], List[Char]])],
                    dp_proof[A, B, List[Char]],
                    dp_proof[A, B,
                              List[Char]]](List('m', 'o', 'n', 'o', 'R', 'e',
         'd', 'P', 'a', 'i', 'r', 'P', 'r', 'o', 'c'),
    ordering_constraint_proofa[lab[A, B]](xml2name),
    (ab: xml) =>
      xml1element[List[(term[lab[A, B], List[Char]],
                         term[lab[A, B], List[Char]])],
                   List[(term[lab[A, B], List[Char]],
                          term[lab[A, B],
                                List[Char]])]](List('d', 'p', 's'),
        rules[lab[A, B]](xml2name),
        id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
        ab),
    (ab: xml) =>
      xml1element[List[(term[lab[A, B], List[Char]],
                         term[lab[A, B], List[Char]])],
                   List[(term[lab[A, B], List[Char]],
                          term[lab[A, B],
                                List[Char]])]](List('t', 'r', 's'),
        rules[lab[A, B]](xml2name),
        id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
        ab),
    (ab: xml) => xml2dp_proof[A, B](xml2name, ab),
    (ab: redtriple_impl[lab[A, B]]) =>
      (b: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]) =>
        (c: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]) =>
          (d: dp_proof[A, B, List[Char]]) =>
            MonoRedPairProc[A, B, List[Char]](ab, b, c, d),
    aa)),
  (List('u', 'n', 'c', 'u', 'r', 'r', 'y', 'P', 'r', 'o', 'c'),
    (aa: xml) =>
      xml_choice[dp_proof[A, B,
                           List[Char]]](List('u', 'n', 'c', 'u', 'r', 'r', 'y',
      'P', 'r', 'o', 'c'),
 List((ab: xml) =>
        xml4elements[(lab[A, B],
                       (List[((lab[A, B], nat), List[lab[A, B]])],
                         (List[(term[lab[A, B], List[Char]],
                                 term[lab[A, B], List[Char]])],
                           List[(term[lab[A, B], List[Char]],
                                  term[lab[A, B], List[Char]])]))),
                      List[(term[lab[A, B], List[Char]],
                             term[lab[A, B], List[Char]])],
                      List[(term[lab[A, B], List[Char]],
                             term[lab[A, B], List[Char]])],
                      dp_proof[A, B, List[Char]],
                      dp_proof[A, B,
                                List[Char]]](List('u', 'n', 'c', 'u', 'r', 'r',
           'y', 'P', 'r', 'o', 'c'),
      uncurry_info[lab[A, B]](xml2name),
      (ac: xml) =>
        xml1element[List[(term[lab[A, B], List[Char]],
                           term[lab[A, B], List[Char]])],
                     List[(term[lab[A, B], List[Char]],
                            term[lab[A, B],
                                  List[Char]])]](List('d', 'p', 's'),
          rules[lab[A, B]](xml2name),
          id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
          ac),
      (ac: xml) =>
        xml1element[List[(term[lab[A, B], List[Char]],
                           term[lab[A, B], List[Char]])],
                     List[(term[lab[A, B], List[Char]],
                            term[lab[A, B],
                                  List[Char]])]](List('t', 'r', 's'),
          rules[lab[A, B]](xml2name),
          id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
          ac),
      (ac: xml) => xml2dp_proof[A, B](xml2name, ac),
      (ac: (lab[A, B],
             (List[((lab[A, B], nat), List[lab[A, B]])],
               (List[(term[lab[A, B], List[Char]],
                       term[lab[A, B], List[Char]])],
                 List[(term[lab[A, B], List[Char]],
                        term[lab[A, B], List[Char]])]))))
        => (b: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])])
             => (c: List[(term[lab[A, B], List[Char]],
                           term[lab[A, B], List[Char]])])
                  => (d: dp_proof[A, B, List[Char]]) =>
                       UncurryProc[A, B, List[Char]](None, ac, b, c, d),
      ab),
       (ab: xml) =>
         xml5elements[nat, (lab[A, B],
                             (List[((lab[A, B], nat), List[lab[A, B]])],
                               (List[(term[lab[A, B], List[Char]],
                                       term[lab[A, B], List[Char]])],
                                 List[(term[lab[A, B], List[Char]],
term[lab[A, B], List[Char]])]))),
                       List[(term[lab[A, B], List[Char]],
                              term[lab[A, B], List[Char]])],
                       List[(term[lab[A, B], List[Char]],
                              term[lab[A, B], List[Char]])],
                       dp_proof[A, B, List[Char]],
                       dp_proof[A, B,
                                 List[Char]]](List('u', 'n', 'c', 'u', 'r', 'r',
            'y', 'P', 'r', 'o', 'c'),
       (ac: xml) =>
         nata(List('a', 'p', 'p', 'l', 'i', 'c', 'a', 't', 'i', 'v', 'e', 'T',
                    'o', 'p'),
               ac),
       uncurry_info[lab[A, B]](xml2name),
       (ac: xml) =>
         xml1element[List[(term[lab[A, B], List[Char]],
                            term[lab[A, B], List[Char]])],
                      List[(term[lab[A, B], List[Char]],
                             term[lab[A, B],
                                   List[Char]])]](List('d', 'p', 's'),
           rules[lab[A, B]](xml2name),
           id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
           ac),
       (ac: xml) =>
         xml1element[List[(term[lab[A, B], List[Char]],
                            term[lab[A, B], List[Char]])],
                      List[(term[lab[A, B], List[Char]],
                             term[lab[A, B],
                                   List[Char]])]](List('t', 'r', 's'),
           rules[lab[A, B]](xml2name),
           id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
           ac),
       (ac: xml) => xml2dp_proof[A, B](xml2name, ac),
       (n: nat) =>
         (ac: (lab[A, B],
                (List[((lab[A, B], nat), List[lab[A, B]])],
                  (List[(term[lab[A, B], List[Char]],
                          term[lab[A, B], List[Char]])],
                    List[(term[lab[A, B], List[Char]],
                           term[lab[A, B], List[Char]])]))))
           => (b: List[(term[lab[A, B], List[Char]],
                         term[lab[A, B], List[Char]])])
                => (c: List[(term[lab[A, B], List[Char]],
                              term[lab[A, B], List[Char]])])
                     => (d: dp_proof[A, B, List[Char]]) =>
                          UncurryProc[A, B,
                                       List[Char]](Some[nat](n), ac, b, c, d),
       ab)),
 aa)),
  (List('f', 'l', 'a', 't', 'C', 'o', 'n', 't', 'e', 'x', 't', 'C', 'l', 'o',
         's', 'u', 'r', 'e', 'P', 'r', 'o', 'c'),
    (aa: xml) =>
      xml5elements[lab[A, B], List[ctxt[lab[A, B], List[Char]]],
                    List[(term[lab[A, B], List[Char]],
                           term[lab[A, B], List[Char]])],
                    List[(term[lab[A, B], List[Char]],
                           term[lab[A, B], List[Char]])],
                    dp_proof[A, B, List[Char]],
                    dp_proof[A, B,
                              List[Char]]](List('f', 'l', 'a', 't', 'C', 'o',
         'n', 't', 'e', 'x', 't', 'C', 'l', 'o', 's', 'u', 'r', 'e', 'P', 'r',
         'o', 'c'),
    (ab: xml) =>
      xml1element[lab[A, B],
                   lab[A, B]](List('f', 'r', 'e', 's', 'h', 'S', 'y', 'm', 'b',
                                    'o', 'l'),
                               xml2name, id[lab[A, B]], ab),
    flat_contexts[lab[A, B]](xml2name),
    (ab: xml) =>
      xml1element[List[(term[lab[A, B], List[Char]],
                         term[lab[A, B], List[Char]])],
                   List[(term[lab[A, B], List[Char]],
                          term[lab[A, B],
                                List[Char]])]](List('d', 'p', 's'),
        rules[lab[A, B]](xml2name),
        id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
        ab),
    (ab: xml) =>
      xml1element[List[(term[lab[A, B], List[Char]],
                         term[lab[A, B], List[Char]])],
                   List[(term[lab[A, B], List[Char]],
                          term[lab[A, B],
                                List[Char]])]](List('t', 'r', 's'),
        rules[lab[A, B]](xml2name),
        id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
        ab),
    (ab: xml) => xml2dp_proof[A, B](xml2name, ab),
    (ab: lab[A, B]) =>
      (b: List[ctxt[lab[A, B], List[Char]]]) =>
        (c: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]) =>
          (d: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])])
            => (e: dp_proof[A, B, List[Char]]) =>
                 FccProc[A, B, List[Char]](ab, b, c, d, e),
    aa)),
  (List('s', 'w', 'i', 't', 'c', 'h', 'I', 'n', 'n', 'e', 'r', 'm', 'o', 's',
         't', 'P', 'r', 'o', 'c'),
    (aa: xml) =>
      xml2elements[Option[List[(term[lab[A, B], List[Char]],
                                 (List[(pos,
 ((term[lab[A, B], List[Char]], term[lab[A, B], List[Char]]),
   term[lab[A, B], List[Char]]))],
                                   (term[lab[A, B], List[Char]],
                                     List[(pos,
    ((term[lab[A, B], List[Char]], term[lab[A, B], List[Char]]),
      term[lab[A, B], List[Char]]))])))]],
                    dp_proof[A, B, List[Char]],
                    dp_proof[A, B,
                              List[Char]]](List('s', 'w', 'i', 't', 'c', 'h',
         'I', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', 'P', 'r', 'o', 'c'),
    wcr_proof[lab[A, B]](xml2name),
    (ab: xml) => xml2dp_proof[A, B](xml2name, ab),
    (ab: Option[List[(term[lab[A, B], List[Char]],
                       (List[(pos, ((term[lab[A, B], List[Char]],
                                      term[lab[A, B], List[Char]]),
                                     term[lab[A, B], List[Char]]))],
                         (term[lab[A, B], List[Char]],
                           List[(pos, ((term[lab[A, B], List[Char]],
 term[lab[A, B], List[Char]]),
term[lab[A, B], List[Char]]))])))]])
      => (b: dp_proof[A, B, List[Char]]) =>
           SwitchInnermostProc[A, B, List[Char]](ab, b),
    aa)),
  (List('s', 'p', 'l', 'i', 't', 'P', 'r', 'o', 'c'),
    (aa: xml) =>
      xml4elements[List[(term[lab[A, B], List[Char]],
                          term[lab[A, B], List[Char]])],
                    List[(term[lab[A, B], List[Char]],
                           term[lab[A, B], List[Char]])],
                    dp_proof[A, B, List[Char]], dp_proof[A, B, List[Char]],
                    dp_proof[A, B,
                              List[Char]]](List('s', 'p', 'l', 'i', 't', 'P',
         'r', 'o', 'c'),
    (ab: xml) =>
      xml1element[List[(term[lab[A, B], List[Char]],
                         term[lab[A, B], List[Char]])],
                   List[(term[lab[A, B], List[Char]],
                          term[lab[A, B],
                                List[Char]])]](List('d', 'p', 's'),
        rules[lab[A, B]](xml2name),
        id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
        ab),
    (ab: xml) =>
      xml1element[List[(term[lab[A, B], List[Char]],
                         term[lab[A, B], List[Char]])],
                   List[(term[lab[A, B], List[Char]],
                          term[lab[A, B],
                                List[Char]])]](List('t', 'r', 's'),
        rules[lab[A, B]](xml2name),
        id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
        ab),
    (ab: xml) => xml2dp_proof[A, B](xml2name, ab),
    (ab: xml) => xml2dp_proof[A, B](xml2name, ab),
    (ab: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]) =>
      (b: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]) =>
        (c: dp_proof[A, B, List[Char]]) =>
          (d: dp_proof[A, B, List[Char]]) =>
            SplitProc[A, B, List[Char]](ab, b, c, d),
    aa)),
  (List('f', 'i', 'n', 'i', 't', 'e', 'n', 'e', 's', 's', 'A', 's', 's', 'u',
         'm', 'p', 't', 'i', 'o', 'n'),
    (aa: xml) =>
      xml1element[(List[(term[lab[A, B], List[Char]],
                          term[lab[A, B], List[Char]])],
                    (strategy[lab[A, B], List[Char]],
                      List[(term[lab[A, B], List[Char]],
                             term[lab[A, B], List[Char]])])),
                   dp_proof[A, B,
                             List[Char]]](List('f', 'i', 'n', 'i', 't', 'e',
        'n', 'e', 's', 's', 'A', 's', 's', 'u', 'm', 'p', 't', 'i', 'o', 'n'),
   xml2dp_input[lab[A, B]](xml2name),
   (ab: (List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])],
          (strategy[lab[A, B], List[Char]],
            List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])])))
     => {
          val (p, (q, r)):
                (List[(term[lab[A, B], List[Char]],
                        term[lab[A, B], List[Char]])],
                  (strategy[lab[A, B], List[Char]],
                    List[(term[lab[A, B], List[Char]],
                           term[lab[A, B], List[Char]])]))
            = ab;
          AssumeFinite[A, B,
                        List[Char]]((p, (Nil,
  (strategy_to_Q[lab[A, B], List[Char]](q, r), (Nil, r)))))
        },
   aa)),
  (List('s', 'i', 'z', 'e', 'C', 'h', 'a', 'n', 'g', 'e', 'P', 'r', 'o', 'c'),
    (xml1many_elements[Option[(redtriple_impl[lab[A, B]],
                                List[(term[lab[A, B], List[Char]],
                                       term[lab[A, B], List[Char]])])],
                        ((term[lab[A, B], List[Char]],
                           term[lab[A, B], List[Char]]),
                          (List[(nat, nat)], List[(nat, nat)])),
                        dp_proof[A, B,
                                  List[Char]]](List('s', 'i', 'z', 'e', 'C',
             'h', 'a', 'n', 'g', 'e', 'P', 'r', 'o', 'c'),
        (aa: xml) =>
          xml_options[Option[(redtriple_impl[lab[A, B]],
                               List[(term[lab[A, B], List[Char]],
                                      term[lab[A, B],
    List[Char]])])]](List((List('s', 'u', 'b', 't', 'e', 'r', 'm', 'C', 'r',
                                 'i', 't', 'e', 'r', 'i', 'o', 'n'),
                            (ab: xml) =>
                              xml0elements[Option[(redtriple_impl[lab[A, B]],
            List[(term[lab[A, B], List[Char]],
                   term[lab[A, B],
                         List[Char]])])]](List('s', 'u', 'b', 't', 'e', 'r',
        'm', 'C', 'r', 'i', 't', 'e', 'r', 'i', 'o', 'n'),
   None, ab)),
                           (List('r', 'e', 'd', 'u', 'c', 't', 'i', 'o', 'n',
                                  'P', 'a', 'i', 'r'),
                             (ab: xml) =>
                               xml2elements[redtriple_impl[lab[A, B]],
     List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])],
     Option[(redtriple_impl[lab[A, B]],
              List[(term[lab[A, B], List[Char]],
                     term[lab[A, B],
                           List[Char]])])]](List('r', 'e', 'd', 'u', 'c', 't',
          'i', 'o', 'n', 'P', 'a', 'i', 'r'),
     ordering_constraint_proofa[lab[A, B]](xml2name),
     (ac: xml) =>
       xml1element[List[(term[lab[A, B], List[Char]],
                          term[lab[A, B], List[Char]])],
                    List[(term[lab[A, B], List[Char]],
                           term[lab[A, B],
                                 List[Char]])]](List('u', 's', 'a', 'b', 'l',
              'e', 'R', 'u', 'l', 'e', 's'),
         rules[lab[A, B]](xml2name),
         id[List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]],
         ac),
     (redp: redtriple_impl[lab[A, B]]) =>
       (ur: List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])]) =>
         Some[(redtriple_impl[lab[A, B]],
                List[(term[lab[A, B], List[Char]],
                       term[lab[A, B], List[Char]])])]((redp, ur)),
     ab))),
                      aa),
        scg[lab[A, B]](xml2name))).apply((aa:
    Option[(redtriple_impl[lab[A, B]],
             List[(term[lab[A, B], List[Char]], term[lab[A, B], List[Char]])])])
   => aa match {
        case None =>
          (ab: List[((term[lab[A, B], List[Char]], term[lab[A, B], List[Char]]),
                      (List[(nat, nat)], List[(nat, nat)]))])
            => SizeChangeSubterm[A, B, List[Char]](ab)
        case Some(redp_ur) =>
          (ab: List[((term[lab[A, B], List[Char]], term[lab[A, B], List[Char]]),
                      (List[(nat, nat)], List[(nat, nat)]))])
            => SizeChangeRedPair[A, B,
                                  List[Char]](fst[redtriple_impl[lab[A, B]],
           List[(term[lab[A, B], List[Char]],
                  term[lab[A, B], List[Char]])]](redp_ur),
       snd[redtriple_impl[lab[A, B]],
            List[(term[lab[A, B], List[Char]],
                   term[lab[A, B], List[Char]])]](redp_ur),
       ab)
      }))),
                                     a),
                                       id[dp_proof[A, B, List[Char]]], x)

def xml2trs_proof[A : equal](xml2name:
                               xml => sum[List[Char], lab[A, List[nat]]],
                              x: xml):
      sum[List[Char], trs_proof[A, List[nat], List[Char]]] =
  xml1element[trs_proof[A, List[nat], List[Char]],
               trs_proof[A, List[nat],
                          List[Char]]](List('t', 'r', 's', 'T', 'e', 'r', 'm',
     'i', 'n', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f'),
(a: xml) =>
  xml_options[trs_proof[A, List[nat],
                         List[Char]]](List((List('r', 'I', 's', 'E', 'm', 'p',
          't', 'y'),
     (aa: xml) =>
       xml0elements[trs_proof[A, List[nat],
                               List[Char]]](List('r', 'I', 's', 'E', 'm', 'p',
          't', 'y'),
     RisEmpty, aa)),
    (List('s', 'e', 'm', 'l', 'a', 'b'),
      (aa: xml) =>
        xml3elements[sl_variant[lab[A, List[nat]], List[Char]],
                      List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]], List[Char]])],
                      trs_proof[A, List[nat], List[Char]],
                      trs_proof[A, List[nat],
                                 List[Char]]](List('s', 'e', 'm', 'l', 'a',
            'b'),
       sl_variant[lab[A, List[nat]]](xml2name),
       (ab: xml) =>
         xml1element[List[(term[lab[A, List[nat]], List[Char]],
                            term[lab[A, List[nat]], List[Char]])],
                      List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]],
                                   List[Char]])]](List('t', 'r', 's'),
           rules[lab[A, List[nat]]](xml2name),
           id[List[(term[lab[A, List[nat]], List[Char]],
                     term[lab[A, List[nat]], List[Char]])]],
           ab),
       (ab: xml) => xml2trs_proof[A](xml2name, ab),
       (sli: sl_variant[lab[A, List[nat]], List[Char]]) =>
         (ab: List[(term[lab[A, List[nat]], List[Char]],
                     term[lab[A, List[nat]], List[Char]])])
           => (b: trs_proof[A, List[nat], List[Char]]) =>
                Semlab[A, List[nat], List[Char]](sli, Nil, ab, b),
       aa)),
    (List('s', 'p', 'l', 'i', 't'),
      (aa: xml) =>
        xml3elements[List[(term[lab[A, List[nat]], List[Char]],
                            term[lab[A, List[nat]], List[Char]])],
                      trs_proof[A, List[nat], List[Char]],
                      trs_proof[A, List[nat], List[Char]],
                      trs_proof[A, List[nat],
                                 List[Char]]](List('s', 'p', 'l', 'i', 't'),
       (ab: xml) =>
         xml1element[List[(term[lab[A, List[nat]], List[Char]],
                            term[lab[A, List[nat]], List[Char]])],
                      List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]],
                                   List[Char]])]](List('t', 'r', 's'),
           rules[lab[A, List[nat]]](xml2name),
           id[List[(term[lab[A, List[nat]], List[Char]],
                     term[lab[A, List[nat]], List[Char]])]],
           ab),
       (ab: xml) => xml2trs_proof[A](xml2name, ab),
       (ab: xml) => xml2trs_proof[A](xml2name, ab),
       (ab: List[(term[lab[A, List[nat]], List[Char]],
                   term[lab[A, List[nat]], List[Char]])])
         => (b: trs_proof[A, List[nat], List[Char]]) =>
              (c: trs_proof[A, List[nat], List[Char]]) =>
                Split[A, List[nat], List[Char]](ab, b, c),
       aa)),
    (List('d', 'p', 'T', 'r', 'a', 'n', 's'),
      (aa: xml) =>
        xml3elements[List[(term[lab[A, List[nat]], List[Char]],
                            term[lab[A, List[nat]], List[Char]])],
                      Boolean, dp_proof[A, List[nat], List[Char]],
                      trs_proof[A, List[nat],
                                 List[Char]]](List('d', 'p', 'T', 'r', 'a', 'n',
            's'),
       (ab: xml) =>
         xml1element[List[(term[lab[A, List[nat]], List[Char]],
                            term[lab[A, List[nat]], List[Char]])],
                      List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]],
                                   List[Char]])]](List('d', 'p', 's'),
           rules[lab[A, List[nat]]](xml2name),
           id[List[(term[lab[A, List[nat]], List[Char]],
                     term[lab[A, List[nat]], List[Char]])]],
           ab),
       (ab: xml) =>
         bool(List('m', 'a', 'r', 'k', 'e', 'd', 'S', 'y', 'm', 'b', 'o', 'l',
                    's'),
               ab),
       (ab: xml) => xml2dp_proof[A, List[nat]](xml2name, ab),
       (dps: List[(term[lab[A, List[nat]], List[Char]],
                    term[lab[A, List[nat]], List[Char]])])
         => (_: Boolean) =>
              (ab: dp_proof[A, List[nat], List[Char]]) =>
                DPTrans[A, List[nat], List[Char]](dps, ab),
       aa)),
    (List('r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l'),
      (aa: xml) =>
        xml3elements[redtriple_impl[lab[A, List[nat]]],
                      List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]], List[Char]])],
                      trs_proof[A, List[nat], List[Char]],
                      trs_proof[A, List[nat],
                                 List[Char]]](List('r', 'u', 'l', 'e', 'R', 'e',
            'm', 'o', 'v', 'a', 'l'),
       ordering_constraint_proofa[lab[A, List[nat]]](xml2name),
       (ab: xml) =>
         xml1element[List[(term[lab[A, List[nat]], List[Char]],
                            term[lab[A, List[nat]], List[Char]])],
                      List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]],
                                   List[Char]])]](List('t', 'r', 's'),
           rules[lab[A, List[nat]]](xml2name),
           id[List[(term[lab[A, List[nat]], List[Char]],
                     term[lab[A, List[nat]], List[Char]])]],
           ab),
       (ab: xml) => xml2trs_proof[A](xml2name, ab),
       (ab: redtriple_impl[lab[A, List[nat]]]) =>
         (b: List[(term[lab[A, List[nat]], List[Char]],
                    term[lab[A, List[nat]], List[Char]])])
           => (c: trs_proof[A, List[nat], List[Char]]) =>
                RuleRemoval[A, List[nat], List[Char]](ab, b, c),
       aa)),
    (List('b', 'o', 'u', 'n', 'd', 's'),
      (xml: xml) =>
        bindb[List[Char], bounds_info[lab[A, List[nat]], List[Char]],
               trs_proof[A, List[nat],
                          List[Char]]]((bounds_info[lab[A,
                 List[nat]]](xml2name)).apply(xml),
comp[trs_proof[A, List[nat], List[Char]],
      sum[List[Char], trs_proof[A, List[nat], List[Char]]],
      bounds_info[lab[A, List[nat]],
                   List[Char]]]((aa: trs_proof[A, List[nat], List[Char]]) =>
                                  Inr[trs_proof[A, List[nat], List[Char]],
                                       List[Char]](aa),
                                 (aa: bounds_info[lab[A, List[nat]],
           List[Char]])
                                   => Bounds[A, List[nat], List[Char]](aa)))),
    (List('s', 't', 'r', 'i', 'n', 'g', 'R', 'e', 'v', 'e', 'r', 's', 'a', 'l'),
      (aa: xml) =>
        xml2elements[List[(term[lab[A, List[nat]], List[Char]],
                            term[lab[A, List[nat]], List[Char]])],
                      trs_proof[A, List[nat], List[Char]],
                      trs_proof[A, List[nat],
                                 List[Char]]](List('s', 't', 'r', 'i', 'n', 'g',
            'R', 'e', 'v', 'e', 'r', 's', 'a', 'l'),
       (ab: xml) =>
         xml1element[List[(term[lab[A, List[nat]], List[Char]],
                            term[lab[A, List[nat]], List[Char]])],
                      List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]],
                                   List[Char]])]](List('t', 'r', 's'),
           rules[lab[A, List[nat]]](xml2name),
           id[List[(term[lab[A, List[nat]], List[Char]],
                     term[lab[A, List[nat]], List[Char]])]],
           ab),
       (ab: xml) => xml2trs_proof[A](xml2name, ab),
       (_: List[(term[lab[A, List[nat]], List[Char]],
                  term[lab[A, List[nat]], List[Char]])])
         => (ab: trs_proof[A, List[nat], List[Char]]) =>
              StringReversal[A, List[nat], List[Char]](ab),
       aa)),
    (List('u', 'n', 'c', 'u', 'r', 'r', 'y'),
      (aa: xml) =>
        xml3elements[(lab[A, List[nat]],
                       (List[((lab[A, List[nat]], nat),
                               List[lab[A, List[nat]]])],
                         (List[(term[lab[A, List[nat]], List[Char]],
                                 term[lab[A, List[nat]], List[Char]])],
                           List[(term[lab[A, List[nat]], List[Char]],
                                  term[lab[A, List[nat]], List[Char]])]))),
                      List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]], List[Char]])],
                      trs_proof[A, List[nat], List[Char]],
                      trs_proof[A, List[nat],
                                 List[Char]]](List('u', 'n', 'c', 'u', 'r', 'r',
            'y'),
       uncurry_info[lab[A, List[nat]]](xml2name),
       (ab: xml) =>
         xml1element[List[(term[lab[A, List[nat]], List[Char]],
                            term[lab[A, List[nat]], List[Char]])],
                      List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]],
                                   List[Char]])]](List('t', 'r', 's'),
           rules[lab[A, List[nat]]](xml2name),
           id[List[(term[lab[A, List[nat]], List[Char]],
                     term[lab[A, List[nat]], List[Char]])]],
           ab),
       (ab: xml) => xml2trs_proof[A](xml2name, ab),
       (ab: (lab[A, List[nat]],
              (List[((lab[A, List[nat]], nat), List[lab[A, List[nat]]])],
                (List[(term[lab[A, List[nat]], List[Char]],
                        term[lab[A, List[nat]], List[Char]])],
                  List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])]))))
         => (b: List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]], List[Char]])])
              => (c: trs_proof[A, List[nat], List[Char]]) =>
                   Uncurry[A, List[nat], List[Char]](ab, b, c),
       aa)),
    (List('f', 'l', 'a', 't', 'C', 'o', 'n', 't', 'e', 'x', 't', 'C', 'l', 'o',
           's', 'u', 'r', 'e'),
      (aa: xml) =>
        xml3elements[List[ctxt[lab[A, List[nat]], List[Char]]],
                      List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]], List[Char]])],
                      trs_proof[A, List[nat], List[Char]],
                      trs_proof[A, List[nat],
                                 List[Char]]](List('f', 'l', 'a', 't', 'C', 'o',
            'n', 't', 'e', 'x', 't', 'C', 'l', 'o', 's', 'u', 'r', 'e'),
       flat_contexts[lab[A, List[nat]]](xml2name),
       (ab: xml) =>
         xml1element[List[(term[lab[A, List[nat]], List[Char]],
                            term[lab[A, List[nat]], List[Char]])],
                      List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]],
                                   List[Char]])]](List('t', 'r', 's'),
           rules[lab[A, List[nat]]](xml2name),
           id[List[(term[lab[A, List[nat]], List[Char]],
                     term[lab[A, List[nat]], List[Char]])]],
           ab),
       (ab: xml) => xml2trs_proof[A](xml2name, ab),
       (ab: List[ctxt[lab[A, List[nat]], List[Char]]]) =>
         (b: List[(term[lab[A, List[nat]], List[Char]],
                    term[lab[A, List[nat]], List[Char]])])
           => (c: trs_proof[A, List[nat], List[Char]]) =>
                Fcc[A, List[nat], List[Char]](ab, b, c),
       aa)),
    (List('s', 'w', 'i', 't', 'c', 'h', 'I', 'n', 'n', 'e', 'r', 'm', 'o', 's',
           't'),
      (aa: xml) =>
        xml2elements[Option[List[(term[lab[A, List[nat]], List[Char]],
                                   (List[(pos,
   ((term[lab[A, List[nat]], List[Char]], term[lab[A, List[nat]], List[Char]]),
     term[lab[A, List[nat]], List[Char]]))],
                                     (term[lab[A, List[nat]], List[Char]],
                                       List[(pos,
      ((term[lab[A, List[nat]], List[Char]],
         term[lab[A, List[nat]], List[Char]]),
        term[lab[A, List[nat]], List[Char]]))])))]],
                      trs_proof[A, List[nat], List[Char]],
                      trs_proof[A, List[nat],
                                 List[Char]]](List('s', 'w', 'i', 't', 'c', 'h',
            'I', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't'),
       wcr_proof[lab[A, List[nat]]](xml2name),
       (ab: xml) => xml2trs_proof[A](xml2name, ab),
       (ab: Option[List[(term[lab[A, List[nat]], List[Char]],
                          (List[(pos, ((term[lab[A, List[nat]], List[Char]],
 term[lab[A, List[nat]], List[Char]]),
term[lab[A, List[nat]], List[Char]]))],
                            (term[lab[A, List[nat]], List[Char]],
                              List[(pos, ((term[lab[A, List[nat]], List[Char]],
    term[lab[A, List[nat]], List[Char]]),
   term[lab[A, List[nat]], List[Char]]))])))]])
         => (b: trs_proof[A, List[nat], List[Char]]) =>
              SwitchInnermost[A, List[nat], List[Char]](ab, b),
       aa)),
    (List('t', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', 'A', 's', 's',
           'u', 'm', 'p', 't', 'i', 'o', 'n'),
      (aa: xml) =>
        xml1element[(strategy[lab[A, List[nat]], List[Char]],
                      (List[(term[lab[A, List[nat]], List[Char]],
                              term[lab[A, List[nat]], List[Char]])],
                        Option[List[(term[lab[A, List[nat]], List[Char]],
                                      term[lab[A, List[nat]], List[Char]])]])),
                     trs_proof[A, List[nat],
                                List[Char]]](List('t', 'e', 'r', 'm', 'i', 'n',
           'a', 't', 'i', 'o', 'n', 'A', 's', 's', 'u', 'm', 'p', 't', 'i', 'o',
           'n'),
      xml2trs_input[lab[A, List[nat]]](xml2name),
      (ab: (strategy[lab[A, List[nat]], List[Char]],
             (List[(term[lab[A, List[nat]], List[Char]],
                     term[lab[A, List[nat]], List[Char]])],
               Option[List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]], List[Char]])]])))
        => {
             val (q, (r, so)):
                   (strategy[lab[A, List[nat]], List[Char]],
                     (List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]], List[Char]])],
                       Option[List[(term[lab[A, List[nat]], List[Char]],
                                     term[lab[A, List[nat]], List[Char]])]]))
               = ab;
             AssumeSN[A, List[nat],
                       List[Char]]((strategy_to_Q[lab[A, List[nat]],
           List[Char]](q, r),
                                     (r, so match {
   case None => Nil
   case Some(s) => s
 })))
           },
      aa))),
                                       a),
id[trs_proof[A, List[nat], List[Char]]], x)

def xml2cr_proof[A : equal](xml2name: xml => sum[List[Char], lab[A, List[nat]]],
                             x: xml):
      sum[List[Char], cr_proof[A, List[nat], List[Char]]] =
  xml1element[cr_proof[A, List[nat], List[Char]],
               cr_proof[A, List[nat],
                         List[Char]]](List('c', 'r', 'P', 'r', 'o', 'o', 'f'),
                                       (a: xml) =>
 xml_options[cr_proof[A, List[nat],
                       List[Char]]](List((List('w', 'c', 'r', 'A', 'n', 'd',
        'S', 'N'),
   (aa: xml) =>
     xml2elements[Option[List[(term[lab[A, List[nat]], List[Char]],
                                (List[(pos,
((term[lab[A, List[nat]], List[Char]], term[lab[A, List[nat]], List[Char]]),
  term[lab[A, List[nat]], List[Char]]))],
                                  (term[lab[A, List[nat]], List[Char]],
                                    List[(pos,
   ((term[lab[A, List[nat]], List[Char]], term[lab[A, List[nat]], List[Char]]),
     term[lab[A, List[nat]], List[Char]]))])))]],
                   trs_proof[A, List[nat], List[Char]],
                   cr_proof[A, List[nat],
                             List[Char]]](List('w', 'c', 'r', 'A', 'n', 'd',
        'S', 'N'),
   wcr_proof[lab[A, List[nat]]](xml2name),
   (ab: xml) => xml2trs_proof[A](xml2name, ab),
   (ab: Option[List[(term[lab[A, List[nat]], List[Char]],
                      (List[(pos, ((term[lab[A, List[nat]], List[Char]],
                                     term[lab[A, List[nat]], List[Char]]),
                                    term[lab[A, List[nat]], List[Char]]))],
                        (term[lab[A, List[nat]], List[Char]],
                          List[(pos, ((term[lab[A, List[nat]], List[Char]],
term[lab[A, List[nat]], List[Char]]),
                                       term[lab[A, List[nat]],
     List[Char]]))])))]])
     => (b: trs_proof[A, List[nat], List[Char]]) =>
          SN_WCR[A, List[nat], List[Char]](ab, b),
   aa))),
                                     a),
                                       id[cr_proof[A, List[nat], List[Char]]],
                                       x)

def rulesd[A, B, C, D](x0: dpp_ops_ext[A, B, C, D]):
      A => List[(term[B, C], term[B, C])] =
  x0 match {
  case Dpp_ops_ext(dpp, p, pw, pairs, q, r, rw, rules, q_empty,
                    rules_no_left_var, rules_non_collapsing, is_QNF,
                    nFQ_subset_NF_rules, rules_map, reverse_rules_map,
                    intersect_pairs, replace_pair, intersect_rules, delete_P_Pw,
                    delete_R_Rw, split_pairs, split_rules, mk, more)
    => rules
}

def is_QNFc[A, B, C, D](x0: dpp_ops_ext[A, B, C, D]):
      A => (term[B, C]) => Boolean =
  x0 match {
  case Dpp_ops_ext(dpp, p, pw, pairs, q, r, rw, rules, q_empty,
                    rules_no_left_var, rules_non_collapsing, is_QNF,
                    nFQ_subset_NF_rules, rules_map, reverse_rules_map,
                    intersect_pairs, replace_pair, intersect_rules, delete_P_Pw,
                    delete_R_Rw, split_pairs, split_rules, mk, more)
    => is_QNF
}

def nFQ_subset_NF_rulesb[A, B, C, D](x0: dpp_ops_ext[A, B, C, D]):
      A => Boolean =
  x0 match {
  case Dpp_ops_ext(dpp, p, pw, pairs, q, r, rw, rules, q_empty,
                    rules_no_left_var, rules_non_collapsing, is_QNF,
                    nFQ_subset_NF_rules, rules_map, reverse_rules_map,
                    intersect_pairs, replace_pair, intersect_rules, delete_P_Pw,
                    delete_R_Rw, split_pairs, split_rules, mk, more)
    => nFQ_subset_NF_rules
}

def icap_impl_dpp[A, B : equal](i: dpp_ops_ext[A, B, List[Char], Unit], d: A):
      (List[term[B, List[Char]]]) =>
        (term[B, List[Char]]) => term[B, sum[Unit, List[Char]]] =
  {
    val qr: Boolean = (nFQ_subset_NF_rulesb[A, B, List[Char], Unit](i)).apply(d)
    val qnf: (term[B, List[Char]]) => Boolean =
      (is_QNFc[A, B, List[Char], Unit](i)).apply(d)
    val r: List[(term[B, List[Char]], term[B, List[Char]])] =
      (rulesd[A, B, List[Char], Unit](i)).apply(d)
    val ic: (List[term[B, List[Char]]]) =>
              ((List[Char]) => Boolean) =>
                (term[B, List[Char]]) => term[B, sum[Unit, List[Char]]]
      = (a: List[term[B, List[Char]]]) =>
          (b: (List[Char]) => Boolean) =>
            (c: term[B, List[Char]]) =>
              icap_impl[B](qr, qnf,
                            mapa[(term[B, List[Char]], term[B, List[Char]]),
                                  term[B,
List[Char]]]((aa: (term[B, List[Char]], term[B, List[Char]])) =>
               fst[term[B, List[Char]], term[B, List[Char]]](aa),
              r),
                            a, b, c);
    (s: List[term[B, List[Char]]]) =>
      {
        val sx: List[List[Char]] =
          remdups[List[Char]](maps[term[B, List[Char]],
                                    List[Char]](vars_term_list[B, List[Char]],
         s));
        (ic(s))((a: List[Char]) => membera[List[Char]](sx, a))
      }
  }

implicit def
  preorder_term[A : equal : order, B : equal : order]: preorder[term[A, B]] =
  new preorder[term[A, B]] {
  val `Ceta.less_eq` = (a: term[A, B], b: term[A, B]) =>
    less_eq_term[A, B](a, b)
  val `Ceta.less` = (a: term[A, B], b: term[A, B]) => less_term[A, B](a, b)
}

implicit def order_term[A : equal : order, B : equal : order]: order[term[A, B]]
  = new order[term[A, B]] {
  val `Ceta.less_eq` = (a: term[A, B], b: term[A, B]) =>
    less_eq_term[A, B](a, b)
  val `Ceta.less` = (a: term[A, B], b: term[A, B]) => less_term[A, B](a, b)
}

abstract sealed class l_poly[+A, +B]
final case class LPoly[B, A](a: B, b: List[(A, B)]) extends l_poly[A, B]

def ident_prob_of_smp[A : equal, B : equal](x0: List[(term[A, B], term[A, B])]):
      List[(term[A, B], term[A, B])] =
  x0 match {
  case Nil => Nil
  case (t, Var(x)) :: other =>
    map_filter[(term[A, B], term[A, B]),
                (term[A, B],
                  term[A, B])]((xa: (term[A, B], term[A, B])) =>
                                 (if ({
val (_, s): (term[A, B], term[A, B]) = xa;
equal_terma[A, B](s, Var[B, A](x))
                                      })
                                   Some[(term[A, B],
  term[A, B])]({
                 val (s, _): (term[A, B], term[A, B]) = xa;
                 (t, s)
               })
                                   else None),
                                other) ++
      ident_prob_of_smp[A, B](other)
}

def sym_collect[A, B](p: (term[A, B]) => Boolean, x1: term[A, B]): List[A] =
  (p, x1) match {
  case (p, Var(x)) => Nil
  case (p, Fun(f, ts)) =>
    (if (p(Fun[A, B](f, ts))) List(f) else Nil) ++
      maps[term[A, B], A]((a: term[A, B]) => sym_collect[A, B](p, a), ts)
}

def fresh_strings_list(name: List[Char], offset: nat, used: List[List[Char]],
                        n: nat):
      List[List[Char]] =
  take[List[Char]](n, filter[List[Char]]((s: List[Char]) =>
   ! (membera[List[Char]](remdups[List[Char]](used), s)),
  mapa[nat, List[Char]]((i: nat) =>
                          name ++
                            (shows_prec_nat(Zero_nat,
     plus_nat(i, offset))).apply(Nil),
                         upt(Zero_nat,
                              plus_nat(n,
size_list[List[Char]](remdups[List[Char]](used)))))))

def rat_gt(delta: rat): rat => rat => Boolean =
  (x: rat) => (y: rat) => less_eq_rat(delta, minus_rat(x, y))

def substpow_impl[A, B](sigma: A => term[B, A], x1: nat): A => term[B, A] =
  (sigma, x1) match {
  case (sigma, Zero_nat) => (a: A) => Var[A, B](a)
  case (sigma, Suc(n)) =>
    subst_compose[A, B, A, A](sigma, substpow_impl[A, B](sigma, n))
}

def funas_term_lista[A, B](x0: term[A, B]): List[(A, nat)] = x0 match {
  case Var(uu) => Nil
  case Fun(f, ts) =>
    (f, size_list[term[A, B]](ts)) ::
      maps[term[A, B],
            (A, nat)]((a: term[A, B]) => funas_term_lista[A, B](a), ts)
}

def funas_rule_list[A, B](r: (term[A, B], term[A, B])): List[(A, nat)] =
  funas_term_lista[A, B](fst[term[A, B], term[A, B]](r)) ++
    funas_term_lista[A, B](snd[term[A, B], term[A, B]](r))

def funas_trs_list[A : equal, B](r: List[(term[A, B], term[A, B])]):
      List[(A, nat)] =
  remdups[(A, nat)](maps[(term[A, B], term[A, B]),
                          (A, nat)]((a: (term[A, B], term[A, B])) =>
                                      funas_rule_list[A, B](a),
                                     r))

def shows_prec_pos(d: nat, p: pos): (List[Char]) => List[Char] = shows_pos(p)

def vars_rule_list[A, B : equal](r: (term[A, B], term[A, B])): List[B] =
  remdups[B](vars_rule_lista[A, B](r))

def ra[A, B, C, D](x0: tp_ops_ext[A, B, C, D]):
      A => List[(term[B, C], term[B, C])] =
  x0 match {
  case Tp_ops_ext(qreltrs, q, r, rw, rules, q_empty, is_QNF,
                   nFQ_subset_NF_rules, rules_map, delete_R_Rw, split_rules, mk,
                   more)
    => r
}

def check_CS_subseteq[A : equal,
                       B : equal](r: List[(term[A, B], term[A, B])],
                                   s: List[(term[A, B], term[A, B])]):
      sum[(term[A, B], term[A, B]), Unit] =
  catcha[((term[A, B], term[A, B]), (term[A, B], term[A, B])), Unit,
          (term[A, B],
            term[A, B])](forallm[(term[A, B], term[A, B]),
                                  (term[A, B],
                                    term[A,
  B])]((a: (term[A, B], term[A, B])) =>
         {
           val (l, ra): (term[A, B], term[A, B]) = a;
           check[(term[A, B],
                   term[A, B])](list_ex[(term[A, B],
  term[A, B])]((aa: (term[A, B], term[A, B])) =>
                 instance_rule[A, B, B]((l, ra), aa),
                s),
                                 (l, ra))
         },
        r),
                          (x: ((term[A, B], term[A, B]),
                                (term[A, B], term[A, B])))
                            => Inl[(term[A, B], term[A, B]),
                                    Unit](snd[(term[A, B], term[A, B]),
       (term[A, B], term[A, B])](x)))

def uncurry_rules[A : equal, B](a: A, sm: A => nat => List[A]):
      (List[(term[A, B], term[A, B])]) => List[(term[A, B], term[A, B])] =
  (b: List[(term[A, B], term[A, B])]) =>
    mapa[(term[A, B], term[A, B]),
          (term[A, B],
            term[A, B])]((ba: (term[A, B], term[A, B])) =>
                           {
                             val (l, r): (term[A, B], term[A, B]) = ba;
                             (uncurry_term[A, B](a, sm, l),
                               uncurry_term[A, B](a, sm, r))
                           },
                          b)

def rwa[A, B, C, D](x0: tp_ops_ext[A, B, C, D]):
      A => List[(term[B, C], term[B, C])] =
  x0 match {
  case Tp_ops_ext(qreltrs, q, r, rw, rules, q_empty, is_QNF,
                   nFQ_subset_NF_rules, rules_map, delete_R_Rw, split_rules, mk,
                   more)
    => rw
}

def eta_closed_rules[A : equal : show,
                      B : equal : show](a: A, sm: A => nat => List[A],
 ra: List[(term[A, B], term[A, B])], r: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[((term[A, B], term[A, B]), (List[Char]) => List[Char]), Unit,
          (List[Char]) =>
            List[Char]](forallm[(term[A, B], term[A, B]),
                                 (List[Char]) =>
                                   List[Char]]((b: (term[A, B], term[A, B])) =>
         {
           val (l, raa): (term[A, B], term[A, B]) = b;
           aarity_term[A, B](a, sm, l) match {
             case None => Inr[Unit, (List[Char]) => List[Char]](())
             case Some(Zero_nat) => Inr[Unit, (List[Char]) => List[Char]](())
             case Some(Suc(_)) =>
               check[(List[Char]) =>
                       List[Char]](list_ex[(term[A, B],
     term[A, B])]((ba: (term[A, B], term[A, B])) =>
                    {
                      val (lll, rrr): (term[A, B], term[A, B]) = ba;
                      (lll, rrr) match {
                        case (Var(_), bb) => false
                        case (Fun(f, Nil), bb) => false
                        case (Fun(f, List(ll)), bb) => false
                        case (Fun(f, List(ll, Var(x))), Var(_)) => false
                        case (Fun(f, List(ll, Var(x))), Fun(g, Nil)) => false
                        case (Fun(f, List(ll, Var(x))), Fun(g, List(rr))) =>
                          false
                        case (Fun(f, List(ll, Var(x))),
                               Fun(g, List(rr, Var(y))))
                          => eq[A](f, a) &&
                               (eq[A](g, a) &&
                                 (eq[B](x, y) &&
                                   (! (membera[B](vars_rule_list[A,
                          B]((ll, rr)),
           x)) &&
                                     instance_rule[A, B,
            B]((l, raa), (ll, rr)))))
                        case (Fun(f, List(ll, Var(x))),
                               Fun(g, rr :: Var(y) :: _ :: _))
                          => false
                        case (Fun(f, List(ll, Var(x))),
                               Fun(g, rr :: Fun(_, _) :: list))
                          => false
                        case (Fun(f, ll :: Var(x) :: _ :: _), bb) => false
                        case (Fun(f, ll :: Fun(_, _) :: list), bb) => false
                      }
                    },
                   r),
                                    comp[List[Char], List[Char],
  List[Char]](shows_prec_list[Char](Zero_nat,
                                     List('e', 't', 'a', ' ', 'e', 'x', 'p',
   'a', 'n', 's', 'i', 'o', 'n', ' ', 'o', 'f', ' ')),
               comp[List[Char], List[Char],
                     List[Char]](shows_rule[A,
     B]((aa: A) => (ba: List[Char]) => shows_prec[A](Zero_nat, aa, ba),
         (aa: B) => (ba: List[Char]) => shows_prec[B](Zero_nat, aa, ba),
         List(' ', '-', '>', ' '), (l, raa)),
                                  shows_prec_list[Char](Zero_nat,
                 List(' ', 'm', 'i', 's', 's', 'i', 'n', 'g')))))
           }
         },
        ra),
                         (x: ((term[A, B], term[A, B]),
                               (List[Char]) => List[Char]))
                           => Inl[(List[Char]) => List[Char],
                                   Unit](snd[(term[A, B], term[A, B]),
      (List[Char]) => List[Char]](x)))

def uncurry_eta_split[A : equal,
                       B : equal](eboth: List[(term[A, B], term[A, B])],
                                   rtest: List[(term[A, B], term[A, B])]):
      (List[(term[A, B], term[A, B])], List[(term[A, B], term[A, B])]) =
  {
    val test: ((term[A, B], term[A, B])) => Boolean =
      (a: (term[A, B], term[A, B])) =>
        {
          val (l, r): (term[A, B], term[A, B]) = a;
          less_nat(Zero_nat, num_args[A, B](l)) &&
            (less_nat(Zero_nat, num_args[A, B](r)) &&
              list_ex[(term[A, B],
                        term[A, B])]((aa: (term[A, B], term[A, B])) =>
                                       eq_rule_mod_vars[A,
                 B]((hd[term[A, B]](args[A, B](l)),
                      hd[term[A, B]](args[A, B](r))),
                     aa),
                                      rtest))
        };
    partition[(term[A, B], term[A, B])](test, eboth)
  }

def sig_list_to_sig_map[A : equal](a: A, sml: List[((A, nat), List[A])],
                                    fmap: (List[((A, nat), List[A])]) =>
    A => nat => A):
      A => nat => List[A] =
  {
    val fm: A => nat => A = fmap(sml);
    (f: A) =>
      (n: nat) =>
        lookupa[(A, nat), List[A]]((f, n), sml) match {
          case None => List((fm(f))(n))
          case Some(xs) => (if (nulla[A](xs)) List((fm(f))(n)) else xs)
        }
  }

def uncurry_of_sig_list[A](a: A, sml: List[((A, nat), List[A])],
                            sm: A => nat => List[A]):
      List[(term[A, List[Char]], term[A, List[Char]])] =
  maps[((A, nat), List[A]),
        (term[A, List[Char]],
          term[A, List[Char]])]((b: ((A, nat), List[A])) =>
                                  {
                                    val (ba, c): ((A, nat), List[A]) = b;
                                    ({
                                       val (f, n): (A, nat) = ba;
                                       (_: List[A]) =>
 {
   val g: nat => A = (aa: nat) => get_symbol[A](sm, f, n, aa);
   mapa[nat, (term[A, List[Char]],
               term[A, List[Char]])]((i: nat) =>
                                       (Fun[A,
     List[Char]](a, List(generate_f_xs[A](g(i), plus_nat(n, i)),
                          Var[List[Char], A](generate_var(plus_nat(n, i))))),
 generate_f_xs[A](g(Suc(i)), plus_nat(n, Suc(i)))),
                                      upt(Zero_nat, aarity[A](sm, f, n)))
 }
                                     })(c)
                                  },
                                 sml)

def uncurry_tt[A, B : equal : show](i: tp_ops_ext[A, B, List[Char], Unit],
                                     info: (B,
     (List[((B, nat), List[B])],
       (List[(term[B, List[Char]], term[B, List[Char]])],
         List[(term[B, List[Char]], term[B, List[Char]])]))),
                                     r: List[(term[B, List[Char]],
       term[B, List[Char]])],
                                     tp: A):
      sum[(List[Char]) => List[Char], A] =
  {
    val (a, (sml, (u, eb))):
          (B, (List[((B, nat), List[B])],
                (List[(term[B, List[Char]], term[B, List[Char]])],
                  List[(term[B, List[Char]], term[B, List[Char]])])))
      = info
    val raa: List[(term[B, List[Char]], term[B, List[Char]])] =
      (ra[A, B, List[Char], Unit](i)).apply(tp)
    val rw: List[(term[B, List[Char]], term[B, List[Char]])] =
      (rwa[A, B, List[Char], Unit](i)).apply(tp)
    val (e, ew):
          (List[(term[B, List[Char]], term[B, List[Char]])],
            List[(term[B, List[Char]], term[B, List[Char]])])
      = uncurry_eta_split[B, List[Char]](eb, raa)
    val r_eta: List[(term[B, List[Char]], term[B, List[Char]])] = e ++ raa
    val rw_eta: List[(term[B, List[Char]], term[B, List[Char]])] = ew ++ rw
    val rb_eta: List[(term[B, List[Char]], term[B, List[Char]])] =
      r_eta ++ rw_eta
    val fmap: (List[((B, nat), List[B])]) => B => nat => B =
      (_: List[((B, nat), List[B])]) => (f: B) => (_: nat) => f
    val sm: B => nat => List[B] = sig_list_to_sig_map[B](a, sml, fmap)
    val uR: List[(term[B, List[Char]], term[B, List[Char]])] =
      (uncurry_rules[B, List[Char]](a, sm)).apply(r_eta)
    val uRw: List[(term[B, List[Char]], term[B, List[Char]])] =
      (uncurry_rules[B, List[Char]](a, sm)).apply(rw_eta);
    {
      val s: List[(term[B, List[Char]], term[B, List[Char]])] =
        uncurry_of_sig_list[B](a, sml, sm);
      bindb[(List[Char]) => List[Char], Unit,
             Unit](catcha[((term[B, List[Char]], term[B, List[Char]]),
                            (List[Char]) => List[Char]),
                           Unit,
                           (List[Char]) =>
                             List[Char]](forallm[(term[B, List[Char]],
           term[B, List[Char]]),
          (List[Char]) =>
            List[Char]]((b: (term[B, List[Char]], term[B, List[Char]])) =>
                          {
                            val (l, _):
                                  (term[B, List[Char]], term[B, List[Char]])
                              = b;
                            check[(List[Char]) =>
                                    List[Char]](hvf_term[B, List[Char]](a, l),
         comp[List[Char], List[Char],
               List[Char]](shows_prec_list[Char](Zero_nat,
          List('h', 'e', 'a', 'd', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e',
                ' ', 'i', 'n', ' ', 'l', 'h', 's', ' ')),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_term[B,
                  List[Char]]((aa: B) =>
                                (ba: List[Char]) =>
                                  shows_prec[B](Zero_nat, aa, ba),
                               (aa: List[Char]) =>
                                 shows_prec_list[Char](Zero_nat, aa),
                               l),
       shows_prec_list[Char](Zero_nat,
                              List(' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o',
                                    'w', 'e', 'd')))))
                          },
                         r_eta),
  (x: ((term[B, List[Char]], term[B, List[Char]]), (List[Char]) => List[Char]))
    => Inl[(List[Char]) => List[Char],
            Unit](snd[(term[B, List[Char]], term[B, List[Char]]),
                       (List[Char]) => List[Char]](x))),
                    (_: Unit) =>
                      bindb[(List[Char]) => List[Char], Unit,
                             Unit](catcha[((term[B, List[Char]],
     term[B, List[Char]]),
    (List[Char]) => List[Char]),
   Unit,
   (List[Char]) =>
     List[Char]](forallm[(term[B, List[Char]], term[B, List[Char]]),
                          (List[Char]) =>
                            List[Char]]((b:
   (term[B, List[Char]], term[B, List[Char]]))
  => {
       val (l, _): (term[B, List[Char]], term[B, List[Char]]) = b;
       check[(List[Char]) =>
               List[Char]](hvf_term[B, List[Char]](a, l),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_prec_list[Char](Zero_nat,
                             List('h', 'e', 'a', 'd', ' ', 'v', 'a', 'r', 'i',
                                   'a', 'b', 'l', 'e', ' ', 'i', 'n', ' ', 'l',
                                   'h', 's', ' ')),
       comp[List[Char], List[Char],
             List[Char]](shows_term[B, List[Char]]((aa: B) =>
             (ba: List[Char]) => shows_prec[B](Zero_nat, aa, ba),
            (aa: List[Char]) => shows_prec_list[Char](Zero_nat, aa), l),
                          shows_prec_list[Char](Zero_nat,
         List(' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e', 'd')))))
     },
 rw_eta),
                  (x: ((term[B, List[Char]], term[B, List[Char]]),
                        (List[Char]) => List[Char]))
                    => Inl[(List[Char]) => List[Char],
                            Unit](snd[(term[B, List[Char]],
term[B, List[Char]]),
                                       (List[Char]) => List[Char]](x))),
                                    (_: Unit) =>
                                      bindb[(List[Char]) => List[Char], Unit,
     Unit](eta_closed_rules[B, List[Char]](a, sm, r_eta, r_eta),
            (_: Unit) =>
              bindb[(List[Char]) => List[Char], Unit,
                     Unit](eta_closed_rules[B,
     List[Char]](a, sm, rb_eta, rb_eta),
                            (_: Unit) =>
                              bindb[(List[Char]) => List[Char], Unit,
                                     Unit](catcha[((term[B, List[Char]],
             term[B, List[Char]]),
            (List[Char]) => List[Char]),
           Unit,
           (List[Char]) =>
             List[Char]](forallm[(term[B, List[Char]], term[B, List[Char]]),
                                  (List[Char]) =>
                                    List[Char]]((aa:
           (term[B, List[Char]], term[B, List[Char]]))
          => {
               val (l, rb): (term[B, List[Char]], term[B, List[Char]]) = aa;
               check[(List[Char]) =>
                       List[Char]](! (is_Var[B, List[Char]](l)),
                                    comp[List[Char], List[Char],
  List[Char]](shows_prec_list[Char](Zero_nat,
                                     List('l', 'h', 's', ' ', 'm', 'u', 's',
   't', ' ', 'n', 'o', 't', ' ', 'b', 'e', ' ', 'a', ' ', 'v', 'a', 'r', 'i',
   'a', 'b', 'l', 'e', ' ', 'i', 'n', ' ', 'r', 'u', 'l', 'e', ' ')),
               shows_rule[B, List[Char]]((ab: B) =>
   (b: List[Char]) => shows_prec[B](Zero_nat, ab, b),
  (ab: List[Char]) => shows_prec_list[Char](Zero_nat, ab),
  List(' ', '-', '>', ' '), (l, rb))))
             },
         rw_eta),
                          (x: ((term[B, List[Char]], term[B, List[Char]]),
                                (List[Char]) => List[Char]))
                            => Inl[(List[Char]) => List[Char],
                                    Unit](snd[(term[B, List[Char]],
        term[B, List[Char]]),
       (List[Char]) => List[Char]](x))),
    (_: Unit) =>
      bindb[(List[Char]) => List[Char], Unit,
             Unit](catcha[(term[B, List[Char]], term[B, List[Char]]), Unit,
                           (List[Char]) =>
                             List[Char]](check_subseteq[(term[B, List[Char]],
                  term[B, List[Char]])](uR, r),
  (x: (term[B, List[Char]], term[B, List[Char]])) =>
    Inl[(List[Char]) => List[Char],
         Unit](comp[List[Char], List[Char],
                     List[Char]](shows_prec_list[Char](Zero_nat,
                List('u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', ' ', 'r', 'u',
                      'l', 'e', ' ')),
                                  comp[List[Char], List[Char],
List[Char]](shows_rule[B, List[Char]]((aa: B) =>
(b: List[Char]) => shows_prec[B](Zero_nat, aa, b),
                                       (aa: List[Char]) =>
 shows_prec_list[Char](Zero_nat, aa),
                                       List(' ', '-', '>', ' '), x),
             shows_prec_list[Char](Zero_nat,
                                    List(' ', 'i', 's', ' ', 'm', 'i', 's', 's',
  'i', 'n', 'g')))))),
                    (_: Unit) =>
                      bindb[(List[Char]) => List[Char], Unit,
                             Unit](catcha[(term[B, List[Char]],
    term[B, List[Char]]),
   Unit,
   (List[Char]) =>
     List[Char]](check_subseteq[(term[B, List[Char]],
                                  term[B, List[Char]])](uRw, r),
                  (x: (term[B, List[Char]], term[B, List[Char]])) =>
                    Inl[(List[Char]) => List[Char],
                         Unit](comp[List[Char], List[Char],
                                     List[Char]](shows_prec_list[Char](Zero_nat,
                                List('u', 'n', 'c', 'u', 'r', 'r', 'i', 'e',
                                      'd', ' ', 'r', 'u', 'l', 'e', ' ')),
          comp[List[Char], List[Char],
                List[Char]](shows_rule[B,
List[Char]]((aa: B) => (b: List[Char]) => shows_prec[B](Zero_nat, aa, b),
             (aa: List[Char]) => shows_prec_list[Char](Zero_nat, aa),
             List(' ', '-', '>', ' '), x),
                             shows_prec_list[Char](Zero_nat,
            List(' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g')))))),
                                    (_: Unit) =>
                                      bindb[(List[Char]) => List[Char], Unit,
     Unit](catcha[(term[B, List[Char]], term[B, List[Char]]), Unit,
                   (List[Char]) =>
                     List[Char]](check_CS_subseteq[B, List[Char]](s, u),
                                  (x: (term[B, List[Char]],
term[B, List[Char]]))
                                    => Inl[(List[Char]) => List[Char],
    Unit](comp[List[Char], List[Char],
                List[Char]](shows_prec_list[Char](Zero_nat,
           List('u', 'n', 'c', 'u', 'r', 'r', 'y', ' ', 'r', 'u', 'l', 'e',
                 ' ')),
                             comp[List[Char], List[Char],
                                   List[Char]](shows_rule[B,
                   List[Char]]((aa: B) =>
                                 (b: List[Char]) =>
                                   shows_prec[B](Zero_nat, aa, b),
                                (aa: List[Char]) =>
                                  shows_prec_list[Char](Zero_nat, aa),
                                List(' ', '-', '>', ' '), x),
        shows_prec_list[Char](Zero_nat,
                               List(' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i',
                                     'n', 'g')))))),
            (_: Unit) =>
              catcha[(term[B, List[Char]], term[B, List[Char]]), Unit,
                      (List[Char]) =>
                        List[Char]](check_subseteq[(term[B, List[Char]],
             term[B, List[Char]])](u, r),
                                     (x:
(term[B, List[Char]], term[B, List[Char]]))
                                       => Inl[(List[Char]) => List[Char],
       Unit](comp[List[Char], List[Char],
                   List[Char]](shows_prec_list[Char](Zero_nat,
              List('u', 'n', 'c', 'u', 'r', 'r', 'y', ' ', 'r', 'u', 'l', 'e',
                    ' ')),
                                comp[List[Char], List[Char],
                                      List[Char]](shows_rule[B,
                      List[Char]]((aa: B) =>
                                    (b: List[Char]) =>
                                      shows_prec[B](Zero_nat, aa, b),
                                   (aa: List[Char]) =>
                                     shows_prec_list[Char](Zero_nat, aa),
                                   List(' ', '-', '>', ' '), x),
           shows_prec_list[Char](Zero_nat,
                                  List(' ', 'i', 's', ' ', 'm', 'i', 's', 's',
'i', 'n', 'g', ' ', 'i', 'n', ' ', 'n', 'e', 'w', ' ', 'T', 'R',
'S'))))))))))))))
    }
      match {
      case Inl(aa) => Inl[(List[Char]) => List[Char], A](aa)
      case Inr(_) =>
        Inr[A, (List[Char]) =>
                 List[Char]]((mka[A, B, List[Char],
                                   Unit](i)).apply(Nil).apply(uR).apply(uRw ++
                                  u))
    }
  }

def apply_af_entry[A, B](uu: A, x1: af_entry, ts: List[term[A, B]]):
      term[A, B] =
  (uu, x1, ts) match {
  case (uu, Collapse(i), ts) => nth[term[A, B]](ts, i)
  case (f, AFList(is), ts) =>
    Fun[A, B](f, mapa[nat, term[A, B]]((a: nat) => nth[term[A, B]](ts, a), is))
}

def afs_term[A, B](pi: A => nat => af_entry, x1: term[A, B]):
      term[filtered[A], B] =
  (pi, x1) match {
  case (pi, Fun(f, ts)) =>
    {
      val l: nat = size_list[term[A, B]](ts);
      apply_af_entry[filtered[A],
                      B](FPair[A](f, l), (pi(f))(l),
                          mapa[term[A, B],
                                term[filtered[A],
                                      B]]((a: term[A, B]) =>
    afs_term[A, B](pi, a),
   ts))
    }
  case (pi, Var(x)) => Var[B, filtered[A]](x)
}

def search_len(t: List[Char], s: List[Char], n: nat):
      Option[(List[Char], List[Char])] =
  (if (equal_lista[Char](take[Char](n, t), s))
    Some[(List[Char], List[Char])]((Nil, drop[Char](n, t)))
    else t match {
           case Nil => None
           case h :: tt =>
             search_len(tt, s, n) match {
               case None => None
               case Some((bef, aft)) =>
                 Some[(List[Char], List[Char])]((h :: bef, aft))
             }
         })

def check_graph_decomp[A](ss: A => (List[Char]) => List[Char],
                           g: ((A, A)) => Boolean,
                           rcs: List[(Boolean, List[A])]):
      sum[(List[Char]) => List[Char], Unit] =
  bindb[(List[Char]) => List[Char], Unit,
         Unit](check_pairwise[List[A],
                               (List[Char]) =>
                                 List[Char]]((a: List[A]) =>
       (b: List[A]) => check_edges[A](ss, g, a, b),
      mapa[(Boolean, List[A]),
            List[A]]((a: (Boolean, List[A])) => snd[Boolean, List[A]](a), rcs)),
                (_: Unit) =>
                  catcha[(List[A], (List[Char]) => List[Char]), Unit,
                          (List[Char]) =>
                            List[Char]](forallm[List[A],
         (List[Char]) =>
           List[Char]]((c: List[A]) => check_edges[A](ss, g, c, c),
                        map_filter[(Boolean, List[A]),
                                    List[A]]((x: (Boolean, List[A])) =>
       (if (! (fst[Boolean, List[A]](x)))
         Some[List[A]](snd[Boolean, List[A]](x)) else None),
      rcs)),
 (x: (List[A], (List[Char]) => List[Char])) =>
   Inl[(List[Char]) => List[Char],
        Unit](snd[List[A], (List[Char]) => List[Char]](x))))

def list_sumI[A](ze: A, pl: A => A => A, x2: List[A]): A = (ze, pl, x2) match {
  case (ze, pl, Nil) => ze
  case (ze, pl, x :: xs) => (pl(x))(list_sumI[A](ze, pl, xs))
}

def mul_pvarsI[A : equal, B](uu: A, uv: A => A => A, a: A, x3: List[(B, A)]):
      List[(B, A)] =
  (uu, uv, a, x3) match {
  case (uu, uv, a, Nil) => Nil
  case (ze, ti, a, (x, b) :: vas) =>
    {
      val p: A = (ti(a))(b)
      val res: List[(B, A)] = mul_pvarsI[A, B](ze, ti, a, vas);
      (if (eq[A](p, ze)) res else (x, p) :: res)
    }
}

def mul_lpolyI[A : equal, B](ze: A, ti: A => A => A, a: A, x3: l_poly[B, A]):
      l_poly[B, A] =
  (ze, ti, a, x3) match {
  case (ze, ti, a, LPoly(b, vas)) =>
    LPoly[A, B]((ti(a))(b), mul_pvarsI[A, B](ze, ti, a, vas))
}

def add_varI[A : equal,
              B : equal](uu: A, uv: A => A => A, x: B, a: A, xa4: List[(B, A)]):
      List[(B, A)] =
  (uu, uv, x, a, xa4) match {
  case (uu, uv, x, a, Nil) => List((x, a))
  case (ze, pl, x, a, (y, b) :: vas) =>
    (if (eq[B](x, y))
      {
        val s: A = (pl(a))(b);
        (if (eq[A](s, ze)) vas else (x, s) :: vas)
      }
      else (y, b) :: add_varI[A, B](ze, pl, x, a, vas))
}

def sum_pvarsI[A : equal,
                B : equal](uu: A, uv: A => A => A, x2: List[(B, A)],
                            vbs: List[(B, A)]):
      List[(B, A)] =
  (uu, uv, x2, vbs) match {
  case (uu, uv, Nil, vbs) => vbs
  case (ze, pl, (x, a) :: vas, vbs) =>
    (if (eq[A](a, ze)) sum_pvarsI[A, B](ze, pl, vas, vbs)
      else sum_pvarsI[A, B](ze, pl, vas, add_varI[A, B](ze, pl, x, a, vbs)))
}

def sum_lpolyI[A : equal,
                B : equal](ze: A, pl: A => A => A, x2: l_poly[B, A],
                            x3: l_poly[B, A]):
      l_poly[B, A] =
  (ze, pl, x2, x3) match {
  case (ze, pl, LPoly(a, vas), LPoly(b, vbs)) =>
    LPoly[A, B]((pl(a))(b), sum_pvarsI[A, B](ze, pl, vas, vbs))
}

def pleftI[A : equal, B,
            C : equal](ze: A, on: A, pl: A => A => A, ti: A => A => A,
                        mz: A => A, pi: B => nat => (A, List[A]),
                        x6: term[B, C]):
      l_poly[C, A] =
  (ze, on, pl, ti, mz, pi, x6) match {
  case (ze, on, pl, ti, mz, pi, Var(x)) => LPoly[A, C](ze, List((x, on)))
  case (ze, on, pl, ti, mz, pi, Fun(f, ts)) =>
    {
      val (c, as): (A, List[A]) = (pi(f))(size_list[term[B, C]](ts));
      sum_lpolyI[A, C](ze, pl, LPoly[A, C](c, Nil),
                        list_sumI[l_poly[C,
  A]](LPoly[A, C](ze, Nil),
       (a: l_poly[C, A]) => (b: l_poly[C, A]) => sum_lpolyI[A, C](ze, pl, a, b),
       mapa[(A, term[B, C]),
             l_poly[C, A]]((at: (A, term[B, C])) =>
                             mul_lpolyI[A,
 C](ze, ti, fst[A, term[B, C]](at),
     pleftI[A, B, C](ze, on, pl, ti, mz, pi, snd[A, term[B, C]](at))),
                            zip[A, term[B, C]](as, ts))))
        match {
        case LPoly(d, Nil) => LPoly[A, C](mz(d), Nil)
        case LPoly(d, ab :: lista) => LPoly[A, C](d, ab :: lista)
      }
    }
}

def check_poly_ge[A : equal,
                   B : ordered_semiring_0](x0: List[(List[(A, nat)], B)],
    q: List[(List[(A, nat)], B)]):
      Boolean =
  (x0, q) match {
  case (Nil, q) =>
    list_all[(List[(A, nat)],
               B)]((a: (List[(A, nat)], B)) =>
                     {
                       val (_, aa): (List[(A, nat)], B) = a;
                       ge[B](zero[B], aa)
                     },
                    q)
  case ((m, c) :: p, q) =>
    extract[(List[(A, nat)],
              B)]((nd: (List[(A, nat)], B)) =>
                    eq_monom[A](fst[List[(A, nat)], B](nd), m),
                   q)
      match {
      case None => ge[B](c, zero[B]) && check_poly_ge[A, B](p, q)
      case Some((q1, ((_, d), q2))) =>
        ge[B](c, d) && check_poly_ge[A, B](p, q1 ++ q2)
    }
}

def check_poly_gt[A : ordered_semiring_0,
                   B : equal](gt: A => A => Boolean,
                               p: List[(List[(B, nat)], A)],
                               q: List[(List[(B, nat)], A)]):
      Boolean =
  {
    val (a1, p1): (A, List[(List[(B, nat)], A)]) = poly_split[B, A](Nil, p)
    val (b1, q1): (A, List[(List[(B, nat)], A)]) = poly_split[B, A](Nil, q);
    (gt(a1))(b1) && check_poly_ge[B, A](p1, q1)
  }

def rpo_strict_unbounded[A : equal : show,
                          B : equal : show](pr:
      (A => nat => A => nat => (Boolean, Boolean), A => Boolean),
     c: A => nat => order_tag):
      ((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit] =
  (a: (term[A, B], term[A, B])) =>
    {
      val (s, t): (term[A, B], term[A, B]) = a;
      check[(List[Char]) =>
              List[Char]](fst[Boolean,
                               Boolean](rpo_unbounded[A,
               B].apply(pr).apply(c).apply(s).apply(t)),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('c', 'o',
                              'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r',
                              'i', 'e', 'n', 't', ' ')),
      comp[List[Char], List[Char],
            List[Char]]((shows_prec_term[A, B](Zero_nat)).apply(s),
                         comp[List[Char], List[Char],
                               List[Char]](shows_string.apply(List(' ', '>',
                            'R', 'P', 'O', ' ')),
    comp[List[Char], List[Char],
          List[Char]]((shows_prec_term[A, B](Zero_nat)).apply(t), shows_nl)))))
    }

def rpo_nstrict_unbounded[A : equal : show,
                           B : equal : show](pr:
       (A => nat => A => nat => (Boolean, Boolean), A => Boolean),
      c: A => nat => order_tag):
      ((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit] =
  (a: (term[A, B], term[A, B])) =>
    {
      val (s, t): (term[A, B], term[A, B]) = a;
      check[(List[Char]) =>
              List[Char]](snd[Boolean,
                               Boolean](rpo_unbounded[A,
               B].apply(pr).apply(c).apply(s).apply(t)),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('c', 'o',
                              'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r',
                              'i', 'e', 'n', 't', ' ')),
      comp[List[Char], List[Char],
            List[Char]]((shows_prec_term[A, B](Zero_nat)).apply(s),
                         comp[List[Char], List[Char],
                               List[Char]](shows_string.apply(List(' ', '>',
                            '=', 'R', 'P', 'O', ' ')),
    comp[List[Char], List[Char],
          List[Char]]((shows_prec_term[A, B](Zero_nat)).apply(t), shows_nl)))))
    }

def create_RPO_redtriple[A : show, B : equal : show,
                          C : equal : show](prec_repr_to_pr:
      (List[((A, nat), (nat, order_tag))]) =>
        (B => nat => nat, B => nat => order_tag),
     pr: List[((A, nat), (nat, order_tag))]):
      redtriple_ext[B, C, Unit] =
  {
    val (p, tau): (B => nat => nat, B => nat => order_tag) = prec_repr_to_pr(pr)
    val ns: ((term[B, C], term[B, C])) => sum[(List[Char]) => List[Char], Unit]
      = rpo_nstrict_unbounded[B, C]((prc_nat[B](p), prl_nat[B](p)), tau);
    Redtriple_ext[B, C,
                   Unit](Inr[Unit, (List[Char]) => List[Char]](()),
                          rpo_strict_unbounded[B,
        C]((prc_nat[B](p), prl_nat[B](p)), tau),
                          ns, ns, (a: (B, nat)) => (b: nat) => full_af[B](a, b),
                          Inr[Unit, (List[Char]) => List[Char]](()),
                          shows_rpo_repr[A](pr), ())
  }

def intersect_values[A : equal,
                      B : linorder](key: A => Option[B], vs: List[A],
                                     m: rbt[B, List[A]]):
      rbt[B, List[A]] =
  foldr[A, rbt[B, List[A]]]((a: A) =>
                              (b: rbt[B, List[A]]) => aux[A, B](key, m, a, b),
                             vs, rm_empty[B, List[A]])

def intersect_rules[A : equal : linorder,
                     B : equal](rs: List[(term[A, B], term[A, B])]):
      (rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]]) =>
        rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]] =
  (a: rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]]) =>
    intersect_values[(Boolean, (term[A, B], term[A, B])),
                      (A, nat)]((aa: (Boolean, (term[A, B], term[A, B]))) =>
                                  key[Boolean, A, B](aa),
                                 mapa[(term[A, B], term[A, B]),
                                       (Boolean,
 (term[A, B], term[A, B]))]((aa: (term[A, B], term[A, B])) => (true, aa), rs) ++
                                   mapa[(term[A, B], term[A, B]),
 (Boolean,
   (term[A, B],
     term[A, B]))]((aa: (term[A, B], term[A, B])) => (false, aa), rs),
                                 a)

def rev_term[A, B](c: ctxt[A, B], x1: term[A, B]): term[A, B] = (c, x1) match {
  case (c, Var(x)) => ctxt_apply[A, B](c, Var[B, A](x))
  case (c, Fun(f, List(t))) => rev_term[A, B](More[A, B](f, Nil, c, Nil), t)
}

def rev_rule[A, B](x0: (term[A, B], term[A, B])): (term[A, B], term[A, B]) = x0
  match {
  case (l, r) => (rev_term[A, B](Hole, l), rev_term[A, B](Hole, r))
}

def sorted_list_subset[A : equal : linorder](x0: List[A], uu: List[A]):
      Option[A] =
  (x0, uu) match {
  case (a :: as, b :: bs) =>
    (if (eq[A](a, b)) sorted_list_subset[A](as, b :: bs)
      else (if (less[A](b, a)) sorted_list_subset[A](a :: as, bs)
             else Some[A](a)))
  case (Nil, uu) => None
  case (a :: uv, Nil) => Some[A](a)
}

def doc_of_string(s: List[Char]): sum[List[Char], xmldoc] =
  bindb[List[Char], (xmldoc, List[Char]),
         xmldoc](parse_doc.apply(s),
                  (a: (xmldoc, List[Char])) =>
                    {
                      val (doc, _): (xmldoc, List[Char]) = a;
                      Inr[xmldoc, List[Char]](doc)
                    })

def filtered_fun[A](x0: filtered[A]): A = x0 match {
  case FPair(f, n) => f
}

def shows_prec_filtered[A : show](d: nat, ff: filtered[A]):
      (List[Char]) => List[Char] =
  (a: List[Char]) => shows_prec[A](Zero_nat, filtered_fun[A](ff), a)

def shows_list_filtered[A : show](ffs: List[filtered[A]]):
      (List[Char]) => List[Char] =
  shows_list_aux[filtered[A]]((a: filtered[A]) =>
                                shows_prec_filtered[A](Zero_nat, a),
                               ffs)

implicit def show_filtered[A : show]: show[filtered[A]] = new show[filtered[A]]
  {
  val `Ceta.shows_prec` = (a: nat, b: filtered[A], c: List[Char]) =>
    (shows_prec_filtered[A](a, b)).apply(c)
  val `Ceta.shows_list` = (a: List[filtered[A]], b: List[Char]) =>
    (shows_list_filtered[A](a)).apply(b)
}

def afs_check[A : show,
               B : show](r: (List[Char]) => List[Char],
                          pi: A => nat => af_entry,
                          g: ((term[filtered[A], B], term[filtered[A], B])) =>
                               sum[(List[Char]) => List[Char], Unit],
                          lr: (term[A, B], term[A, B])):
      sum[(List[Char]) => List[Char], Unit] =
  {
    val pl: term[filtered[A], B] =
      afs_term[A, B](pi, fst[term[A, B], term[A, B]](lr))
    val pr: term[filtered[A], B] =
      afs_term[A, B](pi, snd[term[A, B], term[A, B]](lr));
    catcha[(List[Char]) => List[Char], Unit,
            (List[Char]) =>
              List[Char]](g((pl, pr)),
                           (x: (List[Char]) => List[Char]) =>
                             Inl[(List[Char]) => List[Char],
                                  Unit](comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ', 'n',
   'o', 't', ' ', 'o', 'r', 'i', 'e', 'n', 't', ' ')),
                   comp[List[Char], List[Char],
                         List[Char]]((shows_prec_term[A,
               B](Zero_nat)).apply(fst[term[A, B], term[A, B]](lr)),
                                      comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List(' ')),
                 comp[List[Char], List[Char],
                       List[Char]](r, comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List(' ')),
                 comp[List[Char], List[Char],
                       List[Char]]((shows_prec_term[A,
             B](Zero_nat)).apply(snd[term[A, B], term[A, B]](lr)),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl,
               comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List('p', 'i', '(', ' ')),
                                  comp[List[Char], List[Char],
List[Char]]((shows_prec_term[A, B](Zero_nat)).apply(fst[term[A, B],
                 term[A, B]](lr)),
             comp[List[Char], List[Char],
                   List[Char]](shows_string.apply(List(' ', ')', ' ', '=',
                ' ')),
                                comp[List[Char], List[Char],
                                      List[Char]]((shows_prec_term[filtered[A],
                            B](Zero_nat)).apply(pl),
           comp[List[Char], List[Char],
                 List[Char]](shows_nl,
                              comp[List[Char], List[Char],
                                    List[Char]](shows_string.apply(List('p',
                                 'i', '(', ' ')),
         comp[List[Char], List[Char],
               List[Char]]((shows_prec_term[A,
     B](Zero_nat)).apply(snd[term[A, B], term[A, B]](lr)),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_string.apply(List(' ', ')',
                               ' ', '=', ' ')),
       comp[List[Char], List[Char],
             List[Char]]((shows_prec_term[filtered[A], B](Zero_nat)).apply(pr),
                          comp[List[Char], List[Char],
                                List[Char]](shows_nl, x)))))))))))))))))))
  }

def shows_afs[A : show](af: List[((A, nat), af_entry)]):
      (List[Char]) => List[Char] =
  shows_map[((A, nat),
              af_entry)]((a: ((A, nat), af_entry)) =>
                           {
                             val (aa, b): ((A, nat), af_entry) = a;
                             ({
                                val (f, n): (A, nat) = aa;
                                (e: af_entry) =>
                                  comp[List[Char], List[Char],
List[Char]](shows_string.apply(List('p', 'i', '(')),
             comp[List[Char], List[Char],
                   List[Char]]((ab: List[Char]) =>
                                 shows_prec[A](Zero_nat, f, ab),
                                comp[List[Char], List[Char],
                                      List[Char]](shows_string.apply(List('/')),
           comp[List[Char], List[Char],
                 List[Char]](shows_prec_nat(Zero_nat, n),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_string.apply(List(')',
                                 ' ', '=', ' ')),
         comp[List[Char], List[Char],
               List[Char]](e match {
                             case Collapse(i) =>
                               shows_prec_nat(Zero_nat, Suc(i))
                             case AFList(ids) =>
                               shows_prec_list[nat](Zero_nat,
             mapa[nat, nat]((ab: nat) => Suc(ab), ids))
                           },
                            shows_nl))))))
                              })(b)
                           },
                          af)

def starts_with(t: List[Char], s: List[Char]): Boolean =
  (if (! (nulla[Char](t)) && membera[Char](wspace, hd[Char](t)))
    starts_with(tl[Char](t), s)
    else search_len(t, s, size_list[Char](s)) match {
           case None => false
           case Some(_) => true
         })

def forallm_index_aux[A, B](p: A => nat => sum[B, Unit], i: nat, x2: List[A]):
      sum[((A, nat), B), Unit] =
  (p, i, x2) match {
  case (p, i, Nil) => Inr[Unit, ((A, nat), B)](())
  case (p, i, x :: xs) =>
    bindb[((A, nat), B), Unit,
           Unit](catcha[B, Unit,
                         ((A, nat),
                           B)]((p(x))(i),
                                (xa: B) =>
                                  Inl[((A, nat), B), Unit](((x, i), xa))),
                  (_: Unit) => forallm_index_aux[A, B](p, Suc(i), xs))
}

def forallm_index[A, B](p: A => nat => sum[B, Unit], xs: List[A]):
      sum[((A, nat), B), Unit] =
  forallm_index_aux[A, B](p, Zero_nat, xs)

def equal_gctxta[A : equal, B](x0: gctxt[A, B], x1: gctxt[A, B]): Boolean =
  (x0, x1) match {
  case (GCFun(f, list), GCHole) => false
  case (GCHole, GCFun(f, list)) => false
  case (GCFun(fa, lista), GCFun(f, list)) =>
    eq[A](fa, f) && equal_lista[gctxt[A, B]](lista, list)
  case (GCHole, GCHole) => true
}

implicit def equal_gctxt[A : equal, B]: equal[gctxt[A, B]] = new
  equal[gctxt[A, B]] {
  val `Ceta.equal` = (a: gctxt[A, B], b: gctxt[A, B]) =>
    equal_gctxta[A, B](a, b)
}

implicit def
  linorder_lab[A : equal : linorder, B : equal : linorder]: linorder[lab[A, B]]
  = new linorder[lab[A, B]] {
  val `Ceta.less_eq` = (a: lab[A, B], b: lab[A, B]) => less_eq_lab[A, B](a, b)
  val `Ceta.less` = (a: lab[A, B], b: lab[A, B]) => less_lab[A, B](a, b)
}

def prightI[A : equal, B,
             C : equal](ze: A, on: A, pl: A => A => A, ti: A => A => A,
                         mz: A => A, pi: B => nat => (A, List[A]),
                         x6: term[B, C]):
      l_poly[C, A] =
  (ze, on, pl, ti, mz, pi, x6) match {
  case (ze, on, pl, ti, mz, pi, Var(x)) => LPoly[A, C](ze, List((x, on)))
  case (ze, on, pl, ti, mz, pi, Fun(f, ts)) =>
    {
      val (c, as): (A, List[A]) = (pi(f))(size_list[term[B, C]](ts))
      val (LPoly(d, a)): l_poly[C, A] =
        sum_lpolyI[A, C](ze, pl, LPoly[A, C](c, Nil),
                          list_sumI[l_poly[C,
    A]](LPoly[A, C](ze, Nil),
         (a: l_poly[C, A]) =>
           (b: l_poly[C, A]) => sum_lpolyI[A, C](ze, pl, a, b),
         mapa[(A, term[B, C]),
               l_poly[C, A]]((at: (A, term[B, C])) =>
                               mul_lpolyI[A,
   C](ze, ti, fst[A, term[B, C]](at),
       prightI[A, B, C](ze, on, pl, ti, mz, pi, snd[A, term[B, C]](at))),
                              zip[A, term[B, C]](as, ts))));
      LPoly[A, C](mz(d), a)
    }
}

def match_to_idents_impl[A, B, C : equal,
                          D : equal](simp: (List[A]) =>
     (List[B]) => Option[(List[(term[C, D], term[C, D])], nat)],
                                      ident:
((term[C, D], term[C, D])) => Option[nat],
                                      t_l: A):
      Option[nat] =
  bind[(List[(term[C, D], term[C, D])], nat),
        nat]((simp(List(t_l)))(Nil),
              (a: (List[(term[C, D], term[C, D])], nat)) =>
                {
                  val (smp, i): (List[(term[C, D], term[C, D])], nat) = a
                  val ids: List[Option[nat]] =
                    mapa[(term[C, D], term[C, D]),
                          Option[nat]](ident, ident_prob_of_smp[C, D](smp));
                  (if (membera[Option[nat]](ids, None)) None
                    else Some[nat](plus_nat(max_list(mapa[Option[nat],
                   nat]((aa: Option[nat]) => the[nat](aa), ids)),
     i)))
                })

def match_prob_of_rp_impl[A, B : equal, C : equal, D,
                           E](mu: A => term[B, C], w: (term[B, C]) => List[A],
                               x2: (term[B, C], term[D, E])):
      List[(term[B, C], term[D, E])] =
  (mu, w, x2) match {
  case (mu, w, (t, Var(x))) => List((t, Var[E, D](x)))
  case (mu, w, (t, Fun(v, va))) =>
    {
      val sterms: List[term[B, C]] =
        remdups[term[B, C]](t :: mapa[A, term[B, C]](mu, w(t)))
      val filt: (List[term[B, C]]) => List[term[B, C]] =
        (a: List[term[B, C]]) =>
          filter[term[B, C]]((x: term[B, C]) => ! (is_Var[B, C](x)), a)
      val uterms: List[term[B, C]] =
        maps[term[B, C],
              term[B, C]]((ta: term[B, C]) => filt(supteq_list[B, C](ta)),
                           sterms);
      mapa[term[B, C],
            (term[B, C],
              term[D, E])]((u: term[B, C]) => (u, Fun[D, E](v, va)),
                            remdups[term[B, C]](uterms))
    }
}

def redex_decision[A, B : equal, C : equal, D, E, F, G : equal,
                    H : equal](mu: A => term[B, C],
                                simp: (List[(term[B, C], term[D, E])]) =>
(List[F]) => Option[(List[(term[G, H], term[G, H])], nat)],
                                ident:
                                  ((term[G, H], term[G, H])) => Option[nat],
                                w: (term[B, C]) => List[A],
                                rp: (term[B, C], term[D, E])):
      Boolean =
  {
    val mps: List[(term[B, C], term[D, E])] =
      match_prob_of_rp_impl[A, B, C, D, E](mu, w, rp)
    val a: List[Option[nat]] =
      mapa[(term[B, C], term[D, E]),
            Option[nat]]((a: (term[B, C], term[D, E])) =>
                           match_to_idents_impl[(term[B, C], term[D, E]), F, G,
         H](simp, ident, a),
                          mps);
    list_ex[Option[nat]]((sol: Option[nat]) => ! (is_none[nat](sol)), a)
  }

def poly_vars_list[A : equal, B](p: List[(List[(A, nat)], B)]): List[A] =
  remdups[A](maps[(List[(A, nat)], B),
                   A](comp[List[(A, nat)], List[A],
                            (List[(A, nat)],
                              B)]((a: List[(A, nat)]) =>
                                    mapa[(A, nat),
  A]((aa: (A, nat)) => fst[A, nat](aa), a),
                                   (a: (List[(A, nat)], B)) =>
                                     fst[List[(A, nat)], B](a)),
                       p))

def indent(p: (List[Char]) => List[Char], s: List[Char]): List[Char] =
  maps[Char,
        Char]((c: Char) => (if (c == newline) List(newline, ' ') else List(c)),
               p(s))

def label_s_ns_impl[A : show,
                     B : show](cS: ((term[A, B], term[A, B])) =>
                                     sum[(List[Char]) => List[Char], Unit],
                                cNS: ((term[A, B], term[A, B])) =>
                                       sum[(List[Char]) => List[Char], Unit],
                                s: (term[A, B], nat), t: (term[A, B], nat)):
      (Boolean, Boolean) =
  {
    val (sa, i): (term[A, B], nat) = s
    val (ta, j): (term[A, B], nat) = t;
    (if (isOK[(List[Char]) => List[Char], Unit](cS((sa, ta)))) (true, true)
      else (if (isOK[(List[Char]) => List[Char], Unit](cNS((sa, ta))))
             (less_nat(j, i), less_eq_nat(j, i)) else (false, false)))
  }

def funas_term_list[A : equal, B]: (term[A, B]) => List[(A, nat)] =
  comp[List[(A, nat)], List[(A, nat)],
        term[A, B]]((a: List[(A, nat)]) => remdups[(A, nat)](a),
                     (a: term[A, B]) => funas_term_lista[A, B](a))

def mk_subst_domain[A : equal, B : equal](sigma: List[(A, term[B, A])]):
      List[(A, term[B, A])] =
  {
    val tau: A => term[B, A] = mk_subst[A, B](sigma);
    remdups[(A, term[B, A])](filter[(A, term[B,
      A])]((a: (A, term[B, A])) =>
             {
               val (x, t): (A, term[B, A]) = a;
               ! (equal_terma[B, A](Var[A, B](x), t))
             },
            mapa[(A, term[B, A]),
                  (A, term[B, A])]((a: (A, term[B, A])) =>
                                     {
                                       val (x, _): (A, term[B, A]) = a;
                                       (x, tau(x))
                                     },
                                    sigma)))
  }

def shows_prec_ctxt[A : show, B : show](d: nat):
      (ctxt[A, B]) => (List[Char]) => List[Char] =
  (a: ctxt[A, B]) =>
    shows_ctxt[A, B]((aa: A) =>
                       (b: List[Char]) => shows_prec[A](Zero_nat, aa, b),
                      (aa: B) =>
                        (b: List[Char]) => shows_prec[B](Zero_nat, aa, b),
                      a)

def only_eta_rules[A : equal : show,
                    B : equal : show](e: List[(term[A, B], term[A, B])],
                                       r_eta: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[((term[A, B], term[A, B]), (List[Char]) => List[Char]), Unit,
          (List[Char]) =>
            List[Char]](forallm[(term[A, B], term[A, B]),
                                 (List[Char]) =>
                                   List[Char]]((a: (term[A, B], term[A, B])) =>
         {
           val (l, r): (term[A, B], term[A, B]) = a;
           check[(List[Char]) =>
                   List[Char]]((l, r) match {
                                 case (Var(_), b) => false
                                 case (Fun(f, ls), Var(_)) => false
                                 case (Fun(f, ls), Fun(g, rs)) =>
                                   eq[A](f, g) &&
                                     (equal_nata(size_list[term[A, B]](ls),
          nat(BigInt(2))) &&
                                       (equal_nata(size_list[term[A, B]](rs),
            nat(BigInt(2))) &&
 (equal_terma[A, B](nth[term[A, B]](ls, one_nat),
                     nth[term[A, B]](rs, one_nat)) &&
   list_ex[(term[A, B],
             term[A, B])]((aa: (term[A, B], term[A, B])) =>
                            {
                              val (la, ra): (term[A, B], term[A, B]) = aa;
                              instance_rule[A, B,
     B]((hd[term[A, B]](ls), hd[term[A, B]](rs)), (la, ra))
                            },
                           r_eta))))
                               },
                                comp[List[Char], List[Char],
                                      List[Char]](shows_prec_list[Char](Zero_nat,
                                 List('r', 'u', 'l', 'e', ' ')),
           comp[List[Char], List[Char],
                 List[Char]](shows_rule[A,
 B]((aa: A) => (b: List[Char]) => shows_prec[A](Zero_nat, aa, b),
     (aa: B) => (b: List[Char]) => shows_prec[B](Zero_nat, aa, b),
     List(' ', '-', '>', ' '), (l, r)),
                              shows_prec_list[Char](Zero_nat,
             List(' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', 'n', ' ', '(',
                   'e', 't', 'a', '-', 'e', 'x', 'p', 'a', 'n', 'd', 'e', 'd',
                   ')', ' ', 'o', 'r', 'i', 'g', 'i', 'n', 'a', 'l', ' ', 'r',
                   'u', 'l', 'e')))))
         },
        e),
                         (x: ((term[A, B], term[A, B]),
                               (List[Char]) => List[Char]))
                           => Inl[(List[Char]) => List[Char],
                                   Unit](snd[(term[A, B], term[A, B]),
      (List[Char]) => List[Char]](x)))

def pa[A, B, C, D](x0: dpp_ops_ext[A, B, C, D]):
      A => List[(term[B, C], term[B, C])] =
  x0 match {
  case Dpp_ops_ext(dpp, p, pw, pairs, q, r, rw, rules, q_empty,
                    rules_no_left_var, rules_non_collapsing, is_QNF,
                    nFQ_subset_NF_rules, rules_map, reverse_rules_map,
                    intersect_pairs, replace_pair, intersect_rules, delete_P_Pw,
                    delete_R_Rw, split_pairs, split_rules, mk, more)
    => p
}

def rc[A, B, C, D](x0: dpp_ops_ext[A, B, C, D]):
      A => List[(term[B, C], term[B, C])] =
  x0 match {
  case Dpp_ops_ext(dpp, p, pw, pairs, q, r, rw, rules, q_empty,
                    rules_no_left_var, rules_non_collapsing, is_QNF,
                    nFQ_subset_NF_rules, rules_map, reverse_rules_map,
                    intersect_pairs, replace_pair, intersect_rules, delete_P_Pw,
                    delete_R_Rw, split_pairs, split_rules, mk, more)
    => r
}

def pwa[A, B, C, D](x0: dpp_ops_ext[A, B, C, D]):
      A => List[(term[B, C], term[B, C])] =
  x0 match {
  case Dpp_ops_ext(dpp, p, pw, pairs, q, r, rw, rules, q_empty,
                    rules_no_left_var, rules_non_collapsing, is_QNF,
                    nFQ_subset_NF_rules, rules_map, reverse_rules_map,
                    intersect_pairs, replace_pair, intersect_rules, delete_P_Pw,
                    delete_R_Rw, split_pairs, split_rules, mk, more)
    => pw
}

def rwc[A, B, C, D](x0: dpp_ops_ext[A, B, C, D]):
      A => List[(term[B, C], term[B, C])] =
  x0 match {
  case Dpp_ops_ext(dpp, p, pw, pairs, q, r, rw, rules, q_empty,
                    rules_no_left_var, rules_non_collapsing, is_QNF,
                    nFQ_subset_NF_rules, rules_map, reverse_rules_map,
                    intersect_pairs, replace_pair, intersect_rules, delete_P_Pw,
                    delete_R_Rw, split_pairs, split_rules, mk, more)
    => rw
}

def uncurry_proc[A, B : equal : show](i: dpp_ops_ext[A, B, List[Char], Unit],
                                       info:
 (B, (List[((B, nat), List[B])],
       (List[(term[B, List[Char]], term[B, List[Char]])],
         List[(term[B, List[Char]], term[B, List[Char]])]))),
                                       fmap:
 (List[((B, nat), List[B])]) => B => nat => B,
                                       check_inj:
 B => nat =>
        (List[((B, nat), List[B])]) => sum[(List[Char]) => List[Char], Unit],
                                       p:
 List[(term[B, List[Char]], term[B, List[Char]])],
                                       r:
 List[(term[B, List[Char]], term[B, List[Char]])],
                                       dpp: A):
      sum[(List[Char]) => List[Char], A] =
  {
    val (a, (sml, (u, eb))):
          (B, (List[((B, nat), List[B])],
                (List[(term[B, List[Char]], term[B, List[Char]])],
                  List[(term[B, List[Char]], term[B, List[Char]])])))
      = info
    val paa: List[(term[B, List[Char]], term[B, List[Char]])] =
      (pa[A, B, List[Char], Unit](i)).apply(dpp)
    val pw: List[(term[B, List[Char]], term[B, List[Char]])] =
      (pwa[A, B, List[Char], Unit](i)).apply(dpp)
    val ra: List[(term[B, List[Char]], term[B, List[Char]])] =
      (rc[A, B, List[Char], Unit](i)).apply(dpp)
    val rw: List[(term[B, List[Char]], term[B, List[Char]])] =
      (rwc[A, B, List[Char], Unit](i)).apply(dpp)
    val (e, ew):
          (List[(term[B, List[Char]], term[B, List[Char]])],
            List[(term[B, List[Char]], term[B, List[Char]])])
      = uncurry_eta_split[B, List[Char]](eb, ra)
    val sm: B => nat => List[B] = sig_list_to_sig_map[B](a, sml, fmap)
    val uP: List[(term[B, List[Char]], term[B, List[Char]])] =
      (uncurry_rules[B, List[Char]](a, sm)).apply(paa)
    val uPw: List[(term[B, List[Char]], term[B, List[Char]])] =
      (uncurry_rules[B, List[Char]](a, sm)).apply(pw)
    val r_eta: List[(term[B, List[Char]], term[B, List[Char]])] = e ++ ra
    val rw_eta: List[(term[B, List[Char]], term[B, List[Char]])] = ew ++ rw
    val uR: List[(term[B, List[Char]], term[B, List[Char]])] =
      (uncurry_rules[B, List[Char]](a, sm)).apply(r_eta)
    val uRw: List[(term[B, List[Char]], term[B, List[Char]])] =
      (uncurry_rules[B, List[Char]](a, sm)).apply(rw_eta);
    {
      val s: List[(term[B, List[Char]], term[B, List[Char]])] =
        uncurry_of_sig_list[B](a, sml, sm);
      bindb[(List[Char]) => List[Char], Unit,
             Unit](check[(List[Char]) =>
                           List[Char]](nulla[term[B,
           List[Char]]]((qc[A, B, List[Char], Unit](i)).apply(dpp)),
shows_prec_list[Char](Zero_nat,
                       List('s', 't', 'r', 'a', 't', 'e', 'g', 'y', ' ', 'n',
                             'o', 't', ' ', 's', 'u', 'p', 'p', 'o', 'r', 't',
                             'e', 'd', ' ', 'f', 'o', 'r', ' ', 'u', 'n', 'c',
                             'u', 'r', 'r', 'y', 'i', 'n', 'g'))),
                    (_: Unit) =>
                      bindb[(List[Char]) => List[Char], Unit,
                             Unit](only_eta_rules[B, List[Char]](e, r_eta),
                                    (_: Unit) =>
                                      bindb[(List[Char]) => List[Char], Unit,
     Unit](only_eta_rules[B, List[Char]](ew, rw_eta),
            (_: Unit) =>
              bindb[(List[Char]) => List[Char], Unit,
                     Unit](((check_inj(a))(nat(BigInt(2))))(sml),
                            (_: Unit) =>
                              bindb[(List[Char]) => List[Char], Unit,
                                     Unit](catcha[((term[B, List[Char]],
             term[B, List[Char]]),
            (List[Char]) => List[Char]),
           Unit,
           (List[Char]) =>
             List[Char]](forallm[(term[B, List[Char]], term[B, List[Char]]),
                                  (List[Char]) =>
                                    List[Char]]((b:
           (term[B, List[Char]], term[B, List[Char]]))
          => {
               val (l, _): (term[B, List[Char]], term[B, List[Char]]) = b;
               check[(List[Char]) =>
                       List[Char]](hvf_term[B, List[Char]](a, l),
                                    comp[List[Char], List[Char],
  List[Char]](shows_prec_list[Char](Zero_nat,
                                     List('h', 'e', 'a', 'd', ' ', 'v', 'a',
   'r', 'i', 'a', 'b', 'l', 'e', ' ', 'i', 'n', ' ', 'l', 'h', 's', ' ')),
               comp[List[Char], List[Char],
                     List[Char]](shows_term[B,
     List[Char]]((aa: B) => (ba: List[Char]) => shows_prec[B](Zero_nat, aa, ba),
                  (aa: List[Char]) => shows_prec_list[Char](Zero_nat, aa), l),
                                  shows_prec_list[Char](Zero_nat,
                 List(' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e',
                       'd')))))
             },
         paa),
                          (x: ((term[B, List[Char]], term[B, List[Char]]),
                                (List[Char]) => List[Char]))
                            => Inl[(List[Char]) => List[Char],
                                    Unit](snd[(term[B, List[Char]],
        term[B, List[Char]]),
       (List[Char]) => List[Char]](x))),
    (_: Unit) =>
      bindb[(List[Char]) => List[Char], Unit,
             Unit](catcha[((term[B, List[Char]], term[B, List[Char]]),
                            (List[Char]) => List[Char]),
                           Unit,
                           (List[Char]) =>
                             List[Char]](forallm[(term[B, List[Char]],
           term[B, List[Char]]),
          (List[Char]) =>
            List[Char]]((b: (term[B, List[Char]], term[B, List[Char]])) =>
                          {
                            val (l, _):
                                  (term[B, List[Char]], term[B, List[Char]])
                              = b;
                            check[(List[Char]) =>
                                    List[Char]](hvf_term[B, List[Char]](a, l),
         comp[List[Char], List[Char],
               List[Char]](shows_prec_list[Char](Zero_nat,
          List('h', 'e', 'a', 'd', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e',
                ' ', 'i', 'n', ' ', 'l', 'h', 's', ' ')),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_term[B,
                  List[Char]]((aa: B) =>
                                (ba: List[Char]) =>
                                  shows_prec[B](Zero_nat, aa, ba),
                               (aa: List[Char]) =>
                                 shows_prec_list[Char](Zero_nat, aa),
                               l),
       shows_prec_list[Char](Zero_nat,
                              List(' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o',
                                    'w', 'e', 'd')))))
                          },
                         pw),
  (x: ((term[B, List[Char]], term[B, List[Char]]), (List[Char]) => List[Char]))
    => Inl[(List[Char]) => List[Char],
            Unit](snd[(term[B, List[Char]], term[B, List[Char]]),
                       (List[Char]) => List[Char]](x))),
                    (_: Unit) =>
                      bindb[(List[Char]) => List[Char], Unit,
                             Unit](catcha[((term[B, List[Char]],
     term[B, List[Char]]),
    (List[Char]) => List[Char]),
   Unit,
   (List[Char]) =>
     List[Char]](forallm[(term[B, List[Char]], term[B, List[Char]]),
                          (List[Char]) =>
                            List[Char]]((b:
   (term[B, List[Char]], term[B, List[Char]]))
  => {
       val (l, _): (term[B, List[Char]], term[B, List[Char]]) = b;
       check[(List[Char]) =>
               List[Char]](hvf_term[B, List[Char]](a, l),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_prec_list[Char](Zero_nat,
                             List('h', 'e', 'a', 'd', ' ', 'v', 'a', 'r', 'i',
                                   'a', 'b', 'l', 'e', ' ', 'i', 'n', ' ', 'l',
                                   'h', 's', ' ')),
       comp[List[Char], List[Char],
             List[Char]](shows_term[B, List[Char]]((aa: B) =>
             (ba: List[Char]) => shows_prec[B](Zero_nat, aa, ba),
            (aa: List[Char]) => shows_prec_list[Char](Zero_nat, aa), l),
                          shows_prec_list[Char](Zero_nat,
         List(' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e', 'd')))))
     },
 r_eta),
                  (x: ((term[B, List[Char]], term[B, List[Char]]),
                        (List[Char]) => List[Char]))
                    => Inl[(List[Char]) => List[Char],
                            Unit](snd[(term[B, List[Char]],
term[B, List[Char]]),
                                       (List[Char]) => List[Char]](x))),
                                    (_: Unit) =>
                                      bindb[(List[Char]) => List[Char], Unit,
     Unit](catcha[((term[B, List[Char]], term[B, List[Char]]),
                    (List[Char]) => List[Char]),
                   Unit,
                   (List[Char]) =>
                     List[Char]](forallm[(term[B, List[Char]],
   term[B, List[Char]]),
  (List[Char]) =>
    List[Char]]((b: (term[B, List[Char]], term[B, List[Char]])) =>
                  {
                    val (l, _): (term[B, List[Char]], term[B, List[Char]]) = b;
                    check[(List[Char]) =>
                            List[Char]](hvf_term[B, List[Char]](a, l),
 comp[List[Char], List[Char],
       List[Char]](shows_prec_list[Char](Zero_nat,
  List('h', 'e', 'a', 'd', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ',
        'i', 'n', ' ', 'l', 'h', 's', ' ')),
                    comp[List[Char], List[Char],
                          List[Char]](shows_term[B,
          List[Char]]((aa: B) =>
                        (ba: List[Char]) => shows_prec[B](Zero_nat, aa, ba),
                       (aa: List[Char]) => shows_prec_list[Char](Zero_nat, aa),
                       l),
                                       shows_prec_list[Char](Zero_nat,
                      List(' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w',
                            'e', 'd')))))
                  },
                 rw_eta),
                                  (x: ((term[B, List[Char]],
 term[B, List[Char]]),
(List[Char]) => List[Char]))
                                    => Inl[(List[Char]) => List[Char],
    Unit](snd[(term[B, List[Char]], term[B, List[Char]]),
               (List[Char]) => List[Char]](x))),
            (_: Unit) =>
              bindb[(List[Char]) => List[Char], Unit,
                     Unit](eta_closed_rules[B, List[Char]](a, sm, r_eta, r_eta),
                            (_: Unit) =>
                              bindb[(List[Char]) => List[Char], Unit,
                                     Unit](eta_closed_rules[B,
                     List[Char]](a, sm, rw_eta, rw_eta),
    (_: Unit) =>
      bindb[(List[Char]) => List[Char], Unit,
             Unit](catcha[(term[B, List[Char]], term[B, List[Char]]), Unit,
                           (List[Char]) =>
                             List[Char]](check_subseteq[(term[B, List[Char]],
                  term[B, List[Char]])](uP, p),
  (x: (term[B, List[Char]], term[B, List[Char]])) =>
    Inl[(List[Char]) => List[Char],
         Unit](comp[List[Char], List[Char],
                     List[Char]](shows_prec_list[Char](Zero_nat,
                List('u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', ' ', 'p', 'a',
                      'i', 'r', ' ')),
                                  comp[List[Char], List[Char],
List[Char]](shows_rule[B, List[Char]]((aa: B) =>
(b: List[Char]) => shows_prec[B](Zero_nat, aa, b),
                                       (aa: List[Char]) =>
 shows_prec_list[Char](Zero_nat, aa),
                                       List(' ', '-', '>', ' '), x),
             shows_prec_list[Char](Zero_nat,
                                    List(' ', 'i', 's', ' ', 'm', 'i', 's', 's',
  'i', 'n', 'g')))))),
                    (_: Unit) =>
                      bindb[(List[Char]) => List[Char], Unit,
                             Unit](catcha[(term[B, List[Char]],
    term[B, List[Char]]),
   Unit,
   (List[Char]) =>
     List[Char]](check_subseteq[(term[B, List[Char]],
                                  term[B, List[Char]])](uPw, p),
                  (x: (term[B, List[Char]], term[B, List[Char]])) =>
                    Inl[(List[Char]) => List[Char],
                         Unit](comp[List[Char], List[Char],
                                     List[Char]](shows_prec_list[Char](Zero_nat,
                                List('u', 'n', 'c', 'u', 'r', 'r', 'i', 'e',
                                      'd', ' ', 'p', 'a', 'i', 'r', ' ')),
          comp[List[Char], List[Char],
                List[Char]](shows_rule[B,
List[Char]]((aa: B) => (b: List[Char]) => shows_prec[B](Zero_nat, aa, b),
             (aa: List[Char]) => shows_prec_list[Char](Zero_nat, aa),
             List(' ', '-', '>', ' '), x),
                             shows_prec_list[Char](Zero_nat,
            List(' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g')))))),
                                    (_: Unit) =>
                                      bindb[(List[Char]) => List[Char], Unit,
     Unit](catcha[(term[B, List[Char]], term[B, List[Char]]), Unit,
                   (List[Char]) =>
                     List[Char]](check_subseteq[(term[B, List[Char]],
          term[B, List[Char]])](uR, r),
                                  (x: (term[B, List[Char]],
term[B, List[Char]]))
                                    => Inl[(List[Char]) => List[Char],
    Unit](comp[List[Char], List[Char],
                List[Char]](shows_prec_list[Char](Zero_nat,
           List('u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', ' ', 'r', 'u', 'l',
                 'e', ' ')),
                             comp[List[Char], List[Char],
                                   List[Char]](shows_rule[B,
                   List[Char]]((aa: B) =>
                                 (b: List[Char]) =>
                                   shows_prec[B](Zero_nat, aa, b),
                                (aa: List[Char]) =>
                                  shows_prec_list[Char](Zero_nat, aa),
                                List(' ', '-', '>', ' '), x),
        shows_prec_list[Char](Zero_nat,
                               List(' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i',
                                     'n', 'g')))))),
            (_: Unit) =>
              bindb[(List[Char]) => List[Char], Unit,
                     Unit](catcha[(term[B, List[Char]], term[B, List[Char]]),
                                   Unit,
                                   (List[Char]) =>
                                     List[Char]](check_subseteq[(term[B,
                               List[Char]],
                          term[B, List[Char]])](uRw, r),
          (x: (term[B, List[Char]], term[B, List[Char]])) =>
            Inl[(List[Char]) => List[Char],
                 Unit](comp[List[Char], List[Char],
                             List[Char]](shows_prec_list[Char](Zero_nat,
                        List('u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', ' ',
                              'r', 'u', 'l', 'e', ' ')),
  comp[List[Char], List[Char],
        List[Char]](shows_rule[B, List[Char]]((aa: B) =>
        (b: List[Char]) => shows_prec[B](Zero_nat, aa, b),
       (aa: List[Char]) => shows_prec_list[Char](Zero_nat, aa),
       List(' ', '-', '>', ' '), x),
                     shows_prec_list[Char](Zero_nat,
    List(' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g')))))),
                            (_: Unit) =>
                              bindb[(List[Char]) => List[Char], Unit,
                                     Unit](catcha[(term[B, List[Char]],
            term[B, List[Char]]),
           Unit,
           (List[Char]) =>
             List[Char]](check_CS_subseteq[B, List[Char]](s, u),
                          (x: (term[B, List[Char]], term[B, List[Char]])) =>
                            Inl[(List[Char]) => List[Char],
                                 Unit](comp[List[Char], List[Char],
     List[Char]](shows_prec_list[Char](Zero_nat,
List('u', 'n', 'c', 'u', 'r', 'r', 'y', ' ', 'r', 'u', 'l', 'e', ' ')),
                  comp[List[Char], List[Char],
                        List[Char]](shows_rule[B,
        List[Char]]((aa: B) =>
                      (b: List[Char]) => shows_prec[B](Zero_nat, aa, b),
                     (aa: List[Char]) => shows_prec_list[Char](Zero_nat, aa),
                     List(' ', '-', '>', ' '), x),
                                     shows_prec_list[Char](Zero_nat,
                    List(' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n',
                          'g')))))),
    (_: Unit) =>
      bindb[(List[Char]) => List[Char], Unit,
             Unit](catcha[(term[B, List[Char]], term[B, List[Char]]), Unit,
                           (List[Char]) =>
                             List[Char]](check_CS_subseteq[B, List[Char]](u, s),
  (x: (term[B, List[Char]], term[B, List[Char]])) =>
    Inl[(List[Char]) => List[Char],
         Unit](comp[List[Char], List[Char],
                     List[Char]](shows_prec_list[Char](Zero_nat,
                List('r', 'u', 'l', 'e', ' ')),
                                  comp[List[Char], List[Char],
List[Char]](shows_rule[B, List[Char]]((aa: B) =>
(b: List[Char]) => shows_prec[B](Zero_nat, aa, b),
                                       (aa: List[Char]) =>
 shows_prec_list[Char](Zero_nat, aa),
                                       List(' ', '-', '>', ' '), x),
             shows_prec_list[Char](Zero_nat,
                                    List(' ', 'i', 's', ' ', 'n', 'o', 't', ' ',
  'a', 'n', ' ', 'u', 'n', 'c', 'u', 'r', 'r', 'y', ' ', 'r', 'u', 'l',
  'e')))))),
                    (_: Unit) =>
                      catcha[(term[B, List[Char]], term[B, List[Char]]), Unit,
                              (List[Char]) =>
                                List[Char]](check_subseteq[(term[B, List[Char]],
                     term[B, List[Char]])](u, r),
     (x: (term[B, List[Char]], term[B, List[Char]])) =>
       Inl[(List[Char]) => List[Char],
            Unit](comp[List[Char], List[Char],
                        List[Char]](shows_prec_list[Char](Zero_nat,
                   List('u', 'n', 'c', 'u', 'r', 'r', 'y', ' ', 'r', 'u', 'l',
                         'e', ' ')),
                                     comp[List[Char], List[Char],
   List[Char]](shows_rule[B, List[Char]]((aa: B) =>
   (b: List[Char]) => shows_prec[B](Zero_nat, aa, b),
  (aa: List[Char]) => shows_prec_list[Char](Zero_nat, aa),
  List(' ', '-', '>', ' '), x),
                shows_prec_list[Char](Zero_nat,
                                       List(' ', 'i', 's', ' ', 'm', 'i', 's',
     's', 'i', 'n', 'g', ' ', 'i', 'n', ' ', 'n', 'e', 'w', ' ', 'T', 'R',
     'S'))))))))))))))))))))))
    }
      match {
      case Inl(aa) => Inl[(List[Char]) => List[Char], A](aa)
      case Inr(_) =>
        Inr[A, (List[Char]) =>
                 List[Char]]((mkc[A, B, List[Char],
                                   Unit](i)).apply(uP).apply(uPw).apply(Nil).apply(uR).apply(uRw ++
               u))
    }
  }

def xml2dp_disproof[A](xml2name: xml => sum[List[Char], A], x: xml):
      sum[List[Char], dp_disproof[A, List[Char]]] =
  xml1element[dp_disproof[A, List[Char]],
               dp_disproof[A, List[Char]]](List('d', 'p', 'N', 'o', 'n', 't',
         'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'o',
         'f'),
    (a: xml) =>
      xml_options[dp_disproof[A, List[Char]]](List((List('d', 'p', 'R', 'u',
                  'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l'),
             (aa: xml) =>
               xml3elements[List[(term[A, List[Char]], term[A, List[Char]])],
                             List[(term[A, List[Char]], term[A, List[Char]])],
                             dp_disproof[A, List[Char]],
                             dp_disproof[A,
  List[Char]]](List('d', 'p', 'R', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a',
                     'l'),
                (ab: xml) =>
                  xml1element[List[(term[A, List[Char]], term[A, List[Char]])],
                               List[(term[A, List[Char]],
                                      term[A,
    List[Char]])]](List('d', 'p', 's'), rules[A](xml2name),
                    id[List[(term[A, List[Char]], term[A, List[Char]])]], ab),
                (ab: xml) =>
                  xml1element[List[(term[A, List[Char]], term[A, List[Char]])],
                               List[(term[A, List[Char]],
                                      term[A,
    List[Char]])]](List('t', 'r', 's'), rules[A](xml2name),
                    id[List[(term[A, List[Char]], term[A, List[Char]])]], ab),
                (ab: xml) => xml2dp_disproof[A](xml2name, ab),
                (ab: List[(term[A, List[Char]], term[A, List[Char]])]) =>
                  (b: List[(term[A, List[Char]], term[A, List[Char]])]) =>
                    (c: dp_disproof[A, List[Char]]) =>
                      DP_Rule_Removal[A, List[Char]](ab, b, c),
                aa)),
            (List('l', 'o', 'o', 'p'),
              (aa: xml) =>
                xml_change[(term[A, List[Char]],
                             (List[(pos, ((term[A, List[Char]],
    term[A, List[Char]]),
   (Boolean, term[A, List[Char]])))],
                               (List[(List[Char], term[A, List[Char]])],
                                 ctxt[A, List[Char]]))),
                            dp_disproof[A,
 List[Char]]](loop[A](xml2name),
               (ab: (term[A, List[Char]],
                      (List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                    (Boolean, term[A, List[Char]])))],
                        (List[(List[Char], term[A, List[Char]])],
                          ctxt[A, List[Char]]))))
                 => {
                      val (s, (rseq, (ac, b))):
                            (term[A, List[Char]],
                              (List[(pos, ((term[A, List[Char]],
     term[A, List[Char]]),
    (Boolean, term[A, List[Char]])))],
                                (List[(List[Char], term[A, List[Char]])],
                                  ctxt[A, List[Char]])))
                        = ab;
                      DP_Loop[A, List[Char]](s, rseq, ac, b)
                    },
               aa))),
       a),
    id[dp_disproof[A, List[Char]]], x)

def eqWhiteSpace(s1: List[Char], s2: List[Char]): Boolean =
  debug[Boolean](List('0'),
                  List('c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' ', 'w',
                        'h', 'e', 't', 'h', 'e', 'r', ' ', 'p', 'a', 'r', 's',
                        'e', 'd', ' ', 'i', 'n', 'p', 'u', 't', ' ', 'c', 'o',
                        'r', 'r', 'e', 's', 'p', 'o', 'n', 'd', 's', ' ', 't',
                        'o', ' ', 'r', 'e', 'a', 'l', ' ', 'i', 'n', 'p', 'u',
                        't'),
                  equal_lista[Char](filter[Char]((c: Char) =>
           ! (membera[Char](wspace, c)),
          s1),
                                     filter[Char]((c: Char) =>
            ! (membera[Char](wspace, c)),
           s2)))

def reverse_rules_mapa[A, B, C, D](x0: dpp_ops_ext[A, B, C, D]):
      A => ((B, nat)) => List[(term[B, C], term[B, C])] =
  x0 match {
  case Dpp_ops_ext(dpp, p, pw, pairs, q, r, rw, rules, q_empty,
                    rules_no_left_var, rules_non_collapsing, is_QNF,
                    nFQ_subset_NF_rules, rules_map, reverse_rules_map,
                    intersect_pairs, replace_pair, intersect_rules, delete_P_Pw,
                    delete_R_Rw, split_pairs, split_rules, mk, more)
    => reverse_rules_map
}

def rules_non_collapsinga[A, B, C, D](x0: dpp_ops_ext[A, B, C, D]):
      A => Boolean =
  x0 match {
  case Dpp_ops_ext(dpp, p, pw, pairs, q, r, rw, rules, q_empty,
                    rules_no_left_var, rules_non_collapsing, is_QNF,
                    nFQ_subset_NF_rules, rules_map, reverse_rules_map,
                    intersect_pairs, replace_pair, intersect_rules, delete_P_Pw,
                    delete_R_Rw, split_pairs, split_rules, mk, more)
    => rules_non_collapsing
}

def reverse_capRM_dpp[A, B : equal,
                       C : equal](i: dpp_ops_ext[A, B, C, Unit], dpp: A):
      (term[B, C]) => gctxt[B, C] =
  capRM[B, C]((rules_non_collapsinga[A, B, C, Unit](i)).apply(dpp),
               (reverse_rules_mapa[A, B, C, Unit](i)).apply(dpp))

def icap_impl_dpp_mv[A, B : equal](i: dpp_ops_ext[A, B, List[Char], Unit],
                                    d: A):
      (List[term[B, List[Char]]]) =>
        (term[B, List[Char]]) => term[B, sum[Unit, List[Char]]] =
  {
    val qr: Boolean = (nFQ_subset_NF_rulesb[A, B, List[Char], Unit](i)).apply(d)
    val qnf: (term[B, List[Char]]) => Boolean =
      (is_QNFc[A, B, List[Char], Unit](i)).apply(d)
    val r: List[(term[B, List[Char]], term[B, List[Char]])] =
      (rulesd[A, B, List[Char], Unit](i)).apply(d)
    val ic: (List[term[B, List[Char]]]) =>
              ((List[Char]) => Boolean) =>
                (term[B, List[Char]]) => term[B, sum[Unit, List[Char]]]
      = (a: List[term[B, List[Char]]]) =>
          (b: (List[Char]) => Boolean) =>
            (c: term[B, List[Char]]) =>
              icap_impl[B](qr, qnf,
                            mapa[(term[B, List[Char]], term[B, List[Char]]),
                                  term[B,
List[Char]]]((aa: (term[B, List[Char]], term[B, List[Char]])) =>
               fst[term[B, List[Char]], term[B, List[Char]]](aa),
              r),
                            a, b, c);
    (s: List[term[B, List[Char]]]) =>
      {
        val sa: List[term[B, List[Char]]] =
          mapa[term[B, List[Char]],
                term[B, List[Char]]]((a: term[B, List[Char]]) =>
                                       map_vars[List[Char], List[Char],
         B]((aa: List[Char]) => 'x' :: aa, a),
                                      s)
        val sx: List[List[Char]] =
          remdups[List[Char]](maps[term[B, List[Char]],
                                    List[Char]](vars_term_list[B, List[Char]],
         sa));
        (t: term[B, List[Char]]) =>
          ((ic(sa))((a: List[Char]) =>
                      membera[List[Char]](sx,
   a)))(map_vars[List[Char], List[Char], B]((a: List[Char]) => 'x' :: a, t))
      }
  }

implicit def
  linorder_term[A : equal : linorder, B : equal : linorder]:
    linorder[term[A, B]]
  = new linorder[term[A, B]] {
  val `Ceta.less_eq` = (a: term[A, B], b: term[A, B]) =>
    less_eq_term[A, B](a, b)
  val `Ceta.less` = (a: term[A, B], b: term[A, B]) => less_term[A, B](a, b)
}

abstract sealed class ta_rule_impl[+A, +B]
final case class TA_rule_impl[B, A](a: B, b: List[A], c: A, d: rbt[A, Unit])
  extends ta_rule_impl[A, B]

abstract sealed class ta_impl_ext[+A, +B, +C]
final case class
  Ta_impl_ext[A, B,
               C](a: rbt[A, Unit], b: rbt[(B, nat), List[ta_rule_impl[A, B]]],
                   c: List[A], d: rbt[A, Unit], e: List[(A, A)],
                   f: A => rbt[A, Unit], g: A => rbt[A, Unit], h: C)
  extends ta_impl_ext[A, B, C]

def ta_rqs_set[A, B, C](x0: ta_impl_ext[A, B, C]): rbt[A, Unit] = x0 match {
  case Ta_impl_ext(ta_final_impl, ta_rules_impl, ta_rqs_impl, ta_rqs_set,
                    ta_eps_impl, ta_epss_impl, ta_epsrs_impl, more)
    => ta_rqs_set
}

def rsym_impl[A, B](x0: ta_rule_impl[A, B]): (B, nat) = x0 match {
  case TA_rule_impl(f, qsa, q, qs) => (f, size_list[A](qsa))
}

def conv_ta_rule[A, B](eps: A => rbt[A, Unit], x1: ta_rule[A, B]):
      ta_rule_impl[A, B] =
  (eps, x1) match {
  case (eps, TA_rule(f, qs, q)) => TA_rule_impl[B, A](f, qs, q, eps(q))
}

def rtrancl_rbt_impl[A : linorder]:
      (List[(A, A)]) => (List[A]) => rbt[A, Unit] =
  (a: List[(A, A)]) =>
    rtrancl_impl[A, rbt[A, Unit]]((r: List[(A, A)]) =>
                                    {
                                      val rm: rbt[A, List[(A, A)]] =
elem_list_to_rm[(A, A), A]((aa: (A, A)) => fst[A, A](aa), r);
                                      (as: List[A]) =>
rs_to_list[A].apply(foldl[rbt[A, Unit],
                           A]((aa: rbt[A, Unit]) =>
                                (x: A) =>
                                  rs_union[A].apply(aa).apply(list_to_rs[A].apply(mapa[(A,
         A),
        A]((ab: (A, A)) => snd[A, A](ab),
            (rm_set_lookup[A, (A, A)](rm)).apply(x)))),
                               rs_empty[A], as))
                                    },
                                   (as: List[A]) =>
                                     (bs: rbt[A, Unit]) =>
                                       rs_union[A].apply(bs).apply(list_to_rs[A].apply(as)),
                                   rs_memb[A], rs_empty[A], a)

def memo_rbt_rtrancl[A : linorder](r: List[(A, A)]): A => rbt[A, Unit] =
  {
    val tr: (List[A]) => rbt[A, Unit] = rtrancl_rbt_impl[A].apply(r)
    val rm: rbt[A, rbt[A, Unit]] =
      list_to_rm[A, rbt[A, Unit]].apply(mapa[A,
      (A, rbt[A, Unit])]((a: A) => (a, tr(List(a))),
                          (comp[List[A], List[A],
                                 List[(A,
A)]](comp[rbt[A, Unit], List[A], List[A]](rs_to_list[A], list_to_rs[A]),
      (a: List[(A, A)]) =>
        mapa[(A, A), A]((aa: (A, A)) => fst[A, A](aa), a))).apply(r)));
    (a: A) =>
      rm_lookup[A, rbt[A, Unit]](a, rm) match {
        case None => list_to_rs[A].apply(List(a))
        case Some(as) => as
      }
  }

def generate_ta[A : linorder, B : equal : linorder](x0: tree_automaton[A, B]):
      ta_impl_ext[A, B, Unit] =
  x0 match {
  case Tree_Automaton(fin, rules, eps) =>
    {
      val ep: A => rbt[A, Unit] = memo_rbt_rtrancl[A](eps)
      val epr: A => rbt[A, Unit] =
        memo_rbt_rtrancl[A](mapa[(A, A),
                                  (A, A)]((a: (A, A)) =>
    {
      val (q, qa): (A, A) = a;
      (qa, q)
    },
   eps))
      val rqs_rs: rbt[A, Unit] =
        foldl[rbt[A, Unit],
               ta_rule[A, B]]((a: rbt[A, Unit]) =>
                                (x: ta_rule[A, B]) =>
                                  rs_union[A].apply(a).apply(ep(rq[A, B](x))),
                               rs_empty[A], rules)
      val rrules: List[ta_rule_impl[A, B]] =
        mapa[ta_rule[A, B],
              ta_rule_impl[A, B]]((a: ta_rule[A, B]) =>
                                    conv_ta_rule[A, B](ep, a),
                                   rules);
      Ta_impl_ext[A, B,
                   Unit](list_to_rs[A].apply(fin),
                          elem_list_to_rm[ta_rule_impl[A, B],
   (B, nat)]((a: ta_rule_impl[A, B]) => rsym_impl[A, B](a), rrules),
                          rs_to_list[A].apply(rqs_rs), rqs_rs, eps, ep, epr, ())
    }
}

def check_left_linear_trs[A : show,
                           B : equal : show](trs:
       List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(term[A, B], term[A, B]), Unit,
          (List[Char]) =>
            List[Char]](catcha[((term[A, B], term[A, B]),
                                 (term[A, B], term[A, B])),
                                Unit,
                                (term[A, B],
                                  term[A,
B])](forallm[(term[A, B], term[A, B]),
              (term[A, B],
                term[A, B])]((x: (term[A, B], term[A, B])) =>
                               (if (linear_term[A,
         B](fst[term[A, B], term[A, B]](x)))
                                 Inr[Unit, (term[A, B], term[A, B])](())
                                 else Inl[(term[A, B], term[A, B]), Unit](x)),
                              trs),
      (x: ((term[A, B], term[A, B]), (term[A, B], term[A, B]))) =>
        Inl[(term[A, B], term[A, B]),
             Unit](snd[(term[A, B], term[A, B]), (term[A, B], term[A, B])](x))),
                         (_: (term[A, B], term[A, B])) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_prec_list[(term[A, B], term[A, B])](Zero_nat, trs),
                 comp[List[Char], List[Char],
                       List[Char]](shows_nl,
                                    comp[List[Char], List[Char],
  List[Char]](shows_string.apply(List('i', 's', ' ', 'n', 'o', 't', ' ', 'l',
                                       'e', 'f', 't', '-', 'l', 'i', 'n', 'e',
                                       'a', 'r')),
               shows_nl)))))

def boundstype_fun[A, B : equal](x0: boundstype):
      ((term[A, B], term[A, B])) => (term[A, B]) => Boolean =
  x0 match {
  case Roof => (a: (term[A, B], term[A, B])) => roof[A, B](a)
  case Match => matcha[A, B]
}

def ta_rules_impl[A, B, C](x0: ta_impl_ext[A, B, C]):
      rbt[(B, nat), List[ta_rule_impl[A, B]]] =
  x0 match {
  case Ta_impl_ext(ta_final_impl, ta_rules_impl, ta_rqs_impl, ta_rqs_set,
                    ta_eps_impl, ta_epss_impl, ta_epsrs_impl, more)
    => ta_rules_impl
}

def bounds_condition[A : show,
                      B : equal : show](x0: boundstype,
 uu: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  (x0, uu) match {
  case (Roof, uu) => Inr[Unit, (List[Char]) => List[Char]](())
  case (Match, r) =>
    catcha[(term[A, B], term[A, B]), Unit,
            (List[Char]) =>
              List[Char]](catcha[((term[A, B], term[A, B]),
                                   (term[A, B], term[A, B])),
                                  Unit,
                                  (term[A, B],
                                    term[A,
  B])](forallm[(term[A, B], term[A, B]),
                (term[A, B],
                  term[A, B])]((x: (term[A, B], term[A, B])) =>
                                 (if ({
val (_, a): (term[A, B], term[A, B]) = x;
linear_term[A, B](a)
                                      })
                                   Inr[Unit, (term[A, B], term[A, B])](())
                                   else Inl[(term[A, B], term[A, B]), Unit](x)),
                                r),
        (x: ((term[A, B], term[A, B]), (term[A, B], term[A, B]))) =>
          Inl[(term[A, B], term[A, B]),
               Unit](snd[(term[A, B], term[A, B]),
                          (term[A, B], term[A, B])](x))),
                           (x: (term[A, B], term[A, B])) =>
                             Inl[(List[Char]) => List[Char],
                                  Unit]({
  val (l, ra): (term[A, B], term[A, B]) = x;
  comp[List[Char], List[Char],
        List[Char]](shows_prec_list[Char](Zero_nat,
   List('r', 'u', 'l', 'e', ' ')),
                     comp[List[Char], List[Char],
                           List[Char]](shows_rule[A,
           B]((a: A) => (b: List[Char]) => shows_prec[A](Zero_nat, a, b),
               (a: B) => (b: List[Char]) => shows_prec[B](Zero_nat, a, b),
               List(' ', '-', '>', ' '), (l, ra)),
shows_prec_list[Char](Zero_nat,
                       List(' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'r', 'i',
                             'g', 'h', 't', '-', 'l', 'i', 'n', 'e', 'a', 'r',
                             ' '))))
}))
}

def rqs_impl[A, B](x0: ta_rule_impl[A, B]): List[A] = x0 match {
  case TA_rule_impl(f, qsa, q, qs) => qsa
}

def rqss_impl[A, B](x0: ta_rule_impl[A, B]): rbt[A, Unit] = x0 match {
  case TA_rule_impl(f, qsa, q, qs) => qs
}

def ta_epss_impl[A, B, C](x0: ta_impl_ext[A, B, C]): A => rbt[A, Unit] = x0
  match {
  case Ta_impl_ext(ta_final_impl, ta_rules_impl, ta_rqs_impl, ta_rqs_set,
                    ta_eps_impl, ta_epss_impl, ta_epsrs_impl, more)
    => ta_epss_impl
}

def ta_final_impl[A, B, C](x0: ta_impl_ext[A, B, C]): rbt[A, Unit] = x0 match {
  case Ta_impl_ext(ta_final_impl, ta_rules_impl, ta_rqs_impl, ta_rqs_set,
                    ta_eps_impl, ta_epss_impl, ta_epsrs_impl, more)
    => ta_final_impl
}

def ta_contains_impl[A : equal : linorder,
                      B : equal : linorder](f: List[(A, nat)],
     ta: ta_impl_ext[B, A, Unit], qs: List[B]):
      sum[sum[B, term[A, B]], Unit] =
  {
    val fin: rbt[B, Unit] = ta_final_impl[B, A, Unit](ta)
    val look: ((A, nat)) => List[ta_rule_impl[B, A]] =
      rm_set_lookup[(A, nat),
                     ta_rule_impl[B, A]](ta_rules_impl[B, A, Unit](ta));
    ta_epss_impl[B, A, Unit](ta);
    bindb[sum[B, term[A, B]], Unit,
           Unit](catcha[B, Unit,
                         sum[B, term[A, B]]](catcha[(B, B), Unit,
             B](forallm[B, B]((q: B) =>
                                check[B](rs_memb[B].apply(q).apply(fin), q),
                               qs),
                 (x: (B, B)) => Inl[B, Unit](snd[B, B](x))),
      (x: B) => Inl[sum[B, term[A, B]], Unit](Inl[B, term[A, B]](x))),
                  (_: Unit) =>
                    catcha[term[A, B], Unit,
                            sum[B, term[A,
 B]]](catcha[((A, nat), term[A, B]), Unit,
              term[A, B]](forallm[(A, nat),
                                   term[A,
 B]]((a: (A, nat)) =>
       {
         val (fa, n): (A, nat) = a
         val rules: List[ta_rule_impl[B, A]] = look((fa, n));
         catcha[List[B], Unit,
                 term[A, B]](catcha[(List[B], List[B]), Unit,
                                     List[B]](forallm[List[B],
               List[B]]((x: List[B]) =>
                          (if (list_ex[ta_rule_impl[B,
             A]]((rule: ta_rule_impl[B, A]) =>
                   equal_lista[B](rqs_impl[B, A](rule), x) &&
                     {
                       val qq: rbt[B, Unit] = rqss_impl[B, A](rule);
                       list_ex[B]((q: B) => rs_memb[B].apply(q).apply(qq), qs)
                     },
                  rules))
                            Inr[Unit, List[B]](()) else Inl[List[B], Unit](x)),
                         concat_lists[B](replicate[List[B]](n, qs))),
       (x: (List[B], List[B])) => Inl[List[B], Unit](snd[List[B], List[B]](x))),
                              (x: List[B]) =>
                                Inl[term[A, B],
                                     Unit](Fun[A,
        B](fa, mapa[B, term[A, B]]((aa: B) => Var[B, A](aa), x))))
       },
      f),
                           (x: ((A, nat), term[A, B])) =>
                             Inl[term[A, B],
                                  Unit](snd[(A, nat), term[A, B]](x))),
       (x: term[A, B]) => Inl[sum[B, term[A, B]], Unit](Inr[term[A, B], B](x))))
  }

def flatten_term_enum_filter[A, B](f: (term[A, B]) => Boolean,
                                    x1: term[List[A], B]):
      List[term[A, B]] =
  (f, x1) match {
  case (f, Var(x)) =>
    {
      val tx: term[A, B] = Var[B, A](x);
      (if (f(tx)) List(tx) else Nil)
    }
  case (f, Fun(fs, ts)) =>
    {
      val lts: List[List[term[A, B]]] =
        mapa[term[List[A], B],
              List[term[A, B]]]((a: term[List[A], B]) =>
                                  flatten_term_enum_filter[A, B](f, a),
                                 ts);
      (if (list_ex[List[term[A, B]]]((a: List[term[A, B]]) =>
                                       nulla[term[A, B]](a),
                                      lts))
        Nil else {
                   val ss: List[List[term[A, B]]] =
                     concat_lists[term[A, B]](lts);
                   filter[term[A, B]](f, maps[A,
       term[A, B]]((fa: A) =>
                     mapa[List[term[A, B]],
                           term[A, B]]((a: List[term[A, B]]) =>
 Fun[A, B](fa, a),
ss),
                    fs))
                 })
    }
}

def inverse_base_term_filter[A, B](filt: (term[(A, nat), B]) => Boolean,
                                    l: term[A, B], c: nat):
      List[term[(A, nat), B]] =
  {
    val hs: List[nat] = upt(Zero_nat, Suc(c));
    flatten_term_enum_filter[(A, nat),
                              B](filt,
                                  map_term[A, List[(A, nat)],
    B]((f: A) => mapa[nat, (A, nat)]((h: nat) => lift[A](h, f), hs), l))
  }

def cover_bound_list_filter[A, B](filt: (term[(A, nat), B]) => Boolean,
                                   ff: ((term[A, B], term[A, B])) =>
 (term[A, B]) => Boolean,
                                   c: nat, r: List[(term[A, B], term[A, B])]):
      List[(term[(A, nat), B], term[(A, nat), B])] =
  maps[(term[A, B], term[A, B]),
        (term[(A, nat), B],
          term[(A, nat),
                B])]((a: (term[A, B], term[A, B])) =>
                       {
                         val (l, ra): (term[A, B], term[A, B]) = a
                         val ee: (term[A, B]) => Boolean = ff((l, ra));
                         mapa[term[(A, nat), B],
                               (term[(A, nat), B],
                                 term[(A, nat),
                                       B])]((la: term[(A, nat), B]) =>
      (la, map_term[A, (A, nat),
                     B]((aa: A) =>
                          lift[A](Suc(min_list[nat](mapa[(A, nat),
                  nat]((ab: (A, nat)) => height[A](ab),
                        sym_collect[(A, nat),
                                     B]((t: term[(A, nat), B]) =>
  ee(map_term[(A, nat), A, B]((ab: (A, nat)) => base[A](ab), t)),
 la)))),
                                   aa),
                         ra)),
     inverse_base_term_filter[A, B](filt, l, c))
                       },
                      r)

def ta_res_impl[A : equal : linorder,
                 B : linorder](ta: rbt[(A, nat), List[ta_rule_impl[B, A]]],
                                eps: B => rbt[B, Unit], x2: term[A, B]):
      rbt[B, Unit] =
  (ta, eps, x2) match {
  case (ta, eps, Var(q)) => eps(q)
  case (ta, eps, Fun(f, ts)) =>
    {
      val rec: List[rbt[B, Unit]] =
        mapa[term[A, B],
              rbt[B, Unit]]((a: term[A, B]) => ta_res_impl[A, B](ta, eps, a),
                             ts)
      val n: nat = size_list[term[A, B]](ts)
      val rules: List[ta_rule_impl[B, A]] =
        (rm_set_lookup[(A, nat), ta_rule_impl[B, A]](ta)).apply((f, n))
      val arules: List[ta_rule_impl[B, A]] =
        filter[ta_rule_impl[B, A]]((rule: ta_rule_impl[B, A]) =>
                                     {
                                       val qs: List[B] = rqs_impl[B, A](rule);
                                       all_interval_nat((i: nat) =>
                  rs_memb[B].apply(nth[B](qs,
   i)).apply(nth[rbt[B, Unit]](rec, i)),
                 Zero_nat, n)
                                     },
                                    rules)
      val a: List[rbt[B, Unit]] =
        mapa[ta_rule_impl[B, A],
              rbt[B, Unit]]((a: ta_rule_impl[B, A]) => rqss_impl[B, A](a),
                             arules);
      foldl[rbt[B, Unit], rbt[B, Unit]](rs_union[B], rs_empty[B], a)
    }
}

def ta_rqs_impl[A, B, C](x0: ta_impl_ext[A, B, C]): List[A] = x0 match {
  case Ta_impl_ext(ta_final_impl, ta_rules_impl, ta_rqs_impl, ta_rqs_set,
                    ta_eps_impl, ta_epss_impl, ta_epsrs_impl, more)
    => ta_rqs_impl
}

def rs_subset[A : linorder](as: rbt[A, Unit], bs: rbt[A, Unit]): Option[A] =
  rs_iteratei[Option[A],
               A].apply((a: Option[A]) =>
                          a match {
                            case None => true
                            case Some(_) => false
                          }).apply((a: A) =>
                                     (_: Option[A]) =>
                                       (if (rs_memb[A].apply(a).apply(bs)) None
 else Some[A](a))).apply(as).apply(None)

def ta_epsrs_impl[A, B, C](x0: ta_impl_ext[A, B, C]): A => rbt[A, Unit] = x0
  match {
  case Ta_impl_ext(ta_final_impl, ta_rules_impl, ta_rqs_impl, ta_rqs_set,
                    ta_eps_impl, ta_epss_impl, ta_epsrs_impl, more)
    => ta_epsrs_impl
}

def ta_match_impl[A : equal : linorder, B : equal : linorder,
                   C : equal : linorder](ta:
   rbt[(A, nat), List[ta_rule_impl[B, A]]],
  eps: B => rbt[B, Unit], x2: term[A, C], q: B):
      rbt[List[(C, B)], Unit] =
  (ta, eps, x2, q) match {
  case (ta, eps, Var(x), q) =>
    list_to_rs[List[(C, B)]].apply(mapa[B,
 List[(C, B)]]((qa: B) => List((x, qa)), rs_to_list[B].apply(eps(q))))
  case (ta, eps, Fun(f, ts), q) =>
    {
      val n: nat = size_list[term[A, C]](ts)
      val rules: List[ta_rule_impl[B, A]] =
        (rm_set_lookup[(A, nat), ta_rule_impl[B, A]](ta)).apply((f, n))
      val ep: rbt[B, Unit] = eps(q)
      val fa: (ta_rule_impl[B, A]) => rbt[List[(C, B)], Unit] =
        (rule: ta_rule_impl[B, A]) =>
          list_to_rs[List[(C, B)]].apply({
   val (TA_rule_impl(_, qs, qa, _)): ta_rule_impl[B, A] = rule;
   (if (rs_memb[B].apply(qa).apply(ep))
     {
       val rec: List[List[List[(C, B)]]] =
         mapa[(term[A, C], B),
               List[List[(C, B)]]]((a: (term[A, C], B)) =>
                                     {
                                       val (tsi, qsi): (term[A, C], B) = a;
                                       rs_to_list[List[(C,
                 B)]].apply(ta_match_impl[A, B, C](ta, eps, tsi, qsi))
                                     },
                                    zip[term[A, C], B](ts, qs));
       mapa[List[List[(C, B)]],
             List[(C, B)]]((a: List[List[(C, B)]]) => concat[(C, B)](a),
                            concat_lists[List[(C, B)]](rec))
     }
     else Nil)
 });
      foldl[rbt[List[(C, B)], Unit],
             ta_rule_impl[B, A]]((a: rbt[List[(C, B)], Unit]) =>
                                   (x: ta_rule_impl[B, A]) =>
                                     rs_union[List[(C,
             B)]].apply(a).apply(fa(x)),
                                  rs_empty[List[(C, B)]], rules)
    }
}

def ta_rule_compatible_eff_list[A : equal : linorder, B : equal : linorder,
                                 C : equal : linorder](ta:
                 ta_impl_ext[A, B, Unit],
                x1: (term[B, C], term[B, C])):
      sum[((term[B, A], term[B, A]), A), Unit] =
  (ta, x1) match {
  case (ta, (Fun(f, ls), r)) =>
    {
      val rm: rbt[(B, nat), List[ta_rule_impl[A, B]]] =
        ta_rules_impl[A, B, Unit](ta)
      val eps: A => rbt[A, Unit] = ta_epss_impl[A, B, Unit](ta)
      val epsa: A => rbt[A, Unit] = ta_epsrs_impl[A, B, Unit](ta)
      val n: nat = size_list[term[B, C]](ls)
      val ta_res: (term[B, A]) => rbt[A, Unit] =
        (a: term[B, A]) => ta_res_impl[B, A](rm, eps, a);
      catcha[(ta_rule_impl[A, B], ((term[B, A], term[B, A]), A)), Unit,
              ((term[B, A], term[B, A]),
                A)](forallm[ta_rule_impl[A, B],
                             ((term[B, A], term[B, A]),
                               A)]((a: ta_rule_impl[A, B]) =>
                                     {
                                       val
 (TA_rule_impl(fa, _, q, _)): ta_rule_impl[A, B] = a;
                                       catcha[(List[(C, A)],
        ((term[B, A], term[B, A]), A)),
       Unit,
       ((term[B, A], term[B, A]),
         A)](forallm[List[(C, A)],
                      ((term[B, A], term[B, A]),
                        A)]((sigma: List[(C, A)]) =>
                              {
                                val sigmaa: C => A =
                                  (aa: C) => fun_of[C, A](sigma, aa)
                                val l_sigma: term[B, A] =
                                  map_vars[C, A, B](sigmaa, Fun[B, C](fa, ls))
                                val r_sigma: term[B, A] =
                                  map_vars[C, A, B](sigmaa, r)
                                val qsl: rbt[A, Unit] = ta_res(l_sigma)
                                val qsr: rbt[A, Unit] = ta_res(r_sigma);
                                rs_subset[A](qsl, qsr) match {
                                  case None =>
                                    Inr[Unit, ((term[B, A], term[B, A]), A)](())
                                  case Some(qa) =>
                                    Inl[((term[B, A], term[B, A]), A),
 Unit](((l_sigma, r_sigma), qa))
                                }
                              },
                             rs_to_list[List[(C,
       A)]].apply(ta_match_impl[B, A, C](rm, epsa, Fun[B, C](fa, ls), q))),
              (x: (List[(C, A)], ((term[B, A], term[B, A]), A))) =>
                Inl[((term[B, A], term[B, A]), A),
                     Unit](snd[List[(C, A)], ((term[B, A], term[B, A]), A)](x)))
                                     },
                                    (rm_set_lookup[(B, nat),
            ta_rule_impl[A, B]](rm)).apply((f, n))),
                     (x: (ta_rule_impl[A, B], ((term[B, A], term[B, A]), A))) =>
                       Inl[((term[B, A], term[B, A]), A),
                            Unit](snd[ta_rule_impl[A, B],
                                       ((term[B, A], term[B, A]), A)](x)))
    }
  case (ta, (Var(x), r)) =>
    {
      val rm: rbt[(B, nat), List[ta_rule_impl[A, B]]] =
        ta_rules_impl[A, B, Unit](ta)
      val eps: A => rbt[A, Unit] = ta_epss_impl[A, B, Unit](ta)
      val ta_res: (term[B, A]) => rbt[A, Unit] =
        (a: term[B, A]) => ta_res_impl[B, A](rm, eps, a);
      catcha[(A, ((term[B, A], term[B, A]), A)), Unit,
              ((term[B, A], term[B, A]),
                A)](forallm[A, ((term[B, A], term[B, A]),
                                 A)]((q: A) =>
                                       {
 val r_sigma: term[B, A] = map_vars[C, A, B]((_: C) => q, r);
 check[((term[B, A], term[B, A]),
         A)](rs_memb[A].apply(q).apply(ta_res(r_sigma)),
              ((Var[A, B](q), r_sigma), q))
                                       },
                                      ta_rqs_impl[A, B, Unit](ta)),
                     (xa: (A, ((term[B, A], term[B, A]), A))) =>
                       Inl[((term[B, A], term[B, A]), A),
                            Unit](snd[A, ((term[B, A], term[B, A]), A)](xa)))
    }
}

def ta_trs_compatible_eff_list[A : equal : linorder, B : equal : linorder,
                                C : equal : linorder](ta:
                ta_impl_ext[A, B, Unit],
               r: List[(term[B, C], term[B, C])]):
      sum[((term[B, C], term[B, C]), ((term[B, A], term[B, A]), A)), Unit] =
  {
    val check:
          ((term[B, C], term[B, C])) => sum[((term[B, A], term[B, A]), A), Unit]
      = (a: (term[B, C], term[B, C])) =>
          ta_rule_compatible_eff_list[A, B, C](ta, a);
    catcha[((term[B, C], term[B, C]),
             ((term[B, C], term[B, C]), ((term[B, A], term[B, A]), A))),
            Unit,
            ((term[B, C], term[B, C]),
              ((term[B, A], term[B, A]),
                A))](forallm[(term[B, C], term[B, C]),
                              ((term[B, C], term[B, C]),
                                ((term[B, A], term[B, A]),
                                  A))]((lr: (term[B, C], term[B, C])) =>
 catcha[((term[B, A], term[B, A]), A), Unit,
         ((term[B, C], term[B, C]),
           ((term[B, A], term[B, A]),
             A))](check(lr),
                   (x: ((term[B, A], term[B, A]), A)) =>
                     Inl[((term[B, C], term[B, C]),
                           ((term[B, A], term[B, A]), A)),
                          Unit]((lr, x))),
r),
                      (x: ((term[B, C], term[B, C]),
                            ((term[B, C], term[B, C]),
                              ((term[B, A], term[B, A]), A))))
                        => Inl[((term[B, C], term[B, C]),
                                 ((term[B, A], term[B, A]), A)),
                                Unit](snd[(term[B, C], term[B, C]),
   ((term[B, C], term[B, C]), ((term[B, A], term[B, A]), A))](x)))
  }

def ta_res_impl_all[A : linorder, B : equal : linorder,
                     C](q: rbt[A, Unit],
                         ta: rbt[(B, nat), List[ta_rule_impl[A, B]]],
                         x2: term[B, C]):
      rbt[A, Unit] =
  (q, ta, x2) match {
  case (q, ta, Var(uu)) => q
  case (q, ta, Fun(f, ts)) =>
    {
      val rec: List[rbt[A, Unit]] =
        mapa[term[B, C],
              rbt[A, Unit]]((a: term[B, C]) =>
                              ta_res_impl_all[A, B, C](q, ta, a),
                             ts)
      val n: nat = size_list[term[B, C]](ts)
      val rules: List[ta_rule_impl[A, B]] =
        (rm_set_lookup[(B, nat), ta_rule_impl[A, B]](ta)).apply((f, n))
      val arules: List[ta_rule_impl[A, B]] =
        filter[ta_rule_impl[A, B]]((rule: ta_rule_impl[A, B]) =>
                                     {
                                       val qs: List[A] = rqs_impl[A, B](rule);
                                       all_interval_nat((i: nat) =>
                  rs_memb[A].apply(nth[A](qs,
   i)).apply(nth[rbt[A, Unit]](rec, i)),
                 Zero_nat, n)
                                     },
                                    rules)
      val a: List[rbt[A, Unit]] =
        mapa[ta_rule_impl[A, B],
              rbt[A, Unit]]((a: ta_rule_impl[A, B]) => rqss_impl[A, B](a),
                             arules);
      foldl[rbt[A, Unit], rbt[A, Unit]](rs_union[A], rs_empty[A], a)
    }
}

def ta_rule_compatible_heuristic[A : equal : linorder, B : equal : linorder,
                                  C : linorder](ta: ta_impl_ext[A, B, Unit],
         l: term[B, C]):
      Boolean =
  rs_isEmpty[A].apply(ta_res_impl_all[A, B,
                                       C](ta_rqs_set[A, B, Unit](ta),
   ta_rules_impl[A, B, Unit](ta), l))

def check_bounds[A : equal : linorder : show,
                  B : equal : linorder : show](x0: bounds_info[A, B],
        r: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  (x0, r) match {
  case (Bounds_Info(typea, c, qfin, preTA), r) =>
    {
      val ta: ta_impl_ext[B, (A, nat), Unit] = generate_ta[B, (A, nat)](preTA);
      bindb[(List[Char]) => List[Char], Unit,
             Unit](check_wf_trs[A, B](r),
                    (_: Unit) =>
                      bindb[(List[Char]) => List[Char], Unit,
                             Unit](check_left_linear_trs[A, B](r),
                                    (_: Unit) =>
                                      bindb[(List[Char]) => List[Char], Unit,
     Unit](bounds_condition[A, B](typea, r),
            (_: Unit) =>
              bindb[(List[Char]) => List[Char], Unit,
                     Unit](check[(List[Char]) =>
                                   List[Char]](list_ex[B]((q: B) =>
                    rs_memb[B].apply(q).apply(ta_rqs_set[B, (A, nat),
                  Unit](ta)),
                   qfin),
        shows_prec_list[Char](Zero_nat,
                               List('d', 'i', 'd', ' ', 'n', 'o', 't', ' ', 'f',
                                     'i', 'n', 'd', ' ', 'm', 'e', 'n', 't',
                                     'i', 'o', 'n', 'e', 'd', ' ', 'f', 'i',
                                     'n', 'a', 'l', ' ', 's', 't', 'a', 't',
                                     'e', ' ', 'i', 'n', ' ', 'T', 'A'))),
                            (_: Unit) =>
                              bindb[(List[Char]) => List[Char], Unit,
                                     Unit](catcha[((A, nat), nat), Unit,
           (List[Char]) =>
             List[Char]](catcha[(((A, nat), nat), ((A, nat), nat)), Unit,
                                 ((A, nat),
                                   nat)](forallm[((A, nat), nat),
          ((A, nat),
            nat)]((x: ((A, nat), nat)) =>
                    (if ({
                           val (f, _): ((A, nat), nat) = x;
                           less_eq_nat(height[A](f), c)
                         })
                      Inr[Unit, ((A, nat), nat)](())
                      else Inl[((A, nat), nat), Unit](x)),
                   mapa[(((A, nat), nat), List[ta_rule_impl[B, (A, nat)]]),
                         ((A, nat),
                           nat)]((a: (((A, nat), nat),
                                       List[ta_rule_impl[B, (A, nat)]]))
                                   => fst[((A, nat), nat),
   List[ta_rule_impl[B, (A, nat)]]](a),
                                  rm_to_list[((A, nat), nat),
      List[ta_rule_impl[B, (A, nat)]]].apply(ta_rules_impl[B, (A, nat),
                    Unit](ta)))),
  (x: (((A, nat), nat), ((A, nat), nat))) =>
    Inl[((A, nat), nat), Unit](snd[((A, nat), nat), ((A, nat), nat)](x))),
                          (x: ((A, nat), nat)) =>
                            Inl[(List[Char]) => List[Char],
                                 Unit]({
 val (f, _): ((A, nat), nat) = x;
 comp[List[Char], List[Char],
       List[Char]](shows_prec_prod[A, nat](Zero_nat, f),
                    comp[List[Char], List[Char],
                          List[Char]](shows_prec_list[Char](Zero_nat,
                     List(' ', 'i', 's', ' ', 's', 'y', 'm', 'b', 'o', 'l', ' ',
                           'i', 'n', ' ', 'T', 'A', ' ', 'w', 'i', 't', 'h',
                           ' ', 'h', 'e', 'i', 'g', 'h', 't', ' ', 'l', 'a',
                           'r', 'g', 'e', 'r', ' ', 't', 'h', 'a', 'n', ' ',
                           'c', ' ', '=', ' ')),
                                       shows_prec_nat(Zero_nat, c)))
                                       })),
    (_: Unit) =>
      bindb[(List[Char]) => List[Char], Unit,
             Unit](catcha[sum[B, term[(A, nat), B]], Unit,
                           (List[Char]) =>
                             List[Char]](ta_contains_impl[(A, nat),
                   B](mapa[(A, nat),
                            ((A, nat),
                              nat)]((a: (A, nat)) =>
                                      {
val (f, aa): (A, nat) = a;
(lift[A](Zero_nat, f), aa)
                                      },
                                     funas_trs_list[A, B](r)),
                       ta, qfin),
  (x: sum[B, term[(A, nat), B]]) =>
    Inl[(List[Char]) => List[Char],
         Unit](comp[List[Char], List[Char],
                     List[Char]](shows_prec_list[Char](Zero_nat,
                List('i', 't', ' ', 'c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't',
                      ' ', 'b', 'e', ' ', 'g', 'u', 'a', 'r', 'a', 'n', 't',
                      'e', 'e', 'd', ' ', 't', 'h', 'a', 't', ' ', 'l', 'i',
                      'f', 't', '0', '(', 'T', '(', 'S', 'i', 'g', 'm', 'a',
                      ')', ')', ' ', 'i', 's', ' ', 'a', 'c', 'c', 'e', 'p',
                      't', 'e', 'd', ' ', 'b', 'y', ' ', 'T', 'A')),
                                  comp[List[Char], List[Char],
List[Char]](shows_nl,
             x match {
               case Inl(q) =>
                 comp[List[Char], List[Char],
                       List[Char]](shows_prec_list[Char](Zero_nat,
                  List('s', 't', 'a', 't', 'e', ' ')),
                                    comp[List[Char], List[Char],
  List[Char]]((a: List[Char]) => shows_prec[B](Zero_nat, q, a),
               shows_prec_list[Char](Zero_nat,
                                      List(' ', 'i', 's', ' ', 'n', 'o', 't',
    ' ', 'f', 'i', 'n', 'a', 'l'))))
               case Inr(fqs) =>
                 comp[List[Char], List[Char],
                       List[Char]](shows_prec_list[Char](Zero_nat,
                  List('t', 'h', 'e', 'r', 'e', ' ', 'i', 's', ' ', 'n', 'o',
                        ' ', 't', 'r', 'a', 'n', 's', 'i', 't', 'i', 'o', 'n',
                        ' ', 'f', 'r', 'o', 'm', ' ')),
                                    comp[List[Char], List[Char],
  List[Char]]((shows_prec_term[(A, nat), B](Zero_nat)).apply(fqs),
               shows_prec_list[Char](Zero_nat,
                                      List(' ', 't', 'o', ' ', 'a', ' ', 'f',
    'i', 'n', 'a', 'l', ' ', 's', 't', 'a', 't', 'e'))))
             })))),
                    (_: Unit) =>
                      catcha[((term[(A, nat), B], term[(A, nat), B]),
                               ((term[(A, nat), B], term[(A, nat), B]), B)),
                              Unit,
                              (List[Char]) =>
                                List[Char]](ta_trs_compatible_eff_list[B,
                                (A, nat),
                                B](ta, cover_bound_list_filter[A,
                        B]((l: term[(A, nat), B]) =>
                             ! (ta_rule_compatible_heuristic[B, (A, nat),
                      B](ta, l)),
                            boundstype_fun[A, B](typea), c, r)),
     (x: ((term[(A, nat), B], term[(A, nat), B]),
           ((term[(A, nat), B], term[(A, nat), B]), B)))
       => Inl[(List[Char]) => List[Char],
               Unit]({
                       val (lr, (lrq, q)):
                             ((term[(A, nat), B], term[(A, nat), B]),
                               ((term[(A, nat), B], term[(A, nat), B]), B))
                         = x;
                       comp[List[Char], List[Char],
                             List[Char]](shows_prec_list[Char](Zero_nat,
                        List('T', 'A', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ',
                              'c', 'o', 'm', 'p', 'a', 't', 'i', 'b', 'l', 'e',
                              ' ', 'w', 'i', 't', 'h', ' ', 'R')),
  comp[List[Char], List[Char],
        List[Char]](shows_nl,
                     comp[List[Char], List[Char],
                           List[Char]](shows_prec_list[Char](Zero_nat,
                      List('f', 'o', 'r', ' ', 'r', 'u', 'l', 'e', ' ')),
comp[List[Char], List[Char],
      List[Char]](shows_rule[(A, nat),
                              B]((a: (A, nat)) =>
                                   shows_prec_prod[A, nat](Zero_nat, a),
                                  (a: B) =>
                                    (b: List[Char]) =>
                                      shows_prec[B](Zero_nat, a, b),
                                  List(' ', '-', '>', ' '), lr),
                   comp[List[Char], List[Char],
                         List[Char]](shows_nl,
                                      comp[List[Char], List[Char],
    List[Char]](shows_prec_list[Char](Zero_nat,
                                       List('w', 'h', 'i', 'c', 'h', ' ', 'i',
     's', ' ', 'i', 'n', 's', 't', 'a', 'n', 't', 'i', 'a', 't', 'e', 'd', ' ',
     'b', 'y', ' ', 's', 't', 'a', 't', 'e', 's', ' ', 't', 'o', ' ')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_rule[(A, nat),
       B]((a: (A, nat)) => shows_prec_prod[A, nat](Zero_nat, a),
           (a: B) => (b: List[Char]) => shows_prec[B](Zero_nat, a, b),
           List(' ', '-', '>', ' '), lrq),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl,
               comp[List[Char], List[Char],
                     List[Char]](shows_prec_list[Char](Zero_nat,
                List('t', 'h', 'e', ' ', 's', 't', 'a', 't', 'e', ' ')),
                                  comp[List[Char], List[Char],
List[Char]]((a: List[Char]) => shows_prec[B](Zero_nat, q, a),
             comp[List[Char], List[Char],
                   List[Char]](shows_prec_list[Char](Zero_nat,
              List(' ', 'i', 's', ' ', 'o', 'n', 'l', 'y', ' ', 'r', 'e', 'a',
                    'c', 'h', 'a', 'b', 'l', 'e', ' ', 'f', 'r', 'o', 'm', ' ',
                    't', 'h', 'e', ' ', 'l', 'h', 's')),
                                shows_nl)))))))))))
                     }))))))))
    }
}

def rulesb[A, B, C, D](x0: tp_ops_ext[A, B, C, D]):
      A => List[(term[B, C], term[B, C])] =
  x0 match {
  case Tp_ops_ext(qreltrs, q, r, rw, rules, q_empty, is_QNF,
                   nFQ_subset_NF_rules, rules_map, delete_R_Rw, split_rules, mk,
                   more)
    => rules
}

def bounds_tt[A, B : equal : linorder : show,
               C : equal : linorder : show](i: tp_ops_ext[A, B, C, Unit],
     info: bounds_info[B, C], tp: A):
      sum[(List[Char]) => List[Char], Unit] =
  {
    val a: List[(term[B, C], term[B, C])] =
      (rulesb[A, B, C, Unit](i)).apply(tp);
    check_bounds[B, C](info, a)
  }

def s_label_mul_impl[A : show,
                      B : show](list_ext:
                                  (((term[A, B], nat)) =>
                                    ((term[A, B], nat)) =>
                                      (Boolean, Boolean)) =>
                                    (List[(term[A, B], nat)]) =>
                                      (List[(term[A, B], nat)]) =>
(Boolean, Boolean),
                                 af: ((A, nat)) => List[(nat, nat)],
                                 cS: ((term[A, B], term[A, B])) =>
                                       sum[(List[Char]) => List[Char], Unit],
                                 cNS: ((term[A, B], term[A, B])) =>
sum[(List[Char]) => List[Char], Unit],
                                 st: (term[A, B], term[A, B])):
      sum[(List[Char]) => List[Char], Unit] =
  st match {
    case (Var(_), b) =>
      Inl[(List[Char]) => List[Char],
           Unit](comp[List[Char], List[Char],
                       List[Char]](shows_prec_list[Char](Zero_nat,
                  List('r', 'o', 'o', 't', 's', ' ', 'o', 'f', ' ')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_rule[A, B]((a: A) =>
                                 (ba: List[Char]) =>
                                   shows_prec[A](Zero_nat, a, ba),
                                (a: B) =>
                                  (ba: List[Char]) =>
                                    shows_prec[B](Zero_nat, a, ba),
                                List(' ', '-', '>', ' '), st),
               shows_prec_list[Char](Zero_nat,
                                      List(' ', 'm', 'u', 's', 't', ' ', 'b',
    'e', ' ', 'n', 'o', 'n', '-', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e')))))
    case (Fun(f, ss), Var(_)) =>
      Inl[(List[Char]) => List[Char],
           Unit](comp[List[Char], List[Char],
                       List[Char]](shows_prec_list[Char](Zero_nat,
                  List('r', 'o', 'o', 't', 's', ' ', 'o', 'f', ' ')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_rule[A, B]((a: A) =>
                                 (b: List[Char]) =>
                                   shows_prec[A](Zero_nat, a, b),
                                (a: B) =>
                                  (b: List[Char]) =>
                                    shows_prec[B](Zero_nat, a, b),
                                List(' ', '-', '>', ' '), st),
               shows_prec_list[Char](Zero_nat,
                                      List(' ', 'm', 'u', 's', 't', ' ', 'b',
    'e', ' ', 'n', 'o', 'n', '-', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e')))))
    case (Fun(f, ss), Fun(g, ts)) =>
      check[(List[Char]) =>
              List[Char]](fst[Boolean,
                               Boolean](((list_ext((a: (term[A, B], nat)) =>
             (b: (term[A, B], nat)) =>
               label_s_ns_impl[A, B](cS, cNS, a,
                                      b)))((lterms[A,
            B](af)).apply(Fun[A, B](f, ss))))((lterms[A,
               B](af)).apply(Fun[A, B](g, ts)))),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_prec_list[Char](Zero_nat,
                            List('c', 'a', 'n', 'n', 'o', 't', ' ', 'o', 'r',
                                  'i', 'e', 'n', 't', ' ', 'p', 'a', 'i', 'r',
                                  ' ')),
      comp[List[Char], List[Char],
            List[Char]](shows_rule[A, B]((a: A) =>
   (b: List[Char]) => shows_prec[A](Zero_nat, a, b),
  (a: B) => (b: List[Char]) => shows_prec[B](Zero_nat, a, b),
  List(' ', '-', '>', ' '), st),
                         shows_prec_list[Char](Zero_nat,
        List(' ', 'w', 'e', 'a', 'k', 'l', 'y')))))
  }

def nST_label_mul_impl[A : show,
                        B : show](list_ext:
                                    (((term[A, B], nat)) =>
                                      ((term[A, B], nat)) =>
(Boolean, Boolean)) =>
                                      (List[(term[A, B], nat)]) =>
(List[(term[A, B], nat)]) => (Boolean, Boolean),
                                   af: ((A, nat)) => List[(nat, nat)],
                                   cS: ((term[A, B], term[A, B])) =>
 sum[(List[Char]) => List[Char], Unit],
                                   cNS: ((term[A, B], term[A, B])) =>
  sum[(List[Char]) => List[Char], Unit],
                                   st: (term[A, B], term[A, B])):
      sum[(List[Char]) => List[Char], Unit] =
  st match {
    case (Var(_), b) =>
      Inl[(List[Char]) => List[Char],
           Unit](comp[List[Char], List[Char],
                       List[Char]](shows_prec_list[Char](Zero_nat,
                  List('r', 'o', 'o', 't', 's', ' ', 'o', 'f', ' ')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_rule[A, B]((a: A) =>
                                 (ba: List[Char]) =>
                                   shows_prec[A](Zero_nat, a, ba),
                                (a: B) =>
                                  (ba: List[Char]) =>
                                    shows_prec[B](Zero_nat, a, ba),
                                List(' ', '-', '>', ' '), st),
               shows_prec_list[Char](Zero_nat,
                                      List(' ', 'm', 'u', 's', 't', ' ', 'b',
    'e', ' ', 'n', 'o', 'n', '-', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e')))))
    case (Fun(f, ss), Var(_)) =>
      Inl[(List[Char]) => List[Char],
           Unit](comp[List[Char], List[Char],
                       List[Char]](shows_prec_list[Char](Zero_nat,
                  List('r', 'o', 'o', 't', 's', ' ', 'o', 'f', ' ')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_rule[A, B]((a: A) =>
                                 (b: List[Char]) =>
                                   shows_prec[A](Zero_nat, a, b),
                                (a: B) =>
                                  (b: List[Char]) =>
                                    shows_prec[B](Zero_nat, a, b),
                                List(' ', '-', '>', ' '), st),
               shows_prec_list[Char](Zero_nat,
                                      List(' ', 'm', 'u', 's', 't', ' ', 'b',
    'e', ' ', 'n', 'o', 'n', '-', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e')))))
    case (Fun(f, ss), Fun(g, ts)) =>
      check[(List[Char]) =>
              List[Char]](snd[Boolean,
                               Boolean](((list_ext((a: (term[A, B], nat)) =>
             (b: (term[A, B], nat)) =>
               label_s_ns_impl[A, B](cS, cNS, a,
                                      b)))((lterms[A,
            B](af)).apply(Fun[A, B](f, ss))))((lterms[A,
               B](af)).apply(Fun[A, B](g, ts)))),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_prec_list[Char](Zero_nat,
                            List('c', 'a', 'n', 'n', 'o', 't', ' ', 'o', 'r',
                                  'i', 'e', 'n', 't', ' ', 'p', 'a', 'i', 'r',
                                  ' ')),
      comp[List[Char], List[Char],
            List[Char]](shows_rule[A, B]((a: A) =>
   (b: List[Char]) => shows_prec[A](Zero_nat, a, b),
  (a: B) => (b: List[Char]) => shows_prec[B](Zero_nat, a, b),
  List(' ', '-', '>', ' '), st),
                         shows_prec_list[Char](Zero_nat,
        List(' ', 'w', 'e', 'a', 'k', 'l', 'y')))))
  }

def generate_scnp_rp[A : equal : show, B : show,
                      C](list_ext:
                           (((term[A, B], nat)) =>
                             ((term[A, B], nat)) => (Boolean, Boolean)) =>
                             (List[(term[A, B], nat)]) =>
                               (List[(term[A, B], nat)]) => (Boolean, Boolean),
                          afa: List[((A, nat), List[(nat, nat)])],
                          rti: C => redtriple_ext[A, B, Unit], x: C):
      root_redtriple_ext[A, B, Unit] =
  {
    val rt: redtriple_ext[A, B, Unit] = rti(x)
    val afaa: ((A, nat)) => List[(nat, nat)] =
      (a: (A, nat)) => fun_ofa[List[(nat, nat)], (A, nat)](Nil, afa, a)
    val pi: ((A, nat)) => nat => Boolean = af[A, B, Unit](rt)
    val cS: ((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit]
      = s[A, B, Unit](rt)
    val cNS: ((term[A, B], term[A, B])) => sum[(List[Char]) => List[Char], Unit]
      = ns[A, B, Unit](rt);
    Root_redtriple_ext[A, B,
                        Unit](valid[A, B, Unit](rt),
                               (a: (term[A, B], term[A, B])) =>
                                 s_label_mul_impl[A,
           B](list_ext, afaa, cS, cNS, a),
                               ns[A, B, Unit](rt),
                               (a: (term[A, B], term[A, B])) =>
                                 nST_label_mul_impl[A,
             B](list_ext, afaa, cS, cNS, a),
                               pi, scnp_af_to_af[A](afaa, pi),
                               comp[List[Char], List[Char],
                                     List[Char]](scnp_af_desc[A](afa),
          desca[A, B, Unit](rt)),
                               ())
  }

def eval_lab[A, B, C, D,
              E](i: A => (List[B]) => B, l: A => (List[B]) => C,
                  lc: A => nat => C => D, alpha: E => B, x4: term[A, E]):
      (B, term[D, E]) =
  (i, l, lc, alpha, x4) match {
  case (i, l, lc, alpha, Var(x)) => (alpha(x), Var[E, D](x))
  case (i, l, lc, alpha, Fun(f, ts)) =>
    {
      val clts: List[(B, term[D, E])] =
        mapa[term[A, E],
              (B, term[D, E])]((a: term[A, E]) =>
                                 eval_lab[A, B, C, D, E](i, l, lc, alpha, a),
                                ts)
      val cs: List[B] =
        mapa[(B, term[D, E]),
              B]((a: (B, term[D, E])) => fst[B, term[D, E]](a), clts)
      val c: B = (i(f))(cs)
      val lts: List[term[D, E]] =
        mapa[(B, term[D, E]),
              term[D, E]]((a: (B, term[D, E])) => snd[B, term[D, E]](a), clts);
      (c, Fun[D, E](((lc(f))(size_list[term[A, E]](ts)))((l(f))(cs)), lts))
    }
}

def lab_root[A, B, C, D,
              E](i: A => (List[B]) => B, la: A => (List[B]) => C,
                  l: A => (List[B]) => C, lc: A => nat => C => D, alpha: E => B,
                  x5: term[A, E]):
      term[D, E] =
  (i, la, l, lc, alpha, x5) match {
  case (i, la, l, lc, alpha, Fun(f, ts)) =>
    {
      val clts: List[(B, term[D, E])] =
        mapa[term[A, E],
              (B, term[D, E])]((a: term[A, E]) =>
                                 eval_lab[A, B, C, D, E](i, la, lc, alpha, a),
                                ts)
      val cs: List[B] =
        mapa[(B, term[D, E]),
              B]((a: (B, term[D, E])) => fst[B, term[D, E]](a), clts)
      val a: List[term[D, E]] =
        mapa[(B, term[D, E]),
              term[D, E]]((a: (B, term[D, E])) => snd[B, term[D, E]](a), clts);
      Fun[D, E](((lc(f))(size_list[term[A, E]](ts)))((l(f))(cs)), a)
    }
  case (uu, uv, uw, ux, uy, Var(x)) => Var[E, D](x)
}

abstract sealed class scg[+A, +B]
final case object Null extends scg[Nothing, Nothing]
final case class Scg[A, B](a: A, b: A, c: List[(B, B)], d: List[(B, B)]) extends
  scg[A, B]

def unary_term[A, B](x0: term[A, B]): Boolean = x0 match {
  case Var(x) => true
  case Fun(f, List(t)) => unary_term[A, B](t)
  case Fun(v, Nil) => false
  case Fun(v, vb :: vd :: ve) => false
}

abstract sealed class tp[+A, +B]
final case class
  Tp[A, B](a: (List[term[A, B]],
                (Boolean,
                  (List[(term[A, B], term[A, B])],
                    (List[(term[A, B], term[A, B])],
                      rbt[(A, nat),
                           List[(Boolean, (term[A, B], term[A, B]))]])))))
  extends tp[A, B]

def q_impl[A, B](x0: (List[term[A, B]],
                       (Boolean,
                         (List[(term[A, B], term[A, B])],
                           (List[(term[A, B], term[A, B])],
                             rbt[(A, nat),
                                  List[(Boolean,
 (term[A, B], term[A, B]))]]))))):
      List[term[A, B]] =
  x0 match {
  case (q, uu) => q
}

def impl_ofa[A : linorder, B](x0: tp[A, B]):
      (List[term[A, B]],
        (Boolean,
          (List[(term[A, B], term[A, B])],
            (List[(term[A, B], term[A, B])],
              rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]])))) =
  x0 match {
  case Tp(x) => x
}

def q[A : linorder, B](tp: tp[A, B]): List[term[A, B]] =
  q_impl[A, B](impl_ofa[A, B](tp))

def r_impl[A : equal : linorder,
            B](x0: (List[term[A, B]],
                     (Boolean,
                       (List[(term[A, B], term[A, B])],
                         (List[(term[A, B], term[A, B])],
                           rbt[(A, nat),
                                List[(Boolean, (term[A, B], term[A, B]))]]))))):
      List[(term[A, B], term[A, B])] =
  x0 match {
  case (uu, (uv, (vR, (uw, m)))) =>
    vR ++ rules_with[Boolean, A, B](id[Boolean], m)
}

def r[A : equal : linorder, B](tp: tp[A, B]): List[(term[A, B], term[A, B])] =
  r_impl[A, B](impl_ofa[A, B](tp))

def xml2reltrs_proof[A : equal](xml2name:
                                  xml => sum[List[Char], lab[A, List[nat]]],
                                 x: xml):
      sum[List[Char], trs_proof[A, List[nat], List[Char]]] =
  xml1element[trs_proof[A, List[nat], List[Char]],
               trs_proof[A, List[nat],
                          List[Char]]](List('r', 'e', 'l', 'a', 't', 'i', 'v',
     'e', 'T', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o',
     'o', 'f'),
(a: xml) =>
  xml_options[trs_proof[A, List[nat],
                         List[Char]]](List((List('r', 'I', 's', 'E', 'm', 'p',
          't', 'y'),
     (aa: xml) =>
       xml0elements[trs_proof[A, List[nat],
                               List[Char]]](List('r', 'I', 's', 'E', 'm', 'p',
          't', 'y'),
     RisEmpty, aa)),
    (List('s', 'I', 's', 'E', 'm', 'p', 't', 'y'),
      (aa: xml) =>
        xml1element[trs_proof[A, List[nat], List[Char]],
                     trs_proof[A, List[nat],
                                List[Char]]](List('s', 'I', 's', 'E', 'm', 'p',
           't', 'y'),
      (ab: xml) => xml2trs_proof[A](xml2name, ab),
      id[trs_proof[A, List[nat], List[Char]]], aa)),
    (List('r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l'),
      (aa: xml) =>
        xml4elements[redtriple_impl[lab[A, List[nat]]],
                      List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]], List[Char]])],
                      List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]], List[Char]])],
                      trs_proof[A, List[nat], List[Char]],
                      trs_proof[A, List[nat],
                                 List[Char]]](List('r', 'u', 'l', 'e', 'R', 'e',
            'm', 'o', 'v', 'a', 'l'),
       ordering_constraint_proofa[lab[A, List[nat]]](xml2name),
       (ab: xml) =>
         xml1element[List[(term[lab[A, List[nat]], List[Char]],
                            term[lab[A, List[nat]], List[Char]])],
                      List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]],
                                   List[Char]])]](List('t', 'r', 's'),
           rules[lab[A, List[nat]]](xml2name),
           id[List[(term[lab[A, List[nat]], List[Char]],
                     term[lab[A, List[nat]], List[Char]])]],
           ab),
       (ab: xml) =>
         xml1element[List[(term[lab[A, List[nat]], List[Char]],
                            term[lab[A, List[nat]], List[Char]])],
                      List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]],
                                   List[Char]])]](List('t', 'r', 's'),
           rules[lab[A, List[nat]]](xml2name),
           id[List[(term[lab[A, List[nat]], List[Char]],
                     term[lab[A, List[nat]], List[Char]])]],
           ab),
       (ab: xml) => xml2reltrs_proof[A](xml2name, ab),
       (rp: redtriple_impl[lab[A, List[nat]]]) =>
         (one: List[(term[lab[A, List[nat]], List[Char]],
                      term[lab[A, List[nat]], List[Char]])])
           => (two: List[(term[lab[A, List[nat]], List[Char]],
                           term[lab[A, List[nat]], List[Char]])])
                => (ab: trs_proof[A, List[nat], List[Char]]) =>
                     RuleRemoval[A, List[nat], List[Char]](rp, one ++ two, ab),
       aa)),
    (List('s', 'e', 'm', 'l', 'a', 'b'),
      (aa: xml) =>
        xml4elements[sl_variant[lab[A, List[nat]], List[Char]],
                      List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]], List[Char]])],
                      List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]], List[Char]])],
                      trs_proof[A, List[nat], List[Char]],
                      trs_proof[A, List[nat],
                                 List[Char]]](List('s', 'e', 'm', 'l', 'a',
            'b'),
       sl_variant[lab[A, List[nat]]](xml2name),
       (ab: xml) =>
         xml1element[List[(term[lab[A, List[nat]], List[Char]],
                            term[lab[A, List[nat]], List[Char]])],
                      List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]],
                                   List[Char]])]](List('t', 'r', 's'),
           rules[lab[A, List[nat]]](xml2name),
           id[List[(term[lab[A, List[nat]], List[Char]],
                     term[lab[A, List[nat]], List[Char]])]],
           ab),
       (ab: xml) =>
         xml1element[List[(term[lab[A, List[nat]], List[Char]],
                            term[lab[A, List[nat]], List[Char]])],
                      List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]],
                                   List[Char]])]](List('t', 'r', 's'),
           rules[lab[A, List[nat]]](xml2name),
           id[List[(term[lab[A, List[nat]], List[Char]],
                     term[lab[A, List[nat]], List[Char]])]],
           ab),
       (ab: xml) => xml2reltrs_proof[A](xml2name, ab),
       (sli: sl_variant[lab[A, List[nat]], List[Char]]) =>
         (lr: List[(term[lab[A, List[nat]], List[Char]],
                     term[lab[A, List[nat]], List[Char]])])
           => (lrw: List[(term[lab[A, List[nat]], List[Char]],
                           term[lab[A, List[nat]], List[Char]])])
                => (ab: trs_proof[A, List[nat], List[Char]]) =>
                     Semlab[A, List[nat], List[Char]](sli, Nil, lr ++ lrw, ab),
       aa)),
    (List('u', 'n', 'c', 'u', 'r', 'r', 'y'),
      (aa: xml) =>
        xml4elements[(lab[A, List[nat]],
                       (List[((lab[A, List[nat]], nat),
                               List[lab[A, List[nat]]])],
                         (List[(term[lab[A, List[nat]], List[Char]],
                                 term[lab[A, List[nat]], List[Char]])],
                           List[(term[lab[A, List[nat]], List[Char]],
                                  term[lab[A, List[nat]], List[Char]])]))),
                      List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]], List[Char]])],
                      List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]], List[Char]])],
                      trs_proof[A, List[nat], List[Char]],
                      trs_proof[A, List[nat],
                                 List[Char]]](List('u', 'n', 'c', 'u', 'r', 'r',
            'y'),
       uncurry_info[lab[A, List[nat]]](xml2name),
       (ab: xml) =>
         xml1element[List[(term[lab[A, List[nat]], List[Char]],
                            term[lab[A, List[nat]], List[Char]])],
                      List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]],
                                   List[Char]])]](List('t', 'r', 's'),
           rules[lab[A, List[nat]]](xml2name),
           id[List[(term[lab[A, List[nat]], List[Char]],
                     term[lab[A, List[nat]], List[Char]])]],
           ab),
       (ab: xml) =>
         xml1element[List[(term[lab[A, List[nat]], List[Char]],
                            term[lab[A, List[nat]], List[Char]])],
                      List[(term[lab[A, List[nat]], List[Char]],
                             term[lab[A, List[nat]],
                                   List[Char]])]](List('t', 'r', 's'),
           rules[lab[A, List[nat]]](xml2name),
           id[List[(term[lab[A, List[nat]], List[Char]],
                     term[lab[A, List[nat]], List[Char]])]],
           ab),
       (ab: xml) => xml2reltrs_proof[A](xml2name, ab),
       (i: (lab[A, List[nat]],
             (List[((lab[A, List[nat]], nat), List[lab[A, List[nat]]])],
               (List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]], List[Char]])],
                 List[(term[lab[A, List[nat]], List[Char]],
                        term[lab[A, List[nat]], List[Char]])]))))
         => (r: List[(term[lab[A, List[nat]], List[Char]],
                       term[lab[A, List[nat]], List[Char]])])
              => (s: List[(term[lab[A, List[nat]], List[Char]],
                            term[lab[A, List[nat]], List[Char]])])
                   => (ab: trs_proof[A, List[nat], List[Char]]) =>
                        Uncurry[A, List[nat], List[Char]](i, r ++ s, ab),
       aa)),
    (List('r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'T', 'e', 'r', 'm', 'i', 'n',
           'a', 't', 'i', 'o', 'n', 'A', 's', 's', 'u', 'm', 'p', 't', 'i', 'o',
           'n'),
      (aa: xml) =>
        xml1element[trs_proof[A, List[nat], List[Char]],
                     trs_proof[A, List[nat],
                                List[Char]]](List('r', 'e', 'l', 'a', 't', 'i',
           'v', 'e', 'T', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', 'A',
           's', 's', 'u', 'm', 'p', 't', 'i', 'o', 'n'),
      (ab: xml) =>
        xml1to2elements[List[(term[lab[A, List[nat]], List[Char]],
                               term[lab[A, List[nat]], List[Char]])],
                         List[(term[lab[A, List[nat]], List[Char]],
                                term[lab[A, List[nat]], List[Char]])],
                         trs_proof[A, List[nat],
                                    List[Char]]](List('t', 'r', 's', 'I', 'n',
               'p', 'u', 't'),
          (ac: xml) =>
            xml1element[List[(term[lab[A, List[nat]], List[Char]],
                               term[lab[A, List[nat]], List[Char]])],
                         List[(term[lab[A, List[nat]], List[Char]],
                                term[lab[A, List[nat]],
                                      List[Char]])]](List('t', 'r', 's'),
              rules[lab[A, List[nat]]](xml2name),
              id[List[(term[lab[A, List[nat]], List[Char]],
                        term[lab[A, List[nat]], List[Char]])]],
              ac),
          (ac: xml) =>
            xml1element[List[(term[lab[A, List[nat]], List[Char]],
                               term[lab[A, List[nat]], List[Char]])],
                         List[(term[lab[A, List[nat]], List[Char]],
                                term[lab[A, List[nat]],
                                      List[Char]])]](List('r', 'e', 'l', 'a',
                   't', 'i', 'v', 'e', 'R', 'u', 'l', 'e', 's'),
              rules[lab[A, List[nat]]](xml2name),
              id[List[(term[lab[A, List[nat]], List[Char]],
                        term[lab[A, List[nat]], List[Char]])]],
              ac),
          (r: List[(term[lab[A, List[nat]], List[Char]],
                     term[lab[A, List[nat]], List[Char]])])
            => (ac: Option[List[(term[lab[A, List[nat]], List[Char]],
                                  term[lab[A, List[nat]], List[Char]])]])
                 => ac match {
                      case None =>
                        AssumeSN[A, List[nat], List[Char]]((Nil, (r, Nil)))
                      case Some(rw) =>
                        AssumeSN[A, List[nat], List[Char]]((Nil, (r, rw)))
                    },
          ab),
      id[trs_proof[A, List[nat], List[Char]]], aa))),
                                       a),
id[trs_proof[A, List[nat], List[Char]]], x)

def xml2trs_disproof[A](xml2name: xml => sum[List[Char], A], x: xml):
      sum[List[Char], trs_disproof[A, List[Char]]] =
  xml1element[trs_disproof[A, List[Char]],
               trs_disproof[A, List[Char]]](List('t', 'r', 's', 'N', 'o', 'n',
          't', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o',
          'o', 'f'),
     (a: xml) =>
       xml_options[trs_disproof[A, List[Char]]](List((List('v', 'a', 'r', 'i',
                    'a', 'b', 'l', 'e', 'C', 'o', 'n', 'd', 'i', 't', 'i', 'o',
                    'n', 'V', 'i', 'o', 'l', 'a', 't', 'e', 'd'),
               (aa: xml) =>
                 xml0elements[trs_disproof[A,
    List[Char]]](List('v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', 'C', 'o', 'n',
                       'd', 'i', 't', 'i', 'o', 'n', 'V', 'i', 'o', 'l', 'a',
                       't', 'e', 'd'),
                  TRS_Not_Well_Formed, aa)),
              (List('l', 'o', 'o', 'p'),
                (aa: xml) =>
                  xml_change[(term[A, List[Char]],
                               (List[(pos, ((term[A, List[Char]],
      term[A, List[Char]]),
     (Boolean, term[A, List[Char]])))],
                                 (List[(List[Char], term[A, List[Char]])],
                                   ctxt[A, List[Char]]))),
                              trs_disproof[A,
    List[Char]]](loop[A](xml2name),
                  (ab: (term[A, List[Char]],
                         (List[(pos, ((term[A, List[Char]],
term[A, List[Char]]),
                                       (Boolean, term[A, List[Char]])))],
                           (List[(List[Char], term[A, List[Char]])],
                             ctxt[A, List[Char]]))))
                    => {
                         val (s, (rseq, (ac, b))):
                               (term[A, List[Char]],
                                 (List[(pos,
 ((term[A, List[Char]], term[A, List[Char]]), (Boolean, term[A, List[Char]])))],
                                   (List[(List[Char], term[A, List[Char]])],
                                     ctxt[A, List[Char]])))
                           = ab;
                         TRS_Loop[A, List[Char]](s,
          mapa[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                       (Boolean, term[A, List[Char]]))),
                (pos, ((term[A, List[Char]], term[A, List[Char]]),
                        term[A, List[Char]]))]((ad:
          (pos, ((term[A, List[Char]], term[A, List[Char]]),
                  (Boolean, term[A, List[Char]]))))
         => {
              val (xa, (y, (_, z))):
                    (pos, ((term[A, List[Char]], term[A, List[Char]]),
                            (Boolean, term[A, List[Char]])))
                = ad;
              (xa, (y, z))
            },
        rseq),
          ac, b)
                       },
                  aa)),
              (List('r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l'),
                (aa: xml) =>
                  xml2elements[List[(term[A, List[Char]], term[A, List[Char]])],
                                trs_disproof[A, List[Char]],
                                trs_disproof[A,
      List[Char]]](List('r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l'),
                    (ab: xml) =>
                      xml1element[List[(term[A, List[Char]],
 term[A, List[Char]])],
                                   List[(term[A, List[Char]],
  term[A, List[Char]])]](List('t', 'r', 's'), rules[A](xml2name),
                          id[List[(term[A, List[Char]], term[A, List[Char]])]],
                          ab),
                    (ab: xml) => xml2trs_disproof[A](xml2name, ab),
                    (ab: List[(term[A, List[Char]], term[A, List[Char]])]) =>
                      (b: trs_disproof[A, List[Char]]) =>
                        TRS_Rule_Removal[A, List[Char]](ab, b),
                    aa)),
              (List('d', 'p', 'T', 'r', 'a', 'n', 's'),
                (aa: xml) =>
                  xml3elements[List[(term[A, List[Char]], term[A, List[Char]])],
                                Boolean, dp_disproof[A, List[Char]],
                                trs_disproof[A,
      List[Char]]](List('d', 'p', 'T', 'r', 'a', 'n', 's'),
                    (ab: xml) =>
                      xml1element[List[(term[A, List[Char]],
 term[A, List[Char]])],
                                   List[(term[A, List[Char]],
  term[A, List[Char]])]](List('d', 'p', 's'), rules[A](xml2name),
                          id[List[(term[A, List[Char]], term[A, List[Char]])]],
                          ab),
                    (ab: xml) =>
                      bool(List('m', 'a', 'r', 'k', 'e', 'd', 'S', 'y', 'm',
                                 'b', 'o', 'l', 's'),
                            ab),
                    (ab: xml) => xml2dp_disproof[A](xml2name, ab),
                    (p: List[(term[A, List[Char]], term[A, List[Char]])]) =>
                      (_: Boolean) =>
                        (ab: dp_disproof[A, List[Char]]) =>
                          TRS_DP_Trans[A, List[Char]](p, ab),
                    aa)),
              (List('s', 't', 'r', 'i', 'n', 'g', 'R', 'e', 'v', 'e', 'r', 's',
                     'a', 'l'),
                (aa: xml) =>
                  xml2elements[List[(term[A, List[Char]], term[A, List[Char]])],
                                trs_disproof[A, List[Char]],
                                trs_disproof[A,
      List[Char]]](List('s', 't', 'r', 'i', 'n', 'g', 'R', 'e', 'v', 'e', 'r',
                         's', 'a', 'l'),
                    (ab: xml) =>
                      xml1element[List[(term[A, List[Char]],
 term[A, List[Char]])],
                                   List[(term[A, List[Char]],
  term[A, List[Char]])]](List('t', 'r', 's'), rules[A](xml2name),
                          id[List[(term[A, List[Char]], term[A, List[Char]])]],
                          ab),
                    (ab: xml) => xml2trs_disproof[A](xml2name, ab),
                    (_: List[(term[A, List[Char]], term[A, List[Char]])]) =>
                      (ab: trs_disproof[A, List[Char]]) =>
                        TRS_String_Reversal[A, List[Char]](ab),
                    aa))),
         a),
     id[trs_disproof[A, List[Char]]], x)

def xml2reltrs_disproof[A](xml2name: xml => sum[List[Char], A], x: xml):
      sum[List[Char], reltrs_disproof[A, List[Char]]] =
  xml1element[reltrs_disproof[A, List[Char]],
               reltrs_disproof[A, List[Char]]](List('r', 'e', 'l', 'a', 't',
             'i', 'v', 'e', 'N', 'o', 'n', 't', 'e', 'r', 'm', 'i', 'n', 'a',
             't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f'),
        (a: xml) =>
          xml_options[reltrs_disproof[A, List[Char]]](List((List('v', 'a', 'r',
                          'i', 'a', 'b', 'l', 'e', 'C', 'o', 'n', 'd', 'i', 't',
                          'i', 'o', 'n', 'V', 'i', 'o', 'l', 'a', 't', 'e',
                          'd'),
                     (aa: xml) =>
                       xml0elements[reltrs_disproof[A,
             List[Char]]](List('v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', 'C', 'o',
                                'n', 'd', 'i', 't', 'i', 'o', 'n', 'V', 'i',
                                'o', 'l', 'a', 't', 'e', 'd'),
                           Rel_Not_Well_Formed, aa)),
                    (List('l', 'o', 'o', 'p'),
                      (aa: xml) =>
                        xml_change[(term[A, List[Char]],
                                     (List[(pos,
     ((term[A, List[Char]], term[A, List[Char]]),
       (Boolean, term[A, List[Char]])))],
                                       (List[(List[Char], term[A, List[Char]])],
 ctxt[A, List[Char]]))),
                                    reltrs_disproof[A,
             List[Char]]](loop[A](xml2name),
                           (ab: (term[A, List[Char]],
                                  (List[(pos,
  ((term[A, List[Char]], term[A, List[Char]]),
    (Boolean, term[A, List[Char]])))],
                                    (List[(List[Char], term[A, List[Char]])],
                                      ctxt[A, List[Char]]))))
                             => {
                                  val (s, (rseq, (ac, b))):
(term[A, List[Char]],
  (List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                (Boolean, term[A, List[Char]])))],
    (List[(List[Char], term[A, List[Char]])], ctxt[A, List[Char]])))
                                    = ab;
                                  Rel_Loop[A, List[Char]](s, rseq, ac, b)
                                },
                           aa)),
                    (List('t', 'r', 's', 'N', 'o', 'n', 't', 'e', 'r', 'm', 'i',
                           'n', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'o',
                           'f'),
                      (aa: xml) =>
                        xml_change[trs_disproof[A, List[Char]],
                                    reltrs_disproof[A,
             List[Char]]]((ab: xml) => xml2trs_disproof[A](xml2name, ab),
                           (ab: trs_disproof[A, List[Char]]) =>
                             Rel_R_Not_SN[A, List[Char]](ab),
                           aa)),
                    (List('r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a',
                           'l'),
                      (aa: xml) =>
                        xml3elements[List[(term[A, List[Char]],
    term[A, List[Char]])],
                                      List[(term[A, List[Char]],
     term[A, List[Char]])],
                                      reltrs_disproof[A, List[Char]],
                                      reltrs_disproof[A,
               List[Char]]](List('r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v',
                                  'a', 'l'),
                             (ab: xml) =>
                               xml1element[List[(term[A, List[Char]],
          term[A, List[Char]])],
    List[(term[A, List[Char]],
           term[A, List[Char]])]](List('t', 'r', 's'), rules[A](xml2name),
                                   id[List[(term[A, List[Char]],
     term[A, List[Char]])]],
                                   ab),
                             (ab: xml) =>
                               xml1element[List[(term[A, List[Char]],
          term[A, List[Char]])],
    List[(term[A, List[Char]],
           term[A, List[Char]])]](List('t', 'r', 's'), rules[A](xml2name),
                                   id[List[(term[A, List[Char]],
     term[A, List[Char]])]],
                                   ab),
                             (ab: xml) => xml2reltrs_disproof[A](xml2name, ab),
                             (ab: List[(term[A, List[Char]],
 term[A, List[Char]])])
                               => (b: List[(term[A, List[Char]],
     term[A, List[Char]])])
                                    => (c: reltrs_disproof[A, List[Char]]) =>
 Rel_Rule_Removal[A, List[Char]](ab, b, c),
                             aa))),
               a),
        id[reltrs_disproof[A, List[Char]]], x)

def subsumption_proof[A](xml2name: xml => sum[List[Char], A]):
      xml =>
        sum[List[Char],
             List[((term[A, List[Char]], term[A, List[Char]]),
                    List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                 (Boolean, term[A, List[Char]])))])]] =
  (a: xml) =>
    xml_many_elements[((term[A, List[Char]], term[A, List[Char]]),
                        List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                     (Boolean, term[A, List[Char]])))]),
                       List[((term[A, List[Char]], term[A, List[Char]]),
                              List[(pos, ((term[A, List[Char]],
    term[A, List[Char]]),
   (Boolean,
     term[A, List[Char]])))])]](List('s', 'u', 'b', 's', 'u', 'm', 'p', 't',
                                      'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f'),
                                 (aa: xml) =>
                                   xml2elements[(term[A, List[Char]],
          term[A, List[Char]]),
         List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                      (Boolean, term[A, List[Char]])))],
         ((term[A, List[Char]], term[A, List[Char]]),
           List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                        (Boolean,
                          term[A, List[Char]])))])](List('r', 'u', 'l', 'e',
                  'S', 'u', 'b', 's', 'u', 'm', 'p', 't', 'i', 'o', 'n', 'P',
                  'r', 'o', 'o', 'f'),
             rule[A](xml2name), conversion[A](xml2name),
             (ab: (term[A, List[Char]], term[A, List[Char]])) =>
               (b: List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                (Boolean, term[A, List[Char]])))])
                 => (ab, b),
             aa),
                                 id[List[((term[A, List[Char]],
    term[A, List[Char]]),
   List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                (Boolean, term[A, List[Char]])))])]],
                                 a)

def xml2completion_proof[A : equal](xml2name:
                                      xml =>
sum[List[Char], lab[A, List[nat]]]):
      xml => sum[List[Char], completion_proof[A, List[nat], List[Char]]] =
  (a: xml) =>
    xml3elements[Option[List[(term[lab[A, List[nat]], List[Char]],
                               (List[(pos, ((term[lab[A, List[nat]],
           List[Char]],
      term[lab[A, List[nat]], List[Char]]),
     term[lab[A, List[nat]], List[Char]]))],
                                 (term[lab[A, List[nat]], List[Char]],
                                   List[(pos,
  ((term[lab[A, List[nat]], List[Char]], term[lab[A, List[nat]], List[Char]]),
    term[lab[A, List[nat]], List[Char]]))])))]],
                  trs_proof[A, List[nat], List[Char]],
                  (List[((term[lab[A, List[nat]], List[Char]],
                           term[lab[A, List[nat]], List[Char]]),
                          List[(pos, ((term[lab[A, List[nat]], List[Char]],
term[lab[A, List[nat]], List[Char]]),
                                       (Boolean,
 term[lab[A, List[nat]], List[Char]])))])],
                    Option[List[((term[lab[A, List[nat]], List[Char]],
                                   term[lab[A, List[nat]], List[Char]]),
                                  List[(pos,
 ((term[lab[A, List[nat]], List[Char]], term[lab[A, List[nat]], List[Char]]),
   (Boolean, term[lab[A, List[nat]], List[Char]])))])]]),
                  completion_proof[A, List[nat],
                                    List[Char]]](List('c', 'o', 'm', 'p', 'l',
               'e', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f'),
          wcr_proof[lab[A, List[nat]]](xml2name),
          (aa: xml) => xml2trs_proof[A](xml2name, aa),
          (aa: xml) =>
            xml1to2elements[List[((term[lab[A, List[nat]], List[Char]],
                                    term[lab[A, List[nat]], List[Char]]),
                                   List[(pos,
  ((term[lab[A, List[nat]], List[Char]], term[lab[A, List[nat]], List[Char]]),
    (Boolean, term[lab[A, List[nat]], List[Char]])))])],
                             List[((term[lab[A, List[nat]], List[Char]],
                                     term[lab[A, List[nat]], List[Char]]),
                                    List[(pos,
   ((term[lab[A, List[nat]], List[Char]], term[lab[A, List[nat]], List[Char]]),
     (Boolean, term[lab[A, List[nat]], List[Char]])))])],
                             (List[((term[lab[A, List[nat]], List[Char]],
                                      term[lab[A, List[nat]], List[Char]]),
                                     List[(pos,
    ((term[lab[A, List[nat]], List[Char]], term[lab[A, List[nat]], List[Char]]),
      (Boolean, term[lab[A, List[nat]], List[Char]])))])],
                               Option[List[((term[lab[A, List[nat]],
           List[Char]],
      term[lab[A, List[nat]], List[Char]]),
     List[(pos, ((term[lab[A, List[nat]], List[Char]],
                   term[lab[A, List[nat]], List[Char]]),
                  (Boolean,
                    term[lab[A, List[nat]],
                          List[Char]])))])]])](List('e', 'q', 'u', 'i', 'v',
             'a', 'l', 'e', 'n', 'c', 'e', 'P', 'r', 'o', 'o', 'f'),
        subsumption_proof[lab[A, List[nat]]](xml2name),
        subsumption_proof[lab[A, List[nat]]](xml2name),
        (ab: List[((term[lab[A, List[nat]], List[Char]],
                     term[lab[A, List[nat]], List[Char]]),
                    List[(pos, ((term[lab[A, List[nat]], List[Char]],
                                  term[lab[A, List[nat]], List[Char]]),
                                 (Boolean,
                                   term[lab[A, List[nat]], List[Char]])))])])
          => (b: Option[List[((term[lab[A, List[nat]], List[Char]],
                                term[lab[A, List[nat]], List[Char]]),
                               List[(pos, ((term[lab[A, List[nat]], List[Char]],
     term[lab[A, List[nat]], List[Char]]),
    (Boolean, term[lab[A, List[nat]], List[Char]])))])]])
               => (ab, b),
        aa),
          (w: Option[List[(term[lab[A, List[nat]], List[Char]],
                            (List[(pos, ((term[lab[A, List[nat]], List[Char]],
   term[lab[A, List[nat]], List[Char]]),
  term[lab[A, List[nat]], List[Char]]))],
                              (term[lab[A, List[nat]], List[Char]],
                                List[(pos, ((term[lab[A, List[nat]],
           List[Char]],
      term[lab[A, List[nat]], List[Char]]),
     term[lab[A, List[nat]], List[Char]]))])))]])
            => (t: trs_proof[A, List[nat], List[Char]]) =>
                 (aa: (List[((term[lab[A, List[nat]], List[Char]],
                               term[lab[A, List[nat]], List[Char]]),
                              List[(pos, ((term[lab[A, List[nat]], List[Char]],
    term[lab[A, List[nat]], List[Char]]),
   (Boolean, term[lab[A, List[nat]], List[Char]])))])],
                        Option[List[((term[lab[A, List[nat]], List[Char]],
                                       term[lab[A, List[nat]], List[Char]]),
                                      List[(pos,
     ((term[lab[A, List[nat]], List[Char]],
        term[lab[A, List[nat]], List[Char]]),
       (Boolean, term[lab[A, List[nat]], List[Char]])))])]]))
                   => {
                        val (ab, b):
                              (List[((term[lab[A, List[nat]], List[Char]],
                                       term[lab[A, List[nat]], List[Char]]),
                                      List[(pos,
     ((term[lab[A, List[nat]], List[Char]],
        term[lab[A, List[nat]], List[Char]]),
       (Boolean, term[lab[A, List[nat]], List[Char]])))])],
                                Option[List[((term[lab[A, List[nat]],
            List[Char]],
       term[lab[A, List[nat]], List[Char]]),
      List[(pos, ((term[lab[A, List[nat]], List[Char]],
                    term[lab[A, List[nat]], List[Char]]),
                   (Boolean, term[lab[A, List[nat]], List[Char]])))])]])
                          = aa;
                        SN_WCR_Eq[A, List[nat], List[Char]](w, t, ab, b)
                      },
          a)

def xml2cert_problem[A : equal](xml2uname: xml => sum[List[Char], A],
                                 xml2name:
                                   xml => sum[List[Char], lab[A, List[nat]]],
                                 x2: xml):
      sum[List[Char], cert_problem[A, List[nat], List[Char]]] =
  (xml2uname, xml2name, x2) match {
  case (xml2uname, xml2name, Xml(name, egal, cs, texta)) =>
    (if (equal_lista[Char](name,
                            List('c', 'e', 'r', 't', 'i', 'f', 'i', 'c', 'a',
                                  't', 'i', 'o', 'n', 'P', 'r', 'o', 'b', 'l',
                                  'e', 'm')) &&
           (is_none[List[Char]](texta) &&
             less_eq_nat(nat(BigInt(3)), size_list[xml](cs))))
      bindb[List[Char],
             sum[(List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])],
                   (strategy[lab[A, List[nat]], List[Char]],
                     List[(term[lab[A, List[nat]], List[Char]],
                            term[lab[A, List[nat]], List[Char]])])),
                  sum[(strategy[lab[A, List[nat]], List[Char]],
                        (List[(term[lab[A, List[nat]], List[Char]],
                                term[lab[A, List[nat]], List[Char]])],
                          Option[List[(term[lab[A, List[nat]], List[Char]],
term[lab[A, List[nat]], List[Char]])]])),
                       (List[(term[lab[A, List[nat]], List[Char]],
                               term[lab[A, List[nat]], List[Char]])],
                         List[(term[lab[A, List[nat]], List[Char]],
                                term[lab[A, List[nat]], List[Char]])])]],
             cert_problem[A, List[nat],
                           List[Char]]]((xml2input[lab[A,
                List[nat]]](xml2name)).apply(nth[xml](cs, Zero_nat)),
 (inp: sum[(List[(term[lab[A, List[nat]], List[Char]],
                   term[lab[A, List[nat]], List[Char]])],
             (strategy[lab[A, List[nat]], List[Char]],
               List[(term[lab[A, List[nat]], List[Char]],
                      term[lab[A, List[nat]], List[Char]])])),
            sum[(strategy[lab[A, List[nat]], List[Char]],
                  (List[(term[lab[A, List[nat]], List[Char]],
                          term[lab[A, List[nat]], List[Char]])],
                    Option[List[(term[lab[A, List[nat]], List[Char]],
                                  term[lab[A, List[nat]], List[Char]])]])),
                 (List[(term[lab[A, List[nat]], List[Char]],
                         term[lab[A, List[nat]], List[Char]])],
                   List[(term[lab[A, List[nat]], List[Char]],
                          term[lab[A, List[nat]], List[Char]])])]])
   => bindb[List[Char], List[Char],
             cert_problem[A, List[nat],
                           List[Char]]](text(List('c', 'p', 'f', 'V', 'e', 'r',
           's', 'i', 'o', 'n'),
      nth[xml](cs, one_nat)),
 (_: List[Char]) =>
   inp match {
     case Inl((p, (q, r))) =>
       xml1element[cert_problem[A, List[nat], List[Char]],
                    cert_problem[A, List[nat],
                                  List[Char]]](List('p', 'r', 'o', 'o', 'f'),
        (a: xml) =>
          xml_options[cert_problem[A, List[nat],
                                    List[Char]]](List((List('d', 'p', 'P', 'r',
                     'o', 'o', 'f'),
                (aa: xml) =>
                  xml_change[dp_proof[A, List[nat], List[Char]],
                              cert_problem[A, List[nat],
    List[Char]]]((ab: xml) => xml2dp_proof[A, List[nat]](xml2name, ab),
                  (ab: dp_proof[A, List[nat], List[Char]]) =>
                    DP_Proof[A, List[nat], List[Char]](p, Nil, q, Nil, r, ab),
                  aa)),
               (List('d', 'p', 'N', 'o', 'n', 't', 'e', 'r', 'm', 'i', 'n', 'a',
                      't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f'),
                 (aa: xml) =>
                   xml_change[dp_disproof[lab[A, List[nat]], List[Char]],
                               cert_problem[A, List[nat],
     List[Char]]]((ab: xml) => xml2dp_disproof[lab[A, List[nat]]](xml2name, ab),
                   (ab: dp_disproof[lab[A, List[nat]], List[Char]]) =>
                     DP_Disproof[A, List[nat], List[Char]](p, q, r, ab),
                   aa))),
          a),
        id[cert_problem[A, List[nat], List[Char]]],
        nth[xml](cs, nat(BigInt(2))))
     case Inr(Inl((q, (r, None)))) =>
       xml1element[cert_problem[A, List[nat], List[Char]],
                    cert_problem[A, List[nat],
                                  List[Char]]](List('p', 'r', 'o', 'o', 'f'),
        (a: xml) =>
          xml_options[cert_problem[A, List[nat],
                                    List[Char]]](List((List('t', 'r', 's', 'T',
                     'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', 'P', 'r',
                     'o', 'o', 'f'),
                (aa: xml) =>
                  xml_change[trs_proof[A, List[nat], List[Char]],
                              cert_problem[A, List[nat],
    List[Char]]]((ab: xml) => xml2trs_proof[A](xml2name, ab),
                  (ab: trs_proof[A, List[nat], List[Char]]) =>
                    TRS_Proof[A, List[nat], List[Char]](q, r, None, ab),
                  aa)),
               (List('c', 'r', 'P', 'r', 'o', 'o', 'f'),
                 (if (nulla[term[lab[A, List[nat]],
                                  List[Char]]](strategy_to_Q[lab[A, List[nat]],
                      List[Char]](q, r)))
                   (aa: xml) =>
                     xml_change[cr_proof[A, List[nat], List[Char]],
                                 cert_problem[A, List[nat],
       List[Char]]]((ab: xml) => xml2cr_proof[A](xml2name, ab),
                     (ab: cr_proof[A, List[nat], List[Char]]) =>
                       TRS_Confluence[A, List[nat], List[Char]](r, ab),
                     aa)
                   else (_: xml) =>
                          Inl[List[Char],
                               cert_problem[A, List[nat],
     List[Char]]](List('s', 't', 'r', 'a', 't', 'e', 'g', 'i', 'e', 's', ' ',
                        'f', 'o', 'r', ' ', 'c', 'o', 'n', 'f', 'l', 'u', 'e',
                        'n', 'c', 'e', ' ', 'u', 'n', 's', 'u', 'p', 'p', 'o',
                        'r', 't', 'e', 'd')))),
               (List('t', 'r', 's', 'N', 'o', 'n', 't', 'e', 'r', 'm', 'i', 'n',
                      'a', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f'),
                 (aa: xml) =>
                   xml_change[trs_disproof[lab[A, List[nat]], List[Char]],
                               cert_problem[A, List[nat],
     List[Char]]]((ab: xml) =>
                    xml2trs_disproof[lab[A, List[nat]]](xml2name, ab),
                   (ab: trs_disproof[lab[A, List[nat]], List[Char]]) =>
                     TRS_Disproof[A, List[nat], List[Char]](q, r, ab),
                   aa))),
          a),
        id[cert_problem[A, List[nat], List[Char]]],
        nth[xml](cs, nat(BigInt(2))))
     case Inr(Inl((q, (r, Some(s))))) =>
       xml1element[cert_problem[A, List[nat], List[Char]],
                    cert_problem[A, List[nat],
                                  List[Char]]](List('p', 'r', 'o', 'o', 'f'),
        (a: xml) =>
          xml_options[cert_problem[A, List[nat],
                                    List[Char]]](List((List('r', 'e', 'l', 'a',
                     't', 'i', 'v', 'e', 'T', 'e', 'r', 'm', 'i', 'n', 'a', 't',
                     'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f'),
                (aa: xml) =>
                  xml_change[trs_proof[A, List[nat], List[Char]],
                              cert_problem[A, List[nat],
    List[Char]]]((ab: xml) => xml2reltrs_proof[A](xml2name, ab),
                  (ab: trs_proof[A, List[nat], List[Char]]) =>
                    TRS_Proof[A, List[nat],
                               List[Char]](q, r,
    Some[List[(term[lab[A, List[nat]], List[Char]],
                term[lab[A, List[nat]], List[Char]])]](s),
    ab),
                  aa)),
               (List('r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'N', 'o', 'n', 't',
                      'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', 'P',
                      'r', 'o', 'o', 'f'),
                 (aa: xml) =>
                   xml_change[reltrs_disproof[lab[A, List[nat]], List[Char]],
                               cert_problem[A, List[nat],
     List[Char]]]((ab: xml) =>
                    xml2reltrs_disproof[lab[A, List[nat]]](xml2name, ab),
                   (ab: reltrs_disproof[lab[A, List[nat]], List[Char]]) =>
                     Rel_TRS_Disproof[A, List[nat], List[Char]](q, r, s, ab),
                   aa))),
          a),
        id[cert_problem[A, List[nat], List[Char]]],
        nth[xml](cs, nat(BigInt(2))))
     case Inr(Inr((e, r))) =>
       xml1element[cert_problem[A, List[nat], List[Char]],
                    cert_problem[A, List[nat],
                                  List[Char]]](List('p', 'r', 'o', 'o', 'f'),
        (a: xml) =>
          xml_options[cert_problem[A, List[nat],
                                    List[Char]]](List((List('c', 'o', 'm', 'p',
                     'l', 'e', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f'),
                (aa: xml) =>
                  xml_change[completion_proof[A, List[nat], List[Char]],
                              cert_problem[A, List[nat],
    List[Char]]](xml2completion_proof[A](xml2name),
                  (ab: completion_proof[A, List[nat], List[Char]]) =>
                    Completion_Proof[A, List[nat], List[Char]](e, r, ab),
                  aa))),
          a),
        id[cert_problem[A, List[nat], List[Char]]],
        nth[xml](cs, nat(BigInt(2))))
   }))
      else Inl[List[Char],
                cert_problem[A, List[nat],
                              List[Char]]](List('c', 'o', 'u', 'l', 'd', ' ',
         'n', 'o', 't', ' ', 'p', 'a', 'r', 's', 'e', ' ', 'r', 'o', 'o', 't',
         ' ', 'e', 'l', 'e', 'm', 'e', 'n', 't', ' ', '<') ++
     (name ++ List('>'))))
}

def xml_cert_problem[A : show, B : show,
                      C : show](x0: cert_problem[A, List[B], C]):
      xml =
  x0 match {
  case TRS_Proof(q, r, None, uu) =>
    Xml(List('t', 'r', 's', 'I', 'n', 'p', 'u', 't'), Nil,
         List(xml_rules[A, B, C](List('t', 'r', 's'), r)) ++ xml_Q[A, B, C](q),
         None)
  case TRS_Proof(q, r, Some(s), uv) =>
    Xml(List('t', 'r', 's', 'I', 'n', 'p', 'u', 't'), Nil,
         xml_rules[A, B, C](List('t', 'r', 's'), r) ::
           xml_Q[A, B, C](q) ++
             List(xml_rules[A, B,
                             C](List('r', 'e', 'l', 'a', 't', 'i', 'v', 'e',
                                      'R', 'u', 'l', 'e', 's'),
                                 s)),
         None)
  case TRS_Disproof(q, r, uw) =>
    Xml(List('t', 'r', 's', 'I', 'n', 'p', 'u', 't'), Nil,
         List(xml_rules[A, B, C](List('t', 'r', 's'), r)) ++ xml_Q[A, B, C](q),
         None)
  case Rel_TRS_Disproof(q, r, s, ux) =>
    Xml(List('t', 'r', 's', 'I', 'n', 'p', 'u', 't'), Nil,
         xml_rules[A, B, C](List('t', 'r', 's'), r) ::
           xml_Q[A, B, C](q) ++
             List(xml_rules[A, B,
                             C](List('r', 'e', 'l', 'a', 't', 'i', 'v', 'e',
                                      'R', 'u', 'l', 'e', 's'),
                                 s)),
         None)
  case DP_Proof(p, pw, q, r, rw, uy) =>
    Xml(List('d', 'p', 'I', 'n', 'p', 'u', 't'), Nil,
         xml_rules[A, B, C](List('t', 'r', 's'), rw) ::
           xml_rules[A, B, C](List('d', 'p', 's'), p) ::
             xml_Q[A, B, C](q) ++
               List(Xml(List('m', 'i', 'n', 'i', 'm', 'a', 'l'), Nil, Nil,
                         Some[List[Char]](List('t', 'r', 'u', 'e')))),
         None)
  case DP_Disproof(p, q, r, uz) =>
    Xml(List('d', 'p', 'I', 'n', 'p', 'u', 't'), Nil,
         xml_rules[A, B, C](List('t', 'r', 's'), r) ::
           xml_rules[A, B, C](List('d', 'p', 's'), p) ::
             xml_Q[A, B, C](q) ++
               List(Xml(List('m', 'i', 'n', 'i', 'm', 'a', 'l'), Nil, Nil,
                         Some[List[Char]](List('t', 'r', 'u', 'e')))),
         None)
  case TRS_Confluence(r, va) =>
    Xml(List('t', 'r', 's', 'I', 'n', 'p', 'u', 't'), Nil,
         List(xml_rules[A, B, C](List('t', 'r', 's'), r)), None)
  case Completion_Proof(e, r, vb) =>
    Xml(List('c', 'o', 'm', 'p', 'l', 'e', 't', 'i', 'o', 'n', 'I', 'n', 'p',
              'u', 't'),
         Nil,
         List(xml_rules[A, B,
                         C](List('e', 'q', 'u', 'a', 't', 'i', 'o', 'n', 's'),
                             e),
               xml_rules[A, B, C](List('t', 'r', 's'), r)),
         None)
}

def xmldoc2cert_problem(x0: xmldoc):
      sum[List[Char], cert_problem[List[Char], List[nat], List[Char]]] =
  x0 match {
  case Xmldoc(header, xml) =>
    debug[sum[List[Char],
               cert_problem[List[Char], List[nat],
                             List[Char]]]](List('0'),
    List('p', 'a', 'r', 's', 'i', 'n', 'g', ' ', 'x', 'm', 'l', ' ', 't', 'o',
          ' ', 'c', 'e', 'r', 't', 'i', 'f', 'i', 'c', 'a', 't', 'i', 'o', 'n',
          ' ', 'p', 'r', 'o', 'b', 'l', 'e', 'm'),
    xml2cert_problem[List[Char]]((a: xml) => text(List('n', 'a', 'm', 'e'), a),
                                  (a: xml) => plain_name(a), xml))
}

def parse_cert_problem(s: List[Char]):
      sum[List[Char], cert_problem[List[Char], List[nat], List[Char]]] =
  bindb[List[Char], xmldoc,
         cert_problem[List[Char], List[nat],
                       List[Char]]](doc_of_string(s),
                                     (a: xmldoc) => xmldoc2cert_problem(a))

abstract sealed class dpp[+A, +B]
final case class
  Dpp[A, B](a: (List[(term[A, B], term[A, B])],
                 (List[(term[A, B], term[A, B])],
                   (List[term[A, B]],
                     (Boolean,
                       (Boolean,
                         (List[(term[A, B], term[A, B])],
                           (List[(term[A, B], term[A, B])],
                             (rbt[(A, nat),
                                   List[(Boolean, (term[A, B], term[A, B]))]],
                               rbt[(A, nat),
                                    List[(Boolean,
   (term[A, B], term[A, B]))]])))))))))
  extends dpp[A, B]

def extract_input_proof(s: List[Char]):
      Option[(List[Char], (List[Char], List[Char]))] =
  search_len(s, List('<', 'i', 'n', 'p', 'u', 't', '>'),
              size_list[Char](List('<', 'i', 'n', 'p', 'u', 't', '>')))
    match {
    case None => None
    case Some((_, minput)) =>
      search_len(minput, List('<', '/', 'i', 'n', 'p', 'u', 't', '>'),
                  size_list[Char](List('<', '/', 'i', 'n', 'p', 'u', 't', '>')))
        match {
        case None => None
        case Some((the_input, sa)) =>
          search_len(sa, List('<', 'p', 'r', 'o', 'o', 'f', '>'),
                      size_list[Char](List('<', 'p', 'r', 'o', 'o', 'f', '>')))
            match {
            case None => None
            case Some((version, prf)) =>
              Some[(List[Char],
                     (List[Char], List[Char]))]((the_input, (version, prf)))
          }
      }
  }

def shows_terms[A, B](fun: A => (List[Char]) => List[Char],
                       vara: B => (List[Char]) => List[Char], name: List[Char],
                       ts: List[term[A, B]]):
      (List[Char]) => List[Char] =
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(name),
                     comp[List[Char], List[Char],
                           List[Char]](shows_nl,
comp[List[Char], List[Char],
      List[Char]](shows_list_gen[term[A, B]]((a: term[A, B]) =>
       shows_term[A, B](fun, vara, a),
      Nil, Nil, List(newline), Nil, ts),
                   shows_nl)))

def shows_tp[A, B,
              C](fun: A => (List[Char]) => List[Char],
                  vara: B => (List[Char]) => List[Char],
                  i: tp_ops_ext[C, A, B, Unit], t: C):
      (List[Char]) => List[Char] =
  {
    val r: List[(term[A, B], term[A, B])] = (ra[C, A, B, Unit](i)).apply(t)
    val rw: List[(term[A, B], term[A, B])] = (rwa[C, A, B, Unit](i)).apply(t)
    val q: List[term[A, B]] = (qa[C, A, B, Unit](i)).apply(t);
    comp[List[Char], List[Char],
          List[Char]](shows_trs[A, B](fun, vara,
                                       List('r', 'u', 'l', 'e', 's', ':'),
                                       List(' ', '-', '>', ' '), r),
                       comp[List[Char], List[Char],
                             List[Char]]((if (nulla[(term[A, B],
              term[A, B])](rw))
   id[List[Char]]
   else shows_trs[A, B](fun, vara,
                         List('r', 'e', 'l', 'a', 't', 'i', 'v', 'e', ' ', 'r',
                               'u', 'l', 'e', 's', ':'),
                         List(' ', '-', '>', '=', ' '), rw)),
  (if (nulla[term[A, B]](q)) id[List[Char]]
    else shows_terms[A, B](fun, vara,
                            List('Q', '-', 'c', 'o', 'm', 'p', 'o', 'n', 'e',
                                  'n', 't', ':'),
                            q))))
  }

def shows_dpp[A, B,
               C](fun: A => (List[Char]) => List[Char],
                   vara: B => (List[Char]) => List[Char],
                   i: dpp_ops_ext[C, A, B, Unit], d: C):
      (List[Char]) => List[Char] =
  {
    val p: List[(term[A, B], term[A, B])] = (pa[C, A, B, Unit](i)).apply(d)
    val pw: List[(term[A, B], term[A, B])] = (pwa[C, A, B, Unit](i)).apply(d)
    val r: List[(term[A, B], term[A, B])] = (rc[C, A, B, Unit](i)).apply(d)
    val rw: List[(term[A, B], term[A, B])] = (rwc[C, A, B, Unit](i)).apply(d)
    val q: List[term[A, B]] = (qc[C, A, B, Unit](i)).apply(d);
    comp[List[Char], List[Char],
          List[Char]](shows_trs[A, B](fun, vara,
                                       List('p', 'a', 'i', 'r', 's', ':'),
                                       List(' ', '-', '>', ' '), p),
                       comp[List[Char], List[Char],
                             List[Char]]((if (nulla[(term[A, B],
              term[A, B])](pw))
   id[List[Char]]
   else shows_trs[A, B](fun, vara,
                         List('w', 'e', 'a', 'k', ' ', 'p', 'a', 'i', 'r', 's',
                               ':'),
                         List(' ', '-', '>', '=', ' '), pw)),
  comp[List[Char], List[Char],
        List[Char]]((if (nulla[(term[A, B], term[A, B])](r)) id[List[Char]]
                      else shows_trs[A, B](fun, vara,
    List('s', 't', 'r', 'i', 'c', 't', ' ', 'r', 'u', 'l', 'e', 's', ':'),
    List(' ', '-', '>', '!', ' '), r)),
                     comp[List[Char], List[Char],
                           List[Char]](shows_trs[A,
          B](fun, vara, List('r', 'u', 'l', 'e', 's', ':'),
              List(' ', '-', '>', ' '), rw),
(if (nulla[term[A, B]](q)) id[List[Char]]
  else shows_terms[A, B](fun, vara,
                          List('Q', '-', 'c', 'o', 'm', 'p', 'o', 'n', 'e', 'n',
                                't', ':'),
                          q))))))
  }

def mgu_var_disjoint_generic[A, B : equal, C,
                              D : equal](vu: A => B, wu: C => B, s: term[D, A],
  t: term[D, C]):
      Option[(A => term[D, B], C => term[D, B])] =
  mgu[D, B](map_vars[A, B, D](vu, s), map_vars[C, B, D](wu, t)) match {
    case None => None
    case Some(gamma) =>
      Some[(A => term[D, B],
             C => term[D, B])]((map_subst_domain[A, B, D, B](vu, gamma),
                                 map_subst_domain[C, B, D, B](wu, gamma)))
  }

def qnarrows_impl[A : equal](isnf: (term[A, List[Char]]) => Boolean,
                              r: List[(term[A, List[Char]],
term[A, List[Char]])],
                              t: term[A, List[Char]]):
      List[(term[A, List[Char]], (List[Char]) => term[A, List[Char]])] =
  maps[pos, (term[A, List[Char]],
              (List[Char]) =>
                term[A, List[Char]])]((p: pos) =>
{
  val tp: term[A, List[Char]] = subt_at[A, List[Char]](t, p);
  (if (! (is_Var[A, List[Char]](tp)))
    maps[(term[A, List[Char]], term[A, List[Char]]),
          (term[A, List[Char]],
            (List[Char]) =>
              term[A, List[Char]])]((a: (term[A, List[Char]],
  term[A, List[Char]]))
                                      => {
   val (l, ra): (term[A, List[Char]], term[A, List[Char]]) = a;
   maps[((List[Char]) => term[A, List[Char]],
          (List[Char]) => term[A, List[Char]]),
         (term[A, List[Char]],
           (List[Char]) =>
             term[A, List[Char]])]((aa: ((List[Char]) => term[A, List[Char]],
  (List[Char]) => term[A, List[Char]]))
                                     => {
  val (mu_1, mu_2):
        ((List[Char]) => term[A, List[Char]],
          (List[Char]) => term[A, List[Char]])
    = aa;
  (if (list_all[term[A, List[Char]]](isnf,
                                      args[A,
    List[Char]](subst_apply_term[A, List[Char], List[Char]](l, mu_2))))
    List((ctxt_apply[A, List[Char]](ctxt_of_pos_term[A,
              List[Char]](p, subst_apply_term[A, List[Char],
       List[Char]](t, mu_1)),
                                     subst_apply_term[A, List[Char],
               List[Char]](ra, mu_2)),
           mu_1))
    else Nil)
},
                                    option_to_list[((List[Char]) =>
              term[A, List[Char]],
             (List[Char]) =>
               term[A, List[Char]])](mgu_var_disjoint_generic[List[Char],
                       List[Char], List[Char],
                       A]((aa: List[Char]) => 'x' :: aa,
                           (aa: List[Char]) => 'y' :: aa, tp, l)))
 },
                                     r)
    else Nil)
},
                                       poss_list[A, List[Char]](t))

def pairsa[A, B, C, D](x0: dpp_ops_ext[A, B, C, D]):
      A => List[(term[B, C], term[B, C])] =
  x0 match {
  case Dpp_ops_ext(dpp, p, pw, pairs, q, r, rw, rules, q_empty,
                    rules_no_left_var, rules_non_collapsing, is_QNF,
                    nFQ_subset_NF_rules, rules_map, reverse_rules_map,
                    intersect_pairs, replace_pair, intersect_rules, delete_P_Pw,
                    delete_R_Rw, split_pairs, split_rules, mk, more)
    => pairs
}

def is_iedg_edge_dpp[A, B : equal : linorder : show](i:
               dpp_ops_ext[A, B, List[Char], Unit],
              d: A):
      ((term[B, List[Char]], term[B, List[Char]])) =>
        (term[B, List[Char]]) => Boolean =
  {
    val qnf: (term[B, List[Char]]) => Boolean =
      (is_QNFc[A, B, List[Char], Unit](i)).apply(d)
    val ic: (List[term[B, List[Char]]]) =>
              (term[B, List[Char]]) => term[B, sum[Unit, List[Char]]]
      = icap_impl_dpp_mv[A, B](i, d);
    (a: (term[B, List[Char]], term[B, List[Char]])) =>
      {
        val (s, t): (term[B, List[Char]], term[B, List[Char]]) = a
        val cst: term[B, sum[Unit, List[Char]]] = (ic(List(s)))(t);
        (u: term[B, List[Char]]) =>
          mgu_class[B](cst, u) match {
            case None => false
            case Some(mu) =>
              qnf(subst_apply_term[B, List[Char],
                                    List[Char]](map_vars[List[Char], List[Char],
                  B]((aa: List[Char]) => 'x' :: aa, s),
         mu)) &&
                qnf(subst_apply_term[B, List[Char],
                                      List[Char]](map_vars[List[Char],
                    List[Char], B]((aa: List[Char]) => 'y' :: aa, u),
           mu))
          }
      }
  }

def replace_paira[A, B, C, D](x0: dpp_ops_ext[A, B, C, D]):
      A => ((term[B, C], term[B, C])) => (List[(term[B, C], term[B, C])]) => A =
  x0 match {
  case Dpp_ops_ext(dpp, p, pw, pairs, q, r, rw, rules, q_empty,
                    rules_no_left_var, rules_non_collapsing, is_QNF,
                    nFQ_subset_NF_rules, rules_map, reverse_rules_map,
                    intersect_pairs, replace_pair, intersect_rules, delete_P_Pw,
                    delete_R_Rw, split_pairs, split_rules, mk, more)
    => replace_pair
}

def narrowing_proc[A, B : equal : linorder : show](i:
             dpp_ops_ext[A, B, List[Char], Unit],
            st: (term[B, List[Char]], term[B, List[Char]]), p: pos,
            sts: List[(term[B, List[Char]], term[B, List[Char]])], dpp: A):
      sum[(List[Char]) => List[Char], A] =
  {
    val (s, t): (term[B, List[Char]], term[B, List[Char]]) = st;
    bindb[(List[Char]) => List[Char], Unit,
           Unit](check[(List[Char]) =>
                         List[Char]]((nFQ_subset_NF_rulesb[A, B, List[Char],
                    Unit](i)).apply(dpp) ||
                                       nulla[term[B,
           List[Char]]]((qc[A, B, List[Char], Unit](i)).apply(dpp)) &&
 linear_term[B, List[Char]](t),
                                      shows_prec_list[Char](Zero_nat,
                     List('i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', ' ', 'o',
                           'r', ' ', 'f', 'u', 'l', 'l', ' ', 'r', 'e', 'w',
                           'r', 'i', 't', 'i', 'n', 'g', ' ', 'r', 'e', 'q',
                           'u', 'i', 'r', 'e', 'd', ' ', '(', 'a', 'n', 'd',
                           ' ', 'l', 'i', 'n', 'e', 'a', 'r', 'i', 't', 'y',
                           ' ', 'o', 'f', ' ', 't', ' ', 'i', 'n', ' ', 'f',
                           'u', 'l', 'l', ' ', 'r', 'e', 'w', 'r', 'i', 't',
                           'i', 'n', 'g', ' ', 'c', 'a', 's', 'e', ')'))),
                  (_: Unit) =>
                    {
                      val ic: (List[term[B, List[Char]]]) =>
                                (term[B, List[Char]]) =>
                                  term[B, sum[Unit, List[Char]]]
                        = icap_impl_dpp_mv[A, B](i, dpp)
                      val isnf: (term[B, List[Char]]) => Boolean =
                        (is_QNFc[A, B, List[Char], Unit](i)).apply(dpp)
                      val pairs:
                            List[(term[B, List[Char]], term[B, List[Char]])]
                        = (pairsa[A, B, List[Char], Unit](i)).apply(dpp);
                      bindb[(List[Char]) => List[Char], Unit,
                             Unit](check[(List[Char]) =>
   List[Char]](membera[pos](poss_list[B, sum[Unit,
      List[Char]]]((ic(List(s)))(t)),
                             p),
                comp[List[Char], List[Char],
                      List[Char]](shows_prec_list[Char](Zero_nat,
                 List('p', 'o', 's', 'i', 't', 'i', 'o', 'n', ' ', 'n', 'o',
                       't', ' ', 'c', 'o', 'n', 't', 'a', 'i', 'n', 'e', 'd',
                       ' ', 'i', 'n', ' ', 'c', 'a', 'p', 'p', 'e', 'd', ' ',
                       't', 'e', 'r', 'm', ' ', 'o', 'f', ' ')),
                                   (shows_prec_term[B,
             List[Char]](Zero_nat)).apply(t))),
                                    (_: Unit) =>
                                      {
val tp: term[B, List[Char]] = subt_at[B, List[Char]](t, p)
val narrows: List[(term[B, List[Char]], (List[Char]) => term[B, List[Char]])] =
  qnarrows_impl[B](isnf, (rulesd[A, B, List[Char], Unit](i)).apply(dpp), tp)
val stsa: List[(term[B, List[Char]], term[B, List[Char]])] =
  mapa[(term[B, List[Char]], (List[Char]) => term[B, List[Char]]),
        (term[B, List[Char]],
          term[B, List[Char]])]((a: (term[B, List[Char]],
                                      (List[Char]) => term[B, List[Char]]))
                                  => {
                                       val
 (ta, mu): (term[B, List[Char]], (List[Char]) => term[B, List[Char]]) = a;
                                       (subst_apply_term[B, List[Char],
                  List[Char]](s, mu),
 ctxt_apply[B, List[Char]](ctxt_of_pos_term[B,
     List[Char]](p, subst_apply_term[B, List[Char], List[Char]](t, mu)),
                            ta))
                                     },
                                 narrows);
bindb[(List[Char]) => List[Char], Unit,
       Unit](catcha[((term[B, List[Char]], term[B, List[Char]]),
                      (List[Char]) => List[Char]),
                     Unit,
                     (List[Char]) =>
                       List[Char]](forallm[(term[B, List[Char]],
     term[B, List[Char]]),
    (List[Char]) =>
      List[Char]]((newa: (term[B, List[Char]], term[B, List[Char]])) =>
                    check[(List[Char]) =>
                            List[Char]](list_ex[(term[B, List[Char]],
          term[B, List[Char]])]((a: (term[B, List[Char]], term[B, List[Char]]))
                                  => instance_rule[B, List[Char],
            List[Char]](newa, a),
                                 sts),
 comp[List[Char], List[Char],
       List[Char]](shows_prec_list[Char](Zero_nat,
  List('c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'f', 'i', 'n', 'd',
        ' ', 'n', 'a', 'r', 'r', 'o', 'w', 'e', 'd', ' ', 'p', 'a', 'i', 'r',
        ' ')),
                    shows_rule[B, List[Char]]((a: B) =>
        (b: List[Char]) => shows_prec[B](Zero_nat, a, b),
       (a: List[Char]) => shows_prec_list[Char](Zero_nat, a),
       List(' ', '-', '>', ' '), newa))),
                   stsa),
                                    (x: ((term[B, List[Char]],
   term[B, List[Char]]),
  (List[Char]) => List[Char]))
                                      => Inl[(List[Char]) => List[Char],
      Unit](snd[(term[B, List[Char]], term[B, List[Char]]),
                 (List[Char]) => List[Char]](x))),
              (_: Unit) =>
                {
                  val iedg: (term[B, List[Char]]) => Boolean =
                    (is_iedg_edge_dpp[A, B](i, dpp)).apply((s, t));
                  bindb[(List[Char]) => List[Char], Unit,
                         Unit](catcha[List[Char], Unit,
                                       (List[Char]) =>
 List[Char]](check_subseteq[List[Char]](vars_term_list[B, List[Char]].apply(tp),
 vars_term_list[B, List[Char]].apply(s)),
              (x: List[Char]) =>
                Inl[(List[Char]) => List[Char],
                     Unit](shows_prec_list[Char](Zero_nat,
          List('v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ') ++
            (x ++ List(' ', 'o', 'n', 'l', 'y', ' ', 'o', 'c', 'c', 'u', 'r',
                        's', ' ', 'o', 'n', ' ', 'r', 'h', 's', ' ', 'o', 'f',
                        ' ', 'p', 'a', 'i', 'r'))))),
                                (_: Unit) =>
                                  bindb[(List[Char]) => List[Char], Unit,
 Unit](check[(List[Char]) =>
               List[Char]](membera[(term[B, List[Char]],
                                     term[B,
   List[Char]])]((pa[A, B, List[Char], Unit](i)).apply(dpp), st) ||
                             nulla[(term[B, List[Char]],
                                     term[B,
   List[Char]])]((rc[A, B, List[Char], Unit](i)).apply(dpp)),
                            shows_prec_list[Char](Zero_nat,
           List('s', 't', 'r', 'i', 'c', 't', ' ', 'D', 'P', ' ', 'o', 'r', ' ',
                 'n', 'o', ' ', 's', 't', 'r', 'i', 'c', 't', ' ', 'r', 'u',
                 'l', 'e', 's', ' ', 'r', 'e', 'q', 'u', 'i', 'r', 'e', 'd'))),
        (_: Unit) =>
          (if (isnf(tp))
            catcha[((term[B, List[Char]], term[B, List[Char]]),
                     (List[Char]) => List[Char]),
                    Unit,
                    (List[Char]) =>
                      List[Char]](forallm[(term[B, List[Char]],
    term[B, List[Char]]),
   (List[Char]) =>
     List[Char]]((a: (term[B, List[Char]], term[B, List[Char]])) =>
                   {
                     val (u, v): (term[B, List[Char]], term[B, List[Char]]) = a;
                     bindb[(List[Char]) => List[Char], Unit,
                            Unit](check[(List[Char]) =>
  List[Char]](membera[pos](poss_list[B, List[Char]](u), p),
               comp[List[Char], List[Char],
                     List[Char]](shows_prec_list[Char](Zero_nat,
                List('p', 'o', 's', 'i', 't', 'i', 'o', 'n', ' ', 'n', 'o', 't',
                      ' ', 'c', 'o', 'n', 't', 'a', 'i', 'n', 'e', 'd', ' ',
                      'i', 'n', ' ', 'l', 'h', 's', ' ', 'o', 'f', ' ', 'p',
                      'a', 'i', 'r', ' ')),
                                  shows_rule[B,
      List[Char]]((aa: B) => (b: List[Char]) => shows_prec[B](Zero_nat, aa, b),
                   (aa: List[Char]) => shows_prec_list[Char](Zero_nat, aa),
                   List(' ', '-', '>', ' '), (u, v)))),
                                   (_: Unit) =>
                                     mgu_var_disjoint_generic[List[Char],
                       List[Char], List[Char],
                       B]((aa: List[Char]) => 'x' :: aa,
                           (aa: List[Char]) => 'y' :: aa, tp,
                           subt_at[B, List[Char]](u, p))
                                       match {
                                       case None =>
 Inr[Unit, (List[Char]) => List[Char]](())
                                       case Some((mu_1, mu_2)) =>
 check[(List[Char]) =>
         List[Char]](! (isnf(subst_apply_term[B, List[Char],
       List[Char]](s, mu_1))) ||
                       ! (isnf(subst_apply_term[B, List[Char],
         List[Char]](u, mu_2))),
                      comp[List[Char], List[Char],
                            List[Char]](shows_prec_list[Char](Zero_nat,
                       List('t', ' ', '|', '_', ' ', 'p', ' ', 'a', 'n', 'd',
                             ' ', 'u', ' ', '|', '_', ' ', 'p', ' ', 'u', 'n',
                             'i', 'f', 'y', ' ', 'a', 'n', 'd', ' ', 's', 'a',
                             't', 'i', 's', 'f', 'y', ' ', 'v', 'a', 'r', 'i',
                             'a', 'b', 'l', 'e', ' ', 'c', 'o', 'n', 'd', 'i',
                             't', 'i', 'o', 'n', ' ', 'f', 'o', 'r', ' ', 'p',
                             'a', 'i', 'r', ' ', '(', 'u', ',', 'v', ')', ' ',
                             '=', ' ')),
 shows_rule[B, List[Char]]((aa: B) =>
                             (b: List[Char]) => shows_prec[B](Zero_nat, aa, b),
                            (aa: List[Char]) =>
                              shows_prec_list[Char](Zero_nat, aa),
                            List(' ', '-', '>', ' '), (u, v))))
                                     })
                   },
                  filter[(term[B, List[Char]],
                           term[B, List[Char]])]((a:
            (term[B, List[Char]], term[B, List[Char]]))
           => {
                val (u, _): (term[B, List[Char]], term[B, List[Char]]) = a;
                iedg(u)
              },
          pairs)),
                                   (x: ((term[B, List[Char]],
  term[B, List[Char]]),
 (List[Char]) => List[Char]))
                                     => Inl[(List[Char]) => List[Char],
     Unit](snd[(term[B, List[Char]], term[B, List[Char]]),
                (List[Char]) => List[Char]](x)))
            else Inr[Unit, (List[Char]) => List[Char]](()))))
                })
                                      })
                    })
  }
    match {
    case Inl(a) => Inl[(List[Char]) => List[Char], A](a)
    case Inr(_) =>
      Inr[A, (List[Char]) =>
               List[Char]]((replace_paira[A, B, List[Char],
   Unit](i)).apply(dpp).apply(st).apply(sts))
  }

def critical_pairs_impl[A : equal](p: List[(term[A, List[Char]],
     term[A, List[Char]])],
                                    r: List[(term[A, List[Char]],
      term[A, List[Char]])]):
      List[(Boolean, (term[A, List[Char]], term[A, List[Char]]))] =
  maps[(term[A, List[Char]], term[A, List[Char]]),
        (Boolean,
          (term[A, List[Char]],
            term[A, List[Char]]))]((a: (term[A, List[Char]],
 term[A, List[Char]]))
                                     => {
  val (l, ra): (term[A, List[Char]], term[A, List[Char]]) = a;
  maps[pos, (Boolean,
              (term[A, List[Char]],
                term[A, List[Char]]))]((pa: pos) =>
 {
   val c: ctxt[A, List[Char]] = ctxt_of_pos_term[A, List[Char]](pa, l)
   val la: term[A, List[Char]] = subt_at[A, List[Char]](l, pa)
   val b: Boolean = equal_ctxt[A, List[Char]](c, Hole);
   (if (is_Var[A, List[Char]](la)) Nil
     else maps[(term[A, List[Char]], term[A, List[Char]]),
                (Boolean,
                  (term[A, List[Char]],
                    term[A, List[Char]]))]((aa:
      (term[A, List[Char]], term[A, List[Char]]))
     => {
          val (laa, rb): (term[A, List[Char]], term[A, List[Char]]) = aa;
          mgu_var_disjoint_generic[List[Char], List[Char], List[Char],
                                    A]((ab: List[Char]) => 'x' :: ab,
(ab: List[Char]) => 'y' :: ab, la, laa)
            match {
            case None => Nil
            case Some((sigma, tau)) =>
              List((b, (subst_apply_term[A, List[Char], List[Char]](ra, sigma),
                         ctxt_apply[A, List[Char]](subst_apply_ctxt[A,
                             List[Char], List[Char]](c, sigma),
            subst_apply_term[A, List[Char], List[Char]](rb, tau)))))
          }
        },
    r))
 },
poss_list[A, List[Char]](l))
},
                                    p)

def check_prop_rstep_rule[A : equal : show,
                           B : equal : show](pa:
       (term[A, B]) => sum[(List[Char]) => List[Char], Unit],
      p: pos, rule: (term[A, B], term[A, B]), s: term[A, B], t: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  bindb[(List[Char]) => List[Char], Unit,
         Unit](check[(List[Char]) =>
                       List[Char]](in_poss[A, B](p, s),
                                    comp[List[Char], List[Char],
  List[Char]](shows_prec_pos(Zero_nat, p),
               comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List(' ', 'i', 's', ' ',
                  'n', 'o', 't', ' ', 'a', ' ', 'p', 'o', 's', 'i', 't', 'i',
                  'o', 'n', ' ', 'o', 'f', ' ')),
                                  comp[List[Char], List[Char],
List[Char]]((shows_prec_term[A, B](Zero_nat)).apply(s), shows_nl)))),
                (_: Unit) =>
                  bindb[(List[Char]) => List[Char], Unit,
                         Unit](check[(List[Char]) =>
                                       List[Char]](in_poss[A, B](p, t),
            comp[List[Char], List[Char],
                  List[Char]](shows_prec_pos(Zero_nat, p),
                               comp[List[Char], List[Char],
                                     List[Char]](shows_string.apply(List(' ',
                                  'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ',
                                  'p', 'o', 's', 'i', 't', 'i', 'o', 'n', ' ',
                                  'o', 'f', ' ')),
          comp[List[Char], List[Char],
                List[Char]]((shows_prec_term[A, B](Zero_nat)).apply(t),
                             shows_nl)))),
                                (_: Unit) =>
                                  {
                                    val c: ctxt[A, B] =
                                      ctxt_of_pos_term[A, B](p, s)
                                    val d: ctxt[A, B] =
                                      ctxt_of_pos_term[A, B](p, t)
                                    val u: term[A, B] = subt_at[A, B](s, p)
                                    val v: term[A, B] = subt_at[A, B](t, p);
                                    match_list[A, B,
        B](u, List((u, fst[term[A, B], term[A, B]](rule)),
                    (v, snd[term[A, B], term[A, B]](rule))))
                                      match {
                                      case None =>
Inl[(List[Char]) => List[Char],
     Unit](comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List('t', 'h', 'e', ' ', 't',
              'e', 'r', 'm', ' ')),
                              comp[List[Char], List[Char],
                                    List[Char]]((shows_prec_term[A,
                          B](Zero_nat)).apply(t),
         comp[List[Char], List[Char],
               List[Char]](shows_string.apply(List(' ', 'd', 'o', 'e', 's', ' ',
            'n', 'o', 't', ' ', 'r', 'e', 's', 'u', 'l', 't', ' ', 'f', 'r',
            'o', 'm', ' ', 'a', ' ', 'p', 'r', 'o', 'p', 'e', 'r', ' ', 'a',
            'p', 'p', 'l', 'i', 'c', 'a', 't', 'i', 'o', 'n', ' ', 'o', 'f',
            ' ', 'r', 'u', 'l', 'e', ' ')),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_nl,
       comp[List[Char], List[Char],
             List[Char]](shows_rule[A, B]((a: A) =>
    (b: List[Char]) => shows_prec[A](Zero_nat, a, b),
   (a: B) => (b: List[Char]) => shows_prec[B](Zero_nat, a, b),
   List(' ', '-', '>', ' '), rule),
                          comp[List[Char], List[Char],
                                List[Char]](shows_string.apply(List(' ', 'a',
                             't', ' ', 'p', 'o', 's', 'i', 't', 'i', 'o', 'n',
                             ' ')),
     comp[List[Char], List[Char],
           List[Char]](shows_prec_pos(Zero_nat, p), shows_nl))))))))
                                      case Some(_) =>
bindb[(List[Char]) => List[Char], Unit,
       Unit](catcha[(term[A, B], (List[Char]) => List[Char]), Unit,
                     (List[Char]) =>
                       List[Char]](forallm[term[A, B],
    (List[Char]) => List[Char]](pa, args[A, B](u)),
                                    (x: (term[A, B],
  (List[Char]) => List[Char]))
                                      => Inl[(List[Char]) => List[Char],
      Unit](snd[term[A, B], (List[Char]) => List[Char]](x))),
              (_: Unit) =>
                check[(List[Char]) =>
                        List[Char]](equal_ctxt[A, B](c, d),
                                     comp[List[Char], List[Char],
   List[Char]](shows_string.apply(List('t', 'h', 'e', ' ', 't', 'e', 'r', 'm',
' ')),
                comp[List[Char], List[Char],
                      List[Char]]((shows_prec_term[A, B](Zero_nat)).apply(t),
                                   comp[List[Char], List[Char],
 List[Char]](shows_string.apply(List(' ', 'd', 'o', 'e', 's', ' ', 'n', 'o',
                                      't', ' ', 'r', 'e', 's', 'u', 'l', 't',
                                      ' ', 'f', 'r', 'o', 'm', ' ', 'a', ' ',
                                      'p', 'r', 'o', 'p', 'e', 'r', ' ', 'a',
                                      'p', 'p', 'l', 'i', 'c', 'a', 't', 'i',
                                      'o', 'n', ' ', 'o', 'f', ' ', 'r', 'u',
                                      'l', 'e', ' ')),
              comp[List[Char], List[Char],
                    List[Char]](shows_nl,
                                 comp[List[Char], List[Char],
                                       List[Char]](shows_rule[A,
                       B]((a: A) =>
                            (b: List[Char]) => shows_prec[A](Zero_nat, a, b),
                           (a: B) =>
                             (b: List[Char]) => shows_prec[B](Zero_nat, a, b),
                           List(' ', '-', '>', ' '), rule),
            comp[List[Char], List[Char],
                  List[Char]](shows_string.apply(List(' ', 'a', 't', ' ', 'p',
               'o', 's', 'i', 't', 'i', 'o', 'n', ' ')),
                               comp[List[Char], List[Char],
                                     List[Char]](shows_prec_pos(Zero_nat, p),
          shows_nl)))))))))
                                    }
                                  }))

def check_prop_rstepa[A : equal : show,
                       B : equal : show](pa:
   (term[A, B]) => sum[(List[Char]) => List[Char], Unit],
  r: List[(term[A, B], term[A, B])], p: pos, rule: (term[A, B], term[A, B]),
  s: term[A, B], t: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  bindb[(List[Char]) => List[Char], Unit,
         Unit](check[(List[Char]) =>
                       List[Char]](membera[(term[A, B], term[A, B])](r, rule),
                                    comp[List[Char], List[Char],
  List[Char]](shows_rule[A, B]((a: A) =>
                                 (b: List[Char]) =>
                                   shows_prec[A](Zero_nat, a, b),
                                (a: B) =>
                                  (b: List[Char]) =>
                                    shows_prec[B](Zero_nat, a, b),
                                List(' ', '-', '>', ' '), rule),
               comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List(' ', 'i', 's', ' ',
                  'n', 'o', 't', ' ', 'a', ' ', 'r', 'u', 'l', 'e', ' ', 'o',
                  'f')),
                                  comp[List[Char], List[Char],
List[Char]](shows_nl,
             comp[List[Char], List[Char],
                   List[Char]](shows_trs[A,
  B]((a: A) => (b: List[Char]) => shows_prec[A](Zero_nat, a, b),
      (a: B) => (b: List[Char]) => shows_prec[B](Zero_nat, a, b),
      List('r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'y', 's', 't', 'e', 'm',
            ':'),
      List(' ', '-', '>', ' '), r),
                                shows_nl))))),
                (_: Unit) => check_prop_rstep_rule[A, B](pa, p, rule, s, t))

def check_rstep[A : equal : show, B : equal : show]:
      (List[(term[A, B], term[A, B])]) =>
        pos =>
          ((term[A, B], term[A, B])) =>
            (term[A, B]) =>
              (term[A, B]) => sum[(List[Char]) => List[Char], Unit] =
  (a: List[(term[A, B], term[A, B])]) =>
    (b: pos) =>
      (c: (term[A, B], term[A, B])) =>
        (d: term[A, B]) =>
          (e: term[A, B]) =>
            check_prop_rstepa[A, B]((_: term[A, B]) =>
                                      Inr[Unit, (List[Char]) => List[Char]](()),
                                     a, b, c, d, e)

def show_crit_pair[A : show, B : show](lr: (term[A, B], term[A, B])):
      (List[Char]) => List[Char] =
  comp[List[Char], List[Char],
        List[Char]](shows_prec_list[Char](Zero_nat, List('(')),
                     comp[List[Char], List[Char],
                           List[Char]]((shows_prec_term[A,
                 B](Zero_nat)).apply(fst[term[A, B], term[A, B]](lr)),
comp[List[Char], List[Char],
      List[Char]](shows_prec_list[Char](Zero_nat, List(',', ' ')),
                   comp[List[Char], List[Char],
                         List[Char]]((shows_prec_term[A,
               B](Zero_nat)).apply(snd[term[A, B], term[A, B]](lr)),
                                      shows_prec_list[Char](Zero_nat,
                     List(')'))))))

def critical_pairs_top_impl[A : equal](p:
 List[(term[A, List[Char]], term[A, List[Char]])],
r: List[(term[A, List[Char]], term[A, List[Char]])]):
      List[(term[A, List[Char]], term[A, List[Char]])] =
  maps[(term[A, List[Char]], term[A, List[Char]]),
        (term[A, List[Char]],
          term[A, List[Char]])]((a: (term[A, List[Char]], term[A, List[Char]]))
                                  => {
                                       val
 (l, ra): (term[A, List[Char]], term[A, List[Char]]) = a;
                                       (if (is_Var[A, List[Char]](l)) Nil
 else maps[(term[A, List[Char]], term[A, List[Char]]),
            (term[A, List[Char]],
              term[A, List[Char]])]((aa: (term[A, List[Char]],
   term[A, List[Char]]))
                                      => {
   val (la, rb): (term[A, List[Char]], term[A, List[Char]]) = aa;
   mgu_var_disjoint_generic[List[Char], List[Char], List[Char],
                             A]((ab: List[Char]) => 'x' :: ab,
                                 (ab: List[Char]) => 'y' :: ab, l, la)
     match {
     case None => Nil
     case Some((sigma, tau)) =>
       List((subst_apply_term[A, List[Char], List[Char]](ra, sigma),
              subst_apply_term[A, List[Char], List[Char]](rb, tau)))
   }
 },
                                     r))
                                     },
                                 p)

def check_critical_pairs_innermost[A : equal : show](r:
               List[(term[A, List[Char]], term[A, List[Char]])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[((term[A, List[Char]], term[A, List[Char]]),
           (List[Char]) => List[Char]),
          Unit,
          (List[Char]) =>
            List[Char]](forallm[(term[A, List[Char]], term[A, List[Char]]),
                                 (List[Char]) =>
                                   List[Char]]((a:
          (term[A, List[Char]], term[A, List[Char]]))
         => {
              val (l, ra): (term[A, List[Char]], term[A, List[Char]]) = a;
              check[(List[Char]) =>
                      List[Char]](equal_terma[A, List[Char]](l, ra),
                                   comp[List[Char], List[Char],
 List[Char]](shows_prec_list[Char](Zero_nat,
                                    List('t', 'h', 'e', 'r', 'e', ' ', 'i', 's',
  ' ', 'a', ' ', 'n', 'o', 'n', '-', 't', 'r', 'i', 'v', 'i', 'a', 'l', ' ',
  'c', 'r', 'i', 't', 'i', 'c', 'a', 'l', ' ', 'p', 'a', 'i', 'r', ' ')),
              show_crit_pair[A, List[Char]]((l, ra))))
            },
        critical_pairs_top_impl[A](r, r)),
                         (x: ((term[A, List[Char]], term[A, List[Char]]),
                               (List[Char]) => List[Char]))
                           => Inl[(List[Char]) => List[Char],
                                   Unit](snd[(term[A, List[Char]],
       term[A, List[Char]]),
      (List[Char]) => List[Char]](x)))

def is_ur_closed_term_impl[A : equal, B,
                            C : equal](nfq: (term[A, List[Char]]) => Boolean,
e_cap: (term[A, B]) => term[A, sum[Unit, List[Char]]],
r: List[(term[A, List[Char]], C)], u: List[(term[A, List[Char]], C)],
s: List[term[A, List[Char]]], x5: term[A, B]):
      Boolean =
  (nfq, e_cap, r, u, s, x5) match {
  case (nfq, e_cap, r, u, s, Var(x)) => true
  case (nfq, e_cap, r, u, s, Fun(f, ts)) =>
    list_all[term[A, B]]((a: term[A, B]) =>
                           is_ur_closed_term_impl[A, B,
           C](nfq, e_cap, r, u, s, a),
                          ts) &&
      list_all[(term[A, List[Char]],
                 C)]((a: (term[A, List[Char]], C)) =>
                       {
                         val (l, ra): (term[A, List[Char]], C) = a;
                         membera[(term[A, List[Char]], C)](u, (l, ra)) ||
                           mgu_class[A](Fun[A,
     sum[Unit,
          List[Char]]](f, mapa[term[A, B],
                                term[A, sum[Unit, List[Char]]]](e_cap, ts)),
 l)
                             match {
                             case None => true
                             case Some(mu) =>
                               ! (list_all[term[A,
         List[Char]]]((ua: term[A, List[Char]]) =>
                        nfq(subst_apply_term[A, List[Char],
      List[Char]](map_vars[List[Char], List[Char],
                            A]((aa: List[Char]) => 'y' :: aa, ua),
                   mu)),
                       args[A, List[Char]](l))) ||
                                 ! (list_all[term[A,
           List[Char]]]((ua: term[A, List[Char]]) =>
                          nfq(subst_apply_term[A, List[Char],
        List[Char]](ua, mu)),
                         s))
                           }
                       },
                      r)
}

def is_ur_closed_impl_dpp_mv[A, B : equal : linorder : show](i:
                       dpp_ops_ext[A, B, List[Char], Unit],
                      d: A):
      (List[(term[B, List[Char]], term[B, List[Char]])]) =>
        (List[term[B, List[Char]]]) => (term[B, List[Char]]) => Boolean =
  {
    val ic: (List[term[B, List[Char]]]) =>
              (term[B, List[Char]]) => term[B, sum[Unit, List[Char]]]
      = icap_impl_dpp[A, B](i, d)
    val qnf: (term[B, List[Char]]) => Boolean =
      (is_QNFc[A, B, List[Char], Unit](i)).apply(d)
    val r: List[(term[B, List[Char]], term[B, List[Char]])] =
      (rulesd[A, B, List[Char], Unit](i)).apply(d)
    val urc: (List[term[B, List[Char]]]) =>
               (List[(term[B, List[Char]], term[B, List[Char]])]) =>
                 (List[term[B, List[Char]]]) => (term[B, List[Char]]) => Boolean
      = (s: List[term[B, List[Char]]]) =>
          (a: List[(term[B, List[Char]], term[B, List[Char]])]) =>
            (b: List[term[B, List[Char]]]) =>
              (c: term[B, List[Char]]) =>
                is_ur_closed_term_impl[B, List[Char],
term[B, List[Char]]](qnf, ic(s), r, a, b, c);
    (u: List[(term[B, List[Char]], term[B, List[Char]])]) =>
      (s: List[term[B, List[Char]]]) =>
        {
          val sa: List[term[B, List[Char]]] =
            mapa[term[B, List[Char]],
                  term[B, List[Char]]]((a: term[B, List[Char]]) =>
 map_vars[List[Char], List[Char], B]((aa: List[Char]) => 'x' :: aa, a),
s);
          (t: term[B, List[Char]]) =>
            (((urc(sa))(u))(sa))(map_vars[List[Char], List[Char],
   B]((a: List[Char]) => 'x' :: a, t))
        }
  }

def rewriting_proc[A, B : equal : linorder : show](i:
             dpp_ops_ext[A, B, List[Char], Unit],
            u: List[(term[B, List[Char]], term[B, List[Char]])],
            sta: (term[B, List[Char]], term[B, List[Char]]),
            st: (term[B, List[Char]], term[B, List[Char]]),
            lr: (term[B, List[Char]], term[B, List[Char]]), p: pos, dpp: A):
      sum[(List[Char]) => List[Char], A] =
  bindb[(List[Char]) => List[Char], Unit,
         Unit](check[(List[Char]) =>
                       List[Char]]((nFQ_subset_NF_rulesb[A, B, List[Char],
                  Unit](i)).apply(dpp),
                                    shows_prec_list[Char](Zero_nat,
                   List('i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', ' ', 'r',
                         'e', 'w', 'r', 'i', 't', 'i', 'n', 'g', ' ', 'r', 'e',
                         'q', 'u', 'i', 'r', 'e', 'd'))),
                (_: Unit) =>
                  {
                    val r: List[(term[B, List[Char]], term[B, List[Char]])] =
                      (rulesd[A, B, List[Char], Unit](i)).apply(dpp)
                    val s: term[B, List[Char]] =
                      fst[term[B, List[Char]], term[B, List[Char]]](sta)
                    val t: term[B, List[Char]] =
                      snd[term[B, List[Char]], term[B, List[Char]]](sta)
                    val ta: term[B, List[Char]] =
                      snd[term[B, List[Char]], term[B, List[Char]]](st);
                    bindb[(List[Char]) => List[Char], Unit,
                           Unit](check[(List[Char]) =>
 List[Char]](equal_terma[B, List[Char]](s,
 fst[term[B, List[Char]], term[B, List[Char]]](st)),
              shows_prec_list[Char](Zero_nat,
                                     List('l', 'e', 'f', 't', '-', 'h', 'a',
   'n', 'd', ' ', 's', 'i', 'd', 'e', 's', ' ', 'o', 'f', ' ', 'o', 'l', 'd',
   ' ', 'a', 'n', 'd', ' ', 'n', 'e', 'w', ' ', 'p', 'a', 'i', 'r', ' ', 'd',
   'i', 'f', 'f', 'e', 'r'))),
                                  (_: Unit) =>
                                    bindb[(List[Char]) => List[Char], Unit,
   Unit](check_rstep[B, List[Char]].apply(r).apply(p).apply(lr).apply(t).apply(ta),
          (_: Unit) =>
            bindb[(List[Char]) => List[Char], Unit,
                   Unit](catcha[(term[B, List[Char]], term[B, List[Char]]),
                                 Unit,
                                 (List[Char]) =>
                                   List[Char]](check_subseteq[(term[B,
                             List[Char]],
                        term[B, List[Char]])](u, r),
        (x: (term[B, List[Char]], term[B, List[Char]])) =>
          Inl[(List[Char]) => List[Char],
               Unit](comp[List[Char], List[Char],
                           List[Char]](shows_rule[B,
           List[Char]]((a: B) =>
                         (b: List[Char]) => shows_prec[B](Zero_nat, a, b),
                        (a: List[Char]) => shows_prec_list[Char](Zero_nat, a),
                        List(' ', '-', '>', ' '), x),
shows_prec_list[Char](Zero_nat,
                       List(' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ',
                             'r', 'u', 'l', 'e', ' ', 'o', 'f', ' ', 't', 'h',
                             'e', ' ', 'r', 'e', 'w', 'r', 'i', 't', 'e', ' ',
                             's', 'y', 's', 't', 'e', 'm', ' '))))),
                          (_: Unit) =>
                            bindb[(List[Char]) => List[Char], Unit,
                                   Unit](check[(List[Char]) =>
         List[Char]](membera[(term[B, List[Char]],
                               term[B, List[Char]])]((pa[A, B, List[Char],
                  Unit](i)).apply(dpp),
              sta) ||
                       nulla[(term[B, List[Char]],
                               term[B, List[Char]])]((rc[A, B, List[Char],
                  Unit](i)).apply(dpp)),
                      shows_prec_list[Char](Zero_nat,
     List('s', 't', 'r', 'i', 'c', 't', ' ', 'D', 'P', ' ', 'o', 'r', ' ', 'n',
           'o', ' ', 's', 't', 'r', 'i', 'c', 't', ' ', 'r', 'u', 'l', 'e', 's',
           ' ', 'r', 'e', 'q', 'u', 'i', 'r', 'e', 'd'))),
  (_: Unit) =>
    {
      val tp: term[B, List[Char]] = subt_at[B, List[Char]](t, p)
      val urc: (List[term[B, List[Char]]]) => (term[B, List[Char]]) => Boolean =
        (is_ur_closed_impl_dpp_mv[A, B](i, dpp)).apply(u)
      val check_urc:
            (List[term[B, List[Char]]]) =>
              (term[B, List[Char]]) => sum[(List[Char]) => List[Char], Unit]
        = (sa: List[term[B, List[Char]]]) =>
            (tb: term[B, List[Char]]) =>
              check[(List[Char]) =>
                      List[Char]]((urc(sa))(tb),
                                   comp[List[Char], List[Char],
 List[Char]](shows_prec_list[Char](Zero_nat, List('t', 'e', 'r', 'm', ' ')),
              comp[List[Char], List[Char],
                    List[Char]]((shows_prec_term[B,
          List[Char]](Zero_nat)).apply(tb),
                                 shows_prec_list[Char](Zero_nat,
                List(' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'c', 'l', 'o', 's',
                      'e', 'd', ' ', 'u', 'n', 'd', 'e', 'r', ' ', 'u', 's',
                      'a', 'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e', 's')))));
      bindb[(List[Char]) => List[Char], Unit,
             Unit](catcha[List[Char], Unit,
                           (List[Char]) =>
                             List[Char]](check_subseteq[List[Char]](vars_term_list[B,
    List[Char]].apply(snd[term[B, List[Char]], term[B, List[Char]]](lr)),
                             vars_term_list[B,
     List[Char]].apply(fst[term[B, List[Char]], term[B, List[Char]]](lr))),
  (_: List[Char]) =>
    Inl[(List[Char]) => List[Char],
         Unit](shows_prec_list[Char](Zero_nat,
                                      List('v', 'a', 'r', 'i', 'a', 'b', 'l',
    'e', ' ', 'c', 'o', 'n', 'd', 'i', 't', 'i', 'o', 'n', ' ', 'i', 'n', ' ',
    'r', 'u', 'l', 'e', ' ', 'v', 'i', 'o', 'l', 'a', 't', 'e', 'd')))),
                    (_: Unit) =>
                      bindb[(List[Char]) => List[Char], Unit,
                             Unit](catcha[List[Char], Unit,
   (List[Char]) =>
     List[Char]](check_subseteq[List[Char]](vars_term_list[B,
                    List[Char]].apply(tp),
     vars_term_list[B, List[Char]].apply(s)),
                  (_: List[Char]) =>
                    Inl[(List[Char]) => List[Char],
                         Unit](shows_prec_list[Char](Zero_nat,
              List('v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ', 'c', 'o', 'n',
                    'd', 'i', 't', 'i', 'o', 'n', ' ', 'i', 'n', ' ', 'p', 'a',
                    'i', 'r', ' ', 'v', 'i', 'o', 'l', 'a', 't', 'e', 'd')))),
                                    (_: Unit) =>
                                      bindb[(List[Char]) => List[Char], Unit,
     Unit]((check_urc(List(s)))(tp),
            (_: Unit) =>
              bindb[(List[Char]) => List[Char], Unit,
                     Unit](catcha[((term[B, List[Char]], term[B, List[Char]]),
                                    (List[Char]) => List[Char]),
                                   Unit,
                                   (List[Char]) =>
                                     List[Char]](forallm[(term[B, List[Char]],
                   term[B, List[Char]]),
                  (List[Char]) =>
                    List[Char]]((a: (term[B, List[Char]], term[B, List[Char]]))
                                  => {
                                       val
 (l, aa): (term[B, List[Char]], term[B, List[Char]]) = a;
                                       (check_urc(args[B, List[Char]](l)))(aa)
                                     },
                                 u),
          (x: ((term[B, List[Char]], term[B, List[Char]]),
                (List[Char]) => List[Char]))
            => Inl[(List[Char]) => List[Char],
                    Unit](snd[(term[B, List[Char]], term[B, List[Char]]),
                               (List[Char]) => List[Char]](x))),
                            (_: Unit) =>
                              bindb[(List[Char]) => List[Char], Unit,
                                     Unit](catcha[(List[Char]) => List[Char],
           Unit,
           (List[Char]) =>
             List[Char]](check_critical_pairs_innermost[B](u),
                          (x: (List[Char]) => List[Char]) =>
                            Inl[(List[Char]) => List[Char],
                                 Unit](comp[List[Char], List[Char],
     List[Char]](shows_prec_list[Char](Zero_nat,
List('p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'i', 'n', ' ', 's', 'h', 'o', 'w',
      'i', 'n', 'g', ' ', 'U', 'N', 'F', ' ', 'o', 'f', ' ', 'u', 's', 'a', 'b',
      'l', 'e', ' ', 'r', 'u', 'l', 'e', 's')),
                  comp[List[Char], List[Char], List[Char]](shows_nl, x)))),
    (_: Unit) =>
      catcha[((Boolean, (term[B, List[Char]], term[B, List[Char]])),
               (List[Char]) => List[Char]),
              Unit,
              (List[Char]) =>
                List[Char]](forallm[(Boolean,
                                      (term[B, List[Char]],
term[B, List[Char]])),
                                     (List[Char]) =>
                                       List[Char]]((a:
              (Boolean, (term[B, List[Char]], term[B, List[Char]])))
             => {
                  val (_, (sa, tb)):
                        (Boolean, (term[B, List[Char]], term[B, List[Char]]))
                    = a;
                  check[(List[Char]) =>
                          List[Char]](equal_terma[B, List[Char]](sa, tb),
                                       shows_prec_list[Char](Zero_nat,
                      List('n', 'o', 'n', '-', 't', 'r', 'i', 'v', 'i', 'a',
                            'l', ' ', 'c', 'r', 'i', 't', 'i', 'c', 'a', 'l',
                            ' ', 'p', 'a', 'i', 'r', ' ', 'b', 'e', 't', 'w',
                            'e', 'e', 'n', ' ', 'r', 'u', 'l', 'e', ' ', 't',
                            'o', ' ', 'r', 'e', 'w', 'r', 'i', 't', 'e', ' ',
                            'a', 'n', 'd', ' ', 'u', 's', 'a', 'b', 'l', 'e',
                            ' ', 'r', 'u', 'l', 'e', 's')))
                },
            critical_pairs_impl[B](List(lr), u)),
                             (x: ((Boolean,
                                    (term[B, List[Char]], term[B, List[Char]])),
                                   (List[Char]) => List[Char]))
                               => Inl[(List[Char]) => List[Char],
                                       Unit](snd[(Boolean,
           (term[B, List[Char]], term[B, List[Char]])),
          (List[Char]) => List[Char]](x))))))))
    }))))
                  })
    match {
    case Inl(a) => Inl[(List[Char]) => List[Char], A](a)
    case Inr(_) =>
      Inr[A, (List[Char]) =>
               List[Char]]((replace_paira[A, B, List[Char],
   Unit](i)).apply(dpp).apply(sta).apply(List(st)))
  }

def uncurry_top_rules[A : equal, B](a: A, n: nat, sm: A => nat => List[A]):
      (List[(term[A, B], term[A, B])]) => List[(term[A, B], term[A, B])] =
  (b: List[(term[A, B], term[A, B])]) =>
    mapa[(term[A, B], term[A, B]),
          (term[A, B],
            term[A, B])]((ba: (term[A, B], term[A, B])) =>
                           {
                             val (l, r): (term[A, B], term[A, B]) = ba;
                             (uncurry_top[A, B](a, n, sm, l),
                               uncurry_top[A, B](a, n, sm, r))
                           },
                          b)

def eta_closed_top_rules[A : equal : show,
                          B : equal : show](a: A, n: nat,
     sm: A => nat => List[A], r: List[(term[A, B], term[A, B])],
     p: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[((term[A, B], term[A, B]), (List[Char]) => List[Char]), Unit,
          (List[Char]) =>
            List[Char]](forallm[(term[A, B], term[A, B]),
                                 (List[Char]) =>
                                   List[Char]]((b: (term[A, B], term[A, B])) =>
         {
           val (l, ra): (term[A, B], term[A, B]) = b;
           l match {
             case Var(_) => Inr[Unit, (List[Char]) => List[Char]](())
             case Fun(ff, ls) =>
               check[(List[Char]) =>
                       List[Char]](equal_nata(aarity[A](sm, ff,
                 size_list[term[A, B]](ls)),
       Zero_nat) ||
                                     list_ex[(term[A, B],
       term[A, B])]((ba: (term[A, B], term[A, B])) =>
                      {
                        val (lll, rrr): (term[A, B], term[A, B]) = ba;
                        (lll, rrr) match {
                          case (Var(_), bb) => false
                          case (Fun(f, Nil), bb) => false
                          case (Fun(f, ll :: yy), Var(_)) => false
                          case (Fun(f, ll :: yy), Fun(g, Nil)) => false
                          case (Fun(f, ll :: yy), Fun(g, rr :: zz)) =>
                            eq[A](f, a) &&
                              (eq[A](g, a) &&
                                (equal_lista[term[A, B]](zz, yy) &&
                                  (equal_nata(size_list[term[A, B]](yy),
       minus_nat(n, one_nat)) &&
                                    (distinct[term[A, B]](yy) &&
                                      (list_all[term[A,
              B]]((aa: term[A, B]) => is_Var[A, B](aa), yy) &&
(nulla[B](list_inter[B](mapa[term[A, B],
                              B]((aa: term[A, B]) => the_Var[A, B](aa), yy),
                         vars_rule_list[A, B]((ll, rr)))) &&
  instance_rule[A, B, B]((l, ra), (ll, rr))))))))
                        }
                      },
                     p),
                                    comp[List[Char], List[Char],
  List[Char]](shows_prec_list[Char](Zero_nat,
                                     List('e', 't', 'a', ' ', 'e', 'x', 'p',
   'a', 'n', 's', 'i', 'o', 'n', ' ', 'o', 'f', ' ')),
               comp[List[Char], List[Char],
                     List[Char]](shows_rule[A,
     B]((aa: A) => (ba: List[Char]) => shows_prec[A](Zero_nat, aa, ba),
         (aa: B) => (ba: List[Char]) => shows_prec[B](Zero_nat, aa, ba),
         List(' ', '-', '>', ' '), (l, ra)),
                                  shows_prec_list[Char](Zero_nat,
                 List(' ', 'm', 'i', 's', 's', 'i', 'n', 'g')))))
           }
         },
        r),
                         (x: ((term[A, B], term[A, B]),
                               (List[Char]) => List[Char]))
                           => Inl[(List[Char]) => List[Char],
                                   Unit](snd[(term[A, B], term[A, B]),
      (List[Char]) => List[Char]](x)))

def uncurry_of_top_sig_list[A](a: A, m: nat, sml: List[((A, nat), List[A])],
                                sm: A => nat => List[A]):
      List[(term[A, List[Char]], term[A, List[Char]])] =
  maps[((A, nat), List[A]),
        (term[A, List[Char]],
          term[A, List[Char]])]((b: ((A, nat), List[A])) =>
                                  {
                                    val (ba, c): ((A, nat), List[A]) = b;
                                    ({
                                       val (f, n): (A, nat) = ba;
                                       (_: List[A]) =>
 {
   val g: nat => A = (aa: nat) => get_symbol[A](sm, f, n, aa);
   mapa[nat, (term[A, List[Char]],
               term[A, List[Char]])]((i: nat) =>
                                       (Fun[A,
     List[Char]](a, generate_f_xs[A](g(i), plus_nat(n, i)) ::
                      mapa[nat, term[A, List[Char]]]((ia: nat) =>
               Var[List[Char], A](generate_var(ia)),
              upt(plus_nat(n, i),
                   plus_nat(plus_nat(n, i), minus_nat(m, one_nat))))),
 generate_f_xs[A](g(Suc(i)), plus_nat(plus_nat(n, i), minus_nat(m, one_nat)))),
                                      upt(Zero_nat, aarity[A](sm, f, n)))
 }
                                     })(c)
                                  },
                                 sml)

def uncurry_top_proc[A, B : equal : show](i:
    dpp_ops_ext[A, B, List[Char], Unit],
   info: (B, (List[((B, nat), List[B])],
               (List[(term[B, List[Char]], term[B, List[Char]])],
                 List[(term[B, List[Char]], term[B, List[Char]])]))),
   n: nat, fmap: (List[((B, nat), List[B])]) => B => nat => B,
   check_inj:
     B => nat =>
            (List[((B, nat), List[B])]) =>
              sum[(List[Char]) => List[Char], Unit],
   p: List[(term[B, List[Char]], term[B, List[Char]])],
   r: List[(term[B, List[Char]], term[B, List[Char]])], dpp: A):
      sum[(List[Char]) => List[Char], A] =
  {
    val (a, (sml, (u, eb))):
          (B, (List[((B, nat), List[B])],
                (List[(term[B, List[Char]], term[B, List[Char]])],
                  List[(term[B, List[Char]], term[B, List[Char]])])))
      = info
    val paa: List[(term[B, List[Char]], term[B, List[Char]])] =
      (pa[A, B, List[Char], Unit](i)).apply(dpp)
    val pw: List[(term[B, List[Char]], term[B, List[Char]])] =
      (pwa[A, B, List[Char], Unit](i)).apply(dpp)
    val ra: List[(term[B, List[Char]], term[B, List[Char]])] =
      (rc[A, B, List[Char], Unit](i)).apply(dpp)
    val rw: List[(term[B, List[Char]], term[B, List[Char]])] =
      (rwc[A, B, List[Char], Unit](i)).apply(dpp)
    val (e, ew):
          (List[(term[B, List[Char]], term[B, List[Char]])],
            List[(term[B, List[Char]], term[B, List[Char]])])
      = uncurry_eta_split[B, List[Char]](eb, ra)
    val sm: B => nat => List[B] = sig_list_to_sig_map[B](a, sml, fmap)
    val p_eta: List[(term[B, List[Char]], term[B, List[Char]])] = e ++ paa
    val pw_eta: List[(term[B, List[Char]], term[B, List[Char]])] = ew ++ pw
    val uP: List[(term[B, List[Char]], term[B, List[Char]])] =
      (uncurry_top_rules[B, List[Char]](a, n, sm)).apply(p_eta)
    val uPw: List[(term[B, List[Char]], term[B, List[Char]])] =
      (uncurry_top_rules[B, List[Char]](a, n, sm)).apply(pw_eta)
    val uR: List[(term[B, List[Char]], term[B, List[Char]])] =
      map_rules_wa[B, B,
                    List[Char]]((f: B) =>
                                  (na: nat) =>
                                    get_symbol[B](sm, f, na, Zero_nat),
                                 ra)
    val uRw: List[(term[B, List[Char]], term[B, List[Char]])] =
      map_rules_wa[B, B,
                    List[Char]]((f: B) =>
                                  (na: nat) =>
                                    get_symbol[B](sm, f, na, Zero_nat),
                                 rw);
    bindb[(List[Char]) => List[Char], Unit,
           Unit](check[(List[Char]) =>
                         List[Char]](nulla[term[B,
         List[Char]]]((qc[A, B, List[Char], Unit](i)).apply(dpp)),
                                      shows_prec_list[Char](Zero_nat,
                     List('s', 't', 'r', 'a', 't', 'e', 'g', 'y', ' ', 'c', 'u',
                           'r', 'r', 'e', 'n', 't', 'l', 'y', ' ', 'u', 'n',
                           's', 'u', 'p', 'p', 'o', 'r', 't', 'e', 'd'))),
                  (_: Unit) =>
                    bindb[(List[Char]) => List[Char], Unit,
                           Unit](check[(List[Char]) =>
 List[Char]](! (equal_nata(n, Zero_nat)),
              shows_prec_list[Char](Zero_nat,
                                     List('t', 'h', 'e', ' ', 'a', 'r', 'i',
   't', 'y', ' ', 'o', 'f', ' ', 't', 'h', 'e', ' ', 'u', 'n', 'c', 'u', 'r',
   'r', 'i', 'e', 'd', ' ', 's', 'y', 'm', 'b', 'o', 'l', ' ', 'm', 'u', 's',
   't', ' ', 'b', 'e', ' ', 'a', 't', ' ', 'l', 'e', 'a', 's', 't', ' ', '1'))),
                                  (_: Unit) =>
                                    bindb[(List[Char]) => List[Char], Unit,
   Unit](((check_inj(a))(n))(sml),
          (_: Unit) =>
            {
              val pb: List[(term[B, List[Char]], term[B, List[Char]])] =
                (pairsa[A, B, List[Char], Unit](i)).apply(dpp)
              val is_def: ((B, nat)) => Boolean =
                (fn: (B, nat)) =>
                  ! (nulla[(term[B, List[Char]],
                             term[B, List[Char]])]((rules_mapb[A, B, List[Char],
                        Unit](i)).apply(dpp).apply(fn)))
              val rm: ((B, nat)) =>
                        List[(term[B, List[Char]], term[B, List[Char]])]
                = (rules_mapb[A, B, List[Char], Unit](i)).apply(dpp);
              bindb[(List[Char]) => List[Char], Unit,
                     Unit](catcha[((term[B, List[Char]], term[B, List[Char]]),
                                    (List[Char]) => List[Char]),
                                   Unit,
                                   (List[Char]) =>
                                     List[Char]](forallm[(term[B, List[Char]],
                   term[B, List[Char]]),
                  (List[Char]) =>
                    List[Char]]((b: (term[B, List[Char]], term[B, List[Char]]))
                                  => {
                                       val
 (l, rb): (term[B, List[Char]], term[B, List[Char]]) = b;
                                       bindb[(List[Char]) => List[Char], Unit,
      Unit](check[(List[Char]) =>
                    List[Char]](hvf_top[B, List[Char]](a, n, l),
                                 comp[List[Char], List[Char],
                                       List[Char]](shows_prec_list[Char](Zero_nat,
                                  List('h', 'e', 'a', 'd', ' ', 'v', 'a', 'r',
'i', 'a', 'b', 'l', 'e', ' ', 'i', 'n', ' ', 'l', 'h', 's', ' ')),
            comp[List[Char], List[Char],
                  List[Char]](shows_term[B,
  List[Char]]((aa: B) => (ba: List[Char]) => shows_prec[B](Zero_nat, aa, ba),
               (aa: List[Char]) => shows_prec_list[Char](Zero_nat, aa), l),
                               shows_prec_list[Char](Zero_nat,
              List(' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e',
                    'd'))))),
             (_: Unit) => check_no_var[B, List[Char]](rb))
                                     },
                                 pw_eta ++ p_eta),
          (x: ((term[B, List[Char]], term[B, List[Char]]),
                (List[Char]) => List[Char]))
            => Inl[(List[Char]) => List[Char],
                    Unit](snd[(term[B, List[Char]], term[B, List[Char]]),
                               (List[Char]) => List[Char]](x))),
                            (_: Unit) =>
                              bindb[(List[Char]) => List[Char], Unit,
                                     Unit](catcha[((term[B, List[Char]],
             term[B, List[Char]]),
            (List[Char]) => List[Char]),
           Unit,
           (List[Char]) =>
             List[Char]](forallm[(term[B, List[Char]], term[B, List[Char]]),
                                  (List[Char]) =>
                                    List[Char]]((aa:
           (term[B, List[Char]], term[B, List[Char]]))
          => {
               val (_, rb): (term[B, List[Char]], term[B, List[Char]]) = aa;
               check[(List[Char]) =>
                       List[Char]](! (is_def(the[(B,
           nat)](root[B, List[Char]](rb)))),
                                    comp[List[Char], List[Char],
  List[Char]](shows_prec_list[Char](Zero_nat,
                                     List('r', 'o', 'o', 't', ' ', 'o', 'f',
   ' ')),
               comp[List[Char], List[Char],
                     List[Char]](shows_term[B,
     List[Char]]((ab: B) => (b: List[Char]) => shows_prec[B](Zero_nat, ab, b),
                  (ab: List[Char]) => shows_prec_list[Char](Zero_nat, ab), rb),
                                  shows_prec_list[Char](Zero_nat,
                 List(' ', 'm', 'u', 's', 't', ' ', 'n', 'o', 't', ' ', 'b',
                       'e', ' ', 'd', 'e', 'f', 'i', 'n', 'e', 'd')))))
             },
         pb),
                          (x: ((term[B, List[Char]], term[B, List[Char]]),
                                (List[Char]) => List[Char]))
                            => Inl[(List[Char]) => List[Char],
                                    Unit](snd[(term[B, List[Char]],
        term[B, List[Char]]),
       (List[Char]) => List[Char]](x))),
    (_: Unit) =>
      bindb[(List[Char]) => List[Char], Unit,
             Unit](check[(List[Char]) =>
                           List[Char]](! (is_def((a, n))),
comp[List[Char], List[Char],
      List[Char]](shows_prec_list[Char](Zero_nat,
 List('a', 'p', 'p', 'l', 'i', 'c', 'a', 't', 'i', 'o', 'n', ' ', 's', 'y', 'm',
       'b', 'o', 'l', ' ')),
                   comp[List[Char], List[Char],
                         List[Char]]((b: List[Char]) =>
                                       shows_prec[B](Zero_nat, a, b),
                                      shows_prec_list[Char](Zero_nat,
                     List(' ', 'm', 'u', 's', 't', ' ', 'n', 'o', 't', ' ', 'b',
                           'e', ' ', 'd', 'e', 'f', 'i', 'n', 'e', 'd', ' ',
                           'i', 'n', ' ', 'R'))))),
                    (_: Unit) =>
                      bindb[(List[Char]) => List[Char], Unit,
                             Unit]((if (list_ex[(term[B, List[Char]],
          term[B, List[Char]])]((b: (term[B, List[Char]], term[B, List[Char]]))
                                  => {
                                       val
 (_, rb): (term[B, List[Char]], term[B, List[Char]]) = b;
                                       equal_proda[B,
            nat](the[(B, nat)](root[B, List[Char]](rb)), (a, n)) &&
 equal_gctxta[B, List[Char]](capRM2[B, List[Char]](rm,
            hd[term[B, List[Char]]](args[B, List[Char]](rb))),
                              GCHole)
                                     },
                                 pb))
                                     bindb[(List[Char]) => List[Char], Unit,
    Unit](catcha[(term[B, List[Char]], term[B, List[Char]]), Unit,
                  (List[Char]) =>
                    List[Char]](check_CS_subseteq[B,
           List[Char]](uncurry_of_top_sig_list[B](a, n, sml, sm), u),
                                 (x: (term[B, List[Char]], term[B, List[Char]]))
                                   => Inl[(List[Char]) => List[Char],
   Unit](comp[List[Char], List[Char],
               List[Char]](shows_prec_list[Char](Zero_nat,
          List('u', 'n', 'c', 'u', 'r', 'r', 'y', 'i', 'n', 'g', ' ', 'p', 'a',
                'i', 'r', ' ')),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_rule[B,
                  List[Char]]((aa: B) =>
                                (b: List[Char]) =>
                                  shows_prec[B](Zero_nat, aa, b),
                               (aa: List[Char]) =>
                                 shows_prec_list[Char](Zero_nat, aa),
                               List(' ', '-', '>', ' '), x),
       comp[List[Char], List[Char],
             List[Char]](shows_prec_list[Char](Zero_nat,
        List(' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g', ' ', 'i',
              'n')),
                          comp[List[Char], List[Char],
                                List[Char]](shows_nl,
     shows_rules[B, List[Char]]((aa: B) =>
                                  (b: List[Char]) =>
                                    shows_prec[B](Zero_nat, aa, b),
                                 (aa: List[Char]) =>
                                   shows_prec_list[Char](Zero_nat, aa),
                                 List(' ', '-', '>', ' '), u))))))),
           (_: Unit) =>
             bindb[(List[Char]) => List[Char], Unit,
                    Unit](eta_closed_top_rules[B,
        List[Char]](a, n, sm, ra, p_eta),
                           (_: Unit) =>
                             eta_closed_top_rules[B,
           List[Char]](a, n, sm, rw, pw_eta)))
                                     else Inr[Unit,
       (List[Char]) => List[Char]](())),
                                    (_: Unit) =>
                                      bindb[(List[Char]) => List[Char], Unit,
     Unit](catcha[(term[B, List[Char]], term[B, List[Char]]), Unit,
                   (List[Char]) =>
                     List[Char]](check_subseteq[(term[B, List[Char]],
          term[B, List[Char]])](uP, p),
                                  (x: (term[B, List[Char]],
term[B, List[Char]]))
                                    => Inl[(List[Char]) => List[Char],
    Unit](comp[List[Char], List[Char],
                List[Char]](shows_prec_list[Char](Zero_nat,
           List('u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', ' ', 'p', 'a', 'i',
                 'r', ' ')),
                             comp[List[Char], List[Char],
                                   List[Char]](shows_rule[B,
                   List[Char]]((aa: B) =>
                                 (b: List[Char]) =>
                                   shows_prec[B](Zero_nat, aa, b),
                                (aa: List[Char]) =>
                                  shows_prec_list[Char](Zero_nat, aa),
                                List(' ', '-', '>', ' '), x),
        shows_prec_list[Char](Zero_nat,
                               List(' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i',
                                     'n', 'g')))))),
            (_: Unit) =>
              bindb[(List[Char]) => List[Char], Unit,
                     Unit](catcha[(term[B, List[Char]], term[B, List[Char]]),
                                   Unit,
                                   (List[Char]) =>
                                     List[Char]](check_subseteq[(term[B,
                               List[Char]],
                          term[B, List[Char]])](uPw, p),
          (x: (term[B, List[Char]], term[B, List[Char]])) =>
            Inl[(List[Char]) => List[Char],
                 Unit](comp[List[Char], List[Char],
                             List[Char]](shows_prec_list[Char](Zero_nat,
                        List('u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', ' ',
                              'p', 'a', 'i', 'r', ' ')),
  comp[List[Char], List[Char],
        List[Char]](shows_rule[B, List[Char]]((aa: B) =>
        (b: List[Char]) => shows_prec[B](Zero_nat, aa, b),
       (aa: List[Char]) => shows_prec_list[Char](Zero_nat, aa),
       List(' ', '-', '>', ' '), x),
                     shows_prec_list[Char](Zero_nat,
    List(' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g')))))),
                            (_: Unit) =>
                              bindb[(List[Char]) => List[Char], Unit,
                                     Unit](catcha[(term[B, List[Char]],
            term[B, List[Char]]),
           Unit,
           (List[Char]) =>
             List[Char]](check_subseteq[(term[B, List[Char]],
  term[B, List[Char]])](u, p),
                          (x: (term[B, List[Char]], term[B, List[Char]])) =>
                            Inl[(List[Char]) => List[Char],
                                 Unit](comp[List[Char], List[Char],
     List[Char]](shows_prec_list[Char](Zero_nat,
List('u', 'n', 'c', 'u', 'r', 'r', 'y', 'i', 'n', 'g', ' ', 'p', 'a', 'i', 'r',
      ' ')),
                  comp[List[Char], List[Char],
                        List[Char]](shows_rule[B,
        List[Char]]((aa: B) =>
                      (b: List[Char]) => shows_prec[B](Zero_nat, aa, b),
                     (aa: List[Char]) => shows_prec_list[Char](Zero_nat, aa),
                     List(' ', '-', '>', ' '), x),
                                     shows_prec_list[Char](Zero_nat,
                    List(' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g',
                          ' ', 'i', 'n', ' ', 'n', 'e', 'w', ' ', 'p', 'a', 'i',
                          'r', 's')))))),
    (_: Unit) =>
      bindb[(List[Char]) => List[Char], Unit,
             Unit](catcha[(term[B, List[Char]], term[B, List[Char]]), Unit,
                           (List[Char]) =>
                             List[Char]](check_subseteq[(term[B, List[Char]],
                  term[B, List[Char]])](uR, r),
  (x: (term[B, List[Char]], term[B, List[Char]])) =>
    Inl[(List[Char]) => List[Char],
         Unit](comp[List[Char], List[Char],
                     List[Char]](shows_prec_list[Char](Zero_nat,
                List('r', 'u', 'l', 'e', ' ')),
                                  comp[List[Char], List[Char],
List[Char]](shows_rule[B, List[Char]]((aa: B) =>
(b: List[Char]) => shows_prec[B](Zero_nat, aa, b),
                                       (aa: List[Char]) =>
 shows_prec_list[Char](Zero_nat, aa),
                                       List(' ', '-', '>', ' '), x),
             shows_prec_list[Char](Zero_nat,
                                    List(' ', 'i', 's', ' ', 'm', 'i', 's', 's',
  'i', 'n', 'g', ' ', 'i', 'n', ' ', 'n', 'e', 'w', ' ', 'r', 'u', 'l', 'e',
  's')))))),
                    (_: Unit) =>
                      catcha[(term[B, List[Char]], term[B, List[Char]]), Unit,
                              (List[Char]) =>
                                List[Char]](check_subseteq[(term[B, List[Char]],
                     term[B, List[Char]])](uRw, r),
     (x: (term[B, List[Char]], term[B, List[Char]])) =>
       Inl[(List[Char]) => List[Char],
            Unit](comp[List[Char], List[Char],
                        List[Char]](shows_prec_list[Char](Zero_nat,
                   List('r', 'u', 'l', 'e', ' ')),
                                     comp[List[Char], List[Char],
   List[Char]](shows_rule[B, List[Char]]((aa: B) =>
   (b: List[Char]) => shows_prec[B](Zero_nat, aa, b),
  (aa: List[Char]) => shows_prec_list[Char](Zero_nat, aa),
  List(' ', '-', '>', ' '), x),
                shows_prec_list[Char](Zero_nat,
                                       List(' ', 'i', 's', ' ', 'm', 'i', 's',
     's', 'i', 'n', 'g', ' ', 'i', 'n', ' ', 'n', 'e', 'w', ' ', 'r', 'u', 'l',
     'e', 's'))))))))))))))
            })))
      match {
      case Inl(aa) => Inl[(List[Char]) => List[Char], A](aa)
      case Inr(_) =>
        Inr[A, (List[Char]) =>
                 List[Char]]((mkc[A, B, List[Char],
                                   Unit](i)).apply(uP).apply(uPw ++
                       u).apply(Nil).apply(uR).apply(uRw))
    }
  }

def uncurry_proc_both[A, B : equal : show,
                       C : equal : show](i:
   dpp_ops_ext[A, lab[B, C], List[Char], Unit],
  x1: Option[nat],
  x2: (lab[B, C],
        (List[((lab[B, C], nat), List[lab[B, C]])],
          (List[(term[lab[B, C], List[Char]], term[lab[B, C], List[Char]])],
            List[(term[lab[B, C], List[Char]],
                   term[lab[B, C], List[Char]])])))):
      (List[(term[lab[B, C], List[Char]], term[lab[B, C], List[Char]])]) =>
        (List[(term[lab[B, C], List[Char]], term[lab[B, C], List[Char]])]) =>
          A => sum[(List[Char]) => List[Char], A] =
  (i, x1, x2) match {
  case (i, None, (a, (sml, (u, eb)))) =>
    (b: List[(term[lab[B, C], List[Char]], term[lab[B, C], List[Char]])]) =>
      (c: List[(term[lab[B, C], List[Char]], term[lab[B, C], List[Char]])]) =>
        (d: A) =>
          uncurry_proc[A, lab[B, C]](i, (a, (sml, (u, eb))),
                                      (ba:
 List[((lab[B, C], nat), List[lab[B, C]])])
=> fmap[B, C](a, nat(BigInt(2)), ba),
                                      (aa: lab[B, C]) =>
(ba: nat) =>
  (ca: List[((lab[B, C], nat), List[lab[B, C]])]) =>
    check_inj[B, C](aa, ba, ca),
                                      b, c, d)
  case (i, Some(n), (a, (sml, (u, eb)))) =>
    (b: List[(term[lab[B, C], List[Char]], term[lab[B, C], List[Char]])]) =>
      (c: List[(term[lab[B, C], List[Char]], term[lab[B, C], List[Char]])]) =>
        (d: A) =>
          uncurry_top_proc[A, lab[B, C]](i, (a, (sml, (u, eb))), n,
  (ba: List[((lab[B, C], nat), List[lab[B, C]])]) => fmap[B, C](a, n, ba),
  (aa: lab[B, C]) =>
    (ba: nat) =>
      (ca: List[((lab[B, C], nat), List[lab[B, C]])]) =>
        check_inj[B, C](aa, ba, ca),
  b, c, d)
}

def check_dpp_subsumes[A, B : equal : show, C : equal : show,
                        D : equal : show](i: dpp_ops_ext[A, lab[B, C], D, Unit],
   x1: (List[(term[lab[B, C], D], term[lab[B, C], D])],
         (List[(term[lab[B, C], D], term[lab[B, C], D])],
           (List[term[lab[B, C], D]],
             (List[(term[lab[B, C], D], term[lab[B, C], D])],
               List[(term[lab[B, C], D], term[lab[B, C], D])])))),
   d: A):
      sum[(List[Char]) => List[Char], Unit] =
  (i, x1, d) match {
  case (i, (p, (pw, (q, (r, rw)))), d) =>
    catcha[(List[Char]) => List[Char], Unit,
            (List[Char]) =>
              List[Char]]({
                            val paa: List[(term[lab[B, C], D],
    term[lab[B, C], D])]
                              = (pa[A, lab[B, C], D, Unit](i)).apply(d)
                            val pwaa: List[(term[lab[B, C], D],
     term[lab[B, C], D])]
                              = (pwa[A, lab[B, C], D, Unit](i)).apply(d)
                            val qa: List[term[lab[B, C], D]] =
                              (qc[A, lab[B, C], D, Unit](i)).apply(d)
                            val ra: List[(term[lab[B, C], D],
   term[lab[B, C], D])]
                              = (rc[A, lab[B, C], D, Unit](i)).apply(d)
                            val rwa: List[(term[lab[B, C], D],
    term[lab[B, C], D])]
                              = (rwc[A, lab[B, C], D, Unit](i)).apply(d)
                            val pb: List[(term[lab[B, C], D],
   term[lab[B, C], D])]
                              = p ++ pw
                            val rb: List[(term[lab[B, C], D],
   term[lab[B, C], D])]
                              = ra ++ rwa
                            val rba: List[(term[lab[B, C], D],
    term[lab[B, C], D])]
                              = r ++ rw;
                            bindb[(List[Char]) => List[Char], Unit,
                                   Unit](catcha[(term[lab[B, C], D],
          term[lab[B, C], D]),
         Unit,
         (List[Char]) =>
           List[Char]](check_subseteq[(term[lab[B, C], D],
term[lab[B, C], D])](paa, p),
                        (x: (term[lab[B, C], D], term[lab[B, C], D])) =>
                          Inl[(List[Char]) => List[Char],
                               Unit](toomuch(List('p', 'a', 'i', 'r'),
      shows_rule[lab[B, C],
                  D]((a: lab[B, C]) => shows_prec_lab[B, C](Zero_nat, a),
                      (a: D) =>
                        (b: List[Char]) => shows_prec[D](Zero_nat, a, b),
                      List(' ', '-', '>', ' '), x)))),
  (_: Unit) =>
    bindb[(List[Char]) => List[Char], Unit,
           Unit](catcha[(term[lab[B, C], D], term[lab[B, C], D]), Unit,
                         (List[Char]) =>
                           List[Char]](check_subseteq[(term[lab[B, C], D],
                term[lab[B, C], D])](pwaa, pb),
(x: (term[lab[B, C], D], term[lab[B, C], D])) =>
  Inl[(List[Char]) => List[Char],
       Unit](toomuch(List('w', 'e', 'a', 'k', ' ', 'p', 'a', 'i', 'r'),
                      shows_rule[lab[B, C],
                                  D]((a: lab[B, C]) =>
                                       shows_prec_lab[B, C](Zero_nat, a),
                                      (a: D) =>
(b: List[Char]) => shows_prec[D](Zero_nat, a, b),
                                      List(' ', '-', '>', ' '), x)))),
                  (_: Unit) =>
                    bindb[(List[Char]) => List[Char], Unit,
                           Unit](catcha[term[lab[B, C], D], Unit,
 (List[Char]) =>
   List[Char]](check_subseteq[term[lab[B, C], D]](qa, q),
                (x: term[lab[B, C], D]) =>
                  Inl[(List[Char]) => List[Char],
                       Unit](toomuch(List('t', 'e', 'r', 'm'),
                                      shows_term[lab[B, C],
          D]((a: lab[B, C]) => shows_prec_lab[B, C](Zero_nat, a),
              (a: D) => (b: List[Char]) => shows_prec[D](Zero_nat, a, b), x)))),
                                  (_: Unit) =>
                                    bindb[(List[Char]) => List[Char], Unit,
   Unit](catcha[term[lab[B, C], D], Unit,
                 (List[Char]) =>
                   List[Char]](check_subseteq[term[lab[B, C], D]](q, qa),
                                (x: term[lab[B, C], D]) =>
                                  Inl[(List[Char]) => List[Char],
                                       Unit](missing(List('t', 'e', 'r', 'm'),
              shows_term[lab[B, C],
                          D]((a: lab[B, C]) =>
                               shows_prec_lab[B, C](Zero_nat, a),
                              (a: D) =>
                                (b: List[Char]) =>
                                  shows_prec[D](Zero_nat, a, b),
                              x)))),
          (_: Unit) =>
            bindb[(List[Char]) => List[Char], Unit,
                   Unit](catcha[(term[lab[B, C], D], term[lab[B, C], D]), Unit,
                                 (List[Char]) =>
                                   List[Char]](check_subseteq[(term[lab[B, C],
                             D],
                        term[lab[B, C], D])](ra, r),
        (x: (term[lab[B, C], D], term[lab[B, C], D])) =>
          Inl[(List[Char]) => List[Char],
               Unit](toomuch(List('s', 't', 'r', 'i', 'c', 't', ' ', 'r', 'u',
                                   'l', 'e'),
                              shows_rule[lab[B, C],
  D]((a: lab[B, C]) => shows_prec_lab[B, C](Zero_nat, a),
      (a: D) => (b: List[Char]) => shows_prec[D](Zero_nat, a, b),
      List(' ', '-', '>', ' '), x)))),
                          (_: Unit) =>
                            bindb[(List[Char]) => List[Char], Unit,
                                   Unit](catcha[(term[lab[B, C], D],
          term[lab[B, C], D]),
         Unit,
         (List[Char]) =>
           List[Char]](check_subseteq[(term[lab[B, C], D],
term[lab[B, C], D])](rb, rba),
                        (x: (term[lab[B, C], D], term[lab[B, C], D])) =>
                          Inl[(List[Char]) => List[Char],
                               Unit](toomuch(List('s', 't', 'r', 'i', 'c', 't',
           '/', 'w', 'e', 'a', 'k', ' ', 'r', 'u', 'l', 'e'),
      shows_rule[lab[B, C],
                  D]((a: lab[B, C]) => shows_prec_lab[B, C](Zero_nat, a),
                      (a: D) =>
                        (b: List[Char]) => shows_prec[D](Zero_nat, a, b),
                      List(' ', '-', '>', ' '), x)))),
  (_: Unit) =>
    bindb[(List[Char]) => List[Char], Unit,
           Unit](catcha[(term[lab[B, C], D], term[lab[B, C], D]), Unit,
                         (List[Char]) =>
                           List[Char]](check_subseteq[(term[lab[B, C], D],
                term[lab[B, C], D])](rba, rb),
(x: (term[lab[B, C], D], term[lab[B, C], D])) =>
  Inl[(List[Char]) => List[Char],
       Unit](missing(List('s', 't', 'r', 'i', 'c', 't', '/', 'w', 'e', 'a', 'k',
                           ' ', 'r', 'u', 'l', 'e'),
                      shows_rule[lab[B, C],
                                  D]((a: lab[B, C]) =>
                                       shows_prec_lab[B, C](Zero_nat, a),
                                      (a: D) =>
(b: List[Char]) => shows_prec[D](Zero_nat, a, b),
                                      List(' ', '-', '>', ' '), x)))),
                  (_: Unit) => Inr[Unit, (List[Char]) => List[Char]](()))))))))
                          },
                           (x: (List[Char]) => List[Char]) =>
                             Inl[(List[Char]) => List[Char],
                                  Unit](comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List('f', 'i', 'n', 'i', 't', 'e', 'n',
   'e', 's', 's', ' ', 'o', 'f', ' ', 't', 'h', 'e', ' ', 'p', 'r', 'o', 'b',
   'l', 'e', 'm')),
                   comp[List[Char], List[Char],
                         List[Char]](shows_nl,
                                      comp[List[Char], List[Char],
    List[Char]](shows_dpp[lab[B, C], D,
                           A]((a: lab[B, C]) =>
                                shows_prec_lab[B, C](Zero_nat, a),
                               (a: D) =>
                                 (b: List[Char]) =>
                                   shows_prec[D](Zero_nat, a, b),
                               i, d),
                 comp[List[Char], List[Char],
                       List[Char]](shows_nl,
                                    comp[List[Char], List[Char],
  List[Char]](shows_string.apply(List('m', 'a', 'y', ' ', 'n', 'o', 't', ' ',
                                       'b', 'e', ' ', 'c', 'o', 'n', 'c', 'l',
                                       'u', 'd', 'e', 'd', ' ', 'f', 'r', 'o',
                                       'm', ' ', 'a', 's', 's', 'u', 'm', 'i',
                                       'n', 'g', ' ', 'f', 'i', 'n', 'i', 't',
                                       'e', 'n', 'e', 's', 's', ' ', 'o', 'f',
                                       ' ', 't', 'h', 'e', ' ', 'p', 'r', 'o',
                                       'b', 'l', 'e', 'm')),
               comp[List[Char], List[Char],
                     List[Char]](shows_nl,
                                  comp[List[Char], List[Char],
List[Char]](shows_dpp[lab[B, C], D,
                       A]((a: lab[B, C]) => shows_prec_lab[B, C](Zero_nat, a),
                           (a: D) =>
                             (b: List[Char]) => shows_prec[D](Zero_nat, a, b),
                           i, (mkc[A, lab[B, C], D,
                                    Unit](i)).apply(p).apply(pw).apply(q).apply(r).apply(rw)),
             comp[List[Char], List[Char],
                   List[Char]](shows_nl,
                                comp[List[Char], List[Char],
                                      List[Char]](x, shows_nl)))))))))))
}

def funas_args_term_list[A, B](t: term[A, B]): List[(A, nat)] =
  maps[term[A, B],
        (A, nat)]((a: term[A, B]) => funas_term_lista[A, B](a), args[A, B](t))

def funas_args_rule_list[A, B](r: (term[A, B], term[A, B])): List[(A, nat)] =
  funas_args_term_list[A, B](fst[term[A, B], term[A, B]](r)) ++
    funas_args_term_list[A, B](snd[term[A, B], term[A, B]](r))

def funas_args_trs_list[A : equal, B](r: List[(term[A, B], term[A, B])]):
      List[(A, nat)] =
  remdups[(A, nat)](maps[(term[A, B], term[A, B]),
                          (A, nat)]((a: (term[A, B], term[A, B])) =>
                                      funas_args_rule_list[A, B](a),
                                     r))

def check_flat_ctxt[A : equal : show,
                     B : equal : show](vs: List[A], c: ctxt[B, A]):
      sum[(List[Char]) => List[Char], Unit] =
  c match {
    case Hole =>
      Inl[(List[Char]) => List[Char],
           Unit](comp[List[Char], List[Char],
                       List[Char]]((shows_prec_ctxt[B, A](Zero_nat)).apply(c),
                                    comp[List[Char], List[Char],
  List[Char]](shows_string.apply(List(' ', 'i', 's', ' ', 'n', 'o', 't', ' ',
                                       'a', ' ', 'f', 'l', 'a', 't', ' ', 'c',
                                       'o', 'n', 't', 'e', 'x', 't')),
               shows_nl)))
    case More(_, ss1, Hole, ss2) =>
      {
        val ss: List[term[B, A]] = ss1 ++ ss2;
        bindb[(List[Char]) => List[Char], Unit,
               Unit](check[(List[Char]) =>
                             List[Char]](distinct[term[B, A]](ss),
  comp[List[Char], List[Char],
        List[Char]]((shows_prec_ctxt[B, A](Zero_nat)).apply(c),
                     comp[List[Char], List[Char],
                           List[Char]](shows_string.apply(List(' ', 'c', 'o',
                        'n', 't', 'a', 'i', 'n', 's', ' ', 'd', 'u', 'p', 'l',
                        'i', 'c', 'a', 't', 'e', ' ', 'v', 'a', 'r', 'i', 'a',
                        'b', 'l', 'e', 's')),
shows_nl))),
                      (_: Unit) =>
                        bindb[(List[Char]) => List[Char], Unit,
                               Unit](check[(List[Char]) =>
     List[Char]](list_all[term[B, A]]((a: term[B, A]) => is_Var[B, A](a), ss),
                  comp[List[Char], List[Char],
                        List[Char]]((shows_prec_ctxt[B, A](Zero_nat)).apply(c),
                                     comp[List[Char], List[Char],
   List[Char]](shows_string.apply(List(' ', 'i', 's', ' ', 'n', 'o', 't', ' ',
'f', 'l', 'a', 't', ',', ' ', 'i', '.', 'e', '.', ',', ' ', 'h', 'a', 's', ' ',
'd', 'e', 'p', 't', 'h', ' ', 'g', 'r', 'e', 'a', 't', 'e', 'r', ' ', 't', 'h',
'a', 'n', ' ', 'o', 'n', 'e')),
                shows_nl))),
                                      (_: Unit) =>
check[(List[Char]) =>
        List[Char]](list_all[term[B, A]]((t: term[B, A]) =>
   ! (membera[A](vs, the_Var[B, A](t))),
  ss1 ++ ss2),
                     comp[List[Char], List[Char],
                           List[Char]]((shows_prec_ctxt[B,
                 A](Zero_nat)).apply(c),
comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List(' ', 'h', 'a', 's', ' ', 't', 'o',
   ' ', 'c', 'o', 'n', 't', 'a', 'i', 'n', ' ', 'o', 'n', 'l', 'y', ' ', 'f',
   'r', 'e', 's', 'h', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', 's')),
                   shows_nl)))))
      }
    case More(_, ss1, More(_, _, _, _), ss2) =>
      Inl[(List[Char]) => List[Char],
           Unit](comp[List[Char], List[Char],
                       List[Char]]((shows_prec_ctxt[B, A](Zero_nat)).apply(c),
                                    comp[List[Char], List[Char],
  List[Char]](shows_string.apply(List(' ', 'i', 's', ' ', 'n', 'o', 't', ' ',
                                       'a', ' ', 'f', 'l', 'a', 't', ' ', 'c',
                                       'o', 'n', 't', 'e', 'x', 't')),
               shows_nl)))
  }

def is_flat_ctxt_list[A : equal,
                       B : equal](vs: List[A], fas: List[(B, nat)],
                                   x2: ctxt[B, A]):
      Boolean =
  (vs, fas, x2) match {
  case (vs, fas, More(f, ss1, Hole, ss2)) =>
    {
      val ss: List[term[B, A]] = ss1 ++ ss2;
      membera[(B, nat)](fas, (f, Suc(size_list[term[B, A]](ss)))) &&
        (list_all[term[B, A]]((a: term[B, A]) => is_Var[B, A](a), ss) &&
          (distinct[term[B, A]](ss) &&
            nulla[A](list_inter[A](mapa[term[B, A],
 A]((a: term[B, A]) => the_Var[B, A](a), ss),
                                    vs))))
    }
  case (vs, fas, Hole) => false
  case (vs, fas, More(v, va, More(vd, ve, vf, vg), vc)) => false
}

def check_is_flat_ctxt[A : equal : show,
                        B : equal : show](vs: List[A], fas: List[(B, nat)],
   c: ctxt[B, A]):
      sum[(List[Char]) => List[Char], Unit] =
  check[(List[Char]) =>
          List[Char]](is_flat_ctxt_list[A, B](vs, fas, c),
                       comp[List[Char], List[Char],
                             List[Char]]((shows_prec_ctxt[B,
                   A](Zero_nat)).apply(c),
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List(' ', 'i', 's', ' ', 'n', 'o', 't',
     ' ', 'a', ' ', 'f', 'l', 'a', 't', ' ', 'c', 'o', 'n', 't', 'e', 'x',
     't')),
                     shows_nl)))

def check_rule_preserving[A : equal : show,
                           B : equal : show](fcs: List[ctxt[A, B]],
      rs: List[(term[A, B], term[A, B])], rule: (term[A, B], term[A, B])):
      sum[(List[Char]) => List[Char], Unit] =
  check[(List[Char]) =>
          List[Char]](membera[(term[A, B], term[A, B])](rs, rule) ||
                        list_all[ctxt[A, B]]((c: ctxt[A, B]) =>
       membera[(term[A, B],
                 term[A, B])](rs, (ctxt_apply[A,
       B](c, fst[term[A, B], term[A, B]](rule)),
                                    ctxt_apply[A,
        B](c, snd[term[A, B], term[A, B]](rule)))),
      fcs),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('t', 'h', 'e',
                          ' ', 'r', 'u', 'l', 'e', ' ')),
  comp[List[Char], List[Char],
        List[Char]](shows_rule[A, B]((a: A) =>
                                       (b: List[Char]) =>
 shows_prec[A](Zero_nat, a, b),
                                      (a: B) =>
(b: List[Char]) => shows_prec[B](Zero_nat, a, b),
                                      List(' ', '-', '>', ' '), rule),
                     comp[List[Char], List[Char],
                           List[Char]](shows_string.apply(List(' ', 'i', 's',
                        ' ', 'n', 'e', 'i', 't', 'h', 'e', 'r', ' ', 'c', 'o',
                        'n', 't', 'a', 'i', 'n', 'e', 'd', ' ', 'i', 'n', ' ',
                        't', 'h', 'e', ' ', 'r', 'e', 's', 'u', 'l', 't', 'i',
                        'n', 'g', ' ', 's', 'e', 't', ' ', 'o', 'f', ' ', 'r',
                        'u', 'l', 'e', 's', ' ', 'n', 'o', 'r', ' ', 'c', 'l',
                        'o', 's', 'e', 'd', ' ', 'u', 'n', 'd', 'e', 'r', ' ',
                        'a', 'l', 'l', ' ', 'f', 'l', 'a', 't', ' ', 'c', 'o',
                        'n', 't', 'e', 'x', 't', 's')),
shows_nl))))

def check_rule_reflecting[A : equal : show,
                           B : equal : show](fcs: List[ctxt[A, B]],
      rs: List[(term[A, B], term[A, B])], rule: (term[A, B], term[A, B])):
      sum[(List[Char]) => List[Char], Unit] =
  check[(List[Char]) =>
          List[Char]](list_ex[(term[A, B],
                                term[A, B])]((a: (term[A, B], term[A, B])) =>
       {
         val (l, r): (term[A, B], term[A, B]) = a;
         list_ex[ctxt[A, B]]((c: ctxt[A, B]) =>
                               equal_terma[A,
    B](fst[term[A, B], term[A, B]](rule), ctxt_apply[A, B](c, l)) &&
                                 equal_terma[A,
      B](snd[term[A, B], term[A, B]](rule), ctxt_apply[A, B](c, r)),
                              Hole :: fcs)
       },
      rs),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('t', 'h', 'e',
                          ' ', 'r', 'u', 'l', 'e', ' ')),
  comp[List[Char], List[Char],
        List[Char]](shows_rule[A, B]((a: A) =>
                                       (b: List[Char]) =>
 shows_prec[A](Zero_nat, a, b),
                                      (a: B) =>
(b: List[Char]) => shows_prec[B](Zero_nat, a, b),
                                      List(' ', '-', '>', ' '), rule),
                     comp[List[Char], List[Char],
                           List[Char]](shows_string.apply(List(' ', 'i', 's',
                        ' ', 'n', 'e', 'i', 't', 'h', 'e', 'r', ' ', 'c', 'o',
                        'n', 't', 'a', 'i', 'n', 'e', 'd', ' ', 'i', 'n', ' ',
                        't', 'h', 'e', ' ', 'o', 'r', 'i', 'g', 'i', 'n', 'a',
                        'l', ' ', 's', 'e', 't', ' ', 'o', 'f', ' ', 'r', 'u',
                        'l', 'e', 's', ' ', 'n', 'o', 'r', ' ', 'o', 'b', 't',
                        'a', 'i', 'n', 'e', 'd', ' ', 'b', 'y', ' ', 'a', 'p',
                        'p', 'l', 'y', 'i', 'n', 'g', ' ', 'a', ' ', 'f', 'l',
                        'a', 't', ' ', 'c', 'o', 'n', 't', 'e', 'x', 't')),
shows_nl))))

def hole_at[A : equal, B](n: nat, i: nat, f: A, x3: ctxt[A, B]): Boolean =
  (n, i, f, x3) match {
  case (n, i, f, More(g, ss1, Hole, ss2)) =>
    eq[A](g, f) &&
      (equal_nata(size_list[term[A, B]](ss1), i) &&
        equal_nata(size_list[term[A, B]](ss2),
                    minus_nat(minus_nat(n, i), one_nat)))
  case (n, i, f, Hole) => false
  case (n, i, f, More(v, va, More(vd, ve, vf, vg), vc)) => false
}

def check_flat_ctxt_complete[A : equal : show,
                              B : show](fcs: List[ctxt[A, B]], fa: (A, nat)):
      sum[(List[Char]) => List[Char], Unit] =
  check[(List[Char]) =>
          List[Char]](all_interval_nat((i: nat) =>
 list_ex[ctxt[A, B]]((a: ctxt[A, B]) =>
                       hole_at[A, B](snd[A, nat](fa), i, fst[A, nat](fa), a),
                      fcs),
Zero_nat, snd[A, nat](fa)),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('t', 'h', 'e',
                          ' ', 'l', 'i', 's', 't', ' ', 'o', 'f', ' ', 'f', 'l',
                          'a', 't', ' ', 'c', 'o', 'n', 't', 'e', 'x', 't', 's',
                          ' ', 'i', 's', ' ', 'i', 'n', 'c', 'o', 'm', 'p', 'l',
                          'e', 't', 'e')),
  shows_nl))

def block_term[A, B](f: A, x1: term[A, B]): term[A, B] = (f, x1) match {
  case (f, Var(x)) => Var[B, A](x)
  case (f, Fun(g, ts)) =>
    Fun[A, B](g, mapa[term[A, B],
                       term[A, B]]((t: term[A, B]) => Fun[A, B](f, List(t)),
                                    ts))
}

def block_rule[A, B](f: A, r: (term[A, B], term[A, B])):
      (term[A, B], term[A, B]) =
  (block_term[A, B](f, fst[term[A, B], term[A, B]](r)),
    block_term[A, B](f, snd[term[A, B], term[A, B]](r)))

def check_superset_of_blocked[A : equal : show,
                               B : equal : show](f: A,
          pa: List[(term[A, B], term[A, B])],
          p: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]](catcha[(term[A, B], term[A, B]), Unit,
                                (List[Char]) =>
                                  List[Char]](catcha[((term[A, B], term[A, B]),
               (term[A, B], term[A, B])),
              Unit,
              (term[A, B],
                term[A, B])](forallm[(term[A, B], term[A, B]),
                                      (term[A, B],
term[A, B])]((x: (term[A, B], term[A, B])) =>
               (if (membera[(term[A, B],
                              term[A, B])](pa, block_rule[A, B](f, x)))
                 Inr[Unit, (term[A, B], term[A, B])](())
                 else Inl[(term[A, B], term[A, B]), Unit](x)),
              p),
                              (x: ((term[A, B], term[A, B]),
                                    (term[A, B], term[A, B])))
                                => Inl[(term[A, B], term[A, B]),
Unit](snd[(term[A, B], term[A, B]), (term[A, B], term[A, B])](x))),
       (x: (term[A, B], term[A, B])) =>
         Inl[(List[Char]) => List[Char],
              Unit](comp[List[Char], List[Char],
                          List[Char]](shows_string.apply(List('t', 'h', 'e',
                       ' ', 'r', 'u', 'l', 'e', ' ')),
                                       comp[List[Char], List[Char],
     List[Char]](shows_rule[A, B]((a: A) =>
                                    (b: List[Char]) =>
                                      shows_prec[A](Zero_nat, a, b),
                                   (a: B) =>
                                     (b: List[Char]) =>
                                       shows_prec[B](Zero_nat, a, b),
                                   List(' ', '-', '>', ' '),
                                   block_rule[A, B](f, x)),
                  comp[List[Char], List[Char],
                        List[Char]](shows_string.apply(List(' ', 'i', 's', ' ',
                     'm', 'i', 's', 's', 'i', 'n', 'g')),
                                     shows_nl))))),
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_trs[A, B]((a: A) =>
                                  (b: List[Char]) =>
                                    shows_prec[A](Zero_nat, a, b),
                                 (a: B) =>
                                   (b: List[Char]) =>
                                     shows_prec[B](Zero_nat, a, b),
                                 List('r', 'e', 'w', 'r', 'i', 't', 'e', ' ',
                                       's', 'y', 's', 't', 'e', 'm', ':'),
                                 List(' ', '-', '>', ' '), p),
                 comp[List[Char], List[Char],
                       List[Char]](shows_string.apply(List('i', 's', ' ', 'n',
                    'o', 't', ' ', 'a', ' ', 's', 'u', 'b', 's', 'e', 't', ' ',
                    'o', 'f')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_trs[A, B]((a: A) =>
                                (b: List[Char]) =>
                                  shows_prec[A](Zero_nat, a, b),
                               (a: B) =>
                                 (b: List[Char]) =>
                                   shows_prec[B](Zero_nat, a, b),
                               List('r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's',
                                     'y', 's', 't', 'e', 'm', ':'),
                               List(' ', '-', '>', ' '), pa),
               comp[List[Char], List[Char], List[Char]](x, shows_nl))))))

def check_no_defined_root_defs[A : equal : show,
                                B : show](f: List[(A, nat)], t: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  check[(List[Char]) =>
          List[Char]](! (membera[(A, nat)](f, the[(A, nat)](root[A, B](t)))),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('t', 'h', 'e',
                          ' ', 'r', 'o', 'o', 't', ' ', 'o', 'f', ' ')),
  comp[List[Char], List[Char],
        List[Char]]((shows_prec_term[A, B](Zero_nat)).apply(t),
                     shows_string.apply(List(' ', 'i', 's', ' ', 'd', 'e', 'f',
      'i', 'n', 'e', 'd')))))

def fcc_proc[A, B : equal : show,
              C : equal : show](i: dpp_ops_ext[A, B, C, Unit], f: B,
                                 fcs: List[ctxt[B, C]],
                                 p: List[(term[B, C], term[B, C])],
                                 pw: List[(term[B, C], term[B, C])],
                                 r: List[(term[B, C], term[B, C])], dpp: A):
      sum[(List[Char]) => List[Char], A] =
  {
    val paa: List[(term[B, C], term[B, C])] = (pa[A, B, C, Unit](i)).apply(dpp)
    val pwaa: List[(term[B, C], term[B, C])] =
      (pwa[A, B, C, Unit](i)).apply(dpp)
    val pb: List[(term[B, C], term[B, C])] =
      list_union[(term[B, C], term[B, C])](paa, pwaa)
    val q: List[term[B, C]] = (qc[A, B, C, Unit](i)).apply(dpp)
    val rs: List[(term[B, C], term[B, C])] = (rc[A, B, C, Unit](i)).apply(dpp)
    val ra: List[(term[B, C], term[B, C])] = (rwc[A, B, C, Unit](i)).apply(dpp)
    val fa: (B, nat) = (f, one_nat)
    val cf: ctxt[B, C] = More[B, C](f, Nil, Hole, Nil)
    val fcsa: List[ctxt[B, C]] = cf :: fcs
    val vs: List[C] = vars_trs_list[B, C](ra)
    val fs: List[(B, nat)] =
      list_union[(B, nat)](funas_trs_list[B, C](ra),
                            funas_args_trs_list[B, C](pb))
    val fas: List[(B, nat)] = fa :: fs
    val ds: List[(B, nat)] = defs_list[B, C](ra);
    bindb[(List[Char]) => List[Char], Unit,
           Unit](check[(List[Char]) =>
                         List[Char]](nulla[term[B, C]](q),
                                      comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('Q', ' ', 'i', 's', ' ', 'n', 'o', 't',
 ' ', 'e', 'm', 'p', 't', 'y')),
                 shows_nl)),
                  (_: Unit) =>
                    bindb[(List[Char]) => List[Char], Unit,
                           Unit](check[(List[Char]) =>
 List[Char]](nulla[(term[B, C], term[B, C])](rs),
              comp[List[Char], List[Char],
                    List[Char]](shows_string.apply(List('t', 'h', 'e', ' ', 's',
                 't', 'r', 'i', 'c', 't', ' ', 'r', 'u', 'l', 'e', 's', ' ',
                 'a', 'r', 'e', ' ', 'n', 'o', 't', ' ', 'e', 'm', 'p', 't',
                 'y')),
                                 shows_nl)),
                                  (_: Unit) =>
                                    bindb[(List[Char]) => List[Char], Unit,
   Unit](check[(List[Char]) =>
                 List[Char]](! (membera[(B, nat)](ds, fa)),
                              comp[List[Char], List[Char],
                                    List[Char]]((a: List[Char]) =>
          shows_prec[B](Zero_nat, f, a),
         comp[List[Char], List[Char],
               List[Char]](shows_string.apply(List('i', 's', ' ', 'n', 'o', 't',
            ' ', 'f', 'r', 'e', 's', 'h')),
                            shows_nl))),
          (_: Unit) =>
            bindb[(List[Char]) => List[Char], Unit,
                   Unit](check_left_linear_trs[B, C](ra),
                          (_: Unit) =>
                            bindb[(List[Char]) => List[Char], Unit,
                                   Unit](check_wf_trs[B, C](ra),
  (_: Unit) =>
    bindb[(List[Char]) => List[Char], Unit,
           Unit](catcha[((term[B, C], term[B, C]), (List[Char]) => List[Char]),
                         Unit,
                         (List[Char]) =>
                           List[Char]](forallm[(term[B, C], term[B, C]),
        (List[Char]) =>
          List[Char]]((rb: (term[B, C], term[B, C])) =>
                        bindb[(List[Char]) => List[Char], Unit,
                               Unit](check_no_var[B,
           C](fst[term[B, C], term[B, C]](rb)),
                                      (_: Unit) =>
bindb[(List[Char]) => List[Char], Unit,
       Unit](check_no_var[B, C](snd[term[B, C], term[B, C]](rb)),
              (_: Unit) =>
                check_no_defined_root_defs[B,
    C](ds, snd[term[B, C], term[B, C]](rb)))),
                       pb),
(x: ((term[B, C], term[B, C]), (List[Char]) => List[Char])) =>
  Inl[(List[Char]) => List[Char],
       Unit](snd[(term[B, C], term[B, C]), (List[Char]) => List[Char]](x))),
                  (_: Unit) =>
                    bindb[(List[Char]) => List[Char], Unit,
                           Unit](catcha[(ctxt[B, C],
  (List[Char]) => List[Char]),
 Unit,
 (List[Char]) =>
   List[Char]](forallm[ctxt[B, C],
                        (List[Char]) =>
                          List[Char]]((a: ctxt[B, C]) =>
check_flat_ctxt[C, B](vs, a),
                                       fcsa),
                (x: (ctxt[B, C], (List[Char]) => List[Char])) =>
                  Inl[(List[Char]) => List[Char],
                       Unit](snd[ctxt[B, C], (List[Char]) => List[Char]](x))),
                                  (_: Unit) =>
                                    bindb[(List[Char]) => List[Char], Unit,
   Unit](catcha[(ctxt[B, C], (List[Char]) => List[Char]), Unit,
                 (List[Char]) =>
                   List[Char]](forallm[ctxt[B, C],
(List[Char]) =>
  List[Char]]((a: ctxt[B, C]) => check_is_flat_ctxt[C, B](vs, fas, a), fcsa),
                                (x: (ctxt[B, C], (List[Char]) => List[Char])) =>
                                  Inl[(List[Char]) => List[Char],
                                       Unit](snd[ctxt[B, C],
          (List[Char]) => List[Char]](x))),
          (_: Unit) =>
            bindb[(List[Char]) => List[Char], Unit,
                   Unit](catcha[((B, nat), (List[Char]) => List[Char]), Unit,
                                 (List[Char]) =>
                                   List[Char]](forallm[(B, nat),
                (List[Char]) =>
                  List[Char]]((a: (B, nat)) =>
                                check_flat_ctxt_complete[B, C](fcsa, a),
                               fas),
        (x: ((B, nat), (List[Char]) => List[Char])) =>
          Inl[(List[Char]) => List[Char],
               Unit](snd[(B, nat), (List[Char]) => List[Char]](x))),
                          (_: Unit) =>
                            bindb[(List[Char]) => List[Char], Unit,
                                   Unit](catcha[((term[B, C], term[B, C]),
          (List[Char]) => List[Char]),
         Unit,
         (List[Char]) =>
           List[Char]](forallm[(term[B, C], term[B, C]),
                                (List[Char]) =>
                                  List[Char]]((a: (term[B, C], term[B, C])) =>
        check_rule_preserving[B, C](fcsa, r, a),
       ra),
                        (x: ((term[B, C], term[B, C]),
                              (List[Char]) => List[Char]))
                          => Inl[(List[Char]) => List[Char],
                                  Unit](snd[(term[B, C], term[B, C]),
     (List[Char]) => List[Char]](x))),
  (_: Unit) =>
    bindb[(List[Char]) => List[Char], Unit,
           Unit](catcha[((term[B, C], term[B, C]), (List[Char]) => List[Char]),
                         Unit,
                         (List[Char]) =>
                           List[Char]](forallm[(term[B, C], term[B, C]),
        (List[Char]) =>
          List[Char]]((a: (term[B, C], term[B, C])) =>
                        check_rule_reflecting[B, C](fcsa, ra, a),
                       r),
(x: ((term[B, C], term[B, C]), (List[Char]) => List[Char])) =>
  Inl[(List[Char]) => List[Char],
       Unit](snd[(term[B, C], term[B, C]), (List[Char]) => List[Char]](x))),
                  (_: Unit) =>
                    bindb[(List[Char]) => List[Char], Unit,
                           Unit](check_superset_of_blocked[B, C](f, p, paa),
                                  (_: Unit) =>
                                    check_superset_of_blocked[B,
                       C](f, pw, pwaa)))))))))))))
  }
    match {
    case Inl(a) => Inl[(List[Char]) => List[Char], A](a)
    case Inr(_) =>
      Inr[A, (List[Char]) =>
               List[Char]]((mkc[A, B, C,
                                 Unit](i)).apply(p).apply(pw).apply(Nil).apply(Nil).apply(r))
  }

def redtriple_proc[A, B : show,
                    C : show](i: dpp_ops_ext[A, B, C, Unit],
                               rp: redtriple_ext[B, C, Unit],
                               premove: List[(term[B, C], term[B, C])], dpp: A):
      sum[(List[Char]) => List[Char], A] =
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]]({
                          val (ps, pns):
                                (List[(term[B, C], term[B, C])],
                                  List[(term[B, C], term[B, C])])
                            = (split_pairsa[A, B, C,
     Unit](i)).apply(dpp).apply(premove)
                          val r: List[(term[B, C], term[B, C])] =
                            (rulesd[A, B, C, Unit](i)).apply(dpp);
                          bindb[(List[Char]) => List[Char], Unit,
                                 Unit](valid[B, C, Unit](rp),
(_: Unit) =>
  bindb[(List[Char]) => List[Char], Unit,
         Unit](catcha[(List[Char]) => List[Char], Unit,
                       (List[Char]) =>
                         List[Char]](catcha[((term[B, C], term[B, C]),
      (List[Char]) => List[Char]),
     Unit,
     (List[Char]) =>
       List[Char]](forallm[(term[B, C], term[B, C]),
                            (List[Char]) => List[Char]](ns[B, C, Unit](rp), r),
                    (x: ((term[B, C], term[B, C]), (List[Char]) => List[Char]))
                      => Inl[(List[Char]) => List[Char],
                              Unit](snd[(term[B, C], term[B, C]),
 (List[Char]) => List[Char]](x))),
                                      (x: (List[Char]) => List[Char]) =>
Inl[(List[Char]) => List[Char],
     Unit](comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l',
              'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n',
              't', 'i', 'n', 'g', ' ', 'T', 'R', 'S')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x)))),
                (_: Unit) =>
                  bindb[(List[Char]) => List[Char], Unit,
                         Unit](catcha[(List[Char]) => List[Char], Unit,
                                       (List[Char]) =>
 List[Char]](catcha[((term[B, C], term[B, C]), (List[Char]) => List[Char]),
                     Unit,
                     (List[Char]) =>
                       List[Char]](forallm[(term[B, C], term[B, C]),
    (List[Char]) => List[Char]](nst[B, C, Unit](rp), pns),
                                    (x: ((term[B, C], term[B, C]),
  (List[Char]) => List[Char]))
                                      => Inl[(List[Char]) => List[Char],
      Unit](snd[(term[B, C], term[B, C]), (List[Char]) => List[Char]](x))),
              (x: (List[Char]) => List[Char]) =>
                Inl[(List[Char]) => List[Char],
                     Unit](comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('p', 'r',
                              'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n',
                              ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g',
                              ' ', 'D', 'P', 's')),
      comp[List[Char], List[Char], List[Char]](shows_nl, x)))),
                                (_: Unit) =>
                                  catcha[(List[Char]) => List[Char], Unit,
  (List[Char]) =>
    List[Char]](catcha[((term[B, C], term[B, C]), (List[Char]) => List[Char]),
                        Unit,
                        (List[Char]) =>
                          List[Char]](forallm[(term[B, C], term[B, C]),
       (List[Char]) => List[Char]](s[B, C, Unit](rp), ps),
                                       (x:
  ((term[B, C], term[B, C]), (List[Char]) => List[Char]))
 => Inl[(List[Char]) => List[Char],
         Unit](snd[(term[B, C], term[B, C]), (List[Char]) => List[Char]](x))),
                 (x: (List[Char]) => List[Char]) =>
                   Inl[(List[Char]) => List[Char],
                        Unit](comp[List[Char], List[Char],
                                    List[Char]](shows_string.apply(List('p',
                                 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h',
                                 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n', 't',
                                 'i', 'n', 'g', ' ', 'D', 'P', 's')),
         comp[List[Char], List[Char], List[Char]](shows_nl, x)))))))
                        },
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ', 'n', 'o',
 't', ' ', 'a', 'p', 'p', 'l', 'y', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'd', 'u',
 'c', 't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c', 'e',
 's', 's', 'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h', 'e', ' ', 'f', 'o',
 'l', 'l', 'o', 'w', 'i', 'n', 'g')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_nl,
                                    comp[List[Char], List[Char],
  List[Char]](desca[B, C, Unit](rp),
               comp[List[Char], List[Char], List[Char]](shows_nl, x))))))
    match {
    case Inl(a) => Inl[(List[Char]) => List[Char], A](a)
    case Inr(_) =>
      Inr[A, (List[Char]) =>
               List[Char]]((delete_P_Pwa[A, B, C,
  Unit](i)).apply(dpp).apply(premove).apply(premove))
  }

def check_dep_graph_proc[A, B : equal : linorder : show,
                          C](i: dpp_ops_ext[A, B, List[Char], Unit], dpp: A,
                              dps: List[(Option[C],
  List[(term[B, List[Char]], term[B, List[Char]])])]):
      sum[(List[Char]) => List[Char], Unit] =
  {
    val c: (term[B, List[Char]]) => gctxt[B, List[Char]] =
      capRM_dpp[A, B, List[Char]](i, dpp)
    val rc: (term[B, List[Char]]) => gctxt[B, List[Char]] =
      reverse_capRM_dpp[A, B, List[Char]](i, dpp);
    (nFQ_subset_NF_rulesb[A, B, List[Char], Unit](i)).apply(dpp);
    (is_QNFc[A, B, List[Char], Unit](i)).apply(dpp);
    mapa[(term[B, List[Char]], term[B, List[Char]]),
          term[B, List[Char]]]((a: (term[B, List[Char]], term[B, List[Char]]))
                                 => fst[term[B, List[Char]],
 term[B, List[Char]]](a),
                                (rulesd[A, B, List[Char], Unit](i)).apply(dpp))
    val iedg: ((term[B, List[Char]], term[B, List[Char]])) =>
                (term[B, List[Char]]) => Boolean
      = is_iedg_edge_dpp[A, B](i, dpp)
    val p: List[(term[B, List[Char]], term[B, List[Char]])] =
      (pairsa[A, B, List[Char], Unit](i)).apply(dpp);
    bindb[(List[Char]) => List[Char], Unit,
           Unit](catcha[(term[B, List[Char]], term[B, List[Char]]), Unit,
                         (List[Char]) =>
                           List[Char]](check_subseteq[(term[B, List[Char]],
                term[B, List[Char]])](p, maps[(Option[C],
        List[(term[B, List[Char]], term[B, List[Char]])]),
       (term[B, List[Char]],
         term[B, List[Char]])]((a: (Option[C],
                                     List[(term[B, List[Char]],
    term[B, List[Char]])]))
                                 => snd[Option[C],
 List[(term[B, List[Char]], term[B, List[Char]])]](a),
                                dps)),
(x: (term[B, List[Char]], term[B, List[Char]])) =>
  Inl[(List[Char]) => List[Char],
       Unit](comp[List[Char], List[Char],
                   List[Char]](shows_string.apply(List('D', 'e', 'p', 'e', 'n',
                'd', 'e', 'n', 'c', 'y', ' ', 'P', 'a', 'i', 'r', ' ')),
                                comp[List[Char], List[Char],
                                      List[Char]](shows_rule[B,
                      List[Char]]((a: B) =>
                                    (b: List[Char]) =>
                                      shows_prec[B](Zero_nat, a, b),
                                   (a: List[Char]) =>
                                     shows_prec_list[Char](Zero_nat, a),
                                   List(' ', '-', '>', ' '), x),
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List(' ', 'i', 's', ' ', 'm',
              'i', 's', 's', 'i', 'n', 'g', ' ', 'i', 'n', ' ', 'd', 'e', 'c',
              'o', 'm', 'p', 'o', 's', 'i', 't', 'i', 'o', 'n')),
                              shows_nl))))),
                  (_: Unit) =>
                    catcha[(List[Char]) => List[Char], Unit,
                            (List[Char]) =>
                              List[Char]](check_graph_decomp[((term[B,
                             List[Char]],
                        term[B, List[Char]]),
                       (gctxt[B, List[Char]],
                         (gctxt[B, List[Char]],
                           (term[B, List[Char]]) =>
                             Boolean)))](comp[(term[B, List[Char]],
        term[B, List[Char]]),
       (List[Char]) => List[Char],
       ((term[B, List[Char]], term[B, List[Char]]),
         (gctxt[B, List[Char]],
           (gctxt[B, List[Char]],
             (term[B, List[Char]]) =>
               Boolean)))]((a: (term[B, List[Char]], term[B, List[Char]])) =>
                             shows_prec_prod[term[B, List[Char]],
      term[B, List[Char]]](Zero_nat, a),
                            (a: ((term[B, List[Char]], term[B, List[Char]]),
                                  (gctxt[B, List[Char]],
                                    (gctxt[B, List[Char]],
                                      (term[B, List[Char]]) => Boolean))))
                              => fst[(term[B, List[Char]], term[B, List[Char]]),
                                      (gctxt[B, List[Char]],
(gctxt[B, List[Char]], (term[B, List[Char]]) => Boolean))](a)),
  (a: (((term[B, List[Char]], term[B, List[Char]]),
         (gctxt[B, List[Char]],
           (gctxt[B, List[Char]], (term[B, List[Char]]) => Boolean))),
        ((term[B, List[Char]], term[B, List[Char]]),
          (gctxt[B, List[Char]],
            (gctxt[B, List[Char]], (term[B, List[Char]]) => Boolean)))))
    => {
         val (aa, b):
               (((term[B, List[Char]], term[B, List[Char]]),
                  (gctxt[B, List[Char]],
                    (gctxt[B, List[Char]], (term[B, List[Char]]) => Boolean))),
                 ((term[B, List[Char]], term[B, List[Char]]),
                   (gctxt[B, List[Char]],
                     (gctxt[B, List[Char]], (term[B, List[Char]]) => Boolean))))
           = a;
         ({
            val (ab, ba):
                  ((term[B, List[Char]], term[B, List[Char]]),
                    (gctxt[B, List[Char]],
                      (gctxt[B, List[Char]], (term[B, List[Char]]) => Boolean)))
              = aa;
            ({
               val (_, t): (term[B, List[Char]], term[B, List[Char]]) = ab;
               (ac: (gctxt[B, List[Char]],
                      (gctxt[B, List[Char]], (term[B, List[Char]]) => Boolean)))
                 => {
                      val (_, (ct, ict)):
                            (gctxt[B, List[Char]],
                              (gctxt[B, List[Char]],
                                (term[B, List[Char]]) => Boolean))
                        = ac;
                      (ad: ((term[B, List[Char]], term[B, List[Char]]),
                             (gctxt[B, List[Char]],
                               (gctxt[B, List[Char]],
                                 (term[B, List[Char]]) => Boolean))))
                        => {
                             val (ae, bb):
                                   ((term[B, List[Char]], term[B, List[Char]]),
                                     (gctxt[B, List[Char]],
                                       (gctxt[B, List[Char]],
 (term[B, List[Char]]) => Boolean)))
                               = ad;
                             ({
                                val (u, _):
                                      (term[B, List[Char]], term[B, List[Char]])
                                  = ae;
                                (af: (gctxt[B, List[Char]],
                                       (gctxt[B, List[Char]],
 (term[B, List[Char]]) => Boolean)))
                                  => {
                                       val
 (cu, (_, _)):
   (gctxt[B, List[Char]],
     (gctxt[B, List[Char]], (term[B, List[Char]]) => Boolean))
 = af;
                                       matchc[B, List[Char]](ct, u) &&
 (matchc[B, List[Char]](cu, t) && ict(u))
                                     }
                              })(bb)
                           }
                    }
             })(ba)
          })(b)
       },
  mapa[(Option[C], List[(term[B, List[Char]], term[B, List[Char]])]),
        (Boolean,
          List[((term[B, List[Char]], term[B, List[Char]]),
                 (gctxt[B, List[Char]],
                   (gctxt[B, List[Char]],
                     (term[B, List[Char]]) =>
                       Boolean)))])]((a:
(Option[C], List[(term[B, List[Char]], term[B, List[Char]])]))
                                       => {
    val (real, cs):
          (Option[C], List[(term[B, List[Char]], term[B, List[Char]])])
      = a;
    (! (is_none[C](real)),
      mapa[(term[B, List[Char]], term[B, List[Char]]),
            ((term[B, List[Char]], term[B, List[Char]]),
              (gctxt[B, List[Char]],
                (gctxt[B, List[Char]],
                  (term[B, List[Char]]) =>
                    Boolean)))]((aa: (term[B, List[Char]], term[B, List[Char]]))
                                  => {
                                       val
 (s, t): (term[B, List[Char]], term[B, List[Char]]) = aa;
                                       ((s, t), (rc(s), (c(t), iedg((s, t)))))
                                     },
                                 cs))
  },
                                      dps)),
   (x: (List[Char]) => List[Char]) =>
     Inl[(List[Char]) => List[Char],
          Unit](comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List('o', 'u', 'r', ' ',
                   'e', 's', 't', 'i', 'm', 'a', 't', 'i', 'o', 'n', ' ', '(',
                   'E', 'D', 'G', '*', '*', '*', ' ', '+', ' ', 'I', 'E', 'D',
                   'G', ')', ' ', 'c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't',
                   ' ', 's', 'h', 'o', 'w', ' ', 't', 'h', 'a', 't', ' ', 'y',
                   'o', 'u', ' ', 'h', 'a', 'v', 'e', ' ', 'a', ' ', 'v', 'a',
                   'l', 'i', 'd', ' ', 'd', 'e', 'c', 'o', 'm', 'p', 'o', 's',
                   'i', 't', 'i', 'o', 'n', ' ')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_string.apply(List('d', 'u', 'e', ' ', 't', 'o', ' ', 't',
                                      'h', 'e', ' ', 'f', 'o', 'l', 'l', 'o',
                                      'w', 'i', 'n', 'g', ' ', 'r', 'e', 'a',
                                      's', 'o', 'n')),
              comp[List[Char], List[Char], List[Char]](shows_nl, x))))))
  }

def intersect_pairsa[A, B, C, D](x0: dpp_ops_ext[A, B, C, D]):
      A => (List[(term[B, C], term[B, C])]) => A =
  x0 match {
  case Dpp_ops_ext(dpp, p, pw, pairs, q, r, rw, rules, q_empty,
                    rules_no_left_var, rules_non_collapsing, is_QNF,
                    nFQ_subset_NF_rules, rules_map, reverse_rules_map,
                    intersect_pairs, replace_pair, intersect_rules, delete_P_Pw,
                    delete_R_Rw, split_pairs, split_rules, mk, more)
    => intersect_pairs
}

def dep_graph_proc[A, B : equal : linorder : show,
                    C](i: dpp_ops_ext[A, B, List[Char], Unit], d: A,
                        dps: List[(Option[C],
                                    List[(term[B, List[Char]],
   term[B, List[Char]])])]):
      sum[(List[Char]) => List[Char], List[(C, A)]] =
  check_dep_graph_proc[A, B, C](i, d, dps) match {
    case Inl(a) => Inl[(List[Char]) => List[Char], List[(C, A)]](a)
    case Inr(_) =>
      Inr[List[(C, A)],
           (List[Char]) =>
             List[Char]](map_filter[(Option[C],
                                      List[(term[B, List[Char]],
     term[B, List[Char]])]),
                                     (C, A)]((x:
        (Option[C], List[(term[B, List[Char]], term[B, List[Char]])]))
       => (if (! (is_none[C](fst[Option[C],
                                  List[(term[B, List[Char]],
 term[B, List[Char]])]](x))))
            Some[(C, A)]((the[C](fst[Option[C],
                                      List[(term[B, List[Char]],
     term[B, List[Char]])]](x)),
                           (intersect_pairsa[A, B, List[Char],
      Unit](i)).apply(d).apply(snd[Option[C],
                                    List[(term[B, List[Char]],
   term[B, List[Char]])]](x))))
            else None),
      dps))
  }

def instantiation_proc[A, B : equal : linorder : show](i:
                 dpp_ops_ext[A, B, List[Char], Unit],
                st: (term[B, List[Char]], term[B, List[Char]]),
                sts: List[(term[B, List[Char]], term[B, List[Char]])], dpp: A):
      sum[(List[Char]) => List[Char], A] =
  {
    val ic: (List[term[B, List[Char]]]) =>
              (term[B, List[Char]]) => term[B, sum[Unit, List[Char]]]
      = icap_impl_dpp_mv[A, B](i, dpp)
    val isnf: (term[B, List[Char]]) => Boolean =
      (is_QNFc[A, B, List[Char], Unit](i)).apply(dpp)
    val (s, t): (term[B, List[Char]], term[B, List[Char]]) = st
    val sy: term[B, List[Char]] =
      map_vars[List[Char], List[Char], B]((a: List[Char]) => 'y' :: a, s)
    val ty: term[B, List[Char]] =
      map_vars[List[Char], List[Char], B]((a: List[Char]) => 'y' :: a, t);
    catcha[((term[B, List[Char]], term[B, List[Char]]),
             (List[Char]) => List[Char]),
            Unit,
            (List[Char]) =>
              List[Char]](forallm[(term[B, List[Char]], term[B, List[Char]]),
                                   (List[Char]) =>
                                     List[Char]]((a:
            (term[B, List[Char]], term[B, List[Char]]))
           => {
                val (u, v): (term[B, List[Char]], term[B, List[Char]]) = a;
                mgu_class[B]((ic(List(u)))(v), s) match {
                  case None => Inr[Unit, (List[Char]) => List[Char]](())
                  case Some(mu) =>
                    check[(List[Char]) =>
                            List[Char]](! (isnf(subst_apply_term[B, List[Char],
                          List[Char]](sy, mu))) ||
  (! (isnf(subst_apply_term[B, List[Char],
                             List[Char]](map_vars[List[Char], List[Char],
           B]((aa: List[Char]) => 'x' :: aa, u),
  mu))) ||
    list_ex[(term[B, List[Char]],
              term[B, List[Char]])]((aa: (term[B, List[Char]],
   term[B, List[Char]]))
                                      => instance_rule[B, List[Char],
                List[Char]]((subst_apply_term[B, List[Char],
       List[Char]](sy, mu),
                              subst_apply_term[B, List[Char],
        List[Char]](ty, mu)),
                             aa),
                                     sts)),
 comp[List[Char], List[Char],
       List[Char]](shows_prec_list[Char](Zero_nat,
  List('c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'f', 'i', 'n', 'd',
        ' ', 'i', 'n', 's', 't', 'a', 'n', 'c', 'e', ' ', 'o', 'f', ' ', 'p',
        'a', 'i', 'r', ' ')),
                    comp[List[Char], List[Char],
                          List[Char]](shows_rule[B,
          List[Char]]((aa: B) =>
                        (b: List[Char]) => shows_prec[B](Zero_nat, aa, b),
                       (aa: List[Char]) => shows_prec_list[Char](Zero_nat, aa),
                       List(' ', '-', '>', ' '),
                       (subst_apply_term[B, List[Char], List[Char]](sy, mu),
                         subst_apply_term[B, List[Char], List[Char]](ty, mu))),
                                       comp[List[Char], List[Char],
     List[Char]](shows_nl,
                  comp[List[Char], List[Char],
                        List[Char]](shows_prec_list[Char](Zero_nat,
                   List('w', 'h', 'i', 'c', 'h', ' ', 'r', 'e', 's', 'u', 'l',
                         't', 'e', 'd', ' ', 'f', 'r', 'o', 'm', ' ', 'D', 'P',
                         ' ')),
                                     shows_rule[B,
         List[Char]]((aa: B) =>
                       (b: List[Char]) => shows_prec[B](Zero_nat, aa, b),
                      (aa: List[Char]) => shows_prec_list[Char](Zero_nat, aa),
                      List(' ', '-', '>', ' '), (u, v)))))))
                }
              },
          (pairsa[A, B, List[Char], Unit](i)).apply(dpp)),
                           (x: ((term[B, List[Char]], term[B, List[Char]]),
                                 (List[Char]) => List[Char]))
                             => Inl[(List[Char]) => List[Char],
                                     Unit](snd[(term[B, List[Char]],
         term[B, List[Char]]),
        (List[Char]) => List[Char]](x)))
  }
    match {
    case Inl(a) => Inl[(List[Char]) => List[Char], A](a)
    case Inr(_) =>
      Inr[A, (List[Char]) =>
               List[Char]]((replace_paira[A, B, List[Char],
   Unit](i)).apply(dpp).apply(st).apply(sts))
  }

def mono_redpair_proc[A, B : show,
                       C : show](i: dpp_ops_ext[A, B, C, Unit],
                                  rp: redtriple_ext[B, C, Unit],
                                  premove: List[(term[B, C], term[B, C])],
                                  rremove: List[(term[B, C], term[B, C])],
                                  dpp: A):
      sum[(List[Char]) => List[Char], A] =
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]]({
                          val (ps, pns):
                                (List[(term[B, C], term[B, C])],
                                  List[(term[B, C], term[B, C])])
                            = (split_pairsa[A, B, C,
     Unit](i)).apply(dpp).apply(premove)
                          val (rs, rns):
                                (List[(term[B, C], term[B, C])],
                                  List[(term[B, C], term[B, C])])
                            = (split_rulesc[A, B, C,
     Unit](i)).apply(dpp).apply(rremove);
                          bindb[(List[Char]) => List[Char], Unit,
                                 Unit](valid[B, C, Unit](rp),
(_: Unit) =>
  bindb[(List[Char]) => List[Char], Unit,
         Unit](catcha[(List[Char]) => List[Char], Unit,
                       (List[Char]) =>
                         List[Char]](mono[B, C, Unit](rp),
                                      (x: (List[Char]) => List[Char]) =>
Inl[(List[Char]) => List[Char],
     Unit](comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l',
              'e', 'm', ' ', 'w', 'i', 't', 'h', ' ', 'm', 'o', 'n', 'o', 't',
              'o', 'n', 'i', 'c', 'i', 't', 'y', ' ', 'o', 'f', ' ', 's', 't',
              'r', 'i', 'c', 't', ' ', 'o', 'r', 'd', 'e', 'r')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x)))),
                (_: Unit) =>
                  bindb[(List[Char]) => List[Char], Unit,
                         Unit](catcha[(List[Char]) => List[Char], Unit,
                                       (List[Char]) =>
 List[Char]](catcha[((term[B, C], term[B, C]), (List[Char]) => List[Char]),
                     Unit,
                     (List[Char]) =>
                       List[Char]](forallm[(term[B, C], term[B, C]),
    (List[Char]) => List[Char]](ns[B, C, Unit](rp), rns),
                                    (x: ((term[B, C], term[B, C]),
  (List[Char]) => List[Char]))
                                      => Inl[(List[Char]) => List[Char],
      Unit](snd[(term[B, C], term[B, C]), (List[Char]) => List[Char]](x))),
              (x: (List[Char]) => List[Char]) =>
                Inl[(List[Char]) => List[Char],
                     Unit](comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('p', 'r',
                              'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n',
                              ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g',
                              ' ', 'T', 'R', 'S')),
      comp[List[Char], List[Char], List[Char]](shows_nl, x)))),
                                (_: Unit) =>
                                  bindb[(List[Char]) => List[Char], Unit,
 Unit](catcha[(List[Char]) => List[Char], Unit,
               (List[Char]) =>
                 List[Char]](catcha[((term[B, C], term[B, C]),
                                      (List[Char]) => List[Char]),
                                     Unit,
                                     (List[Char]) =>
                                       List[Char]](forallm[(term[B, C],
                     term[B, C]),
                    (List[Char]) => List[Char]](s[B, C, Unit](rp), rs),
            (x: ((term[B, C], term[B, C]), (List[Char]) => List[Char])) =>
              Inl[(List[Char]) => List[Char],
                   Unit](snd[(term[B, C], term[B, C]),
                              (List[Char]) => List[Char]](x))),
                              (x: (List[Char]) => List[Char]) =>
                                Inl[(List[Char]) => List[Char],
                                     Unit](comp[List[Char], List[Char],
         List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm',
      ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g',
      ' ', 'T', 'R', 'S')),
                      comp[List[Char], List[Char], List[Char]](shows_nl, x)))),
        (_: Unit) =>
          bindb[(List[Char]) => List[Char], Unit,
                 Unit](catcha[(List[Char]) => List[Char], Unit,
                               (List[Char]) =>
                                 List[Char]](catcha[((term[B, C], term[B, C]),
              (List[Char]) => List[Char]),
             Unit,
             (List[Char]) =>
               List[Char]](forallm[(term[B, C], term[B, C]),
                                    (List[Char]) =>
                                      List[Char]](ns[B, C, Unit](rp), pns),
                            (x: ((term[B, C], term[B, C]),
                                  (List[Char]) => List[Char]))
                              => Inl[(List[Char]) => List[Char],
                                      Unit](snd[(term[B, C], term[B, C]),
         (List[Char]) => List[Char]](x))),
      (x: (List[Char]) => List[Char]) =>
        Inl[(List[Char]) => List[Char],
             Unit](comp[List[Char], List[Char],
                         List[Char]](shows_string.apply(List('p', 'r', 'o', 'b',
                      'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r',
                      'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'D', 'P', 's')),
                                      comp[List[Char], List[Char],
    List[Char]](shows_nl, x)))),
                        (_: Unit) =>
                          catcha[(List[Char]) => List[Char], Unit,
                                  (List[Char]) =>
                                    List[Char]](catcha[((term[B, C],
                  term[B, C]),
                 (List[Char]) => List[Char]),
                Unit,
                (List[Char]) =>
                  List[Char]](forallm[(term[B, C], term[B, C]),
                                       (List[Char]) =>
 List[Char]](s[B, C, Unit](rp), ps),
                               (x: ((term[B, C], term[B, C]),
                                     (List[Char]) => List[Char]))
                                 => Inl[(List[Char]) => List[Char],
 Unit](snd[(term[B, C], term[B, C]), (List[Char]) => List[Char]](x))),
         (x: (List[Char]) => List[Char]) =>
           Inl[(List[Char]) => List[Char],
                Unit](comp[List[Char], List[Char],
                            List[Char]](shows_string.apply(List('p', 'r', 'o',
                         'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o',
                         'r', 'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'D', 'P',
                         's')),
 comp[List[Char], List[Char], List[Char]](shows_nl, x)))))))))
                        },
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ', 'n', 'o',
 't', ' ', 'a', 'p', 'p', 'l', 'y', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'd', 'u',
 'c', 't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c', 'e',
 's', 's', 'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h', 'e', ' ', 'f', 'o',
 'l', 'l', 'o', 'w', 'i', 'n', 'g')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_nl,
                                    comp[List[Char], List[Char],
  List[Char]](desca[B, C, Unit](rp),
               comp[List[Char], List[Char], List[Char]](shows_nl, x))))))
    match {
    case Inl(a) => Inl[(List[Char]) => List[Char], A](a)
    case Inr(_) =>
      Inr[A, (List[Char]) =>
               List[Char]]((delete_R_Rwc[A, B, C,
  Unit](i)).apply((delete_P_Pwa[A, B, C,
                                 Unit](i)).apply(dpp).apply(premove).apply(premove)).apply(rremove).apply(rremove))
  }

def create_rseq_map[A : equal,
                     B : equal](rseqs:
                                  List[((term[A, B], term[A, B]),
 List[(pos, ((term[A, B], term[A, B]), term[A, B]))])],
                                 r: (term[A, B], term[A, B])):
      Option[List[(pos, ((term[A, B], term[A, B]), term[A, B]))]] =
  lookupa[(term[A, B], term[A, B]),
           List[(pos, ((term[A, B], term[A, B]), term[A, B]))]](r, rseqs)

def matchCapRMBelow[A : equal,
                     B : equal](rm: ((A, nat)) =>
                                      List[(term[A, B], term[A, B])],
                                 l: term[A, B], x2: term[A, B]):
      Boolean =
  (rm, l, x2) match {
  case (rm, l, Fun(f, ts)) =>
    matchc[A, B](GCFun[A, B](f, mapa[term[A, B],
                                      gctxt[A,
     B]]((a: term[A, B]) => capRM2[A, B](rm, a), ts)),
                  l)
}

def check_ur_closed_term_rm_af[A : equal : show,
                                B : equal : show](uu:
            ((A, nat)) => List[(term[A, B], term[A, B])],
           uv: List[(term[A, B], term[A, B])], uw: ((A, nat)) => nat => Boolean,
           x3: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  (uu, uv, uw, x3) match {
  case (uu, uv, uw, Var(x)) => Inr[Unit, (List[Char]) => List[Char]](())
  case (rm, ur, pi, Fun(f, ts)) =>
    {
      val n: nat = size_list[term[A, B]](ts)
      val pia: nat => Boolean = pi((f, n));
      bindb[(List[Char]) => List[Char], Unit,
             Unit](catcha[((term[A, B], nat), (List[Char]) => List[Char]), Unit,
                           (List[Char]) =>
                             List[Char]](forallm_index[term[A, B],
                (List[Char]) =>
                  List[Char]]((t: term[A, B]) =>
                                (i: nat) =>
                                  (if (pia(i))
                                    check_ur_closed_term_rm_af[A,
                        B](rm, ur, pi, t)
                                    else Inr[Unit,
      (List[Char]) => List[Char]](())),
                               ts),
  (x: ((term[A, B], nat), (List[Char]) => List[Char])) =>
    Inl[(List[Char]) => List[Char],
         Unit](snd[(term[A, B], nat), (List[Char]) => List[Char]](x))),
                    (_: Unit) =>
                      catcha[((term[A, B], term[A, B]),
                               (List[Char]) => List[Char]),
                              Unit,
                              (List[Char]) =>
                                List[Char]](forallm[(term[A, B], term[A, B]),
             (List[Char]) =>
               List[Char]]((lr: (term[A, B], term[A, B])) =>
                             check[(List[Char]) =>
                                     List[Char]](membera[(term[A, B],
                   term[A, B])](ur, lr) ||
           ! (matchCapRMBelow[A, B](rm, fst[term[A, B], term[A, B]](lr),
                                     Fun[A, B](f, ts))),
          comp[List[Char], List[Char],
                List[Char]](shows_string.apply(List('d', 'u', 'e', ' ', 't',
             'o', ' ', 't', 'h', 'e', ' ', 's', 'u', 'b', 't', 'e', 'r', 'm',
             ' ')),
                             comp[List[Char], List[Char],
                                   List[Char]]((shows_prec_term[A,
                         B](Zero_nat)).apply(Fun[A, B](f, ts)),
        comp[List[Char], List[Char],
              List[Char]](shows_string.apply(List(' ', 'o', 'f', ' ', 's', 'o',
           'm', 'e', ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ', 'r', 'h', 's', ',',
           ' ', 'r', 'u', 'l', 'e', ' ')),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_rule[A,
                 B]((a: A) => (b: List[Char]) => shows_prec[A](Zero_nat, a, b),
                     (a: B) => (b: List[Char]) => shows_prec[B](Zero_nat, a, b),
                     List(' ', '-', '>', ' '), lr),
      shows_string.apply(List(' ', 's', 'h', 'o', 'u', 'l', 'd', ' ', 'b', 'e',
                               ' ', 'u', 's', 'a', 'b', 'l', 'e', '.'))))))),
                            rm((f, n))),
     (x: ((term[A, B], term[A, B]), (List[Char]) => List[Char])) =>
       Inl[(List[Char]) => List[Char],
            Unit](snd[(term[A, B], term[A, B]),
                       (List[Char]) => List[Char]](x))))
    }
}

def check_ur_P_closed_rm_af[A : equal : show,
                             B : equal : show](rm:
         ((A, nat)) => List[(term[A, B], term[A, B])],
        ur: List[(term[A, B], term[A, B])], p: List[(term[A, B], term[A, B])],
        pi: ((A, nat)) => nat => Boolean):
      sum[(List[Char]) => List[Char], Unit] =
  bindb[(List[Char]) => List[Char], Unit,
         Unit](catcha[(List[Char]) => List[Char], Unit,
                       (List[Char]) =>
                         List[Char]](catcha[((term[A, B], term[A, B]),
      (List[Char]) => List[Char]),
     Unit,
     (List[Char]) =>
       List[Char]](forallm[(term[A, B], term[A, B]),
                            (List[Char]) =>
                              List[Char]]((lr: (term[A, B], term[A, B])) =>
    check_ur_closed_term_rm_af[A, B](rm, ur, pi,
                                      snd[term[A, B], term[A, B]](lr)),
   ur),
                    (x: ((term[A, B], term[A, B]), (List[Char]) => List[Char]))
                      => Inl[(List[Char]) => List[Char],
                              Unit](snd[(term[A, B], term[A, B]),
 (List[Char]) => List[Char]](x))),
                                      (x: (List[Char]) => List[Char]) =>
Inl[(List[Char]) => List[Char],
     Unit](comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List('e', 'r', 'r', 'o', 'r',
              ' ', 'w', 'h', 'e', 'n', ' ', 'c', 'h', 'e', 'c', 'k', 'i', 'n',
              'g', ' ', 'c', 'l', 'o', 's', 'u', 'r', 'e', ' ', 'p', 'r', 'o',
              'p', 'e', 'r', 't', 'i', 'e', 's', ' ', 'o', 'f', ' ', 'r', 'h',
              's', ' ', 'o', 'f', ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ', 'r',
              'u', 'l', 'e', 's')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x)))),
                (_: Unit) =>
                  catcha[(List[Char]) => List[Char], Unit,
                          (List[Char]) =>
                            List[Char]](catcha[((term[A, B], term[A, B]),
         (List[Char]) => List[Char]),
        Unit,
        (List[Char]) =>
          List[Char]](forallm[(term[A, B], term[A, B]),
                               (List[Char]) =>
                                 List[Char]]((st: (term[A, B], term[A, B])) =>
       check_ur_closed_term_rm_af[A, B](rm, ur, pi,
 snd[term[A, B], term[A, B]](st)),
      p),
                       (x: ((term[A, B], term[A, B]),
                             (List[Char]) => List[Char]))
                         => Inl[(List[Char]) => List[Char],
                                 Unit](snd[(term[A, B], term[A, B]),
    (List[Char]) => List[Char]](x))),
 (x: (List[Char]) => List[Char]) =>
   Inl[(List[Char]) => List[Char],
        Unit](comp[List[Char], List[Char],
                    List[Char]](shows_string.apply(List('e', 'r', 'r', 'o', 'r',
                 ' ', 'w', 'h', 'e', 'n', ' ', 'c', 'h', 'e', 'c', 'k', 'i',
                 'n', 'g', ' ', 'c', 'l', 'o', 's', 'u', 'r', 'e', ' ', 'p',
                 'r', 'o', 'p', 'e', 'r', 't', 'i', 'e', 's', ' ', 'o', 'f',
                 ' ', 'r', 'h', 's', ' ', 'o', 'f', ' ', 'D', 'P', 's')),
                                 comp[List[Char], List[Char],
                                       List[Char]](shows_nl, x)))))

def mono_ur_redpair_proc[A, B : equal : show,
                          C : equal : show](i: dpp_ops_ext[A, B, C, Unit],
     rp: redtriple_ext[B, C, Unit], premove: List[(term[B, C], term[B, C])],
     rremove: List[(term[B, C], term[B, C])],
     ur: List[(term[B, C], term[B, C])], dpp: A):
      sum[(List[Char]) => List[Char], A] =
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]]({
                          val p: List[(term[B, C], term[B, C])] =
                            (pairsa[A, B, C, Unit](i)).apply(dpp)
                          val us: List[(B, nat)] =
                            maps[(term[B, C], term[B, C]),
                                  (B, nat)](comp[term[B, C], List[(B, nat)],
          (term[B, C],
            term[B, C])](funas_term_list[B, C],
                          (a: (term[B, C], term[B, C])) =>
                            snd[term[B, C], term[B, C]](a)),
     p ++ ur)
                          val filt: ((term[B, C], term[B, C])) => Boolean =
                            (lr: (term[B, C], term[B, C])) =>
                              list_all[(B,
 nat)]((a: (B, nat)) => membera[(B, nat)](us, a),
        funas_term_list[B, C].apply(fst[term[B, C], term[B, C]](lr)))
                          val (pms, pns):
                                (List[(term[B, C], term[B, C])],
                                  List[(term[B, C], term[B, C])])
                            = (split_pairsa[A, B, C,
     Unit](i)).apply(dpp).apply(premove)
                          val (ps, pnwf):
                                (List[(term[B, C], term[B, C])],
                                  List[(term[B, C], term[B, C])])
                            = partition[(term[B, C], term[B, C])](filt, pms)
                          val (urms, urns):
                                (List[(term[B, C], term[B, C])],
                                  List[(term[B, C], term[B, C])])
                            = partition[(term[B, C],
  term[B, C])]((a: (term[B, C], term[B, C])) =>
                 membera[(term[B, C], term[B, C])](rremove, a),
                ur)
                          val (urs, urnwf):
                                (List[(term[B, C], term[B, C])],
                                  List[(term[B, C], term[B, C])])
                            = partition[(term[B, C], term[B, C])](filt, urms)
                          val rm: ((B, nat)) => List[(term[B, C], term[B, C])] =
                            (rules_mapb[A, B, C, Unit](i)).apply(dpp);
                          bindb[(List[Char]) => List[Char], Unit,
                                 Unit](valid[B, C, Unit](rp),
(_: Unit) =>
  bindb[(List[Char]) => List[Char], Unit,
         Unit](mono[B, C, Unit](rp),
                (_: Unit) =>
                  bindb[(List[Char]) => List[Char], Unit,
                         Unit](check_ur_P_closed_rm_af[B,
                C](rm, ur, p, (a: (B, nat)) => (b: nat) => full_af[B](a, b)),
                                (_: Unit) =>
                                  bindb[(List[Char]) => List[Char], Unit,
 Unit](catcha[(List[Char]) => List[Char], Unit,
               (List[Char]) =>
                 List[Char]](catcha[((term[B, C], term[B, C]),
                                      (List[Char]) => List[Char]),
                                     Unit,
                                     (List[Char]) =>
                                       List[Char]](forallm[(term[B, C],
                     term[B, C]),
                    (List[Char]) =>
                      List[Char]](ns[B, C, Unit](rp), urns ++ urnwf),
            (x: ((term[B, C], term[B, C]), (List[Char]) => List[Char])) =>
              Inl[(List[Char]) => List[Char],
                   Unit](snd[(term[B, C], term[B, C]),
                              (List[Char]) => List[Char]](x))),
                              (x: (List[Char]) => List[Char]) =>
                                Inl[(List[Char]) => List[Char],
                                     Unit](comp[List[Char], List[Char],
         List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm',
      ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g',
      ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e', 's')),
                      comp[List[Char], List[Char], List[Char]](shows_nl, x)))),
        (_: Unit) =>
          bindb[(List[Char]) => List[Char], Unit,
                 Unit](catcha[(List[Char]) => List[Char], Unit,
                               (List[Char]) =>
                                 List[Char]](catcha[((term[B, C], term[B, C]),
              (List[Char]) => List[Char]),
             Unit,
             (List[Char]) =>
               List[Char]](forallm[(term[B, C], term[B, C]),
                                    (List[Char]) =>
                                      List[Char]](s[B, C, Unit](rp), urs),
                            (x: ((term[B, C], term[B, C]),
                                  (List[Char]) => List[Char]))
                              => Inl[(List[Char]) => List[Char],
                                      Unit](snd[(term[B, C], term[B, C]),
         (List[Char]) => List[Char]](x))),
      (x: (List[Char]) => List[Char]) =>
        Inl[(List[Char]) => List[Char],
             Unit](comp[List[Char], List[Char],
                         List[Char]](shows_string.apply(List('p', 'r', 'o', 'b',
                      'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r',
                      'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'u', 's', 'a',
                      'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e', 's')),
                                      comp[List[Char], List[Char],
    List[Char]](shows_nl, x)))),
                        (_: Unit) =>
                          bindb[(List[Char]) => List[Char], Unit,
                                 Unit](catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](catcha[((term[B, C], term[B, C]),
                              (List[Char]) => List[Char]),
                             Unit,
                             (List[Char]) =>
                               List[Char]](forallm[(term[B, C], term[B, C]),
            (List[Char]) => List[Char]](ns[B, C, Unit](rp), pns ++ pnwf),
    (x: ((term[B, C], term[B, C]), (List[Char]) => List[Char])) =>
      Inl[(List[Char]) => List[Char],
           Unit](snd[(term[B, C], term[B, C]), (List[Char]) => List[Char]](x))),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm', ' ',
                                      'w', 'h', 'e', 'n', ' ', 'o', 'r', 'i',
                                      'e', 'n', 't', 'i', 'n', 'g', ' ', 'D',
                                      'P', 's')),
              comp[List[Char], List[Char], List[Char]](shows_nl, x)))),
(_: Unit) =>
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]](catcha[((term[B, C], term[B, C]),
                                 (List[Char]) => List[Char]),
                                Unit,
                                (List[Char]) =>
                                  List[Char]](forallm[(term[B, C], term[B, C]),
               (List[Char]) => List[Char]](s[B, C, Unit](rp), ps),
       (x: ((term[B, C], term[B, C]), (List[Char]) => List[Char])) =>
         Inl[(List[Char]) => List[Char],
              Unit](snd[(term[B, C], term[B, C]),
                         (List[Char]) => List[Char]](x))),
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm', ' ',
 'w', 'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'D',
 'P', 's')),
                 comp[List[Char], List[Char], List[Char]](shows_nl, x))))))))))
                        },
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ', 'n', 'o',
 't', ' ', 'a', 'p', 'p', 'l', 'y', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'd', 'u',
 'c', 't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c', 'e',
 's', 's', 'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h', 'e', ' ', 'f', 'o',
 'l', 'l', 'o', 'w', 'i', 'n', 'g')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_nl,
                                    comp[List[Char], List[Char],
  List[Char]](desca[B, C, Unit](rp),
               comp[List[Char], List[Char], List[Char]](shows_nl, x))))))
    match {
    case Inl(a) => Inl[(List[Char]) => List[Char], A](a)
    case Inr(_) =>
      Inr[A, (List[Char]) =>
               List[Char]]((delete_R_Rwc[A, B, C,
  Unit](i)).apply((delete_P_Pwa[A, B, C,
                                 Unit](i)).apply(dpp).apply(premove).apply(premove)).apply(rremove).apply(rremove))
  }

def ur_af_redtriple_proc[A, B : equal : show,
                          C : equal : show](i: dpp_ops_ext[A, B, C, Unit],
     rp: redtriple_ext[B, C, Unit], premove: List[(term[B, C], term[B, C])],
     ur: List[(term[B, C], term[B, C])], dpp: A):
      sum[(List[Char]) => List[Char], A] =
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]]({
                          val (ps, pns):
                                (List[(term[B, C], term[B, C])],
                                  List[(term[B, C], term[B, C])])
                            = (split_pairsa[A, B, C,
     Unit](i)).apply(dpp).apply(premove)
                          val p: List[(term[B, C], term[B, C])] =
                            (pairsa[A, B, C, Unit](i)).apply(dpp)
                          val rm: ((B, nat)) => List[(term[B, C], term[B, C])] =
                            (rules_mapb[A, B, C, Unit](i)).apply(dpp);
                          bindb[(List[Char]) => List[Char], Unit,
                                 Unit](valid[B, C, Unit](rp),
(_: Unit) =>
  bindb[(List[Char]) => List[Char], Unit,
         Unit](check_ur_P_closed_rm_af[B, C](rm, ur, p, af[B, C, Unit](rp)),
                (_: Unit) =>
                  bindb[(List[Char]) => List[Char], Unit,
                         Unit](catcha[(List[Char]) => List[Char], Unit,
                                       (List[Char]) =>
 List[Char]](catcha[((term[B, C], term[B, C]), (List[Char]) => List[Char]),
                     Unit,
                     (List[Char]) =>
                       List[Char]](forallm[(term[B, C], term[B, C]),
    (List[Char]) => List[Char]](ns[B, C, Unit](rp), ur),
                                    (x: ((term[B, C], term[B, C]),
  (List[Char]) => List[Char]))
                                      => Inl[(List[Char]) => List[Char],
      Unit](snd[(term[B, C], term[B, C]), (List[Char]) => List[Char]](x))),
              (x: (List[Char]) => List[Char]) =>
                Inl[(List[Char]) => List[Char],
                     Unit](comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('p', 'r',
                              'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n',
                              ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g',
                              ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ', 'r', 'u',
                              'l', 'e', 's')),
      comp[List[Char], List[Char], List[Char]](shows_nl, x)))),
                                (_: Unit) =>
                                  bindb[(List[Char]) => List[Char], Unit,
 Unit](catcha[(List[Char]) => List[Char], Unit,
               (List[Char]) =>
                 List[Char]](catcha[((term[B, C], term[B, C]),
                                      (List[Char]) => List[Char]),
                                     Unit,
                                     (List[Char]) =>
                                       List[Char]](forallm[(term[B, C],
                     term[B, C]),
                    (List[Char]) => List[Char]](nst[B, C, Unit](rp), pns),
            (x: ((term[B, C], term[B, C]), (List[Char]) => List[Char])) =>
              Inl[(List[Char]) => List[Char],
                   Unit](snd[(term[B, C], term[B, C]),
                              (List[Char]) => List[Char]](x))),
                              (x: (List[Char]) => List[Char]) =>
                                Inl[(List[Char]) => List[Char],
                                     Unit](comp[List[Char], List[Char],
         List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm',
      ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g',
      ' ', 'D', 'P', 's')),
                      comp[List[Char], List[Char], List[Char]](shows_nl, x)))),
        (_: Unit) =>
          catcha[(List[Char]) => List[Char], Unit,
                  (List[Char]) =>
                    List[Char]](catcha[((term[B, C], term[B, C]),
 (List[Char]) => List[Char]),
Unit,
(List[Char]) =>
  List[Char]](forallm[(term[B, C], term[B, C]),
                       (List[Char]) => List[Char]](s[B, C, Unit](rp), ps),
               (x: ((term[B, C], term[B, C]), (List[Char]) => List[Char])) =>
                 Inl[(List[Char]) => List[Char],
                      Unit](snd[(term[B, C], term[B, C]),
                                 (List[Char]) => List[Char]](x))),
                                 (x: (List[Char]) => List[Char]) =>
                                   Inl[(List[Char]) => List[Char],
Unit](comp[List[Char], List[Char],
            List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e',
         'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i',
         'n', 'g', ' ', 'D', 'P', 's')),
                         comp[List[Char], List[Char],
                               List[Char]](shows_nl, x))))))))
                        },
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ', 'n', 'o',
 't', ' ', 'a', 'p', 'p', 'l', 'y', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'd', 'u',
 'c', 't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c', 'e',
 's', 's', 'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h', 'e', ' ', 'f', 'o',
 'l', 'l', 'o', 'w', 'i', 'n', 'g')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_nl,
                                    comp[List[Char], List[Char],
  List[Char]](desca[B, C, Unit](rp),
               comp[List[Char], List[Char], List[Char]](shows_nl, x))))))
    match {
    case Inl(a) => Inl[(List[Char]) => List[Char], A](a)
    case Inr(_) =>
      Inr[A, (List[Char]) =>
               List[Char]]((delete_P_Pwa[A, B, C,
  Unit](i)).apply(dpp).apply(premove).apply(premove))
  }

def check_no_defined_root[A : show,
                           B : show](isdef: ((A, nat)) => Boolean,
                                      t: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  check[(List[Char]) =>
          List[Char]](! (isdef(the[(A, nat)](root[A, B](t)))),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('t', 'h', 'e',
                          ' ', 'r', 'o', 'o', 't', ' ', 'o', 'f', ' ')),
  comp[List[Char], List[Char],
        List[Char]]((shows_prec_term[A, B](Zero_nat)).apply(t),
                     shows_string.apply(List(' ', 'i', 's', ' ', 'd', 'e', 'f',
      'i', 'n', 'e', 'd')))))

def root_redtriple_proc[A, B : show,
                         C : show](i: dpp_ops_ext[A, B, C, Unit],
                                    rp: root_redtriple_ext[B, C, Unit],
                                    premove: List[(term[B, C], term[B, C])],
                                    dpp: A):
      sum[(List[Char]) => List[Char], A] =
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]]({
                          val (ps, pns):
                                (List[(term[B, C], term[B, C])],
                                  List[(term[B, C], term[B, C])])
                            = (split_pairsa[A, B, C,
     Unit](i)).apply(dpp).apply(premove)
                          val r: List[(term[B, C], term[B, C])] =
                            (rulesd[A, B, C, Unit](i)).apply(dpp);
                          bindb[(List[Char]) => List[Char], Unit,
                                 Unit](catcha[((term[B, C], term[B, C]),
        (List[Char]) => List[Char]),
       Unit,
       (List[Char]) =>
         List[Char]](forallm[(term[B, C], term[B, C]),
                              (List[Char]) =>
                                List[Char]]((a: (term[B, C], term[B, C])) =>
      {
        val (_, ra): (term[B, C], term[B, C]) = a;
        bindb[(List[Char]) => List[Char], Unit,
               Unit](check_no_var[B, C](ra),
                      (_: Unit) =>
                        check_no_defined_root[B,
       C]((fn: (B, nat)) =>
            ! (nulla[(term[B, C],
                       term[B, C])]((rules_mapb[A, B, C,
         Unit](i)).apply(dpp).apply(fn))),
           ra))
      },
     (pairsa[A, B, C, Unit](i)).apply(dpp)),
                      (x: ((term[B, C], term[B, C]),
                            (List[Char]) => List[Char]))
                        => Inl[(List[Char]) => List[Char],
                                Unit](snd[(term[B, C], term[B, C]),
   (List[Char]) => List[Char]](x))),
(_: Unit) =>
  bindb[(List[Char]) => List[Char], Unit,
         Unit](valida[B, C, Unit](rp),
                (_: Unit) =>
                  bindb[(List[Char]) => List[Char], Unit,
                         Unit](catcha[(List[Char]) => List[Char], Unit,
                                       (List[Char]) =>
 List[Char]](catcha[((term[B, C], term[B, C]), (List[Char]) => List[Char]),
                     Unit,
                     (List[Char]) =>
                       List[Char]](forallm[(term[B, C], term[B, C]),
    (List[Char]) => List[Char]](nsa[B, C, Unit](rp), r),
                                    (x: ((term[B, C], term[B, C]),
  (List[Char]) => List[Char]))
                                      => Inl[(List[Char]) => List[Char],
      Unit](snd[(term[B, C], term[B, C]), (List[Char]) => List[Char]](x))),
              (x: (List[Char]) => List[Char]) =>
                Inl[(List[Char]) => List[Char],
                     Unit](comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('p', 'r',
                              'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n',
                              ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g',
                              ' ', 'T', 'R', 'S')),
      comp[List[Char], List[Char], List[Char]](shows_nl, x)))),
                                (_: Unit) =>
                                  bindb[(List[Char]) => List[Char], Unit,
 Unit](catcha[(List[Char]) => List[Char], Unit,
               (List[Char]) =>
                 List[Char]](catcha[((term[B, C], term[B, C]),
                                      (List[Char]) => List[Char]),
                                     Unit,
                                     (List[Char]) =>
                                       List[Char]](forallm[(term[B, C],
                     term[B, C]),
                    (List[Char]) => List[Char]](nsta[B, C, Unit](rp), pns),
            (x: ((term[B, C], term[B, C]), (List[Char]) => List[Char])) =>
              Inl[(List[Char]) => List[Char],
                   Unit](snd[(term[B, C], term[B, C]),
                              (List[Char]) => List[Char]](x))),
                              (x: (List[Char]) => List[Char]) =>
                                Inl[(List[Char]) => List[Char],
                                     Unit](comp[List[Char], List[Char],
         List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm',
      ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g',
      ' ', 'D', 'P', 's')),
                      comp[List[Char], List[Char], List[Char]](shows_nl, x)))),
        (_: Unit) =>
          catcha[(List[Char]) => List[Char], Unit,
                  (List[Char]) =>
                    List[Char]](catcha[((term[B, C], term[B, C]),
 (List[Char]) => List[Char]),
Unit,
(List[Char]) =>
  List[Char]](forallm[(term[B, C], term[B, C]),
                       (List[Char]) => List[Char]](sa[B, C, Unit](rp), ps),
               (x: ((term[B, C], term[B, C]), (List[Char]) => List[Char])) =>
                 Inl[(List[Char]) => List[Char],
                      Unit](snd[(term[B, C], term[B, C]),
                                 (List[Char]) => List[Char]](x))),
                                 (x: (List[Char]) => List[Char]) =>
                                   Inl[(List[Char]) => List[Char],
Unit](comp[List[Char], List[Char],
            List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e',
         'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i',
         'n', 'g', ' ', 'D', 'P', 's')),
                         comp[List[Char], List[Char],
                               List[Char]](shows_nl, x))))))))
                        },
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ', 'n', 'o',
 't', ' ', 'a', 'p', 'p', 'l', 'y', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'd', 'u',
 'c', 't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c', 'e',
 's', 's', 'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h', 'e', ' ', 'f', 'o',
 'l', 'l', 'o', 'w', 'i', 'n', 'g')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_nl,
                                    comp[List[Char], List[Char],
  List[Char]](descaa[B, C, Unit](rp),
               comp[List[Char], List[Char], List[Char]](shows_nl, x))))))
    match {
    case Inl(a) => Inl[(List[Char]) => List[Char], A](a)
    case Inr(_) =>
      Inr[A, (List[Char]) =>
               List[Char]]((delete_P_Pwa[A, B, C,
  Unit](i)).apply(dpp).apply(premove).apply(premove))
  }

abstract sealed class slm_ops_ext[+A, +B, +C, +D]
final case class
  Slm_ops_ext[A, B, C,
               D](a: A => (List[B]) => C, b: A => (List[B]) => B, c: List[B],
                   d: B, e: A => (List[B]) => C, f: D)
  extends slm_ops_ext[A, B, C, D]

def rl_slm[A : equal,
            B : equal](delt_opt: Option[(lab[A, B], nat)],
                        pre_fs: List[(lab[A, B], nat)],
                        g: List[(lab[A, B], nat)]):
      sum[(List[Char]) => List[Char],
           slm_ops_ext[lab[A, B], lab[A, B], sum[B, List[lab[A, B]]], Unit]] =
  {
    val fs: List[(lab[A, B], nat)] =
      (if (is_none[(lab[A, B], nat)](delt_opt)) pre_fs
        else filter[(lab[A, B],
                      nat)]((f: (lab[A, B], nat)) =>
                              ! (equal_proda[lab[A, B],
      nat](f, the[(lab[A, B], nat)](delt_opt))),
                             pre_fs));
    bindb[(List[Char]) => List[Char], Unit,
           slm_ops_ext[lab[A, B], lab[A, B], sum[B, List[lab[A, B]]],
                        Unit]](check[(List[Char]) =>
                                       List[Char]](! (nulla[(lab[A, B],
                      nat)](fs)),
            comp[List[Char], List[Char],
                  List[Char]](shows_string.apply(List('r', 'o', 'o', 't', '-',
               'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g', ' ', 'r', 'e', 'q', 'u',
               'i', 'r', 'e', 's', ' ', 'a', 't', ' ', 'l', 'e', 'a', 's', 't',
               ' ', 'o', 'n', 'e', ' ', 'f', 'u', 'n', 'c', 't', 'i', 'o', 'n',
               ' ', 's', 'y', 'm', 'b', 'o', 'l', ' ', 'i', 'n', ' ', 't', 'h',
               'e', ' ', 's', 'i', 'g', 'n', 'a', 't', 'u', 'r', 'e')),
                               shows_nl)),
                                (_: Unit) =>
                                  {
                                    val f: lab[A, B] =
                                      fst[lab[A, B],
   nat](hd[(lab[A, B], nat)](fs));
                                    Inr[slm_ops_ext[lab[A, B], lab[A, B],
             sum[B, List[lab[A, B]]], Unit],
 (List[Char]) =>
   List[Char]](Slm_ops_ext[lab[A, B], lab[A, B], sum[B, List[lab[A, B]]],
                            Unit]((_: lab[A, B]) =>
                                    (a: List[lab[A, B]]) =>
                                      Inr[List[lab[A, B]], B](a),
                                   (ga: lab[A, B]) =>
                                     (cs: List[lab[A, B]]) =>
                                       (if (membera[(lab[A, B],
              nat)](fs, (ga, size_list[lab[A, B]](cs))))
 ga else f),
                                   mapa[(lab[A, B], nat),
 lab[A, B]]((a: (lab[A, B], nat)) => fst[lab[A, B], nat](a), fs),
                                   f, (if (is_none[(lab[A, B], nat)](delt_opt))
(_: lab[A, B]) => (a: List[lab[A, B]]) => Inr[List[lab[A, B]], B](a)
else (_: lab[A, B]) =>
       (gs: List[lab[A, B]]) =>
         Inr[List[lab[A, B]],
              B](replicate[lab[A, B]](size_list[lab[A, B]](gs),
                                       fst[lab[A, B],
    nat](the[(lab[A, B], nat)](delt_opt))))),
                                   ()))
                                  })
  }

abstract sealed class sl_ops_ext[+A, +B, +C, +D, +E]
final case class
  Sl_ops_ext[A, B, C, D,
              E](a: A => (List[B]) => C, b: A => nat => C => Boolean,
                  c: A => (List[B]) => B, d: List[B], e: B,
                  f: (List[(term[A, D], term[A, D])]) =>
                       sum[(List[Char]) => List[Char], Unit],
                  g: A => (List[B]) => C, h: A => nat => C => Boolean,
                  i: C => List[C], j: A => nat => List[C], k: E)
  extends sl_ops_ext[A, B, C, D, E]

def qmodel_cge: nat => nat => Boolean =
  (x: nat) => (y: nat) => less_eq_nat(y, x)

def pointwise_ext[A](f: A => A => (Boolean, Boolean), x1: List[A], x2: List[A]):
      (Boolean, Boolean) =
  (f, x1, x2) match {
  case (f, Nil, Nil) => (false, true)
  case (f, uu :: uv, Nil) => (false, false)
  case (f, Nil, uw :: ux) => (false, false)
  case (f, a :: as, b :: bs) =>
    (f(a))(b) match {
      case (stri, true) =>
        pointwise_ext[A](f, as, bs) match {
          case (strir, true) => (stri || strir, true)
          case (strir, false) => (false, false)
        }
      case (stri, false) => (false, false)
    }
}

def qmodel_lge[A, B, C, D](f: A, n: B):
      (sum[List[nat], C]) => (sum[List[nat], D]) => Boolean =
  (l: sum[List[nat], C]) =>
    (r: sum[List[nat], D]) =>
      (l, r) match {
        case (Inl(cs1), Inl(cs2)) =>
          snd[Boolean,
               Boolean](pointwise_ext[nat]((x: nat) =>
     (y: nat) => (less_nat(y, x), less_eq_nat(y, x)),
    cs1, cs2))
        case (Inl(cs1), Inr(_)) => false
        case (Inr(_), b) => false
      }

def qmodel_L[A : equal, B : equal, C, D](sig: List[(lab[A, B], nat)]):
      (lab[A, B]) => (List[C]) => sum[List[C], D] =
  (f: lab[A, B]) =>
    (cs: List[C]) =>
      (if (membera[(lab[A, B], nat)](sig, (f, size_list[C](cs))))
        Inl[List[C], D](cs) else Inl[List[C], D](Nil))

def enum_vectors_nat[A](c: List[A], x1: nat): List[List[A]] = (c, x1) match {
  case (c, Zero_nat) => List(Nil)
  case (c, Suc(n)) =>
    {
      val a: List[List[A]] = enum_vectors_nat[A](c, n);
      maps[List[A],
            List[A]]((vec: List[A]) =>
                       mapa[A, List[A]]((ca: A) => ca :: vec, c),
                      a)
    }
}

def qmodel_LS_gen[A : equal, B, C](sig: List[(A, nat)], ls: List[B]):
      A => nat => List[sum[List[B], C]] =
  (f: A) =>
    (n: nat) =>
      (if (membera[(A, nat)](sig, (f, n)))
        mapa[List[B],
              sum[List[B],
                   C]]((a: List[B]) => Inl[List[B], C](a),
                        enum_vectors_nat[B](ls, n))
        else List(Inl[List[B], C](Nil)))

def qmodel_LS[A : equal, B : equal,
               C : equal](sig: List[(A, nat)], ls: List[B]):
      A => nat => (sum[List[B], C]) => Boolean =
  (f: A) =>
    (n: nat) =>
      set[sum[List[B], C]]((qmodel_LS_gen[A, B, C](sig, ls)).apply(f).apply(n))

def qmodel_LSa[A]:
      (lab[A, List[nat]]) =>
        nat => (sum[List[nat], List[lab[A, List[nat]]]]) => Boolean =
  (_: lab[A, List[nat]]) =>
    (_: nat) =>
      (a: sum[List[nat], List[lab[A, List[nat]]]]) =>
        a match {
          case Inl(_) => true
          case Inr(_) => false
        }

def pointwise_lgen(ns: List[nat]): List[List[nat]] =
  concat_lists[nat](mapa[nat, List[nat]]((n: nat) => upt(Zero_nat, Suc(n)), ns))

def qmodel_lgen[A](l: sum[List[nat], A]): List[sum[List[nat], A]] =
  l match {
    case Inl(ns) =>
      mapa[List[nat],
            sum[List[nat],
                 A]]((a: List[nat]) => Inl[List[nat], A](a), pointwise_lgen(ns))
    case Inr(_) => Nil
  }

def check_decr_present_aux_1[A : equal,
                              B : equal](r:
   List[(term[A, (B, nat)], term[A, (B, nat)])],
  v: B, f1: A, f2: A, n: nat):
      sum[(term[A, (B, nat)], term[A, (B, nat)]), Unit] =
  {
    val vs: List[term[A, (B, nat)]] =
      mapa[nat, term[A, (B, nat)]]((na: nat) => Var[(B, nat), A]((v, na)),
                                    upt(Zero_nat, n))
    val rule: (term[A, (B, nat)], term[A, (B, nat)]) =
      (Fun[A, (B, nat)](f1, vs), Fun[A, (B, nat)](f2, vs));
    check[(term[A, (B, nat)],
            term[A, (B, nat)])](! (is_none[(term[A, (B, nat)],
     term[A, (B, nat)])](find[(term[A, (B, nat)],
                                term[A, (B,
  nat)])]((a: (term[A, (B, nat)], term[A, (B, nat)])) =>
            instance_rule[A, (B, nat), (B, nat)](rule, a),
           r))),
                                 rule)
  }

def check_decr_present_aux_2[A : equal,
                              B : equal](r: List[(term[A, B], term[A, B])],
  v: B, req: List[(A, (A, nat))]):
      sum[(term[A, (B, nat)], term[A, (B, nat)]), Unit] =
  {
    val add_nats: (term[A, B]) => term[A, (B, nat)] =
      (a: term[A, B]) => map_vars[B, (B, nat), A]((va: B) => (va, Zero_nat), a)
    val ra: List[(term[A, (B, nat)], term[A, (B, nat)])] =
      mapa[(term[A, B], term[A, B]),
            (term[A, (B, nat)],
              term[A, (B, nat)])]((a: (term[A, B], term[A, B])) =>
                                    {
                                      val (l, ra): (term[A, B], term[A, B]) = a;
                                      (add_nats(l), add_nats(ra))
                                    },
                                   r);
    catcha[((A, (A, nat)), (term[A, (B, nat)], term[A, (B, nat)])), Unit,
            (term[A, (B, nat)],
              term[A, (B, nat)])](forallm[(A, (A, nat)),
   (term[A, (B, nat)],
     term[A, (B, nat)])]((a: (A, (A, nat))) =>
                           {
                             val (f1, (aa, b)): (A, (A, nat)) = a;
                             check_decr_present_aux_1[A, B](ra, v, f1, aa, b)
                           },
                          req),
                                   (x: ((A, (A, nat)),
 (term[A, (B, nat)], term[A, (B, nat)])))
                                     => Inl[(term[A, (B, nat)],
      term[A, (B, nat)]),
     Unit](snd[(A, (A, nat)), (term[A, (B, nat)], term[A, (B, nat)])](x)))
  }

def check_decr_present[A : equal,
                        B : equal](sig: List[(A, nat)],
                                    l: A => (List[nat]) => A, v: B, c: nat,
                                    r: List[(term[A, B], term[A, B])]):
      sum[(term[A, (B, nat)], term[A, (B, nat)]), Unit] =
  {
    val ca: List[nat] = upt(Zero_nat, Suc(c))
    val ls: ((A, nat)) => List[(A, (A, nat))] =
      (a: (A, nat)) =>
        {
          val (f, n): (A, nat) = a;
          maps[List[nat],
                (A, (A, nat))]((cs: List[nat]) =>
                                 maps[nat, (A,
     (A, nat))]((i: nat) =>
                  {
                    val ci: nat = nth[nat](cs, i);
                    (if (less_nat(ci, c))
                      List(((l(f))(list_update[nat](cs, i, Suc(ci))),
                             ((l(f))(cs), n)))
                      else Nil)
                  },
                 upt(Zero_nat, n)),
                                enum_vectors_nat[nat](ca, n))
        };
    check_decr_present_aux_2[A, B](r, v, maps[(A, nat), (A, (A, nat))](ls, sig))
  }

def qmodel_check_decr[A : equal : show,
                       B : equal](sig: List[(lab[A, List[nat]], nat)], v: B,
                                   c: nat):
      (List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])]) =>
        sum[(List[Char]) => List[Char], Unit] =
  (lR: List[(term[lab[A, List[nat]], B], term[lab[A, List[nat]], B])]) =>
    catcha[(term[lab[A, List[nat]], (B, nat)],
             term[lab[A, List[nat]], (B, nat)]),
            Unit,
            (List[Char]) =>
              List[Char]](check_decr_present[lab[A, List[nat]],
      B](sig, (a: lab[A, List[nat]]) =>
                (b: List[nat]) => Lab[A, List[nat]](a, b),
          v, c, lR),
                           (x: (term[lab[A, List[nat]], (B, nat)],
                                 term[lab[A, List[nat]], (B, nat)]))
                             => Inl[(List[Char]) => List[Char],
                                     Unit]({
     val display:
           (term[lab[A, List[nat]], (B, nat)]) =>
             term[lab[A, List[nat]], List[Char]]
       = (a: term[lab[A, List[nat]], (B, nat)]) =>
           map_vars[(B, nat), List[Char],
                     lab[A, List[nat]]]((aa: (B, nat)) =>
  {
    val (_, n): (B, nat) = aa;
    (comp[List[Char], List[Char],
           List[Char]](shows_prec_list[Char](Zero_nat, List('x')),
                        shows_prec_nat(Zero_nat, n))).apply(Nil)
  },
 a);
     comp[List[Char], List[Char],
           List[Char]](shows_prec_list[Char](Zero_nat,
      List('d', 'e', 'c', 'r', 'e', 'a', 's', 'i', 'n', 'g', ' ', 'r', 'u', 'l',
            'e', ' ')),
                        comp[List[Char], List[Char],
                              List[Char]](shows_rule[lab[A, List[nat]],
              List[Char]]((a: lab[A, List[nat]]) =>
                            shows_prec_lab[A, List[nat]](Zero_nat, a),
                           (a: List[Char]) =>
                             shows_prec_list[Char](Zero_nat, a),
                           List(' ', '-', '>', ' '),
                           (display(fst[term[lab[A, List[nat]], (B, nat)],
 term[lab[A, List[nat]], (B, nat)]](x)),
                             display(snd[term[lab[A, List[nat]], (B, nat)],
  term[lab[A, List[nat]], (B, nat)]](x)))),
   shows_prec_list[Char](Zero_nat,
                          List(' ', 'm', 'i', 's', 's', 'i', 'n', 'g'))))
   }))

def take_default[A](defa: A, x1: List[A], uu: nat): A = (defa, x1, uu) match {
  case (defa, Nil, uu) => defa
  case (uv, x :: xs, Zero_nat) => x
  case (defa, uw :: xs, Suc(i)) => take_default[A](defa, xs, i)
}

def eval_arithFun_unbound(nats: List[nat], x1: arithFun): nat = (nats, x1) match
  {
  case (nats, Arg(i)) => take_default[nat](Zero_nat, nats, i)
  case (nats, Const(n)) => n
  case (nats, Sum(Nil)) => Zero_nat
  case (nats, Sum(f :: fs)) =>
    plus_nat(eval_arithFun_unbound(nats, f),
              eval_arithFun_unbound(nats, Sum(fs)))
  case (nats, Prod(Nil)) => one_nat
  case (nats, Prod(f :: fs)) =>
    times_nat(eval_arithFun_unbound(nats, f),
               eval_arithFun_unbound(nats, Prod(fs)))
  case (nats, Max(List(f))) => eval_arithFun_unbound(nats, f)
  case (nats, Max(f :: v :: va)) =>
    max[nat].apply(eval_arithFun_unbound(nats,
  f)).apply(eval_arithFun_unbound(nats, Max(v :: va)))
  case (nats, Min(List(f))) => eval_arithFun_unbound(nats, f)
  case (nats, Min(f :: v :: va)) =>
    min[nat].apply(eval_arithFun_unbound(nats,
  f)).apply(eval_arithFun_unbound(nats, Min(v :: va)))
}

def eval_arithFun(c: nat, f: arithFun, nats: List[nat]): nat =
  mod_nat(eval_arithFun_unbound(nats, f), c)

def sl_inter_to_inter[A : equal](x0: sl_inter[A], fl: A, cs: List[nat]): nat =
  (x0, fl, cs) match {
  case (SL_Inter(c, ls), fl, cs) =>
    lookupa[(A, nat), arithFun]((fl, size_list[nat](cs)), ls) match {
      case None => Zero_nat
      case Some(f) => eval_arithFun(Suc(c), f, cs)
    }
}

def get_largest_element[A](x0: sl_inter[A]): nat = x0 match {
  case SL_Inter(n, uu) => n
}

def qsli_to_sl_unsafe[A : equal,
                       B : equal : show](v: A,
  f: List[(lab[B, List[nat]], nat)], g: List[(lab[B, List[nat]], nat)],
  sli: sl_inter[lab[B, List[nat]]]):
      sl_ops_ext[lab[B, List[nat]], nat,
                  sum[List[nat], List[lab[B, List[nat]]]], A, Unit] =
  {
    val c: nat = get_largest_element[lab[B, List[nat]]](sli)
    val ca: List[nat] = upt(Zero_nat, Suc(c));
    Sl_ops_ext[lab[B, List[nat]], nat, sum[List[nat], List[lab[B, List[nat]]]],
                A, Unit](qmodel_L[B, List[nat], nat,
                                   List[lab[B, List[nat]]]](f),
                          qmodel_LS[lab[B, List[nat]], nat,
                                     List[lab[B, List[nat]]]](f, ca),
                          (a: lab[B, List[nat]]) =>
                            (b: List[nat]) =>
                              sl_inter_to_inter[lab[B, List[nat]]](sli, a, b),
                          ca, c, qmodel_check_decr[B, A](f, v, c),
                          qmodel_L[B, List[nat], nat,
                                    List[lab[B, List[nat]]]](g),
                          qmodel_LSa[B],
                          (a: sum[List[nat], List[lab[B, List[nat]]]]) =>
                            qmodel_lgen[List[lab[B, List[nat]]]](a),
                          qmodel_LS_gen[lab[B, List[nat]], nat,
 List[lab[B, List[nat]]]](f, ca),
                          ())
  }

def qmodel_check_interpretation(f: arithFun, n: nat, c: nat):
      sum[(List[Char]) => List[Char], Unit] =
  {
    val ca: List[nat] = upt(Zero_nat, Suc(c))
    val css: List[List[nat]] = enum_vectors_nat[nat](ca, n);
    catcha[(List[nat], (List[Char]) => List[Char]), Unit,
            (List[Char]) =>
              List[Char]](forallm[List[nat],
                                   (List[Char]) =>
                                     List[Char]]((cs: List[nat]) =>
           catcha[(nat, (List[Char]) => List[Char]), Unit,
                   (List[Char]) =>
                     List[Char]](forallm[nat,
  (List[Char]) =>
    List[Char]]((i: nat) =>
                  catcha[(nat, (List[Char]) => List[Char]), Unit,
                          (List[Char]) =>
                            List[Char]](forallm[nat,
         (List[Char]) =>
           List[Char]]((l: nat) =>
                         check[(List[Char]) =>
                                 List[Char]](less_eq_nat(eval_arithFun(Suc(c),
                                f, cs),
                  eval_arithFun(Suc(c), f, list_update[nat](cs, i, l))),
      comp[List[Char], List[Char],
            List[Char]](shows_prec_list[Char](Zero_nat,
       List('n', 'o', 't', ' ', 'm', 'o', 'n', 'o', 't', 'o', 'n', 'e', ' ',
             'i', 'n', ' ')),
                         comp[List[Char], List[Char],
                               List[Char]](shows_prec_nat(Zero_nat, Suc(i)),
    shows_prec_list[Char](Zero_nat,
                           List('.', ' ', 'a', 'r', 'g', 'u', 'm', 'e', 'n',
                                 't'))))),
                        upt(nth[nat](cs, i), Suc(c))),
 (x: (nat, (List[Char]) => List[Char])) =>
   Inl[(List[Char]) => List[Char],
        Unit](snd[nat, (List[Char]) => List[Char]](x))),
                 upt(Zero_nat, n)),
                                  (x: (nat, (List[Char]) => List[Char])) =>
                                    Inl[(List[Char]) => List[Char],
 Unit](snd[nat, (List[Char]) => List[Char]](x))),
          css),
                           (x: (List[nat], (List[Char]) => List[Char])) =>
                             Inl[(List[Char]) => List[Char],
                                  Unit](snd[List[nat],
     (List[Char]) => List[Char]](x)))
  }

def qmodel_check_valid[A : show](x0: sl_inter[A]):
      sum[(List[Char]) => List[Char], Unit] =
  x0 match {
  case SL_Inter(c, ls) =>
    catcha[(((A, nat), arithFun), (List[Char]) => List[Char]), Unit,
            (List[Char]) =>
              List[Char]](forallm[((A, nat), arithFun),
                                   (List[Char]) =>
                                     List[Char]]((a: ((A, nat), arithFun)) =>
           {
             val (aa, b): ((A, nat), arithFun) = a;
             ({
                val (f, n): (A, nat) = aa;
                (g: arithFun) =>
                  catcha[(List[Char]) => List[Char], Unit,
                          (List[Char]) =>
                            List[Char]](qmodel_check_interpretation(g, n, c),
 (x: (List[Char]) => List[Char]) =>
   Inl[(List[Char]) => List[Char],
        Unit](comp[List[Char], List[Char],
                    List[Char]](shows_prec_list[Char](Zero_nat,
               List('p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'i', 'n', ' ', 'w',
                     'e', 'a', 'k', '-', 'm', 'o', 'n', 'o', 't', 'o', 'n', 'i',
                     'c', 'i', 't', 'y', ' ', 'o', 'f', ' ', 'i', 'n', 't', 'e',
                     'r', 'p', 'r', 'e', 't', 'a', 't', 'i', 'o', 'n', ' ', 'o',
                     'f', ' ')),
                                 comp[List[Char], List[Char],
                                       List[Char]]((ab: List[Char]) =>
             shows_prec[A](Zero_nat, f, ab),
            comp[List[Char], List[Char], List[Char]](shows_nl, x)))))
              })(b)
           },
          ls),
                           (x: (((A, nat), arithFun),
                                 (List[Char]) => List[Char]))
                             => Inl[(List[Char]) => List[Char],
                                     Unit](snd[((A, nat), arithFun),
        (List[Char]) => List[Char]](x)))
}

def qsli_to_sl[A : equal,
                B : equal : show](v: A, f: List[(lab[B, List[nat]], nat)],
                                   g: List[(lab[B, List[nat]], nat)],
                                   sli: sl_inter[lab[B, List[nat]]]):
      sum[(List[Char]) => List[Char],
           sl_ops_ext[lab[B, List[nat]], nat,
                       sum[List[nat], List[lab[B, List[nat]]]], A, Unit]] =
  bindb[(List[Char]) => List[Char], Unit,
         sl_ops_ext[lab[B, List[nat]], nat,
                     sum[List[nat], List[lab[B, List[nat]]]], A,
                     Unit]](qmodel_check_valid[lab[B, List[nat]]](sli),
                             (_: Unit) =>
                               Inr[sl_ops_ext[lab[B, List[nat]], nat,
       sum[List[nat], List[lab[B, List[nat]]]], A, Unit],
                                    (List[Char]) =>
                                      List[Char]](qsli_to_sl_unsafe[A,
                             B](v, f, g, sli)))

def sli_to_slm[A : equal](sli: sl_inter[lab[A, List[nat]]]):
      slm_ops_ext[lab[A, List[nat]], nat,
                   sum[List[nat], List[lab[A, List[nat]]]], Unit] =
  {
    val c: nat = get_largest_element[lab[A, List[nat]]](sli);
    Slm_ops_ext[lab[A, List[nat]], nat, sum[List[nat], List[lab[A, List[nat]]]],
                 Unit]((_: lab[A, List[nat]]) =>
                         (a: List[nat]) =>
                           Inl[List[nat], List[lab[A, List[nat]]]](a),
                        (a: lab[A, List[nat]]) =>
                          (b: List[nat]) =>
                            sl_inter_to_inter[lab[A, List[nat]]](sli, a, b),
                        upt(Zero_nat, Suc(c)), c,
                        (_: lab[A, List[nat]]) =>
                          (a: List[nat]) =>
                            Inl[List[nat], List[lab[A, List[nat]]]](a),
                        ())
  }

def sl_C[A, B, C, D, E](x0: sl_ops_ext[A, B, C, D, E]): List[B] = x0 match {
  case Sl_ops_ext(sl_La, sl_LSa, sl_I, sl_C, sl_c, sl_check_decr, sl_L, sl_LS,
                   sl_lgen, sl_LS_gen, more)
    => sl_C
}

def sl_I[A, B, C, D, E](x0: sl_ops_ext[A, B, C, D, E]): A => (List[B]) => B = x0
  match {
  case Sl_ops_ext(sl_La, sl_LSa, sl_I, sl_C, sl_c, sl_check_decr, sl_L, sl_LS,
                   sl_lgen, sl_LS_gen, more)
    => sl_I
}

def sl_L[A, B, C, D, E](x0: sl_ops_ext[A, B, C, D, E]): A => (List[B]) => C = x0
  match {
  case Sl_ops_ext(sl_La, sl_LSa, sl_I, sl_C, sl_c, sl_check_decr, sl_L, sl_LS,
                   sl_lgen, sl_LS_gen, more)
    => sl_La
}

def lab_lhs_list[A, B, C, D,
                  E : equal](i: A => (List[B]) => B, l: A => (List[B]) => C,
                              lc: A => nat => C => D, c: List[B],
                              t: term[A, E]):
      List[term[D, E]] =
  mapa[E => B,
        term[D, E]]((alpha: E => B) =>
                      snd[B, term[D, E]](eval_lab[A, B, C, D,
           E](i, l, lc, alpha, t)),
                     mapa[List[(E, B)],
                           E => B]((a: List[(E, B)]) =>
                                     (b: E) => fun_of[E, B](a, b),
                                    enum_vectors[B,
          E](c, vars_term_list[A, E].apply(t))))

def lab_lhss_list[A, B, C, D,
                   E : equal](i: A => (List[B]) => B, l: A => (List[B]) => C,
                               lc: A => nat => C => D, c: List[B],
                               q: List[term[A, E]]):
      List[term[D, E]] =
  maps[term[A, E],
        term[D, E]]((a: term[A, E]) =>
                      lab_lhs_list[A, B, C, D, E](i, l, lc, c, a),
                     q)

def check_sl_Q[A : equal : show, B, C,
                D : equal : show](i: A => (List[B]) => B,
                                   l: A => (List[B]) => C,
                                   lc: A => nat => C => A, c: List[B],
                                   lQ: List[term[A, D]], q: List[term[A, D]]):
      sum[(List[Char]) => List[Char], Unit] =
  bindb[(List[Char]) => List[Char], Unit,
         Unit](check[(List[Char]) =>
                       List[Char]](! (nulla[B](c)),
                                    shows_prec_list[Char](Zero_nat,
                   List('c', 'a', 'r', 'r', 'i', 'e', 'r', ' ', 'm', 'u', 's',
                         't', ' ', 'b', 'e', ' ', 'n', 'o', 'n', '-', 'e', 'm',
                         'p', 't', 'y'))),
                (_: Unit) =>
                  catcha[term[A, D], Unit,
                          (List[Char]) =>
                            List[Char]](check_subseteq[term[A,
                     D]](lab_lhss_list[A, B, C, A, D](i, l, lc, c, q), lQ),
 (x: term[A, D]) =>
   Inl[(List[Char]) => List[Char],
        Unit](comp[List[Char], List[Char],
                    List[Char]](shows_prec_list[Char](Zero_nat,
               List('l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 't', 'e', 'r', 'm',
                     ' ')),
                                 comp[List[Char], List[Char],
                                       List[Char]](shows_term[A,
                       D]((a: A) =>
                            (b: List[Char]) => shows_prec[A](Zero_nat, a, b),
                           (a: D) =>
                             (b: List[Char]) => shows_prec[D](Zero_nat, a, b),
                           x),
            shows_prec_list[Char](Zero_nat,
                                   List(' ', 'i', 's', ' ', 'm', 'i', 's', 's',
 'i', 'n', 'g')))))))

def lab_rule_ass[A, B, C, D,
                  E](i: A => (List[B]) => B, l: A => (List[B]) => C,
                      lc: A => nat => C => D, alpha: E => B,
                      rule: (term[A, E], term[A, E])):
      (term[D, E], term[D, E]) =
  (snd[B, term[D, E]](eval_lab[A, B, C, D,
                                E](i, l, lc, alpha,
                                    fst[term[A, E], term[A, E]](rule))),
    snd[B, term[D, E]](eval_lab[A, B, C, D,
                                 E](i, l, lc, alpha,
                                     snd[term[A, E], term[A, E]](rule))))

def lab_rule_list[A, B, C, D,
                   E : equal](i: A => (List[B]) => B, l: A => (List[B]) => C,
                               lc: A => nat => C => D, c: List[B],
                               lr: (term[A, E], term[A, E])):
      List[(term[D, E], term[D, E])] =
  mapa[E => B,
        (term[D, E],
          term[D, E])]((alpha: E => B) =>
                         lab_rule_ass[A, B, C, D, E](i, l, lc, alpha, lr),
                        mapa[List[(E, B)],
                              E => B]((a: List[(E, B)]) =>
(b: E) => fun_of[E, B](a, b),
                                       enum_vectors[B,
             E](c, vars_rule_list[A, E](lr))))

def lab_trs_list[A, B, C, D,
                  E : equal](i: A => (List[B]) => B, l: A => (List[B]) => C,
                              lc: A => nat => C => D, c: List[B],
                              r: List[(term[A, E], term[A, E])]):
      List[(term[D, E], term[D, E])] =
  maps[(term[A, E], term[A, E]),
        (term[D, E],
          term[D, E])]((a: (term[A, E], term[A, E])) =>
                         lab_rule_list[A, B, C, D, E](i, l, lc, c, a),
                        r)

def check_sl_lab[A : equal : show, B, C,
                  D : equal : show](i: A => (List[B]) => B,
                                     l: A => (List[B]) => C,
                                     lc: A => nat => C => A, c: List[B],
                                     lR: List[(term[A, D], term[A, D])],
                                     r: List[(term[A, D], term[A, D])]):
      sum[(List[Char]) => List[Char], Unit] =
  bindb[(List[Char]) => List[Char], Unit,
         Unit](check[(List[Char]) =>
                       List[Char]](! (nulla[B](c)),
                                    shows_prec_list[Char](Zero_nat,
                   List('c', 'a', 'r', 'r', 'i', 'e', 'r', ' ', 'm', 'u', 's',
                         't', ' ', 'b', 'e', ' ', 'n', 'o', 'n', '-', 'e', 'm',
                         'p', 't', 'y'))),
                (_: Unit) =>
                  catcha[(term[A, D], term[A, D]), Unit,
                          (List[Char]) =>
                            List[Char]](check_subseteq[(term[A, D],
                 term[A, D])](lR, lab_trs_list[A, B, C, A, D](i, l, lc, c, r)),
 (x: (term[A, D], term[A, D])) =>
   Inl[(List[Char]) => List[Char],
        Unit](comp[List[Char], List[Char],
                    List[Char]](shows_prec_list[Char](Zero_nat,
               List('l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'r', 'u', 'l', 'e',
                     ' ')),
                                 comp[List[Char], List[Char],
                                       List[Char]](shows_rule[A,
                       D]((a: A) =>
                            (b: List[Char]) => shows_prec[A](Zero_nat, a, b),
                           (a: D) =>
                             (b: List[Char]) => shows_prec[D](Zero_nat, a, b),
                           List(' ', '-', '>', ' '), x),
            shows_prec_list[Char](Zero_nat,
                                   List(' ', 'i', 's', ' ', 'n', 'o', 't', ' ',
 'a', 'l', 'l', 'o', 'w', 'e', 'd')))))))

def check_sl_Qa[A : equal : show, B,
                 C : equal : show](ld: A => (A, B), lQ: List[term[A, C]],
                                    q: List[term[A, C]]):
      sum[(List[Char]) => List[Char], Unit] =
  {
    val u: A => A = (l: A) => fst[A, B](ld(l));
    catcha[(term[A, C], (List[Char]) => List[Char]), Unit,
            (List[Char]) =>
              List[Char]](forallm[term[A, C],
                                   (List[Char]) =>
                                     List[Char]]((lq: term[A, C]) =>
           check[(List[Char]) =>
                   List[Char]](membera[term[A, C]](q, map_term[A, A, C](u, lq)),
                                comp[List[Char], List[Char],
                                      List[Char]](shows_prec_list[Char](Zero_nat,
                                 List('u', 'n', 'l', 'a', 'b', 'e', 'l', 'i',
                                       'n', 'g', ' ')),
           comp[List[Char], List[Char],
                 List[Char]](shows_term[A,
 C]((a: A) => (b: List[Char]) => shows_prec[A](Zero_nat, a, b),
     (a: C) => (b: List[Char]) => shows_prec[C](Zero_nat, a, b), lq),
                              shows_prec_list[Char](Zero_nat,
             List(' ', 'y', 'i', 'e', 'l', 'd', 's', ' ', 'a', ' ', 't', 'e',
                   'r', 'm', ' ', 'n', 'o', 't', ' ', 'i', 'n', ' ', 'Q'))))),
          lQ),
                           (x: (term[A, C], (List[Char]) => List[Char])) =>
                             Inl[(List[Char]) => List[Char],
                                  Unit](snd[term[A, C],
     (List[Char]) => List[Char]](x)))
  }

def model_splitter[A : equal, B,
                    C : equal](ld: A => (A, B),
                                lAll: List[(term[A, C], term[A, C])],
                                uRw: List[(term[A, C], term[A, C])]):
      (List[(term[A, C], term[A, C])],
        (List[(term[A, C], term[A, C])], List[(term[A, C], term[A, C])])) =
  {
    val unlab: A => A = (lf: A) => fst[A, B](ld(lf))
    val la: List[((term[A, C], term[A, C]), (term[A, C], term[A, C]))] =
      mapa[(term[A, C], term[A, C]),
            ((term[A, C], term[A, C]),
              (term[A, C],
                term[A, C]))]((r: (term[A, C], term[A, C])) =>
                                (r, map_rule[A, A, C](unlab, r)),
                               lAll)
    val (rw, r):
          (List[((term[A, C], term[A, C]), (term[A, C], term[A, C]))],
            List[((term[A, C], term[A, C]), (term[A, C], term[A, C]))])
      = partition[((term[A, C], term[A, C]),
                    (term[A, C],
                      term[A, C]))]((a: ((term[A, C], term[A, C]),
  (term[A, C], term[A, C])))
                                      => {
   val (_, aa): ((term[A, C], term[A, C]), (term[A, C], term[A, C])) = a;
   membera[(term[A, C], term[A, C])](uRw, aa)
 },
                                     la);
    (mapa[((term[A, C], term[A, C]), (term[A, C], term[A, C])),
           (term[A, C],
             term[A, C])]((a: ((term[A, C], term[A, C]),
                                (term[A, C], term[A, C])))
                            => fst[(term[A, C], term[A, C]),
                                    (term[A, C], term[A, C])](a),
                           r),
      (mapa[((term[A, C], term[A, C]), (term[A, C], term[A, C])),
             (term[A, C],
               term[A, C])]((a: ((term[A, C], term[A, C]),
                                  (term[A, C], term[A, C])))
                              => fst[(term[A, C], term[A, C]),
                                      (term[A, C], term[A, C])](a),
                             rw),
        Nil))
  }

def sem_lab_proc[A : equal : show, B, C,
                  D : equal : show](ld: A => (A, B),
                                     i: dpp_ops_ext[C, A, D, Unit],
                                     valid:
                                       sum[(List[Char]) => List[Char], Unit],
                                     check_Q:
                                       (List[term[A, D]]) =>
 (List[term[A, D]]) => sum[(List[Char]) => List[Char], Unit],
                                     check_laba:
                                       (List[(term[A, D], term[A, D])]) =>
 (List[(term[A, D], term[A, D])]) => sum[(List[Char]) => List[Char], Unit],
                                     check_lab:
                                       (List[(term[A, D], term[A, D])]) =>
 (List[(term[A, D], term[A, D])]) => sum[(List[Char]) => List[Char], Unit],
                                     check_model_lab:
                                       (List[(term[A, D], term[A, D])]) =>
 (List[(term[A, D], term[A, D])]) => sum[(List[Char]) => List[Char], Unit],
                                     lPAll: List[(term[A, D], term[A, D])],
                                     lQ: List[term[A, D]],
                                     lRAll: List[(term[A, D], term[A, D])],
                                     dpp: C):
      sum[(List[Char]) => List[Char], C] =
  {
    val r: List[(term[A, D], term[A, D])] = (rc[C, A, D, Unit](i)).apply(dpp)
    val rw: List[(term[A, D], term[A, D])] = (rwc[C, A, D, Unit](i)).apply(dpp)
    val pw: List[(term[A, D], term[A, D])] = (pwa[C, A, D, Unit](i)).apply(dpp)
    val p: List[(term[A, D], term[A, D])] = (pa[C, A, D, Unit](i)).apply(dpp)
    val (lP, (lPw, _)):
          (List[(term[A, D], term[A, D])],
            (List[(term[A, D], term[A, D])], List[(term[A, D], term[A, D])]))
      = model_splitter[A, B, D](ld, lPAll, pw)
    val (lR, (lRw, _)):
          (List[(term[A, D], term[A, D])],
            (List[(term[A, D], term[A, D])], List[(term[A, D], term[A, D])]))
      = model_splitter[A, B, D](ld, lRAll, rw);
    bindb[(List[Char]) => List[Char], Unit,
           Unit](valid,
                  (_: Unit) =>
                    {
                      val q: List[term[A, D]] =
                        (qc[C, A, D, Unit](i)).apply(dpp);
                      catcha[(List[Char]) => List[Char], Unit,
                              (List[Char]) =>
                                List[Char]](bindb[(List[Char]) => List[Char],
           Unit,
           Unit]((check_Q(lQ))(q),
                  (_: Unit) =>
                    bindb[(List[Char]) => List[Char], Unit,
                           Unit](check_sl_Qa[A, B, D](ld, lQ, q),
                                  (_: Unit) =>
                                    bindb[(List[Char]) => List[Char], Unit,
   Unit]((check_laba(lP))(p),
          (_: Unit) =>
            bindb[(List[Char]) => List[Char], Unit,
                   Unit]((check_laba(lPw))(pw),
                          (_: Unit) =>
                            bindb[(List[Char]) => List[Char], Unit,
                                   Unit]((check_model_lab(lR))(r),
  (_: Unit) =>
    bindb[(List[Char]) => List[Char], Unit,
           Unit]((check_model_lab(lRw))(rw),
                  (_: Unit) =>
                    bindb[(List[Char]) => List[Char], Unit,
                           Unit]((check_lab(lR))(r),
                                  (_: Unit) => (check_lab(lRw))(rw)))))))),
     (x: (List[Char]) => List[Char]) =>
       Inl[(List[Char]) => List[Char],
            Unit](comp[List[Char], List[Char],
                        List[Char]](shows_string.apply(List('p', 'r', 'o', 'b',
                     'l', 'e', 'm', ' ', 'd', 'u', 'r', 'i', 'n', 'g', ' ', 'l',
                     'a', 'b', 'e', 'l', 'i', 'n', 'g', ':')),
                                     comp[List[Char], List[Char],
   List[Char]](shows_nl, x))))
                    })
      match {
      case Inl(a) => Inl[(List[Char]) => List[Char], C](a)
      case Inr(_) =>
        Inr[C, (List[Char]) =>
                 List[Char]]((mkc[C, A, D,
                                   Unit](i)).apply(lP).apply(lPw).apply(lQ).apply(lR).apply(lRw))
    }
  }

def check_sl_rule_ass[A : show, B : show, C, D : equal : show,
                       E : equal : show](mc: Boolean, i: A => (List[B]) => B,
  la: A => (List[B]) => C, lc: A => nat => C => D, cge: B => B => Boolean,
  lR: List[(term[D, E], term[D, E])], alpha: E => B,
  x7: (term[A, E], term[A, E])):
      sum[(List[Char]) => List[Char], Unit] =
  (mc, i, la, lc, cge, lR, alpha, x7) match {
  case (mc, i, la, lc, cge, lR, alpha, (l, r)) =>
    {
      val cl_ll: (B, term[D, E]) = eval_lab[A, B, C, D, E](i, la, lc, alpha, l)
      val cr_lr: (B, term[D, E]) = eval_lab[A, B, C, D, E](i, la, lc, alpha, r);
      bindb[(List[Char]) => List[Char], Unit,
             Unit](check[(List[Char]) =>
                           List[Char]]((if (mc)
 (cge(fst[B, term[D, E]](cl_ll)))(fst[B, term[D, E]](cr_lr)) else true),
comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List('r', 'u', 'l', 'e', ' ')),
                   comp[List[Char], List[Char],
                         List[Char]](shows_rule[A,
         E]((a: A) => (b: List[Char]) => shows_prec[A](Zero_nat, a, b),
             (a: E) => (b: List[Char]) => shows_prec[E](Zero_nat, a, b),
             List(' ', '-', '>', ' '), (l, r)),
                                      comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List(' ', 'v', 'i', 'o', 'l', 'a', 't', 'e',
 's', ' ', 't', 'h', 'e', ' ', 'm', 'o', 'd', 'e', 'l', ' ', 'c', 'o', 'n', 'd',
 'i', 't', 'i', 'o', 'n', ',', ' ', '[', 'l', 'h', 's', ']', ' ', '=', ' ')),
                 comp[List[Char], List[Char],
                       List[Char]]((a: List[Char]) =>
                                     shows_prec[B](Zero_nat,
            fst[B, term[D, E]](cl_ll), a),
                                    comp[List[Char], List[Char],
  List[Char]](shows_string.apply(List(',', ' ', '[', 'r', 'h', 's', ']', ' ',
                                       '=', ' ')),
               (a: List[Char]) =>
                 shows_prec[B](Zero_nat, fst[B, term[D, E]](cr_lr), a))))))),
                    (_: Unit) =>
                      check[(List[Char]) =>
                              List[Char]](membera[(term[D, E],
            term[D, E])](lR, (snd[B, term[D, E]](cl_ll),
                               snd[B, term[D, E]](cr_lr))),
   comp[List[Char], List[Char],
         List[Char]](shows_string.apply(List('l', 'a', 'b', 'e', 'l', 'e', 'd',
      ' ', 'r', 'u', 'l', 'e', ' ')),
                      comp[List[Char], List[Char],
                            List[Char]](shows_rule[D,
            E]((a: D) => (b: List[Char]) => shows_prec[D](Zero_nat, a, b),
                (a: E) => (b: List[Char]) => shows_prec[E](Zero_nat, a, b),
                List(' ', '-', '>', ' '),
                (snd[B, term[D, E]](cl_ll), snd[B, term[D, E]](cr_lr))),
 shows_string.apply(List(' ', 'm', 'i', 's', 's', 'i', 'n', 'g'))))))
    }
}

def check_sl_rule[A : show, B : show, C, D : equal : show,
                   E : equal : show](i: A => (List[B]) => B,
                                      l: A => (List[B]) => C,
                                      lc: A => nat => C => D, c: List[B],
                                      cge: B => B => Boolean, mc: Boolean,
                                      lR: List[(term[D, E], term[D, E])],
                                      lr: (term[A, E], term[A, E])):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(E => B, (List[Char]) => List[Char]), Unit,
          (List[Char]) =>
            List[Char]](forallm[E => B,
                                 (List[Char]) =>
                                   List[Char]]((alpha: E => B) =>
         check_sl_rule_ass[A, B, C, D, E](mc, i, l, lc, cge, lR, alpha, lr),
        mapa[List[(E, B)],
              E => B]((a: List[(E, B)]) => (b: E) => fun_of[E, B](a, b),
                       enum_vectors[B, E](c, vars_rule_list[A, E](lr)))),
                         (x: (E => B, (List[Char]) => List[Char])) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](snd[E => B,
   (List[Char]) => List[Char]](x)))

def check_sl_lab_trs[A : equal : show, B : show, C,
                      D : equal : show](i: A => (List[B]) => B,
 l: A => (List[B]) => C, c: List[B], cge: B => B => Boolean,
 label: A => nat => C => A, lP: List[(term[A, D], term[A, D])],
 p: List[(term[A, D], term[A, D])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[((term[A, D], term[A, D]), (List[Char]) => List[Char]), Unit,
          (List[Char]) =>
            List[Char]](forallm[(term[A, D], term[A, D]),
                                 (List[Char]) =>
                                   List[Char]]((a: (term[A, D], term[A, D])) =>
         check_sl_rule[A, B, C, A, D](i, l, label, c, cge, false, lP, a),
        p),
                         (x: ((term[A, D], term[A, D]),
                               (List[Char]) => List[Char]))
                           => Inl[(List[Char]) => List[Char],
                                   Unit](snd[(term[A, D], term[A, D]),
      (List[Char]) => List[Char]](x)))

def check_sl_model_lab_trs[A : equal : show, B : show, C,
                            D : equal : show](i: A => (List[B]) => B,
       l: A => (List[B]) => C, c: List[B], cge: B => B => Boolean,
       label: A => nat => C => A, lR: List[(term[A, D], term[A, D])],
       r: List[(term[A, D], term[A, D])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[((term[A, D], term[A, D]), (List[Char]) => List[Char]), Unit,
          (List[Char]) =>
            List[Char]](forallm[(term[A, D], term[A, D]),
                                 (List[Char]) =>
                                   List[Char]]((a: (term[A, D], term[A, D])) =>
         check_sl_rule[A, B, C, A, D](i, l, label, c, cge, true, lR, a),
        r),
                         (x: ((term[A, D], term[A, D]),
                               (List[Char]) => List[Char]))
                           => Inl[(List[Char]) => List[Char],
                                   Unit](snd[(term[A, D], term[A, D]),
      (List[Char]) => List[Char]](x)))

def sem_lab_fin_proc[A : equal : show, B, C, D : equal : show,
                      E : equal : show](lc: A => nat => B => A, ld: A => (A, B),
 i: dpp_ops_ext[C, A, D, Unit],
 gen: (List[(A, nat)]) =>
        (List[(A, nat)]) =>
          sum[(List[Char]) => List[Char], sl_ops_ext[A, E, B, D, Unit]],
 lPAll: List[(term[A, D], term[A, D])], lQ: List[term[A, D]],
 lRAll: List[(term[A, D], term[A, D])], dp: C):
      sum[(List[Char]) => List[Char], C] =
  bindb[(List[Char]) => List[Char], sl_ops_ext[A, E, B, D, Unit],
         C]((gen(list_union[(A, nat)](funas_trs_list[A,
              D]((rulesd[C, A, D, Unit](i)).apply(dp)),
                                       funas_args_trs_list[A,
                    D]((pairsa[C, A, D, Unit](i)).apply(dp)))))(Nil),
             (ops: sl_ops_ext[A, E, B, D, Unit]) =>
               {
                 val check_q:
                       (List[term[A, D]]) =>
                         (List[term[A, D]]) =>
                           sum[(List[Char]) => List[Char], Unit]
                   = (a: List[term[A, D]]) =>
                       (b: List[term[A, D]]) =>
                         check_sl_Q[A, E, B,
                                     D](sl_I[A, E, B, D, Unit](ops),
 sl_L[A, E, B, D, Unit](ops), lc, sl_C[A, E, B, D, Unit](ops), a, b)
                 val check_ml:
                       (List[(term[A, D], term[A, D])]) =>
                         (List[(term[A, D], term[A, D])]) =>
                           sum[(List[Char]) => List[Char], Unit]
                   = (a: List[(term[A, D], term[A, D])]) =>
                       (b: List[(term[A, D], term[A, D])]) =>
                         check_sl_model_lab_trs[A, E, B,
         D](sl_I[A, E, B, D, Unit](ops), sl_L[A, E, B, D, Unit](ops),
             sl_C[A, E, B, D, Unit](ops), (aa: E) => (ba: E) => eq[E](aa, ba),
             lc, a, b)
                 val check_l:
                       (List[(term[A, D], term[A, D])]) =>
                         (List[(term[A, D], term[A, D])]) =>
                           sum[(List[Char]) => List[Char], Unit]
                   = (a: List[(term[A, D], term[A, D])]) =>
                       (b: List[(term[A, D], term[A, D])]) =>
                         check_sl_lab_trs[A, E, B,
   D](sl_I[A, E, B, D, Unit](ops), sl_L[A, E, B, D, Unit](ops),
       sl_C[A, E, B, D, Unit](ops), (aa: E) => (ba: E) => eq[E](aa, ba), lc, a,
       b)
                 val check_la:
                       (List[(term[A, D], term[A, D])]) =>
                         (List[(term[A, D], term[A, D])]) =>
                           sum[(List[Char]) => List[Char], Unit]
                   = (a: List[(term[A, D], term[A, D])]) =>
                       (b: List[(term[A, D], term[A, D])]) =>
                         check_sl_lab[A, E, B,
                                       D](sl_I[A, E, B, D, Unit](ops),
   sl_L[A, E, B, D, Unit](ops), lc, sl_C[A, E, B, D, Unit](ops), a, b);
                 sem_lab_proc[A, B, C,
                               D](ld, i,
                                   Inr[Unit, (List[Char]) => List[Char]](()),
                                   check_q, check_l, check_la, check_ml, lPAll,
                                   lQ, lRAll, dp)
               })

def slm_C[A, B, C, D](x0: slm_ops_ext[A, B, C, D]): List[B] = x0 match {
  case Slm_ops_ext(slm_La, slm_I, slm_C, slm_c, slm_L, more) => slm_C
}

def slm_I[A, B, C, D](x0: slm_ops_ext[A, B, C, D]): A => (List[B]) => B = x0
  match {
  case Slm_ops_ext(slm_La, slm_I, slm_C, slm_c, slm_L, more) => slm_I
}

def slm_L[A, B, C, D](x0: slm_ops_ext[A, B, C, D]): A => (List[B]) => C = x0
  match {
  case Slm_ops_ext(slm_La, slm_I, slm_C, slm_c, slm_L, more) => slm_La
}

def slm_c[A, B, C, D](x0: slm_ops_ext[A, B, C, D]): B = x0 match {
  case Slm_ops_ext(slm_La, slm_I, slm_C, slm_c, slm_L, more) => slm_c
}

def slm_La[A, B, C, D](x0: slm_ops_ext[A, B, C, D]): A => (List[B]) => C = x0
  match {
  case Slm_ops_ext(slm_La, slm_I, slm_C, slm_c, slm_L, more) => slm_L
}

def slm_to_sl[A, B, C, D](ops: slm_ops_ext[A, B, C, Unit]):
      sl_ops_ext[A, B, C, D, Unit] =
  Sl_ops_ext[A, B, C, D,
              Unit](slm_L[A, B, C, Unit](ops),
                     (_: A) => (_: nat) => top_fun[C, Boolean],
                     slm_I[A, B, C, Unit](ops), slm_C[A, B, C, Unit](ops),
                     slm_c[A, B, C, Unit](ops),
                     (_: List[(term[A, D], term[A, D])]) =>
                       Inr[Unit, (List[Char]) => List[Char]](()),
                     slm_La[A, B, C, Unit](ops),
                     (_: A) => (_: nat) => top_fun[C, Boolean],
                     (l: C) => List(l), (_: A) => (_: nat) => Nil, ())

def slm_gen_to_sl_gen[A, B, C,
                       D](gen: (List[(A, nat)]) =>
                                 (List[(A, nat)]) =>
                                   sum[(List[Char]) => List[Char],
slm_ops_ext[A, B, C, Unit]]):
      (List[(A, nat)]) =>
        (List[(A, nat)]) =>
          sum[(List[Char]) => List[Char], sl_ops_ext[A, B, C, D, Unit]] =
  (f: List[(A, nat)]) =>
    (g: List[(A, nat)]) =>
      bindb[(List[Char]) => List[Char], slm_ops_ext[A, B, C, Unit],
             sl_ops_ext[A, B, C, D,
                         Unit]]((gen(f))(g),
                                 (ops: slm_ops_ext[A, B, C, Unit]) =>
                                   Inr[sl_ops_ext[A, B, C, D, Unit],
(List[Char]) => List[Char]](slm_to_sl[A, B, C, D](ops)))

def sl_La[A, B, C, D, E](x0: sl_ops_ext[A, B, C, D, E]): A => (List[B]) => C =
  x0 match {
  case Sl_ops_ext(sl_La, sl_LSa, sl_I, sl_C, sl_c, sl_check_decr, sl_L, sl_LS,
                   sl_lgen, sl_LS_gen, more)
    => sl_L
}

def funas_root_rule_list[A : equal, B](r: (term[A, B], term[A, B])):
      List[(A, nat)] =
  remdups[(A, nat)](option_to_list[(A, nat)](root[A,
           B](fst[term[A, B], term[A, B]](r))) ++
                      option_to_list[(A, nat)](root[A,
             B](snd[term[A, B], term[A, B]](r))))

def funas_root_rules_list[A : equal, B](r: List[(term[A, B], term[A, B])]):
      List[(A, nat)] =
  remdups[(A, nat)](maps[(term[A, B], term[A, B]),
                          (A, nat)]((a: (term[A, B], term[A, B])) =>
                                      funas_root_rule_list[A, B](a),
                                     r))

def sem_lab_root_proc[A : equal : show, B, C,
                       D : equal : show](ld: A => (A, B),
  i: dpp_ops_ext[C, A, D, Unit], valid: sum[(List[Char]) => List[Char], Unit],
  check_Q:
    (List[term[A, D]]) =>
      (List[term[A, D]]) => sum[(List[Char]) => List[Char], Unit],
  check_laba:
    (List[(term[A, D], term[A, D])]) =>
      (List[(term[A, D], term[A, D])]) => sum[(List[Char]) => List[Char], Unit],
  check_lab:
    (List[(term[A, D], term[A, D])]) =>
      (List[(term[A, D], term[A, D])]) => sum[(List[Char]) => List[Char], Unit],
  check_model_lab:
    (List[(term[A, D], term[A, D])]) =>
      (List[(term[A, D], term[A, D])]) => sum[(List[Char]) => List[Char], Unit],
  lPAll: List[(term[A, D], term[A, D])], lQ: List[term[A, D]],
  lRAll: List[(term[A, D], term[A, D])], dpp: C):
      sum[(List[Char]) => List[Char], C] =
  {
    val r: List[(term[A, D], term[A, D])] = (rc[C, A, D, Unit](i)).apply(dpp)
    val rw: List[(term[A, D], term[A, D])] = (rwc[C, A, D, Unit](i)).apply(dpp)
    val pw: List[(term[A, D], term[A, D])] = (pwa[C, A, D, Unit](i)).apply(dpp)
    val p: List[(term[A, D], term[A, D])] = (pa[C, A, D, Unit](i)).apply(dpp)
    val (lP, (lPw, _)):
          (List[(term[A, D], term[A, D])],
            (List[(term[A, D], term[A, D])], List[(term[A, D], term[A, D])]))
      = model_splitter[A, B, D](ld, lPAll, pw)
    val (lR, (lRw, _)):
          (List[(term[A, D], term[A, D])],
            (List[(term[A, D], term[A, D])], List[(term[A, D], term[A, D])]))
      = model_splitter[A, B, D](ld, lRAll, rw);
    bindb[(List[Char]) => List[Char], Unit,
           Unit](valid,
                  (_: Unit) =>
                    bindb[(List[Char]) => List[Char], Unit,
                           Unit](catcha[((term[A, D], term[A, D]),
  (List[Char]) => List[Char]),
 Unit,
 (List[Char]) =>
   List[Char]](forallm[(term[A, D], term[A, D]),
                        (List[Char]) =>
                          List[Char]]((a: (term[A, D], term[A, D])) =>
{
  val (l, ra): (term[A, D], term[A, D]) = a;
  bindb[(List[Char]) => List[Char], Unit,
         Unit](check_no_var[A, D](l),
                (_: Unit) =>
                  bindb[(List[Char]) => List[Char], Unit,
                         Unit](check_no_var[A, D](ra),
                                (_: Unit) =>
                                  check_no_defined_root[A,
                 D]((fn: (A, nat)) =>
                      ! (nulla[(term[A, D],
                                 term[A, D])]((rules_mapb[C, A, D,
                   Unit](i)).apply(dpp).apply(fn))),
                     ra)))
},
                                       (pairsa[C, A, D, Unit](i)).apply(dpp)),
                (x: ((term[A, D], term[A, D]), (List[Char]) => List[Char])) =>
                  Inl[(List[Char]) => List[Char],
                       Unit](snd[(term[A, D], term[A, D]),
                                  (List[Char]) => List[Char]](x))),
                                  (_: Unit) =>
                                    {
                                      val q: List[term[A, D]] =
(qc[C, A, D, Unit](i)).apply(dpp);
                                      catcha[(List[Char]) => List[Char], Unit,
      (List[Char]) =>
        List[Char]](bindb[(List[Char]) => List[Char], Unit,
                           Unit]((check_Q(lQ))(q),
                                  (_: Unit) =>
                                    bindb[(List[Char]) => List[Char], Unit,
   Unit](check_sl_Qa[A, B, D](ld, lQ, q),
          (_: Unit) =>
            bindb[(List[Char]) => List[Char], Unit,
                   Unit]((check_laba(lP))(p),
                          (_: Unit) =>
                            bindb[(List[Char]) => List[Char], Unit,
                                   Unit]((check_laba(lPw))(pw),
  (_: Unit) =>
    bindb[(List[Char]) => List[Char], Unit,
           Unit]((check_model_lab(lR))(r),
                  (_: Unit) =>
                    bindb[(List[Char]) => List[Char], Unit,
                           Unit]((check_model_lab(lRw))(rw),
                                  (_: Unit) =>
                                    bindb[(List[Char]) => List[Char], Unit,
   Unit]((check_lab(lR))(r), (_: Unit) => (check_lab(lRw))(rw)))))))),
                     (x: (List[Char]) => List[Char]) =>
                       Inl[(List[Char]) => List[Char],
                            Unit](comp[List[Char], List[Char],
List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'd',
                                     'u', 'r', 'i', 'n', 'g', ' ', 'l', 'a',
                                     'b', 'e', 'l', 'i', 'n', 'g', ':')),
             comp[List[Char], List[Char], List[Char]](shows_nl, x))))
                                    }))
      match {
      case Inl(a) => Inl[(List[Char]) => List[Char], C](a)
      case Inr(_) =>
        Inr[C, (List[Char]) =>
                 List[Char]]((mkc[C, A, D,
                                   Unit](i)).apply(lP).apply(lPw).apply(lQ).apply(lR).apply(lRw))
    }
  }

def check_sl_rule_root[A : show, B : show, C, D : equal : show,
                        E : equal : show](i: A => (List[B]) => B,
   la: A => (List[B]) => C, l: A => (List[B]) => C, lc: A => nat => C => D,
   c: List[B], lR: List[(term[D, E], term[D, E])],
   lr: (term[A, E], term[A, E])):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(E => B, (List[Char]) => List[Char]), Unit,
          (List[Char]) =>
            List[Char]](forallm[E => B,
                                 (List[Char]) =>
                                   List[Char]]((alpha: E => B) =>
         {
           val laa: (term[A, E]) => term[D, E] =
             (a: term[A, E]) => lab_root[A, B, C, D, E](i, la, l, lc, alpha, a)
           val lb: term[D, E] = laa(fst[term[A, E], term[A, E]](lr))
           val r: term[D, E] = laa(snd[term[A, E], term[A, E]](lr));
           check[(List[Char]) =>
                   List[Char]](membera[(term[D, E], term[D, E])](lR, (lb, r)),
                                comp[List[Char], List[Char],
                                      List[Char]](shows_prec_list[Char](Zero_nat,
                                 List('l', 'a', 'b', 'e', 'l', 'e', 'd', ' ',
                                       'r', 'u', 'l', 'e', ' ')),
           comp[List[Char], List[Char],
                 List[Char]](shows_rule[D,
 E]((a: D) => (b: List[Char]) => shows_prec[D](Zero_nat, a, b),
     (a: E) => (b: List[Char]) => shows_prec[E](Zero_nat, a, b),
     List(' ', '-', '>', ' '), (lb, r)),
                              shows_prec_list[Char](Zero_nat,
             List(' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g')))))
         },
        mapa[List[(E, B)],
              E => B]((a: List[(E, B)]) => (b: E) => fun_of[E, B](a, b),
                       enum_vectors[B, E](c, vars_rule_list[A, E](lr)))),
                         (x: (E => B, (List[Char]) => List[Char])) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](snd[E => B,
   (List[Char]) => List[Char]](x)))

def check_sl_lab_root_trs[A : show, B : show, C, D : equal : show,
                           E : equal : show](i: A => (List[B]) => B,
      la: A => (List[B]) => C, l: A => (List[B]) => C, c: List[B],
      label: A => nat => C => D, lP: List[(term[D, E], term[D, E])],
      p: List[(term[A, E], term[A, E])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[((term[A, E], term[A, E]), (List[Char]) => List[Char]), Unit,
          (List[Char]) =>
            List[Char]](forallm[(term[A, E], term[A, E]),
                                 (List[Char]) =>
                                   List[Char]]((a: (term[A, E], term[A, E])) =>
         check_sl_rule_root[A, B, C, D, E](i, la, l, label, c, lP, a),
        p),
                         (x: ((term[A, E], term[A, E]),
                               (List[Char]) => List[Char]))
                           => Inl[(List[Char]) => List[Char],
                                   Unit](snd[(term[A, E], term[A, E]),
      (List[Char]) => List[Char]](x)))

def sem_lab_fin_root_proc[A : equal : show, B, C, D : equal : show,
                           E : equal : show](lc: A => nat => B => A,
      ld: A => (A, B), i: dpp_ops_ext[C, A, D, Unit],
      gen: (List[(A, nat)]) =>
             (List[(A, nat)]) =>
               sum[(List[Char]) => List[Char], sl_ops_ext[A, E, B, D, Unit]],
      lPAll: List[(term[A, D], term[A, D])], lQ: List[term[A, D]],
      lRAll: List[(term[A, D], term[A, D])], dp: C):
      sum[(List[Char]) => List[Char], C] =
  {
    val pairs: List[(term[A, D], term[A, D])] =
      (pairsa[C, A, D, Unit](i)).apply(dp);
    bindb[(List[Char]) => List[Char], sl_ops_ext[A, E, B, D, Unit],
           C]((gen(list_union[(A, nat)](funas_trs_list[A,
                D]((rulesd[C, A, D, Unit](i)).apply(dp)),
 funas_args_trs_list[A, D](pairs))))(funas_root_rules_list[A, D](pairs)),
               (ops: sl_ops_ext[A, E, B, D, Unit]) =>
                 {
                   val check_q:
                         (List[term[A, D]]) =>
                           (List[term[A, D]]) =>
                             sum[(List[Char]) => List[Char], Unit]
                     = (a: List[term[A, D]]) =>
                         (b: List[term[A, D]]) =>
                           check_sl_Q[A, E, B,
                                       D](sl_I[A, E, B, D, Unit](ops),
   sl_L[A, E, B, D, Unit](ops), lc, sl_C[A, E, B, D, Unit](ops), a, b)
                   val check_ml:
                         (List[(term[A, D], term[A, D])]) =>
                           (List[(term[A, D], term[A, D])]) =>
                             sum[(List[Char]) => List[Char], Unit]
                     = (a: List[(term[A, D], term[A, D])]) =>
                         (b: List[(term[A, D], term[A, D])]) =>
                           check_sl_model_lab_trs[A, E, B,
           D](sl_I[A, E, B, D, Unit](ops), sl_L[A, E, B, D, Unit](ops),
               sl_C[A, E, B, D, Unit](ops), (aa: E) => (ba: E) => eq[E](aa, ba),
               lc, a, b)
                   val check_l:
                         (List[(term[A, D], term[A, D])]) =>
                           (List[(term[A, D], term[A, D])]) =>
                             sum[(List[Char]) => List[Char], Unit]
                     = (a: List[(term[A, D], term[A, D])]) =>
                         (b: List[(term[A, D], term[A, D])]) =>
                           check_sl_lab_root_trs[A, E, B, A,
          D](sl_I[A, E, B, D, Unit](ops), sl_L[A, E, B, D, Unit](ops),
              sl_La[A, E, B, D, Unit](ops), sl_C[A, E, B, D, Unit](ops), lc, a,
              b)
                   val check_la:
                         (List[(term[A, D], term[A, D])]) =>
                           (List[(term[A, D], term[A, D])]) =>
                             sum[(List[Char]) => List[Char], Unit]
                     = (a: List[(term[A, D], term[A, D])]) =>
                         (b: List[(term[A, D], term[A, D])]) =>
                           check_sl_lab[A, E, B,
 D](sl_I[A, E, B, D, Unit](ops), sl_L[A, E, B, D, Unit](ops), lc,
     sl_C[A, E, B, D, Unit](ops), a, b);
                   sem_lab_root_proc[A, B, C,
                                      D](ld, i,
  Inr[Unit, (List[Char]) => List[Char]](()), check_q, check_l, check_la,
  check_ml, lPAll, lQ, lRAll, dp)
                 })
  }

def sl_LS[A, B, C, D, E](x0: sl_ops_ext[A, B, C, D, E]):
      A => nat => C => Boolean =
  x0 match {
  case Sl_ops_ext(sl_La, sl_LSa, sl_I, sl_C, sl_c, sl_check_decr, sl_L, sl_LS,
                   sl_lgen, sl_LS_gen, more)
    => sl_LSa
}

def sl_lgen[A, B, C, D, E](x0: sl_ops_ext[A, B, C, D, E]): C => List[C] = x0
  match {
  case Sl_ops_ext(sl_La, sl_LSa, sl_I, sl_C, sl_c, sl_check_decr, sl_L, sl_LS,
                   sl_lgen, sl_LS_gen, more)
    => sl_lgen
}

def sl_LS_gen[A, B, C, D, E](x0: sl_ops_ext[A, B, C, D, E]):
      A => nat => List[C] =
  x0 match {
  case Sl_ops_ext(sl_La, sl_LSa, sl_I, sl_C, sl_c, sl_check_decr, sl_L, sl_LS,
                   sl_lgen, sl_LS_gen, more)
    => sl_LS_gen
}

def lge_to_lgr[A, B : equal](lge: A => nat => B => B => Boolean,
                              ls: A => nat => B => Boolean, f: A, n: nat):
      B => B => Boolean =
  {
    val lSfn: B => Boolean = (ls(f))(n)
    val lgefn: B => B => Boolean = (lge(f))(n);
    (l: B) =>
      (la: B) =>
        ! (eq[B](l, la)) &&
          (member[B](l, lSfn) && (member[B](la, lSfn) && (lgefn(l))(la)))
  }

def check_sl_decr_rule[A : equal, B : equal, C : equal,
                        D : equal](lc: A => nat => B => C, ld: C => (A, B),
                                    ls: A => nat => B => Boolean,
                                    lge: A => nat => B => B => Boolean,
                                    x4: (term[C, D], term[C, D])):
      Boolean =
  (lc, ld, ls, lge, x4) match {
  case (lc, ld, ls, lge, (Fun(lf, ts), Fun(lg, us))) =>
    {
      val (f, l1): (A, B) = ld(lf)
      val (g, l2): (A, B) = ld(lg)
      val n: nat = size_list[term[C, D]](ts);
      eq[A](f, g) &&
        (equal_lista[term[C, D]](ts, us) &&
          (eq[C](lf, ((lc(f))(n))(l1)) &&
            (eq[C](lg, ((lc(f))(n))(l2)) &&
              (member[B](l1, (ls(f))(n)) &&
                (member[B](l2, (ls(f))(n)) &&
                  (lge_to_lgr[A, B](lge, ls, f, n)).apply(l1).apply(l2))))))
    }
  case (uu, uv, uw, ux, (Var(vb), va)) => false
  case (uu, uv, uw, ux, (v, Var(vb))) => false
}

def check_sl_decr[A : equal, B : equal, C : equal : show,
                   D : equal : show](lc: A => nat => B => C, ld: C => (A, B),
                                      ls: A => nat => B => Boolean,
                                      lge: A => nat => B => B => Boolean,
                                      d: List[(term[C, D], term[C, D])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[((term[C, D], term[C, D]), (List[Char]) => List[Char]), Unit,
          (List[Char]) =>
            List[Char]](forallm[(term[C, D], term[C, D]),
                                 (List[Char]) =>
                                   List[Char]]((lr: (term[C, D], term[C, D])) =>
         check[(List[Char]) =>
                 List[Char]](check_sl_decr_rule[A, B, C,
         D](lc, ld, ls, lge, lr),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_rule[C,
                    D]((a: C) =>
                         (b: List[Char]) => shows_prec[C](Zero_nat, a, b),
                        (a: D) =>
                          (b: List[Char]) => shows_prec[D](Zero_nat, a, b),
                        List(' ', '-', '>', ' '), lr),
         shows_prec_list[Char](Zero_nat,
                                List(' ', 'i', 's', ' ', 'n', 'o', 't', ' ',
                                      'a', ' ', 'd', 'e', 'c', 'r', 'e', 'a',
                                      's', 'i', 'n', 'g', ' ', 'r', 'u', 'l',
                                      'e')))),
        d),
                         (x: ((term[C, D], term[C, D]),
                               (List[Char]) => List[Char]))
                           => Inl[(List[Char]) => List[Char],
                                   Unit](snd[(term[C, D], term[C, D]),
      (List[Char]) => List[Char]](x)))

def sl_check_decr[A, B, C, D, E](x0: sl_ops_ext[A, B, C, D, E]):
      (List[(term[A, D], term[A, D])]) =>
        sum[(List[Char]) => List[Char], Unit] =
  x0 match {
  case Sl_ops_ext(sl_La, sl_LSa, sl_I, sl_C, sl_c, sl_check_decr, sl_L, sl_LS,
                   sl_lgen, sl_LS_gen, more)
    => sl_check_decr
}

def check_wf_sym_F_all[A, B,
                        C : equal : show](lc: A => nat => B => C,
   ld: C => (A, B), ls: A => nat => B => Boolean):
      ((C, nat)) => sum[(List[Char]) => List[Char], Unit] =
  (a: (C, nat)) =>
    {
      val (lf, n): (C, nat) = a
      val (f, l): (A, B) = ld(lf);
      check[(List[Char]) =>
              List[Char]](member[B](l, (ls(f))(n)) &&
                            eq[C](lf, ((lc(f))(n))(l)),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_prec_list[Char](Zero_nat,
                            List('l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 's',
                                  'y', 'm', 'b', 'o', 'l', ' ')),
      comp[List[Char], List[Char],
            List[Char]]((aa: List[Char]) => shows_prec[C](Zero_nat, lf, aa),
                         shows_prec_list[Char](Zero_nat,
        List(' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e', 'd')))))
    }

def check_wf_terms_F_all[A, B, C : equal : show,
                          D](lc: A => nat => B => C, ld: C => (A, B),
                              ls: A => nat => B => Boolean, lt: term[C, D]):
      sum[(List[Char]) => List[Char], Unit] =
  {
    val lfs: List[(C, nat)] = funas_term_list[C, D].apply(lt);
    catcha[((C, nat), (List[Char]) => List[Char]), Unit,
            (List[Char]) =>
              List[Char]](forallm[(C, nat),
                                   (List[Char]) =>
                                     List[Char]](check_wf_sym_F_all[A, B,
                             C](lc, ld, ls),
          lfs),
                           (x: ((C, nat), (List[Char]) => List[Char])) =>
                             Inl[(List[Char]) => List[Char],
                                  Unit](snd[(C, nat),
     (List[Char]) => List[Char]](x)))
  }

def check_Lab_all_trs[A : equal : show, B,
                       C : equal : show](lc: A => nat => B => A,
  ld: A => (A, B), ls: A => nat => B => Boolean,
  lR: List[(term[A, C], term[A, C])], r: List[(term[A, C], term[A, C])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[((term[A, C], term[A, C]), (List[Char]) => List[Char]), Unit,
          (List[Char]) =>
            List[Char]](forallm[(term[A, C], term[A, C]),
                                 (List[Char]) =>
                                   List[Char]]((a: (term[A, C], term[A, C])) =>
         {
           val (l, ra): (term[A, C], term[A, C]) = a;
           catcha[(List[Char]) => List[Char], Unit,
                   (List[Char]) =>
                     List[Char]](bindb[(List[Char]) => List[Char], Unit,
Unit](check_wf_terms_F_all[A, B, A, C](lc, ld, ls, ra),
       (_: Unit) =>
         check[(List[Char]) =>
                 List[Char]](membera[(term[A, C],
                                       term[A,
     C])](r, map_rule[A, A, C]((lf: A) => fst[A, B](ld(lf)), (l, ra))),
                              shows_prec_list[Char](Zero_nat,
             List('u', 'n', 'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g', ' ', 'o',
                   'f', ' ', 't', 'h', 'e', ' ', 'r', 'u', 'l', 'e', ' ', 'd',
                   'o', 'e', 's', ' ', 'n', 'o', 't', ' ', 'y', 'i', 'e', 'l',
                   'd', ' ', 'o', 'r', 'i', 'g', 'i', 'n', 'a', 'l', ' ', 'r',
                   'u', 'l', 'e')))),
                                  (x: (List[Char]) => List[Char]) =>
                                    Inl[(List[Char]) => List[Char],
 Unit](comp[List[Char], List[Char],
             List[Char]](shows_prec_list[Char](Zero_nat,
        List('p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i', 't', 'h', ' ',
              'l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'r', 'u', 'l', 'e')),
                          comp[List[Char], List[Char],
                                List[Char]](shows_rule[A,
                C]((aa: A) => (b: List[Char]) => shows_prec[A](Zero_nat, aa, b),
                    (aa: C) =>
                      (b: List[Char]) => shows_prec[C](Zero_nat, aa, b),
                    List(' ', '-', '>', ' '), (l, ra)),
     comp[List[Char], List[Char], List[Char]](shows_nl, x)))))
         },
        lR),
                         (x: ((term[A, C], term[A, C]),
                               (List[Char]) => List[Char]))
                           => Inl[(List[Char]) => List[Char],
                                   Unit](snd[(term[A, C], term[A, C]),
      (List[Char]) => List[Char]](x)))

def check_sl_rule_all_ass[A : show, B : show, C, D : equal : show,
                           E : equal : show](i: A => (List[B]) => B,
      lb: A => (List[B]) => C, la: A => (List[B]) => C, lc: A => nat => C => D,
      gen_smaller: C => List[C], lR: List[(term[D, E], term[D, E])],
      alpha: E => B, x7: (term[A, E], term[A, E])):
      sum[(List[Char]) => List[Char], Unit] =
  (i, lb, la, lc, gen_smaller, lR, alpha, x7) match {
  case (i, lb, la, lc, gen_smaller, lR, alpha, (l, Fun(f, ts))) =>
    {
      val ll: term[D, E] = lab_root[A, B, C, D, E](i, lb, la, lc, alpha, l)
      val clts: List[(B, term[D, E])] =
        mapa[term[A, E],
              (B, term[D, E])]((a: term[A, E]) =>
                                 eval_lab[A, B, C, D, E](i, lb, lc, alpha, a),
                                ts)
      val lts: List[term[D, E]] =
        mapa[(B, term[D, E]),
              term[D, E]]((a: (B, term[D, E])) => snd[B, term[D, E]](a), clts)
      val ld: C =
        (la(f))(mapa[(B, term[D, E]),
                      B]((a: (B, term[D, E])) => fst[B, term[D, E]](a), clts))
      val n: nat = size_list[term[A, E]](ts)
      val small: List[C] = gen_smaller(ld);
      catcha[(C, (List[Char]) => List[Char]), Unit,
              (List[Char]) =>
                List[Char]](forallm[C, (List[Char]) =>
 List[Char]]((le: C) =>
               check[(List[Char]) =>
                       List[Char]](membera[(term[D, E],
     term[D, E])](lR, (ll, Fun[D, E](((lc(f))(n))(le), lts))),
                                    comp[List[Char], List[Char],
  List[Char]](shows_string.apply(List('l', 'a', 'b', 'e', 'l', 'e', 'd', ' ',
                                       'r', 'u', 'l', 'e', ' ')),
               comp[List[Char], List[Char],
                     List[Char]](shows_rule[D,
     E]((a: D) => (b: List[Char]) => shows_prec[D](Zero_nat, a, b),
         (a: E) => (b: List[Char]) => shows_prec[E](Zero_nat, a, b),
         List(' ', '-', '>', ' '), (ll, Fun[D, E](((lc(f))(n))(le), lts))),
                                  shows_string.apply(List(' ', 'm', 'i', 's',
                   's', 'i', 'n', 'g'))))),
              small),
                             (x: (C, (List[Char]) => List[Char])) =>
                               Inl[(List[Char]) => List[Char],
                                    Unit](snd[C,
       (List[Char]) => List[Char]](x)))
    }
  case (i, lb, la, lc, gen_smaller, lR, alpha, (l, Var(x))) =>
    {
      val ll: term[D, E] = lab_root[A, B, C, D, E](i, lb, la, lc, alpha, l)
      val lr: term[D, E] =
        lab_root[A, B, C, D, E](i, lb, la, lc, alpha, Var[E, A](x));
      check[(List[Char]) =>
              List[Char]](membera[(term[D, E], term[D, E])](lR, (ll, lr)),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('l', 'a',
                              'b', 'e', 'l', 'e', 'd', ' ', 'r', 'u', 'l', 'e',
                              ' ')),
      comp[List[Char], List[Char],
            List[Char]](shows_rule[D, E]((a: D) =>
   (b: List[Char]) => shows_prec[D](Zero_nat, a, b),
  (a: E) => (b: List[Char]) => shows_prec[E](Zero_nat, a, b),
  List(' ', '-', '>', ' '), (ll, lr)),
                         shows_string.apply(List(' ', 'm', 'i', 's', 's', 'i',
          'n', 'g')))))
    }
}

def check_sl_rule_all[A : show, B : show, C, D : equal : show,
                       E : equal : show](i: A => (List[B]) => B,
  la: A => (List[B]) => C, l: A => (List[B]) => C, lc: A => nat => C => D,
  c: List[B], gen_smaller: C => List[C], lR: List[(term[D, E], term[D, E])],
  lr: (term[A, E], term[A, E])):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(E => B, (List[Char]) => List[Char]), Unit,
          (List[Char]) =>
            List[Char]](forallm[E => B,
                                 (List[Char]) =>
                                   List[Char]]((alpha: E => B) =>
         check_sl_rule_all_ass[A, B, C, D,
                                E](i, la, l, lc, gen_smaller, lR, alpha, lr),
        mapa[List[(E, B)],
              E => B]((a: List[(E, B)]) => (b: E) => fun_of[E, B](a, b),
                       enum_vectors[B, E](c, vars_rule_list[A, E](lr)))),
                         (x: (E => B, (List[Char]) => List[Char])) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](snd[E => B,
   (List[Char]) => List[Char]](x)))

def check_sl_lab_all_trs[A : show, B : show, C, D : equal : show,
                          E : equal : show](i: A => (List[B]) => B,
     la: A => (List[B]) => C, l: A => (List[B]) => C, c: List[B],
     gen: C => List[C], label: A => nat => C => D,
     lP: List[(term[D, E], term[D, E])], p: List[(term[A, E], term[A, E])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[((term[A, E], term[A, E]), (List[Char]) => List[Char]), Unit,
          (List[Char]) =>
            List[Char]](forallm[(term[A, E], term[A, E]),
                                 (List[Char]) =>
                                   List[Char]]((a: (term[A, E], term[A, E])) =>
         check_sl_rule_all[A, B, C, D, E](i, la, l, label, c, gen, lP, a),
        p),
                         (x: ((term[A, E], term[A, E]),
                               (List[Char]) => List[Char]))
                           => Inl[(List[Char]) => List[Char],
                                   Unit](snd[(term[A, E], term[A, E]),
      (List[Char]) => List[Char]](x)))

def flatten_term_enum[A, B](x0: term[List[A], B]): List[term[A, B]] = x0 match {
  case Var(x) => List(Var[B, A](x))
  case Fun(fs, ts) =>
    {
      val lts: List[List[term[A, B]]] =
        mapa[term[List[A], B],
              List[term[A, B]]]((a: term[List[A], B]) =>
                                  flatten_term_enum[A, B](a),
                                 ts)
      val ss: List[List[term[A, B]]] = concat_lists[term[A, B]](lts);
      maps[A, term[A, B]]((f: A) =>
                            mapa[List[term[A, B]],
                                  term[A,
B]]((a: List[term[A, B]]) => Fun[A, B](f, a), ss),
                           fs)
    }
}

def lab_lhss_more_impl[A, B, C,
                        D](lc: A => nat => B => C, lS_gen: A => nat => List[B],
                            q: List[term[A, D]]):
      List[term[C, D]] =
  {
    val f_all: A => nat => List[C] =
      (f: A) => (n: nat) => mapa[B, C]((lc(f))(n), (lS_gen(f))(n));
    maps[term[A, D],
          term[C, D]]((qa: term[A, D]) =>
                        flatten_term_enum[C,
   D](map_term_wa[A, List[C], D](f_all, qa)),
                       q)
  }

def check_sl_lab_lhss_more[A : equal : show, B,
                            C : equal : show](lc: A => nat => B => A,
       lS_gen: A => nat => List[B], lQ: List[term[A, C]], q: List[term[A, C]]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[term[A, C], Unit,
          (List[Char]) =>
            List[Char]](check_subseteq[term[A,
     C]](lab_lhss_more_impl[A, B, A, C](lc, lS_gen, q), lQ),
                         (x: term[A, C]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_term[A, C]((a: A) =>
                                   (b: List[Char]) =>
                                     shows_prec[A](Zero_nat, a, b),
                                  (a: C) =>
                                    (b: List[Char]) =>
                                      shows_prec[C](Zero_nat, a, b),
                                  x),
                 shows_prec_list[Char](Zero_nat,
List(' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g', ' ', 'i', 'n', ' ',
      'l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'Q')))))

def quasi_splitter[A : equal, B,
                    C : equal](ld: A => (A, B),
                                lAll: List[(term[A, C], term[A, C])],
                                uRw: List[(term[A, C], term[A, C])]):
      (List[(term[A, C], term[A, C])],
        (List[(term[A, C], term[A, C])], List[(term[A, C], term[A, C])])) =
  {
    val unlab: A => A = (lf: A) => fst[A, B](ld(lf))
    val la: List[((term[A, C], term[A, C]), (term[A, C], term[A, C]))] =
      mapa[(term[A, C], term[A, C]),
            ((term[A, C], term[A, C]),
              (term[A, C],
                term[A, C]))]((r: (term[A, C], term[A, C])) =>
                                (r, map_rule[A, A, C](unlab, r)),
                               lAll)
    val (d, nD):
          (List[((term[A, C], term[A, C]), (term[A, C], term[A, C]))],
            List[((term[A, C], term[A, C]), (term[A, C], term[A, C]))])
      = partition[((term[A, C], term[A, C]),
                    (term[A, C],
                      term[A, C]))]((a: ((term[A, C], term[A, C]),
  (term[A, C], term[A, C])))
                                      => {
   val (r, ur): ((term[A, C], term[A, C]), (term[A, C], term[A, C])) = a;
   equal_terma[A, C](fst[term[A, C], term[A, C]](ur),
                      snd[term[A, C], term[A, C]](ur)) &&
     ! (equal_terma[A, C](fst[term[A, C], term[A, C]](r),
                           snd[term[A, C], term[A, C]](r)))
 },
                                     la)
    val (rw, r):
          (List[((term[A, C], term[A, C]), (term[A, C], term[A, C]))],
            List[((term[A, C], term[A, C]), (term[A, C], term[A, C]))])
      = partition[((term[A, C], term[A, C]),
                    (term[A, C],
                      term[A, C]))]((a: ((term[A, C], term[A, C]),
  (term[A, C], term[A, C])))
                                      => {
   val (_, aa): ((term[A, C], term[A, C]), (term[A, C], term[A, C])) = a;
   membera[(term[A, C], term[A, C])](uRw, aa)
 },
                                     nD);
    (mapa[((term[A, C], term[A, C]), (term[A, C], term[A, C])),
           (term[A, C],
             term[A, C])]((a: ((term[A, C], term[A, C]),
                                (term[A, C], term[A, C])))
                            => fst[(term[A, C], term[A, C]),
                                    (term[A, C], term[A, C])](a),
                           r),
      (mapa[((term[A, C], term[A, C]), (term[A, C], term[A, C])),
             (term[A, C],
               term[A, C])]((a: ((term[A, C], term[A, C]),
                                  (term[A, C], term[A, C])))
                              => fst[(term[A, C], term[A, C]),
                                      (term[A, C], term[A, C])](a),
                             rw),
        mapa[((term[A, C], term[A, C]), (term[A, C], term[A, C])),
              (term[A, C],
                term[A, C])]((a: ((term[A, C], term[A, C]),
                                   (term[A, C], term[A, C])))
                               => fst[(term[A, C], term[A, C]),
                                       (term[A, C], term[A, C])](a),
                              d)))
  }

def sem_lab_quasi_root_proc[A : equal : show, B, C,
                             D : equal : show](ld: A => (A, B),
        i: dpp_ops_ext[C, A, D, Unit],
        valid: sum[(List[Char]) => List[Char], Unit],
        check_decra:
          (List[(term[A, D], term[A, D])]) =>
            sum[(List[Char]) => List[Char], Unit],
        check_decr:
          (List[(term[A, D], term[A, D])]) =>
            sum[(List[Char]) => List[Char], Unit],
        check_lhss_more:
          (List[term[A, D]]) =>
            (List[term[A, D]]) => sum[(List[Char]) => List[Char], Unit],
        check_lab_all:
          (List[(term[A, D], term[A, D])]) =>
            (List[(term[A, D], term[A, D])]) =>
              sum[(List[Char]) => List[Char], Unit],
        check_lab_all_trs:
          (List[(term[A, D], term[A, D])]) =>
            (List[(term[A, D], term[A, D])]) =>
              sum[(List[Char]) => List[Char], Unit],
        check_model_lab:
          (List[(term[A, D], term[A, D])]) =>
            (List[(term[A, D], term[A, D])]) =>
              sum[(List[Char]) => List[Char], Unit],
        lPAll: List[(term[A, D], term[A, D])], lQ: List[term[A, D]],
        lRAll: List[(term[A, D], term[A, D])], dpp: C):
      sum[(List[Char]) => List[Char], C] =
  {
    val r: List[(term[A, D], term[A, D])] = (rc[C, A, D, Unit](i)).apply(dpp)
    val rw: List[(term[A, D], term[A, D])] = (rwc[C, A, D, Unit](i)).apply(dpp)
    val pw: List[(term[A, D], term[A, D])] = (pwa[C, A, D, Unit](i)).apply(dpp)
    val p: List[(term[A, D], term[A, D])] = (pa[C, A, D, Unit](i)).apply(dpp)
    val (lP, (lPw, _)):
          (List[(term[A, D], term[A, D])],
            (List[(term[A, D], term[A, D])], List[(term[A, D], term[A, D])]))
      = model_splitter[A, B, D](ld, lPAll, pw)
    val (lR, (lRw, d)):
          (List[(term[A, D], term[A, D])],
            (List[(term[A, D], term[A, D])], List[(term[A, D], term[A, D])]))
      = quasi_splitter[A, B, D](ld, lRAll, rw);
    bindb[(List[Char]) => List[Char], Unit,
           Unit](valid,
                  (_: Unit) =>
                    bindb[(List[Char]) => List[Char], Unit,
                           Unit](catcha[((term[A, D], term[A, D]),
  (List[Char]) => List[Char]),
 Unit,
 (List[Char]) =>
   List[Char]](forallm[(term[A, D], term[A, D]),
                        (List[Char]) =>
                          List[Char]]((a: (term[A, D], term[A, D])) =>
{
  val (l, ra): (term[A, D], term[A, D]) = a;
  bindb[(List[Char]) => List[Char], Unit,
         Unit](check_no_var[A, D](l),
                (_: Unit) =>
                  bindb[(List[Char]) => List[Char], Unit,
                         Unit](check_no_var[A, D](ra),
                                (_: Unit) =>
                                  check_no_defined_root[A,
                 D]((fn: (A, nat)) =>
                      ! (nulla[(term[A, D],
                                 term[A, D])]((rules_mapb[C, A, D,
                   Unit](i)).apply(dpp).apply(fn))),
                     ra)))
},
                                       (pairsa[C, A, D, Unit](i)).apply(dpp)),
                (x: ((term[A, D], term[A, D]), (List[Char]) => List[Char])) =>
                  Inl[(List[Char]) => List[Char],
                       Unit](snd[(term[A, D], term[A, D]),
                                  (List[Char]) => List[Char]](x))),
                                  (_: Unit) =>
                                    {
                                      val q: List[term[A, D]] =
(qc[C, A, D, Unit](i)).apply(dpp);
                                      bindb[(List[Char]) => List[Char], Unit,
     Unit](check_decra(d),
            (_: Unit) =>
              bindb[(List[Char]) => List[Char], Unit,
                     Unit](check_decr(d),
                            (_: Unit) =>
                              bindb[(List[Char]) => List[Char], Unit,
                                     Unit](catcha[(term[A, D],
            (List[Char]) => List[Char]),
           Unit,
           (List[Char]) =>
             List[Char]](forallm[term[A, D],
                                  (List[Char]) =>
                                    List[Char]]((qa: term[A, D]) =>
          check[(List[Char]) =>
                  List[Char]](linear_term[A, D](qa),
                               shows_prec_list[Char](Zero_nat,
              List('Q', ' ', 'm', 'u', 's', 't', ' ', 'n', 'o', 't', ' ', 'c',
                    'o', 'n', 't', 'a', 'i', 'n', ' ', 'n', 'o', 'n', '-', 'l',
                    'i', 'n', 'e', 'a', 'r', ' ', 't', 'e', 'r', 'm', 's'))),
         q),
                          (x: (term[A, D], (List[Char]) => List[Char])) =>
                            Inl[(List[Char]) => List[Char],
                                 Unit](snd[term[A, D],
    (List[Char]) => List[Char]](x))),
    (_: Unit) =>
      catcha[(List[Char]) => List[Char], Unit,
              (List[Char]) =>
                List[Char]](bindb[(List[Char]) => List[Char], Unit,
                                   Unit]((check_lhss_more(lQ))(q),
  (_: Unit) =>
    bindb[(List[Char]) => List[Char], Unit,
           Unit](check_sl_Qa[A, B, D](ld, lQ, q),
                  (_: Unit) =>
                    bindb[(List[Char]) => List[Char], Unit,
                           Unit]((check_lab_all(lP))(p),
                                  (_: Unit) =>
                                    bindb[(List[Char]) => List[Char], Unit,
   Unit]((check_lab_all(lPw))(pw),
          (_: Unit) =>
            bindb[(List[Char]) => List[Char], Unit,
                   Unit]((check_model_lab(lR))(r),
                          (_: Unit) =>
                            bindb[(List[Char]) => List[Char], Unit,
                                   Unit]((check_model_lab(lRw))(rw),
  (_: Unit) =>
    bindb[(List[Char]) => List[Char], Unit,
           Unit]((check_lab_all_trs(lR))(r),
                  (_: Unit) => (check_lab_all_trs(lRw))(rw)))))))),
                             (x: (List[Char]) => List[Char]) =>
                               Inl[(List[Char]) => List[Char],
                                    Unit](comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm',
     ' ', 'd', 'u', 'r', 'i', 'n', 'g', ' ', 'l', 'a', 'b', 'e', 'l', 'i', 'n',
     'g', ':')),
                     comp[List[Char], List[Char], List[Char]](shows_nl, x)))))))
                                    }))
      match {
      case Inl(a) => Inl[(List[Char]) => List[Char], C](a)
      case Inr(_) =>
        Inr[C, (List[Char]) =>
                 List[Char]]((mkc[C, A, D,
                                   Unit](i)).apply(lP).apply(lPw).apply(lQ).apply(lR).apply(lRw ++
              d))
    }
  }

def sem_lab_fin_quasi_root_proc[A : equal : show, B : equal, C : show, D,
                                 E : equal : show](lc: A => nat => B => A,
            ld: A => (A, B), cge: C => C => Boolean,
            lge: A => nat => B => B => Boolean, i: dpp_ops_ext[D, A, E, Unit],
            gen: (List[(A, nat)]) =>
                   (List[(A, nat)]) =>
                     sum[(List[Char]) => List[Char],
                          sl_ops_ext[A, C, B, E, Unit]],
            lPAll: List[(term[A, E], term[A, E])], lQ: List[term[A, E]],
            lRAll: List[(term[A, E], term[A, E])], dp: D):
      sum[(List[Char]) => List[Char], D] =
  {
    val pairs: List[(term[A, E], term[A, E])] =
      (pairsa[D, A, E, Unit](i)).apply(dp);
    bindb[(List[Char]) => List[Char], sl_ops_ext[A, C, B, E, Unit],
           D]((gen(list_union[(A, nat)](funas_trs_list[A,
                E]((rulesd[D, A, E, Unit](i)).apply(dp)),
 funas_args_trs_list[A, E](pairs))))(funas_root_rules_list[A, E](pairs)),
               (ops: sl_ops_ext[A, C, B, E, Unit]) =>
                 {
                   val check_d:
                         (List[(term[A, E], term[A, E])]) =>
                           sum[(List[Char]) => List[Char], Unit]
                     = sl_check_decr[A, C, B, E, Unit](ops)
                   val check_da:
                         (List[(term[A, E], term[A, E])]) =>
                           sum[(List[Char]) => List[Char], Unit]
                     = (a: List[(term[A, E], term[A, E])]) =>
                         check_sl_decr[A, B, A,
E](lc, ld, sl_LS[A, C, B, E, Unit](ops), lge, a)
                   val check_q:
                         (List[term[A, E]]) =>
                           (List[term[A, E]]) =>
                             sum[(List[Char]) => List[Char], Unit]
                     = (a: List[term[A, E]]) =>
                         (b: List[term[A, E]]) =>
                           check_sl_lab_lhss_more[A, B,
           E](lc, sl_LS_gen[A, C, B, E, Unit](ops), a, b)
                   val check_ml:
                         (List[(term[A, E], term[A, E])]) =>
                           (List[(term[A, E], term[A, E])]) =>
                             sum[(List[Char]) => List[Char], Unit]
                     = (a: List[(term[A, E], term[A, E])]) =>
                         (b: List[(term[A, E], term[A, E])]) =>
                           check_sl_model_lab_trs[A, C, B,
           E](sl_I[A, C, B, E, Unit](ops), sl_L[A, C, B, E, Unit](ops),
               sl_C[A, C, B, E, Unit](ops), cge, lc, a, b)
                   val check_l:
                         (List[(term[A, E], term[A, E])]) =>
                           (List[(term[A, E], term[A, E])]) =>
                             sum[(List[Char]) => List[Char], Unit]
                     = (a: List[(term[A, E], term[A, E])]) =>
                         (b: List[(term[A, E], term[A, E])]) =>
                           check_sl_lab_all_trs[A, C, B, A,
         E](sl_I[A, C, B, E, Unit](ops), sl_L[A, C, B, E, Unit](ops),
             sl_La[A, C, B, E, Unit](ops), sl_C[A, C, B, E, Unit](ops),
             sl_lgen[A, C, B, E, Unit](ops), lc, a, b)
                   val check_la:
                         (List[(term[A, E], term[A, E])]) =>
                           (List[(term[A, E], term[A, E])]) =>
                             sum[(List[Char]) => List[Char], Unit]
                     = (a: List[(term[A, E], term[A, E])]) =>
                         (b: List[(term[A, E], term[A, E])]) =>
                           check_Lab_all_trs[A, B,
      E](lc, ld, sl_LS[A, C, B, E, Unit](ops), a, b);
                   sem_lab_quasi_root_proc[A, B, D,
    E](ld, i, Inr[Unit, (List[Char]) => List[Char]](()), check_d, check_da,
        check_q, check_l, check_la, check_ml, lPAll, lQ, lRAll, dp)
                 })
  }

def semlab_fin_proc[A, B : equal : show,
                     C : equal : show](j:
 dpp_ops_ext[A, lab[B, List[nat]], C, Unit],
x1: sl_variant[lab[B, List[nat]], C]):
      (List[(term[lab[B, List[nat]], C], term[lab[B, List[nat]], C])]) =>
        (List[term[lab[B, List[nat]], C]]) =>
          (List[(term[lab[B, List[nat]], C], term[lab[B, List[nat]], C])]) =>
            A => sum[(List[Char]) => List[Char], A] =
  (j, x1) match {
  case (j, Rootlab(None)) =>
    (a: List[(term[lab[B, List[nat]], C], term[lab[B, List[nat]], C])]) =>
      (b: List[term[lab[B, List[nat]], C]]) =>
        (c: List[(term[lab[B, List[nat]], C], term[lab[B, List[nat]], C])]) =>
          (d: A) =>
            sem_lab_fin_proc[lab[B, List[nat]],
                              sum[List[nat], List[lab[B, List[nat]]]], A, C,
                              lab[B, List[nat]]]((aa: lab[B, List[nat]]) =>
           (ba: nat) =>
             (ca: sum[List[nat], List[lab[B, List[nat]]]]) =>
               label[B, List[nat]](aa, ba, ca),
          (aa: lab[B, List[nat]]) => label_decomp[B, List[nat]](aa), j,
          slm_gen_to_sl_gen[lab[B, List[nat]], lab[B, List[nat]],
                             sum[List[nat], List[lab[B, List[nat]]]],
                             C]((aa: List[(lab[B, List[nat]], nat)]) =>
                                  (ba: List[(lab[B, List[nat]], nat)]) =>
                                    rl_slm[B, List[nat]](None, aa, ba)),
          a, b, c, d)
  case (j, Rootlab(Some(d))) =>
    (a: List[(term[lab[B, List[nat]], C], term[lab[B, List[nat]], C])]) =>
      (b: List[term[lab[B, List[nat]], C]]) =>
        (c: List[(term[lab[B, List[nat]], C], term[lab[B, List[nat]], C])]) =>
          (e: A) =>
            sem_lab_fin_root_proc[lab[B, List[nat]],
                                   sum[List[nat], List[lab[B, List[nat]]]], A,
                                   C, lab[B,
   List[nat]]]((aa: lab[B, List[nat]]) =>
                 (ba: nat) =>
                   (ca: sum[List[nat], List[lab[B, List[nat]]]]) =>
                     label[B, List[nat]](aa, ba, ca),
                (aa: lab[B, List[nat]]) => label_decomp[B, List[nat]](aa), j,
                slm_gen_to_sl_gen[lab[B, List[nat]], lab[B, List[nat]],
                                   sum[List[nat], List[lab[B, List[nat]]]],
                                   C]((aa: List[(lab[B, List[nat]], nat)]) =>
(ba: List[(lab[B, List[nat]], nat)]) =>
  rl_slm[B, List[nat]](Some[(lab[B, List[nat]], nat)](d), aa, ba)),
                a, b, c, e)
  case (j, Finitelab(sli)) =>
    (a: List[(term[lab[B, List[nat]], C], term[lab[B, List[nat]], C])]) =>
      (b: List[term[lab[B, List[nat]], C]]) =>
        (c: List[(term[lab[B, List[nat]], C], term[lab[B, List[nat]], C])]) =>
          (d: A) =>
            sem_lab_fin_proc[lab[B, List[nat]],
                              sum[List[nat], List[lab[B, List[nat]]]], A, C,
                              nat]((aa: lab[B, List[nat]]) =>
                                     (ba: nat) =>
                                       (ca:
  sum[List[nat], List[lab[B, List[nat]]]])
 => label[B, List[nat]](aa, ba, ca),
                                    (aa: lab[B, List[nat]]) =>
                                      label_decomp[B, List[nat]](aa),
                                    j, slm_gen_to_sl_gen[lab[B, List[nat]], nat,
                  sum[List[nat], List[lab[B, List[nat]]]],
                  C]((_: List[(lab[B, List[nat]], nat)]) =>
                       (_: List[(lab[B, List[nat]], nat)]) =>
                         Inr[slm_ops_ext[lab[B, List[nat]], nat,
  sum[List[nat], List[lab[B, List[nat]]]], Unit],
                              (List[Char]) => List[Char]](sli_to_slm[B](sli))),
                                    a, b, c, d)
  case (j, QuasiFinitelab(sli, v)) =>
    (a: List[(term[lab[B, List[nat]], C], term[lab[B, List[nat]], C])]) =>
      (b: List[term[lab[B, List[nat]], C]]) =>
        (c: List[(term[lab[B, List[nat]], C], term[lab[B, List[nat]], C])]) =>
          (d: A) =>
            sem_lab_fin_quasi_root_proc[lab[B, List[nat]],
 sum[List[nat], List[lab[B, List[nat]]]], nat, A,
 C]((aa: lab[B, List[nat]]) =>
      (ba: nat) =>
        (ca: sum[List[nat], List[lab[B, List[nat]]]]) =>
          label[B, List[nat]](aa, ba, ca),
     (aa: lab[B, List[nat]]) => label_decomp[B, List[nat]](aa), qmodel_cge,
     (aa: lab[B, List[nat]]) =>
       (ba: nat) =>
         qmodel_lge[lab[B, List[nat]], nat, List[lab[B, List[nat]]],
                     List[lab[B, List[nat]]]](aa, ba),
     j, (f: List[(lab[B, List[nat]], nat)]) =>
          (g: List[(lab[B, List[nat]], nat)]) => qsli_to_sl[C, B](v, f, g, sli),
     a, b, c, d)
}

def check_critical_pairs_NF[A : equal : show,
                             B : equal : show](r:
         List[(term[A, B], term[A, B])],
        cp: List[(Boolean, (term[A, B], term[A, B]))]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[((Boolean, (term[A, B], term[A, B])), (List[Char]) => List[Char]),
          Unit,
          (List[Char]) =>
            List[Char]](forallm[(Boolean, (term[A, B], term[A, B])),
                                 (List[Char]) =>
                                   List[Char]]((a:
          (Boolean, (term[A, B], term[A, B])))
         => {
              val (_, (s, t)): (Boolean, (term[A, B], term[A, B])) = a;
              (if (equal_terma[A, B](s, t))
                Inr[Unit, (List[Char]) => List[Char]](())
                else catcha[(List[Char]) => List[Char], Unit,
                             (List[Char]) =>
                               List[Char]](check_join_NF[A, B](r, s, t),
    (x: (List[Char]) => List[Char]) =>
      Inl[(List[Char]) => List[Char],
           Unit](comp[List[Char], List[Char],
                       List[Char]](shows_prec_list[Char](Zero_nat,
                  List('p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e',
                        'n', ' ', 'j', 'o', 'i', 'n', 'i', 'n', 'g', ' ', 'c',
                        'r', 'i', 't', 'i', 'c', 'a', 'l', ' ', 'p', 'a', 'i',
                        'r', ' ')),
                                    comp[List[Char], List[Char],
  List[Char]](show_crit_pair[A, B]((s, t)),
               comp[List[Char], List[Char], List[Char]](shows_nl, x))))))
            },
        cp),
                         (x: ((Boolean, (term[A, B], term[A, B])),
                               (List[Char]) => List[Char]))
                           => Inl[(List[Char]) => List[Char],
                                   Unit](snd[(Boolean,
       (term[A, B], term[A, B])),
      (List[Char]) => List[Char]](x)))

def check_prop_rstep[A : equal : show,
                      B : equal : show](pa:
  (term[A, B]) => sum[(List[Char]) => List[Char], Unit],
 r: List[(term[A, B], term[A, B])], p: pos, rule: (term[A, B], term[A, B]),
 s: term[A, B], t: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  bindb[(List[Char]) => List[Char], Unit,
         Unit](check[(List[Char]) =>
                       List[Char]](list_ex[(term[A, B],
     term[A, B])]((a: (term[A, B], term[A, B])) =>
                    eq_rule_mod_vars[A, B](rule, a),
                   r),
                                    comp[List[Char], List[Char],
  List[Char]](shows_rule[A, B]((a: A) =>
                                 (b: List[Char]) =>
                                   shows_prec[A](Zero_nat, a, b),
                                (a: B) =>
                                  (b: List[Char]) =>
                                    shows_prec[B](Zero_nat, a, b),
                                List(' ', '-', '>', ' '), rule),
               comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List(' ', 'i', 's', ' ',
                  'n', 'o', 't', ' ', 'a', ' ', 'r', 'u', 'l', 'e', ' ', 'o',
                  'f')),
                                  comp[List[Char], List[Char],
List[Char]](shows_nl,
             comp[List[Char], List[Char],
                   List[Char]](shows_trs[A,
  B]((a: A) => (b: List[Char]) => shows_prec[A](Zero_nat, a, b),
      (a: B) => (b: List[Char]) => shows_prec[B](Zero_nat, a, b),
      List('r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'y', 's', 't', 'e', 'm',
            ':'),
      List(' ', '-', '>', ' '), r),
                                shows_nl))))),
                (_: Unit) => check_prop_rstep_rule[A, B](pa, p, rule, s, t))

def check_qrstep[A : equal : show,
                  B : equal : show](nf: (term[A, B]) => Boolean):
      (List[(term[A, B], term[A, B])]) =>
        pos =>
          ((term[A, B], term[A, B])) =>
            (term[A, B]) =>
              (term[A, B]) => sum[(List[Char]) => List[Char], Unit] =
  (a: List[(term[A, B], term[A, B])]) =>
    (b: pos) =>
      (c: (term[A, B], term[A, B])) =>
        (d: term[A, B]) =>
          (e: term[A, B]) =>
            check_prop_rstep[A, B]((t: term[A, B]) =>
                                     check[(List[Char]) =>
     List[Char]](nf(t),
                  comp[List[Char], List[Char],
                        List[Char]]((shows_prec_term[A, B](Zero_nat)).apply(t),
                                     shows_prec_list[Char](Zero_nat,
                    List(' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'i', 'n', ' ',
                          'Q', '-', 'n', 'o', 'r', 'm', 'a', 'l', ' ', 'f', 'o',
                          'r', 'm')))),
                                    a, b, c, d, e)

def check_rqrstep[A : equal : show,
                   B : equal : show](nf: (term[A, B]) => Boolean,
                                      r: List[(term[A, B], term[A, B])],
                                      rule: (term[A, B], term[A, B]),
                                      s: term[A, B], t: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  (check_qrstep[A, B](nf)).apply(r).apply(Empty).apply(rule).apply(s).apply(t)

def check_qsteps[A : equal : show,
                  B : equal : show](nf: (term[A, B]) => Boolean,
                                     p: List[(term[A, B], term[A, B])],
                                     r: List[(term[A, B], term[A, B])],
                                     x3: List[(pos,
        ((term[A, B], term[A, B]), (Boolean, term[A, B])))],
                                     s: term[A, B], u: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  (nf, p, r, x3, s, u) match {
  case (nf, p, r, Nil, s, u) =>
    check[(List[Char]) =>
            List[Char]](equal_terma[A, B](s, u),
                         comp[List[Char], List[Char],
                               List[Char]](shows_string.apply(List('t', 'h',
                            'e', ' ', 'l', 'a', 's', 't', ' ', 't', 'e', 'r',
                            'm', ' ', 'o', 'f', ' ', 't', 'h', 'e', ' ', 'r',
                            'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'e', 'q',
                            'u', 'e', 'n', 'c', 'e')),
    comp[List[Char], List[Char],
          List[Char]](shows_nl,
                       comp[List[Char], List[Char],
                             List[Char]]((shows_prec_term[A,
                   B](Zero_nat)).apply(s),
  comp[List[Char], List[Char],
        List[Char]](shows_nl,
                     comp[List[Char], List[Char],
                           List[Char]](shows_string.apply(List('d', 'o', 'e',
                        's', ' ', 'n', 'o', 't', ' ', 'c', 'o', 'r', 'r', 'e',
                        's', 'p', 'o', 'n', 'd', ' ', 't', 'o', ' ', 't', 'h',
                        'e', ' ', 'g', 'o', 'a', 'l', ' ', 't', 'e', 'r', 'm')),
comp[List[Char], List[Char],
      List[Char]](shows_nl,
                   comp[List[Char], List[Char],
                         List[Char]]((shows_prec_term[A, B](Zero_nat)).apply(u),
                                      shows_nl))))))))
  case (nf, p, ra, (uu, (r, (true, t))) :: prts, s, u) =>
    bindb[(List[Char]) => List[Char], Unit,
           Unit](check_rqrstep[A, B](nf, p, r, s, t),
                  (_: Unit) => check_qsteps[A, B](nf, p, ra, prts, t, u))
  case (nf, pa, ra, (p, (r, (false, t))) :: prts, s, u) =>
    bindb[(List[Char]) => List[Char], Unit,
           Unit]((check_qrstep[A, B](nf)).apply(ra).apply(p).apply(r).apply(s).apply(t),
                  (_: Unit) => check_qsteps[A, B](nf, pa, ra, prts, t, u))
}

def check_qrsteps[A : equal : show,
                   B : equal : show](nf: (term[A, B]) => Boolean,
                                      r: List[(term[A, B], term[A, B])],
                                      prts:
List[(pos, ((term[A, B], term[A, B]), term[A, B]))],
                                      s: term[A, B], u: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  check_qsteps[A, B](nf, Nil, r,
                      mapa[(pos, ((term[A, B], term[A, B]), term[A, B])),
                            (pos, ((term[A, B], term[A, B]),
                                    (Boolean,
                                      term[A,
    B])))]((a: (pos, ((term[A, B], term[A, B]), term[A, B]))) =>
             {
               val (p, (ra, t)): (pos, ((term[A, B], term[A, B]), term[A, B])) =
                 a;
               (p, (ra, (false, t)))
             },
            prts),
                      s, u)

def check_join[A : equal : show,
                B : equal : show](r: List[(term[A, B], term[A, B])],
                                   s: term[A, B],
                                   sseq: List[(pos,
        ((term[A, B], term[A, B]), term[A, B]))],
                                   t: term[A, B],
                                   tseq: List[(pos,
        ((term[A, B], term[A, B]), term[A, B]))]):
      sum[(List[Char]) => List[Char], Unit] =
  {
    val u: term[A, B] =
      last[term[A, B]](s :: mapa[(pos, ((term[A, B], term[A, B]), term[A, B])),
                                  term[A,
B]]((a: (pos, ((term[A, B], term[A, B]), term[A, B]))) =>
      {
        val (_, (_, u)): (pos, ((term[A, B], term[A, B]), term[A, B])) = a;
        u
      },
     sseq))
    val v: term[A, B] =
      last[term[A, B]](t :: mapa[(pos, ((term[A, B], term[A, B]), term[A, B])),
                                  term[A,
B]]((a: (pos, ((term[A, B], term[A, B]), term[A, B]))) =>
      {
        val (_, (_, ua)): (pos, ((term[A, B], term[A, B]), term[A, B])) = a;
        ua
      },
     tseq));
    bindb[(List[Char]) => List[Char], Unit,
           Unit](check[(List[Char]) =>
                         List[Char]](equal_terma[A, B](u, v),
                                      comp[List[Char], List[Char],
    List[Char]](shows_prec_list[Char](Zero_nat,
                                       List('t', 'h', 'e', ' ', 'r', 'e', 'w',
     'r', 'i', 't', 'e', ' ', 's', 'e', 'q', 'u', 'e', 'n', 'c', 'e', 's', ' ',
     'e', 'n', 'd', ' ', 'i', 'n', ' ', 'd', 'i', 'f', 'f', 'e', 'r', 'e', 'n',
     't', ' ', 't', 'e', 'r', 'm', 's', ' ')),
                 comp[List[Char], List[Char],
                       List[Char]]((shows_prec_term[A, B](Zero_nat)).apply(u),
                                    comp[List[Char], List[Char],
  List[Char]](shows_prec_list[Char](Zero_nat, List(' ', 'a', 'n', 'd', ' ')),
               (shows_prec_term[A, B](Zero_nat)).apply(v))))),
                  (_: Unit) =>
                    {
                      val chk: (List[(pos, ((term[A, B], term[A, B]),
     term[A, B]))]) =>
                                 (term[A, B]) =>
                                   (term[A, B]) =>
                                     sum[(List[Char]) => List[Char], Unit]
                        = (a: List[(pos, ((term[A, B], term[A, B]),
   term[A, B]))])
                            => (b: term[A, B]) =>
                                 (c: term[A, B]) =>
                                   check_qrsteps[A,
          B]((_: term[A, B]) => true, r, a, b, c);
                      bindb[(List[Char]) => List[Char], Unit,
                             Unit](((chk(sseq))(s))(u),
                                    (_: Unit) => ((chk(tseq))(t))(v))
                    })
  }

def check_critical_pairs_guided[A : equal : show,
                                 B : equal : show](r:
             List[(term[A, B], term[A, B])],
            cp: List[(Boolean, (term[A, B], term[A, B]))],
            joins:
              List[(term[A, B],
                     (List[(pos, ((term[A, B], term[A, B]), term[A, B]))],
                       (term[A, B],
                         List[(pos, ((term[A, B], term[A, B]),
                                      term[A, B]))])))]):
      sum[(List[Char]) => List[Char], Unit] =
  bindb[(List[Char]) => List[Char], Unit,
         Unit](catcha[((term[A, B],
                         (List[(pos, ((term[A, B], term[A, B]), term[A, B]))],
                           (term[A, B],
                             List[(pos, ((term[A, B], term[A, B]),
  term[A, B]))]))),
                        (List[Char]) => List[Char]),
                       Unit,
                       (List[Char]) =>
                         List[Char]](forallm[(term[A, B],
       (List[(pos, ((term[A, B], term[A, B]), term[A, B]))],
         (term[A, B], List[(pos, ((term[A, B], term[A, B]), term[A, B]))]))),
      (List[Char]) =>
        List[Char]]((a: (term[A, B],
                          (List[(pos, ((term[A, B], term[A, B]), term[A, B]))],
                            (term[A, B],
                              List[(pos, ((term[A, B], term[A, B]),
   term[A, B]))]))))
                      => {
                           val (s, (sseq, (aa, b))):
                                 (term[A, B],
                                   (List[(pos,
   ((term[A, B], term[A, B]), term[A, B]))],
                                     (term[A, B],
                                       List[(pos,
      ((term[A, B], term[A, B]), term[A, B]))])))
                             = a;
                           check_join[A, B](r, s, sseq, aa, b)
                         },
                     joins),
                                      (x:
 ((term[A, B],
    (List[(pos, ((term[A, B], term[A, B]), term[A, B]))],
      (term[A, B], List[(pos, ((term[A, B], term[A, B]), term[A, B]))]))),
   (List[Char]) => List[Char]))
=> Inl[(List[Char]) => List[Char],
        Unit](snd[(term[A, B],
                    (List[(pos, ((term[A, B], term[A, B]), term[A, B]))],
                      (term[A, B],
                        List[(pos, ((term[A, B], term[A, B]), term[A, B]))]))),
                   (List[Char]) => List[Char]](x))),
                (_: Unit) =>
                  {
                    val cpa: List[(term[A, B], term[A, B])] =
                      mapa[(term[A, B],
                             (List[(pos, ((term[A, B], term[A, B]),
   term[A, B]))],
                               (term[A, B],
                                 List[(pos,
((term[A, B], term[A, B]), term[A, B]))]))),
                            (term[A, B],
                              term[A, B])]((a:
      (term[A, B],
        (List[(pos, ((term[A, B], term[A, B]), term[A, B]))],
          (term[A, B], List[(pos, ((term[A, B], term[A, B]), term[A, B]))]))))
     => {
          val (s, (_, (t, _))):
                (term[A, B],
                  (List[(pos, ((term[A, B], term[A, B]), term[A, B]))],
                    (term[A, B],
                      List[(pos, ((term[A, B], term[A, B]), term[A, B]))])))
            = a;
          (s, t)
        },
    joins);
                    catcha[((Boolean, (term[A, B], term[A, B])),
                             (List[Char]) => List[Char]),
                            Unit,
                            (List[Char]) =>
                              List[Char]](forallm[(Boolean,
            (term[A, B], term[A, B])),
           (List[Char]) =>
             List[Char]]((a: (Boolean, (term[A, B], term[A, B]))) =>
                           {
                             val (_, (s, t)):
                                   (Boolean, (term[A, B], term[A, B]))
                               = a;
                             check[(List[Char]) =>
                                     List[Char]](equal_terma[A, B](s, t) ||
           list_ex[(term[A, B],
                     term[A, B])]((st: (term[A, B], term[A, B])) =>
                                    instance_rule[A, B, B]((s, t), st) ||
                                      instance_rule[A, B, B]((t, s), st),
                                   cpa),
          comp[List[Char], List[Char],
                List[Char]](shows_prec_list[Char](Zero_nat,
           List('c', 'r', 'i', 't', 'i', 'c', 'a', 'l', ' ', 'p', 'a', 'i', 'r',
                 ' ')),
                             comp[List[Char], List[Char],
                                   List[Char]](show_crit_pair[A, B]((s, t)),
        shows_prec_list[Char](Zero_nat,
                               List(' ', 'm', 'i', 's', 's', 'i', 'n', 'g')))))
                           },
                          cp),
   (x: ((Boolean, (term[A, B], term[A, B])), (List[Char]) => List[Char])) =>
     Inl[(List[Char]) => List[Char],
          Unit](snd[(Boolean, (term[A, B], term[A, B])),
                     (List[Char]) => List[Char]](x)))
                  })

def check_critical_pairs[A : equal : show](r:
     List[(term[A, List[Char]], term[A, List[Char]])],
    cp: List[(Boolean, (term[A, List[Char]], term[A, List[Char]]))],
    joins_o:
      Option[List[(term[A, List[Char]],
                    (List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                  term[A, List[Char]]))],
                      (term[A, List[Char]],
                        List[(pos, ((term[A, List[Char]], term[A, List[Char]]),
                                     term[A, List[Char]]))])))]]):
      sum[(List[Char]) => List[Char], Unit] =
  joins_o match {
    case None => check_critical_pairs_NF[A, List[Char]](r, cp)
    case Some(a) => check_critical_pairs_guided[A, List[Char]](r, cp, a)
  }

def switch_innermost_proc[A, B : equal : show](i:
         dpp_ops_ext[A, B, List[Char], Unit],
        joins:
          Option[List[(term[B, List[Char]],
                        (List[(pos, ((term[B, List[Char]], term[B, List[Char]]),
                                      term[B, List[Char]]))],
                          (term[B, List[Char]],
                            List[(pos, ((term[B, List[Char]],
  term[B, List[Char]]),
 term[B, List[Char]]))])))]],
        dpp: A):
      sum[(List[Char]) => List[Char], A] =
  {
    val r: List[(term[B, List[Char]], term[B, List[Char]])] =
      (rwc[A, B, List[Char], Unit](i)).apply(dpp)
    val p: List[(term[B, List[Char]], term[B, List[Char]])] =
      (pa[A, B, List[Char], Unit](i)).apply(dpp)
    val pw: List[(term[B, List[Char]], term[B, List[Char]])] =
      (pwa[A, B, List[Char], Unit](i)).apply(dpp);
    bindb[(List[Char]) => List[Char], Unit,
           Unit](check[(List[Char]) =>
                         List[Char]](nulla[term[B,
         List[Char]]]((qc[A, B, List[Char], Unit](i)).apply(dpp)),
                                      shows_prec_list[Char](Zero_nat,
                     List('n', 'o', 'n', '-', 'e', 'm', 'p', 't', 'y', ' ', 'Q',
                           ' ', 'n', 'o', 't', ' ', 'y', 'e', 't', ' ', 's',
                           'u', 'p', 'p', 'o', 'r', 't', 'e', 'd'))),
                  (_: Unit) =>
                    bindb[(List[Char]) => List[Char], Unit,
                           Unit](check[(List[Char]) =>
 List[Char]](nulla[(term[B, List[Char]],
                     term[B, List[Char]])]((rc[A, B, List[Char],
        Unit](i)).apply(dpp)),
              shows_prec_list[Char](Zero_nat,
                                     List('s', 't', 'r', 'i', 'c', 't', ' ',
   'r', 'u', 'l', 'e', 's', ' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w',
   'e', 'd'))),
                                  (_: Unit) =>
                                    bindb[(List[Char]) => List[Char], Unit,
   Unit](check[(List[Char]) =>
                 List[Char]](nulla[(Boolean,
                                     (term[B, List[Char]],
                                       term[B,
     List[Char]]))](critical_pairs_impl[B](p ++ pw, r)),
                              shows_prec_list[Char](Zero_nat,
             List('o', 'v', 'e', 'r', 'l', 'a', 'p', 's', ' ', 'b', 'e', 't',
                   'w', 'e', 'e', 'n', ' ', 'P', ' ', 'a', 'n', 'd', ' ', 'R',
                   ' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e',
                   'd'))),
          (_: Unit) =>
            check_critical_pairs[B](r, critical_pairs_impl[B](r, r), joins))))
      match {
      case Inl(a) => Inl[(List[Char]) => List[Char], A](a)
      case Inr(_) =>
        Inr[A, (List[Char]) =>
                 List[Char]]((mkc[A, B, List[Char],
                                   Unit](i)).apply(p).apply(pw).apply(mapa[(term[B,
  List[Char]],
                                     term[B, List[Char]]),
                                    term[B,
  List[Char]]]((a: (term[B, List[Char]], term[B, List[Char]])) =>
                 fst[term[B, List[Char]], term[B, List[Char]]](a),
                r)).apply(Nil).apply(r))
    }
  }

def ur_af_root_redtriple_proc[A, B : equal : show,
                               C : equal : show](i: dpp_ops_ext[A, B, C, Unit],
          rp: root_redtriple_ext[B, C, Unit],
          premove: List[(term[B, C], term[B, C])],
          ur: List[(term[B, C], term[B, C])], dpp: A):
      sum[(List[Char]) => List[Char], A] =
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]]({
                          val (ps, pns):
                                (List[(term[B, C], term[B, C])],
                                  List[(term[B, C], term[B, C])])
                            = (split_pairsa[A, B, C,
     Unit](i)).apply(dpp).apply(premove)
                          val p: List[(term[B, C], term[B, C])] =
                            (pairsa[A, B, C, Unit](i)).apply(dpp)
                          val rm: ((B, nat)) => List[(term[B, C], term[B, C])] =
                            (rules_mapb[A, B, C, Unit](i)).apply(dpp);
                          (rulesd[A, B, C, Unit](i)).apply(dpp)
                          val pi: ((B, nat)) => nat => Boolean =
                            afa[B, C, Unit](rp)
                          val pia: ((B, nat)) => nat => Boolean =
                            aft[B, C, Unit](rp);
                          bindb[(List[Char]) => List[Char], Unit,
                                 Unit](catcha[((term[B, C], term[B, C]),
        (List[Char]) => List[Char]),
       Unit,
       (List[Char]) =>
         List[Char]](forallm[(term[B, C], term[B, C]),
                              (List[Char]) =>
                                List[Char]]((a: (term[B, C], term[B, C])) =>
      {
        val (l, r): (term[B, C], term[B, C]) = a;
        bindb[(List[Char]) => List[Char], Unit,
               Unit](check_no_var[B, C](l),
                      (_: Unit) =>
                        bindb[(List[Char]) => List[Char], Unit,
                               Unit](check_no_var[B, C](r),
                                      (_: Unit) =>
check_no_defined_root[B, C]((fn: (B, nat)) =>
                              ! (nulla[(term[B, C],
 term[B, C])]((rules_mapb[A, B, C, Unit](i)).apply(dpp).apply(fn))),
                             r)))
      },
     p),
                      (x: ((term[B, C], term[B, C]),
                            (List[Char]) => List[Char]))
                        => Inl[(List[Char]) => List[Char],
                                Unit](snd[(term[B, C], term[B, C]),
   (List[Char]) => List[Char]](x))),
(_: Unit) =>
  bindb[(List[Char]) => List[Char], Unit,
         Unit](valida[B, C, Unit](rp),
                (_: Unit) =>
                  bindb[(List[Char]) => List[Char], Unit,
                         Unit](check_ur_P_closed_rm_af[B,
                C](rm, ur,
                    maps[(term[B, C], term[B, C]),
                          (term[B, C],
                            term[B, C])]((st: (term[B, C], term[B, C])) =>
   {
     val (Fun(f, ts)): term[B, C] = snd[term[B, C], term[B, C]](st)
     val n: nat = size_list[term[B, C]](ts)
     val is: nat => Boolean = pia((f, n));
     map_filter[nat, (term[B, C],
                       term[B, C])]((x: nat) =>
                                      (if (is(x))
Some[(term[B, C],
       term[B, C])]((fst[term[B, C], term[B, C]](st), nth[term[B, C]](ts, x)))
else None),
                                     upt(Zero_nat, n))
   },
  p),
                    pi),
                                (_: Unit) =>
                                  bindb[(List[Char]) => List[Char], Unit,
 Unit](catcha[(List[Char]) => List[Char], Unit,
               (List[Char]) =>
                 List[Char]](catcha[((term[B, C], term[B, C]),
                                      (List[Char]) => List[Char]),
                                     Unit,
                                     (List[Char]) =>
                                       List[Char]](forallm[(term[B, C],
                     term[B, C]),
                    (List[Char]) => List[Char]](nsa[B, C, Unit](rp), ur),
            (x: ((term[B, C], term[B, C]), (List[Char]) => List[Char])) =>
              Inl[(List[Char]) => List[Char],
                   Unit](snd[(term[B, C], term[B, C]),
                              (List[Char]) => List[Char]](x))),
                              (x: (List[Char]) => List[Char]) =>
                                Inl[(List[Char]) => List[Char],
                                     Unit](comp[List[Char], List[Char],
         List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm',
      ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g',
      ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e', 's')),
                      comp[List[Char], List[Char], List[Char]](shows_nl, x)))),
        (_: Unit) =>
          bindb[(List[Char]) => List[Char], Unit,
                 Unit](catcha[(List[Char]) => List[Char], Unit,
                               (List[Char]) =>
                                 List[Char]](catcha[((term[B, C], term[B, C]),
              (List[Char]) => List[Char]),
             Unit,
             (List[Char]) =>
               List[Char]](forallm[(term[B, C], term[B, C]),
                                    (List[Char]) =>
                                      List[Char]](nsta[B, C, Unit](rp), pns),
                            (x: ((term[B, C], term[B, C]),
                                  (List[Char]) => List[Char]))
                              => Inl[(List[Char]) => List[Char],
                                      Unit](snd[(term[B, C], term[B, C]),
         (List[Char]) => List[Char]](x))),
      (x: (List[Char]) => List[Char]) =>
        Inl[(List[Char]) => List[Char],
             Unit](comp[List[Char], List[Char],
                         List[Char]](shows_string.apply(List('p', 'r', 'o', 'b',
                      'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r',
                      'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'D', 'P', 's')),
                                      comp[List[Char], List[Char],
    List[Char]](shows_nl, x)))),
                        (_: Unit) =>
                          catcha[(List[Char]) => List[Char], Unit,
                                  (List[Char]) =>
                                    List[Char]](catcha[((term[B, C],
                  term[B, C]),
                 (List[Char]) => List[Char]),
                Unit,
                (List[Char]) =>
                  List[Char]](forallm[(term[B, C], term[B, C]),
                                       (List[Char]) =>
 List[Char]](sa[B, C, Unit](rp), ps),
                               (x: ((term[B, C], term[B, C]),
                                     (List[Char]) => List[Char]))
                                 => Inl[(List[Char]) => List[Char],
 Unit](snd[(term[B, C], term[B, C]), (List[Char]) => List[Char]](x))),
         (x: (List[Char]) => List[Char]) =>
           Inl[(List[Char]) => List[Char],
                Unit](comp[List[Char], List[Char],
                            List[Char]](shows_string.apply(List('p', 'r', 'o',
                         'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o',
                         'r', 'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'D', 'P',
                         's')),
 comp[List[Char], List[Char], List[Char]](shows_nl, x)))))))))
                        },
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ', 'n', 'o',
 't', ' ', 'a', 'p', 'p', 'l', 'y', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'd', 'u',
 'c', 't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c', 'e',
 's', 's', 'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h', 'e', ' ', 'f', 'o',
 'l', 'l', 'o', 'w', 'i', 'n', 'g')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_nl,
                                    comp[List[Char], List[Char],
  List[Char]](descaa[B, C, Unit](rp),
               comp[List[Char], List[Char], List[Char]](shows_nl, x))))))
    match {
    case Inl(a) => Inl[(List[Char]) => List[Char], A](a)
    case Inr(_) =>
      Inr[A, (List[Char]) =>
               List[Char]]((delete_P_Pwa[A, B, C,
  Unit](i)).apply(dpp).apply(premove).apply(premove))
  }

def check_wf_afs[A : show](pi: List[((A, nat), af_entry)]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(((A, nat), af_entry), (List[Char]) => List[Char]), Unit,
          (List[Char]) =>
            List[Char]](forallm[((A, nat), af_entry),
                                 (List[Char]) =>
                                   List[Char]]((a: ((A, nat), af_entry)) =>
         {
           val (aa, b): ((A, nat), af_entry) = a;
           ({
              val (f, n): (A, nat) = aa;
              (e: af_entry) =>
                catcha[nat, Unit,
                        (List[Char]) =>
                          List[Char]](e match {
case Collapse(i) => check[nat](less_nat(i, n), i)
case AFList(ids) =>
  catcha[(nat, nat), Unit,
          nat](forallm[nat, nat]((x: nat) =>
                                   (if (less_nat(x, n)) Inr[Unit, nat](())
                                     else Inl[nat, Unit](x)),
                                  ids),
                (x: (nat, nat)) => Inl[nat, Unit](snd[nat, nat](x)))
                                      },
                                       (x: nat) =>
 Inl[(List[Char]) => List[Char],
      Unit](comp[List[Char], List[Char],
                  List[Char]](shows_string.apply(List('i', 'n', 'd', 'e', 'x',
               ' ')),
                               comp[List[Char], List[Char],
                                     List[Char]](shows_prec_nat(Zero_nat, x),
          comp[List[Char], List[Char],
                List[Char]](shows_string.apply(List(' ', 't', 'o', 'o', ' ',
             'h', 'i', 'g', 'h', ' ', 'f', 'o', 'r', ' ', 'A', 'F', '-', 'e',
             'n', 't', 'r', 'y', ' ', 'o', 'f')),
                             comp[List[Char], List[Char],
                                   List[Char]](shows_prec_prod[A,
                        nat](Zero_nat, (f, n)),
        shows_nl))))))
            })(b)
         },
        pi),
                         (x: (((A, nat), af_entry), (List[Char]) => List[Char]))
                           => Inl[(List[Char]) => List[Char],
                                   Unit](snd[((A, nat), af_entry),
      (List[Char]) => List[Char]](x)))

def afs_list_to_af[A : equal](pi: List[((A, nat), af_entry)], fn: (A, nat)):
      nat => Boolean =
  lookupa[(A, nat), af_entry](fn, pi) match {
    case None => (_: nat) => true
    case Some(Collapse(a)) => (b: nat) => equal_nata(a, b)
    case Some(AFList(a)) => (b: nat) => membera[nat](a, b)
  }

def mono_af_entry(n: nat, x1: af_entry): Boolean = (n, x1) match {
  case (n, Collapse(i)) => less_eq_nat(n, one_nat)
  case (n, AFList(ids)) =>
    all_interval_nat((a: nat) => membera[nat](ids, a), Zero_nat, n)
}

def check_mono_afs_help[A : show, B, C](x0: ((A, B), C)):
      (List[Char]) => List[Char] =
  x0 match {
  case ((f, a), e) =>
    comp[List[Char], List[Char],
          List[Char]](shows_string.apply(List('e', 'n', 't', 'r', 'y', ' ', 'f',
       'o', 'r', ' ')),
                       comp[List[Char], List[Char],
                             List[Char]]((aa: List[Char]) =>
   shows_prec[A](Zero_nat, f, aa),
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List(' ', 'n', 'o', 't', ' ', 'm', 'o',
     'n', 'o', 't', 'o', 'n', 'i', 'c')),
                     shows_nl)))
}

def check_mono_afs[A : show](pi: List[((A, nat), af_entry)]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[((A, nat), af_entry), Unit,
          (List[Char]) =>
            List[Char]](catcha[(((A, nat), af_entry), ((A, nat), af_entry)),
                                Unit,
                                ((A, nat),
                                  af_entry)](forallm[((A, nat), af_entry),
              ((A, nat),
                af_entry)]((x: ((A, nat), af_entry)) =>
                             (if (mono_af_entry(snd[A,
             nat](fst[(A, nat), af_entry](x)),
         snd[(A, nat), af_entry](x)))
                               Inr[Unit, ((A, nat), af_entry)](())
                               else Inl[((A, nat), af_entry), Unit](x)),
                            pi),
      (x: (((A, nat), af_entry), ((A, nat), af_entry))) =>
        Inl[((A, nat), af_entry),
             Unit](snd[((A, nat), af_entry), ((A, nat), af_entry)](x))),
                         (x: ((A, nat), af_entry)) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](check_mono_afs_help[A, nat, af_entry](x)))

def default_af_entry(n: nat): af_entry = AFList(upt(Zero_nat, n))

def afs_list_to_afs[A : equal](pi: List[((A, nat), af_entry)], f: A, n: nat):
      af_entry =
  lookupa[(A, nat), af_entry]((f, n), pi) match {
    case None => default_af_entry(n)
    case Some(entry) => entry
  }

def af_redtriple[A : equal : show,
                  B : show](pi: List[((A, nat), af_entry)],
                             rp: redtriple_ext[filtered[A], B, Unit]):
      redtriple_ext[A, B, Unit] =
  {
    val af: ((A, nat)) => nat => Boolean =
      (a: (A, nat)) => afs_list_to_af[A](pi, a)
    val afs: A => nat => af_entry =
      (a: A) => (b: nat) => afs_list_to_afs[A](pi, a, b);
    Redtriple_ext[A, B,
                   Unit](bindb[(List[Char]) => List[Char], Unit,
                                Unit](check_wf_afs[A](pi),
                                       (_: Unit) =>
 valid[filtered[A], B, Unit](rp)),
                          (a: (term[A, B], term[A, B])) =>
                            afs_check[A, B](shows_string.apply(List('>')), afs,
     s[filtered[A], B, Unit](rp), a),
                          (a: (term[A, B], term[A, B])) =>
                            afs_check[A, B](shows_string.apply(List('>', '=')),
     afs, ns[filtered[A], B, Unit](rp), a),
                          (a: (term[A, B], term[A, B])) =>
                            afs_check[A, B](shows_string.apply(List('>', '=')),
     afs, nst[filtered[A], B, Unit](rp), a),
                          af, bindb[(List[Char]) => List[Char], Unit,
                                     Unit](check_mono_afs[A](pi),
    (_: Unit) => mono[filtered[A], B, Unit](rp)),
                          comp[List[Char], List[Char],
                                List[Char]](shows_string.apply(List('A', 'r',
                             'g', 'u', 'm', 'e', 'n', 't', ' ', 'F', 'i', 'l',
                             't', 'e', 'r', ':', ' ')),
     comp[List[Char], List[Char],
           List[Char]](shows_nl,
                        comp[List[Char], List[Char],
                              List[Char]](shows_afs[A](pi),
   comp[List[Char], List[Char],
         List[Char]](shows_nl, desca[filtered[A], B, Unit](rp))))),
                          ())
  }

def equal_filtereda[A : equal](x0: filtered[A], x1: filtered[A]): Boolean =
  (x0, x1) match {
  case (FPair(fa, nata), FPair(f, nat)) => eq[A](fa, f) && equal_nata(nata, nat)
}

implicit def equal_filtered[A : equal]: equal[filtered[A]] = new
  equal[filtered[A]] {
  val `Ceta.equal` = (a: filtered[A], b: filtered[A]) =>
    equal_filtereda[A](a, b)
}

def shows_arctic(x0: arctic): (List[Char]) => List[Char] = x0 match {
  case Num_arc(i) => shows_prec_int(Zero_nat, i)
  case MinInfty => shows_string.apply(List('-', 'i', 'n', 'f'))
}

def shows_prec_arctic(d: nat, ai: arctic): (List[Char]) => List[Char] =
  shows_arctic(ai)

def shows_list_arctic(as: List[arctic]): (List[Char]) => List[Char] =
  shows_list_aux[arctic]((a: arctic) => shows_prec_arctic(Zero_nat, a), as)

implicit def show_arctic: show[arctic] = new show[arctic] {
  val `Ceta.shows_prec` = (a: nat, b: arctic, c: List[Char]) =>
    (shows_prec_arctic(a, b)).apply(c)
  val `Ceta.shows_list` = (a: List[arctic], b: List[Char]) =>
    (shows_list_arctic(a)).apply(b)
}

implicit def ordered_semiring_0_int: ordered_semiring_0[BigInt] = new
  ordered_semiring_0[BigInt] {
  val `Ceta.ge` = (a: BigInt, b: BigInt) => ge_int(a, b)
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
  val `Ceta.zero` = zero_inta
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
}

implicit def ordered_semiring_1_int: ordered_semiring_1[BigInt] = new
  ordered_semiring_1[BigInt] {
  val `Ceta.ge` = (a: BigInt, b: BigInt) => ge_int(a, b)
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
  val `Ceta.zero` = zero_inta
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
  val `Ceta.one` = one_inta
}

implicit def max_ordered_semiring_1_int: max_ordered_semiring_1[BigInt] = new
  max_ordered_semiring_1[BigInt] {
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
  val `Ceta.one` = one_inta
  val `Ceta.max0` = (a: BigInt) => max0_int(a)
  val `Ceta.ge` = (a: BigInt, b: BigInt) => ge_int(a, b)
  val `Ceta.zero` = zero_inta
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
}

implicit def poly_carrier_int: poly_carrier[BigInt] = new poly_carrier[BigInt] {
  val `Ceta.max0` = (a: BigInt) => max0_int(a)
  val `Ceta.ge` = (a: BigInt, b: BigInt) => ge_int(a, b)
  val `Ceta.zero` = zero_inta
  val `Ceta.plus` = (a: BigInt, b: BigInt) => a + b
  val `Ceta.one` = one_inta
  val `Ceta.times` = (a: BigInt, b: BigInt) => a * b
}

implicit def ordered_semiring_0_rat: ordered_semiring_0[rat] = new
  ordered_semiring_0[rat] {
  val `Ceta.ge` = (a: rat, b: rat) => ge_rat(a, b)
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
  val `Ceta.zero` = zero_rata
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
}

implicit def ordered_semiring_1_rat: ordered_semiring_1[rat] = new
  ordered_semiring_1[rat] {
  val `Ceta.ge` = (a: rat, b: rat) => ge_rat(a, b)
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
  val `Ceta.zero` = zero_rata
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
  val `Ceta.one` = one_rata
}

implicit def max_ordered_semiring_1_rat: max_ordered_semiring_1[rat] = new
  max_ordered_semiring_1[rat] {
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
  val `Ceta.one` = one_rata
  val `Ceta.max0` = (a: rat) => max0_rat(a)
  val `Ceta.ge` = (a: rat, b: rat) => ge_rat(a, b)
  val `Ceta.zero` = zero_rata
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
}

implicit def poly_carrier_rat: poly_carrier[rat] = new poly_carrier[rat] {
  val `Ceta.max0` = (a: rat) => max0_rat(a)
  val `Ceta.ge` = (a: rat, b: rat) => ge_rat(a, b)
  val `Ceta.zero` = zero_rata
  val `Ceta.plus` = (a: rat, b: rat) => plus_rata(a, b)
  val `Ceta.one` = one_rata
  val `Ceta.times` = (a: rat, b: rat) => times_rata(a, b)
}

def shows_arctic_rat(x0: arctic_rat): (List[Char]) => List[Char] = x0 match {
  case Num_arc_rat(i) => shows_prec_rat(Zero_nat, i)
  case MinInfty_rat => shows_string.apply(List('-', 'i', 'n', 'f'))
}

def shows_prec_arctic_rat(d: nat, ari: arctic_rat): (List[Char]) => List[Char] =
  shows_arctic_rat(ari)

def shows_list_arctic_rat(as: List[arctic_rat]): (List[Char]) => List[Char] =
  shows_list_aux[arctic_rat]((a: arctic_rat) =>
                               shows_prec_arctic_rat(Zero_nat, a),
                              as)

implicit def show_arctic_rat: show[arctic_rat] = new show[arctic_rat] {
  val `Ceta.shows_prec` = (a: nat, b: arctic_rat, c: List[Char]) =>
    (shows_prec_arctic_rat(a, b)).apply(c)
  val `Ceta.shows_list` = (a: List[arctic_rat], b: List[Char]) =>
    (shows_list_arctic_rat(a)).apply(b)
}

def poly_inter_to_af[A, B](i: A => nat => List[(List[(nat, nat)], B)],
                            fn: (A, nat)):
      nat => Boolean =
  (a: nat) => poly_var[nat, B]((i(fst[A, nat](fn)))(snd[A, nat](fn)), a)

def check_poly_weak_mono_all[A, B : ordered_semiring_0](p:
                  List[(List[(A, nat)], B)]):
      Boolean =
  list_all[(List[(A, nat)],
             B)]((a: (List[(A, nat)], B)) =>
                   {
                     val (_, c): (List[(A, nat)], B) = a;
                     ge[B](c, zero[B])
                   },
                  p)

def check_poly_weak_mono_discrete[A : equal,
                                   B : equal : poly_carrier](p:
                       List[(List[(A, nat)], B)],
                      v: A):
      Boolean =
  check_poly_ge[A, B](poly_subst[A, A,
                                  B]((w: A) =>
                                       poly_of[A,
        B]((if (eq[A](w, v)) PSum[A, B](List(PNum[B, A](one[B]), PVar[A, B](v)))
             else PVar[A, B](w))),
                                      p),
                       p)

def check_poly_weak_mono_and_pos[A : equal,
                                  B : equal : poly_carrier](discrete: Boolean,
                     p: List[(List[(A, nat)], B)]):
      Boolean =
  (if (discrete)
    list_all[A]((a: A) => check_poly_weak_mono_discrete[A, B](p, a),
                 poly_vars_list[A, B](p)) &&
      ge[B](eval_poly[A, B]((_: A) => zero[B], p), zero[B])
    else check_poly_weak_mono_all[A, B](p))

def check_poly_inter_list[A, B : equal : poly_carrier](discrete: Boolean,
                i: List[((A, nat), List[(List[(nat, nat)], B)])]):
      sum[(A, List[(List[(nat, nat)], B)]), Unit] =
  catcha[((A, nat), List[(List[(nat, nat)], B)]), Unit,
          (A, List[(List[(nat, nat)],
                     B)])](catcha[(((A, nat), List[(List[(nat, nat)], B)]),
                                    ((A, nat), List[(List[(nat, nat)], B)])),
                                   Unit,
                                   ((A, nat),
                                     List[(List[(nat, nat)],
    B)])](forallm[((A, nat), List[(List[(nat, nat)], B)]),
                   ((A, nat),
                     List[(List[(nat, nat)],
                            B)])]((x: ((A, nat), List[(List[(nat, nat)], B)]))
                                    => (if ({
      val (_, a): ((A, nat), List[(List[(nat, nat)], B)]) = x;
      check_poly_weak_mono_and_pos[nat, B](discrete, a)
    })
 Inr[Unit, ((A, nat), List[(List[(nat, nat)], B)])](())
 else Inl[((A, nat), List[(List[(nat, nat)], B)]), Unit](x)),
                                   i),
           (x: (((A, nat), List[(List[(nat, nat)], B)]),
                 ((A, nat), List[(List[(nat, nat)], B)])))
             => Inl[((A, nat), List[(List[(nat, nat)], B)]),
                     Unit](snd[((A, nat), List[(List[(nat, nat)], B)]),
                                ((A, nat), List[(List[(nat, nat)], B)])](x))),
                            (x: ((A, nat), List[(List[(nat, nat)], B)])) =>
                              Inl[(A, List[(List[(nat, nat)], B)]),
                                   Unit]({
   val (a, b): ((A, nat), List[(List[(nat, nat)], B)]) = x;
   ({
      val (f, _): (A, nat) = a;
      (aa: List[(List[(nat, nat)], B)]) => (f, aa)
    })(b)
 }))

def poly_inter_list_to_inter[A : poly_carrier,
                              B : equal](defa: A,
  i: List[((B, nat), List[(List[(nat, nat)], A)])]):
      B => nat => List[(List[(nat, nat)], A)] =
  (f: B) =>
    (n: nat) =>
      lookupa[(B, nat), List[(List[(nat, nat)], A)]]((f, n), i) match {
        case None => default_I[A](defa, n)
        case Some(p) => p
      }

def check_monom_strict_mono[A : equal](pm: Boolean, m: List[(A, nat)], v: A):
      Boolean =
  ! (nulla[(A, nat)](m)) &&
    (nulla[(A, nat)](tl[(A, nat)](m)) &&
      (eq[A](fst[A, nat](hd[(A, nat)](m)), v) &&
        (if (pm) less_eq_nat(one_nat, snd[A, nat](hd[(A, nat)](m)))
          else equal_nata(snd[A, nat](hd[(A, nat)](m)), one_nat))))

def check_poly_strict_mono[A : equal,
                            B : poly_carrier](pm: Boolean,
       p: List[(List[(A, nat)], B)], v: A):
      Boolean =
  list_ex[(List[(A, nat)],
            B)]((a: (List[(A, nat)], B)) =>
                  {
                    val (m, c): (List[(A, nat)], B) = a;
                    ge[B](c, one[B]) && check_monom_strict_mono[A](pm, m, v)
                  },
                 p)

def check_poly_strict_mono_discrete[A : equal : poly_carrier,
                                     B : equal](gt: A => A => Boolean,
         p: List[(List[(B, nat)], A)], v: B):
      Boolean =
  check_poly_gt[A, B](gt, poly_subst[B, B,
                                      A]((w: B) =>
   poly_of[B, A]((if (eq[B](w, v))
                   PSum[B, A](List(PNum[A, B](one[A]), PVar[B, A](v)))
                   else PVar[B, A](w))),
  p),
                       p)

def check_poly_strict_mono_smart[A : equal : poly_carrier,
                                  B : equal](discrete: Boolean, pm: Boolean,
      gt: A => A => Boolean, p: List[(List[(B, nat)], A)], v: B):
      Boolean =
  (if (discrete) check_poly_strict_mono_discrete[A, B](gt, p, v)
    else check_poly_strict_mono[B, A](pm, p, v))

def create_nlpoly_redtriple[A : equal : poly_carrier : show, B : equal : show,
                             C : equal : show](defa: A, gt: A => A => Boolean,
        power_mono: Boolean, discrete: Boolean,
        i: List[((B, nat), List[(List[(nat, nat)], A)])]):
      redtriple_ext[B, C, Unit] =
  {
    val j: B => nat => List[(List[(nat, nat)], A)] =
      poly_inter_list_to_inter[A, B](defa, i)
    val x: (List[(List[(nat, nat)], A)]) => List[(List[(List[Char], nat)], A)] =
      (a: List[(List[(nat, nat)], A)]) =>
        poly_subst[nat, List[Char],
                    A]((n: nat) =>
                         poly_of[List[Char],
                                  A](PVar[List[Char],
   A](List('x', '_') ++ (shows_prec_nat(Zero_nat, n)).apply(Nil))),
                        a)
    val ns: ((term[B, C], term[B, C])) => sum[(List[Char]) => List[Char], Unit]
      = (a: (term[B, C], term[B, C])) =>
          {
            val (s, t): (term[B, C], term[B, C]) = a
            val p: List[(List[(C, nat)], A)] = eval_term[B, A, C](j, s)
            val q: List[(List[(C, nat)], A)] = eval_term[B, A, C](j, t);
            check[(List[Char]) =>
                    List[Char]](check_poly_ge[C, A](p, q),
                                 comp[List[Char], List[Char],
                                       List[Char]](shows_string.apply(List('c',
                                    'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ',
                                    'e', 'n', 's', 'u', 'r', 'e', ' ')),
            comp[List[Char], List[Char],
                  List[Char]](shows_poly[C, A](p),
                               comp[List[Char], List[Char],
                                     List[Char]](shows_string.apply(List(' ',
                                  '>', '=', ' ')),
          shows_poly[C, A](q)))))
          };
    Redtriple_ext[B, C,
                   Unit](catcha[(B, List[(List[(nat, nat)], A)]), Unit,
                                 (List[Char]) =>
                                   List[Char]](check_poly_inter_list[B,
                              A](discrete, i),
        (xa: (B, List[(List[(nat, nat)], A)])) =>
          Inl[(List[Char]) => List[Char],
               Unit]({
                       val (f, p): (B, List[(List[(nat, nat)], A)]) = xa;
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('i', 'n', 't',
                          'e', 'r', 'p', 'r', 'e', 't', 'a', 't', 'i', 'o', 'n',
                          ' ')),
  comp[List[Char], List[Char],
        List[Char]](shows_poly[List[Char], A](x(p)),
                     comp[List[Char], List[Char],
                           List[Char]](shows_string.apply(List(' ', 'o', 'f',
                        ' ')),
comp[List[Char], List[Char],
      List[Char]]((a: List[Char]) => shows_prec[B](Zero_nat, f, a),
                   shows_string.apply(List(' ', 'i', 'n', 'v', 'a', 'l', 'i',
    'd', ' '))))))
                     })),
                          (a: (term[B, C], term[B, C])) =>
                            {
                              val (s, t): (term[B, C], term[B, C]) = a
                              val p: List[(List[(C, nat)], A)] =
                                eval_term[B, A, C](j, s)
                              val q: List[(List[(C, nat)], A)] =
                                eval_term[B, A, C](j, t);
                              check[(List[Char]) =>
                                      List[Char]](check_poly_gt[A, C](gt, p, q),
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd',
              ' ', 'n', 'o', 't', ' ', 'e', 'n', 's', 'u', 'r', 'e', ' ')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_poly[C, A](p),
         comp[List[Char], List[Char],
               List[Char]](shows_string.apply(List(' ', '>', ' ')),
                            shows_poly[C, A](q)))))
                            },
                          ns, ns, (a: (B, nat)) => poly_inter_to_af[B, A](j, a),
                          catcha[((B, nat), List[(List[(nat, nat)], A)]), Unit,
                                  (List[Char]) =>
                                    List[Char]](catcha[(((B, nat),
                  List[(List[(nat, nat)], A)]),
                 ((B, nat), List[(List[(nat, nat)], A)])),
                Unit,
                ((B, nat),
                  List[(List[(nat, nat)],
                         A)])](forallm[((B, nat), List[(List[(nat, nat)], A)]),
((B, nat),
  List[(List[(nat, nat)],
         A)])]((xa: ((B, nat), List[(List[(nat, nat)], A)])) =>
                 (if ({
                        val (a, b): ((B, nat), List[(List[(nat, nat)], A)]) =
                          xa;
                        ({
                           val (_, n): (B, nat) = a;
                           (p: List[(List[(nat, nat)], A)]) =>
                             all_interval_nat((aa: nat) =>
        check_poly_strict_mono_smart[A, nat](discrete, power_mono, gt, p, aa),
       Zero_nat, n)
                         })(b)
                      })
                   Inr[Unit, ((B, nat), List[(List[(nat, nat)], A)])](())
                   else Inl[((B, nat), List[(List[(nat, nat)], A)]), Unit](xa)),
                i),
                                (xa: (((B, nat), List[(List[(nat, nat)], A)]),
                                       ((B, nat), List[(List[(nat, nat)], A)])))
                                  => Inl[((B, nat),
   List[(List[(nat, nat)], A)]),
  Unit](snd[((B, nat), List[(List[(nat, nat)], A)]),
             ((B, nat), List[(List[(nat, nat)], A)])](xa))),
         (xa: ((B, nat), List[(List[(nat, nat)], A)])) =>
           Inl[(List[Char]) => List[Char],
                Unit]({
                        val (a, b): ((B, nat), List[(List[(nat, nat)], A)]) =
                          xa;
                        ({
                           val (f, _): (B, nat) = a;
                           (p: List[(List[(nat, nat)], A)]) =>
                             comp[List[Char], List[Char],
                                   List[Char]](shows_string.apply(List('c', 'o',
                                'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'e',
                                'n', 's', 'u', 'r', 'e', ' ', 'm', 'o', 'n',
                                'o', 't', 'o', 'n', 'i', 'c', 't', 'y', ' ',
                                'o', 'f', ' ')),
        comp[List[Char], List[Char],
              List[Char]](shows_poly[List[Char], A](x(p)),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List(' ', 'a',
                              's', ' ', 'i', 'n', 't', 'e', 'r', 'p', 'r', 'e',
                              't', 'a', 't', 'i', 'o', 'n', ' ', 'o', 'f',
                              ' ')),
      (aa: List[Char]) => shows_prec[B](Zero_nat, f, aa))))
                         })(b)
                      })),
                          comp[List[Char], List[Char],
                                List[Char]](shows_string.apply(List('p', 'o',
                             'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l', ' ', 'i',
                             'n', 't', 'e', 'r', 'p', 'r', 'e', 't', 'a', 't',
                             'i', 'o', 'n')),
     comp[List[Char], List[Char],
           List[Char]](shows_nl,
                        shows_sep[((B, nat),
                                    List[(List[(nat, nat)],
   A)])]((a: ((B, nat), List[(List[(nat, nat)], A)])) =>
           {
             val (aa, b): ((B, nat), List[(List[(nat, nat)], A)]) = a;
             ({
                val (f, n): (B, nat) = aa;
                (p: List[(List[(nat, nat)], A)]) =>
                  comp[List[Char], List[Char],
                        List[Char]](shows_string.apply(List('P', 'o', 'l',
                     '(')),
                                     comp[List[Char], List[Char],
   List[Char]]((ab: List[Char]) => shows_prec[B](Zero_nat, f, ab),
                comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List('/')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_prec_nat(Zero_nat, n),
              comp[List[Char], List[Char],
                    List[Char]](shows_string.apply(List(')', ' ', '=', ' ')),
                                 shows_poly[List[Char], A](x(p)))))))
              })(b)
           },
          shows_nl, i))),
                          ())
  }

def create_afI[A : equal,
                B : equal](ze: A, x1: List[((B, nat), (A, List[A]))],
                            x2: (B, nat)):
      nat => Boolean =
  (ze, x1, x2) match {
  case (ze, Nil, (f, n)) => (_: nat) => true
  case (ze, ((g, m), (c, coeffs)) :: rest, (f, n)) =>
    (if (eq[B](f, g) && equal_nata(m, n))
      {
        val len: nat = size_list[A](coeffs);
        (i: nat) => less_nat(i, len) && ! (eq[A](nth[A](coeffs, i), ze))
      }
      else create_afI[A, B](ze, rest, (f, n)))
}

def show_pvars[A : equal : show, B : show](on: A, x1: List[(B, A)]):
      List[List[Char]] =
  (on, x1) match {
  case (on, Nil) => Nil
  case (on, (x, c) :: vas) =>
    (if (eq[A](on, c)) Nil else shows_prec[A](Zero_nat, c, Nil)) ++
      shows_prec[B](Zero_nat, x, Nil) ::
      show_pvars[A, B](on, vas)
}

def shows_lpoly[A : equal : show, B : show](ze: A, on: A, x2: l_poly[B, A]):
      (List[Char]) => List[Char] =
  (ze, on, x2) match {
  case (ze, on, LPoly(c, cs)) =>
    show_pvars[A, B](on, cs) match {
      case Nil => (a: List[Char]) => shows_prec[A](Zero_nat, c, a)
      case a :: list =>
        comp[List[Char], List[Char],
              List[Char]]((if (eq[A](c, ze)) id[List[Char]]
                            else comp[List[Char], List[Char],
                                       List[Char]]((aa: List[Char]) =>
             shows_prec[A](Zero_nat, c, aa),
            shows_string.apply(List(' ', '+', ' ')))),
                           shows_list_gen[List[Char]]((aa: List[Char]) =>
                shows_prec_list[Char](Zero_nat, aa),
               Nil, Nil, List(' ', '+', ' '), Nil, a :: list))
    }
}

def lookup_rest[A : equal, B](x: A, xa1: List[(A, B)]):
      Option[(B, List[(A, B)])] =
  (x, xa1) match {
  case (x, Nil) => None
  case (x, (y, c) :: ycs) =>
    (if (eq[A](x, y)) Some[(B, List[(A, B)])]((c, ycs))
      else lookup_rest[A, B](x, ycs) match {
             case None => None
             case Some((d, yccs)) =>
               Some[(B, List[(A, B)])]((d, (y, c) :: yccs))
           })
}

def check_pvarsI[A, B : equal](ze: A, ge: A => A => Boolean, vas: List[(B, A)],
                                x3: List[(B, A)]):
      sum[B, Unit] =
  (ze, ge, vas, x3) match {
  case (ze, ge, vas, Nil) =>
    catcha[((B, A), B), Unit,
            B](forallm[(B, A),
                        B]((va: (B, A)) =>
                             check[B]((ge(snd[B, A](va)))(ze), fst[B, A](va)),
                            vas),
                (x: ((B, A), B)) => Inl[B, Unit](snd[(B, A), B](x)))
  case (ze, ge, vas, (x, b) :: vbs) =>
    {
      val (a, vasa): (A, List[(B, A)]) =
        lookup_rest[B, A](x, vas) match {
          case None => (ze, vas)
          case Some((a, ba)) => (a, ba)
        };
      bindb[B, Unit,
             Unit](check[B]((ge(a))(b), x),
                    (_: Unit) => check_pvarsI[A, B](ze, ge, vasa, vbs))
    }
}

def check_lpoly_sI[A : equal : show,
                    B : equal : show](mode: Boolean, ze: A, on: A,
                                       ge: A => A => Boolean,
                                       gt: A => A => Boolean, x5: l_poly[B, A],
                                       x6: l_poly[B, A]):
      sum[(List[Char]) => List[Char], Unit] =
  (mode, ze, on, ge, gt, x5, x6) match {
  case (mode, ze, on, ge, gt, LPoly(a, vas), LPoly(b, vbs)) =>
    catcha[(List[Char]) => List[Char], Unit,
            (List[Char]) =>
              List[Char]](bindb[(List[Char]) => List[Char], Unit,
                                 Unit](check[(List[Char]) =>
       List[Char]]((gt(a))(b),
                    shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm',
     ' ', 'w', 'h', 'e', 'n', ' ', 'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g',
     ' ', 'c', 'o', 'n', 's', 't', 'a', 'n', 't', ' ', 'p', 'a', 'r', 't'))),
(_: Unit) =>
  catcha[B, Unit,
          (List[Char]) =>
            List[Char]](check_pvarsI[A, B](ze, (if (mode) ge else gt), vas,
    vbs),
                         (x: B) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm', ' ',
 'w', 'h', 'e', 'n', ' ', 'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' ', 'c',
 'o', 'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n', 't', 's', ' ', 'o', 'f', ' ', 'v',
 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ')),
                 (aa: List[Char]) => shows_prec[B](Zero_nat, x, aa))))),
                           (x: (List[Char]) => List[Char]) =>
                             Inl[(List[Char]) => List[Char],
                                  Unit](comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm',
   ' ', 'w', 'h', 'e', 'n', ' ', 'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g',
   ' ')),
                   comp[List[Char], List[Char],
                         List[Char]](shows_lpoly[A,
          B](ze, on, LPoly[A, B](a, vas)),
                                      comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List(' ', '>', ' ')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_lpoly[A,
        B](ze, on, LPoly[A, B](b, vbs)),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl, x)))))))
}

def check_polo_sI[A : equal : show, B : show,
                   C : equal : show](mode: Boolean, ze: A, on: A,
                                      pl: A => A => A, ti: A => A => A,
                                      mz: A => A, ge: A => A => Boolean,
                                      gt: A => A => Boolean,
                                      pi: B => nat => (A, List[A]),
                                      x9: (term[B, C], term[B, C])):
      sum[(List[Char]) => List[Char], Unit] =
  (mode, ze, on, pl, ti, mz, ge, gt, pi, x9) match {
  case (mode, ze, on, pl, ti, mz, ge, gt, pi, (s, t)) =>
    {
      val left: l_poly[C, A] = pleftI[A, B, C](ze, on, pl, ti, mz, pi, s)
      val right: l_poly[C, A] = prightI[A, B, C](ze, on, pl, ti, mz, pi, t);
      catcha[(List[Char]) => List[Char], Unit,
              (List[Char]) =>
                List[Char]](check_lpoly_sI[A,
    C](mode, ze, on, ge, gt, left, right),
                             (x: (List[Char]) => List[Char]) =>
                               Inl[(List[Char]) => List[Char],
                                    Unit](comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ', 'n',
     'o', 't', ' ', 'e', 'n', 's', 'u', 'r', 'e', ' ')),
                     comp[List[Char], List[Char],
                           List[Char]]((shows_prec_term[B,
                 C](Zero_nat)).apply(s),
comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List(' ', '>', ' ')),
                   comp[List[Char], List[Char],
                         List[Char]]((shows_prec_term[B, C](Zero_nat)).apply(t),
                                      comp[List[Char], List[Char],
    List[Char]](shows_nl, x)))))))
    }
}

def check_lpoly_nsI[A : equal : show,
                     B : equal : show](ze: A, on: A, ge: A => A => Boolean,
x3: l_poly[B, A], x4: l_poly[B, A]):
      sum[(List[Char]) => List[Char], Unit] =
  (ze, on, ge, x3, x4) match {
  case (ze, on, ge, LPoly(a, vas), LPoly(b, vbs)) =>
    catcha[(List[Char]) => List[Char], Unit,
            (List[Char]) =>
              List[Char]](bindb[(List[Char]) => List[Char], Unit,
                                 Unit](check[(List[Char]) =>
       List[Char]]((ge(a))(b),
                    shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm',
     ' ', 'w', 'h', 'e', 'n', ' ', 'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g',
     ' ', 'c', 'o', 'n', 's', 't', 'a', 'n', 't', ' ', 'p', 'a', 'r', 't',
     's'))),
(_: Unit) =>
  catcha[B, Unit,
          (List[Char]) =>
            List[Char]](check_pvarsI[A, B](ze, ge, vas, vbs),
                         (x: B) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm', ' ',
 'w', 'h', 'e', 'n', ' ', 'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' ', 'c',
 'o', 'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n', 't', 's', ' ', 'o', 'f', ' ', 'v',
 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ')),
                 (aa: List[Char]) => shows_prec[B](Zero_nat, x, aa))))),
                           (x: (List[Char]) => List[Char]) =>
                             Inl[(List[Char]) => List[Char],
                                  Unit](comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm',
   ' ', 'w', 'h', 'e', 'n', ' ', 'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g',
   ' ')),
                   comp[List[Char], List[Char],
                         List[Char]](shows_lpoly[A,
          B](ze, on, LPoly[A, B](a, vas)),
                                      comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List(' ', '>', '=', ' ')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_lpoly[A,
        B](ze, on, LPoly[A, B](b, vbs)),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl, x)))))))
}

def check_polo_nsI[A : equal : show, B : show,
                    C : equal : show](ze: A, on: A, pl: A => A => A,
                                       ti: A => A => A, mz: A => A,
                                       ge: A => A => Boolean,
                                       pi: B => nat => (A, List[A]),
                                       x7: (term[B, C], term[B, C])):
      sum[(List[Char]) => List[Char], Unit] =
  (ze, on, pl, ti, mz, ge, pi, x7) match {
  case (ze, on, pl, ti, mz, ge, pi, (s, t)) =>
    {
      val left: l_poly[C, A] = pleftI[A, B, C](ze, on, pl, ti, mz, pi, s)
      val right: l_poly[C, A] = prightI[A, B, C](ze, on, pl, ti, mz, pi, t);
      catcha[(List[Char]) => List[Char], Unit,
              (List[Char]) =>
                List[Char]](check_lpoly_nsI[A, C](ze, on, ge, left, right),
                             (x: (List[Char]) => List[Char]) =>
                               Inl[(List[Char]) => List[Char],
                                    Unit](comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ', 'n',
     'o', 't', ' ', 'e', 'n', 's', 'u', 'r', 'e', ' ')),
                     comp[List[Char], List[Char],
                           List[Char]]((shows_prec_term[B,
                 C](Zero_nat)).apply(s),
comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List(' ', '>', '=', ' ')),
                   comp[List[Char], List[Char],
                         List[Char]]((shows_prec_term[B, C](Zero_nat)).apply(t),
                                      comp[List[Char], List[Char],
    List[Char]](shows_nl, x)))))))
    }
}

def to_lpoly_inter[A, B : equal](d: A, on: A, i: List[((B, nat), (A, List[A]))],
                                  f: B, n: nat):
      (A, List[A]) =
  lookupa[(B, nat), (A, List[A])]((f, n), i) match {
    case None => (d, replicate[A](n, on))
    case Some(cas) => cas
  }

def check_poly_monoI[A : show,
                      B : show](ze: A, ge: A => A => Boolean, cm: A => Boolean):
      (List[((B, nat), (A, List[A]))]) =>
        sum[(List[Char]) => List[Char], Unit] =
  (xs: List[((B, nat), (A, List[A]))]) =>
    catcha[(((B, nat), (A, List[A])), (List[Char]) => List[Char]), Unit,
            (List[Char]) =>
              List[Char]](forallm[((B, nat), (A, List[A])),
                                   (List[Char]) =>
                                     List[Char]]((a: ((B, nat), (A, List[A])))
           => {
                val (aa, b): ((B, nat), (A, List[A])) = a;
                ({
                   val (f, n): (B, nat) = aa;
                   (ab: (A, List[A])) =>
                     {
                       val (c, cs): (A, List[A]) = ab;
                       catcha[(List[Char]) => List[Char], Unit,
                               (List[Char]) =>
                                 List[Char]](bindb[(List[Char]) => List[Char],
            Unit,
            Unit](check[(List[Char]) =>
                          List[Char]]((ge(c))(ze),
                                       comp[List[Char], List[Char],
     List[Char]](shows_string.apply(List('c', 'o', 'n', 's', 't', 'a', 'n', 't',
  ' ', 'p', 'a', 'r', 't', ' ')),
                  comp[List[Char], List[Char],
                        List[Char]]((ac: List[Char]) =>
                                      shows_prec[A](Zero_nat, c, ac),
                                     comp[List[Char], List[Char],
   List[Char]](shows_string.apply(List(' ', 'm', 'u', 's', 't', ' ', 'b', 'e',
' ', 'a', 't', ' ', 'l', 'e', 'a', 's', 't', ' ')),
                comp[List[Char], List[Char],
                      List[Char]]((ac: List[Char]) =>
                                    shows_prec[A](Zero_nat, ze, ac),
                                   shows_nl))))),
                   (_: Unit) =>
                     bindb[(List[Char]) => List[Char], Unit,
                            Unit](check[(List[Char]) =>
  List[Char]](less_eq_nat(n, size_list[A](cs)),
               comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List('t', 'h', 'e', ' ',
                  'l', 'a', 's', 't', ' ', 'a', 'r', 'g', 'u', 'm', 'e', 'n',
                  't', ' ', 'i', 's', ' ', 'i', 'g', 'n', 'o', 'r', 'e', 'd')),
                                  shows_nl)),
                                   (_: Unit) =>
                                     catcha[(A, (List[Char]) => List[Char]),
     Unit,
     (List[Char]) =>
       List[Char]](forallm[A, (List[Char]) =>
                                List[Char]]((d: A) =>
      check[(List[Char]) =>
              List[Char]](cm(d),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('c', 'o',
                              'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n', 't',
                              ' ')),
      comp[List[Char], List[Char],
            List[Char]]((ac: List[Char]) => shows_prec[A](Zero_nat, d, ac),
                         comp[List[Char], List[Char],
                               List[Char]](shows_string.apply(List(' ', 'i',
                            's', ' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o',
                            'w', 'e', 'd')),
    shows_nl)))),
     cs),
                    (x: (A, (List[Char]) => List[Char])) =>
                      Inl[(List[Char]) => List[Char],
                           Unit](snd[A, (List[Char]) => List[Char]](x))))),
      (x: (List[Char]) => List[Char]) =>
        Inl[(List[Char]) => List[Char],
             Unit](comp[List[Char], List[Char],
                         List[Char]](shows_string.apply(List('p', 'r', 'o', 'b',
                      'l', 'e', 'm', ' ', 'w', 'i', 't', 'h', ' ', 'm', 'o',
                      'n', 'o', 't', 'o', 'n', 'i', 'c', 'i', 't', 'y', ' ',
                      'd', 'u', 'e', ' ', 't', 'o', ' ', 'i', 'n', 't', 'e',
                      'r', 'p', 'r', 'e', 't', 'a', 't', 'i', 'o', 'n', ' ',
                      'o', 'f', ' ')),
                                      comp[List[Char], List[Char],
    List[Char]]((ac: List[Char]) => shows_prec[B](Zero_nat, f, ac),
                 comp[List[Char], List[Char],
                       List[Char]](shows_string.apply(List('/')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_prec_nat(Zero_nat, n),
               comp[List[Char], List[Char], List[Char]](shows_nl, x)))))))
                     }
                 })(b)
              },
          xs),
                           (x: (((B, nat), (A, List[A])),
                                 (List[Char]) => List[Char]))
                             => Inl[(List[Char]) => List[Char],
                                     Unit](snd[((B, nat), (A, List[A])),
        (List[Char]) => List[Char]](x)))

def create_lpoly_repr[A : equal : show,
                       B : equal : show](c: poly_c_ext[A, Unit],
  i: List[((B, nat), (A, List[A]))]):
      (List[Char]) => List[Char] =
  {
    val pi: B => nat => (A, List[A]) =
      (a: B) =>
        (b: nat) =>
          to_lpoly_inter[A, B](default[A, Unit](c), onea[A, Unit](c), i, a, b);
    comp[List[Char], List[Char],
          List[Char]](shows_string.apply(List('p', 'o', 'l', 'y', 'n', 'o', 'm',
       'i', 'a', 'l', ' ', 'i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't', 'r',
       'a', 't', 'i', 'o', 'n', ' ', 'o', 'v', 'e', 'r', ' ')),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(description[A,
                                 Unit](c)),
  comp[List[Char], List[Char],
        List[Char]](shows_nl,
                     shows_sep[(B, nat)]((a: (B, nat)) =>
   {
     val (f, n): (B, nat) = a
     val t: term[B, List[Char]] =
       Fun[B, List[Char]](f, mapa[List[Char],
                                   term[B,
 List[Char]]]((aa: List[Char]) => Var[List[Char], B](aa),
               fresh_strings_list(List('x', '_'), one_nat, Nil, n)));
     comp[List[Char], List[Char],
           List[Char]](shows_string.apply(List('P', 'o', 'l', '(')),
                        comp[List[Char], List[Char],
                              List[Char]]((shows_prec_term[B,
                    List[Char]](Zero_nat)).apply(t),
   comp[List[Char], List[Char],
         List[Char]](shows_string.apply(List(')', ' ', '=', ' ')),
                      shows_lpoly[A, List[Char]](zeroa[A, Unit](c),
          onea[A, Unit](c),
          pleftI[A, B,
                  List[Char]](zeroa[A, Unit](c), onea[A, Unit](c),
                               plusa[A, Unit](c), timesa[A, Unit](c),
                               max0a[A, Unit](c), pi, t)))))
   },
  shows_nl,
  remdups[(B, nat)](mapa[((B, nat), (A, List[A])),
                          (B, nat)]((a: ((B, nat), (A, List[A]))) =>
                                      fst[(B, nat), (A, List[A])](a),
                                     i))))))
  }

def check_lpoly_coeffs[A : show,
                        B : show](wfe: A => Boolean, apos: A => Boolean,
                                   ch: A => Boolean,
                                   i: List[((B, nat), (A, List[A]))]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(((B, nat), (A, List[A])), (List[Char]) => List[Char]), Unit,
          (List[Char]) =>
            List[Char]](forallm[((B, nat), (A, List[A])),
                                 (List[Char]) =>
                                   List[Char]]((a: ((B, nat), (A, List[A]))) =>
         {
           val (aa, b): ((B, nat), (A, List[A])) = a;
           ({
              val (f, n): (B, nat) = aa;
              (ab: (A, List[A])) =>
                {
                  val (c, cs): (A, List[A]) = ab;
                  catcha[(List[Char]) => List[Char], Unit,
                          (List[Char]) =>
                            List[Char]](bindb[(List[Char]) => List[Char], Unit,
       Unit](check[(List[Char]) =>
                     List[Char]](wfe(c),
                                  comp[List[Char], List[Char],
List[Char]](shows_string.apply(List('c', 'o', 'n', 's', 't', 'a', 'n', 't', ' ',
                                     'p', 'a', 'r', 't', ' ')),
             comp[List[Char], List[Char],
                   List[Char]]((ac: List[Char]) =>
                                 shows_prec[A](Zero_nat, c, ac),
                                comp[List[Char], List[Char],
                                      List[Char]](shows_string.apply(List(' ',
                                   'i', 's', ' ', 'n', 'o', 't', ' ', 'w', 'e',
                                   'l', 'l', '-', 'f', 'o', 'r', 'm', 'e',
                                   'd')),
           shows_nl)))),
              (_: Unit) =>
                bindb[(List[Char]) => List[Char], Unit,
                       Unit](check[(List[Char]) =>
                                     List[Char]](apos(c) ||
           list_ex[A](apos, take[A](n, cs)),
          comp[List[Char], List[Char],
                List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd',
             ' ', 'n', 'o', 't', ' ', 'f', 'i', 'n', 'd', ' ', 'p', 'o', 's',
             'i', 't', 'i', 'v', 'e', ' ', 'e', 'n', 't', 'r', 'y', ' ', 'w',
             'h', 'i', 'c', 'h', ' ', 'i', 's', ' ', 'r', 'e', 'q', 'u', 'i',
             'r', 'e', 'd', ' ', 'f', 'o', 'r', ' ', 'a', 'r', 'c', 't', 'i',
             'c', ' ', 'i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't', 'a', 't',
             'i', 'o', 'n', 's')),
                             shows_nl)),
                              (_: Unit) =>
                                catcha[(A, (List[Char]) => List[Char]), Unit,
(List[Char]) =>
  List[Char]](forallm[A, (List[Char]) =>
                           List[Char]]((ac: A) =>
 check[(List[Char]) =>
         List[Char]](ch(ac) && wfe(ac),
                      comp[List[Char], List[Char],
                            List[Char]](shows_string.apply(List('c', 'o', 'e',
                         'f', 'f', 'i', 'c', 'i', 'e', 'n', 't', ' ')),
 comp[List[Char], List[Char],
       List[Char]]((ba: List[Char]) => shows_prec[A](Zero_nat, ac, ba),
                    comp[List[Char], List[Char],
                          List[Char]](shows_string.apply(List(' ', 'i', 's',
                       ' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e',
                       'd')),
                                       shows_nl)))),
cs),
               (x: (A, (List[Char]) => List[Char])) =>
                 Inl[(List[Char]) => List[Char],
                      Unit](snd[A, (List[Char]) => List[Char]](x))))),
 (x: (List[Char]) => List[Char]) =>
   Inl[(List[Char]) => List[Char],
        Unit](comp[List[Char], List[Char],
                    List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l',
                 'e', 'm', ' ', 'w', 'i', 't', 'h', ' ', 'i', 'n', 't', 'e',
                 'r', 'p', 'r', 'e', 't', 'a', 't', 'i', 'o', 'n', ' ', 'o',
                 'f', ' ')),
                                 comp[List[Char], List[Char],
                                       List[Char]]((ac: List[Char]) =>
             shows_prec[B](Zero_nat, f, ac),
            comp[List[Char], List[Char],
                  List[Char]](shows_string.apply(List('/')),
                               comp[List[Char], List[Char],
                                     List[Char]](shows_prec_nat(Zero_nat, n),
          comp[List[Char], List[Char], List[Char]](shows_nl, x)))))))
                }
            })(b)
         },
        i),
                         (x: (((B, nat), (A, List[A])),
                               (List[Char]) => List[Char]))
                           => Inl[(List[Char]) => List[Char],
                                   Unit](snd[((B, nat), (A, List[A])),
      (List[Char]) => List[Char]](x)))

def create_poly_redtriple[A : equal : show, B : equal : show,
                           C : equal : show](cI:
       (List[((A, nat), (B, List[B]))]) =>
         sum[(List[Char]) => List[Char], Unit],
      c: poly_c_ext[B, Unit], i: List[((A, nat), (B, List[B]))]):
      redtriple_ext[A, C, Unit] =
  {
    val pi: A => nat => (B, List[B]) =
      (a: A) =>
        (b: nat) =>
          to_lpoly_inter[B, A](default[B, Unit](c), onea[B, Unit](c), i, a, b)
    val ns: ((term[A, C], term[A, C])) => sum[(List[Char]) => List[Char], Unit]
      = (a: (term[A, C], term[A, C])) =>
          check_polo_nsI[B, A,
                          C](zeroa[B, Unit](c), onea[B, Unit](c),
                              plusa[B, Unit](c), timesa[B, Unit](c),
                              max0a[B, Unit](c), gea[B, Unit](c), pi, a);
    Redtriple_ext[A, C,
                   Unit](bindb[(List[Char]) => List[Char], Unit,
                                Unit](cI(i),
                                       (_: Unit) =>
 check_lpoly_coeffs[B, A](wf[B, Unit](c), arc_pos[B, Unit](c),
                           (x: B) =>
                             (gea[B, Unit](c)).apply(x).apply(zeroa[B,
                             Unit](c)),
                           i)),
                          (a: (term[A, C], term[A, C])) =>
                            check_polo_sI[B, A,
   C](monoa[B, Unit](c), zeroa[B, Unit](c), onea[B, Unit](c), plusa[B, Unit](c),
       timesa[B, Unit](c), max0a[B, Unit](c), gea[B, Unit](c), gt[B, Unit](c),
       pi, a),
                          ns, ns,
                          (a: (A, nat)) =>
                            create_afI[B, A](zeroa[B, Unit](c), i, a),
                          bindb[(List[Char]) => List[Char], Unit,
                                 Unit](check[(List[Char]) =>
       List[Char]](monoa[B, Unit](c),
                    shows_string.apply(List('m', 'o', 'n', 'o', 't', 'o', 'n',
     'e', ' ', 'p', 'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l', 's', ' ', 'n',
     'o', 't', ' ', 's', 'u', 'p', 'p', 'o', 'r', 't', ' ', 'f', 'o', 'r', ' ',
     't', 'h', 'i', 's', ' ', 'd', 'o', 'm', 'a', 'i', 'n'))),
(_: Unit) =>
  (check_poly_monoI[B, A](zeroa[B, Unit](c), gea[B, Unit](c),
                           check_mono[B, Unit](c))).apply(i)),
                          create_lpoly_repr[B, A](c, i), ())
  }

def prec_repr_to_pr[A : equal](prs: List[((A, nat), (nat, order_tag))],
                                x1: filtered[A], uu: nat):
      nat =
  (prs, x1, uu) match {
  case (prs, FPair(f, a), uu) =>
    lookupa[(A, nat), (nat, order_tag)]((f, a), prs) match {
      case None => Zero_nat
      case Some(aa) => fst[nat, order_tag](aa)
    }
}

def faulty_redtriple[A : show, B : show](s: List[Char]):
      redtriple_ext[A, B, Unit] =
  Redtriple_ext[A, B,
                 Unit](Inl[(List[Char]) => List[Char],
                            Unit](shows_string.apply(s)),
                        (_: (term[A, B], term[A, B])) =>
                          Inr[Unit, (List[Char]) => List[Char]](()),
                        (_: (term[A, B], term[A, B])) =>
                          Inr[Unit, (List[Char]) => List[Char]](()),
                        (_: (term[A, B], term[A, B])) =>
                          Inr[Unit, (List[Char]) => List[Char]](()),
                        (a: (A, nat)) => (b: nat) => full_af[A](a, b),
                        Inr[Unit, (List[Char]) => List[Char]](()),
                        id[List[Char]], ())

def prec_repr_to_status[A : equal](prs: List[((A, nat), (nat, order_tag))],
                                    x1: filtered[A], uu: nat):
      order_tag =
  (prs, x1, uu) match {
  case (prs, FPair(f, a), uu) =>
    lookupa[(A, nat), (nat, order_tag)]((f, a), prs) match {
      case None => Lex
      case Some(aa) => snd[nat, order_tag](aa)
    }
}

def get_redtriple[A : equal : show, B : equal : show](x0: redtriple_impl[A]):
      redtriple_ext[A, B, Unit] =
  x0 match {
  case Int_carrier(i) =>
    create_poly_redtriple[A, BigInt,
                           B]((_: List[((A, nat), (BigInt, List[BigInt]))]) =>
                                Inr[Unit, (List[Char]) => List[Char]](()),
                               int_components, i)
  case Int_nl_carrier(i) =>
    create_nlpoly_redtriple[BigInt, A,
                             B](BigInt(1), (x: BigInt) => (y: BigInt) => y < x,
                                 true, true, i)
  case Rat_carrier(i) =>
    create_poly_redtriple[A, rat,
                           B]((_: List[((A, nat), (rat, List[rat]))]) =>
                                Inr[Unit, (List[Char]) => List[Char]](()),
                               rat_components, i)
  case Rat_nl_carrier(d, i) =>
    (if (less_rat(fract(BigInt(0), BigInt(1)), d))
      create_nlpoly_redtriple[rat, A,
                               B](d, rat_gt(d),
                                   less_eq_rat(fract(BigInt(1), BigInt(1)), d),
                                   false, i)
      else faulty_redtriple[A, B](List('d', 'e', 'l', 't', 'a', ' ', 'm', 'u',
's', 't', ' ', 'b', 'e', ' ', 'l', 'a', 'r', 'g', 'e', 'r', ' ', 't', 'h', 'a',
'n', ' ', 'z', 'e', 'r', 'o')))
  case Arctic_carrier(i) =>
    create_poly_redtriple[A, arctic,
                           B]((_: List[((A, nat), (arctic, List[arctic]))]) =>
                                Inr[Unit, (List[Char]) => List[Char]](()),
                               arctic_components, i)
  case Arctic_rat_carrier(i) =>
    create_poly_redtriple[A, arctic_rat,
                           B]((_: List[((A, nat),
 (arctic_rat, List[arctic_rat]))])
                                => Inr[Unit, (List[Char]) => List[Char]](()),
                               arctic_rat_components, i)
  case Int_mat_carrier(n, sd, i) =>
    create_poly_redtriple[A, List[List[BigInt]],
                           B]((_: List[((A, nat),
 (List[List[BigInt]], List[List[List[BigInt]]]))])
                                => check[(List[Char]) =>
   List[Char]](less_nat(Zero_nat, sd) && less_eq_nat(sd, n),
                shows_string.apply(List('i', 'n', 'v', 'a', 'l', 'i', 'd', ' ',
 'd', 'i', 'm', 'e', 'n', 's', 'i', 'o', 'n'))),
                               Poly_c_ext[List[List[BigInt]],
   Unit]((a: List[List[BigInt]]) => mat[BigInt](n, n, a),
          mat0I[BigInt](BigInt(0), n, n),
          mat1I[BigInt](BigInt(0), BigInt(1), n),
          mat1I[BigInt](BigInt(0), default[BigInt, Unit](int_components), n),
          (a: List[List[BigInt]]) =>
            (b: List[List[BigInt]]) =>
              mat_plusI[BigInt]((aa: BigInt) => (ba: BigInt) => aa + ba, a, b),
          (a: List[List[BigInt]]) =>
            (b: List[List[BigInt]]) =>
              mat_multI[BigInt](BigInt(0),
                                 (aa: BigInt) => (ba: BigInt) => aa + ba,
                                 (aa: BigInt) => (ba: BigInt) => aa * ba, n, a,
                                 b),
          mat_max0[BigInt], mat_ge[BigInt],
          (a: List[List[BigInt]]) =>
            (b: List[List[BigInt]]) =>
              mat_gtI[BigInt](gt[BigInt, Unit](int_components), sd, a, b),
          monoa[BigInt, Unit](int_components),
          concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil), List('x'),
                             (shows_prec_nat(Zero_nat, n)).apply(Nil),
                             List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's',
                                   ' ', 'w', 'i', 't', 'h', ' ', 's', 't', 'r',
                                   'i', 'c', 't', ' ', 'd', 'i', 'm', 'e', 'n',
                                   's', 'i', 'o', 'n', ' '),
                             (shows_prec_nat(Zero_nat, sd)).apply(Nil),
                             List(' ', 'o', 'v', 'e', 'r', ' '),
                             description[BigInt, Unit](int_components))),
          (_: List[List[BigInt]]) => true,
          mat_mono[BigInt](sd, check_mono[BigInt, Unit](int_components)), ()),
                               i)
  case Rat_mat_carrier(n, sd, i) =>
    create_poly_redtriple[A, List[List[rat]],
                           B]((_: List[((A, nat),
 (List[List[rat]], List[List[List[rat]]]))])
                                => check[(List[Char]) =>
   List[Char]](less_nat(Zero_nat, sd) && less_eq_nat(sd, n),
                shows_string.apply(List('i', 'n', 'v', 'a', 'l', 'i', 'd', ' ',
 'd', 'i', 'm', 'e', 'n', 's', 'i', 'o', 'n'))),
                               Poly_c_ext[List[List[rat]],
   Unit]((a: List[List[rat]]) => mat[rat](n, n, a),
          mat0I[rat](fract(BigInt(0), BigInt(1)), n, n),
          mat1I[rat](fract(BigInt(0), BigInt(1)), fract(BigInt(1), BigInt(1)),
                      n),
          mat1I[rat](fract(BigInt(0), BigInt(1)),
                      default[rat, Unit](rat_components), n),
          (a: List[List[rat]]) =>
            (b: List[List[rat]]) =>
              mat_plusI[rat]((aa: rat) => (ba: rat) => plus_rata(aa, ba), a, b),
          (a: List[List[rat]]) =>
            (b: List[List[rat]]) =>
              mat_multI[rat](fract(BigInt(0), BigInt(1)),
                              (aa: rat) => (ba: rat) => plus_rata(aa, ba),
                              (aa: rat) => (ba: rat) => times_rata(aa, ba), n,
                              a, b),
          mat_max0[rat], mat_ge[rat],
          (a: List[List[rat]]) =>
            (b: List[List[rat]]) =>
              mat_gtI[rat](gt[rat, Unit](rat_components), sd, a, b),
          monoa[rat, Unit](rat_components),
          concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil), List('x'),
                             (shows_prec_nat(Zero_nat, n)).apply(Nil),
                             List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's',
                                   ' ', 'w', 'i', 't', 'h', ' ', 's', 't', 'r',
                                   'i', 'c', 't', ' ', 'd', 'i', 'm', 'e', 'n',
                                   's', 'i', 'o', 'n', ' '),
                             (shows_prec_nat(Zero_nat, sd)).apply(Nil),
                             List(' ', 'o', 'v', 'e', 'r', ' '),
                             description[rat, Unit](rat_components))),
          (_: List[List[rat]]) => true,
          mat_mono[rat](sd, check_mono[rat, Unit](rat_components)), ()),
                               i)
  case Arc_mat_carrier(n, i) =>
    create_poly_redtriple[A, List[List[arctic]],
                           B]((_: List[((A, nat),
 (List[List[arctic]], List[List[List[arctic]]]))])
                                => check[(List[Char]) =>
   List[Char]](less_nat(Zero_nat, n),
                shows_string.apply(List('i', 'n', 'v', 'a', 'l', 'i', 'd', ' ',
 'd', 'i', 'm', 'e', 'n', 's', 'i', 'o', 'n'))),
                               Poly_c_ext[List[List[arctic]],
   Unit]((a: List[List[arctic]]) => mat[arctic](n, n, a),
          mat0I[arctic](zero_arctic, n, n),
          mat1I[arctic](zero_arctic, one_arctic, n),
          mat1I[arctic](zero_arctic, default[arctic, Unit](arctic_components),
                         n),
          (a: List[List[arctic]]) =>
            (b: List[List[arctic]]) =>
              mat_plusI[arctic]((aa: arctic) =>
                                  (ba: arctic) => plus_arctic(aa, ba),
                                 a, b),
          (a: List[List[arctic]]) =>
            (b: List[List[arctic]]) =>
              mat_multI[arctic](zero_arctic,
                                 (aa: arctic) =>
                                   (ba: arctic) => plus_arctic(aa, ba),
                                 (aa: arctic) =>
                                   (ba: arctic) => times_arctic(aa, ba),
                                 n, a, b),
          id[List[List[arctic]]], mat_ge[arctic],
          (a: List[List[arctic]]) =>
            (b: List[List[arctic]]) =>
              mat_comp_all[arctic](gt[arctic, Unit](arctic_components), a, b),
          false,
          concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil), List('x'),
                             (shows_prec_nat(Zero_nat, n)).apply(Nil),
                             List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's',
                                   ' ', 'o', 'v', 'e', 'r', ' '),
                             description[arctic, Unit](arctic_components))),
          (a: List[List[arctic]]) =>
            mat_arc_posI[arctic](arc_pos[arctic, Unit](arctic_components), a),
          (_: List[List[arctic]]) => false, ()),
                               i)
  case Arc_rat_mat_carrier(n, i) =>
    create_poly_redtriple[A, List[List[arctic_rat]],
                           B]((_: List[((A, nat),
 (List[List[arctic_rat]], List[List[List[arctic_rat]]]))])
                                => check[(List[Char]) =>
   List[Char]](less_nat(Zero_nat, n),
                shows_string.apply(List('i', 'n', 'v', 'a', 'l', 'i', 'd', ' ',
 'd', 'i', 'm', 'e', 'n', 's', 'i', 'o', 'n'))),
                               Poly_c_ext[List[List[arctic_rat]],
   Unit]((a: List[List[arctic_rat]]) => mat[arctic_rat](n, n, a),
          mat0I[arctic_rat](zero_arctic_rat, n, n),
          mat1I[arctic_rat](zero_arctic_rat, one_arctic_rat, n),
          mat1I[arctic_rat](zero_arctic_rat,
                             default[arctic_rat, Unit](arctic_rat_components),
                             n),
          (a: List[List[arctic_rat]]) =>
            (b: List[List[arctic_rat]]) =>
              mat_plusI[arctic_rat]((aa: arctic_rat) =>
                                      (ba: arctic_rat) =>
plus_arctic_rat(aa, ba),
                                     a, b),
          (a: List[List[arctic_rat]]) =>
            (b: List[List[arctic_rat]]) =>
              mat_multI[arctic_rat](zero_arctic_rat,
                                     (aa: arctic_rat) =>
                                       (ba: arctic_rat) =>
 plus_arctic_rat(aa, ba),
                                     (aa: arctic_rat) =>
                                       (ba: arctic_rat) =>
 times_arctic_rat(aa, ba),
                                     n, a, b),
          id[List[List[arctic_rat]]], mat_ge[arctic_rat],
          (a: List[List[arctic_rat]]) =>
            (b: List[List[arctic_rat]]) =>
              mat_comp_all[arctic_rat](gt[arctic_rat,
   Unit](arctic_rat_components),
a, b),
          false,
          concat[Char](List((shows_prec_nat(Zero_nat, n)).apply(Nil), List('x'),
                             (shows_prec_nat(Zero_nat, n)).apply(Nil),
                             List(' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's',
                                   ' ', 'o', 'v', 'e', 'r', ' '),
                             description[arctic_rat,
  Unit](arctic_rat_components))),
          (a: List[List[arctic_rat]]) =>
            mat_arc_posI[arctic_rat](arc_pos[arctic_rat,
      Unit](arctic_rat_components),
                                      a),
          (_: List[List[arctic_rat]]) => false, ()),
                               i)
  case Rpo(prec_tau, pi) =>
    af_redtriple[A, B](pi, create_RPO_redtriple[A, filtered[A],
         B]((pr: List[((A, nat), (nat, order_tag))]) =>
              ((a: filtered[A]) => (b: nat) => prec_repr_to_pr[A](pr, a, b),
                (a: filtered[A]) =>
                  (b: nat) => prec_repr_to_status[A](pr, a, b)),
             prec_tau))
}

def intersect_rulesb[A, B, C, D](x0: dpp_ops_ext[A, B, C, D]):
      A => (List[(term[B, C], term[B, C])]) => A =
  x0 match {
  case Dpp_ops_ext(dpp, p, pw, pairs, q, r, rw, rules, q_empty,
                    rules_no_left_var, rules_non_collapsing, is_QNF,
                    nFQ_subset_NF_rules, rules_map, reverse_rules_map,
                    intersect_pairs, replace_pair, intersect_rules, delete_P_Pw,
                    delete_R_Rw, split_pairs, split_rules, mk, more)
    => intersect_rules
}

def usable_rules_proc[A, B : equal : linorder : show](i:
                dpp_ops_ext[A, B, List[Char], Unit],
               u: List[(term[B, List[Char]], term[B, List[Char]])], dpp: A):
      sum[(List[Char]) => List[Char], A] =
  bindb[(List[Char]) => List[Char], Unit,
         Unit](check[(List[Char]) =>
                       List[Char]]((nFQ_subset_NF_rulesb[A, B, List[Char],
                  Unit](i)).apply(dpp),
                                    shows_prec_list[Char](Zero_nat,
                   List('i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', ' ', 'r',
                         'e', 'w', 'r', 'i', 't', 'i', 'n', 'g', ' ', 'r', 'e',
                         'q', 'u', 'i', 'r', 'e', 'd'))),
                (_: Unit) =>
                  {
                    val p: List[(term[B, List[Char]], term[B, List[Char]])] =
                      (pairsa[A, B, List[Char], Unit](i)).apply(dpp)
                    val urc: (List[term[B, List[Char]]]) =>
                               (term[B, List[Char]]) => Boolean
                      = (is_ur_closed_impl_dpp_mv[A, B](i, dpp)).apply(u)
                    val check_urc:
                          (List[term[B, List[Char]]]) =>
                            (term[B, List[Char]]) =>
                              sum[(List[Char]) => List[Char], Unit]
                      = (s: List[term[B, List[Char]]]) =>
                          (t: term[B, List[Char]]) =>
                            check[(List[Char]) =>
                                    List[Char]]((urc(s))(t),
         comp[List[Char], List[Char],
               List[Char]](shows_prec_list[Char](Zero_nat,
          List('t', 'e', 'r', 'm', ' ')),
                            comp[List[Char], List[Char],
                                  List[Char]]((shows_prec_term[B,
                        List[Char]](Zero_nat)).apply(t),
       shows_prec_list[Char](Zero_nat,
                              List(' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'c',
                                    'l', 'o', 's', 'e', 'd', ' ', 'u', 'n', 'd',
                                    'e', 'r', ' ', 'u', 's', 'a', 'b', 'l', 'e',
                                    ' ', 'r', 'u', 'l', 'e', 's')))));
                    bindb[(List[Char]) => List[Char], Unit,
                           Unit](catcha[((term[B, List[Char]],
   term[B, List[Char]]),
  (List[Char]) => List[Char]),
 Unit,
 (List[Char]) =>
   List[Char]](forallm[(term[B, List[Char]], term[B, List[Char]]),
                        (List[Char]) =>
                          List[Char]]((a:
 (term[B, List[Char]], term[B, List[Char]]))
=> {
     val (l, r): (term[B, List[Char]], term[B, List[Char]]) = a;
     bindb[(List[Char]) => List[Char], Unit,
            Unit](catcha[List[Char], Unit,
                          (List[Char]) =>
                            List[Char]](check_subseteq[List[Char]](vars_term_list[B,
   List[Char]].apply(r),
                            vars_term_list[B, List[Char]].apply(l)),
 (_: List[Char]) =>
   Inl[(List[Char]) => List[Char],
        Unit](shows_prec_list[Char](Zero_nat,
                                     List('v', 'a', 'r', 'i', 'a', 'b', 'l',
   'e', ' ', 'c', 'o', 'n', 'd', 'i', 't', 'i', 'o', 'n', ' ', 'i', 'n', ' ',
   'P', ' ', 'v', 'i', 'o', 'l', 'a', 't', 'e', 'd')))),
                   (_: Unit) => (check_urc(List(l)))(r))
   },
                                       p),
                (x: ((term[B, List[Char]], term[B, List[Char]]),
                      (List[Char]) => List[Char]))
                  => Inl[(List[Char]) => List[Char],
                          Unit](snd[(term[B, List[Char]], term[B, List[Char]]),
                                     (List[Char]) => List[Char]](x))),
                                  (_: Unit) =>
                                    catcha[((term[B, List[Char]],
      term[B, List[Char]]),
     (List[Char]) => List[Char]),
    Unit,
    (List[Char]) =>
      List[Char]](forallm[(term[B, List[Char]], term[B, List[Char]]),
                           (List[Char]) =>
                             List[Char]]((a:
    (term[B, List[Char]], term[B, List[Char]]))
   => {
        val (l, aa): (term[B, List[Char]], term[B, List[Char]]) = a;
        (check_urc(args[B, List[Char]](l)))(aa)
      },
  u),
                   (x: ((term[B, List[Char]], term[B, List[Char]]),
                         (List[Char]) => List[Char]))
                     => Inl[(List[Char]) => List[Char],
                             Unit](snd[(term[B, List[Char]],
 term[B, List[Char]]),
(List[Char]) => List[Char]](x))))
                  })
    match {
    case Inl(a) => Inl[(List[Char]) => List[Char], A](a)
    case Inr(_) =>
      Inr[A, (List[Char]) =>
               List[Char]]((intersect_rulesb[A, B, List[Char],
      Unit](i)).apply(dpp).apply(u))
  }

def forward_instantiation_proc[A, B : equal : linorder : show](i:
                         dpp_ops_ext[A, B, List[Char], Unit],
                        st: (term[B, List[Char]], term[B, List[Char]]),
                        sts: List[(term[B, List[Char]], term[B, List[Char]])],
                        u_opt:
                          Option[List[(term[B, List[Char]],
term[B, List[Char]])]],
                        dpp: A):
      sum[(List[Char]) => List[Char], A] =
  {
    val isnf: (term[B, List[Char]]) => Boolean =
      (is_QNFc[A, B, List[Char], Unit](i)).apply(dpp)
    val (s, t): (term[B, List[Char]], term[B, List[Char]]) = st
    val sy: term[B, List[Char]] =
      map_vars[List[Char], List[Char], B]((a: List[Char]) => 'y' :: a, s)
    val ty: term[B, List[Char]] =
      map_vars[List[Char], List[Char], B]((a: List[Char]) => 'y' :: a, t)
    val u: List[(term[B, List[Char]], term[B, List[Char]])] =
      u_opt match {
        case None => (rulesd[A, B, List[Char], Unit](i)).apply(dpp)
        case Some(u) => u
      };
    bindb[(List[Char]) => List[Char], Unit,
           Unit]((if (is_none[List[(term[B, List[Char]],
                                     term[B, List[Char]])]](u_opt))
                   Inr[Unit, (List[Char]) => List[Char]](())
                   else {
                          val urc: (List[term[B, List[Char]]]) =>
                                     (term[B, List[Char]]) => Boolean
                            = (is_ur_closed_impl_dpp_mv[A, B](i, dpp)).apply(u)
                          val check_urc:
                                (List[term[B, List[Char]]]) =>
                                  (term[B, List[Char]]) =>
                                    sum[(List[Char]) => List[Char], Unit]
                            = (sa: List[term[B, List[Char]]]) =>
                                (ta: term[B, List[Char]]) =>
                                  check[(List[Char]) =>
  List[Char]]((urc(sa))(ta),
               comp[List[Char], List[Char],
                     List[Char]](shows_prec_list[Char](Zero_nat,
                List('t', 'e', 'r', 'm', ' ')),
                                  comp[List[Char], List[Char],
List[Char]]((shows_prec_term[B, List[Char]](Zero_nat)).apply(ta),
             shows_prec_list[Char](Zero_nat,
                                    List(' ', 'i', 's', ' ', 'n', 'o', 't', ' ',
  'c', 'l', 'o', 's', 'e', 'd', ' ', 'u', 'n', 'd', 'e', 'r', ' ', 'u', 's',
  'a', 'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e', 's')))));
                          bindb[(List[Char]) => List[Char], Unit,
                                 Unit](check[(List[Char]) =>
       List[Char]]((nFQ_subset_NF_rulesb[A, B, List[Char], Unit](i)).apply(dpp),
                    shows_prec_list[Char](Zero_nat,
   List('i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', ' ', 'r', 'e', 'w', 'r',
         'i', 't', 'i', 'n', 'g', ' ', 'r', 'e', 'q', 'u', 'i', 'r', 'e',
         'd'))),
(_: Unit) =>
  bindb[(List[Char]) => List[Char], Unit,
         Unit](catcha[((term[B, List[Char]], term[B, List[Char]]),
                        (List[Char]) => List[Char]),
                       Unit,
                       (List[Char]) =>
                         List[Char]](forallm[(term[B, List[Char]],
       term[B, List[Char]]),
      (List[Char]) =>
        List[Char]]((a: (term[B, List[Char]], term[B, List[Char]])) =>
                      {
                        val (l, aa): (term[B, List[Char]], term[B, List[Char]])
                          = a;
                        (check_urc(args[B, List[Char]](l)))(aa)
                      },
                     u),
                                      (x:
 ((term[B, List[Char]], term[B, List[Char]]), (List[Char]) => List[Char]))
=> Inl[(List[Char]) => List[Char],
        Unit](snd[(term[B, List[Char]], term[B, List[Char]]),
                   (List[Char]) => List[Char]](x))),
                (_: Unit) =>
                  bindb[(List[Char]) => List[Char], Unit,
                         Unit]((check_urc(List(s)))(t),
                                (_: Unit) =>
                                  catcha[List[Char], Unit,
  (List[Char]) =>
    List[Char]](check_subseteq[List[Char]](vars_term_list[B,
                   List[Char]].apply(t),
    vars_term_list[B, List[Char]].apply(s)),
                 (_: List[Char]) =>
                   Inl[(List[Char]) => List[Char],
                        Unit](shows_prec_list[Char](Zero_nat,
             List('v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ', 'c', 'o', 'n',
                   'd', 'i', 't', 'i', 'o', 'n', ' ', 'i', 'n', ' ', 'p', 'a',
                   'i', 'r', ' ', 'v', 'i', 'o', 'l', 'a', 't', 'e', 'd')))))))
                        }),
                  (_: Unit) =>
                    {
                      val ur: List[(term[B, List[Char]], term[B, List[Char]])] =
                        mapa[(term[B, List[Char]], term[B, List[Char]]),
                              (term[B, List[Char]],
                                term[B, List[Char]])]((a:
                 (term[B, List[Char]], term[B, List[Char]]))
                => {
                     val (l, r): (term[B, List[Char]], term[B, List[Char]]) = a;
                     (r, l)
                   },
               u)
                      val ic: (term[B, List[Char]]) =>
                                term[B, sum[Unit, List[Char]]]
                        = (icap_impla[B](ur, Nil)).apply(Nil);
                      catcha[((term[B, List[Char]], term[B, List[Char]]),
                               (List[Char]) => List[Char]),
                              Unit,
                              (List[Char]) =>
                                List[Char]](forallm[(term[B, List[Char]],
              term[B, List[Char]]),
             (List[Char]) =>
               List[Char]]((a: (term[B, List[Char]], term[B, List[Char]])) =>
                             {
                               val (ua, v):
                                     (term[B, List[Char]], term[B, List[Char]])
                                 = a;
                               mgu_class[B](ic(ua), t) match {
                                 case None =>
                                   Inr[Unit, (List[Char]) => List[Char]](())
                                 case Some(mu) =>
                                   check[(List[Char]) =>
   List[Char]](! (isnf(subst_apply_term[B, List[Char], List[Char]](sy, mu))) ||
                 (! (isnf(subst_apply_term[B, List[Char],
    List[Char]](map_vars[List[Char], List[Char],
                          B]((aa: List[Char]) => 'x' :: aa, ua),
                 mu))) ||
                   list_ex[(term[B, List[Char]],
                             term[B, List[Char]])]((aa:
              (term[B, List[Char]], term[B, List[Char]]))
             => instance_rule[B, List[Char],
                               List[Char]]((subst_apply_term[B, List[Char],
                      List[Char]](sy, mu),
     subst_apply_term[B, List[Char], List[Char]](ty, mu)),
    aa),
            sts)),
                comp[List[Char], List[Char],
                      List[Char]](shows_prec_list[Char](Zero_nat,
                 List('c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'f',
                       'i', 'n', 'd', ' ', 'i', 'n', 's', 't', 'a', 'n', 'c',
                       'e', ' ', 'o', 'f', ' ', 'p', 'a', 'i', 'r', ' ')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_rule[B, List[Char]]((aa: B) =>
 (b: List[Char]) => shows_prec[B](Zero_nat, aa, b),
(aa: List[Char]) => shows_prec_list[Char](Zero_nat, aa),
List(' ', '-', '>', ' '),
(subst_apply_term[B, List[Char], List[Char]](sy, mu),
  subst_apply_term[B, List[Char], List[Char]](ty, mu))),
              comp[List[Char], List[Char],
                    List[Char]](shows_nl,
                                 comp[List[Char], List[Char],
                                       List[Char]](shows_prec_list[Char](Zero_nat,
                                  List('w', 'h', 'i', 'c', 'h', ' ', 'r', 'e',
's', 'u', 'l', 't', 'e', 'd', ' ', 'f', 'r', 'o', 'm', ' ', 'D', 'P', ' ')),
            shows_rule[B, List[Char]]((aa: B) =>
(b: List[Char]) => shows_prec[B](Zero_nat, aa, b),
                                       (aa: List[Char]) =>
 shows_prec_list[Char](Zero_nat, aa),
                                       List(' ', '-', '>', ' '), (ua, v)))))))
                               }
                             },
                            (pairsa[A, B, List[Char], Unit](i)).apply(dpp)),
     (x: ((term[B, List[Char]], term[B, List[Char]]),
           (List[Char]) => List[Char]))
       => Inl[(List[Char]) => List[Char],
               Unit](snd[(term[B, List[Char]], term[B, List[Char]]),
                          (List[Char]) => List[Char]](x)))
                    })
  }
    match {
    case Inl(a) => Inl[(List[Char]) => List[Char], A](a)
    case Inr(_) =>
      Inr[A, (List[Char]) =>
               List[Char]]((replace_paira[A, B, List[Char],
   Unit](i)).apply(dpp).apply(st).apply(sts))
  }

def check_weak[A : equal : show,
                B : equal : show](p: ((A, nat)) => nat,
                                   r: (term[A, B], term[A, B])):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]](check[(List[Char]) =>
                                List[Char]](equal_terma[A,
                 B](proj_term[A, B](p, fst[term[A, B], term[A, B]](r)),
                     proj_term[A, B](p, snd[term[A, B], term[A, B]](r))),
     comp[List[Char], List[Char],
           List[Char]](shows_string.apply(List('t', 'h', 'e', ' ', 'p', 'r',
        'o', 'j', 'e', 'c', 't', 'e', 'd', ' ', 'l', 'h', 's', ' ', 'i', 's',
        ' ', 'n', 'o', 't', ' ', 'e', 'q', 'u', 'a', 'l', ' ', 't', 'o', ' ',
        't', 'h', 'e', ' ', 'p', 'r', 'o', 'j', 'e', 'c', 't', 'e', 'd', ' ',
        'r', 'h', 's')),
                        shows_nl)),
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('C', 'o', 'u', 'l', 'd', ' ', 'n', 'o',
 't', ' ', 'o', 'r', 'i', 'e', 'n', 't', ' ', 'r', 'u', 'l', 'e', ' ')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_rule[A,
       B]((a: A) => (b: List[Char]) => shows_prec[A](Zero_nat, a, b),
           (a: B) => (b: List[Char]) => shows_prec[B](Zero_nat, a, b),
           List(' ', '-', '>', ' '), r),
                                    comp[List[Char], List[Char],
  List[Char]](shows_string.apply(List(',', ' ', 's', 'i', 'n', 'c', 'e')),
               comp[List[Char], List[Char],
                     List[Char]](shows_nl,
                                  comp[List[Char], List[Char],
List[Char]]((shows_prec_term[A, B](Zero_nat)).apply(proj_term[A,
                       B](p, fst[term[A, B], term[A, B]](r))),
             comp[List[Char], List[Char],
                   List[Char]](shows_string.apply(List(' ', '!', '=', ' ')),
                                comp[List[Char], List[Char],
                                      List[Char]]((shows_prec_term[A,
                            B](Zero_nat)).apply(proj_term[A,
                   B](p, snd[term[A, B], term[A, B]](r))),
           comp[List[Char], List[Char], List[Char]](shows_nl, x))))))))))

def q_emptyc[A, B, C, D](x0: dpp_ops_ext[A, B, C, D]): A => Boolean = x0 match {
  case Dpp_ops_ext(dpp, p, pw, pairs, q, r, rw, rules, q_empty,
                    rules_no_left_var, rules_non_collapsing, is_QNF,
                    nFQ_subset_NF_rules, rules_map, reverse_rules_map,
                    intersect_pairs, replace_pair, intersect_rules, delete_P_Pw,
                    delete_R_Rw, split_pairs, split_rules, mk, more)
    => q_empty
}

def check_strict_rstep[A : equal : show,
                        B : equal : show](ra: List[(term[A, B], term[A, B])],
   rseqm:
     ((term[A, B], term[A, B])) =>
       Option[List[(pos, ((term[A, B], term[A, B]), term[A, B]))]],
   p: ((A, nat)) => nat, r: (term[A, B], term[A, B])):
      sum[(List[Char]) => List[Char], Unit] =
  {
    val s: term[A, B] = proj_term[A, B](p, fst[term[A, B], term[A, B]](r))
    val t: term[A, B] = proj_term[A, B](p, snd[term[A, B], term[A, B]](r));
    rseqm(r) match {
      case None => check_supt[A, B](s, t)
      case Some(rseq) =>
        (if (equal_nata(size_list[(pos, ((term[A, B], term[A, B]),
  term[A, B]))](rseq),
                         Zero_nat))
          check_supt[A, B](s, t)
          else {
                 val u: term[A, B] =
                   snd[(term[A, B], term[A, B]),
                        term[A, B]](snd[pos,
 ((term[A, B], term[A, B]),
   term[A, B])](last[(pos, ((term[A, B], term[A, B]), term[A, B]))](rseq)));
                 bindb[(List[Char]) => List[Char], Unit,
                        Unit](check_qrsteps[A,
     B]((_: term[A, B]) => true, ra, rseq, s, u),
                               (_: Unit) => check_supteq[A, B](u, t))
               })
    }
  }

def check_strict_qrstep[A : equal : show,
                         B : equal : show](nf: (term[A, B]) => Boolean,
    ra: List[(term[A, B], term[A, B])],
    rseqm:
      ((term[A, B], term[A, B])) =>
        Option[List[(pos, ((term[A, B], term[A, B]), term[A, B]))]],
    p: ((A, nat)) => nat, r: (term[A, B], term[A, B])):
      sum[(List[Char]) => List[Char], Unit] =
  {
    val s: term[A, B] = proj_term[A, B](p, fst[term[A, B], term[A, B]](r))
    val t: term[A, B] = proj_term[A, B](p, snd[term[A, B], term[A, B]](r));
    rseqm(r) match {
      case None => check_supt[A, B](s, t)
      case Some(Nil) =>
        Inl[(List[Char]) => List[Char],
             Unit](shows_string.apply(List('m', 'o', 'r', 'e', ' ', 't', 'h',
    'a', 'n', ' ', 'a', ' ', 's', 'i', 'n', 'g', 'l', 'e', ' ', 'Q', '-', 'r',
    'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 't', 'e', 'p', ' ', 'i', 's', ' ',
    'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e', 'd')))
      case Some(List((pos, (rule, u)))) =>
        bindb[(List[Char]) => List[Char], Unit,
               Unit]((check_qrstep[A, B](nf)).apply(ra).apply(pos).apply(rule).apply(s).apply(u),
                      (_: Unit) => check_supteq[A, B](u, t))
      case Some((pos, (rule, u)) :: _ :: _) =>
        Inl[(List[Char]) => List[Char],
             Unit](shows_string.apply(List('m', 'o', 'r', 'e', ' ', 't', 'h',
    'a', 'n', ' ', 'a', ' ', 's', 'i', 'n', 'g', 'l', 'e', ' ', 'Q', '-', 'r',
    'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 't', 'e', 'p', ' ', 'i', 's', ' ',
    'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e', 'd')))
    }
  }

def subterm_criterion_proc[A, B : equal : show,
                            C : equal : show](i: dpp_ops_ext[A, B, C, Unit],
       p: ((B, nat)) => nat,
       rseqm:
         ((term[B, C], term[B, C])) =>
           Option[List[(pos, ((term[B, C], term[B, C]), term[B, C]))]],
       prm: List[(term[B, C], term[B, C])], dpp: A):
      sum[(List[Char]) => List[Char], A] =
  {
    val pa: List[(term[B, C], term[B, C])] =
      (pairsa[A, B, C, Unit](i)).apply(dpp)
    val r: List[(term[B, C], term[B, C])] =
      (rulesd[A, B, C, Unit](i)).apply(dpp)
    val pb: List[(term[B, C], term[B, C])] =
      snd[List[(term[B, C], term[B, C])],
           List[(term[B, C],
                  term[B, C])]]((split_pairsa[A, B, C,
       Unit](i)).apply(dpp).apply(prm));
    bindb[(List[Char]) => List[Char], Unit,
           Unit](catcha[((term[B, C], term[B, C]), (List[Char]) => List[Char]),
                         Unit,
                         (List[Char]) =>
                           List[Char]](forallm[(term[B, C], term[B, C]),
        (List[Char]) =>
          List[Char]]((a: (term[B, C], term[B, C])) =>
                        {
                          val (l, ra): (term[B, C], term[B, C]) = a;
                          bindb[(List[Char]) => List[Char], Unit,
                                 Unit](check_no_var[B, C](l),
(_: Unit) =>
  bindb[(List[Char]) => List[Char], Unit,
         Unit](check_no_var[B, C](ra),
                (_: Unit) =>
                  check_no_defined_root[B,
 C]((fn: (B, nat)) =>
      ! (nulla[(term[B, C],
                 term[B, C])]((rules_mapb[A, B, C,
   Unit](i)).apply(dpp).apply(fn))),
     ra)))
                        },
                       pa),
(x: ((term[B, C], term[B, C]), (List[Char]) => List[Char])) =>
  Inl[(List[Char]) => List[Char],
       Unit](snd[(term[B, C], term[B, C]), (List[Char]) => List[Char]](x))),
                  (_: Unit) =>
                    bindb[(List[Char]) => List[Char], Unit,
                           Unit]((if ((q_emptyc[A, B, C, Unit](i)).apply(dpp))
                                   catcha[((term[B, C], term[B, C]),
    (List[Char]) => List[Char]),
   Unit,
   (List[Char]) =>
     List[Char]](forallm[(term[B, C], term[B, C]),
                          (List[Char]) =>
                            List[Char]]((a: (term[B, C], term[B, C])) =>
  check_strict_rstep[B, C](r, rseqm, p, a),
 prm),
                  (x: ((term[B, C], term[B, C]), (List[Char]) => List[Char])) =>
                    Inl[(List[Char]) => List[Char],
                         Unit](snd[(term[B, C], term[B, C]),
                                    (List[Char]) => List[Char]](x)))
                                   else catcha[((term[B, C], term[B, C]),
         (List[Char]) => List[Char]),
        Unit,
        (List[Char]) =>
          List[Char]](forallm[(term[B, C], term[B, C]),
                               (List[Char]) =>
                                 List[Char]]((a: (term[B, C], term[B, C])) =>
       check_strict_qrstep[B, C]((is_QNFc[A, B, C, Unit](i)).apply(dpp), r,
                                  rseqm, p, a),
      prm),
                       (x: ((term[B, C], term[B, C]),
                             (List[Char]) => List[Char]))
                         => Inl[(List[Char]) => List[Char],
                                 Unit](snd[(term[B, C], term[B, C]),
    (List[Char]) => List[Char]](x)))),
                                  (_: Unit) =>
                                    catcha[((term[B, C], term[B, C]),
     (List[Char]) => List[Char]),
    Unit,
    (List[Char]) =>
      List[Char]](forallm[(term[B, C], term[B, C]),
                           (List[Char]) =>
                             List[Char]]((a: (term[B, C], term[B, C])) =>
   check_weak[B, C](p, a),
  pb),
                   (x: ((term[B, C], term[B, C]), (List[Char]) => List[Char]))
                     => Inl[(List[Char]) => List[Char],
                             Unit](snd[(term[B, C], term[B, C]),
(List[Char]) => List[Char]](x)))))
  }
    match {
    case Inl(a) => Inl[(List[Char]) => List[Char], A](a)
    case Inr(_) =>
      Inr[A, (List[Char]) =>
               List[Char]]((delete_P_Pwa[A, B, C,
  Unit](i)).apply(dpp).apply(prm).apply(prm))
  }

def min_set_ext[A](s_ns: A => A => (Boolean, Boolean)):
      (List[A]) => (List[A]) => (Boolean, Boolean) =
  (as: List[A]) =>
    (bs: List[A]) =>
      (! (nulla[A](bs)) &&
         list_all[A]((a: A) =>
                       list_ex[A]((b: A) => fst[Boolean, Boolean]((s_ns(a))(b)),
                                   bs),
                      as),
        list_all[A]((a: A) =>
                      list_ex[A]((b: A) => snd[Boolean, Boolean]((s_ns(a))(b)),
                                  bs),
                     as))

def or2(x0: (Boolean, Boolean), x1: (Boolean, Boolean)): (Boolean, Boolean) =
  (x0, x1) match {
  case ((b1, b2), (b3, b4)) => (b1 || b3, b2 || b4)
}

def lifta(x0: (Boolean, Boolean)): (Boolean, Boolean) = x0 match {
  case (uu, b) => (b, b)
}

def ex_dominating[A : equal](f: A => A => (Boolean, Boolean), y: A, x2: List[A],
                              uw: List[A], ux: List[A]):
      (Boolean, Boolean) =
  (f, y, x2, uw, ux) match {
  case (f, y, Nil, uw, ux) => (false, false)
  case (f, y, b :: bs, xs, ys) =>
    (f(b))(y) match {
      case (true, ns) =>
        or2(lifta(mul_ext_impl[A](f, remove1[A](b, xs),
                                   filter[A]((ya: A) =>
       ! (fst[Boolean, Boolean]((f(b))(ya))),
      ys))),
             ex_dominating[A](f, y, bs, xs, ys))
      case (false, true) =>
        or2(mul_ext_impl[A](f, remove1[A](b, xs), ys),
             ex_dominating[A](f, y, bs, xs, ys))
      case (false, false) => ex_dominating[A](f, y, bs, xs, ys)
    }
}

def mul_ext_impl[A : equal](f: A => A => (Boolean, Boolean), x1: List[A],
                             x2: List[A]):
      (Boolean, Boolean) =
  (f, x1, x2) match {
  case (f, Nil, Nil) => (false, true)
  case (f, Nil, v :: va) => (false, false)
  case (f, v :: va, Nil) => (true, true)
  case (f, v :: va, y :: ys) => ex_dominating[A](f, y, v :: va, v :: va, ys)
}

def list_ext[A : equal](x0: list_order_type):
      (A => A => (Boolean, Boolean)) =>
        (List[A]) => (List[A]) => (Boolean, Boolean) =
  x0 match {
  case MS_Ext =>
    (a: A => A => (Boolean, Boolean)) =>
      (b: List[A]) => (c: List[A]) => mul_ext_impl[A](a, b, c)
  case Max_Ext => (a: A => A => (Boolean, Boolean)) => set_ext[A](a)
  case Min_Ext => (a: A => A => (Boolean, Boolean)) => min_set_ext[A](a)
}

def get_root_redtriple[A : equal : show,
                        B : equal : show](x0: root_redtriple_impl[A]):
      root_redtriple_ext[A, B, Unit] =
  x0 match {
  case Scnp(typea, af, rp) =>
    generate_scnp_rp[A, B,
                      redtriple_impl[A]](list_ext[(term[A, B], nat)](typea), af,
  (a: redtriple_impl[A]) => get_redtriple[A, B](a), rp)
}

def combine[A, B : equal](x0: scg[A, B], s: scg[A, B]): scg[A, B] = (x0, s)
  match {
  case (Scg(pa, qa, stra, wka), Scg(p, q, str, wk)) =>
    Scg[A, B](pa, qa, remdups[(B, B)](stra ++ str), remdups[(B, B)](wka ++ wk))
  case (Null, s) => Null
  case (Scg(v, va, vb, vc), Null) => Null
}

def in_situ[A, B : equal](x0: scg[A, B]): Boolean = x0 match {
  case Null => true
  case Scg(p, q, str, wk) =>
    list_ex[(B, B)]((a: (B, B)) => {
                                     val (aa, b): (B, B) = a;
                                     eq[B](aa, b)
                                   },
                     str)
}

def compa[A : equal](esa: List[(A, A)], es: List[(A, A)]): List[(A, A)] =
  remdups[(A, A)](maps[(A, A),
                        (A, A)]((a: (A, A)) =>
                                  {
                                    val (x, y): (A, A) = a;
                                    maps[(A, A),
  (A, A)]((aa: (A, A)) =>
            {
              val (ya, z): (A, A) = aa;
              (if (eq[A](y, ya)) List((x, z)) else Nil)
            },
           es)
                                  },
                                 esa))

def subtract_list[A : equal](xs: List[A], ys: List[A]): List[A] =
  filter[A]((x: A) => ! (membera[A](ys, x)), xs)

def scg_comp[A, B : equal](conn: A => A => Boolean, x1: scg[A, B],
                            g: scg[A, B]):
      scg[A, B] =
  (conn, x1, g) match {
  case (conn, Scg(pa, qa, stra, wka), Scg(p, q, str, wk)) =>
    (if (! ((conn(qa))(p))) Null
      else {
             val strs: List[(B, B)] =
               remdups[(B, B)](compa[B](stra, str) ++
                                 (compa[B](stra, wk) ++ compa[B](wka, str)))
             val a: List[(B, B)] =
               subtract_list[(B, B)](remdups[(B, B)](compa[B](wka, wk)), strs);
             Scg[A, B](pa, q, strs, a)
           })
  case (conn, Null, g) => Null
  case (conn, Scg(v, va, vb, vc), Null) => Null
}

def subsumes[A : equal, B : equal](g: scg[A, B], x1: scg[A, B]): Boolean =
  (g, x1) match {
  case (Scg(pa, qa, stra, wka), Scg(p, q, str, wk)) =>
    eq[A](pa, p) &&
      (eq[A](qa, q) &&
        (list_all[(B, B)]((a: (B, B)) => membera[(B, B)](str, a), stra) &&
          list_all[(B, B)]((a: (B, B)) => membera[(B, B)](str ++ wk, a), wka)))
  case (g, Null) => true
  case (Null, Scg(v, va, vb, vc)) => false
}

def sagiv[A : equal, B : equal](conn: A => A => Boolean, g: scg[A, B]):
      Boolean =
  (if (in_situ[A, B](g)) true
    else {
           val gg: scg[A, B] = scg_comp[A, B](conn, g, g);
           (if (subsumes[A, B](gg, g)) false
             else sagiv[A, B](conn, combine[A, B](g, gg)))
         })

def equal_scg[A : equal, B : equal](x0: scg[A, B], x1: scg[A, B]): Boolean =
  (x0, x1) match {
  case (Scg(pp1, pp2, list1, list2), Null) => false
  case (Null, Scg(pp1, pp2, list1, list2)) => false
  case (Scg(pp1a, pp2a, list1a, list2a), Scg(pp1, pp2, list1, list2)) =>
    eq[A](pp1a, pp1) &&
      (eq[A](pp2a, pp2) &&
        (equal_lista[(B, B)](list1a, list1) &&
          equal_lista[(B, B)](list2a, list2)))
  case (Null, Null) => true
}

def insert_scg[A : equal, B : equal](g: scg[A, B], gs: List[scg[A, B]]):
      Option[List[scg[A, B]]] =
  (if (equal_scg[A, B](g, Null)) None
    else (if (list_ex[scg[A, B]]((ga: scg[A, B]) => subsumes[A, B](ga, g), gs))
           None
           else Some[List[scg[A, B]]](g ::
filter[scg[A, B]]((ga: scg[A, B]) => ! (subsumes[A, B](g, ga)), gs))))

def mk_trancl[A : equal,
               B : equal](conn: A => A => Boolean, x1: List[scg[A, B]],
                           s: List[scg[A, B]]):
      List[scg[A, B]] =
  (conn, x1, s) match {
  case (conn, Nil, s) => s
  case (conn, g :: gs, s) =>
    insert_scg[A, B](g, s) match {
      case None => mk_trancl[A, B](conn, gs, s)
      case Some(sa) =>
        mk_trancl[A, B](conn,
                         mapa[scg[A, B],
                               scg[A, B]]((a: scg[A, B]) =>
    scg_comp[A, B](conn, g, a),
   sa) ++
                           (mapa[scg[A, B],
                                  scg[A, B]]((f: scg[A, B]) =>
       scg_comp[A, B](conn, f, g),
      sa) ++
                             gs),
                         sa)
    }
}

def check_SCT[A : equal,
               B : equal](conn: A => A => Boolean, gs: List[scg[A, B]]):
      Boolean =
  {
    val a: List[scg[A, B]] = mk_trancl[A, B](conn, gs, Nil);
    list_all[scg[A, B]]((aa: scg[A, B]) => sagiv[A, B](conn, aa), a)
  }

def get_arg[A, B](t: term[A, B], x1: nat): term[A, B] = (t, x1) match {
  case (t, Zero_nat) => t
  case (t, Suc(n)) => nth[term[A, B]](args[A, B](t), n)
}

def check_sct_ur_af_entry[A : equal : show, B : equal : show, C, D, E,
                           F](is_def: ((A, nat)) => Boolean,
                               rm: ((A, nat)) => List[(term[A, B], term[A, B])],
                               ur: List[(term[A, B], term[A, B])],
                               pi: ((A, nat)) => nat => Boolean,
                               sa: ((term[A, B], term[A, B])) => sum[C, D],
                               nst: ((term[A, B], term[A, B])) => sum[E, F],
                               s: term[A, B], t: term[A, B],
                               stri: List[(nat, nat)], nstri: List[(nat, nat)]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]](bindb[(List[Char]) => List[Char], Unit,
                               Unit](check_no_var[A, B](s),
                                      (_: Unit) =>
bindb[(List[Char]) => List[Char], Unit,
       Unit](check_no_var[A, B](t),
              (_: Unit) =>
                bindb[(List[Char]) => List[Char], Unit,
                       Unit](check_no_defined_root[A, B](is_def, t),
                              (_: Unit) =>
                                {
                                  val m: nat = num_args[A, B](t)
                                  val n: nat = num_args[A, B](s);
                                  bindb[(List[Char]) => List[Char], Unit,
 Unit](catcha[(nat, (List[Char]) => List[Char]), Unit,
               (List[Char]) =>
                 List[Char]](forallm[nat, (List[Char]) =>
    List[Char]]((i: nat) =>
                  check[(List[Char]) =>
                          List[Char]](less_eq_nat(i, n),
                                       comp[List[Char], List[Char],
     List[Char]](shows_string.apply(List('l', 'e', 'f', 't', '-', 'i', 'n', 'd',
  'e', 'x', ' ', 't', 'o', ' ', 'l', 'a', 'r', 'g', 'e')),
                  comp[List[Char], List[Char],
                        List[Char]](shows_prec_nat(Zero_nat, i), shows_nl))),
                 remdups[nat](mapa[(nat, nat),
                                    nat]((a: (nat, nat)) => fst[nat, nat](a),
  stri ++ nstri))),
                              (x: (nat, (List[Char]) => List[Char])) =>
                                Inl[(List[Char]) => List[Char],
                                     Unit](snd[nat,
        (List[Char]) => List[Char]](x))),
        (_: Unit) =>
          bindb[(List[Char]) => List[Char], Unit,
                 Unit](catcha[(nat, (List[Char]) => List[Char]), Unit,
                               (List[Char]) =>
                                 List[Char]](forallm[nat,
              (List[Char]) =>
                List[Char]]((j: nat) =>
                              check[(List[Char]) =>
                                      List[Char]](less_eq_nat(j, m) &&
            isOK[(List[Char]) => List[Char],
                  Unit](check_ur_closed_term_rm_af[A,
            B](rm, ur, pi, get_arg[A, B](t, j))),
           comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List('r', 'i', 'g', 'h', 't',
              '-', 'i', 'n', 'd', 'e', 'x', ' ', 't', 'o', ' ', 'l', 'a', 'r',
              'g', 'e', ' ', 'o', 'r', ' ', 'a', 'r', 'g', 'u', 'm', 'e', 'n',
              't', ' ', 'v', 'i', 'o', 'l', 'a', 't', 'e', 's', ' ', 'u', 's',
              'a', 'b', 'l', 'e', '-', 'r', 'u', 'l', 'e', 's', ' ', 'c', 'o',
              'n', 'd', 'i', 't', 'i', 'o', 'n')),
                              shows_prec_nat(Zero_nat, j))),
                             remdups[nat](mapa[(nat, nat),
        nat]((a: (nat, nat)) => snd[nat, nat](a), stri ++ nstri))),
      (x: (nat, (List[Char]) => List[Char])) =>
        Inl[(List[Char]) => List[Char],
             Unit](snd[nat, (List[Char]) => List[Char]](x))),
                        (_: Unit) =>
                          {
                            args[A, B](s);
                            args[A, B](t);
                            bindb[(List[Char]) => List[Char], Unit,
                                   Unit](catcha[((nat, nat),
          (List[Char]) => List[Char]),
         Unit,
         (List[Char]) =>
           List[Char]](forallm[(nat, nat),
                                (List[Char]) =>
                                  List[Char]]((a: (nat, nat)) =>
        {
          val (i, j): (nat, nat) = a;
          check[(List[Char]) =>
                  List[Char]](isOK[C, D](sa((get_arg[A, B](s, i),
      get_arg[A, B](t, j)))),
                               comp[List[Char], List[Char],
                                     List[Char]](shows_string.apply(List('p',
                                  'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i',
                                  't', 'h', ' ', 'e', 'd', 'g', 'e', ' ')),
          comp[List[Char], List[Char],
                List[Char]](shows_prec_nat(Zero_nat, i),
                             comp[List[Char], List[Char],
                                   List[Char]](shows_string.apply(List(' ', '-',
                                'S', '-', '>', ' ')),
        shows_prec_nat(Zero_nat, j)))))
        },
       stri),
                        (x: ((nat, nat), (List[Char]) => List[Char])) =>
                          Inl[(List[Char]) => List[Char],
                               Unit](snd[(nat, nat),
  (List[Char]) => List[Char]](x))),
  (_: Unit) =>
    catcha[((nat, nat), (List[Char]) => List[Char]), Unit,
            (List[Char]) =>
              List[Char]](forallm[(nat, nat),
                                   (List[Char]) =>
                                     List[Char]]((a: (nat, nat)) =>
           {
             val (i, j): (nat, nat) = a;
             check[(List[Char]) =>
                     List[Char]](isOK[E, F](nst((get_arg[A, B](s, i),
          get_arg[A, B](t, j)))),
                                  comp[List[Char], List[Char],
List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                     'i', 't', 'h', ' ', 'e', 'd', 'g', 'e',
                                     ' ')),
             comp[List[Char], List[Char],
                   List[Char]](shows_prec_nat(Zero_nat, i),
                                comp[List[Char], List[Char],
                                      List[Char]](shows_string.apply(List(' ',
                                   '-', 'N', 'S', '-', '>', ' ')),
           shows_prec_nat(Zero_nat, j)))))
           },
          nstri),
                           (x: ((nat, nat), (List[Char]) => List[Char])) =>
                             Inl[(List[Char]) => List[Char],
                                  Unit](snd[(nat, nat),
     (List[Char]) => List[Char]](x))))
                          }))
                                }))),
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l', 'e', 'm', 's',
 ' ', 'w', 'i', 't', 'h', ' ', 'D', 'P', ' ')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_rule[A,
       B]((a: A) => (b: List[Char]) => shows_prec[A](Zero_nat, a, b),
           (a: B) => (b: List[Char]) => shows_prec[B](Zero_nat, a, b),
           List(' ', '-', '>', ' '), (s, t)),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl, x)))))

def sct_ur_af_proc[A, B : equal : show,
                    C : equal : show](i: dpp_ops_ext[A, B, C, Unit],
                                       rp: redtriple_ext[B, C, Unit],
                                       gs:
 List[((term[B, C], term[B, C]), (List[(nat, nat)], List[(nat, nat)]))],
                                       ur: List[(term[B, C], term[B, C])],
                                       dpp: A):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]]({
                          val p: List[(term[B, C], term[B, C])] =
                            (pairsa[A, B, C, Unit](i)).apply(dpp)
                          val rm: ((B, nat)) => List[(term[B, C], term[B, C])] =
                            (rules_mapb[A, B, C, Unit](i)).apply(dpp)
                          val is_def: ((B, nat)) => Boolean =
                            (fn: (B, nat)) =>
                              ! (nulla[(term[B, C],
 term[B, C])]((rules_mapb[A, B, C, Unit](i)).apply(dpp).apply(fn)));
                          (rulesd[A, B, C, Unit](i)).apply(dpp)
                          val ggs: List[((term[B, C], term[B, C]),
  (List[(nat, nat)], List[(nat, nat)]))]
                            = filter[((term[B, C], term[B, C]),
                                       (List[(nat, nat)],
 List[(nat, nat)]))]((g: ((term[B, C], term[B, C]),
                           (List[(nat, nat)], List[(nat, nat)])))
                       => membera[(term[B, C],
                                    term[B,
  C])](p, fst[(term[B, C], term[B, C]),
               (List[(nat, nat)], List[(nat, nat)])](g)),
                      gs)
                          val pi: ((B, nat)) => nat => Boolean =
                            af[B, C, Unit](rp)
                          val sa: ((term[B, C], term[B, C])) =>
                                    sum[(List[Char]) => List[Char], Unit]
                            = s[B, C, Unit](rp)
                          val nsa: ((term[B, C], term[B, C])) =>
                                     sum[(List[Char]) => List[Char], Unit]
                            = ns[B, C, Unit](rp)
                          val nsta: ((term[B, C], term[B, C])) =>
                                      sum[(List[Char]) => List[Char], Unit]
                            = nst[B, C, Unit](rp)
                          val cap1: (term[B, C]) => gctxt[B, C] =
                            capRM_dpp[A, B, C](i, dpp)
                          val cap2: (term[B, C]) => gctxt[B, C] =
                            reverse_capRM_dpp[A, B, C](i, dpp);
                          bindb[(List[Char]) => List[Char], Unit,
                                 Unit](valid[B, C, Unit](rp),
(_: Unit) =>
  bindb[(List[Char]) => List[Char], Unit,
         Unit](catcha[(List[Char]) => List[Char], Unit,
                       (List[Char]) =>
                         List[Char]](catcha[((term[B, C], term[B, C]),
      (List[Char]) => List[Char]),
     Unit,
     (List[Char]) =>
       List[Char]](forallm[(term[B, C], term[B, C]),
                            (List[Char]) =>
                              List[Char]]((lr: (term[B, C], term[B, C])) =>
    check_ur_closed_term_rm_af[B, C](rm, ur, pi,
                                      snd[term[B, C], term[B, C]](lr)),
   ur),
                    (x: ((term[B, C], term[B, C]), (List[Char]) => List[Char]))
                      => Inl[(List[Char]) => List[Char],
                              Unit](snd[(term[B, C], term[B, C]),
 (List[Char]) => List[Char]](x))),
                                      (x: (List[Char]) => List[Char]) =>
Inl[(List[Char]) => List[Char],
     Unit](comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List('e', 'r', 'r', 'o', 'r',
              ' ', 'w', 'h', 'e', 'n', ' ', 'c', 'h', 'e', 'c', 'k', 'i', 'n',
              'g', ' ', 'c', 'l', 'o', 's', 'u', 'r', 'e', ' ', 'p', 'r', 'o',
              'p', 'e', 'r', 't', 'i', 'e', 's', ' ', 'o', 'f', ' ', 'r', 'h',
              's', ' ', 'o', 'f', ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ', 'r',
              'u', 'l', 'e', 's')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x)))),
                (_: Unit) =>
                  bindb[(List[Char]) => List[Char], Unit,
                         Unit](catcha[(term[B, C], term[B, C]), Unit,
                                       (List[Char]) =>
 List[Char]](check_subseteq[(term[B, C],
                              term[B, C])](p,
    mapa[((term[B, C], term[B, C]), (List[(nat, nat)], List[(nat, nat)])),
          (term[B, C],
            term[B, C])]((a: ((term[B, C], term[B, C]),
                               (List[(nat, nat)], List[(nat, nat)])))
                           => fst[(term[B, C], term[B, C]),
                                   (List[(nat, nat)], List[(nat, nat)])](a),
                          ggs)),
              (x: (term[B, C], term[B, C])) =>
                Inl[(List[Char]) => List[Char],
                     Unit](comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('t', 'h',
                              'e', 'r', 'e', ' ', 'i', 's', ' ', 'n', 'o', ' ',
                              's', 'i', 'z', 'e', '-', 'c', 'h', 'a', 'n', 'g',
                              'e', ' ', 'g', 'r', 'a', 'p', 'h', ' ', 'f', 'o',
                              'r', ' ', 'D', 'P', ' ')),
      shows_rule[B, C]((a: B) =>
                         (b: List[Char]) => shows_prec[B](Zero_nat, a, b),
                        (a: C) =>
                          (b: List[Char]) => shows_prec[C](Zero_nat, a, b),
                        List(' ', '-', '>', ' '), x)))),
                                (_: Unit) =>
                                  bindb[(List[Char]) => List[Char], Unit,
 Unit](catcha[(((term[B, C], term[B, C]), (List[(nat, nat)], List[(nat, nat)])),
                (List[Char]) => List[Char]),
               Unit,
               (List[Char]) =>
                 List[Char]](forallm[((term[B, C], term[B, C]),
                                       (List[(nat, nat)], List[(nat, nat)])),
                                      (List[Char]) =>
List[Char]]((a: ((term[B, C], term[B, C]),
                  (List[(nat, nat)], List[(nat, nat)])))
              => {
                   val (aa, b):
                         ((term[B, C], term[B, C]),
                           (List[(nat, nat)], List[(nat, nat)]))
                     = a;
                   ({
                      val (saa, t): (term[B, C], term[B, C]) = aa;
                      (ab: (List[(nat, nat)], List[(nat, nat)])) =>
                        {
                          val (ac, ba): (List[(nat, nat)], List[(nat, nat)]) =
                            ab;
                          check_sct_ur_af_entry[B, C,
         (List[Char]) => List[Char], Unit, (List[Char]) => List[Char],
         Unit](is_def, rm, ur, pi, sa, nsta, saa, t, ac, ba)
                        }
                    })(b)
                 },
             ggs),
                              (x: (((term[B, C], term[B, C]),
                                     (List[(nat, nat)], List[(nat, nat)])),
                                    (List[Char]) => List[Char]))
                                => Inl[(List[Char]) => List[Char],
Unit](snd[((term[B, C], term[B, C]), (List[(nat, nat)], List[(nat, nat)])),
           (List[Char]) => List[Char]](x))),
        (_: Unit) =>
          bindb[(List[Char]) => List[Char], Unit,
                 Unit](catcha[(List[Char]) => List[Char], Unit,
                               (List[Char]) =>
                                 List[Char]](catcha[((term[B, C], term[B, C]),
              (List[Char]) => List[Char]),
             Unit,
             (List[Char]) =>
               List[Char]](forallm[(term[B, C], term[B, C]),
                                    (List[Char]) => List[Char]](nsa, ur),
                            (x: ((term[B, C], term[B, C]),
                                  (List[Char]) => List[Char]))
                              => Inl[(List[Char]) => List[Char],
                                      Unit](snd[(term[B, C], term[B, C]),
         (List[Char]) => List[Char]](x))),
      (x: (List[Char]) => List[Char]) =>
        Inl[(List[Char]) => List[Char],
             Unit](comp[List[Char], List[Char],
                         List[Char]](shows_string.apply(List('p', 'r', 'o', 'b',
                      'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r',
                      'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'u', 's', 'a',
                      'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e', 's')),
                                      comp[List[Char], List[Char],
    List[Char]](shows_nl, x)))),
                        (_: Unit) =>
                          check[(List[Char]) =>
                                  List[Char]](check_SCT[(term[B, C],
                  (term[B, C], (gctxt[B, C], gctxt[B, C]))),
                 nat]((a: (term[B, C],
                            (term[B, C], (gctxt[B, C], gctxt[B, C]))))
                        => {
                             val (_, (t, (_, ct))):
                                   (term[B, C],
                                     (term[B, C], (gctxt[B, C], gctxt[B, C])))
                               = a;
                             (aa: (term[B, C],
                                    (term[B, C], (gctxt[B, C], gctxt[B, C]))))
                               => {
                                    val (u, (_, (cu, _))):
  (term[B, C], (term[B, C], (gctxt[B, C], gctxt[B, C])))
                                      = aa;
                                    matchc[B, C](ct, u) && matchc[B, C](cu, t)
                                  }
                           },
                       mapa[((term[B, C], term[B, C]),
                              (List[(nat, nat)], List[(nat, nat)])),
                             scg[(term[B, C],
                                   (term[B, C], (gctxt[B, C], gctxt[B, C]))),
                                  nat]]((a:
   ((term[B, C], term[B, C]), (List[(nat, nat)], List[(nat, nat)])))
  => {
       val (aa, b):
             ((term[B, C], term[B, C]), (List[(nat, nat)], List[(nat, nat)]))
         = a;
       ({
          val (sb, t): (term[B, C], term[B, C]) = aa;
          (ab: (List[(nat, nat)], List[(nat, nat)])) =>
            {
              val (stri, nstri): (List[(nat, nat)], List[(nat, nat)]) = ab
              val e: (term[B, C], (term[B, C], (gctxt[B, C], gctxt[B, C]))) =
                (sb, (t, (cap2(sb), cap1(t))));
              Scg[(term[B, C], (term[B, C], (gctxt[B, C], gctxt[B, C]))),
                   nat](e, e, stri, nstri)
            }
        })(b)
     },
 ggs)),
       comp[List[Char], List[Char],
             List[Char]](shows_string.apply(List('s', 'i', 'z', 'e', '-', 'c',
          'h', 'a', 'n', 'g', 'e', ' ', 'a', 'n', 'a', 'l', 'y', 's', 'i', 's',
          ' ', 'f', 'a', 'i', 'l', 'e', 'd')),
                          shows_nl)))))))
                        },
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ', 'n', 'o',
 't', ' ', 'a', 'p', 'p', 'l', 'y', ' ', 't', 'h', 'e', ' ', 's', 'i', 'z', 'e',
 '-', 'c', 'h', 'a', 'n', 'g', 'e', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o',
 'r', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l', 'l', 'o',
 'w', 'i', 'n', 'g', ' ')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_nl,
                                    comp[List[Char], List[Char],
  List[Char]](desca[B, C, Unit](rp),
               comp[List[Char], List[Char],
                     List[Char]](shows_nl,
                                  comp[List[Char], List[Char],
List[Char]](shows_string.apply(List('f', 'o', 'r', ' ', 't', 'h', 'e', ' ', 'f',
                                     'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g',
                                     ' ', 'r', 'e', 'a', 's', 'o', 'n')),
             comp[List[Char], List[Char], List[Char]](shows_nl, x))))))))

def sct_subterm_proc[A, B : equal : show,
                      C : equal : show](i: dpp_ops_ext[A, B, C, Unit],
 gs: List[((term[B, C], term[B, C]), (List[(nat, nat)], List[(nat, nat)]))],
 dpp: A):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]]({
                          val p: List[(term[B, C], term[B, C])] =
                            (pairsa[A, B, C, Unit](i)).apply(dpp)
                          val is_def: ((B, nat)) => Boolean =
                            (fn: (B, nat)) =>
                              ! (nulla[(term[B, C],
 term[B, C])]((rules_mapb[A, B, C, Unit](i)).apply(dpp).apply(fn)))
                          val cap1: (term[B, C]) => gctxt[B, C] =
                            capRM_dpp[A, B, C](i, dpp)
                          val cap2: (term[B, C]) => gctxt[B, C] =
                            reverse_capRM_dpp[A, B, C](i, dpp);
                          bindb[(List[Char]) => List[Char], Unit,
                                 Unit](catcha[(term[B, C], term[B, C]), Unit,
       (List[Char]) =>
         List[Char]](check_subseteq[(term[B, C],
                                      term[B,
    C])](p, mapa[((term[B, C], term[B, C]),
                   (List[(nat, nat)], List[(nat, nat)])),
                  (term[B, C],
                    term[B, C])]((a: ((term[B, C], term[B, C]),
                                       (List[(nat, nat)], List[(nat, nat)])))
                                   => fst[(term[B, C], term[B, C]),
   (List[(nat, nat)], List[(nat, nat)])](a),
                                  gs)),
                      (x: (term[B, C], term[B, C])) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](shows_string.apply(List('t', 'h', 'e', 'r', 'e', ' ', 'i', 's',
                                      ' ', 'n', 'o', ' ', 's', 'i', 'z', 'e',
                                      '-', 'c', 'h', 'a', 'n', 'g', 'e', ' ',
                                      'g', 'r', 'a', 'p', 'h', ' ', 'f', 'o',
                                      'r', ' ', 't', 'h', 'e', ' ', 'p', 'a',
                                      'i', 'r', ' ')),
              shows_rule[B, C]((a: B) =>
                                 (b: List[Char]) =>
                                   shows_prec[B](Zero_nat, a, b),
                                (a: C) =>
                                  (b: List[Char]) =>
                                    shows_prec[C](Zero_nat, a, b),
                                List(' ', '-', '>', ' '), x)))),
(_: Unit) =>
  bindb[(List[Char]) => List[Char], Unit,
         Unit](catcha[(((term[B, C], term[B, C]),
                         (List[(nat, nat)], List[(nat, nat)])),
                        (List[Char]) => List[Char]),
                       Unit,
                       (List[Char]) =>
                         List[Char]](forallm[((term[B, C], term[B, C]),
       (List[(nat, nat)], List[(nat, nat)])),
      (List[Char]) =>
        List[Char]]((a: ((term[B, C], term[B, C]),
                          (List[(nat, nat)], List[(nat, nat)])))
                      => {
                           val (aa, b):
                                 ((term[B, C], term[B, C]),
                                   (List[(nat, nat)], List[(nat, nat)]))
                             = a;
                           ({
                              val (s, t): (term[B, C], term[B, C]) = aa;
                              (ab: (List[(nat, nat)], List[(nat, nat)])) =>
                                {
                                  val (stri, nstri):
(List[(nat, nat)], List[(nat, nat)])
                                    = ab;
                                  catcha[(List[Char]) => List[Char], Unit,
  (List[Char]) =>
    List[Char]](bindb[(List[Char]) => List[Char], Unit,
                       Unit](check_no_var[B, C](s),
                              (_: Unit) =>
                                bindb[(List[Char]) => List[Char], Unit,
                                       Unit](check_no_var[B, C](t),
      (_: Unit) =>
        bindb[(List[Char]) => List[Char], Unit,
               Unit](check_no_defined_root[B, C](is_def, t),
                      (_: Unit) =>
                        {
                          val m: nat = num_args[B, C](t)
                          val n: nat = num_args[B, C](s);
                          bindb[(List[Char]) => List[Char], Unit,
                                 Unit](catcha[((nat, nat),
        (List[Char]) => List[Char]),
       Unit,
       (List[Char]) =>
         List[Char]](forallm[(nat, nat),
                              (List[Char]) =>
                                List[Char]]((ac: (nat, nat)) =>
      {
        val (ia, j): (nat, nat) = ac;
        check[(List[Char]) =>
                List[Char]](less_eq_nat(ia, n) &&
                              (less_eq_nat(j, m) &&
                                isOK[(List[Char]) => List[Char],
                                      Unit](check_supt[B,
                C](get_arg[B, C](s, ia), get_arg[B, C](t, j)))),
                             comp[List[Char], List[Char],
                                   List[Char]](shows_string.apply(List('p', 'r',
                                'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i', 't',
                                'h', ' ', 'e', 'd', 'g', 'e', ' ')),
        comp[List[Char], List[Char],
              List[Char]](shows_prec_nat(Zero_nat, ia),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List(' ', '|',
                              '>', ' ')),
      comp[List[Char], List[Char],
            List[Char]](shows_prec_nat(Zero_nat, j), shows_nl)))))
      },
     stri),
                      (x: ((nat, nat), (List[Char]) => List[Char])) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](snd[(nat, nat),
(List[Char]) => List[Char]](x))),
(_: Unit) =>
  catcha[((nat, nat), (List[Char]) => List[Char]), Unit,
          (List[Char]) =>
            List[Char]](forallm[(nat, nat),
                                 (List[Char]) =>
                                   List[Char]]((ac: (nat, nat)) =>
         {
           val (ia, j): (nat, nat) = ac;
           check[(List[Char]) =>
                   List[Char]](less_eq_nat(ia, n) &&
                                 (less_eq_nat(j, m) &&
                                   isOK[(List[Char]) => List[Char],
 Unit](check_supteq[B, C](get_arg[B, C](s, ia), get_arg[B, C](t, j)))),
                                comp[List[Char], List[Char],
                                      List[Char]](shows_string.apply(List('p',
                                   'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i',
                                   't', 'h', ' ', 'e', 'd', 'g', 'e', ' ')),
           comp[List[Char], List[Char],
                 List[Char]](shows_prec_nat(Zero_nat, ia),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_string.apply(List(' ',
                                 '|', '>', '=', ' ')),
         comp[List[Char], List[Char],
               List[Char]](shows_prec_nat(Zero_nat, j), shows_nl)))))
         },
        nstri),
                         (x: ((nat, nat), (List[Char]) => List[Char])) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](snd[(nat, nat),
   (List[Char]) => List[Char]](x))))
                        }))),
                 (x: (List[Char]) => List[Char]) =>
                   Inl[(List[Char]) => List[Char],
                        Unit](comp[List[Char], List[Char],
                                    List[Char]](shows_string.apply(List('p',
                                 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i',
                                 't', 'h', ' ', 'p', 'a', 'i', 'r', ' ')),
         comp[List[Char], List[Char],
               List[Char]](shows_rule[B, C]((ac: B) =>
      (ba: List[Char]) => shows_prec[B](Zero_nat, ac, ba),
     (ac: C) => (ba: List[Char]) => shows_prec[C](Zero_nat, ac, ba),
     List(' ', '-', '>', ' '), (s, t)),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_nl, x)))))
                                }
                            })(b)
                         },
                     gs),
                                      (x:
 (((term[B, C], term[B, C]), (List[(nat, nat)], List[(nat, nat)])),
   (List[Char]) => List[Char]))
=> Inl[(List[Char]) => List[Char],
        Unit](snd[((term[B, C], term[B, C]),
                    (List[(nat, nat)], List[(nat, nat)])),
                   (List[Char]) => List[Char]](x))),
                (_: Unit) =>
                  check[(List[Char]) =>
                          List[Char]](check_SCT[(term[B, C],
          (term[B, C], (gctxt[B, C], gctxt[B, C]))),
         nat]((a: (term[B, C], (term[B, C], (gctxt[B, C], gctxt[B, C])))) =>
                {
                  val (_, (t, (_, ct))):
                        (term[B, C], (term[B, C], (gctxt[B, C], gctxt[B, C])))
                    = a;
                  (aa: (term[B, C], (term[B, C], (gctxt[B, C], gctxt[B, C]))))
                    => {
                         val (u, (_, (cu, _))):
                               (term[B, C],
                                 (term[B, C], (gctxt[B, C], gctxt[B, C])))
                           = aa;
                         matchc[B, C](ct, u) && matchc[B, C](cu, t)
                       }
                },
               mapa[((term[B, C], term[B, C]),
                      (List[(nat, nat)], List[(nat, nat)])),
                     scg[(term[B, C], (term[B, C], (gctxt[B, C], gctxt[B, C]))),
                          nat]]((a: ((term[B, C], term[B, C]),
                                      (List[(nat, nat)], List[(nat, nat)])))
                                  => {
                                       val
 (aa, b): ((term[B, C], term[B, C]), (List[(nat, nat)], List[(nat, nat)])) = a;
                                       ({
  val (s, t): (term[B, C], term[B, C]) = aa;
  (ab: (List[(nat, nat)], List[(nat, nat)])) =>
    {
      val (stri, nstri): (List[(nat, nat)], List[(nat, nat)]) = ab
      val e: (term[B, C], (term[B, C], (gctxt[B, C], gctxt[B, C]))) =
        (s, (t, (cap2(s), cap1(t))));
      Scg[(term[B, C], (term[B, C], (gctxt[B, C], gctxt[B, C]))),
           nat](e, e, stri, nstri)
    }
})(b)
                                     },
                                 gs)),
                                       comp[List[Char], List[Char],
     List[Char]](shows_string.apply(List('s', 'i', 'z', 'e', '-', 'c', 'h', 'a',
  'n', 'g', 'e', ' ', 'a', 'n', 'a', 'l', 'y', 's', 'i', 's', ' ', 'f', 'a',
  'i', 'l', 'e', 'd')),
                  shows_nl))))
                        },
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ', 'n', 'o',
 't', ' ', 'a', 'p', 'p', 'l', 'y', ' ', 't', 'h', 'e', ' ', 's', 'i', 'z', 'e',
 '-', 'c', 'h', 'a', 'n', 'g', 'e', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o',
 'r', ' ', 'b', 'a', 's', 'e', 'd', ' ', 'o', 'n', ' ', 't', 'h', 'e', ' ', 's',
 'u', 'b', 't', 'e', 'r', 'm', '-', 'r', 'e', 'l', 'a', 't', 'i', 'o', 'n')),
                 comp[List[Char], List[Char], List[Char]](shows_nl, x))))

def check_dp_proof[A, B : equal : linorder : show](assms: Boolean,
            ia: (List[Char]) => List[Char],
            i: dpp_ops_ext[A, lab[B, List[nat]], List[Char], Unit], dpp: A,
            x4: dp_proof[B, List[nat], List[Char]]):
      sum[(List[Char]) => List[Char], Unit] =
  (assms, ia, i, dpp, x4) match {
  case (assms, ia, i, dpp, PisEmpty) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('P', ' ', 'i', 's', ' ', 'e', 'm', 'p', 't', 'y'),
                       (if (nulla[(term[lab[B, List[nat]], List[Char]],
                                    term[lab[B, List[nat]],
  List[Char]])]((pa[A, lab[B, List[nat]], List[Char], Unit](i)).apply(dpp)) &&
                              (nulla[(term[lab[B, List[nat]], List[Char]],
                                       term[lab[B, List[nat]],
     List[Char]])]((pwa[A, lab[B, List[nat]], List[Char],
                         Unit](i)).apply(dpp)) ||
                                nulla[(term[lab[B, List[nat]], List[Char]],
term[lab[B, List[nat]],
      List[Char]])]((rc[A, lab[B, List[nat]], List[Char],
                         Unit](i)).apply(dpp))))
                         Inr[Unit, (List[Char]) => List[Char]](())
                         else Inl[(List[Char]) => List[Char],
                                   Unit](comp[List[Char], List[Char],
       List[Char]](ia, comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List(':', ' ', 'P',
                          ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'e', 'm', 'p',
                          't', 'y', ' ', 'i', 'n', ' ', 't', 'h', 'e', ' ', 'f',
                          'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g', ' ', 'D', 'P',
                          '-', 'p', 'r', 'o', 'b', 'l', 'e', 'm', ' ')),
  comp[List[Char], List[Char],
        List[Char]](shows_nl,
                     shows_dpp[lab[B, List[nat]], List[Char],
                                A]((a: lab[B, List[nat]]) =>
                                     shows_prec_lab[B, List[nat]](Zero_nat, a),
                                    (a: List[Char]) =>
                                      shows_prec_list[Char](Zero_nat, a),
                                    i, dpp)))))))
  case (assms, ia, i, dpp, SubtermCriterionProc(p, rseq, rP, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('S', 'u', 'b', 't', 'e', 'r', 'm', 'C', 'r', 'i',
                             't', 'e', 'r', 'i', 'o', 'n', 'P', 'r', 'o', 'c'),
                       {
                         val paa: List[(term[lab[B, List[nat]], List[Char]],
 term[lab[B, List[nat]], List[Char]])]
                           = (pairsa[A, lab[B, List[nat]], List[Char],
                                      Unit](i)).apply(dpp)
                         val pr: List[(term[lab[B, List[nat]], List[Char]],
term[lab[B, List[nat]], List[Char]])]
                           = list_diff[(term[lab[B, List[nat]], List[Char]],
 term[lab[B, List[nat]], List[Char]])](paa, rP);
                         bindb[(List[Char]) => List[Char], A,
                                Unit](catcha[(List[Char]) => List[Char], A,
      (List[Char]) =>
        List[Char]](subterm_criterion_proc[A, lab[B, List[nat]],
    List[Char]](i, (a: (lab[B, List[nat]], nat)) =>
                     create_proj[lab[B, List[nat]]](p, a),
                 (a: (term[lab[B, List[nat]], List[Char]],
                       term[lab[B, List[nat]], List[Char]]))
                   => create_rseq_map[lab[B, List[nat]], List[Char]](rseq, a),
                 pr, dpp),
                     (x: (List[Char]) => List[Char]) =>
                       Inl[(List[Char]) => List[Char],
                            A](comp[List[Char], List[Char],
                                     List[Char]](ia,
          comp[List[Char], List[Char],
                List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
             'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
             'i', 'n', 'g', ' ', 't', 'h', 'e', ' ', 's', 'u', 'b', 't', 'e',
             'r', 'm', ' ', 'c', 'r', 'i', 't', 'e', 'r', 'i', 'o', 'n', ' ',
             't', 'o', ' ', 't', 'h', 'e', ' ', 'D', 'P', ' ', 'p', 'r', 'o',
             'b', 'l', 'e', 'm')),
                             comp[List[Char], List[Char],
                                   List[Char]](shows_nl,
        comp[List[Char], List[Char],
              List[Char]](shows_dpp[lab[B, List[nat]], List[Char],
                                     A]((a: lab[B, List[nat]]) =>
  shows_prec_lab[B, List[nat]](Zero_nat, a),
 (a: List[Char]) => shows_prec_list[Char](Zero_nat, a), i, dpp),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_nl,
      comp[List[Char], List[Char],
            List[Char]](shows_string.apply(List('a', 'n', 'd', ' ', 't', 'r',
         'y', 'i', 'n', 'g', ' ', 't', 'o', ' ', 'r', 'e', 'm', 'o', 'v', 'e',
         ' ', 't', 'h', 'e', ' ', 'p', 'a', 'i', 'r', 's', ' ')),
                         comp[List[Char], List[Char],
                               List[Char]](shows_nl,
    comp[List[Char], List[Char],
          List[Char]](shows_rules[lab[B, List[nat]],
                                   List[Char]]((a: lab[B, List[nat]]) =>
         shows_prec_lab[B, List[nat]](Zero_nat, a),
        (a: List[Char]) => shows_prec_list[Char](Zero_nat, a),
        List(' ', '-', '>', ' '), pr),
                       comp[List[Char], List[Char],
                             List[Char]](shows_nl, x))))))))))),
                                       (dppa: A) =>
 catcha[(List[Char]) => List[Char], Unit,
         (List[Char]) =>
           List[Char]](check_dp_proof[A, B](assms,
     comp[List[Char], List[Char],
           List[Char]](ia, shows_string.apply(List('.', '1'))),
     i, dppa, prf),
                        (x: (List[Char]) => List[Char]) =>
                          Inl[(List[Char]) => List[Char],
                               Unit](comp[List[Char], List[Char],
   List[Char]](ia, comp[List[Char], List[Char],
                         List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                      'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't',
                      'h', 'e', ' ', 's', 'u', 'b', 't', 'e', 'r', 'm', ' ',
                      'c', 'r', 'i', 't', 'e', 'r', 'i', 'o', 'n')),
                                      comp[List[Char], List[Char],
    List[Char]](shows_nl, (a: List[Char]) => indent(x, a)))))))
                       })
  case (assms, ia, i, dpp, RedPairProc(redp, rP, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('R', 'e', 'd', 'P', 'a', 'i', 'r', 'P', 'r', 'o',
                             'c'),
                       {
                         val p: List[(term[lab[B, List[nat]], List[Char]],
                                       term[lab[B, List[nat]], List[Char]])]
                           = (pairsa[A, lab[B, List[nat]], List[Char],
                                      Unit](i)).apply(dpp)
                         val pr: List[(term[lab[B, List[nat]], List[Char]],
term[lab[B, List[nat]], List[Char]])]
                           = list_diff[(term[lab[B, List[nat]], List[Char]],
 term[lab[B, List[nat]], List[Char]])](p, rP)
                         val proc: (List[(term[lab[B, List[nat]], List[Char]],
   term[lab[B, List[nat]], List[Char]])]) =>
                                     A => sum[(List[Char]) => List[Char], A]
                           = redp match {
                               case Inl(rrp) =>
                                 (a: List[(term[lab[B, List[nat]], List[Char]],
    term[lab[B, List[nat]], List[Char]])])
                                   => (b: A) =>
root_redtriple_proc[A, lab[B, List[nat]],
                     List[Char]](i, get_root_redtriple[lab[B, List[nat]],
                List[Char]](rrp),
                                  a, b)
                               case Inr(rp) =>
                                 (a: List[(term[lab[B, List[nat]], List[Char]],
    term[lab[B, List[nat]], List[Char]])])
                                   => (b: A) =>
redtriple_proc[A, lab[B, List[nat]],
                List[Char]](i, get_redtriple[lab[B, List[nat]], List[Char]](rp),
                             a, b)
                             };
                         bindb[(List[Char]) => List[Char], A,
                                Unit](catcha[(List[Char]) => List[Char], A,
      (List[Char]) =>
        List[Char]]((proc(pr))(dpp),
                     (x: (List[Char]) => List[Char]) =>
                       Inl[(List[Char]) => List[Char],
                            A](comp[List[Char], List[Char],
                                     List[Char]](ia,
          comp[List[Char], List[Char],
                List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
             'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
             'i', 'n', 'g', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'd', 'u', 'c',
             't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o',
             'c', 'e', 's', 's', 'o', 'r', ' ', 't', 'o', ' ', 'r', 'e', 'm',
             'o', 'v', 'e', ' ', 'f', 'r', 'o', 'm', ' ', 't', 'h', 'e', ' ',
             'D', 'P', ' ', 'p', 'r', 'o', 'b', 'l', 'e', 'm')),
                             comp[List[Char], List[Char],
                                   List[Char]](shows_nl,
        comp[List[Char], List[Char],
              List[Char]](shows_dpp[lab[B, List[nat]], List[Char],
                                     A]((a: lab[B, List[nat]]) =>
  shows_prec_lab[B, List[nat]](Zero_nat, a),
 (a: List[Char]) => shows_prec_list[Char](Zero_nat, a), i, dpp),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_nl,
      comp[List[Char], List[Char],
            List[Char]](shows_string.apply(List(' ', 't', 'h', 'e', ' ', 'p',
         'a', 'i', 'r', 's', ' ')),
                         comp[List[Char], List[Char],
                               List[Char]](shows_nl,
    comp[List[Char], List[Char],
          List[Char]](shows_rules[lab[B, List[nat]],
                                   List[Char]]((a: lab[B, List[nat]]) =>
         shows_prec_lab[B, List[nat]](Zero_nat, a),
        (a: List[Char]) => shows_prec_list[Char](Zero_nat, a),
        List(' ', '-', '>', ' '), pr),
                       comp[List[Char], List[Char],
                             List[Char]](shows_nl, x))))))))))),
                                       (dppa: A) =>
 catcha[(List[Char]) => List[Char], Unit,
         (List[Char]) =>
           List[Char]](check_dp_proof[A, B](assms,
     comp[List[Char], List[Char],
           List[Char]](ia, shows_string.apply(List('.', '1'))),
     i, dppa, prf),
                        (x: (List[Char]) => List[Char]) =>
                          Inl[(List[Char]) => List[Char],
                               Unit](comp[List[Char], List[Char],
   List[Char]](ia, comp[List[Char], List[Char],
                         List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                      'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't',
                      'h', 'e', ' ', 'r', 'e', 'd', 'u', 'c', 't', 'i', 'o',
                      'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c',
                      'e', 's', 's', 'o', 'r')),
                                      comp[List[Char], List[Char],
    List[Char]](shows_nl, (a: List[Char]) => indent(x, a)))))))
                       })
  case (assms, ia, i, dpp, UsableRulesProc(u, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('U', 's', 'a', 'b', 'l', 'e', 'R', 'u', 'l', 'e',
                             's', 'P', 'r', 'o', 'c'),
                       bindb[(List[Char]) => List[Char], A,
                              Unit](catcha[(List[Char]) => List[Char], A,
    (List[Char]) =>
      List[Char]](usable_rules_proc[A, lab[B, List[nat]]](i, u, dpp),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          A](comp[List[Char], List[Char],
                                   List[Char]](ia,
        comp[List[Char], List[Char],
              List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o',
           'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n',
           'g', ' ', 't', 'h', 'e', ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ', 'r',
           'u', 'l', 'e', 's', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r',
           ' ', 't', 'o', ' ', 'r', 'e', 's', 't', 'r', 'i', 'c', 't', ' ', 't',
           'h', 'e', ' ', 'D', 'P', ' ', 'p', 'r', 'o', 'b', 'l', 'e', 'm')),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_nl,
      comp[List[Char], List[Char],
            List[Char]](shows_dpp[lab[B, List[nat]], List[Char],
                                   A]((a: lab[B, List[nat]]) =>
shows_prec_lab[B, List[nat]](Zero_nat, a),
                                       (a: List[Char]) =>
 shows_prec_list[Char](Zero_nat, a),
                                       i, dpp),
                         comp[List[Char], List[Char],
                               List[Char]](shows_nl,
    comp[List[Char], List[Char],
          List[Char]](shows_string.apply(List(' ', 't', 'o', ' ', 't', 'h', 'e',
       ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e', 's', ' ')),
                       comp[List[Char], List[Char],
                             List[Char]](shows_nl,
  comp[List[Char], List[Char],
        List[Char]](shows_rules[lab[B, List[nat]],
                                 List[Char]]((a: lab[B, List[nat]]) =>
       shows_prec_lab[B, List[nat]](Zero_nat, a),
      (a: List[Char]) => shows_prec_list[Char](Zero_nat, a),
      List(' ', '-', '>', ' '), u),
                     comp[List[Char], List[Char],
                           List[Char]](shows_nl, x))))))))))),
                                     (dppa: A) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](check_dp_proof[A, B](assms,
   comp[List[Char], List[Char],
         List[Char]](ia, shows_string.apply(List('.', '1'))),
   i, dppa, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](ia, comp[List[Char], List[Char],
                       List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                    'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                    'e', ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ', 'r', 'u', 'l',
                    'e', 's', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o',
                    'r')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (assms, ia, i, dpp, MonoRedPairProc(redp, rP, rR, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('M', 'o', 'n', 'o', 'R', 'e', 'd', 'P', 'a', 'i',
                             'r', 'P', 'r', 'o', 'c'),
                       {
                         val p: List[(term[lab[B, List[nat]], List[Char]],
                                       term[lab[B, List[nat]], List[Char]])]
                           = (pairsa[A, lab[B, List[nat]], List[Char],
                                      Unit](i)).apply(dpp)
                         val pr: List[(term[lab[B, List[nat]], List[Char]],
term[lab[B, List[nat]], List[Char]])]
                           = list_diff[(term[lab[B, List[nat]], List[Char]],
 term[lab[B, List[nat]], List[Char]])](p, rP)
                         val r: List[(term[lab[B, List[nat]], List[Char]],
                                       term[lab[B, List[nat]], List[Char]])]
                           = (rulesd[A, lab[B, List[nat]], List[Char],
                                      Unit](i)).apply(dpp)
                         val rr: List[(term[lab[B, List[nat]], List[Char]],
term[lab[B, List[nat]], List[Char]])]
                           = list_diff[(term[lab[B, List[nat]], List[Char]],
 term[lab[B, List[nat]], List[Char]])](r, rR);
                         bindb[(List[Char]) => List[Char], A,
                                Unit](catcha[(List[Char]) => List[Char], A,
      (List[Char]) =>
        List[Char]](mono_redpair_proc[A, lab[B, List[nat]],
                                       List[Char]](i,
            get_redtriple[lab[B, List[nat]], List[Char]](redp), pr, rr, dpp),
                     (x: (List[Char]) => List[Char]) =>
                       Inl[(List[Char]) => List[Char],
                            A](comp[List[Char], List[Char],
                                     List[Char]](ia,
          comp[List[Char], List[Char],
                List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
             'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
             'i', 'n', 'g', ' ', 't', 'h', 'e', ' ', 'm', 'o', 'n', 'o', ' ',
             'r', 'e', 'd', 'u', 'c', 't', 'i', 'o', 'n', ' ', 'p', 'a', 'i',
             'r', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r', ' ', 't',
             'o', ' ', 'r', 'e', 'm', 'o', 'v', 'e', ' ', 'f', 'r', 'o', 'm',
             ' ', 't', 'h', 'e', ' ', 'D', 'P', ' ', 'p', 'r', 'o', 'b', 'l',
             'e', 'm')),
                             comp[List[Char], List[Char],
                                   List[Char]](shows_nl,
        comp[List[Char], List[Char],
              List[Char]](shows_dpp[lab[B, List[nat]], List[Char],
                                     A]((a: lab[B, List[nat]]) =>
  shows_prec_lab[B, List[nat]](Zero_nat, a),
 (a: List[Char]) => shows_prec_list[Char](Zero_nat, a), i, dpp),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_nl,
      comp[List[Char], List[Char],
            List[Char]](shows_string.apply(List(' ', 't', 'h', 'e', ' ', 'p',
         'a', 'i', 'r', 's', ' ')),
                         comp[List[Char], List[Char],
                               List[Char]](shows_nl,
    comp[List[Char], List[Char],
          List[Char]](shows_rules[lab[B, List[nat]],
                                   List[Char]]((a: lab[B, List[nat]]) =>
         shows_prec_lab[B, List[nat]](Zero_nat, a),
        (a: List[Char]) => shows_prec_list[Char](Zero_nat, a),
        List(' ', '-', '>', ' '), pr),
                       comp[List[Char], List[Char],
                             List[Char]](shows_nl,
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List(' ', 'a', 'n', 'd', ' ', 't', 'h',
     'e', ' ', 'r', 'u', 'l', 'e', 's', ' ')),
                     comp[List[Char], List[Char],
                           List[Char]](shows_nl,
comp[List[Char], List[Char],
      List[Char]](shows_rules[lab[B, List[nat]],
                               List[Char]]((a: lab[B, List[nat]]) =>
     shows_prec_lab[B, List[nat]](Zero_nat, a),
    (a: List[Char]) => shows_prec_list[Char](Zero_nat, a),
    List(' ', '-', '>', ' '), rr),
                   comp[List[Char], List[Char],
                         List[Char]](shows_nl, x))))))))))))))),
                                       (dppa: A) =>
 catcha[(List[Char]) => List[Char], Unit,
         (List[Char]) =>
           List[Char]](check_dp_proof[A, B](assms,
     comp[List[Char], List[Char],
           List[Char]](ia, shows_string.apply(List('.', '1'))),
     i, dppa, prf),
                        (x: (List[Char]) => List[Char]) =>
                          Inl[(List[Char]) => List[Char],
                               Unit](comp[List[Char], List[Char],
   List[Char]](ia, comp[List[Char], List[Char],
                         List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                      'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't',
                      'h', 'e', ' ', 'm', 'o', 'n', 'o', ' ', 'r', 'e', 'd',
                      'u', 'c', 't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r',
                      ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r')),
                                      comp[List[Char], List[Char],
    List[Char]](shows_nl, (a: List[Char]) => indent(x, a)))))))
                       })
  case (assms, ia, i, dpp, DepGraphProc(edpts)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('D', 'e', 'p', 'G', 'r', 'a', 'p', 'h', 'P', 'r',
                             'o', 'c'),
                       bindb[(List[Char]) => List[Char],
                              List[(dp_proof[B, List[nat], List[Char]], A)],
                              Unit](catcha[(List[Char]) => List[Char],
    List[(dp_proof[B, List[nat], List[Char]], A)],
    (List[Char]) =>
      List[Char]](dep_graph_proc[A, lab[B, List[nat]],
                                  dp_proof[B, List[nat],
    List[Char]]](i, dpp, edpts),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          List[(dp_proof[B, List[nat], List[Char]],
                                 A)]](comp[List[Char], List[Char],
    List[Char]](ia, comp[List[Char], List[Char],
                          List[Char]](shows_string.apply(List(':', ' ', 'e',
                       'r', 'r', 'o', 'r', ' ', 'w', 'h', 'i', 'l', 'e', ' ',
                       't', 'r', 'y', 'i', 'n', 'g', ' ', 't', 'o', ' ', 'p',
                       'e', 'r', 'f', 'o', 'r', 'm', ' ', 'S', 'C', 'C', '-',
                       'd', 'e', 'c', 'o', 'm', 'p', 'o', 's', 'i', 't', 'i',
                       'o', 'n', ' ', ' ', 'o', 'n', ' ')),
                                       comp[List[Char], List[Char],
     List[Char]](shows_nl,
                  comp[List[Char], List[Char],
                        List[Char]](shows_dpp[lab[B, List[nat]], List[Char],
       A]((a: lab[B, List[nat]]) => shows_prec_lab[B, List[nat]](Zero_nat, a),
           (a: List[Char]) => shows_prec_list[Char](Zero_nat, a), i, dpp),
                                     comp[List[Char], List[Char],
   List[Char]](shows_nl, x))))))),
                                     (pdpps:
List[(dp_proof[B, List[nat], List[Char]], A)])
                                       => catcha[(List[Char]) => List[Char],
          Unit,
          (List[Char]) =>
            List[Char]](catcha[(((dp_proof[B, List[nat], List[Char]], A), nat),
                                 (List[Char]) => List[Char]),
                                Unit,
                                (List[Char]) =>
                                  List[Char]](forallm_index[(dp_proof[B,
                               List[nat], List[Char]],
                      A),
                     (List[Char]) =>
                       List[Char]]((a: (dp_proof[B, List[nat], List[Char]], A))
                                     => {
  val (prof, dppa): (dp_proof[B, List[nat], List[Char]], A) = a;
  (j: nat) =>
    check_dp_proof[A, B](assms,
                          comp[List[Char], List[Char],
                                List[Char]](ia,
     comp[List[Char], List[Char],
           List[Char]](shows_string.apply(List('.')),
                        shows_prec_nat(Zero_nat, Suc(j)))),
                          i, dppa, prof)
},
                                    pdpps),
       (x: (((dp_proof[B, List[nat], List[Char]], A), nat),
             (List[Char]) => List[Char]))
         => Inl[(List[Char]) => List[Char],
                 Unit](snd[((dp_proof[B, List[nat], List[Char]], A), nat),
                            (List[Char]) => List[Char]](x))),
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](ia, comp[List[Char], List[Char],
                          List[Char]](shows_string.apply(List(':', ' ', 'e',
                       'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ',
                       't', 'h', 'e', ' ', 'd', 'e', 'p', 'e', 'n', 'd', 'e',
                       'n', 'c', 'y', ' ', 'g', 'r', 'a', 'p', 'h', ' ', 'p',
                       'r', 'o', 'c', 'e', 's', 's', 'o', 'r')),
                                       comp[List[Char], List[Char],
     List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (assms, ia, i, dpp, RedPairUrProc(redp, rP, ur, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('R', 'e', 'd', 'P', 'a', 'i', 'r', 'U', 'r', 'P',
                             'r', 'o', 'c'),
                       {
                         val p: List[(term[lab[B, List[nat]], List[Char]],
                                       term[lab[B, List[nat]], List[Char]])]
                           = (pairsa[A, lab[B, List[nat]], List[Char],
                                      Unit](i)).apply(dpp)
                         val pr: List[(term[lab[B, List[nat]], List[Char]],
term[lab[B, List[nat]], List[Char]])]
                           = list_diff[(term[lab[B, List[nat]], List[Char]],
 term[lab[B, List[nat]], List[Char]])](p, rP)
                         val proc: (List[(term[lab[B, List[nat]], List[Char]],
   term[lab[B, List[nat]], List[Char]])]) =>
                                     (List[(term[lab[B, List[nat]], List[Char]],
     term[lab[B, List[nat]], List[Char]])]) =>
                                       A => sum[(List[Char]) => List[Char], A]
                           = redp match {
                               case Inl(rrp) =>
                                 (a: List[(term[lab[B, List[nat]], List[Char]],
    term[lab[B, List[nat]], List[Char]])])
                                   => (b:
 List[(term[lab[B, List[nat]], List[Char]],
        term[lab[B, List[nat]], List[Char]])])
=> (c: A) =>
     ur_af_root_redtriple_proc[A, lab[B, List[nat]],
                                List[Char]](i,
     get_root_redtriple[lab[B, List[nat]], List[Char]](rrp), a, b, c)
                               case Inr(rp) =>
                                 (a: List[(term[lab[B, List[nat]], List[Char]],
    term[lab[B, List[nat]], List[Char]])])
                                   => (b:
 List[(term[lab[B, List[nat]], List[Char]],
        term[lab[B, List[nat]], List[Char]])])
=> (c: A) =>
     ur_af_redtriple_proc[A, lab[B, List[nat]],
                           List[Char]](i,
get_redtriple[lab[B, List[nat]], List[Char]](rp), a, b, c)
                             };
                         bindb[(List[Char]) => List[Char], A,
                                Unit](catcha[(List[Char]) => List[Char], A,
      (List[Char]) =>
        List[Char]](((proc(pr))(ur))(dpp),
                     (x: (List[Char]) => List[Char]) =>
                       Inl[(List[Char]) => List[Char],
                            A](comp[List[Char], List[Char],
                                     List[Char]](ia,
          comp[List[Char], List[Char],
                List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
             'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
             'i', 'n', 'g', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'd', 'u', 'c',
             't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o',
             'c', 'e', 's', 's', 'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 'u',
             's', 'a', 'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e', 's', ' ', 't',
             'o', ' ', 'r', 'e', 'm', 'o', 'v', 'e', ' ', 'f', 'r', 'o', 'm',
             ' ', 't', 'h', 'e', ' ', 'D', 'P', ' ', 'p', 'r', 'o', 'b', 'l',
             'e', 'm')),
                             comp[List[Char], List[Char],
                                   List[Char]](shows_nl,
        comp[List[Char], List[Char],
              List[Char]](shows_dpp[lab[B, List[nat]], List[Char],
                                     A]((a: lab[B, List[nat]]) =>
  shows_prec_lab[B, List[nat]](Zero_nat, a),
 (a: List[Char]) => shows_prec_list[Char](Zero_nat, a), i, dpp),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_nl,
      comp[List[Char], List[Char],
            List[Char]](shows_string.apply(List(' ', 't', 'h', 'e', ' ', 'p',
         'a', 'i', 'r', 's', ' ')),
                         comp[List[Char], List[Char],
                               List[Char]](shows_nl,
    comp[List[Char], List[Char],
          List[Char]](shows_rules[lab[B, List[nat]],
                                   List[Char]]((a: lab[B, List[nat]]) =>
         shows_prec_lab[B, List[nat]](Zero_nat, a),
        (a: List[Char]) => shows_prec_list[Char](Zero_nat, a),
        List(' ', '-', '>', ' '), pr),
                       comp[List[Char], List[Char],
                             List[Char]](shows_nl, x))))))))))),
                                       (dppa: A) =>
 catcha[(List[Char]) => List[Char], Unit,
         (List[Char]) =>
           List[Char]](check_dp_proof[A, B](assms,
     comp[List[Char], List[Char],
           List[Char]](ia, shows_string.apply(List('.', '1'))),
     i, dppa, prf),
                        (x: (List[Char]) => List[Char]) =>
                          Inl[(List[Char]) => List[Char],
                               Unit](comp[List[Char], List[Char],
   List[Char]](ia, comp[List[Char], List[Char],
                         List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                      'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't',
                      'h', 'e', ' ', 'r', 'e', 'd', 'u', 'c', 't', 'i', 'o',
                      'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c',
                      'e', 's', 's', 'o', 'r')),
                                      comp[List[Char], List[Char],
    List[Char]](shows_nl, (a: List[Char]) => indent(x, a)))))))
                       })
  case (assms, ia, i, dpp, MonoRedPairUrProc(redp, rP, rR, ur, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('M', 'o', 'n', 'o', 'R', 'e', 'd', 'P', 'a', 'i',
                             'r', 'U', 'r', 'P', 'r', 'o', 'c'),
                       {
                         val p: List[(term[lab[B, List[nat]], List[Char]],
                                       term[lab[B, List[nat]], List[Char]])]
                           = (pairsa[A, lab[B, List[nat]], List[Char],
                                      Unit](i)).apply(dpp)
                         val pr: List[(term[lab[B, List[nat]], List[Char]],
term[lab[B, List[nat]], List[Char]])]
                           = list_diff[(term[lab[B, List[nat]], List[Char]],
 term[lab[B, List[nat]], List[Char]])](p, rP)
                         val r: List[(term[lab[B, List[nat]], List[Char]],
                                       term[lab[B, List[nat]], List[Char]])]
                           = (rulesd[A, lab[B, List[nat]], List[Char],
                                      Unit](i)).apply(dpp)
                         val rr: List[(term[lab[B, List[nat]], List[Char]],
term[lab[B, List[nat]], List[Char]])]
                           = list_diff[(term[lab[B, List[nat]], List[Char]],
 term[lab[B, List[nat]], List[Char]])](r, rR);
                         bindb[(List[Char]) => List[Char], A,
                                Unit](catcha[(List[Char]) => List[Char], A,
      (List[Char]) =>
        List[Char]](mono_ur_redpair_proc[A, lab[B, List[nat]],
  List[Char]](i, get_redtriple[lab[B, List[nat]], List[Char]](redp), pr, rr, ur,
               dpp),
                     (x: (List[Char]) => List[Char]) =>
                       Inl[(List[Char]) => List[Char],
                            A](comp[List[Char], List[Char],
                                     List[Char]](ia,
          comp[List[Char], List[Char],
                List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
             'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
             'i', 'n', 'g', ' ', 't', 'h', 'e', ' ', 'm', 'o', 'n', 'o', ' ',
             'r', 'e', 'd', 'u', 'c', 't', 'i', 'o', 'n', ' ', 'p', 'a', 'i',
             'r', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r', ' ', 'w',
             'i', 't', 'h', ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ', 'r', 'u',
             'l', 'e', 's', ' ', 't', 'o', ' ', 'r', 'e', 'm', 'o', 'v', 'e',
             ' ', 'f', 'r', 'o', 'm', ' ', 't', 'h', 'e', ' ', 'D', 'P', ' ',
             'p', 'r', 'o', 'b', 'l', 'e', 'm')),
                             comp[List[Char], List[Char],
                                   List[Char]](shows_nl,
        comp[List[Char], List[Char],
              List[Char]](shows_dpp[lab[B, List[nat]], List[Char],
                                     A]((a: lab[B, List[nat]]) =>
  shows_prec_lab[B, List[nat]](Zero_nat, a),
 (a: List[Char]) => shows_prec_list[Char](Zero_nat, a), i, dpp),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_nl,
      comp[List[Char], List[Char],
            List[Char]](shows_string.apply(List(' ', 't', 'h', 'e', ' ', 'p',
         'a', 'i', 'r', 's', ' ')),
                         comp[List[Char], List[Char],
                               List[Char]](shows_nl,
    comp[List[Char], List[Char],
          List[Char]](shows_rules[lab[B, List[nat]],
                                   List[Char]]((a: lab[B, List[nat]]) =>
         shows_prec_lab[B, List[nat]](Zero_nat, a),
        (a: List[Char]) => shows_prec_list[Char](Zero_nat, a),
        List(' ', '-', '>', ' '), pr),
                       comp[List[Char], List[Char],
                             List[Char]](shows_nl,
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List(' ', 'a', 'n', 'd', ' ', 't', 'h',
     'e', ' ', 'r', 'u', 'l', 'e', 's', ' ')),
                     comp[List[Char], List[Char],
                           List[Char]](shows_nl,
comp[List[Char], List[Char],
      List[Char]](shows_rules[lab[B, List[nat]],
                               List[Char]]((a: lab[B, List[nat]]) =>
     shows_prec_lab[B, List[nat]](Zero_nat, a),
    (a: List[Char]) => shows_prec_list[Char](Zero_nat, a),
    List(' ', '-', '>', ' '), rr),
                   comp[List[Char], List[Char],
                         List[Char]](shows_nl, x))))))))))))))),
                                       (dppa: A) =>
 catcha[(List[Char]) => List[Char], Unit,
         (List[Char]) =>
           List[Char]](check_dp_proof[A, B](assms,
     comp[List[Char], List[Char],
           List[Char]](ia, shows_string.apply(List('.', '1'))),
     i, dppa, prf),
                        (x: (List[Char]) => List[Char]) =>
                          Inl[(List[Char]) => List[Char],
                               Unit](comp[List[Char], List[Char],
   List[Char]](ia, comp[List[Char], List[Char],
                         List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                      'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't',
                      'h', 'e', ' ', 'm', 'o', 'n', 'o', ' ', 'r', 'e', 'd',
                      'u', 'c', 't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r',
                      ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r', ' ',
                      'w', 'i', 't', 'h', ' ', 'u', 's', 'a', 'b', 'l', 'e',
                      ' ', 'r', 'u', 'l', 'e', 's')),
                                      comp[List[Char], List[Char],
    List[Char]](shows_nl, (a: List[Char]) => indent(x, a)))))))
                       })
  case (assms, ia, i, dpp, UncurryProc(mode, u_info, p, r, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('U', 'n', 'c', 'u', 'r', 'r', 'y', 'P', 'r', 'o',
                             'c'),
                       bindb[(List[Char]) => List[Char], A,
                              Unit](catcha[(List[Char]) => List[Char], A,
    (List[Char]) =>
      List[Char]]((uncurry_proc_both[A, B,
                                      List[nat]](i, mode,
          u_info)).apply(p).apply(r).apply(dpp),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          A](comp[List[Char], List[Char],
                                   List[Char]](ia,
        comp[List[Char], List[Char],
              List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o',
           'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n',
           'g', ' ', 't', 'h', 'e', ' ', 'u', 'n', 'c', 'u', 'r', 'r', 'y', 'i',
           'n', 'g', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r', ' ', 'o',
           'n', ' ', 't', 'h', 'e', ' ', 'D', 'P', ' ', 'p', 'r', 'o', 'b', 'l',
           'e', 'm')),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_nl,
      comp[List[Char], List[Char],
            List[Char]](shows_dpp[lab[B, List[nat]], List[Char],
                                   A]((a: lab[B, List[nat]]) =>
shows_prec_lab[B, List[nat]](Zero_nat, a),
                                       (a: List[Char]) =>
 shows_prec_list[Char](Zero_nat, a),
                                       i, dpp),
                         comp[List[Char], List[Char],
                               List[Char]](shows_nl, x))))))),
                                     (dppa: A) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](check_dp_proof[A, B](assms,
   comp[List[Char], List[Char],
         List[Char]](ia, shows_string.apply(List('.', '1'))),
   i, dppa, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](ia, comp[List[Char], List[Char],
                       List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                    'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                    'e', ' ', 'u', 'n', 'c', 'u', 'r', 'r', 'y', 'i', 'n', 'g',
                    ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (assms, ia, i, dpp, SizeChangeSubterm(graphs)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('S', 'i', 'z', 'e', 'C', 'h', 'a', 'n', 'g', 'e',
                             'S', 'u', 'b', 't', 'e', 'r', 'm'),
                       sct_subterm_proc[A, lab[B, List[nat]],
 List[Char]](i, graphs, dpp))
  case (assms, ia, i, dpp, SizeChangeRedPair(redp, ur, graphs)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('S', 'i', 'z', 'e', 'C', 'h', 'a', 'n', 'g', 'e',
                             'R', 'e', 'd', 'P', 'a', 'i', 'r'),
                       sct_ur_af_proc[A, lab[B, List[nat]],
                                       List[Char]](i,
            get_redtriple[lab[B, List[nat]], List[Char]](redp), graphs, ur,
            dpp))
  case (assms, ia, i, dpp, FccProc(f, fcs, p, r, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil), List('F', 'c', 'c', 'P', 'r', 'o', 'c'),
                       bindb[(List[Char]) => List[Char], A,
                              Unit](catcha[(List[Char]) => List[Char], A,
    (List[Char]) =>
      List[Char]](fcc_proc[A, lab[B, List[nat]],
                            List[Char]](i, f, fcs, p, Nil, r, dpp),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          A](comp[List[Char], List[Char],
                                   List[Char]](ia,
        comp[List[Char], List[Char],
              List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o',
           'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n',
           'g', ' ', 't', 'h', 'e', ' ', 'f', 'l', 'a', 't', ' ', 'c', 'o', 'n',
           't', 'e', 'x', 't', ' ', 'c', 'l', 'o', 's', 'u', 'r', 'e', ' ', 'p',
           'r', 'o', 'c', 'e', 's', 's', 'o', 'r', ' ', 'o', 'n', ' ', 't', 'h',
           'e', ' ', 'D', 'P', ' ', 'p', 'r', 'o', 'b', 'l', 'e', 'm')),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_nl,
      comp[List[Char], List[Char],
            List[Char]](shows_dpp[lab[B, List[nat]], List[Char],
                                   A]((a: lab[B, List[nat]]) =>
shows_prec_lab[B, List[nat]](Zero_nat, a),
                                       (a: List[Char]) =>
 shows_prec_list[Char](Zero_nat, a),
                                       i, dpp),
                         comp[List[Char], List[Char],
                               List[Char]](shows_nl, x))))))),
                                     (dppa: A) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](check_dp_proof[A, B](assms,
   comp[List[Char], List[Char],
         List[Char]](ia, shows_string.apply(List('.', '1'))),
   i, dppa, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](ia, comp[List[Char], List[Char],
                       List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                    'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                    'e', ' ', 'f', 'l', 'a', 't', ' ', 'c', 'o', 'n', 't', 'e',
                    'x', 't', ' ', 'c', 'l', 'o', 's', 'u', 'r', 'e', ' ', 'p',
                    'r', 'o', 'c', 'e', 's', 's', 'o', 'r')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (assms, ia, i, dpp, SplitProc(prem, rrem, prf1, prf2)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('S', 'p', 'l', 'i', 't', 'P', 'r', 'o', 'c'),
                       {
                         val (dpp1, dpp2): (A, A) =
                           split_proc[A, lab[B, List[nat]],
                                       List[Char]](i, dpp, prem, rrem);
                         bindb[(List[Char]) => List[Char], Unit,
                                Unit](catcha[(List[Char]) => List[Char], Unit,
      (List[Char]) =>
        List[Char]](check_dp_proof[A, B](assms,
  comp[List[Char], List[Char],
        List[Char]](ia, shows_string.apply(List('.', '1'))),
  i, dpp1, prf1),
                     (x: (List[Char]) => List[Char]) =>
                       Inl[(List[Char]) => List[Char],
                            Unit](comp[List[Char], List[Char],
List[Char]](ia, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                   'e', ' ', 's', 'p', 'l', 'i', 't', ' ', 'p', 'r', 'o', 'c',
                   'e', 's', 's', 'o', 'r')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (a: List[Char]) => indent(x, a)))))),
                                       (_: Unit) =>
 catcha[(List[Char]) => List[Char], Unit,
         (List[Char]) =>
           List[Char]](check_dp_proof[A, B](assms,
     comp[List[Char], List[Char],
           List[Char]](ia, shows_string.apply(List('.', '2'))),
     i, dpp2, prf2),
                        (x: (List[Char]) => List[Char]) =>
                          Inl[(List[Char]) => List[Char],
                               Unit](comp[List[Char], List[Char],
   List[Char]](ia, comp[List[Char], List[Char],
                         List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                      'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't',
                      'h', 'e', ' ', 's', 'p', 'l', 'i', 't', ' ', 'p', 'r',
                      'o', 'c', 'e', 's', 's', 'o', 'r')),
                                      comp[List[Char], List[Char],
    List[Char]](shows_nl, (a: List[Char]) => indent(x, a)))))))
                       })
  case (assms, ia, i, dpp, SemlabProc(sli, lP, lQ, lR, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('S', 'e', 'm', 'l', 'a', 'b', 'P', 'r', 'o', 'c'),
                       bindb[(List[Char]) => List[Char], A,
                              Unit](catcha[(List[Char]) => List[Char], A,
    (List[Char]) =>
      List[Char]]((semlab_fin_proc[A, B,
                                    List[Char]](i,
         sli)).apply(lP).apply(lQ).apply(lR).apply(dpp),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          A](comp[List[Char], List[Char],
                                   List[Char]](ia,
        comp[List[Char], List[Char],
              List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o',
           'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n',
           'g', ' ', 't', 'h', 'e', ' ', 's', 'e', 'm', 'l', 'a', 'b', ' ', 'p',
           'r', 'o', 'c', 'e', 's', 's', 'o', 'r', ' ', 'o', 'n', ' ', 't', 'h',
           'e', ' ', 'D', 'P', ' ', 'p', 'r', 'o', 'b', 'l', 'e', 'm')),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_nl,
      comp[List[Char], List[Char],
            List[Char]](shows_dpp[lab[B, List[nat]], List[Char],
                                   A]((a: lab[B, List[nat]]) =>
shows_prec_lab[B, List[nat]](Zero_nat, a),
                                       (a: List[Char]) =>
 shows_prec_list[Char](Zero_nat, a),
                                       i, dpp),
                         comp[List[Char], List[Char],
                               List[Char]](shows_nl, x))))))),
                                     (dppa: A) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](check_dp_proof[A, B](assms,
   comp[List[Char], List[Char],
         List[Char]](ia, shows_string.apply(List('.', '1'))),
   i, dppa, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](ia, comp[List[Char], List[Char],
                       List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                    'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                    'e', ' ', 's', 'e', 'm', 'l', 'a', 'b', ' ', 'p', 'r', 'o',
                    'c', 'e', 's', 's', 'o', 'r')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (assms, ia, i, dpp, SwitchInnermostProc(joins, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('S', 'w', 'i', 't', 'c', 'h', 'I', 'n', 'n', 'e',
                             'r', 'm', 'o', 's', 't', 'P', 'r', 'o', 'c'),
                       bindb[(List[Char]) => List[Char], A,
                              Unit](catcha[(List[Char]) => List[Char], A,
    (List[Char]) =>
      List[Char]](switch_innermost_proc[A, lab[B, List[nat]]](i, joins, dpp),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          A](comp[List[Char], List[Char],
                                   List[Char]](ia,
        comp[List[Char], List[Char],
              List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o',
           'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n',
           'g', ' ', 't', 'h', 'e', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o',
           'r', ' ', 't', 'o', ' ', 's', 'w', 'i', 't', 'c', 'h', ' ', 't', 'o',
           ' ', 'i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', ' ', 'o', 'n', ' ',
           't', 'h', 'e', ' ', 'D', 'P', ' ', 'p', 'r', 'o', 'b', 'l', 'e',
           'm')),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_nl,
      comp[List[Char], List[Char],
            List[Char]](shows_dpp[lab[B, List[nat]], List[Char],
                                   A]((a: lab[B, List[nat]]) =>
shows_prec_lab[B, List[nat]](Zero_nat, a),
                                       (a: List[Char]) =>
 shows_prec_list[Char](Zero_nat, a),
                                       i, dpp),
                         comp[List[Char], List[Char],
                               List[Char]](shows_nl, x))))))),
                                     (dppa: A) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](check_dp_proof[A, B](assms,
   comp[List[Char], List[Char],
         List[Char]](ia, shows_string.apply(List('.', '1'))),
   i, dppa, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](ia, comp[List[Char], List[Char],
                       List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                    'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                    'e', ' ', 's', 'w', 'i', 't', 'c', 'h', ' ', 't', 'o', ' ',
                    'i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', ' ', 'p', 'r',
                    'o', 'c', 'e', 's', 's', 'o', 'r')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (assms, ia, i, dppa, AssumeFinite(dpp)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('F', 'i', 'n', 'i', 't', 'e', 'n', 'e', 's', 's',
                             ' ', 'A', 's', 's', 'u', 'm', 'p', 't', 'i', 'o',
                             'n'),
                       (if (assms)
                         catcha[(List[Char]) => List[Char], Unit,
                                 (List[Char]) =>
                                   List[Char]](check_dpp_subsumes[A, B,
                           List[nat], List[Char]](i, dpp, dppa),
        (x: (List[Char]) => List[Char]) =>
          Inl[(List[Char]) => List[Char],
               Unit](comp[List[Char], List[Char],
                           List[Char]](ia,
comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o', 'r',
   ' ', 'i', 'n', ' ', 'f', 'i', 'n', 'i', 't', 'e', 'n', 'e', 's', 's', ' ',
   'a', 's', 's', 'u', 'm', 'p', 't', 'i', 'o', 'n')),
                   comp[List[Char], List[Char],
                         List[Char]](shows_nl,
                                      comp[List[Char], List[Char],
    List[Char]](x, shows_nl))))))
                         else Inl[(List[Char]) => List[Char],
                                   Unit](comp[List[Char], List[Char],
       List[Char]](ia, comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List(':', ' ', 't',
                          'h', 'e', ' ', 'p', 'r', 'o', 'o', 'f', ' ', 'c', 'o',
                          'n', 't', 'a', 'i', 'n', 's', ' ', 'a', ' ', 'l', 'o',
                          'c', 'a', 'l', ' ', 'f', 'i', 'n', 'i', 't', 'e', 'n',
                          'e', 's', 's', ' ', 'a', 's', 's', 'u', 'm', 'p', 't',
                          'i', 'o', 'n')),
  shows_nl)))))
  case (assms, ia, i, dpp, RewritingProc(u, sta, st, lr, p, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('R', 'e', 'w', 'r', 'i', 't', 'i', 'n', 'g', 'P',
                             'r', 'o', 'c'),
                       bindb[(List[Char]) => List[Char], A,
                              Unit](catcha[(List[Char]) => List[Char], A,
    (List[Char]) =>
      List[Char]](rewriting_proc[A, lab[B,
 List[nat]]](i, u, sta, st, lr, p, dpp),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          A](comp[List[Char], List[Char],
                                   List[Char]](ia,
        comp[List[Char], List[Char],
              List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o',
           'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n',
           'g', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'w', 'r', 'i', 't', 'i', 'n',
           'g', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r', ' ', 't', 'o',
           ' ', 'r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 't', 'h', 'e', ' ', 'p',
           'a', 'i', 'r', ' ')),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_nl,
      comp[List[Char], List[Char],
            List[Char]](shows_rule[lab[B, List[nat]],
                                    List[Char]]((a: lab[B, List[nat]]) =>
          shows_prec_lab[B, List[nat]](Zero_nat, a),
         (a: List[Char]) => shows_prec_list[Char](Zero_nat, a),
         List(' ', '-', '>', ' '), sta),
                         comp[List[Char], List[Char],
                               List[Char]](shows_nl,
    comp[List[Char], List[Char],
          List[Char]](shows_string.apply(List(' ', 't', 'o', ' ', 't', 'h', 'e',
       ' ', 'p', 'a', 'i', 'r', ' ')),
                       comp[List[Char], List[Char],
                             List[Char]](shows_nl,
  comp[List[Char], List[Char],
        List[Char]](shows_rule[lab[B, List[nat]],
                                List[Char]]((a: lab[B, List[nat]]) =>
      shows_prec_lab[B, List[nat]](Zero_nat, a),
     (a: List[Char]) => shows_prec_list[Char](Zero_nat, a),
     List(' ', '-', '>', ' '), st),
                     comp[List[Char], List[Char],
                           List[Char]](shows_nl, x))))))))))),
                                     (dppa: A) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](check_dp_proof[A, B](assms,
   comp[List[Char], List[Char],
         List[Char]](ia, shows_string.apply(List('.', '1'))),
   i, dppa, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](ia, comp[List[Char], List[Char],
                       List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                    'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                    'e', ' ', 'r', 'e', 'w', 'r', 'i', 't', 'i', 'n', 'g', ' ',
                    'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (assms, ia, i, dpp, NarrowingProc(st, p, sts, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('N', 'a', 'r', 'r', 'o', 'w', 'i', 'n', 'g', 'P',
                             'r', 'o', 'c'),
                       bindb[(List[Char]) => List[Char], A,
                              Unit](catcha[(List[Char]) => List[Char], A,
    (List[Char]) =>
      List[Char]](narrowing_proc[A, lab[B, List[nat]]](i, st, p, sts, dpp),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          A](comp[List[Char], List[Char],
                                   List[Char]](ia,
        comp[List[Char], List[Char],
              List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o',
           'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n',
           'g', ' ', 't', 'h', 'e', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o',
           'r', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r', ' ', 't', 'o',
           ' ', 'n', 'a', 'r', 'r', 'o', 'w', ' ', 't', 'h', 'e', ' ', 'p', 'a',
           'i', 'r', ' ')),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_nl,
      comp[List[Char], List[Char],
            List[Char]](shows_rule[lab[B, List[nat]],
                                    List[Char]]((a: lab[B, List[nat]]) =>
          shows_prec_lab[B, List[nat]](Zero_nat, a),
         (a: List[Char]) => shows_prec_list[Char](Zero_nat, a),
         List(' ', '-', '>', ' '), st),
                         comp[List[Char], List[Char],
                               List[Char]](shows_nl,
    comp[List[Char], List[Char],
          List[Char]](shows_string.apply(List(' ', 't', 'o', ' ', 't', 'h', 'e',
       ' ', 'p', 'a', 'i', 'r', 's', ' ')),
                       comp[List[Char], List[Char],
                             List[Char]](shows_nl,
  comp[List[Char], List[Char],
        List[Char]](shows_trs[lab[B, List[nat]],
                               List[Char]]((a: lab[B, List[nat]]) =>
     shows_prec_lab[B, List[nat]](Zero_nat, a),
    (a: List[Char]) => shows_prec_list[Char](Zero_nat, a),
    List('r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'y', 's', 't', 'e', 'm',
          ':'),
    List(' ', '-', '>', ' '), sts),
                     comp[List[Char], List[Char],
                           List[Char]](shows_nl, x))))))))))),
                                     (dppa: A) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](check_dp_proof[A, B](assms,
   comp[List[Char], List[Char],
         List[Char]](ia, shows_string.apply(List('.', '1'))),
   i, dppa, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](ia, comp[List[Char], List[Char],
                       List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                    'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                    'e', ' ', 'n', 'a', 'r', 'r', 'o', 'w', 'i', 'n', 'g', ' ',
                    'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (assms, ia, i, dpp, InstantiationProc(st, sts, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('I', 'n', 's', 't', 'a', 'n', 't', 'i', 'a', 't',
                             'i', 'o', 'n', 'P', 'r', 'o', 'c'),
                       bindb[(List[Char]) => List[Char], A,
                              Unit](catcha[(List[Char]) => List[Char], A,
    (List[Char]) =>
      List[Char]](instantiation_proc[A, lab[B, List[nat]]](i, st, sts, dpp),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          A](comp[List[Char], List[Char],
                                   List[Char]](ia,
        comp[List[Char], List[Char],
              List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o',
           'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n',
           'g', ' ', 't', 'h', 'e', ' ', 'i', 'n', 's', 't', 'a', 'n', 't', 'i',
           'a', 't', 'i', 'o', 'n', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o',
           'r', ' ', 't', 'o', ' ', 'i', 'n', 's', 't', 'a', 'n', 't', 'i', 'a',
           't', 'e', ' ', 't', 'h', 'e', ' ', 'p', 'a', 'i', 'r', ' ')),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_nl,
      comp[List[Char], List[Char],
            List[Char]](shows_rule[lab[B, List[nat]],
                                    List[Char]]((a: lab[B, List[nat]]) =>
          shows_prec_lab[B, List[nat]](Zero_nat, a),
         (a: List[Char]) => shows_prec_list[Char](Zero_nat, a),
         List(' ', '-', '>', ' '), st),
                         comp[List[Char], List[Char],
                               List[Char]](shows_nl,
    comp[List[Char], List[Char],
          List[Char]](shows_string.apply(List(' ', 't', 'o', ' ', 't', 'h', 'e',
       ' ', 'p', 'a', 'i', 'r', 's', ' ')),
                       comp[List[Char], List[Char],
                             List[Char]](shows_nl,
  comp[List[Char], List[Char],
        List[Char]](shows_trs[lab[B, List[nat]],
                               List[Char]]((a: lab[B, List[nat]]) =>
     shows_prec_lab[B, List[nat]](Zero_nat, a),
    (a: List[Char]) => shows_prec_list[Char](Zero_nat, a),
    List('r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'y', 's', 't', 'e', 'm',
          ':'),
    List(' ', '-', '>', ' '), sts),
                     comp[List[Char], List[Char],
                           List[Char]](shows_nl, x))))))))))),
                                     (dppa: A) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](check_dp_proof[A, B](assms,
   comp[List[Char], List[Char],
         List[Char]](ia, shows_string.apply(List('.', '1'))),
   i, dppa, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](ia, comp[List[Char], List[Char],
                       List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                    'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                    'e', ' ', 'i', 'n', 's', 't', 'a', 'n', 't', 'i', 'a', 't',
                    'i', 'o', 'n', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o',
                    'r')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
  case (assms, ia, i, dpp, ForwardInstantiationProc(st, sts, u_opt, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('F', 'o', 'r', 'w', 'a', 'r', 'd', 'I', 'n', 's',
                             't', 'a', 'n', 't', 'i', 'a', 't', 'i', 'o', 'n',
                             'P', 'r', 'o', 'c'),
                       bindb[(List[Char]) => List[Char], A,
                              Unit](catcha[(List[Char]) => List[Char], A,
    (List[Char]) =>
      List[Char]](forward_instantiation_proc[A,
      lab[B, List[nat]]](i, st, sts, u_opt, dpp),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          A](comp[List[Char], List[Char],
                                   List[Char]](ia,
        comp[List[Char], List[Char],
              List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o',
           'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n',
           'g', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'r', 'w', 'a', 'r', 'd', '_',
           'i', 'n', 's', 't', 'a', 'n', 't', 'i', 'a', 't', 'i', 'o', 'n', ' ',
           'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r', ' ', 't', 'o', ' ', 'i',
           'n', 's', 't', 'a', 'n', 't', 'i', 'a', 't', 'e', ' ', 't', 'h', 'e',
           ' ', 'p', 'a', 'i', 'r', ' ')),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_nl,
      comp[List[Char], List[Char],
            List[Char]](shows_rule[lab[B, List[nat]],
                                    List[Char]]((a: lab[B, List[nat]]) =>
          shows_prec_lab[B, List[nat]](Zero_nat, a),
         (a: List[Char]) => shows_prec_list[Char](Zero_nat, a),
         List(' ', '-', '>', ' '), st),
                         comp[List[Char], List[Char],
                               List[Char]](shows_nl,
    comp[List[Char], List[Char],
          List[Char]](shows_string.apply(List(' ', 't', 'o', ' ', 't', 'h', 'e',
       ' ', 'p', 'a', 'i', 'r', 's', ' ')),
                       comp[List[Char], List[Char],
                             List[Char]](shows_nl,
  comp[List[Char], List[Char],
        List[Char]](shows_trs[lab[B, List[nat]],
                               List[Char]]((a: lab[B, List[nat]]) =>
     shows_prec_lab[B, List[nat]](Zero_nat, a),
    (a: List[Char]) => shows_prec_list[Char](Zero_nat, a),
    List('r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'y', 's', 't', 'e', 'm',
          ':'),
    List(' ', '-', '>', ' '), sts),
                     comp[List[Char], List[Char],
                           List[Char]](shows_nl, x))))))))))),
                                     (dppa: A) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](check_dp_proof[A, B](assms,
   comp[List[Char], List[Char],
         List[Char]](ia, shows_string.apply(List('.', '1'))),
   i, dppa, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](ia, comp[List[Char], List[Char],
                       List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                    'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                    'e', ' ', 'f', 'o', 'r', 'w', 'a', 'r', 'd', ' ', 'i', 'n',
                    's', 't', 'a', 'n', 't', 'i', 'a', 't', 'i', 'o', 'n', ' ',
                    'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
}

def check_tp_subsumes[A, B : equal : show, C : equal : show,
                       D : equal : show](i: tp_ops_ext[A, lab[B, C], D, Unit],
  x1: (List[term[lab[B, C], D]],
        (List[(term[lab[B, C], D], term[lab[B, C], D])],
          List[(term[lab[B, C], D], term[lab[B, C], D])])),
  tp: A):
      sum[(List[Char]) => List[Char], Unit] =
  (i, x1, tp) match {
  case (i, (q, (r, rw)), tp) =>
    catcha[(List[Char]) => List[Char], Unit,
            (List[Char]) =>
              List[Char]]({
                            val qaa: List[term[lab[B, C], D]] =
                              (qa[A, lab[B, C], D, Unit](i)).apply(tp)
                            val raa: List[(term[lab[B, C], D],
    term[lab[B, C], D])]
                              = (ra[A, lab[B, C], D, Unit](i)).apply(tp)
                            val rwaa: List[(term[lab[B, C], D],
     term[lab[B, C], D])]
                              = (rwa[A, lab[B, C], D, Unit](i)).apply(tp)
                            val rb: List[(term[lab[B, C], D],
   term[lab[B, C], D])]
                              = r ++ rw;
                            bindb[(List[Char]) => List[Char], Unit,
                                   Unit](catcha[term[lab[B, C], D], Unit,
         (List[Char]) =>
           List[Char]](check_subseteq[term[lab[B, C], D]](qaa, q),
                        (x: term[lab[B, C], D]) =>
                          Inl[(List[Char]) => List[Char],
                               Unit](toomuch(List('t', 'e', 'r', 'm'),
      shows_term[lab[B, C],
                  D]((a: lab[B, C]) => shows_prec_lab[B, C](Zero_nat, a),
                      (a: D) =>
                        (b: List[Char]) => shows_prec[D](Zero_nat, a, b),
                      x)))),
  (_: Unit) =>
    bindb[(List[Char]) => List[Char], Unit,
           Unit](catcha[term[lab[B, C], D], Unit,
                         (List[Char]) =>
                           List[Char]](check_subseteq[term[lab[B, C],
                    D]](q, qaa),
(x: term[lab[B, C], D]) =>
  Inl[(List[Char]) => List[Char],
       Unit](missing(List('t', 'e', 'r', 'm'),
                      shows_term[lab[B, C],
                                  D]((a: lab[B, C]) =>
                                       shows_prec_lab[B, C](Zero_nat, a),
                                      (a: D) =>
(b: List[Char]) => shows_prec[D](Zero_nat, a, b),
                                      x)))),
                  (_: Unit) =>
                    bindb[(List[Char]) => List[Char], Unit,
                           Unit](catcha[(term[lab[B, C], D],
  term[lab[B, C], D]),
 Unit,
 (List[Char]) =>
   List[Char]](check_subseteq[(term[lab[B, C], D], term[lab[B, C], D])](raa, r),
                (x: (term[lab[B, C], D], term[lab[B, C], D])) =>
                  Inl[(List[Char]) => List[Char],
                       Unit](toomuch(List('r', 'u', 'l', 'e'),
                                      shows_rule[lab[B, C],
          D]((a: lab[B, C]) => shows_prec_lab[B, C](Zero_nat, a),
              (a: D) => (b: List[Char]) => shows_prec[D](Zero_nat, a, b),
              List(' ', '-', '>', ' '), x)))),
                                  (_: Unit) =>
                                    bindb[(List[Char]) => List[Char], Unit,
   Unit](catcha[(term[lab[B, C], D], term[lab[B, C], D]), Unit,
                 (List[Char]) =>
                   List[Char]](check_subseteq[(term[lab[B, C], D],
        term[lab[B, C], D])](rwaa, rb),
                                (x: (term[lab[B, C], D], term[lab[B, C], D])) =>
                                  Inl[(List[Char]) => List[Char],
                                       Unit](toomuch(List('r', 'e', 'l', 'a',
                   't', 'i', 'v', 'e', ' ', 'r', 'u', 'l', 'e'),
              shows_rule[lab[B, C],
                          D]((a: lab[B, C]) =>
                               shows_prec_lab[B, C](Zero_nat, a),
                              (a: D) =>
                                (b: List[Char]) =>
                                  shows_prec[D](Zero_nat, a, b),
                              List(' ', '-', '>', ' '), x)))),
          (_: Unit) => Inr[Unit, (List[Char]) => List[Char]](())))))
                          },
                           (x: (List[Char]) => List[Char]) =>
                             Inl[(List[Char]) => List[Char],
                                  Unit](comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List('t', 'e', 'r', 'm', 'i', 'n', 'a',
   't', 'i', 'o', 'n', ' ', 'o', 'f', ' ', 't', 'h', 'e', ' ', 'p', 'r', 'o',
   'b', 'l', 'e', 'm')),
                   comp[List[Char], List[Char],
                         List[Char]](shows_nl,
                                      comp[List[Char], List[Char],
    List[Char]](shows_tp[lab[B, C], D,
                          A]((a: lab[B, C]) =>
                               shows_prec_lab[B, C](Zero_nat, a),
                              (a: D) =>
                                (b: List[Char]) =>
                                  shows_prec[D](Zero_nat, a, b),
                              i, tp),
                 comp[List[Char], List[Char],
                       List[Char]](shows_nl,
                                    comp[List[Char], List[Char],
  List[Char]](shows_string.apply(List('m', 'a', 'y', ' ', 'n', 'o', 't', ' ',
                                       'b', 'e', ' ', 'c', 'o', 'n', 'c', 'l',
                                       'u', 'd', 'e', 'd', ' ', 'f', 'r', 'o',
                                       'm', ' ', 'a', 's', 's', 'u', 'm', 'i',
                                       'n', 'g', ' ', 't', 'e', 'r', 'm', 'i',
                                       'n', 'a', 't', 'i', 'o', 'n', ' ', 'o',
                                       'f', ' ', 't', 'h', 'e', ' ', 'p', 'r',
                                       'o', 'b', 'l', 'e', 'm')),
               comp[List[Char], List[Char],
                     List[Char]](shows_nl,
                                  comp[List[Char], List[Char],
List[Char]](shows_tp[lab[B, C], D,
                      A]((a: lab[B, C]) => shows_prec_lab[B, C](Zero_nat, a),
                          (a: D) =>
                            (b: List[Char]) => shows_prec[D](Zero_nat, a, b),
                          i, (mka[A, lab[B, C], D,
                                   Unit](i)).apply(q).apply(r).apply(rw)),
             comp[List[Char], List[Char],
                   List[Char]](shows_nl,
                                comp[List[Char], List[Char],
                                      List[Char]](x, shows_nl)))))))))))
}

def fcc_tt[A, B : equal : show,
            C : equal : show](i: tp_ops_ext[A, B, C, Unit],
                               fcs: List[ctxt[B, C]],
                               cR: List[(term[B, C], term[B, C])], tp: A):
      sum[(List[Char]) => List[Char], A] =
  {
    val r: List[(term[B, C], term[B, C])] = (ra[A, B, C, Unit](i)).apply(tp)
    val q: List[term[B, C]] = (qa[A, B, C, Unit](i)).apply(tp)
    val vs: List[C] = vars_trs_list[B, C](r)
    val fas: List[(B, nat)] = funas_trs_list[B, C](r);
    bindb[(List[Char]) => List[Char], Unit,
           Unit](check[(List[Char]) =>
                         List[Char]](nulla[term[B, C]](q),
                                      comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('Q', ' ', 'i', 's', ' ', 'n', 'o', 't',
 ' ', 'e', 'm', 'p', 't', 'y')),
                 shows_nl)),
                  (_: Unit) =>
                    bindb[(List[Char]) => List[Char], Unit,
                           Unit](check[(List[Char]) =>
 List[Char]](nulla[(term[B, C], term[B, C])]((rwa[A, B, C, Unit](i)).apply(tp)),
              comp[List[Char], List[Char],
                    List[Char]](shows_string.apply(List('w', 'e', 'a', 'k', ' ',
                 'r', 'u', 'l', 'e', 's', ' ', 'a', 'r', 'e', ' ', 'n', 'o',
                 't', ' ', 'e', 'm', 'p', 't', 'y')),
                                 shows_nl)),
                                  (_: Unit) =>
                                    bindb[(List[Char]) => List[Char], Unit,
   Unit](check[(List[Char]) =>
                 List[Char]](! (nulla[ctxt[B, C]](fcs)),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_string.apply(List('a',
                                 't', ' ', 'l', 'e', 'a', 's', 't', ' ', 'o',
                                 'n', 'e', ' ', 'f', 'l', 'a', 't', ' ', 'c',
                                 'o', 'n', 't', 'e', 'x', 't', ' ', 'i', 's',
                                 ' ', 'r', 'e', 'q', 'u', 'i', 'r', 'e', 'd',
                                 ' ', 'f', 'o', 'r', ' ', 'f', 'l', 'a', 't',
                                 ' ', 'c', 'o', 'n', 't', 'e', 'x', 't', ' ',
                                 'c', 'l', 'o', 's', 'u', 'r', 'e')),
         shows_nl)),
          (_: Unit) =>
            bindb[(List[Char]) => List[Char], Unit,
                   Unit](catcha[(ctxt[B, C], (List[Char]) => List[Char]), Unit,
                                 (List[Char]) =>
                                   List[Char]](forallm[ctxt[B, C],
                (List[Char]) =>
                  List[Char]]((a: ctxt[B, C]) => check_flat_ctxt[C, B](vs, a),
                               fcs),
        (x: (ctxt[B, C], (List[Char]) => List[Char])) =>
          Inl[(List[Char]) => List[Char],
               Unit](snd[ctxt[B, C], (List[Char]) => List[Char]](x))),
                          (_: Unit) =>
                            bindb[(List[Char]) => List[Char], Unit,
                                   Unit](catcha[(ctxt[B, C],
          (List[Char]) => List[Char]),
         Unit,
         (List[Char]) =>
           List[Char]](forallm[ctxt[B, C],
                                (List[Char]) =>
                                  List[Char]]((a: ctxt[B, C]) =>
        check_is_flat_ctxt[C, B](vs, fas, a),
       fcs),
                        (x: (ctxt[B, C], (List[Char]) => List[Char])) =>
                          Inl[(List[Char]) => List[Char],
                               Unit](snd[ctxt[B, C],
  (List[Char]) => List[Char]](x))),
  (_: Unit) =>
    bindb[(List[Char]) => List[Char], Unit,
           Unit](catcha[((B, nat), (List[Char]) => List[Char]), Unit,
                         (List[Char]) =>
                           List[Char]](forallm[(B, nat),
        (List[Char]) =>
          List[Char]]((a: (B, nat)) => check_flat_ctxt_complete[B, C](fcs, a),
                       fas),
(x: ((B, nat), (List[Char]) => List[Char])) =>
  Inl[(List[Char]) => List[Char],
       Unit](snd[(B, nat), (List[Char]) => List[Char]](x))),
                  (_: Unit) =>
                    bindb[(List[Char]) => List[Char], Unit,
                           Unit](catcha[((term[B, C], term[B, C]),
  (List[Char]) => List[Char]),
 Unit,
 (List[Char]) =>
   List[Char]](forallm[(term[B, C], term[B, C]),
                        (List[Char]) =>
                          List[Char]]((a: (term[B, C], term[B, C])) =>
check_rule_preserving[B, C](fcs, cR, a),
                                       r),
                (x: ((term[B, C], term[B, C]), (List[Char]) => List[Char])) =>
                  Inl[(List[Char]) => List[Char],
                       Unit](snd[(term[B, C], term[B, C]),
                                  (List[Char]) => List[Char]](x))),
                                  (_: Unit) =>
                                    catcha[((term[B, C], term[B, C]),
     (List[Char]) => List[Char]),
    Unit,
    (List[Char]) =>
      List[Char]](forallm[(term[B, C], term[B, C]),
                           (List[Char]) =>
                             List[Char]]((a: (term[B, C], term[B, C])) =>
   check_rule_reflecting[B, C](fcs, r, a),
  cR),
                   (x: ((term[B, C], term[B, C]), (List[Char]) => List[Char]))
                     => Inl[(List[Char]) => List[Char],
                             Unit](snd[(term[B, C], term[B, C]),
(List[Char]) => List[Char]](x))))))))))
      match {
      case Inl(a) => Inl[(List[Char]) => List[Char], A](a)
      case Inr(_) =>
        Inr[A, (List[Char]) =>
                 List[Char]]((mka[A, B, C,
                                   Unit](i)).apply(Nil).apply(cR).apply(Nil))
    }
  }

def rule_removal_tt[A, B : show,
                     C : show](i: tp_ops_ext[A, B, C, Unit],
                                rp: redtriple_ext[B, C, Unit],
                                rremove: List[(term[B, C], term[B, C])],
                                trs: A):
      sum[(List[Char]) => List[Char], A] =
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]]({
                          val (rs, rns):
                                (List[(term[B, C], term[B, C])],
                                  List[(term[B, C], term[B, C])])
                            = (split_rulesa[A, B, C,
     Unit](i)).apply(trs).apply(rremove);
                          bindb[(List[Char]) => List[Char], Unit,
                                 Unit](valid[B, C, Unit](rp),
(_: Unit) =>
  bindb[(List[Char]) => List[Char], Unit,
         Unit](catcha[(List[Char]) => List[Char], Unit,
                       (List[Char]) =>
                         List[Char]](mono[B, C, Unit](rp),
                                      (x: (List[Char]) => List[Char]) =>
Inl[(List[Char]) => List[Char],
     Unit](comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List('p', 'r', 'o', 'b', 'l',
              'e', 'm', ' ', 'w', 'i', 't', 'h', ' ', 'm', 'o', 'n', 'o', 't',
              'o', 'n', 'i', 'c', 'i', 't', 'y', ' ', 'o', 'f', ' ', 's', 't',
              'r', 'i', 'c', 't', ' ', 'o', 'r', 'd', 'e', 'r')),
                              comp[List[Char], List[Char],
                                    List[Char]](shows_nl, x)))),
                (_: Unit) =>
                  bindb[(List[Char]) => List[Char], Unit,
                         Unit](catcha[(List[Char]) => List[Char], Unit,
                                       (List[Char]) =>
 List[Char]](catcha[((term[B, C], term[B, C]), (List[Char]) => List[Char]),
                     Unit,
                     (List[Char]) =>
                       List[Char]](forallm[(term[B, C], term[B, C]),
    (List[Char]) => List[Char]](ns[B, C, Unit](rp), rns),
                                    (x: ((term[B, C], term[B, C]),
  (List[Char]) => List[Char]))
                                      => Inl[(List[Char]) => List[Char],
      Unit](snd[(term[B, C], term[B, C]), (List[Char]) => List[Char]](x))),
              (x: (List[Char]) => List[Char]) =>
                Inl[(List[Char]) => List[Char],
                     Unit](comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('p', 'r',
                              'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n',
                              ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g',
                              ' ', 'T', 'R', 'S')),
      comp[List[Char], List[Char], List[Char]](shows_nl, x)))),
                                (_: Unit) =>
                                  catcha[(List[Char]) => List[Char], Unit,
  (List[Char]) =>
    List[Char]](catcha[((term[B, C], term[B, C]), (List[Char]) => List[Char]),
                        Unit,
                        (List[Char]) =>
                          List[Char]](forallm[(term[B, C], term[B, C]),
       (List[Char]) => List[Char]](s[B, C, Unit](rp), rs),
                                       (x:
  ((term[B, C], term[B, C]), (List[Char]) => List[Char]))
 => Inl[(List[Char]) => List[Char],
         Unit](snd[(term[B, C], term[B, C]), (List[Char]) => List[Char]](x))),
                 (x: (List[Char]) => List[Char]) =>
                   Inl[(List[Char]) => List[Char],
                        Unit](comp[List[Char], List[Char],
                                    List[Char]](shows_string.apply(List('p',
                                 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h',
                                 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n', 't',
                                 'i', 'n', 'g', ' ', 'T', 'R', 'S')),
         comp[List[Char], List[Char], List[Char]](shows_nl, x)))))))
                        },
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ', 'n', 'o',
 't', ' ', 'a', 'p', 'p', 'l', 'y', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'd', 'u',
 'c', 't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c', 'e',
 's', 's', 'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h', 'e', ' ', 'f', 'o',
 'l', 'l', 'o', 'w', 'i', 'n', 'g')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_nl,
                                    comp[List[Char], List[Char],
  List[Char]](desca[B, C, Unit](rp),
               comp[List[Char], List[Char], List[Char]](shows_nl, x))))))
    match {
    case Inl(a) => Inl[(List[Char]) => List[Char], A](a)
    case Inr(_) =>
      Inr[A, (List[Char]) =>
               List[Char]]((delete_R_Rwa[A, B, C,
  Unit](i)).apply(trs).apply(rremove).apply(rremove))
  }

def sem_lab_rel_tt[A : equal : show, B : equal : show, C,
                    D](splitter:
                         (List[(term[A, B], term[A, B])]) =>
                           (List[(term[A, B], term[A, B])]) =>
                             (List[(term[A, B], term[A, B])],
                               (List[(term[A, B], term[A, B])],
                                 List[(term[A, B], term[A, B])])),
                        ld: A => (A, C), i: tp_ops_ext[D, A, B, Unit],
                        valid: sum[(List[Char]) => List[Char], Unit],
                        check_decr:
                          (List[(term[A, B], term[A, B])]) =>
                            sum[(List[Char]) => List[Char], Unit],
                        check_model_lab:
                          (List[(term[A, B], term[A, B])]) =>
                            (List[(term[A, B], term[A, B])]) =>
                              sum[(List[Char]) => List[Char], Unit],
                        lQ: List[term[A, B]],
                        lAll: List[(term[A, B], term[A, B])], tp: D):
      sum[(List[Char]) => List[Char], D] =
  {
    val r: List[(term[A, B], term[A, B])] = (ra[D, A, B, Unit](i)).apply(tp)
    val rw: List[(term[A, B], term[A, B])] = (rwa[D, A, B, Unit](i)).apply(tp)
    val (lR, (lRw, d)):
          (List[(term[A, B], term[A, B])],
            (List[(term[A, B], term[A, B])], List[(term[A, B], term[A, B])]))
      = (splitter(lAll))(rw);
    bindb[(List[Char]) => List[Char], Unit,
           Unit](valid,
                  (_: Unit) =>
                    {
                      val q: List[term[A, B]] =
                        (qa[D, A, B, Unit](i)).apply(tp);
                      catcha[(List[Char]) => List[Char], Unit,
                              (List[Char]) =>
                                List[Char]](bindb[(List[Char]) => List[Char],
           Unit,
           Unit](check_decr(d),
                  (_: Unit) =>
                    bindb[(List[Char]) => List[Char], Unit,
                           Unit](check_sl_Qa[A, C, B](ld, lQ, q),
                                  (_: Unit) =>
                                    bindb[(List[Char]) => List[Char], Unit,
   Unit]((check_model_lab(lR))(r), (_: Unit) => (check_model_lab(lRw))(rw)))),
     (x: (List[Char]) => List[Char]) =>
       Inl[(List[Char]) => List[Char],
            Unit](comp[List[Char], List[Char],
                        List[Char]](shows_string.apply(List('p', 'r', 'o', 'b',
                     'l', 'e', 'm', ' ', 'w', 'i', 't', 'h', ' ', 'l', 'a', 'b',
                     'e', 'l', 'e', 'd', ' ', 'T', 'R', 'S', ':')),
                                     comp[List[Char], List[Char],
   List[Char]](shows_nl, x))))
                    })
      match {
      case Inl(a) => Inl[(List[Char]) => List[Char], D](a)
      case Inr(_) =>
        Inr[D, (List[Char]) =>
                 List[Char]]((mka[D, A, B,
                                   Unit](i)).apply(lQ).apply(lR).apply(lRw ++
                                 d))
    }
  }

def sem_lab_fin_tt[A : equal : show, B : equal : show, C, D : show,
                    E](splitter:
                         (List[(term[A, B], term[A, B])]) =>
                           (List[(term[A, B], term[A, B])]) =>
                             (List[(term[A, B], term[A, B])],
                               (List[(term[A, B], term[A, B])],
                                 List[(term[A, B], term[A, B])])),
                        lc: A => nat => C => A, ld: A => (A, C),
                        cge: D => D => Boolean, i: tp_ops_ext[E, A, B, Unit],
                        gen: (List[(A, nat)]) =>
                               (List[(A, nat)]) =>
                                 sum[(List[Char]) => List[Char],
                                      sl_ops_ext[A, D, C, B, Unit]],
                        lQ: List[term[A, B]],
                        lAll: List[(term[A, B], term[A, B])], tp: E):
      sum[(List[Char]) => List[Char], E] =
  bindb[(List[Char]) => List[Char], sl_ops_ext[A, D, C, B, Unit],
         E]((gen(funas_trs_list[A, B]((rulesb[E, A, B,
       Unit](i)).apply(tp))))(Nil),
             (ops: sl_ops_ext[A, D, C, B, Unit]) =>
               {
                 val check_ml:
                       (List[(term[A, B], term[A, B])]) =>
                         (List[(term[A, B], term[A, B])]) =>
                           sum[(List[Char]) => List[Char], Unit]
                   = (a: List[(term[A, B], term[A, B])]) =>
                       (b: List[(term[A, B], term[A, B])]) =>
                         check_sl_model_lab_trs[A, D, C,
         B](sl_I[A, D, C, B, Unit](ops), sl_L[A, D, C, B, Unit](ops),
             sl_C[A, D, C, B, Unit](ops), cge, lc, a, b)
                 val check_d:
                       (List[(term[A, B], term[A, B])]) =>
                         sum[(List[Char]) => List[Char], Unit]
                   = sl_check_decr[A, D, C, B, Unit](ops);
                 sem_lab_rel_tt[A, B, C,
                                 E](splitter, ld, i,
                                     Inr[Unit, (List[Char]) => List[Char]](()),
                                     check_d, check_ml, lQ, lAll, tp)
               })

def semlab_fin_tt[A, B : equal : show,
                   C : equal : show](j: tp_ops_ext[A, lab[B, List[nat]], C,
            Unit],
                                      x1: sl_variant[lab[B, List[nat]], C]):
      (List[term[lab[B, List[nat]], C]]) =>
        (List[(term[lab[B, List[nat]], C], term[lab[B, List[nat]], C])]) =>
          A => sum[(List[Char]) => List[Char], A] =
  (j, x1) match {
  case (j, Rootlab(uu)) =>
    (a: List[term[lab[B, List[nat]], C]]) =>
      (b: List[(term[lab[B, List[nat]], C], term[lab[B, List[nat]], C])]) =>
        (c: A) =>
          sem_lab_fin_tt[lab[B, List[nat]], C,
                          sum[List[nat], List[lab[B, List[nat]]]],
                          lab[B, List[nat]],
                          A]((aa: List[(term[lab[B, List[nat]], C],
 term[lab[B, List[nat]], C])])
                               => (ba: List[(term[lab[B, List[nat]], C],
      term[lab[B, List[nat]], C])])
                                    => model_splitter[lab[B, List[nat]],
               sum[List[nat], List[lab[B, List[nat]]]],
               C]((ab: lab[B, List[nat]]) => label_decomp[B, List[nat]](ab), aa,
                   ba),
                              (aa: lab[B, List[nat]]) =>
                                (ba: nat) =>
                                  (ca: sum[List[nat], List[lab[B, List[nat]]]])
                                    => label[B, List[nat]](aa, ba, ca),
                              (aa: lab[B, List[nat]]) =>
                                label_decomp[B, List[nat]](aa),
                              (aa: lab[B, List[nat]]) =>
                                (ba: lab[B, List[nat]]) =>
                                  equal_laba[B, List[nat]](aa, ba),
                              j, slm_gen_to_sl_gen[lab[B, List[nat]],
            lab[B, List[nat]], sum[List[nat], List[lab[B, List[nat]]]],
            C]((aa: List[(lab[B, List[nat]], nat)]) =>
                 (ba: List[(lab[B, List[nat]], nat)]) =>
                   rl_slm[B, List[nat]](None, aa, ba)),
                              a, b, c)
  case (j, Finitelab(sli)) =>
    (a: List[term[lab[B, List[nat]], C]]) =>
      (b: List[(term[lab[B, List[nat]], C], term[lab[B, List[nat]], C])]) =>
        (c: A) =>
          sem_lab_fin_tt[lab[B, List[nat]], C,
                          sum[List[nat], List[lab[B, List[nat]]]], nat,
                          A]((aa: List[(term[lab[B, List[nat]], C],
 term[lab[B, List[nat]], C])])
                               => (ba: List[(term[lab[B, List[nat]], C],
      term[lab[B, List[nat]], C])])
                                    => model_splitter[lab[B, List[nat]],
               sum[List[nat], List[lab[B, List[nat]]]],
               C]((ab: lab[B, List[nat]]) => label_decomp[B, List[nat]](ab), aa,
                   ba),
                              (aa: lab[B, List[nat]]) =>
                                (ba: nat) =>
                                  (ca: sum[List[nat], List[lab[B, List[nat]]]])
                                    => label[B, List[nat]](aa, ba, ca),
                              (aa: lab[B, List[nat]]) =>
                                label_decomp[B, List[nat]](aa),
                              (aa: nat) => (ba: nat) => equal_nata(aa, ba), j,
                              slm_gen_to_sl_gen[lab[B, List[nat]], nat,
         sum[List[nat], List[lab[B, List[nat]]]],
         C]((_: List[(lab[B, List[nat]], nat)]) =>
              (_: List[(lab[B, List[nat]], nat)]) =>
                Inr[slm_ops_ext[lab[B, List[nat]], nat,
                                 sum[List[nat], List[lab[B, List[nat]]]], Unit],
                     (List[Char]) => List[Char]](sli_to_slm[B](sli))),
                              a, b, c)
  case (j, QuasiFinitelab(sli, v)) =>
    (a: List[term[lab[B, List[nat]], C]]) =>
      (b: List[(term[lab[B, List[nat]], C], term[lab[B, List[nat]], C])]) =>
        (c: A) =>
          sem_lab_fin_tt[lab[B, List[nat]], C,
                          sum[List[nat], List[lab[B, List[nat]]]], nat,
                          A]((aa: List[(term[lab[B, List[nat]], C],
 term[lab[B, List[nat]], C])])
                               => (ba: List[(term[lab[B, List[nat]], C],
      term[lab[B, List[nat]], C])])
                                    => quasi_splitter[lab[B, List[nat]],
               sum[List[nat], List[lab[B, List[nat]]]],
               C]((ab: lab[B, List[nat]]) => label_decomp[B, List[nat]](ab), aa,
                   ba),
                              (aa: lab[B, List[nat]]) =>
                                (ba: nat) =>
                                  (ca: sum[List[nat], List[lab[B, List[nat]]]])
                                    => label[B, List[nat]](aa, ba, ca),
                              (aa: lab[B, List[nat]]) =>
                                label_decomp[B, List[nat]](aa),
                              qmodel_cge, j,
                              (f: List[(lab[B, List[nat]], nat)]) =>
                                (g: List[(lab[B, List[nat]], nat)]) =>
                                  qsli_to_sl[C, B](v, f, g, sli),
                              a, b, c)
}

def check_unary_signature[A : show,
                           B : show](r: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(term[A, B], term[A, B]), Unit,
          (List[Char]) =>
            List[Char]](catcha[((term[A, B], term[A, B]),
                                 (term[A, B], term[A, B])),
                                Unit,
                                (term[A, B],
                                  term[A,
B])](forallm[(term[A, B], term[A, B]),
              (term[A, B],
                term[A, B])]((x: (term[A, B], term[A, B])) =>
                               (if ({
                                      val (l, ra): (term[A, B], term[A, B]) = x;
                                      unary_term[A, B](l) &&
unary_term[A, B](ra)
                                    })
                                 Inr[Unit, (term[A, B], term[A, B])](())
                                 else Inl[(term[A, B], term[A, B]), Unit](x)),
                              r),
      (x: ((term[A, B], term[A, B]), (term[A, B], term[A, B]))) =>
        Inl[(term[A, B], term[A, B]),
             Unit](snd[(term[A, B], term[A, B]), (term[A, B], term[A, B])](x))),
                         (x: (term[A, B], term[A, B])) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('t', 'h', 'e', ' ', 'r', 'u', 'l', 'e',
 ' ')),
                 comp[List[Char], List[Char],
                       List[Char]](shows_nl,
                                    comp[List[Char], List[Char],
  List[Char]](shows_rule[A, B]((a: A) =>
                                 (b: List[Char]) =>
                                   shows_prec[A](Zero_nat, a, b),
                                (a: B) =>
                                  (b: List[Char]) =>
                                    shows_prec[B](Zero_nat, a, b),
                                List(' ', '-', '>', ' '), x),
               comp[List[Char], List[Char],
                     List[Char]](shows_nl,
                                  comp[List[Char], List[Char],
List[Char]](shows_string.apply(List('v', 'i', 'o', 'l', 'a', 't', 'e', 's', ' ',
                                     't', 'h', 'e', ' ', 'c', 'o', 'n', 'd',
                                     'i', 't', 'i', 'o', 'n', ' ', 't', 'h',
                                     'a', 't', ' ', 'a', 'l', 'l', ' ', 'f',
                                     'u', 'n', 'c', 't', 'i', 'o', 'n', ' ',
                                     's', 'y', 'm', 'b', 'o', 'l', 's')),
             comp[List[Char], List[Char],
                   List[Char]](shows_nl,
                                shows_prec_list[Char](Zero_nat,
               List('h', 'a', 'v', 'e', ' ', 't', 'o', ' ', 'b', 'e', ' ', 'u',
                     'n', 'a', 'r', 'y'))))))))))

def string_reversal_tt[A, B : show,
                        C : show](i: tp_ops_ext[A, B, C, Unit], trs: A):
      sum[(List[Char]) => List[Char], A] =
  {
    val r: List[(term[B, C], term[B, C])] =
      (rulesb[A, B, C, Unit](i)).apply(trs);
    check_unary_signature[B, C](r) match {
      case Inl(a) => Inl[(List[Char]) => List[Char], A](a)
      case Inr(_) =>
        Inr[A, (List[Char]) =>
                 List[Char]]((mka[A, B, C,
                                   Unit](i)).apply(Nil).apply(mapa[(term[B, C],
                             term[B, C]),
                            (term[B, C],
                              term[B, C])]((a: (term[B, C], term[B, C])) =>
     rev_rule[B, C](a),
    r)).apply(Nil))
    }
  }

def switch_innermost_tt[A, B : equal : show](i:
       tp_ops_ext[A, B, List[Char], Unit],
      joins_o:
        Option[List[(term[B, List[Char]],
                      (List[(pos, ((term[B, List[Char]], term[B, List[Char]]),
                                    term[B, List[Char]]))],
                        (term[B, List[Char]],
                          List[(pos, ((term[B, List[Char]],
term[B, List[Char]]),
                                       term[B, List[Char]]))])))]],
      trs: A):
      sum[(List[Char]) => List[Char], A] =
  {
    val r: List[(term[B, List[Char]], term[B, List[Char]])] =
      (rulesb[A, B, List[Char], Unit](i)).apply(trs);
    {
      val cp: List[(Boolean, (term[B, List[Char]], term[B, List[Char]]))] =
        critical_pairs_impl[B](r, r);
      bindb[(List[Char]) => List[Char], Unit,
             Unit](catcha[((Boolean,
                             (term[B, List[Char]], term[B, List[Char]])),
                            (List[Char]) => List[Char]),
                           Unit,
                           (List[Char]) =>
                             List[Char]](forallm[(Boolean,
           (term[B, List[Char]], term[B, List[Char]])),
          (List[Char]) =>
            List[Char]]((a: (Boolean,
                              (term[B, List[Char]], term[B, List[Char]])))
                          => {
                               val (b, _):
                                     (Boolean,
                                       (term[B, List[Char]],
 term[B, List[Char]]))
                                 = a;
                               check[(List[Char]) =>
                                       List[Char]](b,
            shows_prec_list[Char](Zero_nat,
                                   List('r', 'u', 'l', 'e', 's', ' ', 'a', 'r',
 'e', ' ', 'n', 'o', 't', ' ', 'o', 'v', 'e', 'r', 'l', 'a', 'y')))
                             },
                         cp),
  (x: ((Boolean, (term[B, List[Char]], term[B, List[Char]])),
        (List[Char]) => List[Char]))
    => Inl[(List[Char]) => List[Char],
            Unit](snd[(Boolean, (term[B, List[Char]], term[B, List[Char]])),
                       (List[Char]) => List[Char]](x))),
                    (_: Unit) => check_critical_pairs[B](r, cp, joins_o))
    }
      match {
      case Inl(a) => Inl[(List[Char]) => List[Char], A](a)
      case Inr(_) =>
        Inr[A, (List[Char]) =>
                 List[Char]]((mka[A, B, List[Char],
                                   Unit](i)).apply(mapa[(term[B, List[Char]],
                  term[B, List[Char]]),
                 term[B, List[Char]]]((a:
 (term[B, List[Char]], term[B, List[Char]]))
=> fst[term[B, List[Char]], term[B, List[Char]]](a),
                                       r)).apply(r).apply(Nil))
    }
  }

def dP_list[A : equal,
             B : equal](shp: A => A, r: List[(term[A, B], term[A, B])],
                         d_list: List[(A, nat)]):
      List[(term[A, B], term[A, B])] =
  maps[(term[A, B], term[A, B]),
        (term[A, B],
          term[A, B])]((lr: (term[A, B], term[A, B])) =>
                         {
                           val l: term[A, B] = fst[term[A, B], term[A, B]](lr)
                           val s: term[A, B] = sharp_term[A, B](shp, l);
                           map_filter[term[A, B],
                                       (term[A, B],
 term[A, B])]((x: term[A, B]) =>
                (if (! (member[(term[A, B], term[A, B])]((l, x), supt[A, B])) &&
                       (! (is_Var[A, B](x)) &&
                         membera[(A, nat)](d_list,
    the[(A, nat)](root[A, B](x)))))
                  Some[(term[A, B], term[A, B])]((s, sharp_term[A, B](shp, x)))
                  else None),
               supteq_list[A, B](snd[term[A, B], term[A, B]](lr)))
                         },
                        r)

def is_QNFa[A, B, C, D](x0: tp_ops_ext[A, B, C, D]):
      A => (term[B, C]) => Boolean =
  x0 match {
  case Tp_ops_ext(qreltrs, q, r, rw, rules, q_empty, is_QNF,
                   nFQ_subset_NF_rules, rules_map, delete_R_Rw, split_rules, mk,
                   more)
    => is_QNF
}

def check_subseteq_sorted[A : equal : linorder](xs: List[A], ys: List[A]):
      sum[A, Unit] =
  sorted_list_subset[A]((sort_key[A, A]((x: A) => x)).apply(xs),
                         (sort_key[A, A]((x: A) => x)).apply(ys))
    match {
    case None => Inr[Unit, A](())
    case Some(a) => Inl[A, Unit](a)
  }

def useful_rule_impla[A, B](q: (term[A, B]) => Boolean,
                             lr: (term[A, B], term[A, B])):
      Boolean =
  list_all[term[A, B]](q, args[A, B](fst[term[A, B], term[A, B]](lr)))

def dependency_pairs_tt[A, B : equal : linorder : show,
                         C : equal : linorder : show,
                         D](i: tp_ops_ext[A, B, C, Unit],
                             j: dpp_ops_ext[D, B, C, Unit], tp: A, shp: B => B,
                             p: List[(term[B, C], term[B, C])]):
      sum[(List[Char]) => List[Char], D] =
  {
    val r: List[(term[B, C], term[B, C])] = (rulesb[A, B, C, Unit](i)).apply(tp)
    val q: List[term[B, C]] = (qa[A, B, C, Unit](i)).apply(tp)
    val iQ: (term[B, C]) => Boolean = (is_QNFa[A, B, C, Unit](i)).apply(tp)
    val u: List[(term[B, C], term[B, C])] =
      filter[(term[B, C],
               term[B, C])]((a: (term[B, C], term[B, C])) =>
                              useful_rule_impla[B, C](iQ, a),
                             r);
    catcha[(List[Char]) => List[Char], Unit,
            (List[Char]) =>
              List[Char]](bindb[(List[Char]) => List[Char], Unit,
                                 Unit](check_wf_trs[B, C](u),
(_: Unit) =>
  bindb[(List[Char]) => List[Char], Unit,
         Unit](catcha[(term[B, C], (List[Char]) => List[Char]), Unit,
                       (List[Char]) =>
                         List[Char]](forallm[term[B, C],
      (List[Char]) => List[Char]]((a: term[B, C]) => check_no_var[B, C](a), q),
                                      (x:
 (term[B, C], (List[Char]) => List[Char]))
=> Inl[(List[Char]) => List[Char],
        Unit](snd[term[B, C], (List[Char]) => List[Char]](x))),
                (_: Unit) =>
                  {
                    val qr: List[(B, nat)] =
                      mapa[term[B, C],
                            (B, nat)]((a: term[B, C]) =>
{
  val (Fun(f, ss)): term[B, C] = a;
  (f, size_list[term[B, C]](ss))
},
                                       q)
                    val d: List[(B, nat)] = defs_list[B, C](u);
                    bindb[(List[Char]) => List[Char], Unit,
                           Unit](catcha[((B, nat), (List[Char]) => List[Char]),
 Unit,
 (List[Char]) =>
   List[Char]](forallm[(B, nat),
                        (List[Char]) =>
                          List[Char]]((a: (B, nat)) =>
{
  val (f, n): (B, nat) = a;
  check[(List[Char]) =>
          List[Char]](! (membera[(B, nat)](d, (shp(f), n))),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('s', 'h', 'a',
                          'r', 'p', 'i', 'n', 'g', ' ')),
  comp[List[Char], List[Char],
        List[Char]]((aa: List[Char]) => shows_prec[B](Zero_nat, f, aa),
                     comp[List[Char], List[Char],
                           List[Char]](shows_string.apply(List(' ', 'y', 'i',
                        'e', 'l', 'd', 's', ' ', 't', 'h', 'e', ' ', 'd', 'e',
                        'f', 'i', 'n', 'e', 'd', ' ', 's', 'y', 'm', 'b', 'o',
                        'l', ' ')),
(aa: List[Char]) => shows_prec[B](Zero_nat, shp(f), aa)))))
},
                                       d),
                (x: ((B, nat), (List[Char]) => List[Char])) =>
                  Inl[(List[Char]) => List[Char],
                       Unit](snd[(B, nat), (List[Char]) => List[Char]](x))),
                                  (_: Unit) =>
                                    bindb[(List[Char]) => List[Char], Unit,
   Unit](catcha[((B, nat), (List[Char]) => List[Char]), Unit,
                 (List[Char]) =>
                   List[Char]](forallm[(B, nat),
(List[Char]) =>
  List[Char]]((a: (B, nat)) =>
                {
                  val (f, n): (B, nat) = a;
                  check[(List[Char]) =>
                          List[Char]](! (membera[(B, nat)](qr, (shp(f), n))),
                                       comp[List[Char], List[Char],
     List[Char]](shows_string.apply(List('s', 'h', 'a', 'r', 'p', 'i', 'n', 'g',
  ' ')),
                  comp[List[Char], List[Char],
                        List[Char]]((aa: List[Char]) =>
                                      shows_prec[B](Zero_nat, f, aa),
                                     comp[List[Char], List[Char],
   List[Char]](shows_string.apply(List(' ', 'y', 'i', 'e', 'l', 'd', 's', ' ',
't', 'h', 'e', ' ', 's', 'y', 'm', 'b', 'o', 'l', ' ')),
                comp[List[Char], List[Char],
                      List[Char]]((aa: List[Char]) =>
                                    shows_prec[B](Zero_nat, shp(f), aa),
                                   shows_string.apply(List(' ', 'w', 'h', 'i',
                    'c', 'h', ' ', 'i', 's', ' ', 'a', ' ', 'r', 'o', 'o', 't',
                    ' ', 'o', 'f', ' ', 'Q')))))))
                },
               d),
                                (x: ((B, nat), (List[Char]) => List[Char])) =>
                                  Inl[(List[Char]) => List[Char],
                                       Unit](snd[(B, nat),
          (List[Char]) => List[Char]](x))),
          (_: Unit) =>
            catcha[(term[B, C], term[B, C]), Unit,
                    (List[Char]) =>
                      List[Char]](check_subseteq_sorted[(term[B, C],
                  term[B, C])](dP_list[B, C](shp, u, d), p),
                                   (x: (term[B, C], term[B, C])) =>
                                     Inl[(List[Char]) => List[Char],
  Unit](comp[List[Char], List[Char],
              List[Char]](shows_string.apply(List('t', 'h', 'e', ' ', 'D', 'P',
           ' ')),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_rule[B,
                 C]((a: B) => (b: List[Char]) => shows_prec[B](Zero_nat, a, b),
                     (a: C) => (b: List[Char]) => shows_prec[C](Zero_nat, a, b),
                     List(' ', '-', '>', ' '), x),
      comp[List[Char], List[Char],
            List[Char]](shows_string.apply(List(' ', 'd', 'o', 'e', 's', ' ',
         'n', 'o', 't', ' ', 'a', 'p', 'p', 'e', 'a', 'r', ' ', 'i', 'n', ' ',
         't', 'h', 'e', ' ', 'D', 'P', ' ', 'p', 'r', 'o', 'b', 'l', 'e', 'm')),
                         shows_nl)))))))
                  })),
                           (x: (List[Char]) => List[Char]) =>
                             Inl[(List[Char]) => List[Char],
                                  Unit](comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List('t', 'h', 'e', ' ', 'D', 'P', '-',
   't', 'r', 'a', 'n', 's', 'f', 'o', 'r', 'm', 'a', 't', 'i', 'o', 'n', ' ',
   'i', 's', ' ', 'n', 'o', 't', ' ', 'a', 'p', 'p', 'l', 'i', 'e', 'd', ' ',
   'c', 'o', 'r', 'r', 'e', 'c', 't', 'l', 'y', '.')),
                   comp[List[Char], List[Char], List[Char]](shows_nl, x))))
      match {
      case Inl(a) => Inl[(List[Char]) => List[Char], D](a)
      case Inr(_) =>
        Inr[D, (List[Char]) =>
                 List[Char]]((mkc[D, B, C,
                                   Unit](j)).apply(p).apply(Nil).apply(q).apply(Nil).apply(r))
    }
  }

def check_trs_proof[A, B : equal : linorder : show,
                     C](a: Boolean, ia: (List[Char]) => List[Char],
                         i: tp_ops_ext[A, lab[B, List[nat]], List[Char], Unit],
                         j: dpp_ops_ext[C, lab[B, List[nat]], List[Char], Unit],
                         tp: A, x5: trs_proof[B, List[nat], List[Char]]):
      sum[(List[Char]) => List[Char], Unit] =
  (a, ia, i, j, tp, x5) match {
  case (a, ia, i, j, tp, RisEmpty) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('R', ' ', 'i', 's', ' ', 'e', 'm', 'p', 't', 'y'),
                       (if (nulla[(term[lab[B, List[nat]], List[Char]],
                                    term[lab[B, List[nat]],
  List[Char]])]((ra[A, lab[B, List[nat]], List[Char], Unit](i)).apply(tp)))
                         Inr[Unit, (List[Char]) => List[Char]](())
                         else Inl[(List[Char]) => List[Char],
                                   Unit](comp[List[Char], List[Char],
       List[Char]](ia, comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List(':', ' ', 'R',
                          ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'e', 'm', 'p',
                          't', 'y', ' ', 'i', 'n', ' ', 't', 'h', 'e', ' ', 'f',
                          'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g', ' ', 't', 'e',
                          'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', '-', 'p',
                          'r', 'o', 'b', 'l', 'e', 'm', ' ')),
  comp[List[Char], List[Char],
        List[Char]](shows_nl,
                     shows_tp[lab[B, List[nat]], List[Char],
                               A]((aa: lab[B, List[nat]]) =>
                                    shows_prec_lab[B, List[nat]](Zero_nat, aa),
                                   (aa: List[Char]) =>
                                     shows_prec_list[Char](Zero_nat, aa),
                                   i, tp)))))))
  case (a, ia, i, j, tp, RuleRemoval(redp, rR, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('R', 'u', 'l', 'e', ' ', 'R', 'e', 'm', 'o', 'v',
                             'a', 'l'),
                       {
                         val r: List[(term[lab[B, List[nat]], List[Char]],
                                       term[lab[B, List[nat]], List[Char]])]
                           = (rulesb[A, lab[B, List[nat]], List[Char],
                                      Unit](i)).apply(tp)
                         val rr: List[(term[lab[B, List[nat]], List[Char]],
term[lab[B, List[nat]], List[Char]])]
                           = list_diff[(term[lab[B, List[nat]], List[Char]],
 term[lab[B, List[nat]], List[Char]])](r, rR);
                         bindb[(List[Char]) => List[Char], A,
                                Unit](catcha[(List[Char]) => List[Char], A,
      (List[Char]) =>
        List[Char]](rule_removal_tt[A, lab[B, List[nat]],
                                     List[Char]](i,
          get_redtriple[lab[B, List[nat]], List[Char]](redp), rr, tp),
                     (x: (List[Char]) => List[Char]) =>
                       Inl[(List[Char]) => List[Char],
                            A](comp[List[Char], List[Char],
                                     List[Char]](ia,
          comp[List[Char], List[Char],
                List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r',
             'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
             'i', 'n', 'g', ' ', 't', 'h', 'e', ' ', 'r', 'u', 'l', 'e', ' ',
             'r', 'e', 'm', 'o', 'v', 'a', 'l', ' ', 't', 'e', 'c', 'h', 'n',
             'i', 'q', 'u', 'e', ' ', 'o', 'n', ' ')),
                             comp[List[Char], List[Char],
                                   List[Char]](shows_nl,
        comp[List[Char], List[Char],
              List[Char]](shows_tp[lab[B, List[nat]], List[Char],
                                    A]((aa: lab[B, List[nat]]) =>
 shows_prec_lab[B, List[nat]](Zero_nat, aa),
(aa: List[Char]) => shows_prec_list[Char](Zero_nat, aa), i, tp),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_nl,
      comp[List[Char], List[Char],
            List[Char]](shows_string.apply(List(' ', 't', 'o', ' ', 'r', 'e',
         'm', 'o', 'v', 'e', ' ', 't', 'h', 'e', ' ', 'r', 'u', 'l', 'e', 's',
         ' ')),
                         comp[List[Char], List[Char],
                               List[Char]](shows_nl,
    comp[List[Char], List[Char],
          List[Char]](shows_rules[lab[B, List[nat]],
                                   List[Char]]((aa: lab[B, List[nat]]) =>
         shows_prec_lab[B, List[nat]](Zero_nat, aa),
        (aa: List[Char]) => shows_prec_list[Char](Zero_nat, aa),
        List(' ', '-', '>', ' '), rr),
                       comp[List[Char], List[Char],
                             List[Char]](shows_nl, x))))))))))),
                                       (tpa: A) =>
 catcha[(List[Char]) => List[Char], Unit,
         (List[Char]) =>
           List[Char]](check_trs_proof[A, B,
C](a, comp[List[Char], List[Char],
            List[Char]](ia, shows_string.apply(List('.', '1'))),
    i, j, tpa, prf),
                        (x: (List[Char]) => List[Char]) =>
                          Inl[(List[Char]) => List[Char],
                               Unit](comp[List[Char], List[Char],
   List[Char]](ia, comp[List[Char], List[Char],
                         List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                      'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't',
                      'h', 'e', ' ', 'r', 'u', 'l', 'e', ' ', 'r', 'e', 'm',
                      'o', 'v', 'a', 'l', ' ', 't', 'e', 'c', 'h', 'n', 'i',
                      'q', 'u', 'e')),
                                      comp[List[Char], List[Char],
    List[Char]](shows_nl, (aa: List[Char]) => indent(x, aa)))))))
                       })
  case (a, ia, i, j, tp, DPTrans(p, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil), List('D', 'P', ' ', 't', 'r', 'a', 'n', 's'),
                       bindb[(List[Char]) => List[Char], C,
                              Unit](catcha[(List[Char]) => List[Char], C,
    (List[Char]) =>
      List[Char]](dependency_pairs_tt[A, lab[B, List[nat]], List[Char],
                                       C](i, j, tp,
   (aa: lab[B, List[nat]]) => Sharp[B, List[nat]](aa), p),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          C](comp[List[Char], List[Char],
                                   List[Char]](ia,
        comp[List[Char], List[Char],
              List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o',
           'r', ' ', 'w', 'h', 'e', 'n', ' ', 's', 'w', 'i', 't', 'c', 'h', 'i',
           'n', 'g', ' ', 'f', 'r', 'o', 'm', ' ', 't', 'h', 'e', ' ', 'T', 'R',
           'S')),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_nl,
      comp[List[Char], List[Char],
            List[Char]](shows_tp[lab[B, List[nat]], List[Char],
                                  A]((aa: lab[B, List[nat]]) =>
                                       shows_prec_lab[B,
               List[nat]](Zero_nat, aa),
                                      (aa: List[Char]) =>
shows_prec_list[Char](Zero_nat, aa),
                                      i, tp),
                         comp[List[Char], List[Char],
                               List[Char]](shows_nl,
    comp[List[Char], List[Char],
          List[Char]](shows_string.apply(List('t', 'o', ' ', 't', 'h', 'e', ' ',
       'i', 'n', 'i', 't', 'i', 'a', 'l', ' ', 'D', 'P', ' ', 'p', 'r', 'o',
       'b', 'l', 'e', 'm', ' ', 'w', 'i', 't', 'h', ' ', 'p', 'a', 'i', 'r',
       's', ' ')),
                       comp[List[Char], List[Char],
                             List[Char]](shows_nl,
  comp[List[Char], List[Char],
        List[Char]](shows_rules[lab[B, List[nat]],
                                 List[Char]]((aa: lab[B, List[nat]]) =>
       shows_prec_lab[B, List[nat]](Zero_nat, aa),
      (aa: List[Char]) => shows_prec_list[Char](Zero_nat, aa),
      List(' ', '-', '>', ' '), p),
                     comp[List[Char], List[Char],
                           List[Char]](shows_nl, x))))))))))),
                                     (dpp: C) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](check_dp_proof[C, B](a,
   comp[List[Char], List[Char],
         List[Char]](ia, shows_string.apply(List('.', '1'))),
   j, dpp, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](ia, comp[List[Char], List[Char],
                       List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                    'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 's', 'w',
                    'i', 't', 'c', 'h', ' ', 't', 'o', ' ', 'd', 'e', 'p', 'e',
                    'n', 'd', 'e', 'n', 'c', 'y', ' ', 'p', 'a', 'i', 'r',
                    's')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl, (aa: List[Char]) => indent(x, aa))))))))
  case (a, ia, i, j, tp, StringReversal(prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('S', 't', 'r', 'i', 'n', 'g', ' ', 'R', 'e', 'v',
                             'e', 'r', 's', 'a', 'l'),
                       bindb[(List[Char]) => List[Char], A,
                              Unit](catcha[(List[Char]) => List[Char], A,
    (List[Char]) =>
      List[Char]](string_reversal_tt[A, lab[B, List[nat]], List[Char]](i, tp),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          A](comp[List[Char], List[Char],
                                   List[Char]](ia,
        comp[List[Char], List[Char],
              List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o',
           'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n',
           'g', ' ', 's', 't', 'r', 'i', 'n', 'g', ' ', 'r', 'e', 'v', 'e', 'r',
           's', 'a', 'l', ' ', 'o', 'n', ' ')),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_nl,
      comp[List[Char], List[Char],
            List[Char]](shows_tp[lab[B, List[nat]], List[Char],
                                  A]((aa: lab[B, List[nat]]) =>
                                       shows_prec_lab[B,
               List[nat]](Zero_nat, aa),
                                      (aa: List[Char]) =>
shows_prec_list[Char](Zero_nat, aa),
                                      i, tp),
                         comp[List[Char], List[Char],
                               List[Char]](shows_nl, x))))))),
                                     (tpa: A) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](check_trs_proof[A, B,
                                      C](a,
  comp[List[Char], List[Char],
        List[Char]](ia, shows_string.apply(List('.', '1'))),
  i, j, tpa, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](ia, comp[List[Char], List[Char],
                       List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                    'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                    'e', ' ', 's', 't', 'r', 'i', 'n', 'g', ' ', 'r', 'e', 'v',
                    'e', 'r', 's', 'a', 'l', ' ', 't', 'e', 'c', 'h', 'n', 'i',
                    'q', 'u', 'e')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl, (aa: List[Char]) => indent(x, aa))))))))
  case (a, ia, i, j, tp, Semlab(sli, lQ, lAll, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil), List('S', 'e', 'm', 'l', 'a', 'b'),
                       bindb[(List[Char]) => List[Char], A,
                              Unit](catcha[(List[Char]) => List[Char], A,
    (List[Char]) =>
      List[Char]]((semlab_fin_tt[A, B,
                                  List[Char]](i,
       sli)).apply(lQ).apply(lAll).apply(tp),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          A](comp[List[Char], List[Char],
                                   List[Char]](ia,
        comp[List[Char], List[Char],
              List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o',
           'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n',
           'g', ' ', 's', 'e', 'm', 'a', 'n', 't', 'i', 'c', ' ', 'l', 'a', 'b',
           'e', 'l', 'l', 'i', 'n', 'g', ' ', 'o', 'n', ' ')),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_nl,
      comp[List[Char], List[Char],
            List[Char]](shows_tp[lab[B, List[nat]], List[Char],
                                  A]((aa: lab[B, List[nat]]) =>
                                       shows_prec_lab[B,
               List[nat]](Zero_nat, aa),
                                      (aa: List[Char]) =>
shows_prec_list[Char](Zero_nat, aa),
                                      i, tp),
                         comp[List[Char], List[Char],
                               List[Char]](shows_nl, x))))))),
                                     (tpa: A) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](check_trs_proof[A, B,
                                      C](a,
  comp[List[Char], List[Char],
        List[Char]](ia, shows_string.apply(List('.', '1'))),
  i, j, tpa, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](ia, comp[List[Char], List[Char],
                       List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                    'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                    'e', ' ', 's', 'e', 'm', '.', 'l', 'a', 'b', ' ', 't', 'e',
                    'c', 'h', 'n', 'i', 'q', 'u', 'e')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl, (aa: List[Char]) => indent(x, aa))))))))
  case (a, ia, i, j, tp, Bounds(info)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil), List('B', 'o', 'u', 'n', 'd', 's'),
                       catcha[(List[Char]) => List[Char], Unit,
                               (List[Char]) =>
                                 List[Char]](bounds_tt[A, lab[B, List[nat]],
                List[Char]](i, info, tp),
      (x: (List[Char]) => List[Char]) =>
        Inl[(List[Char]) => List[Char],
             Unit](comp[List[Char], List[Char],
                         List[Char]](ia, comp[List[Char], List[Char],
       List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o', 'r',
    ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
    'b', 'o', 'u', 'n', 'd', 's', ' ', 'o', 'n', ' ', 't', 'h', 'e', ' ', 't',
    'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', ' ', 'p', 'r', 'o', 'b',
    'l', 'e', 'm', ' ')),
                    comp[List[Char], List[Char],
                          List[Char]](shows_nl,
                                       comp[List[Char], List[Char],
     List[Char]](shows_tp[lab[B, List[nat]], List[Char],
                           A]((aa: lab[B, List[nat]]) =>
                                shows_prec_lab[B, List[nat]](Zero_nat, aa),
                               (aa: List[Char]) =>
                                 shows_prec_list[Char](Zero_nat, aa),
                               i, tp),
                  comp[List[Char], List[Char], List[Char]](shows_nl, x))))))))
  case (a, ia, i, j, tp, Uncurry(u_info, rR, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil), List('U', 'n', 'c', 'u', 'r', 'r', 'y'),
                       bindb[(List[Char]) => List[Char], A,
                              Unit](catcha[(List[Char]) => List[Char], A,
    (List[Char]) =>
      List[Char]](uncurry_tt[A, lab[B, List[nat]]](i, u_info, rR, tp),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          A](comp[List[Char], List[Char],
                                   List[Char]](ia,
        comp[List[Char], List[Char],
              List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o',
           'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n',
           'g', ' ', 'c', 'u', 'r', 'r', 'y', 'i', 'n', 'g', ' ', 't', 'r', 'a',
           'n', 's', 'f', 'o', 'r', 'm', 'a', 't', 'i', 'o', 'n')),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_nl, x))))),
                                     (tpa: A) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](check_trs_proof[A, B,
                                      C](a,
  comp[List[Char], List[Char],
        List[Char]](ia, shows_string.apply(List('.', '1'))),
  i, j, tpa, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](ia, comp[List[Char], List[Char],
                       List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                    'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                    'e', ' ', 'u', 'n', 'c', 'u', 'r', 'r', 'y', 'i', 'n', 'g',
                    ' ', 't', 'e', 'c', 'h', 'n', 'i', 'q', 'u', 'e')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl, (aa: List[Char]) => indent(x, aa))))))))
  case (a, ia, i, j, tp, Fcc(fcs, r, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil), List('F', 'c', 'c'),
                       bindb[(List[Char]) => List[Char], A,
                              Unit](catcha[(List[Char]) => List[Char], A,
    (List[Char]) =>
      List[Char]](fcc_tt[A, lab[B, List[nat]], List[Char]](i, fcs, r, tp),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          A](comp[List[Char], List[Char],
                                   List[Char]](ia,
        comp[List[Char], List[Char],
              List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o',
           'r', ' ', 'w', 'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n',
           'g', ' ', 'f', 'l', 'a', 't', ' ', 'c', 'o', 'n', 't', 'e', 'x', 't',
           ' ', 'c', 'l', 'o', 's', 'u', 'r', 'e')),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_nl, x))))),
                                     (tpa: A) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](check_trs_proof[A, B,
                                      C](a,
  comp[List[Char], List[Char],
        List[Char]](ia, shows_string.apply(List('.', '1'))),
  i, j, tpa, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](ia, comp[List[Char], List[Char],
                       List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                    'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 'f', 'l',
                    'a', 't', ' ', 'c', 'o', 'n', 't', 'e', 'x', 't', ' ', 'c',
                    'l', 'o', 's', 'u', 'r', 'e')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl, (aa: List[Char]) => indent(x, aa))))))))
  case (a, ia, i, j, tp, Split(rrem, prf1, prf2)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil), List('S', 'p', 'l', 'i', 't'),
                       {
                         val (tp1, tp2): (A, A) =
                           split_tt[A, lab[B, List[nat]],
                                     List[Char]](i, tp, rrem);
                         bindb[(List[Char]) => List[Char], Unit,
                                Unit](catcha[(List[Char]) => List[Char], Unit,
      (List[Char]) =>
        List[Char]](check_trs_proof[A, B,
                                     C](a,
 comp[List[Char], List[Char],
       List[Char]](ia, shows_string.apply(List('.', '1'))),
 i, j, tp1, prf1),
                     (x: (List[Char]) => List[Char]) =>
                       Inl[(List[Char]) => List[Char],
                            Unit](comp[List[Char], List[Char],
List[Char]](ia, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                   'e', ' ', 's', 'p', 'l', 'i', 't', 't', 'i', 'n', 'g')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (aa: List[Char]) => indent(x, aa)))))),
                                       (_: Unit) =>
 catcha[(List[Char]) => List[Char], Unit,
         (List[Char]) =>
           List[Char]](check_trs_proof[A, B,
C](a, comp[List[Char], List[Char],
            List[Char]](ia, shows_string.apply(List('.', '2'))),
    i, j, tp2, prf2),
                        (x: (List[Char]) => List[Char]) =>
                          Inl[(List[Char]) => List[Char],
                               Unit](comp[List[Char], List[Char],
   List[Char]](ia, comp[List[Char], List[Char],
                         List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                      'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't',
                      'h', 'e', ' ', 's', 'p', 'l', 'i', 't', 't', 'i', 'n',
                      'g')),
                                      comp[List[Char], List[Char],
    List[Char]](shows_nl, (aa: List[Char]) => indent(x, aa)))))))
                       })
  case (a, ia, i, j, tp, SwitchInnermost(joins, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('S', 'w', 'i', 't', 'c', 'h', ' ', 'I', 'n', 'n',
                             'e', 'r', 'm', 'o', 's', 't'),
                       bindb[(List[Char]) => List[Char], A,
                              Unit](catcha[(List[Char]) => List[Char], A,
    (List[Char]) =>
      List[Char]](switch_innermost_tt[A, lab[B, List[nat]]](i, joins, tp),
                   (x: (List[Char]) => List[Char]) =>
                     Inl[(List[Char]) => List[Char],
                          A](comp[List[Char], List[Char],
                                   List[Char]](ia,
        comp[List[Char], List[Char],
              List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o',
           'r', ' ', 'w', 'h', 'e', 'n', ' ', 's', 'w', 'i', 't', 'c', 'h', 'i',
           'n', 'g', ' ', 't', 'o', ' ', 'i', 'n', 'n', 'e', 'r', 'm', 'o', 's',
           't', ' ', 'o', 'n', ' ')),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_nl,
      comp[List[Char], List[Char],
            List[Char]](shows_tp[lab[B, List[nat]], List[Char],
                                  A]((aa: lab[B, List[nat]]) =>
                                       shows_prec_lab[B,
               List[nat]](Zero_nat, aa),
                                      (aa: List[Char]) =>
shows_prec_list[Char](Zero_nat, aa),
                                      i, tp),
                         comp[List[Char], List[Char],
                               List[Char]](shows_nl, x))))))),
                                     (tpa: A) =>
                                       catcha[(List[Char]) => List[Char], Unit,
       (List[Char]) =>
         List[Char]](check_trs_proof[A, B,
                                      C](a,
  comp[List[Char], List[Char],
        List[Char]](ia, shows_string.apply(List('.', '1'))),
  i, j, tpa, prf),
                      (x: (List[Char]) => List[Char]) =>
                        Inl[(List[Char]) => List[Char],
                             Unit](comp[List[Char], List[Char],
 List[Char]](ia, comp[List[Char], List[Char],
                       List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                    'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                    'e', ' ', 's', 'w', 'i', 't', 'c', 'h', ' ', 't', 'o', ' ',
                    'i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't')),
                                    comp[List[Char], List[Char],
  List[Char]](shows_nl, (aa: List[Char]) => indent(x, aa))))))))
  case (a, ia, i, j, tpa, AssumeSN(tp)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('T', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o',
                             'n', ' ', 'A', 's', 's', 'u', 'm', 'p', 't', 'i',
                             'o', 'n'),
                       (if (a)
                         catcha[(List[Char]) => List[Char], Unit,
                                 (List[Char]) =>
                                   List[Char]](check_tp_subsumes[A, B,
                          List[nat], List[Char]](i, tp, tpa),
        (x: (List[Char]) => List[Char]) =>
          Inl[(List[Char]) => List[Char],
               Unit](comp[List[Char], List[Char],
                           List[Char]](ia,
comp[List[Char], List[Char],
      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o', 'r',
   ' ', 'i', 'n', ' ', 't', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n',
   ' ', 'a', 's', 's', 'u', 'm', 'p', 't', 'i', 'o', 'n')),
                   comp[List[Char], List[Char],
                         List[Char]](shows_nl,
                                      comp[List[Char], List[Char],
    List[Char]](x, shows_nl))))))
                         else Inl[(List[Char]) => List[Char],
                                   Unit](comp[List[Char], List[Char],
       List[Char]](ia, comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List(':', ' ', 't',
                          'h', 'e', ' ', 'p', 'r', 'o', 'o', 'f', ' ', 'c', 'o',
                          'n', 't', 'a', 'i', 'n', 's', ' ', 'a', ' ', 'l', 'o',
                          'c', 'a', 'l', ' ', 't', 'e', 'r', 'm', 'i', 'n', 'a',
                          't', 'i', 'o', 'n', ' ', 'a', 's', 's', 'u', 'm', 'p',
                          't', 'i', 'o', 'n')),
  shows_nl)))))
}

def check_cr_proof[A, B : equal : linorder : show,
                    C](a: Boolean, ia: (List[Char]) => List[Char],
                        i: tp_ops_ext[A, lab[B, List[nat]], List[Char], Unit],
                        j: dpp_ops_ext[C, lab[B, List[nat]], List[Char], Unit],
                        r: List[(term[lab[B, List[nat]], List[Char]],
                                  term[lab[B, List[nat]], List[Char]])],
                        x5: cr_proof[B, List[nat], List[Char]]):
      sum[(List[Char]) => List[Char], Unit] =
  (a, ia, i, j, r, x5) match {
  case (a, ia, i, j, r, SN_WCR(joins, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil), List('S', 'N', '_', 'W', 'C', 'R'),
                       {
                         val tp: A =
                           (mka[A, lab[B, List[nat]], List[Char],
                                 Unit](i)).apply(Nil).apply(r).apply(Nil);
                         bindb[(List[Char]) => List[Char], Unit,
                                Unit](catcha[(List[Char]) => List[Char], Unit,
      (List[Char]) =>
        List[Char]](check_trs_proof[A, B,
                                     C](a,
 comp[List[Char], List[Char],
       List[Char]](ia, shows_string.apply(List('.', '1'))),
 i, j, tp, prf),
                     (x: (List[Char]) => List[Char]) =>
                       Inl[(List[Char]) => List[Char],
                            Unit](comp[List[Char], List[Char],
List[Char]](ia, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 's', 't',
                   'r', 'o', 'n', 'g', ' ', 'n', 'o', 'r', 'm', 'a', 'l', 'i',
                   'z', 'a', 't', 'i', 'o', 'n', ' ', '+', ' ', 'w', 'c', 'r')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (aa: List[Char]) => indent(x, aa)))))),
                                       (_: Unit) =>
 catcha[(List[Char]) => List[Char], Unit,
         (List[Char]) =>
           List[Char]](check_critical_pairs[lab[B,
         List[nat]]](r, critical_pairs_impl[lab[B, List[nat]]](r, r), joins),
                        (_: (List[Char]) => List[Char]) =>
                          Inl[(List[Char]) => List[Char],
                               Unit](comp[List[Char], List[Char],
   List[Char]](ia, comp[List[Char], List[Char],
                         List[Char]](shows_prec_list[Char](Zero_nat,
                    List('e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ',
                          'p', 'r', 'o', 'v', 'i', 'n', 'g', ' ', 'l', 'o', 'c',
                          'a', 'l', ' ', 'c', 'o', 'n', 'f', 'l', 'u', 'e', 'n',
                          'c', 'e', ' ', 'o', 'f', ' ')),
                                      shows_tp[lab[B, List[nat]], List[Char],
        A]((aa: lab[B, List[nat]]) =>
             shows_prec_lab[B, List[nat]](Zero_nat, aa),
            (aa: List[Char]) => shows_prec_list[Char](Zero_nat, aa), i, tp))))))
                       })
}

def check_subsumption_NF[A : equal : show,
                          B : equal : show](e: List[(term[A, B], term[A, B])],
     r: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[((term[A, B], term[A, B]), (List[Char]) => List[Char]), Unit,
          (List[Char]) =>
            List[Char]](forallm[(term[A, B], term[A, B]),
                                 (List[Char]) =>
                                   List[Char]]((ea: (term[A, B], term[A, B])) =>
         catcha[(List[Char]) => List[Char], Unit,
                 (List[Char]) =>
                   List[Char]](check_join_NF[A,
      B](r, fst[term[A, B], term[A, B]](ea), snd[term[A, B], term[A, B]](ea)),
                                (x: (List[Char]) => List[Char]) =>
                                  Inl[(List[Char]) => List[Char],
                                       Unit](comp[List[Char], List[Char],
           List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd', ' ',
        'n', 'o', 't', ' ', 'j', 'o', 'i', 'n', ' ', 'e', 'q', 'u', 'a', 't',
        'i', 'o', 'n', ' ')),
                        comp[List[Char], List[Char],
                              List[Char]](shows_rule[A,
              B]((a: A) => (b: List[Char]) => shows_prec[A](Zero_nat, a, b),
                  (a: B) => (b: List[Char]) => shows_prec[B](Zero_nat, a, b),
                  List(' ', '=', ' '), ea),
   comp[List[Char], List[Char], List[Char]](shows_nl, x))))),
        e),
                         (x: ((term[A, B], term[A, B]),
                               (List[Char]) => List[Char]))
                           => Inl[(List[Char]) => List[Char],
                                   Unit](snd[(term[A, B], term[A, B]),
      (List[Char]) => List[Char]](x)))

def check_estep[A : equal : show,
                 B : equal : show](e: List[(term[A, B], term[A, B])], p: pos,
                                    rule: (term[A, B], term[A, B]),
                                    l_to_r: Boolean, s: term[A, B],
                                    t: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  bindb[(List[Char]) => List[Char], Unit,
         Unit](check[(List[Char]) =>
                       List[Char]](list_ex[(term[A, B],
     term[A, B])]((a: (term[A, B], term[A, B])) =>
                    eq_rule_mod_vars[A, B](rule, a),
                   e),
                                    comp[List[Char], List[Char],
  List[Char]](shows_rule[A, B]((a: A) =>
                                 (b: List[Char]) =>
                                   shows_prec[A](Zero_nat, a, b),
                                (a: B) =>
                                  (b: List[Char]) =>
                                    shows_prec[B](Zero_nat, a, b),
                                List(' ', '-', '>', ' '), rule),
               comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List(' ', 'i', 's', ' ',
                  'n', 'o', 't', ' ', 'a', 'n', ' ', 'e', 'q', 'u', 'a', 't',
                  'i', 'o', 'n', ' ', 'o', 'f')),
                                  comp[List[Char], List[Char],
List[Char]](shows_nl,
             comp[List[Char], List[Char],
                   List[Char]](shows_trs[A,
  B]((a: A) => (b: List[Char]) => shows_prec[A](Zero_nat, a, b),
      (a: B) => (b: List[Char]) => shows_prec[B](Zero_nat, a, b),
      List('e', 'q', 'u', 'a', 't', 'i', 'o', 'n', 'a', 'l', ' ', 's', 'y', 's',
            't', 'e', 'm', ':'),
      List(' ', '=', ' '), e),
                                shows_nl))))),
                (_: Unit) =>
                  bindb[(List[Char]) => List[Char], Unit,
                         Unit](check[(List[Char]) =>
                                       List[Char]](in_poss[A, B](p, s),
            comp[List[Char], List[Char],
                  List[Char]](shows_prec_pos(Zero_nat, p),
                               comp[List[Char], List[Char],
                                     List[Char]](shows_string.apply(List(' ',
                                  'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ',
                                  'p', 'o', 's', 'i', 't', 'i', 'o', 'n', ' ',
                                  'o', 'f', ' ')),
          comp[List[Char], List[Char],
                List[Char]]((shows_prec_term[A, B](Zero_nat)).apply(s),
                             shows_nl)))),
                                (_: Unit) =>
                                  bindb[(List[Char]) => List[Char], Unit,
 Unit](check[(List[Char]) =>
               List[Char]](in_poss[A, B](p, t),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_prec_pos(Zero_nat, p),
       comp[List[Char], List[Char],
             List[Char]](shows_string.apply(List(' ', 'i', 's', ' ', 'n', 'o',
          't', ' ', 'a', ' ', 'p', 'o', 's', 'i', 't', 'i', 'o', 'n', ' ', 'o',
          'f', ' ')),
                          comp[List[Char], List[Char],
                                List[Char]]((shows_prec_term[A,
                      B](Zero_nat)).apply(t),
     shows_nl)))),
        (_: Unit) =>
          {
            val c: ctxt[A, B] = ctxt_of_pos_term[A, B](p, s)
            val d: ctxt[A, B] = ctxt_of_pos_term[A, B](p, t)
            val u: term[A, B] = subt_at[A, B](s, p)
            val v: term[A, B] = subt_at[A, B](t, p)
            val rrule: (term[A, B], term[A, B]) =
              (if (l_to_r) rule
                else (snd[term[A, B], term[A, B]](rule),
                       fst[term[A, B], term[A, B]](rule)));
            match_list[A, B,
                        B](u, List((u, fst[term[A, B], term[A, B]](rrule)),
                                    (v, snd[term[A, B], term[A, B]](rrule))))
              match {
              case None =>
                Inl[(List[Char]) => List[Char],
                     Unit](comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('t', 'h',
                              'e', ' ', 't', 'e', 'r', 'm', ' ')),
      comp[List[Char], List[Char],
            List[Char]]((shows_prec_term[A, B](Zero_nat)).apply(t),
                         comp[List[Char], List[Char],
                               List[Char]](shows_string.apply(List(' ', 'd',
                            'o', 'e', 's', ' ', 'n', 'o', 't', ' ', 'r', 'e',
                            's', 'u', 'l', 't', ' ', 'f', 'r', 'o', 'm', ' ',
                            'a', ' ', 'p', 'r', 'o', 'p', 'e', 'r', ' ', 'a',
                            'p', 'p', 'l', 'i', 'c', 'a', 't', 'i', 'o', 'n',
                            ' ', 'o', 'f', ' ', 'e', 'q', 'u', 'a', 't', 'i',
                            'o', 'n', ' ')),
    comp[List[Char], List[Char],
          List[Char]](shows_nl,
                       comp[List[Char], List[Char],
                             List[Char]](shows_rule[A,
             B]((a: A) => (b: List[Char]) => shows_prec[A](Zero_nat, a, b),
                 (a: B) => (b: List[Char]) => shows_prec[B](Zero_nat, a, b),
                 List(' ', '=', ' '), rrule),
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(List(' ', 'a', 't', ' ', 'p', 'o', 's',
     'i', 't', 'i', 'o', 'n', ' ')),
                     comp[List[Char], List[Char],
                           List[Char]](shows_prec_pos(Zero_nat, p),
shows_nl))))))))
              case Some(_) =>
                check[(List[Char]) =>
                        List[Char]](equal_ctxt[A, B](c, d),
                                     comp[List[Char], List[Char],
   List[Char]](shows_string.apply(List('t', 'h', 'e', ' ', 't', 'e', 'r', 'm',
' ')),
                comp[List[Char], List[Char],
                      List[Char]]((shows_prec_term[A, B](Zero_nat)).apply(t),
                                   comp[List[Char], List[Char],
 List[Char]](shows_string.apply(List(' ', 'd', 'o', 'e', 's', ' ', 'n', 'o',
                                      't', ' ', 'r', 'e', 's', 'u', 'l', 't',
                                      ' ', 'f', 'r', 'o', 'm', ' ', 'a', ' ',
                                      'p', 'r', 'o', 'p', 'e', 'r', ' ', 'a',
                                      'p', 'p', 'l', 'i', 'c', 'a', 't', 'i',
                                      'o', 'n', ' ', 'o', 'f', ' ', 'e', 'q',
                                      'u', 'a', 't', 'i', 'o', 'n', ' ')),
              comp[List[Char], List[Char],
                    List[Char]](shows_nl,
                                 comp[List[Char], List[Char],
                                       List[Char]](shows_rule[A,
                       B]((a: A) =>
                            (b: List[Char]) => shows_prec[A](Zero_nat, a, b),
                           (a: B) =>
                             (b: List[Char]) => shows_prec[B](Zero_nat, a, b),
                           List(' ', '=', ' '), rrule),
            comp[List[Char], List[Char],
                  List[Char]](shows_string.apply(List(' ', 'a', 't', ' ', 'p',
               'o', 's', 'i', 't', 'i', 'o', 'n', ' ')),
                               comp[List[Char], List[Char],
                                     List[Char]](shows_prec_pos(Zero_nat, p),
          shows_nl))))))))
            }
          })))

def check_conversion[A : equal : show,
                      B : equal : show](e: List[(term[A, B], term[A, B])],
 x1: List[(pos, ((term[A, B], term[A, B]), (Boolean, term[A, B])))],
 s: term[A, B], u: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  (e, x1, s, u) match {
  case (e, Nil, s, u) =>
    check[(List[Char]) =>
            List[Char]](equal_terma[A, B](s, u),
                         comp[List[Char], List[Char],
                               List[Char]](shows_string.apply(List('t', 'h',
                            'e', ' ', 'l', 'a', 's', 't', ' ', 't', 'e', 'r',
                            'm', ' ', 'o', 'f', ' ', 't', 'h', 'e', ' ', 'c',
                            'o', 'n', 'v', 'e', 'r', 's', 'i', 'o', 'n', ' ')),
    comp[List[Char], List[Char],
          List[Char]](shows_nl,
                       comp[List[Char], List[Char],
                             List[Char]]((shows_prec_term[A,
                   B](Zero_nat)).apply(s),
  comp[List[Char], List[Char],
        List[Char]](shows_nl,
                     comp[List[Char], List[Char],
                           List[Char]](shows_string.apply(List('d', 'o', 'e',
                        's', ' ', 'n', 'o', 't', ' ', 'c', 'o', 'r', 'r', 'e',
                        's', 'p', 'o', 'n', 'd', ' ', 't', 'o', ' ', 't', 'h',
                        'e', ' ', 'g', 'o', 'a', 'l', ' ', 't', 'e', 'r', 'm')),
comp[List[Char], List[Char],
      List[Char]](shows_nl,
                   comp[List[Char], List[Char],
                         List[Char]]((shows_prec_term[A, B](Zero_nat)).apply(u),
                                      shows_nl))))))))
  case (e, (p, (r, (l_to_r, t))) :: c, s, u) =>
    bindb[(List[Char]) => List[Char], Unit,
           Unit](check_estep[A, B](e, p, r, l_to_r, s, t),
                  (_: Unit) => check_conversion[A, B](e, c, t, u))
}

def check_subsumption_guided[A : equal : show,
                              B : equal : show](ea:
          List[(term[A, B], term[A, B])],
         e: List[(term[A, B], term[A, B])],
         convs:
           List[((term[A, B], term[A, B]),
                  List[(pos, ((term[A, B], term[A, B]),
                               (Boolean, term[A, B])))])]):
      sum[(List[Char]) => List[Char], Unit] =
  bindb[(List[Char]) => List[Char], Unit,
         Unit](catcha[(term[A, B], term[A, B]), Unit,
                       (List[Char]) =>
                         List[Char]](check_subseteq[(term[A, B],
              term[A, B])](ea, mapa[((term[A, B], term[A, B]),
                                      List[(pos,
     ((term[A, B], term[A, B]), (Boolean, term[A, B])))]),
                                     (term[A, B],
                                       term[A,
     B])]((a: ((term[A, B], term[A, B]),
                List[(pos, ((term[A, B], term[A, B]), (Boolean, term[A, B])))]))
            => fst[(term[A, B], term[A, B]),
                    List[(pos, ((term[A, B], term[A, B]),
                                 (Boolean, term[A, B])))]](a),
           convs)),
                                      (x: (term[A, B], term[A, B])) =>
Inl[(List[Char]) => List[Char],
     Unit](comp[List[Char], List[Char],
                 List[Char]](shows_string.apply(List('c', 'o', 'u', 'l', 'd',
              ' ', 'n', 'o', 't', ' ', 'f', 'i', 'n', 'd', ' ', 'c', 'o', 'n',
              'v', 'e', 'r', 's', 'i', 'o', 'n', ' ', 'f', 'o', 'r', ' ', 'e',
              'q', 'u', 'a', 't', 'i', 'o', 'n', ' ')),
                              shows_rule[A,
  B]((a: A) => (b: List[Char]) => shows_prec[A](Zero_nat, a, b),
      (a: B) => (b: List[Char]) => shows_prec[B](Zero_nat, a, b),
      List(' ', '=', ' '), x)))),
                (_: Unit) =>
                  catcha[(((term[A, B], term[A, B]),
                            List[(pos, ((term[A, B], term[A, B]),
 (Boolean, term[A, B])))]),
                           (List[Char]) => List[Char]),
                          Unit,
                          (List[Char]) =>
                            List[Char]](forallm[((term[A, B], term[A, B]),
          List[(pos, ((term[A, B], term[A, B]), (Boolean, term[A, B])))]),
         (List[Char]) =>
           List[Char]]((a: ((term[A, B], term[A, B]),
                             List[(pos, ((term[A, B], term[A, B]),
  (Boolean, term[A, B])))]))
                         => {
                              val (eaa, seq):
                                    ((term[A, B], term[A, B]),
                                      List[(pos,
     ((term[A, B], term[A, B]), (Boolean, term[A, B])))])
                                = a;
                              catcha[(List[Char]) => List[Char], Unit,
                                      (List[Char]) =>
List[Char]](check_conversion[A, B](e, seq, fst[term[A, B], term[A, B]](eaa),
                                    snd[term[A, B], term[A, B]](eaa)),
             (x: (List[Char]) => List[Char]) =>
               Inl[(List[Char]) => List[Char],
                    Unit](comp[List[Char], List[Char],
                                List[Char]](shows_string.apply(List('p', 'r',
                             'o', 'b', 'l', 'e', 'm', ' ', 'i', 'n', ' ', 'c',
                             'o', 'n', 'v', 'e', 'r', 's', 'i', 'o', 'n', ' ',
                             'f', 'o', 'r', ' ', 'e', 'q', 'u', 'a', 't', 'i',
                             'o', 'n', ' ')),
     comp[List[Char], List[Char],
           List[Char]](shows_rule[A, B]((aa: A) =>
  (b: List[Char]) => shows_prec[A](Zero_nat, aa, b),
 (aa: B) => (b: List[Char]) => shows_prec[B](Zero_nat, aa, b),
 List(' ', '=', ' '), eaa),
                        comp[List[Char], List[Char],
                              List[Char]](shows_nl, x)))))
                            },
                        convs),
 (x: (((term[A, B], term[A, B]),
        List[(pos, ((term[A, B], term[A, B]), (Boolean, term[A, B])))]),
       (List[Char]) => List[Char]))
   => Inl[(List[Char]) => List[Char],
           Unit](snd[((term[A, B], term[A, B]),
                       List[(pos, ((term[A, B], term[A, B]),
                                    (Boolean, term[A, B])))]),
                      (List[Char]) => List[Char]](x))))

def check_subsumption[A : equal : show,
                       B : equal : show](e: List[(term[A, B], term[A, B])],
  r: List[(term[A, B], term[A, B])],
  convs_o:
    Option[List[((term[A, B], term[A, B]),
                  List[(pos, ((term[A, B], term[A, B]),
                               (Boolean, term[A, B])))])]]):
      sum[(List[Char]) => List[Char], Unit] =
  convs_o match {
    case None => check_subsumption_NF[A, B](e, r)
    case Some(a) => check_subsumption_guided[A, B](e, r, a)
  }

def check_completion_proof[A, B : equal : linorder : show,
                            C](a: Boolean, ia: (List[Char]) => List[Char],
                                i: tp_ops_ext[A, lab[B, List[nat]], List[Char],
       Unit],
                                j: dpp_ops_ext[C, lab[B, List[nat]], List[Char],
        Unit],
                                e: List[(term[lab[B, List[nat]], List[Char]],
  term[lab[B, List[nat]], List[Char]])],
                                r: List[(term[lab[B, List[nat]], List[Char]],
  term[lab[B, List[nat]], List[Char]])],
                                x6: completion_proof[B, List[nat], List[Char]]):
      sum[(List[Char]) => List[Char], Unit] =
  (a, ia, i, j, e, r, x6) match {
  case (a, ia, i, j, e, r, SN_WCR_Eq(joins, prf, conv1, conv2)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('S', 'N', '_', 'W', 'C', 'R', '_', 'E', 'q'),
                       {
                         val tp: A =
                           (mka[A, lab[B, List[nat]], List[Char],
                                 Unit](i)).apply(Nil).apply(r).apply(Nil);
                         bindb[(List[Char]) => List[Char], Unit,
                                Unit](catcha[(List[Char]) => List[Char], Unit,
      (List[Char]) =>
        List[Char]](check_trs_proof[A, B,
                                     C](a,
 comp[List[Char], List[Char],
       List[Char]](ia, shows_string.apply(List('.', '1'))),
 i, j, tp, prf),
                     (x: (List[Char]) => List[Char]) =>
                       Inl[(List[Char]) => List[Char],
                            Unit](comp[List[Char], List[Char],
List[Char]](ia, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'b', 'e', 'l', 'o', 'w', ' ', 's', 't',
                   'r', 'o', 'n', 'g', ' ', 'n', 'o', 'r', 'm', 'a', 'l', 'i',
                   'z', 'a', 't', 'i', 'o', 'n', ' ', '+', ' ', 'w', 'c', 'r')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl, (aa: List[Char]) => indent(x, aa)))))),
                                       (_: Unit) =>
 bindb[(List[Char]) => List[Char], Unit,
        Unit](catcha[(List[Char]) => List[Char], Unit,
                      (List[Char]) =>
                        List[Char]](check_subsumption_guided[lab[B, List[nat]],
                      List[Char]](r, e, conv1),
                                     (x: (List[Char]) => List[Char]) =>
                                       Inl[(List[Char]) => List[Char],
    Unit](comp[List[Char], List[Char],
                List[Char]](ia, comp[List[Char], List[Char],
                                      List[Char]](shows_prec_list[Char](Zero_nat,
                                 List(':', ' ', 'e', 'r', 'r', 'o', 'r', ' ',
                                       'w', 'h', 'e', 'n', ' ', 's', 'h', 'o',
                                       'w', 'i', 'n', 'g', ' ', 't', 'h', 'a',
                                       't', ' ', 'r', 'e', 'w', 'r', 'i', 't',
                                       'e', ' ', 'r', 'e', 'l', 'a', 't', 'i',
                                       'o', 'n', ' ', 'c', 'a', 'n', ' ', 'b',
                                       'e', ' ', 's', 'i', 'm', 'u', 'l', 'a',
                                       't', 'e', 'd', ' ', 'b', 'y', ' ', 'e',
                                       'q', 'u', 'a', 't', 'i', 'o', 'n', 's')),
           (aa: List[Char]) => indent(x, aa))))),
               (_: Unit) =>
                 bindb[(List[Char]) => List[Char], Unit,
                        Unit](catcha[(List[Char]) => List[Char], Unit,
                                      (List[Char]) =>
List[Char]](check_subsumption[lab[B, List[nat]], List[Char]](e, r, conv2),
             (x: (List[Char]) => List[Char]) =>
               Inl[(List[Char]) => List[Char],
                    Unit](comp[List[Char], List[Char],
                                List[Char]](ia,
     comp[List[Char], List[Char],
           List[Char]](shows_prec_list[Char](Zero_nat,
      List(':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 's',
            'h', 'o', 'w', 'i', 'n', 'g', ' ', 't', 'h', 'a', 't', ' ', 'e',
            'q', 'u', 'a', 't', 'i', 'o', 'n', 's', ' ', 'c', 'a', 'n', ' ',
            'b', 'e', ' ', 's', 'i', 'm', 'u', 'l', 'a', 't', 'e', 'd', ' ',
            'b', 'y', ' ', 'r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'y',
            's', 't', 'e', 'm')),
                        (aa: List[Char]) => indent(x, aa))))),
                               (_: Unit) =>
                                 catcha[(List[Char]) => List[Char], Unit,
 (List[Char]) =>
   List[Char]](check_critical_pairs[lab[B,
 List[nat]]](r, critical_pairs_impl[lab[B, List[nat]]](r, r), joins),
                (x: (List[Char]) => List[Char]) =>
                  Inl[(List[Char]) => List[Char],
                       Unit](comp[List[Char], List[Char],
                                   List[Char]](ia,
        comp[List[Char], List[Char],
              List[Char]](shows_prec_list[Char](Zero_nat,
         List(':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ',
               'p', 'r', 'o', 'v', 'i', 'n', 'g', ' ', 'l', 'o', 'c', 'a', 'l',
               ' ', 'c', 'o', 'n', 'f', 'l', 'u', 'e', 'n', 'c', 'e', ' ', 'o',
               'f', ' ')),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_tp[lab[B, List[nat]],
               List[Char],
               A]((aa: lab[B, List[nat]]) =>
                    shows_prec_lab[B, List[nat]](Zero_nat, aa),
                   (aa: List[Char]) => shows_prec_list[Char](Zero_nat, aa), i,
                   tp),
      comp[List[Char], List[Char],
            List[Char]](shows_nl, (aa: List[Char]) => indent(x, aa))))))))))
                       })
}

def contains_var_term[A : equal, B](x: A, xa1: term[B, A]): Boolean = (x, xa1)
  match {
  case (x, Var(y)) => eq[A](x, y)
  case (x, Fun(uu, ts)) =>
    list_ex[term[B, A]]((a: term[B, A]) => contains_var_term[A, B](x, a), ts)
}

def ident_solve_impla[A : equal,
                       B : equal](mu: A => term[B, A], v_incr: List[A],
                                   cps: List[(term[B, A], (term[B, A], nat))],
                                   stn: (term[B, A], (term[B, A], nat))):
      Option[nat] =
  {
    val cp: List[(term[B, A], (term[B, A], nat))] =
      remdups[(term[B, A], (term[B, A], nat))](conflict_pos[A, B](mu, stn));
    (if (list_ex[(term[B, A],
                   (term[B, A],
                     nat))]((a: (term[B, A], (term[B, A], nat))) =>
                              {
                                val (u, (_, _)): (term[B, A], (term[B, A], nat))
                                  = a;
                                ! (is_Var[B, A](u))
                              },
                             cp))
      None
      else (if (list_ex[(term[B, A],
                          (term[B, A],
                            nat))]((a: (term[B, A], (term[B, A], nat))) =>
                                     {
                                       val
 (u, (v, _)): (term[B, A], (term[B, A], nat)) = a;
                                       is_Var[B, A](u) &&
 (! (is_Var[B, A](v)) &&
   (! (membera[A](v_incr, the_Var[B, A](u))) ||
     contains_var_term[A, B](the_Var[B, A](u), v)))
                                     },
                                    cp))
             None
             else (if (list_ex[(term[B, A],
                                 (term[B, A],
                                   nat))]((a: (term[B, A], (term[B, A], nat)))
    => {
         val (u, (v, m)): (term[B, A], (term[B, A], nat)) = a;
         is_Var[B, A](u) &&
           (! (is_Var[B, A](v)) &&
             list_ex[(term[B, A],
                       (term[B, A],
                         nat))]((aa: (term[B, A], (term[B, A], nat))) =>
                                  {
                                    val (ua, (va, ma)):
  (term[B, A], (term[B, A], nat))
                                      = aa;
                                    equal_terma[B, A](ua, u) &&
                                      (! (is_Var[B, A](va)) &&
is_none[A => term[B, A]](mgu[B, A](subst_apply_term[B, A,
             A](v, substpow_impl[A, B](mu, m)),
                                    subst_apply_term[B, A,
              A](va, substpow_impl[A, B](mu, ma)))))
                                  },
                                 cps))
       },
   cp))
                    None
                    else (if (list_ex[(term[B, A],
(term[B, A],
  nat))]((a: (term[B, A], (term[B, A], nat))) =>
           {
             val (u, (v, _)): (term[B, A], (term[B, A], nat)) = a;
             membera[(term[B, A],
                       term[B, A])](mapa[(term[B, A], (term[B, A], nat)),
  (term[B, A],
    term[B, A])]((aa: (term[B, A], (term[B, A], nat))) =>
                   {
                     val (ua, (va, _)): (term[B, A], (term[B, A], nat)) = aa;
                     (ua, va)
                   },
                  cps),
                                     (u, v))
           },
          cp))
                           None
                           else bind[List[nat],
                                      nat](mapM[(term[B, A], (term[B, A], nat)),
         nat]((a: (term[B, A], (term[B, A], nat))) =>
                {
                  val (u, (v, m)): (term[B, A], (term[B, A], nat)) = a;
                  ident_solve_impla[A, B](mu, v_incr, (u, (v, m)) :: cps,
   (subst_apply_term[B, A, A](u, mu), (v, Suc(m))))
                },
               cp),
    (is: List[nat]) =>
      Some[nat](max_list(mapa[nat, nat]((a: nat) => Suc(a), is))))))))
  }

def ident_solve_impl[A : equal, B : equal]:
      (A => term[B, A]) =>
        (List[A]) => ((term[B, A], term[B, A])) => Option[nat] =
  (mu: A => term[B, A]) =>
    (vincr: List[A]) =>
      (a: (term[B, A], term[B, A])) =>
        {
          val (s, t): (term[B, A], term[B, A]) = a;
          ident_solve_impla[A, B](mu, vincr, Nil, (s, (t, Zero_nat)))
        }

def redex_rps_decision[A : equal, B : equal](mu: List[(A, term[B, A])]):
      (List[(term[B, A], term[B, A])]) => sum[(term[B, A], term[B, A]), Unit] =
  {
    val mua: A => term[B, A] = mk_subst[A, B](mu)
    val d: List[(A, term[B, A])] = mk_subst_domain[A, B](mu)
    val v: List[A] = v_incr_impl[A, B](d)
    val simp: (List[(term[B, A], term[B, A])]) =>
                (List[(term[B, A], term[B, A])]) =>
                  Option[(List[(term[B, A], term[B, A])], nat)]
      = simplify_mp[A, B](mu, v)
    val ident: ((term[B, A], term[B, A])) => Option[nat] =
      ident_solve_impl[A, B].apply(mua).apply(v)
    val w: (term[B, A]) => List[A] = w_impl[A, B](d)
    val main: ((term[B, A], term[B, A])) => Boolean =
      (a: (term[B, A], term[B, A])) =>
        redex_decision[A, B, A, B, A, (term[B, A], term[B, A]), B,
                        A](mua, simp, ident, w, a);
    (xs: List[(term[B, A], term[B, A])]) =>
      catcha[((term[B, A], term[B, A]), (term[B, A], term[B, A])), Unit,
              (term[B, A],
                term[B, A])](forallm[(term[B, A], term[B, A]),
                                      (term[B, A],
term[B, A])]((tl: (term[B, A], term[B, A])) =>
               check[(term[B, A], term[B, A])](! (main(tl)), tl),
              xs),
                              (x: ((term[B, A], term[B, A]),
                                    (term[B, A], term[B, A])))
                                => Inl[(term[B, A], term[B, A]),
Unit](snd[(term[B, A], term[B, A]), (term[B, A], term[B, A])](x)))
  }

def check_NF_iteration[A : equal, B : equal](mu: List[(A, term[B, A])]):
      (List[term[B, A]]) =>
        (term[B, A]) => sum[(term[B, A], term[B, A]), Unit] =
  {
    val dec: (List[(term[B, A], term[B, A])]) =>
               sum[(term[B, A], term[B, A]), Unit]
      = redex_rps_decision[A, B](mu);
    (q: List[term[B, A]]) =>
      (t: term[B, A]) =>
        dec(mapa[term[B, A],
                  (term[B, A], term[B, A])]((a: term[B, A]) => (t, a), q))
  }

def check_qrstep_subst[A : equal : show,
                        B : equal : show](cni:
    (term[A, B]) => sum[(term[A, B], term[A, B]), Unit]):
      (List[(term[A, B], term[A, B])]) =>
        pos =>
          ((term[A, B], term[A, B])) =>
            (term[A, B]) =>
              (term[A, B]) => sum[(List[Char]) => List[Char], Unit] =
  {
    val main: (List[(term[A, B], term[A, B])]) =>
                pos =>
                  ((term[A, B], term[A, B])) =>
                    (term[A, B]) =>
                      (term[A, B]) => sum[(List[Char]) => List[Char], Unit]
      = (a: List[(term[A, B], term[A, B])]) =>
          (b: pos) =>
            (c: (term[A, B], term[A, B])) =>
              (d: term[A, B]) =>
                (e: term[A, B]) =>
                  check_prop_rstep[A, B]((t: term[A, B]) =>
   catcha[(term[A, B], term[A, B]), Unit,
           (List[Char]) =>
             List[Char]](cni(t),
                          (_: (term[A, B], term[A, B])) =>
                            Inl[(List[Char]) => List[Char],
                                 Unit](comp[List[Char], List[Char],
     List[Char]]((shows_prec_term[A, B](Zero_nat)).apply(t),
                  shows_prec_list[Char](Zero_nat,
 List(' ', 'm', 'u', ' ', '^', '^', ' ', 'i', ' ', 'i', 's', ' ', 'n', 'o', 't',
       ' ', 'i', 'n', ' ', 'Q', '-', 'n', 'o', 'r', 'm', 'a', 'l', ' ', 'f',
       'o', 'r', 'm', ' ', 'f', 'o', 'r', ' ', 'a', 'l', 'l', ' ', 'i'))))),
  a, b, c, d, e);
    (r: List[(term[A, B], term[A, B])]) =>
      (p: pos) =>
        (ra: (term[A, B], term[A, B])) =>
          (s: term[A, B]) =>
            (t: term[A, B]) =>
              bindb[(List[Char]) => List[Char], Unit,
                     Unit](check[(List[Char]) =>
                                   List[Char]](! (is_Var[A,
                  B](fst[term[A, B], term[A, B]](ra))),
        shows_prec_list[Char](Zero_nat,
                               List('l', 'o', 'o', 'p', ' ', 'c', 'h', 'e', 'c',
                                     'k', ' ', 'r', 'e', 'q', 'u', 'i', 'r',
                                     'e', 's', ' ', 'l', 'h', 's', 's', ' ',
                                     't', 'o', ' ', 'b', 'e', ' ', 'n', 'o',
                                     'n', '-', 'v', 'a', 'r', 'i', 'a', 'b',
                                     'l', 'e'))),
                            (_: Unit) => ((((main(r))(p))(ra))(s))(t))
  }

def check_rqrstep_subst[A : equal : show,
                         B : equal : show](cni:
     (term[A, B]) => sum[(term[A, B], term[A, B]), Unit]):
      (List[(term[A, B], term[A, B])]) =>
        ((term[A, B], term[A, B])) =>
          (term[A, B]) =>
            (term[A, B]) => sum[(List[Char]) => List[Char], Unit] =
  (r: List[(term[A, B], term[A, B])]) =>
    (check_qrstep_subst[A, B](cni)).apply(r).apply(Empty)

def check_qsteps_subst[A : equal : show,
                        B : equal : show](cni:
    (term[A, B]) => sum[(term[A, B], term[A, B]), Unit],
   p: List[(term[A, B], term[A, B])], r: List[(term[A, B], term[A, B])],
   x3: List[(pos, ((term[A, B], term[A, B]), (Boolean, term[A, B])))],
   s: term[A, B], u: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  (cni, p, r, x3, s, u) match {
  case (cni, p, r, Nil, s, u) =>
    check[(List[Char]) =>
            List[Char]](equal_terma[A, B](s, u),
                         comp[List[Char], List[Char],
                               List[Char]](shows_string.apply(List('t', 'h',
                            'e', ' ', 'l', 'a', 's', 't', ' ', 't', 'e', 'r',
                            'm', ' ', 'o', 'f', ' ', 't', 'h', 'e', ' ', 'r',
                            'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'e', 'q',
                            'u', 'e', 'n', 'c', 'e')),
    comp[List[Char], List[Char],
          List[Char]](shows_nl,
                       comp[List[Char], List[Char],
                             List[Char]]((shows_prec_term[A,
                   B](Zero_nat)).apply(s),
  comp[List[Char], List[Char],
        List[Char]](shows_nl,
                     comp[List[Char], List[Char],
                           List[Char]](shows_string.apply(List('d', 'o', 'e',
                        's', ' ', 'n', 'o', 't', ' ', 'c', 'o', 'r', 'r', 'e',
                        's', 'p', 'o', 'n', 'd', ' ', 't', 'o', ' ', 't', 'h',
                        'e', ' ', 'g', 'o', 'a', 'l', ' ', 't', 'e', 'r', 'm')),
comp[List[Char], List[Char],
      List[Char]](shows_nl,
                   comp[List[Char], List[Char],
                         List[Char]]((shows_prec_term[A, B](Zero_nat)).apply(u),
                                      shows_nl))))))))
  case (cni, p, ra, (uu, (r, (true, t))) :: prts, s, u) =>
    bindb[(List[Char]) => List[Char], Unit,
           Unit]((check_rqrstep_subst[A, B](cni)).apply(p).apply(r).apply(s).apply(t),
                  (_: Unit) => check_qsteps_subst[A, B](cni, p, ra, prts, t, u))
  case (cni, pa, ra, (p, (r, (false, t))) :: prts, s, u) =>
    bindb[(List[Char]) => List[Char], Unit,
           Unit]((check_qrstep_subst[A, B](cni)).apply(ra).apply(p).apply(r).apply(s).apply(t),
                  (_: Unit) =>
                    check_qsteps_subst[A, B](cni, pa, ra, prts, t, u))
}

def check_qrsteps_subst[A : equal : show,
                         B : equal : show](cni:
     (term[A, B]) => sum[(term[A, B], term[A, B]), Unit],
    r: List[(term[A, B], term[A, B])],
    prts: List[(pos, ((term[A, B], term[A, B]), term[A, B]))], s: term[A, B],
    u: term[A, B]):
      sum[(List[Char]) => List[Char], Unit] =
  check_qsteps_subst[A, B](cni, Nil, r,
                            mapa[(pos, ((term[A, B], term[A, B]), term[A, B])),
                                  (pos, ((term[A, B], term[A, B]),
  (Boolean,
    term[A, B])))]((a: (pos, ((term[A, B], term[A, B]), term[A, B]))) =>
                     {
                       val (p, (ra, t)):
                             (pos, ((term[A, B], term[A, B]), term[A, B]))
                         = a;
                       (p, (ra, (false, t)))
                     },
                    prts),
                            s, u)

def check_loop[A : equal : show,
                B : equal : show](q: List[term[A, B]],
                                   nf: (term[A, B]) => Boolean, s: term[A, B],
                                   rseq: List[(pos,
        ((term[A, B], term[A, B]), term[A, B]))],
                                   sigma: List[(B, term[A, B])], c: ctxt[A, B],
                                   r: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  bindb[(List[Char]) => List[Char], Unit,
         Unit](check[(List[Char]) =>
                       List[Char]](! (nulla[(pos,
      ((term[A, B], term[A, B]), term[A, B]))](rseq)),
                                    shows_prec_list[Char](Zero_nat,
                   List('r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'e', 'q',
                         'u', 'e', 'n', 'c', 'e', ' ', 'm', 'u', 's', 't', ' ',
                         'b', 'e', ' ', 'n', 'o', 'n', '-', 'e', 'm', 'p', 't',
                         'y'))),
                (_: Unit) =>
                  (if (nulla[term[A, B]](q))
                    check_qrsteps[A, B]((_: term[A, B]) => true, r, rseq, s,
 ctxt_apply[A, B](c, subst_apply_term[A, B, B](s, mk_subst[B, A](sigma))))
                    else check_qrsteps_subst[A,
      B]((check_NF_iteration[B, A](sigma)).apply(q), r, rseq, s,
          ctxt_apply[A, B](c, subst_apply_term[A, B,
        B](s, mk_subst[B, A](sigma))))))

def check_dp_loop[A : equal : show,
                   B : equal : show](q: List[term[A, B]],
                                      nf: (term[A, B]) => Boolean,
                                      s: term[A, B],
                                      prseq:
List[(pos, ((term[A, B], term[A, B]), (Boolean, term[A, B])))],
                                      sigma: List[(B, term[A, B])],
                                      c: ctxt[A, B],
                                      p: List[(term[A, B], term[A, B])],
                                      r: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  (if (equal_ctxt[A, B](c, Hole))
    bindb[(List[Char]) => List[Char], Unit,
           Unit](check[(List[Char]) =>
                         List[Char]](! (nulla[(pos,
        ((term[A, B], term[A, B]), (Boolean, term[A, B])))](prseq)),
                                      shows_prec_list[Char](Zero_nat,
                     List('r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'e', 'q',
                           'u', 'e', 'n', 'c', 'e', ' ', 'm', 'u', 's', 't',
                           ' ', 'b', 'e', ' ', 'n', 'o', 'n', '-', 'e', 'm',
                           'p', 't', 'y'))),
                  (_: Unit) =>
                    (if (nulla[term[A, B]](q))
                      check_qsteps[A, B](nf, p, r, prseq, s,
  subst_apply_term[A, B, B](s, mk_subst[B, A](sigma)))
                      else check_qsteps_subst[A,
       B]((check_NF_iteration[B, A](sigma)).apply(q), p, r, prseq, s,
           subst_apply_term[A, B, B](s, mk_subst[B, A](sigma)))))
    else check_loop[A, B](q, nf, s,
                           mapa[(pos, ((term[A, B], term[A, B]),
(Boolean, term[A, B]))),
                                 (pos, ((term[A, B], term[A, B]),
 term[A, B]))]((a: (pos, ((term[A, B], term[A, B]), (Boolean, term[A, B])))) =>
                 {
                   val (x, (y, (_, z))):
                         (pos, ((term[A, B], term[A, B]),
                                 (Boolean, term[A, B])))
                     = a;
                   (x, (y, z))
                 },
                prseq),
                           sigma, c, r))

def check_dp_disproof[A, B : equal : show,
                       C : equal : show](ia: (List[Char]) => List[Char],
  i: dpp_ops_ext[A, B, C, Unit], dpp: A, x3: dp_disproof[B, C]):
      sum[(List[Char]) => List[Char], Unit] =
  (ia, i, dpp, x3) match {
  case (ia, i, dpp, DP_Loop(s, rseq, sigma, c)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil), List('L', 'o', 'o', 'p'),
                       {
                         val p: List[(term[B, C], term[B, C])] =
                           (pairsa[A, B, C, Unit](i)).apply(dpp)
                         val r: List[(term[B, C], term[B, C])] =
                           (rulesd[A, B, C, Unit](i)).apply(dpp)
                         val nf: (term[B, C]) => Boolean =
                           (is_QNFc[A, B, C, Unit](i)).apply(dpp)
                         val q: List[term[B, C]] =
                           (qc[A, B, C, Unit](i)).apply(dpp);
                         check_dp_loop[B, C](q, nf, s, rseq, sigma, c, p, r)
                       })
  case (ia, i, dpp, DP_Rule_Removal(p, r, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('R', 'u', 'l', 'e', ' ', 'R', 'e', 'm', 'o', 'v',
                             'a', 'l'),
                       {
                         val pa: List[(term[B, C], term[B, C])] =
                           (pairsa[A, B, C, Unit](i)).apply(dpp)
                         val prm: List[(term[B, C], term[B, C])] =
                           list_diff[(term[B, C], term[B, C])](pa, p)
                         val ra: List[(term[B, C], term[B, C])] =
                           (rulesd[A, B, C, Unit](i)).apply(dpp)
                         val rrm: List[(term[B, C], term[B, C])] =
                           list_diff[(term[B, C], term[B, C])](ra, r);
                         check_dp_disproof[A, B,
    C](comp[List[Char], List[Char],
             List[Char]](ia, shows_string.apply(List('.', '1'))),
        i, (delete_R_Rwc[A, B, C,
                          Unit](i)).apply((delete_P_Pwa[A, B, C,
                 Unit](i)).apply(dpp).apply(prm).apply(prm)).apply(rrm).apply(rrm),
        prf)
                       })
}

def unsharp[A, B](x0: lab[A, B]): lab[A, B] = x0 match {
  case Sharp(f) => f
  case Lab(v, va) => Lab[A, B](v, va)
  case FunLab(v, va) => FunLab[A, B](v, va)
  case UnLab(v) => UnLab[A, B](v)
}

def check_dps[A : equal : show,
               B : equal : show](unshp: A => A,
                                  r: List[(term[A, B], term[A, B])],
                                  p: List[(term[A, B], term[A, B])]):
      sum[(term[A, B], term[A, B]), Unit] =
  {
    val d: List[(A, nat)] = defs_list[A, B](r);
    catcha[((term[A, B], term[A, B]), (term[A, B], term[A, B])), Unit,
            (term[A, B],
              term[A, B])](forallm[(term[A, B], term[A, B]),
                                    (term[A, B],
                                      term[A,
    B])]((x: (term[A, B], term[A, B])) =>
           (if ({
                  val (ll, rr): (term[A, B], term[A, B]) = x;
                  ! (is_Var[A, B](rr)) &&
                    (! (membera[(A, nat)](d, the[(A, nat)](root[A, B](rr)))) &&
                      list_ex[(term[A, B],
                                term[A, B])]((a: (term[A, B], term[A, B])) =>
       {
         val (l, ra): (term[A, B], term[A, B]) = a;
         equal_terma[A, B](l, sharp_term[A, B](unshp, ll)) &&
           member[(term[A, B],
                    term[A, B])]((ra, sharp_term[A, B](unshp, rr)),
                                  supteq[A, B])
       },
      r))
                })
             Inr[Unit, (term[A, B], term[A, B])](())
             else Inl[(term[A, B], term[A, B]), Unit](x)),
          p),
                            (x: ((term[A, B], term[A, B]),
                                  (term[A, B], term[A, B])))
                              => Inl[(term[A, B], term[A, B]),
                                      Unit](snd[(term[A, B], term[A, B]),
         (term[A, B], term[A, B])](x)))
  }

def q_emptya[A, B, C, D](x0: tp_ops_ext[A, B, C, D]): A => Boolean = x0 match {
  case Tp_ops_ext(qreltrs, q, r, rw, rules, q_empty, is_QNF,
                   nFQ_subset_NF_rules, rules_map, delete_R_Rw, split_rules, mk,
                   more)
    => q_empty
}

def useful_rule_impl[A, B](nf: (term[A, B]) => Boolean,
                            r: (term[A, B], term[A, B])):
      Boolean =
  list_all[term[A, B]](nf, args[A, B](fst[term[A, B], term[A, B]](r)))

def check_wwf_qtrs[A : show,
                    B : equal : show](nf: (term[A, B]) => Boolean,
                                       r: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  catcha[(List[Char]) => List[Char], Unit,
          (List[Char]) =>
            List[Char]](catcha[((term[A, B], term[A, B]),
                                 (List[Char]) => List[Char]),
                                Unit,
                                (List[Char]) =>
                                  List[Char]](forallm[(term[A, B], term[A, B]),
               (List[Char]) =>
                 List[Char]]((ra: (term[A, B], term[A, B])) =>
                               (if (useful_rule_impl[A, B](nf, ra))
                                 catcha[(List[Char]) => List[Char], Unit,
 (List[Char]) =>
   List[Char]](bindb[(List[Char]) => List[Char], Unit,
                      Unit](check[(List[Char]) =>
                                    List[Char]](! (is_Var[A,
                   B](fst[term[A, B], term[A, B]](ra))),
         shows_string.apply(List('v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ',
                                  'l', 'e', 'f', 't', '-', 'h', 'a', 'n', 'd',
                                  ' ', 's', 'i', 'd', 'e', ' ', 'i', 'n'))),
                             (_: Unit) =>
                               catcha[B, Unit,
                                       (List[Char]) =>
 List[Char]](check_subseteq[B](vars_term_list[A,
       B].apply(snd[term[A, B], term[A, B]](ra)),
                                vars_term_list[A,
        B].apply(fst[term[A, B], term[A, B]](ra))),
              (x: B) =>
                Inl[(List[Char]) => List[Char],
                     Unit](comp[List[Char], List[Char],
                                 List[Char]](shows_string.apply(List('f', 'r',
                              'e', 'e', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l',
                              'e', ' ')),
      comp[List[Char], List[Char],
            List[Char]]((a: List[Char]) => shows_prec[B](Zero_nat, x, a),
                         shows_string.apply(List(' ', 'i', 'n', ' ', 'r', 'i',
          'g', 'h', 't', '-', 'h', 'a', 'n', 'd', ' ', 's', 'i', 'd', 'e', ' ',
          'o', 'f'))))))),
                (x: (List[Char]) => List[Char]) =>
                  Inl[(List[Char]) => List[Char],
                       Unit](comp[List[Char], List[Char],
                                   List[Char]](x,
        comp[List[Char], List[Char],
              List[Char]](shows_string.apply(List(' ', 'r', 'u', 'l', 'e',
           ' ')),
                           comp[List[Char], List[Char],
                                 List[Char]](shows_rule[A,
                 B]((a: A) => (b: List[Char]) => shows_prec[A](Zero_nat, a, b),
                     (a: B) => (b: List[Char]) => shows_prec[B](Zero_nat, a, b),
                     List(' ', '-', '>', ' '), ra),
      shows_nl)))))
                                 else Inr[Unit,
   (List[Char]) => List[Char]](())),
                              r),
       (x: ((term[A, B], term[A, B]), (List[Char]) => List[Char])) =>
         Inl[(List[Char]) => List[Char],
              Unit](snd[(term[A, B], term[A, B]),
                         (List[Char]) => List[Char]](x))),
                         (x: (List[Char]) => List[Char]) =>
                           Inl[(List[Char]) => List[Char],
                                Unit](comp[List[Char], List[Char],
    List[Char]](shows_string.apply(List('t', 'h', 'e', ' ', 'Q', '-', 'T', 'R',
 'S', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'w', 'e', 'a', 'k', 'l', 'y', ' ',
 'w', 'e', 'l', 'l', '-', 'f', 'o', 'r', 'm', 'e', 'd')),
                 comp[List[Char], List[Char], List[Char]](shows_nl, x))))

def check_not_wwf_qtrs[A : show,
                        B : equal : show](nf: (term[A, B]) => Boolean,
   r: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  check[(List[Char]) =>
          List[Char]](! (isOK[(List[Char]) => List[Char],
                               Unit](check_wwf_qtrs[A, B](nf, r))),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('T', 'h', 'e',
                          ' ', 'Q', '-', 'T', 'R', 'S', ' ', 'i', 's', ' ', 'w',
                          'e', 'l', 'l', ' ', 'f', 'o', 'r', 'm', 'e', 'd')),
  shows_nl))

def check_trs_disproof[A, B : equal : show, C : equal : show, D : equal : show,
                        E](ia: (List[Char]) => List[Char],
                            i: tp_ops_ext[A, lab[B, C], D, Unit],
                            j: dpp_ops_ext[E, lab[B, C], D, Unit], tp: A,
                            x4: trs_disproof[lab[B, C], D]):
      sum[(List[Char]) => List[Char], Unit] =
  (ia, i, j, tp, x4) match {
  case (ia, i, j, tp, TRS_Loop(s, rseq, sigma, c)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil), List('L', 'o', 'o', 'p'),
                       {
                         val r: List[(term[lab[B, C], D], term[lab[B, C], D])] =
                           (rulesb[A, lab[B, C], D, Unit](i)).apply(tp)
                         val nf: (term[lab[B, C], D]) => Boolean =
                           (is_QNFa[A, lab[B, C], D, Unit](i)).apply(tp)
                         val q: List[term[lab[B, C], D]] =
                           (qa[A, lab[B, C], D, Unit](i)).apply(tp);
                         check_loop[lab[B, C], D](q, nf, s, rseq, sigma, c, r)
                       })
  case (ia, i, j, tp, TRS_Not_Well_Formed) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('N', 'o', 't', ' ', 'W', 'e', 'l', 'l', '-', 'F',
                             'o', 'r', 'm', 'e', 'd'),
                       catcha[(List[Char]) => List[Char], Unit,
                               (List[Char]) =>
                                 List[Char]](check_not_wwf_qtrs[lab[B, C],
                         D]((is_QNFa[A, lab[B, C], D, Unit](i)).apply(tp),
                             (rulesb[A, lab[B, C], D, Unit](i)).apply(tp)),
      (x: (List[Char]) => List[Char]) =>
        Inl[(List[Char]) => List[Char],
             Unit](comp[List[Char], List[Char],
                         List[Char]](ia, comp[List[Char], List[Char],
       List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o', 'r',
    ' ', 'i', 'n', ' ', 'n', 'o', 't', ' ', 'w', 'e', 'l', 'l', '-', 'f', 'o',
    'r', 'm', 'e', 'd', ' ', 'p', 'r', 'o', 'o', 'f')),
                    comp[List[Char], List[Char], List[Char]](shows_nl, x))))))
  case (ia, i, j, tp, TRS_Rule_Removal(r, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('R', 'u', 'l', 'e', ' ', 'R', 'e', 'm', 'o', 'v',
                             'a', 'l'),
                       {
                         val ra: List[(term[lab[B, C], D], term[lab[B, C], D])]
                           = (rulesb[A, lab[B, C], D, Unit](i)).apply(tp)
                         val rrm: List[(term[lab[B, C], D], term[lab[B, C], D])]
                           = list_diff[(term[lab[B, C], D],
 term[lab[B, C], D])](ra, r);
                         check_trs_disproof[A, B, C, D,
     E](comp[List[Char], List[Char],
              List[Char]](ia, shows_string.apply(List('.', '1'))),
         i, j,
         (delete_R_Rwa[A, lab[B, C], D,
                        Unit](i)).apply(tp).apply(rrm).apply(rrm),
         prf)
                       })
  case (ia, i, j, tp, TRS_String_Reversal(prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('S', 't', 'r', 'i', 'n', 'g', ' ', 'R', 'e', 'v',
                             'e', 'r', 's', 'a', 'l'),
                       {
                         val r: List[(term[lab[B, C], D], term[lab[B, C], D])] =
                           (rulesb[A, lab[B, C], D, Unit](i)).apply(tp);
                         bindb[(List[Char]) => List[Char], Unit,
                                Unit](check[(List[Char]) =>
      List[Char]]((q_emptya[A, lab[B, C], D, Unit](i)).apply(tp),
                   shows_prec_list[Char](Zero_nat,
  List('Q', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'e', 'm', 'p', 't', 'y'))),
                                       (_: Unit) =>
 bindb[(List[Char]) => List[Char], A,
        Unit](catcha[(List[Char]) => List[Char], A,
                      (List[Char]) =>
                        List[Char]](check_unary_signature[lab[B, C], D](r) match
                                      {
                                      case Inl(a) =>
Inl[(List[Char]) => List[Char], A](a)
                                      case Inr(_) =>
Inr[A, (List[Char]) =>
         List[Char]]((mka[A, lab[B, C], D,
                           Unit](i)).apply(Nil).apply(mapa[(term[lab[B, C], D],
                     term[lab[B, C], D]),
                    (term[lab[B, C], D],
                      term[lab[B, C],
                            D])]((a: (term[lab[B, C], D], term[lab[B, C], D]))
                                   => rev_rule[lab[B, C], D](a),
                                  r)).apply(Nil))
                                    },
                                     (x: (List[Char]) => List[Char]) =>
                                       Inl[(List[Char]) => List[Char],
    A](comp[List[Char], List[Char],
             List[Char]](ia, comp[List[Char], List[Char],
                                   List[Char]](shows_string.apply(List(':', ' ',
                                'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                                'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n',
                                'g', ' ', 's', 't', 'r', 'i', 'n', 'g', ' ',
                                'r', 'e', 'v', 'e', 'r', 's', 'a', 'l', ' ',
                                'o', 'n')),
        comp[List[Char], List[Char],
              List[Char]](shows_nl,
                           comp[List[Char], List[Char],
                                 List[Char]](shows_tp[lab[B, C], D,
               A]((a: lab[B, C]) => shows_prec_lab[B, C](Zero_nat, a),
                   (a: D) => (b: List[Char]) => shows_prec[D](Zero_nat, a, b),
                   i, tp),
      comp[List[Char], List[Char], List[Char]](shows_nl, x))))))),
               (tpa: A) =>
                 catcha[(List[Char]) => List[Char], Unit,
                         (List[Char]) =>
                           List[Char]](check_trs_disproof[A, B, C, D,
                   E](comp[List[Char], List[Char],
                            List[Char]](ia, shows_string.apply(List('.', '1'))),
                       i, j, tpa, prf),
(x: (List[Char]) => List[Char]) =>
  Inl[(List[Char]) => List[Char],
       Unit](comp[List[Char], List[Char],
                   List[Char]](ia, comp[List[Char], List[Char],
 List[Char]](shows_string.apply(List(':', ' ', 'e', 'r', 'r', 'o', 'r', ' ',
                                      'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                                      'e', ' ', 's', 't', 'r', 'i', 'n', 'g',
                                      ' ', 'r', 'e', 'v', 'e', 'r', 's', 'a',
                                      'l', ' ', 't', 'e', 'c', 'h', 'n', 'i',
                                      'q', 'u', 'e')),
              comp[List[Char], List[Char],
                    List[Char]](shows_nl, (a: List[Char]) => indent(x, a))))))))
                       })
  case (ia, i, j, tp, TRS_DP_Trans(p, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('D', 'P', ' ', 'T', 'r', 'a', 'n', 's', 'f', 'o',
                             'r', 'm', 'a', 't', 'i', 'o', 'n'),
                       {
                         val r: List[(term[lab[B, C], D], term[lab[B, C], D])] =
                           (rulesb[A, lab[B, C], D, Unit](i)).apply(tp)
                         val q: List[term[lab[B, C], D]] =
                           (qa[A, lab[B, C], D, Unit](i)).apply(tp);
                         bindb[(List[Char]) => List[Char], Unit,
                                Unit](catcha[(term[lab[B, C], D],
       term[lab[B, C], D]),
      Unit,
      (List[Char]) =>
        List[Char]](check_dps[lab[B, C],
                               D]((a: lab[B, C]) => unsharp[B, C](a), r, p),
                     (x: (term[lab[B, C], D], term[lab[B, C], D])) =>
                       Inl[(List[Char]) => List[Char],
                            Unit](comp[List[Char], List[Char],
List[Char]](ia, comp[List[Char], List[Char],
                      List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                   'r', 'o', 'r', ' ', 'i', 'n', ' ', 'D', 'P', ' ', 't', 'r',
                   'a', 'n', 's', 'f', 'o', 'r', 'm', 'a', 't', 'i', 'o', 'n')),
                                   comp[List[Char], List[Char],
 List[Char]](shows_nl,
              comp[List[Char], List[Char],
                    List[Char]](shows_prec_list[Char](Zero_nat,
               List('p', 'r', 'o', 'b', 'l', 'e', 'm', 'a', 't', 'i', 'c', ' ',
                     'r', 'u', 'l', 'e', ' ')),
                                 shows_rule[lab[B, C],
     D]((a: lab[B, C]) => shows_prec_lab[B, C](Zero_nat, a),
         (a: D) => (b: List[Char]) => shows_prec[D](Zero_nat, a, b),
         List(' ', '-', '>', ' '), x))))))),
                                       (_: Unit) =>
 {
   val dpp: E =
     (mkc[E, lab[B, C], D,
           Unit](j)).apply(p).apply(Nil).apply(q).apply(r).apply(Nil);
   check_dp_disproof[E, lab[B, C],
                      D](comp[List[Char], List[Char],
                               List[Char]](ia,
    shows_string.apply(List('.', '1'))),
                          j, dpp, prf)
 })
                       })
}

def check_rel_seq[A : equal : show,
                   B : equal : show](r: List[(term[A, B], term[A, B])],
                                      sa: List[(term[A, B], term[A, B])],
                                      x2:
List[(pos, ((term[A, B], term[A, B]), (Boolean, term[A, B])))],
                                      s: term[A, B], u: term[A, B], b: Boolean):
      sum[(List[Char]) => List[Char], Unit] =
  (r, sa, x2, s, u, b) match {
  case (r, sa, Nil, s, u, true) =>
    check[(List[Char]) =>
            List[Char]](equal_terma[A, B](s, u),
                         comp[List[Char], List[Char],
                               List[Char]](shows_string.apply(List('t', 'h',
                            'e', ' ', 'l', 'a', 's', 't', ' ', 't', 'e', 'r',
                            'm', ' ', 'o', 'f', ' ', 't', 'h', 'e', ' ', 'r',
                            'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'e', 'q',
                            'u', 'e', 'n', 'c', 'e')),
    comp[List[Char], List[Char],
          List[Char]](shows_nl,
                       comp[List[Char], List[Char],
                             List[Char]]((shows_prec_term[A,
                   B](Zero_nat)).apply(s),
  comp[List[Char], List[Char],
        List[Char]](shows_string.apply(shows_nl.apply(List('d', 'o', 'e', 's',
                    ' ', 'n', 'o', 't', ' ', 'c', 'o', 'r', 'r', 'e', 's', 'p',
                    'o', 'n', 'd', ' ', 't', 'o', ' ', 't', 'h', 'e', ' ', 'g',
                    'o', 'a', 'l', ' ', 't', 'e', 'r', 'm'))),
                     comp[List[Char], List[Char],
                           List[Char]](shows_nl,
comp[List[Char], List[Char],
      List[Char]]((shows_prec_term[A, B](Zero_nat)).apply(u), shows_nl)))))))
  case (r, sa, Nil, s, u, false) =>
    Inl[(List[Char]) => List[Char],
         Unit](shows_prec_list[Char](Zero_nat,
                                      List('d', 'i', 'd', ' ', 'n', 'o', 't',
    ' ', 'f', 'i', 'n', 'd', ' ', 's', 't', 'r', 'i', 'c', 't', ' ', 's', 't',
    'e', 'p', ' ', 'i', 'n', ' ', 'r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's',
    'e', 'q', 'u', 'e', 'n', 'c', 'e')))
  case (ra, sa, (p, (r, (true, t))) :: prts, s, u, b) =>
    bindb[(List[Char]) => List[Char], Unit,
           Unit]((check_qrstep[A, B]((_: term[A, B]) =>
                                       true)).apply(ra).apply(p).apply(r).apply(s).apply(t),
                  (_: Unit) => check_rel_seq[A, B](ra, sa, prts, t, u, true))
  case (ra, sa, (p, (r, (false, t))) :: prts, s, u, b) =>
    bindb[(List[Char]) => List[Char], Unit,
           Unit]((check_qrstep[A, B]((_: term[A, B]) =>
                                       true)).apply(sa).apply(p).apply(r).apply(s).apply(t),
                  (_: Unit) => check_rel_seq[A, B](ra, sa, prts, t, u, b))
}

def check_rel_loop[A : equal : show,
                    B : equal : show](sa: term[A, B],
                                       rseq:
 List[(pos, ((term[A, B], term[A, B]), (Boolean, term[A, B])))],
                                       sigma: List[(B, term[A, B])],
                                       c: ctxt[A, B],
                                       r: List[(term[A, B], term[A, B])],
                                       s: List[(term[A, B], term[A, B])]):
      sum[(List[Char]) => List[Char], Unit] =
  check_rel_seq[A, B](r, s, rseq, sa,
                       ctxt_apply[A, B](c,
 subst_apply_term[A, B, B](sa, mk_subst[B, A](sigma))),
                       false)

def check_wf_reltrs[A : show,
                     B : equal : show](x0:
 (List[(term[A, B], term[A, B])], List[(term[A, B], term[A, B])])):
      sum[(List[Char]) => List[Char], Unit] =
  x0 match {
  case (r, s) =>
    bindb[(List[Char]) => List[Char], Unit,
           Unit](check_wf_trs[A, B](r),
                  (_: Unit) =>
                    (if (nulla[(term[A, B], term[A, B])](r))
                      Inr[Unit, (List[Char]) => List[Char]](())
                      else catcha[((term[A, B], term[A, B]),
                                    (List[Char]) => List[Char]),
                                   Unit,
                                   (List[Char]) =>
                                     List[Char]](forallm[(term[A, B],
                   term[A, B]),
                  (List[Char]) =>
                    List[Char]]((rule: (term[A, B], term[A, B])) =>
                                  catcha[(List[Char]) => List[Char], Unit,
  (List[Char]) =>
    List[Char]](catcha[B, Unit,
                        (List[Char]) =>
                          List[Char]](check_subseteq[B](vars_term_list[A,
                                B].apply(snd[term[A, B], term[A, B]](rule)),
                 vars_term_list[A, B].apply(fst[term[A, B], term[A, B]](rule))),
                                       (x: B) =>
 Inl[(List[Char]) => List[Char],
      Unit](comp[List[Char], List[Char],
                  List[Char]](shows_string.apply(List('f', 'r', 'e', 'e', ' ',
               'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ')),
                               comp[List[Char], List[Char],
                                     List[Char]]((a: List[Char]) =>
           shows_prec[B](Zero_nat, x, a),
          shows_string.apply(List(' ', 'i', 'n', ' ', 'r', 'i', 'g', 'h', 't',
                                   '-', 'h', 'a', 'n', 'd', ' ', 's', 'i', 'd',
                                   'e', ' ', 'o', 'f')))))),
                 (x: (List[Char]) => List[Char]) =>
                   Inl[(List[Char]) => List[Char],
                        Unit](comp[List[Char], List[Char],
                                    List[Char]](x,
         comp[List[Char], List[Char],
               List[Char]](shows_string.apply(List(' ', 'r', 'u', 'l', 'e',
            ' ')),
                            comp[List[Char], List[Char],
                                  List[Char]](shows_rule[A,
                  B]((a: A) => (b: List[Char]) => shows_prec[A](Zero_nat, a, b),
                      (a: B) =>
                        (b: List[Char]) => shows_prec[B](Zero_nat, a, b),
                      List(' ', '-', '>', ' '), rule),
       shows_nl))))),
                                 s),
          (x: ((term[A, B], term[A, B]), (List[Char]) => List[Char])) =>
            Inl[(List[Char]) => List[Char],
                 Unit](snd[(term[A, B], term[A, B]),
                            (List[Char]) => List[Char]](x)))))
}

def check_not_wf_reltrs[A : show,
                         B : equal : show](rs:
     (List[(term[A, B], term[A, B])], List[(term[A, B], term[A, B])])):
      sum[(List[Char]) => List[Char], Unit] =
  check[(List[Char]) =>
          List[Char]](! (isOK[(List[Char]) => List[Char],
                               Unit](check_wf_reltrs[A, B](rs))),
                       comp[List[Char], List[Char],
                             List[Char]](shows_string.apply(List('T', 'h', 'e',
                          ' ', 'T', 'R', 'S', 's', ' ', 'R', ' ', 'a', 'n', 'd',
                          ' ', 'S', ' ', 'a', 'r', 'e', ' ', 'w', 'e', 'l', 'l',
                          ' ', 'f', 'o', 'r', 'm', 'e', 'd')),
  shows_nl))

def check_reltrs_disproof[A, B : equal : show, C : equal : show,
                           D : equal : show,
                           E](ia: (List[Char]) => List[Char],
                               i: tp_ops_ext[A, lab[B, C], D, Unit],
                               j: dpp_ops_ext[E, lab[B, C], D, Unit], tp: A,
                               x4: reltrs_disproof[lab[B, C], D]):
      sum[(List[Char]) => List[Char], Unit] =
  (ia, i, j, tp, x4) match {
  case (ia, i, j, tp, Rel_Loop(s, rseq, sigma, c)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil), List('L', 'o', 'o', 'p'),
                       {
                         val r: List[(term[lab[B, C], D], term[lab[B, C], D])] =
                           (ra[A, lab[B, C], D, Unit](i)).apply(tp)
                         val sa: List[(term[lab[B, C], D], term[lab[B, C], D])]
                           = (rwa[A, lab[B, C], D, Unit](i)).apply(tp);
                         bindb[(List[Char]) => List[Char], Unit,
                                Unit](check[(List[Char]) =>
      List[Char]]((q_emptya[A, lab[B, C], D, Unit](i)).apply(tp),
                   shows_prec_list[Char](Zero_nat,
  List('Q', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'e', 'm', 'p', 't', 'y'))),
                                       (_: Unit) =>
 catcha[(List[Char]) => List[Char], Unit,
         (List[Char]) =>
           List[Char]](check_rel_loop[lab[B, C], D](s, rseq, sigma, c, r, sa),
                        (x: (List[Char]) => List[Char]) =>
                          Inl[(List[Char]) => List[Char],
                               Unit](comp[List[Char], List[Char],
   List[Char]](ia, comp[List[Char], List[Char],
                         List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                      'r', 'o', 'r', ' ', 'i', 'n', ' ', 'l', 'o', 'o', 'p',
                      ' ', 'p', 'r', 'o', 'o', 'f')),
                                      comp[List[Char], List[Char],
    List[Char]](shows_nl, x))))))
                       })
  case (ia, i, j, tp, Rel_R_Not_SN(prf)) =>
    {
      val q: List[term[lab[B, C], D]] = (qa[A, lab[B, C], D, Unit](i)).apply(tp)
      val r: List[(term[lab[B, C], D], term[lab[B, C], D])] =
        (ra[A, lab[B, C], D, Unit](i)).apply(tp)
      val tpa: A = (mka[A, lab[B, C], D, Unit](i)).apply(q).apply(r).apply(Nil);
      check_trs_disproof[A, B, C, D,
                          E](comp[List[Char], List[Char],
                                   List[Char]](ia,
        shows_string.apply(List('.', '1'))),
                              i, j, tpa, prf)
    }
  case (ia, i, j, tp, Rel_Not_Well_Formed) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('N', 'o', 't', ' ', 'W', 'e', 'l', 'l', '-', 'F',
                             'o', 'r', 'm', 'e', 'd'),
                       {
                         val r: List[(term[lab[B, C], D], term[lab[B, C], D])] =
                           (ra[A, lab[B, C], D, Unit](i)).apply(tp)
                         val s: List[(term[lab[B, C], D], term[lab[B, C], D])] =
                           (rwa[A, lab[B, C], D, Unit](i)).apply(tp);
                         bindb[(List[Char]) => List[Char], Unit,
                                Unit](check[(List[Char]) =>
      List[Char]]((q_emptya[A, lab[B, C], D, Unit](i)).apply(tp),
                   shows_prec_list[Char](Zero_nat,
  List('Q', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'e', 'm', 'p', 't', 'y'))),
                                       (_: Unit) =>
 catcha[(List[Char]) => List[Char], Unit,
         (List[Char]) =>
           List[Char]](check_not_wf_reltrs[lab[B, C], D]((r, s)),
                        (x: (List[Char]) => List[Char]) =>
                          Inl[(List[Char]) => List[Char],
                               Unit](comp[List[Char], List[Char],
   List[Char]](ia, comp[List[Char], List[Char],
                         List[Char]](shows_string.apply(List(':', ' ', 'e', 'r',
                      'r', 'o', 'r', ' ', 'i', 'n', ' ', 'n', 'o', 't', '-',
                      'w', 'e', 'l', 'l', '-', 'f', 'o', 'r', 'm', 'e', 'd',
                      ' ', 'p', 'r', 'o', 'o', 'f')),
                                      comp[List[Char], List[Char],
    List[Char]](shows_nl, x))))))
                       })
  case (ia, i, j, tp, Rel_Rule_Removal(r, s, prf)) =>
    debug[sum[(List[Char]) => List[Char],
               Unit]](ia(Nil),
                       List('R', 'u', 'l', 'e', ' ', 'R', 'e', 'm', 'o', 'v',
                             'a', 'l'),
                       {
                         val raa: List[(term[lab[B, C], D], term[lab[B, C], D])]
                           = (ra[A, lab[B, C], D, Unit](i)).apply(tp)
                         val rrm: List[(term[lab[B, C], D], term[lab[B, C], D])]
                           = list_diff[(term[lab[B, C], D],
 term[lab[B, C], D])](raa, r)
                         val sa: List[(term[lab[B, C], D], term[lab[B, C], D])]
                           = (rwa[A, lab[B, C], D, Unit](i)).apply(tp)
                         val srm: List[(term[lab[B, C], D], term[lab[B, C], D])]
                           = list_diff[(term[lab[B, C], D],
 term[lab[B, C], D])](sa, s);
                         check_reltrs_disproof[A, B, C, D,
        E](comp[List[Char], List[Char],
                 List[Char]](ia, shows_string.apply(List('.', '1'))),
            i, j,
            (delete_R_Rwa[A, lab[B, C], D,
                           Unit](i)).apply(tp).apply(rrm).apply(srm),
            prf)
                       })
}

def certify_cert_problem[A, B : equal : linorder : show,
                          C](a: Boolean,
                              i: tp_ops_ext[A, lab[B, List[nat]], List[Char],
     Unit],
                              j: dpp_ops_ext[C, lab[B, List[nat]], List[Char],
      Unit],
                              x3: cert_problem[B, List[nat], List[Char]]):
      cert_result =
  (a, i, j, x3) match {
  case (a, i, j, TRS_Proof(q, r, None, prf)) =>
    check_trs_proof[A, B,
                     C](a, shows_string.apply(List('1')), i, j,
                         (mka[A, lab[B, List[nat]], List[Char],
                               Unit](i)).apply(strategy_to_Q[lab[B, List[nat]],
                      List[Char]](q, r)).apply(r).apply(Nil),
                         prf)
      match {
      case Inl(err) =>
        Error((comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List('t', 'e', 'r', 'm',
                  'i', 'n', 'a', 't', 'i', 'o', 'n', ' ', 'p', 'r', 'o', 'o',
                  'f', ' ', 'n', 'o', 't', ' ', 'a', 'c', 'c', 'e', 'p', 't',
                  'e', 'd')),
                                  comp[List[Char], List[Char],
List[Char]](shows_nl, err))).apply(Nil))
      case Inr(_) =>
        Certified(xml_tag[B, List[nat],
                           List[Char]](TRS_Proof[B, List[nat],
          List[Char]](q, r, None, prf)))
    }
  case (a, i, j, TRS_Proof(q, r, Some(s), prf)) =>
    check_trs_proof[A, B,
                     C](a, shows_string.apply(List('1')), i, j,
                         (mka[A, lab[B, List[nat]], List[Char],
                               Unit](i)).apply(strategy_to_Q[lab[B, List[nat]],
                      List[Char]](q, r)).apply(r).apply(s),
                         prf)
      match {
      case Inl(err) =>
        Error((comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List('r', 'e', 'l', 'a',
                  't', 'i', 'v', 'e', ' ', 't', 'e', 'r', 'm', 'i', 'n', 'a',
                  't', 'i', 'o', 'n', ' ', 'p', 'r', 'o', 'o', 'f', ' ', 'n',
                  'o', 't', ' ', 'a', 'c', 'c', 'e', 'p', 't', 'e', 'd')),
                                  comp[List[Char], List[Char],
List[Char]](shows_nl, err))).apply(Nil))
      case Inr(_) =>
        Certified(xml_tag[B, List[nat],
                           List[Char]](TRS_Proof[B, List[nat],
          List[Char]](q, r,
                       Some[List[(term[lab[B, List[nat]], List[Char]],
                                   term[lab[B, List[nat]], List[Char]])]](s),
                       prf)))
    }
  case (a, i, j, TRS_Disproof(q, r, prf)) =>
    check_trs_disproof[A, B, List[nat], List[Char],
                        C](shows_string.apply(List('1')), i, j,
                            (mka[A, lab[B, List[nat]], List[Char],
                                  Unit](i)).apply(strategy_to_Q[lab[B,
                             List[nat]],
                         List[Char]](q, r)).apply(r).apply(Nil),
                            prf)
      match {
      case Inl(err) =>
        Error((comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List('n', 'o', 'n', 't',
                  'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', ' ', 'p',
                  'r', 'o', 'o', 'f', ' ', 'n', 'o', 't', ' ', 'a', 'c', 'c',
                  'e', 'p', 't', 'e', 'd')),
                                  comp[List[Char], List[Char],
List[Char]](shows_nl, err))).apply(Nil))
      case Inr(_) =>
        Certified(xml_tag[B, List[nat],
                           List[Char]](TRS_Disproof[B, List[nat],
             List[Char]](q, r, prf)))
    }
  case (a, i, j, Rel_TRS_Disproof(q, r, s, prf)) =>
    check_reltrs_disproof[A, B, List[nat], List[Char],
                           C](shows_string.apply(List('1')), i, j,
                               (mka[A, lab[B, List[nat]], List[Char],
                                     Unit](i)).apply(strategy_to_Q[lab[B,
                                List[nat]],
                            List[Char]](q, r)).apply(r).apply(s),
                               prf)
      match {
      case Inl(err) =>
        Error((comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List('r', 'e', 'l', 'a',
                  't', 'i', 'v', 'e', ' ', 'n', 'o', 'n', 't', 'e', 'r', 'm',
                  'i', 'n', 'a', 't', 'i', 'o', 'n', ' ', 'p', 'r', 'o', 'o',
                  'f', ' ', 'n', 'o', 't', ' ', 'a', 'c', 'c', 'e', 'p', 't',
                  'e', 'd')),
                                  comp[List[Char], List[Char],
List[Char]](shows_nl, err))).apply(Nil))
      case Inr(_) =>
        Certified(xml_tag[B, List[nat],
                           List[Char]](Rel_TRS_Disproof[B, List[nat],
                 List[Char]](q, r, s, prf)))
    }
  case (a, i, j, DP_Proof(p, pw, q, r, rw, prf)) =>
    check_dp_proof[C, B](a, shows_string.apply(List('1')), j,
                          (mkc[C, lab[B, List[nat]], List[Char],
                                Unit](j)).apply(p).apply(pw).apply(strategy_to_Q[lab[B,
      List[nat]],
  List[Char]](q, r ++ rw)).apply(r).apply(rw),
                          prf)
      match {
      case Inl(err) =>
        Error((comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List('f', 'i', 'n', 'i',
                  't', 'e', 'n', 'e', 's', 's', ' ', 'p', 'r', 'o', 'o', 'f',
                  ' ', 'n', 'o', 't', ' ', 'a', 'c', 'c', 'e', 'p', 't', 'e',
                  'd')),
                                  comp[List[Char], List[Char],
List[Char]](shows_nl, err))).apply(Nil))
      case Inr(_) =>
        Certified(xml_tag[B, List[nat],
                           List[Char]](DP_Proof[B, List[nat],
         List[Char]](p, pw, q, r, rw, prf)))
    }
  case (a, i, j, DP_Disproof(p, q, r, prf)) =>
    check_dp_disproof[C, lab[B, List[nat]],
                       List[Char]](shows_string.apply(List('1')), j,
                                    (mkc[C, lab[B, List[nat]], List[Char],
  Unit](j)).apply(p).apply(Nil).apply(strategy_to_Q[lab[B, List[nat]],
             List[Char]](q, r)).apply(Nil).apply(r),
                                    prf)
      match {
      case Inl(err) =>
        Error((comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List('i', 'n', 'f', 'i',
                  'n', 'i', 't', 'e', 'n', 'e', 's', 's', ' ', 'p', 'r', 'o',
                  'o', 'f', ' ', 'n', 'o', 't', ' ', 'a', 'c', 'c', 'e', 'p',
                  't', 'e', 'd')),
                                  comp[List[Char], List[Char],
List[Char]](shows_nl, err))).apply(Nil))
      case Inr(_) =>
        Certified(xml_tag[B, List[nat],
                           List[Char]](DP_Disproof[B, List[nat],
            List[Char]](p, q, r, prf)))
    }
  case (a, i, j, TRS_Confluence(r, prf)) =>
    check_cr_proof[A, B, C](a, shows_string.apply(List('1')), i, j, r, prf)
      match {
      case Inl(err) =>
        Error((comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List('c', 'o', 'n', 'f',
                  'l', 'u', 'e', 'n', 'c', 'e', ' ', 'p', 'r', 'o', 'o', 'f',
                  ' ', 'n', 'o', 't', ' ', 'a', 'c', 'c', 'e', 'p', 't', 'e',
                  'd')),
                                  comp[List[Char], List[Char],
List[Char]](shows_nl, err))).apply(Nil))
      case Inr(_) =>
        Certified(xml_tag[B, List[nat],
                           List[Char]](TRS_Confluence[B, List[nat],
               List[Char]](r, prf)))
    }
  case (a, i, j, Completion_Proof(e, r, prf)) =>
    check_completion_proof[A, B,
                            C](a, shows_string.apply(List('1')), i, j, e, r,
                                prf)
      match {
      case Inl(err) =>
        Error((comp[List[Char], List[Char],
                     List[Char]](shows_string.apply(List('c', 'o', 'm', 'p',
                  'l', 'e', 't', 'i', 'o', 'n', ' ', 'p', 'r', 'o', 'o', 'f',
                  ' ', 'n', 'o', 't', ' ', 'a', 'c', 'c', 'e', 'p', 't', 'e',
                  'd')),
                                  comp[List[Char], List[Char],
List[Char]](shows_nl, err))).apply(Nil))
      case Inr(_) =>
        Certified(xml_tag[B, List[nat],
                           List[Char]](Completion_Proof[B, List[nat],
                 List[Char]](e, r, prf)))
    }
}

def mk_impl[A : equal : linorder,
             B : equal](q: List[term[A, B]], r: List[(term[A, B], term[A, B])],
                         rw: List[(term[A, B], term[A, B])]):
      (List[term[A, B]],
        (Boolean,
          (List[(term[A, B], term[A, B])],
            (List[(term[A, B], term[A, B])],
              rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]])))) =
  {
    val (vr, ra):
          (List[(term[A, B], term[A, B])], List[(term[A, B], term[A, B])])
      = partition[(term[A, B],
                    term[A, B])](comp[term[A, B], Boolean,
                                       (term[A, B],
 term[A, B])]((a: term[A, B]) => is_Var[A, B](a),
               (a: (term[A, B], term[A, B])) => fst[term[A, B], term[A, B]](a)),
                                  r)
    val (vrw, rwa):
          (List[(term[A, B], term[A, B])], List[(term[A, B], term[A, B])])
      = partition[(term[A, B],
                    term[A, B])](comp[term[A, B], Boolean,
                                       (term[A, B],
 term[A, B])]((a: term[A, B]) => is_Var[A, B](a),
               (a: (term[A, B], term[A, B])) => fst[term[A, B], term[A, B]](a)),
                                  rw);
    (q, (nF_subset[A, B](mapa[(term[A, B], term[A, B]),
                               term[A, B]]((a: (term[A, B], term[A, B])) =>
     fst[term[A, B], term[A, B]](a),
    r ++ rw),
                          q),
          (vr, (vrw, (insert_rules[Boolean, A,
                                    B](true,
ra)).apply((insert_rules[Boolean, A,
                          B](false,
                              rwa)).apply(rm_empty[(A, nat),
            List[(Boolean, (term[A, B], term[A, B]))]]))))))
  }

def mk[A : equal : linorder,
        B : equal](q: List[term[A, B]], r: List[(term[A, B], term[A, B])],
                    rw: List[(term[A, B], term[A, B])]):
      tp[A, B] =
  Tp[A, B](mk_impl[A, B](q, r, rw))

def rw_impl[A : equal : linorder,
             B](x0: (List[term[A, B]],
                      (Boolean,
                        (List[(term[A, B], term[A, B])],
                          (List[(term[A, B], term[A, B])],
                            rbt[(A, nat),
                                 List[(Boolean,
(term[A, B], term[A, B]))]]))))):
      List[(term[A, B], term[A, B])] =
  x0 match {
  case (uu, (uv, (uw, (vRw, m)))) =>
    vRw ++ rules_with[Boolean, A, B]((a: Boolean) => ! a, m)
}

def rw[A : equal : linorder, B](tp: tp[A, B]): List[(term[A, B], term[A, B])] =
  rw_impl[A, B](impl_ofa[A, B](tp))

def rules_impl[A : equal : linorder,
                B](x0: (List[term[A, B]],
                         (Boolean,
                           (List[(term[A, B], term[A, B])],
                             (List[(term[A, B], term[A, B])],
                               rbt[(A, nat),
                                    List[(Boolean,
   (term[A, B], term[A, B]))]]))))):
      List[(term[A, B], term[A, B])] =
  x0 match {
  case (uu, (uv, (vR, (vRw, m)))) =>
    vR ++ (vRw ++
            mapa[(Boolean, (term[A, B], term[A, B])),
                  (term[A, B],
                    term[A, B])]((a: (Boolean, (term[A, B], term[A, B]))) =>
                                   snd[Boolean, (term[A, B], term[A, B])](a),
                                  values[(A, nat),
  (Boolean, (term[A, B], term[A, B]))](m)))
}

def rulesa[A : equal : linorder, B](tp: tp[A, B]):
      List[(term[A, B], term[A, B])] =
  rules_impl[A, B](impl_ofa[A, B](tp))

def nF_terms_list[A : equal, B : equal,
                   C : equal](ls: List[term[A, B]], t: term[A, C]):
      Boolean =
  list_all[term[A, C]]((s: term[A, C]) =>
                         list_all[term[A,
B]]((l: term[A, B]) => ! (matches[A, C, B](s, l)), ls),
                        supteq_list[A, C](t))

def is_QNF_impl[A : equal,
                 B : equal](x0: (List[term[A, B]],
                                  (Boolean,
                                    (List[(term[A, B], term[A, B])],
                                      (List[(term[A, B], term[A, B])],
rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]]))))):
      (term[A, B]) => Boolean =
  x0 match {
  case (q, uu) => (a: term[A, B]) => nF_terms_list[A, B, B](q, a)
}

def is_QNF[A : equal : linorder, B : equal](tp: tp[A, B]):
      (term[A, B]) => Boolean =
  is_QNF_impl[A, B](impl_ofa[A, B](tp))

def q_empty_impl[A, B](x0: (List[term[A, B]],
                             (Boolean,
                               (List[(term[A, B], term[A, B])],
                                 (List[(term[A, B], term[A, B])],
                                   rbt[(A, nat),
List[(Boolean, (term[A, B], term[A, B]))]]))))):
      Boolean =
  x0 match {
  case (q, uu) => nulla[term[A, B]](q)
}

def q_empty[A : linorder, B](tp: tp[A, B]): Boolean =
  q_empty_impl[A, B](impl_ofa[A, B](tp))

def qreltrs_impl[A : equal : linorder,
                  B : equal](x0: (List[term[A, B]],
                                   (Boolean,
                                     (List[(term[A, B], term[A, B])],
                                       (List[(term[A, B], term[A, B])],
 rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]]))))):
      ((term[A, B]) => Boolean,
        (((term[A, B], term[A, B])) => Boolean,
          ((term[A, B], term[A, B])) => Boolean)) =
  x0 match {
  case (q, (uu, (vR, (vRw, m)))) =>
    (set[term[A, B]](q),
      (set[(term[A, B],
             term[A, B])](vR ++ rules_with[Boolean, A, B](id[Boolean], m)),
        set[(term[A, B],
              term[A, B])](vRw ++
                             rules_with[Boolean, A,
 B]((a: Boolean) => ! a, m))))
}

def qreltrs[A : equal : linorder, B : equal](tp: tp[A, B]):
      ((term[A, B]) => Boolean,
        (((term[A, B], term[A, B])) => Boolean,
          ((term[A, B], term[A, B])) => Boolean)) =
  qreltrs_impl[A, B](impl_ofa[A, B](tp))

def rules_map_impl[A : equal : linorder,
                    B](x0: (List[term[A, B]],
                             (Boolean,
                               (List[(term[A, B], term[A, B])],
                                 (List[(term[A, B], term[A, B])],
                                   rbt[(A, nat),
List[(Boolean, (term[A, B], term[A, B]))]])))),
                        fn: (A, nat)):
      List[(term[A, B], term[A, B])] =
  (x0, fn) match {
  case ((uu, (uv, (uw, (ux, m)))), fn) =>
    rm_lookup[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]](fn, m) match
      {
      case None => Nil
      case Some(a) =>
        mapa[(Boolean, (term[A, B], term[A, B])),
              (term[A, B],
                term[A, B])]((aa: (Boolean, (term[A, B], term[A, B]))) =>
                               snd[Boolean, (term[A, B], term[A, B])](aa),
                              a)
    }
}

def rules_map[A : equal : linorder, B](tp: tp[A, B]):
      ((A, nat)) => List[(term[A, B], term[A, B])] =
  (a: (A, nat)) => rules_map_impl[A, B](impl_ofa[A, B](tp), a)

def delete_R_Rw_impl[A : equal : linorder,
                      B : equal](x0: (List[term[A, B]],
                                       (Boolean,
 (List[(term[A, B], term[A, B])],
   (List[(term[A, B], term[A, B])],
     rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]])))),
                                  r: List[(term[A, B], term[A, B])],
                                  rw: List[(term[A, B], term[A, B])]):
      (List[term[A, B]],
        (Boolean,
          (List[(term[A, B], term[A, B])],
            (List[(term[A, B], term[A, B])],
              rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]])))) =
  (x0, r, rw) match {
  case ((q, (uu, (vR, (vRw, m)))), r, rw) =>
    {
      val (vr, ra):
            (List[(term[A, B], term[A, B])], List[(term[A, B], term[A, B])])
        = partition[(term[A, B],
                      term[A, B])](comp[term[A, B], Boolean,
 (term[A, B],
   term[A, B])]((a: term[A, B]) => is_Var[A, B](a),
                 (a: (term[A, B], term[A, B])) =>
                   fst[term[A, B], term[A, B]](a)),
                                    r)
      val (vrw, rwa):
            (List[(term[A, B], term[A, B])], List[(term[A, B], term[A, B])])
        = partition[(term[A, B],
                      term[A, B])](comp[term[A, B], Boolean,
 (term[A, B],
   term[A, B])]((a: term[A, B]) => is_Var[A, B](a),
                 (a: (term[A, B], term[A, B])) =>
                   fst[term[A, B], term[A, B]](a)),
                                    rw)
      val vra: List[(term[A, B], term[A, B])] =
        list_diff[(term[A, B], term[A, B])](vR, vr)
      val vrwa: List[(term[A, B], term[A, B])] =
        list_diff[(term[A, B], term[A, B])](vRw, vrw)
      val ma: rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]] =
        (delete_rules[Boolean, A,
                       B](true,
                           ra)).apply((delete_rules[Boolean, A,
             B](false, rwa)).apply(m));
      (q, (nF_subset[A, B](mapa[(term[A, B], term[A, B]),
                                 term[A, B]]((a: (term[A, B], term[A, B])) =>
       fst[term[A, B], term[A, B]](a),
      vra ++
        (vrwa ++
          mapa[(Boolean, (term[A, B], term[A, B])),
                (term[A, B],
                  term[A, B])]((a: (Boolean, (term[A, B], term[A, B]))) =>
                                 snd[Boolean, (term[A, B], term[A, B])](a),
                                values[(A, nat),
(Boolean, (term[A, B], term[A, B]))](ma)))),
                            q),
            (vra, (vrwa, ma))))
    }
}

def delete_R_Rw[A : equal : linorder,
                 B : equal](tp: tp[A, B], r: List[(term[A, B], term[A, B])],
                             rw: List[(term[A, B], term[A, B])]):
      tp[A, B] =
  Tp[A, B](delete_R_Rw_impl[A, B](impl_ofa[A, B](tp), r, rw))

def split_rules_impl[A : equal : linorder,
                      B : equal](tp: (List[term[A, B]],
                                       (Boolean,
 (List[(term[A, B], term[A, B])],
   (List[(term[A, B], term[A, B])],
     rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]])))),
                                  rs: List[(term[A, B], term[A, B])]):
      (List[(term[A, B], term[A, B])], List[(term[A, B], term[A, B])]) =
  partition[(term[A, B],
              term[A, B])]((a: (term[A, B], term[A, B])) =>
                             membera[(term[A, B], term[A, B])](rs, a),
                            rules_impl[A, B](tp))

def split_rules[A : equal : linorder, B : equal](tp: tp[A, B]):
      (List[(term[A, B], term[A, B])]) =>
        (List[(term[A, B], term[A, B])], List[(term[A, B], term[A, B])]) =
  (a: List[(term[A, B], term[A, B])]) =>
    split_rules_impl[A, B](impl_ofa[A, B](tp), a)

def nFQ_subset_NF_rules_impl[A, B](x0: (List[term[A, B]],
 (Boolean,
   (List[(term[A, B], term[A, B])],
     (List[(term[A, B], term[A, B])],
       rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]]))))):
      Boolean =
  x0 match {
  case (uu, (b, uv)) => b
}

def nFQ_subset_NF_rules[A : linorder, B](tp: tp[A, B]): Boolean =
  nFQ_subset_NF_rules_impl[A, B](impl_ofa[A, B](tp))

def tp_rbt_impl[A : equal : linorder, B : equal]:
      tp_ops_ext[tp[A, B], A, B, Unit] =
  Tp_ops_ext[tp[A, B], A, B,
              Unit]((a: tp[A, B]) => qreltrs[A, B](a),
                     (a: tp[A, B]) => q[A, B](a), (a: tp[A, B]) => r[A, B](a),
                     (a: tp[A, B]) => rw[A, B](a),
                     (a: tp[A, B]) => rulesa[A, B](a),
                     (a: tp[A, B]) => q_empty[A, B](a),
                     (a: tp[A, B]) => is_QNF[A, B](a),
                     (a: tp[A, B]) => nFQ_subset_NF_rules[A, B](a),
                     (a: tp[A, B]) => rules_map[A, B](a),
                     (a: tp[A, B]) =>
                       (b: List[(term[A, B], term[A, B])]) =>
                         (c: List[(term[A, B], term[A, B])]) =>
                           delete_R_Rw[A, B](a, b, c),
                     (a: tp[A, B]) => split_rules[A, B](a),
                     (a: List[term[A, B]]) =>
                       (b: List[(term[A, B], term[A, B])]) =>
                         (c: List[(term[A, B], term[A, B])]) =>
                           mk[A, B](a, b, c),
                     ())

def p_impl[A, B](x0: (List[(term[A, B], term[A, B])],
                       (List[(term[A, B], term[A, B])],
                         (List[term[A, B]],
                           (Boolean,
                             (Boolean,
                               (List[(term[A, B], term[A, B])],
                                 (List[(term[A, B], term[A, B])],
                                   (rbt[(A, nat),
 List[(Boolean, (term[A, B], term[A, B]))]],
                                     rbt[(A, nat),
  List[(Boolean, (term[A, B], term[A, B]))]]))))))))):
      List[(term[A, B], term[A, B])] =
  x0 match {
  case (p, uu) => p
}

def impl_ofb[A : linorder, B](x0: dpp[A, B]):
      (List[(term[A, B], term[A, B])],
        (List[(term[A, B], term[A, B])],
          (List[term[A, B]],
            (Boolean,
              (Boolean,
                (List[(term[A, B], term[A, B])],
                  (List[(term[A, B], term[A, B])],
                    (rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]],
                      rbt[(A, nat),
                           List[(Boolean, (term[A, B], term[A, B]))]])))))))) =
  x0 match {
  case Dpp(x) => x
}

def p[A : linorder, B](d: dpp[A, B]): List[(term[A, B], term[A, B])] =
  p_impl[A, B](impl_ofb[A, B](d))

def q_impla[A, B](x0: (List[(term[A, B], term[A, B])],
                        (List[(term[A, B], term[A, B])],
                          (List[term[A, B]],
                            (Boolean,
                              (Boolean,
                                (List[(term[A, B], term[A, B])],
                                  (List[(term[A, B], term[A, B])],
                                    (rbt[(A, nat),
  List[(Boolean, (term[A, B], term[A, B]))]],
                                      rbt[(A, nat),
   List[(Boolean, (term[A, B], term[A, B]))]]))))))))):
      List[term[A, B]] =
  x0 match {
  case (uu, (uv, (q, uw))) => q
}

def qb[A : linorder, B](d: dpp[A, B]): List[term[A, B]] =
  q_impla[A, B](impl_ofb[A, B](d))

def r_impla[A : equal : linorder,
             B](x0: (List[(term[A, B], term[A, B])],
                      (List[(term[A, B], term[A, B])],
                        (List[term[A, B]],
                          (Boolean,
                            (Boolean,
                              (List[(term[A, B], term[A, B])],
                                (List[(term[A, B], term[A, B])],
                                  (rbt[(A, nat),
List[(Boolean, (term[A, B], term[A, B]))]],
                                    rbt[(A, nat),
 List[(Boolean, (term[A, B], term[A, B]))]]))))))))):
      List[(term[A, B], term[A, B])] =
  x0 match {
  case (uu, (uv, (uw, (ux, (uy, (vr, (uz, (m, va)))))))) =>
    vr ++ rules_with[Boolean, A, B](id[Boolean], m)
}

def rb[A : equal : linorder, B](d: dpp[A, B]): List[(term[A, B], term[A, B])] =
  r_impla[A, B](impl_ofb[A, B](d))

def mk_impla[A : equal : linorder,
              B : equal](p: List[(term[A, B], term[A, B])],
                          pw: List[(term[A, B], term[A, B])],
                          q: List[term[A, B]],
                          r: List[(term[A, B], term[A, B])],
                          rw: List[(term[A, B], term[A, B])]):
      (List[(term[A, B], term[A, B])],
        (List[(term[A, B], term[A, B])],
          (List[term[A, B]],
            (Boolean,
              (Boolean,
                (List[(term[A, B], term[A, B])],
                  (List[(term[A, B], term[A, B])],
                    (rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]],
                      rbt[(A, nat),
                           List[(Boolean, (term[A, B], term[A, B]))]])))))))) =
  {
    val (vr, ra):
          (List[(term[A, B], term[A, B])], List[(term[A, B], term[A, B])])
      = partition[(term[A, B],
                    term[A, B])](comp[term[A, B], Boolean,
                                       (term[A, B],
 term[A, B])]((a: term[A, B]) => is_Var[A, B](a),
               (a: (term[A, B], term[A, B])) => fst[term[A, B], term[A, B]](a)),
                                  r)
    val (vrw, rwa):
          (List[(term[A, B], term[A, B])], List[(term[A, B], term[A, B])])
      = partition[(term[A, B],
                    term[A, B])](comp[term[A, B], Boolean,
                                       (term[A, B],
 term[A, B])]((a: term[A, B]) => is_Var[A, B](a),
               (a: (term[A, B], term[A, B])) => fst[term[A, B], term[A, B]](a)),
                                  rw)
    val rs: List[(term[A, B], term[A, B])] = r ++ rw;
    (p, (pw, (q, (nF_subset[A, B](mapa[(term[A, B], term[A, B]),
term[A, B]]((a: (term[A, B], term[A, B])) => fst[term[A, B], term[A, B]](a),
             rs),
                                   q),
                   (list_all[(term[A, B],
                               term[A, B])]((a: (term[A, B], term[A, B])) =>
      {
        val (_, rb): (term[A, B], term[A, B]) = a;
        ! (is_Var[A, B](rb))
      },
     rs),
                     (vr, (vrw, ((insert_rules[Boolean, A,
        B](true,
            ra)).apply((insert_rules[Boolean, A,
                                      B](false,
  rwa)).apply(rm_empty[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]])),
                                  (insert_rules[Boolean, A,
         B](true,
             mapa[(term[A, B], term[A, B]),
                   (term[A, B],
                     term[A, B])]((a: (term[A, B], term[A, B])) =>
                                    {
                                      val (x, y): (term[A, B], term[A, B]) = a;
                                      (y, x)
                                    },
                                   r))).apply((insert_rules[Boolean, A,
                     B](false,
                         mapa[(term[A, B], term[A, B]),
                               (term[A, B],
                                 term[A, B])]((a: (term[A, B], term[A, B])) =>
        {
          val (x, y): (term[A, B], term[A, B]) = a;
          (y, x)
        },
       rw))).apply(rm_empty[(A, nat),
                             List[(Boolean,
                                    (term[A, B], term[A, B]))]]))))))))))
  }

def mkb[A : equal : linorder,
         B : equal](p: List[(term[A, B], term[A, B])],
                     pw: List[(term[A, B], term[A, B])], q: List[term[A, B]],
                     r: List[(term[A, B], term[A, B])],
                     rw: List[(term[A, B], term[A, B])]):
      dpp[A, B] =
  Dpp[A, B](mk_impla[A, B](p, pw, q, r, rw))

def pw_impl[A, B](x0: (List[(term[A, B], term[A, B])],
                        (List[(term[A, B], term[A, B])],
                          (List[term[A, B]],
                            (Boolean,
                              (Boolean,
                                (List[(term[A, B], term[A, B])],
                                  (List[(term[A, B], term[A, B])],
                                    (rbt[(A, nat),
  List[(Boolean, (term[A, B], term[A, B]))]],
                                      rbt[(A, nat),
   List[(Boolean, (term[A, B], term[A, B]))]]))))))))):
      List[(term[A, B], term[A, B])] =
  x0 match {
  case (uu, (pw, uv)) => pw
}

def pw[A : linorder, B](d: dpp[A, B]): List[(term[A, B], term[A, B])] =
  pw_impl[A, B](impl_ofb[A, B](d))

def rw_impla[A : equal : linorder,
              B](x0: (List[(term[A, B], term[A, B])],
                       (List[(term[A, B], term[A, B])],
                         (List[term[A, B]],
                           (Boolean,
                             (Boolean,
                               (List[(term[A, B], term[A, B])],
                                 (List[(term[A, B], term[A, B])],
                                   (rbt[(A, nat),
 List[(Boolean, (term[A, B], term[A, B]))]],
                                     rbt[(A, nat),
  List[(Boolean, (term[A, B], term[A, B]))]]))))))))):
      List[(term[A, B], term[A, B])] =
  x0 match {
  case (uu, (uv, (uw, (ux, (uy, (uz, (vrw, (m, va)))))))) =>
    vrw ++ rules_with[Boolean, A, B]((a: Boolean) => ! a, m)
}

def rwb[A : equal : linorder, B](d: dpp[A, B]): List[(term[A, B], term[A, B])] =
  rw_impla[A, B](impl_ofb[A, B](d))

def dpp_impl[A : equal : linorder,
              B : equal](x0: (List[(term[A, B], term[A, B])],
                               (List[(term[A, B], term[A, B])],
                                 (List[term[A, B]],
                                   (Boolean,
                                     (Boolean,
                                       (List[(term[A, B], term[A, B])],
 (List[(term[A, B], term[A, B])],
   (rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]],
     rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]]))))))))):
      (((term[A, B], term[A, B])) => Boolean,
        (((term[A, B], term[A, B])) => Boolean,
          ((term[A, B]) => Boolean,
            (((term[A, B], term[A, B])) => Boolean,
              ((term[A, B], term[A, B])) => Boolean)))) =
  x0 match {
  case (p, (pw, (q, (uu, (uv, (vr, (vrw, (m, uw)))))))) =>
    (set[(term[A, B], term[A, B])](p),
      (set[(term[A, B], term[A, B])](pw),
        (set[term[A, B]](q),
          (set[(term[A, B],
                 term[A, B])](vr ++ rules_with[Boolean, A, B](id[Boolean], m)),
            set[(term[A, B],
                  term[A, B])](vrw ++
                                 rules_with[Boolean, A,
     B]((a: Boolean) => ! a, m))))))
}

def dppa[A : equal : linorder, B : equal](d: dpp[A, B]):
      (((term[A, B], term[A, B])) => Boolean,
        (((term[A, B], term[A, B])) => Boolean,
          ((term[A, B]) => Boolean,
            (((term[A, B], term[A, B])) => Boolean,
              ((term[A, B], term[A, B])) => Boolean)))) =
  dpp_impl[A, B](impl_ofb[A, B](d))

def pairs_impl[A, B](x0: (List[(term[A, B], term[A, B])],
                           (List[(term[A, B], term[A, B])],
                             (List[term[A, B]],
                               (Boolean,
                                 (Boolean,
                                   (List[(term[A, B], term[A, B])],
                                     (List[(term[A, B], term[A, B])],
                                       (rbt[(A, nat),
     List[(Boolean, (term[A, B], term[A, B]))]],
 rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]]))))))))):
      List[(term[A, B], term[A, B])] =
  x0 match {
  case (p, (pw, uu)) => p ++ pw
}

def pairs[A : linorder, B](d: dpp[A, B]): List[(term[A, B], term[A, B])] =
  pairs_impl[A, B](impl_ofb[A, B](d))

def rules_impla[A : equal : linorder,
                 B](x0: (List[(term[A, B], term[A, B])],
                          (List[(term[A, B], term[A, B])],
                            (List[term[A, B]],
                              (Boolean,
                                (Boolean,
                                  (List[(term[A, B], term[A, B])],
                                    (List[(term[A, B], term[A, B])],
                                      (rbt[(A, nat),
    List[(Boolean, (term[A, B], term[A, B]))]],
rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]]))))))))):
      List[(term[A, B], term[A, B])] =
  x0 match {
  case (uu, (uv, (uw, (ux, (uy, (vr, (vrw, (m, uz)))))))) =>
    vr ++ (vrw ++
            mapa[(Boolean, (term[A, B], term[A, B])),
                  (term[A, B],
                    term[A, B])]((a: (Boolean, (term[A, B], term[A, B]))) =>
                                   snd[Boolean, (term[A, B], term[A, B])](a),
                                  values[(A, nat),
  (Boolean, (term[A, B], term[A, B]))](m)))
}

def rulesc[A : equal : linorder, B](d: dpp[A, B]):
      List[(term[A, B], term[A, B])] =
  rules_impla[A, B](impl_ofb[A, B](d))

def is_QNF_impla[A : equal, B : equal,
                  C : equal](x0: (List[(term[A, B], term[A, B])],
                                   (List[(term[A, B], term[A, B])],
                                     (List[term[A, B]],
                                       (Boolean,
 (Boolean,
   (List[(term[A, B], term[A, B])],
     (List[(term[A, B], term[A, B])],
       (rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]],
         rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]]))))))))):
      (term[A, C]) => Boolean =
  x0 match {
  case (uu, (uv, (q, uw))) => (a: term[A, C]) => nF_terms_list[A, B, C](q, a)
}

def is_QNFb[A : equal : linorder, B : equal](d: dpp[A, B]):
      (term[A, B]) => Boolean =
  is_QNF_impla[A, B, B](impl_ofb[A, B](d))

def q_empty_impla[A, B](x0: (List[(term[A, B], term[A, B])],
                              (List[(term[A, B], term[A, B])],
                                (List[term[A, B]],
                                  (Boolean,
                                    (Boolean,
                                      (List[(term[A, B], term[A, B])],
(List[(term[A, B], term[A, B])],
  (rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]],
    rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]]))))))))):
      Boolean =
  x0 match {
  case (uu, (uv, (q, uw))) => nulla[term[A, B]](q)
}

def q_emptyb[A : linorder, B](d: dpp[A, B]): Boolean =
  q_empty_impla[A, B](impl_ofb[A, B](d))

def rules_map_impla[A : equal : linorder,
                     B](x0: (List[(term[A, B], term[A, B])],
                              (List[(term[A, B], term[A, B])],
                                (List[term[A, B]],
                                  (Boolean,
                                    (Boolean,
                                      (List[(term[A, B], term[A, B])],
(List[(term[A, B], term[A, B])],
  (rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]],
    rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]])))))))),
                         fn: (A, nat)):
      List[(term[A, B], term[A, B])] =
  (x0, fn) match {
  case ((uu, (uv, (uw, (ux, (uy, (uz, (va, (m, vb)))))))), fn) =>
    rm_lookup[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]](fn, m) match
      {
      case None => Nil
      case Some(a) =>
        mapa[(Boolean, (term[A, B], term[A, B])),
              (term[A, B],
                term[A, B])]((aa: (Boolean, (term[A, B], term[A, B]))) =>
                               snd[Boolean, (term[A, B], term[A, B])](aa),
                              a)
    }
}

def rules_mapa[A : equal : linorder, B](d: dpp[A, B]):
      ((A, nat)) => List[(term[A, B], term[A, B])] =
  (a: (A, nat)) => rules_map_impla[A, B](impl_ofb[A, B](d), a)

def delete_P_Pw_impl[A : equal,
                      B : equal](x0: (List[(term[A, B], term[A, B])],
                                       (List[(term[A, B], term[A, B])],
 (List[term[A, B]],
   (Boolean,
     (Boolean,
       (List[(term[A, B], term[A, B])],
         (List[(term[A, B], term[A, B])],
           (rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]],
             rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]])))))))),
                                  pd: List[(term[A, B], term[A, B])],
                                  pwd: List[(term[A, B], term[A, B])]):
      (List[(term[A, B], term[A, B])],
        (List[(term[A, B], term[A, B])],
          (List[term[A, B]],
            (Boolean,
              (Boolean,
                (List[(term[A, B], term[A, B])],
                  (List[(term[A, B], term[A, B])],
                    (rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]],
                      rbt[(A, nat),
                           List[(Boolean, (term[A, B], term[A, B]))]])))))))) =
  (x0, pd, pwd) match {
  case ((p, (pw, rest)), pd, pwd) =>
    (list_diff[(term[A, B], term[A, B])](p, pd),
      (list_diff[(term[A, B], term[A, B])](pw, pwd), rest))
}

def delete_P_Pw[A : equal : linorder,
                 B : equal](d: dpp[A, B], p: List[(term[A, B], term[A, B])],
                             pw: List[(term[A, B], term[A, B])]):
      dpp[A, B] =
  Dpp[A, B](delete_P_Pw_impl[A, B](impl_ofb[A, B](d), p, pw))

def delete_R_Rw_impla[A : equal : linorder,
                       B : equal](d: (List[(term[A, B], term[A, B])],
                                       (List[(term[A, B], term[A, B])],
 (List[term[A, B]],
   (Boolean,
     (Boolean,
       (List[(term[A, B], term[A, B])],
         (List[(term[A, B], term[A, B])],
           (rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]],
             rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]])))))))),
                                   r: List[(term[A, B], term[A, B])],
                                   rw: List[(term[A, B], term[A, B])]):
      (List[(term[A, B], term[A, B])],
        (List[(term[A, B], term[A, B])],
          (List[term[A, B]],
            (Boolean,
              (Boolean,
                (List[(term[A, B], term[A, B])],
                  (List[(term[A, B], term[A, B])],
                    (rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]],
                      rbt[(A, nat),
                           List[(Boolean, (term[A, B], term[A, B]))]])))))))) =
  {
    val (p, (pw, (q, (nfq, (nc, (vR, (vRw, (m, rm)))))))):
          (List[(term[A, B], term[A, B])],
            (List[(term[A, B], term[A, B])],
              (List[term[A, B]],
                (Boolean,
                  (Boolean,
                    (List[(term[A, B], term[A, B])],
                      (List[(term[A, B], term[A, B])],
                        (rbt[(A, nat),
                              List[(Boolean, (term[A, B], term[A, B]))]],
                          rbt[(A, nat),
                               List[(Boolean,
                                      (term[A, B], term[A, B]))]]))))))))
      = d
    val (vr, ra):
          (List[(term[A, B], term[A, B])], List[(term[A, B], term[A, B])])
      = partition[(term[A, B],
                    term[A, B])](comp[term[A, B], Boolean,
                                       (term[A, B],
 term[A, B])]((a: term[A, B]) => is_Var[A, B](a),
               (a: (term[A, B], term[A, B])) => fst[term[A, B], term[A, B]](a)),
                                  r)
    val (vrw, rwa):
          (List[(term[A, B], term[A, B])], List[(term[A, B], term[A, B])])
      = partition[(term[A, B],
                    term[A, B])](comp[term[A, B], Boolean,
                                       (term[A, B],
 term[A, B])]((a: term[A, B]) => is_Var[A, B](a),
               (a: (term[A, B], term[A, B])) => fst[term[A, B], term[A, B]](a)),
                                  rw)
    val vra: List[(term[A, B], term[A, B])] =
      list_diff[(term[A, B], term[A, B])](vR, vr)
    val vrwa: List[(term[A, B], term[A, B])] =
      list_diff[(term[A, B], term[A, B])](vRw, vrw)
    val ma: rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]] =
      (delete_rules[Boolean, A,
                     B](true,
                         ra)).apply((delete_rules[Boolean, A,
           B](false, rwa)).apply(m))
    val rma: rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]] =
      (delete_rules[Boolean, A,
                     B](true,
                         mapa[(term[A, B], term[A, B]),
                               (term[A, B],
                                 term[A, B])]((a: (term[A, B], term[A, B])) =>
        {
          val (x, y): (term[A, B], term[A, B]) = a;
          (y, x)
        },
       r))).apply((delete_rules[Boolean, A,
                                 B](false,
                                     mapa[(term[A, B], term[A, B]),
   (term[A, B],
     term[A, B])]((a: (term[A, B], term[A, B])) =>
                    {
                      val (x, y): (term[A, B], term[A, B]) = a;
                      (y, x)
                    },
                   rw))).apply(rm))
    val rs: List[(term[A, B], term[A, B])] =
      vra ++
        (vrwa ++
          mapa[(Boolean, (term[A, B], term[A, B])),
                (term[A, B],
                  term[A, B])]((a: (Boolean, (term[A, B], term[A, B]))) =>
                                 snd[Boolean, (term[A, B], term[A, B])](a),
                                values[(A, nat),
(Boolean, (term[A, B], term[A, B]))](ma)));
    (p, (pw, (q, (nfq ||
                    nF_subset[A, B](mapa[(term[A, B], term[A, B]),
  term[A, B]]((a: (term[A, B], term[A, B])) => fst[term[A, B], term[A, B]](a),
               rs),
                                     q),
                   (nc || list_all[(term[A, B],
                                     term[A,
   B])]((a: (term[A, B], term[A, B])) =>
          {
            val (_, rb): (term[A, B], term[A, B]) = a;
            ! (is_Var[A, B](rb))
          },
         rs),
                     (vra, (vrwa, (ma, rma))))))))
  }

def delete_R_Rwb[A : equal : linorder,
                  B : equal](d: dpp[A, B], r: List[(term[A, B], term[A, B])],
                              rw: List[(term[A, B], term[A, B])]):
      dpp[A, B] =
  Dpp[A, B](delete_R_Rw_impla[A, B](impl_ofb[A, B](d), r, rw))

def split_pairs_impl[A : equal,
                      B : equal](d: (List[(term[A, B], term[A, B])],
                                      (List[(term[A, B], term[A, B])],
(List[term[A, B]],
  (Boolean,
    (Boolean,
      (List[(term[A, B], term[A, B])],
        (List[(term[A, B], term[A, B])],
          (rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]],
            rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]])))))))),
                                  ps: List[(term[A, B], term[A, B])]):
      (List[(term[A, B], term[A, B])], List[(term[A, B], term[A, B])]) =
  partition[(term[A, B],
              term[A, B])]((a: (term[A, B], term[A, B])) =>
                             membera[(term[A, B], term[A, B])](ps, a),
                            pairs_impl[A, B](d))

def split_pairs[A : equal : linorder, B : equal](d: dpp[A, B]):
      (List[(term[A, B], term[A, B])]) =>
        (List[(term[A, B], term[A, B])], List[(term[A, B], term[A, B])]) =
  (a: List[(term[A, B], term[A, B])]) =>
    split_pairs_impl[A, B](impl_ofb[A, B](d), a)

def split_rules_impla[A : equal : linorder,
                       B : equal](d: (List[(term[A, B], term[A, B])],
                                       (List[(term[A, B], term[A, B])],
 (List[term[A, B]],
   (Boolean,
     (Boolean,
       (List[(term[A, B], term[A, B])],
         (List[(term[A, B], term[A, B])],
           (rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]],
             rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]])))))))),
                                   rs: List[(term[A, B], term[A, B])]):
      (List[(term[A, B], term[A, B])], List[(term[A, B], term[A, B])]) =
  partition[(term[A, B],
              term[A, B])]((a: (term[A, B], term[A, B])) =>
                             membera[(term[A, B], term[A, B])](rs, a),
                            rules_impla[A, B](d))

def split_rulesb[A : equal : linorder, B : equal](d: dpp[A, B]):
      (List[(term[A, B], term[A, B])]) =>
        (List[(term[A, B], term[A, B])], List[(term[A, B], term[A, B])]) =
  (a: List[(term[A, B], term[A, B])]) =>
    split_rules_impla[A, B](impl_ofb[A, B](d), a)

def replace_pair_impl[A : equal,
                       B : equal](x0: (List[(term[A, B], term[A, B])],
(List[(term[A, B], term[A, B])],
  (List[term[A, B]],
    (Boolean,
      (Boolean,
        (List[(term[A, B], term[A, B])],
          (List[(term[A, B], term[A, B])],
            (rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]],
              rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]])))))))),
                                   pair: (term[A, B], term[A, B]),
                                   ps: List[(term[A, B], term[A, B])]):
      (List[(term[A, B], term[A, B])],
        (List[(term[A, B], term[A, B])],
          (List[term[A, B]],
            (Boolean,
              (Boolean,
                (List[(term[A, B], term[A, B])],
                  (List[(term[A, B], term[A, B])],
                    (rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]],
                      rbt[(A, nat),
                           List[(Boolean, (term[A, B], term[A, B]))]])))))))) =
  (x0, pair, ps) match {
  case ((p, (pw, rest)), pair, ps) =>
    (replace_impl[(term[A, B], term[A, B])](pair, ps, p),
      (replace_impl[(term[A, B], term[A, B])](pair, ps, pw), rest))
}

def replace_pair[A : equal : linorder,
                  B : equal](d: dpp[A, B], pair: (term[A, B], term[A, B]),
                              ps: List[(term[A, B], term[A, B])]):
      dpp[A, B] =
  Dpp[A, B](replace_pair_impl[A, B](impl_ofb[A, B](d), pair, ps))

def intersect_pairs_impl[A : equal,
                          B : equal](x0: (List[(term[A, B], term[A, B])],
   (List[(term[A, B], term[A, B])],
     (List[term[A, B]],
       (Boolean,
         (Boolean,
           (List[(term[A, B], term[A, B])],
             (List[(term[A, B], term[A, B])],
               (rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]],
                 rbt[(A, nat),
                      List[(Boolean, (term[A, B], term[A, B]))]])))))))),
                                      ps: List[(term[A, B], term[A, B])]):
      (List[(term[A, B], term[A, B])],
        (List[(term[A, B], term[A, B])],
          (List[term[A, B]],
            (Boolean,
              (Boolean,
                (List[(term[A, B], term[A, B])],
                  (List[(term[A, B], term[A, B])],
                    (rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]],
                      rbt[(A, nat),
                           List[(Boolean, (term[A, B], term[A, B]))]])))))))) =
  (x0, ps) match {
  case ((p, (pw, rest)), ps) =>
    (list_inter[(term[A, B], term[A, B])](p, ps),
      (list_inter[(term[A, B], term[A, B])](pw, ps), rest))
}

def intersect_pairs[A : equal : linorder,
                     B : equal](d: dpp[A, B],
                                 ps: List[(term[A, B], term[A, B])]):
      dpp[A, B] =
  Dpp[A, B](intersect_pairs_impl[A, B](impl_ofb[A, B](d), ps))

def intersect_rules_impl[A : equal : linorder,
                          B : equal](d: (List[(term[A, B], term[A, B])],
  (List[(term[A, B], term[A, B])],
    (List[term[A, B]],
      (Boolean,
        (Boolean,
          (List[(term[A, B], term[A, B])],
            (List[(term[A, B], term[A, B])],
              (rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]],
                rbt[(A, nat),
                     List[(Boolean, (term[A, B], term[A, B]))]])))))))),
                                      ri: List[(term[A, B], term[A, B])]):
      (List[(term[A, B], term[A, B])],
        (List[(term[A, B], term[A, B])],
          (List[term[A, B]],
            (Boolean,
              (Boolean,
                (List[(term[A, B], term[A, B])],
                  (List[(term[A, B], term[A, B])],
                    (rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]],
                      rbt[(A, nat),
                           List[(Boolean, (term[A, B], term[A, B]))]])))))))) =
  {
    val (p, (pw, (q, (nfq, (nc, (vR, (vRw, (m, rm)))))))):
          (List[(term[A, B], term[A, B])],
            (List[(term[A, B], term[A, B])],
              (List[term[A, B]],
                (Boolean,
                  (Boolean,
                    (List[(term[A, B], term[A, B])],
                      (List[(term[A, B], term[A, B])],
                        (rbt[(A, nat),
                              List[(Boolean, (term[A, B], term[A, B]))]],
                          rbt[(A, nat),
                               List[(Boolean,
                                      (term[A, B], term[A, B]))]]))))))))
      = d
    val (vri, ria):
          (List[(term[A, B], term[A, B])], List[(term[A, B], term[A, B])])
      = partition[(term[A, B],
                    term[A, B])](comp[term[A, B], Boolean,
                                       (term[A, B],
 term[A, B])]((a: term[A, B]) => is_Var[A, B](a),
               (a: (term[A, B], term[A, B])) => fst[term[A, B], term[A, B]](a)),
                                  ri)
    val vr: List[(term[A, B], term[A, B])] =
      list_inter[(term[A, B], term[A, B])](vR, vri)
    val vrw: List[(term[A, B], term[A, B])] =
      list_inter[(term[A, B], term[A, B])](vRw, vri)
    val ma: rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]] =
      (intersect_rules[A, B](ria)).apply(m)
    val rma: rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]] =
      (intersect_rules[A, B](mapa[(term[A, B], term[A, B]),
                                   (term[A, B],
                                     term[A,
   B])]((a: (term[A, B], term[A, B])) =>
          {
            val (x, y): (term[A, B], term[A, B]) = a;
            (y, x)
          },
         ri))).apply(rm)
    val rs: List[(term[A, B], term[A, B])] =
      vr ++ (vrw ++
              mapa[(Boolean, (term[A, B], term[A, B])),
                    (term[A, B],
                      term[A, B])]((a: (Boolean, (term[A, B], term[A, B]))) =>
                                     snd[Boolean, (term[A, B], term[A, B])](a),
                                    values[(A, nat),
    (Boolean, (term[A, B], term[A, B]))](ma)));
    (p, (pw, (q, (nfq ||
                    nF_subset[A, B](mapa[(term[A, B], term[A, B]),
  term[A, B]]((a: (term[A, B], term[A, B])) => fst[term[A, B], term[A, B]](a),
               rs),
                                     q),
                   (nc || list_all[(term[A, B],
                                     term[A,
   B])]((a: (term[A, B], term[A, B])) =>
          {
            val (_, r): (term[A, B], term[A, B]) = a;
            ! (is_Var[A, B](r))
          },
         rs),
                     (vr, (vrw, (ma, rma))))))))
  }

def intersect_rulesa[A : equal : linorder,
                      B : equal](d: dpp[A, B],
                                  rs: List[(term[A, B], term[A, B])]):
      dpp[A, B] =
  Dpp[A, B](intersect_rules_impl[A, B](impl_ofb[A, B](d), rs))

def reverse_rules_map_impl[A : equal : linorder,
                            B](x0: (List[(term[A, B], term[A, B])],
                                     (List[(term[A, B], term[A, B])],
                                       (List[term[A, B]],
 (Boolean,
   (Boolean,
     (List[(term[A, B], term[A, B])],
       (List[(term[A, B], term[A, B])],
         (rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]],
           rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]])))))))),
                                fn: (A, nat)):
      List[(term[A, B], term[A, B])] =
  (x0, fn) match {
  case ((uu, (uv, (uw, (ux, (uy, (uz, (va, (vb, m)))))))), fn) =>
    rm_lookup[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]](fn, m) match
      {
      case None => Nil
      case Some(a) =>
        mapa[(Boolean, (term[A, B], term[A, B])),
              (term[A, B],
                term[A, B])]((aa: (Boolean, (term[A, B], term[A, B]))) =>
                               snd[Boolean, (term[A, B], term[A, B])](aa),
                              a)
    }
}

def reverse_rules_map[A : equal : linorder, B](d: dpp[A, B]):
      ((A, nat)) => List[(term[A, B], term[A, B])] =
  (a: (A, nat)) => reverse_rules_map_impl[A, B](impl_ofb[A, B](d), a)

def rules_no_left_var_impl[A, B](x0: (List[(term[A, B], term[A, B])],
                                       (List[(term[A, B], term[A, B])],
 (List[term[A, B]],
   (Boolean,
     (Boolean,
       (List[(term[A, B], term[A, B])],
         (List[(term[A, B], term[A, B])],
           (rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]],
             rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]]))))))))):
      Boolean =
  x0 match {
  case (uu, (uv, (uw, (ux, (uy, (Nil, (Nil, uz))))))) => true
  case (v, (vc, (ve, (vg, (vi, (vm :: vn, vl)))))) => false
  case (v, (vc, (ve, (vg, (vi, (vk, (vo :: vp, vn))))))) => false
}

def rules_no_left_var[A : linorder, B](d: dpp[A, B]): Boolean =
  rules_no_left_var_impl[A, B](impl_ofb[A, B](d))

def nFQ_subset_NF_rules_impla[A, B](x0: (List[(term[A, B], term[A, B])],
  (List[(term[A, B], term[A, B])],
    (List[term[A, B]],
      (Boolean,
        (Boolean,
          (List[(term[A, B], term[A, B])],
            (List[(term[A, B], term[A, B])],
              (rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]],
                rbt[(A, nat),
                     List[(Boolean, (term[A, B], term[A, B]))]]))))))))):
      Boolean =
  x0 match {
  case (uu, (uv, (uw, (b, ux)))) => b
}

def nFQ_subset_NF_rulesa[A : linorder, B](d: dpp[A, B]): Boolean =
  nFQ_subset_NF_rules_impla[A, B](impl_ofb[A, B](d))

def rules_non_collapsing_impl[A, B](x0: (List[(term[A, B], term[A, B])],
  (List[(term[A, B], term[A, B])],
    (List[term[A, B]],
      (Boolean,
        (Boolean,
          (List[(term[A, B], term[A, B])],
            (List[(term[A, B], term[A, B])],
              (rbt[(A, nat), List[(Boolean, (term[A, B], term[A, B]))]],
                rbt[(A, nat),
                     List[(Boolean, (term[A, B], term[A, B]))]]))))))))):
      Boolean =
  x0 match {
  case (uu, (uv, (uw, (ux, (nc, uy))))) => nc
}

def rules_non_collapsing[A : linorder, B](d: dpp[A, B]): Boolean =
  rules_non_collapsing_impl[A, B](impl_ofb[A, B](d))

def dpp_rbt_impl[A : equal : linorder, B : equal]:
      dpp_ops_ext[dpp[A, B], A, B, Unit] =
  Dpp_ops_ext[dpp[A, B], A, B,
               Unit]((a: dpp[A, B]) => dppa[A, B](a),
                      (a: dpp[A, B]) => p[A, B](a),
                      (a: dpp[A, B]) => pw[A, B](a),
                      (a: dpp[A, B]) => pairs[A, B](a),
                      (a: dpp[A, B]) => qb[A, B](a),
                      (a: dpp[A, B]) => rb[A, B](a),
                      (a: dpp[A, B]) => rwb[A, B](a),
                      (a: dpp[A, B]) => rulesc[A, B](a),
                      (a: dpp[A, B]) => q_emptyb[A, B](a),
                      (a: dpp[A, B]) => rules_no_left_var[A, B](a),
                      (a: dpp[A, B]) => rules_non_collapsing[A, B](a),
                      (a: dpp[A, B]) => is_QNFb[A, B](a),
                      (a: dpp[A, B]) => nFQ_subset_NF_rulesa[A, B](a),
                      (a: dpp[A, B]) => rules_mapa[A, B](a),
                      (a: dpp[A, B]) => reverse_rules_map[A, B](a),
                      (a: dpp[A, B]) =>
                        (b: List[(term[A, B], term[A, B])]) =>
                          intersect_pairs[A, B](a, b),
                      (a: dpp[A, B]) =>
                        (b: (term[A, B], term[A, B])) =>
                          (c: List[(term[A, B], term[A, B])]) =>
                            replace_pair[A, B](a, b, c),
                      (a: dpp[A, B]) =>
                        (b: List[(term[A, B], term[A, B])]) =>
                          intersect_rulesa[A, B](a, b),
                      (a: dpp[A, B]) =>
                        (b: List[(term[A, B], term[A, B])]) =>
                          (c: List[(term[A, B], term[A, B])]) =>
                            delete_P_Pw[A, B](a, b, c),
                      (a: dpp[A, B]) =>
                        (b: List[(term[A, B], term[A, B])]) =>
                          (c: List[(term[A, B], term[A, B])]) =>
                            delete_R_Rwb[A, B](a, b, c),
                      (a: dpp[A, B]) => split_pairs[A, B](a),
                      (a: dpp[A, B]) => split_rulesb[A, B](a),
                      (a: List[(term[A, B], term[A, B])]) =>
                        (b: List[(term[A, B], term[A, B])]) =>
                          (c: List[term[A, B]]) =>
                            (d: List[(term[A, B], term[A, B])]) =>
                              (e: List[(term[A, B], term[A, B])]) =>
                                mkb[A, B](a, b, c, d, e),
                      ())

def certify_proof(a: Boolean, problemString: List[Char]): cert_result =
  parse_cert_problem(problemString) match {
    case Inl(err) =>
      Unsupported(List('e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'i', 'l', 'e',
                        ' ', 'p', 'a', 'r', 's', 'i', 'n', 'g') ++
                    (List(newline) ++ err))
    case Inr(cp) =>
      extract_input_proof(problemString) match {
        case None =>
          Error(List('c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'f', 'i',
                      'n', 'd', ' ', 'i', 'n', 'p', 'u', 't', ' ', 'a', 'n',
                      'd', ' ', 'p', 'r', 'o', 'o', 'f', ' ', 'i', 'n', ' ',
                      's', 't', 'r', 'i', 'n', 'g', ' ', 'f', 'o', 'r', ' ',
                      'c', 'o', 'm', 'p', 'a', 'r', 'i', 's', 'o', 'n', ' ',
                      'w', 'i', 't', 'h', ' ', 'p', 'a', 'r', 's', 'e', 'd',
                      ' ', 'r', 'e', 's', 'u', 'l', 't'))
        case Some((the_input, (_, the_proof))) =>
          (if (eqWhiteSpace(the_input,
                             (shows_prec_xml(Zero_nat,
      xml_cert_problem[List[Char], nat, List[Char]](cp))).apply(Nil)))
            certify_cert_problem[tp[lab[List[Char], List[nat]], List[Char]],
                                  List[Char],
                                  dpp[lab[List[Char], List[nat]],
                                       List[Char]]](a,
             tp_rbt_impl[lab[List[Char], List[nat]], List[Char]],
             dpp_rbt_impl[lab[List[Char], List[nat]], List[Char]], cp)
              match {
              case Certified(tag) =>
                (if (starts_with(the_proof, tag)) Certified(tag)
                  else Error(List('p', 'r', 'o', 'v', 'e', 'n', ' ', 'p', 'r',
                                   'o', 'p', 'e', 'r', 't', 'y', ' ') ++
                               (tag ++
                                 (List(' ', 'd', 'o', 'e', 's', ' ', 'n', 'o',
't', ' ', 'c', 'o', 'r', 'r', 'e', 's', 'p', 'o', 'n', 'd', ' ', 't', 'o', ' ',
'p', 'r', 'o', 'o', 'f', ' ', 'i', 'n', ' ', 'i', 'n', 'p', 'u', 't', ' ', 's',
't', 'r', 'i', 'n', 'g', ':', ' ') ++
                                   take[Char](nat(BigInt(20)), the_proof)))))
              case Unsupported(aa) => Unsupported(aa)
              case Error(aa) => Error(aa)
            }
            else Error(concat[Char](List(List('p', 'a', 'r', 's', 'e', 'd', ' ',
       'p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'd', 'o', 'e', 's', ' ', 'n',
       'o', 't', ' ', 'c', 'o', 'r', 'r', 'e', 's', 'p', 'o', 'n', 'd', ' ',
       't', 'o', ' ', 'i', 'n', 'p', 'u', 't'),
  List(newline), List('i', 'n', 'p', 'u', 't', ':'), List(newline),
  filter[Char]((c: Char) => ! (membera[Char](wspace, c)), the_input),
  List(newline), List(newline), List('p', 'a', 'r', 's', 'e', 'd', ':'),
  List(newline),
  filter[Char]((c: Char) => ! (membera[Char](wspace, c)),
                (shows_prec_xml(Zero_nat,
                                 xml_cert_problem[List[Char], nat,
           List[Char]](cp))).apply(Nil)),
  List(newline)))))
      }
  }

} /* object Ceta */
